/*
 *  bf.c
 *  
 *  This file contains functions necessary to install the following branch
 *  functions:
 *      bf_jump() - called by former jumps,
 *      bf_call() - called by former calls,
 *  and to modify jumps/calls to call those branch functions. The bf_jump()
 *  initially contains only the first instruction - pushf, the bf_call()
 *  initially contains a nop instruction. Both branch functions are completed
 *  later, after the final addresses of instructions are known (see finalize.c
 *  for details). The purpose of the branch function is to derive the target
 *  of the original control transfer instruction using the address of the
 *  instruction following the call (the return address pushed onto the stack)
 *  and the perfect hash function (see perfect.c for details). Then branch
 *  function overwrites the return address with the derived target address
 *  and returns to it.
 */
#include "obf_global.h"
#include "bf.h"
#include "obf_macros.h"
#include "flowgraph.h"
#include "utilities.h"
#include "linkage.h"
#include "sysdep.h"



static void CreateBranchFunctions();
static Function *InsertBranchFunction(char *name);
static void ConvertControlTransfersToBFCalls();
static void JumpToBFCall(Instruction *jump);
static void CallToBFCall(Instruction *call);



/*
 *  InstallBranchFunctionCalls() --
 *
 *  This is a driver for the obfuscation using branch functions. First, both
 *  branch functions are created (not finished). Second, we convert the control
 *  transfer instructions (jumps, calls) at candidate locations to the calls
 *  to the branch functions.
 */
void InstallBranchFunctionCalls() {

    REPORT(RP_DETAIL, "Creating branch functions");
    CreateBranchFunctions();

    REPORT(RP_DETAIL, "Converting control transfers to branch function calls");
    ConvertControlTransfersToBFCalls();

    return;
}



/*
 *  CreateBranchFunctions() --
 *
 *  Creates two branch functions each containing only one instruction. They
 *  will be finalized later, when the final addresses of the instructions are
 *  known (see CompleteJumpBranchFunction() and CompleteCallBranchFunction()
 *  in finalize.c).
 */
static void CreateBranchFunctions() {
    Function *bf;

    bf = InsertBranchFunction("bf_jump");
    bf_jump_first_bbl = bf->f_first;

    if (BF_CONVERT_CALLS) {
        bf = InsertBranchFunction("bf_call");
        bf_call_first_bbl = bf->f_first;
    }

    return;
}



/*
 *  InsertBranchFunction(name) --
 *
 *  Creates and inserts at the end of the binary a one-instruction branch
 *  function with the specified name.
 */
static Function *InsertBranchFunction(char *name) {
    Function *bf;
    Block *bf_first_bbl;
    Instruction *new_instr;

    bf_first_bbl = BblNew();
    bf_first_bbl->b_type = BT_NORMAL;

    bf = FunNew(); //takes care of insertion at the end of f_list
    bf->f_name = name;
    bf->f_first = bf_first_bbl;
    bf->f_last = bf_first_bbl;
    AddNewFunctionSymbol(bf); //!use only for debugging
    FUNCTION_SET_FLAG(bf, FUNCTION_DO_NOT_OBFUSCATE_SIGNAL);

    if (strcmp(name, "bf_jump") == 0) {
        new_instr = InstrCreateSave();
    } else {
        new_instr = InstrCreateNOP();
    }
    bf_first_bbl->b_first = bf_first_bbl->b_last = new_instr;
    INSTR_SET_FLAG(bf_first_bbl->b_first, BBL_START);
    INSTR_SET_FLAG(bf_first_bbl->b_first, BBL_END);
    bf_first_bbl->b_first->i_Bblock = bf_first_bbl;

    bf_first_bbl->b_first->i_prev = bf_first_bbl->b_prev->b_last;
    bf_first_bbl->b_prev->b_last->i_next = bf_first_bbl->b_first;

    bf_first_bbl->b_fun = bf->f_entry->b_fun;

    REPORT(RP_SMLOOP, "Inserted the first block of %s()", bf->f_name);
    return bf;
}



/*
 *  ConvertControlTransfersToBFCalls() --
 *
 *  Converts the control transfer instructions (jumps, calls) at the end of
 *  the selected basic blocks to calls to the branch functions.
 */
static void ConvertControlTransfersToBFCalls() {
    Block *biterator;
    int num_jumps_converted = 0;
    int num_calls_converted = 0;
    int num_converted = 0;

    FOREACH_BBL(biterator) {

        if (BBL_HAS_FLAG(biterator, BBL_BF_FLAG)) {

            if (INSTR_IS_DIRECT_JUMP(biterator->b_last)) {
                JumpToBFCall(biterator->b_last);
                num_jumps_converted++;
            } else if (INSTR_IS_DIRECT_CALL(biterator->b_last)) {
                CallToBFCall(biterator->b_last);
                num_calls_converted++;
            } else {
                WARNING("Attempt to convert to bf call instruction"
                    " other than jump or call");
                num_converted--;
            }

            num_converted++;
        }
    }

    REPORT(RP_DETAIL, "Jumps converted:   %6d", num_jumps_converted);
    REPORT(RP_DETAIL, "Calls converted:   %6d", num_calls_converted);
    REPORT(RP_DETAIL, "Total converted:   %6d", num_converted);

    return;
}



/*
 *  JumpToBFCall(jump) --
 *
 *  Converts a direct unconditional jump received as a parameter to the call
 *  to the branch function. It actually only replaces a jump with a call, all
 *  edge and block information will be taken care of later, after the final
 *  addresses are known.
 */
static void JumpToBFCall(Instruction *jump) {
    Instruction *new_instr = InstrCreateCALLDisp();
    InsertInstrBeforeInstr(new_instr, jump);
    INSTR_SET_FLAG(new_instr, INSTR_JBF_CALL_FLAG);
    InstrDisconnect(jump);
    //InstrFree(jump);
    return;
}



/*
 *  CallToBFCall(call) --
 *
 *  Converts a direct call received as a parameter to the call to the branch
 *  function. It actually only checks that the block containing this call
 *  instruction has both outgoing edges. The address in the instruction (the
 *  callee's address) will be updated later, after the final addresses are
 *  known.
 */
static void CallToBFCall(Instruction *call) {
    Edge *edge;

    edge = BBLGetFunlinkEdge(call->i_Bblock);
    if (edge == NULL) {
        WARNING("Could not convert call to bf call, no funlink edge. %x: %s",
            call->i_address, PrintInstruction(call));
        return;
    }

    edge = BBLGetFunDirectEdge(call->i_Bblock);
    if (edge == NULL) {
        WARNING("Could not convert call to bf call, no fundirect edge. %x: %s",
            call->i_address, PrintInstruction(call));
        return;
    }

    INSTR_SET_FLAG(call, INSTR_CBF_CALL_FLAG);
    return;
}

