/*
 *  extend.c
 *
 *  Functions for extending the number of candidate locations for obfuscation.
 *  In general, the candidates are pc modifying instructions without a
 *  fallthrough edge. In particular, most of the functions here serve the
 *  purpose of increasing the number of unconditional jumps in the binary.
 */
#include "obf_global.h"
#include "extend.h"
#include "sysdep.h"
#include "utilities.h"
#include "flowgraph.h"
#include "linkage.h"
#include "x86-asm.h"



static void EqualizeBlocks();
static void InvertBranches();
static bool InvertBranch(Instruction *);
static void DetachDisassemblyEntries();
static void IsolateHotRegions();



/*
 *  ExtendCandidates() --
 *
 *  Applies various techniques to extend the number of control transfer
 *  instructions - the main candidates for the control flow obfuscation.
 */
void ExtendCandidates() {

    if (EQUALIZE_BLOCKS) { // -q
        REPORT(RP_MINOR, "Equalizing Basic Blocks");
        //EqualizeBlocks(max_bbl_size);
        EqualizeBlocks(); //add parameter
        BBLOutputStats();
        CheckConsistency();
    }

    if (INVERT_CONDITIONAL_BRANCHES) { // -f (flip)
        REPORT(RP_MINOR, "Inverting (Flipping) Conditional Branches");
        InvertBranches();
        BBLOutputStats();
        CheckConsistency();
    }

    if (CUT_OFF_STARTS) { // -a now, change later
        REPORT(RP_MINOR, "Cutting Off Disassembly Starting Points");
        DetachDisassemblyEntries(); //add parameter to specify: lin, rec, both
        BBLOutputStats();
        CheckConsistency();
    }

    if (ISOLATE_HOT_REGIONS) {
        REPORT(RP_MINOR, "Isolating Hot Regions");
        IsolateHotRegions();
        BBLOutputStats();
        CheckConsistency();
    }

    return;
}



/*
 *  EqualizeBlocks() --
 *
 *  ...
 */
void EqualizeBlocks() {
    WARNING("[EqualizeBlocks]: just called");
    WARNING("[EqualizeBlocks]: returning");
    return;
}



/*
 *  InvertBranches() --
 *
 *  InvertBranches locates the conditional branches and flips the condition
 *  to the opposite by calling InvertBranch(). Transformation is only applied
 *  to the non-hot code.
 */
void InvertBranches() {
    Block *biterator, *next;
    Instruction *instr;
    bool inverted = false;
    int hot_skipped = 0;
    int num_branches = 0;
    int num_inverted = 0;
    int num_not_inverted = 0;

    biterator = bbl_list;
    while (biterator != NULL) {
        next = biterator->b_next;
        if (!INVERT_HOT_BRANCHES && BBL_HAS_FLAG(biterator, BBL_HOT_FLAG)) {
            hot_skipped++;
            biterator = next;
            continue;
        }

        instr = biterator->b_last;
        if (instr == NULL) {
            biterator = next;
            continue;
        }

        if (isConditionalInstr(instr) &&
            !(isCallClass(instr) || isReturnClass(instr))) {
            num_branches++;
            inverted = InvertBranch(instr);
  
            if (inverted) {
                num_inverted++;
            } else {
                num_not_inverted++;
            }
        }
        biterator = next;
    }

    REPORT(RP_DETAIL, "Total Hot Skipped BBLs:  %6d", hot_skipped);
    REPORT(RP_DETAIL, "Total Non-Hot Branches:  %6d", num_branches);
    REPORT(RP_DETAIL, "Total Inverted Branches: %6d", num_inverted);

    return;
}



/*
 *  InvertBranch(cond) --
 *
 *  InvertBranch takes a conditional branch and replaces it with its counter
 *  conditional branch and adds an absolute jump after the branch that actually
 *  jumps to the target of the original branch. The new conditional branch
 *  now branches to the instruction after the jump, eg:
 *          Addr0: je   Target
 *          Addr1: next_instr
 *  becomes
 *          Addr0: jne  Addr2
 *          Addr1: jmp  Target
 *          Addr2: next_instr
 */
static bool InvertBranch(Instruction *cond_instr) {
    Block *jump_bbl;
    Instruction *jump_instr = NULL;
    Instruction *target_instr;
    Instruction *next_instr;
    EDGE_WEIGHT_TYPE true_edge_weight, false_edge_weight;
    Edge *true_edge, *false_edge, *uncond_edge;

    /*
     * Get the target instruction and next instruction of the branch and ensure
     * that target is not next or itself.
     */
    true_edge = BBLGetTrueEdge(cond_instr->i_Bblock);
    if (true_edge == NULL) {
        WARNING("%s", "No TRUE edge");
        return false;
    }
    true_edge_weight = true_edge->e_edgeWeight;

    target_instr = true_edge->e_to->b_first;
    if (target_instr == NULL) {
        WARNING("%s", "Target of branch instruction at 0x%x is NULL",
            cond_instr->i_address);
        return false;
    }
    if (target_instr == cond_instr) {
        WARNING("Branch at 0x%x in bbl %d targets itself",
            cond_instr->i_address, cond_instr->i_Bblock->b_id);
        return false;
    }

    false_edge = BBLGetFalseEdge(cond_instr->i_Bblock);
    if (false_edge == NULL) {
        WARNING("%s", "No FALSE edge");
        return false;
    }
    false_edge_weight = false_edge->e_edgeWeight;

    next_instr = cond_instr->i_next;
    if (target_instr == next_instr) {
        WARNING("Branch at 0x%x targets the next (fallthrough) instruction",
            cond_instr->i_next);
        return false;
    }

    if (false_edge->e_to->b_first != next_instr) {
        WARNING("False edge goes to instruction other than the next one");
        return false;
    }

    /*
     *  Invert the branch condition, insert jump in its own bbl,
     *  set the appropriate flags, edges and weights.
     */
    InstrSwitchCondition(cond_instr);
    INSTR_SET_FLAG(cond_instr, INSTR_CLASS_BRANCH);
    BBLRemoveEdges(cond_instr->i_Bblock);

    jump_bbl = BBLCreate();
    jump_bbl->b_type = BT_UNCOND;
    BBLAddEdge(cond_instr->i_Bblock, next_instr->i_Bblock, ET_TRUE);
    BBLAddEdge(cond_instr->i_Bblock, jump_bbl, ET_FALSE);
    BBLAddEdge(jump_bbl, target_instr->i_Bblock, ET_UNCOND);
    InsertBBLAfterBBL(jump_bbl, cond_instr->i_Bblock);

    true_edge = BBLGetTrueEdge(cond_instr->i_Bblock);
    true_edge->e_edgeWeight = false_edge_weight;

    false_edge = BBLGetFalseEdge(cond_instr->i_Bblock);
    false_edge->e_edgeWeight = true_edge_weight;

    jump_bbl->b_weight = true_edge_weight;
    if (jump_bbl->b_weight >= lowest_hot_weight && lowest_hot_weight != 0) {
        BBL_SET_FLAG(jump_bbl, BBL_HOT_FLAG);
    }

    uncond_edge = BBLGetEdgeToBBL(jump_bbl, target_instr->i_Bblock);
    uncond_edge->e_edgeWeight = true_edge_weight;

    jump_instr = InstrCreateJMP();
    INSTR_SET_FLAG(jump_instr, INSTR_CLASS_JUMP);
    BBLPrependInstr(jump_bbl, jump_instr);

    return true;
}



/*
 *  DetachDisassemblyEntries() --
 *
 *  Inserts jumps in the very beginning of the instruction stream (before the
 *  _init() function) and after the first instruction of the _start() function.
 *  The idea is that the above jumps will get obfuscated and disassembler will
 *  get confused right from the start.
 */
void DetachDisassemblyEntries() {
    Instruction *tmp_instr, *jump;
    Block *old_bbl, *new_bbl;

    //cut off disassembly at the very start with a fake jump
    //this is to confuse linear disassembly
    new_bbl = BBLCreate();
    new_bbl->b_type = BT_UNCOND;
    new_bbl->b_weight = bbl_list->b_weight;

    jump = InstrCreateJMP();
    INSTR_SET_FLAG(jump, BBL_START);
    INSTR_SET_FLAG(jump, BBL_END);
    BblAppendInstr(new_bbl, jump);
    BBLAddEdge(new_bbl, bbl_list, ET_UNCOND);

    old_bbl = bbl_list;
    InsertBBLBeforeBBL(new_bbl, bbl_list);

    if (old_bbl->b_fun->f_first == old_bbl) {
        new_bbl->b_fun->f_first = new_bbl;
    }

    if (old_bbl->b_fun->f_last == old_bbl) {
        new_bbl->b_fun->f_last = new_bbl;
    }

    if (BBL_HAS_FLAG(old_bbl, FUN_START)) {
        BBL_REMOVE_FLAG(old_bbl, FUN_START);
        BBL_SET_FLAG(new_bbl, FUN_START);
    }

    if (BBL_HAS_FLAG(old_bbl, FUN_END)) {
        BBL_REMOVE_FLAG(old_bbl, FUN_END);
        BBL_SET_FLAG (new_bbl, FUN_END);
    }

    //cut off the first recursive block (the start instruction block)
    //this is to confuse recursive disassembly
    tmp_instr = startins->i_next;
    BBLDivideBlock(tmp_instr->i_Bblock, tmp_instr);
    jump = InstrCreateJMP();
    BblAppendInstr(startins->i_Bblock, jump);
    BBLRemoveEdge(startins->i_Bblock, tmp_instr->i_Bblock);
    BBLAddEdge(startins->i_Bblock, tmp_instr->i_Bblock, ET_UNCOND);

    return;
}



/*
 *  IsolateHotRegions() --
 *
 *  ...
 */
void IsolateHotRegions() {
    WARNING("[IsolateHotRegions]: just called");
    WARNING("[IsolateHotRegions]: returning");
    return;
}

