/*
 *  finalize.c
 *  
 *  This file contains the functions necessary to finalize:
 *    1. Branch function called by the former jumps - bf_jump(),
 *    2. Branch function called by the former calls - bf_call(),
 *    3. Restore function used in signal based obfuscation - my_restore().
 *  Before execution comes here the functions above only have one instruction
    !!!CHANGE
 *  (pushf, nop, nop respectively). The main task of the functions in this file
 *  is to complete the above 3 functions with the following:
 *    - the perfect hash function code which takes the address from the known
 *      location on the stack and generates the index into the table located
 *      in the data section,
 *    - the code to use the output of the perfect hash function to fetch the
 *      appropriate return address(es),
 *    - the code to place the fetched address(es) into known stack location,
 *  and to insert the table with the target addresses into the data section.
 */
#include "obf_global.h"
#include "obf.h"
#include "finalize.h"
#include "obf_macros.h"
#include "utilities.h"
#include "linkage.h"
#include "flowgraph.h"
#include "sysdep.h"
#include "x86-registers.h"
#include "upm.h"
#include "relocate.h"
#include "perfect.h"
#include "liveness.h"
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <clone.h>



static void FinalizeJumpBranchFunction();
static void FinalizeCallBranchFunction();
static void FinalizeMyRestore();

static void GenerateIndices(LinkedList *addr_list, const char *addr_fname,
    const char *addr_ind_fname);
static void InsertHashFunction(const char *obj_file_from, Block *block_to);

static void CollectBFCallRetAddresses(LinkedList **addr_list,
    const char *addr_fname, LinkedList **bf_calls_list);
static void FillBFTargetTable(const char *addr_ind_fname,
    LinkedList *bf_calls_list, int num_columns);
static void CompleteJumpBranchFunction(int num_columns);
static void CopyJumpBranchFunction();
static void CompleteCallBranchFunction();

static void CollectTrapAddresses(LinkedList **addr_list,
    const char *addr_fname, LinkedList **traps_list);
static void FillTrapTargetTable(const char *addr_ind_fname,
    LinkedList *traps_list, int num_columns);
static void CompleteMyRestore();
static void CompleteShortMyRestore();
static void CompleteLongMyRestore();



/*
 *  ObfuscateFinalize() --
 *
 *  This function finalizes the instruction stream by completing the three
 *  functions at the end of the stream that require the knowledge of final
 *  instruction addresses. The name is a bit misleading in the way that after
 *  this function executes it is still OK to modify the bytes in the
 *  instruction stream (ex. addresses embedded in instructions, offsets,
 *  constants). What has to remain constant is the length of the instructions,
 *  since changing the instruction's length will change their addresses, which,
 *  in turn, will invalidate the perfect hash functions incorporated into
 *  two branch functions and one restore function finalized here.
 */
void ObfuscateFinalize() {

    RecalculateAddresses();

    if (INSERT_BRANCH_FUNCTIONS) {
        REPORT(RP_MINOR, "Finalizing branch function called by former jumps");
        FinalizeJumpBranchFunction();

        if (BF_CONVERT_CALLS) {
            REPORT(RP_MINOR, "Finalizing branch function called by"
                " former calls");
            FinalizeCallBranchFunction();
        }
    }

    if (INSTALL_TRAPS) {
        REPORT(RP_MINOR, "Finalizing restore");
        FinalizeMyRestore();
    }

    return;
}



/*
 *  FinalizeJumpBranchFunction() --
 *
 *  Inserts the table (bf_target_table) into the data section.
 *  Finalizes bf_jump() function to include:
 *   1. The perfect hash function, which gets the return address currently on
 *      the stack as input and returns the index into the above table at which
 *      the target address(es) XORed with this current return address is/are
 *      stored.
 *   2. The sequence of instructions that, using the data from the table,
 *      overwrites the current return addresses with the target address.
 *  Inserts data used by the perfect hash function into the data section.
 */
static void FinalizeJumpBranchFunction() {
    LinkedList *addr_list;
    LinkedList *bf_calls_list;
    char *addr_fname = "tmp_addr";
    char *addr_ind_fname = "tmp_addr_ind";

    int num_columns = 1;
    if (BF_CONVERT_CALLS) {
        num_columns++;
    }

    REPORT(RP_DETAIL, "Collecting the return addresses of"
        " branch function calls");
    CollectBFCallRetAddresses(&addr_list, addr_fname, &bf_calls_list);

    /* no point to proceed if no input for perfect hashing */
    if (ll_size(bf_calls_list) <= 0) {
        WARNING("No branch function calls found\n");
        return;
    }

    REPORT(RP_DETAIL, "Generating unique index for each return address");
    GenerateIndices(addr_list, addr_fname, addr_ind_fname);

    REPORT(RP_DETAIL, "Filling the table of branch function targets");
    FillBFTargetTable(addr_ind_fname, bf_calls_list, num_columns);

    REPORT(RP_DETAIL, "Inserting the perfect hash function");
    InsertHashFunction("phash.o", bf_jump_first_bbl);

    REPORT(RP_DETAIL, "Completing the jump branch function");
    CompleteJumpBranchFunction(num_columns);

    remove("phash.o");
    remove(addr_fname);
    remove(addr_ind_fname);

    REPORT(RP_DETAIL, "Jump branch function is finalized");

    //PrintControlTransferEdges();

    return;
}



/*
 *  FinalizeCallBranchFunction() --
 *
 *  Copies the contents of bf_jump() function into the bf_call() function.
 *  Adds to bf_call() the instructions that will derive the second target
 *  address from the bf_target_table and overwrite the deeper return address
 *  with it.
 */
static void FinalizeCallBranchFunction() {

    REPORT(RP_DETAIL, "Cloning the jump branch function");
    CopyJumpBranchFunction();

    REPORT(RP_DETAIL, "Completing the call branch function");
    CompleteCallBranchFunction();

    REPORT(RP_DETAIL, "Call branch function is finalized");

    //PrintControlTransferEdges();

    return;
}



/*
 *  FinalizeMyRestore() --
 *
 *  Inserts the table (trap_target_table) into the data section.
 *  Finalizes my restore stub to include:
 *   1. The perfect hash function, which gets the trap address as input and
 *      returns the index into the above table at which the target address(es)
 *      XORed with the trap address is/are stored.
 *   2. The sequence of instructions that, using the data from the table,
 *      overwrites the 0, 1 or 2 return addresses depending on whether
 *      the original instructon was return, jump or call, respectively.
 *  Inserts data used by the perfect hash function into the data section.
 */
static void FinalizeMyRestore() {
    LinkedList *addr_list;
    LinkedList *traps_list;
    char *addr_fname = "tmp_addr";
    char *addr_ind_fname = "tmp_addr_ind";

    int num_columns = 1;
    if (TRAPS_CONVERT_CALLS | TRAPS_CONVERT_RETURNS) {
        num_columns = 3;
    }

    REPORT(RP_DETAIL, "Collecting the addresses of traps");
    CollectTrapAddresses(&addr_list, addr_fname, &traps_list);

    /* no point to proceed if no input for perfect hashing */
    if (ll_size(traps_list) <= 0) {
        WARNING("No traps found\n");
        return;
    }

    REPORT(RP_DETAIL, "Generating unique index for each trap address");
    GenerateIndices(addr_list, addr_fname, addr_ind_fname);

    REPORT(RP_DETAIL, "Filling the table of trap targets");
    FillTrapTargetTable(addr_ind_fname, traps_list, num_columns);

    REPORT(RP_DETAIL, "Inserting the perfect hash function");
    InsertHashFunction("phash.o", my_restore_bbls[0]); //2nd argument not used

    REPORT(RP_DETAIL, "Completing the restore function");
    CompleteMyRestore();

    remove("phash.o");
    remove(addr_fname);
    remove(addr_ind_fname);

    REPORT(RP_DETAIL, "Restore function is finalized");

    //PrintControlTransferEdges();

    return;
}



/*
 *  GenerateIndices(addr_list, addr_fname, addr_ind_fname) --
 *
 *  Creates phash.o file (object file containing the pefect hash function)
 *  and, using it, generates unique index for every address in the file with
 *  the name addr_fname. File with the name addr_ind_fname will contain pairs
 *  <address, index>.
 */
static void GenerateIndices(LinkedList *addr_list, const char *addr_fname,
    const char *addr_ind_fname) {

    char *command;
    char *options = "gcc -O3 -fomit-frame-pointer";
    char *hash_fun_name = "tmp_phash";
    char *space = " ";

    REPORT(RP_SMLOOP, "Building perfect hash table");
    /* create the phash.c, phash.h and phash_main.c files */
    create_hash(addr_list);

    /* create phash.o file */
    command = zalloc(80);
    strcat(command, options);
    strcat(command, space);
    strcat(command, "-c phash.c");
    REPORT(RP_SMLOOP, "Executing: %s", command);

    if (system(command) == -1) {
        PFATAL("phash.o not created");
    }
    free(command);

    /* create tmp_phash file */
    command = zalloc(80);
    strcat(command, options);
    strcat(command, space);
    strcat(command, "-o");
    strcat(command, space);
    strcat(command, hash_fun_name);
    strcat(command, space);
    strcat(command, "phash.o phash_main.c");
    REPORT(RP_SMLOOP, "Executing: %s", command);

    if (system(command) == -1) {
        PFATAL("tmp_phash not created");
    }
    free(command);

    /* given file with addresses generate file with addresses and indices */
    command = zalloc(80);
    strcat(command, hash_fun_name);
    strcat(command, space);
    strcat(command, addr_fname);
    strcat(command, space);
    strcat(command, addr_ind_fname);
    REPORT(RP_SMLOOP, "Executing: %s", command);

    if (system(command) == -1) {
        PFATAL("tmp_phash failed");
    }
    free(command);

    /* remove tmp files created in this function that are not needed */
    remove("phash.c");
    remove("phash.h");
    remove("phash_main.c");
    remove(hash_fun_name);

    REPORT(RP_SMLOOP, "Generated unique index for each address");

    return;
}



//get rid of block_to
/*
 *  InsertHashFunction(obj_file_from, block_to) --
 *
 *  Inserts the data and text of the perfect hash function from the file with
 *  the name obj_file_from. Text is inserted into the block pointed by
 *  block_to.
 */
static void InsertHashFunction(const char *obj_file_from, Block *block_to) {
    Elf32_Ehdr *hdr;
    Elf32_Shdr *shdr;
    Elf32_Shdr *sym_hdr;
    Elf32_Shdr *rel_hdr;
    Elf32_Shdr *txt_hdr;
    Elf32_Rel *rel;
    Elf32_Sym *symt;
    int symt_size;
    int rel_size;
    char *strn_tbl;
    char *sym_strn_tbl;
    int fd;
    UINT8 *buf;
    struct stat file_stat;
    int i;
    ADDRESS tab_addr = -1;
    ADDRESS scr_addr = -1;
    Instruction *tmp_instr;
    Instruction *first_nop_instr;
    Iterator *iterator;

    if (stat(obj_file_from, &file_stat) < 0) {
        PFATAL("problem statting file");
    }

    i = file_stat.st_size;
    buf = zalloc(i);

    fd = open(obj_file_from, O_RDONLY);
    if (fd < 0) {
        PFATAL("problem opening binary");
    }
    if (read(fd, buf, i) < 0) {
        PFATAL("problem reading binary");
    }
    close(fd);

    /* Insert Data */

    hdr = (Elf32_Ehdr *)buf;
    shdr = (Elf32_Shdr *)&buf[hdr->e_shoff];
    strn_tbl = (char *)(buf + shdr[hdr->e_shstrndx].sh_offset);

    for (i = 0; i < hdr->e_shnum; i++) {
        //get the symbol table header and symbol table
        if (shdr[i].sh_type == SHT_SYMTAB) {
            sym_hdr = &shdr[i];
            symt = (Elf32_Sym *)(buf + sym_hdr->sh_offset);
        }
        //get the symbol table string table and it's header
        if (shdr[i].sh_type == SHT_STRTAB) {
            sym_strn_tbl = (char *)(buf + shdr[i].sh_offset);
        }
        //get the text section header
        if (strcmp(".text", &strn_tbl[shdr[i].sh_name]) == 0) {
            txt_hdr = &shdr[i];
        }
        //get the relocations for the text section (addrs for tab and/or
        //scramble patches)
        if (strcmp(".rel.text", &strn_tbl[shdr[i].sh_name]) == 0) {
            rel_hdr = &shdr[i];
            rel = (Elf32_Rel *)(buf + rel_hdr->sh_offset);
        }
    }

    symt_size = sym_hdr->sh_size / sizeof(Elf32_Sym);
    for (i = 0; i < symt_size; i++) {
        //get the tab symbol
        if (strcmp("tab", &sym_strn_tbl[symt[i].st_name]) == 0) {
            tab_addr = SECTION_REF_DATA->s_newdatasize;
            IncreaseSectionSize(SECTION_REF_DATA, symt[i].st_size);
            memcpy(&(SECTION_REF_DATA->s_newdata)[tab_addr],
                buf + shdr[symt[i].st_shndx].sh_offset + symt[i].st_value,
                symt[i].st_size);
        //get the scramble symbol
        } else if (strcmp("scramble", &sym_strn_tbl[symt[i].st_name]) == 0) {
            scr_addr = SECTION_REF_DATA->s_newdatasize;
            IncreaseSectionSize(SECTION_REF_DATA, symt[i].st_size);
            memcpy(&(SECTION_REF_DATA->s_newdata)[scr_addr],
                buf + shdr[symt[i].st_shndx].sh_offset + symt[i].st_value,
                symt[i].st_size);
        }
    }

    rel_size = rel_hdr->sh_size / sizeof(Elf32_Rel);

    //disassemble each instruction into the struct Instruction so that we
    //can create the actual hash function

    /* Insert Text */

    iterator = ll_iterator(my_restore_bbls_lists[0]);
    FOREACH_ITEM(iterator, block_to, Block *) {
        i = 0;
        first_nop_instr = block_to->b_first->i_next;
        while (i < shdr[1].sh_size) {
            tmp_instr = InstrCreate();
            tmp_instr->i_length = DisassembleInstr(0,
                buf + txt_hdr->sh_offset + i, tmp_instr);
            tmp_instr->i_address = i;
            InstrSetClassFlags(tmp_instr);
            InsertInstrBeforeInstr(tmp_instr, first_nop_instr);
            i += tmp_instr->i_length;
        }

        //WARNING("Inserted %d bytes", i); //including ret, removed later
        RemoveNOPs(first_nop_instr, i - 1);

        tmp_instr = block_to->b_first;
        while (tmp_instr->i_Bblock == block_to) {
            for (i = 0; i < rel_size; i++) {
                if (rel[i].r_offset >= tmp_instr->i_address &&
                    rel[i].r_offset < tmp_instr->i_address +
                    tmp_instr->i_length) {

                    int new_off = rel[i].r_offset - tmp_instr->i_address;
                    Operand *op;

                    op = GetOperandForInstrOffset(tmp_instr, new_off);
                    if (strcmp("tab", &sym_strn_tbl[symt[
                        ELF32_R_SYM(rel[i].r_info)].st_name]) == 0) {
                        AddPatchSectionAddressIntoInstr(SECTION_REF_DATA,
                            tab_addr, tmp_instr, op);
                    } else if (strcmp("scramble", &sym_strn_tbl[symt[
                        ELF32_R_SYM(rel[i].r_info)].st_name]) == 0) {
                        AddPatchSectionAddressIntoInstr(SECTION_REF_DATA,
                            scr_addr, tmp_instr, op);
                    }
                }
            }

            tmp_instr->i_address = 0xfeedface;

            if (tmp_instr->i_next != NULL && isReturnClass(tmp_instr->i_next)) {
                InstrDisconnect(tmp_instr->i_next);
            }
            //tmp_instr->i_address = 0xfeedface;
            tmp_instr = tmp_instr->i_next;
        }

        //adjust the displacement of the "mov ...(%esp), some_reg" to 0x20
        FOREACH_INSTR_IN_BBL(block_to, tmp_instr) {
            if (tmp_instr->i_src &&
                tmp_instr->i_src->a_base_reg == (reg_32 | reg_sp)) {
                tmp_instr->i_src->a_displacement = 0x20;
            }
        }
    }

    return;
}



/*
 *  CollectBFCallRetAddresses(addr_list, addr_fname, bf_calls_list) --
 *
 *  Iterates through the binary and collects the set of addresses of
 *  instructions following the branch function calls into the addr_list
 *  and file with the name addr_fname. Also collects the branch function
 *  call instructions into the bf_calls_list.
 */
static void CollectBFCallRetAddresses(LinkedList **addr_list,
    const char *addr_fname, LinkedList **bf_calls_list) {

    FILE *addr_file;
    Block *biterator;

    *addr_list = ll_new(LIST_UNSORTED);
    *bf_calls_list = ll_new(LIST_UNSORTED);
    
    addr_file = fopen(addr_fname, "w");
    if (!addr_file) {
        PFATAL("Could not create temp file");
    }

    FOREACH_BBL(biterator) {
        if ((INSTR_HAS_FLAG(biterator->b_last, INSTR_JBF_CALL_FLAG))
            || (INSTR_HAS_FLAG(biterator->b_last, INSTR_CBF_CALL_FLAG))) {
            //next instr final address into addr_list and addr_file
            ll_add(*addr_list, (void *)(biterator->b_last->i_final_address +
                biterator->b_last->i_length));
            fprintf(addr_file, "%x\n", biterator->b_last->i_final_address +
                biterator->b_last->i_length);
            //instr itself into bf_calls_list
            ll_add(*bf_calls_list, biterator->b_last);
        }
    }
    fclose(addr_file);

    REPORT(RP_SMLOOP, "Collected %d final addresses of instructions"
        " following bf calls", ll_size(*bf_calls_list));

    return;
}



/*
 *  FillBFTargetTable(addr_ind_fname, bf_calls_list, num_columns) --
 *
 *  Allocates a table (bf_target_table) in the data section and, at each index
 *  taken from the file with the name addr_ind_fname, stores the corresponding
 *  address of the instruction following the branch function call XORed with
 *  the original target of jump replaced by this branch function call.
 */
static void FillBFTargetTable(const char *addr_ind_fname,
    LinkedList *bf_calls_list, int num_columns) {

    FILE *addr_ind_file;
    Iterator *iterator;
    Edge *funlink_edge;
    Edge *fundirect_edge;
    Instruction *bf_call_instr;
    Instruction *target_instr;
    Instruction *return_instr;
    int new_index;
    ADDRESS addr;

    addr_ind_file = fopen(addr_ind_fname, "r");
    if (!addr_ind_file) {
        PFATAL("open failed for map file");
    }

    /*
     * At this point we will create a table of addresses such that at index i
     * taken from the above file we will have either
     *  a) one entry:
     *         1. target_addr ^ next_instr_addr,
     *     if only jumps are converted, or
     *  b) two entries:
     *         1. target_addr ^ next_instr_addr,
     *         2. return_addr ^ next_instr_addr,
     *     if calls are converted along with jumps.
     */
    bf_target_table_offset = SECTION_REF_DATA->s_newdatasize;
    IncreaseSectionSize(SECTION_REF_DATA,
        ll_size(bf_calls_list) * sizeof(ADDRESS) * num_columns);

    iterator = ll_iterator(bf_calls_list);
    FOREACH_ITEM(iterator, bf_call_instr, Instruction *) {
        if (fscanf(addr_ind_file, "%x %d", &addr, &new_index) <= 0) {
            PFATAL("problem scanning index from file");
        }

        /* if former jump, fill only slot 1 */
        if (INSTR_HAS_FLAG(bf_call_instr, INSTR_JBF_CALL_FLAG)) {

            /* get the original target of this jump */
            target_instr = INSTR_TARGET(bf_call_instr);

            /* write <target_addr ^ next_instr_addr> into the table[index] */
            storeLE(&(SECTION_REF_DATA->s_newdata)[bf_target_table_offset +
                new_index * sizeof(ADDRESS) * num_columns],
                target_instr->i_final_address ^
                (bf_call_instr->i_final_address + bf_call_instr->i_length));

            /* adjust BBL_TYPE and edges */
            BBLRemoveEdgeTyped(bf_call_instr->i_Bblock,
                target_instr->i_Bblock, ET_UNCOND);
            bf_call_instr->i_Bblock->b_type = BT_CALL;
            BBLAddEdge(bf_call_instr->i_Bblock,
                bf_jump_first_bbl, ET_FUNDIRECT);
            if (bf_call_instr->i_Bblock->b_next != NULL) {
                BBLAddEdge(bf_call_instr->i_Bblock,
                    bf_call_instr->i_Bblock->b_next, ET_FUNLINK);
            }

        /* if former call, fill both slot 1 and slot 2*/
        } else if (INSTR_HAS_FLAG(bf_call_instr, INSTR_CBF_CALL_FLAG)) {

            /* get the original target of this call */
            fundirect_edge = BBLGetFunDirectEdge(bf_call_instr->i_Bblock);
            target_instr = fundirect_edge->e_to->b_first;

            /* write <target_addr^next_instr_addr> into the table[index + 0] */
            storeLE(&(SECTION_REF_DATA->s_newdata)[bf_target_table_offset +
                new_index * sizeof(ADDRESS) * num_columns],
                target_instr->i_final_address ^
                (bf_call_instr->i_final_address + bf_call_instr->i_length));

            /* get the original return of this call */
            funlink_edge = BBLGetFunlinkEdge(bf_call_instr->i_Bblock);
            return_instr = funlink_edge->e_to->b_first;

            /* write <return_addr^next_instr_addr> into the table[index + 1] */
            storeLE(&(SECTION_REF_DATA->s_newdata)[bf_target_table_offset +
                new_index * sizeof(ADDRESS) * num_columns]
                + sizeof(ADDRESS) * (num_columns - 1),
                return_instr->i_final_address ^
                (bf_call_instr->i_final_address + bf_call_instr->i_length));

            /* adjust edges */
            BBLRemoveEdgeTyped(bf_call_instr->i_Bblock,
                target_instr->i_Bblock, ET_FUNDIRECT);
            BBLAddEdge(bf_call_instr->i_Bblock,
                bf_call_first_bbl, ET_FUNDIRECT);

        } else {
            MALFUNCTION("Non control transfer instr in the list of bf calls");
        }
    }

    fclose(addr_ind_file);

    if (num_columns == 1) {
        REPORT(RP_SMLOOP, "Table of <&target^&next_instr> entries filled");
    } else {
        REPORT(RP_SMLOOP, "Table of "
            "<&target^&next_instr, &ret_instr^&next_instr> entries filled");
    }

    return;
}



/*
 *  CompleteJumpBranchFunction(num_columns) --
 *
 *  Adds to the bf_jump() function the instructons that will retrieve the
 *  target_address from the bf_target_table and place it on the top of the
 *  stack, so that ret instruction will return to it.
 */
static void CompleteJumpBranchFunction(int num_columns) {
    Block *new_bbl;
    Instruction *tmp_instr;
    REGISTER_BITVECTOR defined_regs;
    int i;

    new_bbl = bf_jump_first_bbl;

    //imul $0x<num_columns * 4>, %eax, %eax
    tmp_instr = InstrCreateIMulImm8ByReg(num_columns *
        sizeof(ADDRESS), reg_a | reg_32);
    InsertInstrBeforeInstr(tmp_instr, new_bbl->b_last);

    //mov <&bf_target_table_offset>(%eax), %eax
    tmp_instr = InstrCreateMovMemDisp32BaseToReg(0x0, reg_a | reg_32,
        reg_a | reg_32);
    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA,
        bf_target_table_offset, tmp_instr, tmp_instr->i_src);
    tmp_instr->i_literal_data = (UINT8 *)0xbabecafe; //need for call_bf()
    InsertInstrBeforeInstr(tmp_instr, new_bbl->b_last);

    //temporarily disconnect the pushf from the basic block, otherwise
    //ComputeRegistersDefinedInBBL mistakenly identifies this as defining %esi
    tmp_instr = InstrDisconnect(new_bbl->b_first);
    defined_regs = ComputeRegistersDefinedInBBL(new_bbl);
    InsertInstrBeforeInstr(tmp_instr, new_bbl->b_first);

    int num_pushes = 1; //start at 1 since we've already pushed the flags
    for (i = 0; i < 6; i++) {
        if (defined_regs & (1 << i)) {
            //push %<reg_i>
            tmp_instr = InstrCreatePush(reg_32 | BitRegisterToRegister(i));
            InsertInstrAfterInstr(tmp_instr, new_bbl->b_first);
            num_pushes++;
        }
    }

    //xor %eax, <num_pushes * 4>(%esp)
    tmp_instr = InstrCreateXorRegWithMemDisp8Base(reg_a | reg_32,
        num_pushes * sizeof(ADDRESS), reg_sp | reg_32);
    InsertInstrBeforeInstr(tmp_instr, new_bbl->b_last);

    //adjust the offset from %esp
    FOREACH_INSTR_IN_BBL(new_bbl, tmp_instr) {
        if (tmp_instr->i_src &&
            (tmp_instr->i_src->a_base_reg == (reg_32 | reg_sp))) {
            tmp_instr->i_src->a_displacement = num_pushes * sizeof(ADDRESS);
        }
    }

    for (i = 0; i < 6; i++) {
        if (defined_regs & (1 << i)) {
            //pop %<reg_i>
            tmp_instr = InstrCreatePop(reg_32 | BitRegisterToRegister(i));
            InsertInstrBeforeInstr(tmp_instr, new_bbl->b_last);
        }
    }

    //popf
    tmp_instr = InstrCreateRestore();
    InsertInstrBeforeInstr(tmp_instr, new_bbl->b_last);

    new_bbl->b_first->i_prev = new_bbl->b_prev->b_last;
    new_bbl->b_prev->b_last->i_next = new_bbl->b_first;

    return;
}



/*
 *  CopyJumpBranchFunction() --
 *
 *  Copies the instructions in the bf_jump() function into the bf_call()
 *  function.
 */
static void CopyJumpBranchFunction() {
    Instruction *iiterator;
    Instruction *new_instr;

    FOREACH_INSTR_IN_BBL(bf_jump_first_bbl, iiterator) {
        new_instr = InstrClone(iiterator);
        BblAppendInstr(bf_call_first_bbl, new_instr);
    }

    return;
}



/*
 *  CompleteCallBranchFunction() --
 *
 *  Adds to the bf_call() function the instructions that will get the second
 *  target address (the address of the instruction to which the original call
 *  returned) and write it as a deeper return address.
 */
static void CompleteCallBranchFunction() {
    Instruction *new_instr;
    Instruction *iiterator;
    Patch *upm;
    PatchSectionToInstr *patch;
    LinkedList *upms;

    //push (%esp) //Duplicate the last value on the stack
    new_instr = InstrCreatePushMemBase(reg_32 | reg_sp);
    InsertInstrAfterInstr(new_instr, bf_call_first_bbl->b_first);

    FOREACH_INSTR_IN_BBL(bf_call_first_bbl, iiterator) {
        if (iiterator->i_literal_data == (UINT8 *)0xbabecafe) {
            break;
        }
    }

    //mov %eax, %edx
    new_instr = InstrCreateMovRegToReg(reg_32 | reg_a, reg_32 | reg_d);
    InsertInstrBeforeInstr(new_instr, iiterator);

    iiterator = iiterator->i_next;

    //mov <&bf_target_table_offset+0x4>(%edx), %eax
    new_instr = InstrClone(iiterator->i_prev);
    upms = UPMTableGetByDest(new_instr->i_src);
    upm = ll_get(upms, 0);
    ll_free(upms);
    if (upm == NULL) {
        MALFUNCTION("UPM was not properly cloned!");
    }
    patch = upm->p_patch;
    patch->p_src_offset += sizeof(ADDRESS);
    new_instr->i_src->a_base_reg = reg_32 | reg_d;
    InsertInstrAfterInstr(new_instr, iiterator);

    iiterator = iiterator->i_next;

    //xor %eax, <deeper_ret_addr_offset>(%esp)
    new_instr = InstrClone(iiterator->i_prev);
    new_instr->i_dest->a_displacement += sizeof(ADDRESS);
    InsertInstrAfterInstr(new_instr, iiterator);

    bf_call_first_bbl->b_first->i_prev = bf_call_first_bbl->b_prev->b_last;
    bf_call_first_bbl->b_prev->b_last->i_next = bf_call_first_bbl->b_first;

    return;
}



/*
 *  CollectTrapAddresses(addr_list, addr_fname, traps_list) --
 *
 *  Iterates through the binary and collects the set of addresses of trap
 *  instructions into the addr_list and file with the name addr_fname. Also
 *  collects the trap instructions into the traps_list.
 */
static void CollectTrapAddresses(LinkedList **addr_list,
    const char *addr_fname, LinkedList **traps_list) {

    FILE *addr_file;
    Block *biterator;

    *addr_list = ll_new(LIST_UNSORTED);
    *traps_list = ll_new(LIST_UNSORTED);

    addr_file = fopen(addr_fname, "w");
    if (!addr_file) {
        PFATAL("Could not create temp file");
    }

    FOREACH_BBL(biterator) {
        if ((INSTR_HAS_FLAG(biterator->b_last, INSTR_SIGILL_FLAG))
            || (INSTR_HAS_FLAG(biterator->b_last, INSTR_SIGFPE_FLAG))
            || (INSTR_HAS_FLAG(biterator->b_last, INSTR_SIGSEGV_FLAG))) {
            //instr final address into addr_list and addr_file
            ll_add(*addr_list, (void *)(biterator->b_last->i_final_address));
            fprintf(addr_file, "%x\n", biterator->b_last->i_final_address);
            //instr itself into traps_list
            ll_add(*traps_list, biterator->b_last);
        }
    }
    fclose(addr_file);

    REPORT(RP_SMLOOP, "Collected %d final addresses of traps",
        ll_size(*traps_list));

    return;
}



/*
 *  FillTrapTargetTable(addr_ind_fname, traps_list, num_columns) --
 *
 *  Allocates a table (trap_target_table) in the data section and, at each
 *  index from the file with the name addr_ind_fname, stores the corresponding
 *  trap instruction address XORed with the target of that trap instruction.
 */
static void FillTrapTargetTable(const char *addr_ind_fname,
    LinkedList *traps_list, int num_columns) {

    FILE *addr_ind_file;
    Iterator *iterator;
    Edge *funlink_edge;
    Edge *fundirect_edge;
    Instruction *fault_instr;
    Instruction *target_instr;
    Instruction *return_instr;
    int new_index;
    ADDRESS addr;

    addr_ind_file = fopen(addr_ind_fname, "r");
    if (!addr_ind_file) {
        PFATAL("open failed for map file");
    }

    /*
     * At this point we will create a table of addresses such that at index i
     * taken from the above file we will have either
     *  a) one entry:
     *         1. target_addr ^ fault_addr,
     *     if only jumps are converted, or
     *  b) the following three entries:
     *         1. target_addr ^ fault_addr
     *         2. return_addr ^ fault_addr
     *         3. 0, if instruction at fault_addr used to be a return,
     *            1, if instruction at fault_addr used to be a jump,
     *            2, if instruction at fault_addr used to be a call,
     *     if calls/returns are converted in addition to jumps
     * For b), having 0 as entry 3 means that we do not need entries 1 and 2,
     * because the target (return) address is already on the stack. Having 1
     * means that we only need entry 1 to derive the target_addr using the
     * fault_addr. Having 2 means that we need both 1 and 2, to derive the
     * target of the former call and the return as well.
     */
    trap_target_table_offset = SECTION_REF_DATA->s_newdatasize;
    IncreaseSectionSize(SECTION_REF_DATA,
        ll_size(traps_list) * sizeof(ADDRESS) * num_columns);

    iterator = ll_iterator(traps_list);
    if (num_columns == 1) { //only jmp are converted, no call, no ret
        FOREACH_ITEM(iterator, fault_instr, Instruction *) {
            if (fscanf(addr_ind_file, "%x %d", &addr, &new_index) <= 0) {
                PFATAL("problem scanning index from file");
            }

            /* fill the only slot - slot 1 */

            //get the original target of this jump
            target_instr = INSTR_TARGET(fault_instr);

            //write <target_addr XOR fault_addr> into the table
            storeLE(&(SECTION_REF_DATA->s_newdata)[trap_target_table_offset +
                new_index * sizeof(ADDRESS)],
                target_instr->i_final_address ^ fault_instr->i_final_address);

            /* adjust BBL_TYPE and edges */
            BBLRemoveEdgeTyped(fault_instr->i_Bblock, target_instr->i_Bblock,
                ET_UNCOND);
            fault_instr->i_Bblock->b_type = BT_NORMAL;
            if (fault_instr->i_Bblock->b_next != NULL) {
                BBLAddEdge(fault_instr->i_Bblock,
                    fault_instr->i_Bblock->b_next, ET_NORMAL);
            }
        }
    } else { //call/ret are converted along with jmp
        FOREACH_ITEM(iterator, fault_instr, Instruction *) {
            if (fscanf(addr_ind_file, "%x %d", &addr, &new_index) <= 0) {
                PFATAL("problem scanning index from file");
            }

            if (fault_instr->i_Bblock->b_type == BT_CALL) {

                /* fill slot 1 */

                //get the original target of this call
                fundirect_edge = BBLGetFunDirectEdge(fault_instr->i_Bblock);
                target_instr = fundirect_edge->e_to->b_first;

                //write <target_addr XOR fault_addr> into the table[index+0]
                storeLE(&(SECTION_REF_DATA->s_newdata)[trap_target_table_offset
                    + new_index * sizeof(ADDRESS) * num_columns],
                    target_instr->i_final_address ^
                    fault_instr->i_final_address);

                /* fill slot 2 */

                //get the instruction to which execution returned after the call
                funlink_edge = BBLGetFunlinkEdge(fault_instr->i_Bblock);
                return_instr = funlink_edge->e_to->b_first;

                //write <return_addr XOR fault_addr> into the table[index+1]
                storeLE(&(SECTION_REF_DATA->s_newdata)[trap_target_table_offset
                    + new_index * sizeof(ADDRESS) * num_columns
                    + sizeof(ADDRESS) * (num_columns - 2)],
                    return_instr->i_final_address ^
                    fault_instr->i_final_address);

                /* fill slot 3 with 0x2 */

                //write 0x2 into the table[index+2]
                storeLE(&(SECTION_REF_DATA->s_newdata)[trap_target_table_offset
                    + new_index * sizeof(ADDRESS) * num_columns
                    + sizeof(ADDRESS) * (num_columns - 1)], 2);

                /* adjust BBL_TYPE and edges */
                BBLRemoveOutgoingEdges(fault_instr->i_Bblock);
                fault_instr->i_Bblock->b_type = BT_NORMAL;
                if (fault_instr->i_Bblock->b_next != NULL) {
                    BBLAddEdge(fault_instr->i_Bblock,
                        fault_instr->i_Bblock->b_next, ET_NORMAL);
                }

            } else if (fault_instr->i_Bblock->b_type == BT_UNCOND) {

                /* fill slot 1 */

                //get the original target of this jump
                target_instr = INSTR_TARGET(fault_instr);

                //write <target_addr XOR fault_addr> into the table[index+0]
                storeLE(&(SECTION_REF_DATA->s_newdata)[trap_target_table_offset
                    + new_index * sizeof(ADDRESS) * num_columns],
                    target_instr->i_final_address ^
                    fault_instr->i_final_address);

                /* do not fill slot 2, of fill with some junk if need to */

                /* fill slot 3 with 0x1 */

                //write 0x1 into the table[index+2]
                storeLE(&(SECTION_REF_DATA->s_newdata)[trap_target_table_offset
                    + new_index * sizeof(ADDRESS) * num_columns
                    + sizeof(ADDRESS) * (num_columns - 1)], 1);

                /* adjust BBL_TYPE and edges */
                BBLRemoveEdgeTyped(fault_instr->i_Bblock,
                    target_instr->i_Bblock, ET_UNCOND);
                fault_instr->i_Bblock->b_type = BT_NORMAL;
                if (fault_instr->i_Bblock->b_next != NULL) {
                    BBLAddEdge(fault_instr->i_Bblock,
                        fault_instr->i_Bblock->b_next, ET_NORMAL);
                }

            } else if (fault_instr->i_Bblock->b_type == BT_RETURN) {
                //all 3 slots should have initially contained zeros

                /* do not fill slot 1, of fill with some junk if need to */

                /* do not fill slot 2, of fill with some junk if need to */

                /* fill slot 3 with 0x0 */

                //write 0x0 into the table[index+2]
                //storeLE(&(SECTION_REF_DATA->s_newdata)[
                //    trap_target_table_offset
                //    + new_index * sizeof(ADDRESS) * num_columns
                //    + sizeof(ADDRESS) * (num_columns - 1)], 0);

                /* adjust BBL_TYPE and edges */
                BBLRemoveOutgoingEdges(fault_instr->i_Bblock);
                fault_instr->i_Bblock->b_type = BT_NORMAL;
                if (fault_instr->i_Bblock->b_next != NULL) {
                    BBLAddEdge(fault_instr->i_Bblock,
                        fault_instr->i_Bblock->b_next, ET_NORMAL);
                }

            } else {
                MALFUNCTION("Block type must be one of the following: "
                    "BT_CALL, BT_UNCOND, BT_RETURN");
            }
        }
    }

    fclose(addr_ind_file);

    if (num_columns == 1) {
        REPORT(RP_SMLOOP, "Table of <&target^&fault> entries filled");
    } else {
        REPORT(RP_SMLOOP, "Table of "
            "<&target^&fault, &ret^&fault, used_to_be> entries filled");
    }

    return;
}



/*
 *  CompleteMyRestore() --
 *
 *  Adds to the my_restore() function the instructons that will retrieve 0, 1
 *  or 2 target addresses from the trap_target_table and place them (if 1 or 2)
 *  on the top of the stack, so that ret instruction will return to the
 *  target of the original instruction. If the original instruction was call
 *  then the second target address is the address of the instruction to which
 *  the original call returned.
 */
static void CompleteMyRestore() {
    if (TRAPS_CONVERT_CALLS || TRAPS_CONVERT_RETURNS) {
        CompleteLongMyRestore();
    } else {
        CompleteShortMyRestore();
    }

    return;
}



/*
 *  CompleteShortMyRestore() --
 *
 *  This one is used when only jumps are converted to traps.
 */
static void CompleteShortMyRestore() {
    Instruction *new_instr;
    Iterator *iterator;
    Block *my_restore_bbl1;

    iterator = ll_iterator(my_restore_bbls_lists[1]);
    FOREACH_ITEM(iterator, my_restore_bbl1, Block *) {

        //BBL1 (BT_RETURN)
        RemoveNOPs(my_restore_bbl1->b_first->i_next, 13);

        //mov <&trap_target_table_offset>(%eax), %eax
        new_instr = InstrCreateMovMemDisp32BaseToReg(0x0, reg_a | reg_32,
            reg_a | reg_32);
        AddPatchSectionAddressIntoInstr(SECTION_REF_DATA,
            trap_target_table_offset, new_instr, new_instr->i_src);
        InsertInstrAfterInstr(new_instr, my_restore_bbl1->b_last);

        //xor %eax, 0x20(%esp)
        new_instr = InstrCreateXorRegWithMemDisp8Base(reg_a | reg_32,
            0x20, reg_sp | reg_32);
        InsertInstrAfterInstr(new_instr, my_restore_bbl1->b_last);

        //popa
        new_instr = InstrCreatePopAll();
        InsertInstrAfterInstr(new_instr, my_restore_bbl1->b_last);

        //ret
        new_instr = InstrCreateReturn();
        InsertInstrAfterInstr(new_instr, my_restore_bbl1->b_last);
        INSTR_SET_FLAG(new_instr, INSTR_CLASS_RETURN);
    }

    REPORT(RP_SMLOOP, "Completed the short my_restore() function(s)");
    return;
}



/*
 *  CompleteLongMyRestore() --
 *
 *  This one is used when jumps, calls and returns are converted to traps.
 */
static void CompleteLongMyRestore() {
    Instruction *mov_instr;
    Instruction *xor_instr;
    Instruction *cmp_instr;
    Instruction *new_instr;
    Patch *upm;
    PatchSectionToInstr *patch;
    LinkedList *upms;
    Iterator *iterator;
    Block *my_restore_bbl1;
    Block *my_restore_bbl3;
    Block *my_restore_bbl4;

    iterator = ll_iterator(my_restore_bbls_lists[1]);
    FOREACH_ITEM(iterator, my_restore_bbl1, Block *) {

        //BBL1 (BT_BRANCH)
        RemoveNOPs(my_restore_bbl1->b_first->i_next, 13);

        //cmp $0x0, <&trap_target_table_offset+0x8>(%eax)
        cmp_instr = new_instr = InstrCreateCmpImm8WithMemDisp32Base(0x0,
            0x12345, reg_a | reg_32);
        AddPatchSectionAddressIntoInstr(SECTION_REF_DATA,
            trap_target_table_offset, new_instr, new_instr->i_src);
        upms = UPMTableGetByDest(new_instr->i_src);
        upm = ll_get(upms, 0);
        ll_free(upms);
        if (upm == NULL) {
            MALFUNCTION("UPM was not properly cloned");
        }
        patch = upm->p_patch;
        patch->p_src_offset += 2 * sizeof(ADDRESS);
        InsertInstrAfterInstr(new_instr, my_restore_bbl1->b_last);

        //jne <&my_restore_bbls[3]>
        new_instr = InstrCreateJne();
        InsertInstrAfterInstr(new_instr, my_restore_bbl1->b_last);
    }

    iterator = ll_iterator(my_restore_bbls_lists[3]);
    FOREACH_ITEM(iterator, my_restore_bbl3, Block *) {

        //BBL3 (BT_BRANCH)
        RemoveNOPs(my_restore_bbl3->b_first->i_next, 22);

        //mov %eax, %edx
        new_instr = InstrCreateMovRegToReg(reg_a | reg_32, reg_d | reg_32);
        InsertInstrAfterInstr(new_instr, my_restore_bbl3->b_last);

        //mov <&trap_target_table_offset>(%eax), %eax
        mov_instr = new_instr = InstrCreateMovMemDisp32BaseToReg(0x0,
            reg_a | reg_32, reg_a | reg_32);
        AddPatchSectionAddressIntoInstr(SECTION_REF_DATA,
            trap_target_table_offset, new_instr, new_instr->i_src);
        InsertInstrAfterInstr(new_instr, my_restore_bbl3->b_last);

        //xor %eax, 0x20(%esp)
        xor_instr = new_instr = InstrCreateXorRegWithMemDisp8Base(
            reg_a | reg_32, 0x20, reg_sp | reg_32);
        InsertInstrAfterInstr(new_instr, my_restore_bbl3->b_last);

        //cmp $0x1, <&trap_target_table_offset+0x8>(%edx)
        new_instr = InstrClone(cmp_instr);
        new_instr->i_src2->a_displacement = 0x1;
        new_instr->i_src->a_base_reg = reg_32 | reg_d;
        InsertInstrAfterInstr(new_instr, my_restore_bbl3->b_last);

        //je <&my_restore_bbls[5]>
        new_instr = InstrCreateJe();
        InsertInstrAfterInstr(new_instr, my_restore_bbl3->b_last);
    }

    iterator = ll_iterator(my_restore_bbls_lists[4]);
    FOREACH_ITEM(iterator, my_restore_bbl4, Block *) {

        //BBL4 (BT_NORMAL)
        RemoveNOPs(my_restore_bbl4->b_first->i_next, 11);

        //mov <&trap_target_table_offset+0x4>(%edx), %eax
        new_instr = InstrClone(mov_instr);
        upms = UPMTableGetByDest(new_instr->i_src);
        upm = ll_get(upms, 0);
        ll_free(upms);
        if (upm == NULL) {
            MALFUNCTION("UPM was not properly cloned");
        }
        patch = upm->p_patch;
        patch->p_src_offset += 1 * sizeof(ADDRESS);
        new_instr->i_src->a_base_reg = reg_32 | reg_d;
        InsertInstrAfterInstr(new_instr, my_restore_bbl4->b_last);

        //xor %eax, 0x20(%esp)
        new_instr = InstrClone(xor_instr);
        new_instr->i_dest->a_displacement += sizeof(ADDRESS);
        InsertInstrAfterInstr(new_instr, my_restore_bbl4->b_last);
    }

    REPORT(RP_SMLOOP, "Completed the long my_restore() function(s)");
    return;
}

