/*
 *  inserts.c
 *  
 *  TO DO: 1. Look at every function TO DO list.
 *         2. Make as many passes of {recalculate addresses; reinsert junk;} as
 *            needed such that:
 *            a. the junk length is optimal, 
 *            b. there are no invalid bytes as per linear disassembly.
 *
 *  This file contains functions related to the junk insertion. The purpose of
 *  this type of obfuscation is to confuse the linear disassembly by
 *  introducing the junk bytes which are the partial instructions after the
 *  non fall-through locations (unconditional jumps, returns, etc.).
 */
#include "obf_global.h"
#include "inserts.h"
#include "obf.h"
#include "linkage.h"
#include "x86-classes.h"
#include "utilities.h"
#include "sysdep.h"



static bool IsValidJunkInsertionLocation(Instruction *last_instr_in_bbl);
static Instruction *GenerateJunk();
static UINT8 PickJunkByte1();
static UINT8 PickJunkByte2();
static int GetOptimalJunkLength(UINT8 *junk_bytes, int num_junk_bytes,
    Instruction *i_after_junk);
static int InstructionsToBytes(UINT8 *buffer_start, int buffer_size,
    Instruction *instr);
static int GetNumBytesConfused(UINT8 *dis_buffer, int num_junk_bytes);
static Instruction *PickJunkBranchNear();
static Instruction *PickJunkBranchShort();



/*
 *  InsertJunk(highest_bbl_id) --
 *
 *  TO DO: 1. input parameter. //grep for "INPUT"
 *         2. tmp hack. //grep for "tmp hack"
 *
 *  This is a driver for the junk insertion. It inserts a basic block containing
 *  a conditional branch, or a partial instruction or both after every selected
 *  non fall-through location. The purpose of partial instruction is to confuse
 *  linear disassembly. The purpose of conditional branch is to introduce
 *  control flow edges that do not exist.
 */
void InsertJunk(int highest_bbl_id) {
    Block *block_iterator;
    Block *b_after_junk;
    Block *bbl_last;
    Instruction *i_before_junk;
    Instruction *i_after_junk;
    int num_bbls = 0;
    int num_ji_loc = 0;
//    int skipped = 0;

    FOREACH_BBL(block_iterator) {
        bbl_last = block_iterator;
    }

//INPUT
    //Think how to set the pass direction from the command line.
    //FOREACH_BBL(block_iterator) {
    //this is from first to last
    //for (block_iterator = bbl_list; block_iterator != NULL;
    //    block_iterator = block_iterator->b_next) {
    //this is from last to first
    for (block_iterator = bbl_last; block_iterator != NULL;
        block_iterator = block_iterator->b_prev) {
#if 0
        printf("b_id = %d", block_iterator->b_id);
        if (BBL_HAS_FLAG(block_iterator, BBL_JUNK_FLAG)) {
            printf(" - junk bbl: skipping\n");
            continue;
        }
        printf("\n");
#endif
        num_bbls++;

        b_after_junk = block_iterator->b_next;
        if (b_after_junk == NULL) {
            //WARNING("Should only see this once. (no b_next)");
            continue;
        }

        i_after_junk = b_after_junk->b_first;
        if (i_after_junk == NULL) {
            //WARNING("Should never see this. (no b_first)");
            continue;
        }

        i_before_junk = block_iterator->b_last;
        if (!IsValidJunkInsertionLocation(i_before_junk)) {
            continue;
        }

//        if (block_iterator->b_id > highest_bbl_id) {
//            skipped++;
//            continue;
//        }

        num_ji_loc++;

        Block *junk_bbl = BBLCreate();
        BBL_SET_FLAG(junk_bbl, BBL_JUNK_FLAG);
        InsertBBLAfterBBL(junk_bbl, block_iterator);

        if (JUNK_IN_JUNK) {
            Instruction *junk = GenerateJunk(i_after_junk);
            if (junk != NULL) {
                //INSTR_SET_FLAG(junk, INSTR_JUNK_FLAG);
                BBLPrependInstr(junk_bbl, junk);
            } else {
                WARNING("No junk has been generated.");
                //have to either ensure that junk is never NULL,
                //or to create a bbl in different place only if needed,
                //so that we never have an empty bbl.
                BBLPrependInstr(junk_bbl, PickJunkBranchNear()); //tmp hack
            }
        }

        if (JUMP_IN_JUNK && ((!JUMP_IN_JUNK_SKIP_UNCOND) ||
            (JUMP_IN_JUNK_SKIP_UNCOND &&
            (i_before_junk->i_type != TYPE_UNCONDITIONAL) ))
            ) {

            Instruction *junk_branch;
            if (JUMP_IN_JUNK_NEAR_0 || JUMP_IN_JUNK_NEAR_1 ||
                JUMP_IN_JUNK_NEAR_2) {
                //WARNING("NEAR!!!");
                junk_branch = PickJunkBranchNear();
            } else if (JUMP_IN_JUNK_SHORT_0 || JUMP_IN_JUNK_SHORT_1) {
                //WARNING("SHORT!!!");
                junk_branch = PickJunkBranchShort();
            } else { // if JUMP_IN_JUNK_NEAR_2_SHORT_1
                if (rand() % 2) {
                    junk_branch = PickJunkBranchNear();
                } else {
                    junk_branch = PickJunkBranchShort();
                }
            }

            INSTR_SET_FLAG(junk_branch, INSTR_JUNK_FLAG);
            BBLPrependInstr(junk_bbl, junk_branch);
        }
    }

    REPORT(RP_DETAIL, "Total blocks = %d, junk insertion candidates = %d",
        num_bbls, num_ji_loc);
//    REPORT(RP_DETAIL, "Skipped %d blocks with id > %d",
//        skipped, highest_bbl_id);

    return;
}



/*
 *  IsValidJunkInsertionLocation(last_instr_in_bbl) --
 *
 *  TO DO: 1. In x86/table.c change "ret" and "iret" TYPE_NONE to
 *            TYPE_UNCONDITIONAL
 *
 *  Returns true if the location after the bbl_last_instr could be used for
 *  junk insertion.
 */
static bool IsValidJunkInsertionLocation(Instruction *last_instr_in_bbl) {
    return (last_instr_in_bbl->i_type == TYPE_UNCONDITIONAL &&
        !isCallClass(last_instr_in_bbl))
        || isHaltClass(last_instr_in_bbl)
        || isReturnClass(last_instr_in_bbl)  //won't need this once TYPE of
                                            //"ret" in x86/table.c is fixed
        //don't need, covered by TYPE_UNCONDITIONAL above
        || INSTR_HAS_FLAG(last_instr_in_bbl, INSTR_SPOOF_JT_FLAG)

        || INSTR_HAS_FLAG(last_instr_in_bbl, INSTR_JBF_CALL_FLAG)
        || INSTR_HAS_FLAG(last_instr_in_bbl, INSTR_CBF_CALL_FLAG)

        || INSTR_HAS_FLAG(last_instr_in_bbl, INSTR_SIGILL_FLAG)
        || INSTR_HAS_FLAG(last_instr_in_bbl, INSTR_SIGFPE_FLAG)
        || INSTR_HAS_FLAG(last_instr_in_bbl, INSTR_SIGSEGV_FLAG);
}



/*
 *  GenerateJunk(i_after_junk) --
 *
 *  TO DO: 1. Extend to more than just 6-byte instruction. (10-11 byte with two
 *            memory addresses or immediate)
 *         2. Find optimal length not after the instruction is chosen, but
 *            for every possible instruction choice GetOptimalJunkLength, and
 *            then use the one (the proper piece of one) that gives the best
 *            confusion.
 *
 *  Generates the junk instruction.
 */
static Instruction *GenerateJunk(Instruction *i_after_junk) {
    Instruction *junk_instr;
    int optimal_junk_length;
    UINT8 buffer[6] = { 0 };

    buffer[0] = PickJunkByte1();
    buffer[1] = PickJunkByte2();

    optimal_junk_length = GetOptimalJunkLength(buffer, 6, i_after_junk);

    if (optimal_junk_length == 0) {
        //WARNING("[GenerateJunk]: returning NULL");
        return NULL;
    }

    junk_instr = CreateCustomInstr(buffer, optimal_junk_length);
    
    return junk_instr;
}



//just temporary
//static UINT8 PickJunkByte11() {
//    UINT8 instr[2];
//    int pick;
//
//    /* mov */
//    instr[0] = 0x8b; /*   to reg32 */
//    instr[1] = 0x89; /* from reg32 */
//
//    pick = rand() % 2;
//
//    return instr[pick];
//}

/*
 *  PickJunkByte1() --
 *
 *  Randomly picks the first byte for a 6-byte instruction used as a basis for
 *  generating junk.
 */
static UINT8 PickJunkByte1() {
    UINT8 instr[20];
    int pick;

    /* add */
    instr[0] = 0x02; /*   to reg8 */
    instr[1] = 0x00; /* from reg8 */
    instr[2] = 0x03; /*   to reg32 */
    instr[3] = 0x01; /* from reg32 */
    /* sub */
    instr[4] = 0x2a; /*   to reg8 */
    instr[5] = 0x28; /* from reg8 */
    instr[6] = 0x2b; /*   to reg32 */
    instr[7] = 0x29; /* from reg32 */
    /* and */
    instr[8] = 0x22; /*   to reg8 */
    instr[9] = 0x20; /* from reg8 */
    instr[10] = 0x23; /*   to reg32 */
    instr[11] = 0x21; /* from reg32 */
    /* or */
    instr[12] = 0x0a; /*   to reg8 */
    instr[13] = 0x08; /* from reg8 */
    instr[14] = 0x0b; /*   to reg32 */
    instr[15] = 0x09; /* from reg32 */
    /* mov */
    instr[16] = 0x8a; /*   to reg8 */
    instr[17] = 0x88; /* from reg8 */
    instr[18] = 0x8b; /*   to reg32 */
    instr[19] = 0x89; /* from reg32 */

    pick = rand() % 20;

    return instr[pick];
}



/*
 *  PickJunkByte2() --
 *
 *  Randomly picks the second byte for a 6-byte instruction used as a basis for
 *  generating junk.
 */
static UINT8 PickJunkByte2() {
    UINT8 instr[4];
    int pick;

    instr[0] = 0x05; /* ax, mem32 */
    instr[1] = 0x0d; /* cx, mem32 */
    instr[2] = 0x15; /* dx, mem32 */
    instr[3] = 0x1d; /* bx, mem32 */

    pick = rand() % 4;

    return instr[pick];
}



/*
 *  GetOptimalJunkLength(junk_bytes, num_junk_bytes, i_after_junk) --
 *
 *  TO DO: 1. Keep track of the number of instructions confused, too.
 *            (Currently only bytes).
 *
 *  Returns the number of bytes taken from junk_bytes that will cause the
 *  maximum confusion in the stream starting at i_after_junk.
 */
static int GetOptimalJunkLength(UINT8 *junk_bytes, int num_junk_bytes,
    Instruction *i_after_junk) {

    int i;
    int num_instr;
    UINT8 junk_buffer[JUNK_DISASSEMBLY_SIZE];
    int junk_size;
    int num_bytes_confused;
    int max_bytes_confused = 0;
    int opt_junk_length = 0;
    static int total_bytes_confused;
    static int total_max_bytes_confused;

    for (i = 0; i < JUNK_DISASSEMBLY_SIZE; i++) {
        junk_buffer[i] = 0x90;
    }

    num_instr = InstructionsToBytes(&junk_buffer[num_junk_bytes],
        JUNK_DISASSEMBLY_SIZE - num_junk_bytes, i_after_junk);
#if 0
    printf("bbl_id: %d, junk_buffer (%d instr):",
        i_after_junk->i_Bblock->b_id, num_instr);
    for (i = 0; i < JUNK_DISASSEMBLY_SIZE; i++) {
        printf(" 0x%x", junk_buffer[i]);
    }
    printf("\n");
#endif
    //now buffers are initialized.

    //for every junk length from 1 to (length-1)
    for (junk_size = 1; junk_size < num_junk_bytes; junk_size++) {
    //for (junk_size = 1; junk_size <= num_junk_bytes; junk_size++) {
        //write the junk right before the start of i_after_junk
        for (i = junk_size; i > 0; i--) {
            junk_buffer[num_junk_bytes - i] = junk_bytes[junk_size - i];
        }
#if 0
        printf("bbl_id: %d, junk_buffer (%d junk bytes):",
            i_after_junk->i_Bblock->b_id, junk_size);
        for (i = 0; i < 8; i++) {
            printf(" 0x%x", junk_buffer[i]);
        }
        printf("\n");
#endif
        num_bytes_confused = GetNumBytesConfused(junk_buffer, num_junk_bytes);
        if (num_bytes_confused > max_bytes_confused) {
            max_bytes_confused = num_bytes_confused;
            opt_junk_length = junk_size;
        }
    }

    total_bytes_confused += max_bytes_confused;
    if (max_bytes_confused > total_max_bytes_confused) {
        total_max_bytes_confused = max_bytes_confused;
    }
    //REPORT
    //printf("Total bytes confused: %d, max confused so far: %d\n",
    //    total_bytes_confused, total_max_bytes_confused);
    
    return opt_junk_length;
}



/*
 *  InstructionsToBytes(buffer_start, buffer_size, instr) --
 *
 *  TO DO: 1. Look inside (INSTR_DATA_BYTE)
 *
 *  Fills up the buffer that starts at buffer_start with the bytes from the
 *  instruction stream that starts at instr.
 */
static int InstructionsToBytes(UINT8 *buffer_start, int buffer_size,
    Instruction *instr) {

    int num_instr = 0;
    int offset = 0;  //this many bytes were written into buffer
    int length;

    while (instr != NULL && offset < buffer_size) {
        UINT8 curr_instr_bytes[1000];
//        int i;

        //might want to treat this specially (never know how long it is)
        //if (INSTR_HAS_FLAG(instr, INSTR_DATA_BYTE))

        length = InstrAssemble(curr_instr_bytes, instr);
#if 0        
        printf("length = %d: [", length);
        for (i = 0; i < length; i++) {
            printf(" 0x%x", curr_instr_bytes[i]);
        }
        printf("]\n");
#endif
        if (offset + length > buffer_size) {
//            printf("!!!Offset + length + 6 = %d\n", offset + length + 6);
            break;
        }

        memcpy(buffer_start + offset, curr_instr_bytes, length);

        offset += length;
        num_instr++;
        instr = instr->i_next;
    }

//    printf("Last index into instr bytes: %d\n", offset - 1 + 6);
    
    return num_instr;
}



/*
 *  GetNumBytesConfused(dis_buffer, num_junk_bytes) --
 *
 *  TO DO: 1. Instead of making JUNK_DISASSEMBLY_SIZE ureachably big, ensure
 *            that [junk, act]_dis_index do not overflow the dis_buffer.
 *         2. Keep track of the number of instructions confused.
 *
 *  Returns the number of bytes incorrectly disassembled.
 */
static int GetNumBytesConfused(UINT8 *dis_buffer, int num_junk_bytes) {
    Instruction *instr = InstrCreate();
    int junk_dis_index = 0;
    int act_dis_index = num_junk_bytes;

    while (junk_dis_index < JUNK_DISASSEMBLY_SIZE &&
        dis_buffer[junk_dis_index] == 0x90) {
        junk_dis_index++;
    }

    //REPORT
    //printf("Junk disassembly index: %d, actual disassembly_index: %d\n",
    //    junk_dis_index, act_dis_index);

    while (junk_dis_index != act_dis_index) {
        int length;
        //REPORT
        //printf("entering while, junk_dis_index = %d, act_dis_index = %d\n",
        //    junk_dis_index, act_dis_index);
        if (junk_dis_index < act_dis_index) {
            length = DisassembleInstr(0, &dis_buffer[junk_dis_index], instr);
            if (length == -1) {
                //This is where we would want to return 0 in the case, when we
                //want the stream including junk to be fully valid (no bytes
                //considered to be invalid before the resynchronization point)
                //!or make it an INPUT parameter.
                //return 0;
                //REPORT
                //printf("Invalid instruction start: 0x%x",
                //    dis_buffer[junk_dis_index]);
                junk_dis_index++;
            } else {
                junk_dis_index += length;
            }
        } else if (junk_dis_index > act_dis_index) {
            length = DisassembleInstr(0, &dis_buffer[act_dis_index], instr);
            if (length == -1) {
                //WARNING or MALFUNCTION
            //printf("!!THIS CAN'T HAPPEN: Invalid instruction start: 0x%x\n",
                //    dis_buffer[act_dis_index]);
                act_dis_index++;
            } else {
                act_dis_index += length;
            }
        }
    }

    //REPORT
    //printf ("Resynchronized at index: a: %d, j: %d\n",
    //    act_dis_index, junk_dis_index);
    //printf ("Number of bytes confused: %d\n", act_dis_index - num_junk_bytes);

    return act_dis_index - num_junk_bytes;
}



/*
 *  PickJunkBranchNear() --
 *
 *  Returns one of the 12 randomly chosen near branches. Branches are chosen
 *  at about the same probability as they usually appear in specint2000.
 */
static Instruction *PickJunkBranchNear() {
    Instruction *junk_branch;
    int pick = rand() % 26;

    switch (pick) {
        case 0:
            junk_branch = InstrCreateNearJb();
            break;
        case 1:
            junk_branch = InstrCreateNearJae();
            break;
        case 2:
            junk_branch = InstrCreateNearJbe();
            break;
        case 3:
            junk_branch = InstrCreateNearJa();
            break;
        case 4:
            junk_branch = InstrCreateNearJs();
            break;
        case 5:
            junk_branch = InstrCreateNearJns();
            break;
        case 6:
            junk_branch = InstrCreateNearJl();
            break;
        case 7:
            junk_branch = InstrCreateNearJge();
            break;
        case 8:
            junk_branch = InstrCreateNearJle();
            break;
        case 9:
            junk_branch = InstrCreateNearJg();
            break;
        case 10:
        case 11:
        case 12:
        case 13:
        case 14:
        case 15:
        case 16:
        case 17:
        case 18:
        case 19:
            junk_branch = InstrCreateJe();
            break;
        case 20:
        case 21:
            junk_branch = InstrCreateJne();
            break;
        case 22:
        case 23:
        case 24:
        case 25:
            if (INVERT_CONDITIONAL_BRANCHES) {
                junk_branch = InstrCreateJe();
            } else {
                junk_branch = InstrCreateJne();
            }
            break;
    }

    if (junk_branch->i_src) {
        if (JUMP_IN_JUNK_NEAR_0) {
            junk_branch->i_src->a_displacement = 0;
        } else if (JUMP_IN_JUNK_NEAR_1) {
            junk_branch->i_src->a_displacement = (rand() % 256) - 128;
        } else {
            junk_branch->i_src->a_displacement = (rand() % 65536) - 32768;
        }
    }

    return junk_branch;
}



/*
 *  PickJunkBranchShort() --
 *
 *  Returns one of the 12 randomly chosen short branches. Branches are chosen
 *  at about the same probability as they usually appear in specint2000.
 */
static Instruction *PickJunkBranchShort() {
    Instruction *junk_branch;
    int pick = rand() % 26;

    switch (pick) {
        case 0:
            junk_branch = InstrCreateShortJb();
            break;
        case 1:
            junk_branch = InstrCreateShortJae();
            break;
        case 2:
            junk_branch = InstrCreateShortJbe();
            break;
        case 3:
            junk_branch = InstrCreateShortJa();
            break;
        case 4:
            junk_branch = InstrCreateShortJs();
            break;
        case 5:
            junk_branch = InstrCreateShortJns();
            break;
        case 6:
            junk_branch = InstrCreateShortJl();
            break;
        case 7:
            junk_branch = InstrCreateShortJge();
            break;
        case 8:
            junk_branch = InstrCreateShortJle();
            break;
        case 9:
            junk_branch = InstrCreateShortJg();
            break;
        case 10:
        case 11:
        case 12:
        case 13:
        case 14:
        case 15:
        case 16:
        case 17:
        case 18:
        case 19:
            junk_branch = InstrCreateShortJe();
            break;
        case 20:
        case 21:
            junk_branch = InstrCreateShortJne();
            break;
        case 22:
        case 23:
        case 24:
        case 25:
            if (INVERT_CONDITIONAL_BRANCHES) {
                junk_branch = InstrCreateShortJe();
            } else {
                junk_branch = InstrCreateShortJne();
            }
            break;
    }

    if (junk_branch->i_src) {
        if (JUMP_IN_JUNK_SHORT_0) {
            junk_branch->i_src->a_displacement = 0;
        } else {
            junk_branch->i_src->a_displacement = (rand() % 256) - 128; 
        }
    }

    return junk_branch;
}

