/*
 *  interleave.c
 *
 *  This file contains functions, necessary to interleave hot and cold blocks
 *  in the binary.
 */
#include "obf_global.h"
#include "interleave.h"
#include "linkage.h"
#include "opt.layout.h"



static LinkedList *cold_obf_blocks;
static LinkedList *hot_blocks;



static void CollectBlocks();
static void InterleaveBlocks();
static void BBLBeforeBBL(Block *first, Block *second);



/*
 *  InterleaveHotCold() --
 *
 *  Entry point into this file's functionality. Interleaves hot and cold blocks
 *  in four steps using four helper functions.
 */
void InterleaveHotCold() {

    CollectBlocks();
    InterleaveBlocks();
    InstrRelinkBasedOnBBL();
    InstrUpdateEdges();

    return;
}



/*
 *  CollectBlocks() --
 *
 *  Collects all hot blocks into the hot_blocks and all cold blocks marked with
 *  BBL_TRAP_FLAG into the cold_obf_blocks.
 */
static void CollectBlocks() {
    Block *biterator;
    Instruction *iiterator;
    cold_obf_blocks = ll_new(LIST_UNSORTED);
    hot_blocks = ll_new(LIST_UNSORTED);
    
    int num_blocks = 0;
    int num_hot_blocks = 0;
    int num_hot_instr = 0;
    int num_cold_blocks = 0;
    int num_marked_blocks = 0;
    int num_trap_blocks = 0;
    int num_trap_instr = 0; //number of instructions in bbls ending with traps
   
    FOREACH_BBL(biterator) {
        num_blocks++;

        if (BBL_HAS_FLAG(biterator, BBL_HOT_FLAG)) {
            num_hot_blocks++;
            FOREACH_INSTR_IN_BBL(biterator, iiterator) {
                num_hot_instr++;
            }
            ll_add(hot_blocks, biterator);
        } else {
            num_cold_blocks++;
        }

        if (BBL_HAS_FLAG(biterator, BBL_TRAP_FLAG)) {
            num_marked_blocks++;

            //if (INSTR_HAS_FLAG(biterator->b_last, INSTR_SIGILL_FLAG) ||
            //    INSTR_HAS_FLAG(biterator->b_last, INSTR_SIGFPE_FLAG) ||
            //    INSTR_HAS_FLAG(biterator->b_last, INSTR_SIGSEGV_FLAG)) {
                num_trap_blocks++;
                FOREACH_INSTR_IN_BBL(biterator, iiterator) {
                    num_trap_instr++;
                }
                ll_add(cold_obf_blocks, biterator);
            //}
        }
    }

    REPORT(RP_DETAIL, "Total blocks: %6d", num_blocks);
    REPORT(RP_DETAIL, "        hot:  %6d (%1.2f instr per block)",
        num_hot_blocks, (double)num_hot_instr / num_hot_blocks);
    REPORT(RP_DETAIL, "       cold:  %6d", num_cold_blocks);
    REPORT(RP_DETAIL, "      traps:  %6d (%1.2f instr per block)",
        num_trap_blocks, (double)num_trap_instr / num_trap_blocks);

    return;
}



/*
 *  InterleaveBlocks() --
 *
 *  Interleaves blocks found in cold_obf_blocks between blocks found in
 *  hot_blocks.
 */
static void InterleaveBlocks() {
    Block *hot;
    Block *cold;

    while (!ll_isEmpty(cold_obf_blocks) && !ll_isEmpty(hot_blocks)) {
        cold = ll_removeFirst(cold_obf_blocks);
        hot = ll_removeFirst(hot_blocks);
        BBLBeforeBBL(cold, hot);
    }

    return;
}



/*
 *  BBLBeforeBBL(first, second) --
 *
 *  Inserts first block before second block.
 */
void BBLBeforeBBL(Block *first, Block *second) {

    //unlink first from old location
    if (first->b_prev) {
        first->b_prev->b_next = first->b_next;
    }
    if (first->b_next) {
        first->b_next->b_prev = first->b_prev;
    }

    //link first into new location (before second)
    if (second->b_prev) {
        second->b_prev->b_next = first;
    } else {
        if (bbl_list != second) {
            WARNING("not the first bbl does not have b_prev");
        }
        bbl_list = first;
    }
    first->b_prev = second->b_prev;

    second->b_prev = first;
    first->b_next = second;

    return;
}

