/*
 *  jt.c
 *
 *  This file contains functions related to the jump table spoofing. The
 *  purpose of this type of obfuscation is to replace an unconditional
 *  control transfer instruction (jump) with the sequence of instructions
 *  that transfer control through the jump table. When the jump through the
 *  jump table is present in the instruction stream, the index into the table
 *  is usually computed in one of the registers and the bounds check, to ensure
 *  that index is not greater than the size of the jump table, preceeds that
 *  jump. The jump table itself resides in the data section. When replacing a
 *  jump with the jump through the jump table, there are two possible ways to
 *  proceed:
 *      1. When computing the index into the jump table ensure that it is
 *         always greater than the size of the table and so that bounds check
 *         always fails. The default case to which execution branches on bounds
 *         miss will be the original target of the former jump.
 *      2. When computing the index into the jump table ensure that it always
 *         equals to some predefined integer (or several integers) less than
 *         the size of the jump table. At that location(s) in the jump table
 *         we should place the address of the original target of the former
 *         jump.
 *  The unused locations of the jump table could be populated with some junk
 *  addresses that we want the disassembler to see, but that are not the actual
 *  instruction addresses (the middle of instruction, etc.). The bounds miss
 *  target in case 2 could also point into the middle of an instruction.
 *  In addition, instead of branching to the original target after the bounds
 *  miss (case 1) or after the jump through the jump table (case 2) execution
 *  could be rerouted to the direct jump, that jumps to the original target.
 *  This jump could subsequently be converted to either branch function call
 *  (see bf.c for details) or trap (see signals.c for details).
 *  Currently case 1 is implemented.
 */
#include "obf_global.h"
#include "jt.h"
#include "obf_macros.h"
#include "utilities.h"
#include "flowgraph.h"
#include "linkage.h"
#include "sysdep.h"
#include "x86-registers.h"
#include "upm.h"



static void JumpToJumpTable(Instruction *jump);
static Block *InsertRestoreBBL(Instruction *from_jump);

static int num_restore_bbls;
static int num_bypass_bbls;
static int num_bf_obf;
static int num_trap_obf;



/*
 *  SpoofJumpTables() --
 *
 *  This is the driver for the jump table spoofing. It converts the previously
 *  marked jumps into the jumps through the jump tables.
 */
void SpoofJumpTables() {
    Block *biterator;
    int num_spoofs = 0;

    FOREACH_BBL(biterator) {
        if (BBL_HAS_FLAG(biterator, BBL_JT_FLAG)) {
            JumpToJumpTable(biterator->b_last);
            num_spoofs++;
        }
    }

    REPORT(RP_SMLOOP, "Number of restore blocks inserted: %6d",
        num_restore_bbls);
    REPORT(RP_SMLOOP, "Number of bypass blocks inserted:  %6d",
        num_bypass_bbls);
    REPORT(RP_DETAIL, "Total converted:   %6d", num_spoofs);

    if (JT_TARGET_BF || JT_TARGET_TRAP) {
        REPORT(RP_DETAIL, "Total marked for further obfuscation:   %6d",
            num_bf_obf + num_trap_obf);
        REPORT(RP_SMLOOP, "    for bf calls: %6d", num_bf_obf);
        REPORT(RP_SMLOOP, "    for traps:    %6d", num_trap_obf);
    }

    return;
}



/*
 *  JumpToJumpTable(jump) --
 *
 *  This function takes a direct unconditional jump and converts it to the
 *  following:
 *    - code to save the state and registers used in the next step
 *    - sequence of instructions to compute an index (always >= 512) in %eax
 *    - instruction that checks the index against the actual size of the jump
 *      table (TO DO: size will be determined by a different function).
 *    - instruction that jumps through the jump table (start address is used
 *      as a displacement, no base register, %eax - index, scale = 4 which is
 *      the size of an address in bytes).
 *  It also inserts an empty jump table into the data section and a restore
 *  block before the target to restore the state and registers saved.
 */
static void JumpToJumpTable(Instruction *jump) {
    Block *from_bbl;
    Block *to_bbl;
    Block *restore_bbl;
    Block *jt_bbl;
    Block *bf_or_trap_bbl;
    ADDRESS jt_offset;
    int jt_size = 10;
    Instruction *new_instr;
    Instruction *cond_jump;

    from_bbl = jump->i_Bblock;
    to_bbl = INSTR_TARGET_BBL(jump);
    restore_bbl = InsertRestoreBBL(jump);

    /* insert state saving instructons before the jump */

    //push %eax
    new_instr = InstrCreatePush(reg_a | reg_32);
    InsertInstrBeforeInstr(new_instr, jump);

    //push %ebx
    new_instr = InstrCreatePush(reg_b | reg_32);
    InsertInstrBeforeInstr(new_instr, jump);

    //pushf
    new_instr = InstrCreateSave();
    InsertInstrBeforeInstr(new_instr, jump);

    /* insert state restore instructions into the restore block */

    //pop %eax
    new_instr = InstrCreatePop(reg_a | reg_32);
    BBLPrependInstr(restore_bbl, new_instr);

    //pop %ebx
    new_instr = InstrCreatePop(reg_b | reg_32);
    BBLPrependInstr(restore_bbl, new_instr);

    //popf
    new_instr = InstrCreateRestore();
    BBLPrependInstr(restore_bbl, new_instr);

    /* calculate the index in the %eax (the so-called opaque expression) */

    //mov %edx, %ebx    //%ebx = anything
    //new_instr = InstrCreateMovRegToReg(reg_d | reg_32, reg_b | reg_32);
    //InsertInstrBeforeInstr(new_instr, jump);

    //imul %ebx, %ebx   //%ebx >= 0
    new_instr = InstrCreateIMULRegReg(reg_b | reg_32, reg_b | reg_32);
    InsertInstrBeforeInstr(new_instr, jump);

    //add $0x1, %ebx    //%ebx >= 1
    new_instr = InstrCreateAddImm8ToReg(1, reg_b | reg_32);
    InsertInstrBeforeInstr(new_instr, jump);

    //shl $0x8, %ebx    //%ebx >= 512
    new_instr = InstrCreate32BitShlImm(reg_b | reg_32, 8);
    InsertInstrBeforeInstr(new_instr, jump);

    //mov %ebx, %eax    //%eax >= 512
    new_instr = InstrCreateMovRegToReg(reg_b | reg_32, reg_a | reg_32);
    InsertInstrBeforeInstr(new_instr, jump);

    //cmp $<jt_size>, %eax  //if jt_size <= 512, then see next
    new_instr = InstrCreateCmpRegToValue(reg_a | reg_32, jt_size);
    InsertInstrBeforeInstr(new_instr, jump);

    //jae <&bounds_miss>    //    always jump
    cond_jump = InstrCreateNearJae();
    InsertInstrAfterInstr(cond_jump, jump);
    InstrDisconnect(jump);
    //InstrFree(jump);

    /* insert new block with a jump through the spoofed jump table */

    jt_bbl = BBLCreate();
    jt_bbl->b_type = BT_SWITCH;
    InsertBBLAfterBBL(jt_bbl, cond_jump->i_Bblock);

    //jmp *<jt_offset>(, %eax, 4)
    new_instr = InstrCreateJmpNearAbsDisp32Index(0, reg_32 | reg_a , 4);
    BBLPrependInstr(jt_bbl, new_instr);

    jt_offset = SECTION_REF_DATA->s_newdatasize;
    IncreaseSectionSize(SECTION_REF_DATA, jt_size * sizeof(ADDRESS));

    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, jt_offset, new_instr,
        new_instr->i_src);

    //TO DO: store the addresses to be filled
    //...foo(jt_offset, jt_size)...

    /* adjust outgoing edges and type of from_bbl */
    BBLRemoveEdgeTyped(from_bbl, restore_bbl, ET_UNCOND);
    BBLAddEdge(from_bbl, restore_bbl, ET_TRUE);
    BBLAddEdge(from_bbl, cond_jump->i_next->i_Bblock, ET_FALSE);
    from_bbl->b_type = BT_BRANCH;

    /*
     * Jump to unconditional jump on bounds miss. This unconditional jump will
     * then get converted to either branch function call or trap.
     */
    if (JT_TARGET_BF || (JT_TARGET_TRAP && !FUNCTION_HAS_FLAG(from_bbl->b_fun,
        FUNCTION_DO_NOT_OBFUSCATE_SIGNAL))) {

        bf_or_trap_bbl = BBLCreate();
        bf_or_trap_bbl->b_type = BT_UNCOND;
        BBLAddEdge(bf_or_trap_bbl, restore_bbl, ET_UNCOND);
        InsertBBLAfterBBL(bf_or_trap_bbl, jt_bbl);
        new_instr = InstrCreateJMP();
        BBLPrependInstr(bf_or_trap_bbl, new_instr);
        BBLRemoveEdgeTyped(from_bbl, restore_bbl, ET_TRUE);
        BBLAddEdge(from_bbl, bf_or_trap_bbl, ET_TRUE);
        /* mark for obfuscation */
        if (JT_TARGET_BF) {
            BBL_SET_FLAG(bf_or_trap_bbl, BBL_BF_FLAG);
            num_bf_obf++;
        } else if (JT_TARGET_TRAP) {
            BBL_SET_FLAG(bf_or_trap_bbl, BBL_TRAP_FLAG);
            num_trap_obf++;
        } 
    }
}



/*
 *  InsertRestoreBBL(from_jump) --
 *
 *  Inserts a restore block before the block targeted by the from_jump
 *  instruction. If the target block had a fall-through predecessor, the
 *  control flow from it to the target is rerouted using an unconditional
 *  jump instruction.
 */
static Block *InsertRestoreBBL(Instruction *from_jump) {
    Block *from_bbl;
    Block *to_bbl;
    Block *fallthru_bbl;
    Block *bypass_bbl;
    Block *restore_bbl;
    Instruction *bypass_jump;
    EdgeNode *enode;
    EDGE_WEIGHT_TYPE fallthru_edge_weight;

    from_bbl = from_jump->i_Bblock;
    to_bbl = INSTR_TARGET_BBL(from_jump);

    fallthru_bbl = NULL;
    /* See if any of to_bbl's predecessors is a fall-through block. */
    FOREACH_PRED_IN_BBL(to_bbl, enode) {
        Block *pred_bbl = NODE_FROM(enode);
        fallthru_edge_weight = NODE_WEIGHT(enode);
        if (pred_bbl->b_type == BT_NORMAL ||
            (pred_bbl->b_type == BT_BRANCH
            && to_bbl == BBLGetFalseEdge(pred_bbl)->e_to) ||
            (pred_bbl->b_type == BT_CALL
            && to_bbl == BBLGetFunlinkEdge(pred_bbl)->e_to)) {
            fallthru_bbl = pred_bbl;
            break;
        }
    }

    /* Create and insert restore bbl such that it falls through to b_to. */
    restore_bbl = BBLCreate();
    restore_bbl->b_weight = from_bbl->b_weight;
    restore_bbl->b_type = BT_NORMAL;
    InsertBBLBeforeBBL(restore_bbl, to_bbl);
    num_restore_bbls++;

    /* set up control flow edges */
    BBLRemoveEdges(from_bbl);
    BBLAddEdge(from_bbl, restore_bbl, ET_UNCOND);
    BBLAddEdge(restore_bbl, to_bbl, ET_NORMAL);

    /*
     * If control flow fell through from fallthru_bbl into to_bbl it will now
     * jump to it to bypass the newly inserted restore_bbl. To do so bypass_bbl
     * is inserted between fallthru_bbl and restore_bbl which contains a jump
     * to the to_bbl.
     */
    if (fallthru_bbl != NULL) {
        bypass_bbl = BBLCreate();
        bypass_bbl->b_weight = fallthru_edge_weight;
        bypass_bbl->b_type = BT_UNCOND;
        InsertBBLAfterBBL(bypass_bbl, fallthru_bbl);
        num_bypass_bbls++;

        bypass_jump = InstrCreateJMP();
        BBLPrependInstr(bypass_bbl, bypass_jump);
        BBLAddEdge(bypass_bbl, to_bbl, ET_UNCOND);

        /* adjust the fallthru edge from fallthru_bbl to now go to bypass_bbl */
        if (fallthru_bbl->b_type == BT_NORMAL) {
            BBLRemoveEdgeTyped(fallthru_bbl, to_bbl, ET_NORMAL);
            BBLAddEdge(fallthru_bbl, bypass_bbl, ET_NORMAL);
        } else if (fallthru_bbl->b_type == BT_BRANCH) {
            BBLRemoveEdgeTyped(fallthru_bbl, to_bbl, ET_FALSE);
            BBLAddEdge(fallthru_bbl, bypass_bbl, ET_FALSE);
        } else if (fallthru_bbl->b_type == BT_CALL) {
            BBLRemoveEdgeTyped(fallthru_bbl, to_bbl, ET_FUNLINK);
            BBLAddEdge(fallthru_bbl, bypass_bbl, ET_FUNLINK);
        } else {
            WARNING("Fall-through edge of unexpected type.");
        }
    }

    return restore_bbl;
}

