/*
 *  obf.c
 *
 *  Drives the obfuscation processes. Contains definitions of functions called
 *  from main.c. Also contains definitions of some functions used in more than
 *  one file in the obfuscator.
 */
#define OBF_GLOBAL      // causes obfuscator globals to be defined by this file

#include "obf_global.h"
#include "obf.h"
#include "obf_macros.h"
#include "utilities.h"
#include "extend.h"
#include "jt.h"
#include "bf.h"
#include "signals.h"
#include "interleave.h"
#include "inserts.h"
#include "linkage.h"
#include "flowgraph.h"
#include "sysdep.h"
#include "relocate.h"
#include "liveness.h"
#include "x86-liveness.h"



static void IndirectControlFlow(int highest_bbl_id);
static void BBLAssignObfuscation();
static void BBLAssignJTObf();
static void BBLAssignBFObf();
static void MarkNoTrapObfuscation();
static void markNoTrapObf(Function *signal_function);
static void BBLAssignTrapObf();

//!!!both have to be input parameters
static int percent_jt;
static int percent_bf;




/*
 *  ObfuscateModify() --
 *
 *  Calls various helper functions to obfuscate the input binary.
 */
void ObfuscateModify() {
    Block *biterator;
    int highest_bbl_id = 0;

    BBLOutputStats();

    if (EXTEND_CANDIDATES) { // -q (equalize), -f (flip), -a (cut off starts)
        REPORT(RP_MAJOR, "Extending the number of candidate locations");
        ExtendCandidates();
    }

//    if (INDIRECT_CONTROL_FLOW) { // -j (jt)  -b (bf)  -s (traps)
//        REPORT(RP_MAJOR, "Indirecting control flow");
//        IndirectControlFlow(); 
//    }

    //tmp hack, 02/24/06
    //BBLAssignObfuscation();

    FOREACH_BBL(biterator) {
        if (biterator->b_id > highest_bbl_id) {
            highest_bbl_id = biterator->b_id;
        }
    }

    printf("After flipping branches highest bbl_id = %d\n", highest_bbl_id);

    if (INTERLEAVE_HOT_COLD) { // -m (mix)
        if (INSTALL_TRAPS) {
            REPORT(RP_SMLOOP, "Marking functions to be excluded from"
                " installing traps");
            MarkNoTrapObfuscation();
            REPORT(RP_SMLOOP, "Marking blocks for installing traps");
            BBLAssignTrapObf();
        } else {
            INSTALL_TRAPS = true;
            TRAPS_CONVERT_CALLS = true;
            TRAPS_CONVERT_RETURNS = true;
            REPORT(RP_SMLOOP, "Marking traps only for the purpose of"
                " interleaving");
            REPORT(RP_SMLOOP, "Marking functions to be excluded from"
                " installing traps");
            MarkNoTrapObfuscation();
            REPORT(RP_SMLOOP, "Marking blocks for installing traps");
            BBLAssignTrapObf();
            INSTALL_TRAPS = false;
            TRAPS_CONVERT_CALLS = false;
            TRAPS_CONVERT_RETURNS = false;
        }
        REPORT(RP_MAJOR, "Interleaving hot code with obfuscated cold code");
        InterleaveHotCold();
    }

    // hack, does not work when applied twice //03/09 works now, applied once
    if (INDIRECT_CONTROL_FLOW) { // -j (jt)  -b (bf)  -s (traps)
        REPORT(RP_MAJOR, "Indirecting control flow");
        IndirectControlFlow(highest_bbl_id);
    }

    if (INSERT_JUNK) { // -i
        REPORT(RP_MAJOR, "Inserting junk");
        InsertJunk(highest_bbl_id);
        BBLOutputStats();
    }

    return;
}



/*
 *  IndirectControlFlow() --
 *
 *  Modifies the program's control flow by applying the following types of
 *  obfuscation:
 *    - changing jumps to jumps through the spoofed jump tables
 *      (see jt.c for description),
 *    - changing jumps/calls to calls to the branch functions
 *      (see bf.c for description),
 *    - changing jumps/calls/returns to the sequence of instructions that
 *      cause a trap into the kernel (see signals.c for description)
 */
static void IndirectControlFlow(int highest_bbl_id) {

    percent_jt = 10;
    percent_bf = 50;

    REPORT(RP_MINOR, "Marking blocks for control flow obfuscation");
    BBLAssignObfuscation();

    if (SPOOF_JUMP_TABLES) { // -j
        REPORT(RP_MINOR, "Spoofing jump tables");
        SpoofJumpTables();
        BBLOutputStats();
        CheckConsistency();
    }

    if (INSERT_BRANCH_FUNCTIONS) { // -b
        if (!BF_CONVERT_CALLS) {
            REPORT(RP_MINOR, "Redirecting jumps through the branch function");
        } else {
            REPORT(RP_MINOR, "Redirecting jumps and calls through the "
                "branch functions");
        }
        InstallBranchFunctionCalls();
        BBLOutputStats();
        CheckConsistency();
    }

    if (INSTALL_TRAPS) { // -s
        REPORT(RP_MINOR, "Installing traps");
        InstallTraps(highest_bbl_id);
        BBLOutputStats();
        CheckConsistency();
    }

    return;
}



/*
 *  BBLAssignObfuscation() --
 *
 *  Iterates through the list of basic blocks and marks the chosen ones for the
 *  particular type of obfuscation.
 */
static void BBLAssignObfuscation() {
    Block *biterator;
    int obf_jt = 0;
    int obf_bf = 0;
    int obf_trap = 0;
    int obf_total;

    if (SPOOF_JUMP_TABLES) {
        REPORT(RP_SMLOOP, "Marking blocks for jump table spoofing");
        BBLAssignJTObf();
    }

    if (INSERT_BRANCH_FUNCTIONS) {
        REPORT(RP_SMLOOP, "Marking blocks for branch function calls");
        BBLAssignBFObf();
    }

    if (INSTALL_TRAPS && !INTERLEAVE_HOT_COLD) {
    //if (INSTALL_TRAPS) {
        REPORT(RP_SMLOOP, "Marking functions to be excluded from"
            " installing traps");
        MarkNoTrapObfuscation();
        REPORT(RP_SMLOOP, "Marking blocks for installing traps");
        BBLAssignTrapObf();
    }

    FOREACH_BBL(biterator) {
        if (BBL_HAS_FLAG(biterator, BBL_JT_FLAG)) {
            obf_jt++;
        } else if (BBL_HAS_FLAG(biterator, BBL_BF_FLAG)) {
            obf_bf++;
        } else if (BBL_HAS_FLAG(biterator, BBL_TRAP_FLAG)) {
            obf_trap++;
        }
    }

    obf_total = obf_jt + obf_bf + obf_trap;

    REPORT(RP_DETAIL, "Marked for jump table spoofing:   %6d", obf_jt);
    REPORT(RP_DETAIL, "Marked for branch function calls: %6d", obf_bf);
    REPORT(RP_DETAIL, "Marked for installing traps:      %6d", obf_trap);
    REPORT(RP_DETAIL, "Total blocks marked:              %6d", obf_total);

    return;
}



/*
 *  BBLAssignJTObf() --
 *
 *  Marks the blocks satisfying certain conditions for the jump table spoofing.
 */
static void BBLAssignJTObf() {
    Block *biterator;
    int hot_skipped = 0;
    int already_marked_else = 0;
    int has_interproc_pred = 0;
    int has_hell_succ = 0;
    int no_target_skipped = 0;
    int exit_fun_skipped = 0;
    int marked = 0;

    FOREACH_BBL(biterator) {
        if (INSTR_IS_DIRECT_JUMP(biterator->b_last)) {

            if (BBL_HAS_FLAG(biterator, BBL_HOT_FLAG)) {
                hot_skipped++;
                continue;
            }

            if (BBL_HAS_FLAG(biterator, BBL_BF_FLAG) ||
                BBL_HAS_FLAG(biterator, BBL_TRAP_FLAG)) {
                already_marked_else++;
                continue;
            }

            if (BBLHasInterproceduralPredecessor(biterator)) {
                has_interproc_pred++;
                continue;
            }

            if (BBLHasHellSuccessor(biterator)) {
                has_hell_succ++;
                continue;
            }

            if (!INSTR_TARGET(biterator->b_last)) {
                no_target_skipped++;
                continue;
            }

            if (strcmp(biterator->b_fun->f_name, "__libc_csu_fini") == 0) {
                exit_fun_skipped++;
                continue;
            }

            /* mark only some of the candidates */
            if (INSERT_BRANCH_FUNCTIONS || INSTALL_TRAPS) {
                if (rand() % 100 < percent_jt) {
                    BBL_SET_FLAG(biterator, BBL_JT_FLAG);
                    marked++;
                }
            /* mark all candidates */
            } else {
                BBL_SET_FLAG(biterator, BBL_JT_FLAG);
                marked++;
            }
        }
    }

    REPORT(RP_SMDETAIL, "Hot skipped:                       %6d", hot_skipped);
    REPORT(RP_SMDETAIL, "Already marked for different obf:  %6d",
        already_marked_else);
    REPORT(RP_SMDETAIL, "Have interprocedural predecessors: %6d",
        has_interproc_pred);
    REPORT(RP_SMDETAIL, "Have HELL successors:              %6d",
        has_hell_succ);
    REPORT(RP_SMDETAIL, "Jumps with no target:              %6d",
        no_target_skipped);
    REPORT(RP_SMDETAIL, "Jumps in exit functions:           %6d",
        exit_fun_skipped);
    REPORT(RP_SMDETAIL, "Marked:                            %6d", marked);

    return;
}



/*
 *  BBLAssignBFObf() --
 *
 *  Marks the blocks satisfying certain conditions for the obfuscation using
 *  branch functions.
 */
static void BBLAssignBFObf() {
    Block *biterator;
    int hot_skipped = 0;
    int already_marked_else = 0;
    //int has_interproc_pred = 0;
    int has_hell_succ = 0;
    int no_target_skipped = 0;
    int pic_call = 0;
    int marked = 0;

    FOREACH_BBL(biterator) {
        if (INSTR_IS_DIRECT_JUMP(biterator->b_last) ||
            (INSTR_IS_DIRECT_CALL(biterator->b_last) && BF_CONVERT_CALLS)) {

            if (BBL_HAS_FLAG(biterator, BBL_HOT_FLAG)) {
                hot_skipped++;
                continue;
            }

            if (BBL_HAS_FLAG(biterator, BBL_JT_FLAG) ||
                BBL_HAS_FLAG(biterator, BBL_TRAP_FLAG)) {
                already_marked_else++;
                continue;
            }

            //if (BBLHasInterproceduralPredecessor(biterator)) {
            //    has_interproc_pred++;
            //    continue;
            //} 

            if (BBLHasHellSuccessor(biterator)) {
                has_hell_succ++;
                continue;
            }

            if (INSTR_IS_DIRECT_JUMP(biterator->b_last) &&
                !INSTR_TARGET(biterator->b_last)) {
                no_target_skipped++;
                continue;
            }

            if (INSTR_IS_DIRECT_CALL(biterator->b_last) &&
                IsPICCall(biterator->b_last)) {
                pic_call++;
                continue;
            }

            /* mark only some of the candidates */
            if (INSTALL_TRAPS) {
                if (rand() % 100 < percent_bf) {
                    BBL_SET_FLAG(biterator, BBL_BF_FLAG);
                    marked++;
                }

            /* mark all candidates */
            } else {
                BBL_SET_FLAG(biterator, BBL_BF_FLAG);
                marked++;
            }
        }
    }

    REPORT(RP_SMDETAIL, "Hot skipped:                       %6d", hot_skipped);
    REPORT(RP_SMDETAIL, "Already marked for different obf:  %6d",
        already_marked_else);
    //REPORT(RP_SMDETAIL, "Have interprocedural predecessors: %6d",
    //    has_interproc_pred);
    REPORT(RP_SMDETAIL, "Have HELL successors:              %6d",
        has_hell_succ);
    REPORT(RP_SMDETAIL, "Jumps with no target:              %6d",
        no_target_skipped);
    REPORT(RP_SMDETAIL, "PIC calls:                         %6d", pic_call);
    REPORT(RP_SMDETAIL, "Marked:                            %6d", marked);

    return;
}



/*
 *  MarkNoTrapObfuscation() --
 *
 *  Locates the signal() or __bsd_signal() function that is used to install a
 *  signal handler and marks it and every function in its call tree with the
 *  FUNCTION_DO_NOT_OBFUSCATE_SIGNAL flag. Instructions in these functions will
 *  be excluded from the signal based obfuscation. Uses markNoTrapObf() to
 *  recursively mark the call tree.
 */
static void MarkNoTrapObfuscation() {
    Function *fiterator;

    /* Mark _start(). */
    FOREACH_FUNCTION(fiterator) {
        if (strcmp(fiterator->f_name, "_start") == 0) {
            FUNCTION_SET_FLAG(fiterator, FUNCTION_DO_NOT_OBFUSCATE_SIGNAL);
            REPORT(RP_SMDETAIL, "%s()", fiterator->f_name);
            break;
        }
    }

    /* Locate signal() or __bsd_signal(). */
    FOREACH_FUNCTION(fiterator) {
        if ((strcmp(fiterator->f_name, "__bsd_signal") == 0)
            || (strcmp(fiterator->f_name, "signal") == 0)) {
            break;
        }
    }

    /* Locate __sigaction() if above are not present. */
    if (fiterator == NULL) {
        FOREACH_FUNCTION(fiterator) {
            if (strcmp(fiterator->f_name, "__sigaction") == 0) {
                break;
            }
        }
    }

    if (fiterator == NULL) {
        WARNING("Neither signal() nor __bsd_signal() nor "
            "__sigaction() found.\n");
    } else {
        /* Mark its call tree. */
        markNoTrapObf(fiterator);
    }
}



/*
 *  markNoTrapObf(signal_function) --
 *
 *  The helper of the MarkNoTrapObfuscation() function. Marks the
 *  signal_function and every function in its call tree with the
 *  FUNCTION_DO_NOT_OBFUSCATE_SIGNAL flag. Instructions in these functions will
 *  be excluded from the signal based obfuscation.
 */
static void markNoTrapObf(Function *signal_function) {
    Block *biterator;
    Edge *fundirect_edge;

    FUNCTION_SET_FLAG(signal_function, FUNCTION_DO_NOT_OBFUSCATE_SIGNAL);
    REPORT(RP_SMDETAIL, "%s()", signal_function->f_name);
    FOREACH_BBL_IN_FUN(signal_function, biterator) {
        if (isCallClass(biterator->b_last)) {
            fundirect_edge = BBLGetFunDirectEdge(biterator);
            markNoTrapObf(fundirect_edge->e_to->b_fun);
        }
    }

    return;
}



/*
 *  BBLAssignTrapObf() --
 *
 *  Marks the blocks satisfying certain conditions for the signal based
 *  obfuscation.
 */
static void BBLAssignTrapObf() {
    Block *biterator;
    int hot_skipped = 0;
    int already_marked_else = 0;
    int already_marked_same = 0;
    int has_interproc_pred = 0;
    int has_hell_succ = 0;
    int no_target_skipped = 0;
    int pic_call = 0;
    int excluded = 0;
    int psw_live_skipped = 0;
    int regs_live_skipped = 0;
    int marked = 0;

//---
    ComputePSWLiveness(1);
    ComputeRegisterLiveness(1);
//---

    FOREACH_BBL(biterator) {
        if (INSTR_IS_DIRECT_JUMP(biterator->b_last) ||
            (INSTR_IS_DIRECT_CALL(biterator->b_last) && TRAPS_CONVERT_CALLS) ||
            (isReturnClass(biterator->b_last) && TRAPS_CONVERT_RETURNS)) {

            if (BBL_HAS_FLAG(biterator, BBL_HOT_FLAG)) {
                hot_skipped++;
                continue;
            }

            if (BBL_HAS_FLAG(biterator, BBL_BF_FLAG) ||
                BBL_HAS_FLAG(biterator, BBL_JT_FLAG)) {
                already_marked_else++;
                continue;
            }

            //if (BBLHasInterproceduralPredecessor(biterator)) {
            //    has_interproc_pred++;
            //    continue;
            //} 

            if (BBLHasHellSuccessor(biterator)) {
                has_hell_succ++;
                continue;
            }

            if (INSTR_IS_DIRECT_JUMP(biterator->b_last) &&
                !INSTR_TARGET(biterator->b_last)) {
                no_target_skipped++;
                continue;
            }

            if (INSTR_IS_DIRECT_CALL(biterator->b_last) &&
                IsPICCall(biterator->b_last)) {
                pic_call++;
                continue;
            }

            if (INSTR_IS_DIRECT_CALL(biterator->b_last)
                && BBLHasInterproceduralPredecessor(biterator)
                && (biterator->b_first == biterator->b_last)) {
                has_interproc_pred++;
                continue;
            }

            if (FUNCTION_HAS_FLAG(biterator->b_fun,
                FUNCTION_DO_NOT_OBFUSCATE_SIGNAL)) {
                excluded++;
                continue;
            }

            //
            if (BBL_HAS_FLAG(biterator, BBL_TRAP_FLAG)) {
                already_marked_same++;
            }

//---
            if (PSWLiveBitvectorAtInstr(biterator->b_last) & 0x003fffff) {
                psw_live_skipped++;
                continue;
            }

            if ((RegsLiveBitvectorAtInstr(biterator->b_last) &
                BITVECTOR_REGISTER_EAX) &&
                (RegsLiveBitvectorAtInstr(biterator->b_last) &
                BITVECTOR_REGISTER_ECX) &&
                (RegsLiveBitvectorAtInstr(biterator->b_last) &
                BITVECTOR_REGISTER_EDX)) {
                regs_live_skipped++;
                //03/31 - this is why no illegal instructions were inserted!!!
                //continue;
            }

//---

            BBL_SET_FLAG(biterator, BBL_TRAP_FLAG);
            marked++;
        }
    }

    REPORT(RP_SMDETAIL, "Hot skipped:                       %6d", hot_skipped);
    REPORT(RP_SMDETAIL, "Already marked for different obf:  %6d",
        already_marked_else);
    //REPORT(RP_SMDETAIL, "Have interprocedural predecessors: %6d",
    //    has_interproc_pred);
    REPORT(RP_SMDETAIL, "Have HELL successors:              %6d",
        has_hell_succ);
    REPORT(RP_SMDETAIL, "Jumps with no target:              %6d",
        no_target_skipped);
    REPORT(RP_SMDETAIL, "PIC calls:                         %6d", pic_call);
    REPORT(RP_SMDETAIL, "1 instr call bbl w/interproc pred: %6d",
        has_interproc_pred);
    REPORT(RP_SMDETAIL, "Excluded:                          %6d", excluded);
    REPORT(RP_SMDETAIL, "PSW live skipped:                  %6d",
        psw_live_skipped);
                                        /*03/31!!!!!*/
    REPORT(RP_SMDETAIL, "eax, ecx, edx live, NOT skipped:   %6d",
        regs_live_skipped);
    REPORT(RP_SMDETAIL, "Marked:                            %6d", marked);
    REPORT(RP_SMDETAIL, " out of which were already marked: %6d",
        already_marked_same);

    return;
}



/*
 *  DumpControlTransferEdges() --
 *
 *  Prints out the number of control transfer instructions by type and the
 *  total number of the control transfer edges.
 */
static void DumpControlTransferEdges() {
    Block *biterator;
    int num_jumps = 0;
    int num_branches = 0;
    int num_returns = 0;
    int num_calls = 0;
    int num_syscalls = 0;
    int num_control_transfer_edges;

    FOREACH_BBL(biterator) {
        //if (INSTR_HAS_FLAG(biterator->b_last, INSTR_CLASS_JUMP)) {
        if ((biterator->b_type == BT_UNCOND) ||
            (biterator->b_type == BT_SWITCH)) {
            num_jumps++;
        //} //else if (INSTR_HAS_FLAG(biterator->b_last, INSTR_CLASS_BRANCH)) {
        } else if (biterator->b_type == BT_BRANCH) {
            num_branches++;
        } else if (INSTR_HAS_FLAG(biterator->b_last, INSTR_CLASS_RETURN)) {
            num_returns++;
        } else if (INSTR_HAS_FLAG(biterator->b_last, INSTR_CLASS_CALL)) {
            num_calls++;
        } else if (INSTR_HAS_FLAG(biterator->b_last, INSTR_CLASS_SYSCALL)) {
            num_syscalls++;
        }
    }

    num_control_transfer_edges = num_jumps + 2 * num_branches +
        //num_returns +
        2 * num_calls + 2 * num_syscalls;

    REPORT(RP_DETAIL, "Edge statistics:");
    REPORT(RP_DETAIL, "%10d jumps\n", num_jumps);
    REPORT(RP_DETAIL, "%10d branches\n", num_branches);
    REPORT(RP_DETAIL, "%10d returns\n", num_returns);
    REPORT(RP_DETAIL, "%10d calls\n", num_calls);
    REPORT(RP_DETAIL, "%10d syscalls\n", num_syscalls);
    REPORT(RP_DETAIL, "%10d total control transfer edges\n",
        num_control_transfer_edges);

    return;
}



/*
 *  CreateCustomInstr(bytes, nBytes) --
 *
 *  Creates and returns an instruction containing nBytes from the array
 *  bytes. This functon is used when bytes contains an invalid instruction, so
 *  that the conventinal method calling the xDisassembleInstr() will not work.
 */
Instruction *CreateCustomInstr(UINT8 *bytes, int nBytes) {
    Instruction *instr = InstrCreate();

    instr->i_length = nBytes;
    instr->i_address = MAGIC_NUMBER;

    INSTR_SET_FLAG(instr, INSTR_LITERAL_BYTES_FLAG);
    instr->i_literal_data = (UINT8 *)zalloc(nBytes);
    memcpy(instr->i_literal_data, bytes, nBytes);

    return instr;
}



/*
 *  GetNonFallThroughBBL() --
 *
 *  Returns a pointer to some randomly chosen non-fallthrough basic block.
 */
Block *GetNonFallThroughBBL() {
    Block *biterator;
    int num_blocks = 0;

    //count all blocks
    FOREACH_BBL(biterator) {
        if ((biterator->b_last->i_type == TYPE_UNCONDITIONAL &&
            !isCallClass(biterator->b_last))
            || isHaltClass(biterator->b_last)
            || isReturnClass(biterator->b_last)) {

            num_blocks++;
        }
    }

    //printf("Total non fall through candidates = %d\n", num_blocks);

    int pick = rand() % (num_blocks - 20) + 10;

    //REPORT(SM_LOOP, " picked num %d, ", pick);

    num_blocks = 0;
    FOREACH_BBL(biterator) {
        if ((biterator->b_last->i_type == TYPE_UNCONDITIONAL &&
            !isCallClass(biterator->b_last))
            || isHaltClass(biterator->b_last)
            || isReturnClass(biterator->b_last)) {

            num_blocks++;
            if (num_blocks == pick) {
                break;
            }
        }
    }

    REPORT(RP_SMDETAIL, "Picked non-fallthrough bbl with bid = %d",
        biterator->b_id);
    //printf("with id %d\n", biterator->b_id);

    return biterator;
}



/*
 *  InsertNOPs(after, num_nops) --
 *
 *  Inserts the num_nops number of NOPs after the Instruction pointed to by
 *  after. Returns the last NOP inserted.
 */
Instruction *InsertNOPs(Instruction *after, int num_nops) {
    Instruction *nop = NULL;
    int i;
    for (i = 0; i < num_nops; i++) {
        nop = InstrCreateNOP();
        InsertInstrAfterInstr(nop, after);
    }
    return nop;
}



/*
 *  RemoveNOPs(from, num_nops) --
 *
 *  Removess the num_nops number of NOPs starting with the one pointed to by
 *  from. Returns the instruction imediately preceeding the series of removed
 *  NOPs.
 */
Instruction *RemoveNOPs(Instruction *from, int num_nops) {
    Instruction *prev = from->i_prev;
    int i;
    for (i = 0; i < num_nops; i++) {
        //if (/* not nop*/) {
        ////error
        //}
        InstrDisconnect(prev->i_next);
    }
    return prev;
}



/*
 *  CallEdgesExist(call) --
 *
 *  Returns true if call instruction has both FUNLINK and FUNDIRECT edges,
 *  false otherwise.
 */
bool CallEdgesExist(Instruction *call) {
    Edge *edge;

    edge = BBLGetFunlinkEdge(call->i_Bblock);
    if (edge == NULL) {
        WARNING("Could not convert call to trap, no funlink edge. %x: %s",
            call->i_address, PrintInstruction(call));
        return false;
    }

    edge = BBLGetFunDirectEdge(call->i_Bblock);
    if (edge == NULL) {
        WARNING("Could not convert call to trap, no fundirect edge. %x: %s",
            call->i_address, PrintInstruction(call)); 
        return false;
    }

    return true;
}

