/*
 *  config.c
 *
 *  Initialization of the PLTO configuration, including command
 *  option handling and configuration file reading.
 *
 *  TO DO: change OPTSTRING to contain the binobf used options only.
 */

#include <unistd.h>

#include "obf_global.h"
#include "liveness.h"
#include "opt.layout.h"

#define CLMAX 1000		// maximum size of a configuration line



/*
 *  Private configuration variables.
 */
static bool listUsage;		// if -h selected
static bool listConfig;		// if -l selected

static char *curfile;		// current file name, if any, for diagnostics



/*
 *  Local function declarations
 */
static void ConfigError(char *msg, char *arg1, char *arg2);
static void InitializeGlobals();
static void Options(int argc, char *argv[]);
static void ParseLine(char *s);
static void PrintHelp(FILE *f);
static void PrintConfig(FILE *f);
static void ReadConfig(char *filename);
static void ScanArgs(int argc, char *argv[]);
static void SetParameter(Parameter *p, char *value);
static void SetFlags(char c, char *s, int type, bool b);
static void UsageError(void);



/*
 *  Option string for getopt().  When modifying this, be sure to also modify
 *	(1) the man page in ../doc/man/plto.1
 *	(2) the helpText[] array at the end of this source file
 *	(3) the processing code in Options()
 */
//added a,b:,f:,i:,j:,m,q,s: //have not done any of the above
#define OPTSTRING "+ab:c:d:D:e:E:f:g:hi:j:klmno:p:Pqs:tv:wx:X:"



/*
 *  Macros for defining parameter table entries
 */
#define P_SPACER()			{ PT_SPACER,  0, "",   0, "", 0 }
#define P_SWITCH(ch, name, dflt)	{ PT_SWITCH, ch, name, L_Bool, 0, dflt }
#define P_OPTIM(ch, name, dflt)		{ PT_OPTIM,  ch, name, L_Bool, 0, dflt }
#define P_DEBUG(ch, name, dflt)		{ PT_DEBUG,  ch, name, L_Bool, 0, dflt }
#define P_INTEGER(ch, name, dflt)	{ PT_INTEGER, ch, name, 0, 0, dflt }
#define P_FLOAT(ch, name, dflt)		{ PT_FLOAT,  ch, name, 0, 0, dflt }
#define P_STRING(ch, name, dflt)	{ PT_STRING, ch, name, 0, dflt, 0 }
#define P_CHOICE(ch, name, list, dflt)	{ PT_CHOICE, ch, name, list, dflt, 0 }



/*
 *  Lists of valid choices for parameters that don't take arbitrary strings
 */
static char* L_Bool[] = { "OFF", "ON", 0 };
static char *L_Sens[] = { "NONE", "INSENSITIVE", "SENSITIVE", 0 };
static char *L_Layo[] = { "OFF", "PETTIS-HANSEN", "GREEDY", "ALTO", "FUNS", 0 };
static char *L_Otyp[] =
    { "CORE", "SHARED", "PLTO_RELOCATABLE", "RELOCATABLE", "EXECUTABLE", 0 };



/*
 *  Table of recognized parameters (in the order printed by plto -l)
 */
static Parameter paramList[] = {

    P_SPACER(),
    P_OPTIM('a', "AlignFunctions",			true),
    P_OPTIM('a', "AlignLoopHeaders",			true),
    P_OPTIM('c', "PropagateConstants",			true),
    P_OPTIM('C', "Compact",				true),
    P_OPTIM('d', "RemoveDeadRegisterStores",		false),
    P_OPTIM('d', "RemoveDeadStackStores",		true),
    P_OPTIM('D', "ComputeDominators",			true),
    P_OPTIM('f', "ForwardValues",			true),
    P_OPTIM('i', "InlineFunctions",			true),
    P_OPTIM('j', "OptimizeJumpTables",			false),
    P_OPTIM('k', "AnalyzeStacks",			false),
    P_OPTIM('p', "PeepholeCoalesceAdds",		true),
    P_OPTIM('p', "PeepholeConditionalMove",		true),
    P_OPTIM('p', "PeepholeStupidJumps",			true),
    P_OPTIM('p', "PeepholeStupidMath",			true),
    P_OPTIM('p', "PeepholeStupidMoves",			true),
    P_OPTIM('p', "PeepholeUnusedPushes",		true),
    P_OPTIM('P', "FixPrologues",			false),
    P_OPTIM('s', "ScheduleInstructions",		false),
    P_OPTIM('u', "RemoveDeadCode",			true),
    P_OPTIM('v', "SpecializeValues",			false),

    P_SPACER(),
    P_CHOICE(0,   "Layout",		L_Layo,		"PETTIS-HANSEN"),
    P_CHOICE(0,   "RegisterLiveness",	L_Sens,		"SENSITIVE"),
    P_CHOICE(0,   "StackLiveness",	L_Sens,		"INSENSITIVE"),
    P_CHOICE(0,   "PSWLiveness",	L_Sens,		"INSENSITIVE"),
    P_CHOICE(0,   "OutputType",		L_Otyp,		"EXECUTABLE"),

    P_SPACER(),
    P_INTEGER(0, "PeepholeIterations",			2),
    P_INTEGER(0, "HotInstrsNeededToInline",		15),
    P_INTEGER(0, "MaxInstructionLength",		16),
    P_INTEGER(0, "MaxSizeFuncInlined",			2500),
    P_FLOAT(0,   "InliningHotThreshold",		0.60),
    P_FLOAT(0,   "InliningMultiplicityIgnore",		0.01),
    P_FLOAT(0,   "JTSingleTestCutoff",			0.75),
    P_FLOAT(0,   "BBLHotThreshold",			0.90),

    P_SPACER(),
    P_SWITCH('k', "KernelMode",				false),
    P_SWITCH(0,   "ReadFakeRelocations",		false),
    P_SWITCH(0,   "LabelBasicBlocks",			true),
    P_SWITCH(0,   "GenerateSymbolTable",		true),

    P_SPACER(),
    P_STRING(0,   "OptOutFile",				"*.opt"),
    P_STRING(0,   "ProfOutFile",			"*.prof"),
    P_STRING(0,   "ValueProfileOutFile",		"*.prof"),
    P_STRING(0,   "CountsOutFile",			"*.counts"),
    P_STRING(0,   "ValueProfileCountsOutFile",		"*.Values"),
    P_STRING(0,   "AddressTranslationFile",		"*.addr"),
    P_STRING(0,   "GDBScriptFile",			"*.gdb"),
    P_STRING(0,   "StaticHistogramOutFile",		"*.histo"),
    P_STRING(0,   "WeightedHistogramOutFile",		"*.histw"),

    P_SPACER(),
    P_INTEGER('v', "Verbosity",				RP_MINOR),  // -vn
    P_DEBUG(DB_FLOW,   "DumpControlFlowGraph",		false),  // -df
    P_DEBUG(DB_STKG,   "DumpStackGain",			false),  // -dg
    P_DEBUG(DB_WRTGDB, "WriteGDBScript",		false),  // -dG
    P_DEBUG(DB_STATHG, "GenerateStaticHistograms",	false),  // -dH
    P_DEBUG(DB_INFAIL, "DumpInlineFailures",		false),  // -di
    P_DEBUG(DB_JUMPS,  "DumpJumpAdjustments",		false),  // -dj
    P_DEBUG(DB_KILLS,  "DumpKilledInstructions",	false),  // -dk
    P_DEBUG(DB_ODBG,   "OutputDebuggingInfo",		false),  // -do
    P_DEBUG(DB_CPFAIL, "DumpConstantPropagationFailures", false),// -dp
    P_DEBUG(DB_DEADREG, "DumpDeadRegisterStores",	false),  // -dr
    P_DEBUG(DB_DEADSTK, "DumpDeadStackStores",		false),  // -ds
    P_DEBUG(DB_WADDR,  "WriteAddressTranslations",	false),  // -dT
    P_DEBUG(DB_UNBUF,  "UnbufferOutput",		false),  // -du
    P_DEBUG(DB_WGTHG,  "GenerateWeightedHistograms",	false),  // -dW

    P_SPACER(),
};



/*
 *  Configure(argc,argv) -- process command options and read config file.
 */
void ConfigureObfuscator(int argc, char *argv[]) {
    int i;
    ProfileFile *last_profile;

    // scan for -c (configfile) argument
    ScanArgs(argc, argv);

    // read configuration file
    if (cfgFilename != NULL) {
	ReadConfig(cfgFilename);		// from command line
    } else if ((cfgFilename = getenv(PLTO_ENVVAR)) != NULL
	    && strlen(cfgFilename) > 0) {
	ReadConfig(cfgFilename);		// from env variable
    } else {
	cfgFilename = "[none]";
    }

    // rescan and process the rest of the options
    // now that the configuration is initialized
    optind = 1;
    Options(argc, argv);

    // save debugging options in an array for quick access
    for (i = 0; i < ARRAYSIZE(paramList); i++) {
	Parameter *p = &paramList[i];
	if (p->pm_type == PT_DEBUG) {
	    SET_DEBUG(p->pm_char, (bool)p->pm_nval);
	}
    }

    // print command help, if asked, or if no input file named
    if (listUsage || (optind == argc && ! listConfig)) {
	PrintHelp(stdout);
    }

    // print the resulting configuration, if asked
    if (listConfig) {
	PrintConfig(stdout);
    }

    // if no binary input, quit
    binFilename = argv[optind++];
    if (binFilename == NULL) {
	exit(0);
    }

    // initialize global variables that depend on configuration
    InitializeGlobals();

    // set output file names
    if (PROFILE_BBLS || PROFILE_EDGES) {
	profFilename = MakeFilename("CountsOutFile", binFilename);
	if (newFilename == NULL) {
	    newFilename = MakeFilename("ProfOutFile", binFilename);
	}
    } else if (PROFILE_VALUES) {
	profFilename = MakeFilename("ValueProfileCountsOutFile", binFilename);
	if (newFilename == NULL) {
	    newFilename = MakeFilename("ValueProfileOutFile", binFilename);
	}
    } else {
	if (newFilename == NULL) {
	    newFilename = MakeFilename("OptOutFile", binFilename);
	}
    }

    // record profiling input file names, if any
    profileFiles = NULL;
    while (optind < argc) {
        ProfileFile *new_profile = zalloc(sizeof(ProfileFile));
        new_profile->pf_name = argv[optind++];
        if (profileFiles) {
            last_profile->pf_next = new_profile;
            new_profile->pf_prev = last_profile;
        } else {
            profileFiles = new_profile;
        }
        last_profile = new_profile;
    }

    // diagnose extra arguments
    if (optind < argc) {
	UsageError();
    }
}



/*
 *  ScanArgs(argc,argv) -- prescan options to process -c option.
 */
static void ScanArgs(int argc, char *argv[]) {
    int c;

    while ((c = getopt(argc, argv, OPTSTRING)) != -1) {
	switch (c) {
	    case 'c':			// -c configfile
		cfgFilename = optarg;
		break;
	    case '?':
		UsageError();
	}
    }
}



/*
 *  Options(argc,argv) -- process command options (except -c, already handled).
 */
static void Options(int argc, char *argv[]) {
    int c;
    char name[CLMAX], value[CLMAX];
    Parameter *p;

    while ((c = getopt(argc, argv, OPTSTRING)) != -1) {
	switch (c) {

            case 'a':
                EXTEND_CANDIDATES = true;
                CUT_OFF_STARTS = true;
                break;

            case 'b':
                INDIRECT_CONTROL_FLOW = true;
                INSERT_BRANCH_FUNCTIONS = true;
                if (strcmp(optarg, "j") == 0) {        // -b j: jumps
                    BF_CONVERT_CALLS = false;
                } else if (strcmp(optarg, "c") == 0) { // -b c: jumps + calls
                    BF_CONVERT_CALLS = true;
                } else {
                    FATAL("Invalid parameter: %s for option: %c", optarg, c);
                }
                break;

	    case 'c':			// -c configfile
		// nothing: handled earlier
		break;

	    case 'd':			// -d xyz: enable debugging x,y,z
		SetFlags(c, optarg, PT_DEBUG, true);
		break;

	    case 'D':			// -D XxXxx=value: set config param
		if (strlen(optarg) >= CLMAX) {
		    FATAL("Value too long: -D %s", optarg);
		}
		if (sscanf(optarg, " %[_a-zA-Z0-9]=%s", name, value) != 2) {
		    FATAL("Malformed definition: -D %s", optarg);
		}
		p = GetParameter(name);
		if (p == NULL) {
		    FATAL("Unrecognized parameter name: -D %s=...", name);
		}
		SetParameter(p, value);
		break;

	    case 'e':			// -e xyz: enable optimizations x,y,z
		SetFlags(c, optarg, PT_OPTIM, true);
		break;

	    case 'E':			// -E XxXxx: enable optimization
	    case 'X':			// -X XxXxx: disable optimization
		p = GetParameter(optarg);
		if (p == NULL) {
		    FATAL("Unrecognized optimization name: %s", optarg);
		}
		SetParameter(p, (c == 'E') ? "ON" : "OFF");
		break;

            case 'f':
                EXTEND_CANDIDATES = true;
                INVERT_CONDITIONAL_BRANCHES = true;
                if (strcmp(optarg, "c") == 0) {        // -f c: invert cold only
                    INVERT_HOT_BRANCHES = false;
                } else if (strcmp(optarg, "h") == 0) { // -f h: invert cold+hot
                    INVERT_HOT_BRANCHES = true;
                } else {
                    FATAL("Invalid parameter: %s for option: %c", optarg, c);
                }
                break;

	    case 'g':			// -g filename: specify sloop output
		if (! (sloopfile = fopen(optarg, "w"))) {
		    PFATAL(optarg);
		}
		sloopFilename = optarg;
		break;

	    case 'h':			// -h: print help
		listUsage = true;
		break;
/*orig code start
            case 'i':
                INSERT_JUNK = true;
                if (strcmp(optarg, "p") == 0) {        // -i p: jump
                    JUMP_IN_JUNK = true;
                    JUNK_IN_JUNK = false;
                } else if (strcmp(optarg, "k") == 0) { // -i k: junk
                    JUMP_IN_JUNK = false;
                    JUNK_IN_JUNK = true;
                } else if (strcmp(optarg, "a") == 0) { // -i a: jump + junk
                    JUMP_IN_JUNK = true;
                    JUNK_IN_JUNK = true;
                } else {
                    FATAL("Invalid parameter: %s for option: %c", optarg, c);
                }
                break;
orig code end*/
//new code - insert 5 types of conditional jumps into junk 03/16/06
//---new code start
            case 'i':
                INSERT_JUNK = true;

                JUMP_IN_JUNK_NEAR_0 = false;  // a
                JUMP_IN_JUNK_NEAR_1 = false;  // b
                JUMP_IN_JUNK_NEAR_2 = false;  // c
                JUMP_IN_JUNK_SHORT_0 = false; // d
                JUMP_IN_JUNK_SHORT_1 = false; // e
                JUMP_IN_JUNK_NEAR_2_SHORT_1 = false; // f
                JUMP_IN_JUNK_SKIP_UNCOND = false; // u

                if (strcmp(optarg, "p") == 0) {        // -i p: jump
                    JUMP_IN_JUNK = true;
                    JUNK_IN_JUNK = false;
                    JUMP_IN_JUNK_NEAR_2_SHORT_1 = true; //0.5 near + 0.5 short
                } else if (strcmp(optarg, "k") == 0) { // -i k: junk
                    JUMP_IN_JUNK = false;
                    JUNK_IN_JUNK = true;
                } else if (strcmp(optarg, "a") == 0) { // -i a: jump + junk
                    JUMP_IN_JUNK = true;
                    JUNK_IN_JUNK = true;
                    JUMP_IN_JUNK_NEAR_0 = true; //near, target = 0
                } else if (strcmp(optarg, "b") == 0) { // -i b: jump + junk
                    JUMP_IN_JUNK = true;
                    JUNK_IN_JUNK = true;
                    JUMP_IN_JUNK_NEAR_1 = true; //near, target = [80h..7fh]
                } else if (strcmp(optarg, "c") == 0) { // -i c: jump + junk
                    JUMP_IN_JUNK = true;
                    JUNK_IN_JUNK = true;
                    JUMP_IN_JUNK_NEAR_2 = true; //near, target = [8000h..7fffh]
                } else if (strcmp(optarg, "d") == 0) { // -i d: jump + junk
                    JUMP_IN_JUNK = true;
                    JUNK_IN_JUNK = true;
                    JUMP_IN_JUNK_SHORT_0 = true; //short, target = 0
                } else if (strcmp(optarg, "e") == 0) { // -i e: jump + junk
                    JUMP_IN_JUNK = true;
                    JUNK_IN_JUNK = true;
                    JUMP_IN_JUNK_SHORT_1 = true; //short, target = [80h..7fh]
                } else if (strcmp(optarg, "f") == 0) { // -i f: 0.5c + 0.5e
                    JUMP_IN_JUNK = true;
                    JUNK_IN_JUNK = true;
                    JUMP_IN_JUNK_NEAR_2_SHORT_1 = true; //0.5 near + 0.5 short
                } else if (strcmp(optarg, "u") == 0) { // -i u: like above,
                    JUMP_IN_JUNK = true;               //       but no jcc
                    JUNK_IN_JUNK = true;               //       after jmp
                    JUMP_IN_JUNK_NEAR_2_SHORT_1 = true; //0.5 near + 0.5 short
                    JUMP_IN_JUNK_SKIP_UNCOND = true;
                } else {
                    FATAL("Invalid parameter: %s for option: %c", optarg, c);
                }
                break;
//---new code end
            case 'j':
                INDIRECT_CONTROL_FLOW = true;
                SPOOF_JUMP_TABLES = true;
                /* what to do with the actual target */
                if (strcmp(optarg, "n") == 0) {        // -j n: do nothing
                    JT_TARGET_BF = false;
                    JT_TARGET_TRAP = false;
                } else if (strcmp(optarg, "b") == 0) { // -j b: convert to bf
                    JT_TARGET_BF = true;
                    JT_TARGET_TRAP = false;
                } else if (strcmp(optarg, "s") == 0) { // -j t: convert to trap
                    JT_TARGET_BF = false;
                    JT_TARGET_TRAP = true;
                } else {
                    FATAL("Invalid parameter: %s for option: %c", optarg, c);
                }
                break;

	    case 'k':
	    	SetParameter(GetParameter("KernelMode"), "ON");
	    	break;

	    case 'l':			// -l: print config
		listConfig = true;
		break;

            case 'm':
                INTERLEAVE_HOT_COLD = true;
                break;

	    case 'n':			// -n: no optimization
		OPT_DISABLE_ALL = true;
		break;

            case 'o':                   // -o filename: specify binary output
                newFilename = optarg;
                break;

	    case 'p':			// -p x: add profiling specified by x
		if (strcmp(optarg, "a") == 0) {        // -p a: profile both
		    PROFILE_BBLS = true;
		    PROFILE_EDGES = true;
		} else if (strcmp(optarg, "b") == 0) { // -p b: profile bbls
		    PROFILE_BBLS = true;
		    PROFILE_EDGES = false;
		} else if (strcmp(optarg, "e") == 0) { // -p e: profile edges
		    PROFILE_BBLS = false;
                    PROFILE_EDGES = true;
		} else {
		    FATAL("Invalid parameter: %s for option: %c", optarg, c);
		}
		break;

	    case 'P':			// -P: add value prof
		PROFILE_VALUES = true;
		break;

            case 'q':
                EXTEND_CANDIDATES = true;
                EQUALIZE_BLOCKS = true;
                break;

            case 's':
                INDIRECT_CONTROL_FLOW = true;
                INSTALL_TRAPS = true;
                if (strcmp(optarg, "j") == 0) {        // -sj: jumps
                    TRAPS_CONVERT_CALLS = false;
                    TRAPS_CONVERT_RETURNS = false;
                } else if (strcmp(optarg, "c") == 0) { // -sc: jumps + calls
                    TRAPS_CONVERT_CALLS = true;
                    TRAPS_CONVERT_RETURNS = false;
                } else if (strcmp(optarg, "r") == 0) { // -sr: jumps + rets
                    TRAPS_CONVERT_CALLS = false;
                    TRAPS_CONVERT_RETURNS = true;
                } else if (strcmp(optarg, "a") == 0) { // -sa: jumps+calls+rets
                    TRAPS_CONVERT_CALLS = true;
                    TRAPS_CONVERT_RETURNS = true;
                } else {
                    FATAL("Invalid parameter: %s for option: %c", optarg, c);
                }
                break;

	    case 't':
		showTimestamps = true;
		break;

	    case 'v':
		SetParameter(GetParameter("Verbosity"), optarg);
		break;

	    case 'w':			// -w: disable warnings
		noWarnings = true;
		break;

	    case 'x':			// -x xyz: disable optimizations x,y,z
		SetFlags(c, optarg, PT_OPTIM, false);
		break;

	    default:
		MALFUNCTION("Accepted option -%c not handled", c);
	}
    }
}



/*
 *  InitializeGlobals() -- initialize some globals that depend on options.
 */
static void InitializeGlobals() {

    if (PROFILE_EDGES || PROFILE_VALUES) {
	OPT_DISABLE_ALL = true;	// if instrumenting, inhibit optimization
    }

    verbosityLevel = IntegerParameter("Verbosity");
    if (verbosityLevel == 0) {
	noWarnings = true;
    }

    MaxInstrLength = IntegerParameter("MaxInstructionLength");

    OPT_ELIMINATE_DEAD_REGISTERS = BooleanParameter("RemoveDeadRegisterStores");
    OPT_ELIMINATE_DEAD_MEMORY = BooleanParameter("RemoveDeadStackStores");

    PEEPHOLE_STUPID_MOVES = BooleanParameter("PeepholeStupidMoves");
    PEEPHOLE_STUPID_JUMPS = BooleanParameter("PeepholeStupidJumps");
    PEEPHOLE_COALESCE_ADDS = BooleanParameter("PeepholeCoalesceAdds");
    PEEPHOLE_STUPID_MATH = BooleanParameter("PeepholeStupidMath");
    PEEPHOLE_UNUSED_PUSHES = BooleanParameter("PeepholeUnusedPushes");
    PEEPHOLE_CONDITIONAL_MOVE = BooleanParameter("PeepholeConditionalMove");

    OPT_PEEPHOLE = PEEPHOLE_STUPID_MOVES || PEEPHOLE_STUPID_JUMPS ||
	PEEPHOLE_COALESCE_ADDS || PEEPHOLE_STUPID_MATH ||
	PEEPHOLE_UNUSED_PUSHES || PEEPHOLE_CONDITIONAL_MOVE;

    INLINING_MAX_FUNSIZE = IntegerParameter("MaxSizeFuncInlined");

    READ_FAKE_RELOCATIONS = BooleanParameter("ReadFakeRelocations");

    KERNEL_MODE = BooleanParameter("KernelMode");

    switch (StringParameter("Layout")[0]) {
	case 'O':  BBL_LAYOUT_SCHEME = BBL_LAYOUT_OFF;		break;
	case 'P':  BBL_LAYOUT_SCHEME = BBL_LAYOUT_PH;		break;
	case 'G':  BBL_LAYOUT_SCHEME = BBL_LAYOUT_GREEDY;	break;
	case 'A':  BBL_LAYOUT_SCHEME = BBL_LAYOUT_ALTO;		break;
	case 'F':  BBL_LAYOUT_SCHEME = BBL_LAYOUT_FUNS_ONLY;	break;
	default:   MALFUNCTION("unrecognized Layout spec");
    }

    switch (StringParameter("OutputType")[0]) {
	case 'C':  output_binary_type = BINARY_CORE_TYPE;		break;
	case 'S':  output_binary_type = BINARY_SHARED_OBJECT_TYPE;	break;
	case 'R':  output_binary_type = BINARY_RELOCATABLE_TYPE;	break;
	case 'E':  output_binary_type = BINARY_EXECUTABLE_TYPE;		break;
	case 'P':  output_binary_type = BINARY_PLTO_RELOCATABLE_TYPE;	break;
	default:   MALFUNCTION("unrecognized OutputType spec");
    }

    InitializeLivenessParameters();
}



/*
 *  ReadConfig(filename) -- read configuration file.
 */
static void ReadConfig(char *filename) {
    FILE *f;
    char linebuf[CLMAX];

    f = fopen(filename, "r");
    if (f == NULL) {
	PFATAL(filename);
    }
    curfile = filename;

    while (fgets(linebuf, CLMAX, f) != NULL) {
	ParseLine(linebuf);
    }

    fclose(f);
    curfile = NULL;
}



/*
 *  ParseLine(s) -- parse one line of a configuration file.
 *
 *  Interprets one line of a configuration file and stores the setting.
 *  Aborts in case of error.
 */
static void ParseLine(char *s) {
    Parameter *p;
    char name[CLMAX], value[CLMAX];

    if (sscanf(s, " %1[^#]", name) < 1) {
	return;			// comment line
    }
    if (sscanf(s, " %[_a-zA-Z0-9] : %[^# \t\n]", name, value) != 2) {
	ConfigError("Malformed configuration line: %s", s, NULL);
    }
    p = GetParameter(name);
    if (p == NULL) {
	ConfigError("Unrecognized parameter name in %s: %s", cfgFilename, name);
    }
    SetParameter(p, value);
}



/*
 *  MakeFilename(cfgname, basename) -- make file name.
 *
 *  cfgname is the name of a configuration parameter to be treated as a pattern.
 *  If the pattern is of the form "*.extn", then the resulting name is the base
 *  filename with the pattern's extension substituted or appended.  Otherwise,
 *  the pattern is treated as a simple file name and returned verbatim.
 */
/*char *MakeFilename(char *cfgname, char *basename) {
    char *pattern, *extn, *s;

    pattern = StringParameter(cfgname);
    if (strncmp(pattern, "*.", 2) == 0) {
	s = alloc(strlen(pattern) + strlen(basename));
	strcpy(s, basename);
	if ((extn = strrchr(s, '.')) != NULL) {
	    *extn = '\0';
	}
	return strcat(s, pattern + 1);
    } else {
	return pattern;
    }
}
*/


/*
 *  GetParameter(name) -- return a pointer to a Parameter struct.
 *
 *  Returns null if name is unrecognized.
 */
/*Parameter *GetParameter(char *name) {
    int i;
    for (i = 0; i < ARRAYSIZE(paramList); i++) {
	Parameter *p = &paramList[i];
	if (strcmp(p->pm_name, name) == 0) {
	    return p;
	}
    }
    return NULL;
}
*/


/*
 *  GetTypedParameter(name, mask) -- return a pointer to a Parameter struct.
 *
 *  mask: bitmask indicating which types are legal
 *
 *  Aborts if name is unrecognized or if type is wrong.
 */
/*static Parameter *GetTypedParameter(char *name, int mask) {
    Parameter *p = GetParameter(name);
    if (p == NULL) {
	MALFUNCTION("Unrecognized parameter name: %s", name); //MALFUNCTION
    } else if (((mask >> p->pm_type) & 1) == 0) {
	MALFUNCTION("Wrong parameter type %d for %s", p->pm_type, name);
    }
    return p;
}
*/


/*
 *  BooleanParameter(name) -- get a boolean parameter value.
 */
/*bool BooleanParameter(char *name) {
    int mask = (1 << PT_SWITCH) | (1 << PT_OPTIM) | (1 << PT_DEBUG);
    return (bool) GetTypedParameter(name, mask)->pm_nval;
}
*/


/*
 *  IntegerParameter(name) -- get an integer parameter value.
 */
/*long IntegerParameter(char *name) {
    int mask = 1 << PT_INTEGER;
    return (long) GetTypedParameter(name, mask)->pm_nval;
}
*/


/*
 *  FloatParameter(name) -- get a floating parameter value.
 */
/*double FloatParameter(char *name) {
    int mask = 1 << PT_FLOAT;
    return GetTypedParameter(name, mask)->pm_nval;
}
*/


/*
 *  StringParameter(name) -- get a string parameter value.
 */
/*char *StringParameter(char *name) {
    int mask = (1 << PT_STRING) | (1 << PT_CHOICE);
    return GetTypedParameter(name, mask)->pm_sval;
}
*/


/*
 *  PrintConfig(f) -- list all configuration values on file f.
 */
static void PrintConfig(FILE *f) {
    int i;
    char *v;
    char lbuf[1000], nbuf[20], sbuf[20];

    fprintf(f, "\n");
    fprintf(f, "#  PLTO configuration\n");

    for (i = 0; i < ARRAYSIZE(paramList); i++) {
	Parameter *p = &paramList[i];

	switch (p->pm_type) {
	    case PT_SWITCH:
	    case PT_OPTIM:
	    case PT_DEBUG:
		v = p->pm_nval ? "ON" : "OFF";
		break;
	    case PT_INTEGER:
		sprintf(v = nbuf, "%ld", (long) p->pm_nval);
		break;
	    case PT_FLOAT:
		sprintf(v = nbuf, "%.3f", p->pm_nval);
		break;
	    case PT_SPACER:
	    case PT_STRING:
	    case PT_CHOICE:
		v = p->pm_sval;
		break;
	}

	strcpy(sbuf, "");
	if (p->pm_char) {
	    switch (p->pm_type) {
		case PT_SWITCH:
		    sprintf(sbuf, "\t\t# -%c", p->pm_char);
		    break;
		case PT_OPTIM:
		    sprintf(sbuf, "\t\t# -e%c", p->pm_char);
		    break;
		case PT_DEBUG:
		    sprintf(sbuf, "\t\t# -d%c", p->pm_char);
		    break;
		case PT_INTEGER:
		    sprintf(sbuf, "\t\t# -%c nnn", p->pm_char);
		    break;
		case PT_FLOAT:
		    sprintf(sbuf, "\t\t# -%c nn.nn", p->pm_char);
		    break;
	    }
	}

	sprintf(lbuf, "%-32s %s%s\n", p->pm_name, v ? v : "", sbuf);
	if (*p->pm_name) {
	    *strchr(lbuf, ' ') = ':';
	}
	fputs(lbuf, f);
    }
}



/*
 *  SetParameter(param,value) -- set a parameter to the given value.
 *
 *  Aborts if the value is illegal (unrecognized choice or malformed number).
 */
static void SetParameter(Parameter *p, char *value) {
    int i;
    char c[2];
    char **list;

    p->pm_sval = salloc(value);
    switch (p->pm_type) {

	case PT_STRING:
	    break;	// nothing more to do

	case PT_INTEGER:
	case PT_FLOAT:
	    if ((sscanf(value, "%lf%c", &p->pm_nval, c) != 1)
	    || (p->pm_type == PT_INTEGER && p->pm_nval != (long) p->pm_nval)) {
		ConfigError("Illegal value for %s: %s", p->pm_name, value);
	    }
	    break;

	case PT_OPTIM:
	case PT_DEBUG:
	case PT_SWITCH:
	case PT_CHOICE:
	    for (i = 0, list = p->pm_list; *list != NULL; i++, list++) {
		if (strcmp(*list, value) == 0) {
		    p->pm_sval = *list;
		    p->pm_nval = i;
		    break;
		}
	    }
	    if (*list == NULL) {
		ConfigError("Illegal value for %s: %s", p->pm_name, value);
	    }
	    break;

	case PT_SPACER:
	default:
	    MALFUNCTION("Invalid parameter type %d for %s",
		p->pm_type, p->pm_name);
    }
}



/*
 *  SetFlags(c,s,type,b) -- set flags for -d, -e, or -x.
 */
static void SetFlags(char c, char *s, int type, bool b) {
    int i, n;

    for (; *s != '\0'; s++) {
	for (i = n = 0; i < ARRAYSIZE(paramList); i++) {
	    Parameter *p = &paramList[i];
	    if (p->pm_type == type && p->pm_char == *s) {
		p->pm_nval = b;
		p->pm_sval = b ? "ON" : "OFF";
		n++;
	    }
	}
	if (n == 0) {
	    FATAL("Unrecognized flag: -%c %c", c, *s);
	}
    }
}



/*
 *  ConfigError(msg,arg1,arg2) -- abort on error that may be from a config file.
 */
static void ConfigError(char *msg, char *arg1, char *arg2) {
    if (curfile != NULL) {
	fprintf(stderr, "In configuration file %s:\n", curfile);
    }
    FATAL(msg, arg1, arg2);
}



/*
 *  UsageError() -- abort due to usage error.
 */
static void UsageError() {
    fprintf(stderr, "Usage: plto [options] binaryfile [profiledata]"
	"    (plto -h for details)\n");
    exit(1);
}



/*
 *  List of strings that are output in response to "plto -h".
 */
static char *helpText[] = {
"",
"Usage: plto [options] binaryfile [profiledata]",
"",
"",
"CONTROL OPTIONS",
"",
"-k             treat the input as a Linux kernel image",
"-p x           instrument for bbl/edge profiling specified by x, suppressing",
"               optimization",
"        -p a   profile both (basic blocks and edges)",
"        -p b   profile basic blocks",
"        -p e   profile edges",
"-P             instrument for value profiling, suppressing optimization",
"",
"-o outputfile  specify output file for executable (default is *.opt/*.prof)",
"",
"-g sloopfile   write graph information for use by sloop",
"",
"",
"OPTIMIZATION OPTIONS",
"",
"-c configfile  specify configuration; overrides $" PLTO_ENVVAR,
"",
"-e xyz         enable optimizations abbreviated by characters x, y, z",
"-x xyz         exclude optimizations abbreviated by characters x, y, z",
"",
"-E XxxXxxx     enable optimization named XxxXxxx",
"-X XxxXxxx     exclude optimization named XxxXxxx",
"",
"-D XxXxx=vvvv  define parameter value",
"",
"-n             no optimization; master cutoff switch",
"",
"Command arguments override any configuration file values.",
"",
"",
"OUTPUT OPTIONS",
"",
"-h             print usage help (this page)",
"-l             list the effective configuration, showing all possible options",
"-t             add time stamps to commentary on standard output",
"-w             suppress warning messages",
"",
"-d xyz         enable debugging options x, y, z",
"",
"-v n           control verbosity of commentary on standard output",
"        -v 0   silent (except for fatal errors)",
"        -v 1   major milestones (loading, disassembling, optimizing, ...)",
"        -v 2   minor milestones (reading elf, merging sects, symtab, ...)",
"        -v 3   more detail (section sizes, disassembly statistics, ...)",
"        -v 4   reports from within small loops (typically < 100 iterations)",
"        -v 5   more detail within small loops",
"               [ beyond this point the output can grow very large ]",
"        -v 6   reports from within larger loops",
"        -v 7   more detail within larger loops",
"        -v 8   the most trivial details",
"        -v 9   all possible output including all debugging options",
"",
};



/*
 *  PrintHelp(f) -- print long help message on file f.
 */
static void PrintHelp(FILE *f) {
    int i;
    for (i = 0; i < ARRAYSIZE(helpText); i++) {
	fprintf(f, "%s\n", helpText[i]);
    }
}
