/*
 *  perfect.c
 *
 *  This file was taken from the source cited below. It was then modified to
 *  use in binary obfuscator to create a perfect hash function for two types
 *  of obfuscation:
 *      - signal based obfuscation
 *      - changing many jumps/calls to call to one branch function
 *
 *  --------------------------------------------------------------------------
 *  perfect.c: code to generate code for a hash for perfect hashing.
 *  (c) Bob Jenkins, September 1996, December 1999
 *  You may use this code in any way you wish, and it is free.  No warranty.
 *  I hereby place this in the public domain.
 *  Source is http://burtleburtle.net/bob/c/perfect.c
 *
 *  This generates a minimal perfect hash function.  That means, given a
 *  set of n keys, this determines a hash function that maps each of
 *  those keys into a value in 0..n-1 with no collisions.
 *
 *  The perfect hash function first uses a normal hash function on the key
 *  to determine (a,b) such that the pair (a,b) is distinct for all
 *  keys, then it computes a^scramble[tab[b]] to get the final perfect hash.
 *  tab[] is an array of 1-byte values and scramble[] is a 256-term array of 
 *  2-byte or 4-byte values.  If there are n keys, the length of tab[] is a 
 *  power of two between n/3 and n.
 *
 *  I found the idea of computing distinct (a,b) values in "Practical minimal 
 *  perfect hash functions for large databases", Fox, Heath, Chen, and Daoud, 
 *  Communications of the ACM, January 1992.  They found the idea in Chichelli 
 *  (CACM Jan 1980).  Beyond that, our methods differ.
 *
 *  The key is hashed to a pair (a,b) where a in 0..*alen*-1 and b in
 *  0..*blen*-1.  A fast hash function determines both a and b
 *  simultaneously.  Any decent hash function is likely to produce
 *  hashes so that (a,b) is distinct for all pairs.  I try the hash
 *  using different values of *salt* until all pairs are distinct.
 *
 *  The final hash is (a XOR scramble[tab[b]]).  *scramble* is a
 *  predetermined mapping of 0..255 into 0..smax-1.  *tab* is an
 *  array that we fill in in such a way as to make the hash perfect.
 *
 *  First we fill in all values of *tab* that are used by more than one
 *  key.  We try all possible values for each position until one works.
 *
 *  This leaves m unmapped keys and m values that something could hash to.
 *  If you treat unmapped keys as lefthand nodes and unused hash values
 *  as righthand nodes, and draw a line connecting each key to each hash
 *  value it could map to, you get a bipartite graph.  We attempt to
 *  find a perfect matching in this graph.  If we succeed, we have
 *  determined a perfect hash for the whole set of keys.
 *
 *  *scramble* is used because (a^tab[i]) clusters keys around *a*.
 *  --------------------------------------------------------------------------
 */
#include "standard.h"
#include "lookupa.h"
#include "recycle.h"
#include "perfect.h"
#include "global.h"


/*
 * ---------------------------------------------------------------------------
 *  Find the mapping that will produce a perfect hash
 * ---------------------------------------------------------------------------
 */

/*
 *  log_2(val) --
 *
 *  Returns the ceiling of the log (base 2) of val.
 */
UINT32 log_2(UINT32 val) {
    UINT32 i;
    for (i = 0; ((UINT32) 1 << i) < val; ++i) {
    };
    return i;
}



/*
 *  permute(x, nbits) --
 *
 *  Computes p(x), where p is a permutation of 0..(1 << nbits) - 1
 *  permute(0) = 0.  This is intended and useful.
 *
 *  Formal parameters:
 *      x     - input, a value in some range,
 *      nbits - input, number of bits in range.
 */
static UINT32 permute(UINT32 x, UINT32 nbits) {
    int i;
    int mask = ((UINT32) 1 << nbits) - 1;     /* all ones */
    int const2 = 1 + nbits / 2;
    int const3 = 1 + nbits / 3;
    int const4 = 1 + nbits / 4;
    int const5 = 1 + nbits / 5;
    for (i = 0; i < 20; ++i) {
        x = (x + (x << const2)) & mask;
        x = (x ^ (x >> const3));
        x = (x + (x << const4)) & mask;
        x = (x ^ (x >> const5));
    }
    return x;
}



/*
 *  scrambleinit(scramble, smax) --
 *
 *  Initializes scramble[] with distinct random values in 0..smax-1.
 *
 *  Formal parameters:
 *      scramble - hash is a^scramble[tab[b]],
 *      smax     - scramble values should be in 0..smax-1.
 */
static void scrambleinit(UINT32 *scramble, UINT32 smax) {
    UINT32 i;

    /* fill scramble[] with distinct random integers in 0..smax-1 */
    for (i = 0; i < SCRAMBLE_LEN; ++i) {
        scramble[i] = permute(i, log_2(smax));
    }
}



/*
 *  checkdup(key1, key2, form) --
 *
 *  Checks if key1 and key2 are the same.
 *  We already checked (a,b) are the same.
 */
static void checkdup(key *key1, key *key2, hashform *form) {
    if (key1->hash_k == key2->hash_k) {
        fprintf(stderr, "perfect.c: Duplicate keys!  %.8x\n", key1->hash_k);
        exit(SUCCESS);
    }
}



/*
 *  inittab(tabb, blen, keys, form, complete) --
 *
 *  Puts keys in tabb according to key->b_k.
 *  Checks if the initial hash might work.
 *
 *  Formal parameters:
 *      tabb     - output, list of keys with b for (a,b),
 *      blen     - length of tabb,
 *      keys     - list of keys already hashed,
 *      form     - user directives,
 *      complete - true means to complete init despite collisions.
 */
static int inittab(bstuff *tabb, UINT32 blen, key *keys, hashform *form,
    int complete) {

    int nocollision = true;
    key *mykey;

    memset((void *)tabb, 0, (size_t) (sizeof(bstuff) * blen));

    /* Two keys with the same (a,b) guarantees a collision */
    for (mykey = keys; mykey; mykey = mykey->next_k) {
        key *otherkey;

        for (otherkey = tabb[mykey->b_k].list_b; otherkey;
            otherkey = otherkey->nextb_k) {
            if (mykey->a_k == otherkey->a_k) {
                nocollision = false;
                checkdup(mykey, otherkey, form);
                if (!complete) {
                    return false;
                }
            }
        }
        ++tabb[mykey->b_k].listlen_b;
        mykey->nextb_k = tabb[mykey->b_k].list_b;
        tabb[mykey->b_k].list_b = mykey;
    }

    /* no two keys have the same (a,b) pair */
    return nocollision;
}



/*
 *  initnorm(keys, alen, blen, smax, salt, final) --
 *
 *  Does the initial hash for normal mode (uses lookup and checksum).
 *
 *  Formal parameters:
 *      keys  - list of all keys,
 *      alen  - (a,b) has a in 0..alen-1, a power of 2,
 *      blen  - (a,b) has b in 0..blen-1, a power of 2,
 *      smax  - maximum range of computable hash values,
 *      salt  - used to initialize the hash function,
 *      final - output, code for the final hash.
 */
static void initnorm(key *keys, UINT32 alen, UINT32 blen, UINT32 smax,
    UINT32 salt, gencode *final) {

    key *mykey;
    if (log_2(alen) + log_2(blen) > UB4BITS) {
        /* the golden ratio; an arbitrary value */
        UINT32 initlev = salt * 0x9e3779b9;

        for (mykey = keys; mykey; mykey = mykey->next_k) {
            UINT32 i, state[CHECKSTATE];
            for (i = 0; i < CHECKSTATE; ++i) {
                state[i] = initlev;
            }
            checksum(mykey->name_k, mykey->len_k, state);
            mykey->a_k = state[0] & (alen - 1);
            mykey->b_k = state[1] & (blen - 1);
        }
        final->used = 4;
        sprintf(final->line[0], "    ub4 i, state[CHECKSTATE], rsl;\n");
        sprintf(final->line[1],
            "    for (i = 0; i < CHECKSTATE; ++i) state[i] = 0x%x;\n", initlev);
        sprintf(final->line[2], "    checksum(key, len, state);\n");
        sprintf(final->line[3],
            "    rsl = ((state[0] & 0x%x) ^ scramble[tab[state[1] & 0x%x]]);\n",
            alen - 1, blen - 1);
    } else {
        UINT32 loga = log_2(alen); /* log based 2 of blen */
        /* the golden ratio; an arbitrary value */
        UINT32 initlev = salt * 0x9e3779b9;

        for (mykey = keys; mykey; mykey = mykey->next_k) {
            UINT32 hash = lookup(mykey->name_k, mykey->len_k, initlev);
            mykey->a_k = (loga > 0) ? hash >> (UB4BITS - loga) : 0;
            mykey->b_k = (blen > 1) ? hash & (blen - 1) : 0;
        }
        final->used = 2;
        sprintf(final->line[0],
            "    ub4 rsl, val = lookup(key, len, 0x%x);\n", initlev);
        if (smax <= 1) {
            sprintf(final->line[1], "    rsl = 0;\n");
        } else if (blen < USE_SCRAMBLE) {
            sprintf(final->line[1], "    rsl = ((val>>%d) ^ tab[val&0x%x]);\n",
                UB4BITS - log_2(alen), blen - 1);
        } else {
            sprintf(final->line[1],
                "    rsl = ((val >> %d) ^ scramble[tab[val & 0x%x]]);\n",
                UB4BITS - log_2(alen), blen - 1);
        }
    }
}



/*
 *  initinl(keys, alen, blen, smax, salt, final) --
 *
 *  Does initial hash for inline mode.
 *
 *  Formal parameters:
 *      keys  - list of all keys,
 *      alen  - (a,b) has a in 0..alen-1, a power of 2,
 *      blen  - (a,b) has b in 0..blen-1, a power of 2,
 *      smax  - range of computable hash values,
 *      salt  - used to initialize the hash function,
 *      final - generated code for final hash.
 */
static void initinl(key *keys, UINT32 alen, UINT32 blen, UINT32 smax,
    UINT32 salt, gencode *final) {

    key *mykey;
    UINT32 amask = alen - 1;
    UINT32 blog = log_2(blen);
    /* the golden ratio; an arbitrary value */
    UINT32 initval = salt * 0x9e3779b9;

    /* It's more important to have b uniform than a, so b is the low bits */
    for (mykey = keys; mykey != (key *)0; mykey = mykey->next_k) {
        UINT32 hash = initval;
        UINT32 i;
        for (i = 0; i < mykey->len_k; ++i) {
            hash = (mykey->name_k[i] ^ hash) +
                ((hash << (UB4BITS - 6)) + (hash >> 6));
        }
        mykey->hash_k = hash;
        mykey->a_k = (alen > 1) ? (hash & amask) : 0;
        mykey->b_k = (blen > 1) ? (hash >> (UB4BITS - blog)) : 0;
    }
    final->used = 1;
    if (smax <= 1) {
        sprintf(final->line[0], "    ub4 rsl = 0;\n");
    } else if (blen < USE_SCRAMBLE) {
        sprintf(final->line[0],
            "    ub4 rsl = ((val & 0x%x) ^ tab[val >> %d]);\n",
            amask, UB4BITS - blog);
    } else {
        sprintf(final->line[0],
            "    ub4 rsl = ((val & 0x%x) ^ scramble[tab[val >> %d]]);\n",
            amask, UB4BITS - blog);
    }
}



/*
 *  initkey(keys, nkeys, tabb, alen, blen, smax, salt, form, final) --
 *
 *  Runs a hash function on the key to get a and b.
 *
 *  Formal parameters:
 *      keys  - list of all keys,
 *      nkeys - total number of keys,
 *      tabb  - stuff indexed by b,
 *      alen  - (a,b) has a in 0..alen-1, a power of 2,
 *      blen  - (a,b) has b in 0..blen-1, a power of 2,
 *      smax  - range of computable hash values,
 *      salt  - used to initialize the hash function,
 *      form  - user directives,
 *      final - code for final hash.
 *
 *  Returns:
 *      0: didn't find distinct (a,b) for all keys,
 *      1: found distinct (a,b) for all keys, put keys in tabb[],
 *      2: found a perfect hash, no need to do any more work.
 */
static UINT32 initkey(key *keys, UINT32 nkeys, bstuff *tabb, UINT32 alen,
    UINT32 blen, UINT32 smax, UINT32 salt, hashform *form, gencode *final) {

    UINT32 finished;

    /* Do the initial hash of the keys */
    switch (form->mode) {
        case NORMAL_HM:
            initnorm(keys, alen, blen, smax, salt, final);
            break;
        case INLINE_HM:
            initinl(keys, alen, blen, smax, salt, final);
            break;
        case HEX_HM:
        case DECIMAL_HM:
            finished =
                inithex(keys, nkeys, alen, blen, smax, salt, final, form);
            if (finished) {
                return 2;
            }
            break;
        default:
            fprintf(stderr, "fatal error: illegal mode\n");
            exit(1);
    }

    if (nkeys <= 1) {
        final->used = 1;
        sprintf(final->line[0], "    ub4 rsl = 0;\n");
        return 2;
    }

    return inittab(tabb, blen, keys, form, false);
}



/*
 *  duplicates(tabb, blen, keys, form) --
 *
 *  Prints an error message and exits if there are duplicates.
 *
 *  Formal parameters:
 *      tabb - array of lists of keys with the same b,
 *      blen - length of tabb, a power of 2,
 *      keys
 *      form - user directives.
 */
static void duplicates(bstuff *tabb, UINT32 blen, key *keys, hashform *form) {
    UINT32 i;
    key *key1;
    key *key2;

    inittab(tabb, blen, keys, form, true);

    /* for each b, do nested loops through key list looking for duplicates */
    for (i = 0; i < blen; ++i) {
        for (key1 = tabb[i].list_b; key1; key1 = key1->nextb_k) {
            for (key2 = key1->nextb_k; key2; key2 = key2->nextb_k) {
                checkdup(key1, key2, form);
            }
        }
    }
}



/*
 *  apply(tabb, tabh, tabq, blen, scramble, tail, rollback) --
 *
 *  Tries to apply an augmenting list. When rollback is false - applies
 *  augmenting path, when true - rolls back.
 */
static int apply(bstuff *tabb, hstuff *tabh, qstuff *tabq, UINT32 blen,
    UINT32 *scramble, UINT32 tail, int rollback) {

    UINT32 hash;
    key *mykey;
    bstuff *pb;
    UINT32 child;
    UINT32 parent;
    UINT32 stabb;                 /* scramble[tab[b]] */

    /* walk from child to parent */
    for (child = tail - 1; child; child = parent) {
        parent = tabq[child].parent_q;            /* find child's parent */
        pb = tabq[parent].b_q;         /* find parent's list of siblings */

        /* erase old hash values */
        stabb = scramble[pb->val_b];
        for (mykey = pb->list_b; mykey; mykey = mykey->nextb_k) {
            hash = mykey->a_k ^ stabb;
            if (mykey == tabh[hash].key_h) {
                /* erase hash for all of child's siblings */
                tabh[hash].key_h = (key *)0;
            }
        }

        /* change pb->val_b, it will change the hashes of all parent siblings */
        pb->val_b = (rollback ? tabq[child].oldval_q : tabq[child].newval_q);

        /* set new hash values */
        stabb = scramble[pb->val_b];
        for (mykey = pb->list_b; mykey; mykey = mykey->nextb_k) {
            hash = mykey->a_k ^ stabb;
            if (rollback) {
                if (parent == 0) {
                    continue;                /* root never had a hash */
                }
            } else if (tabh[hash].key_h) {
                /* very rare: roll back any changes */
                apply(tabb, tabh, tabq, blen, scramble, tail, true);
                return false;               /* failure, collision */
            }
            tabh[hash].key_h = mykey;
        }
    }
    return true;
}



/*
 *  augment(tabb, tabh, tabq, blen, scramble, smax, item, nkeys, highwater,
 *      form) --
 *
 *  Adds item to the mapping.
 *
 *  Constructs a spanning tree of *b*s with *item* as root, where each
 *  parent can have all its hashes changed (by some new val_b) with 
 *  at most one collision, and each child is the b of that collision.
 *
 *  This is from Tarjan's "Data Structures and Network Algorithms".  The
 *  path from *item* to a *b* that can be remapped with no collision is 
 *  an "augmenting path".  Change values of tab[b] along the path so that 
 *  the unmapped key gets mapped and the unused hash value gets used.
 *
 *  Assuming 1 key per b, if m out of n hash values are still unused, 
 *  you should expect the transitive closure to cover n/m nodes before 
 *  an unused node is found.  Sum(i=1..n)(n/i) is about nlogn, so expect
 *  this approach to take about nlogn time to map all single-key b's.
 *
 *  Formal parameters:
 *      tabb      - stuff indexed by b,
 *      tabh      - which key is associated with which hash, indexed by hash,
 *      tabq      - queue of *b* values, this is the spanning tree,
 *      blen      - length of tabb,
 *      scramble  - final hash is a^scramble[tab[b]],
 *      smax      - highest value in scramble,
 *      item      - &tabb[b] for the b to be mapped,
 *      nkeys     - final hash must be in 0..nkeys-1,
 *      highwater - a value higher than any now in tabb[].water_b,
 *      form      - true if we should do a minimal perfect hash.
 */
static int augment(bstuff *tabb, hstuff *tabh, qstuff *tabq, UINT32 blen,
    UINT32 *scramble, UINT32 smax, bstuff *item, UINT32 nkeys,
    UINT32 highwater, hashform *form) {

    UINT32 q;             /* current position walking through the queue */
    UINT32 tail;          /* tail of the queue.  0 is the head of the queue. */
    UINT32 limit = ((blen < USE_SCRAMBLE) ? smax : UB1MAXVAL + 1);
    UINT32 highhash = ((form->perfect == MINIMAL_HP) ? nkeys : smax);
    int trans = (form->speed == SLOW_HS || form->perfect == MINIMAL_HP);

    /* initialize the root of the spanning tree */
    tabq[0].b_q = item;
    tail = 1;

    /* construct the spanning tree by walking the queue, add children to tail */
    for (q = 0; q < tail; ++q) {
        bstuff *myb = tabq[q].b_q;         /* the b for this node */
        UINT32 i;                          /* possible value for myb->val_b */

        if (!trans && (q == 1)) {
            break;                         /* don't do transitive closure */
        }

        for (i = 0; i < limit; ++i) {
            bstuff *childb = (bstuff *)0;  /* the b that this i maps to */
            key *mykey;                    /* for walking through myb's keys */

            for (mykey = myb->list_b; mykey; mykey = mykey->nextb_k) {
                key *childkey;
                UINT32 hash = mykey->a_k ^ scramble[i];

                if (hash >= highhash) {
                    break;                 /* out of bounds */
                }
                childkey = tabh[hash].key_h;

                if (childkey) {
                    bstuff *hitb = &tabb[childkey->b_k];

                    if (childb) {
                        if (childb != hitb) {
                            break;         /* hit at most one child b */
                        }
                    } else {
                        childb = hitb;     /* remember this as childb */
                        if (childb->water_b == highwater) {
                            break;         /* already explored */
                        }
                    }
                }
            }
            if (mykey) {
                continue;               /* myb with i has multiple collisions */
            }

            /* add childb to the queue of reachable things */
            if (childb) {
                childb->water_b = highwater;
            }
            tabq[tail].b_q = childb;
            tabq[tail].newval_q = i;/* how to make parent (myb) use this hash */
            tabq[tail].oldval_q = myb->val_b; /* need this for rollback */
            tabq[tail].parent_q = q;
            ++tail;

            if (!childb) {           /* found an *i* with no collisions? */
                /* try to apply the augmenting path */
                if (apply(tabb, tabh, tabq, blen, scramble, tail, false)) {
                    /* success, item was added to the perfect hash */
                    return true;
                }
                --tail;          /* don't know how to handle such a child! */
            }
        }
    }
    return false;
}



/*
 *  perfect( tabb, tabh, tabq, blen, smax, scramble, nkeys, form ) --
 *
 *  Finds a mapping that makes this a perfect hash.
 */
static int perfect(bstuff *tabb, hstuff *tabh, qstuff *tabq, UINT32 blen,
    UINT32 smax, UINT32 *scramble, UINT32 nkeys, hashform *form) {

    UINT32 maxkeys;               /* maximum number of keys for any b */
    UINT32 i, j;

    /* clear any state from previous attempts */
    memset((void *)tabh, 0, (size_t)(sizeof(hstuff) *
        ((form->perfect == MINIMAL_HP) ? nkeys : smax)));

    memset((void *)tabq, 0, (size_t)(sizeof(qstuff) * (blen + 1)));

    for (maxkeys = 0, i = 0; i < blen; ++i) {
        if (tabb[i].listlen_b > maxkeys) {
            maxkeys = tabb[i].listlen_b;
        }
    }

    /* In descending order by number of keys, map all *b*s */
    for (j = maxkeys; j > 0; --j) {
        for (i = 0; i < blen; ++i) {
            if (tabb[i].listlen_b == j) {
                if (!augment(tabb, tabh, tabq, blen, scramble, smax, &tabb[i],
                    nkeys, i + 1, form)) {
                    REPORT(RP_SMDETAIL, "fail to map group of size"
                        " %d for tab size %d", j, blen);
                    return false;
                }
            }
        }
    }

    /* Success!  We found a perfect hash of all keys into 0..nkeys-1. */
    return true;
}



/*
 *  hash_ab(tabb, alen, blen, salt, final, scramble, smax, keys, nkeys, form) --
 *
 *  Simple case: user gave (a,b).  No more mixing, no guessing alen or blen. 
 *  This assumes a,b reside in (key->a_k, key->b_k), and final->form == AB_HK.
 *
 *  Formal parameters:
 *      tabb     - output, tab[] of the perfect hash, length *blen,
 *      alen     - output, 0..alen-1 is range for a of (a,b),
 *      blen     - output, 0..blen-1 is range for b of (a,b),
 *      salt     - output, initializes initial hash,
 *      final    - code for final hash,
 *      scramble - input, hash = a^scramble[tab[b]],
 *      smax     - input, scramble[i] in 0..smax-1,
 *      keys     - input, keys to hash,
 *      nkeys    - input, number of keys being hashed,
 *      form     - user directives.
 */
static void hash_ab(bstuff **tabb, UINT32 *alen, UINT32 *blen, UINT32 *salt,
    gencode *final, UINT32 *scramble, UINT32 *smax, key *keys, UINT32 nkeys,
    hashform *form) {

    hstuff *tabh;
    qstuff *tabq;
    key *mykey;
    UINT32 i;
    int used_tab;

    /* initially make smax the first power of two bigger than nkeys */
    *smax = ((UINT32)1 << log_2(nkeys));
    scrambleinit(scramble, *smax);

    /* set *alen and *blen based on max A and B from user */
    *alen = 1;
    *blen = 1;
    for (mykey = keys; mykey != (key *)0; mykey = mykey->next_k) {
        while (*alen <= mykey->a_k) {
            *alen *= 2;
        }
        while (*blen <= mykey->b_k) {
            *blen *= 2;
        }
    }
    if (*alen > 2 * *smax) {
        fprintf(stderr,
            "perfect.c: Can't deal with (A,B) having A bigger than twice \n");
        fprintf(stderr, "  the smallest power of two greater or equal "
            "to any legal hash.\n");
        exit(SUCCESS);
    }

    /* allocate working memory */
    *tabb = (bstuff *)alloc((size_t)(sizeof(bstuff) * (*blen)));
    tabq = (qstuff *)remalloc(sizeof(qstuff) * (*blen + 1), "perfect.c, tabq");
    tabh = (hstuff *)remalloc(sizeof(hstuff) *
        (form->perfect == MINIMAL_HP ? nkeys : *smax), "perfect.c, tabh");

    /* check that (a,b) are distinct and put them in tabb indexed by b */
    inittab(*tabb, *blen, keys, form, false);

    /* try with smax */
    if (!perfect(*tabb, tabh, tabq, *blen, *smax, scramble, nkeys, form)) {
        if (form->perfect == MINIMAL_HP) {
            printf("fatal error: "
                "Cannot find perfect hash for user (A,B) pairs\n");
            exit(SUCCESS);
        } else {
            /* try with 2*smax */
            free((void *)tabh);
            *smax = *smax * 2;
            scrambleinit(scramble, *smax);
            tabh = (hstuff *)remalloc(sizeof(hstuff) *
                (form->perfect == MINIMAL_HP ? nkeys : *smax),
                "perfect.c, tabh");
            if (!perfect(*tabb, tabh, tabq, *blen, *smax, scramble, nkeys,
                form)) {
                printf ("fatal error: "
                    "Cannot find perfect hash for user (A,B) pairs\n");
                exit(SUCCESS);
            }
        }
    }

    /* check if tab[] was really needed */
    for (i = 0; i < *blen; ++i) {
        if ((*tabb)[i].val_b != 0) {
            break;                    /* assumes permute(0) == 0 */
        }
    }
    used_tab = (i < *blen);

    /* write the code for the perfect hash */
    *salt = 1;
    final->used = 1;
    if (!used_tab) {
        sprintf(final->line[0], "    ub4 rsl = a;\n");
    } else if (*blen < USE_SCRAMBLE) {
        sprintf(final->line[0], "    ub4 rsl = (a ^ tab[b]);\n");
    } else {
        sprintf(final->line[0], "    ub4 rsl = (a ^ scramble[tab[b]]);\n");
    }

    free((void *)tabq);
    free((void *)tabh);
}



/*
 *  initalen(alen, blen, smax, nkeys, form) --
 *
 *  Guesses the initial values for alen and blen. They are found empirically.
 *  Some factors:
 *
 *  If smax < 256 there is no scramble, so tab[b] needs to cover 0..smax-1.
 *
 *  alen and blen must be powers of 2 because the values in 0..alen-1 and
 *  0..blen-1 are produced by applying a bitmask to the initial hash function.
 *
 *  alen must be less than smax, in fact less than nkeys, because otherwise
 *  there would often be no i such that a^scramble[i] is in 0..nkeys-1 for
 *  all the *a*s associated with a given *b*, so there would be no legal
 *  value to assign to tab[b].  This only matters when we're doing a minimal
 *  perfect hash.
 *
 *  It takes around 800 trials to find distinct (a,b) with nkey = smax * (5/8)
 *  and alen * blen = smax * smax/32.
 *
 *  Values of blen less than smax/4 never work, and smax/2 always works.
 *
 *  We want blen as small as possible because it is the number of bytes in
 *  the huge array we must create for the perfect hash.
 *
 *  When nkey <= smax * (5/8), blen = smax/4 works much more often with
 *  alen = smax/8 than with alen = smax/4.  Above smax * (5/8), blen = smax/4
 *  doesn't seem to care whether alen = smax/8 or alen = smax/4. I think it
 *  has something to do with 5/8 = 1/8 * 5. For example examine 80000, 
 *  85000, and 90000 keys with different values of alen.  This only matters
 *  if we're doing a minimal perfect hash.
 *
 *  When alen * blen <= 1 << UB4BITS, the initial hash must produce one
 *  integer. Bigger than that it must produce two integers, which increases
 *  the cost of the hash per character hashed.
 *
 *  Formal parameters:
 *      alen  - output, initial alen,
 *      blen  - output, initial blen,
 *      smax  - input, power of two greater or equal to max hash value,
 *      nkeys - number of keys being hashed,
 *      form  - user directives.
 */
static void initalen(UINT32 *alen, UINT32 *blen, UINT32 *smax, UINT32 nkeys,
    hashform *form) {

    if (form->perfect == NORMAL_HP) {
        if ((form->speed == FAST_HS) && (nkeys > *smax * 0.8)) {
            *smax = *smax * 2;
        }

        *alen = ((form->hashtype == INT_HT) && *smax > 131072) ?
            /* distinct keys => distinct (A,B) */
            ((UINT32)1 << (UB4BITS - log_2(*blen))) :
            *smax;                   /* no reason to restrict alen to smax/2 */
        if ((form->hashtype == INT_HT) && *smax < 32) {
            *blen = *smax;                /* go for function speed not space */
        } else if (*smax / 4 <= (1 << 14)) {
            *blen = ((nkeys <= *smax * 0.56) ? *smax / 32 :
                (nkeys <= *smax * 0.74) ? *smax / 16 : *smax / 8);
        } else {
            *blen = ((nkeys <= *smax * 0.6) ? *smax / 16 :
                (nkeys <= *smax * 0.8) ? *smax / 8 : *smax / 4);
        }

        if ((form->speed == FAST_HS) && (*blen < *smax / 8)) {
            *blen = *smax / 8;
        }

        if (*alen < 1) {
            *alen = 1;
        }
        if (*blen < 1) {
            *blen = 1;
        }
    } else {
        switch (log_2(*smax)) {
            case 0:
                *alen = 1;
                *blen = 1;
            case 1:
            case 2:
            case 3:
            case 4:
            case 5:
            case 6:
            case 7:
            case 8:
                *alen = (form->perfect == NORMAL_HP) ? *smax : *smax / 2;
                *blen = *smax / 2;
                break;
            case 9:
            case 10:
            case 11:
            case 12:
            case 13:
            case 14:
            case 15:
            case 16:
            case 17:
                if (form->speed == FAST_HS) {
                    *alen = *smax / 2;
                    *blen = *smax / 4;
                } else if (*smax / 4 < USE_SCRAMBLE) {
                    *alen = ((nkeys <= *smax * 0.52) ? *smax / 8 : *smax / 4);
                    *blen = ((nkeys <= *smax * 0.52) ? *smax / 8 : *smax / 4);
                } else {
                    *alen = ((nkeys <= *smax * (5.0 / 8.0)) ? *smax / 8 :
                        (nkeys <= *smax * (3.0 / 4.0)) ? *smax / 4 : *smax / 2);
                    *blen = *smax / 4;  /* always give the small size a shot */
                }
                break;
            case 18:
                if (form->speed == FAST_HS) {
                    *alen = *smax / 2;
                    *blen = *smax / 2;
                } else {
                    *alen = *smax / 8;   /* never require the multiword hash */
                    *blen = (nkeys <= *smax * (5.0 / 8.0)) ? *smax / 4 :
                        *smax / 2;
                }
                break;
            case 19:
            case 20:
                *alen = (nkeys <= *smax * (5.0 / 8.0)) ? *smax / 8 : *smax / 2;
                *blen = (nkeys <= *smax * (5.0 / 8.0)) ? *smax / 4 : *smax / 2;
                break;
            default:
                *alen = *smax / 2;  /* just find a hash as quick as possible */
                /* we'll be thrashing virtual memory at this size */
                *blen = *smax / 2;
                break;
        }
    }
}



/*
 *  findhash(tabb, alen, blen, salt, final, scramble, smax, keys,
 *      nkeys, form) --
 *
 *  Given the keys, scramble[], and hash mode, finds the perfect hash.
 *
 *  Formal parameters:
 *      tabb     - output, tab[] of the perfect hash, length *blen,
 *      alen     - output, 0..alen-1 is range for a of (a,b),
 *      blen     - output, 0..blen-1 is range for b of (a,b),
 *      salt     - output, initializes initial hash,
 *      final    - code for final hash,
 *      scramble - input, hash = a^scramble[tab[b]],
 *      smax     - input, scramble[i] in 0..smax-1,
 *      keys     - input, keys to hash,
 *      nkeys    - input, number of keys being hashed,
 *      form     - user directives.
 */
void findhash(bstuff **tabb, UINT32 *alen, UINT32 *blen, UINT32 *salt,
    gencode *final, UINT32 *scramble, UINT32 *smax, key *keys, UINT32 nkeys,
    hashform *form) {

    UINT32 bad_initkey;           /* how many times did initkey fail? */
    UINT32 bad_perfect;           /* how many times did perfect fail? */
    UINT32 trysalt;               /* trial initializer for initial hash */
    UINT32 maxalen;
    hstuff *tabh;                 /* table of keys indexed by hash value */
    qstuff *tabq;  /* table of stuff indexed by queue value, used by augment */

    /* The case of (A,B) supplied by the user is a special case */
    if (form->hashtype == AB_HT) {
        hash_ab(tabb, alen, blen, salt, final, scramble, smax, keys,
            nkeys, form);
        return;
    }

    /* guess initial values for smax, alen and blen */
    *smax = ((UINT32)1 << log_2(nkeys));
    initalen(alen, blen, smax, nkeys, form);

    scrambleinit(scramble, *smax);

    maxalen = (form->perfect == MINIMAL_HP) ? *smax / 2 : *smax;

    /* allocate working memory */
    *tabb = (bstuff *)remalloc((size_t)(sizeof(bstuff) * (*blen)),
        "perfect.c, tabb");
    tabq = (qstuff *)remalloc(sizeof(qstuff) * (*blen + 1), "perfect.c, tabq");
    tabh = (hstuff *)remalloc(sizeof(hstuff) *
        (form->perfect == MINIMAL_HP ? nkeys : *smax), "perfect.c, tabh");

    /* Actually find the perfect hash */
    *salt = 0;
    bad_initkey = 0;
    bad_perfect = 0;
    for (trysalt = 1; ; ++trysalt) {
        UINT32 rslinit;
        /* Try to find distinct (A,B) for all keys */
        rslinit = initkey(keys, nkeys, *tabb, *alen, *blen, *smax, trysalt,
            form, final);

        if (rslinit == 2) {
            /* initkey actually found a perfect hash, not just distinct (a,b) */
            *salt = 1;
            *blen = 0;
            break;
        } else if (rslinit == 0) {
            /* didn't find distinct (a,b) */
            if (++bad_initkey >= RETRY_INITKEY) {
                /*
                 *  Try to put more bits in (A,B) to make distinct (A,B)
                 *  more likely.
                 */
                if (*alen < maxalen) {
                    *alen *= 2;
                } else if (*blen < *smax) {
                    *blen *= 2;
                    free(tabq);
                    free(*tabb);
                    *tabb = (bstuff *)alloc((size_t)(sizeof(bstuff) *
                        (*blen)));
                    tabq = (qstuff *)alloc((size_t)(sizeof(qstuff) *
                        (*blen + 1)));
                } else {
                    /* check for duplicates */
                    duplicates(*tabb, *blen, keys, form);
                    printf("fatal error: "
                        "Cannot perfect hash: cannot find distinct (A,B)\n");
                    exit(SUCCESS);
                }
                bad_initkey = 0;
                bad_perfect = 0;
            }
            continue;                 /* two keys have same (a,b) pair */
        }

        /* Given distinct (A,B) for all keys, build a perfect hash */
        if (!perfect(*tabb, tabh, tabq, *blen, *smax, scramble, nkeys, form)) {
            if ((form->hashtype != INT_HT && ++bad_perfect >= RETRY_PERFECT) ||
                (form->hashtype == INT_HT && ++bad_perfect >= RETRY_HEX)) {
                if (*blen < *smax) {
                    *blen *= 2;
                    free(*tabb);
                    free(tabq);
                    *tabb = (bstuff *)alloc((size_t)(sizeof(bstuff) *
                        (*blen)));
                    tabq = (qstuff *)alloc((size_t)(sizeof(qstuff) *
                        (*blen + 1)));
                    --trysalt;       /* we know this salt got distinct (A,B) */
                } else {
                    printf("fatal error: "
                        "Cannot perfect hash: cannot build tab[]\n");
                    exit(SUCCESS);
                }
                bad_perfect = 0;
            }
            continue;
        }

        *salt = trysalt;
        break;
    }

    REPORT(RP_SMLOOP, "Built perfect hash table of size %d", *blen);

    /* free working memory */
    free((void *)tabh);
    free((void *)tabq);
}



/*
 *  getkeys(keys, nkeys, textroot, keyroot, key_list) --
 */
static void getkeys(key **keys, UINT32 *nkeys, reroot *textroot,
    reroot *keyroot, LinkedList *key_list) {
    key *mykey;
    char *mytext;
    Iterator *iterator;
    void *viterator;

    mytext = (char *)renew(textroot);
    *keys = 0;
    *nkeys = 0;

    iterator = ll_iterator(key_list);
    FOREACH_ITEM(iterator, viterator, void *) {
        mykey = (key *)renew(keyroot);
        mykey->hash_k = (UINT32)viterator;
        mykey->next_k = *keys;
        *keys = mykey;
        ++*nkeys;
    }
    redel(textroot, mytext);

    free(iterator);
}



/*
 *  make_h(blen, smax, nkeys, salt) --
 *
 *  Creates phash.h file.
 */
static void make_h(UINT32 blen, UINT32 smax, UINT32 nkeys, UINT32 salt) {
    FILE *f;

    f = fopen("phash.h", "w");
    if (!f) {
        PFATAL("Could not open file");
    }

    fprintf(f, "#ifndef __PHASH_H__\n");
    fprintf(f, "#define __PHASH_H__\n");
    fprintf(f, "/*\n");
    fprintf(f, " *  --------------------------------------------------------"
        "------------------\n");
    fprintf(f, " *  By Bob Jenkins, September 1996.\n");
    fprintf(f, " *  lookupa.h, a hash function for table lookup, "
        "same function as lookup.c.\n");
    fprintf(f,  " *  Use this code in any way you wish. Public Domain. "
        "It has no warranty.\n");
    fprintf(f, " *  Source is http://burtleburtle.net/bob/c/lookupa.h\n");
    fprintf(f, " *  --------------------------------------------------------"
        "------------------\n");
    fprintf(f, " */\n");
    fprintf(f, "#include <stdio.h>\n");
    fprintf(f, "#include <stddef.h>\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "#define CHECKSTATE 8\n");
    fprintf(f, "#define hashsize(n) ((ub4)1<<(n))\n");
    fprintf(f, "#define hashmask(n) (hashsize(n)-1)\n");
    fprintf(f, "#define bis(target,mask)  ((target) |=  (mask))\n");
    fprintf(f, "#define bic(target,mask)  ((target) &= ~(mask))\n");
    fprintf(f, "#define bit(target,mask)  ((target) &   (mask))\n");
    fprintf(f, "#ifndef min\n");
    fprintf(f, "# define min(a,b) (((a)<(b)) ? (a) : (b))\n");
    fprintf(f, "#endif /* min */\n");
    fprintf(f, "#ifndef max\n");
    fprintf(f, "# define max(a,b) (((a)<(b)) ? (b) : (a))\n");
    fprintf(f, "#endif /* max */\n");
    fprintf(f, "#ifndef align\n");
    fprintf(f, "# define align(a) (((ub4)a+(sizeof(void *)-1))&"
        "(~(sizeof(void *)-1)))\n");
    fprintf(f, "#endif /* align */\n");
    fprintf(f, "#define SUCCESS 0  /* 1 on VAX */\n");
    fprintf(f, "#define UB8MAXVAL 0xffffffffffffffffLL\n");
    fprintf(f, "#define UB8BITS 64\n");
    fprintf(f, "#define SB8MAXVAL 0x7fffffffffffffffLL\n");
    fprintf(f, "#define UB4MAXVAL 0xffffffff\n");
    fprintf(f, "#define UB4BITS 32\n");
    fprintf(f, "#define SB4MAXVAL 0x7fffffff\n");
    fprintf(f, "#define UB2MAXVAL 0xffff\n");
    fprintf(f, "#define UB2BITS 16\n");
    fprintf(f, "#define SB2MAXVAL 0x7fff\n");
    fprintf(f, "#define UB1MAXVAL 0xff\n");
    fprintf(f, "#define UB1BITS 8\n");
    fprintf(f, "#define SB1MAXVAL 0x7f\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "/*\n");
    fprintf(f, " *  --------------------------------------------------------"
        "------------------\n");
    fprintf(f, " *  mix -- mix 3 32-bit values reversibly.\n");
    fprintf(f, " *  For every delta with one or two bit set,"
        " and the deltas of all three\n");
    fprintf(f, " *  high bits or all three low bits,"
        " whether the original value of a,b,c\n");
    fprintf(f, " *  is almost all zero or is uniformly distributed,\n");
    fprintf(f, " *  * If mix() is run forward or backward,"
        " at least 32 bits in a,b,c\n");
    fprintf(f, " *    have at least 1/4 probability of changing.\n");
    fprintf(f, " *  * If mix() is run forward, every bit of c will"
        " change between 1/3 and\n");
    fprintf(f, " *    2/3 of the time. (Well, 22/100 and 78/100 for"
        " some 2-bit deltas.)\n");
    fprintf(f, " *  mix() was built out of 36 single-cycle latency"
        " instructions in a\n");
    fprintf(f, " *  structure that could supported 2x parallelism,"
        " like so:\n");
    fprintf(f, " *        a -= b;\n");
    fprintf(f, " *        a -= c; x = (c >> 13);\n");
    fprintf(f, " *        b -= c; a ^= x;\n");
    fprintf(f, " *        b -= a; x = (a << 8);\n");
    fprintf(f, " *        c -= a; b ^= x;\n");
    fprintf(f, " *        c -= b; x = (b >> 13);\n");
    fprintf(f, " *        ...\n");
    fprintf(f, " *  Unfortunately, superscalar Pentiums and Sparcs"
        " can't take advantage\n");
    fprintf(f, " *  of that parallelism. They've also turned some"
        " of those single-cycle\n");
    fprintf(f, " *  latency instructions into multi-cycle latency"
        " instructions. Still,\n");
    fprintf(f, " *  this is the fastest good hash I could find."
        " There were about 2^^68\n");
    fprintf(f, " *  to choose from. I only looked at a billion or so.\n");
    fprintf(f, " *  --------------------------------------------------------"
        "------------------\n");
    fprintf(f, " */\n");
    fprintf(f, "#define mix(a,b,c) { \\\n");
    fprintf(f, "    a -= b; a -= c; a ^= (c >> 13); \\\n");
    fprintf(f, "    b -= c; b -= a; b ^= (a << 8); \\\n");
    fprintf(f, "    c -= a; c -= b; c ^= (b >> 13); \\\n");
    fprintf(f, "    a -= b; a -= c; a ^= (c >> 12); \\\n");
    fprintf(f, "    b -= c; b -= a; b ^= (a << 16); \\\n");
    fprintf(f, "    c -= a; c -= b; c ^= (b >> 5); \\\n");
    fprintf(f, "    a -= b; a -= c; a ^= (c >> 3); \\\n");
    fprintf(f, "    b -= c; b -= a; b ^= (a << 10); \\\n");
    fprintf(f, "    c -= a; c -= b; c ^= (b >> 15); \\\n");
    fprintf(f, "}\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "/*\n");
    fprintf(f, " *  --------------------------------------------------------"
        "------------------\n");
    fprintf(f, " *  mixc -- mixc 8 4-bit values as quickly and"
        " thoroughly as possible.\n");
    fprintf(f, " *  Repeating mix() three times achieves avalanche.\n");
    fprintf(f, " *  Repeating mix() four times eliminates all funnels and\n");
    fprintf(f, " *  all characteristics stronger than 2^{-11}.\n");
    fprintf(f, " *  --------------------------------------------------------"
        "------------------\n");
    fprintf(f, " */\n");
    fprintf(f, "#define mixc(a,b,c,d,e,f,g,h) { \\\n");
    fprintf(f, "    a ^= b << 11; d += a; b += c; \\\n");
    fprintf(f, "    b ^= c >> 2;  e += b; c += d; \\\n");
    fprintf(f, "    c ^= d << 8;  f += c; d += e; \\\n");
    fprintf(f, "    d ^= e >> 16; g += d; e += f; \\\n");
    fprintf(f, "    e ^= f << 10; h += e; f += g; \\\n");
    fprintf(f, "    f ^= g >> 4;  a += f; g += h; \\\n");
    fprintf(f, "    g ^= h << 8;  b += g; h += a; \\\n");
    fprintf(f, "    h ^= a >> 9;  c += h; a += b; \\\n");
    fprintf(f, "}\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "typedef  unsigned       char ub1;\n");
    fprintf(f, "typedef  unsigned short int  ub2;\n");
    fprintf(f, "typedef  unsigned long  int  ub4;\n");
    fprintf(f, "typedef  unsigned long long  ub8;\n");
    fprintf(f, "typedef    signed       char sb1;   "
        "/* signed 1-byte quantities */\n");
    fprintf(f, "typedef    signed short int  sb2;\n");
    fprintf(f, "typedef    signed long  int  sb4;\n");
    fprintf(f, "typedef    signed long long  sb8;\n");
    fprintf(f, "typedef                 int  word;  "
        "/* fastest type available */\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "/*\n");
    fprintf(f, " *  --------------------------------------------------------"
        "------------------\n");
    fprintf(f, " *  lookup() -- hash a variable-length key into"
        " a 32-bit value\n");
    fprintf(f, " *    k     : the key (the unaligned variable-length"
        " array of bytes)\n");
    fprintf(f, " *    len   : the length of the key, counting by bytes\n");
    fprintf(f, " *    level : can be any 4-byte value\n");
    fprintf(f, " *  Returns a 32-bit value. Every bit of the key"
        " affects every bit of\n");
    fprintf(f, " *  the return value. Every 1-bit and 2-bit delta"
        " achieves avalanche.\n");
    fprintf(f, " *  About 6len+35 instructions.\n");
    fprintf(f, " *\n");
    fprintf(f, " *  The best hash table sizes are powers of 2."
        " There is no need to do\n");
    fprintf(f, " *  mod a prime (mod is sooo slow!). If you need"
        " less than 32 bits,\n");
    fprintf(f, " *  use a bitmask. For example, if you need only"
        " 10 bits, do\n");
    fprintf(f, " *    h = (h & hashmask(10));\n");
    fprintf(f, " *  In which case, the hash table should have"
        " hashsize(10) elements.\n");
    fprintf(f, " *\n");
    fprintf(f, " *  If you are hashing n strings (ub1 **)k,"
        " do it like this:\n");
    fprintf(f, " *    for (i = 0, h = 0; i < n; ++i)"
        " h = lookup(k[i], len[i], h);\n");
    fprintf(f, " *\n");
    fprintf(f, " *  By Bob Jenkins, 1996.  bob_jenkins@burtleburtle.net."
        " You may use this\n");
    fprintf(f, " *  code any way you wish, private, educational,"
        " or commercial.\n");
    fprintf(f, " *\n");
    fprintf(f, " *  See http://burtleburtle.net/bob/hash/evahash.html\n");
    fprintf(f, " *  Use for hash table lookup, or anything where"
        " one collision in 2^32 is\n");
    fprintf(f, " *  acceptable.  Do NOT use for cryptographic purposes.\n");
    fprintf(f, " *  --------------------------------------------------------"
        "------------------\n");
    fprintf(f, " */\n");
    fprintf(f, "ub4 lookup(register ub1 *k, register ub4 length,"
        " register ub4 level);\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "/*\n");
    fprintf(f, " *  --------------------------------------------------------"
        "------------------\n");
    fprintf(f, " *  checksum() -- hash a variable-length key"
        " into a 256-bit value\n");
    fprintf(f, " *    k     : the key (the unaligned variable-length"
        " array of bytes)\n");
    fprintf(f, " *    len   : the length of the key, counting by bytes\n");
    fprintf(f, " *    state : an array of CHECKSTATE 4-byte values"
        " (256 bits)\n");
    fprintf(f, " *  The state is the checksum."
        " Every bit of the key affects every bit of\n");
    fprintf(f, " *  the state. There are no funnels."
        " About 112+6.875len instructions.\n");
    fprintf(f, " *\n");
    fprintf(f, " *  If you are hashing n strings (ub1 **)k,"
        " do it like this:\n");
    fprintf(f, " *    for (i = 0; i < 8; ++i) state[i] = 0x9e3779b9;\n");
    fprintf(f, " *    for (i = 0, h = 0; i < n; ++i)"
        " checksum(k[i], len[i], state);\n");
    fprintf(f, " *\n");
    fprintf(f, " *  (c) Bob Jenkins, 1996. bob_jenkins@burtleburtle.net."
        " You may use this\n");
    fprintf(f, " *  code any way you wish, private, educational,"
        " or commercial, as long\n");
    fprintf(f, " *  as this whole comment accompanies it.\n");
    fprintf(f, " *\n");
    fprintf(f, " *  See http://burtleburtle.net/bob/hash/evahash.html\n");
    fprintf(f, " *  Use to detect changes between revisions of"
        " documents, assuming nobody\n");
    fprintf(f, " *  is trying to cause collisions."
        " Do NOT use for cryptography.\n");
    fprintf(f, " *  --------------------------------------------------------"
        "------------------\n");
    fprintf(f, " */\n");
    fprintf(f, "void checksum(register ub1 *k, register ub4 len,"
        " register ub4 *state);\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "\n");

    if (blen > 0) {
        if (blen <= UB1MAXVAL + 1 || blen >= USE_SCRAMBLE) {
            fprintf(f, "extern ub1 tab[];\n");
        } else {
            fprintf(f, "extern ub2 tab[];\n");
            fprintf(f, "\n");
            fprintf(f, "\n");
            fprintf(f, "\n");
            if (smax <= UB2MAXVAL + 1) {
                fprintf(f, "extern ub2 scramble[];\n");
            } else {
                fprintf(f, "extern ub4 scramble[];\n");
            }
        }
        fprintf(f, "\n");
        fprintf(f, "\n");
        fprintf(f, "\n");
        fprintf(f, "#define PHASHLEN 0x%x  "
            "/* length of hash mapping table */\n", blen);
    }

    fprintf(f, "#define PHASHNKEYS %d  "
        "/* How many keys were hashed */\n", nkeys);
    fprintf(f, "#define PHASHRANGE %d  "
        "/* Range any input might map to */\n", smax);
    fprintf(f, "#define PHASHSALT 0x%.8x  "
        "/* internal, initialize normal hash */\n", salt * 0x9e3779b9);
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "\n");
    fprintf(f, "ub4 phash();\n");
    fprintf(f, "#endif\n");
    fprintf(f, "\n");

    fclose(f);
}



/*
 *  make_c(tab, smax, blen, scramble, final, form) --
 *
 *  Creates phash.c file.
 */
static void make_c(bstuff *tab, UINT32 smax, UINT32 blen, UINT32 *scramble,
    gencode *final, hashform *form) {
    UINT32 i;
    FILE *f;

    f = fopen("phash.c", "w");
    if (!f) {
        PFATAL("Could not open file");
    }

    fprintf(f, "#include \"phash.h\"\n");
    fprintf(f, "\n");

    if (blen >= USE_SCRAMBLE) {
        fprintf(f, "/* A way to make the 1-byte values in tab bigger */\n");
        if (smax > UB2MAXVAL + 1) {
            fprintf(f, "ub4 scramble[] = {\n");
            for (i = 0; i <= UB1MAXVAL; i += 4) {
                fprintf(f, "0x%.8x, 0x%.8x, 0x%.8x, 0x%.8x,\n",
                    scramble[i + 0], scramble[i + 1], scramble[i + 2],
                    scramble[i + 3]);
            }
        } else {
            fprintf(f, "ub2 scramble[] = {\n");
            for (i = 0; i <= UB1MAXVAL; i += 8) {
                fprintf(f, "0x%.4x, 0x%.4x, 0x%.4x, 0x%.4x, 0x%.4x, 0x%.4x,"
                    " 0x%.4x, 0x%.4x,\n", scramble[i + 0], scramble[i + 1],
                    scramble[i + 2], scramble[i + 3], scramble[i + 4],
                    scramble[i + 5], scramble[i + 6], scramble[i + 7]);
            }
        }
        fprintf(f, "};\n");
        fprintf(f, "\n");
    }

    if (blen > 0) {
        fprintf(f, "/* small adjustments to _a_ to make values distinct */\n");

        if (smax <= UB1MAXVAL + 1 || blen >= USE_SCRAMBLE) {
            fprintf(f, "ub1 tab[] = {\n");
        } else {
            fprintf(f, "ub2 tab[] = {\n");
        }

        if (blen < 16) {
            for (i = 0; i < blen; ++i) {
                fprintf(f, "%3d, ", scramble[tab[i].val_b]);
            }
        } else if (blen <= 1024) {
            for (i = 0; i < blen; i += 16) {
                fprintf(f, "%d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d,"
                    " %d, %d, %d, %d,\n",
                    scramble[tab[i + 0].val_b], scramble[tab[i + 1].val_b],
                    scramble[tab[i + 2].val_b], scramble[tab[i + 3].val_b],
                    scramble[tab[i + 4].val_b], scramble[tab[i + 5].val_b],
                    scramble[tab[i + 6].val_b], scramble[tab[i + 7].val_b],
                    scramble[tab[i + 8].val_b], scramble[tab[i + 9].val_b],
                    scramble[tab[i + 10].val_b], scramble[tab[i + 11].val_b],
                    scramble[tab[i + 12].val_b], scramble[tab[i + 13].val_b],
                    scramble[tab[i + 14].val_b], scramble[tab[i + 15].val_b]);
            }
        } else if (blen < USE_SCRAMBLE) {
            for (i = 0; i < blen; i += 8) {
                fprintf(f, "%d, %d, %d, %d, %d, %d, %d, %d,\n",
                    scramble[tab[i + 0].val_b], scramble[tab[i + 1].val_b],
                    scramble[tab[i + 2].val_b], scramble[tab[i + 3].val_b],
                    scramble[tab[i + 4].val_b], scramble[tab[i + 5].val_b],
                    scramble[tab[i + 6].val_b], scramble[tab[i + 7].val_b]);
            }
        } else {
            for (i = 0; i < blen; i += 16) {
                fprintf(f, "%d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d,"
                    " %d, %d, %d, %d,\n",
                    tab[i + 0].val_b, tab[i + 1].val_b,
                    tab[i + 2].val_b, tab[i + 3].val_b,
                    tab[i + 4].val_b, tab[i + 5].val_b,
                    tab[i + 6].val_b, tab[i + 7].val_b,
                    tab[i + 8].val_b, tab[i + 9].val_b,
                    tab[i + 10].val_b, tab[i + 11].val_b,
                    tab[i + 12].val_b, tab[i + 13].val_b,
                    tab[i + 14].val_b, tab[i + 15].val_b);
            }
        }
        fprintf(f, "};\n");
        fprintf(f, "\n");
    }

    fprintf(f, "/* The hash function */\n");
    fprintf(f, "ub4 phash(ub4 val) {\n");
    for (i = 0; i < final->used; ++i) {
        fprintf(f, final->line[i]);
    }
    fprintf(f, "    return rsl;\n");
    fprintf(f, "}\n");
    fprintf(f, "\n");

    fclose(f);
}



/*
 *  make_main() --
 *
 *  Creates phash_main.c file.
 */
void make_main() {
    FILE *f;

    f = fopen("phash_main.c", "w");
    if (!f) {
        PFATAL("Could not open file");
    }

    fprintf(f, "#include <stdlib.h>\n");
    fprintf(f, "#include <stdio.h>\n");
    fprintf(f, "#include \"phash.h\"\n");
    fprintf(f, "\n");
    fprintf(f, "ub4 lookup(register ub1 *k, register ub4 length, "
        "register ub4 level) {\n");
    fprintf(f, "    register ub4 a, b, c, len;\n");
    fprintf(f, "\n");
    fprintf(f, "    /* Set up the internal state */\n");
    fprintf(f, "    len = length;\n");
    fprintf(f, "    a = b = 0x9e3779b9; "
        "/* the golden ratio; an arbitrary value */\n");
    fprintf(f, "    c = level;          /* the previous hash value */\n");
    fprintf(f, "\n");
    fprintf(f, "    /*-------------------------- handle most of the key */\n");
    fprintf(f, "    while (len >= 12) {\n");
    fprintf(f, "        a += (k[0] + ((ub4)k[1] << 8) + ((ub4)k[2] << 16) +"
        " ((ub4)k[3] << 24));\n");
    fprintf(f, "        b += (k[4] + ((ub4)k[5] << 8) + ((ub4)k[6] << 16) +"
        " ((ub4)k[7] << 24));\n");
    fprintf(f, "        c += (k[8] + ((ub4)k[9] << 8) +\n");
    fprintf(f, "            ((ub4)k[10] << 16) + ((ub4)k[11] << 24));\n");
    fprintf(f, "        mix(a, b, c);\n");
    fprintf(f, "        k += 12;\n");
    fprintf(f, "        len -= 12;\n");
    fprintf(f, "    }\n");
    fprintf(f, "\n");
    fprintf(f, "    /*------------------------ handle the last 11 bytes */\n");
    fprintf(f, "    c += length;\n");
    fprintf(f, "    switch (len) {   /* all the case stmts fall through */\n");
    fprintf(f, "        case 11: c += ((ub4)k[10] << 24);\n");
    fprintf(f, "        case 10: c += ((ub4)k[9] << 16);\n");
    fprintf(f, "        case 9 : c += ((ub4)k[8] << 8);\n");
    fprintf(f, "        /* first byte of c is reserved for the length */\n");
    fprintf(f, "        case 8 : b += ((ub4)k[7] << 24);\n");
    fprintf(f, "        case 7 : b += ((ub4)k[6] << 16);\n");
    fprintf(f, "        case 6 : b += ((ub4)k[5] << 8);\n");
    fprintf(f, "        case 5 : b += k[4];\n");
    fprintf(f, "        case 4 : a += ((ub4)k[3] << 24);\n");
    fprintf(f, "        case 3 : a += ((ub4)k[2] << 16);\n");
    fprintf(f, "        case 2 : a += ((ub4)k[1] << 8);\n");
    fprintf(f, "        case 1 : a += k[0];\n");
    fprintf(f, "        /* case 0: nothing left to add */\n");
    fprintf(f, "    }\n");
    fprintf(f, "    mix(a, b, c);\n");
    fprintf(f, "    /*------------------------------- report the result */\n");
    fprintf(f, "    return c;\n");
    fprintf(f, "}\n");
    fprintf(f, "\n");
    fprintf(f, "void checksum(register ub1 *k, register ub4 len, "
        "register ub4 *state) {\n");
    fprintf(f, "    register ub4 a, b, c, d, e, f, g, h, length;\n");
    fprintf(f, "\n");
    fprintf(f, "    /* Use length and level; add in the golden ratio. */\n");
    fprintf(f, "    length = len;\n");
    fprintf(f, "    a = state[0];\n");
    fprintf(f, "    b = state[1];\n");
    fprintf(f, "    c = state[2];\n");
    fprintf(f, "    d = state[3];\n");
    fprintf(f, "    e = state[4];\n");
    fprintf(f, "    f = state[5];\n");
    fprintf(f, "    g = state[6];\n");
    fprintf(f, "    h = state[7];\n");
    fprintf(f, "\n");
    fprintf(f, "    /*-------------------------- handle most of the key */\n");
    fprintf(f, "    while (len >= 32) {\n");
    fprintf(f, "        a += (k[0] + (k[1] << 8) + (k[2] << 16) +"
        " (k[3] << 24));\n");
    fprintf(f, "        b += (k[4] + (k[5] << 8) + (k[6] << 16) +"
        " (k[7] << 24));\n");
    fprintf(f, "        c += (k[8] + (k[9] << 8) + (k[10] << 16) +"
        " (k[11] << 24));\n");
    fprintf(f, "        d += (k[12] + (k[13] << 8) + (k[14] << 16) +"
        " (k[15] << 24));\n");
    fprintf(f, "        e += (k[16] + (k[17] << 8) + (k[18] << 16) +"
        " (k[19] << 24));\n");
    fprintf(f, "        f += (k[20] + (k[21] << 8) + (k[22] << 16) +"
        " (k[23] << 24));\n");
    fprintf(f, "        g += (k[24] + (k[25] << 8) + (k[26] << 16) +"
        " (k[27] << 24));\n");
    fprintf(f, "        h += (k[28] + (k[29] << 8) + (k[30] << 16) +"
        " (k[31] << 24));\n");
    fprintf(f, "        mixc(a, b, c, d, e, f, g, h);\n");
    fprintf(f, "        mixc(a, b, c, d, e, f, g, h);\n");
    fprintf(f, "        mixc(a, b, c, d, e, f, g, h);\n");
    fprintf(f, "        mixc(a, b, c, d, e, f, g, h);\n");
    fprintf(f, "        k += 32;\n");
    fprintf(f, "        len -= 32;\n");
    fprintf(f, "    }\n");
    fprintf(f, "\n");
    fprintf(f, "    /*------------------------ handle the last 31 bytes */\n");
    fprintf(f, "    h += length;\n");
    fprintf(f, "    switch (len) {\n");
    fprintf(f, "        case 31: h += (k[30] << 24);\n");
    fprintf(f, "        case 30: h += (k[29] << 16);\n");
    fprintf(f, "        case 29: h += (k[28] << 8);\n");
    fprintf(f, "        case 28: g += (k[27] << 24);\n");
    fprintf(f, "        case 27: g += (k[26] << 16);\n");
    fprintf(f, "        case 26: g += (k[25] << 8);\n");
    fprintf(f, "        case 25: g += k[24];\n");
    fprintf(f, "        case 24: f += (k[23] << 24);\n");
    fprintf(f, "        case 23: f += (k[22] << 16);\n");
    fprintf(f, "        case 22: f += (k[21] << 8);\n");
    fprintf(f, "        case 21: f += k[20];\n");
    fprintf(f, "        case 20: e += (k[19] << 24);\n");
    fprintf(f, "        case 19: e += (k[18] << 16);\n");
    fprintf(f, "        case 18: e += (k[17] << 8);\n");
    fprintf(f, "        case 17: e += k[16];\n");
    fprintf(f, "        case 16: d += (k[15] << 24);\n");
    fprintf(f, "        case 15: d += (k[14] << 16);\n");
    fprintf(f, "        case 14: d += (k[13] << 8);\n");
    fprintf(f, "        case 13: d += k[12];\n");
    fprintf(f, "        case 12: c += (k[11] << 24);\n");
    fprintf(f, "        case 11: c += (k[10] << 16);\n");
    fprintf(f, "        case 10: c += (k[9] << 8);\n");
    fprintf(f, "        case 9 : c += k[8];\n");
    fprintf(f, "        case 8 : b += (k[7] << 24);\n");
    fprintf(f, "        case 7 : b += (k[6] << 16);\n");
    fprintf(f, "        case 6 : b += (k[5] << 8);\n");
    fprintf(f, "        case 5 : b += k[4];\n");
    fprintf(f, "        case 4 : a += (k[3] << 24);\n");
    fprintf(f, "        case 3 : a += (k[2] << 16);\n");
    fprintf(f, "        case 2 : a += (k[1] << 8);\n");
    fprintf(f, "        case 1 : a += k[0];\n");
    fprintf(f, "    }\n");
    fprintf(f, "    mixc(a, b, c, d, e, f, g, h);\n");
    fprintf(f, "    mixc(a, b, c, d, e, f, g, h);\n");
    fprintf(f, "    mixc(a, b, c, d, e, f, g, h);\n");
    fprintf(f, "    mixc(a, b, c, d, e, f, g, h);\n");
    fprintf(f, "\n");
    fprintf(f, "    /*------------------------------- report the result */\n");
    fprintf(f, "    state[0] = a;\n");
    fprintf(f, "    state[1] = b;\n");
    fprintf(f, "    state[2] = c;\n");
    fprintf(f, "    state[3] = d;\n");
    fprintf(f, "    state[4] = e;\n");
    fprintf(f, "    state[5] = f;\n");
    fprintf(f, "    state[6] = g;\n");
    fprintf(f, "    state[7] = h;\n");
    fprintf(f, "}\n");
    fprintf(f, "\n");
    fprintf(f, "int main(int argc, char **argv) {\n");
    fprintf(f, "    unsigned int addr;\n");
    fprintf(f, "    int hash_val;\n");
    fprintf(f, "    FILE *f1,\n");
    fprintf(f, "         *f2;\n");
    fprintf(f, "\n");
    fprintf(f, "    f1 = fopen(argv[1], \"r\");\n");
    fprintf(f, "    if (!f1) {\n");
    fprintf(f, "        perror(\"could not open f1\");\n");
    fprintf(f, "        exit(-1);\n");
    fprintf(f, "    }\n");
    fprintf(f, "    f2 = fopen(argv[2], \"w\");\n");
    fprintf(f, "    if (!f2) {\n");
    fprintf(f, "        perror(\"could not open f2\");\n");
    fprintf(f, "        exit(-1);\n");
    fprintf(f, "    }\n");
    fprintf(f, "\n");
    fprintf(f, "    while (fscanf(f1, \"%%x\", &addr) > 0) {\n");
    fprintf(f, "        fprintf(f2, \"%%.7x %%d\\n\", addr, phash(addr));\n");
    fprintf(f, "    }\n");
    fprintf(f, "\n");
    fprintf(f, "    fclose(f1);\n");
    fprintf(f, "    fclose(f2);\n");
    fprintf(f, "\n");
    fprintf(f, "    return 0;\n");
    fprintf(f, "}\n");
    fprintf(f, "\n");

    fclose(f);
}



/*
 *  create_hash(key_list) --
 *
 *  Creates the perfect hashing function (phash.{c,h}).
 */
void create_hash(LinkedList *key_list) {
    hashform __form;
    hashform *form;
    UINT32 nkeys;            /* number of keys */
    key *keys;               /* head of list of keys */
    bstuff *tab;             /* table indexed by b */
    UINT32 smax;             /* scramble[] values in 0..smax-1, a power of 2 */
    UINT32 alen;             /* a in 0..alen-1, a power of 2 */
    UINT32 blen;             /* b in 0..blen-1, a power of 2 */
    UINT32 salt;             /* a parameter to the hash function */
    reroot *textroot;        /* MAXKEYLEN-character text lines */
    reroot *keyroot;         /* source of keys */
    gencode final;           /* code for final hash */
    UINT32 i;
    UINT32 scramble[SCRAMBLE_LEN];            /* used in final hash function */
    char buf[10][80];        /* buffer for generated code */
    char *buf2[10];          /* also for generated code */

    form = &__form;
    form->speed = SLOW_HS;
    form->mode = HEX_HM;
    form->hashtype = INT_HT;
    form->perfect = MINIMAL_HP;

    textroot = remkroot((size_t)MAXKEYLEN);
    keyroot = remkroot(sizeof(key));

    final.line = buf2;
    final.used = 0;
    final.len = 10;
    for (i = 0; i < 10; ++i) {
        final.line[i] = buf[i];
    }

    getkeys(&keys, &nkeys, textroot, keyroot, key_list);

    findhash(&tab, &alen, &blen, &salt, &final, scramble, &smax,
        keys, nkeys, form);

    make_h(blen, smax, nkeys, salt);
    make_c(tab, smax, blen, scramble, &final, form);
    make_main();

    refree(textroot);
    refree(keyroot);
    free((void *)tab);
}

