/*
 *  recycle.c
 *
 *  This file was taken from the source cited below. It was then modified to
 *  use in binary obfuscator to create a perfect hash function for two types
 *  of obfuscation:
 *      - signal based obfuscation
 *      - changing many jumps/calls to call to one branch function
 *
 *  --------------------------------------------------------------------------
 *  By Bob Jenkins, September 1996.  recycle.c
 *  You may use this code in any way you wish, and it is free.  No warranty.
 *
 *  This manages memory for commonly-allocated structures.
 *  It allocates RESTART to REMAX items at a time.
 *  Timings have shown that, if malloc is used for every new structure,
 *    malloc will consume about 90% of the time in a program.  This
 *    module cuts down the number of mallocs by an order of magnitude.
 *  This also decreases memory fragmentation, and freeing structures
 *    only requires freeing the root.
 *  --------------------------------------------------------------------------
 */
#include "standard.h"
#include "recycle.h"
#include "global.h"



/*
 *  remkroot(size) --
 *
 *  Makes a new recycling root.
 */
reroot *remkroot(size_t size) {
    reroot *r = (reroot *)remalloc(sizeof(reroot), "recycle.c, root");
    r->list = (recycle *)0;
    r->trash = (recycle *)0;
    r->size = align(size);
    r->logsize = RESTART;
    r->numleft = 0;
    return r;
}



/*
 *  refree(r) --
 *
 *  Frees a recycling root and all the items it has made.
 */
void refree(struct reroot *r) {
    recycle *temp;
    if ((temp = (r->list))) {
        while (r->list) {
            temp = r->list->next;
            free((char *)r->list);
            r->list = temp;
        }
    }
    free((char *)r);
    return;
}



/*
 *  renewx(r) --
 *
 *  To be called from the macro renew only.
 */
char *renewx(struct reroot *r) {
    recycle *temp;
    if (r->trash) {  /* pull a node off the trash heap */
        temp = r->trash;
        r->trash = temp->next;
        memset((void *)temp, 0, r->size);
    } else {          /* allocate a new block of nodes */
        r->numleft = r->size * ((UINT32)1 << r->logsize);
        if (r->numleft < REMAX) {
            ++r->logsize;
        }
        temp = (recycle *)remalloc(sizeof(recycle) + r->numleft,
            "recycle.c, data");
        temp->next = r->list;
        r->list = temp;
        r->numleft -= r->size;
        temp = (recycle *)((char *)(r->list + 1) + r->numleft);
    }
    return (char *)temp;
}



/*
 *  remalloc(len, purpose) --
 *
 *  Uses malloc, but complains to stderr and exits program if no joy.
 *  Use plain free() to free memory allocated by remalloc().
 */
char *remalloc(size_t len, char *purpose) {
    char *x = (char *)alloc(len);
    if (!x) {
        fprintf(stderr, "malloc of %d failed for %s\n", len, purpose);
        exit(SUCCESS);
    }
    return x;
}

