//!!!Have to put a note regarding InstrFree() here and at every place
//!!!where it was supposed to be used.
/*
 *  signals.c
 *  
 *  This file contains functions related to the signal based obfuscation. The
 *  purpose of this type of obfuscation is to replace a control transfer
 *  instruction (jump, call, return) with the sequence of instructions that
 *  would raise a signal and cause a trap into the kernel when executed. The
 *  kernel would then give control to the handler installed for the particular
 *  signal. Normally, after the handler returns, execution goes to the kernel's
 *  restore function, and then back to the location at which the trap occured.
 *  The purpose of our handler would be to overwrite the trap address with the
 *  address of our restore function - my_restore(), so that kernel's restore
 *  returns to my_restore(). my_restore() will derive the target of the
 *  original control transfer instruction using the trap address and the
 *  perfect hash function (see perfect.c for details) and will return to it.
 *  Only the first instruction of my_restore() - nop is created here. The
 *  function is completed later, after the final addresses of instructions
 *  are known (see finalize.c for details).
 */
#include "obf_global.h"
#include "signals.h"
#include "obf.h"
#include "obf_macros.h"
#include "obf_print.h"
#include "utilities.h"
#include "linkage.h"
#include "flowgraph.h"
#include "sysdep.h"
#include "x86-registers.h"
#include "upm.h"
#include "relocate.h"
#include "liveness.h"
#include "x86-liveness.h"



static void ModifyOriginalHandlersInstallation();
static Function *Insert__bsd_signal();

static void InsertMyRestore();
static void InsertShortMyRestore();
static void InsertLongMyRestore();
static void InsertMyRestore1(int num_int);
static void InsertShortMyRestore2(int num_int);
static void InsertLongMyRestore2(int num_int);
static void InsertLongMyRestore3(int num_int);
static void MyRestoreAddEdges(int bbl_num_from, int num_copies_from,
    int bbl_num_to, int num_copies_to, int edge_type);

static void InstallMySignalHandler();
static void InsertMyHandler(Function *signal_function);
static void InsertMyHandler1(int num_int);
static void InsertMyHandler2(int num_int, Function *signal_function);
static void InsertMyHandler3(int num_int, Function *signal_function);
static void InsertMyHandler4(int num_int, Function *signal_function);
static void MyHandlerAddEdges(int bbl_num_from, int num_copies_from,
    int bbl_num_to, int num_copies_to, int edge_type);

static Block *InstallHandler(Function *signal_function,
    Block *handler_first_bbl, Block *prev_bbl, Block *next_bbl,
    int signal_number);

static void ConvertControlTransfersToTraps(int highest_bbl_id);
static void JumpToTrap1Reg(Instruction *jump, UINT32 reg1);
static void JumpToTrap2Regs(Instruction *jump, UINT32 reg1, UINT32 reg2);
static void JumpToTrap3Regs(Instruction *jump, UINT32 reg1, UINT32 reg2,
    UINT32 reg3);
static void JumpToTrap4Regs(Instruction *jump, UINT32 reg1, UINT32 reg2,
    UINT32 reg3, UINT32 reg4);

static void CallToTrap1Reg(Instruction *call, UINT32 reg1);
static void CallToTrap2Regs(Instruction *call, UINT32 reg1, UINT32 reg2);
static void CallToTrap3Regs(Instruction *call, UINT32 reg1, UINT32 reg2,
    UINT32 reg3);
static void CallToTrap4Regs(Instruction *call, UINT32 reg1, UINT32 reg2,
    UINT32 reg3, UINT32 reg4);

static void JumpToSIGILL(Instruction *jump);
//static void JumpToSIGFPE(Instruction *jump);
//static void JumpToSIGSEGV(Instruction *jump);
static void CallToSIGILL(Instruction *call);
//static void CallToSIGFPE(Instruction *call);
//static void CallToSIGSEGV(Instruction *call);
static void RetToSIGILL(Instruction *ret);
//static void RetToSIGFPE(Instruction *ret);
//static void RetToSIGSEGV(Instruction *ret);

static void InsertInstrDecStack(int num_bytes, Instruction *next_instr);
static void InsertInstrPush(Instruction *next_instr);
static void InsertInstrSetCommunicationFlag(Instruction *next_instr);
static void InsertUseZeroRegInstr(UINT32 zero_reg, UINT32 dest_reg,
    Instruction *next_instr);
static void InsertIllegalInstr(Instruction *next_instr);
static UINT8 PickUndefinedInstrByte2();



/*
 * The offset to the base of the table containing the pointers to the
 * currently installed signal handlers.
 */
static ADDRESS sh_table_offset;



/*
 * The offset to the beginning of the array of global variables telling where
 * the signal is raised from. One randomly chosen variable is set to some
 * random non-zero value right before rasing the signal from my location. It is
 * reset to zero inside the handler. When handler is called, it checks all
 * the above variables until one has a non-zero value. If such found, the
 * handler ensures that control is eventualy (using my_restore) transferred
 * to the original target. Otherwise, the signal was raised from some other
 * location and will be handled the default (or user installed) way.
 */
static ADDRESS raised_from;



/*
 * Each entry with index n in this array contains a linked list of pointers to
 * the copies of nth basic block of my handler function.
 */
static LinkedList *my_handler_bbls_lists[14];



/*
 *  InstallTraps() --
 *
 *  This is a driver for the signal based obfuscation. The process is split
 *  into three parts. First, we need to tap into the process of installing
 *  user defined handlers (if any) to know which ones he/she wanted to install.
 *  We need this to ensure that the signals for which we want to install
 *  handlers do not have handlers installed by user. Second, we need to install
 *  our handlers for the signals that will be raised by our traps (currently
 *  SIGILL, SIGFPE and SIGSEGV). In this implementation one handler handles
 *  all three types of signals. Third, we convert the control transfer
 *  instructions (jumps, calls, returns) at candidate locations to the
 *  sequences of instructions that would raise one of the three signals
 *  described above.
 */
void InstallTraps(int highest_bbl_id) {

    REPORT(RP_DETAIL, "Installing a wrapper to signal() function");
    ModifyOriginalHandlersInstallation();

    REPORT(RP_DETAIL, "Inserting my_restore() function");
    InsertMyRestore();

    REPORT(RP_DETAIL, "Installing my signal handler");
    InstallMySignalHandler();

    PrintControlTransferEdges();
    
    REPORT(RP_DETAIL, "Converting control transfers to traps");
    ConvertControlTransfersToTraps(highest_bbl_id);

    return;
}



/*
 *  ModifyOriginalHandlersInstallation() --
 *
 *  Creates and inserts a wrapper to a function signal(<sig_num>, <sig_handler>)
 *  from signal.h library. It only intercepts calls to the above function, does
 *  not intercept returns from it. If a user tries to install a handler for the
 *  signal, for which we also want to install a handler, it simply stores the
 *  pointer to the user-defined handler instead of actually installing it. The
 *  only handler installed in this case is our handler. If a user tries to
 *  install a handler for another signal, execution proceeds to signal() or
 *  __bsd_signal().
 */
static void ModifyOriginalHandlersInstallation() {
    Function *fiterator;
    Function *signal_wrapper;
    Function *signal_function;
    Block *bsd_signal_first_bbl;
    Block *bsd_signal_entry_bbl;
    Block *bsd_signal_exit_bbl;
    Block *my1_bbl;
    Block *my2_bbl;
    Block *my3_bbl;
    Block *my4_bbl;
    Block *my5_bbl;
    Block *my6_bbl;
    Block *caller_bbl;
    Instruction *new_instr;
    EdgeNode *enode;
    LinkedList *callers;

    /*
     *  Allocate space for the table of pointers to signal handlers in the
     *  .data section
     */
    sh_table_offset = SECTION_REF_DATA->s_newdatasize;
    IncreaseSectionSize(SECTION_REF_DATA, MAX_SIGNAL_NUMBER * sizeof(ADDRESS));

    /* Locate signal() or __bsd_signal() function. If not found, insert it. */
    FOREACH_FUNCTION(fiterator) {
        if ((strcmp(fiterator->f_name, "__bsd_signal") == 0)
            || (strcmp(fiterator->f_name, "signal") == 0)) {
            break;
        }
    }
    if (fiterator == NULL) {
        REPORT(RP_SMLOOP, "Neither __bsd_signal() nor signal() found.");
        signal_function = Insert__bsd_signal();
        FUNCTION_SET_FLAG(signal_function, FUNCTION_DO_NOT_OBFUSCATE_SIGNAL);
    } else {
        REPORT(RP_SMLOOP, "Found %s()", fiterator->f_name);
        signal_function = fiterator;
    }
    bsd_signal_first_bbl = signal_function->f_first;
    bsd_signal_entry_bbl = signal_function->f_entry;
    bsd_signal_exit_bbl = signal_function->f_exit;

    /*
     * Have all the links now, inject the code before the __bsd_signal(), or,
     * if it was not found, before our substitution for it - my__bsd_signal().
     */
    my1_bbl = BBLCreate();
    my1_bbl->b_type = BT_BRANCH;
    my2_bbl = BBLCreate();
    my2_bbl->b_type = BT_BRANCH;
    my3_bbl = BBLCreate();
    my3_bbl->b_type = BT_BRANCH;
    my4_bbl = BBLCreate();
    my4_bbl->b_type = BT_CALL;
    my5_bbl = BBLCreate();
    my5_bbl->b_type = BT_NORMAL;
    my6_bbl = BBLCreate();
    my6_bbl->b_type = BT_RETURN;

    signal_wrapper = FunCreate();
    signal_wrapper->f_name = "my_signal";
    signal_wrapper->f_first = my1_bbl;
    signal_wrapper->f_last = my6_bbl;
    if (BooleanParameter("LabelBasicBlocks")) {
        AddNewFunctionSymbol(signal_wrapper);
    }
    InsertFunctionAfterFunction(signal_wrapper, signal_function->f_prev);
    FUNCTION_SET_FLAG(signal_wrapper, FUNCTION_DO_NOT_OBFUSCATE_SIGNAL);

    BBLAddEdge(signal_wrapper->f_entry, my1_bbl, ET_ENTRY);
    BBLAddEdge(my1_bbl, my2_bbl, ET_FALSE);
    BBLAddEdge(my1_bbl, my6_bbl, ET_TRUE);
    BBLAddEdge(my2_bbl, my3_bbl, ET_FALSE);
    BBLAddEdge(my2_bbl, my6_bbl, ET_TRUE);
    BBLAddEdge(my3_bbl, my4_bbl, ET_FALSE);
    BBLAddEdge(my3_bbl, my6_bbl, ET_TRUE);
    BBLAddEdge(my4_bbl, my5_bbl, ET_FUNLINK);
    BBLAddEdge(my5_bbl, my6_bbl, ET_NORMAL);
    BBLAddEdge(my6_bbl, signal_wrapper->f_exit, ET_EXIT);

    //BBL1 (BT_BRANCH)
    InsertBBLBeforeBBL(my1_bbl, bsd_signal_first_bbl);
    my1_bbl->b_fun = signal_wrapper->f_entry->b_fun;

    //push %ebp
    new_instr = InstrCreatePush(reg_bp | reg_32);
    BBLPrependInstr(my1_bbl, new_instr);

    //mov %esp, %ebp
    new_instr = InstrCreateMovRegToReg(reg_sp | reg_32, reg_bp | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //pushf
    new_instr = InstrCreateSave();
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //push %ebx
    new_instr = InstrCreatePush(reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //push %ecx
    new_instr = InstrCreatePush(reg_c | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //push %edx
    new_instr = InstrCreatePush(reg_d | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //move $<sh_table_offset>, %edx
    new_instr = InstrCreateMovImmToReg(reg_d | reg_32, 123);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);
    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, sh_table_offset,
        new_instr, new_instr->i_src);

    //mov 0x8(%ebp), %ecx
    new_instr = InstrCreateMovMemDisp8BaseToReg(0x8, reg_bp | reg_32,
        reg_c | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //mov 0xb(%ebp), %ebx
    new_instr = InstrCreateMovMemDisp8BaseToReg(0xb, reg_bp | reg_32,
        reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);


    //mov (%edx, %ecx, 4), %eax //save pointer to old handler in return register
    new_instr = InstrCreateMovMemBaseIndexToReg(reg_d | reg_32, reg_c | reg_32,
        4, reg_a | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //mov %ebx, (%edx, %ecx, 4)  //store pointer to new handler
    new_instr = InstrCreateMovRegToMemBaseIndex(reg_b | reg_32, reg_d | reg_32,
        reg_c | reg_32, 4);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //cmp %ecx, $0x4
    new_instr = InstrCreateCmpImm8WithReg(0x4, reg_c | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //je <&BBL6>
    new_instr = InstrCreateJe();
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL2 (BT_BRANCH)
    InsertBBLAfterBBL(my2_bbl, my1_bbl);

    //cmp %ecx, $0x8
    new_instr = InstrCreateCmpImm8WithReg(0x8, reg_c | reg_32);
    BBLPrependInstr(my2_bbl, new_instr);

    //je <&BBL6>
    new_instr = InstrCreateJe();
    InsertInstrAfterInstr(new_instr, my2_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL3 (BT_BRANCH)
    InsertBBLAfterBBL(my3_bbl, my2_bbl);

    //cmp %ecx, $0xb
    new_instr = InstrCreateCmpImm8WithReg(0xb, reg_c | reg_32);
    BBLPrependInstr(my3_bbl, new_instr);

    //je <&BBL6>
    new_instr = InstrCreateJe();
    InsertInstrAfterInstr(new_instr, my3_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //When extending to more signals for which we install handlers, repeat
    //the above two instructions for every new signal. They will also go into
    //a new BBL (see BBL2 and BBL3)

    //BBL4 (BT_CALL)
    InsertBBLAfterBBL(my4_bbl, my3_bbl);

    //sub $0x8, %esp
    new_instr = InstrCreateSubImm8FromReg(0x8, reg_sp | reg_32);
    BBLPrependInstr(my4_bbl, new_instr);

    //push %ebx
    new_instr = InstrCreatePush(reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my4_bbl->b_last);

    //push %ecx
    new_instr = InstrCreatePush(reg_c | reg_32);
    InsertInstrAfterInstr(new_instr, my4_bbl->b_last);

    //call __bsd_signal
    new_instr = InstrCreateCALLDisp();
    InsertInstrAfterInstr(new_instr, my4_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_CALL);

    //BBL5 (BT_NORMAL)
    InsertBBLAfterBBL(my5_bbl, my4_bbl);

    //add $0x10, %esp
    new_instr = InstrCreateAddImm8ToReg(0x10, reg_sp | reg_32);
    BBLPrependInstr(my5_bbl, new_instr);

    //BBL6 (BT_RETURN)
    InsertBBLAfterBBL(my6_bbl, my5_bbl);

    //pop %edx
    new_instr = InstrCreatePop(reg_d | reg_32);
    BBLPrependInstr(my6_bbl, new_instr);

    //pop %ecx
    new_instr = InstrCreatePop(reg_c | reg_32);
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);

    //pop %ebx
    new_instr = InstrCreatePop(reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);

    //popf
    new_instr = InstrCreateRestore();
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);

    //leave
    new_instr = InstrCreateLeave();
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);

    //ret
    new_instr = InstrCreateReturn();
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_RETURN);

    //if we injected the my__bsd_signal(),
    //    return;

    /* rerout all callers to call the wrapper we've just created */
    callers = ll_new(LIST_UNSORTED);
    FOREACH_PRED_IN_BBL(bsd_signal_first_bbl, enode) {
        if (NODE_TYPE(enode) == ET_HELL || NODE_TYPE(enode) == ET_HELLMAYBE) {
            //|| NODE_TYPE(enode) == ET_ENTRY) {
            //TO DO: We'll have to stop here, assuming that any handler could
            //be installed. Return -1 meaning that we should give up at this
            //point. In specint2000 no program had __bsd_signal called from
            //HELL, so this was not expanded further.
            WARNING("HELLMAYBE edge to %s. Cannot proceed with signal based "
                "obfuscation", signal_function->f_name);
        } else if (NODE_TYPE(enode) == ET_FUNDIRECT) {
            caller_bbl = NODE_FROM(enode);
            ll_add(callers, (void *)(caller_bbl));
        } else if (NODE_TYPE(enode) == ET_UNCOND) {
            //need to deal with this, too, if it happens
            //01/26/06 - it does happen in perlbmk!!!
            WARNING("Jump to %s instead of calling it. Not rerouted.",
                signal_function->f_name);
        }
    }

    while (!ll_isEmpty(callers)) {
        caller_bbl = ll_removeFirst(callers);
        BBLRemoveEdgeTyped(caller_bbl, bsd_signal_entry_bbl, ET_FUNCALL);
        BBLRemoveEdgeTyped(caller_bbl, bsd_signal_first_bbl, ET_FUNDIRECT);
        BBLAddEdge(caller_bbl, signal_wrapper->f_entry, ET_FUNCALL);
        BBLAddEdge(caller_bbl, my1_bbl, ET_FUNDIRECT);
///        printf("caller_bbl->b_id         = %d\n", caller_bbl->b_id);
///        printf("caller_bbl->b_next       = 0x%x\n", caller_bbl->b_next);
///        printf("caller_bbl->b_next->b_id = %d\n", caller_bbl->b_next->b_id);

///        BBLRemoveEdgeTyped(bsd_signal_exit_bbl, caller_bbl->b_next, ET_RETURN);
///        BBLAddEdge(signal_wrapper->f_exit, caller_bbl->b_next, ET_RETURN);
    }

    BBLAddEdge(my4_bbl, bsd_signal_entry_bbl, ET_FUNCALL);
    BBLAddEdge(my4_bbl, bsd_signal_first_bbl, ET_FUNDIRECT);
    BBLAddEdge(bsd_signal_exit_bbl, my5_bbl, ET_RETURN);

    return;
}



/*
 *  Insert__bsd_signal() --
 *
 *  Inserts the code that resembles the __bsd_signal function from signal.h
 *  library.
 */
static Function *Insert__bsd_signal() {
    Function *fiterator;
    Function *f__errno_location;
    Function *f__sigaction;
    Function *my__bsd_signal;
    Block *errno_location_first_bbl;
    Block *errno_location_entry_bbl;
    Block *errno_location_exit_bbl;
    Block *sigaction_first_bbl;
    Block *sigaction_entry_bbl;
    Block *sigaction_exit_bbl;
    Block *my1_bbl;
    Block *my2_bbl;
    Block *my3_bbl;
    Block *my4_bbl;
    Block *my5_bbl;
    Block *my6_bbl;
    Block *my7_bbl;
    Block *my8_bbl;
    Block *my9_bbl;
    Block *my10_bbl;
    Block *my11_bbl;
    Instruction *new_instr;

    /* find or insert __errno_location() */
    FOREACH_FUNCTION(fiterator) {
        if (strcmp(fiterator->f_name, "__errno_location") == 0) {
            break;
        }
    }
    if (fiterator == NULL) {
        WARNING("__errno_location() not found.\n");
        //f__errno_location = Insert__errno_location();
    } else {
        REPORT(RP_SMLOOP, "Found %s()", fiterator->f_name);
        f__errno_location = fiterator;
    }
    errno_location_first_bbl = f__errno_location->f_first;
    errno_location_entry_bbl = f__errno_location->f_entry;
    errno_location_exit_bbl = f__errno_location->f_exit;

    /* find or insert __sigaction() */
    FOREACH_FUNCTION(fiterator) {
        if (strcmp(fiterator->f_name, "__sigaction") == 0) {
            break;
        }
    }
    if (fiterator == NULL) {
        WARNING("__sigaction() not found.\n");
        //f__sigaction = Insert__sigaction();
    } else {
        REPORT(RP_SMLOOP, "Found %s()", fiterator->f_name);
        f__sigaction = fiterator;
    }
    sigaction_first_bbl = f__sigaction->f_first;
    sigaction_entry_bbl = f__sigaction->f_entry;
    sigaction_exit_bbl = f__sigaction->f_exit;

    /*
     * Have all the links now, insert the code for my__bsd_signal() before the
     * __sigaction(), or, if it was not found, before our substitution for it -
     *  my__sigaction().
     */
    my1_bbl = BBLCreate();
    my1_bbl->b_type = BT_BRANCH;
    my2_bbl = BBLCreate();
    my2_bbl->b_type = BT_BRANCH;
    my3_bbl = BBLCreate();
    my3_bbl->b_type = BT_BRANCH;
    my4_bbl = BBLCreate();
    my4_bbl->b_type = BT_CALL;
    my5_bbl = BBLCreate();
    my5_bbl->b_type = BT_NORMAL;
    my6_bbl = BBLCreate();
    my6_bbl->b_type = BT_RETURN;
    my7_bbl = BBLCreate();
    my7_bbl->b_type = BT_NORMAL;
    my8_bbl = BBLCreate();
    my8_bbl->b_type = BT_BRANCH;
    my9_bbl = BBLCreate();
    my9_bbl->b_type = BT_CALL;
    my10_bbl = BBLCreate();
    my10_bbl->b_type = BT_BRANCH;
    my11_bbl = BBLCreate();
    my11_bbl->b_type = BT_UNCOND;

    my__bsd_signal = FunCreate();
    my__bsd_signal->f_name = "my__bsd_signal";
    my__bsd_signal->f_first = my1_bbl;
    my__bsd_signal->f_last = my11_bbl;
    if (BooleanParameter("LabelBasicBlocks")) {
        AddNewFunctionSymbol(my__bsd_signal);
    }
    InsertFunctionAfterFunction(my__bsd_signal, f__sigaction->f_prev);
    FUNCTION_SET_FLAG(my__bsd_signal, FUNCTION_DO_NOT_OBFUSCATE_SIGNAL);

    BBLAddEdge(my__bsd_signal->f_entry, my1_bbl, ET_ENTRY);
    BBLAddEdge(my1_bbl, my2_bbl, ET_FALSE);
    BBLAddEdge(my1_bbl, my4_bbl, ET_TRUE);
    BBLAddEdge(my2_bbl, my3_bbl, ET_FALSE);
    BBLAddEdge(my2_bbl, my4_bbl, ET_TRUE);
    BBLAddEdge(my3_bbl, my4_bbl, ET_FALSE);
    BBLAddEdge(my3_bbl, my7_bbl, ET_TRUE);
    BBLAddEdge(my4_bbl, errno_location_entry_bbl, ET_FUNCALL);
    BBLAddEdge(my4_bbl, errno_location_first_bbl, ET_FUNDIRECT);
    BBLAddEdge(errno_location_exit_bbl, my5_bbl, ET_RETURN);
    BBLAddEdge(my4_bbl, my5_bbl, ET_FUNLINK);
    BBLAddEdge(my5_bbl, my6_bbl, ET_NORMAL);
    BBLAddEdge(my6_bbl, my__bsd_signal->f_exit, ET_EXIT);
    BBLAddEdge(my7_bbl, my8_bbl, ET_NORMAL);
    BBLAddEdge(my8_bbl, my8_bbl, ET_TRUE);
    BBLAddEdge(my8_bbl, my9_bbl, ET_FALSE);
    BBLAddEdge(my9_bbl, sigaction_entry_bbl, ET_FUNCALL);
    BBLAddEdge(my9_bbl, sigaction_first_bbl, ET_FUNDIRECT);
    BBLAddEdge(sigaction_exit_bbl, my10_bbl, ET_RETURN);
    BBLAddEdge(my9_bbl, my10_bbl, ET_FUNLINK);
    BBLAddEdge(my10_bbl, my6_bbl, ET_TRUE);
    BBLAddEdge(my10_bbl, my11_bbl, ET_FALSE);
    BBLAddEdge(my11_bbl, my6_bbl, ET_UNCOND);

    //BBL1 (BT_BRANCH)
    InsertBBLBeforeBBL(my1_bbl, sigaction_first_bbl);
    my1_bbl->b_fun = my__bsd_signal->f_entry->b_fun;

    //push %ebp
    new_instr = InstrCreatePush(reg_bp | reg_32);
    BBLPrependInstr(my1_bbl, new_instr);

    //mov %esp, %ebp
    new_instr = InstrCreateMovRegToReg(reg_sp | reg_32, reg_bp | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //push %edi
    new_instr = InstrCreatePush(reg_di | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //push %esi
    new_instr = InstrCreatePush(reg_si | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //push %ebx
    new_instr = InstrCreatePush(reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //sub $0x12c, %esp
    new_instr = InstrCreateSubImm32FromReg(0x12c, reg_sp | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //mov $0xc(%ebp), %eax
    new_instr = InstrCreateMovMemDisp8BaseToReg(0xc, reg_bp | reg_32,
        reg_a | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //cmp $0xffffffff, %eax
    new_instr = InstrCreateCmpImm8WithReg(0xff, reg_a | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //mov $0x8(%ebp), %edi
    new_instr = InstrCreateMovMemDisp8BaseToReg(0x8, reg_bp | reg_32,
        reg_di | reg_32);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //je <&BBL4>
    new_instr = InstrCreateJe();
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL2 (BT_BRANCH)
    InsertBBLAfterBBL(my2_bbl, my1_bbl);

    //test %edi, %edi
    new_instr = InstrCreateTestRegWithReg(reg_di | reg_32, reg_di | reg_32);
    BBLPrependInstr(my2_bbl, new_instr);

    //jle <&BBL4>
    new_instr = InstrCreateShortJle();
    InsertInstrAfterInstr(new_instr, my2_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL3 (BT_BRANCH)
    InsertBBLAfterBBL(my3_bbl, my2_bbl);

    //cmp $0x3f, %edi
    new_instr = InstrCreateCmpImm8WithReg(0x3f, reg_di | reg_32);
    BBLPrependInstr(my3_bbl, new_instr);

    //jle <&BBL7>
    new_instr = InstrCreateShortJle();
    InsertInstrAfterInstr(new_instr, my3_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL4 (BT_CALL)
    InsertBBLAfterBBL(my4_bbl, my3_bbl);

    //call __errno_location
    new_instr = InstrCreateCALLDisp();
    BBLPrependInstr(my4_bbl, new_instr);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_CALL);

    //BBL5 (BT_NORMAL)
    InsertBBLAfterBBL(my5_bbl, my4_bbl);

    //mov $0xffffffff, %edx
    new_instr = InstrCreateMovImmToReg(reg_d | reg_32, 0xff);
    BBLPrependInstr(my5_bbl, new_instr);

    //mov $0x16, (%eax)
    new_instr = InstrCreateMovImmToMemBase(0x16, reg_a | reg_32);
    InsertInstrAfterInstr(new_instr, my5_bbl->b_last);

    //BBL6 (BT_RETURN)
    InsertBBLAfterBBL(my6_bbl, my5_bbl);

    //lea 0xfffffff4(%ebp), %esp
    new_instr = InstrCreateLea(reg_bp | reg_32, 0xf4, reg_sp | reg_32);
    BBLPrependInstr(my6_bbl, new_instr);

    //pop %ebx
    new_instr = InstrCreatePop(reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);

    //pop %esi
    new_instr = InstrCreatePop(reg_si | reg_32);
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);

    //mov %edx, %eax
    new_instr = InstrCreateMovRegToReg(reg_d | reg_32, reg_a | reg_32);
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);

    //pop %edi
    new_instr = InstrCreatePop(reg_di | reg_32);
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);

    //leave
    new_instr = InstrCreateLeave();
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);

    //ret
    new_instr = InstrCreateReturn();
    InsertInstrAfterInstr(new_instr, my6_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_RETURN);

    //BBL7 (BT_NORM)
    InsertBBLAfterBBL(my7_bbl, my6_bbl);

    //mov %eax, 0xffffff58(%ebp)
    new_instr = InstrCreateMovRegToMemDisp32Base(reg_a | reg_32, 0xffffff58,
        reg_bp | reg_32);
    BBLPrependInstr(my7_bbl, new_instr);

    //lea 0xffffff5c(%ebp), %esi
    new_instr = InstrCreateLea(reg_bp | reg_32, 0xffffff5c, reg_si | reg_32);
    InsertInstrAfterInstr(new_instr, my7_bbl->b_last);

    //mov $0x1f, %eax
    new_instr = InstrCreateMovImmToReg(reg_a | reg_32, 0x1f);
    InsertInstrAfterInstr(new_instr, my7_bbl->b_last);

    //nop
    new_instr = InstrCreateNOP();
    InsertInstrAfterInstr(new_instr, my7_bbl->b_last);

    //BBL8 (BT_BRANCH)
    InsertBBLAfterBBL(my8_bbl, my7_bbl);

    //mov $0x0, (%esi, %eax, 4)
    new_instr = InstrCreateMovImmToMemBaseIndex(0x0, reg_si | reg_32,
        reg_a | reg_32, 4);
    BBLPrependInstr(my8_bbl, new_instr);

    //dec %eax
    new_instr = InstrCreateDec(reg_a | reg_32);
    InsertInstrAfterInstr(new_instr, my8_bbl->b_last);

    //jns <&BBL8>
    new_instr = InstrCreateShortJns();
    InsertInstrAfterInstr(new_instr, my8_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL9 (BT_CALL)
    InsertBBLAfterBBL(my9_bbl, my8_bbl);

    //lea 0xffffffff(%edi), %eax
    new_instr = InstrCreateLea(reg_di | reg_32, 0xff, reg_a | reg_32);
    BBLPrependInstr(my9_bbl, new_instr);

    //mov %eax, %ecx
    new_instr = InstrCreateMovRegToReg(reg_a | reg_32, reg_c | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //and $0x1f, %ecx
    new_instr = InstrCreateAndImmWithReg(0x1f, reg_c | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //mov $0x1, %edx
    new_instr = InstrCreateMovImmToReg(reg_d | reg_32, 0x1);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //shr $0x5, %eax
    new_instr = InstrCreate32BitShr(reg_a | reg_32, 0x5);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //shl %cl, %edx
    new_instr = InstrCreate32BitShlCL(reg_d | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //or %edx, (%esi, %eax, 4)
    new_instr = InstrCreateOrRegWithMemBaseIndex(reg_d | reg_32,
        reg_si | reg_32, reg_a | reg_32, 4);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //test $0x0, %edx
    new_instr = InstrCreateTestImmWithReg(0x0, reg_d | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //sete %cl
    new_instr = InstrCreateSeteReg(reg_c | reg_16);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //movzbl %cl, %edx
    new_instr = InstrCreateMovzReg8ToReg32(reg_c | reg_16, reg_d | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //dec %edx
    new_instr = InstrCreateDec(reg_d | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //and $0xf0000000, %edx
    new_instr = InstrCreateAndImmWithReg(0xf0000000, reg_d | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //add $0x10000000, %edx
    new_instr = InstrCreateAddImm32ToReg(0x10000000, reg_d | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //push %eax
    new_instr = InstrCreatePush(reg_a | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //mov %edx, 0xffffffdc(%ebp)
    new_instr = InstrCreateMovRegToMemDisp8Base(reg_d | reg_32, 0xdc,
        reg_bp | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //lea 0xfffffec8(%ebp), %edx
    new_instr = InstrCreateLea(reg_bp | reg_32, 0xfffffec8, reg_d | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //push %edx
    new_instr = InstrCreatePush(reg_d | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //lea 0xffffff58(%ebp), %ebx
    new_instr = InstrCreateLea(reg_bp | reg_32, 0xffffff58, reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //push %ebx
    new_instr = InstrCreatePush(reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //push %edi
    new_instr = InstrCreatePush(reg_di | reg_32);
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);

    //call __sigaction
    new_instr = InstrCreateCALLDisp();
    InsertInstrAfterInstr(new_instr, my9_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_CALL);

    //BBL10 (BT_BRANCH)
    InsertBBLAfterBBL(my10_bbl, my9_bbl);

    //add $0x10, %esp
    new_instr = InstrCreateAddImm8ToReg(0x10, reg_sp | reg_32);
    BBLPrependInstr(my10_bbl, new_instr);

    //test %eax, %eax
    new_instr = InstrCreateTestRegWithReg(reg_a | reg_32, reg_a | reg_32);
    InsertInstrAfterInstr(new_instr, my10_bbl->b_last);

    //move $0xffffffff, %edx
    new_instr = InstrCreateMovImmToReg(reg_d | reg_32, 0xff);
    InsertInstrAfterInstr(new_instr, my10_bbl->b_last);

    //js <&BBL6>
    new_instr = InstrCreateShortJs();
    InsertInstrAfterInstr(new_instr, my10_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL11 (BT_UNCOND)
    InsertBBLAfterBBL(my11_bbl, my10_bbl);

    //mov 0xfffffec8(%ebp), %edx
    new_instr = InstrCreateMovMemDisp32BaseToReg(0xfffffec8, reg_bp | reg_32,
        reg_d | reg_32);
    BBLPrependInstr(my11_bbl, new_instr);

    //jmp <&BBL6>
    new_instr = InstrCreateJMP();
    InsertInstrAfterInstr(new_instr, my11_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_JUMP);

    REPORT(RP_SMLOOP, "Inserted %s() to replace __bsd_signal()",
        my__bsd_signal->f_name);
    return my__bsd_signal;
}



/*
 *  InsertMyRestore() --
 *
 *  Creates multiple copies of my_restore() function, which will get executed
 *  after __restore() and before the instruction at the original jump target.
 *  For each basic block it now contains at least the first instruction.
 *  Some basic blocks contain all instructions, some basic blocks after the
 *  first instruction contain a sequence of nops. For all blocks with nops,
 *  except the first one, the length of nops is equal to the length of the
 *  instructions that will replace nops later. For the first basic block the
 *  length of nops should be not less that the maximum possible length of the
 *  perfect hash function that will later replace nops. my_restore() will be
 *  finalized later, when the final addresses of the instructions are known
 *  (see CompleteMyRestore() in finalize.c).
 */
static void InsertMyRestore() {
    if (TRAPS_CONVERT_CALLS || TRAPS_CONVERT_RETURNS) {
        InsertLongMyRestore();
    } else {
        InsertShortMyRestore();
    }
}



/*
 *  InsertShortMyRestore() --
 *
 *  This one is used when only jumps are converted to traps. The short
 *  my_restore() function contains 2 basic blocks and, for the purpose of
 *  spreading it around the code, is split into 2 "pieces" ending in control
 *  transfers that do not have a fallthrough edge (jump, ret). Each of these
 *  two pieces is duplicated the number of times specified by the
 *  corresponding constant and is inserted after some random non-fallthrough
 *  location.
 */ 
static void InsertShortMyRestore() {
    int i;

    /* create empty lists for each of the 2 bbls of my_restore */
    for (i = 0; i < 2; i++) {
        my_restore_bbls_lists[i] = ll_new(LIST_UNSORTED);
    }

    /*
     * Insert NUMBER_OF_RESTORE1_COPIES copies of the first piece
     * (contains bbl #0) of my_restore.
     */
    for (i = 0; i < NUMBER_OF_RESTORE1_COPIES; i++) {
        InsertMyRestore1(i);
    }

    /*
     * Insert NUMBER_OF_RESTORE2_COPIES copies of the second piece
     * (contains bbls #1) of my_restore.
     */
    for (i = 0; i < NUMBER_OF_RESTORE2_COPIES; i++) {
        InsertShortMyRestore2(i);
    }

    MyRestoreAddEdges(0, NUMBER_OF_RESTORE1_COPIES,
        1, NUMBER_OF_RESTORE2_COPIES, ET_UNCOND);

    return;
}



/*
 *  InsertLongMyRestore() --
 *
 *  This one is used only when jumps, calls and returns are converted to traps.
 *  The my_restore() function contains 6 basic blocks and, for the purpose of
 *  spreading it around the code, is split into 3 "pieces" ending in control
 *  transfers that do not have a fallthrough edge (jump, ret). Each of these
 *  three pieces is duplicated the number of times specified by the
 *  corresponding constant and is inserted after some random non-fallthrough
 *  location.
 */
static void InsertLongMyRestore() {
    int i;

    /* create empty lists for each of the 6 bbls of my_restore */
    for (i = 0; i < 6; i++) {
        my_restore_bbls_lists[i] = ll_new(LIST_UNSORTED);
    }

    /*
     * Insert NUMBER_OF_RESTORE1_COPIES copies of the first piece
     * (contains bbl #0) of my_restore.
     */
    for (i = 0; i < NUMBER_OF_RESTORE1_COPIES; i++) {
        InsertMyRestore1(i);
    }

    /*
     * Insert NUMBER_OF_RESTORE2_COPIES copies of the second piece
     * (contains bbls #1 and #2) of my_restore.
     */
    for (i = 0; i < NUMBER_OF_RESTORE2_COPIES; i++) {
        InsertLongMyRestore2(i);
    }

    /*
     * Insert NUMBER_OF_RESTORE3_COPIES copies of the third piece
     * (contains bbls #3, #4, #5) of my_restore.
     */
    for (i = 0; i < NUMBER_OF_RESTORE3_COPIES; i++) {
        InsertLongMyRestore3(i);
    }

    //MyRestoreAddEdgesBbl3ToBbl5(); //think if reasonable
    MyRestoreAddEdges(1, NUMBER_OF_RESTORE2_COPIES,
        3, NUMBER_OF_RESTORE3_COPIES, ET_TRUE);
    MyRestoreAddEdges(0, NUMBER_OF_RESTORE1_COPIES,
        1, NUMBER_OF_RESTORE2_COPIES, ET_UNCOND);

    return;
}



/*
 *  InsertMyRestore1(num_int) --
 *
 *  Inserts one copy of the first piece of the my_restore() function after
 *  some randomly chosen non-fallthrough location.
 */
static void InsertMyRestore1(int num_int) {
    Function *my_restore;
    Block *non_fall_through_bbl;
    Block *my_restore_bbl0;
    Instruction *new_instr;
    char *num_str;

    my_restore_bbl0 = BBLCreate();
    my_restore_bbl0->b_type = BT_UNCOND;
    ll_add(my_restore_bbls_lists[0], my_restore_bbl0);

    my_restore = FunNew(); //takes care of insertion at the end of f_list
    my_restore->f_name = zalloc(15);
    strcpy(my_restore->f_name, "my_restore");
    num_str = zalloc(3);
    sprintf(num_str, "%d", num_int);
    strcat(my_restore->f_name, num_str);
    my_restore->f_first = my_restore_bbl0;
    my_restore->f_last = my_restore_bbl0; //will be replaced if more bbls
    if (BooleanParameter("LabelBasicBlocks")) {
        AddNewFunctionSymbol(my_restore);
    }
    FUNCTION_SET_FLAG(my_restore, FUNCTION_DO_NOT_OBFUSCATE_SIGNAL);

    BBLAddEdge(my_restore->f_entry, my_restore_bbl0, ET_ENTRY);

    non_fall_through_bbl = GetNonFallThroughBBL();

    //BBL0 (BT_UNCOND) - incomplete, filled with nops
    InsertBBLAfterBBL(my_restore_bbl0, non_fall_through_bbl);
    my_restore_bbl0->b_fun = my_restore->f_entry->b_fun;

    //pusha
    new_instr = InstrCreatePushAll();
    BBLPrependInstr(my_restore_bbl0, new_instr);

    //space for the perfect hash function inserted in finalize.c
    InsertNOPs(new_instr, 70);

    //jmp <&my_restore_bbls[1]>
    new_instr = InstrCreateJMP();
    InsertInstrAfterInstr(new_instr, my_restore_bbl0->b_last);

    REPORT(RP_SMLOOP, "Inserted some instructions and nops for long"
        " %s(), bbl 0", my_restore->f_name);

    return;
}



/*
 *  InsertShortMyRestore2(num_int) --
 *
 *  Inserts one copy of the second piece of the my_restore() function after
 *  some randomly chosen non-fallthrough location.
 */
static void InsertShortMyRestore2(int num_int) {
    Function *my_restore;
    Block *non_fall_through_bbl;
    Block *my_restore_bbl0;
    Block *my_restore_bbl1;
    Instruction *new_instr;
    int num_columns = 1;

    my_restore_bbl1 = BBLCreate();
    my_restore_bbl1->b_type = BT_RETURN;
    ll_add(my_restore_bbls_lists[1], my_restore_bbl1);

    my_restore_bbl0 = ll_get(my_restore_bbls_lists[0], num_int);
    my_restore = my_restore_bbl0->b_fun;
    my_restore->f_last = my_restore_bbl1;

    BBLAddEdge(my_restore_bbl1, my_restore->f_exit, ET_EXIT);

    non_fall_through_bbl = GetNonFallThroughBBL();

    //BBL1 (BT_BRANCH) - incomplete, filled with nops
    InsertBBLAfterBBL(my_restore_bbl1, non_fall_through_bbl);
    my_restore_bbl1->b_fun = my_restore->f_entry->b_fun;

    //imul $0x<num_columns * 4>, %eax, %eax
    new_instr = InstrCreateIMulImm8ByReg(num_columns *
        sizeof(ADDRESS), reg_a | reg_32);
    BBLPrependInstr(my_restore_bbl1, new_instr);

    //space for 4 more instructions inserted in finalize.c
    InsertNOPs(new_instr, 13);

    REPORT(RP_SMLOOP, "Inserted some instructions and nops for short"
        " %s(), bbl 1", my_restore->f_name);

    return;
}



/*
 *  InsertLongMyRestore2(num_int) --
 *
 *  Inserts one copy of the second piece of the my_restore() function after
 *  some randomly chosen non-fallthrough location.
 */
static void InsertLongMyRestore2(int num_int) {
    Function *my_restore;
    Block *non_fall_through_bbl;
    Block *my_restore_bbl0;
    Block *my_restore_bbl1;
    Block *my_restore_bbl2;
    Instruction *new_instr;
    int num_columns = 3;

    my_restore_bbl1 = BBLCreate();
    my_restore_bbl1->b_type = BT_BRANCH;
    ll_add(my_restore_bbls_lists[1], my_restore_bbl1);
    my_restore_bbl2 = BBLCreate();
    my_restore_bbl2->b_type = BT_RETURN;
    ll_add(my_restore_bbls_lists[2], my_restore_bbl2);

    my_restore_bbl0 = ll_get(my_restore_bbls_lists[0], num_int);
    my_restore = my_restore_bbl0->b_fun;

    BBLAddEdge(my_restore_bbl1, my_restore_bbl2, ET_FALSE);
    BBLAddEdge(my_restore_bbl2, my_restore->f_exit, ET_EXIT);

    non_fall_through_bbl = GetNonFallThroughBBL();

    //BBL1 (BT_BRANCH) - incomplete, filled with nops
    InsertBBLAfterBBL(my_restore_bbl1, non_fall_through_bbl);
    my_restore_bbl1->b_fun = my_restore->f_entry->b_fun;

    //imul $0x<num_columns * 4>, %eax, %eax
    new_instr = InstrCreateIMulImm8ByReg(num_columns *
        sizeof(ADDRESS), reg_a | reg_32);
    BBLPrependInstr(my_restore_bbl1, new_instr);

    //space for 2 more instructions inserted in finalize.c
    InsertNOPs(new_instr, 13);

    //BBL2 (BT_RETURN) - completed here
    InsertBBLAfterBBL(my_restore_bbl2, my_restore_bbl1);

    //popa
    new_instr = InstrCreatePopAll();
    BBLPrependInstr(my_restore_bbl2, new_instr);

    //add $0x4, %esp
    new_instr = InstrCreateAddImm8ToReg(0x4, reg_sp | reg_32);
    InsertInstrAfterInstr(new_instr, my_restore_bbl2->b_last);

    //ret
    new_instr = InstrCreateReturn();
    InsertInstrAfterInstr(new_instr, my_restore_bbl2->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_RETURN);

    REPORT(RP_SMLOOP, "Inserted some instructions and nops for long"
        " %s(), bbls 1, 2", my_restore->f_name);

    return;
}



/*
 *  InsertLongMyRestore3(num_int) --
 *
 *  Inserts one copy of the third piece of the my_restore() function after
 *  some randomly chosen non-fallthrough location.
 */
static void InsertLongMyRestore3(int num_int) {
    Function *my_restore;
    Block *non_fall_through_bbl;
    Block *my_restore_bbl0;
    Block *my_restore_bbl3;
    Block *my_restore_bbl4;
    Block *my_restore_bbl5;
    Instruction *new_instr;

    my_restore_bbl3 = BBLCreate();
    my_restore_bbl3->b_type = BT_BRANCH;
    ll_add(my_restore_bbls_lists[3], my_restore_bbl3);
    my_restore_bbl4 = BBLCreate();
    my_restore_bbl4->b_type = BT_NORMAL;
    ll_add(my_restore_bbls_lists[4], my_restore_bbl4);
    my_restore_bbl5 = BBLCreate();
    my_restore_bbl5->b_type = BT_RETURN;
    ll_add(my_restore_bbls_lists[5], my_restore_bbl5);

    my_restore_bbl0 = ll_get(my_restore_bbls_lists[0], num_int);
    my_restore = my_restore_bbl0->b_fun;
    my_restore->f_last = my_restore_bbl5;

    BBLAddEdge(my_restore_bbl3, my_restore_bbl5, ET_TRUE);
    BBLAddEdge(my_restore_bbl3, my_restore_bbl4, ET_FALSE);
    BBLAddEdge(my_restore_bbl4, my_restore_bbl5, ET_NORMAL);
    BBLAddEdge(my_restore_bbl5, my_restore->f_exit, ET_EXIT);

    non_fall_through_bbl = GetNonFallThroughBBL();

    //BBL3 (BT_BRANCH) - incomplete, filled with nops
    InsertBBLAfterBBL(my_restore_bbl3, non_fall_through_bbl);
    my_restore_bbl3->b_fun = my_restore->f_entry->b_fun;

    //mov 0x20(%esp), %ecx
    new_instr = InstrCreateMovMemDisp8BaseToReg(0x20, reg_sp | reg_32,
        reg_c | reg_32);
    BBLPrependInstr(my_restore_bbl3, new_instr);
    
    //space for 5 more instructions inserted in finalize.c
    InsertNOPs(new_instr, 22);
    
    //BBL4 (BT_NORMAL) - incomplete, filled with nops
    InsertBBLAfterBBL(my_restore_bbl4, my_restore_bbl3);

    //mov %ecx, $0x24(%esp)
    new_instr = InstrCreateMovRegToMemDisp8Base(reg_c | reg_32,
        0x24, reg_sp | reg_32);
    BBLPrependInstr(my_restore_bbl4, new_instr);
    
    //space for 2 more instructions inserted in finalize.c
    InsertNOPs(new_instr, 11);

    //BBL5 (BT_RETURN) - completed here
    InsertBBLAfterBBL(my_restore_bbl5, my_restore_bbl4);

    //popa
    new_instr = InstrCreatePopAll();
    BBLPrependInstr(my_restore_bbl5, new_instr);

    //ret
    new_instr = InstrCreateReturn(); 
    InsertInstrAfterInstr(new_instr, my_restore_bbl5->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_RETURN);

    REPORT(RP_SMLOOP, "Inserted some instructions and nops for long"
        " %s(), bbls 3 - 5", my_restore->f_name);

    return;
}



/*  
 *  MyRestoreAddEdges(bbl_num_from, num_copies_from, bbl_num_to, num_copies_to,
 *      edge_type) --
 *
 *  Adds the edge_type edges from all copies of bbl #bbl_num_from of
 *  my_restore() function to the randomly chosen copies of bbl #bbl_num_to of
 *  my_restore().
 */
static void MyRestoreAddEdges(int bbl_num_from, int num_copies_from,
    int bbl_num_to, int num_copies_to, int edge_type) {
    int i;

    for (i = 0; i < num_copies_from; i++) {
        Block *bbl_from = ll_get(my_restore_bbls_lists[bbl_num_from], i);
        Block *bbl_to = ll_get(my_restore_bbls_lists[bbl_num_to],
            rand() % num_copies_to);
        BBLAddEdge(bbl_from, bbl_to, edge_type);
    }

    return;
}



/*
 *  InstallMySignalHandler() --
 *
 *  Inserts the instructions that install handlers for three signals: SIGILL,
 *  SIGFPE and SIGSEGV. The same handler - my_handler() is installed in each
 *  case.
 */
static void InstallMySignalHandler() {
    //Function *my_handler;
    Function *fiterator;
    Function *signal_function = NULL;
    Block *install_bbl;
    Block *my_handler_bbl0;

    /* Locate function used to install a signal handler */
    FOREACH_FUNCTION(fiterator) {
        if ((strcmp(fiterator->f_name, "__bsd_signal") == 0)
            || (strcmp(fiterator->f_name, "signal") == 0)
            || (strcmp(fiterator->f_name, "my__bsd_signal") == 0)) {
            break;
        }
    }
    if (fiterator == NULL) {
        //at least my__bsd_signal() must be in the code now
        WARNING("No signal() function found. my_handler() not installed.");
        return;
    }
    REPORT(RP_SMLOOP, "Found %s()", fiterator->f_name);
    signal_function = fiterator;

    /* Allocate space for the raised_from flags in the .data section */
    raised_from = SECTION_REF_DATA->s_newdatasize;
    IncreaseSectionSize(SECTION_REF_DATA,
        NUMBER_OF_COMMUNICATION_FLAGS * sizeof(ADDRESS));
    
    InsertMyHandler(signal_function);

    //need a place, where to insert signal installing instructions,
    //for now using a place in _start before the call to __libc_start_main
    FOREACH_FUNCTION(fiterator) {
        if (strcmp(fiterator->f_name, "_start") == 0) {
            FUNCTION_SET_FLAG(fiterator, FUNCTION_DO_NOT_OBFUSCATE_SIGNAL);
            break;
        }
    }
    if (fiterator == NULL) {
        //if _start() is not there have to find some other place in the
        //beginning of execution from where to install handlers
        WARNING("No _start() function found. my_handler() not installed.");
        return;
    }
    REPORT(RP_SMLOOP, "Found %s()", fiterator->f_name);

    BBLDivideBlock(fiterator->f_first, fiterator->f_first->b_last);

    my_handler_bbl0 = ll_get(my_handler_bbls_lists[0],
        rand() % NUMBER_OF_HANDLER1_COPIES);
    install_bbl = InstallHandler(signal_function, my_handler_bbl0,
        fiterator->f_first, fiterator->f_first->b_next, 4);  //SIGILL
    REPORT(RP_SMLOOP, "Installed %s() for SIGILL",
        my_handler_bbl0->b_fun->f_name);

    my_handler_bbl0 = ll_get(my_handler_bbls_lists[0],
        rand() % NUMBER_OF_HANDLER1_COPIES);
    install_bbl = InstallHandler(signal_function, my_handler_bbl0,
        install_bbl->b_next, install_bbl->b_next->b_next, 8);  //SIGFPE
    REPORT(RP_SMLOOP, "Installed %s() for SIGFPE",
        my_handler_bbl0->b_fun->f_name);

    my_handler_bbl0 = ll_get(my_handler_bbls_lists[0],
        rand() % NUMBER_OF_HANDLER1_COPIES);
    install_bbl = InstallHandler(signal_function, my_handler_bbl0,
        install_bbl->b_next, install_bbl->b_next->b_next, 11);  //SIGSEGV
    REPORT(RP_SMLOOP, "Installed %s() for SIGSEGV",
        my_handler_bbl0->b_fun->f_name);

    return;
}



/*
 *  InsertMyHandler() --
 *
 *  Creates multiple copies of my_handler() function, which will get executed
 *  when one of the following three signals are raised:
 *    SIGILL - raised when an illegal instruction is encountered,
 *    SIGFPE - raised when an instruction performs division by 0,
 *    SIGSEGV - raised when an instruction tries to read/write an illegal
 *              memory address.
 *  The my_handler() function contains 17 basic blocks and, for the purpose of
 *  spreading it around the code, is split into 4 "pieces" ending in a control
 *  transfers that do not have a fallthrough edge (jump, ret). Each of these
 *  four pieces is duplicated the number of times specified by the
 *  corresponding constant and is inserted after some random non-fallthrough
 *  location.
 */
static void InsertMyHandler(Function *signal_function) {
    int i;

    /* create empty lists for each of the 6 bbls of my_restore */
    for (i = 0; i < 14; i++) {
        my_handler_bbls_lists[i] = ll_new(LIST_UNSORTED);
    }

    /*
     * Insert NUMBER_OF_HANDLER1_COPIES copies of the first piece
     * (contains bbls #0, #1, #2, #3) of my_handler.
     */
    for (i = 0; i < NUMBER_OF_HANDLER1_COPIES; i++) {
        InsertMyHandler1(i);
    }

    /*
     * Insert NUMBER_OF_HANDLER2_COPIES copies of the second piece
     * (contains bbls #4, #5) of my_handler.
     */
    for (i = 0; i < NUMBER_OF_HANDLER2_COPIES; i++) {
        InsertMyHandler2(i, signal_function);
    }

    /*
     * Insert NUMBER_OF_HANDLER3_COPIES copies of the third piece
     * (contains bbls #6, #7, #8, #9, #10, #13) of my_handler.
     */
    for (i = 0; i < NUMBER_OF_HANDLER3_COPIES; i++) {
        InsertMyHandler3(i, signal_function);
    }

    /*
     * Insert NUMBER_OF_HANDLER4_COPIES copies of the fourth piece
     * (contains bbls #11, #12) of my_handler.
     */
    for (i = 0; i < NUMBER_OF_HANDLER4_COPIES; i++) {
        InsertMyHandler4(i, signal_function);
    }

    MyHandlerAddEdges(12, NUMBER_OF_HANDLER4_COPIES,
        10, NUMBER_OF_HANDLER3_COPIES, ET_UNCOND);
    MyHandlerAddEdges(7, NUMBER_OF_HANDLER3_COPIES,
        11, NUMBER_OF_HANDLER4_COPIES, ET_TRUE);
    MyHandlerAddEdges(6, NUMBER_OF_HANDLER3_COPIES,
        13, NUMBER_OF_HANDLER3_COPIES, ET_TRUE);
    MyHandlerAddEdges(2, NUMBER_OF_HANDLER1_COPIES,
        6, NUMBER_OF_HANDLER3_COPIES, ET_TRUE);
    MyHandlerAddEdges(1, NUMBER_OF_HANDLER1_COPIES,
        4, NUMBER_OF_HANDLER2_COPIES, ET_TRUE);

    return;
}



/*  
 *  InsertMyHandler1(num_int) --
 *
 *  Inserts one copy of the first piece of the my_handler() function after
 *  some randomly chosen non-fallthrough location.
 */ 
static void InsertMyHandler1(int num_int) {
    Function *my_handler;
    Block *non_fall_through_bbl;
    Block *my_handler_bbl0;
    Block *my_handler_bbl1;
    Block *my_handler_bbl2;
    Block *my_handler_bbl3;
    Instruction *new_instr;
    char *num_str;
    Patch *upm;
    PatchSectionToInstr *patch;
    LinkedList *upms;
    int curr_offset;

    my_handler_bbl0 = BBLCreate();
    my_handler_bbl0->b_type = BT_NORMAL;
    ll_add(my_handler_bbls_lists[0], my_handler_bbl0);
    my_handler_bbl1 = BBLCreate();
    my_handler_bbl1->b_type = BT_BRANCH;
    ll_add(my_handler_bbls_lists[1], my_handler_bbl1);
    my_handler_bbl2 = BBLCreate();
    my_handler_bbl2->b_type = BT_BRANCH;
    ll_add(my_handler_bbls_lists[2], my_handler_bbl2);
    my_handler_bbl3 = BBLCreate();
    my_handler_bbl3->b_type = BT_UNCOND;
    ll_add(my_handler_bbls_lists[3], my_handler_bbl3);

    my_handler = FunNew(); //takes care of insertion at the end of f_list
    my_handler->f_name = zalloc(15);
    strcpy(my_handler->f_name, "my_handler");
    num_str = zalloc(3);
    sprintf(num_str, "%d", num_int);
    strcat(my_handler->f_name, num_str);
    my_handler->f_first = my_handler_bbl0;
    my_handler->f_last = my_handler_bbl3; //will be replaced if more bbls
    if (BooleanParameter("LabelBasicBlocks")) {
        AddNewFunctionSymbol(my_handler);
    }
    FUNCTION_SET_FLAG(my_handler, FUNCTION_DO_NOT_OBFUSCATE_SIGNAL);

    BBLAddEdge(my_handler->f_entry, my_handler_bbl0, ET_ENTRY);
    BBLAddEdge(my_handler_bbl0, my_handler_bbl1, ET_NORMAL);
    BBLAddEdge(my_handler_bbl1, my_handler_bbl2, ET_FALSE);
    BBLAddEdge(my_handler_bbl2, my_handler_bbl3, ET_FALSE);
    BBLAddEdge(my_handler_bbl3, my_handler_bbl1, ET_UNCOND);

    non_fall_through_bbl = GetNonFallThroughBBL();

    //BBL0 (BT_BRANCH)
    InsertBBLAfterBBL(my_handler_bbl0, non_fall_through_bbl);
    my_handler_bbl0->b_fun = my_handler->f_entry->b_fun;

    //push %ebp 
    new_instr = InstrCreatePush(reg_bp | reg_32);
    BBLPrependInstr(my_handler_bbl0, new_instr);

    //mov %esp, %ebp
    new_instr = InstrCreateMovRegToReg(reg_sp | reg_32, reg_bp | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl0->b_last);

    curr_offset = rand() % COMMUNICATION_FLAGS_OFFSET -
        COMMUNICATION_FLAGS_OFFSET / 2;

    //mov <&raised_from - curr_offset * 4>, %eax
    new_instr = InstrCreateMovImmToReg(reg_a | reg_32, 0x0);
    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, raised_from, new_instr,
        new_instr->i_src);
    upms = UPMTableGetByDest(new_instr->i_src);
    upm = ll_get(upms, 0);
    ll_free(upms);
    if (upm == NULL) {
        MALFUNCTION("UPM was not properly cloned");
    }
    patch = upm->p_patch;
    patch->p_src_offset -= curr_offset * sizeof(ADDRESS);
    InsertInstrAfterInstr(new_instr, my_handler_bbl0->b_last);

    //mov $(curr_offset), %ebx
    new_instr = InstrCreateMovImmToReg(reg_b | reg_32, curr_offset);
    InsertInstrAfterInstr(new_instr, my_handler_bbl0->b_last);

    //BBL1 (BT_BRANCH)
    InsertBBLAfterBBL(my_handler_bbl1, my_handler_bbl0);

    //cmpl $0x0, (%eax,%ebx,4)
    new_instr = InstrCreateCmpImm8WithMemBaseIndex(0x0, reg_a | reg_32,
        reg_b | reg_32, 4);
    BBLPrependInstr(my_handler_bbl1, new_instr);

    //jne <&reinstall1_bbl>
    new_instr = InstrCreateShortJne();
    InsertInstrAfterInstr(new_instr, my_handler_bbl1->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL2 (BT_BRANCH)
    InsertBBLAfterBBL(my_handler_bbl2, my_handler_bbl1);

    //inc %ebx
    new_instr = InstrCreateInc(reg_b | reg_32);
    BBLPrependInstr(my_handler_bbl2, new_instr);

    //cmp $(NUMBER_OF_COMMUNICATION_FLAGS + curr_offset), %ebx
    new_instr = InstrCreateCmpImm8WithReg(NUMBER_OF_COMMUNICATION_FLAGS +
        curr_offset, reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl2->b_last);

    //je <&my3_bbl>
    new_instr = InstrCreateShortJe();
    InsertInstrAfterInstr(new_instr, my_handler_bbl2->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL3 (BT_UNCOND)
    InsertBBLAfterBBL(my_handler_bbl3, my_handler_bbl2);

    //jmp <&my_handler_bbl1>
    new_instr = InstrCreateJMP();
    BBLPrependInstr(my_handler_bbl3, new_instr);

    REPORT(RP_SMLOOP, "Inserted instructions for %s(), bbls 0 - 3",
        my_handler->f_name);

    return;
}



/*  
 *  InsertMyHandler2(num_int, signal_function) --
 *  
 *  Inserts one copy of the second piece of the my_handler() function after
 *  some randomly chosen non-fallthrough location.
 */ 
static void InsertMyHandler2(int num_int, Function *signal_function) {
    Function *my_handler;
    Block *non_fall_through_bbl;
    Block *my_handler_bbl0;
    Block *my_handler_bbl4;
    Block *my_handler_bbl5;
    Block *my_restore_bbl0;
    Instruction *new_instr;

    my_handler_bbl4 = BBLCreate();
    my_handler_bbl4->b_type = BT_NORMAL;
    ll_add(my_handler_bbls_lists[4], my_handler_bbl4);
    my_handler_bbl5 = BBLCreate();
    my_handler_bbl5->b_type = BT_RETURN;
    ll_add(my_handler_bbls_lists[5], my_handler_bbl5);

    my_handler_bbl0 = ll_get(my_handler_bbls_lists[0], num_int);
    my_handler = my_handler_bbl0->b_fun;

    BBLAddEdge(my_handler_bbl4, my_handler_bbl5, ET_NORMAL);

    non_fall_through_bbl = GetNonFallThroughBBL();

    //BBL4 (BT_NORMAL)
    InsertBBLAfterBBL(my_handler_bbl4, non_fall_through_bbl);
    my_handler_bbl4->b_fun = my_handler->f_entry->b_fun;

    //mov $0x0, (%eax,%ebx,4)
    new_instr = InstrCreateMovImmToMemBaseIndex(0x0, reg_a | reg_32,
        reg_b | reg_32, 4);
    BBLPrependInstr(my_handler_bbl4, new_instr);

    //Has to reinstall itself... The code will be inserted
    //here, but call to insert it will be made later.

    //BBL5 (BT_RETURN)
    InsertBBLAfterBBL(my_handler_bbl5, my_handler_bbl4);

    //leave
    new_instr = InstrCreateLeave();
    BBLPrependInstr(my_handler_bbl5, new_instr);

    //mov 0x40(%esp), %eax
    new_instr = InstrCreateMovMemDisp8BaseToReg(0x40, reg_sp | reg_32,
        reg_a | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl5->b_last);

    //mov 0x24(%esp), %ebx
    new_instr = InstrCreateMovMemDisp8BaseToReg(0x24, reg_sp | reg_32,
        reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl5->b_last);

    //mov %eax, (%ebx)
    new_instr = InstrCreateMovRegToMemBase(reg_a | reg_32, reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl5->b_last);

    //mov $<&my_restore_bbl0>, 0x40(%esp)
    new_instr = InstrCreateMovImmToMemDisp8Base(12345, 0x40, reg_sp | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl5->b_last);
    my_restore_bbl0 = ll_get(my_restore_bbls_lists[0],
        rand() % NUMBER_OF_RESTORE1_COPIES);
    AddPatchInstrAddressIntoInstr(my_restore_bbl0->b_first, new_instr,
        new_instr->i_src);

    //ret
    new_instr = InstrCreateReturn();
    InsertInstrAfterInstr(new_instr, my_handler_bbl5->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_RETURN);

    //insert two more bbls between 4 and 5 that will reinstall this handler
    my_handler_bbl0 = ll_get(my_handler_bbls_lists[0],
        rand() % NUMBER_OF_HANDLER1_COPIES);
    InstallHandler(signal_function, my_handler_bbl0,
        my_handler_bbl4, my_handler_bbl5, 0);

    REPORT(RP_SMLOOP, "Inserted instructions for %s(), bbls 4, 5",
        my_handler->f_name);

    return;
}



/*  
 *  InsertMyHandler3(num_int, signal_function) --
 *  
 *  Inserts one copy of the third piece of the my_handler() function after
 *  some randomly chosen non-fallthrough location.
 */ 
static void InsertMyHandler3(int num_int, Function *signal_function) {
    Function *my_handler;
    Block *non_fall_through_bbl;
    Block *my_handler_bbl0;
    Block *my_handler_bbl6;
    Block *my_handler_bbl7;
    Block *my_handler_bbl8;
    Block *my_handler_bbl9;
    Block *my_handler_bbl10;
    Block *reinstall_bbl;
    Instruction *new_instr;

    my_handler_bbl6 = BBLCreate();
    my_handler_bbl6->b_type = BT_BRANCH;
    ll_add(my_handler_bbls_lists[6], my_handler_bbl6);
    my_handler_bbl7 = BBLCreate();
    my_handler_bbl7->b_type = BT_BRANCH;
    ll_add(my_handler_bbls_lists[7], my_handler_bbl7);
    my_handler_bbl8 = BBLCreate();
    my_handler_bbl8->b_type = BT_CALL;
    ll_add(my_handler_bbls_lists[8], my_handler_bbl8);
    my_handler_bbl9 = BBLCreate();
    my_handler_bbl9->b_type = BT_NORMAL;
    ll_add(my_handler_bbls_lists[9], my_handler_bbl9);
    my_handler_bbl10 = BBLCreate();
    my_handler_bbl10->b_type = BT_RETURN;
    ll_add(my_handler_bbls_lists[10], my_handler_bbl10);

    my_handler_bbl0 = ll_get(my_handler_bbls_lists[0], num_int);
    my_handler = my_handler_bbl0->b_fun;
    my_handler->f_last = my_handler_bbl10;

    BBLAddEdge(my_handler_bbl6, my_handler_bbl7, ET_FALSE);
    BBLAddEdge(my_handler_bbl7, my_handler_bbl8, ET_FALSE);
    BBLAddEdge(my_handler_bbl8, my_handler_bbl9, ET_FUNLINK);
    BBLAddEdge(my_handler_bbl9, my_handler_bbl10, ET_NORMAL); //temporarily
    BBLAddEdge(my_handler_bbl10, my_handler->f_exit, ET_EXIT);

    non_fall_through_bbl = GetNonFallThroughBBL();

    //BBL6 (BT_BRANCH)
    InsertBBLAfterBBL(my_handler_bbl6, non_fall_through_bbl);
    my_handler_bbl6->b_fun = my_handler->f_entry->b_fun;

    //move $<sh_table_offset>, %edx
    new_instr = InstrCreateMovImmToReg(reg_d | reg_32, 123);
    BBLPrependInstr(my_handler_bbl6, new_instr);
    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, sh_table_offset,
        new_instr, new_instr->i_src);

    //mov 0x8(%ebp), %ebx  //move the handler number into ebx 
    new_instr = InstrCreateMovMemDisp8BaseToReg(0x8, reg_bp | reg_32,
        reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl6->b_last);

    //mov %ebx, %ecx
    new_instr = InstrCreateMovRegToReg(reg_b | reg_32, reg_c | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl6->b_last);

    //shl $0x2, %ebx
    new_instr = InstrCreate32BitShlImm(reg_b | reg_32, 0x2);
    InsertInstrAfterInstr(new_instr, my_handler_bbl6->b_last);

    //add %edx, %ebx
    new_instr = InstrCreateAddRegToReg(reg_d | reg_32, reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl6->b_last);

    //cmp $0x1, (%ebx)
    new_instr = InstrCreateCmpImm8WithMemBase(0x1, reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl6->b_last);

    //je <reinstall itself>
    new_instr = InstrCreateJe();
    InsertInstrAfterInstr(new_instr, my_handler_bbl6->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL7 (BT_BRANCH)
    InsertBBLAfterBBL(my_handler_bbl7, my_handler_bbl6);

    //cmp $0x0, (%ebx)
    new_instr = InstrCreateCmpImm8WithMemBase(0x0, reg_b | reg_32);
    BBLPrependInstr(my_handler_bbl7, new_instr);

    //je <install default>
    new_instr = InstrCreateJe();
    InsertInstrAfterInstr(new_instr, my_handler_bbl7->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_BRANCH);

    //BBL8 (BT_CALL)
    InsertBBLAfterBBL(my_handler_bbl8, my_handler_bbl7);

    //sub $0xc, %esp
    new_instr = InstrCreateSubImm8FromReg(0xc, reg_sp | reg_32);
    BBLPrependInstr(my_handler_bbl8, new_instr);

    //push %ecx
    new_instr = InstrCreatePush(reg_c | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl8->b_last);

    //call *(%ebx)
    new_instr = InstrCreateCallNearMemBase(reg_b | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl8->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_CALL);

    //BBL9 (BT_NORMAL)
    InsertBBLAfterBBL(my_handler_bbl9, my_handler_bbl8);

    //add $0x10, %esp
    new_instr = InstrCreateAddImm8ToReg(0x10, reg_sp | reg_32);
    BBLPrependInstr(my_handler_bbl9, new_instr);

    //Has to reinstall itself before returning... The code will be inserted
    //here, but call to insert it will be made later.

    //BBL10 (BT_RETURN)
    InsertBBLAfterBBL(my_handler_bbl10, my_handler_bbl9);

    //leave
    new_instr = InstrCreateLeave();
    BBLPrependInstr(my_handler_bbl10, new_instr);

    //ret
    new_instr = InstrCreateReturn();
    InsertInstrAfterInstr(new_instr, my_handler_bbl10->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_RETURN);

    //insert two more bbls between 9 and 10 that will reinstall this handler
    reinstall_bbl = InstallHandler(signal_function, my_handler_bbl0,
        my_handler_bbl9, my_handler_bbl10, 0);
    ll_add(my_handler_bbls_lists[13], reinstall_bbl);

    REPORT(RP_SMLOOP, "Inserted instructions for %s(), bbls 6 - 10",
        my_handler->f_name);

    return;
}



/*  
 *  InsertMyHandler4(num_int, signal_function) --
 *  
 *  Inserts one copy of the fourth piece of the my_handler() function after
 *  some randomly chosen non-fallthrough location.
 */ 
static void InsertMyHandler4(int num_int, Function *signal_function) {
    Function *my_handler;
    Block *non_fall_through_bbl;
    Block *my_handler_bbl0;
    Block *my_handler_bbl11;
    Block *my_handler_bbl12;
    Instruction *new_instr;

    my_handler_bbl11 = BBLCreate();
    my_handler_bbl11->b_type = BT_CALL;
    ll_add(my_handler_bbls_lists[11], my_handler_bbl11);
    my_handler_bbl12 = BBLCreate();
    my_handler_bbl12->b_type = BT_UNCOND;
    ll_add(my_handler_bbls_lists[12], my_handler_bbl12);

    my_handler_bbl0 = ll_get(my_handler_bbls_lists[0], num_int);
    my_handler = my_handler_bbl0->b_fun;
    my_handler->f_last = my_handler_bbl12;

    BBLAddEdge(my_handler_bbl11, signal_function->f_entry, ET_FUNCALL);
    BBLAddEdge(my_handler_bbl11, signal_function->f_first, ET_FUNDIRECT);
    BBLAddEdge(my_handler_bbl11, my_handler_bbl12, ET_FUNLINK);
    BBLAddEdge(signal_function->f_exit, my_handler_bbl12, ET_RETURN);

    non_fall_through_bbl = GetNonFallThroughBBL();

    //BBL11 (BT_CALL)
    InsertBBLAfterBBL(my_handler_bbl11, non_fall_through_bbl);
    my_handler_bbl11->b_fun = my_handler->f_entry->b_fun;

    //sub $0x8, %esp
    new_instr = InstrCreateSubImm8FromReg(0x8, reg_sp | reg_32);
    BBLPrependInstr(my_handler_bbl11, new_instr);

    //push $0x0
    new_instr = InstrCreatePushImm(0);
    InsertInstrAfterInstr(new_instr, my_handler_bbl11->b_last);

    //push $0x8(%ebp)
    new_instr = InstrCreatePushMemDisp8Base(0x8, reg_bp | reg_32);
    InsertInstrAfterInstr(new_instr, my_handler_bbl11->b_last);

    //call __bsd_signal
    new_instr = InstrCreateCALLDisp();
    InsertInstrAfterInstr(new_instr, my_handler_bbl11->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_CALL);

    //BBL9 (BT_UNCOND)
    InsertBBLAfterBBL(my_handler_bbl12, my_handler_bbl11);

    //add $0x10, %esp
    new_instr = InstrCreateAddImm8ToReg(0x10, reg_sp | reg_32);
    BBLPrependInstr(my_handler_bbl12, new_instr);

    //jmp <&my_handler_bbl10>
    new_instr = InstrCreateJMP();
    InsertInstrAfterInstr(new_instr, my_handler_bbl12->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_JUMP);

    REPORT(RP_SMLOOP, "Inserted instructions for %s(), bbls 11, 12",
        my_handler->f_name);

    return;
}



/*  
 *  MyHandlerAddEdges(bbl_num_from, num_copies_from, bbl_num_to, num_copies_to,
 *      edge_type) --
 *
 *  Adds the edge_type edges from all copies of bbl #bbl_num_from of
 *  my_handler() function to the randomly chosen copies of bbl #bbl_num_to of
 *  my_handler().
 */
static void MyHandlerAddEdges(int bbl_num_from, int num_copies_from,
    int bbl_num_to, int num_copies_to, int edge_type) {
    int i;

    for (i = 0; i < num_copies_from; i++) {
        Block *bbl_from = ll_get(my_handler_bbls_lists[bbl_num_from], i);
        Block *bbl_to = ll_get(my_handler_bbls_lists[bbl_num_to],
            rand() % num_copies_to);
        BBLAddEdge(bbl_from, bbl_to, edge_type);
    }

    return;
}



/*
 *  InstallHandler(signal_function, handler_first_bbl, prev_bbl, next_bbl,
 *      signal_number) --
 *
 *  Inserts a code that installs a signal handler by calling the
 *  signal_function (pointer to signal() or __bsd_signal() or my__bsd_signal()).
 *  The code (2 bbls) is inserted between prev_bbl and next_bbl, which
 *  have to be connected by a BT_NORMAL edge. The signal number is
 *  contained in either signal_number if it is > 0, or in 0x8(%ebp) at runtime
 *  if signal_number == 0.
 */
Block *InstallHandler(Function *signal_function, Block *handler_first_bbl,
    Block *prev_bbl, Block *next_bbl, int signal_number) {

    Block *my1_bbl;
    Block *my2_bbl;
    Instruction *new_instr;

    my1_bbl = BBLCreate();
    my1_bbl->b_type = BT_CALL;
    my2_bbl = BBLCreate();
    my2_bbl->b_type = BT_NORMAL;

    BBLRemoveEdgeTyped(prev_bbl, next_bbl, ET_NORMAL);
    BBLAddEdge(prev_bbl, my1_bbl, ET_NORMAL);
    BBLAddEdge(my1_bbl, my2_bbl, ET_FUNLINK);
    BBLAddEdge(my1_bbl, signal_function->f_entry, ET_FUNCALL);
    BBLAddEdge(my1_bbl, signal_function->f_first, ET_FUNDIRECT);
    BBLAddEdge(my2_bbl, next_bbl, ET_NORMAL);
    BBLAddEdge(signal_function->f_exit, my2_bbl, ET_RETURN);

    //BBL1 (BT_CALL)
    InsertBBLAfterBBL(my1_bbl, prev_bbl);
    my1_bbl->b_weight = prev_bbl->b_weight;

    //sub $0x8, %esp
    new_instr = InstrCreateSubImm8FromReg(0x8, reg_sp | reg_32);
    BBLPrependInstr(my1_bbl, new_instr);

    //push <&handler_first_instr>
    new_instr = InstrCreatePushImm(123);
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);
    AddPatchInstrAddressIntoInstr(handler_first_bbl->b_first, new_instr,
        new_instr->i_src);

    if (signal_number == 0) {  //called to reinstall itself from the handler
        //push 0x8(%ebp)
        new_instr = InstrCreatePushMemDisp8Base(0x8, reg_bp | reg_32);
    } else {//called with actual number to install a handler for the first time
        //push <signal_number>
        new_instr = InstrCreatePushImm(signal_number);
    }
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);

    //call __bsd_signal
    new_instr = InstrCreateCALLDisp();
    InsertInstrAfterInstr(new_instr, my1_bbl->b_last);
    INSTR_SET_FLAG(new_instr, INSTR_CLASS_CALL);

    //BBL2 (BT_NORMAL)
    InsertBBLAfterBBL(my2_bbl, my1_bbl);
    my2_bbl->b_weight = my1_bbl->b_weight;

    //add $0x10, %esp
    new_instr = InstrCreateAddImm8ToReg(0x10, reg_sp | reg_32);
    BBLPrependInstr(my2_bbl, new_instr);

    return my1_bbl;
}



/*
 *  ConvertControlTransfersToTraps() --
 *
 *  Converts control transfer instructions (jumps, calls, returns) to the
 *  sequence of instructions that cause the SIGILL, SIGFPE or SIGSEGV signal to
 *  be raised. It does so by calling one of the 9 <instr>To<signal> functions
 *  on the corresponding control transfer instruction.
 *
 *  Instruction              | Signal raised
 *  -------------------------|----------------
 *  Illegal instruction      |   SIGILL (4)
 *  Division by zero         |   SIGFPE (8)
 *  Illegal memory reference |   SIGSEGV (11)
 */
void ConvertControlTransfersToTraps(int highest_bbl_id) {
    Block *biterator;
//    int choice;

    int num_jumps_marked = 0;
    int num_jumps_psw_live = 0;
    int num_jumps_converted = 0;
//    int num_jumps_conv_SIGILL = 0;
//    int num_jumps_conv_SIGFPE = 0;
//    int num_jumps_conv_SIGSEGV = 0;
    int num_jumps_conv_0regs = 0;
    int num_jumps_conv_1reg = 0;
    int num_jumps_conv_2regs = 0;
    int num_jumps_conv_3regs = 0;
    int num_jumps_conv_4regs = 0;
    int jumps_num_regs_dead[5] = {0,0,0,0,0};

    int num_calls_marked = 0;
    int num_calls_psw_live = 0;
    int num_calls_converted = 0;
//    int num_calls_conv_SIGILL = 0;
//    int num_calls_conv_SIGFPE = 0;
//    int num_calls_conv_SIGSEGV = 0;
    int num_calls_conv_0regs = 0;
    int num_calls_conv_1reg = 0;
    int num_calls_conv_2regs = 0;
    int num_calls_conv_3regs = 0;
    int num_calls_conv_4regs = 0;
    int calls_num_regs_dead[5] = {0,0,0,0,0};

    int num_returns_marked = 0;
    int num_returns_psw_live = 0;
    int num_returns_converted = 0;
//    int num_returns_conv_SIGILL = 0;
//    int num_returns_conv_SIGFPE = 0;
//    int num_returns_conv_SIGSEGV = 0;
    int num_returns_conv_0regs = 0;
    int num_returns_conv_1reg = 0;
    int num_returns_conv_2regs = 0;
    int num_returns_conv_3regs = 0;
    int num_returns_conv_4regs = 0;
    int returns_num_regs_dead[5] = {0,0,0,0,0};

    int num_marked = 0;
    int num_psw_live = 0;
    int num_converted = 0;

    int num_first_bbl_jumps_a = 0;
    int num_first_bbl_jumps_b = 0;

    //Function *fiterator;
    //int num_blocks;
    //int num_instr;

    //!!!declare global
    //junk_branches_to_patch = ll_new(LIST_UNSORTED);

    ComputePSWLiveness(1);
    ComputeRegisterLiveness(1);

#if 0
    FOREACH_FUNCTION(fiterator) {
        if (strcmp(fiterator->f_name, "__printf_fp") == 0) {
            Block *biterator;
            Instruction *iiterator;

            num_instr = 0;
            num_blocks = 0;
            printf("__printf_fp stats:\n");
            FOREACH_BBL_IN_FUN(fiterator, biterator) {
                num_blocks++;
                FOREACH_INSTR_IN_BBL(biterator, iiterator) {
                    num_instr++;
                }
            }
            printf("Blocks: %d, instructions: %d", num_blocks, num_instr);
        }
    }
#endif

#if 0
    printf("\n__overflow() follows:\n");
    FOREACH_BBL(biterator) {
        if (strcmp(biterator->b_fun->f_name, "__overflow") == 0) {
            PrintBbl(biterator);
        }
    }
    printf("__overflow() ends.\n\n");
#endif

    FOREACH_BBL(biterator) {
        //PrintBbl(biterator);
/*
//to obfuscate the first jump in hot fun introduced by interleaving
        if ((biterator->b_id > highest_bbl_id) &&
            //(biterator->b_last == biterator->b_first) &&
            (INSTR_IS_DIRECT_JUMP(biterator->b_last)) &&
            ((biterator->b_last == biterator->b_first) &&
             (biterator->b_prev->b_fun->f_first == biterator->b_prev))
            ) {
            num_first_bbl_jumps_a++;
            BBL_SET_FLAG(biterator, BBL_TRAP_FLAG);
        }

        if ((INSTR_IS_DIRECT_JUMP(biterator->b_last)) &&
            (BBL_HAS_FLAG(biterator, BBL_HOT_FLAG)) &&
            (BBL_HAS_FLAG(biterator->b_next, BBL_TRAP_FLAG)) &&
            (biterator->b_next->b_next) &&
            (INSTR_TARGET_BBL(biterator->b_last) == biterator->b_next->b_next)
           ) {
            num_first_bbl_jumps_b++;
            BBL_SET_FLAG(biterator, BBL_TRAP_FLAG);
        }
*/
        if (BBL_HAS_FLAG(biterator, BBL_TRAP_FLAG)) {
            num_marked++;

            if (PSWLiveBitvectorAtInstr(biterator->b_last) & 0x003fffff) {
                num_psw_live++;
            }
#if 0
            //!!!tmp hack
            //change for better condition:
            //if (free_reg is ebx && ebx was not saved on function entry)
            //    assume ebx is live
            if (strcmp(biterator->b_fun->f_name, "__overflow") == 0) {
                WARNING("block in __overflow() excluded: ebx may be live");
                PrintBbl(biterator);
                continue;
            }
#endif
            //choice = rand() % 3;
            //choice = 0;

            if (INSTR_IS_DIRECT_JUMP(biterator->b_last)) {
                int num_regs_available = 3;
                UINT32 regs_available[4] =
                    {BITVECTOR_REGISTER_NONE, BITVECTOR_REGISTER_NONE,
                    BITVECTOR_REGISTER_NONE, BITVECTOR_REGISTER_NONE};
                int next_slot = 0;
                num_jumps_marked++;

                //printf("Live regs: 0x%x\n",
                //    RegsLiveBitvectorAtInstr(biterator->b_last));
                //printf(" free register: 0x%x (%d)\n",
                //    GetFreeRegisterAtInstr(biterator->b_last),
                //    GetFreeRegisterAtInstr(biterator->b_last));
//                printf(" PSW: Ox%x\n",
//                    PSWLiveBitvectorAtInstr(biterator->b_last));

                if (PSWLiveBitvectorAtInstr(biterator->b_last) & 0x003fffff) {
                    num_jumps_psw_live++;
                    continue;
                }

                /* check %eax availability */
                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_EAX) {
                    num_regs_available--;
                } else {
                    //printf("  eax available\n");
                    regs_available[next_slot++] = reg_a;
                }

                /* check %ebx availability */
/*                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_EBX) {
                    num_regs_available--;
                } else {
                    //printf("  ebx available\n");
                    regs_available[next_slot++] = reg_b;
                }
*/
                /* check %ecx availability */
                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_ECX) {
                    num_regs_available--;
                } else {
                    //printf("  ecx available\n");
                    regs_available[next_slot++] = reg_c;
                }

                /* check %edx availability */
                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_EDX) {
                    num_regs_available--;
                } else {
                    //printf("  edx available\n");
                    regs_available[next_slot++] = reg_d;
                }
                //printf("-- %d available--\n", num_regs_available);

                jumps_num_regs_dead[num_regs_available]++;

                if (num_regs_available == 0) {
                    JumpToSIGILL(biterator->b_last);
                    num_jumps_conv_0regs++;
                } else if (num_regs_available == 1) {
                    JumpToTrap1Reg(biterator->b_last, regs_available[0]);
                    num_jumps_conv_1reg++;
                } else if (num_regs_available == 2) {
                    JumpToTrap2Regs(biterator->b_last, regs_available[0],
                        regs_available[1]);
                    num_jumps_conv_2regs++;
                } else if (num_regs_available == 3) {
                    JumpToTrap3Regs(biterator->b_last, regs_available[0],
                        regs_available[1], regs_available[2]);
                    num_jumps_conv_3regs++;
                } else if (num_regs_available == 4) {
                    JumpToTrap4Regs(biterator->b_last, regs_available[0],
                        regs_available[1], regs_available[2],
                        regs_available[3]);
                    num_jumps_conv_4regs++;
                }
/*old
                if (choice == 0) {
                    JumpToSIGILL(biterator->b_last);
                    num_jumps_conv_SIGILL++;
                } else if (choice == 1) {
                    JumpToSIGFPE(biterator->b_last);
                    num_jumps_conv_SIGFPE++;
                } else {
                    JumpToSIGSEGV(biterator->b_last);
                    num_jumps_conv_SIGSEGV++;
                }
*/
                num_jumps_converted++;

            } else if (INSTR_IS_DIRECT_CALL(biterator->b_last)) {
                int num_regs_available = 3;
                UINT32 regs_available[4] =
                    {BITVECTOR_REGISTER_NONE, BITVECTOR_REGISTER_NONE,
                    BITVECTOR_REGISTER_NONE, BITVECTOR_REGISTER_NONE};
                int next_slot = 0;
                //Edge *edge;
                num_calls_marked++;

                if (PSWLiveBitvectorAtInstr(biterator->b_last) & 0x003fffff) {
                    num_calls_psw_live++;
                    continue;
                }
#if 0
                //!!!tmp hack
                edge = BBLGetFunDirectEdge(biterator);
                if (edge != NULL) {
                    Function *callee = edge->e_to->b_fun;
                    if (strcmp(callee->f_name, "__overflow") == 0) {
                        WARNING("call to %s() from %s() excluded: "
                            "ebx may be live", callee->f_name,
                            biterator->b_fun->f_name);
                        PrintBbl(biterator);
                        //continue; 
                    }
                }
#endif
                /* check %eax availability */
                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_EAX) {
                    num_regs_available--;
                } else {
                    //printf("  eax available\n");
                    regs_available[next_slot++] = reg_a;
                }

                /* check %ebx availability */
/*                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_EBX) {
                    num_regs_available--;
                } else {
                    //printf("  ebx available\n");
                    regs_available[next_slot++] = reg_b;
                }
*/
                /* check %ecx availability */
                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_ECX) {
                    num_regs_available--;
                } else {
                    //printf("  ecx available\n");
                    regs_available[next_slot++] = reg_c;
                }

                /* check %edx availability */
                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_EDX) {
                    num_regs_available--;
                } else {
                    //printf("  edx available\n");
                    regs_available[next_slot++] = reg_d;
                }
                //printf("-- %d available--\n", num_regs_available);

                calls_num_regs_dead[num_regs_available]++;

                if (num_regs_available == 0) {
                    CallToSIGILL(biterator->b_last);
                    num_calls_conv_0regs++;
                } else if (num_regs_available == 1) {
                    CallToTrap1Reg(biterator->b_last, regs_available[0]);
                    num_calls_conv_1reg++;
                } else if (num_regs_available == 2) {
                    CallToTrap2Regs(biterator->b_last, regs_available[0],
                        regs_available[1]);
                    num_calls_conv_2regs++;
                } else if (num_regs_available == 3) {
                    CallToTrap3Regs(biterator->b_last, regs_available[0],
                        regs_available[1], regs_available[2]);
                    num_calls_conv_3regs++;
                } else if (num_regs_available == 4) {
                    CallToTrap4Regs(biterator->b_last, regs_available[0],
                        regs_available[1], regs_available[2],
                        regs_available[3]);
                    num_calls_conv_4regs++;
                }
/*old
                if (choice == 0) {
                    CallToSIGILL(biterator->b_last);
                    num_calls_conv_SIGILL++;
                } else if (choice == 1) {
                    CallToSIGFPE(biterator->b_last);
                    num_calls_conv_SIGFPE++;
                } else {
                    CallToSIGSEGV(biterator->b_last);
                    num_calls_conv_SIGSEGV++;
                }
*/
                num_calls_converted++;

            //!!!have to exclude iret, think how
            } else if (isReturnClass(biterator->b_last)) {
                int num_regs_available = 3;
                UINT32 regs_available[4] =
                    {BITVECTOR_REGISTER_NONE, BITVECTOR_REGISTER_NONE,
                    BITVECTOR_REGISTER_NONE, BITVECTOR_REGISTER_NONE};
                int next_slot = 0;
                num_returns_marked++;

                if (PSWLiveBitvectorAtInstr(biterator->b_last) & 0x003fffff) {
                    num_returns_psw_live++;
                    continue;
                }

                /* check %eax availability */
                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_EAX) {
                    num_regs_available--;
                } else {
                    //printf("  eax available\n");
                    regs_available[next_slot++] = reg_a;
                }

                /* check %ebx availability */
/*                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_EBX) {
                    num_regs_available--;
                } else {
                    //printf("  ebx available\n");
                    regs_available[next_slot++] = reg_b;
                }
*/
                /* check %ecx availability */
                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_ECX) {
                    num_regs_available--;
                } else {
                    //printf("  ecx available\n");
                    regs_available[next_slot++] = reg_c;
                }

                /* check %edx availability */
                if (RegsLiveBitvectorAtInstr(biterator->b_last) &
                    BITVECTOR_REGISTER_EDX) {
                    num_regs_available--;
                } else {
                    //printf("  edx available\n");
                    regs_available[next_slot++] = reg_d;
                }
                //printf("-- %d available--\n", num_regs_available);

                returns_num_regs_dead[num_regs_available]++;

                if (num_regs_available == 0) {
                    RetToSIGILL(biterator->b_last);
                    num_returns_conv_0regs++;
                } else if (num_regs_available == 1) {
                    //RetToTrap1Reg(biterator->b_last, regs_available[0]);
                    JumpToTrap1Reg(biterator->b_last, regs_available[0]);
                    num_returns_conv_1reg++;
                } else if (num_regs_available == 2) {
                    //RetToTrap2Regs(biterator->b_last, regs_available[0],
                    //    regs_available[1]);
                    JumpToTrap2Regs(biterator->b_last, regs_available[0],
                        regs_available[1]);
                    num_returns_conv_2regs++;
                } else if (num_regs_available == 3) {
                    //RetToTrap3Regs(biterator->b_last, regs_available[0],
                    //    regs_available[1], regs_available[2]);
                    JumpToTrap3Regs(biterator->b_last, regs_available[0],
                        regs_available[1], regs_available[2]);
                    num_returns_conv_3regs++;
                } else if (num_regs_available == 4) {
                    //RetToTrap4Regs(biterator->b_last, regs_available[0],
                    //    regs_available[1], regs_available[2],
                    //    regs_available[3]);
                    JumpToTrap4Regs(biterator->b_last, regs_available[0],
                        regs_available[1], regs_available[2],
                        regs_available[3]);
                    num_returns_conv_4regs++;
                }
/*old
                if (choice == 0) {
                    RetToSIGILL(biterator->b_last);
                    num_returns_conv_SIGILL++;
                } else if (choice == 1) {
                    RetToSIGFPE(biterator->b_last);
                    num_returns_conv_SIGFPE++;
                } else {
                    RetToSIGSEGV(biterator->b_last);
                    num_returns_conv_SIGSEGV++;
                }
*/
                num_returns_converted++;

            } else {
                WARNING("Attempt to convert to trap instruction"
                    " other than jump, call or ret");
                num_converted--;
            }

            num_converted++;
        } //end if (BBL_HAS_FLAG(biterator, BBL_TRAP_FLAG))
    } //end FOREACH_BBL(biterator)

    REPORT(RP_DETAIL, "Total jumps converted (marked):   %6d (%d) %1.4f",
        num_jumps_converted, num_jumps_marked,
        (double)num_jumps_converted/num_jumps_marked);
//    REPORT(RP_DETAIL, "    to SIGILL:  %6d", num_jumps_conv_SIGILL);
//    REPORT(RP_DETAIL, "    to SIGFPE:  %6d", num_jumps_conv_SIGFPE);
//    REPORT(RP_DETAIL, "    to SIGSEGV: %6d", num_jumps_conv_SIGSEGV);
    REPORT(RP_DETAIL, "    using 0 regs: %6d", num_jumps_conv_0regs);
    REPORT(RP_DETAIL, "    using 1 reg : %6d", num_jumps_conv_1reg);
    REPORT(RP_DETAIL, "    using 2 regs: %6d", num_jumps_conv_2regs);
    REPORT(RP_DETAIL, "    using 3 regs: %6d", num_jumps_conv_3regs);
    //REPORT(RP_DETAIL, "    using 4 regs: %6d", num_jumps_conv_4regs);

    REPORT(RP_DETAIL, " Number of registers and locations available:");
    REPORT(RP_DETAIL, "                   0  -  %5d (%d) %1.4f",
        jumps_num_regs_dead[0], num_jumps_marked - num_jumps_psw_live,
        (double)jumps_num_regs_dead[0]/(num_jumps_marked - num_jumps_psw_live));
    REPORT(RP_DETAIL, "                   1  -  %5d (%d) %1.4f",
        jumps_num_regs_dead[1], num_jumps_marked - num_jumps_psw_live,
        (double)jumps_num_regs_dead[1]/(num_jumps_marked - num_jumps_psw_live));
    REPORT(RP_DETAIL, "                   2  -  %5d (%d) %1.4f",
        jumps_num_regs_dead[2], num_jumps_marked - num_jumps_psw_live,
        (double)jumps_num_regs_dead[2]/(num_jumps_marked - num_jumps_psw_live));
    REPORT(RP_DETAIL, "                   3  -  %5d (%d) %1.4f",
        jumps_num_regs_dead[3], num_jumps_marked - num_jumps_psw_live,
        (double)jumps_num_regs_dead[3]/(num_jumps_marked - num_jumps_psw_live));
    //REPORT(RP_DETAIL, "                   4  -  %5d (%d) %1.4f",
    //    jumps_num_regs_dead[4], num_jumps_marked - num_jumps_psw_live,
    //    (double)jumps_num_regs_dead[4]/(num_jumps_marked - num_jumps_psw_live));

    REPORT(RP_DETAIL, "Total calls converted (marked):   %6d (%d) %1.4f",
        num_calls_converted, num_calls_marked,
        (double)num_calls_converted/num_calls_marked);
//    REPORT(RP_DETAIL, "    to SIGILL:  %6d", num_calls_conv_SIGILL);
//    REPORT(RP_DETAIL, "    to SIGFPE:  %6d", num_calls_conv_SIGFPE);
//    REPORT(RP_DETAIL, "    to SIGSEGV: %6d", num_calls_conv_SIGSEGV);
    REPORT(RP_DETAIL, "    using 0 regs: %6d", num_calls_conv_0regs);
    REPORT(RP_DETAIL, "    using 1 reg : %6d", num_calls_conv_1reg);
    REPORT(RP_DETAIL, "    using 2 regs: %6d", num_calls_conv_2regs);
    REPORT(RP_DETAIL, "    using 3 regs: %6d", num_calls_conv_3regs);
    //REPORT(RP_DETAIL, "    using 4 regs: %6d", num_calls_conv_4regs);
    REPORT(RP_DETAIL, " Number of registers and locations available:");
    REPORT(RP_DETAIL, "                   0  -  %5d (%d) %1.4f",
        calls_num_regs_dead[0], num_calls_marked - num_calls_psw_live,
        (double)calls_num_regs_dead[0]/(num_calls_marked - num_calls_psw_live));
    REPORT(RP_DETAIL, "                   1  -  %5d (%d) %1.4f",
        calls_num_regs_dead[1], num_calls_marked - num_calls_psw_live,
        (double)calls_num_regs_dead[1]/(num_calls_marked - num_calls_psw_live));
    REPORT(RP_DETAIL, "                   2  -  %5d (%d) %1.4f",
        calls_num_regs_dead[2], num_calls_marked - num_calls_psw_live,
        (double)calls_num_regs_dead[2]/(num_calls_marked - num_calls_psw_live));
    REPORT(RP_DETAIL, "                   3  -  %5d (%d) %1.4f",
        calls_num_regs_dead[3], num_calls_marked - num_calls_psw_live,
        (double)calls_num_regs_dead[3]/(num_calls_marked - num_calls_psw_live));
    //REPORT(RP_DETAIL, "                   4  -  %5d (%d) %1.4f",
    //    calls_num_regs_dead[4], num_calls_marked - num_calls_psw_live,
    //    (double)calls_num_regs_dead[4]/(num_calls_marked - num_calls_psw_live));

    REPORT(RP_DETAIL, "Total returns converted (marked): %6d (%d) %1.4f",
        num_returns_converted, num_returns_marked,
        (double)num_returns_converted/num_returns_marked);
//    REPORT(RP_DETAIL, "    to SIGILL:  %6d", num_returns_conv_SIGILL);
//    REPORT(RP_DETAIL, "    to SIGFPE:  %6d", num_returns_conv_SIGFPE);
//    REPORT(RP_DETAIL, "    to SIGSEGV: %6d", num_returns_conv_SIGSEGV);
    REPORT(RP_DETAIL, "    using 0 regs: %6d", num_returns_conv_0regs);
    REPORT(RP_DETAIL, "    using 1 reg : %6d", num_returns_conv_1reg);
    REPORT(RP_DETAIL, "    using 2 regs: %6d", num_returns_conv_2regs);
    REPORT(RP_DETAIL, "    using 3 regs: %6d", num_returns_conv_3regs);
    //REPORT(RP_DETAIL, "    using 4 regs: %6d", num_returns_conv_4regs);
    REPORT(RP_DETAIL, " Number of registers and locations available:");
    REPORT(RP_DETAIL, "                   0  -  %5d (%d) %1.4f",
        returns_num_regs_dead[0], num_returns_marked - num_returns_psw_live,
        (double)returns_num_regs_dead[0]/(num_returns_marked -
        num_returns_psw_live));
    REPORT(RP_DETAIL, "                   1  -  %5d (%d) %1.4f",
        returns_num_regs_dead[1], num_returns_marked - num_returns_psw_live,
        (double)returns_num_regs_dead[1]/(num_returns_marked -
        num_returns_psw_live));
    REPORT(RP_DETAIL, "                   2  -  %5d (%d) %1.4f",
        returns_num_regs_dead[2], num_returns_marked - num_returns_psw_live,
        (double)returns_num_regs_dead[2]/(num_returns_marked -
        num_returns_psw_live));
    REPORT(RP_DETAIL, "                   3  -  %5d (%d) %1.4f",
        returns_num_regs_dead[3], num_returns_marked - num_returns_psw_live,
        (double)returns_num_regs_dead[3]/(num_returns_marked -
        num_returns_psw_live));
    //REPORT(RP_DETAIL, "                   4  -  %5d (%d) %1.4f",
    //    returns_num_regs_dead[4], num_returns_marked - num_returns_psw_live,
    //    (double)returns_num_regs_dead[4]/(num_returns_marked -
    //    num_returns_psw_live));

    REPORT(RP_DETAIL, "Total converted:         %6d", num_converted);
    REPORT(RP_DETAIL, "Converted/Marked = %d/%d = %1.4f",
        num_converted, num_marked, (double)num_converted/num_marked);
    REPORT(RP_DETAIL, "At least 1 reg     %d/%d = %1.4f",
        num_converted - num_jumps_conv_0regs - num_calls_conv_0regs -
        num_returns_conv_0regs, num_marked, (double)(num_converted -
        num_jumps_conv_0regs - num_calls_conv_0regs -
        num_returns_conv_0regs)/num_marked);
    REPORT(RP_DETAIL, "num_first_bbl_jumps = %d + %d = %d",
        num_first_bbl_jumps_a, num_first_bbl_jumps_b,
        num_first_bbl_jumps_a + num_first_bbl_jumps_b);

    return;
}



void JumpToTrap1Reg(Instruction *jump, UINT32 reg1) {
    Instruction *new_instr;

    if (rand() % 2) {
        InsertInstrDecStack(sizeof(ADDRESS), jump);
        InsertInstrSetCommunicationFlag(jump);
    } else {
        InsertInstrSetCommunicationFlag(jump);
        InsertInstrDecStack(sizeof(ADDRESS), jump);
    }

    //this will change to load the opaque value into the chosen register
    //such that this value is always 0.
    //mov $0x0, %<reg1>
    new_instr = InstrCreateMovImmToReg(reg1 | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, jump);

    //idiv %<reg1>
    //new_instr = InstrCreate32bitIDivByReg(reg1 | reg_32);
    //InsertInstrBeforeInstr(new_instr, jump);
    //INSTR_SET_FLAG(new_instr, INSTR_SIGFPE_FLAG);
    InsertUseZeroRegInstr(reg1, reg1, jump);

    /* get rid of jump */
    InstrDisconnect(jump);
//    InstrFree(jump);

    return;
}



void JumpToTrap2Regs(Instruction *jump, UINT32 reg1, UINT32 reg2) {
    Instruction *new_instr;

    if (rand() % 2) {
        InsertInstrDecStack(sizeof(ADDRESS), jump);
        InsertInstrSetCommunicationFlag(jump);
    } else {
        InsertInstrSetCommunicationFlag(jump);
        InsertInstrDecStack(sizeof(ADDRESS), jump);
    }

    //mov $0x0, %<reg1>
    new_instr = InstrCreateMovImmToReg(reg1 | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, jump);

    //mov $0x1234, %<reg2>
    new_instr = InstrCreateMovImmToReg(reg2 | reg_32, 0x1234);
    InsertInstrBeforeInstr(new_instr, jump);

    //mov (%<reg1>), %<reg2>
    //new_instr = InstrCreateMovMemBaseToReg(reg1 | reg_32, reg2 | reg_32);
    //InsertInstrBeforeInstr(new_instr, jump);
    //INSTR_SET_FLAG(new_instr, INSTR_SIGSEGV_FLAG);
    InsertUseZeroRegInstr(reg1, reg2, jump);

    /* get rid of jump */
    InstrDisconnect(jump);
//    InstrFree(jump);

    return;
}



void JumpToTrap3Regs(Instruction *jump, UINT32 reg1, UINT32 reg2,
    UINT32 reg3) {

    Instruction *new_instr;

    if (rand() % 2) {
        InsertInstrDecStack(sizeof(ADDRESS), jump);
        InsertInstrSetCommunicationFlag(jump);
    } else {
        InsertInstrSetCommunicationFlag(jump);
        InsertInstrDecStack(sizeof(ADDRESS), jump);
    }

    //this will change to load the opaque value into the chosen register
    //such that this value is always 0.
    //mov $0x0, %<reg1>
    new_instr = InstrCreateMovImmToReg(reg1 | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, jump);

    //mov $0x1234, %<reg2>
    new_instr = InstrCreateMovImmToReg(reg2 | reg_32, 0x1234);
    InsertInstrBeforeInstr(new_instr, jump);

    //mov $0x12345, %<reg3>
    new_instr = InstrCreateMovImmToReg(reg3 | reg_32, 0x12345);
    InsertInstrBeforeInstr(new_instr, jump);

    //idiv %<reg1>
    //new_instr = InstrCreate32bitIDivByReg(reg1 | reg_32);
    //InsertInstrBeforeInstr(new_instr, jump);
    //INSTR_SET_FLAG(new_instr, INSTR_SIGFPE_FLAG);
    InsertUseZeroRegInstr(reg1, reg3, jump);

    /* get rid of jump */
    InstrDisconnect(jump);
//    InstrFree(jump);

    return;
}



void JumpToTrap4Regs(Instruction *jump, UINT32 reg1, UINT32 reg2,
    UINT32 reg3, UINT32 reg4) {

    Instruction *new_instr;

    if (rand() % 2) {
        InsertInstrDecStack(sizeof(ADDRESS), jump);
        InsertInstrSetCommunicationFlag(jump);
    } else {
        InsertInstrSetCommunicationFlag(jump);
        InsertInstrDecStack(sizeof(ADDRESS), jump);
    }

    //mov $0x0, %<reg1>
    new_instr = InstrCreateMovImmToReg(reg1 | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, jump);

    //mov $0x1234, %<reg2>
    new_instr = InstrCreateMovImmToReg(reg2 | reg_32, 0x1234);
    InsertInstrBeforeInstr(new_instr, jump);

    //mov $0x12345, %<reg3>
    new_instr = InstrCreateMovImmToReg(reg3 | reg_32, 0x12345);
    InsertInstrBeforeInstr(new_instr, jump);

    //mov $0x123456, %<reg4>
    new_instr = InstrCreateMovImmToReg(reg4 | reg_32, 0x123456);
    InsertInstrBeforeInstr(new_instr, jump);

    //mov (%<reg1>), %<reg4>
    //new_instr = InstrCreateMovMemBaseToReg(reg1 | reg_32, reg4 | reg_32);
    //InsertInstrBeforeInstr(new_instr, jump);
    //INSTR_SET_FLAG(new_instr, INSTR_SIGSEGV_FLAG);
    InsertUseZeroRegInstr(reg1, reg4, jump);

    /* get rid of jump */
    InstrDisconnect(jump);
//    InstrFree(jump);

    return;
}



void CallToTrap1Reg(Instruction *call, UINT32 reg1) {
    Instruction *new_instr;

    if (!CallEdgesExist(call)) {
        return;
    }

    if (rand() % 2) {
        InsertInstrDecStack(2 * sizeof(ADDRESS), call);
        InsertInstrSetCommunicationFlag(call);
    } else {
        InsertInstrSetCommunicationFlag(call);
        InsertInstrDecStack(2 * sizeof(ADDRESS), call);
    }

    //this will change to load the opaque value into the chosen register
    //such that this value is always 0.
    //mov $0x0, %<reg1>
    new_instr = InstrCreateMovImmToReg(reg1 | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, call);

    //idiv %<reg1>
    //new_instr = InstrCreate32bitIDivByReg(reg1 | reg_32);
    //InsertInstrBeforeInstr(new_instr, call);
    //INSTR_SET_FLAG(new_instr, INSTR_SIGFPE_FLAG);
    InsertUseZeroRegInstr(reg1, reg1, call);

    /* get rid of call */
    InstrDisconnect(call);
//    InstrFree(call);

    return;
}



void CallToTrap2Regs(Instruction *call, UINT32 reg1, UINT32 reg2) {
    Instruction *new_instr;

    if (!CallEdgesExist(call)) {
        return;
    }

    if (rand() % 2) {
        InsertInstrDecStack(2 * sizeof(ADDRESS), call);
        InsertInstrSetCommunicationFlag(call);
    } else {
        InsertInstrSetCommunicationFlag(call);
        InsertInstrDecStack(2 * sizeof(ADDRESS), call);
    }

    //mov $0x0, %<reg1>
    new_instr = InstrCreateMovImmToReg(reg1 | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, call);

    //mov $0x1234, %<reg2>
    new_instr = InstrCreateMovImmToReg(reg2 | reg_32, 0x1234);
    InsertInstrBeforeInstr(new_instr, call);

    //mov (%<reg1>), %<reg2>
    //new_instr = InstrCreateMovMemBaseToReg(reg1 | reg_32, reg2 | reg_32);
    //InsertInstrBeforeInstr(new_instr, call);
    //INSTR_SET_FLAG(new_instr, INSTR_SIGSEGV_FLAG);
    InsertUseZeroRegInstr(reg1, reg2, call);

    /* get rid of call */
    InstrDisconnect(call);
//    InstrFree(call);

    return;
}



void CallToTrap3Regs(Instruction *call, UINT32 reg1, UINT32 reg2,
    UINT32 reg3) {

    Instruction *new_instr;

    if (!CallEdgesExist(call)) {
        return;
    }

    if (rand() % 2) {
        InsertInstrDecStack(2 * sizeof(ADDRESS), call);
        InsertInstrSetCommunicationFlag(call);
    } else {
        InsertInstrSetCommunicationFlag(call);
        InsertInstrDecStack(2 * sizeof(ADDRESS), call);
    }

    //this will change to load the opaque value into the chosen register
    //such that this value is always 0.
    //mov $0x0, %<reg1>
    new_instr = InstrCreateMovImmToReg(reg1 | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, call);

    //mov $0x1234, %<reg2>
    new_instr = InstrCreateMovImmToReg(reg2 | reg_32, 0x1234);
    InsertInstrBeforeInstr(new_instr, call);

    //mov $0x12345, %<reg3>
    new_instr = InstrCreateMovImmToReg(reg3 | reg_32, 0x12345);
    InsertInstrBeforeInstr(new_instr, call);

    //idiv %<reg1>
    //new_instr = InstrCreate32bitIDivByReg(reg1 | reg_32);
    //InsertInstrBeforeInstr(new_instr, call);
    //INSTR_SET_FLAG(new_instr, INSTR_SIGFPE_FLAG);
    InsertUseZeroRegInstr(reg1, reg3, call);

    /* get rid of call */
    InstrDisconnect(call);
//    InstrFree(call);

    return;
}



void CallToTrap4Regs(Instruction *call, UINT32 reg1, UINT32 reg2,
    UINT32 reg3, UINT32 reg4) {

    Instruction *new_instr;

    if (!CallEdgesExist(call)) {
        return;
    }

    if (rand() % 2) {
        InsertInstrDecStack(2 * sizeof(ADDRESS), call);
        InsertInstrSetCommunicationFlag(call);
    } else {
        InsertInstrSetCommunicationFlag(call);
        InsertInstrDecStack(2 * sizeof(ADDRESS), call);
    }

    //mov $0x0, %<reg1>
    new_instr = InstrCreateMovImmToReg(reg1 | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, call);

    //mov $0x1234, %<reg2>
    new_instr = InstrCreateMovImmToReg(reg2 | reg_32, 0x1234);
    InsertInstrBeforeInstr(new_instr, call);

    //mov $0x12345, %<reg3>
    new_instr = InstrCreateMovImmToReg(reg3 | reg_32, 0x12345);
    InsertInstrBeforeInstr(new_instr, call);

    //mov $0x123456, %<reg4>
    new_instr = InstrCreateMovImmToReg(reg4 | reg_32, 0x123456);
    InsertInstrBeforeInstr(new_instr, call);

    //mov (%<reg1>), %<reg4>
    //new_instr = InstrCreateMovMemBaseToReg(reg1 | reg_32, reg4 | reg_32);
    //InsertInstrBeforeInstr(new_instr, call);
    //INSTR_SET_FLAG(new_instr, INSTR_SIGSEGV_FLAG);
    InsertUseZeroRegInstr(reg1, reg4, call);

    /* get rid of call */
    InstrDisconnect(call);
//    InstrFree(call);

    return;
}



/*
 *  JumpToSIGILL() --
 *
 *  Converts a direct unconditonal jump received as a parameter to the sequence
 *  of instructions that
 *    - reserve 4 bytes on the stack for the target of the original jump,
 *      (fault address will be written there by the handler and used by
 *      restore to derive, using the perfect hashing, the target address,
 *      which will then be placed there)
 *    - save the state of the registers before jump on the stack,
 *    - cause a SIGILL signal to be raised.
 *  As a result the handler for the SIGILL signal is called, which modifies
 *  the address of the faulting instruction on the stack with the address
 *  of the restore block. The restore block then restores the information
 *  from the stack and returns to the target of the original jump.
 */
void JumpToSIGILL(Instruction *jump) {

    if (rand() % 2) {
        InsertInstrDecStack(sizeof(ADDRESS), jump);
        InsertInstrSetCommunicationFlag(jump);
    } else {
        InsertInstrSetCommunicationFlag(jump);
        InsertInstrDecStack(sizeof(ADDRESS), jump);
    }

    InsertIllegalInstr(jump);

    /* get rid of jump */
    InstrDisconnect(jump);
//    InstrFree(jump);

    return;
}



/*
 *  JumpToSIGFPE() --
 *
 *  Converts a direct unconditonal jump received as a parameter to the sequence
 *  of instructions that
 *    - reserve 4 bytes on the stack for the target of the original jump,
 *      (fault address will be written there by the handler and used by
 *      restore to derive, using the perfect hashing, the target address,
 *      which will then be placed there)
 *    - save the state of the registers before jump on the stack,
 *    - cause a SIGFPE signal to be raised (divide by zero).
 *  As a result the handler for the SIGFPE signal is called, which modifies
 *  the address of the faulting instruction on the stack with the address
 *  of the restore block. The restore block then restores the information
 *  from the stack and returns to the target of the original jump.
 */
/*
void JumpToSIGFPE(Instruction *jump) {
    Instruction *new_instr;

    InsertInstrDecStack(sizeof(ADDRESS), jump);
    InsertInstrSetCommunicationFlag(jump);

    //might want to pick the below register randomly later on
    //pick idiv or div randomly, too
    //mov 0, div32 <--> shl 8 (mul 256), div8

    //this will change to load the opaque value into the chosen register
    //such that this value is always 0.
    //mov $0x0, %ebx
    new_instr = InstrCreateMovImmToReg(reg_b | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, jump);

    //idiv %ebx
    new_instr = InstrCreate32bitIDivByReg(reg_b | reg_32);
    InsertInstrBeforeInstr(new_instr, jump);
    INSTR_SET_FLAG(new_instr, INSTR_SIGFPE_FLAG);
*/
    /* get rid of jump */
/*
    InstrDisconnect(jump);
//    InstrFree(jump);

    return;
}
*/


/*
 *  JumpToSIGSEGV() --
 *
 *  Converts a direct unconditonal jump received as a parameter to the sequence
 *  of instructions that
 *    - reserve 4 bytes on the stack for the target of the original jump,
 *      (fault address will be written there by the handler and used by
 *      restore to derive, using the perfect hashing, the target address,
 *      which will then be placed there)
 *    - save the state of the registers before jump on the stack,
 *    - cause a SIGSEGV signal to be raised (reference illegal memory).
 *  As a result the handler for the SIGSEGV signal is called, which modifies
 *  the address of the faulting instruction on the stack with the address
 *  of the restore block. The restore block then restores the information
 *  from the stack and returns to the target of the original jump.
 */
/*
void JumpToSIGSEGV(Instruction *jump) {
    Instruction *new_instr;
    UINT32 reg_src = reg_a;
    UINT32 reg_dest = reg_b;

//    regs[0] = reg_a;
//    regs[1] = reg_b;
//    regs[2] = reg_c;
//    regs[3] = reg_d;

//    reg_src = regs[rand() % 4];
//    reg_dest = regs[rand() %4];

    InsertInstrDecStack(sizeof(ADDRESS), jump);
    InsertInstrSetCommunicationFlag(jump);

    //might want to pick the below two registers randomly later on.

    //this will change to load the opaque value into the chosen register
    //such that this value is always beyond the memory addresses available
    //for the program.    //LoadIllegalAddressIntoReg(reg_src, jump);
    //mov $0x0, %eax
    new_instr = InstrCreateMovImmToReg(reg_src | reg_32, 0x0);
    ////shr $0x20, %eax
    //new_instr = InstrCreate32BitShrImm(reg_a | reg_32, rand() % 31 + 1);
    InsertInstrBeforeInstr(new_instr, jump);

    //this will change to randomly pick the instruction that uses the illegal
    //memory: add, sub, mov, ... whatever //IndirectRegister(reg_src, reg_dest);
    //mov (%eax), %ebx
    new_instr = InstrCreateMovMemBaseToReg(reg_src | reg_32, reg_dest | reg_32);

    InsertInstrBeforeInstr(new_instr, jump);
    INSTR_SET_FLAG(new_instr, INSTR_SIGSEGV_FLAG);
*/
    /* get rid of jump */
/*
    InstrDisconnect(jump);
//    InstrFree(jump);

    return;
}
*/


/*
 *  CallToSIGILL() --
 *
 *  Converts a direct call received as a parameter to the sequence
 *  of instructions that
 *    - reserve 8 bytes on the stack for
 *      1. the target of the original call
 *          (fault address will be written there by the handler and used by
 *          restore to derive, using the perfect hashing, the target address,
 *          which will then be placed there),
 *      2. the return address
 *          (derived from the fault address using the perfect hashing and
 *          placed there by restore)
 *    - save the state of the registers before call on the stack,
 *    - cause a SIGILL signal to be raised.
 *  As a result the handler for the SIGILL signal is called, which modifies
 *  the address of the faulting instruction on the stack with the address
 *  of the restore block. The restore block then restores the information
 *  from the stack and returns to the target of the original call, at which
 *  point the top of the stack will contain the return address of the original
 *  call.
 */
void CallToSIGILL(Instruction *call) {

    if (!CallEdgesExist(call)) {
        return;
    }

    if (rand() % 2) {
        InsertInstrDecStack(2 * sizeof(ADDRESS), call);
        InsertInstrSetCommunicationFlag(call);
    } else {
        InsertInstrSetCommunicationFlag(call);
        InsertInstrDecStack(2 * sizeof(ADDRESS), call);
    }

    InsertIllegalInstr(call);

    /* get rid of call */
    InstrDisconnect(call);
//    InstrFree(call);

    return;
}



/*
 *  Converts a direct call received as a parameter to the sequence
 *  of instructions that
 *    - reserve 8 bytes on the stack for
 *      1. the target of the original call
 *          (fault address will be written there by the handler and used by
 *          restore to derive, using the perfect hashing, the target address,
 *          which will then be placed there),
 *      2. the return address
 *          (derived from the fault address using the perfect hashing and
 *          placed there by restore)
 *    - save the state of the registers before call on the stack,
 *    - cause a SIGFPE signal to be raised (divide by zero).
 *  As a result the handler for the SIGFPE signal is called, which modifies
 *  the address of the faulting instruction on the stack with the address
 *  of the restore block. The restore block then restores the information
 *  from the stack and returns to the target of the original call, at which
 *  point the top of the stack will contain the return address of the original
 *  call.
 */
/*
void CallToSIGFPE(Instruction *call) {
    Instruction *new_instr;

    if (!CallEdgesExist(call)) {
        return;
    }

    InsertInstrDecStack(2 * sizeof(ADDRESS), call);
    InsertInstrSetCommunicationFlag(call);

    //might want to pick the below register randomly later on
    //pick idiv or div randomly, too
    //mov 0, div32 <--> shl 8 (mul 256), div8

    //this will change to load the opaque value into the chosen register
    //such that this value is always 0.
    //mov $0x0, %ebx
    new_instr = InstrCreateMovImmToReg(reg_b | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, call);

    //idiv %ebx
    new_instr = InstrCreate32bitIDivByReg(reg_b | reg_32);
    InsertInstrBeforeInstr(new_instr, call);
    INSTR_SET_FLAG(new_instr, INSTR_SIGFPE_FLAG);
*/
    /* get rid of call */
/*
    InstrDisconnect(call);
//    InstrFree(call);

    return;
}
*/


/*
 *  Converts a direct call received as a parameter to the sequence
 *  of instructions that
 *    - reserve 8 bytes on the stack for
 *      1. the target of the original call
 *          (fault address will be written there by the handler and used by
 *          restore to derive, using the perfect hashing, the target address,
 *          which will then be placed there),
 *      2. the return address
 *          (derived from the fault address using the perfect hashing and
 *          placed there by restore)
 *    - save the state of the registers before call on the stack,
 *    - cause a SIGSEGV signal to be raised (reference illegal memory).
 *  As a result the handler for the SIGSEGV signal is called, which modifies
 *  the address of the faulting instruction on the stack with the address
 *  of the restore block. The restore block then restores the information
 *  from the stack and returns to the target of the original call, at which
 *  point the top of the stack will contain the return address of the original
 *  call.
 */
/*
void CallToSIGSEGV(Instruction *call) {
    Instruction *new_instr;
    UINT32 reg_src = reg_a;
    UINT32 reg_dest = reg_b;

//    regs[0] = reg_a;
//    regs[1] = reg_b;
//    regs[2] = reg_c;
//    regs[3] = reg_d;

//    reg_src = regs[rand() % 4];
//    reg_dest = regs[rand() %4];

    if (!CallEdgesExist(call)) {
        return;
    }

    InsertInstrDecStack(2 * sizeof(ADDRESS), call);
    InsertInstrSetCommunicationFlag(call);

    //might want to pick the below two registers randomly later on.

    //this will change to load the opaque value into the chosen register
    //such that this value is always beyond the memory addresses available
    //for the program.    //LoadIllegalAddressIntoReg(reg_src, jump);
    //mov $0x0, %eax
    new_instr = InstrCreateMovImmToReg(reg_src | reg_32, 0x0);
    ////shr $0x20, %eax
    //new_instr = InstrCreate32BitShrImm(reg_a | reg_32, rand() % 31 + 1);
    InsertInstrBeforeInstr(new_instr, call);

    //this will change to randomly pick the instruction that uses the illegal
    //memory: add, sub, mov, ... whatever //IndirectRegister(reg_src, reg_dest);
    //mov (%eax), %ebx
    new_instr = InstrCreateMovMemBaseToReg(reg_src | reg_32, reg_dest | reg_32);

    InsertInstrBeforeInstr(new_instr, call);
    INSTR_SET_FLAG(new_instr, INSTR_SIGSEGV_FLAG);
*/
    /* get rid of jump */
/*
    InstrDisconnect(call);
//    InstrFree(call);

    return;
}
*/


/*
 *  Converts a return received as a parameter to the sequence
 *  of instructions that
 *    - reserve 4 bytes on the stack,
 *      (fault address will be written there by the handler and used by
 *      restore to determine, using the perfect hashing, what instruction
 *      was here originally)
 *    - save the state of the registers before return on the stack,
 *    - cause a SIGILL signal to be raised.
 *  As a result the handler for the SIGILL signal is called, which modifies
 *  the address of the faulting instruction on the stack with the address
 *  of the restore block. The restore block then restores the information
 *  from the stack and returns to the target of the original return.
 */
void RetToSIGILL(Instruction *ret) {


    if (rand() % 2) {
        InsertInstrDecStack(sizeof(ADDRESS), ret);
        InsertInstrSetCommunicationFlag(ret);
    } else {
        InsertInstrSetCommunicationFlag(ret);
        InsertInstrDecStack(sizeof(ADDRESS), ret);
    }

    InsertIllegalInstr(ret);

    /* get rid of ret */
    InstrDisconnect(ret);
//    InstrFree(ret);

    return;
}



/*
 *  RetToSIGFPE() --
 *
 *  Converts a return received as a parameter to the sequence
 *  of instructions that
 *    - reserve 4 bytes on the stack,
 *      (fault address will be written there by the handler and used by
 *      restore to determine, using the perfect hashing, what instruction
 *      was here originally)
 *    - save the state of the registers before return on the stack,
 *    - cause a SIGFPE signal to be raised (divide by zero).
 *  As a result the handler for the SIGFPE signal is called, which modifies
 *  the address of the faulting instruction on the stack with the address
 *  of the restore block. The restore block then restores the information
 *  from the stack and returns to the target of the original return.
 */
/*
void RetToSIGFPE(Instruction *ret) {
    Instruction *new_instr;

    InsertInstrDecStack(sizeof(ADDRESS), ret);
    InsertInstrSetCommunicationFlag(ret);

    //might want to pick the below register randomly later on
    //pick idiv or div randomly, too
    //mov 0, div32 <--> shl 8 (mul 256), div8

    //this will change to load the opaque value into the chosen register
    //such that this value is always 0.
    //mov $0x0, %ebx
    new_instr = InstrCreateMovImmToReg(reg_b | reg_32, 0x0);
    InsertInstrBeforeInstr(new_instr, ret);

    //idiv %ebx
    new_instr = InstrCreate32bitIDivByReg(reg_b | reg_32);
    InsertInstrBeforeInstr(new_instr, ret);
    INSTR_SET_FLAG(new_instr, INSTR_SIGFPE_FLAG);
*/
    /* get rid of ret */
/*
    InstrDisconnect(ret);
//    InstrFree(ret);

    return;
}
*/


/*
 *  RetToSIGSEGV() --
 *
 *  Converts a return received as a parameter to the sequence
 *  of instructions that
 *    - reserve 4 bytes on the stack,
 *      (fault address will be written there by the handler and used by
 *      restore to determine, using the perfect hashing, what instruction
 *      was here originally)
 *    - save the state of the registers before return on the stack,
 *    - cause a SIGSEGV signal to be raised (reference illegal memory).
 *  As a result the handler for the SIGSEGV signal is called, which modifies
 *  the address of the faulting instruction on the stack with the address
 *  of the restore block. The restore block then restores the information
 *  from the stack and returns to the target of the original return.
 */
/*
void RetToSIGSEGV(Instruction *ret) {
    Instruction *new_instr;
    UINT32 reg_src = reg_a;
    UINT32 reg_dest = reg_b;

//    regs[0] = reg_a;
//    regs[1] = reg_b;
//    regs[2] = reg_c;
//    regs[3] = reg_d;

//    reg_src = regs[rand() % 4];
//    reg_dest = regs[rand() %4];

    InsertInstrDecStack(sizeof(ADDRESS), ret);
    InsertInstrSetCommunicationFlag(ret);

    //might want to pick the below two registers randomly later on.

    //this will change to load the opaque value into the chosen register
    //such that this value is always beyond the memory addresses available
    //for the program.    //LoadIllegalAddressIntoReg(reg_src, jump);
    //mov $0x0, %eax
    new_instr = InstrCreateMovImmToReg(reg_src | reg_32, 0x0);
    ////shr $0x20, %eax
    //new_instr = InstrCreate32BitShrImm(reg_a | reg_32, rand() % 31 + 1);
    InsertInstrBeforeInstr(new_instr, ret);

    //this will change to randomly pick the instruction that uses the illegal
    //memory: add, sub, mov, ... whatever //IndirectRegister(reg_src, reg_dest);
    //mov (%eax), %ebx
    new_instr = InstrCreateMovMemBaseToReg(reg_src | reg_32, reg_dest | reg_32);

    InsertInstrBeforeInstr(new_instr, ret);
    INSTR_SET_FLAG(new_instr, INSTR_SIGSEGV_FLAG);
*/
    /* get rid of ret */
/*
    InstrDisconnect(ret);
//    InstrFree(ret);

    return;
}
*/


/*
 *  InsertInstrDecStack(num_bytes, next_instr) --
 *
 *  Creates one or two instructions that will decrement the %esp by the
 *  num_bytes bytes. Instruction(s) created are inserted into the binary before
 *  the next_instr.
 */
void InsertInstrDecStack(int num_bytes, Instruction *next_instr) {
    Instruction *new_instr;
    int pick;
    
    //trap replaces jump or ret
    if (num_bytes == 4) {
        InsertInstrPush(next_instr);

    //trap replaces call
    } else { // num_bytes == 8
        pick = rand() % 11;
        if (pick == 0) {
            new_instr = InstrCreateSubImm8FromReg(0x8, reg_sp | reg_32);
            InsertInstrBeforeInstr(new_instr, next_instr);
        } else {
            InsertInstrPush(next_instr);
            InsertInstrPush(next_instr);
        }
    }

    return;
}



/*
 *  InsertInstrPush(next_instr) --
 *
 *  Creates a push instruction. Has eight possible choices for push. Six of
 *  them push a register, one pushes immediate, and one pushes the contents of
 *  memory. One of the eight different pushes is chosen with about the same
 *  probability with which they appear in specint2000.
 */
void InsertInstrPush(Instruction *next_instr) {
    Instruction *new_instr;
    int pick = rand() % 33;

    switch (pick) {
        case 0:
        case 1:
        case 2:
        case 3:
            new_instr = InstrCreatePush(reg_a | reg_32);
            break;
        case 4:
        case 5:
        case 6:
        case 7:
        case 8:
        case 9:
            new_instr = InstrCreatePush(reg_b | reg_32);
            break;
        case 10:
        case 11:
            new_instr = InstrCreatePush(reg_c | reg_32);
            break;
        case 12:
        case 13:
            new_instr = InstrCreatePush(reg_d | reg_32);
            break;
        case 14:
        case 15:
        case 16:
        case 17:
            new_instr = InstrCreatePush(reg_si | reg_32);
            break;
        case 18:
        case 19:
        case 20:
            new_instr = InstrCreatePush(reg_di | reg_32);
            break;
        case 21: {
            Patch *upm;
            PatchSectionToInstr *patch;
            LinkedList *upms;

            //the memory referenced will be in the .data section
            new_instr = xInstrCreatePushMem(0x0);
            AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, 0,
                new_instr, new_instr->i_src);

            upms = UPMTableGetByDest(new_instr->i_src);
            upm = ll_get(upms, 0);
            ll_free(upms);
            if (upm == NULL) {
                MALFUNCTION("UPM was not properly cloned");
            }
            patch = upm->p_patch;
            patch->p_src_offset += (rand() %
                (SECTION_REF_DATA->s_newdatasize / 4)) * sizeof(ADDRESS);

            break;
        }
        case 22:
        case 23:
        case 24:
        case 25:
        case 26:
        case 27:
        case 28:
        case 29:
        case 30:
        case 31:
        case 32: {
            Patch *upm;
            PatchSectionToInstr *patch;
            LinkedList *upms;

            //the immediate is some address in the .data section
            new_instr = xInstrCreatePushImm(0x0);
            AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, 0,
                new_instr, new_instr->i_src);

            upms = UPMTableGetByDest(new_instr->i_src);
            upm = ll_get(upms, 0);
            ll_free(upms);
            if (upm == NULL) {
                MALFUNCTION("UPM was not properly cloned");
            }
            patch = upm->p_patch;
            patch->p_src_offset += (rand() %
                (SECTION_REF_DATA->s_newdatasize / 4)) * sizeof(ADDRESS);

            break;
        }
    }
    InsertInstrBeforeInstr(new_instr, next_instr);

    return;
}



/*
 *  InsertInstrSetCommunicationFlag(next_instr) --
 *
 *  Creates a mov instruction that sets one, randomly chosen, communication
 *  flag to some randomly chosen non-zero value. Instruction created is
 *  inserted into the binary before the next_instr.
 */
void InsertInstrSetCommunicationFlag(Instruction *next_instr) {
    Patch *upm;
    PatchSectionToInstr *patch;
    LinkedList *upms;
    int flag_value = rand() % RAND_MAX + 1;
    Instruction *new_instr = InstrCreate32BitMovImmToMem(0, flag_value);

    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, raised_from, new_instr,
        new_instr->i_dest);
    upms = UPMTableGetByDest(new_instr->i_dest);
    upm = ll_get(upms, 0);
    ll_free(upms);
    if (upm == NULL) {
        MALFUNCTION("UPM was not properly cloned");
    }
    patch = upm->p_patch;
    patch->p_src_offset += (rand() % NUMBER_OF_COMMUNICATION_FLAGS) *
        sizeof(ADDRESS);

    InsertInstrBeforeInstr(new_instr, next_instr);

    return;
}



/*
 *  InsertUseZeroRegInstr(zero_reg, dest_reg, next_instr) --
 *
 *  Creates one of the two instructions that uses the zero_reg. First choice
 *  is mov that dereferences the zero_reg. Execution of it at runtime will
 *  raise a SIGSEGV signal. dest_reg is the destination of this mov
 *  instruction. Second choice is div/idiv by zero_reg. Execution of it at
 *  runtime will raise a SIGFPE signal. In both cases new instruction is
 *  inserted before the next_instr.
 *  TO DO: There are many possible ways to dereference zero_reg. Could use
 *  almost any instruction that has this addressing mode. cmp and test look
 *  very appealing since the next instruction will be conditional jump. Very
 *  frequently used pair. Adds stealth.
 */
void InsertUseZeroRegInstr(UINT32 zero_reg, UINT32 dest_reg,
    Instruction *next_instr) {

    Instruction *new_instr;

    if (rand() % 20) {

        //mov (%<zero_reg>), %<dest_reg>
        new_instr = InstrCreateMovMemBaseToReg(zero_reg | reg_32,
            dest_reg | reg_32);
        InsertInstrBeforeInstr(new_instr, next_instr);
        INSTR_SET_FLAG(new_instr, INSTR_SIGSEGV_FLAG);

    } else {

        if (rand() % 2) {
            //idiv %<zero_reg>
            new_instr = InstrCreate32bitIDivByReg(zero_reg | reg_32);
        } else {
            //div %<zero_reg>
            new_instr = InstrCreate32bitDivByReg(zero_reg | reg_32);
        }
        InsertInstrBeforeInstr(new_instr, next_instr);
        INSTR_SET_FLAG(new_instr, INSTR_SIGFPE_FLAG);

    }

    return;
}



/*
 *  InsertIllegalInstr(next_instr) --
 *
 *  Creates an illegal instruction. Execution of such illegal instruction at
 *  runtime will raise a SIGILL signal. Instruction created is inserted into
 *  the binary before the next_instr.
 */
void InsertIllegalInstr(Instruction *next_instr) {
    Instruction *new_instr;
    UINT8 buffer[4];

    //TO DO:
    //pick the illegal instruction (not just second byte) randomly later on
    buffer[0] = 0x0f;
    buffer[1] = PickUndefinedInstrByte2();

    new_instr = CreateCustomInstr(buffer, 2);
    InsertInstrBeforeInstr(new_instr, next_instr);
    INSTR_SET_FLAG(new_instr, INSTR_SIGILL_FLAG);

    return;
}



/*
 *  PickUndefinedInstrByte2() --
 *
 *  Randomly picks and returns the second byte of the illegal instruction
 *  starting at 0f.
 */
UINT8 PickUndefinedInstrByte2() {
    UINT8 instr[25];
    int pick;

    instr[0] = 0x04;
    instr[1] = 0x05;
    instr[2] = 0x07;
    instr[3] = 0x0a;
    instr[4] = 0x0c;
    instr[5] = 0x0d;
    instr[6] = 0x0e;
    instr[7] = 0x0f;
    instr[8] = 0x25;
    instr[9] = 0x27;
    instr[10] = 0x36;
    instr[11] = 0x37;
    instr[12] = 0x38;
    instr[13] = 0x39;
    instr[14] = 0x3a;
    instr[15] = 0x3b;
    instr[16] = 0x3d;
    instr[17] = 0x3e;
    instr[18] = 0x3f;
    instr[19] = 0xa6;
    instr[20] = 0xa7;
    instr[21] = 0xb8;
    instr[22] = 0xd0;
    instr[23] = 0xf0;
    instr[24] = 0xff;

    pick = rand() % 25;

    return instr[pick];
}

