/*
 *  convert.c
 */
#include "protos.h"
#include "constants.h"
#include "utilities.h"
#include "linkage.h"
#include "hash.h"
#include "relocate.h"

/*
 *  Forward declarations and external symbols
 */
static Symbol *GetCIdentSectionSymbol(HashTable *symbol_hash_table,
    Section *section, bool start);
static void UpdateSymbolsForVirtualAddresses();
static void ResolveSpecialCaseSymbols();
static Section *SectionIndexToSection(Section *section_list, int elf_index);
static Symbol *SymbolIndexToSymbol(Symbol *symbols, int elf_index);
static Relocation *GetPltoRelocations(Section *sections, Symbol *symbols);
static Symbol *GetPltoSymbols(Elf32File *elf_file, Section *sections);
static Relocation *GetFakePltoRelocations(Elf32File *elf_file,
    Section *sections);



/*
 *  UpdateSymbolsForVirtualAddresses() --
 *
 *  Updates all symbols to use their virtual values instead of their
 *  section-offset values.
 */
static void UpdateSymbolsForVirtualAddresses() {
    Symbol *symiterator;

    FOREACH_SYMBOL(symiterator) {
        if (!strcmp(symiterator->sy_name, "_etext") ||
            !strcmp(symiterator->sy_name, "_edata")) {
            continue;       
        }
	if (symiterator->sy_section != NULL) {
	    symiterator->sy_value += symiterator->sy_section->s_address;
	} 
    }
}



/*
 *  GetCIdentSectionSymbol(symbol_hash_table,section,start) --
 *
 *  Returns the __start or __stop symbol for a c-identifier style section.
 *
 *  symbol_hash_table: a hash table of all Symbols
 *  section: the Section for which the symbol should be identified
 *  start: true if the start symbol should be extracted, false if
 *  the stop symbol should be extracted
 */
static Symbol *GetCIdentSectionSymbol(HashTable *symbol_hash_table,
    Section *section, bool start) {
    Symbol *sym;
    char *symbol_name, *prefix;

    prefix = start ? CIDENT_SECTION_START_PREFIX : CIDENT_SECTION_STOP_PREFIX;

    symbol_name =
	zalloc(sizeof(char *) * (strlen(section->s_name) + strlen(prefix)));
    sprintf(symbol_name, "%s%s", prefix, section->s_name);

    sym = (Symbol *) ht_get(symbol_hash_table, symbol_name);
    if (sym == NULL) {
	sym = zalloc(sizeof(Symbol));
	sym->sy_name = salloc(symbol_name);
    }

    sym->sy_section = section;
    sym->sy_flags = 0;
    sym->sy_flags |= SYMBOL_GLOBAL_FLAG;
    sym->sy_flags |= SYMBOL_OBJECT_FLAG;
    sym->sy_flags |= SYMBOL_ABSOLUTE_FLAG;
    sym->sy_value = 0;

    if (!start) {
	sym->sy_value += section->s_datasize;
	sym->sy_flags |= SYMBOL_OBJECT_END_FLAG;
    }

    free(symbol_name);

    return sym;
}



/*
 *  ResolveSpecialCaseSymbols() --
 *
 *  Resolves values for special case symbols such as the __start_, __stop_
 *  glibc symbols.
 */
static void ResolveSpecialCaseSymbols() {
    Section *seciterator, *preinit_section, *init_section, *fini_section;
    Symbol *sym;
    HashTable *symtbl;

    REPORT(RP_MINOR, WITHFUNC("Resolving special case symbols"));

    symtbl = ht_new(HASH_STR, HASH_PTR, 0xfff, .7);

    FOREACH_SYMBOL(sym) {
	// typically it's not safe to check for symbol identity by name
	// since local symbols are allowed to have the same names,
	// however, since we only care about specific weak undefined symbols
	// we are guaranteed problem free usage of this name matching mechanism.
	ht_put(symtbl, sym->sy_name, sym);
    }

    // well gcc or libc or ld or something does this really neat thing that
    // when a section name can be represented as a c identifier, it creates
    // two symbols to reference that section -- for example, a section
    // __hello__ would have two symbols created:
    // __start___hello__ and __stop___hello, where each symbol is just
    // __start__ and __stop__ appended to the section name. The first, __start_
    // points will have the address of the section and __stop_ will have the
    // address of the section plus it's size
    //
    // rather than trying to pattern match each symbol against each section,
    // we will wait until all symbols are read in, then scan all sections for
    // valid C identifier names and then finish defining these symbols. I just
    // have to say that it kind of stinks that we have to resort to string
    // pattern matching...
    FOREACH_SECTION(seciterator) {
	if (IsValidCIdentifier(seciterator->s_name)) {
	    sym = GetCIdentSectionSymbol(symtbl, seciterator, true);
	    if (ht_get(symtbl, sym->sy_name) == NULL) {
		InsertSymbolAfterSymbol(sym, s_list);
	    }
	    REPORT(RP_SMLOOP, "Resolved %s for %s",
		sym->sy_name, seciterator->s_name);

	    sym = GetCIdentSectionSymbol(symtbl, seciterator, false);
	    if (ht_get(symtbl, sym->sy_name) == NULL) {
		InsertSymbolAfterSymbol(sym, s_list);
	    }
	    REPORT(RP_SMLOOP, "Resolved %s for %s",
		sym->sy_name, seciterator->s_name);
	}
    }					// done resolving cident style symbols

    preinit_section = init_section = fini_section = NULL;
    FOREACH_SECTION(seciterator) {
	if (strcmp(seciterator->s_name, ".preinit_array") == 0) {
	    preinit_section = seciterator;
	} else if (strcmp(seciterator->s_name, ".init_array") == 0) {
	    init_section = seciterator;
	} else if (strcmp(seciterator->s_name, ".fini_array") == 0) {
	    fini_section = seciterator;
	}
    }

    FOREACH_SYMBOL(sym) {
	if ((sym->sy_flags & SYMBOL_NOTYPE_FLAG) &&
	    (sym->sy_flags & SYMBOL_GLOBAL_FLAG) &&
	    (sym->sy_flags & SYMBOL_UNDEFINED_FLAG)) {
	    if (strcmp(sym->sy_name, "__preinit_array_start") == 0) {
		sym->sy_flags &= ~SYMBOL_UNDEFINED_FLAG;
		sym->sy_flags |= SYMBOL_ABSOLUTE_FLAG;
		if (preinit_section != NULL) {
		    sym->sy_value = preinit_section->s_address;
		    REPORT(RP_SMLOOP, "Resolved %s to %s",
			sym->sy_name, preinit_section->s_name);
		} else {
		    sym->sy_value = SECTION_REF_DATA->s_address;
		    REPORT(RP_SMLOOP, "Resolved %s to %s",
			sym->sy_name, SECTION_REF_DATA->s_name);
		}
	    } else if (strcmp(sym->sy_name, "__preinit_array_end") == 0) {
		sym->sy_flags &= ~SYMBOL_UNDEFINED_FLAG;
		sym->sy_flags |= SYMBOL_ABSOLUTE_FLAG;
		if (preinit_section != NULL) {
		    sym->sy_value = preinit_section->s_address +
			preinit_section->s_datasize;
		    REPORT(RP_SMLOOP, "Resolved %s to end of %s",
			sym->sy_name, preinit_section->s_name);
		} else {
		    sym->sy_value = SECTION_REF_DATA->s_address;
		    REPORT(RP_SMLOOP, "Resolved %s to %s",
			sym->sy_name, SECTION_REF_DATA->s_name);
		}
	    } else if (strcmp(sym->sy_name, "__init_array_start") == 0) {
		sym->sy_flags &= ~SYMBOL_UNDEFINED_FLAG;
		sym->sy_flags |= SYMBOL_ABSOLUTE_FLAG;
		if (preinit_section != NULL) {
		    sym->sy_value = init_section->s_address;
		    REPORT(RP_SMLOOP, "Resolved %s to %s",
			sym->sy_name, init_section->s_name);
		} else {
		    sym->sy_value = SECTION_REF_DATA->s_address;
		    REPORT(RP_SMLOOP, "Resolved %s to %s",
			sym->sy_name, SECTION_REF_DATA->s_name);
		}
	    } else if (strcmp(sym->sy_name, "__init_array_end") == 0) {
		sym->sy_flags &= ~SYMBOL_UNDEFINED_FLAG;
		sym->sy_flags |= SYMBOL_ABSOLUTE_FLAG;
		if (preinit_section != NULL) {
		    sym->sy_value =
			init_section->s_address + init_section->s_datasize;
		    REPORT(RP_SMLOOP, "Resolved %s to end of %s",
			sym->sy_name, init_section->s_name);
		} else {
		    sym->sy_value = SECTION_REF_DATA->s_address;
		    REPORT(RP_SMLOOP, "Resolved %s to %s",
			sym->sy_name, SECTION_REF_DATA->s_name);
		}
	    } else if (strcmp(sym->sy_name, "__fini_array_start") == 0) {
		sym->sy_flags &= ~SYMBOL_UNDEFINED_FLAG;
		sym->sy_flags |= SYMBOL_ABSOLUTE_FLAG;
		if (preinit_section != NULL) {
		    sym->sy_value = fini_section->s_address;
		    REPORT(RP_SMLOOP, "Resolved %s to %s",
			sym->sy_name, fini_section->s_name);
		} else {
		    sym->sy_value = SECTION_REF_DATA->s_address;
		    REPORT(RP_SMLOOP, "Resolved %s to %s",
			sym->sy_name, SECTION_REF_DATA->s_name);
		}
	    } else if (strcmp(sym->sy_name, "__fini_array_end") == 0) {
		sym->sy_flags &= ~SYMBOL_UNDEFINED_FLAG;
		sym->sy_flags |= SYMBOL_ABSOLUTE_FLAG;
		if (preinit_section != NULL) {
		    sym->sy_value =
			fini_section->s_address + fini_section->s_datasize;
		    REPORT(RP_SMLOOP, "Resolved %s to end of %s",
			sym->sy_name, fini_section->s_name);
		} else {
		    sym->sy_value = SECTION_REF_DATA->s_address;
		    REPORT(RP_SMLOOP, "Resolved %s to %s",
			sym->sy_name, SECTION_REF_DATA->s_name);
		}
	    }
	}
    }

    ht_clear(symtbl);
    ht_free(symtbl);
}



/*
 *  SectionIndexToSection(section_list,elf_index) --
 *
 *  Returns the Section whose original Elf_Shdr section header index was
 *  elf_index. This is done by checking the s_misc field of each Section
 *  in section_list for a value matching elf_index.  During GetPltoSections
 *  each Section's s_misc value is set to its original ELF section header index.
 *
 *  section_list: the list of sections to check in for elf_index
 *  elf_index: the index of the section to retrieve
 *  returns: the Section whose original section index (in the ELF file) was
 *  elf_index
 */
static Section *SectionIndexToSection(Section *section_list, int elf_index) {
    while (section_list != NULL) {
	if (section_list->s_misc == elf_index) {
	    return section_list;
	}
	section_list = section_list->s_next;
    }
    return NULL;
}



/*
 *  SymbolIndexToSymbol(symbols,elf_index) --
 *
 *  Returns the Symbol whose original ELF symbol table index was
 *  elf_index. This is done by checking to see if the symbol's misc field is set
 *  to elf_index. Each symbol's misc field is set to it's original ELF symbol
 *  index during GetPltoSymbols.
 *
 *  symbols: the list of symbols to search through
 *  elf_index: the index of the Elf32_Sym whose Symbol we want
 *  returns: the Symbol whose original Elf32_Sym symbol had elf_index as
 *  it's symbol table index
 */
static Symbol *SymbolIndexToSymbol(Symbol *symbols, int elf_index) {
    while (symbols != NULL) {
	if ((int) symbols->sy_temp == elf_index) {
	    return symbols;
	}

	symbols = symbols->sy_next;
    }

    return NULL;
}



/*
 *  GetPltoSections(elf_file) --
 *
 *  Collects all meaningful sections from elf_file and converts them to the
 *  corresponding PLTO Section, returning the list of sections.
 *
 *  elf_file: the Elf32File that represents the relocatable to be read in
 *  returns: the pointer to the front of the linked list of all Sections
 */
Section *GetPltoSections(Elf32File *elf_file) {
    Section *sections, *new_section, *last_section;
    Elf32_Shdr *cur_shdr;

    sections = NULL;
    elf_file->ef_sect_map =
	zalloc(elf_file->ef_hdr->e_shnum * sizeof(Section *));

    REPORT(RP_SMLOOP, WITHFUNC("num offset  address   size flags name "));

    FOREACH_ELF32_SECTION(elf_file, cur_shdr) {

	// ignore the null section and others specially listed
	if (cur_shdr->sh_type == SHT_NULL
	    || IsIgnoredElfSection(elf_file, indexval)) {
	    continue;
	}
	// create the new section and fill in all of the necessary
	// information
	new_section = zalloc(sizeof(Section));
	elf_file->ef_sect_map[indexval] = new_section;
	new_section->s_name = salloc(&elf_file->ef_strn_tbl[cur_shdr->sh_name]);
	new_section->s_flags = GetPltoFlagsFromElf32_Shdr(cur_shdr);

	new_section->s_datasize = new_section->s_newdatasize =
	    cur_shdr->sh_size;

	if (SECTION_HAS_FLAG(new_section,
		SECTION_OCCUPIES_FILESPACE_FLAG | SECTION_RELOCATIONS_FLAG |
		SECTION_SYMBOL_TABLE_FLAG | SECTION_NOTE_FLAG)) {
	    new_section->s_data = alloc(new_section->s_datasize);
	    memcpy(new_section->s_data,
		&elf_file->ef_file_data[cur_shdr->sh_offset],
		new_section->s_datasize);
	} else {
	    new_section->s_data = NULL;
	    new_section->s_newdata = NULL;
	}

	new_section->s_address = cur_shdr->sh_addr;
	new_section->s_alignment = cur_shdr->sh_addralign;
	new_section->s_info = cur_shdr->sh_info;
	new_section->s_flags = GetPltoFlagsFromElf32_Shdr(cur_shdr);

	// we save the section's original section header index
	// so that we can use it to tie symbols and relocations
	// to it in GetPltoSymbols and GetPltoRelocations
	new_section->s_number = indexval;
	new_section->s_misc = indexval;
	new_section->s_name_offset = cur_shdr->sh_offset;

	REPORT(RP_SMLOOP, "%3d %6X %8X %6X %5X %s",
	    new_section->s_number, cur_shdr->sh_offset,
	    new_section->s_address, cur_shdr->sh_size,
	    new_section->s_flags, new_section->s_name);

	if (IsDebuggingInfoSection(elf_file, indexval)) {
	    SECTION_SET_FLAG(new_section, SECTION_DEBUGGING_FLAG);

	    // if we should exclude debugging sections
	    if (! WANT_DEBUG(DB_ODBG)) {	// if !OutputDebuggingInfo
		SECTION_SET_FLAG(new_section, SECTION_DROPPED_FLAG);
	    }
	}

	if (sections == NULL) {
	    sections = last_section = new_section;
	    new_section->s_next = new_section->s_prev = NULL;
	} else {
	    InsertSectionAfterSection(new_section, last_section);
	    last_section = new_section;
	}
    }

    return sections;
}


/*
 * maps offset to a section based on original section addresses
 * 
 * NOTE : Assumes a valid offset is going to be passed in.
 */
static Section * mapOffsetToSection(Section **sections,
    int num_sections, ADDRESS  offset) {
    Section *seciterator = 0;
    Section *next_section = 0;
    int i, j;

    for (i = 1; i < num_sections - 1; i++) {
        seciterator  = sections[i];

        if (!seciterator) {
            continue;
        }
    
        j = i + 1;
        while (!sections[j] ||
            SECTION_HAS_FLAG(sections[j], SECTION_RELOCATIONS_FLAG)) {
            j++;     
        }	
        next_section = sections[j];
        if ((offset >= seciterator->s_address - base_addr) &&
            (offset < next_section->s_address - base_addr)) {
            return seciterator;
        }
    }
    if (offset > (sections[i]->s_address - base_addr)) {
        return sections[i];
    }
    
    return NULL;
}
	

/*
 *  GetPltoRelocations(sections,symbols) --
 *
 *  Goes through each (non-ignored) relocation section in elf_file and retrieves
 *  all relocations then converts them into Relocations and also adds
 *  them to r_list (the global list of relocations).
 *
 *  elf_file: Elf32File from which the relocations should be extracted
 *  sections: list of Sections that were retrieved from elf_file
 *  symbols: list of Symbols that were retrieved from elf_file
 *  returns: pointer to the front of the linked list of all Relocations
 *  for elf_file
 */
static Relocation *GetPltoRelocations(Section *sections, Symbol *symbols) {
    Relocation *relocations, *new_relocation;
    Section *section, *seciterator;
    Symbol *symbol;
    Elf32_Rel *relocs, *cur_reloc;
    int numents, i;

    REPORT(RP_MINOR, WITHFUNC("Reading relocations"));
    relocations = NULL;

    FOREACH_SECTION_IN_LIST(sections, seciterator) {

	// if this is a relocation section
	if (SECTION_HAS_FLAG(seciterator, SECTION_RELOCATIONS_FLAG)) {
	    numents = seciterator->s_datasize / sizeof(*cur_reloc);

	    relocs = (Elf32_Rel *) seciterator->s_data;
	    section = SectionIndexToSection(sections, seciterator->s_info);

	    // if we removed this section already (debug, comment, ...) then
	    // we don't need the relocations for it
	    if (section == NULL) {
		continue;
	    }
	    // for each relocation in this section
	    REPORT(RP_SMLOOP, "%8d relocations in %s",numents,section->s_name);
	    for (i = 0; i < numents; i++) {
		cur_reloc = &relocs[i];

		symbol = SymbolIndexToSymbol(symbols,
		    ELF32_R_SYM(cur_reloc->r_info));

		// if there is no symbol then we will assume that we removed
		// the section that contained the symbol and just continue
		// along our merry little way
		if (symbol == NULL) {
		    WARNING("Relocation could not find a symbol it needed,"
			" symbol index: %d", ELF32_R_SYM(cur_reloc->r_info));
		    continue;
		}

		new_relocation = zalloc(sizeof(Relocation));
		new_relocation->r_flags = GetPltoFlagsFromElf32_Rel(cur_reloc);
		new_relocation->r_section = section;
		new_relocation->r_offset = cur_reloc->r_offset;
		new_relocation->r_symbol = symbol;
		new_relocation->r_addend =
		    loadLE(section->s_data + cur_reloc->r_offset);
		new_relocation->r_type = ELF32_R_TYPE(cur_reloc->r_info);
		new_relocation->r_next = new_relocation->r_prev = NULL;

		if (relocations == NULL) {
		    relocations = new_relocation;
		} else {
		    InsertRelocationAfterRelocation(new_relocation,
			relocations);
		}
	    }
	}
    }

    return relocations;
}



/*
 *  DiscardSection(section) -- unlink and free a section.
 *  Returns a pointer to the next section.
 */
static Section *DiscardSection(Section *section) {
    Symbol *symiterator;

    // unlink from the section list
    Section *next = section->s_next;
    if (section->s_prev != NULL) {
	section->s_prev->s_next = section->s_next;
    }
    if (section->s_next != NULL) {
	section->s_next->s_prev = section->s_prev;
    }

    // remove any references from within the symbol table
    FOREACH_SYMBOL(symiterator) {
	if (symiterator->sy_section == section) {
	    symiterator->sy_section = NULL;
	}
    }

    // before freeing data, fill with garbage value
    // to ensure no inadvertent later reuse
    memset(section->s_data, 0x22, section->s_datasize);
    free(section->s_data);
    memset(section, 0x22, sizeof(Section));
    free(section);

    return next;
}



/*
 *  DeleteSections(sections,flag) -- unlink sections matching a flag bit.
 */
static void DeleteSections(Section *sections, int flag) {
    Section *seciterator;
    for (seciterator = sections; seciterator != NULL;) {
	if (SECTION_HAS_FLAG(seciterator, flag)) {
	    seciterator = DiscardSection(seciterator);
	} else {
	    seciterator = seciterator->s_next;
	}
    }
}



/*
 *  GetPltoSymbols(elf_file,sections) --
 *
 *  Goes through the symbol table (.symtab) for elf_file and creates and returns
 *  the list of all Symbols that correspond to those Elf32_Sym symbols.
 *
 *  elf_file: Elf32File from which the symbols should be extracted
 *  sections: list of Sections that were retrieved from elf_file
 *  returns: list of all Symbols that can be extracted from Elf32_Sym
 *  structures in elf_file
 */
static Symbol *GetPltoSymbols(Elf32File *elf_file, Section *sections) {
    Symbol *symbols, *new_symbol, *prev_symbol;
    Section *seciterator, *ref_section;
    Elf32_Sym *cur_sym;
    int symcount, j;

    symbols = NULL;
    prev_symbol = NULL;

    REPORT(RP_MINOR, WITHFUNC("Converting symbol table"));

    FOREACH_SECTION_IN_LIST(sections, seciterator) {
	if (SECTION_HAS_FLAG(seciterator, SECTION_SYMBOL_TABLE_FLAG)) {

	    cur_sym = (Elf32_Sym *) seciterator->s_data;
	    symcount = seciterator->s_datasize / sizeof(*cur_sym);
	    cur_sym++;			// skip initial null entry

	    for (j = 1; j < symcount; j++, cur_sym++) {

		// don't copy over useless file symbols
		if (ELF32_ST_TYPE(cur_sym->st_info) == STT_FILE) {
		    continue;
		}

		new_symbol = zalloc(sizeof(Symbol));
		new_symbol->sy_flags = 0;
		if (cur_sym->st_shndx < SHN_LORESERVE) {
		    // sectmap maps section number to the correct Section struct
		    ref_section = elf_file->ef_sect_map[cur_sym->st_shndx];
		} else {
		    ref_section = NULL;
		    if (cur_sym->st_shndx == SHN_ABS) {
			if (cur_sym->st_value > base_addr) {
			    cur_sym->st_value -= base_addr;	
			    ref_section = mapOffsetToSection(
				elf_file->ef_sect_map,
				elf_file->ef_hdr->e_shnum,
				cur_sym->st_value);
			    if (ref_section) {
				cur_sym->st_value +=
				    base_addr - ref_section->s_address;
			    }
			}
			new_symbol->sy_flags |= SYMBOL_ABSOLUTE_FLAG;
		    }
		}

		new_symbol->sy_value = cur_sym->st_value;
		new_symbol->sy_size = cur_sym->st_size;


	// PATCH FOR MERGED SECTIONS -- Thank you !!!
		new_symbol->sy_section = ref_section;
		if (ref_section != NULL) {
		    new_symbol->sy_value += ref_section->s_delta;
		    if (ref_section->s_label == NULL) {
			ref_section->s_label = new_symbol;
		    }
		    if (ref_section->s_mergeto != NULL) {
			new_symbol->sy_section = ref_section->s_mergeto;
		    }
		}
		// we assign the symbol the index it had in the ELF file
		// (we use this later to link relocations up to their
		// respective symbols so as not to rely on name pattern
		// matching which is unreliable since local symbols can
		// have the same names...)
		new_symbol->sy_temp = (void *) j;

		if (new_symbol->sy_section == NULL) {
		    switch (cur_sym->st_shndx) {
		    case SHN_LORESERVE:	// also SHN_LOPROC
		    case SHN_HIPROC:
		    case SHN_HIRESERVE:
			break;
		    case SHN_ABS:
			new_symbol->sy_flags |= SYMBOL_ABSOLUTE_FLAG;
			break;
		    case SHN_UNDEF:
			new_symbol->sy_flags |= SYMBOL_UNDEFINED_FLAG;
			break;
		    case SHN_COMMON:
			new_symbol->sy_flags |= SYMBOL_COMMON_FLAG;
			break;
		    default:
			// if we are missing the section that this refers to
			// then we will assume that it was a section or
			// warning symbol which is fine and we can just
			// discard the symbol (i.e., not copy it into a
			// respective PLTO data structure)
			free(new_symbol);
			continue;
		    }
		}
		// we need to know what binding this symbol needs in order
		// to figure out how to handle it later
		switch (ELF32_ST_BIND(cur_sym->st_info)) {
		case STB_LOCAL:
		    new_symbol->sy_flags |= SYMBOL_LOCAL_FLAG;
		    break;
		case STB_GLOBAL:
		    new_symbol->sy_flags |= SYMBOL_GLOBAL_FLAG;
		    break;
		case STB_WEAK:
		    new_symbol->sy_flags |= SYMBOL_WEAK_FLAG;
		    break;
		}

		// figure out what type of symbol this is
		switch (ELF32_ST_TYPE(cur_sym->st_info)) {
		case STT_SECTION:
		    new_symbol->sy_flags |= SYMBOL_SECTION_FLAG;
		    break;
		case STT_OBJECT:
		    new_symbol->sy_flags |= SYMBOL_OBJECT_FLAG;
		    break;
		case STT_FUNC:
		    new_symbol->sy_flags |= SYMBOL_FUNCTION_FLAG;
		    break;
		case STT_NOTYPE:
		    new_symbol->sy_flags |= SYMBOL_NOTYPE_FLAG;
		    break;
		}

		// get the name of the symbol: if it is a section symbol
		// then the name must be retreived from the section header
		// string table, otherwise the name
		// can be retreived from the symbol string table
		if (new_symbol->sy_flags & SYMBOL_SECTION_FLAG) {
		    new_symbol->sy_name = salloc(ref_section->s_name);
		    // temporarily turn a merged-section label into a
		    // "normal" symbol
		    if (new_symbol == ref_section->s_label) {
			new_symbol->sy_flags &= ~SYMBOL_SECTION_FLAG;
			new_symbol->sy_flags |= SYMBOL_NOTYPE_FLAG;
		    }
		} else {
		    new_symbol->sy_name =
			salloc(&elf_file->ef_sym_strn_tbl[cur_sym->st_name]);
		}


		// check to see if this is the symbol for the global offset
		// table (.got) section.  if it is, we need to set its
		// section reference to the global offset table
		if (strcmp(new_symbol->sy_name, GOT_NAME) == 0) {
		    new_symbol->sy_section = SECTION_REF_GOT;
		    new_symbol->sy_flags &= ~SYMBOL_UNDEFINED_FLAG;
		}
		// add the new symbol to the list of symbols
		if (symbols == NULL) {
		    symbols = new_symbol;
		    symbols->sy_next = NULL;
		    symbols->sy_prev = NULL;
		} else {
		    InsertSymbolAfterSymbol(new_symbol, prev_symbol);
		}
		prev_symbol = new_symbol;

	    }				// end symbol loop
	}
    }

    return symbols;
}



/*
 *  LabelFragments(mrg_list,sym_list) --
 *
 *  Finds the first symbol associated with each merged-out section,
 *  and links it to the section header.
 */
static void LabelFragments(Section *mrg_list, Symbol *sym_list) {
    Section *sec;
    Symbol *sym;

    FOREACH_SECTION_IN_LIST(mrg_list, sec) {
	FOREACH_SYMBOL_IN_LIST(sym_list, sym) {
	    if (sym->sy_section == SECTION_REF_CODE
		&& sym->sy_value >= sec->s_delta) {
		sec->s_label = sym;
		break;
	    }
	}
    }
}



/*
 *  GetFakePltoRelocations(elf_file,sections) --
 *
 *  Given an Elf32File and a list of sections, this function scans the
 *  file for pseudo-relocation sections inserted on a previous execution
 *  and returns the Relocation list containing the equivalent
 *  relocations.
 *
 *  elf_file: the Elf32File to look in
 *  sections: the list of Sections to associate any found
 *  relocations with
 */
static Relocation *GetFakePltoRelocations(Elf32File *elf_file,
    Section *sections) {
    Relocation *relocations, *new_relocation;
    Section *section;
    Elf32_Shdr *cur_shdr;
    Elf32_Rel *relocs, *cur_reloc;
    int numents, i;

    REPORT(RP_MINOR, WITHFUNC("Reading fake relocations"));
    relocations = NULL;

    FOREACH_ELF32_SECTION(elf_file, cur_shdr) {
	// if this is a relocation section

	if (cur_shdr->sh_type == SHT_REL) {

	    REPORT(RP_SMLOOP, "Relocatable section, start offset = %d (0x%X)",
		cur_shdr->sh_offset, cur_shdr->sh_offset);
	    numents = cur_shdr->sh_size / cur_shdr->sh_entsize;

	    relocs = (Elf32_Rel *) &elf_file->ef_file_data[cur_shdr->sh_offset];
	    section = SectionIndexToSection(sections, cur_shdr->sh_info);

	    // if we removed this section already (debug, comment, ...) then
	    // we don't need the relocations for it
	    if (section == NULL) {
		REPORT(RP_SMDETAIL, "(Discarded)");
		continue;
	    }

	    REPORT(RP_SMDETAIL, "Fixing relocations for section %s",
		section->s_name);

	    // for each relocation in this section
	    for (i = 0; i < numents; i++) {
		cur_reloc = &relocs[i];

		new_relocation = zalloc(sizeof(Relocation));
		new_relocation->r_section = section;
		new_relocation->r_offset = cur_reloc->r_offset;
		new_relocation->r_symbol = NULL;
		new_relocation->r_addend = loadLE(&elf_file->ef_file_data
		    [elf_file->ef_shdr_tbl[cur_shdr->sh_info].sh_offset +
			cur_reloc->r_offset]);
		new_relocation->r_type = ELF32_R_TYPE(cur_reloc->r_info);

		if (relocations == NULL) {
		    relocations = new_relocation;
		    new_relocation->r_next = new_relocation->r_prev = NULL;
		} else {
		    InsertRelocationAfterRelocation(new_relocation,
			relocations);
		}
	    }
	}
    }

    return relocations;
}



/*
 *  SetupPltoDataStructures() -- load an ELF file into memory.
 *  All parameters and results are passed through global variables.
 *
 *  Initializes the base PLTO data structures: sec_list (Section list),
 *  s_list (Symbol list), and r_list (Relocation list) as well as
 *  sets PAGE_SIZE, ENTRY_SYMBOL (the symbol to the first function that will be
 *  called), EXIT_SYMBOL (the symbol to the last function that will be called),
 *  and startAddress (the first address that will be executed).
 */
void SetupPltoDataStructures() {
    Elf32File *elf_file;

    PAGE_SIZE = MAXPAGESIZE;

    if (KERNEL_MODE) {
	base_addr = 0xc0100000;
    } else {
	base_addr = 0x08048000;
    }

    // read in the binary, parse and normalize sections
    elf_file = GetInitializedBinary(binFilename);


    
    // create a GOT section if one was not found
    if (!KERNEL_MODE && SECTION_REF_GOT == NULL) {
	REPORT(RP_DETAIL, "Adding a GOT section");
	SECTION_REF_GOT = CreateGOTSection();
	InsertSectionLast(SECTION_REF_GOT, sec_list);
    }

    
    s_list = NULL;

    if (elf_file->ef_symt != NULL) {
	s_list = GetPltoSymbols(elf_file, sec_list);
	LabelFragments(mrg_list, s_list);
	DeleteSections(sec_list, SECTION_SYMBOL_TABLE_FLAG);

	EXIT_SYMBOL = GetElfSymbol(s_list, EXIT_SYMBOL_NAME);
	if (EXIT_SYMBOL == NULL) {
	    WARNING("No exit symbol defined");
	}

	ENTRY_SYMBOL = GetElfSymbol(s_list, ENTRY_SYMBOL_NAME);
	if (ENTRY_SYMBOL == NULL) {
	    WARNING("No entry symbol defined");
	}
    }

    r_list = NULL;

    if (READ_FAKE_RELOCATIONS) {
	// get our fake relocations from
	r_list = GetFakePltoRelocations(elf_file, sec_list);

	// now all of the relocation values should be ready to go and we can
	// build the relocation hash table
	BuildRelocationHashTable();
    } else if (elf_file->ef_type == BINARY_RELOCATABLE_TYPE) {
	Section *seciterator;
	int offset;

	r_list = GetPltoRelocations(sec_list, s_list);
	DeleteSections(sec_list, SECTION_RELOCATIONS_FLAG);

	// set the initial addresses for each of the sections
	REPORT(RP_MINOR, WITHFUNC("Assigning initial section addresses"));
	offset = base_addr;
	FOREACH_SECTION_IN_LIST(sec_list, seciterator) {
	    if (SECTION_HAS_FLAG(seciterator, SECTION_OCCUPIES_VM_FLAG)) {
		// preserve section address if specified;
		// otherwise concatenate beginning at base_addr
		if (seciterator->s_address == 0) {
		    seciterator->s_newaddress = seciterator->s_address = offset;
		    offset += seciterator->s_datasize;
		} else {
		    seciterator->s_newaddress = offset = seciterator->s_address;
		}
		REPORT(RP_SMLOOP, "address=%08X size=%08X %s",
		    seciterator->s_newaddress,
		    seciterator->s_datasize, seciterator->s_name);
	    } else {
		seciterator->s_newaddress = seciterator->s_address = 0;
	    }

	}


	// find any GOT32 relocations and give their symbols indices into
	// the global offset table and update the size of the global offset
	// table
        if (!KERNEL_MODE) {
	    UpdateGOTForRelocations();
	    UpdateSectionSizes(false);
        }

	// go through the symbols and allocate space for any COMMON symbols
	// in the .bss section and update it's size accordingly
	UpdateBSSForSymbols();
    
	UpdateSectionSizes(false);
	// attempt to resolve any special case symbol values (such as the cident
	// section symbols)
	ResolveSpecialCaseSymbols();

	// now we need to update all of our symbols to have absolute values
	// instead of section + offset values because we technically have an
	// executable right now (just need to apply the relocations)
	UpdateSymbolsForVirtualAddresses();
    
	// update all of the relocations so that the disassembly can be done
	ApplyElfRelocations(r_list);

	// now all of the relocation values should be ready to go and we can
	// build the relocation hash table
	BuildRelocationHashTable();
    }
    // add the (()->s_newdata) to each section so that we can manipulate it
    // during the course of binary manipulation
    SetupSectionNewData(sec_list);
    
//  if (((elf_file)->ef_symt) != NULL) {
//    // calculated the address of the first instruction to be executed
//    startAddress = (ENTRY_SYMBOL)->sy_value;
//  } else {
//   // KERNEL START ADDRESS
    startAddress = elf_file->ef_hdr->e_entry;
    if (startAddress == 0) {
	startAddress = ENTRY_SYMBOL->sy_value;
    }
//  }

    // rename the code section .text so the linker won't make another one
    SECTION_REF_CODE->s_name = TEXT_SECTION_NAME;

    // set the global input type to the input type we just read in
    input_binary_type = elf_file->ef_type;

    // before freeing the original file image data,
    // fill with garbage to ensure no inadvertent later reuse
    memset(elf_file->ef_file_data, 0x11, elf_file->ef_file_size);
    free(elf_file->ef_file_data);
    memset(elf_file, 0x11, sizeof(*elf_file));
    free(elf_file);
}



/*
 *  GetPltoFlagsFromElf32_Shdr(elf_shdr) --
 *
 *  Given a pointer to an Elf32_Shdr, this function returns the equivalent
 *  Section flags.
 *
 *  elf_shdr: a pointer to the Elf32_Shdr that whose properties will
 *  be extracted into the Section flags
 *  returns: the Section flags derived from elf_shdr
 */
UINT32 GetPltoFlagsFromElf32_Shdr(Elf32_Shdr *elf_shdr) {
    UINT32 flags;

    flags = 0;

    if (elf_shdr->sh_flags & SHF_ALLOC) {
	flags |= SECTION_OCCUPIES_VM_FLAG;
    }

    if (elf_shdr->sh_flags & SHF_WRITE) {
	flags |= SECTION_WRITABLE_FLAG;
    }

    if (elf_shdr->sh_flags & SHF_EXECINSTR) {
	flags |= SECTION_EXECUTABLE_FLAG;
    }

    if (elf_shdr->sh_flags & SHF_MERGE) {
	flags |= SECTION_MERGE_FLAG;
    }

    if (elf_shdr->sh_type == SHT_PROGBITS) {
	flags |= SECTION_OCCUPIES_FILESPACE_FLAG;
    }

    if (elf_shdr->sh_type == SHT_NOTE) {
	flags |= SECTION_NOTE_FLAG;
    }

    if (elf_shdr->sh_type == SHT_STRTAB) {
	flags |= SECTION_STRING_TABLE_FLAG;
    }

    if (elf_shdr->sh_type == SHT_SYMTAB) {
	flags |= SECTION_SYMBOL_TABLE_FLAG;
    }

    if (elf_shdr->sh_type == SHT_REL) {
	flags |= SECTION_RELOCATIONS_FLAG;
    }

    return flags;
}



/*
 *  FillElf32_ShdrFromPltoFlags(shdr,plto_flags) --
 *
 *  Given the flags for a Section, this function fills in any
 *  fields of the specified Elf32_Shdr that it can based on those flags.
 *
 *  shdr: a pointer to the Elf32_Shdr whose fields should be updated
 *  plto_flags: the Section flags to use in updating shdr
 */
void FillElf32_ShdrFromPltoFlags(Elf32_Shdr *shdr, UINT32 plto_flags) {
    shdr->sh_flags = 0;

    if (plto_flags & SECTION_WRITABLE_FLAG) {
	shdr->sh_flags |= SHF_WRITE;
    }

    if (plto_flags & SECTION_EXECUTABLE_FLAG) {
	shdr->sh_flags |= SHF_EXECINSTR;
    }

    if (plto_flags & SECTION_OCCUPIES_VM_FLAG) {
	shdr->sh_flags |= SHF_ALLOC;
    }

    if (plto_flags & SECTION_OCCUPIES_FILESPACE_FLAG) {
	shdr->sh_type = SHT_PROGBITS;
    } else {
	shdr->sh_type = SHT_NOBITS;
    }

    if (plto_flags & SECTION_IDS_FLAG) {
	shdr->sh_type = SHT_IDS;
    }

    if (plto_flags & SECTION_NOTE_FLAG) {
	shdr->sh_type = SHT_NOTE;
    }

    if (plto_flags & SECTION_RELOCATIONS_FLAG) {
	shdr->sh_type = SHT_REL;
    }

    if (plto_flags & SECTION_STRING_TABLE_FLAG) {
	shdr->sh_type = SHT_STRTAB;
    }

    if (plto_flags & SECTION_SYMBOL_TABLE_FLAG) {
	shdr->sh_type = SHT_SYMTAB;
    }
}



/*
 *  FillElf32_SymFromPltoFlags(sym,sym_flags) --
 *
 *  Given the flags for a Symbol, this function fills in any
 *  fields of the specified Elf32_Sym that it can based on those flags.
 *
 *  sym: a pointer to the Elf32_Sym whose fields should be updated
 *  plto_flags: the Symbol flags to use in updating shdr
 */
void FillElf32_SymFromPltoFlags(Elf32_Sym *sym, UINT32 sym_flags) {
    UINT8 type, binding;

    type = binding = 0;
    if (sym_flags & SYMBOL_LOCAL_FLAG) {
	binding = STB_LOCAL;
    } else if (sym_flags & SYMBOL_GLOBAL_FLAG) {
	binding = STB_GLOBAL;
    } else if (sym_flags & SYMBOL_WEAK_FLAG) {
	binding = STB_WEAK;
    }

    if (sym_flags & SYMBOL_FUNCTION_FLAG) {
	type = STT_FUNC;
    } else if (sym_flags & SYMBOL_SECTION_FLAG) {
	type = STT_SECTION;
    } else if (sym_flags & SYMBOL_OBJECT_FLAG) {
	type = STT_OBJECT;
    } else if (sym_flags & SYMBOL_NOTYPE_FLAG) {
	type = STT_NOTYPE;
    }

    sym->st_info = ELF32_ST_INFO(binding, type);

    if (sym_flags & SYMBOL_ABSOLUTE_FLAG) {
	sym->st_shndx = SHN_ABS;
    } else if (sym_flags & SYMBOL_COMMON_FLAG) {
	sym->st_shndx = SHN_COMMON;
    } else if (sym_flags & SYMBOL_UNDEFINED_FLAG) {
	sym->st_shndx = SHN_UNDEF;
    }
}



/*
 *  BinaryTypeToElfType(type) --
 *
 *  Given a BINARY_TYPE, returns the equivalent ELF binary type.
 *
 *  type: the BINARY_TYPE to be converted
 *  returns: the ELF binary type equivalent to the specified BINARY_TYPE
 */
UINT16 BinaryTypeToElfType(BINARY_TYPE type) {
    if (type == BINARY_RELOCATABLE_TYPE) {
	return ET_REL;
    }

    if (type == BINARY_EXECUTABLE_TYPE) {
	return ET_EXEC;
    }

    if (type == BINARY_PLTO_RELOCATABLE_TYPE) {
	return ET_EXEC;
    }

    if (type == BINARY_SHARED_OBJECT_TYPE) {
	return ET_DYN;
    }

    if (type == BINARY_CORE_TYPE) {
	return ET_CORE;
    }

    return -1;
}



/*
 *  GetPltoFlagsFromElf32_Rel(elf_rel) --
 *
 *  Returns the plto flags derived from the specified elf relocation.
 *
 *  elf_rel: a pointer to the elf relocation to decipher flags from
 *  returns: the flags that are the plto analogs of the properties of the
 *  specified elf relocation
 */
UINT32 GetPltoFlagsFromElf32_Rel(Elf32_Rel *elf_rel) {
    UINT32 flags;
    int type;

    flags = 0;

    type = ELF32_R_TYPE(elf_rel->r_info);

    switch (type) {
    case R_386_NONE:
	break;
    case R_386_32:
	break;
    case R_386_PC32:
	flags |= RELOCATION_PC_RELATIVE_FLAG;
	break;
    case R_386_GOT32:
	flags |= RELOCATION_PC_RELATIVE_FLAG;
	break;
    case R_386_PLT32:
	break;
    case R_386_COPY:
	break;
    case R_386_GLOB_DAT:
	break;
    case R_386_JMP_SLOT:
	break;
    case R_386_RELATIVE:
	break;
    case R_386_GOTOFF:
	break;
    case R_386_GOTPC:
	flags |= RELOCATION_GOTPC_FLAG;
	break;
    case R_386_32PLT:
	break;
    default:
	MALFUNCTION("Unknown Relocation Type");
    }

    return flags;
}
