/*
 *  read.c
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include "protos.h"
#include "constants.h"
#include "linkage.h"
#include "utilities.h"



/*
 *  CorrectRelocations(data,length,delta) --
 *
 *  Updates a relocation section by adding a given delta to the offsets.
 */
static void CorrectRelocations(void *data, int length, int delta) {
    Elf32_Rel *entry = data;
    int n = length / sizeof(*entry);
    int i;
    for (i = 0; i < n; i++, entry++) {
	entry->r_offset += delta;
    }
}



/*
 *  GetFileData(filename,file_size) -- load program from binFile.
 *
 *  Returns the raw data (byte array) contained in the file whose path is
 *  filename and sets *file_size to be the total number of bytes read into the
 *  byte array.
 *
 *  filename: the path to the file whose data should be collected
 *  file_size: the pointer to an int whose value will be set
 *  to the size of the byte array
 *  returns: the raw data (byte array) contained in the file whose path is
 *  filename
 */
static UINT8 *GetFileData(char *filename, int *file_size) {
    UINT8 *buf;
    struct stat file_stat;

    if (stat(filename, &file_stat) < 0) {
	PFATAL(filename);
    }
    *file_size = file_stat.st_size;
    buf = alloc(*file_size);

    if (fread(buf, 1, *file_size, binFile) < *file_size) {
	PFATAL(filename);
    }

    return buf;
}



/*
 *  UpdateSymbols(sec, offset) -- increment address of symbols in sec by offset.
 */
void UpdateSymbols(Section *sec, int offset) {
    Symbol *sym;

    FOREACH_SYMBOL(sym) {
	if ((sym->sy_value > sec->s_address) &&
		(sym->sy_value <= sec->s_address + sec->s_datasize)) {
	    sym->sy_value += offset;
	}
    }
}



/*
 *  MergeSections(sec_list,flag) --
 *
 *  Concatenates the data of sections with the given flag,
 *  replacing the data of the first such section, and returning that section.
 *  (Other sections are marked "merged" but are NOT deallocated or unlinked.)
 *
 *  If the flag is SECTION_RELOCATIONS_FLAG, then the section is merged
 *  only if the previous section was marked with SECTION_EXECUTABLE_FLAG;
 *  and if so, CorrectRelocations is called to fix the merged data.
 */
static Section *MergeSections(Section *sec_list, int flag) {
    Section *seciterator, *first_section;
    UINT8 *allcode;
    int offset;

    // total the sizes and note the first section
    first_section = NULL;
    offset = 0;
    FOREACH_SECTION_IN_LIST(sec_list, seciterator) {
	if (SECTION_HAS_FLAG(seciterator, flag)) {
	    // ignore those relocation sections which are not preceded by
	    // executable sections
	    if (flag == SECTION_RELOCATIONS_FLAG
		&& !(SECTION_HAS_FLAG(seciterator->s_prev,
			SECTION_EXECUTABLE_FLAG))) {
		continue;
	    }
	    // round the section by its alignment constraints .. before merge
	    offset = roundup(offset, seciterator->s_alignment) +
		seciterator->s_datasize;
	    if (first_section == NULL) {
		first_section = seciterator;
	    } else {
		seciterator->s_mergeto = first_section;
	    }
	}
    }
    if (first_section == NULL) {
	return NULL;
    }
    // concatenate the code sections in a new block of memory
    allcode = alloc(offset);
    offset = 0;
    FOREACH_SECTION_IN_LIST(sec_list, seciterator) {
	if (SECTION_HAS_FLAG(seciterator, flag)) {
	    if (flag == SECTION_RELOCATIONS_FLAG
		&& !(SECTION_HAS_FLAG(seciterator->s_prev,
			SECTION_EXECUTABLE_FLAG))) {
		continue;
	    }
	    int gap = roundup(offset, seciterator->s_alignment) - offset;
	    memset(allcode + offset, PAD, gap);
	    offset += gap;
	    REPORT(RP_SMLOOP, "offset=%06X size=%06X %s",
		offset, seciterator->s_datasize, seciterator->s_name);
	    memcpy(allcode + offset, seciterator->s_data,
		seciterator->s_datasize);
	    seciterator->s_delta = offset;
	    if (flag == SECTION_RELOCATIONS_FLAG) {
		CorrectRelocations(allcode + offset,
		    seciterator->s_datasize, seciterator->s_prev->s_delta);
	    }
	    offset += seciterator->s_datasize;
	}
    }

    // replace the first section's data block
    // (trash the old copy to ensure against reuse)
    memset(first_section->s_data, 0x33, first_section->s_datasize);
    free(first_section->s_data);
    first_section->s_data = allcode;
    first_section->s_datasize = offset;
    return first_section;
}



/*
 *  NormalizeCodeSections(elf_file,sec_list) --
 *
 *  Normalizes (merges) executable code sections into the first code section,
 *  updating relocation and symbol information where necessary.
 *
 *  Assumes that:
 *  -- any relocation section immediately follows the affected section
 *  -- gaps between executable sections can be filled with NOP bytes (0x90)
 *
 *  elf_file: the Elf32File for which sections will be merged
 */
static void NormalizeCodeSections(Elf32File *elf_file, Section *sec_list) {
    Section *seciterator, *code_section, *rel_section, *tmp;

    REPORT(RP_MINOR, WITHFUNC("Normalizing code sections"));

    code_section = MergeSections(sec_list, SECTION_EXECUTABLE_FLAG);
    rel_section = MergeSections(sec_list, SECTION_RELOCATIONS_FLAG);

    CASSERT(code_section != NULL, "No code section found");

    // unlink and free the merged sections,
    // saving executable-section headers on mrg_list
    mrg_list = NULL;
    for (seciterator = sec_list; seciterator != NULL;
	seciterator = seciterator->s_next) {
	if (seciterator->s_mergeto != NULL) {
	    tmp = seciterator;
	    seciterator = seciterator->s_prev;
	    tmp->s_prev->s_next = tmp->s_next;
	    tmp->s_next->s_prev = tmp->s_prev;
	    free(tmp->s_data);
	    if (SECTION_HAS_FLAG(tmp, SECTION_EXECUTABLE_FLAG) &&
            SECTION_HAS_FLAG(tmp, SECTION_OCCUPIES_FILESPACE_FLAG)) {
		if (mrg_list == NULL) {
		    tmp->s_prev = tmp->s_next = NULL;
		    mrg_list = tmp;
		} else {
		    InsertSectionLast(tmp, mrg_list);
		}
	    } else {
		elf_file->ef_sect_map[tmp->s_number] = NULL;
		free(tmp);
	    }
	}
    }
}



/*
 *  ElfFile_Parse(file_data,file_size) --
 *
 *  Returns the Elf32File that is parsed from the raw bytes in file_data.
 *  Assumes that the data sizes and endianness of the file match
 *  the data structures of the host C compiler implementation.
 *
 *  file_data: the raw bytes from the ELF binary
 *  file_size: the number of bytes in file_data
 *  returns: the Elf32File that file_data represents
 */
static Elf32File *ElfFile_Parse(UINT8 *file_data, int file_size) {
    Elf32_Shdr *cur_shdr;
    Elf32File *elf_file;
    int i;

    elf_file = zalloc(sizeof(Elf32File));

    elf_file->ef_file_data = file_data;
    elf_file->ef_file_size = file_size;
    elf_file->ef_hdr = (Elf32_Ehdr *) file_data;  // header is at front of data

    switch (elf_file->ef_hdr->e_type) {

    case ET_REL:
	elf_file->ef_type = BINARY_RELOCATABLE_TYPE;
	break;
    case ET_EXEC:
	elf_file->ef_type = BINARY_EXECUTABLE_TYPE;
	break;
    case ET_DYN:
	elf_file->ef_type = BINARY_SHARED_OBJECT_TYPE;
	break;
    case ET_CORE:
	elf_file->ef_type = BINARY_CORE_TYPE;
	break;
    default:
	FATAL("Unsupported or unrecognized ELF binary type");
    }

    // get the section header table
    elf_file->ef_shdr_tbl =
	(Elf32_Shdr *) &file_data[elf_file->ef_hdr->e_shoff];
    if (elf_file->ef_shdr_tbl == NULL) {
	FATAL("No section header table found");
    }
    // get the main string table
    elf_file->ef_strn_tbl = (char *) (file_data +
	elf_file->ef_shdr_tbl[elf_file->ef_hdr->e_shstrndx].sh_offset);

    // find the symbol table and symbol string table
    for (i = 0; i < elf_file->ef_hdr->e_shnum; i++) {
	cur_shdr = &elf_file->ef_shdr_tbl[i];

	if (cur_shdr->sh_type == SHT_SYMTAB) {
	    elf_file->ef_sym_hdr = cur_shdr;
	    elf_file->ef_symt =
		(Elf32_Sym *) (file_data + elf_file->ef_sym_hdr->sh_offset);

	    elf_file->ef_sym_strn_tbl = (char *) (file_data +
		elf_file->ef_shdr_tbl[elf_file->ef_sym_hdr->sh_link].sh_offset);
	}
    }

    if (elf_file->ef_symt == NULL) {
	FATAL("No symbol table found");
    }
    if (elf_file->ef_sym_strn_tbl == NULL) {
	FATAL("No symbol string table found");
    }

    return elf_file;
}



/*
 *  GetInitializedBinary(infile_name) --
 *
 *  Reads an ELF input file into data structures.
 *  Merges executable sections into one.
 *  Relocations are adjusted here; symbols are handled later.
 *
 *  infile_name: the name of the file to read
 *  returns: an Elf32File struct, adjusted by merging sections
 */
Elf32File *GetInitializedBinary(char *infile_name) {
    Elf32File *elf_file1;
    UINT8 *file_data;
    int file_size;

    // get the original relocatable
    file_data = GetFileData(infile_name, &file_size);

    REPORT(RP_MINOR, "Parsing ELF file");
    elf_file1 = ElfFile_Parse(file_data, file_size);

    // parse the necessary sections out of the ELF file
    sec_list = GetPltoSections(elf_file1);

    // set up the fundamental section references
    SECTION_REF_CODE = SectionNameToSection(sec_list, CODE_SECTION_NAME);
    SECTION_REF_DATA = SectionNameToSection(sec_list, DATA_SECTION_NAME);
    SECTION_REF_RODATA = SectionNameToSection(sec_list, RODATA_SECTION_NAME);
    SECTION_REF_BSS = SectionNameToSection(sec_list, BSS_SECTION_NAME);
    SECTION_REF_SYM = SectionNameToSection(sec_list, SYM_SECTION_NAME);
    SECTION_REF_GOT = SectionNameToSection(sec_list, GOT_SECTION_NAME);

    if (!READ_FAKE_RELOCATIONS) {
	// normalize (merge) sections to make a unified code section
	NormalizeCodeSections(elf_file1, sec_list);
	if (SECTION_REF_CODE->s_mergeto != NULL) {
	    SECTION_REF_CODE = SECTION_REF_CODE->s_mergeto;
	}
    }

    return elf_file1;
}
