/*
 *  relocate.c
 *
 *  This file contains functionality for Elf32_Rel relocations. There is
 *  currently no support for Elf32_Rela relocations, but gcc doesn't currently
 *  put those out for the x86 as far as I can tell. The Elf manual is
 *  particularly handy for understanding this file, especially the section
 *  labeled 'Relocation'.
 *
 *  The following excerpt is from the ELF manual: Relocation Types
 *  (Section 1-23):
 *
 *  A  This means the addend used to compute the value of the relocatable field.
 *  B  This means the base address at which a shared object has been loaded into
 *     memory during execution. Generally, a shared object file is built with a
 *     0 base virtual address, but the execution address will be different.
 *  G  This means the offset into the global offset table at which the address
 *     of the relocation entry's symbol will reside during execution.
 *     See Global Offset Table in Part 2 for more information.
 *  GOT This means the address of the global offset table.
 *     See Global Offset Table in Part 2 for more information.
 *  L  This means the place (section offset or address) of the procedure linkage
 *     table entry for a symbol. A procedure linkage table entry redirects a
 *     function call to the proper destination. The link editor builds the
 *     initial procedure linkage table, and the dynamic linker modifies the
 *     entries during execution. See Procedure Linkage Table in Part 2 for
 *     more information.
 *  P  This means the place (section offset or address) of the storage unit
 *     being relocated (computed using r_offset).
 *  S  This means the value of the symbol whose index resides in the
 *     relocation entry.
 *
 *  A relocation entry's r_offset value affected storage unit.
 *  The relocation type
 *  The SYSTEM V architecture uses only addend. In all cases, the addend and
 *  the computed result use the same byte order.
 */

#include "protos.h"
#include "utilities.h"
#include "constants.h"
#include "upm.h"
#include "relocate.h"
#include "flowgraph.h"
#include "elfio.h"



/*
 *  ApplyElfRelocations(relocations) --
 *
 *  Iterates through the specified list of relocations applying each of them.
 *
 *  relocations: the list of relocations that should be applied
 */
void ApplyElfRelocations(Relocation *relocations) {
    Relocation *riterator;

    REPORT(RP_MINOR, WITHFUNC("Applying relocations"));
    FOREACH_RELOCATION_IN_LIST(relocations, riterator) {
	switch (riterator->r_type) {
	case R_386_NONE:
	    break;
	case R_386_32:
	    ApplyRelocation_32(riterator);
	    break;
	case R_386_PC32:
	    ApplyRelocation_PC32(riterator);
	    break;
	case R_386_GOT32:
	    ApplyRelocation_GOT32(riterator);
	    break;
	case R_386_PLT32:
	    ApplyRelocation_PLT32(riterator);
	    break;
	case R_386_COPY:
	    ApplyRelocation_COPY(riterator);
	    break;
	case R_386_GLOB_DAT:
	    ApplyRelocation_GLOB_DAT(riterator);
	    break;
	case R_386_JMP_SLOT:
	    ApplyRelocation_JMP_SLOT(riterator);
	    break;
	case R_386_RELATIVE:
	    ApplyRelocation_RELATIVE(riterator);
	    break;
	case R_386_GOTOFF:
	    ApplyRelocation_GOTOFF(riterator);
	    break;
	case R_386_GOTPC:
	    ApplyRelocation_GOTPC(riterator);
	    break;
	default:
	    MALFUNCTION("Unrecognized relocation type: %x", riterator->r_type);
	    return;
	}
    }

}



/*
 *  ApplyRelocation_32(rel) --
 */
void ApplyRelocation_32(Relocation *rel) {
    // Relocation Computation: (S + A)
    Symbol *sym;

    sym = rel->r_symbol;
    storeLE(
        &rel->r_section->s_data[rel->r_offset], sym->sy_value + rel->r_addend);

}



/*
 *  ApplyRelocation_PC32(rel) --
 */
void ApplyRelocation_PC32(Relocation *rel) {
    // Relocation Computation: (S + A - P)

    Symbol *sym;

    sym = rel->r_symbol;

    storeLE(&rel->r_section->s_data[rel->r_offset],
	sym->sy_value + rel->r_addend - (rel->r_offset +
	    rel->r_section->s_address));
}



/*
 *  ApplyRelocation_GOT32(rel) --
 */
void ApplyRelocation_GOT32(Relocation *rel) {
    // Relocation Computation: G + A - P
    // Also, if there is no global offset table one must be constructed

    CASSERT(SECTION_REF_GOT != NULL, "No global offset table (.got) section");

    storeLE(&rel->r_section->s_data[rel->r_offset], rel->r_symbol->sy_value);
}



/*
 *  ApplyRelocation_PLT32(rel) --
 */
void ApplyRelocation_PLT32(Relocation *rel) {
    // Relocation Computation: L + A - P

    FATAL("Unimplemented R_386_PLT32 relocation");
}



/*
 *  ApplyRelocation_COPY(rel) --
 */
void ApplyRelocation_COPY(Relocation *rel) {
    // Relocation Computation: None
}



/*
 *  ApplyRelocation_GLOB_DAT(rel) --
 */
void ApplyRelocation_GLOB_DAT(Relocation *rel) {
    // Relocation Computation: S

    Symbol *sym;

    sym = rel->r_symbol;

    storeLE(&rel->r_section->s_data[rel->r_offset], sym->sy_value);
}



/*
 *  ApplyRelocation_JMP_SLOT(rel) --
 */
void ApplyRelocation_JMP_SLOT(Relocation *rel) {
    // Relocation Computation: S

    FATAL("Unimplemented R_386_JMP_SLOT relocation");
    storeLE(&rel->r_section->s_data[rel->r_offset], rel->r_symbol->sy_value);
}



/*
 *  ApplyRelocation_RELATIVE(rel) --
 */
void ApplyRelocation_RELATIVE(Relocation *rel) {
    // Relocation Computation: B + A

    FATAL("Unimplemented R_386_RELATIVE relocation");
}



/*
 *  ApplyRelocation_GOTOFF(rel) --
 */
void ApplyRelocation_GOTOFF(Relocation *rel) {
    // Relocation Computation: S + A - GOT

    Symbol *sym;

    CASSERT(SECTION_REF_GOT != NULL, "No global offset table (.got) section");

    sym = rel->r_symbol;

    storeLE(&rel->r_section->s_data[rel->r_offset],
	sym->sy_value + rel->r_addend - SECTION_REF_GOT->s_address);
}



/*
 *  ApplyRelocation_GOTPC(rel) --
 */
void ApplyRelocation_GOTPC(Relocation *rel) {
    // Relocation Computation: GOT + A - P

    CASSERT(SECTION_REF_GOT != NULL, "No global offset table (.got) section");

    storeLE(&rel->r_section->s_data[rel->r_offset],
	SECTION_REF_GOT->s_address + rel->r_addend -
	(rel->r_offset + rel->r_section->s_address));
}



/*
 *  UpdateGOTForRelocations() --
 *
 *  Scans through all relocations allocating memory in the global offset table
 *  for any relocations that need it.
 */
void UpdateGOTForRelocations() {
    Relocation *rel;
    Symbol *sym;

    FOREACH_RELOCATION(rel) {
	sym = rel->r_symbol;

	if (rel->r_type == R_386_GOT32) {
	    int index;

	    if (sym->sy_value != 0) {
		index = sym->sy_value;
	    } else {
		index = AddGotEntry(0x0);
		sym->sy_value = index;
	    }
	}
    }

}



/*
 *  AllocateCommonSymbol(sym,bss_offset) -- assign address in BSS.
 *
 *  Allocates aligned space in the BSS section for a common symbol.
 *  On input, sym->sy_value gives the required alignment.
 *  On output, sym->sy_value is the assigned offset in the BSS section,
 *  and the alignment is in sym->sy_alignment.
 */
int AllocateCommonSymbol(Symbol *sym, int bss_offset) {

    // pad .bss for proper alignment
    bss_offset = roundup(bss_offset, sym->sy_value);
    if (SECTION_REF_BSS->s_alignment < sym->sy_value) {
	SECTION_REF_BSS->s_alignment = sym->sy_value;
    }

    // update symbol data
    sym->sy_section = SECTION_REF_BSS;
    sym->sy_alignment = sym->sy_value;
    sym->sy_value = bss_offset;
    sym->sy_flags &= ~SYMBOL_COMMON_FLAG;

    // allocate space for object
    bss_offset += sym->sy_size;
    return bss_offset;
}



/*
 *  UpdateBSSForSymbols() --
 *
 *  Scans through the global list of symbols and allocates memory
 *  in the .bss section for any COMMON symbols.
 */
void UpdateBSSForSymbols() {
    Symbol *sym;
    int bss_offset, added_syms;

    REPORT(RP_MINOR, WITHFUNC("Allocating space for common symbols"));

    CASSERT(SECTION_REF_BSS != NULL, "No .bss section");
    CASSERT(s_list != NULL, "Symbol table is null");

    bss_offset = SECTION_REF_BSS->s_datasize;
    added_syms = 0;

    // get to the rear of the list
    sym = s_list;
    while (sym->sy_next != NULL) {
	sym = sym->sy_next;
    }

    while (sym != NULL) {
	if (sym->sy_flags & SYMBOL_COMMON_FLAG) {
	    bss_offset = AllocateCommonSymbol(sym, bss_offset);
	    added_syms++;
	}
	sym = sym->sy_prev;
    }

    REPORT(RP_DETAIL, "Added %d bytes to .bss for %d symbols",
	bss_offset - SECTION_REF_BSS->s_datasize, added_syms);
	
    SECTION_REF_BSS->s_datasize = bss_offset;
}



/*
 *  FindPreviousSection(addr) -- find section containing addr (?!)
 *
 *  An address falling between sections is considered to belong to the
 *  preceding section.
 */
Section *FindPreviousSection(ADDRESS src_addr) {
    Section *temp;

    FOREACH_SECTION(temp) {
	if (temp->s_address <= src_addr) {
	    if (temp->s_next == NULL) {
		// this means it's the last section.
		return temp;
	    } else if (temp->s_next->s_address > src_addr) {
		return temp;
	    }
	}
    }
    return NULL;
}



/*
 *  InitializeUPMs() --
 *
 *  Iterates through the global list of Relocations converting
 *  all relocations into UPMs.
 *  see also: r_list
 */
void InitializeUPMs() {
    LinkedList *pic_calls;
    Relocation *riterator;
    Operand *op;
    Section *rel_section, *src_section;
    Instruction *src_instr, *dest_instr;
    ADDRESS rel_offset, rel_addr, src_addr;
    FList *src_flist, *dest_flist;
    int rel_contents, total, word_add, pc_changing, call_zeroes;
    int dest_patch_offset, src_patch_offset = 0;

    call_zeroes = pc_changing = total = 0;
    pic_calls = GetPICCalls();

    REPORT(RP_MINOR, WITHFUNC("Converting relocations to UPMs"));

    FOREACH_RELOCATION(riterator) {
	total++;
	src_section = NULL;
	src_instr = dest_instr = NULL;
	op = NULL;
        src_patch_offset = 0;
        dest_patch_offset = 0;
	word_add = PC32_OFFSET;
	src_flist = dest_flist = NULL;

	rel_section = riterator->r_section;
	rel_offset = riterator->r_offset;
	rel_contents = loadLE(&(rel_section->s_data)[rel_offset]);
	rel_addr = rel_offset + rel_section->s_address;
    
	// If the relocation is in the code section then, it must (??) be
	// in an instruction. Find the instruction.
	if (rel_section == SECTION_REF_CODE) {
	    // A side-effect of this is that the i_temp field indicates how far
	    // in the instruction the relocation needs to be applied.
	    dest_instr = BacktraceToInstr(rel_addr);

	    // if the instruction is NULL then something is WRONG !!!
	    // every thing in the code section should be either
	    // instrs or data instrs.
	    if (dest_instr == NULL) {
		MALFUNCTION("No instruction in text section for relocation");
	    } else if (INSTR_HAS_FLAG(dest_instr, INSTR_DATA_BYTE)) {
		dest_patch_offset = dest_instr->i_temp;
	    } else {
		op = GetOperandForOperandAddress(dest_instr, rel_addr);
		if (op == NULL) {
		    REPORT(RP_TRIVIAL,
		    "Could not find operand for instruction %x at address %x",
			dest_instr->i_address, rel_addr);
		    continue;
		}
	    }
	} else {
	    word_add = 0;
	}
    
	/*
	 * Patch creation of absolute symbols to instruction
	 * This is different because absolute symbols are
	 * bindings.
	 */
	if ((riterator->r_symbol->sy_flags & SYMBOL_ABSOLUTE_FLAG)) {
	    if (dest_instr != NULL) {
                AddPatchSymbolValueIntoInstr(riterator->r_symbol,
                    riterator->r_addend, dest_instr, op);
	    } else {
                AddPatchSymbolValueIntoSectionAddress(riterator->r_symbol,
                    riterator->r_addend, rel_section, rel_offset);
	    }
            continue;
	}
    
	// locate the source instruction
	if (riterator->r_type == R_386_PC32) {
	    src_addr = rel_contents + rel_addr + word_add;
	} else {
	    src_addr = rel_contents;
	}
	src_instr = AddressToInstr(src_addr);
	src_section = AddressToSection(src_addr);
    
	// is the symbol is associated with the end of an object (e.g. section)?
	if (!!(riterator->r_symbol->sy_flags & SYMBOL_OBJECT_END_FLAG)) {
	    src_section = AddressToSection(riterator->r_symbol->sy_value - 1);
	    if (riterator->r_type != R_386_32) {
		WARNING("End symbol relocation with non R_386_32 type %s",
                    RelocationTypeToString(riterator->r_type));
	    } else if (dest_instr != NULL) {
		AddPatchSectionEndIntoInstr(src_section, dest_instr,
		    op, riterator->r_addend);
	    } else {
		AddPatchSectionEndIntoSectionAddress(src_section, rel_section,
		    rel_offset);
	    }
	    continue;
	}

	if (riterator->r_type == R_386_GOT32) {
	    // we just skip this one since once the symbol's index into the
	    // GOT has been determined, relocating doesn't have any effect
	    total--;
	    continue;
	}
	if (riterator->r_type == R_386_PLT32) {
	    WARNING("Need to implement PLT32 Relocation!");
	    continue;
	}
	if (riterator->r_type == R_386_GOTOFF) {
	    WARNING("Need to implement GOTOFF Relocation!");
	    continue;
	}
	if (riterator->r_type == R_386_GOTPC) {

	    if (SECTION_REF_GOT == NULL) {
		WARNING("Attempt to convert GOTPC relocation with no .got");
		continue;
	    } else if (dest_instr != NULL) {
		Iterator *itr;
		Instruction *aux_instr = NULL;

		//here we don't actually want to add the aux_instr UPM for the
		//dest instr, instead we actually want to use the call that
		//originates the PIC to be the aux_instr
		itr = ll_iterator(pic_calls);
		while (itr->it_hasNext(itr)) {
		    aux_instr = (Instruction *) itr->it_next(itr);

		    if (aux_instr->i_next->i_next == dest_instr) {
			break;
		    }

		    aux_instr = NULL;
		}
		free(itr);

		if (aux_instr == NULL) {
		    aux_instr = dest_instr;
		    WARNING(
		    "Could not locate PIC call for GOTPC relocation at: %x",
			dest_instr->i_address);
		}

		AddPatchGOTOffsetIntoInstr(aux_instr, dest_instr, op);
		INSTR_REMOVE_FLAG(dest_instr, INSTR_RELOCATION_FLAG);
		op->a_relocation = NULL;
		continue;
	    } else {
		AddPatchSectionSectionOffsetIntoSectionAddress(SECTION_REF_GOT,
		    0, rel_section, rel_offset, rel_section, rel_offset);
		continue;
	    }
	}
	// peculiarity in the kernel: sometimes they want to hardcode addresses
	// and they use relocations in a CHEAP :-| way to do so.
	if ((src_instr == NULL) && (src_section == NULL)) {
	    // MR--note: An alternative here is to create a new relocation,
	    // where in the end, the address is offset accordingly.
	    // Ok so what happens in FC3 -- in several relocations what seems
	    // to happen is that the src_addr falls in the padding between
	    // sections -- aka unassigned space
	    //      sec_addr + sec_size < src_address
	    // so SECTION_CONTAINS returns a null section.
	    // Patch -- find the closest section and patch in and compute
	    // the offset.
	    // I've put in a sanity check to make this FC3 specific.
	    src_section = FindPreviousSection(src_addr);
	    if (!src_addr) {
		continue;
	    }
	    if (!((FindPreviousSection(src_addr))->s_next)) {
		continue;
	    } else {
		if (FindPreviousSection(src_addr)->s_next->s_alignment != 32) {
		    continue;
		}
	    }
	}

    if (src_section != riterator->r_symbol->sy_section &&
        riterator->r_symbol->sy_section) {
        if (dest_instr != NULL && op) {
            AddPatchSymbolValueIntoInstr(riterator->r_symbol,
                riterator->r_addend, dest_instr, op);
            continue;
        } 
    }
    
	// if AddressToInstr could not find the source instr, then probably,
	// the relocation points into undisassembled code [a.k.a data_insts]
	if ((src_instr == NULL) && (src_section == SECTION_REF_CODE)) {
	    src_flist = GetFlistFunAtAddress(src_addr);
	    if (src_flist) {
		src_instr = BacktraceToInstr(src_addr);
		src_patch_offset = src_instr->i_temp;
	    }
	}

	// If there are calls to sections that will be deleted, ignore them.
	if ((src_section != NULL) && (src_section->s_address == 0)) {
	    call_zeroes++;
	    continue;
	}

	// if the target is an instruction
	if (dest_instr != NULL) {
	    if (INSTR_HAS_FLAG(dest_instr, INSTR_DATA_BYTE)) {
                if (src_instr != NULL) {
                    if (INSTR_HAS_FLAG(src_instr, INSTR_DATA_BYTE)) {
                        if (riterator->r_type == R_386_PC32) {
                            WARNING("Test patch & remove this comment. %x-->%x",
                                dest_instr->i_address,
                                src_instr->i_address);
                            AddPatchDataInstrToDataInstrRelative(src_instr,
                                src_patch_offset,
                                dest_instr, dest_patch_offset);	
                        } else {
                            AddPatchDataInstrToDataInstr(src_instr,
                                src_patch_offset, dest_instr,
                                dest_patch_offset);
                            continue;
                        }
                    } else {
                        if (riterator->r_type == R_386_PC32) {
                            AddPatchDataInstrToDataInstrRelative(src_instr, 0,
                                dest_instr, dest_patch_offset);	
                        } else {
                            AddPatchInstrAddressIntoSectionAddress(src_instr,
                                rel_section,
                                dest_instr->i_address - rel_section->s_address +
                                dest_patch_offset);
                        }
                        continue;
                    }
            } else {
                if (src_section != riterator->r_symbol->sy_section) {
                    WARNING(WITHFUN("src section != symbol-section %s %s\n"),
                        src_section->s_name,
                        riterator->r_symbol->sy_section->s_name);
                    src_section = riterator->r_symbol->sy_section;
                }

                AddPatchDataSectionOffsetIntoDataInstr(src_section,
                    rel_contents - src_section->s_address, dest_instr,
                    dest_patch_offset);
                continue;
            }
	    }

	    if (src_instr != NULL) {
		if (INSTR_HAS_FLAG(src_instr, INSTR_DATA_BYTE)) {
		    REPORT(RP_TRIVIAL,
			"Src instr is data instr at %x with offset %x to %x\n", 
			src_instr->i_address, src_patch_offset,
			dest_instr->i_address);
		}

		// if it is a jump or a call then we want to let the final
		// address be determined by flowgraph edges
		if (isPCChangingClass(dest_instr)) {
		    pc_changing++;
		    continue;
		}
		// Can we patch it based on Instruction Offset
		if (riterator->r_type == R_386_PC32) {
                    
		    AddPatchInstrInstrOffsetIntoInstr(src_instr, dest_instr,
			dest_instr, op);
		    INSTR_REMOVE_FLAG(dest_instr, INSTR_RELOCATION_FLAG);
		    op->a_relocation = NULL;
		    continue;

		} else {
		    // or we could use an address
		    if (INSTR_HAS_FLAG(src_instr, INSTR_DATA_BYTE)) {
			// This is a fix for the time being.
			// Create a new patch type later. 
			AddPatchDataInstrAddressToInstr(src_instr,
			    src_patch_offset, dest_instr, op);
		    } else {
			AddPatchInstrAddressIntoInstr(src_instr,dest_instr,op);
		        INSTR_REMOVE_FLAG(dest_instr, INSTR_RELOCATION_FLAG);
		    }
		    op->a_relocation = NULL;
		    continue;
		}
	    } else {
	    // If we don't know the src_instr it is probably in a data segment.
		if (riterator->r_type == R_386_PC32) {
		    AddPatchInstrSectionOffsetIntoInstr(src_section,
			rel_contents + dest_instr->i_address +
			dest_instr->i_length - src_section->s_address,
			dest_instr, dest_instr, op);
		    INSTR_REMOVE_FLAG(dest_instr, INSTR_RELOCATION_FLAG);
		    op->a_relocation = NULL;
		    continue;
		} else {
		    AddPatchSectionAddressIntoInstr(src_section,
			rel_contents - src_section->s_address, dest_instr, op);
		    INSTR_REMOVE_FLAG(dest_instr, INSTR_RELOCATION_FLAG);
		    op->a_relocation = NULL;
		    continue;
		}
	    }

	// now to handle the case where the dest_instr is outside of the text
	// section that is the relocation entry lies outside the text section.

	} else {

	    if (src_instr != NULL) {

		// CASE 1: src_instr, what the relocation points to is known.
		if (INSTR_HAS_FLAG(src_instr, INSTR_DATA_BYTE)) {
		    if (riterator->r_type == R_386_PC32) {
			AddPatchInstrSectionOffsetIntoSectionAddress(src_instr,
			    src_section, src_patch_offset, rel_section,
			    rel_offset);
			continue;
		    } else {
			AddPatchDataInstrToSectionAddress(src_instr,
			    src_patch_offset, rel_section, rel_offset);
			continue;;
		    }
		}
		if (riterator->r_type == R_386_PC32) {
		    AddPatchInstrSectionOffsetIntoSectionAddress(src_instr,
			rel_section, rel_offset, rel_section, rel_offset);
		    continue;
		} else {
		    AddPatchInstrAddressIntoSectionAddress(src_instr,
			rel_section, rel_offset);
		    continue;
		}
	    } else {

		// CASE 2: src_instr, what the relocation points to is NOT
		//    known, or in other words...  data to data relocation.
		if (riterator->r_type == R_386_PC32) {
		    int src_offset, dest_offset;

		    src_offset = rel_contents + rel_addr + PC32_OFFSET -
			src_section->s_address;
		    dest_offset = rel_offset;
		    AddPatchSectionSectionOffsetIntoSectionAddress(src_section,
			src_offset,
			rel_section, dest_offset, rel_section, dest_offset);
		    continue;
		} else {
		    int src_offset, dest_offset;

		    src_offset = rel_contents - src_section->s_address;
		    dest_offset = rel_offset;
		    AddPatchSectionAddressIntoSectionAddress(src_section,
			src_offset, rel_section, dest_offset);
		    continue;
		}
	    }
	}

	WARNING("Unhandled Relocation -- "
	    "rel_section: %s, rel_offset: %x, rel_type: %x",
	    rel_section->s_name, rel_offset, riterator->r_type);
    }

    ll_clear(pic_calls);
    ll_free(pic_calls);

    REPORT(RP_DETAIL, "Converted %d of %d relocations to UPMs",
	ll_size(upm2s_list), total);
    REPORT(RP_DETAIL,
	"Removed %d PCrel relocations for PC changing instructions",
	pc_changing);
    REPORT(RP_DETAIL, "Removed %d relocations for call 0 instructions",
	call_zeroes);
    REPORT(RP_DETAIL, "Total unhandled relocations: %d",
	total - (call_zeroes + pc_changing + ll_size(upm2s_list)));
}



/*
 *  GetFakeRelocationInfoForUPM(upm,type,offset,section) --
 *
 *  Given a UPM, fills in values for the type, offset and section for the
 *  equivalent pseudo-relocations to be written out in a pseudo-relocatable.
 *
 *  upm: the UPM to 'convert' to a pseudo-relocation
 *  type: a pointer to a UINT32 that will be filled in with the native
 *  relocation type corresponding to upm
 *  offset: a pointer to an ADDRESS that will be filled in with the
 *  virtual address of the destination of the specified upm
 *  section: a pointer to the Section that will be filled in with a
 *  pointer to the section associated with upm
 */
void GetFakeRelocationInfoForUPM(Patch *upm, UINT32 *type, ADDRESS *offset,
    Section **section) {
    *type = -1;
    *offset = -1;
    *section = NULL;

    if (upm->p_Apply == patchInstrAddressIntoInstr) {
	PatchInstrToInstr *patch;

	patch = (PatchInstrToInstr *) upm->p_patch;

	*offset = patch->p_dest_instr->i_final_address -
	    SECTION_REF_CODE->s_newaddress +
	    OperandOffset(patch->p_dest_instr, patch->p_dest_op);
	*type = R_386_32;
	*section = SECTION_REF_CODE;
    } else if (upm->p_Apply == patchSectionAddressIntoInstr) {
	PatchSectionToInstr *patch;

	patch = (PatchSectionToInstr *) upm->p_patch;

	*offset = patch->p_dest_instr->i_final_address -
	    SECTION_REF_CODE->s_newaddress +
	    OperandOffset(patch->p_dest_instr, patch->p_dest_op);
	*type = R_386_32;
	*section = SECTION_REF_CODE;
    } else if (upm->p_Apply == patchInstrInstrOffsetIntoInstr) {
	PatchInstrToInstrAux *patch;

	patch = (PatchInstrToInstrAux *) upm->p_patch;

	*offset = patch->p_dest_instr->i_final_address -
	    SECTION_REF_CODE->s_newaddress +
	    OperandOffset(patch->p_dest_instr, patch->p_dest_op);
	*type = R_386_PC32;
	*section = SECTION_REF_CODE;
    } else if (upm->p_Apply == patchInstrSectionOffsetIntoInstr) {
	PatchSectionInstrToInstr *patch;

	patch = (PatchSectionInstrToInstr *) upm->p_patch;

	*offset = patch->p_dest_instr->i_final_address -
	    SECTION_REF_CODE->s_newaddress +
	    OperandOffset(patch->p_dest_instr, patch->p_dest_op);
	*type = R_386_PC32;
	*section = SECTION_REF_CODE;
    } else if (upm->p_Apply == patchInstrSectionOffsetIntoSectionAddress) {
	PatchSectionInstrToSection *patch;

	patch = (PatchSectionInstrToSection *) upm->p_patch;

	*offset = patch->p_dest_offset;
	*type = R_386_PC32;
	*section = patch->p_dest_section;
    } else if (upm->p_Apply == patchInstrAddressIntoSectionAddress) {
	PatchInstrToSection *patch;

	patch = (PatchInstrToSection *) upm->p_patch;

	*offset = patch->p_dest_offset;
	*type = R_386_32;
	*section = patch->p_dest_section;
    } else if (upm->p_Apply == patchSectionSectionOffsetIntoSectionAddress) {
	PatchSectionToSectionAux *patch;

	patch = (PatchSectionToSectionAux *) upm->p_patch;

	*offset = patch->p_dest_offset;
	*type = R_386_PC32;
	*section = patch->p_dest_section;
    } else if (upm->p_Apply == patchSectionAddressIntoSectionAddress) {
	PatchSectionToSection *patch;

	patch = (PatchSectionToSection *) upm->p_patch;

	*offset = patch->p_dest_offset;
	*type = R_386_32;
	*section = patch->p_dest_section;
    }
}
