/*
 *  utils.c
 */
#include <elf.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "file.h"
#include "protos.h"
#include "constants.h"
#include "utilities.h"
#include "linkage.h"

/*
 *  there are all debugging or useless sections or sections
 *  that need to be rebuilt before writing the binary out again
 */
static char *ignored_sec_prefixes[] = {
    ".strtab",
    ".shstrtab",
    ".stab",
    ".rel.stab",
    ".stabstr"
    // ".gnu.warning",  (preserve these -- don't remember why #%#%)
    // ".comment",	(preserve this -- don't remember why #%#%)
};



/*
 *  IsDebuggingInfoSection(elf_file,shdr_index) --
 *
 *  Returns true if a section name begins with DEBUGGING_SECTION_PREFIX.
 *
 *  elf_file: the Elf32File the section in question is located in
 *  shdr_index: the index into the section header table for the section in
 *  question
 *  returns: true if the section at the specified index in the specified
 *  Elf32File is a debugging section
 */
bool IsDebuggingInfoSection(Elf32File *elf_file, int shdr_index) {
    Elf32_Shdr *cur_shdr = &elf_file->ef_shdr_tbl[shdr_index];
    char *name = elf_file->ef_strn_tbl + cur_shdr->sh_name;
    int n = strlen(DEBUGGING_SECTION_PREFIX);
    return strncmp(name, DEBUGGING_SECTION_PREFIX, n) == 0;
}



/*
 *  IsIgnoredElfSection(elf_file,shdr_index) --
 *
 *  Returns whether or not the section whose index into the section header table
 *  for the specified Elf32File is an 'ignored' section.  To be an ignored
 *  section a section must either be regenerated by PLTO or be meaningless to
 *  the binary (i.e., the .gnu.warning section exists only for warning strings
 *  and can be omitted).
 *
 *  elf_file: the Elf32File containing the section header table to check
 *  shdr_index: the index into the section header table for the section in
 *  question
 */
bool IsIgnoredElfSection(Elf32File *elf_file, int shdr_index) {
    int i;
    Elf32_Shdr *cur_shdr = &elf_file->ef_shdr_tbl[shdr_index];
    char *name = &elf_file->ef_strn_tbl[cur_shdr->sh_name];

    for (i = 0; i < ARRAYSIZE(ignored_sec_prefixes); i++) {
	char *prefix = ignored_sec_prefixes[i];
	if (strncmp(name, prefix, strlen(prefix)) == 0) {
	    return true;
	}
    }
    return false;
}



/*
 *  CreateGOTSection() --
 *
 *  Creates and returns a section that can be used as the .got section.
 *  returns: a Section that has the correct properties to be a .got section
 */
Section *CreateGOTSection() {
    Section *section;

    section = zalloc(sizeof(Section));
    section->s_datasize = GOT_SECTION_SIZE;
    section->s_data = zalloc(GOT_SECTION_SIZE);
    memset(section->s_data, 0x00, GOT_SECTION_SIZE);
    section->s_address = 0x0;
    section->s_name = salloc(GOT_SECTION_NAME);
    section->s_alignment = 0x4;
    section->s_flags = SECTION_OCCUPIES_VM_FLAG | SECTION_WRITABLE_FLAG |
	SECTION_OCCUPIES_FILESPACE_FLAG;
    section->s_misc = -1;
    section->s_growth = 0;

    return section;
}



/*
 *  AddGotEntry(address) --
 *
 *  This function writes the specified address into the global offset table
 *  (growing it appropriately) and returns the offset into the table at which it
 *  was written.
 *
 *  address: the address (value) to write into the global offset table
 *  returns: the offset into the global offset table where address was written
 */
ADDRESS AddGotEntry(ADDRESS address) {
    ADDRESS offset;
    UINT8 *new_buf;

    offset = SECTION_REF_GOT->s_datasize;
    new_buf = zalloc(offset + sizeof(ADDRESS));
    memcpy(new_buf, SECTION_REF_GOT->s_data, SECTION_REF_GOT->s_datasize);
    storeLE(&new_buf[offset], address);
    SECTION_REF_GOT->s_datasize += sizeof(ADDRESS);

    free(SECTION_REF_GOT->s_data);
    SECTION_REF_GOT->s_data = new_buf;

    return offset;
}



/*
 *  SetupSectionNewData(section_list) --
 *
 *  Iterates through every section in the specified list of sections and
 *  allocates space for the NEW_DATA and copies the DATA into it.
 *
 *  section_list: the list of sections for which NEWDATA should be created
 */
void SetupSectionNewData(Section *section_list) {
    Section *seciterator;

    FOREACH_SECTION_IN_LIST(section_list, seciterator) {
	if (SECTION_HAS_FLAG(seciterator,
		SECTION_OCCUPIES_FILESPACE_FLAG | SECTION_NOTE_FLAG)) {
	    seciterator->s_newdata = zalloc(seciterator->s_datasize);
	    memcpy(seciterator->s_newdata, seciterator->s_data,
		seciterator->s_datasize);
	}

	seciterator->s_newdatasize = seciterator->s_datasize;
    }
}



/*
 *  GetElfSymbol(symbol_list,symbol_name) --
 *
 *  Searches a list of ELF symbols to find a particular one.
 *
 *  symbol_list: the list of symbols to search
 *  symbol_name: the name of the desired symbol
 *  returns: the matching Symbol, or NULL if not found
 */
Symbol *GetElfSymbol(Symbol *symbol_list, char *symbol_name) {
    Symbol *siterator;

    FOREACH_SYMBOL_IN_LIST(symbol_list, siterator) {
	if (strcmp(siterator->sy_name, symbol_name) == 0) {
	    return siterator;
	}
    }
    return NULL;
}



/*
 *  IsValidCIdentifier(str) --
 *
 *  Returns whether or not the specified character string represents a valid C
 *  identifier.
 *
 *  name: the character string to be checked
 *  returns: true if name is a valid C identifier
 */
bool IsValidCIdentifier(char *str) {
    if ((*str != '_') &&
	((*str < 'a') || (*str > 'z')) && ((*str < 'A') || (*str > 'Z'))) {
	return false;
    }

    str++;

    while (*str != '\0') {
	if ((*str != '_') &&
	    ((*str < 'a') || (*str > 'z')) &&
	    ((*str < 'A') || (*str > 'Z')) && ((*str < '0') || (*str > '9'))) {
	    return false;
	}

	str++;
    }

    return true;
}



/*
 *  RelocationTypeToString(type) --
 *
 *  Returns the string representation of the specified Elf relocation type.
 *
 *  type: the Elf relocation type to be converted
 *  returns: the string representation of the specified relocation type
 */
char *RelocationTypeToString(int type) {
    switch (type) {
    case R_386_NONE:
	return "R_386_NONE";
    case R_386_32:
	return "R_386_32";
    case R_386_PC32:
	return "R_386_PC32";
    case R_386_GOT32:
	return "R_386_GOT32";
    case R_386_PLT32:
	return "R_386_PLT32";
    case R_386_COPY:
	return "R_386_COPY";
    case R_386_GLOB_DAT:
	return "R_386_GLOB_DAT";
    case R_386_JMP_SLOT:
	return "R_386_JMP_SLOT";
    case R_386_RELATIVE:
	return "R_386_RELATIVE";
    case R_386_GOTOFF:
	return "R_386_GOTOFF";
    case R_386_GOTPC:
	return "R_386_GOTPC";
    case R_386_32PLT:
	return "R_386_32PLT";
	/*
	 * This stuff is too new for redhat's below 8,
	 * so we'll leave them out for now...
	 *
	 * case R_386_TLS_TPOFF: return "R_386_TLS_TPOFF";
	 * case R_386_TLS_IE: return "R_386_TLS_IE";
	 * case R_386_TLS_GOTIE: return "R_386_TLS_GOTIE";
	 * case R_386_TLS_LE: return "R_386_TLS_LE";
	 * case R_386_TLS_GD: return "R_386_TLS_GD";
	 * case R_386_TLS_LDM: return "R_386_TLS_LDM";
	 * case R_386_16: return "R_386_16";
	 * case R_386_PC16: return "R_386_PC16";
	 * case R_386_8: return "R_386_8";
	 * case R_386_PC8: return "R_386_PC8";
	 * case R_386_TLS_GD_32: return "R_386_TLS_GD_32";
	 * case R_386_TLS_GD_PUSH: return "R_386_TLS_GD_PUSH";
	 * case R_386_TLS_GD_CALL: return "R_386_TLS_GD_CALL";
	 * case R_386_TLS_GD_POP: return "R_386_TLS_GD_POP";
	 * case R_386_TLS_LDM_32: return "R_386_TLS_LDM_32";
	 * case R_386_TLS_LDM_PUSH: return "R_386_TLS_LDM_PUSH";
	 * case R_386_TLS_LDM_CALL: return "R_386_TLS_LDM_CALL";
	 * case R_386_TLS_LDM_POP: return "R_386_TLS_LDM_POP";
	 * case R_386_TLS_LDO_32: return "R_386_TLS_LDO_32";
	 * case R_386_TLS_IE_32: return "R_386_TLS_IE_32";
	 * case R_386_TLS_LE_32: return "R_386_TLS_LE_32";
	 * case R_386_TLS_DTPMOD32: return "R_386_TLS_DTPMOD32";
	 * case R_386_TLS_DTPOFF32: return "R_386_TLS_DTPOFF32";
	 * case R_386_TLS_TPOFF32: return "R_386_TLS_TPOFF32";
	 */
    default:
	return "Unknown Relocation Type";
    }
}



/*
 *  GetSectionOrderScore(section) --
 *
 *  Returns a scoring of the specified Section to be used in determining
 *  the placement of the section in the final layout of the binary.
 *  Scoring is based on section flags, with ties broken by section number
 *  (which reflects original file order for sections that are not new).
 */
int GetSectionOrderScore(Section *section) {

    int n = section->s_number ? section->s_number : 999;

    if (SECTION_HAS_FLAG(section, SECTION_NOTE_FLAG)) {
	return 50;
    }

    if (KERNEL_MODE) {
	return n;
    }

    if (SECTION_HAS_FLAG(section, SECTION_EXECUTABLE_FLAG) &&
	SECTION_HAS_FLAG(section, SECTION_OCCUPIES_VM_FLAG)) {
	return 1000 + n;
    }

    if (SECTION_HAS_FLAG(section, SECTION_OCCUPIES_VM_FLAG) &&
	!SECTION_HAS_FLAG(section, SECTION_WRITABLE_FLAG)) {
	return 2000 + n;
    }

    if (section == SECTION_REF_DATA) {
	return 3000 + n;
    }

    if (SECTION_HAS_FLAG(section, SECTION_OCCUPIES_VM_FLAG) &&
	SECTION_HAS_FLAG(section, SECTION_WRITABLE_FLAG) &&
	SECTION_HAS_FLAG(section, SECTION_OCCUPIES_FILESPACE_FLAG)) {
	return 4000 + n;
    }

    if (SECTION_HAS_FLAG(section, SECTION_OCCUPIES_VM_FLAG) &&
	SECTION_HAS_FLAG(section, SECTION_WRITABLE_FLAG) &&
	!SECTION_HAS_FLAG(section, SECTION_OCCUPIES_FILESPACE_FLAG)) {
	return 5000 + n;
    }

    if (SECTION_HAS_FLAG(section, SECTION_SYMBOL_TABLE_FLAG)) {
	return 7000 + n;
    }

    if (SECTION_HAS_FLAG(section, SECTION_STRING_TABLE_FLAG)) {
	return 8000 + n;
    }

    return 6000 + n;
}



/*
 *  SortSectionsByWriteOrder() --
 *
 *  Sorts the elements of the global Section list by the order in which
 *  they should be written out.
 */
void SortSectionsByWriteOrder() {
    int num_sections, low, i, j;
    Section **list;
    Section *seciterator;

    REPORT(RP_DETAIL, "Sorting sections by write order");
    num_sections = 0;

    // assign order scores to the misc field for each section
    FOREACH_SECTION(seciterator) {
	seciterator->s_misc = GetSectionOrderScore(seciterator);
	num_sections++;
    }

    list = zalloc(sizeof(Section *) * num_sections);
    num_sections = 0;

    // put each section into an array index
    FOREACH_SECTION(seciterator) {
	list[num_sections++] = seciterator;
    }

    // do a selection sort to order the sections
    for (i = 0; i < num_sections; i++) {
	low = i;
	for (j = i + 1; j < num_sections; j++) {
	    if (list[j]->s_misc < list[low]->s_misc) {
		low = j;
	    }
	}

	if (low != i) {
	    seciterator = list[i];
	    list[i] = list[low];
	    list[low] = seciterator;
	}
    }

    // put the sections back into sec_list
    sec_list = list[0];
    sec_list->s_next = sec_list->s_prev = NULL;
    for (i = 0; i < num_sections; i++) {
	REPORT(RP_SMDETAIL,
	    "%2d.  (%4d)  %s", i+1, list[i]->s_misc, list[i]->s_name);
	if (i > 0) {
	    InsertSectionAfterSection(list[i], list[i - 1]);
	}
    }

}



/*
 *  WriteFileData(filename,data,data_size) --
 *
 *  Writes the bytes in data to a file filename of size data_size.
 *
 *  filename: the path to the file that should be created
 *  data: the array of bytes to be written to the file with path filename
 *  data_size: the size in bytes of the file that will be created
 */
void WriteFileData(char *filename, UINT8 *data, int data_size) {
    UINT32 mode;
    int fd;

    mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
    if ((output_binary_type == BINARY_EXECUTABLE_TYPE) ||
	(output_binary_type == BINARY_PLTO_RELOCATABLE_TYPE)) {
	mode |= S_IXUSR | S_IXGRP | S_IXOTH;
    }

    fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, mode);
    if (fd == -1) {
	PFATAL(filename);
    }

    write(fd, data, data_size);
    close(fd);
}
