/*
 *  write.c
 */
#include "global.h"
#include "utilities.h"
#include "relocate.h"
#include "sys/types.h"
#include "sys/stat.h"
#include "linkage.h"
#include "upm.h"
#include "protos.h"
#include "constants.h"

static bool REQUIRES_PROGRAM_HEADER;

typedef struct ElfSectionPatch {
    Section *esp_src_section;
    Elf32_Sym *esp_dest_elf_sym;
} ElfSectionPatch;


#define SECTION_FINAL_INDEX(section) ((section)->s_misc)

#define UPDATE_ALIGNMENT(alignment, offset) \
    while ((offset % alignment) != 0) offset++;

static int AddStringToTable(StringTable *tbl, char *new_str);
static Elf32_Ehdr *CreateNewElfHeader(int total_sections,
    ADDRESS shdr_tbl_offset);
static Elf32_Addr ComputePHOffset();
static Elf32_Half ComputeNoOfSegments();
static void TextGetSizes(Elf32_Phdr *pheader);
static void DataGetSizes(Elf32_Phdr *pheader);
static void NoteGetSizes(Elf32_Phdr *pheader);
static void NoteSegmentUpdateVirtualAndOffset(Elf32_Phdr *pheader,
    Elf32_Shdr *section_header_table);
static void DataSegmentUpdateVirtualAndOffset(Elf32_Phdr *pheader,
    Elf32_Shdr *section_header_table);
static void TextSegmentUpdateVirtualAndOffset(Elf32_Phdr *pheader,
    Elf32_Shdr *section_header_table);
static Elf32_Phdr *CreateProgramHeaderTable();
static void StoreSectionFileOffsets();
static void StoreSectionFinalIndices();
static void FinalizeSectionList();
static void AddPatchSectionIndexIntoElf_Sym(Section *src_section,
    Elf32_Sym *dest_sym, LinkedList *patch_list);
static Elf32_Shdr *CreateSectionHeaderTable(int *num_entries);
static void ShowSectionContents();
static void KernelFirstSegmentGetSizes(Elf32_Phdr *pheader);
static void KernelSecondSegmentGetSizes(Elf32_Phdr *pheader);
static void KernelFirstSegmentUpdateVirtualAndOffset(Elf32_Phdr *pheader,
    Elf32_Shdr *section_header_table); 
static void KernelSecondSegmentUpdateVirtualAndOffset(Elf32_Phdr *pheader,
    Elf32_Shdr *section_header_table);



/*
 *  CreateStringTable() --
 */
static StringTable *CreateStringTable() {
    StringTable *strtab;

    strtab = zalloc(sizeof(StringTable));
    strtab->st_data = zalloc(STRTAB_DEFAULT_SIZE);
    strtab->st_data[0] = '\0';
    strtab->st_entries = 0;
    strtab->st_next_offset = 1;
    strtab->st_datasize = STRTAB_DEFAULT_SIZE;

    return strtab;
}



/*
 *  CreateNewElfHeader(total_sections,shdr_tbl_offset) --
 *
 *  Creates the Elf header for the output binary file.
 */
static Elf32_Ehdr *CreateNewElfHeader(int total_sections,
    ADDRESS shdr_tbl_offset) {
    Elf32_Ehdr *new_header = zalloc(sizeof(Elf32_Ehdr));
    Section *seciterator;

    new_header->e_ident[EI_MAG0] = 0x7f;
    new_header->e_ident[EI_MAG1] = 'E';
    new_header->e_ident[EI_MAG2] = 'L';
    new_header->e_ident[EI_MAG3] = 'F';
    new_header->e_ident[EI_CLASS] = ELFCLASS32;
    new_header->e_ident[EI_DATA] = ELFDATA2LSB;
    new_header->e_ident[EI_VERSION] = EV_CURRENT;
    new_header->e_ident[EI_OSABI] = ELFOSABI_SYSV;
    new_header->e_ident[EI_ABIVERSION] = 0;
    new_header->e_ident[EI_PAD] = 0;
    new_header->e_type = BinaryTypeToElfType(output_binary_type);
    new_header->e_machine = EM_386;
    new_header->e_version = EV_CURRENT;

    if ((output_binary_type == BINARY_EXECUTABLE_TYPE) ||
	(output_binary_type == BINARY_PLTO_RELOCATABLE_TYPE)) {
	// get the virtual address of the first instruction
	// to be executed in the binary
	new_header->e_entry = startins->i_final_address;
    } else {
	new_header->e_entry = 0;
    }

    // assuming all headers have the standard length.
    new_header->e_ehsize = sizeof(Elf32_Ehdr);

    // inheriting the flags from the original file.
    new_header->e_flags = 0;

    // now to build the header tables.
    //
    //          ----------
    //          | ELF_HDR |
    //          -----------
    //          |  P H T  |
    //          -----------
    //          |  Sec 1  |
    //          -----------
    //          |  Sec n  |
    //          -----------
    //          |  S H T  |
    //          -----------
    //

    // Program header first.
    if (REQUIRES_PROGRAM_HEADER) {
	new_header->e_phentsize = sizeof(Elf32_Phdr);
	new_header->e_phnum = ComputeNoOfSegments();
	new_header->e_phoff = ComputePHOffset();
    } else {
	new_header->e_phentsize = 0;
	new_header->e_phnum = 0;
	new_header->e_phoff = 0;
    }
    // Section Header Next.
    new_header->e_shentsize = sizeof(Elf32_Shdr);
    new_header->e_shnum = total_sections;
    // compute the offset to the section header string table
    new_header->e_shoff = shdr_tbl_offset;
    while ((new_header->e_shoff % 4) != 0) {
	new_header->e_shoff++;
    }

    FOREACH_SECTION(seciterator) {
	if (strcmp(seciterator->s_name,SECTION_HEADER_STRING_TABLE_NAME) == 0) {
	    new_header->e_shstrndx = SECTION_FINAL_INDEX(seciterator);
	}
    }

    return new_header;
}



/*
 *  ComputePHOffset() --
 *  Compute the offset at which the program header table starts.
 *
 *  Algorithm
 *  If its required, assuming its the same as
 *  elf_header->ph_ehsize -- so as to make it
 *  immediately follow the elf header.
 */
static Elf32_Addr ComputePHOffset() {

    if (REQUIRES_PROGRAM_HEADER) {
	return sizeof(Elf32_Ehdr);
    } else {
	return 0;
    }
}



/*
 *  ComputeNoOfSegments() --
 *  Returns the number of segments in the file
 *  [Ref. CreateProgramHeaderTable()]
 */
static Elf32_Half ComputeNoOfSegments() {
    return 2;
}



/*
 *  CreateProgramHeaderTable() --
 *
 *  Creates program header table based on the original
 *  program. The assumptions we make and the section to
 *  segment mappings are described below :
 *
 *  _SEGMENT_ | _Type_ | _PERMS_ |_SECTIONS_
 *    Text    |  LOAD  |   R,X   | All sections with FLAGS A,(!W) & X
 *    Data    |  LOAD  |   R,W   | All sections with FLAGS A,W
 *    Note    |  NOTE  |    R    | All section_type = NOTE.
 *
 *  Not all values are filled here. This function simply
 *  returns a template -- dumpExecutable is where the meat is.
 *  TBF = To Be Filled.
 */
static Elf32_Phdr *CreateProgramHeaderTable() {
    Elf32_Phdr *ph_table = zalloc(sizeof(Elf32_Phdr) * ComputeNoOfSegments());

    if (!KERNEL_MODE) {
        // Layout the Text segment.
        ph_table[0].p_type = PT_LOAD;
        ph_table[0].p_offset = 0x00;           	// TBF
        ph_table[0].p_vaddr = 0x00;		// TBF
        ph_table[0].p_paddr = 0x00;
        TextGetSizes(&ph_table[0]);
        ph_table[0].p_flags = PF_R + PF_X;	// R&W
        ph_table[0].p_align = 0x1000;	// Page boundary

        // Layout the data segment.
        ph_table[1].p_type = PT_LOAD;
        ph_table[1].p_offset = 0x00;	        // TBF
        ph_table[1].p_vaddr = 0x00;		// TBF
        ph_table[1].p_paddr = 0x00;
        DataGetSizes(&ph_table[1]);
        ph_table[1].p_flags = PF_R + PF_W;	// R&W
        ph_table[1].p_align = 0x1000;	// Page boundary

        // Layout the note segment.
        ph_table[2].p_type = PT_NOTE;
        ph_table[2].p_offset = 0x00;	                // TBF
        ph_table[2].p_vaddr = 0x00;		        // TBF
        ph_table[2].p_paddr = 0x00;		        // ignored
        NoteGetSizes(&ph_table[2]);
        ph_table[2].p_flags = 0x4;		        // Read only perms.
        ph_table[2].p_align = 0x04;
    } else {
        // Layout the Text segment.
        ph_table[0].p_type = PT_LOAD;
        ph_table[0].p_offset = 0x00;	                // TBF
        ph_table[0].p_vaddr = 0x00;		        // TBF
        ph_table[0].p_paddr = 0x00;
        //KernelFirstSegmentGetSizes(&ph_table[0]);
        KernelSecondSegmentGetSizes(&ph_table[0]);
        ph_table[0].p_flags = PF_R + PF_W + PF_X;       // RWX
        ph_table[0].p_align = 0x1000;	                // Page boundary

        // Layout the data segment.
        //ph_table[1].p_type = PT_LOAD;
        ph_table[1].p_type = PT_GNU_STACK;
        ph_table[1].p_offset = 0x00;	                // TBF
        ph_table[1].p_vaddr = 0x00;		        // TBF
        ph_table[1].p_paddr = 0x00;
        //KernelSecondSegmentGetSizes(&ph_table[1]);
        ph_table[1].p_flags = PF_R + PF_W + PF_X;       // RWX 
        ph_table[1].p_align = 0x1000;	                // Page boundary

        // Layout the note segment.
        //ph_table[2].p_type = PT_GNU_STACK;
        //ph_table[2].p_offset = 0x00;	                // TBF
        //ph_table[2].p_vaddr = 0x00;		        // TBF
        //ph_table[2].p_paddr = 0x00;		        // ignored
        //ph_table[2].p_flags = PF_R + PF_W + PF_X;       // RWX
        //ph_table[2].p_align = 0x04;
    }



    return ph_table;
}



/*
 *  TextGetSizes(pheader) --
 *  Go through all sections in the TEXT Segment and compute sizes.
 */
static void TextGetSizes(Elf32_Phdr *pheader) {

    int mem_size = 0;
    Section *sec_iter, *last_section, *first_section;

    first_section = NULL;
    FOREACH_SECTION(sec_iter) {
	if (SECTION_HAS_FLAG(sec_iter, SECTION_OCCUPIES_VM_FLAG)) {
	    if (SECTION_HAS_FLAG(sec_iter, SECTION_WRITABLE_FLAG)) {
		break;
	    }

	    if (first_section == NULL) {
		first_section = sec_iter;
	    }

	    mem_size += sec_iter->s_newdatasize;
	    last_section = sec_iter;
	}
    }

    pheader->p_filesz =
	last_section->s_file_offset + last_section->s_newdatasize;
    pheader->p_memsz = last_section->s_newaddress +
	last_section->s_newdatasize - base_addr;
}



/*
 *  DataGetSizes(pheader) --
 *  Go through all sections in the DATA Segment and compute sizes.
 */
static void DataGetSizes(Elf32_Phdr *pheader) {

    int mem_size = 0;
    Section *sec_iter, *first_section, *last_section;

    first_section = NULL;
    FOREACH_SECTION(sec_iter) {
	if (SECTION_HAS_FLAG(sec_iter, SECTION_OCCUPIES_VM_FLAG) &&
	    SECTION_HAS_FLAG(sec_iter, SECTION_WRITABLE_FLAG)) {
	    if (first_section == NULL) {
		first_section = sec_iter;
	    }

	    mem_size += sec_iter->s_newdatasize;
	    last_section = sec_iter;
	}
    }

/*  This computation results in an incorrect filesz.  Experimentally, the
 *  file offset of the section after last_section is not equal to the file
 *  offset of the last_section + its size.
 *  pheader->p_filesz =
 *      last_section->s_next->s_file_offset -- first_section->s_file_offset
 */
    pheader->p_filesz = last_section->s_file_offset +
	last_section->s_newdatasize - first_section->s_file_offset;
    pheader->p_memsz = last_section->s_newaddress +
	last_section->s_newdatasize - first_section->s_newaddress;
}



/*      
 * KernelFirstSegmentGetSizes(pheader) -- 
 * Compute the size of the first kernel segment -- this segment
 * at the moment, statically maps all sections upto and including 
 * the .data section into the first segment. 
 */
static void KernelFirstSegmentGetSizes(Elf32_Phdr *pheader) {

    int mem_size = 0;
    Section *sec_iter, *first_section, *last_section;

    first_section = NULL;
    FOREACH_SECTION(sec_iter) {
        if (SECTION_HAS_FLAG(sec_iter, SECTION_OCCUPIES_VM_FLAG)) {
            if (first_section == NULL) {
                first_section = sec_iter;
            }
            mem_size += sec_iter->s_newdatasize;
            last_section = sec_iter;
            REPORT(RP_LGLOOP,
                "Mapping %s into segment 1", sec_iter->s_name);
            if (strcmp(sec_iter->s_name, ".bss") == 0) {
                break;
            }
        }
    }
    pheader->p_filesz = last_section->s_file_offset +
	last_section->s_newdatasize - first_section->s_file_offset;
    pheader->p_memsz = last_section->s_newaddress +
	last_section->s_newdatasize - first_section->s_newaddress;
}



/*      
 * KernelSecondSegmentGetSizes(pheader) --
 * Compute the size of the first kernel segment -- this segment
 * at the moment, statically maps all sections starting from and 
 * including the .data.init_task section into the second segment. 
 */
static void KernelSecondSegmentGetSizes(Elf32_Phdr *pheader) {
    int mem_size = 0;
    Section *sec_iter, *first_section, *last_section, *last_file_section;
    bool start = false;
    
    first_section = NULL;
    FOREACH_SECTION(sec_iter) {
        if (SECTION_HAS_FLAG(sec_iter, SECTION_OCCUPIES_VM_FLAG)) {
            if (first_section == NULL) {
                start = true;
                first_section = sec_iter;
            }
            if (start) {
                REPORT(RP_LGLOOP,
                    "Mapping %s into segment 2", sec_iter->s_name);
                mem_size += sec_iter->s_newdatasize;
                last_section = sec_iter;

                // the .bss section which is the last section mapped into 
                // this segment, does not occupy filespace. So don't include 
                // when computing file size
                if (SECTION_HAS_FLAG(sec_iter,
                        SECTION_OCCUPIES_FILESPACE_FLAG)) {
                    last_file_section = sec_iter;
                }
                
                if (strcmp(sec_iter->s_name, ".bss") == 0) {
                    break;
                }
            
            }
        }
    }
    pheader->p_filesz = last_file_section->s_file_offset +
	last_file_section->s_newdatasize - first_section->s_file_offset;
    pheader->p_memsz = last_section->s_newaddress +
	last_section->s_newdatasize - first_section->s_newaddress;
}



/*
 *  NoteGetSizes(pheader) --
 *  Go through all NOTE segments and compute the total memory size.
 */
static void NoteGetSizes(Elf32_Phdr *pheader) {
    int file_size = 0, mem_size = 0;
    Section *sec_iter;

    FOREACH_SECTION(sec_iter) {
	if (SECTION_HAS_FLAG(sec_iter, SECTION_NOTE_FLAG)) {
	    if (SECTION_HAS_FLAG(sec_iter, SECTION_OCCUPIES_FILESPACE_FLAG)) {
		file_size += sec_iter->s_newdatasize;
	    }

	    if (SECTION_HAS_FLAG(sec_iter, SECTION_OCCUPIES_VM_FLAG)) {
		mem_size += sec_iter->s_newdatasize;
	    }
	}
    }

    pheader->p_filesz = file_size;
    pheader->p_memsz = mem_size;
}



/*
 *  ComputeFirstSectionOffset() --
 */
ADDRESS ComputeFirstSectionOffset() {
    ADDRESS addr;

    addr = sizeof(Elf32_Ehdr);

    // Set global flags.
    REQUIRES_PROGRAM_HEADER = (output_binary_type == BINARY_EXECUTABLE_TYPE) ||
	(output_binary_type == BINARY_PLTO_RELOCATABLE_TYPE);

    // In case of an executable a PHT entry is required.
    if (REQUIRES_PROGRAM_HEADER) {
	addr += sizeof(Elf32_Phdr) * ComputeNoOfSegments();
    }

    if (KERNEL_MODE) {
	addr = 0x1000;		// for kernel, pad to page 1 boundary
    }

    return addr;
}



/*
 *  ComputeFirstSectionAddress() -- returns the first section address.
 *
 *  Computes the address of the first section based on the current section
 *  layout, sizes, and output binary type.
 */
ADDRESS ComputeFirstSectionAddress() {
    if (KERNEL_MODE) {
	return base_addr;
    } else {
	return ComputeFirstSectionOffset() + base_addr;
    }
}



/*
 *  WriteOutBinary() -- the "main" function for this part of the code.
 */
void WriteOutBinary() {
    unsigned char *outbuff;
    int fptr = 0;		// used as file pointer.
    int filesize = 0;
    int i;
    int shdr_tbl_offset;
    int total_sections;

    Elf32_Ehdr *elf_header;
    Elf32_Shdr *section_header_table;
    Elf32_Phdr *program_header_table;

    Section *sec_iter;

    REPORT(RP_MAJOR, "Preparing ELF binary output");
    FinalizeSectionList();
    ShowSectionContents();
    section_header_table = CreateSectionHeaderTable(&total_sections);

    // get the number of bytes that will be occupied by all of the sections,
    // and the program header table
    filesize = section_header_table[total_sections - 1].sh_offset;
    filesize += section_header_table[total_sections - 1].sh_size;
    while ((filesize % 4) != 0) {
	filesize++;
    }

    // this is the place where the section header table should reside
    shdr_tbl_offset = filesize;

    // add the space taken up by the section header table
    filesize += sizeof(Elf32_Shdr) * total_sections;

    REPORT(RP_MINOR, WITHFUNC("Building file image"));
    outbuff = zalloc(filesize);

    // Now onwards, I assume that all addresses
    // computed are final and the only thing left is to
    // create an executable image.

    // Layout data in the file.
    elf_header = CreateNewElfHeader(total_sections, shdr_tbl_offset);

    memcpy(outbuff, elf_header, sizeof(Elf32_Ehdr));
    fptr += sizeof(Elf32_Ehdr);

    if (REQUIRES_PROGRAM_HEADER) {
	program_header_table = CreateProgramHeaderTable();
	memcpy(outbuff + fptr, program_header_table,
	    sizeof(Elf32_Phdr) * ComputeNoOfSegments());
	fptr += sizeof(Elf32_Phdr) * ComputeNoOfSegments();
    }
    // MR -- At the moment I trust that the Section provides
    //       all the info I need -- if it doesn't, this can be
    //       modified trivially by adding stuff to the next loop.

    sec_iter = sec_list;
    for (i = 1; i < total_sections; i++) {

	if (SECTION_HAS_FLAG(sec_iter, SECTION_OCCUPIES_FILESPACE_FLAG) ||
	    SECTION_HAS_FLAG(sec_iter, SECTION_SYMBOL_TABLE_FLAG) ||
	    SECTION_HAS_FLAG(sec_iter, SECTION_STRING_TABLE_FLAG) ||
	    SECTION_HAS_FLAG(sec_iter, SECTION_NOTE_FLAG) ||
	    SECTION_HAS_FLAG(sec_iter, SECTION_RELOCATIONS_FLAG)) {

	    if (SECTION_HAS_FLAG(sec_iter, SECTION_OCCUPIES_VM_FLAG) ||
		SECTION_HAS_FLAG(sec_iter, SECTION_SYMBOL_TABLE_FLAG) ||
		SECTION_HAS_FLAG(sec_iter, SECTION_STRING_TABLE_FLAG) ||
		SECTION_HAS_FLAG(sec_iter, SECTION_RELOCATIONS_FLAG)) {

	    	REPORT(RP_TRIVIAL, "memcpying   section %s", sec_iter->s_name);
		memcpy(outbuff + section_header_table[i].sh_offset,
		    sec_iter->s_newdata, sec_iter->s_newdatasize);
	    } else {
	    	REPORT(RP_TRIVIAL, "not copying section %s", sec_iter->s_name);
	    }
	} else {
	    	REPORT(RP_TRIVIAL, "ignoring    section %s", sec_iter->s_name);
	}

	sec_iter = sec_iter->s_next;
    }

    // Now that we know the address that segments reside in
    // update the program header table in memory.
    if (REQUIRES_PROGRAM_HEADER) {

        if (KERNEL_MODE) {
            KernelFirstSegmentUpdateVirtualAndOffset(
                &program_header_table[0], section_header_table);
        } else {
	    TextSegmentUpdateVirtualAndOffset(&program_header_table[0],
	        section_header_table);
	    DataSegmentUpdateVirtualAndOffset(&program_header_table[1],
	        section_header_table);
	    NoteSegmentUpdateVirtualAndOffset(&program_header_table[2],
	        section_header_table);
        }
    }

    if (REQUIRES_PROGRAM_HEADER) {
	// make this hardcoded *VVVVVV* value a global.
	memcpy(outbuff + sizeof(Elf32_Ehdr), program_header_table,
	    sizeof(Elf32_Phdr) * ComputeNoOfSegments());
    }

    memcpy(outbuff + shdr_tbl_offset,
	section_header_table, sizeof(Elf32_Shdr) * total_sections);

    REPORT(RP_MINOR, "Writing binary file: %s", newFilename);
    WriteFileData(newFilename, outbuff, filesize);
}



/*
 *  KernelFirstSegmentUpdateVirtualAndOffset(pheader,section_header_table) --
 */
static void KernelFirstSegmentUpdateVirtualAndOffset(Elf32_Phdr *pheader,
    Elf32_Shdr *section_header_table) {

    Section *sec_iter;

    // Find the first kernel section.
    FOREACH_SECTION(sec_iter) {
        if (strcmp(sec_iter->s_name, ".text") == 0) {
            pheader->p_offset= section_header_table[sec_iter->s_misc].sh_offset;
	    pheader->p_vaddr = section_header_table[sec_iter->s_misc].sh_addr;
	    pheader->p_paddr = section_header_table[sec_iter->s_misc].sh_addr;
	    return;
	}
    }
}



/*
 *  KernelSecondSegmentUpdateVirtualAndOffset(pheader,section_header_table) --
 */
static void KernelSecondSegmentUpdateVirtualAndOffset(Elf32_Phdr *pheader,
    Elf32_Shdr *section_header_table) {

    Section *sec_iter;

    // Find the first section for the second kernel segment
    FOREACH_SECTION(sec_iter) {
        if (strcmp(sec_iter->s_name, ".data") == 0) {
            pheader->p_offset= section_header_table[sec_iter->s_misc].sh_offset;
	    pheader->p_vaddr = section_header_table[sec_iter->s_misc].sh_addr;
	    pheader->p_paddr = section_header_table[sec_iter->s_misc].sh_addr;
	    return;
	}
    }
}



/*
 *  TextSegmentUpdateVirtualAndOffset(pheader,section_header_table) --
 */
static void TextSegmentUpdateVirtualAndOffset(Elf32_Phdr *pheader,
    Elf32_Shdr *section_header_table) {

    Section *sec_iter;

    // Find the first text section.
    FOREACH_SECTION(sec_iter) {
	if (SECTION_HAS_FLAG(sec_iter, SECTION_OCCUPIES_VM_FLAG) &&
	    !SECTION_HAS_FLAG(sec_iter, SECTION_NOTE_FLAG) &&
	    !SECTION_HAS_FLAG(sec_iter, SECTION_WRITABLE_FLAG)) {

	    pheader->p_offset= section_header_table[sec_iter->s_misc].sh_offset;
	    pheader->p_vaddr = section_header_table[sec_iter->s_misc].sh_addr;
	    pheader->p_paddr = section_header_table[sec_iter->s_misc].sh_addr;
	    return;
	}
    }
}



/*
 *  DataSegmentUpdateVirtualAndOffset(pheader,section_header_table) --
 */
static void DataSegmentUpdateVirtualAndOffset(Elf32_Phdr *pheader,
    Elf32_Shdr *section_header_table) {

    Section *sec_iter;

    // Find the first data section.
    FOREACH_SECTION(sec_iter) {
	if (SECTION_HAS_FLAG(sec_iter, SECTION_OCCUPIES_VM_FLAG) &&
	    SECTION_HAS_FLAG(sec_iter, SECTION_WRITABLE_FLAG)) {
	    pheader->p_offset =section_header_table[sec_iter->s_misc].sh_offset;
	    pheader->p_vaddr = section_header_table[sec_iter->s_misc].sh_addr;
	    pheader->p_paddr = section_header_table[sec_iter->s_misc].sh_addr;
	    return;
	}

    }
}



/*
 *  NoteSegmentUpdateVirtualAndOffset(pheader,section_header_table) --
 */
static void NoteSegmentUpdateVirtualAndOffset(Elf32_Phdr *pheader,
    Elf32_Shdr *section_header_table) {
    Section *sec_iter;


    // Find the first note section.
    FOREACH_SECTION(sec_iter) {
	if (SECTION_HAS_FLAG(sec_iter, SECTION_NOTE_FLAG)) {
	    pheader->p_offset =
		section_header_table[sec_iter->s_misc].sh_offset;
	    pheader->p_vaddr = section_header_table[sec_iter->s_misc].sh_addr;
	    pheader->p_paddr = section_header_table[sec_iter->s_misc].sh_addr;
	    return;
	}

    }
}



/*
 *  AddStringToTable(tbl,new_str) --
 */
static int AddStringToTable(StringTable *tbl, char *new_str) {
    int offset;
    UINT8 *new_buffer;

    CASSERT(tbl != NULL, "Cannot add to a NULL string table");
    CASSERT(new_str != NULL, "Cannot add a NULL entry to string table");

    offset = tbl->st_next_offset;

    // if we need more room for the strings we will grow our string
    if (offset + strlen(new_str) >= tbl->st_datasize) {
	new_buffer = zalloc(tbl->st_datasize * 2);

	memcpy(new_buffer, tbl->st_data, tbl->st_datasize);
	tbl->st_datasize *= 2;

	free(tbl->st_data);
	tbl->st_data = new_buffer;
    }

    strcpy((char *) &tbl->st_data[tbl->st_next_offset], new_str);
    tbl->st_next_offset += strlen(new_str) + 1;
    tbl->st_entries++;

    return offset;
}



/*
 *  SortSymbolsForTable() --
 */
void SortSymbolsForTable() {
    Symbol *front, *back, *tmp;

    REPORT(RP_DETAIL, "Sorting symbols for output symbol table");

    // put the back pointer on the last symbol
    FOREACH_SYMBOL(back) {
	if (back->sy_next == NULL) {
	    break;
	}
    }

    // go from both ends of the list and replace any non-local symbols with a
    // local symbol from the end of the list -- by the time the pointers meet,
    // we know that there are only non-local symbols after the front pointer
    // and no non-local symbols before it
    front = s_list;
    while (front != back) {
	if (!(front->sy_flags & SYMBOL_LOCAL_FLAG)) {
	    if (back->sy_flags & SYMBOL_LOCAL_FLAG) {
		// remove the back node
		tmp = back->sy_prev;
		RemoveSymbol(back);

		InsertSymbolBeforeSymbol(back, front);
		back = tmp;
	    } else {
		back = back->sy_prev;
	    }
	} else {
	    front = front->sy_next;
	}
    }

}



/*
 *  CreateSymbolTable(symstrtab_ptr,symbol_patches) --
 */
Section *CreateSymbolTable(Section **symstrtab_ptr, LinkedList *symbol_patches){
    StringTable *struct_strtab;
    Symbol *symiterator;
    Section *symstrtab, *symtab;
    Elf32_Sym *sym;
    int total_syms, offset, last_local;

    REPORT(RP_MINOR, WITHFUNC("Creating symbol table"));

    SortSymbolsForTable();

    struct_strtab = CreateStringTable();

    last_local = total_syms = 0;
    FOREACH_SYMBOL(symiterator) {
	symiterator->sy_temp = (void *) total_syms;
	if (symiterator->sy_flags & SYMBOL_LOCAL_FLAG) {
	    last_local = total_syms;
	}

	total_syms++;
    }

    // account for first UNDEF symbol
    total_syms++;
    last_local++;

    REPORT(RP_DETAIL, "Total symbols: %d", total_syms);
    REPORT(RP_DETAIL, "Last local symbol: %d", last_local);

    symtab = zalloc(sizeof(Section));
    symtab->s_name = salloc(SECTION_SYMBOL_TABLE_NAME);
    symtab->s_flags = SECTION_SYMBOL_TABLE_FLAG;
    symtab->s_newdatasize = total_syms * sizeof(Elf32_Sym);
    symtab->s_newdata = zalloc(symtab->s_newdatasize);
    symtab->s_lsend = last_local;
    symtab->s_alignment = 4;

    offset = 0;
    sym = (Elf32_Sym *) &symtab->s_newdata[offset];
    sym->st_name = 0;
    sym->st_value = 0;
    sym->st_size = 0;
    sym->st_info = ELF32_ST_INFO(STB_LOCAL, STT_NOTYPE);
    sym->st_other = 0;
    sym->st_shndx = SHN_UNDEF;

    symiterator = s_list;
    offset += sizeof(Elf32_Sym);
    while (offset < symtab->s_newdatasize) {
	sym = (Elf32_Sym *) &symtab->s_newdata[offset];

	sym->st_name = AddStringToTable(struct_strtab, symiterator->sy_name);
	sym->st_value = symiterator->sy_value;
	sym->st_size = symiterator->sy_size;
	sym->st_other = 0;
	FillElf32_SymFromPltoFlags(sym, symiterator->sy_flags);

	if ((symiterator->sy_section != NULL) &&
	    !(symiterator->sy_flags & SYMBOL_COMMON_FLAG) &&
	    !(symiterator->sy_flags & SYMBOL_ABSOLUTE_FLAG)) {
	    AddPatchSectionIndexIntoElf_Sym(symiterator->sy_section, sym,
		symbol_patches);
	}

	offset += sizeof(Elf32_Sym);
	symiterator = symiterator->sy_next;
    }

    symstrtab = zalloc(sizeof(Section));
    symstrtab->s_name = salloc(SECTION_SYMBOL_STRING_TABLE_NAME);
    symstrtab->s_flags = SECTION_STRING_TABLE_FLAG;
    symstrtab->s_newdata = struct_strtab->st_data;
    symstrtab->s_newdatasize = struct_strtab->st_next_offset;
    symstrtab->s_alignment = 1;

    *symstrtab_ptr = symstrtab;
    return symtab;
}



/*
 *  InsertSectionHeaderStringTable() --
 */
void InsertSectionHeaderStringTable() {
    Section *strtab, *seciterator, *last_section;
    StringTable *struct_strtab;

    strtab = zalloc(sizeof(Section));
    strtab->s_name = salloc(SECTION_HEADER_STRING_TABLE_NAME);
    strtab->s_flags = SECTION_STRING_TABLE_FLAG;
    strtab->s_alignment = 1;

    struct_strtab = CreateStringTable();

    FOREACH_SECTION(seciterator) {
	seciterator->s_name_offset =
	    AddStringToTable(struct_strtab, seciterator->s_name);
	last_section = seciterator;
    }

    strtab->s_name_offset = AddStringToTable(struct_strtab, strtab->s_name);

    // set up the new data for the string table
    strtab->s_newdata = struct_strtab->st_data;
    strtab->s_newdatasize = struct_strtab->st_next_offset;

    InsertSectionAfterSection(strtab, last_section);
    last_section = last_section->s_next;
    REPORT(RP_DETAIL, "Inserted string section %s with size %x",
	last_section->s_name, last_section->s_newdatasize);
}



/*
 *  InsertPltoRelocationSections(upm_list) --
 *
 *  NOTE: Ignores its argument.
 */
void InsertPltoRelocationSections(LinkedList *upm_list) {
    Patch *uiterator;
    Iterator *itr;
    HashTable *sections, *rel_syms;
    LinkedList *list, *new_sections;
    Elf32_Rel *rel;
    Section *seciterator, *section;
    Symbol *symbol;
    ADDRESS offset;
    UINT32 type;
    int data_offset;

    REPORT(RP_MINOR, WITHFUNC("Creating relocation sections"));

    sections = ht_new(HASH_PTR, HASH_PTR, 0xff, .7);
    rel_syms = ht_new(HASH_PTR, HASH_PTR, 0xfff, .7);
    new_sections = ll_new(LIST_UNSORTED);

    FOREACH_SECTION(seciterator) {
	ht_put(sections, seciterator->s_name, ll_new(LIST_UNSORTED));
    }

    itr = ll_iterator(upm2s_list);

    while (itr->it_hasNext(itr)) {
	uiterator = (Patch *) itr->it_next(itr);

	GetFakeRelocationInfoForUPM(uiterator, &type, &offset, &section);

	if ((type == -1) ||
	    (offset == -1) || (section == NULL) || (symbol == NULL)) {
	    WARNING("Could not convert UPM into relocation!");
	    continue;
	}

	rel = zalloc(sizeof(Elf32_Rel));
	rel->r_offset = offset;
	rel->r_info = ELF32_R_INFO(0, type);

	ll_add((LinkedList *) ht_get(sections, section->s_name), rel);
	ht_put(rel_syms, rel, symbol);
    }

    FOREACH_SECTION(seciterator) {
	list = (LinkedList *) ht_get(sections, seciterator->s_name);

	if (ll_size(list) > 0) {
	    Section *new_section;

	    new_section = zalloc(sizeof(Section));
	    new_section->s_name = zalloc(strlen(seciterator->s_name) +
		    strlen(RELOCATION_SECTION_PREFIX) + 1);
	    sprintf(new_section->s_name, "%s%s",
		RELOCATION_SECTION_PREFIX, seciterator->s_name);
	    new_section->s_flags = SECTION_RELOCATIONS_FLAG;
	    new_section->s_alignment = 4;
	    new_section->s_newdatasize = new_section->s_datasize =
		sizeof(Elf32_Rel) * ll_size(list);
	    new_section->s_newdata = new_section->s_data =
		zalloc(new_section->s_newdatasize);
	    new_section->s_relsect = seciterator;

	    data_offset = 0;

	    itr = ll_iterator(list);
	    while (itr->it_hasNext(itr)) {
		rel = (Elf32_Rel *) itr->it_next(itr);

		memcpy(&new_section->s_newdata[data_offset], rel,
		    sizeof(Elf32_Rel));
		data_offset += sizeof(Elf32_Rel);
	    }
	    free(itr);

	    ll_add(new_sections, new_section);
	}
    }

    itr = ll_iterator(new_sections);
    while (itr->it_hasNext(itr)) {
	seciterator = (Section *) itr->it_next(itr);

	InsertSectionLast(seciterator, sec_list);
    }

    // clean up our hashtable and lists
    itr = ht_iterator(sections);
    while (itr->it_hasNext(itr)) {
	list = (LinkedList *) itr->it_next(itr);

	ll_clear(list);
	ll_free(list);
    }

    ht_clear(sections);
    ht_free(sections);
    ht_free(rel_syms);
    free(new_sections);
}



/*
 *  FinalizeSectionList() --
 *
 *  This function creates and adds the section header string table to the list
 *  of Sections and creates and adds the symbol and symbol string table
 *  as well if a symbol table is being generated.
 */
void FinalizeSectionList() {
    Section *seciterator, *symtab, *symstrtab;
    LinkedList *symbol_upms;

    symbol_upms = NULL;

    if (output_binary_type == BINARY_PLTO_RELOCATABLE_TYPE) {
	// add relocation sections
	symbol_upms = ll_new(LIST_UNSORTED);

	InsertPltoRelocationSections(symbol_upms);
    }
    // get the last section
    FOREACH_SECTION(seciterator) {
	if (seciterator->s_next == NULL) {
	    break;
	}
    }

    if (BooleanParameter("GenerateSymbolTable")) {
	if (symbol_upms == NULL) {
	    symbol_upms = ll_new(LIST_UNSORTED);
	}

	symtab = CreateSymbolTable(&symstrtab, symbol_upms);

	InsertSectionAfterSection(symtab, seciterator);
	InsertSectionAfterSection(symstrtab, symtab);
    }

    InsertSectionHeaderStringTable();

    StoreSectionFinalIndices();
    StoreSectionFileOffsets();

    if (BooleanParameter("GenerateSymbolTable")) {
	ApplyUPMs(symbol_upms);
    }
}



/*
 *  CreateSectionHeaderTable(num_entries) --
 */
Elf32_Shdr *CreateSectionHeaderTable(int *num_entries) {
    Section *seciterator;
    Elf32_Shdr *tbl;
    int i;

    *num_entries = 0;
    FOREACH_SECTION(seciterator) {
	(*num_entries)++;
    }

    // save room for the first section
    (*num_entries)++;

    tbl = zalloc(sizeof(Elf32_Shdr) * (*num_entries));

    // Create the first element -- the NULL Entry.
    tbl[0].sh_name = 0;
    tbl[0].sh_type = SHT_NULL;
    tbl[0].sh_flags = 0;
    tbl[0].sh_addr = 0;
    tbl[0].sh_offset = 0;
    tbl[0].sh_size = 0;
    tbl[0].sh_link = SHN_UNDEF;
    tbl[0].sh_info = 0;
    tbl[0].sh_addralign = 0;
    tbl[0].sh_entsize = 0;

    seciterator = sec_list;
    for (i = 1; i < *num_entries; i++) {
	tbl[i].sh_name = seciterator->s_name_offset;
	FillElf32_ShdrFromPltoFlags(&tbl[i], seciterator->s_flags);
	if ((output_binary_type == BINARY_EXECUTABLE_TYPE) ||
	    (output_binary_type == BINARY_PLTO_RELOCATABLE_TYPE)) {
	    tbl[i].sh_addr = seciterator->s_newaddress;
	} else {
	    tbl[i].sh_addr = 0;
	}
	tbl[i].sh_offset = seciterator->s_file_offset;
	tbl[i].sh_size = seciterator->s_newdatasize;
	tbl[i].sh_addralign = seciterator->s_alignment;
	if (SECTION_HAS_FLAG(seciterator, SECTION_SYMBOL_TABLE_FLAG)) {
	    tbl[i].sh_entsize = sizeof(Elf32_Sym);
	    tbl[i].sh_link = i + 1;
	    tbl[i].sh_info = seciterator->s_lsend;
	} else if (SECTION_HAS_FLAG(seciterator, SECTION_RELOCATIONS_FLAG)) {
	    tbl[i].sh_entsize = sizeof(Elf32_Rel);
	    tbl[i].sh_link = *num_entries - 3;
	    tbl[i].sh_info = SECTION_FINAL_INDEX(seciterator->s_relsect);
	} else {
	    // <actually we need to account for other types of sections that
	    // might need this such as relocation sections -- which haven't
	    // been added just yet>
	    // <cml>
	    tbl[i].sh_entsize = 0;
	}

//  PrintSectionHeader(&tbl[i], ((seciterator)->s_name));

	seciterator = seciterator->s_next;
    }

    return tbl;
}



/*
 *  PatchSectionIndexIntoElf_Sym(upm) --
 */
static bool PatchSectionIndexIntoElf_Sym(Patch *upm) {
    ElfSectionPatch *patch;

    patch = (ElfSectionPatch *) upm->p_patch;

    CASSERT(patch->esp_src_section != NULL, "Patching a NULL section");
    CASSERT(patch->esp_dest_elf_sym != NULL, "Patching a NULL symbol");

    patch->esp_dest_elf_sym->st_shndx =
	SECTION_FINAL_INDEX(patch->esp_src_section);

    return false;
}



/*
 *  AddPatchSectionIndexIntoElf_Sym(src_section,dest_sym,patch_list) --
 */
static void AddPatchSectionIndexIntoElf_Sym(Section *src_section,
    Elf32_Sym *dest_sym, LinkedList *patch_list) {
    Patch *upm;
    ElfSectionPatch *patch;

    patch = zalloc(sizeof(ElfSectionPatch));
    patch->esp_src_section = src_section;
    patch->esp_dest_elf_sym = dest_sym;

    upm = UPMCreate(patch, &PatchSectionIndexIntoElf_Sym, NULL, NULL, NULL);

    ll_add(patch_list, upm);
}



/*
 *  StoreSectionFileOffsets() --
 */
static void StoreSectionFileOffsets() {
    Section *seciterator;
    ADDRESS offset;

    offset = ComputeFirstSectionOffset();

    if (offset > (sec_list->s_newaddress - base_addr)) {
	REPORT(RP_DETAIL, "Updating %s address from %x", sec_list->s_name,
	    sec_list->s_newaddress);
	if (!KERNEL_MODE) {
	    sec_list->s_newaddress = base_addr + offset;
	}
	UpdateSectionSizes(true);
	REPORT(RP_DETAIL, "Updated %s address to %x", sec_list->s_name,
	    sec_list->s_newaddress);
    }

    REPORT(RP_DETAIL, "First section offset = 0x%8.8x  [%s]", offset, __func__);

    FOREACH_SECTION(seciterator) {

	UPDATE_ALIGNMENT(seciterator->s_alignment, offset);

	while ((offset % PAGE_SIZE) != (seciterator->s_newaddress % PAGE_SIZE)){
	    offset++;
	}
	// ((seciterator)->s_newaddress) = offset + base_addr;

	seciterator->s_file_offset = offset;

	if (SECTION_HAS_FLAG(seciterator, SECTION_OCCUPIES_FILESPACE_FLAG) ||
	    SECTION_HAS_FLAG(seciterator, SECTION_SYMBOL_TABLE_FLAG) ||
	    SECTION_HAS_FLAG(seciterator, SECTION_STRING_TABLE_FLAG) ||
	    SECTION_HAS_FLAG(seciterator, SECTION_RELOCATIONS_FLAG)) {
	    offset += seciterator->s_newdatasize;
	}
    }
}



/*
 *  ShowSectionContents() --
 *  Utility function that shows the state of the sections before writeout.
 */
void ShowSectionContents() {
    Section *seciterator;
    int i = 0;

    REPORT(RP_SMLOOP, WITHFUNC("Offset  Address   Size  Ix  Name "));
    FOREACH_SECTION(seciterator) {
	REPORT(RP_SMLOOP, "%6X %8X %6X %3d  %s",
	    seciterator->s_file_offset, seciterator->s_newaddress,
	    seciterator->s_newdatasize, ++i, seciterator->s_name);
    }
}



/*
 *  StoreSectionFinalIndices() --
 */
static void StoreSectionFinalIndices() {
    Section *seciterator;
    int i = 1;

    FOREACH_SECTION(seciterator) {
	SECTION_FINAL_INDEX(seciterator) = i++;
    }
}
