/*
 *  This file contains stuff specific to the Block data structure
 *  which represents a single basic block. Each basic block associates a
 *  linear sequence of instructions for which there is a single control
 *  flow entry point and a single control flow exit point.
 */
#ifndef BBL_H
#define BBL_H

#include "plto_types.h"
#include "instr.h"



/*
 *  The representation used for a Basic Block.
 */
typedef struct Block {
    struct BlockInfo *b_auxinfo;   // auxiliary info
    UINT16 b_type;		   // basic block type, e.g. normal, branch
    INT64 b_flags;		   // assorted flags
    EDGE_WEIGHT_TYPE b_weight;	   // block weight (profiled execution count)
    struct Instruction *b_first;   // first instruction within this block
    struct Instruction *b_last;	   // last instruction within this block
    struct Block *b_next;	   // next basic block in the list
    struct Block *b_prev;	   // previous basic block in the list
    struct EdgeNode *b_pred;	   // list of bbls that can transfer to here
    struct EdgeNode *b_succ;	   // list of bbls to which control can flow
    struct Function *b_fun;	   // function to which this bbl belongs
    struct FList *b_flist;	   // ???
    UINT32 b_id;		   // ID number for this basic block
} Block;

/*
 *  Macros for accessing fields in the basic blocks
 */
#define BBL_SET_FLAG(block, z) (((block)->b_flags) |= z)
#define BBL_HAS_FLAG(block, z) ((bool)(((block)->b_flags) & z))
#define BBL_REMOVE_FLAG(block, z) (((block)->b_flags) &= (~z))

/*
 *  Basic Block Types
 */
#define BT_INVALID 0	// BBL of this type has been set up incorrectly
#define BT_UNCOND 1	// BBL ending with an unconditional jump
#define BT_NORMAL 2	// BBL with no control transfer at the end
#define BT_BRANCH 3	// BBL ending with a conditional jump instruction
#define BT_RETURN 4	// BBL ending with a return instruction
#define BT_SWITCH 5	// BBL which ends with a jump through a jump table
#define BT_HELL 6	// HELL
#define BT_ENTRY 7	// abstract pseudo entry block for a function
#define BT_EXIT 8	// abstract pseudo exit block for a function
#define BT_CALL 9	// BBL ending with a function call instruction
#define BT_HALT 10	// BBL in which the program terminates
#define BT_SYSCALL 11	// basic block that ends in a system call
#define BT_DATABLOCK 12	// basic block that ends in a system call



/*
 *  Auxiliary information for a basic block.
 */
typedef struct BlockInfo {
    int bi_tmp;                    // temporary storage space
    long bi_fprint;                // fingerprint for code compaction
    struct Value *bi_stackshot;
    struct Value *bi_regshot;
    struct Block *bi_chainnext;
    struct Block *bi_chainprev;
    struct Block *bi_chainhead;
    struct Block *bi_headnext;
    struct Block *bi_headprev;
    struct Block *bi_nexttochain;
    REGISTER_BITVECTOR bi_may_use_in;
    REGISTER_BITVECTOR bi_bypass_in;
    REGISTER_BITVECTOR bi_may_use_out;
    REGISTER_BITVECTOR bi_bypass_out;
    REGISTER_BITVECTOR bi_ubd;
    REGISTER_BITVECTOR bi_def;
    double bi_frequency;           // frequency
    struct Block *bi_preDom;	   // pre-dominator of this block
    struct Block *bi_postDom;	   // post-dominator of this block
    /* The following fields are related to blocks that are contained in loops */
    struct Block *bi_loopHead;	   // head (entry block) for the loop
    struct Block *bi_loopNext;	   // next block in the loop
    struct Block *bi_loopFirst;	   // first block in loop (if loop header)
    struct Block *bi_clone;	   // bbl from which this was cloned
    REGISTER_BITVECTOR bi_liveIn;  // live registers going into the bbl
    REGISTER_BITVECTOR bi_liveOut; // live registers leaving the bbl
    PSW_BITVECTOR bi_liveInPSW;	   // live PSW bits going into the bbl
    PSW_BITVECTOR bi_liveOutPSW;   // live PSW bits leaving the bbl
    UINT32 bi_BPSPdiff;		   // difference between ESP and EBP at entry
    UINT32 bi_stack_cont;	   // net stack contribution of the bbl
} BlockInfo;

/*
 *  Shortcut macros allowing BlockInfo fields to be accessed
 *  as if part of the main Block struct.
 */
#define	b_tmp		b_auxinfo->bi_tmp
#define b_fprint        b_auxinfo->bi_fprint
#define	b_stackshot	b_auxinfo->bi_stackshot
#define	b_regshot	b_auxinfo->bi_regshot
#define	b_chainnext	b_auxinfo->bi_chainnext
#define	b_chainprev	b_auxinfo->bi_chainprev
#define	b_chainhead	b_auxinfo->bi_chainhead
#define	b_headnext	b_auxinfo->bi_headnext
#define	b_headprev	b_auxinfo->bi_headprev
#define	b_nexttochain	b_auxinfo->bi_nexttochain
#define	b_may_use_in	b_auxinfo->bi_may_use_in
#define	b_bypass_in	b_auxinfo->bi_bypass_in
#define	b_may_use_out	b_auxinfo->bi_may_use_out
#define	b_bypass_out	b_auxinfo->bi_bypass_out
#define	b_ubd		b_auxinfo->bi_ubd
#define	b_def		b_auxinfo->bi_def
#define	b_frequency	b_auxinfo->bi_frequency
#define	b_preDom	b_auxinfo->bi_preDom
#define	b_postDom	b_auxinfo->bi_postDom
#define	b_loopHead	b_auxinfo->bi_loopHead
#define	b_loopNext	b_auxinfo->bi_loopNext
#define	b_loopFirst	b_auxinfo->bi_loopFirst
#define	b_clone		b_auxinfo->bi_clone
#define	b_liveIn	b_auxinfo->bi_liveIn
#define	b_liveOut	b_auxinfo->bi_liveOut
#define	b_liveInPSW	b_auxinfo->bi_liveInPSW
#define	b_liveOutPSW	b_auxinfo->bi_liveOutPSW
#define	b_BPSPdiff	b_auxinfo->bi_BPSPdiff
#define	b_stack_cont	b_auxinfo->bi_stack_cont



/*
 *  FUNCTIONS
 */
extern void BBLFree(Block *bbl);
extern Block *BBLCreate();
extern BlockInfo *BBLCreateAuxinfo();
extern Block *BBLCreateBase();

#endif
