/*
 *  An implementation of a directed, weighted graph.
 *
 *  Pretty memory heavy as it takes O(num_vertices ^ 2), but it should
 *  be pretty fast with constant time edge insertion and edge lookup.
 *  Predecessor iteration and successor iteration should run in O(n)
 *  where n is the number of predecessors or successors.
 *
 *  The number of vertices must be specified at initialization time.
 */
#ifndef D_GRAPH_H
#define D_GRAPH_H

#include "global.h"

/*
 *  DigraphVertex
 */
typedef struct DigraphVertex {
    UINT32 dv_id;			// vertex id in 0 ... n-1
    struct DigraphVertex *dv_next;	// next vertex
    struct SchedPoints *dv_data;	// data values
} DigraphVertex;


/*
 *  DigraphEdge
 */
typedef struct DigraphEdge {
    DigraphVertex *di_source;			// source vertex
    DigraphVertex *di_target;			// target vertex
    UINT8 di_weight;				// edge weight
    struct DigraphEdge *di_next_successor;	// next successor of source
    struct DigraphEdge *di_next_predecessor;	// next predecessor of target
    bool di_exists;		// true if this edge has been added to the graph
} DigraphEdge;


/*
 *  Digraph
 */
typedef struct Digraph {
    UINT32 d_size;			// maximum number of vertices
    DigraphVertex *d_vertices;		// array of vertices
    DigraphEdge *d_edges;		// 2-D array of edges
    DigraphEdge **d_successors;		// first successor for each vertex
    DigraphEdge **d_predecessors;	// first predecessor for each vertex
} Digraph;

#define DGRAPH_GET_EDGE(x, v1, v2) \
    (&((x)->d_edges[(v1)->dv_id * x->d_size + (v2)->dv_id]))
#define DGRAPH_HAS_EDGE(x, v1, v2) \
    ((x)->d_edges[(v1)->dv_id * x->d_size + (v2)->dv_id].di_exists == true)
UINT32 DGRAPH_IN_DEGREE(Digraph *graph, DigraphVertex *s);
UINT32 DGRAPH_OUT_DEGREE(Digraph *graph, DigraphVertex *s);



/*
 *  FUNCTIONS
 */

Digraph *CreateDGraph(int max_size);
void ClearDGraph(Digraph *graph);
void DestroyDGraph(Digraph *graph);
void CreateEdgeDGraph(Digraph *graph, DigraphVertex *s, DigraphVertex *d,
    UINT8 weight);
void PrintGraph(Digraph *graph);

#endif
