/*
 *  This file contains functions for computing pre and post dominator trees as
 *  well as loop information, and storing these relationships to basic blocks.
 *  There are also some general purpose dominator functions for distinguishing
 *  the relationships between given basic blocks and nodes in a dominator tree.
 */
#ifndef DOMINATOR_H
#define DOMINATOR_H

#include "global.h"

/*
 *  Represents a single node in a dominator tree with children, a parent,
 *  back edges and children.
 */
typedef struct TreeNode {
    struct TreeNode *t_children;	// children of this node
    struct TreeNode *t_parent;		// parent of this node
    struct TreeNode *t_next_child;	// next child in the linked list
    Block *t_bbl;			// basic block for the node
} TreeNode;

/*
 *  A shortcut to the last child in the child list of some dominator node.
 */
#define DOMINATOR_TREE_LAST_CHILD(x, y) \
    for (y = (x)->t_children; (y)->t_next_child != NULL; y = (y)->t_next_child);

/*
 *  A macro for iterating over a list of children nodes for a given node.
 */
#define FOREACH_NODE_CHILD(node, itr) \
    for (itr = (node)->t_children; itr != NULL; itr = (itr)->t_next_child)



/*
 *  FUNCTIONS
 */

extern TreeNode *GetRootNode(TreeNode *tree);
extern TreeNode *GetExistingNodeFor(TreeNode *tree, Block *bbl);
extern TreeNode *GetNearestCommonAncestor(TreeNode *node1, TreeNode *node2);
extern bool BBLIsAncestorOfNode(TreeNode *tree, Block *bbl);
extern void AddChildNode(TreeNode *tree, TreeNode *new_node);
extern void RemoveChildNode(TreeNode *parent, TreeNode *child);
extern TreeNode *DominatorNodeCreate();
extern LinkedList *FindAllNodesForFun(TreeNode *tree,
    bool (*is_the_one) (TreeNode *node));
extern TreeNode *FindFirstNodeForFun(TreeNode *tree,
    bool (*is_the_one) (TreeNode *node));
extern TreeNode *FunCreatePredominatorTree(Function *fun);
extern TreeNode *FunCreatePostdominatorTree(Function *fun);
extern bool BBLDominatesBBL(TreeNode *tree, Block *bbl1, Block *bbl2);
extern bool NodeDominatesNode(TreeNode *node1, TreeNode *node2);
extern void StoreLoopInfo(TreeNode *tree);
extern LinkedList *GetDepthFirstTraversal(TreeNode *tree);
extern LinkedList *GetBreadthFirstTraversal(TreeNode *tree);
extern bool BBLIsInALoop(Block *bbl);
extern bool BBLIsInSameLoopAsBBL(Block *bbl1, Block *bbl2);
extern void DominatorTreeFree(TreeNode *tree);
extern void StoreDominatorInfoForFun(Function *fun);
extern void StoreDominatorInfo();
extern void StoreLoopInfo(TreeNode *tree);
extern void BBLAddToLoop(Block *bbl, Block *head);

#endif
