/*
 *  Functions for CFG manipulation.
 */
#ifndef FLOWGRAPH_H
#define FLOWGRAPH_H

#include "global.h"

/*
 *  This struct represents a jump table lying in some readonly section.
 */
typedef struct JumpTable {
    struct JumpTable *jt_next;	// next one in the global list
    struct JumpTable *jt_prev;	// previous one in the global list
    Section *jt_section;	// section that the jump table lies in
    ADDRESS jt_base_offset;	// offset from containing section
    int jt_size;		// number of entries in the table
    LinkedList *jt_users;	// instructions that use this jump table
} JumpTable;

/*
 *  Adds an edge between 'src' and 'dest' of type 'type'.
 *  See bbl.type.h for a list of types
 *  The newly-added edge is then returned.
 */
extern Edge *BBLAddEdge(Block *src, Block *dest, UINT8 type);

/* Writes a textual representation of the control flow graph to standard out */
extern void BBLOutputGraph();

/* Spits out the number of instrs, bbls, funs and unresolved references
 * (HELL edges) in the CFG */
extern void BBLOutputStats();

/* Removes all successors from 'ablock */
extern void BBLRemoveEdges(Block *ablock);

/* Removes all predecessors from 'ablock' */
extern void BBLRemovePredecessors(Block *ablock);

/* Removes the edge connecting 'src' to 'dest', and frees memory */
extern void BBLRemoveEdge(Block *src, Block *dest);

/* Removes the edge connecting block src to dest of type type, and frees mem */
extern void BBLRemoveEdgeTyped(Block *src, Block *dest, int edge_type);

/*
 *  Divides a basic block at 'instr'. 'instr' appears as the first instruction
 *  in a new basic block that is linked to 'src' by an ET_NORMAL edge.  All the
 *  original successors of 'src' now becomes the successors of instr->i_Bblock
 */
extern void BBLDivideBlock(Block *src, Instruction *instr);

/* Copies the successor edges from the basic block 'src' to 'dest' */
extern void BBLCopyEdges(Block *src, Block *dest);

/* Returns true if an edge exists from 'src' to 'dest'. */
extern bool BBLEdgeExists(Block *src, Block *dest);

/* Returns true if 'bbl' has HELL as one of its predecessors. */
extern bool BBLHasHellPredecessor(Block *bbl);

/* Returns true if 'bbl' has an interprocedural predecessor edge */
extern bool BBLHasInterproceduralPredecessor(Block *bbl);

/* Returns true if 'bbl' has HELL as one of its successors. */
extern bool BBLHasHellSuccessor(Block *bbl);

/* Returns true if 'bbl' has a pseudo BT_ENTRY block as a predecessor */
extern bool BBLHasEntryPredecessor(Block *bbl);

extern bool BBLHasExitSuccessor(Block *bbl);

/* Returns true if 'bbl' is the first basic block within its function. */
extern bool BBLIsFirstBBLInFun(Block *bbl);

/* Returns true if 'bbl' is the last basic block within its function. */
extern bool BBLIsLastBBLInFun(Block *bbl);

/* Returns the number of predecessors of 'bbl' */
extern int BBLNumberOfPredecessors(Block *bbl);

/* Returns the number of successors of 'bbl' */
extern int BBLNumberOfSuccessors(Block *bbl);

/* Returns true if 'block' has an incoming edge of type ET_FUNLINK */
extern bool BBLHasFunlinkPredecessor(Block *block);

/* Returns the first ET_FUNLINK edge in 'block', or NULL if none exists */
extern Edge *BBLGetFunlinkEdge(Block *block);

/* Returns the first ET_UNCOND edge out of the basic block */
Edge *BBLGetUncondEdge(Block *block);

/* Returns the first ET_NORMAL edge out of the basic block */
Edge *BBLGetNormalEdge(Block *block);

/* Returns the first edge of type ET_TRUE in the basic block */
extern Edge *BBLGetTrueEdge(Block *block);

/* Returns the first edge of type ET_FALSE in the basic block */
extern Edge *BBLGetFalseEdge(Block *block);

/* Returns the first ET_FUNDIRECT edge in 'block', or NULL if none exists */
extern Edge *BBLGetFunDirectEdge(Block *block);

/* Returns the first ET_FUNCALL edge in 'block', or NULL if none exists */
extern Edge *BBLGetFuncallEdge(Block *block);

/* Returns the return edge going into 'bbl' */
extern Edge *BBLGetReturnPredecessor(Block *block);

/* Returns the first edge of any type between the source and destination BBL */
extern Edge *BBLGetEdgeToBBL(Block *src, Block *dest);

/* Returns the number of instructions in 'bbl' */
extern int BBLNumberOfInstructions(Block *bbl);

/* Removes all edges coming into the basic block */
extern void BBLRemoveIncomingEdges(Block *bbl);

/* Sums the weight of all incoming edges and returns the result */
extern EDGE_WEIGHT_TYPE BBLSumIncomingEdges(Block *bbl);

/* Sums the weights of all outgoing edges and returns the result */
extern EDGE_WEIGHT_TYPE BBLSumOutgoingEdges(Block *bbl);

/* Returns the number of instructions that 'fun' contains */
extern int FunNumberOfInstructions(Function *fun);

/* Returns the number of callsites in 'fun' */
extern int FunNumberOfCallSites(Function *fun);

/* Returns true if the edge goes across function boundaries */
extern bool isInterproceduralEdge(Edge *edge);

extern void BuildControlFlowGraph(Instruction *list);

/*
 *  CheckConsistency is an expensive function that will make sure all data
 *  structures agree with each other.  Any abnormalities will be reported.
 */
extern void CheckConsistency();

/*
 *  Counts up the number of instructions, basic blocks and functions
 *  and sets global variables based on what the results are
 */
extern void CountStructures();

/*
 *  This calls BBLCreate then puts the bbl at the end of the basic
 *  block list
 */
extern Block *BblNew();

/*
 *  This calls FunCreate, then puts the function at the end of the
 *  function list
 */
extern Function *FunNew();

/* Returns a new Edge object */
extern Edge *EdgeCreate();

/* Returns a new Operand object */
extern Operand *OperandCreate();

/* Returns a new EdgeNode object */
extern EdgeNode *EdgeNodeCreate();

/*
 *  Returns the total number of known function calls that will be made at
 *  runtime if the current representation of the CFG is written out as an
 *  executable
 */
extern long long CountFuncalls();

/*
 *  Returns a pointer to the instruction that will be executed after 'instr',
 *  or NULL if no such instruction exists (or cannot be determined)
 */
extern Instruction *InstrExecutedAfterInstr(Instruction *instr);

/*
 *  Merges basic blocks in the control flow graph. Two blocks are merged if they
 *  are connected by an edge of type ET_NORMAL, the first block has only one
 *  successor and the second block has a single predecessor
 */
extern void MergeBasicBlocks();

/* Returns the number of edges from bbl1 to bbl2 */
extern int BBLNumberOfEdgesToBBL(Block *bbl1, Block *bbl2);

/* Returns the number of times bbl2 appears in bbl1's predecessors list */
extern int BBLNumberOfPredecessorsFromBBL(Block *bbl1, Block *bbl2);

/* Returns true if the BBL is the target of an indirect call */
extern bool BBLIsIndirectTarget(Block *bbl);

extern void BBLRemoveOutgoingEdges(Block *bbl);
extern Edge *BBLGetEdgeTyped(Block *block, UINT32 type);
extern bool InstrIsEscapingJump(Instruction *instr);
extern Edge *BBLGetEdgeToBBLTyped(Block *src, Block *dest, ETYPE type);
extern void MarkFlistCalls();
extern void ReplaceHaltBranchesWithHalts();
extern LinkedList *GetPICCalls();
extern JumpTable *InstrGetJT(Instruction *instr);
extern bool FunHasEsc2(Function *fun);
extern bool BBLHasEsc2(Block *bbl);
extern Edge *BBLGetEscapingEdge(Block *bbl);
extern void RemoveEscapingEdges();
extern JumpTable *GetJumpTableForInstr(Instruction *instr);
extern JumpTable *AddJumpTable(Section *section, ADDRESS offset, int size);
extern Edge *BBLGetSyscallLinkEdge(Block *bbl);


/*
 *  Reports an error related to an instruction.
 */
extern void ReportInstrError(char *problem, Instruction *instruction);

#endif				// FLOWGRAPH_H
