/*
 *  global.h -- definitions of global variables and constants.
 *
 *  Also serves as a "master" include file by its inclusion
 *  of other commonly used definition files.
 */
#ifndef GLOBAL_H
#define GLOBAL_H

#ifndef GLOBAL		// if undefined, declare globals without defining
    #define GLOBAL extern
#endif



/*
 *  Standard usage libraries.
 */
#include <assert.h>
#include <ctype.h>
#include <elf.h>
#include <limits.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/times.h>
#include <time.h>

/*
 *  Some includes for things that pretty much any other file will need.
 */
#include "config.h"
#include "macros.h"
#include "util.h"

#include "bbl.h"
#include "edge.h"
#include "edge.type.h"
#include "flags.h"
#include "flist.h"
#include "function.h"
#include "instr.h"
#include "operand.h"
#include "plto_types.h"
#include "section.h"
#include "string_table.h"



/*
 *  GLOBAL VARIABLES
 *
 *  Global variables should only be defined once.  The hack that we use
 *  is to prefix all of the declarations with GLOBAL.  If GLOBAL is not
 *  set, then it is set to be extern.  The file that should define all
 *  these global variables should set GLOBAL to be nothing.  This way, these
 *  variables are only defined once.
 */

GLOBAL int numFunctions;		// number of functions in the program
GLOBAL int numBlocks;			// number of bbls in the program
GLOBAL int numInstructions;		// number of instrs in the program
GLOBAL int numSections;			// number of sections in the program
GLOBAL EDGE_WEIGHT_TYPE lowest_hot_weight;  // smallest weight considered hot

GLOBAL ADDRESS DEFAULT_TEXT_START_ADDR;	// base address of text section
GLOBAL ADDRESS base_addr;	// base address of binary, default 0x8048000
GLOBAL ADDRESS startAddress;	// address of the program entry point

GLOBAL UINT32 bblProfiles;              // number of bbl profiles we generated

GLOBAL UINT32 globalEID;		// global counter for assigning edge IDs
GLOBAL UINT32 edgeProfiles;		// number of edge profiles we generated

GLOBAL BINARY_TYPE input_binary_type;	// type of the input binary
GLOBAL BINARY_TYPE output_binary_type;	// type of the output binary

GLOBAL UINT32 PAGE_SIZE;		// size of a virtual memory page

GLOBAL int verbosityLevel;		// selected output verbosity level
GLOBAL bool noWarnings;			// suppress warnings?
GLOBAL bool showTimestamps;		// show timestamps on output?
GLOBAL bool debugFlags[128];		// selected debugging classes

GLOBAL ProfConfig profile;		// for edge profiling

GLOBAL Block *bbl_list;			// linked list of all basic blocks
GLOBAL Block *HELL;			// the worst-case HELL block

GLOBAL Function *f_list;		// linked list of all functions

GLOBAL Instruction *i_list;		// list of internal insn representations
GLOBAL Instruction *startins;		// starting instruction for the program

GLOBAL Relocation *r_list;		// linked list of all relocations

GLOBAL Symbol *s_list;			// linked list of all symbols
GLOBAL Symbol *ENTRY_SYMBOL;		// last function to execute
GLOBAL Symbol *EXIT_SYMBOL;		// last function to execute

GLOBAL FList *dontTouchFuns;		// list of functions we won't alter

GLOBAL LinkedList *upm2s_list;		// all UPMS to apply before writeout

GLOBAL int MaxInstrLength;		// maximum instruction size ever seen



/*
 *  command-line options
 */
GLOBAL bool PROFILE_BBLS;               // -pb: instrument for bbl profiling
GLOBAL bool PROFILE_EDGES;		// -pe: instrument for edge profiling
GLOBAL bool PROFILE_VALUES;		// -P: instrument for value profiling
GLOBAL bool PROFILE_CALLSTACK;		// -P: instrument for value profiling
GLOBAL bool OPT_DISABLE_ALL;		// -n: disable optimization

GLOBAL bool READ_FAKE_RELOCATIONS;
GLOBAL bool KERNEL_MODE;



/*
 *  files
 */
GLOBAL char *cfgFilename;		// configuration filename

GLOBAL char *binFilename;		// binary input filename
GLOBAL FILE *binFile;			// binary input file

GLOBAL char *newFilename;		// binary output filename
GLOBAL char *profFilename;		// profile output filename

//REMOVE ME - SOMU
GLOBAL char *valProfileFilename;		// profile output filename

GLOBAL ProfileFile *profileFiles;	// linked list of input profile files

GLOBAL ProfileFile *valProfileFiles;	// linked list of input value 
                                        // profile files              

GLOBAL ProfileFile *callStackProfileFiles;     // linked list of input 
                                               // call stack profile files

GLOBAL ProfileFile *sysValProfileFiles;     // linked list of input 
                                            // system call parameter files

GLOBAL char *sloopFilename;		// sloop output filename
GLOBAL FILE *sloopfile;			// sloop output file



/*
 *  optimization parameters
 */
GLOBAL int BBL_LAYOUT_SCHEME;		// value of "Layout" parameter

GLOBAL INT32 INLINING_MAX_FUNSIZE;

GLOBAL bool OPT_ELIMINATE_DEAD_REGISTERS;
GLOBAL bool OPT_ELIMINATE_DEAD_MEMORY;

GLOBAL bool OPT_PEEPHOLE;		// set if any of the following are on
GLOBAL bool PEEPHOLE_STUPID_MOVES;
GLOBAL bool PEEPHOLE_STUPID_JUMPS;
GLOBAL bool PEEPHOLE_COALESCE_ADDS;
GLOBAL bool PEEPHOLE_STUPID_MATH;
GLOBAL bool PEEPHOLE_UNUSED_PUSHES;
GLOBAL bool PEEPHOLE_CONDITIONAL_MOVE;



/*
 *  well-known sections
 */
GLOBAL Section *SECTION_REF_DATA;	// initialized data section
GLOBAL Section *SECTION_REF_CODE;	// primary executable code section
GLOBAL Section *SECTION_REF_RODATA;	// read-only data section
GLOBAL Section *SECTION_REF_GOT;	// Global Offset Table
GLOBAL Section *SECTION_REF_BSS;	// uninitialized data section
GLOBAL Section *SECTION_REF_SYM;	// symbol table section

GLOBAL Section *sec_list;	// linked list of sections in the binary
GLOBAL Section *mrg_list;	// linked list of sections that were merged out



/*
 *  DEFINITIONS
 */

#define FLIST_SECTION_INITIAL_ADDRESS (0x7fffffff)

#define FLIST_SECTION_NAME         "plto_flists"
#define MAGIC_NUMBER               0xfeedface
#define PLTO_ENVVAR                "PLTO_CONFIG_FILE"
#define MAXIMUM_INSTRUCTION_LENGTH 16
#define BBL_PROFILE_DATA_SIZE      12
#define EDGE_PROFILE_DATA_SIZE     20

#define ALTO_MAGIC 0xdeadbeef
#define PROFILE_INIT_SIZE 12*4
#define CONFIG_NENTRIES 20

#endif				// GLOBAL_H
