/*
 *  This file contains stuff specific to the Instruction data structure
 *  which represents a single instruction.
 */
#ifndef INSTR_H
#define INSTR_H

#include "plto_types.h"
#include "operand.h"

/*
 *  STRUCTURES
 *
 *  The representation used for an instruction.
 */
typedef struct Instruction {
    struct Instruction *i_next;	    // following instruction
    struct Instruction *i_prev;	    // preceding instruction
    INT64 i_flags;		    // instruction flags
    Operand *i_src;		    // first source operand
    Operand *i_src2;		    // second source operand
    Operand *i_dest;		    // destination operand
    ADDRESS i_address;		    // original location
    ADDRESS i_final_address;	    // final location
    struct Block *i_Bblock;	    // basic block that contains this insn
    struct Instruction *i_hashNext; // hash chain
    int i_temp;			    // temporary: reloc offset or stack usage
    UINT32 i_id;		    // ID number
    UINT16 i_prefixes;		    // instruction prefix
    UINT16 i_class;		    // instruction class (e.g. add, pop, imul)
    UINT32 i_length;		    // length in bytes
    UINT8 i_type;		    // instruction type (e.g. cond, uncond...)
    UINT8 *i_literal_data;	    // literal bytes associated with this insn
					// (literal bytes flag must be set)
    UINT32 i_alignment;             // alignment constraints for instructions
} Instruction;

/* Macros for accessing fields in the instruction */
#define INSTR_NEXT_NONNULL(inst)    ((inst) == NULL ? (inst) : (inst)->i_next)
#define INSTR_BBL_NEXT(inst)        ((inst)->i_Bblock->b_next)
#define INSTR_BBL_PREV(inst)        ((inst)->i_Bblock->b_prev)
#define INSTR_FUNCTION(inst)        ((inst)->i_Bblock->b_fun)
#define INSTR_BBL_PRED(inst)        ((inst)->i_Bblock->b_pred)
#define INSTR_BBL_SUCC(inst)        ((inst)->i_Bblock->b_succ)

#define INSTR_SET_FLAG(instr, flag)    (((instr)->i_flags) |= (flag))
#define INSTR_HAS_FLAG(instr, flag)    ((bool)(((instr)->i_flags) & (flag)))
#define INSTR_REMOVE_FLAG(instr, flag) (((instr)->i_flags) &= ~(flag))

/*
 *  struct DataBlock
 *
 *  A list of data that appears in the code segment. Items in this list are
 *  ignored during disassembly, and are identified based on entries in the
 *  symbol table.
 */
typedef struct DataBlock {
    struct DataBlock *db_next;
    struct DataBlock *db_prev;
    ADDRESS db_startAddress;
    int db_size;
    UINT8 *db_data;
} DataBlock;



extern Instruction *InstrCreate();
extern void InstrFree(Instruction *instr);

#endif
