/*
 *  liveness.h
 *
 *  Functions for performing a register liveness analysis on the CFG
 */
#ifndef LIVENESS_H
#define LIVENESS_H

#define LIVENESS_NONE        0
#define LIVENESS_INSENSITIVE 1
#define LIVENESS_SENSITIVE   2

/* Initializes parameters */
extern void InitializeLivenessParameters();

/* Sets the bitvectors for the registers which are live
 * at the start and end of each BBL */
extern void ComputeRegisterLiveness(int full);

/* Sets the bitvectors for the PSW flags which are live
 * at the start and end of each BBL */
extern void ComputePSWLiveness(int full);

/* Sets the live bits in the stack snapshots for stack slots that are live
 * at the start of each BBL */
extern void ComputeStackLiveness(int full);

/* Tests the register liveness analysis by setting all dead registers to -1 */
extern void TestRegisterLiveness();

/* Removes instructions who write to dead locations */
extern void RemoveDeadCode();

/* Returns the total number of dead registers,
 * at the end of each basic block, in the entire program */
extern int CountDeadRegisters();

/* Returns a bitvector with the bits set for each register which this
 * basic block uses before defining */
extern REGISTER_BITVECTOR ComputeRegistersUsedBeforeDefinedInBBL(Block *bbl);

/* Returns a bitvector with the bits set for each register
 * which this basic block defines */
extern REGISTER_BITVECTOR ComputeRegistersDefinedInBBL(Block *bbl);

/* Computes the depth to which functions may overwrite the stack */
extern void ComputeFunctionTrashDepths();

/* Computes the depth to which functions access the stack */
extern void ComputeFunctionUseDepths();

/* Computes and stores in each function a register bitvector
 * specifying what registers  the function may define */
extern void ComputeRegsThatFunctionsDefineLiberal();

extern PSW_BITVECTOR PSWLiveBitvectorAtInstr(Instruction *instr);

extern bool PSWBitvectorIsLiveAfterInstr(PSW_BITVECTOR vec, Instruction *instr);

/*
 *  Returns whether or not the specified PSW bitvector is live at the specified
 *  instruction (directly before the specified instruction is executed.
 *
 *  vec: the PSW bitvector to check liveness against
 *  instr: the instruction directly prior to which the PSW bitvector
 *  liveness will be tested
 *  returns: true if the specified PSW bitvector is live at the specified
 *  instruction
 */
extern bool PSWBitvectorIsLiveAtInstr(PSW_BITVECTOR vec, Instruction *instr);

extern PSW_BITVECTOR PSWLiveBitvectorAfterInstr(Instruction *instr);

extern REGISTER_BITVECTOR RegsLiveBitvectorAtInstr(Instruction *instr);

extern int GetFreeRegisterAtInstr(Instruction *instr);

#define MAXIMUM_SIZE_FOR_STACK_LIVENESS   25000
#define ALL_REGISTERS 0x000000ff    // Machine dependent: 8 register for P6
#define ALL_PSW       0xffffffff    // Machine dependent: 32 bits in PSW for P6

#endif				// LIVENESS_H
