/*
 *  stack.analyze.h
 *
 *  Functions for performing a stack analysis
 */
#ifndef STACK_ANALYZE_H
#define STACK_ANALYZE_H

/*
 *  MarkWellbehavedFunctions() sets the FUNCTION_WELL_BEHAVED_FLAG in all
 *  functions that leave the stack as they found it. A LEAVE/RET combo means
 *  a function is well-behaved, as does a guarantee that all execution paths
 *  through the function result in a net gain of 0 bytes.
 */
void MarkWellBehavedFunctions();

/*
 *  ComputeStackFrameSize() will return the size of the stack frame for 'fun'.
 *  The size is determined by looking at the largest the stack grows during
 *  execution
 */
int ComputeStackFrameSize(Function *fun);

/*
 *  Calculates the difference between the BP and SP at basic block entry
 *  points, and stores the result in ->b_BPSPdiff.
 */
void CalculateSPBPDifference();

/* Returns true if this instruction aliases the stack or frame pointer */
bool InstrAliasesStackPointer(Instruction *instr);

/* Returns true if the instruction potentially writes to the stack */
bool InstrPotentiallyWritesToStack(Instruction *instr);

/* Returns the difference between the BP and SP right before the instruction */
int InstrBPSPDiff(Instruction *instr);

/*
 *  Stores the result of calling InstrBPSPDiff() on every instruction in the
 *  instruction's temp field
 */
void StoreInstrBPSPDiff();

/*
 *  Stores the result of calling InstrBPSPDiff() on every instruction in the
 *  temp field for the given function
 */
void StoreInstrBPSPDiffForFunction(Function *fun);

/*
 *  Sets the FUNCTION_ALIASES_STACK_POINTER_FLAG in all functions who have
 *  instructions that may alias the stack or frame pointer
 */
void MarkFunsAliasingStackPointer();

/*
 *  Sets the FUNCTION_REDEFINES_FP_FLAG in all functions that define
 *  the frame pointer outside of epilogue/prologue.  Sets the
 *  FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG in all functions
 *  that define %ebp outside of the epilogue and prologue, and the
 *  instructions that do so were not inserted during the inlining phase.
 */
void MarkFunsDefiningFramePointer();

/* Updates the flags in 'fun', and re-calculates the stack height */
void UpdateFunction(Function *fun);

/* Updates all flags in all functions and re-calculates their stack heights */
void UpdateFunctions();

/* Sets the (()->f_stackFrameSize) and (()->f_callerFrameSize) fields
 * in each function */
void ComputeFunctionAndCallerSizes();

/* Returns the weighted count of all memory operations in the executable */
long long CountMemoryOperations();

/* Returns a number in bytes that represents the net gain 'instr'
 * contributes to the stack */
int ComputeMyStackGainForInstr(Instruction *instr);

/*
 *  Returns the number of arguments that the function 'fun' takes.
 *  This is an (accurate) estimate but not always correct.
 *  This function looks for loads from below the stack pointer to determine
 *  how deep the function accesses.
 */
int FunNumberOfArguments(Function *fun);

/*
 *  Returns whether or not the specified basic block contains a
 *  recognizable exit combo, such as a leave return.
 *
 *  bbl: the basic block to check for an exit combo
 *  returns: true if the specified basic block contains an exit combo
 */
bool BBLContainsExitCombo(Block *bbl);

bool BBLContainsPrologue(Block *bbl);

extern int InstrSynchronizesStack(Instruction *instr);
int StackJoin(int a, int b);
int StackAdd(int a, int b);
void RemoveEscapingEdges();
void MarkWellBehavedFunctions2();
int ComputeStackContributionForInstr(Instruction *instr);
void StoreInstrBPSPDiffForFun(Function *fun);
void MarkPrologueFuns();
void MarkExitComboFuns();
void StoreInstrBPSPDiffForAllFuns();
void StoreNetStackContributionsForFun(Function *fun);
void PrintStackAnalysisSummary();

#define STACK_UNKNOWN       0xfbca001f
#define STACK_UNDEFINED     0xfcaaa0aa

#endif				// STACK_ANALYZE_H
