/*
 *  This file contains definitions for functions used to manage the
 *  Universal Patching Mechanism II (UPM).
 */
#ifndef UPM_H
#define UPM_H

/*
 *  Structures for UPM handling routines.
 *
 *  An enumeration used to identify operand types for a UPM. For each
 *  type that this enumeration defines, there should be a macro for the
 *  UPM_OPERAND which does an automatic cast to that type, e.g., for
 *  UPM_INSTR_TYPE there is a UPM_OPERAND macro UPM_OPERAND_INSTR which
 *  casts the operand to an instruction.
 */
typedef enum {
    UPM_INSTR_MEMBER_TYPE,		  // The member is a Instruction
    UPM_SECTION_MEMBER_TYPE,		  // The member is a Section
    UPM_OFFSET_MEMBER_TYPE,		  // The member is an ADDRESS offset
    UPM_SYMBOL_MEMBER_TYPE,		  // The member is a Symbol
    UPM_OPERAND_MEMBER_TYPE,		  // The member is a Operand
    UPM_FUNCTION_MEMBER_TYPE		  // The member is a Function
} UMTYPE;

/*
 *  An member for a UPM is either a source (something that is being used
 *  to calculate a value to be placed into a destination) or a destination
 *  (where a
 *  calculated value is placed).
 */
typedef struct Location {
    UMTYPE l_type;		// type of the member
    void *l_member;		// UPM member this structure encapsulates
} Location;



#define UPM_MEMBER_INSTR(op) ((op)->l_type == UPM_INSTR_MEMBER_TYPE ? \
    ((Instruction **)((op)->l_member)) : NULL)

#define UPM_MEMBER_SECTION(op) ((op)->l_type == UPM_SECTION_MEMBER_TYPE ? \
    ((Section **)((op)->l_member)) : NULL)

#define UPM_MEMBER_OFFSET(op) ((op)->l_type == UPM_OFFSET_MEMBER_TYPE ? \
    ((ADDRESS *)((op)->l_member)) : NULL)

#define UPM_MEMBER_SYMBOL(op) ((op)->l_type == UPM_SYMBOL_MEMBER_TYPE ? \
    ((Symbol **)((op)->l_member)) : NULL)

#define UPM_MEMBER_OPERAND(op) ((op)->l_type == UPM_OPERAND_MEMBER_TYPE ? \
    ((Operand **)((op)->l_member)) : NULL)

#define UPM_MEMBER_FUNCTION(op) ((op)->l_type == UPM_FUNCTION_MEMBER_TYPE ? \
    ((Function **)((op)->l_member)) : NULL)



/*
 *  A structure that manages a list of UPM_MEMBERs
 *  see also: Location
 */
typedef struct UPMList {
    Location *u_members;	// array of UPM_MEMBERs
    int u_size;			// number of members in this list
} UPMList;


#define BEGIN_FOREACH_UPM_DEST_MEMBER(upm, mbr_itr) \
    { UPMList *list; \
    int indexval; \
    list = ((upm)->p_GetDestinations)(upm); \
    for (indexval = 0; indexval < ((list)->u_size); indexval++) { \
	mbr_itr = &((list)->u_members)[indexval];

#define BEGIN_FOREACH_UPM_SOURCE_MEMBER(upm, mbr_itr) \
    { UPMList *list; \
    int indexval; \
    list = ((upm)->p_GetSources)(upm); \
    for (indexval = 0; indexval < ((list)->u_size); indexval++) { \
	mbr_itr = &((list)->u_members)[indexval];

#define END_FOREACH_UPM_MEMBER } UPMMemberListDestroy(list); }

/*
 *  A Patch acts as (as the name indicates) a universal patching
 *  mechanism.  Specifically this struct is used to patch some final
 *  property (that which is present directly before the binary is output
 *  after all addresses and layout is determined) into something. One
 *  example is patching the final address of one instruction into the
 *  operand of another instruction (see PatchInstrAddressToInstrSource1).
 *  There are currently several patches already available that can be used
 *  to patch final addresses of section locations and instructions into
 *  other instructions and section locations and patch these addresses
 *  into symbols.
 *
 *  The convention for creating new UPMs is to create a patch function
 *  (i.e. the function that will actually do the patching in the end) that
 *  should be compatible with the Apply function pointer of this UPM
 *  struct and should be named 'Patch<sources>Into<destinations>'
 *  replacing <sources> and <destinations> with the types of the sources
 *  and destinations. One should also create
 *  Patch<sources>Into<destinations>Get[Sources|Destinations] functions
 *  that return UPM_MEMBER_LISTs for each of their sources/destinations
 *  respectively.  Finally, if the UPM has any instruction targets a
 *  clone function of the form Patch<sources>Into<destinations>Clone
 *  function should also be created. There is one primary UPM list that
 *  UPMs are typically added to, this list is a LinkedList global
 *  accessible as upm2s_list.
 *
 *  The get sources and get destinations functions need to return lists
 *  populated in the following manner:
 *  -- any operand must be preceded by the instruction it lies in
 *  -- any offset must be preceded by the section it lies in
 *  see also: UPMCreate
 *  see also: UPMFree
 *  see also: LinkedList
 */
typedef struct Patch {
    void *p_patch;		// patch structure that the UPM should use
    bool (*p_Apply) (struct Patch *upm);
				// function to apply the patch
    void (*p_CloneToInstr) (struct Patch *upm, Instruction *dest_instr);
				// function to clone the UPM
    UPMList *(*p_GetSources) (struct Patch *upm);
				// function to return all sources
    UPMList *(*p_GetDestinations) (struct Patch *upm);
				// function to return all destinations
    unsigned char p_flags;	// flags
} Patch;

/*
 *  A structure used to patch some property of a section into some place in
 *  another section (possibly the same section).
 */
typedef struct PatchSectionToSection {
    Section	*p_src_section;
    ADDRESS	 p_src_offset;
    Section	*p_dest_section;
    ADDRESS	 p_dest_offset;
} PatchSectionToSection;

/*
 *  A structure used to patch some relationship between two section locations
 *  into another section location (possibly the same location as one of the
 *  sources).
 */
typedef struct PatchSectionToSectionAux {
    Section	*p_src_section;
    ADDRESS	 p_src_offset;
    Section	*p_dest_section;
    ADDRESS	 p_dest_offset;
    Section	*p_aux_section;
    ADDRESS	 p_aux_offset;
} PatchSectionToSectionAux;

/*
 *  A structure used to patch some property of a given section into an
 *  instruction.
 */
typedef struct PatchSectionToInstr {
    Section	*p_src_section;
    ADDRESS	 p_src_offset;
    Instruction	*p_dest_instr;
    Operand	*p_dest_op;
} PatchSectionToInstr;

/*
 *  A structure used to patch some property of an instruction into a section.
 */
typedef struct PatchInstrToSection {
    Instruction	*p_src_instr;
    Section	*p_dest_section;
    ADDRESS	 p_dest_offset;
} PatchInstrToSection;

/*
 *  A structure used to patch some property of one instruction into another
 *  instruction (possibly the same instruction).
 */
typedef struct PatchInstrToInstr {
    Instruction	*p_src_instr;
    Instruction	*p_dest_instr;
    Operand	*p_dest_op;
} PatchInstrToInstr;

/*
 *  A structure used to patch some relationship between two instructions into
 *  another instruction (possibly the same instruction).
 */
typedef struct PatchInstrToInstrAux {
    Instruction	*p_src_instr;
    Instruction	*p_aux_instr;
    Instruction	*p_dest_instr;
    Operand	*p_dest_op;
} PatchInstrToInstrAux;

/*
 *  A structure used to patch some relationship between two instructions into
 *  some section location.
 */
typedef struct PatchInstrInstrToSection {
    Instruction	*p_src_instr;
    Instruction	*p_aux_instr;
    Section	*p_dest_section;
    ADDRESS	 p_dest_offset;
} PatchInstrInstrToSection;

/*
 *  A structure used to patch some relationship between two instructions
 *  into a section location.
 */
typedef struct PatchSectionInstrToSection {
    Section	*p_dest_section;
    ADDRESS	 p_dest_offset;
    Section	*p_src_section;
    ADDRESS	 p_src_offset;
    Instruction	*p_src_instr;
} PatchSectionInstrToSection;

/*
 *  A structure used to patch some relationship between a section and an
 *  instruction into an instruction.
 */
typedef struct PatchSectionInstrToInstr {
    Section	*p_src_section;
    ADDRESS	 p_src_offset;
    Instruction	*p_src_instr;
    Instruction	*p_dest_instr;
    Operand	*p_dest_op;
} PatchSectionInstrToInstr;

/*
 *  A structure used to patch some relationship between two sections into
 *  an instruction.
 */
typedef struct PatchSectionSectionToInstr {
    Section	*p_src_section;
    ADDRESS	 p_src_offset;
    Section	*p_aux_section;
    ADDRESS	 p_aux_offset;
    Instruction	*p_dest_instr;
    Operand	*p_dest_op;
} PatchSectionSectionToInstr;

/*
 *  A structure used to patch some aspect of an instruction into a symbol.
 */
typedef struct PatchInstrToSymbol {
    Instruction	*p_src_instr;
    Symbol	*p_dest_sym;
} PatchInstrToSymbol;

/*
 *  A structure used to patch some aspect of a function into a symbol.
 */
typedef struct PatchFunctionToSymbol {
    Function	*p_src_function;
    Symbol	*p_dest_sym;
} PatchFunctionToSymbol;

/*
 *  A structure used to patch some aspect of a section location into a
 *  symbol.
 */
typedef struct PatchSectionToSymbol {
    Section	*p_src_section;
    ADDRESS	 p_src_offset;
    Symbol	*p_dest_sym;
} PatchSectionToSymbol;

/*
 * Patch to a fixed offset from src instruction. 
 */
typedef struct PatchInstrOffsetToInstr { 
    Instruction *p_src_instr;
    int 	 p_src_offset;
    Instruction *p_dest_instr;
    Operand     *p_dest_op;
} PatchInstrOffsetToInstr;


/*
 * Patch to a fixed address from src instruction.
 * Needed to patch when src instruction is a
 * data instruction and dest instruction is a normal
 * instruction with an absolute address into data instruction
 */
typedef struct PatchDataInstrAddressToInstr { 
    Instruction *p_src_instr;
    int 	 p_src_offset;
    Instruction *p_dest_instr;
    Operand     *p_dest_op;
} PatchDataInstrAddressToInstr;
 
/*
 * A structure used to patch data instructions.
 */
typedef struct PatchDataInstrToDataInstr {
    Instruction	*p_src_instr;
    int		 p_src_patch_offset;
    Instruction	*p_dest_instr;
    int		 p_dest_patch_offset;
} PatchDataInstrToDataInstr;

/*
 *  Used to patch data instructions with relocations to data sections.
 */
typedef struct PatchDataSectionOffsetIntoDataInstr {
    Section	*p_src_section;
    int		 p_src_offset;
    Instruction	*p_dest_instr;
    int		 p_dest_offset;
} PatchDataSectionOffsetIntoDataInstr;

/*
 *  A structure used to patch data instructions into section addresses.
 */
typedef struct PatchDataInstrToSectionAddress {
    Instruction	*p_src_instr;
    int		 p_src_patch_offset;
    Section	*p_dest_section;
    int		 p_dest_patch_offset;
} PatchDataInstrToSectionAddress;

/*
 *  A structure used to patch some aspect of a section location into a symbol.
 */
typedef struct PatchSymbolValueToInstr {
    Symbol	*p_src_sym;
    ADDRESS	 p_src_offset;
    Instruction	*p_dest_instr;
    Operand	*p_dest_op;
} PatchSymbolValueToInstr;

/*
 *  A structure used to patch some aspect of a section location into a symbol.
 */
typedef struct PatchSymbolValueToSectionAddress {
    Symbol	*p_src_sym;
    ADDRESS	 p_src_offset;
    Section	*p_dest_section;
    int		 p_dest_patch_offset;
} PatchSymbolValueToSectionAddress;

 
/*
 *  FUNCTIONS
 */

extern Patch *AddPatchInstrAddressIntoInstr(
Instruction *src_instr, Instruction *dest_instr, Operand *dest_op);

extern Patch *AddPatchSectionAddressIntoInstr(Section *src_section,
ADDRESS src_offset, Instruction *dest_instr, Operand *dest_op);

extern Patch *AddPatchInstrInstrOffsetIntoInstr(Instruction *src_instr,
Instruction *aux_instr, Instruction *dest_instr, Operand *dest_op);

extern Patch *AddPatchInstrSectionOffsetIntoInstr(Section *src_section,
ADDRESS src_offset, Instruction *src_instr, Instruction *dest_instr,
Operand *dest_op);

extern Patch *AddPatchInstrSectionOffsetIntoSectionAddress(
Instruction *src_instr, Section *src_section, ADDRESS src_offset,
Section *dest_section, ADDRESS dest_offset);

extern Patch *AddPatchInstrAddressIntoSectionAddress(
Instruction *src_instr, Section *dest_section, ADDRESS dest_offset);

extern Patch *AddPatchSectionSectionOffsetIntoSectionAddress(
Section *src_section, ADDRESS src_offset, Section *aux_section,
ADDRESS aux_offset, Section *dest_section, ADDRESS dest_offset);

extern Patch *AddPatchSectionAddressIntoSectionAddress(Section *src_section,
ADDRESS src_offset, Section *dest_section, ADDRESS dest_offset);

extern Patch *AddPatchFunctionAddressIntoSymbol(
Function *src_fun, Symbol *dest_sym);

extern Patch *AddPatchInstrAddressIntoSymbol(
Instruction *src_instr, Symbol *dest_sym);

extern Patch *AddPatchSectionAddressIntoSymbol(
Section *src_section, ADDRESS src_offset, Symbol *dest_sym);

extern Patch *AddPatchSectionEndIntoInstr(
Section *src_section, Instruction *dest_instr, Operand *dest_op, int);

extern Patch *AddPatchSectionEndIntoSectionAddress(
Section *src_section, Section *dest_section, ADDRESS dest_offset);

extern Patch *AddPatchSectionEndIntoSymbol(
Section *src_section, Symbol *dest_sym);

extern Patch *AddPatchGOTOffsetIntoInstr(
Instruction *pic_call, Instruction *dest_instr, Operand *dest_op);

extern Patch *AddPatchInstrEndAddressIntoSectionAddress(
Instruction *src_instr, Section *dest_section, ADDRESS dest_offset);

extern Patch *AddPatchDataInstrToDataInstr(Instruction *src_instr,
int src_offset, Instruction *dest_instr, int dest_offset);

extern Patch *AddPatchDataInstrToDataInstrRelative(Instruction *src_instr,
int src_offset, Instruction *dest_instr, int dest_offset);
 
extern Patch *AddPatchDataSectionOffsetIntoDataInstr(
Section *src_section, int src_offset, Instruction *dest_instr, int dest_offset);

extern Patch *AddPatchDataInstrToSectionAddress(
Instruction *src_instr, int src_offset, Section *dest_section, int dest_offset);

extern Patch *AddPatchInstrOffsetToInstr(Instruction *src_instr,
int src_offset, Instruction *dest_instr, Operand *dest_op);

extern Patch *AddPatchDataInstrAddressToInstr(Instruction *src_instr,
int src_offset, Instruction *dest_instr, Operand *dest_op);

extern Patch *AddPatchSymbolValueIntoInstr(Symbol *sym, ADDRESS offset,
    Instruction *instr, Operand *op);
 
extern Patch *AddPatchSymbolValueIntoSectionAddress(Symbol *sym, ADDRESS addend,
    Section *dest_section, int dest_offset); 
extern bool patchInstrAddressIntoInstr(Patch *upm);
extern bool patchSectionAddressIntoInstr(Patch *upm);
extern bool patchInstrInstrOffsetIntoInstr(Patch *upm);
extern bool patchInstrSectionOffsetIntoInstr(Patch *upm);
extern bool patchInstrSectionOffsetIntoSectionAddress(Patch *upm);
extern bool patchInstrAddressIntoSectionAddress(Patch *upm);
extern bool patchSectionSectionOffsetIntoSectionAddress(Patch *upm);
extern bool patchSectionAddressIntoSectionAddress(Patch *upm);
extern bool patchFunctionAddressIntoSymbol(Patch *upm);
extern bool patchInstrAddressIntoSymbol(Patch *upm);
extern bool patchSectionAddressIntoSymbol(Patch *upm);
extern bool patchSectionEndIntoInstr(Patch *upm);
extern bool patchSectionEndIntoSectionAddress(Patch *upm);
extern bool patchSectionEndIntoSymbol(Patch *upm);
extern bool patchGOTOffsetIntoInstr(Patch *upm);
extern bool patchInstrEndAddressIntoSectionAddress(Patch *upm);
extern bool patchInstrDataInstrAddressIntoInstr(Patch *upm); 
extern bool patchSymbolValueIntoInstr(Patch *upm);
extern bool patchDataInstrToDataInstrRelative(Patch *upm);
 
extern void UPMFree(Patch *upm);

extern Patch *UPMCreate(void *patch, bool (*Apply) (Patch *upm),
void (*CloneToInstr) (Patch *upm, Instruction *dest_instr),
UPMList *(*GetSources) (Patch *upm), UPMList *(*GetDestinations) (Patch *upm));

extern int ApplyUPMs(LinkedList *upm2_list);

extern void CloneInstrUPMsToInstr(
Instruction *src_instr, Instruction *dest_instr);

extern void AddSymbolUPMs();

extern bool UPMChangeSourceFromInstrToInstr(
Patch *upm, Instruction *old, Instruction *new_instr);

extern void UPMTableInsertBySource(void *key, Patch *upm);
extern void UPMTableInsertByDest(void *key, Patch *upm);
extern LinkedList *UPMTableGetBySource(void *key);
extern Patch *UPMTableGetFirstBySource(void *key);
extern LinkedList *UPMTableGetByDest(void *key);

extern void UPMTableInsertByDestAddress(
Section *section, ADDRESS offset, Patch *upm);

extern void UPMTableInsertBySourceAddress(
Section *section, ADDRESS offset, Patch *upm);

extern LinkedList *UPMTableGetBySourceAddress(Section *section, ADDRESS offset);
extern LinkedList *UPMTableGetByDestAddress(Section *section, ADDRESS offset);
extern Instruction *UPMGetDestInstr(Patch *upm);
extern Section *UPMGetDestSection(Patch *upm, ADDRESS *offset);
extern Section *UPMGetSourceSection(Patch *upm, ADDRESS *offset);
extern void InstrDestroyUPMs(Instruction *instr);
extern void FunDestroyUPMs(Function *fun);
extern UPMList *UPMMemberListCreate(int num_members);
extern void UPMMemberListDestroy(UPMList *list);

extern void UPMChangeSourceFromSectionAddressToSectionAddress(Patch *upm,
Section *old_sec, ADDRESS old_offset, Section *new_sec, ADDRESS new_offset);

extern void UPMChangeDestFromSectionAddressToSectionAddress(Patch *upm,
Section *old_sec, ADDRESS old_offset, Section *new_sec, ADDRESS new_offset);

extern void UPMChangeDestFromInstrToInstr(Patch *upm, Instruction *old_instr,
Operand *old_op, Instruction *new_instr, Operand *new_op);

extern void UPMChangeAllSourcesFromInstrToInstr(
Instruction *old_instr, Instruction *new_instr);

extern void UPMChangeAllDestinationsFromInstrToInstr(
Instruction *old_instr, Instruction *new_instr);

extern bool AddressIsUPMSource(Section *section, ADDRESS offset);
extern bool AddressIsUPMDest(Section *section, ADDRESS offset);
extern bool InstrIsUPMSource(Instruction *instr);

#endif
