/*
 *  util.h -- include file for using the util subdirectory
 */

#ifndef UTIL_H
#define UTIL_H

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>



/*  simple functions  */

void *alloc(int amount);
void *zalloc(int amount);
void *salloc(char *string);

int roundup(int value, int quantum);
bool memberOf(int item, int arr[], int size);

int loadLE(void *address);
void storeLE(void *address, int value);
void storeLE2(void *address, int value);



/*
 *  Output macros.  These should be used for all text output from PLTO.
 *  "fmt" and "args" are as in printf(fmt, args).
 *
 *  Errors and warnings are sent to stderr, and are also echoed on stdout
 *  if it is not the terminal (where presumably the stderr copy suffices).
 *
 *  Reports and debugging messages go to stdout, and are output or
 *  suppressed depending on globals set by command-line options.
 *  Note that these macros execute code conditionally, and so their
 *  arguments should not produce side effects (e.g. code like "round++").
 *
 *  Reports are classified by "level" (see definitions below) ranging from
 *  RP_MAJOR to RP_TRIVIAL.  Choose levels carefully for new messages.
 *  Each message should make sense whether or not less-important messages
 *  are output.
 */

//  WARNING(fmt,args) -- issue a warning message, and continue.
#define WARNING(...) warning(__VA_ARGS__)

//  FATAL(fmt,args) -- diagnose a fatal error, and abort.
//  For unrecoverable errors that do not indicate a bug in PLTO.
#define FATAL(...) fatal(__VA_ARGS__)

//  PFATAL(msg) -- diagnose a fatal error due to a failed system call.
//  msg (typically a file name) is printed with the system failure message.
#define PFATAL(msg) pfatal(msg)

//  MALFUNCTION(fmt,args) -- diagnose a malfunction, and abort.
//  A malfunction is an error in PLTO itself, rather than its input.
#define MALFUNCTION(...) malfunction(__func__, __FILE__, __LINE__, __VA_ARGS__)

//  REPORT(n,fmt,args) -- issue a report if level n messages are wanted
#define REPORT(n, ...) (WANT_REPORT(n) ? report(n, __VA_ARGS__) : (void)0)

//  DEBUG(c,fmt,args) -- issue a debugging message if class c is selected
#define DEBUG(c, ...) (WANT_DEBUG(c) ? debug(c, __VA_ARGS__) : (void)0)

//  for printing function name after a string
#define WITHFUNC(msg) msg " [%s]" , __func__

//  true if a report message of level n should be written to output
#define WANT_REPORT(n) (verbosityLevel >= (n))

//  true if character c designates a debugging class that has been selected
#define WANT_DEBUG(c) \
    (WANT_REPORT(RP_DEBUG) || ((c) < ' ' || (c) > '~' || debugFlags[(int)c]))

//  set debugging on or off for class c
#define SET_DEBUG(c,v) (debugFlags[(int)c] = (bool)(v))

// Prepends the function name to the string
#define WITHFUN(x) ("[%s] " x),__func__



/*
 *  Output function declarations.
 *  Use the macros above instead of calling these directly.
 */
void warning(char *fmt, ...);
void fatal(char *fmt, ...);
void pfatal(char *mess);
void malfunction(const char *func, char *file, int line, char *fmt, ...);
void report(int level, char *fmt, ...);
void debug(char c, char *fmt, ...);



/*
 *  Level definitions for issuing reports.
 */
#define RP_NONE      0	// nothing but errors
#define RP_MAJOR     1	// major milestones
#define RP_MINOR     2	// minor milestones
#define RP_DETAIL    3	// additional details at milestones
#define RP_SMLOOP    4	// small-loop reports (typically < 100 iterations)
#define RP_SMDETAIL  5	// small-loop additional details
#define RP_LGLOOP    6	// large-loop reports (voluminous)
#define RP_LGDETAIL  7	// large-loop additional details
#define RP_TRIVIAL   8	// the most trivial details
#define RP_DEBUG     9	// all debugging messages


/*
 *  Debugging flag characters.  See config.c for long names used equivalently.
 */
#define DB_FLOW		'f'	// dump control flow graph
#define DB_STKG		'g'	// stack gain
#define DB_WRTGDB	'G'	// write GDB script
#define DB_STATHG	'H'	// generate static histograms
#define DB_INFAIL	'i'	// inlining failures
#define DB_JUMPS	'j'	// jump adjustments
#define DB_KILLS	'k'	// killed instructions
#define DB_ODBG		'o'	// output debugging sections
#define DB_CPFAIL	'p'	// constant propagation failures
#define DB_DEADREG	'r'	// dead register stores
#define DB_DEADSTK	's'	// dead stack stores
#define DB_WADDR	'T'	// write address translations
#define DB_UNBUF	'u'	// unbuffer output
#define DB_WGTHG	'W'	// generate weighted histograms



/*
 *  Most of the following Java-inspired functions
 *  were originally from Martin Stepp.
 */



/*  helper functions  */

int intCompare(void *o1, void *o2);
int refCompare(void *o1, void *o2);
int strCompare(void *o1, void *o2);
void intPrint(void *item);
void refPrint(void *item);
void strPrint(void *item);



/* Iterators */

typedef struct Iterator {
    void *it_dataStructure;
    void *it_current;
    void *it_opaque1;
    void *it_opaque2;
    void *(*it_next) (struct Iterator *);
    int (*it_hasNext) (struct Iterator *);
    void *(*it_remove) (struct Iterator *);
} Iterator;



/* ArrayLists */

typedef enum { AL_LIST_UNSORTED, AL_LIST_SORTED, AL_LIST_SET } ATYPE;

typedef struct ArrayList {
    void **al_elements;
    int al_size;
    int al_capacity;
    ATYPE al_type;
    int al_lock;
    int (*al_compare) (void *, void *);
    void (*al_print) (void *);
    void (*al_free) (void *);
} ArrayList;

ArrayList *al_new(void);
ArrayList *al_newT(ATYPE type);
ArrayList *al_newInt(ATYPE type);
ArrayList *al_newPtr(ATYPE type);
ArrayList *al_newStr(ATYPE type);
ArrayList *al_newGeneric(ATYPE type, int (*compareFunc) (void *, void *),
    void (*printFunc) (void *), void (*freeFunc) (void *));
void al_free(ArrayList *list);
void al_freeWithElements(ArrayList *list);

void *al_add(ArrayList *list, void *val);
void al_addAll(ArrayList *list1, ArrayList *list2);
void al_backmap(ArrayList *list, void (*func) (void **));
void al_clear(ArrayList *list);
void al_clearAndFreeElements(ArrayList *list);
ArrayList *al_clone(ArrayList *list);
int al_contains(ArrayList *list, void *val);
int al_equals(ArrayList *list1, ArrayList *list2);
void *al_first(ArrayList *list);
void *al_get(ArrayList *list, int index);
int al_indexOf(ArrayList *list, void *val);
void *al_insert(ArrayList *list, void *val);
void *al_insertAt(ArrayList *list, void *val, int index);
void *al_insertSorted(ArrayList *list, void *val);
int al_isEmpty(ArrayList *list);
Iterator *al_iterator(ArrayList *list);
void *al_last(ArrayList *list);
void al_map(ArrayList *list, void (*func) (void **));
void al_print(ArrayList *list);
void *al_remove(ArrayList *list, void *val);
void *al_removeAt(ArrayList *list, int index);
void *al_removeFirst(ArrayList *list);
void *al_removeLast(ArrayList *list);
int al_size(ArrayList *list);
void al_sort(ArrayList *list);

ArrayList *al_difference(ArrayList *list1, ArrayList *list2);
ArrayList *al_intersection(ArrayList *list1, ArrayList *list2);
ArrayList *al_union(ArrayList *list1, ArrayList *list2);
bool al_setEquals(ArrayList *list1, ArrayList *list2);



/* LinkedLists */

typedef enum { LIST_UNSORTED, LIST_SORTED, LIST_SET } LTYPE;

typedef struct LinkedList {
    struct Link *ll_head;		// dummy header
    struct Link *ll_tail;		// dummy tail
    int ll_size;			// # nodes
    LTYPE ll_type;			// flag
    int ll_lock;
    int (*ll_compare) (void *, void *);	// function to compare two nodes
    void (*ll_print) (void *);
    void (*ll_free) (void *);
} LinkedList;

extern LinkedList *ll_new(LTYPE type);
extern LinkedList *ll_newInt(LTYPE type);
extern LinkedList *ll_newPtr(LTYPE type);
extern LinkedList *ll_newStr(LTYPE type);
extern LinkedList *ll_newGeneric(LTYPE type, int (*compareFunc)(void *,void *),
    void (*printFunc) (void *), void (*freeFunc) (void *));
extern void ll_free(LinkedList *list);
extern void ll_freeWithElements(LinkedList *list);
extern void *ll_add(LinkedList *list, void *val);
extern void *ll_addLast(LinkedList *list, void *val);
extern void *ll_addFirst(LinkedList *list, void *val);
extern void ll_addAll(LinkedList *list1, LinkedList *list2);
extern void ll_backmap(LinkedList *list, void (*func) (void **));
extern void ll_clear(LinkedList *list);
extern void ll_clearAndFreeElements(LinkedList *list);
extern int ll_contains(LinkedList *list, void *val);
extern LinkedList *ll_difference(LinkedList *list1, LinkedList *list2);
extern int ll_equals(LinkedList *list1, LinkedList *list2);
extern void *ll_first(LinkedList *list);
extern void *ll_get(LinkedList *list, int index);
extern LinkedList *ll_intersection(LinkedList *list1, LinkedList *list2);
extern int ll_indexOf(LinkedList *list, void *val);
extern void *ll_insert(LinkedList *list, void *val);
extern void *ll_insertAt(LinkedList *list, void *val, int index);
extern void *ll_insertSorted(LinkedList *list, void *val);
extern int ll_isEmpty(LinkedList *list);
extern Iterator *ll_iterator(LinkedList *list);
extern void *ll_last(LinkedList *list);
extern void ll_map(LinkedList *list, void (*func) (void **));
extern void ll_print(LinkedList *list);
extern void *ll_remove(LinkedList *list, void *val);
extern void *ll_removeAt(LinkedList *list, int index);
extern void *ll_removeFirst(LinkedList *list);
extern void *ll_removeLast(LinkedList *list);
extern void *ll_push(LinkedList *list, void *val);
extern void *ll_pop(LinkedList *list);
extern void *ll_enqueue(LinkedList *list, void *val);
extern void *ll_dequeue(LinkedList *list);
extern int ll_size(LinkedList *list);
extern void ll_sort(LinkedList *list);
extern void ll_sort2(LinkedList *list);
extern LinkedList *ll_union(LinkedList *list1, LinkedList *list2);



/* HashTables */

#define DEFAULT_HASH_SIZE 10
#define DEFAULT_HASH_LOAD 0.75

typedef enum { HASH_INT, HASH_PTR, HASH_STR } HTYPE;

typedef struct HashEntry {
    int h_hash;
    void *h_key;
    void *h_value;
    struct HashEntry *h_next;
} HashEntry;

typedef struct HashTable {
    HashEntry *ht_values;
    int ht_capacity;
    int ht_size;
    double ht_load;
    int (*ht_hashCode) (void *ptr);
    int (*ht_compareKey) (void *, void *);
    int (*ht_compareValue) (void *, void *);
    void (*ht_printKey) (void *);
    void (*ht_printValue) (void *);
    void (*ht_free) (void *);
} HashTable;

HashTable *ht_new(HTYPE keyType, HTYPE valueType, int capacity, double load);

HashTable *ht_newGenericKey(HTYPE valueType,
    int capacity,
    double load,
    int (*hashFunc) (void *),
    int (*compareKeyFunc) (void *, void *), void (*printKeyFunc) (void *));

HashTable *ht_newGenericValue(HTYPE keyType,
    int capacity,
    double load,
    int (*compareValueFunc) (void *, void *),
    void (*printValueFunc) (void *), void (*freeValueFunc) (void *));

HashTable *ht_newGeneric(int capacity,
    double load,
    int (*hashFunc) (void *),
    int (*compareKeyFunc) (void *, void *),
    int (*compareValueFunc) (void *, void *),
    void (*printKeyFunc) (void *),
    void (*printValueFunc) (void *), void (*freeValueFunc) (void *));

void ht_free(HashTable *table);
void ht_freeWithElements(HashTable *table);
int refHashCode(void *obj);
int intHashCode(void *ptr);
int strHashCode(void *ptr);
void ht_clear(HashTable *table);
void ht_clearAndFreeElements(HashTable *table);
int ht_containsKey(HashTable *table, void *key);
int ht_containsValue(HashTable *table, void *value);
int ht_equals(HashTable *table1, HashTable *table2);
void *ht_get(HashTable *table, void *key);
int ht_isEmpty(HashTable *table);
Iterator *ht_iterator(HashTable *table);
void ht_print(HashTable *table);
void *ht_put(HashTable *table, void *key, void *value);
void ht_putAll(HashTable *table, HashTable *otherTable);
void ht_rehash(HashTable *table);
void *ht_remove(HashTable *table, void *key);
int ht_size(HashTable *table);



#endif				// UTIL_H
