/*
 *  General-usage utility functions.
 */
#ifndef UTILITIES_H
#define UTILITIES_H

#include "global.h"

/*
 *  When we mis-parsed the instruction stream (maybe there was data in it)
 *  or something, then we need to clear up all the wrong instructions. This
 *  function starts with the given instruction and deletes it, followed by
 *  everything after it.  All memory is returned to the OS
 */
void DeleteAllInstructionsStartingWith(Instruction *instr);

void RelocationFree(Relocation *reloc);

/* Conversions from types to string representations */
char *edgeTypeToString(int type);
char *blockTypeToString(int type);
char *lookupFunctionName(Function *fun);

/* Returns true if the operand uses a displacement */
bool operandHasDisplacement(Operand *op);

/* Conversion from a string to a section/symbol */
Section *nameToSection(const char *name);
Symbol *nameToSymbol(char *name);

bool DonotTouchMe(Function *fun);
/* Useful functions to convert an address to an instruction/data/section */
Instruction *FinalAddrToInstr(UINT32 address);
Instruction *AddressToInstr(UINT32 address);
UINT32 AddressToData(UINT32 address);
Section *AddressToSection(UINT32 address);
Section *NewAddressToSection(ADDRESS address);
Section *AddressToPrevSection(ADDRESS address);
Instruction *InstrContainingAddress(UINT32 destination);

/*
 *  Converts an address to an instruction by tracing backward
 *  byte-by-byte in the instruction stream until a valid instruction is
 *  encountered
 */
extern Instruction *BacktraceToInstr(UINT32 addr);
extern int PointsIntoInstruction(ADDRESS targetAddress);


/* Returns the first instruction in the function */
Instruction *functionFirstInstr(Function *fun);

/* Returns the last instruction in the function */
Instruction *functionLastInstr(Function *fun);

bool isOnlyInstrInBBL(Instruction *instr);

Instruction *IDToInstr(INT32 id);

/*
 *  Finds the symbol at some address.  Optionally name may be used to
 *  disambiguate different symbols at the same address.
 */
Symbol *AddressToSymbol(ADDRESS addr, char *name);

bool isOPOFFClass(Instruction *instr);
bool isOPDIRClass(Instruction *instr);
bool isJumpClass(Instruction *instr);
bool isBranchClass(Instruction *instr);
bool isPCChangingClass(Instruction *instr);
bool isNOPClass(Instruction *instr);
bool isFPClass(Instruction *instr);
bool isLoopClass(Instruction *instr);
bool isCallClass(Instruction *instr);
bool isReturnClass(Instruction *instr);
bool isLeaveClass(Instruction *instr);
bool isLeaClass(Instruction *instr);
bool isAddClass(Instruction *instr);
bool isAndClass(Instruction *instr);
bool isSubClass(Instruction *instr);
bool isHaltClass(Instruction *instr);
bool isMovClass(Instruction *instr);
bool isSyscallClass(Instruction *instr);
bool isDataInstruction(Instruction *instr);

long long CountInstructionsExecuted();

void BBLClearQuadwords();
void InstrClearTemp();

void BblAppendInstr(Block *ibbl, Instruction *iins);

Function *nameToFunction(char *str);

/*
 *  Increases the size of the section by nbytes, and copies the old
 *  data into the new space at the beginning
 */
void IncreaseSectionSize(Section *section, int nbytes);

Section *SectionNameToSection(Section *sections, char *name);
Section *GetContainingSection(ADDRESS address, bool useNewAddress);
void DumpAllSections();
int OperandOffset(Instruction *instr, Operand *op);
bool isPopClass(Instruction *instr);
bool isPushClass(Instruction *instr);
Operand *GetOperandForOperandAddress(Instruction *instr,
    ADDRESS operand_address);
Operand *GetOperandForInstrOffset(Instruction *instr, int offset);
void AddBasicBlockSymbols();
Symbol *AddNewFunctionSymbol(Function *fun);
Symbol *AddNewSymbolForInstr(char *name, Instruction *instr, UINT32 flags,
    Symbol *after);
Symbol *AddNewObjectSymbol(char *name, Section *section, ADDRESS offset,
    UINT32 flags);
char *GetSymbolSectionName(Symbol *symbol);

FList *GetFlistAtOffset(ADDRESS offset);

Symbol *GetStrongestSymbolForValue(ADDRESS value);
Symbol *GetSymbolForSectionAndValue(Section *section, ADDRESS value);

extern void WriteAddressTranslationFile(char *filename);

extern Section *SectionCreate();

/*
 *  Write out a gdb script file that can be used to debug this thing.
 *
 *  file_name: the name of the file that the gdb script should be written to
 *  extra_info_fun: a function to add any extra info to the gdb script
 */
extern void WriteGDBScript(char *file_name,
    void (*extra_info_fun) (FILE *file));

#endif				// UTILITIES_H
