/*
 *  x86-table-defs.h
 *
 *  The header file that most modules will want to include.
 */
#ifndef X86_TABLE_DEFS_H
#define X86_TABLE_DEFS_H

#include "x86-util.h"

typedef int (*op_rtn) (int bytemode, int aflag, int dflag);

/*
 *  BYTEMODES
 *  b_mode: 8 bit operand
 *  v_mode: 16 or 32 bit operand (check aflag/dflag)
 *  w_mode: 16 bit operand
 *  d_mode: 32 bit operand
 */
#define b_mode 1
#define v_mode 2
#define w_mode 3
#define d_mode 4

/*
 *  32 BIT REGISTERS
 */
#define all_regs  98
#define no_reg    99
#define es_reg   100
#define cs_reg   101
#define ss_reg   102
#define ds_reg   103
#define fs_reg   104
#define gs_reg   105
#define eAX_reg  107
#define eCX_reg  108
#define eDX_reg  109
#define eBX_reg  110
#define eSP_reg  111
#define eBP_reg  112
#define eSI_reg  113
#define eDI_reg  114
#define lptr     115

/*
 *  8 BIT REGISTERS
 */
#define al_reg 116
#define cl_reg 117
#define dl_reg 118
#define bl_reg 119
#define ah_reg 120
#define ch_reg 121
#define dh_reg 122
#define bh_reg 123

/*
 *  16 BIT REGISTERS
 */
#define ax_reg 124
#define cx_reg 125
#define dx_reg 126
#define bx_reg 127
#define sp_reg 128
#define bp_reg 129
#define si_reg 130
#define di_reg 131
#define indir_dx_reg 150

/*
 *  Floating Point Registers
 */
#define st0_reg  151
#define st1_reg  152
#define st2_reg  153
#define st3_reg  154
#define st4_reg  155
#define st5_reg  156
#define st6_reg  157
#define st7_reg  158
#define st8_reg  159
#define fpu_regs 160

/*
 *  X86op
 *  There are the entries that appear in the opcode table.  Each entry
 *  contains the type, class and name of the instruction, as well as
 *  function pointers to routines that are written specifically for its
 *  genre of instruction. Also appearing is a set of flags and implicit
 *  registers that the instruction may use or define.
 */
typedef struct X86op {
    int x_type;			// type of instruction: mov, add, etc
    int x_class;		// class of instruction: cond. jump, move, etc
    char *x_name;		// mnemonic for the instruction
    int x_uops;			// number of micro-ops used by the instruction
    op_rtn x_op1;		// operand handling routine for the dest. op
    int x_bytemode1;		// size of the dest. op
    op_rtn x_op2;		// operand handling routine for the source op
    int x_bytemode2;		// size of the source op
    op_rtn x_op3;		// operand handling routine for the source2 op
    int x_bytemode3;		// size of the source2 op
    int x_implicitDest1;	// first implicit destination register, if any
    int x_implicitDest2;	// second implicit destinatinon register, if any
    int x_implicitSource1;	// first implicit source register, if any
    int x_implicitSource2;	// second implicit source register, if any
    unsigned char x_pushBytes;	// number of bytes pushed on the stack
    unsigned char x_popBytes;	// number of bytes popped from the stack
    long long x_psw;		// PSW/FPU bits that can change
    long long x_pswUsed;	// PSW/FPU bits that are used
} X86op;

#define DIS386_IMPLICIT_DEST(dis)     (ConvertRegister(dis->x_implicitDest1))
#define DIS386_IMPLICIT_DEST2(dis)    (ConvertRegister(dis->x_implicitDest2))
#define DIS386_IMPLICIT_DEST1(dis)    (ConvertRegister(dis->x_implicitDest1))
#define DIS386_IMPLICIT_SOURCE(dis)   (ConvertRegister(dis->x_implicitSource1))
#define DIS386_IMPLICIT_SOURCE1(dis)  (ConvertRegister(dis->x_implicitSource1))
#define DIS386_IMPLICIT_SOURCE2(dis)  (ConvertRegister(dis->x_implicitSource2))

#define PSW_CF_INDEX             0
#define PSW_PF_INDEX             2
#define PSW_AF_INDEX             4
#define PSW_ZF_INDEX             6
#define PSW_SF_INDEX             7
#define PSW_TF_INDEX             8
#define PSW_IF_INDEX             9
#define PSW_DF_INDEX             10
#define PSW_OF_INDEX             11
#define PSW_IOPL1_INDEX          12
#define PSW_IOPL2_INDEX          13
#define PSW_NT_INDEX             14
#define PSW_RF_INDEX             16
#define PSW_VM_INDEX             17
#define PSW_AC_INDEX             18
#define PSW_VIF_INDEX            19
#define PSW_VIP_INDEX            20
#define PSW_ID_INDEX             21
#define FPU_C0_INDEX             22
#define FPU_C1_INDEX             23
#define FPU_C2_INDEX             24
#define FPU_C3_INDEX             25
#define DUF_INDEX                26	// Destination Used Flag
#define S1D_INDEX                27	// Source1 Defined Flag
#define S2D_INDEX                28	// Source2 Defined flag

#define PSW_CF                   (1 << PSW_CF_INDEX)
#define PSW_PF                   (1 << PSW_PF_INDEX)
#define PSW_AF                   (1 << PSW_AF_INDEX)
#define PSW_ZF                   (1 << PSW_ZF_INDEX)
#define PSW_SF                   (1 << PSW_SF_INDEX)
#define PSW_TF                   (1 << PSW_TF_INDEX)
#define PSW_IF                   (1 << PSW_IF_INDEX)
#define PSW_DF                   (1 << PSW_DF_INDEX)
#define PSW_OF                   (1 << PSW_OF_INDEX)
#define PSW_IOPL1                (1 << PSW_IOPL1_INDEX)
#define PSW_IOPL2                (1 << PSW_IOPL2_INDEX)
#define PSW_NT                   (1 << PSW_NT_INDEX)
#define PSW_RF                   (1 << PSW_RF_INDEX)
#define PSW_VM                   (1 << PSW_VM_INDEX)
#define PSW_AC                   (1 << PSW_AC_INDEX)
#define PSW_VIF                  (1 << PSW_VIF_INDEX)
#define PSW_VIP                  (1 << PSW_VIP_INDEX)
#define PSW_ID                   (1 << PSW_ID_INDEX)
#define PSW_ALL                  (0x003fffff)
#define FPU_C0                   (1 << FPU_C0_INDEX)
#define FPU_C1                   (1 << FPU_C1_INDEX)
#define FPU_C2                   (1 << FPU_C2_INDEX)
#define FPU_C3                   (1 << FPU_C3_INDEX)
#define DUF                      (1 << DUF_INDEX)	// Destination Used Flag
#define S1D                      (1 << S1D_INDEX)	// Source1 Defined Flag
#define S2D                      (1 << S2D_INDEX)	// Source2 Defined flag
#define FPU_ALL                  (FPU_C0 | FPU_C1 | FPU_C2 | FPU_C3)
#define PSW_PSEUDO_FLAGS         (DUF | S1D | S2D)

#define COMPLEX                  -1	/* The uops field of dis386 is COMPLEX
					 * if the number of uops cannot be
					 * determined or does not apply */

/*
 *  onebyte_has_modrm
 *  An entry in this table is 1 if the corresponding index in the table
 *  dis386[] uses a MOD/RM byte. An entry in this table is 0 otherwise.
 */
static const unsigned char onebyte_has_modrm[256] = {
    /* 00 */  1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0,  /* 0f */
    /* 10 */  1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0,  /* 1f */
    /* 20 */  1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0,  /* 2f */
    /* 30 */  1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0,  /* 3f */
    /* 40 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 4f */
    /* 50 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 5f */
    /* 60 */  0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0,  /* 6f */
    /* 70 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 7f */
    /* 80 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /* 8f */
    /* 90 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 9f */
    /* a0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* af */
    /* b0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* bf */
    /* c0 */  1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,  /* cf */
    /* d0 */  1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1,  /* df */
    /* e0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* ef */
    /* f0 */  0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1   /* ff */
};

/*
 *  twobyte_has_modrm
 *  An entry in this table is 1 if the corresponding index in the table
 *  dis386_twobyte[] uses a MOD/RM byte. An entry in this table is 0 otherwise.
 */
static const unsigned char twobyte_has_modrm[256] = {
    /* 00 */  1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1,  /* 0f */
    /* 10 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 1f */
    /* 20 */  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,  /* 2f */
    /* 30 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 3f */
    /* 40 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /* 4f */
    /* 50 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 5f */
    /* 60 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1,  /* 6f */
    /* 70 */  0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1,  /* 7f */
    /* 80 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  /* 8f */
    /* 90 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /* 9f */
    /* a0 */  0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1,  /* af */
    /* b0 */  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1,  /* bf */
    /* c0 */  1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,  /* cf */
    /* d0 */  0, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1,  /* df */
    /* e0 */  0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1,  /* ef */
    /* f0 */  0, 1, 1, 1, 0, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0   /* ff */
};

/*
 *  OPERAND HANDLING ROUTINES
 */
extern int OP_E(int, int, int);
extern int OP_G(int, int, int);
extern int OP_I(int, int, int);
extern int OP_indirE(int, int, int);
extern int OP_sI(int, int, int);
extern int OP_REG(int, int, int);
extern int OP_J(int, int, int);
extern int OP_DIR(int, int, int);
extern int OP_OFF(int, int, int);
extern int OP_ESDI(int, int, int);
extern int OP_DSSI(int, int, int);
extern int OP_SEG(int, int, int);
extern int OP_C(int, int, int);
extern int OP_D(int, int, int);
extern int OP_T(int, int, int);
extern int OP_rm(int, int, int);
extern int OP_ST(int, int, int);
extern int OP_STi(int, int, int);
extern int OP_MMX(int, int, int);
extern int OP_EM(int, int, int);
extern int OP_MS(int, int, int);
extern int OP_3DNow(int, int, int);
extern int OP_3DNowSuffix(int, int, int);
extern int OP_PREFETCH(int, int, int);

/*
 *  OPCODE TABLES
 */
extern X86op dis386[];
extern X86op dis386_twobyte[];
extern X86op grps[][8];
extern X86op float_mem[];
extern X86op float_reg[][8];
extern X86op fgrps[][8];

#endif				// X86_TABLE_DEFS_H
