#include <stdio.h>
#include <stdlib.h>
#include "global.h"
#include "NDG.h"

static int debug_it = 0;

static void print_elem(NDG *graph, NDGElem *elems) {
    NDGElem *elem;

    for (elem = elems; elem; elem = elem->next) {
	graph->print_data (elem->data);
    }
}

static void print_node(NDG *graph, NDGNode *node) {
    printf ("(");
    print_elem (graph, node->elems);
    printf (")");
}

static void print_edge(NDG *graph, NDGEdge *edge) {
    print_node (graph, edge->v1);
    printf (" -%lld-> ", edge->weight);
    print_node (graph, edge->v2);

}

/*
 * NDGCreate ()
 *
 * Create a Non-directed graph.
 */
NDG *NDGCreate() {
    NDG *graph = (NDG *) alloc (sizeof (*graph));

    graph->nodes = NULL;
    graph->num_nodes = 0;
    graph->edges = NULL;
    graph->num_edges = 0;
    graph->Weight = NULL;
    graph->table_size = 0;
    
    return graph;
}

/*
 * FindNode ()
 * 
 * Find a node which has an element that contains the input data.
 */
NDGNode *FindNode(NDG *graph, void *data) {
    NDGNode *node;
    NDGElem *elem;

    for (node = graph->nodes; node; node = node->next) {
	for (elem = node->elems; elem; elem = elem->next) {
	    if (graph->data_cmp (elem->data, data) == 0) {
		return node;
	    }
	}
    }
    return NULL;
}

/*
 * FindEdge ()
 *
 * Find an edge which connect node v1 and node v2
 */
NDGEdge *FindEdge(NDG *graph, NDGNode *v1, NDGNode *v2) {
    NDGEdge *edge;

    for (edge = graph->edges; edge; edge = edge->next) {
	if (edge->v1 == v1 && edge->v2 == v2) {
	    return edge;	
	}
    }
    return NULL;
}

static void add_elem(NDGNode *node, void *data) {
    NDGElem *elem = (NDGElem *) alloc (sizeof (*elem));
    elem->data = data;
    elem->next = NULL;
    elem->prev = NULL;

    elem->next = node->elems;
    if (node->elems) {
	node->elems->prev = elem;
    }
    node->elems = elem;
}

static NDGElem *get_last_elem(NDGElem *elems) {
    NDGElem *elem;
    
    for (elem = elems; elem->next; elem = elem->next)
	;

    return elem;
}

static NDGElem *reverse_elem_list(NDGElem *elems) {
    NDGElem *elem, *next, *prev;

    elem = elems;
    next = NULL;
    prev = NULL;
    while (elem) {
	next = elem->next;
	prev = elem->prev;
	elem->prev = next;
	elem->next = prev;
	elem = next;
    }
    
    /* prev is the head of elem_list */
    return prev;
}

static NDGNode *new_node() {
    NDGNode *node = (NDGNode *) alloc (sizeof(*node));

    node->elems = NULL;
    node->idx = 0;
    node->next = NULL;
    
    return node;
}

static void delete_node(NDG *graph, NDGNode * node) {
    NDGNode *prev, *cur;

    prev = NULL;
    for (cur = graph->nodes; cur; prev = cur, cur = cur->next) {
        if (cur == node) {
            break;
	}
    }

    if (cur == NULL) {
        return;
    }

    if (prev == NULL) {
        graph->nodes = graph->nodes->next;
    } else {
        prev->next = cur->next;
    }
    graph->num_nodes--;	
}

static void add_node(NDG *graph, NDGNode *node) {
    node->next = graph->nodes;
    graph->nodes = node;
    graph->num_nodes++;	
}

static NDGEdge *new_edge(NDGNode *v1, NDGNode *v2, WINT weight) {
    NDGEdge *edge = (NDGEdge *) alloc (sizeof (*edge));

    edge->v1 = v1;
    edge->v2 = v2;
    edge->weight = weight;
    edge->next = NULL;

    return edge;
}

static void add_edge(NDG *graph, NDGEdge *edge) {
    edge->id = graph->num_edges;
    edge->next = graph->edges;
    graph->edges = edge;
    graph->num_edges++;   
}

/*
 * NDGAddEdge ()
 *
 * Add an edge into the graph. At first, we create a node for each new data, 
 * i.e., this node contains only one data. 
 */
void NDGAddEdge(NDG *graph, void *data1, void *data2, WINT weight) {
    NDGNode *v1, *v2;
    NDGEdge *edge;

    if (debug_it) {
	printf (" - add edge:"); 
	graph->print_data (data1); 
	printf (" -%lld-> ", weight); 
	graph->print_data(data2); 
	printf ("\n");
    }

    if ((v1 = FindNode (graph, data1)) == NULL) {
	v1 = new_node ();
	add_elem(v1, data1);
	add_node(graph, v1);
    }

    if ((v2 = FindNode (graph, data2)) == NULL) {
	v2 = new_node ();
	add_elem(v2, data2);
	add_node(graph, v2);
    }
    
    if (v1 == v2) {
	/* avoid edge between itself */
	return;
    }

    if ((edge = FindEdge (graph, v1, v2)) != NULL) {
	/* 
	 * it could be that v1 call v2 from different place in v1,
	 * therefore, there could be already an edge from v1->v2,
	 * if there is, add the weight of new call from v1->v2 onto
	 * the existing edge between v1->v2 
	 */
	edge->weight += weight;
    } else {
	edge = new_edge (v1, v2, weight);
	add_edge (graph, edge);
    }
}

void NDGInitWeight(NDG *graph) {
    int i, j;
    NDGEdge *edge;
    NDGNode *v1, *v2;

    graph->table_size = graph->num_nodes;
    if (debug_it) {
	printf (" -- NDG table size =%d\n", graph->num_nodes);
    }

    graph->nodes_index = 
	(NDGNode **) alloc(sizeof (NDGNode *) * graph->num_nodes);

    /* set index of all nodes */
    for (i = 0, v1 = graph->nodes; v1; v1 = v1->next, i++) {
	v1->idx = i;
	/* initially each node has only one element */
	v1->elems->idx = i;
	graph->nodes_index[i] = v1;
    }
			    
    /* init all weight as 0 */
    graph->OrigWeight = (WINT **)alloc(sizeof(WINT *) * graph->num_nodes);
    graph->Weight = (WINT **)alloc(sizeof(WINT *) * graph->num_nodes);

    for (i = 0; i < graph->num_nodes; i++) {
	(graph->OrigWeight)[i] = 
	    (WINT *) alloc (sizeof (WINT) * graph->num_nodes);
	(graph->Weight)[i] = 
	    (WINT *) alloc (sizeof (WINT) * graph->num_nodes);

	for (j = 0; j < graph->num_nodes; j++) {
	    graph->Weight[i][j] = 0;
	    graph->OrigWeight[i][j] = 0;
	}
    }

    /* 
     * Update the weight table based on edge
     * Note: 
     * since graph is a non-direct graph, 
     *    (graph->Weight)[x][y] = (graph->Weight)[y][x]
     *
     * However, the edges given are directed edge.
     * It could be the case that there are two edges: A->B and B->A
     * We simply just take the sum of both edges' weight.
     */
    for (edge = graph->edges; edge; edge = edge->next) {
	v1 = edge->v1; 
	v2 = edge->v2;

	(graph->Weight)[v1->idx][v2->idx] = edge->weight;
	(graph->Weight)[v2->idx][v1->idx] = edge->weight;

	(graph->OrigWeight)[v1->idx][v2->idx] = edge->weight;
	(graph->OrigWeight)[v2->idx][v1->idx] = edge->weight;

    }
}

static inline NDGNode *find_node_by_idx(NDG *graph, int idx) {
    return graph->nodes_index[idx];
}

static int find_node(NDG *graph, NDGNode *fnode) {
    NDGNode *node;

    for (node = graph->nodes; node; node = node->next) {
	if (node == fnode) {
	    return 1;
	}
    }
    return 0;
}

static int compare_edge(const void *e1, const void *e2) {
    return (int)((((NDGEdge *)e2)->weight - ((NDGEdge *)e1)->weight));
}

/*
 * GetEdges ()
 *
 * Search the current Weight table and find all the existing edges in the table.
 * Sort these edges in decreasing order and return the list of the sorted edges.
 */ 
static NDGEdge *GetEdges(NDG *graph) {
    int i, j, k;    
    int num_edges;
    NDGEdge *E, *array;
    NDGNode *v1, *v2;

    num_edges = 0;
    for (i = 0; i < graph->table_size; i++) {
	for (j = 0; j < graph->table_size; j++) {
	    if (i < j) {
		/* only look at the up half matrix */
		if (find_node_by_idx (graph, i) && 
		    find_node_by_idx (graph, j) && 
		    graph->Weight[i][j] > 0) {
		    num_edges++;
		}
	    }
	}
    }   
 
    if (num_edges == 0) {	
	return NULL;
    }

    array = (struct _edge_list *) alloc (sizeof (struct _edge_list)*num_edges);
    k = 0;
    for (i = 0; i < graph->table_size; i++) {
	for (j = 0; j < graph->table_size; j++) {
	    if (i < j) {
		v1 = find_node_by_idx (graph, i);
		v2 = find_node_by_idx (graph, j);
		if (v1 && v2 && graph->Weight[i][j] > 0) {
		    array[k].v1 = v1;
		    array[k].v2 = v2;
		    array[k].weight = graph->Weight[i][j];
		    k++;
		}
	    }
	}
    }   

    qsort (array, num_edges, sizeof(NDGEdge), compare_edge);	

    E = &array[0];
    for (i = 0; i < num_edges - 1; i++) {
        array[i].next = &array[i+1];
    }
    array[i].next = NULL;
    
    return E;
}

/*
 * MergeEdge ()
 *
 * Merge nodes based on currect Weight table. More details see Pettis-Hansen 
 * paper 3.2. 
 */
static void MergeEdge(NDG *graph, NDGEdge *edge) {    
    NDGNode *v1, *v2, *node;
    NDGElem *head;
    NDGElem *h1, *h2, *t1, *t2;
    int i;
    WINT w1, w2, w3, w4;

    v1 = edge->v1; 
    v2 = edge->v2; 

    if (!find_node (graph, v1) || !find_node (graph, v2)) {
	/* 
	 * make sure nodes are still avaible, not been 
	 * merged by previous action 
	 */
	return;
    }

    /* there can be four different ways to connect list of v1 and
     * list of v2:
     *   1 [h1, t1]:[h2, t2] 
     *   2 [h1, t1]:[t2, h2]
     *   3 [t1, h1]:[h2, t2]
     *   4 [t1, h1]:[t2, h2]
     */
    h1 = v1->elems;
    h2 = v2->elems;

    t1 = get_last_elem (h1); 
    t2 = get_last_elem (h2); 

    w1 = graph->OrigWeight[t1->idx][h2->idx];
    w2 = graph->OrigWeight[t1->idx][t2->idx];
    w3 = graph->OrigWeight[h1->idx][h2->idx];
    w4 = graph->OrigWeight[h1->idx][t2->idx];	
    
    if (w1 >= w2 && w1 >= w3 && w1 >= w4) {
	head = h1;
	t1->next = h2;
	h2->prev = t1;
    } else if (w2  >=  w1 && w2 >= w3 && w2 >= w4) {
	head = h1;
	/* reverse [h2,t2]. after head become tail, tail become head 
	 * so the new list is [t2, h2]
	 */
	reverse_elem_list (h2);
	t1->next = t2;
	t2->prev = t1;
    } else if (w3 >= w1 && w3 >= w2 && w3 >= w4) {
	/* reverse [h1,t1] to [t1, h1] */	
	reverse_elem_list (h1);
	head = t1;
	h1->next = h2;
	h2->prev = h1;
    } else if (w4 >= w1 && w4 >= w2 && w4 >= w3) {
	reverse_elem_list (h1);
	head = t1;
	reverse_elem_list (h2);
	h1->next = t2;
	t2->prev = h1;
    } else {
	/* impossible because w1<w2 and w2<w1 are impossible*/
    }

    /*
     * Create a new node. For the purpose of reuse weight table,
     * we use v1->idx as the idx of the new node
     */

    /* add new node into nodes list */
    node = new_node ();
    node->idx = v1->idx;
    node->elems = head;
        
    add_node (graph, node);

    /* now delete v1 and v2 from nodes list */
    delete_node (graph, v1);
    delete_node (graph, v2);

    /* update nodes_index */
    graph->nodes_index[v1->idx] = node;
    graph->nodes_index[v2->idx] = NULL;

    /* ToDo: free v1 and v2 */
    
    
    /* update weight table in the graph */
    /* traverse all the edges associated with v2 */

    for (i = 0; i < graph->table_size; i++) {
	if (graph->Weight[v2->idx][i] && i != v1->idx) {
	    /* found an edge between v2-V and V is not v1 */
	    /* add the weight of v2-V onto v1-V */
	    graph->Weight[v1->idx][i] += graph->Weight[v2->idx][i];
	    graph->Weight[i][v1->idx] += graph->Weight[v2->idx][i];	    
	}
    }

}

void MergeEdgeAll(NDG *graph, NDGEdge *edge) {
    NDGEdge *largest;

    for (largest = edge; largest; largest = largest->next) {
	MergeEdge (graph, largest);
    }
}

/*
 * NDGSort ()
 * 
 * Apply Pettis-Hansen procedure positioning on given call graph.
 *
 */
void NDGSort(NDG *graph) {
    NDGEdge *E;
    int round = 0;

    while (1) {

	if (debug_it) {
	    printf (" -- round=%d, num_nodes=%d\n", round, graph->num_nodes);
	}

	E = GetEdges (graph);
	if (E == NULL) {
	    break;
	}
	if (debug_it) {	    
	    printf (" found edges:"); print_edge(graph, E); printf ("\n");
	}
	MergeEdge (graph, E);
	if (debug_it) {
	    NDGPrintGraph (graph);
	}
	round ++;
    }
    if (debug_it) {
	NDGPrintGraph (graph);
    }
}

void NDGPrintGraph(NDG *graph) {
    NDGNode *v1, *v2;

    if (debug_it) {
    for (v1 = graph->nodes; v1; v1 = v1->next) {
	printf ("\t");
	print_node (graph, v1);
    }
    }
    printf ("[NDG Weight]:\n");

    for (v1 = graph->nodes; v1; v1 = v1->next) {
	print_node(graph, v1);
	for (v2 = graph->nodes; v2; v2 = v2->next) {
	    printf ("\t%lld", graph->Weight[v1->idx][v2->idx]);
	}
	printf ("\n\n");
    }        
}
