/*
 *  cache.analyze.c
 *
 *  Functions for performing icache/dcache analysis
 *
 *  Notes: All numbers should be taken with a Grain of Salt, as some
 *  loop sizes cannot be computed precisely until the final binary is
 *  written. Instruction lengths and padding for alignment, among other
 *  things, are subject to change.
 */
#include "global.h"
#include "math.h"
#include "cache.analyze.h"
#include "flowgraph.h"
#include "sysdep.h"
#include "dominator.h"
#include "utilities.h"

/*
 *  We treat all instruction caches as 70% of their real capacity, so that
 *  any errors we may make will probably not result in terrible cache overflows
 */
#define ICACHE_INLINE_THRESHOLD .70

static void ComputeLoopSizeHelper(Block *header);

static void AddBasicBlockSizes(Block *bbl, double *size);
static void AddFunSize(Function *fun, double *size);

/* A helper function for ComputeFunSizeRecursively */
static void ComputeFunSizeRecursivelyPartTwo(Function *fun,
    EDGE_WEIGHT_TYPE weight);

/* Returns the number of calls to 'fun' in the loop given by 'loopHeader' */
static int CountMultiplicityOfCallInLoop(Function *fun, Block *loopHeader);

/* Returns the number of calls to 'checkForCallsToFun' in 'fun' */
static int CountCallsInFun(Function *checkForCallsToFun, Function *fun,
    double scale);

/* Helper routine to count multiplicities of calls */
static int CountsCallsInFunHelper(Function *check, Block *bbl, int calls,
    double scale);

/* Frees up the memory used by the looplist */
static void DeallocateLooplist(LoopList *l);

static void ZeroOutFrequenciesForReachableBlocks(Block *loopHeader);
static void ZeroOutFrequenciesForReachableBlocksHelper(Block *loopHeader);
static void ZeroOutFunRecursively(Function *fun);

static EDGE_WEIGHT_TYPE headerWeight;
static EDGE_WEIGHT_TYPE largestWeight;



/*
 *  ZeroOutFrequenciesForReachableBlocks(loopHeader) --
 *
 *  Zeroes out the frequencies in all reachable blocks
 *  from the loop body given by loopHeader
 */
static void ZeroOutFrequenciesForReachableBlocks(Block *loopHeader) {
    Function *fiterator;

    FOREACH_FUNCTION(fiterator) {
	FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_TEMP_FLAG);
    }

    ZeroOutFrequenciesForReachableBlocksHelper(loopHeader);

    return;
}



/*
 *  ZeroOutFrequenciesForReachableBlocksHelper(loopHeader) --
 */
static void ZeroOutFrequenciesForReachableBlocksHelper(Block *loopHeader) {
    Block *iterator = loopHeader;

    while (iterator != NULL) {
	iterator->b_frequency = 0;

	if ((iterator->b_type == BT_CALL)
		&& (BBLGetFunDirectEdge(iterator) != NULL)) {
	    Function *callee = BBLGetFunDirectEdge(iterator)->e_to->b_fun;
	    ZeroOutFunRecursively(callee);
	}

	if (iterator == loopHeader) {
	    iterator = iterator->b_loopFirst;
	} else {
	    iterator = iterator->b_loopNext;
	}

	if (iterator != NULL && BBL_HAS_FLAG(iterator, BBL_LOOP_HEADER_FLAG)) {
	    ZeroOutFrequenciesForReachableBlocks(iterator);
	}
    }

    return;
}



/*
 *  ZeroOutFunRecursively(fun) --
 */
static void ZeroOutFunRecursively(Function *fun) {
    Block *biterator;

    if (FUNCTION_HAS_FLAG(fun, FUNCTION_TEMP_FLAG)) {
	return;
    }

    FUNCTION_SET_FLAG(fun, FUNCTION_TEMP_FLAG);

    FOREACH_BBL_IN_FUN(fun, biterator) {
	biterator->b_frequency = 0;
	if ((biterator->b_type == BT_CALL)
		&& (BBLGetFunDirectEdge(biterator) != NULL)) {
	    Function *callee = BBLGetFunDirectEdge(biterator)->e_to->b_fun;
	    ZeroOutFunRecursively(callee);
	}

    }

    return;
}



/*
 *  ComputeLoopSize(loopHeader) --
 *
 *  ComputeLoopSize takes, as an argument, the header basic block for a loop.
 *  It then computes the size that the loop will occupy in the icache by adding
 *  up instruction lengths
 */
int ComputeLoopSize(Block *loopHeader) {
    Function *fiterator;
    double size = 0;

    ZeroOutFrequenciesForReachableBlocks(loopHeader);
    ComputeLoopSizeHelper(loopHeader);

    FOREACH_FUNCTION(fiterator) {
	FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_TEMP_FLAG);
    }

    AddBasicBlockSizes(loopHeader, &size);

    return (int) size;
}



/*
 *  ComputeLoopSizeHelper(loopHeader) --
 *
 *  A helper function for ComputeLoopSize
 */
static void ComputeLoopSizeHelper(Block *loopHeader) {
    Block *iterator;
    double scalewt;

    headerWeight = loopHeader->b_weight;

    iterator = loopHeader;
    while (iterator != NULL) {
	scalewt = (double) iterator->b_weight / (double) headerWeight;

	if (scalewt > 1.0) {
	    iterator->b_frequency = 1.0;
	} else {
	    iterator->b_frequency = scalewt;
	}

	if ((iterator->b_type == BT_CALL)
		&& (BBLGetFunDirectEdge(iterator) != NULL)) {
	    Function *callee = BBLGetFunDirectEdge(iterator)->e_to->b_fun;
	    ComputeFunSizeRecursively(callee, iterator->b_weight);
	}

	if (iterator == loopHeader) {
	    iterator = iterator->b_loopFirst;
	} else {
	    iterator = iterator->b_loopNext;
	}

	if (iterator != NULL && BBL_HAS_FLAG(iterator, BBL_LOOP_HEADER_FLAG)) {
	    ComputeLoopSizeHelper(iterator);
	}
    }

    return;
}



/*
 *  AddBasicBlockSizes(loopHeader,size) --
 */
static void AddBasicBlockSizes(Block *loopHeader, double *size) {
    Block *iterator;

    iterator = loopHeader;
    while (iterator != NULL) {
	*size = *size + (ComputeBBLSize(iterator) * iterator->b_frequency);
	if ((iterator->b_type == BT_CALL)
		&& (BBLGetFunDirectEdge(iterator) != NULL)) {
	    Function *callee = BBLGetFunDirectEdge(iterator)->e_to->b_fun;
	    AddFunSize(callee, size);
	}

	if (iterator == loopHeader) {
	    iterator = iterator->b_loopFirst;
	} else {
	    iterator = iterator->b_loopNext;
	}

	if (iterator != NULL && BBL_HAS_FLAG(iterator, BBL_LOOP_HEADER_FLAG)) {
	    AddBasicBlockSizes(iterator, size);
	}
    }

    return;
}



/*
 *  ComputeFunSize(fun) --
 *
 *  Adds the lengths of the instructions in the function and returns the total.
 *  Any subfunctions that may be called are not examined here.
 */
int ComputeFunSize(Function *fun) {
    int funSize = 0;
    Block *biterator;

    FOREACH_BBL_IN_FUN(fun, biterator) {
	funSize += ComputeBBLSize(biterator);
    }

    return funSize;
}



/*
 *  ComputeFunSizeRecursively(fun,weight) --
 *
 *  Adds the lengths of the instructions in the function and any subfunctions,
 *  returning the total.  A depth-first descent is made into subfunctions,
 *  so that anything called from within the function will also be considered.
 */
void ComputeFunSizeRecursively(Function *fun, EDGE_WEIGHT_TYPE weight) {
    Function *fiterator;

    FOREACH_FUNCTION(fiterator) {
	FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_TEMP_FLAG);
    }

    ComputeFunSizeRecursivelyPartTwo(fun, weight);
}



/*
 *  AddFunSize(fun,size) --
 */
static void AddFunSize(Function *fun, double *size) {
    Block *biterator;

    if (FUNCTION_HAS_FLAG(fun, FUNCTION_TEMP_FLAG)) {
	return;
    }

    FUNCTION_SET_FLAG(fun, FUNCTION_TEMP_FLAG);

    FOREACH_BBL_IN_FUN(fun, biterator) {
	*size = *size + (ComputeBBLSize(biterator) * biterator->b_frequency);
	if ((biterator->b_type == BT_CALL)
		&& (BBLGetFunDirectEdge(biterator) != NULL)) {
	    Function *callee = BBLGetFunDirectEdge(biterator)->e_to->b_fun;
	    AddFunSize(callee, size);
	}

    }

    return;
}



/*
 *  ComputeFunSizeRecursivelyPartTwo(fun,weight) --
 *
 *  ComputeFunSizeRecursivelyPartTwo is used by ComputeFunSizeRecursively to
 *  perform a DFS on the function's CFG.
 */
static void ComputeFunSizeRecursivelyPartTwo(Function *fun,
    EDGE_WEIGHT_TYPE weight) {
    Block *biterator;
    double callFactor;

    if (FUNCTION_HAS_FLAG(fun, FUNCTION_TEMP_FLAG)) {
	return;
    }

    FUNCTION_SET_FLAG(fun, FUNCTION_TEMP_FLAG);

    callFactor = 0.0;
    if (fun->f_first->b_weight != 0) {
	callFactor = (double) weight / fun->f_first->b_weight;
    }

    FOREACH_BBL_IN_FUN(fun, biterator) {
	double scalewt =
	    (double) (callFactor * biterator->b_weight) / headerWeight;

	if (scalewt > 1.0) {
	    biterator->b_frequency = 1.0;
	} else {
	    biterator->b_frequency = scalewt;
	}

	if ((biterator->b_type == BT_CALL)
		&& (BBLGetFunDirectEdge(biterator) != NULL)) {
	    Function *callee = BBLGetFunDirectEdge(biterator)->e_to->b_fun;
	    ComputeFunSizeRecursivelyPartTwo(callee,
		callFactor * biterator->b_weight);
	}

    }

    return;
}



/*
 *  ComputeBBLSize(bbl) -- returns the size of the basic block, in bytes.
 */
int ComputeBBLSize(Block *bbl) {
    Instruction *iterator;
    int bblSize = 0;

    if (bbl->b_weight == 0) {
	return bblSize;
    }

    FOREACH_INSTR_IN_BBL(bbl, iterator) {
	if (iterator->i_length <= 0) {
	    UpdateInstrLength(iterator);
	}
	bblSize += iterator->i_length;
    }

    return bblSize;
}



/*
 *  GoodIdeaToInlineFunFromCallsite(fun,callSite) --
 *
 *  Returns true if inlining the function 'fun' from 'callsite' would be a good
 *  thing to do.  An icache-model is used to determined how many extra cache
 *  misses would be incurred. This number is compared with the savings from
 *  removing the call and return instruction, and a decision is made.
 */
bool GoodIdeaToInlineFunFromCallsite(Function *fun, Instruction *callSite) {
    int multiplicity, icacheCapacity, funSize, oldLoopSize;
    LoopList *l = GetLoops(callSite);
    Function *funs;

    FOREACH_FUNCTION(funs) {
	FUNCTION_REMOVE_FLAG(fun, FUNCTION_TEMP_FLAG);
    }

    icacheCapacity = getIcacheSize() * ICACHE_INLINE_THRESHOLD;

    if (l == NULL) {
	return true;
    }

    while (l != NULL) {
	largestWeight = -1;
	oldLoopSize = ComputeLoopSize(l->oo_header);
	multiplicity = CountMultiplicityOfCallInLoop(fun, l->oo_header);
	funSize = ComputeFunSize(fun);

	if (multiplicity > 1) {
	    if ((funSize + oldLoopSize) > icacheCapacity) {
		return false;
	    }
	}

	if (funSize > INLINING_MAX_FUNSIZE) {
	    return false;
	}

	l = l->oo_next;
    }

    DeallocateLooplist(l);
    return true;

}



/*
 *  GetLoops(callInstr) --
 *
 *  Returns a linked-list of all the loops containing 'callInstr'. Currently
 *  the list only includes loops detected within the function that contains
 *  'callInstr'.
 */
LoopList *GetLoops(Instruction *callInstr) {
    LoopList *l = NULL;
    Block *bbl = callInstr->i_Bblock;

    if (bbl->b_loopFirst == NULL) {
	bbl = bbl->b_loopHead;
    }

    if (bbl == NULL) {
	return NULL;
    }

    while (1) {
	if (bbl && bbl->b_loopFirst) {
	    if (l == NULL) {
		l = zalloc(sizeof(LoopList));
		l->oo_next = l->oo_prev = NULL;
		l->oo_header = bbl;
	    } else {
		LoopList *newLoop = zalloc(sizeof(LoopList));
		newLoop->oo_next = l;
		l->oo_prev = newLoop;
		newLoop->oo_header = bbl;
		l = newLoop;
	    }
	    bbl = bbl->b_loopHead;
	} else {
	    break;
	}
    }

    return l;
}



/*
 *  DeallocateLooplist(l) --
 *
 *  Deallocates a LoopList linked list.
 */
static void DeallocateLooplist(LoopList *l) {
    LoopList *next;

    while (l != NULL) {
	next = l->oo_next;
	free(l);
	l = next;
    }

    return;
}



/*
 *  CountMultiplicityOfCallInLoop(fun,loopHeader) --
 *
 *  Returns the number of calls to 'fun' inside the loop for which 'loopHeader'
 *  is the header basic block.
 */
static int CountMultiplicityOfCallInLoop(Function *fun, Block *loopHeader) {
    Block *biterator = loopHeader;
    int numberOfCalls = 0;

    FUNCTION_SET_FLAG(loopHeader->b_fun, FUNCTION_TEMP_FLAG);

    while (biterator != NULL) {
	if ((biterator->b_type == BT_CALL) && (biterator->b_weight > 0)) {
	    numberOfCalls =
		CountsCallsInFunHelper(fun, biterator, numberOfCalls, 1.0);
	} else if (biterator->b_loopFirst && (biterator != loopHeader)) {
	    numberOfCalls =
		numberOfCalls + CountMultiplicityOfCallInLoop(fun, biterator);
	}
	if (biterator != loopHeader) {
	    biterator = biterator->b_loopNext;
	} else {
	    biterator = biterator->b_loopFirst;
	}
    }

    return numberOfCalls;
}



/*
 *  CountCallsInFun(checkForCallsToFun,fun,scale) --
 *
 *  Marks 'fun' as being visited, and checks for any calls to
 *  'checkForCallsToFun' contained in 'fun'.
 */
static int CountCallsInFun(Function *checkForCallsToFun, Function *fun,
    double scale) {
    int numberOfCalls = 0;
    Block *biterator;

    if (FUNCTION_HAS_FLAG(fun, FUNCTION_TEMP_FLAG)) {
	return 0;
    }

    FUNCTION_SET_FLAG(fun, FUNCTION_TEMP_FLAG);

    FOREACH_BBL_IN_FUN(fun, biterator) {
	if ((biterator->b_type == BT_CALL) && (biterator->b_weight > 0)) {
	    numberOfCalls = CountsCallsInFunHelper(
		checkForCallsToFun, biterator, numberOfCalls, scale);
	}
    }

    return numberOfCalls;
}



/*
 *  CountsCallsInFunHelper(check,biterator,calls,scale) --
 */
static int CountsCallsInFunHelper(Function *check, Block *biterator, int calls,
    double scale) {
    Edge *funDirect = BBLGetFunDirectEdge(biterator);
    Function *callee;
    double scale2;
    double ignore = FloatParameter("InliningMultiplicityIgnore");

    if (funDirect == NULL) {
	return calls;
    }

    callee = funDirect->e_to->b_fun;
    if (BBLSumIncomingEdges(callee->f_first) > 0) {
	scale2 = biterator->b_weight / BBLSumIncomingEdges(callee->f_first);
    } else {
	scale2 = 0;
    }
    calls += CountCallsInFun(check, callee, scale * scale2);

    if (check == callee) {
	if (largestWeight == -1) {
	    largestWeight = biterator->b_weight + 1;
	    calls++;
	} else if (largestWeight > biterator->b_weight) {
	    if ((biterator->b_weight / (float) largestWeight) >= ignore) {
		calls++;
	    }
	} else if (biterator->b_weight >= largestWeight) {
	    if ((largestWeight / (float) biterator->b_weight) >= ignore) {
		calls++;
	    }
	    largestWeight = biterator->b_weight;
	}
    }

    return calls;
}
