/*
 *  This file contains definitions for functions used in cloning different
 *  structures.
 */
#include "global.h"
#include "clone.h"
#include "utilities.h"
#include "upm.h"
#include "flowgraph.h"
#include "cp.h"
#include "relocate.h"
#include "linkage.h"

static void CloneJumpTables(Function *fun);



/*
 *  FunClone(fun) --
 *
 *  FunClone creates a duplicate copy of a function and returns a pointer
 *  to the new copy.  The basic blocks and instructions will be linked in
 *  the cloned copy.  Edges will also be duplicated in the copy.
 *
 *  fun: the function to clone
 *  returns: a pointer to the clone function
 */
Function *FunClone(Function *fun) {
    Function *clone;
    EdgeNode *successors;
    EdgeNode *predecessors;
    Block *blockClones;
    Block *biterator;
    char *name;

    clone = FunCreate();

    /*
     * Mark each basic block to be cloned, and then call the function to
     * clone marked blocks.
     */
    FOREACH_BBL_IN_FUN(fun, biterator) {
	BBL_SET_FLAG(biterator, BBL_MARKED_FLAG);
    }

    biterator = blockClones = FunCloneMarkedBlocks(fun);

    /*
     * Set each block's function pointer to be 'clone', and set up the
     * first and last basic block for the function.
     */
    while (biterator != NULL) {
	Block *next = biterator->b_next;
	biterator->b_next = NULL;
	biterator->b_prev = NULL;
	biterator->b_fun = clone;

	if (clone->f_first == NULL) {
	    clone->f_first = clone->f_last = biterator;
	} else {
	    InsertBBLAfterBBL(biterator, clone->f_last);
	    clone->f_last = biterator;
	}

	biterator = next;
    }

    /*
     * Unfortunately FunCloneMarkedBlocks() directs ENTRY and EXIT edges
     * to the function from which our clone was made; we actually want
     * them to go to cloned function's pseudo entry and exit nodes. The
     * following two loops remove all ET_EXIT and ET_ENTRY edges from the
     * the cloned region. The two loops after them re-add the ET_EXIT and
     * ET_ENTRY edges, but direct them to the clone's entry and exit
     * nodes instead.
     */

    FOREACH_BBL_IN_FUN(clone, biterator) {
	successors = biterator->b_succ;
	while (successors != NULL) {
	    EdgeNode *next = successors->en_next;
	    if (NODE_TYPE(successors) == ET_EXIT) {
		BBLRemoveEdgeTyped(biterator, NODE_TO(successors), ET_EXIT);
	    }
	    successors = next;
	}
    }					/* end iteration over basic blocks */

    FOREACH_BBL_IN_FUN(clone, biterator) {
	predecessors = biterator->b_pred;
	while (predecessors != NULL) {
	    EdgeNode *next = predecessors->en_next;
	    if (NODE_TYPE(predecessors) == ET_ENTRY) {
		BBLRemoveEdgeTyped(NODE_FROM(predecessors), biterator,
		    ET_ENTRY);
	    }
	    predecessors = next;
	}
    }					/* end iteration over basic blocks */

    FOREACH_BBL_IN_FUN(fun, biterator) {
	FOREACH_SUCC_IN_BBL(biterator, successors) {
	    if (NODE_TYPE(successors) == ET_EXIT) {
		BBLAddEdge(BBLOriginalToClone(biterator, clone),
		    clone->f_exit, ET_EXIT);
	    }
	}
    }					/* end iteration over basic blocks */

    FOREACH_BBL_IN_FUN(fun, biterator) {
	FOREACH_PRED_IN_BBL(biterator, predecessors) {
	    if (NODE_TYPE(predecessors) == ET_ENTRY) {
		BBLAddEdge(clone->f_entry, BBLOriginalToClone(biterator,
			clone), ET_ENTRY);
	    }
	}
    }					/* end iteration over basic blocks */

    /*
     * Copy the flags into the clone, set up the function's name, and
     * relink the instruction list based on the basic block list.
     */

    clone->f_flags = fun->f_flags;
    FUNCTION_REMOVE_FLAG(clone, FUNCTION_NOCLONE_FLAGS);
    name = alloc(strlen(fun->f_name) + 10);
    sprintf(name, "%s_clone", fun->f_name);
    clone->f_name = name;

    InstrRelinkBasedOnBBL();

    CloneJumpTables(clone);

    return clone;
}



/*
 *  BBLClone(bbl) --
 *
 *  BBLClone creates a duplicate copy of a basic block and returns a
 *  pointer to the new copy.  The instructions in the basic block will be
 *  linked in the cloned copy.
 *
 *  bbl: the basic block to be cloned
 *  returns: a pointer to the clone basic block
 */
Block *BBLClone(Block *bbl) {
    Block *clone;
    Instruction *iterator;

    clone = BBLCreate();

    FOREACH_INSTR_IN_BBL(bbl, iterator) {
	Instruction *newInstr = InstrClone(iterator);
	newInstr->i_Bblock = clone;
	if (clone->b_first == NULL) {
	    clone->b_first = clone->b_last = newInstr;
	    /* make sure we stick an end flag on this instruction -- but
	     * it's temporary */
   	    INSTR_SET_FLAG(newInstr, BBL_START);        
	    INSTR_SET_FLAG(newInstr, BBL_END);
	} else {
	    InsertInstrAfterInstr(newInstr, clone->b_last);
	    clone->b_last = newInstr;
	}
    }

    clone->b_flags = bbl->b_flags;
    BBL_REMOVE_FLAG(clone, BBL_NOCLONE_FLAGS);

    clone->b_type = bbl->b_type;
    clone->b_weight = bbl->b_weight;

    clone->b_fun = bbl->b_fun;
    
    if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
	BBLAllocateRegshot(clone);
    }
    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	BBLAllocateStackshot(clone);
    }

    /* Why would have somebody put this piece of code
     * in here
     */
    //    clone->b_fun = NULL;

    if (bbl->b_flist != NULL) {
	clone->b_flist = bbl->b_flist;
    }

    return clone;
}



/*
 *  FunCloneMarkedBlocks(fun) --
 *
 *  This function clones only those basic blocks in fun that have been
 *  marked, and returns the resulting clone function.
 *
 *  fun: the function that should be cloned
 *  returns: a pointer to the clone function
 *  see also: FunClone
 *  see also: BBL_MARKED_FLAG
 */
Block *FunCloneMarkedBlocks(Function *fun) {
    Block *biterator = NULL, *bblClone = NULL;
    Block *succClone = NULL, *newBbl = NULL;
    Instruction *jmp = NULL;
    EdgeNode *niterator;
    
    Block *clonedList = NULL, **clonedListTail = NULL;
    Block *prevClone = NULL;

    clonedListTail = &clonedList;
    /* Clone all marked blocks */
    FOREACH_BBL_IN_FUN(fun, biterator) {
	if (BBL_HAS_FLAG(biterator, BBL_MARKED_FLAG)) {
	    Block *newBlock = BBLClone(biterator);
	    newBlock->b_clone = biterator;

	    /* Stick the cloned block onto the end of the cloned list */
            *clonedListTail = newBlock;
            newBlock->b_prev = prevClone;
            prevClone = newBlock;
            clonedListTail = &newBlock->b_next;        
	}
    }

    if (!clonedList) {
        return NULL;
    }
    /*
     * Now we have all the instructions, operands and bbls cloned. We need
     * to add edges appropriately. For intra-procedural edges, we will
     * connect cloned basic blocks with cloned basic blocks iff the edge
     * was going between two original blocks that were marked. If the edge
     * is between a cloned block and a non cloned block, we will add an edge
     * from the cloned region to the non-cloned region. The same applies for
     * interprocedural edges.
     */

    /* First duplicate the function entry and exit edges */
    FOREACH_BBL_IN_FUN(fun, biterator) {
        bblClone = BBLOriginalToCloneFromList(biterator, clonedList);
        
        if (!bblClone) {
            continue;
        }
        
	/* Find all entry points of the function */
	FOREACH_PRED_IN_BBL(biterator, niterator) {
	    if (NODE_TYPE(niterator) == ET_ENTRY) {
                BBLAddEdge(fun->f_entry, bblClone, ET_ENTRY);
	    }
	}

	/* Find all exit points of the function */
	FOREACH_SUCC_IN_BBL(biterator, niterator) {
	    if (NODE_TYPE(niterator) == ET_EXIT) {
                BBLAddEdge(bblClone, fun->f_exit, ET_EXIT);
	    }
	}
    }

    REPORT(RP_TRIVIAL, WITHFUN("Fixed function entry and exit edges"));

    FOREACH_BBL_IN_FUN(fun, biterator) {
        int hellPreds = BBLNumberOfPredecessorsFromBBL(biterator, HELL);
        int i;
        bblClone = BBLOriginalToCloneFromList(biterator, clonedList);
        
        if (!bblClone) {
            continue;
        }
        
        for (i = 0; i < hellPreds; i++) {
            BBLAddEdge(HELL, bblClone, ET_HELLMAYBE);
        }
    }

    /* Next handle all non entry/exit edges */
    FOREACH_BBL_IN_FUN(fun, biterator) {
        Block *bblClone;

        if (!BBL_HAS_FLAG(biterator, BBL_MARKED_FLAG)) {
            continue;
        }

        bblClone = BBLOriginalToCloneFromList(biterator, clonedList);

        CASSERT ((bblClone != NULL),
            "[opt.specialize.c]Block Marked but no clone\n");
        
        if (!bblClone) {
            continue;
        }

        /*
         * For function calls, we add 4 edges:
         * ET_FUNCALL, ET_FUNLINK, ET_FUNDIRECT and an ET_RETURN edge.
         * The weights of the edges are also duplicated here.
         */
        if (((biterator->b_type == BT_CALL)
                && (BBLGetFuncallEdge(biterator) != NULL))) {

            Block *funLinkClone = NULL;
            funLinkClone = BBLOriginalToCloneFromList(
                BBLGetFunlinkEdge(biterator)->e_to,
                    clonedList);

            /* If the block after function call is not cloned
             * jump to it. This is so that the function call can
             * return to a point only after the function call
             * so the funlink edge and block has to be cloned
             */
            
            if (!funLinkClone) {
                jmp = InstrCreateJMP();
                newBbl = BBLCreate();
                newBbl->b_fun = bblClone->b_fun;
                newBbl->b_first = jmp;
                newBbl->b_last  = jmp;

                INSTR_SET_FLAG(jmp, BBL_START);
                INSTR_SET_FLAG(jmp, BBL_END);                
                jmp->i_Bblock = newBbl;

                newBbl->b_next = bblClone->b_next;
                newBbl->b_prev = bblClone;

                funLinkClone = newBbl;
                
                if (bblClone->b_next) {                            
                    bblClone->b_next->b_prev = newBbl;
                } else {
                    // Else it is bblClone was the
                    // last block on the cloned list
                    // Hence now funLinkClone has to
                    // be the last block
                    *clonedListTail = funLinkClone; 
                    funLinkClone->b_prev = prevClone; 
                    funLinkClone->b_clone = NULL;
                    clonedListTail = &(funLinkClone->b_next); 
                    prevClone = funLinkClone; 
                }

                bblClone->b_next = newBbl;
                
                newBbl->b_type = BT_UNCOND;        

                
                /* Add a ET_FUNLINK edge between the the cloned
                 * block and the jmp block
                 */
                BBLAddEdge(funLinkClone,
                    BBLGetFunlinkEdge(biterator)->e_to,
                    ET_UNCOND);
            }
            
            if (bblClone->b_next != NULL) {
                BBLAddEdge(bblClone, funLinkClone, ET_FUNLINK);
                BBLAddEdge((BBLGetFuncallEdge(biterator)->e_to->b_fun->f_exit),
                    bblClone->b_next, ET_RETURN);

                BBLGetFunlinkEdge(bblClone)->e_edgeWeight =
                    BBLGetFunlinkEdge(biterator)->e_edgeWeight;
                BBLGetEdgeToBBL(
                    BBLGetFuncallEdge(biterator)->e_to->b_fun->f_exit,
                    BBLGetFunlinkEdge(bblClone)->e_to)->e_edgeWeight =
                    BBLGetFunDirectEdge(biterator)->e_edgeWeight;
            }

            BBLAddEdge(bblClone, BBLGetFunDirectEdge(biterator)->e_to,
                ET_FUNDIRECT);
            BBLAddEdge(bblClone, BBLGetFuncallEdge(biterator)->e_to,
                ET_FUNCALL);

            BBLGetFuncallEdge(bblClone)->e_edgeWeight =
                BBLGetFunDirectEdge(biterator)->e_edgeWeight;
            BBLGetFunDirectEdge(bblClone)->e_edgeWeight =
                BBLGetFunDirectEdge(biterator)->e_edgeWeight;
            continue;        
        }

        if ((biterator->b_type == BT_CALL) &&
            BBLHasHellSuccessor(biterator)) {

            Block *funLinkClone = NULL;

            funLinkClone = BBLOriginalToCloneFromList(
                BBLGetFunlinkEdge(biterator)->e_to,
                    clonedList);

            /* If the block after function call is not cloned
             * jump to it. This is so that the function call can
             * return to a point only after the function call
             * so the funlink edge and block has to be cloned
             */
            
            if (!funLinkClone) {
                jmp = InstrCreateJMP();
                newBbl = BBLCreate();
                newBbl->b_fun = bblClone->b_fun;
                newBbl->b_first = jmp;
                newBbl->b_last  = jmp;

                INSTR_SET_FLAG(jmp, BBL_START);
                INSTR_SET_FLAG(jmp, BBL_END);                
                jmp->i_Bblock = newBbl;

                newBbl->b_next = bblClone->b_next;
                newBbl->b_prev = bblClone;

                funLinkClone = newBbl;
                
                if (bblClone->b_next) {                            
                    bblClone->b_next->b_prev = newBbl;
                } else {
                    // Else it is bblClone was the
                    // last block on the cloned list
                    // Hence now funLinkClone has to
                    // be the last block
                    *clonedListTail = funLinkClone; 
                    funLinkClone->b_prev = prevClone; 
                    funLinkClone->b_clone = NULL;
                    clonedListTail = &(funLinkClone->b_next); 
                    prevClone = funLinkClone; 
                }

                bblClone->b_next = newBbl;
                
                newBbl->b_type = BT_UNCOND;        

                
                /* Add a ET_FUNLINK edge between the the cloned
                 * block and the jmp block
                 */
                BBLAddEdge(funLinkClone,
                    BBLGetFunlinkEdge(biterator)->e_to,
                    ET_UNCOND);
            }
            
            
            BBLAddEdge(bblClone, funLinkClone, ET_FUNLINK);
            BBLAddEdge(bblClone, HELL, ET_HELL);                
            BBLAddEdge(HELL, bblClone->b_next, ET_HELLMAYBE);
            continue; 
        }

        FOREACH_SUCC_IN_BBL(biterator, niterator) {

            /*
             * We've already handled all FUNCALL, FUNDIRECT, EXIT edges.
             */

            if (NODE_TYPE(niterator) == ET_FUNCALL) {
                continue;
            }
            if (NODE_TYPE(niterator) == ET_FUNDIRECT) {
                continue;
            }
            if (NODE_TYPE(niterator) == ET_EXIT) {
                continue;
            }

            if (NODE_TYPE(niterator) == ET_HELL) {
                BBLAddEdge(bblClone, HELL, ET_HELL);
            } else if (isInterproceduralEdge(niterator->en_edge)) {

                // WARNING("[OPT:Inline] Interprocedural non-funcall"
                //  " edge duplicated. Source: 0x%8.8x, Dest: 0x%8.8x",
                // NODE_FROM(niterator)->b_first->i_address,
                // NODE_TO(niterator)->b_first->i_address);

                BBLAddEdge(bblClone, NODE_TO(niterator),
                        NODE_TYPE(niterator))->e_edgeWeight =
                    NODE_WEIGHT(niterator);
            } else {
                if (!((NODE_TYPE(niterator) == ET_FALSE)
                        || (NODE_TYPE(niterator) == ET_TRUE)
                        || (NODE_TYPE(niterator) == ET_UNCOND)
                        || (NODE_TYPE(niterator) == ET_NORMAL)
                        || (NODE_TYPE(niterator) == ET_SYSCALL_LINK)                        
                        || (NODE_TYPE(niterator) == ET_FUNLINK)
                        || (NODE_TYPE(niterator) == ET_SWITCH)
                        || (NODE_TYPE(niterator) == ET_DATALINK))) {
                    WARNING("Encountered unexpected edge type that is"
                        " non-interprocedural: %x", NODE_TYPE(niterator));
                }

                if (BBL_HAS_FLAG(NODE_TO(niterator), BBL_MARKED_FLAG)) {
                    BBLAddEdge(bblClone,
                            BBLOriginalToCloneFromList(NODE_TO(niterator),
                                clonedList), NODE_TYPE(niterator))
                                ->e_edgeWeight = NODE_WEIGHT(niterator);
                } else {
                    if (NODE_TYPE(niterator) == ET_FALSE) {
                        
                        jmp = InstrCreateJMP();
                        newBbl = BBLCreate();
                        newBbl->b_fun = bblClone->b_fun;
                        
                        newBbl->b_first = jmp;
                        newBbl->b_last  = jmp;
                        INSTR_SET_FLAG(jmp, BBL_START);
                        INSTR_SET_FLAG(jmp, BBL_END);
                        jmp->i_Bblock = newBbl;
                        
                        /*  newBbl->b_next = bblClone->b_next; */
                        /*  newBbl->b_prev = bblClone; */
                        /*  if (bblClone->b_next) {      */
                        /*      bblClone->b_next->b_prev = newBbl; */
                        /*  } */
                        /*  bblClone->b_next = newBbl; */
                        
                        *clonedListTail = newBbl; 
                        newBbl->b_prev = prevClone; 
                        newBbl->b_clone = NULL;
                        clonedListTail = &(newBbl->b_next); 
                        prevClone = newBbl;
                        newBbl->b_type = BT_UNCOND;
                        
                        BBLAddEdge(bblClone, jmp->i_Bblock,
                            ET_FALSE)->e_edgeWeight = 
                            NODE_WEIGHT(niterator);

                        jmp->i_Bblock->b_type = BT_UNCOND;
                        BBLAddEdge(jmp->i_Bblock, NODE_TO(niterator),
                            ET_UNCOND)->e_edgeWeight =
                            NODE_WEIGHT(niterator);                        
                    } else {
                        BBLAddEdge(bblClone, NODE_TO(niterator),
                            NODE_TYPE(niterator))->e_edgeWeight =
                            NODE_WEIGHT(niterator);
                    }
                }

            }

        }  /* For all the edges in the block */
    }  /* For all the blocks in a function */

    // hook all of the instruction next/prev stuff set up correctly for
    // the cloned basic blocks
    bblClone = clonedList;
    succClone = bblClone->b_next;

    while (succClone) {
        bblClone->b_last->i_next = succClone->b_first;
        succClone->b_first->i_prev = bblClone->b_last;
        bblClone = succClone;
        succClone = bblClone->b_next;
    }

    // handle GOT UPMs specially...
    FOREACH_BBL_IN_FUN(fun, biterator) {
	Block *bblClone;

	bblClone = BBLOriginalToCloneFromList(biterator, clonedList);

	if (IsPICCall(biterator->b_last) && bblClone) {
	    Instruction *got_instr;

	    got_instr = GetGOTInstrForPICCall(biterator->b_last);
	    if (got_instr != NULL) {
		LinkedList *upms;
		Patch *upm;
		PatchInstrToInstrAux *patch;
		Iterator *itr;

		got_instr = GetGOTInstrForPICCall(bblClone->b_last);

		upms = UPMTableGetBySource(biterator->b_last->i_next);
		itr = ll_iterator(upms);
		while (itr->it_hasNext(itr)) {
		    upm = (Patch *) itr->it_next(itr);

		    patch = upm->p_patch;
		    if (patch->p_dest_instr == got_instr) {
			UPMChangeSourceFromInstrToInstr(upm,
			    patch->p_src_instr, bblClone->b_last->i_next);
			break;
		    }
		}
		free(itr);
		ll_free(upms);
	    } else {
		WARNING("Could not find GOT instr for cloned instruction");
	    }
	}
    }

    return clonedList;
}



/*
 *  OperandClone(op) --
 *
 *  Clones an operand and returns a pointer to the copy.  The operand is
 *  is filled in with the same displacements, registers, etc.. as the
 *  original.  This may not be correct, but will later be fixed when
 *  instruction patching is performed.
 *
 *  op: the operand to clone
 *  returns: a pointer to the clone operand
 */
Operand *OperandClone(Operand *op) {
    Operand *newop = OperandCreate();

    newop->a_flags = op->a_flags;
    newop->a_base_reg = op->a_base_reg;
    newop->a_index_reg = op->a_index_reg;
    newop->a_scale = op->a_scale;
    newop->a_displacement = op->a_displacement;

    OPERAND_REMOVE_FLAG(newop, OPERAND_NOCLONE_FLAGS);

    return newop;
}



/*
 *  InstrClone(instr) --
 *
 *  Clones an instruction and returns a pointer to the copy. Operands for
 *  the instruction are also cloned, and the instruction is given the same
 *  flags as the original. Note that these flags may not correct for the
 *  clone (namely, BBL_START, BBL_END may no longer be true. These flags
 *  are not used outside of flowgraph construction though.
 *
 *  instr: the instruction to be cloned
 *  returns: a pointer to the clone instruction
 */
Instruction *InstrClone(Instruction *instr) {
    Instruction *newInstr = InstrCreate();

    newInstr->i_class = instr->i_class;
    newInstr->i_type = instr->i_type;
    newInstr->i_prefixes = instr->i_prefixes;
    newInstr->i_flags = instr->i_flags;
    INSTR_REMOVE_FLAG(newInstr, INSTR_NOCLONE_FLAG);
    INSTR_SET_FLAG(newInstr, INSTR_CLONED);

    if (INSTR_HAS_FLAG(instr, DEST_FLAG)) {
	newInstr->i_dest = OperandClone(instr->i_dest);
    }
    if (INSTR_HAS_FLAG(instr, SOURCE1_FLAG)) {
	newInstr->i_src = OperandClone(instr->i_src);
    }
    if (INSTR_HAS_FLAG(instr, SOURCE2_FLAG)) {
	newInstr->i_src2 = OperandClone(instr->i_src2);
    }

    CloneInstrUPMsToInstr(instr, newInstr);

    newInstr->i_literal_data = instr->i_literal_data;

    return newInstr;
}



/*
 *  BBLOriginalToCloneFromList(original,list) --
 *
 *  Given an original basic block and a cloned list of blocks,
 *  BBLOriginalToClone returns a pointer to the bbl contained in the list
 *  that was derived from the original bbl. Returns NULL if no bbl is
 *  found.
 *
 *  original: the original basic block
 *  list: the list of clone basic blocks the check for the clone of original in
 *  returns: the clone of the specified original block or NULL if one does
 *  not exist in list
 */
Block *BBLOriginalToCloneFromList(Block *original, Block *list) {
    Block *biterator;

    biterator = list;
    while (biterator != NULL) {
	if (biterator->b_clone == original) {
	    return biterator;
	}
	biterator = biterator->b_next;
    }

    return NULL;
}



/*
 *  BBLOriginalToClone(original,clonedFun) --
 *
 *  Given an original basic block and a cloned function,
 *  BBLOriginalToClone will return a pointer to a bbl contained in the
 *  cloned function that was derived from the original bbl. Returns NULL
 *  if no basic block is found satisfying these conditions.
 *
 *  original: the original basic block
 *  clonedFun: the clone function to look for the clone of original in
 *  returns: the clone basic block of original if it exists, NULL otherwise
 */
Block *BBLOriginalToClone(Block *original, Function *clonedFun) {
    Block *biterator;

    FOREACH_BBL_IN_FUN(clonedFun, biterator) {
	if (biterator->b_clone == original) {
	    return biterator;
	}
    }

    return NULL;
}



/*
 *  CloneJumpTables(fun) --
 *
 *  Given a cloned function, CloneJumpTables will scan through and find
 *  indirect jumps through memory locations. These memory locations will
 *  be duplicated and changed if they point to a location inside the
 *  cloned function (e.g. most jump tables).
 *
 *  fun: the function to clone jump tables for
 */
void CloneJumpTables(Function *fun) {
    Instruction *iterator;
    Instruction *instrWithJTBase;
    Block *biterator;
    Block *blocks;
    Patch *upm;
    LinkedList *upms;
    Iterator *itr;
    JumpTable *jt;
    int i;
    int jumpTableCloneBase;      // offset to jump table clones in .rodata

    FOREACH_BBL_IN_FUN(fun, biterator) {

	BBL_REMOVE_FLAG(biterator, BBL_TEMP_FLAG);

	if (!BBL_HAS_FLAG(biterator, BBL_MARKED_FLAG)) {
	    continue;
	}

	FOREACH_INSTR_IN_BBL(biterator, iterator) {
	    if (!isJumpClass(iterator)) {
		continue;
	    }
	    if (!INSTR_HAS_FLAG(iterator, INDIRECT)) {
		continue;
	    }
	    if (biterator->b_type != BT_SWITCH) {
		continue;
	    }
	    CASSERT(!BBLHasHellSuccessor(biterator),
		"Failed to clone jump table correctly due to HELL edge.");
	    CASSERT(INSTR_HAS_FLAG(iterator, SOURCE1_FLAG),
		"Indirect jump without source operand.");
	    CASSERT(!INSTR_HAS_FLAG(iterator, SOURCE2_FLAG),
		"Indirect jump with second source operand.");
	    CASSERT(!INSTR_HAS_FLAG(iterator, DEST_FLAG),
		"Indirect jump with destination operand.");
	    CASSERT(iterator->i_Bblock->b_type == BT_SWITCH,
		"Cloning jump table on non-switch block.");

	    jumpTableCloneBase =
		SECTION_REF_RODATA->s_growth + SECTION_REF_RODATA->s_datasize;

	    /*
	     * We want to create a patch entry for the jump through the jump
	     * table. More precisely, we want to patch it to a cloned jump
	     * table.  A patch entry has already been constructed when the
	     * instruction was cloned.  We need to find the entry, remove it,
	     * and then add our new one.
	     */

	    instrWithJTBase = InstrLocateJTBaseInstr(iterator);

	    CASSERT(instrWithJTBase != NULL,
		"Failed to patch instruction jumping through cloned JT.");

	    upms = UPMTableGetByDest(instrWithJTBase);
	    itr = ll_iterator(upms);
	    while (itr->it_hasNext(itr)) {
		upm = (Patch *) itr->it_next(itr);

		if (upm->p_Apply == patchSectionAddressIntoInstr) {
		    PatchSectionToInstr *patch =
			(PatchSectionToInstr *) upm->p_patch;
		    patch->p_src_offset = jumpTableCloneBase;
		}
	    }
	    free(itr);
	    ll_free(upms);

	    /* We want to reroute the switch edges through the new jump table.
	     * First delete the existing edges, then add in the new ones */
	    BBLRemoveEdges(iterator->i_Bblock);

	    jt = GetJumpTableForInstr(iterator);
	    if (jt == NULL) {
		MALFUNCTION("Could not duplicate jump table. Instr %x: %s",
		    iterator->i_address, PrintInstruction(iterator));
	    }

	    /*
	     * JTBase = InstrLocateJTBaseAddress(iterator);
	     * JTSize = InstrLocateJTSize(iterator);
	     *
	     * if ((JTBase == 0) || (JTSize == 0)) {
	     * MALFUNCTION("Did not successfully duplicate jump table.");
	     * }
	     */

	    /*
	     * We look at all the addresses in the jump table and find where
	     * they are located in our cloned function.  This mapping is used
	     * to construct a new jump table for the instruction.
	     */

	    for (i = 0; i < jt->jt_size; i++) {
		UINT32 pointer;
		Instruction *target;
		Block *clone = NULL;
		Block *temp;

		pointer =
		    loadLE(&jt->jt_section->s_newdata[(jt->jt_base_offset)]);
		// pointer = AddressToData(JTBase + (ADDRESS_SIZE * i));
                // target = AddressToInstr(pointer);
		target = FinalAddrToInstr(pointer);

		CASSERT(target != NULL, "No JT target. Clone Failed.");

		FOREACH_BBL_IN_FUN(fun, blocks) {
		    if (BBL_HAS_FLAG(blocks, BBL_CLONE_SWITCH_RESOLVED_FLAG)) {
			continue;
		    }

		    temp = blocks->b_clone;
		    while ((temp) && (temp->b_clone != NULL)) {
			temp = temp->b_clone;
		    }

		    if (temp == target->i_Bblock) {
			clone = blocks;
			BBL_SET_FLAG(clone, BBL_TEMP_FLAG);
			break;
		    }

		}

		if (clone == NULL) {
		    FOREACH_BBL_IN_FUN(fun, blocks) {
			temp = blocks->b_clone;
			while ((temp) && (temp->b_clone != NULL)) {
			    temp = temp->b_clone;
			}
			if (temp == target->i_Bblock) {
			    clone = blocks;
			    BBL_SET_FLAG(clone, BBL_TEMP_FLAG);
			    break;
			}
		    }
		}

		CASSERT(clone != NULL,
		    "Could not find cloned basic block for jump table target.");

		AddPatchInstrAddressIntoSectionAddress(clone->b_first,
		    SECTION_REF_RODATA,
		    SECTION_REF_RODATA->s_growth +
		    SECTION_REF_RODATA->s_datasize);


		(BBLAddEdge(iterator->i_Bblock, clone,
			ET_SWITCH)->e_jumpTableSlot) = i;
		IncreaseSectionSize(SECTION_REF_RODATA, ADDRESS_SIZE);
	    }

	    AddJumpTable(SECTION_REF_RODATA, SECTION_REF_RODATA->s_growth,
		jt->jt_size);

	    FOREACH_BBL_IN_FUN(fun, blocks) {
		if (BBL_HAS_FLAG(blocks, BBL_TEMP_FLAG)) {
		    BBL_REMOVE_FLAG(blocks, BBL_TEMP_FLAG);
		    BBL_SET_FLAG(blocks, BBL_CLONE_SWITCH_RESOLVED_FLAG);
		}
	    }


	}				/* Foreach instruction */
    }					/* Foreach basic block */

    return;
}
