/*
 *  cp.c
 *
 *  Constant Propagator and Intel Instruction Evaluator
 *
 *  [1] The goal of the constant propagator is to find immediate values that are
 *  being moved in/out of registers/stack and replace their uses with constants.
 *  The instruction evaluator evaluates instructions with constant arguments.
 *
 *  [2] There has been an attempt to make constant propagation work over all of
 *  the general purpose registers, even the 8 and 16 bit versions.
 */
#include "global.h"
#include "cp.h"
#include "dominator.h"
#include "stack.analyze.h"
#include "utilities.h"
#include "x86-classes.h"
#include "x86-table-defs.h"
#include "x86-asm.h"
#include "x86-registers.h"
#include "opt.layout.h"
#include "flowgraph.h"
#include "sysdep.h"
#include "opt.h"
#include "liveness.h"
#include "opt.unreachable.h"
#include "upm.h"
#include "hash.h"

/*
 *  Macros to access the fields of a Value structure.
 */
#define REG_CONTENTS(bbl, reg)	(RegContents(bbl, reg))
#define REG_TYPE(bbl, reg)	(bbl->b_regshot[RegisterToIndex(reg)].v_type)
#define REG_FLAGS(bbl, reg)	(bbl->b_regshot[RegisterToIndex(reg)].v_flags)
#define STACK_CONTENTS(bbl, base)   (StackContents(bbl, base))
#define STACK_TYPE(bbl, base)       (StackType(bbl, base))
#define STACK_FLAGS(bbl, base)      (StackFlags(bbl, base))

#define INSTR_IMPLICIT_DEST(instr) \
    DIS386_IMPLICIT_DEST(InstrGetOpcodeEntry(instr))
#define INSTR_IMPLICIT_DEST1(instr) \
    DIS386_IMPLICIT_DEST(InstrGetOpcodeEntry(instr))
#define INSTR_IMPLICIT_DEST2(instr) \
    DIS386_IMPLICIT_DEST2(InstrGetOpcodeEntry(instr))
#define INSTR_IMPLICIT_SOURCE(instr) \
    DIS386_IMPLICIT_SOURCE(InstrGetOpcodeEntry(instr))
#define INSTR_IMPLICIT_SOURCE1(instr) \
    DIS386_IMPLICIT_SOURCE(InstrGetOpcodeEntry(instr))
#define INSTR_IMPLICIT_SOURCE2(instr) \
    DIS386_IMPLICIT_SOURCE2(InstrGetOpcodeEntry(instr))

/*
 *  Bitmasks for dealing with 8, 16 and 32 bit registers.
 */
#define REGISTER_8L_BITS (0x000000ff)
#define REGISTER_8H_BITS (0x0000ff00)
#define REGISTER_16_BITS (0x0000ffff)
#define REGISTER_32_BITS (0xffffffff)

/*
 *  We model the PSW in the instruction evaluator. PSW_OFF means the bit is
 *  not set.  PSW_SET means the bit is set.  PSW_UNKNOWN means the bit
 *  may or may not be set.  The index of a certain bit, say (1 << i),
 *  can be accessed by examining the integer psw[i].
 *  It will be one of these three values.
 */
#define PSW_OFF     0
#define PSW_SET     1
#define PSW_UNKNOWN 2

/*
 *  REPORT_CP_HEADER and REPORT_CP_TRAILER are two magic numbers
 *  that can be passed to ReportCPResults() to signify the start and end
 *  of constant propagation.  These cause various headers and
 *  summary statistics to be printed out.
 */
#define REPORT_CP_HEADER  -1565656
#define REPORT_CP_TRAILER -5645656

/*
 *  Performs a join for the slot at offset 'stack' at the top of 'bbl'. The
 *  value 'constantVal', which is of type 'type' is joined with whatever
 *  resides in the basic block. The 'flags' fields are also joined.
 *  If the 'first' flag is passed as true, then the fields are simply
 *  assigned without performing the semantics of a join.
 *  This can be used when merging the first predecessor.
 *  Finally, 'ptrOff' is the offset into the stack frame which this slot
 *  is aliased to. If it does not alias anything, 0 can safely be passed.
 */
static void MergeStack(Block *bbl, int stack, int constantVal, int type,
    int flags, bool first, int ptrOff);

static void MergeRegister(Block *bbl, int reg, int constantVal, int type,
    int flags, bool first, int ptrOff);

/*
 *  We use functions to access the field of the Value so that bounds
 *  checking on the arrays can be wrapped in. RegContents() and StackContents()
 *  return the value of the constant residing in 'bbl' at the supplied index.
 *  StackType() and REG_TYPE() return the type (SLOT_TYPE_UNKNOWN,
 *  SLOT_TYPE_UNDEFINED, SLOT_TYPE_CONSTANT) of the location. StackFlags() and
 *  REG_FLAGS() return the flags bitvector for the slot.
 */
/* Returns the contents of the register 'someReg' in 'bbl' */
static int RegContents(Block *bbl, int someReg);

/* Returns the contents of the runtime stack in 'bbl' at index 'idx' */
static int StackContents(Block *bbl, int idx);

/* Returns the slot type of the runtime stack in 'bbl' at index 'idx' */
static int StackType(Block *bbl, int idx);

/* Returns the flags associated with the slot at index 'idx'
 * in bbl's view of the stack */
static UINT32 StackFlags(Block *bbl, int idx);

static void MergePredecessors(Block *bbl);

/*
 *  Merges the predecessors, evaluates all instructions, and then returns true
 *  if something changed in the bbl's view of the stack/registers
 */
static bool PropagateForBBL(Block *bbl);

/* Initializes the model for the registers and runtime stack in 'bbl' */
static void InitializeStackAndRegs(Block *bbl);

/* Writes 'constantVal' into the stack slot at offset 'stack' */
static void AssignStack(Block *bbl, int stack, int constantVal, int type,
    UINT32 flags);

/* Assigns 'flags' to the stack location at offset 'stackLoc' in 'bbl'. */
static void AssignStackPtr(Block *bbl, int stackLoc, int flags, int stackPtr);

/* Assigns 'flags' to the register index 'reg' in 'bbl' */
static void AssignRegPtr(Block *BBL, int regIndex, int flags, int stackPtr);

/* Writes 'constantVal' into the register slot for register 'reg' */
static void AssignRegister(Block *bbl, int reg, int constantVal, int type,
    UINT32 flags);

/* Returns true if the register 'aReg' is undefined or unknown in the
 * instructions bbl */
static bool RegisterIsUndefinedOrUnknown(Instruction *instr, int aReg);

/*
 *  Returns true if the operand 'op' in instruction 'instr' is of type
 *  SLOT_TYPE_UNDEFINED or SLOT_TYPE_UNKNOWN
 */
static bool OperandIsUndefinedOrUnknown(Instruction *instr, Operand *op);

/* Sets all the stack slots to unknown in 'bbl' */
static void BBLStackBecomesUnknown(Block *bbl);

/* Sets all the register slots to unknown in 'bbl' */
static void BBLRegsBecomeUnknown(Block *bbl);

/* Sets all stack and register slots to unknown in 'bbl' */
static void BBLStackAndRegistersBecomeUnknown(Block *bbl);

/* Returns the slot flag corresponding to the supplied register */
static int RegisterToSlotFlag(int aReg);

/* Returns the register corresponding to the supplied slot flag */
static int FlagsToReg(int flags);

/* Updates the slot flags in the bbl for the supplied instruction */
static void UpdateFlags(Instruction *instr);

/*
 *  Propagate information about stack/registers from 'bbl's predecesors into
 *  bbl that can be gleaned from comparisons and conditional jumps
 */
static void BBLGetConditionalEqualityInformation(Block *bbl);

/* Evaluates the instruction and returns true if stack/regs were changed */
static Block * EvaluateInstruction(Instruction *instr);

/* Attempts to replace a register/stack use with a constant value */
static int ReplaceOperandWithConstant(Instruction *instr, Operand *op);

/* Attempts to replace a stack use with a register */
static int ReplaceOperandWithRegister(Instruction *instr, Operand *op);

/* Returns a copy to a pointer of the register snapshot
 * associated with the basic block */
static Value *BBLCopyRegshot(Block *bbl);

/* Returns a copy to a pointer of the runtime stack snapshot
 * associated with the basic block */
static Value *BBLCopyStackshot(Block *bbl);

/* Returns true if the contents of the register snapshot for 'bbl'
 * are the same as the  contents pointed to by the value slot array 'regs' */
static bool BBLRegContentsAreEqual(Block *bbl, Value *regs);

/* Returns true if the contents of the stack snapshot for 'bbl'
 * are the same as the contents pointed to by the value slot array 'stack' */
static bool BBLStackContentsAreEqual(Block *bbl, Value *stack);

/* Returns true if the two value slots are equal */
static bool ValueSlotsAreEqual(Value *x, Value *y);

/* Replaces the stack reference in 'op' by the register 'reg' */
static bool ReplaceStackWithRegister(Instruction *instr, Operand *op, int reg);

/* Replaces the stack reference in 'op' by the constant value 'val' */
static bool ReplaceStack(Instruction *instr, Operand *op, int val);

/* Replaces the register use by 'op' with the constant value 'val' */
static bool ReplaceReg(Instruction *instr, Operand *op, int val);

/* Reports summary information for the constant propagation phase */
static void ReportCPResults(char *name, int constants, int registers,
    int branches);

/* Creates a patch entry for the address that was propagated
 * into the instruction's operand 'op' */
static void CreateUPMForAddressPropagation(Instruction *instr, Operand *op,
    int val);

/*
 *  Global variables for constant propagation
 */
static int myBase;			// index into slot array for ebp
static int stackBase;			// index into slot array for esp
static int base;			// temporary index into slot array
static int numInstructionsEvaluated;	// number of instructions evaluated
static int numInstructionsNotEvaluated;	// number that could not be evaluated
static int roundNumber;			// number of CP rounds elapsed
static bool resolveConditionalJumps;	// set if to eliminate conditional jumps
static int conditionalsResolved;	// count of conditional jumps eliminated
static int psw[32];			// model of the program status word

/*
 *  Macros used by the instruction evaluator to pull out of evaluation when an
 *  address is encountered.
 */
#define NO_EVAL_IF_DEST_IS_ADDRESS() \
    if (destReg \
    && (REG_FLAGS(bbl, ((instr)->i_dest->a_base_reg)) & SLOT_ADDRESS_FLAG)) { \
	goto noEval; \
    } else if (destStack) { \
	base = myBase - ((instr)->i_dest->a_displacement) / ADDRESS_SIZE; \
	if (STACK_FLAGS(bbl, base) & SLOT_ADDRESS_FLAG) \
	    goto noEval; \
    } else if (destImm && destAddr) \
	goto noEval;

#define NO_EVAL_IF_SOURCE1_IS_ADDRESS() \
    if (src1Reg \
    && (REG_FLAGS(bbl, ((instr)->i_src->a_base_reg)) & SLOT_ADDRESS_FLAG)) { \
	goto noEval; \
    } else if (src1Stack) { \
	base = myBase - ((instr)->i_src->a_displacement) / ADDRESS_SIZE; \
	if (STACK_FLAGS(bbl, base) & SLOT_ADDRESS_FLAG) \
	    goto noEval; \
    } else if (src1Imm && src1Addr) \
	goto noEval;

#define NO_EVAL_IF_SOURCE2_IS_ADDRESS() \
    if (src2Reg \
    && (REG_FLAGS(bbl, ((instr)->i_src2->a_base_reg)) & SLOT_ADDRESS_FLAG)) { \
	goto noEval; \
    } else if (src2Stack) { \
	base = myBase - ((instr)->i_src2->a_displacement) / ADDRESS_SIZE; \
	if (STACK_FLAGS(bbl, base) & SLOT_ADDRESS_FLAG) \
	    goto noEval; \
    } else if (src2Imm && src2Addr) \
	goto noEval;



/*
 *  MergeStack(bbl,stack,constantVal,type,flags,first,ptrOff) --
 *
 *  Performs a merge operation on a stack value slot.  The semantics
 *  of merge are that (undefined,anything) = (anything), (unknown,anything) =
 *  unknown, (constant1, constant2) = constant1 if constant1 = constant2, or
 *  unknown if constant1 != constant2.
 */
static void MergeStack(Block *bbl, int stack, int constantVal, int type,
    int flags, bool first, int ptrOff) {

    Function *fun = bbl->b_fun;

    /*
     * If this is the first merge taking place, then we don't need to do a join.
     * The constants, types and flags can simply be assigned to this block.
     */

    if (first == true) {
	AssignStack(bbl, stack, constantVal, type, flags);

	if ((stack >= 0) && (stack <
		(fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE)
	    && (bbl->b_stackshot != NULL)) {
		bbl->b_stackshot[stack].v_stackOffset = ptrOff;
	}

	return;
    }

    if (stack < 0) {
	return;
    }
    CASSERT(stack <
	((fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE),
	"[MergeStack] Stack index larger than allocated stack region!\n");

    if (bbl->b_stackshot == NULL) {
	return;
    }

    if (type == SLOT_TYPE_UNDEFINED) {
	bbl->b_stackshot[stack].v_flags &= (flags | SLOT_ADDRESS_FLAG);
	return;
    } else if (type == SLOT_TYPE_UNKNOWN) {
	bbl->b_stackshot[stack].v_type = SLOT_TYPE_UNKNOWN;
	bbl->b_stackshot[stack].v_flags &= flags & ~SLOT_ADDRESS_FLAG;
    } else if (bbl->b_stackshot[stack].v_type == SLOT_TYPE_UNKNOWN) {
	bbl->b_stackshot[stack].v_type = SLOT_TYPE_UNKNOWN;
	bbl->b_stackshot[stack].v_flags &= flags & ~SLOT_ADDRESS_FLAG;
    }

    if (bbl->b_stackshot[stack].v_type == SLOT_TYPE_UNDEFINED) {
	bbl->b_stackshot[stack].v_type = type;
	bbl->b_stackshot[stack].v_value = constantVal;
	bbl->b_stackshot[stack].v_flags = flags;
    } else if (bbl->b_stackshot[stack].v_type == SLOT_TYPE_CONSTANT) {
	if (bbl->b_stackshot[stack].v_value != constantVal) {
	    bbl->b_stackshot[stack].v_type = SLOT_TYPE_UNKNOWN;
	    bbl->b_stackshot[stack].v_flags = 0;
	} else {
	    bbl->b_stackshot[stack].v_flags &= flags;
	}
    }

    if (SLOT_HAS_FLAG_DIRECT(bbl->b_stackshot[stack], SLOT_STACK_FLAG)) {
	if (bbl->b_stackshot[stack].v_stackOffset != ptrOff) {
	    SLOT_REMOVE_FLAG_DIRECT(bbl->b_stackshot[stack], SLOT_STACK_FLAG);
	    bbl->b_stackshot[stack].v_stackOffset = 0;
	}
    }

    return;
}



/*
 *  MergeRegister(bbl,reg,constantVal,type,flags,first,ptrOff) --
 *
 *  Performs a merge of a register value with the current value.
 *  The semantics are equivalent to those above in MergeStack.
 */
/* This function is fragile. Needs revamp in the future
 * -Somu
 */

static void MergeRegister(Block *bbl, int reg, int constantVal, int type,
    int flags, bool first, int ptrOff) {
    int regIndex = reg;

    if (first == true) {
        AssignRegister(bbl, IndexToRegister(reg), constantVal, type, flags);
        bbl->b_regshot[regIndex].v_stackOffset = ptrOff;
        return;
    }

    if (type == SLOT_TYPE_UNDEFINED) {
        bbl->b_regshot[regIndex].v_flags &= flags | SLOT_ADDRESS_FLAG;
        return;
    }

    if (type == SLOT_TYPE_UNKNOWN) {
        bbl->b_regshot[regIndex].v_type = SLOT_TYPE_UNKNOWN;
        bbl->b_regshot[regIndex].v_flags &= flags & ~SLOT_ADDRESS_FLAG;
    } else if (bbl->b_regshot[regIndex].v_type == SLOT_TYPE_UNKNOWN) {
        bbl->b_regshot[regIndex].v_type = SLOT_TYPE_UNKNOWN;
        bbl->b_regshot[regIndex].v_flags &= flags & ~SLOT_ADDRESS_FLAG;
    } else if (bbl->b_regshot[regIndex].v_type == SLOT_TYPE_UNDEFINED) {
        bbl->b_regshot[regIndex].v_value = constantVal;
        bbl->b_regshot[regIndex].v_type = type;
        bbl->b_regshot[regIndex].v_flags = flags;
    } else if (bbl->b_regshot[regIndex].v_type == SLOT_TYPE_CONSTANT) {
        if (bbl->b_regshot[regIndex].v_value != constantVal &&
            type == SLOT_TYPE_CONSTANT) {
            bbl->b_regshot[regIndex].v_type = SLOT_TYPE_UNKNOWN;
            bbl->b_regshot[regIndex].v_flags = 0;
        } else {
            bbl->b_regshot[regIndex].v_flags &= flags;
        }
    } 

    if (SLOT_HAS_FLAG_DIRECT(bbl->b_regshot[regIndex], SLOT_STACK_FLAG)) {
        if (bbl->b_regshot[regIndex].v_stackOffset != ptrOff) {
            SLOT_REMOVE_FLAG_DIRECT(bbl->b_regshot[regIndex], SLOT_STACK_FLAG);
            bbl->b_regshot[regIndex].v_stackOffset = 0;
        }
    }

    return;
}



/*
 *  MergePredecessors(bbl) --
 *
 *  Sets 'bbl's register and stack snapshots to the result attained from
 *  merging its predecessors.
 */
static void MergePredecessors(Block *bbl) {
    EdgeNode *predecessors;
    bool first = true;

    if (!bbl) {
        return;
    }
    
    if (BBLEdgeExists(bbl, bbl) &&
        !EDGE_HAS_FLAG(BBLGetEdgeToBBL(bbl, bbl),
            EDGE_NO_FLOW_FLAG)) {
        first = false;
    }

    /*
     * Take the information about all our predecessors and use it
     * to propagate constants through out the basic block.
     */

    FOREACH_PRED_IN_BBL(bbl, predecessors) {
	Block *pred = NODE_FROM(predecessors);
	Function *callee = NULL;
	int sz = (bbl->b_fun->f_stackFrameSize +
	    bbl->b_fun->f_callerFrameSize) / ADDRESS_SIZE;
	int mybase = bbl->b_fun->f_callerFrameSize / ADDRESS_SIZE;
	int predsz = (pred->b_fun->f_stackFrameSize +
	    pred->b_fun->f_callerFrameSize) / ADDRESS_SIZE;
	int predbase = pred->b_fun->f_callerFrameSize / ADDRESS_SIZE;
	int i, p, type, value, flags, stackDirect;
	int unknownIndex, callOffset;
	bool directAssign = first;
    Edge *from = NULL;
    
	if (pred == bbl) {
	    continue;
	}

    from = BBLGetEdgeToBBL(pred, bbl);

    /*  if (EDGE_HAS_FLAG(from, EDGE_NO_FLOW_FLAG) && */
    /*         !(NODE_TYPE(predecessors) == ET_HELL)) { */
    /*         continue; */
    /*     } */

	/*
	 * ET_FUNDIRECT edges are discarded in constant propagation, because
	 * control is routed through the entry block to the function, and
	 * we can propagate the constants through that instead. Note that
	 * this may lead to a less precise analysis; e.g, the function entry
	 * block may be less precise than simply using the direct
	 * predecessor which makes the call.
	 */

	if (NODE_TYPE(predecessors) == ET_FUNDIRECT) {
	    continue;
	}

	if (NODE_TYPE(predecessors) == ET_RETURN) {
	    first = false;
	    callee = pred->b_fun;
	    callOffset = InstrBPSPDiff(bbl->b_first->i_prev);

	    if ((callee->f_trashDepth == -1) || (callOffset == STACK_UNKNOWN)) {
		BBLStackBecomesUnknown(bbl);
	    } else {
		/*
		 * Otherwise, we can see how far down into the stack the callee
		 * writes, and set only the slots above there to UNKNOWN.
		 */
		unknownIndex = (callOffset / ADDRESS_SIZE) + 2 -
		    (callee->f_trashDepth / ADDRESS_SIZE) +
		    (bbl->b_fun->f_callerFrameSize / ADDRESS_SIZE);

		for (i = 0; i <= sz; i++) {
		    if (i >= unknownIndex) {
			AssignStack(bbl, i, 0, SLOT_TYPE_UNKNOWN, 0);
		    }
		}
	    }
	    continue;
	}

	if (!BBL_HAS_FLAG(pred, BBL_REGSHOT_FLAG)) {
	    BBLRegsBecomeUnknown(bbl);
	}
    
	if (!BBL_HAS_FLAG(pred, BBL_STACKSHOT_FLAG)) {
	    BBLStackBecomesUnknown(bbl);
	}

	/*
	 * All blocks with incoming hell edges are initialized to UNKNOWN
	 * so we don't need to worry about merging predecessors from HELL.
	 */

	if ((NODE_TYPE(predecessors) == ET_HELL)
	    || (NODE_TYPE(predecessors) == ET_HELLMAYBE)) {
	    if (first) {
		first = false;
	    }
	    BBLStackAndRegistersBecomeUnknown(bbl);
	    continue;
	}

	if (first) {
	    first = false;
	}

	/*
	 * First we'll handle propagating the registers to this basic
	 * block. Special cases are functionlink edge (ET_FUNLINK) and
	 * return edges (ET_RETURN). Since control does not flow along a
	 * function-link edge, we only want to propagate registers through
	 * if they were definitely unchanged by the callee. Similarly, for
	 * function return edges, we going to propagate registers that
	 * the callee might have changed. This will definitely allow us to
	 * get any return value that is passed in a register, and will
	 * also allow registers that are not used by the callee but are
	 * used by the caller to pass through the function link edge.
	 */

	if (NODE_TYPE(predecessors) == ET_FUNLINK) {
	    Edge *direct = BBLGetFunDirectEdge(pred);
	    Edge *ret = BBLGetReturnPredecessor(bbl);
	    Block *returnPred;
	    REGISTER_BITVECTOR changed = ALL_REGISTERS;
	    if (direct) {
		changed = direct->e_to->b_fun->f_regsDefined;
	    }

	    for (p = 0; p < NUMBER_OF_GP_REGISTERS; p++) {
		if (!((1 << p) & changed)) {
		    MergeRegister(bbl, p, pred->b_regshot[p].v_value,
			pred->b_regshot[p].v_type, pred->b_regshot[p].v_flags,
			directAssign, pred->b_regshot[p].v_stackOffset);
		}
	    }

	    if (ret != NULL) {
		returnPred = ret->e_from;

		CASSERT(direct != NULL, "[MergePredecessors]"
		    " ET_FUNDIRECT edge not found for predecessor block"
		    " when bbl has a return edge coming in.\n");

		for (p = 0; p < NUMBER_OF_GP_REGISTERS; p++) {
		    if (((1 << p) & changed)) {
			MergeRegister(bbl, p, returnPred->b_regshot[p].v_value,
			    returnPred->b_regshot[p].v_type,
			    returnPred->b_regshot[p].v_flags, directAssign,
			    returnPred->b_regshot[p].v_stackOffset);
		    }
		}
	    }
	} else {
	    for (p = 0; p < NUMBER_OF_GP_REGISTERS; p++) {
		MergeRegister(bbl, p, pred->b_regshot[p].v_value,
		    pred->b_regshot[p].v_type, pred->b_regshot[p].v_flags,
		    directAssign, pred->b_regshot[p].v_stackOffset);
	    }
	}

	/*
	 * We have merged the registers for this predecessor; now comes
	 * the stack. The interesting (read--special) cases are function
	 * call and return edges, where we merge different parts of each
	 * function's stack frame.
	 */

	if (!BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	    continue;
	}

	if (NODE_TYPE(predecessors) == ET_FUNCALL) {
	    callOffset = InstrBPSPDiff(pred->b_last);
	    for (p = mybase - 2; p >= 0; p--) {
		int predIndex;

		if (!BBL_HAS_FLAG(pred, BBL_STACKSHOT_FLAG)) {
		    predIndex = -1;
		} else if (callOffset != STACK_UNKNOWN) {
		    predIndex = predbase +
			(callOffset / ADDRESS_SIZE) - (mybase - 2 - p);
		} else {
		    predIndex = -1;
		}

		CASSERT(callOffset != STACK_UNDEFINED, "[MergePredecessors]"
		    " Instr offset into runtime stack is undefined.\n");

		if (predIndex < 0) {
		    value = 0;
		    type = SLOT_TYPE_UNKNOWN;
		    flags = 0;
		    stackDirect = 0;
		} else if (predIndex > predsz) {
		    value = 0;
		    type = SLOT_TYPE_UNKNOWN;
		    flags = 0;
		    stackDirect = 0;
		} else {
		    value = pred->b_stackshot[predIndex].v_value;
		    type = pred->b_stackshot[predIndex].v_type;
		    flags = pred->b_stackshot[predIndex].v_flags;
		    stackDirect = pred->b_stackshot[predIndex].v_stackOffset;
		}

		MergeStack(bbl, p, value, type, flags, directAssign,
		    stackDirect);
	    }

	/*
	 * For RETURN edges, merge the bottom of the callee's stack
	 * frame with our stack frame. This should account for any
	 * changes that the callee may have made to our frame.
	 */

	} else if (NODE_TYPE(predecessors) == ET_RETURN) {

	    for (p = mybase; p < sz; p++) {
		int predIndex = predbase - (p - mybase);

		CASSERT(isCallClass(bbl->b_first->i_prev),
		    "[MergePredecessors] Error merging ET_RETURN edge.\n");

		if (!BBL_HAS_FLAG(pred, BBL_STACKSHOT_FLAG)) {
		    type = SLOT_TYPE_UNKNOWN;
		    value = flags = stackDirect = 0;
		} else {
		    type = pred->b_stackshot[predIndex].v_type;
		    value = pred->b_stackshot[predIndex].v_value;
		    flags = pred->b_stackshot[predIndex].v_flags;
		    stackDirect = pred->b_stackshot[predIndex].v_stackOffset;

		}

		CASSERT(predIndex >= 0, "[MergePredecessors] predIndex < 0.\n");
		CASSERT(predIndex <= predsz,
		    "[MergePredecessors] predIndex > predsz.\n");

		MergeStack(bbl, p, value, type, flags, directAssign,
		    stackDirect);
	    }


	/*
	 * Otherwise the edge is intra-procedural, and we can simply
	 * transfer the contents of the stack slots into this bbl.
	 */

	} else {
	    for (p = 0; p < sz; p++) {
		if ((!BBL_HAS_FLAG(pred, BBL_STACKSHOT_FLAG)) || (p > predsz)) {
		    type = SLOT_TYPE_UNKNOWN;
		    value = flags = stackDirect = 0;
		} else {
		    type = pred->b_stackshot[p].v_type;
		    value = pred->b_stackshot[p].v_value;
		    flags = pred->b_stackshot[p].v_flags;
		    stackDirect = pred->b_stackshot[p].v_stackOffset;
		}

		MergeStack(bbl, p, value, type, flags, directAssign,
		    stackDirect);
	    }
	}

    }					/* end iteration over preds in bbl */
    return;
}



/*
 *  AssignStack(bbl,stack,constantVal,type,flags) --
 *
 *  Assigns the value 'constantVal' to the stack location at 'stack'.
 */
static void AssignStack(Block *bbl, int stack, int constantVal, int type,
    UINT32 flags) {
    Function *fun = bbl->b_fun;

    if (bbl->b_stackshot == NULL) {
	return;
    }

    if (stack < 0) {
	return;
    }
    if (stack >=
	((fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE)) {
	return;
    }

    bbl->b_stackshot[stack].v_type = type;
    bbl->b_stackshot[stack].v_value = constantVal;
    bbl->b_stackshot[stack].v_flags = flags;

    return;
}



/*
 *  AssignRegister(bbl,reg,constantVal,type,flags) --
 *
 *  Contains the semantics for an assignment to a runtime register.
 */
static void AssignRegister(Block *bbl, int reg, int constantVal, int type,
    UINT32 flags) {
    int regIndex = RegisterToIndex(reg);
    int oldVal = bbl->b_regshot[regIndex].v_value;
    int newVal = constantVal;
    int combinedVal = oldVal;

    CASSERT(regIndex <= NUMBER_OF_GP_REGISTERS,
	"[AssignRegister] Invalid register index returned.\n");

    CASSERT(regIndex >= 0,
	"[AssignRegister] Invalid register index returned.\n");

    if (regIndex < 0) {
	return;
    }

    if ((reg & (reg_8h))) {
	if (bbl->b_regshot[regIndex].v_type == SLOT_TYPE_UNKNOWN) {
	    type = SLOT_TYPE_UNKNOWN;
	}
	newVal = newVal << 8;
	combinedVal = combinedVal & (~REGISTER_8H_BITS);
    } else if ((reg & (reg_8l))) {
	if (bbl->b_regshot[regIndex].v_type == SLOT_TYPE_UNKNOWN) {
	    type = SLOT_TYPE_UNKNOWN;
	}
	combinedVal = combinedVal & (~REGISTER_8L_BITS);
    } else if ((reg & (reg_16))) {
	if (bbl->b_regshot[regIndex].v_type == SLOT_TYPE_UNKNOWN) {
	    type = SLOT_TYPE_UNKNOWN;
	}
	combinedVal = combinedVal & (~REGISTER_16_BITS);
    } else {
	combinedVal = combinedVal & (~REGISTER_32_BITS);
    }

    combinedVal = combinedVal + newVal;

    bbl->b_regshot[regIndex].v_value = combinedVal;
    bbl->b_regshot[regIndex].v_type = type;
    bbl->b_regshot[regIndex].v_flags = flags;

    return;
}



/*
 *  RegContents(bbl,someReg) --
 *
 *  Returns the contents of the register 'someReg' as seen by 'bbl'.
 */
static int RegContents(Block *bbl, int someReg) {
    int regIndex = RegisterToIndex(someReg);
    int oldVal = bbl->b_regshot[regIndex].v_value;
    int newVal = oldVal;

    CASSERT(regIndex <= NUMBER_OF_GP_REGISTERS,
	"[AssignRegister] Invalid register index returned.\n");

    if ((someReg & (reg_8h))) {
	newVal = newVal & (REGISTER_8H_BITS);
    } else if ((someReg & (reg_8l))) {
	newVal = newVal & (REGISTER_8L_BITS);
    } else if ((someReg & (reg_16))) {
	newVal = newVal & (REGISTER_16_BITS);
    } else {
	newVal = newVal & (REGISTER_32_BITS);
    }

    return newVal;
}



/*
 *  StackContents(bbl,idx) --
 *
 *  Returns the contents of the runtime stack as seen by 'bbl' at the slot
 *  indexed by 'idx'.
 */
static int StackContents(Block *bbl, int idx) {
    Function *fun = bbl->b_fun;
    int size = (fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE;

    if (idx >= size) {
	return 0;
    }
    if (idx < 0) {
	return 0;
    }
    return bbl->b_stackshot[idx].v_value;
}



/*
 *  StackType(bbl,idx) --
 *
 *  Returns the slot type of the runtime stack as seen by 'bbl' at the
 *  slot indexed by 'idx'
 */
static int StackType(Block *bbl, int idx) {
    Function *fun = bbl->b_fun;
    int size = (fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE;

    if (idx >= size) {
	return SLOT_TYPE_UNKNOWN;
    }
    if (idx < 0) {
	return SLOT_TYPE_UNKNOWN;
    }
    return bbl->b_stackshot[idx].v_type;
}



/*
 *  StackFlags(bbl,idx) --
 *
 *  Returns the slot flags of the runtime stack as seen by 'bbl' at the
 *  slot indexed by 'idx'.
 */
static UINT32 StackFlags(Block *bbl, int idx) {
    Function *fun = bbl->b_fun;
    int size = (fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE;

    if (idx >= size) {
	return 0;
    }
    if (idx < 0) {
	return 0;
    }
    return bbl->b_stackshot[idx].v_flags;
}



/*
 *  AllocateSnapshots() --
 *
 *  Allocates space for stack and register snapshots within all basic
 *  blocks. If space is successfully allocated for the stack, the flag
 *  BBL_STACKSHOT_FLAG will be set in the basic block. BBL_REGSHOT_FLAG
 *  will be set for basic blocks in which a register snapshot was
 *  successfully set up.
 */
void AllocateSnapshots() {
    Function *fun;
    Block *bbl;
    int totalSize;

    UpdateFunctions();
    ComputeFunctionAndCallerSizes();
    MarkFunsDefiningFramePointer();

    /*
     * First, free memory that is used by any existing snapshots.
     */

    FOREACH_FUNCTION(fun) {
	FOREACH_BBL_IN_FUN(fun, bbl) {
	    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
		free(bbl->b_stackshot);
		BBL_REMOVE_FLAG(bbl, BBL_STACKSHOT_FLAG);
	    }
	    if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
		free(bbl->b_regshot);
		BBL_REMOVE_FLAG(bbl, BBL_REGSHOT_FLAG);
	    }
	}
	FUNCTION_REMOVE_FLAG(fun, FUNCTION_SNAPSHOT_FLAG);
    }

    FOREACH_FUNCTION(fun) {
	bool fpDef =
	    FUNCTION_HAS_FLAG(fun, FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG);
	if (((fun->f_stackFrameSize > MAXIMUM_AR_SIZE)
		|| (fun->f_stackFrameSize <= 0)) && !fpDef) {
	    fun->f_stackFrameSize = UNKNOWN_STACK_SIZE;
	    FOREACH_BBL_IN_FUN(fun, bbl) {
		if ((bbl->b_type == BT_CALL) && BBLGetFunDirectEdge(bbl)) {
		    Function *callee = BBLGetFunDirectEdge(bbl)->e_to->b_fun;
		    if (callee->f_callerFrameSize < fun->f_stackFrameSize) {
			callee->f_callerFrameSize = fun->f_stackFrameSize;
		    }
		}
	    }
	}
    }

    FOREACH_FUNCTION(fun) {
	if (fun->f_callerFrameSize > MAXIMUM_AR_SIZE) {
	    fun->f_callerFrameSize = UNKNOWN_STACK_SIZE;
	}
    }

    FOREACH_FUNCTION(fun) {
	bool fpDef =
	    FUNCTION_HAS_FLAG(fun, FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG);
	totalSize = fun->f_stackFrameSize + fun->f_callerFrameSize;

	if ((fun->f_stackFrameSize <= 0) && !fpDef) {
	    fun->f_stackFrameSize = UNKNOWN_STACK_SIZE;
	}

	if ((fun->f_stackFrameSize > 0) && (totalSize <= (MAXIMUM_AR_SIZE * 2))
	    && !fpDef) {
	    FOREACH_BBL_IN_FUN(fun, bbl) {
		bbl->b_stackshot = NULL;
		BBLAllocateStackshot(bbl);
	    }
	    BBLAllocateStackshot(fun->f_entry);
	    BBLAllocateStackshot(fun->f_exit);
	    FUNCTION_SET_FLAG(fun, FUNCTION_SNAPSHOT_FLAG);
	} else {
	    fun->f_stackFrameSize = 0;
	    fun->f_callerFrameSize = 0;
	}

	FOREACH_BBL_IN_FUN(fun, bbl) {
	    BBLAllocateRegshot(bbl);
	}
	BBLAllocateRegshot(fun->f_entry);
	BBLAllocateRegshot(fun->f_exit);
    }

    BBLAllocateRegshot(HELL);

    return;
}

void DeallocateSnapShots() {
    Function *fun = 0;
    Block *bbl;
    
    FOREACH_FUN (fun) {
        FOREACH_BBL_IN_FUN(fun, bbl) {
            if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
                free(bbl->b_stackshot);
                BBL_REMOVE_FLAG (bbl, BBL_STACKSHOT_FLAG);
            }
            if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
                free(bbl->b_regshot);
                BBL_REMOVE_FLAG (bbl, BBL_REGSHOT_FLAG);                
            }
        }
        
        if (BBL_HAS_FLAG(fun->f_entry,
                BBL_REGSHOT_FLAG)) {
            free(fun->f_entry->b_regshot);
            BBL_REMOVE_FLAG(fun->f_entry,
                BBL_REGSHOT_FLAG);            
        }

        if (BBL_HAS_FLAG(fun->f_entry,
                BBL_STACKSHOT_FLAG)) {
        free(fun->f_entry->b_stackshot);
        BBL_REMOVE_FLAG(fun->f_entry,
            BBL_STACKSHOT_FLAG);                                
        }

        if (BBL_HAS_FLAG(fun->f_exit,
                BBL_REGSHOT_FLAG)) {
            free(fun->f_exit->b_regshot);
            BBL_REMOVE_FLAG(fun->f_exit,
                BBL_REGSHOT_FLAG);
        }

        if (BBL_HAS_FLAG(fun->f_exit,
                BBL_STACKSHOT_FLAG)) {
            free(fun->f_exit->b_stackshot);
            BBL_REMOVE_FLAG(fun->f_exit,
                BBL_STACKSHOT_FLAG);            
        }
    }
}

static long long GetInCount(Block *bbl) {
    long long retVal = 0;
    EdgeNode *pred = 0;

    FOREACH_PRED_IN_BBL(bbl, pred) {
        if (NODE_TYPE(pred) == ET_FUNLINK ||
            NODE_TYPE(pred) == ET_ENTRY) {
            continue;
        }
        retVal += NODE_WEIGHT(pred);
    }    

    return retVal;
}

static long long GetOutCount(Block *bbl) {
    long long retVal = 0;
    EdgeNode *succ = 0;

    FOREACH_SUCC_IN_BBL(bbl, succ) {
        if (NODE_TYPE(succ) == ET_FUNLINK ||
            NODE_TYPE(succ) == ET_FUNCALL) {
            continue;
        }        
        retVal += NODE_WEIGHT(succ);
    }    

    return retVal;
}


/*
 * AdjustWeightsForFunction() - reweights the
 * basic blocks for a function.  
 */

static void AdjustWeightsForFunction(Function *fun) {
    bool change = true, visit = true;;
    EdgeNode *succ = NULL, *pred = NULL;
    Block *biterator = NULL;
    Edge *tedge = NULL, *fedge = NULL, *tempEdge = NULL;
    double fratio = 0.0, tratio = 0.0;
    long long inCount = 0, outCount = 0;

    TreeNode *tnode = 0;

    tnode = FunCreatePredominatorTree(fun);
    
    FOREACH_BBL_IN_FUN (fun, biterator) {
        if (!BBLNumberOfPredecessors(biterator) &&
            biterator != fun->f_first) {
            FOREACH_SUCC_IN_BBL(biterator, succ) {
                NODE_WEIGHT(succ) = 0;
            }
        }
    }

    FOREACH_BBL_IN_FUN (fun, biterator) {
        BBL_REMOVE_FLAG(biterator, BBL_TEMP_FLAG);
    }
        
    while (change) {
        change = false;

        FOREACH_BBL_IN_FUN(fun, biterator) {

            if (!BBLNumberOfPredecessors(biterator) ||
                biterator == fun->f_first) {
                BBL_SET_FLAG(biterator, BBL_TEMP_FLAG);
                continue;
            }

            visit = true;
            /* Visit a bbl only if all its predecessors have been visited */
            FOREACH_PRED_IN_BBL(biterator, pred) {
                if (BBLDominatesBBL(tnode, biterator, NODE_FROM(pred))) {
                    continue;
                }

                if (!BBL_HAS_FLAG(NODE_FROM(pred), BBL_TEMP_FLAG)) {
                    visit = false;
                }
            }

            if (!visit) {
                continue;
            }
            
            inCount = GetInCount(biterator);
            outCount = GetOutCount(biterator);

            if (inCount != outCount) {
                switch (biterator->b_type) {
                case BT_NORMAL:
                    tempEdge = BBLGetEdgeTyped(biterator, ET_NORMAL);
                    tempEdge->e_edgeWeight = inCount;
                    BBL_REMOVE_FLAG(tempEdge->e_to, BBL_TEMP_FLAG);
                    change = true;
                    break;                    
                case BT_UNCOND:
                    tempEdge = BBLGetEdgeTyped(biterator, ET_UNCOND);
                    if (tempEdge) {
                        tempEdge->e_edgeWeight = inCount;
                        BBL_REMOVE_FLAG(tempEdge->e_to, BBL_TEMP_FLAG);
                        change = true;
                    }
                    break;
                case BT_RETURN:
                    tempEdge = BBLGetEdgeTyped(biterator, ET_EXIT);
                    tempEdge->e_edgeWeight = inCount;
                    BBL_REMOVE_FLAG(tempEdge->e_to, BBL_TEMP_FLAG);
                    change = true;
                    break;                    
                case BT_CALL:
                    FOREACH_SUCC_IN_BBL(biterator,  succ) {
                        NODE_WEIGHT(succ) = inCount;
                    }
                    change = true;
                    break;
                case BT_BRANCH:
                    tedge = BBLGetTrueEdge(biterator);
                    fedge = BBLGetFalseEdge(biterator);
                    if (!tedge && !fedge) {
                        tempEdge = BBLGetEdgeTyped(biterator, ET_UNCOND);
                        tempEdge->e_edgeWeight = inCount;
                        break;                        
                    }

                    if (!tedge || !fedge) {
                        printf("[AWF] Edge missing in bbl %d at %s\n",
                            biterator->b_type, fun->f_name);
                        break;
                    }

                    tedge = BBLGetTrueEdge(biterator);
                    fedge = BBLGetFalseEdge(biterator);

                    tratio = (tedge->e_edgeWeight * 1.0) /
                        (tedge->e_edgeWeight + fedge->e_edgeWeight);

                    fratio = (fedge->e_edgeWeight * 1.0) /
                        (tedge->e_edgeWeight + fedge->e_edgeWeight);
                    
                    tedge = BBLGetTrueEdge(biterator);
                    fedge = BBLGetFalseEdge(biterator);

                    tedge->e_edgeWeight = (unsigned long)(tratio * inCount);

                    fedge->e_edgeWeight = (unsigned long)(fratio * inCount);

                    BBL_REMOVE_FLAG(fedge->e_to, BBL_TEMP_FLAG);
                    BBL_REMOVE_FLAG(tedge->e_to, BBL_TEMP_FLAG);
                    
                    while (inCount >
                        (tedge->e_edgeWeight + fedge->e_edgeWeight)) {
                        tedge->e_edgeWeight++;
                    }
                    
                    while (inCount <
                        (tedge->e_edgeWeight + fedge->e_edgeWeight)) {
                        tedge->e_edgeWeight--;
                    }
                    
                    break;
                default:
                    change = true;
                    break;
                }
            } else if (inCount < outCount) {
                printf("[AWF] incount lesser than outcount for bbl %d in %s\n",
                    biterator->b_id, fun->f_name);
            }
            BBL_SET_FLAG(biterator, BBL_TEMP_FLAG);
        }
    }
}
    
/*
 *  PropagateConstants() --
 *
 *  Invokes the constant propagator. The propagator will first construct
 *  snapshots of the stack and registers at the basic block level, then
 *  propagate the information around the control flow graph.
 */
void PropagateConstants() {
    Block *bbl;
    Function *fun;
    Instruction *instr;
    bool change = true;
    static int timesCalled = 1;
    int numSimplified, forwards;
    EdgeNode *successors = NULL;
    int i = 0;
    
    REPORT(RP_MAJOR, "Propagating Constants round :%d", timesCalled++);
    MarkWellBehavedFunctions(true);
    
    REPORT(RP_DETAIL, "[%s] Stacks analyzed", __func__);

    CalculateSPBPDifference();
    REPORT(RP_DETAIL,
        "[%s] Calculated (%%ESP-%%EBP) for all blocks", __func__);

    StoreInstrBPSPDiff();
    REPORT(RP_DETAIL,
        "[%s] Calculated (%%ESP-%%EBP) calculated for all instructions",
        __func__);

    MarkFunsAliasingStackPointer();

    AllocateSnapshots();
    REPORT(RP_DETAIL, "[%s] Snapshots allocated", __func__);

    ComputeFunctionTrashDepths();
    REPORT(RP_DETAIL, "[%s] Computed trash depths for functions", __func__);

    ComputeRegsThatFunctionsDefineLiberal();
    REPORT(RP_DETAIL, "[%s] Computed function-defined registers", __func__);
    
    REPORT(RP_DETAIL, "Beginning interprocedural fixpoint computation");    
    /*
     *         Start out all registers and stack slots as unknown
     */
    FOREACH_FUNCTION(fun) {
        myBase = fun->f_callerFrameSize / ADDRESS_SIZE;

        FOREACH_BBL_IN_FUN(fun, bbl) {
            BBL_SET_FLAG(bbl, BBL_CP_FLAG);
            InitializeStackAndRegs(bbl);
        } 

        InitializeStackAndRegs(fun->f_entry);
        InitializeStackAndRegs(fun->f_exit);
        BBL_SET_FLAG(fun->f_entry, BBL_CP_FLAG);
        BBL_SET_FLAG(fun->f_exit, BBL_CP_FLAG);
    }

    /*
     * Begin the fixpoint computation. For all basic blocks, including
     * pseudo blocks (BT_ENTRY, BT_EXIT), we propagate information around
     * (including interprocedural edges) until nothing changes.
     */

    change = true;

    /*
     * If the constant propagator knows the condition codes in the PSW,
     * then it can potentially evaluate the outcome of a conditional jump.
     * The problem is that we don't want to delete the edge until the CP
     * reaches its fixpoint, otherwise we might incorrectly get rid of
     * an edge just because we didn't give a value enough chance to
     * propagate downward. 'resolveConditionalJumps' tells the evaluator
     * to delete edges if it knows the outcome of the jump. We set it to
     * false here, and later set it to true once a fixpoint is attained.
     */

    resolveConditionalJumps = false;

    roundNumber = 0;
    while (change) {
        change = false;
        REPORT(RP_SMLOOP,"Interprocedural computations: round %d",roundNumber);
        i = 0;
        FOREACH_FUNCTION(fun) {
            i++;
            if (!FUNCTION_HAS_FLAG(fun, FUNCTION_WELL_BEHAVED_FLAG) ||
                !FUNCTION_HAS_FLAG(fun, FUNCTION_BALANCED_STACK_FLAG)
/*                 (i > 0.15 *numFunctions) || */
/*                 (i < 0 *numFunctions) */) {

                FOREACH_BBL_IN_FUN(fun, bbl) {
                    BBLRegsBecomeUnknown(bbl);
                    BBLStackBecomesUnknown(bbl);
                    FOREACH_SUCC_IN_BBL(bbl, successors) {
                        BBL_SET_FLAG(NODE_TO(successors), BBL_CP_FLAG);
                    }            
                }
		
                BBLRegsBecomeUnknown(fun->f_entry);
                BBLRegsBecomeUnknown(fun->f_exit);
                BBLStackBecomesUnknown(fun->f_entry);
                BBLStackBecomesUnknown(fun->f_exit);

                /*
                 *  There were cases were a non-constant 
                 *  variable was getting propagated. So 
                 *  fixed it here to make sure that if a 
                 *  a register/variable becomes UNKNOWN
                 *  that info is propagated in the next
                 *  round. Without the CP FLAG it will 
                 *  not propagate.
                 */
                FOREACH_SUCC_IN_BBL(fun->f_exit, successors) {
                    BBL_SET_FLAG(NODE_TO(successors), BBL_CP_FLAG);
                }		
                continue;
            }

            myBase = fun->f_callerFrameSize / ADDRESS_SIZE;

            if (PropagateForBBL(fun->f_entry)) {
                change = true;
            }

            FOREACH_BBL_IN_FUN (fun, bbl) {
                if (PropagateForBBL(bbl)) {
                    change = true;
                }
            }

            if (PropagateForBBL(fun->f_exit)) {
                change = true;
            }

        }			/* end iteration over funs */
    }		/* end fixpoint constant propagation over basic blocks */

    REPORT(RP_SMLOOP, "Interprocedural computations: done");    

        /*
         * Now, replace all uses of known locations with their constants.
         */
    resolveConditionalJumps = true;

    ReportCPResults("foo", REPORT_CP_HEADER, REPORT_CP_HEADER,
                    REPORT_CP_HEADER);

    i = 0;
    FOREACH_FUNCTION(fun) {
        i++;
/*         if ( i > 0.15*numFunctions || */
/*              i < 0 *numFunctions ) */
/*             continue; */
        /*
         * We do not want to propagate constants across functions
         * which are not well behaved or which do not have a 
         * balanced stack
         */
        if (!FUNCTION_HAS_FLAG(fun, FUNCTION_WELL_BEHAVED_FLAG)) {
            continue;
        }

        if (!FUNCTION_HAS_FLAG(fun, FUNCTION_BALANCED_STACK_FLAG)) {
            continue;
        }
	  
        if (FUNCTION_HAS_FLAG(fun, FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG)) {
            continue;
        }
	
        myBase = fun->f_callerFrameSize / ADDRESS_SIZE;
        numSimplified = conditionalsResolved = forwards = 0;
        FOREACH_BBL_IN_FUN(fun, bbl) {

            /* The order of Merge and BBLGetConditionalEqualityInformation
             * is strict. Because Merge would merge the values from the
             * predecessors, so doing the BBLGet.. before merge will
             * nullify the effects of Merge
             */
        
            MergePredecessors(bbl);
            BBLGetConditionalEqualityInformation(bbl);
        
            instr = bbl->b_first;
            while (instr && (instr->i_Bblock == bbl)) {
                Instruction *next = instr->i_next;
                int diff = InstrBPSPDiff(instr);
                if (diff != STACK_UNKNOWN) {
                    stackBase = myBase + diff / ADDRESS_SIZE;
                } else {
                    stackBase = STACK_UNKNOWN;
                }

                if (instr->i_prev && isCallClass(instr->i_prev)) {
                    UpdateFlags(instr);
                }

                if (!isFPClass(instr) && !instr->i_prefixes) {
                    if (instr->i_src) {
                        numSimplified +=
                            ReplaceOperandWithConstant(instr, instr->i_src);
                        forwards +=
                            ReplaceOperandWithRegister(instr, instr->i_src);
                    }
                    if (instr->i_src2) {
                        numSimplified +=
                            ReplaceOperandWithConstant(instr, instr->i_src2);
                        forwards +=
                            ReplaceOperandWithRegister(instr, instr->i_src2);
                    }
                }

                UpdateFlags(instr);
                EvaluateInstruction(instr);
                instr = next;
            }			/* end iteration over instructions */
        }		/* end iteration over basic blocks in function */

        ReportCPResults(fun->f_name, numSimplified, forwards,
                        conditionalsResolved);
        numSimplified = conditionalsResolved = forwards = 0;
    }					/* end iteration over functions */


    ReportCPResults("foo", REPORT_CP_TRAILER, REPORT_CP_TRAILER,
                    REPORT_CP_TRAILER);
    DeallocateSnapShots();
    
    return;
}



/*
 *  PropagateForBBL(bbl) --
 *
 *  Merges predecessors of the bbl, evaluates all the instructions in the
 *  block, and then returns true if something has changed.
 */
static bool PropagateForBBL(Block *bbl) {
    bool change = false;
    Value *tempStack = NULL;
    Value *tempRegs = NULL;
    Instruction *instr;
    EdgeNode *successors;
    Block *succBlock = NULL;

    if (bbl->b_type == BT_DATABLOCK) {
        BBLRegsBecomeUnknown(bbl);
        BBLStackBecomesUnknown(bbl);
        FOREACH_SUCC_IN_BBL(bbl, successors) {
            BBL_SET_FLAG(NODE_TO(successors), BBL_CP_FLAG);
        }
        return true;
    }
    
    if (!BBL_HAS_FLAG(bbl, BBL_CP_FLAG)) {
        return false;
    }
    
    BBL_REMOVE_FLAG(bbl, BBL_CP_FLAG);

    if (FUNCTION_HAS_FLAG(bbl->b_fun, FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG)) {
        return false;
    }

    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
        tempStack = BBLCopyStackshot(bbl);
    }
    if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
        tempRegs = BBLCopyRegshot(bbl);
    }

    MergePredecessors(bbl);

    /*
     * We should also try and get any information we can about
     * equality-based conditional jumps. The invocation of
     * BBLGetConditionalEqualityInformation() will look at all
     * predecessors and try to propagate information about registers
     * or the stack into this block.
     */

    BBLGetConditionalEqualityInformation(bbl);

    /*
     * With the current view of the runtime stack and the registers,
     * the instructions can now be evaluated, and the view of the
     * world can be constructed at the end of this basic block.
     */

    FOREACH_INSTR_IN_BBL(bbl, instr) {
        int diff = InstrBPSPDiff(instr);
        if (diff != STACK_UNKNOWN) {
            stackBase = myBase + diff / ADDRESS_SIZE;
        } else {
            stackBase = STACK_UNKNOWN;
        }
        UpdateFlags(instr);
        succBlock = EvaluateInstruction(instr);
    }

    if (!change) {
        if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
            if (!BBLStackContentsAreEqual(bbl, tempStack)) {
                change = true;
            }
        }
        if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
            if (!BBLRegContentsAreEqual(bbl, tempRegs)) {
                change = true;
            }
        }
    }

    if (tempRegs != NULL) {
        free(tempRegs);
    }
    if (tempStack != NULL) {
        free(tempStack);
    }

    if (change == true) {
        /* if succBlock == NULL then you are not sure which
         * way control may flow. So mark all successor blocks.
         * else mark only the succBlock
         */
        if (!succBlock) {
            FOREACH_SUCC_IN_BBL(bbl, successors) {
                BBL_SET_FLAG(NODE_TO(successors), BBL_CP_FLAG);
            }
        } else {
            BBL_SET_FLAG(succBlock, BBL_CP_FLAG);            
        }
    }

    return change;
}



/*
 *  BBLAllocateStackshot(bbl) --
 *
 *  Allocates an array of Value records, one for each slot in
 *  the stack frame that 'fun' accesses, as well as one for each slot in
 *  the stack frame of its largest-framed callee.
 */
void BBLAllocateStackshot(Block *bbl) {
    Function *fun = bbl->b_fun;
    int stackSize = fun->f_stackFrameSize + fun->f_callerFrameSize;

    if (stackSize == 0) {
        return;
    }

    bbl->b_stackshot = zalloc(((stackSize / ADDRESS_SIZE) + 1) * sizeof(Value));

    BBL_SET_FLAG(bbl, BBL_STACKSHOT_FLAG);

    return;
}



/*
 *  BBLAllocateRegshot(bbl) --
 *
 *  Allocates an array of Value records, one for each general
 *  purpose register. The idea is that the basic blocks will keep track of
 *  the contents of registers at their entry points, and we can
 *  reconstruct the values at their exit points.
 */
void BBLAllocateRegshot(Block *bbl) {

    bbl->b_regshot = zalloc(sizeof(Value) * NUMBER_OF_GP_REGISTERS);
    BBL_SET_FLAG(bbl, BBL_REGSHOT_FLAG);
    return;
}



/*
 *  EvaluateInstruction(instr) --
 *
 *  The root of the constant propagator is the instruction evaluation
 *  capabilities. If we can determine the outcome of an instruction given
 *  some constant arguments then we've gained a lot. This routine acts as
 *  a virtual machine and emulates the instructions in software. Not all
 *  instructions appear below, but the top 99.9% do (based on opcode
 *  histograms).
 */
static Block *EvaluateInstruction(Instruction *instr) {
    bool eval = true;			// true if the insn will be evaluated
    Block *bbl = instr->i_Bblock;	// basic block for the supplied insn
    int pushBytes, baseReg, indexReg, i, value;
    int regContents, stackContents;
    bool destStack, src1Stack, src2Stack;
    bool destImm, src1Imm, src2Imm;
    bool destAddr, src1Addr, src2Addr;
    bool destReg, src1Reg, src2Reg;
    bool source1Unknown, source2Unknown, destUnknown;
    bool destFullReg, src1FullReg, src2FullReg;
    int bblStackshot;
    long long signedSource1 = 0;
    long long signedSource2 = 0;
    long long unsignedSource1 = 0;
    long long unsignedSource2 = 0;
    long long signedResult;
    bool knownSource1 = true;
    bool knownSource2 = true;
    Edge *deleteEdge = NULL;
    Edge *keepEdge = NULL;
    Block *target = NULL;
    PSW_BITVECTOR pswBitvector;
    char *instrText;
	Instruction *jinstr = NULL;
    
    destStack = src1Stack = src2Stack = false;
    destImm = src1Imm = src2Imm = false;
    destAddr = src1Addr = src2Addr = false;
    destReg = src1Reg = src2Reg = false;
    source1Unknown = source2Unknown = destUnknown = false;
    destFullReg = src1FullReg = src2FullReg = false;

    if (instr->i_class == CLASS_INT) {
        BBLRegsBecomeUnknown(bbl);	  
        BBLStackBecomesUnknown(bbl);
        return target;
    }
	
    if (instr->i_dest) {
        if (OPERAND_HAS_FLAG((instr->i_dest), OPERAND_VMODE_FLAG)) {
            destStack = OperandAccessesStackFromFramePointer(instr->i_dest);
        }
        destReg = OperandIsGPRegister(instr->i_dest);
        destImm = OperandIsImmediate(instr->i_dest);
        destAddr = OperandIsAddress(instr->i_dest);
    }
	
    if (instr->i_src) {
        if (OPERAND_HAS_FLAG((instr->i_src), OPERAND_VMODE_FLAG)) {
            src1Stack = OperandAccessesStackFromFramePointer(instr->i_src);
        }
        src1Reg = OperandIsGPRegister(instr->i_src);
        src1Imm = OperandIsImmediate(instr->i_src);
        src1Addr = OperandIsAddress(instr->i_src);
    }
    
    if (instr->i_src2) {
        if (OPERAND_HAS_FLAG((instr->i_src2), OPERAND_VMODE_FLAG)) {
            src2Stack = OperandAccessesStackFromFramePointer(instr->i_src2);
        }
        src2Reg = OperandIsGPRegister(instr->i_src2);
        src2Imm = OperandIsImmediate(instr->i_src2);
        src2Addr = OperandIsAddress(instr->i_src2);
    }

    bblStackshot = BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG);
    if (destReg) {
        destFullReg = RegisterIs32Bit(instr->i_dest->a_base_reg);
    }
    if (src1Reg) {
        src1FullReg = RegisterIs32Bit(instr->i_src->a_base_reg);
    }
    if (src2Reg) {
        src2FullReg = RegisterIs32Bit(instr->i_src2->a_base_reg);
    }

    /* This fix has to go. I need to fix the opcode table for plto
     * to build STRUCT_INSTR properly ot reflect the implicit and
     * registers used in these operations - SOMU
     */	
	/* Added by Somu */
	if (instr->i_class == CLASS_ADD  ||
	    instr->i_class == CLASS_SUBB ||
	    instr->i_class == CLASS_SUB  ||			 
	    instr->i_class == CLASS_ADDC) {
        if (OperandIsImmediate(instr->i_src) &&
			instr->i_src2 == NULL) {
            src2Reg = OperandIsGPRegister(instr->i_dest);
            src2FullReg = RegisterIs32Bit(instr->i_dest->a_base_reg);	   
	    }
	}

	if (instr->i_class == CLASS_LEA) {
        if (OPERAND_HAS_FLAG(instr->i_src, OPERAND_REGISTER_INDIRECT)) {
	        if (OPERAND_HAS_FLAG(instr->i_src, OPERAND_BASE_FLAG)) {
	            src1Reg = true;
                src1FullReg = RegisterIs32Bit(instr->i_src->a_base_reg);
            }
		
            if (OPERAND_HAS_FLAG(instr->i_src, OPERAND_INDEX_FLAG)) {
                src2Reg = true;
                src2FullReg = RegisterIs32Bit(instr->i_src->a_index_reg);
            } else {
                src2Reg = false;
            }
	    }
	}
	
	  
    /*
     * If we have an instruction that writes indirectly through some
     * register, than it can potentially write anywhere to memory, so
     * we'll set the entire stack to unknown. The only exception is if
     * it writes indirectly through EBP, in which case we might know
     * where it's going.
     */
    if (instr->i_dest) {
        if (OPERAND_HAS_FLAG((instr->i_dest), OPERAND_INDIRECT_FLAG)) {
            BBLStackBecomesUnknown(bbl);
        }
        if (OPERAND_HAS_FLAG(instr->i_dest, OPERAND_REGISTER_INDIRECT)) {
            if (!destStack) {
                BBLStackBecomesUnknown(bbl);
            }
        }
    }

    /*
     * For compare instructions, the PSW will be set inside the switch
     * on the instruction class. For every other type of instruction,
     * the PSW will get the PSW_UKNOWN bit set, to reflect the fact that
     * we don't know what effect the instruction has on the status word.
     * If an instruction does not define PSW let the PSW be. Do not set
     * it to UNKNOWN.
     */
    if (ComputePSWDefined(instr)) {
        pswBitvector = ComputePSWDefined(instr);
        for (i = 0; i < 32; i++) {
            if ((1 <<i) & pswBitvector) {
                psw[i] = PSW_UNKNOWN;
            }
        }
    }

    /*
     * If the destination operand writes to the stack at some weird
     * offset, e.g, one that is not a multiple of ADDRESS_SIZE, then
     * we'll set the slots it may write to as UNKNOWN.
     */
    if ((instr->i_dest) && (destStack) &&
        ((instr->i_dest->a_displacement % ADDRESS_SIZE) != 0)) {
        base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
        AssignStack(bbl, base, 0, SLOT_TYPE_UNKNOWN, 0);
        AssignStack(bbl, base + 1, 0, SLOT_TYPE_UNKNOWN, 0);
        eval = false;
    }

    /*
     * If any of the source operand access the stack at weird offsets,
     * such as one that is not a multiple of ADDRESS_SIZE, then we'll
     * opt not to try and evaluate the instruction. This happens so
     * rarely that we won't lose anything by doing this.
     */
    if ((instr->i_src) && (src1Stack) &&
        ((instr->i_src->a_displacement % ADDRESS_SIZE) != 0)) {
        eval = false;
    }
    if ((instr->i_src2) && (src2Stack) &&
        ((instr->i_src2->a_displacement % ADDRESS_SIZE) != 0)) {
        eval = false;
    }

    /*
     * Other cases when we can't/choose not to evaluate an instruction:
     *   1) A source operand is UNKNOWN
     *   2) The instruction has an implicit floating point dest.
     *   3) The destination operand is used, and UNKNOWN
     *   4) The instruction is a floating point instruction
     */
    if (OperandIsUndefinedOrUnknown(instr, instr->i_src)) {
        source1Unknown = true;
    }
	
    if (OperandIsUndefinedOrUnknown(instr, instr->i_src2)) {
        source2Unknown = true;
    }

	/* Added by Somu */
    if (instr->i_class == CLASS_ADD  ||
        instr->i_class == CLASS_SUBB ||
        instr->i_class == CLASS_SUB  ||
        instr->i_class == CLASS_ADDC) {
        if (OperandIsImmediate(instr->i_src) &&
            instr->i_src2 == NULL) {
            if (OperandIsUndefinedOrUnknown(instr, instr->i_dest)) {
                source2Unknown = true;
	        }
        }
    }

    if (instr->i_class == CLASS_LEA) { 
        if (OPERAND_HAS_FLAG(instr->i_src, OPERAND_REGISTER_INDIRECT)) {
            source1Unknown = false;
            source2Unknown = false;

            if (OPERAND_HAS_FLAG(instr->i_src, OPERAND_BASE_FLAG)) {
                if (REG_TYPE(bbl, instr->i_src->a_base_reg)==SLOT_TYPE_UNKNOWN||
                    REG_TYPE(bbl, instr->i_src->a_base_reg)==SLOT_TYPE_UNDEFINED) {
                    source1Unknown = true;
                }
            }

            if (OPERAND_HAS_FLAG(instr->i_src, OPERAND_INDEX_FLAG)) {
                if (REG_TYPE(bbl, instr->i_src->a_index_reg)==SLOT_TYPE_UNKNOWN||
                    REG_TYPE(bbl, instr->i_src->a_index_reg)==SLOT_TYPE_UNDEFINED) {
       	            source2Unknown = true;
                }
            } 
        }
    }

	
    if (InstrHasImplicitSrc(instr)
        && RegisterIsUndefinedOrUnknown(instr, InstrImplicitSrc(instr))
        && (instr->i_class != CLASS_PUSH)
        && (instr->i_class != CLASS_POP)) {
	    eval = false;
    }
    if (InstrHasImplicitSrc2(instr)
        && RegisterIsUndefinedOrUnknown(instr, InstrImplicitSrc2(instr))
        && (instr->i_class != CLASS_PUSH)
        && (instr->i_class != CLASS_POP)) {
	    eval = false;
    }
    if (InstrHasImplicitDest(instr)
        && RegisterIsFloatingPoint(InstrImplicitDest(instr))) {
	    eval = false;
    }
    if (InstrHasImplicitDest2(instr)
        && RegisterIsFloatingPoint(InstrImplicitDest2(instr))) {
	    eval = false;
    }
    if ((InstrGetOpcodeEntry(instr)->x_psw & DUF)) {
        if (OperandIsUndefinedOrUnknown(instr, instr->i_dest)) {
            destUnknown = true;
        }
    }
    if (isFPClass(instr)) {
        eval = false;
    }
    if (INSTR_HAS_FLAG(instr, INSTR_USES_FP_BUT_COMPENSATED_FLAG)) {
        eval = false;
    }

    if (source1Unknown && !destUnknown && !source2Unknown) {
        INSTR_SET_FLAG(instr, INSTR_EVAL_IF_SOURCE1);
    }
    if (source2Unknown && !destUnknown && !source1Unknown) {
        INSTR_SET_FLAG(instr, INSTR_EVAL_IF_SOURCE2);
    }
    if (destUnknown && !source1Unknown && !source2Unknown) {
        INSTR_SET_FLAG(instr, INSTR_EVAL_IF_DEST);
    }

    if (source1Unknown || source2Unknown || destUnknown) {
        eval = false;
    }

    /*
     * If the instruction is not going to be evaluated, then the slots
     * for the destination operand, whether they be register slots or
     * stack slots, need to take on UNKNOWN. If the instruction pushes
     * to the stack, then the top of the stack should also take on an
     * UNKNOWN value. Finally, if the instruction has prefixes, then all
     * the registers become unknown.
     */

    if (eval == false) {
        if (instr->i_dest) {
            /* else was missing. Added it here -Somu
             */
            if (destReg) {
                AssignRegister(bbl, instr->i_dest->a_base_reg, 0,
                    SLOT_TYPE_UNKNOWN, 0);
            } else if (destStack) {
                base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
                AssignStack(bbl, base, 0, SLOT_TYPE_UNKNOWN, 0);
                if (INSTR_HAS_FLAG(instr, INSTR_USES_FP_BUT_COMPENSATED_FLAG)) {
                    BBLStackBecomesUnknown(bbl);
                }
            }
        }

        if (InstrHasImplicitDest(instr)) {
            if (!RegisterIsFloatingPoint(INSTR_IMPLICIT_DEST1(instr))) {
                AssignRegister(bbl, INSTR_IMPLICIT_DEST1(instr), 0,
                    SLOT_TYPE_UNKNOWN, 0);
            }
        }

        if (InstrHasImplicitDest2(instr)) {
            if (!RegisterIsFloatingPoint(INSTR_IMPLICIT_DEST2(instr))) {
                AssignRegister(bbl, INSTR_IMPLICIT_DEST2(instr), 0,
                    SLOT_TYPE_UNKNOWN, 0);
            }
        }

        pushBytes = InstrPushBytes(instr);
        if (pushBytes != 0) {
            for (i = 0; i < pushBytes / ADDRESS_SIZE; i++) {
                AssignStack(bbl, stackBase + 1 + i, 0, SLOT_TYPE_UNKNOWN, 0);
            }
        }

        if (instr->i_prefixes) {
            BBLRegsBecomeUnknown(bbl);
        }

    }

    /*
     * Some instructions can be evaluated without knowing the contents of the
     * stack or register location, e.g; xor eax, eax.
     */

    if (eval == false) {
        Operand *one = instr->i_dest;
        Operand *two = instr->i_src;

        switch (instr->i_class) {

            /*
             *                                  XOR
             */

        case CLASS_XOR:
            if (one->a_base_reg != two->a_base_reg) {
                break;
            }
            if (one->a_index_reg != two->a_index_reg) {
                break;
            }
            if (one->a_base_reg != two->a_scale) {
                break;
            }
            if (one->a_displacement != two->a_displacement) {
                break;
            }
            if (destStack) {
                base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
                AssignStack(bbl, base, 0, SLOT_TYPE_CONSTANT, 0);
            } else if (destReg) {
                AssignRegister(bbl, one->a_base_reg, 0, SLOT_TYPE_CONSTANT, 0);
            }
            break;

            /*
             *                                  MOV
             */

        case CLASS_MOV:

            if (destStack && bblStackshot) {
                base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
                if (src1Stack) {
                    int base2 =
                        myBase - instr->i_src->a_displacement / ADDRESS_SIZE;
                    AssignStackPtr(bbl, base, SLOT_STACK_FLAG, base2);
                } else if (src1Reg && src1FullReg) {
                    AssignStackPtr(bbl, base,
                        RegisterToSlotFlag(instr->i_src->a_base_reg), 0);
                }
            } else if (destReg && destFullReg) {
                int reg = RegisterToIndex(instr->i_dest->a_base_reg);
                if (src1Stack && bblStackshot) {
                    int base2 =
                        myBase - instr->i_src->a_displacement / ADDRESS_SIZE;
                    AssignRegPtr(bbl, reg, SLOT_STACK_FLAG, base2);
                } else if (src1Reg && src1FullReg) {
                    AssignRegPtr(bbl, reg,
                        RegisterToSlotFlag(instr->i_src->a_base_reg), 0);
                }
            }

            break;

            /*
             *                                 PUSH
             */

        case CLASS_PUSH:

            if (stackBase == STACK_UNKNOWN) {
                break;
            }

            if (src1Reg && src1FullReg && bblStackshot) {
                AssignStackPtr(bbl, stackBase + 1,
                    RegisterToSlotFlag(instr->i_src->a_base_reg), 0);
            }

            break;

            /*
             *                                 POP
             */

        case CLASS_POP:

            if (stackBase == STACK_UNKNOWN) {
                break;
            }

            if (destReg && destFullReg) {
                int reg = RegisterToIndex(instr->i_dest->a_base_reg);
                AssignRegPtr(bbl, reg, SLOT_STACK_FLAG, stackBase);
            }

            break;

        }

        return target;
    }

    numInstructionsEvaluated++;

    /*
     * The heart of the instruction evaluation lies below. The following
     * instructions are currently interpreted:
     *
     *         1) MOV
     *         2) PUSH
     *         3) POP
     *         4) ADD
     *         5) CALL
     */

    switch (instr->i_class) {


        /*
         *                                  CALL
         */

    case CLASS_CALL:

        break;

        /*
         *                                  PUSH
         */

    case CLASS_PUSH:
        if (stackBase == STACK_UNKNOWN) {
            BBLStackBecomesUnknown(bbl);
        } else {
            if (src1Imm) {
                AssignStack(bbl, stackBase + 1,
                    instr->i_src->a_displacement,
                    SLOT_TYPE_CONSTANT, src1Addr ? SLOT_ADDRESS_FLAG : 0);
            } else if (src1Reg) {
                AssignStack(bbl, stackBase + 1,
                    REG_CONTENTS(bbl, instr->i_src->a_base_reg),
                    REG_TYPE(bbl, instr->i_src->a_base_reg),
                    (REG_FLAGS(bbl, instr->i_src->a_base_reg) &
                        SLOT_ADDRESS_FLAG));
                if (src1FullReg) {
                    AssignStackPtr(bbl, stackBase + 1,
                        RegisterToSlotFlag(instr->i_src->a_base_reg) |
                        (REG_FLAGS(bbl, instr->i_src->a_base_reg)), 0);
                }
            } else {
                goto noEval;
            }
        }
        break;


        /*
         *                                   POP
         */

    case CLASS_POP:
        if (destReg) {
            if (stackBase == STACK_UNKNOWN) {
                AssignRegister(bbl, instr->i_dest->a_base_reg, 0,
                    SLOT_TYPE_UNKNOWN, 0);
            } else {
                int reg = RegisterToIndex(instr->i_dest->a_base_reg);
                AssignRegister(bbl, instr->i_dest->a_base_reg,
                    STACK_CONTENTS(bbl, stackBase), STACK_TYPE(bbl, stackBase),
                    STACK_FLAGS(bbl, stackBase) & SLOT_ADDRESS_FLAG);
                AssignRegPtr(bbl, reg,
                    SLOT_STACK_FLAG | (STACK_FLAGS(bbl, stackBase)), stackBase);
            }
        } else {
            goto noEval;
        }
        break;


        /*
         *                                   CMP
         */

    case CLASS_CMP:

        if (src1Reg) {
            signedSource1 = REG_CONTENTS(bbl, instr->i_src->a_base_reg);
        } else if (src1Imm) {
            signedSource1 = instr->i_src->a_displacement;
        } else if (src1Stack) {
            base = myBase - instr->i_src->a_displacement / ADDRESS_SIZE;
            signedSource1 = STACK_CONTENTS(bbl, base);
        } else {
            knownSource1 = false;
        }

        if (src2Reg) {
            signedSource2 = REG_CONTENTS(bbl, instr->i_src2->a_base_reg);
        } else if (src2Imm) {
            signedSource2 = instr->i_src2->a_displacement;
        } else if (src2Stack) {
            base = myBase - instr->i_src2->a_displacement / ADDRESS_SIZE;
            signedSource2 = STACK_CONTENTS(bbl, base);
        } else {
            knownSource2 = false;
        }

        if (knownSource1 && knownSource2) {
            unsignedSource1 = (unsigned long long) signedSource1;
            unsignedSource2 = (unsigned long long) signedSource2;

            if (signedSource1 == signedSource2) {
                psw[PSW_ZF_INDEX] = PSW_SET;
            } else {
                psw[PSW_ZF_INDEX] = PSW_OFF;
            }

            if ((signedSource1 - signedSource2) < 0) {
                psw[PSW_SF_INDEX] = PSW_SET;
            } else {
                psw[PSW_SF_INDEX] = PSW_OFF;
            }

            if (unsignedSource1 < unsignedSource2) {
                psw[PSW_CF_INDEX] = PSW_SET;
            } else {
                psw[PSW_CF_INDEX] = PSW_OFF;
            }

            signedResult = signedSource1 - signedSource2;
            if (signedResult > ((long long) 0x7fffffff)) {
                psw[PSW_OF_INDEX] = PSW_SET;
            } else {
                psw[PSW_OF_INDEX] = PSW_OFF;
            }
            if (signedResult < ((long long) 0x7fffffff) * -1) {
                psw[PSW_OF_INDEX] = PSW_SET;
            } else {
                psw[PSW_OF_INDEX] = PSW_OFF;
            }
        } else {
            goto noEval;
        }

        break;


        /*
         *                                   NOP
         */

    case CLASS_NOP:
        break;


        /*
         *                                  MOVZB
         */

    case CLASS_MOVZB:
        goto noEval;
        break;


	/*
	 *                                   SAR
	 */

    case CLASS_SAR:
	goto noEval;
	break;


	/*
	 *                                   SET
	 */

    case CLASS_SETO:
    case CLASS_SETAE:
    case CLASS_SETBE:
    case CLASS_SETNS:
    case CLASS_SETGE:
    case CLASS_SETNO:
    case CLASS_SETE:
    case CLASS_SETA:
    case CLASS_SETP:
    case CLASS_SETLE:
    case CLASS_SETB:
    case CLASS_SETNE:
    case CLASS_SETS:
    case CLASS_SETL:
    case CLASS_SETG:
	goto noEval;
	break;


	/*
	 *                                   LEA
	 */

    case CLASS_LEA:
	baseReg = instr->i_src->a_base_reg;
	indexReg = instr->i_src->a_index_reg;
	value = instr->i_src->a_displacement;

	if (baseReg) {
	    if ((REG_TYPE(bbl, baseReg) == SLOT_TYPE_UNKNOWN)) {
		goto noEval;
	    }
	    if ((REG_TYPE(bbl, baseReg) == SLOT_TYPE_UNDEFINED)) {
		goto noEval;
	    }
	    if ((REG_FLAGS(bbl, baseReg) & SLOT_ADDRESS_FLAG)) {
		goto noEval;
	    }
	    value = value + REG_CONTENTS(bbl, baseReg);
	}

	if (indexReg) {
	    if ((REG_TYPE(bbl, indexReg) == SLOT_TYPE_UNKNOWN)) {
		goto noEval;
	    }
	    if ((REG_TYPE(bbl, indexReg) == SLOT_TYPE_UNDEFINED)) {
		goto noEval;
	    }
	    if ((REG_FLAGS(bbl, indexReg) & SLOT_ADDRESS_FLAG)) {
		goto noEval;
	    }
	    value = value +
		(REG_CONTENTS(bbl, indexReg) * instr->i_src->a_scale);
	}

	AssignRegister(bbl, instr->i_dest->a_base_reg, value,
	    SLOT_TYPE_CONSTANT, 0);
	break;

	/*
	 *                                   RET
	 */

    case CLASS_RET:
	break;


	/*
	 *                                   HALT
	 */

    case CLASS_HALT:
	break;


	/*
	 *                                  TEST
	 */

    case CLASS_TEST:

	if (src1Reg) {
	    unsignedSource1 = REG_CONTENTS(bbl, instr->i_src->a_base_reg);
	} else if (src1Imm) {
	    unsignedSource1 = instr->i_src->a_displacement;
	} else if (src1Stack) {
	    base = myBase - instr->i_src->a_displacement / ADDRESS_SIZE;
	    unsignedSource1 = STACK_CONTENTS(bbl, base);
	} else {
	    knownSource1 = false;
	}

	if (src2Reg) {
	    unsignedSource2 = REG_CONTENTS(bbl, instr->i_src2->a_base_reg);
	} else if (src2Imm) {
	    unsignedSource2 = instr->i_src2->a_displacement;
	} else if (src2Stack) {
	    base = myBase - instr->i_src2->a_displacement / ADDRESS_SIZE;
	    unsignedSource2 = STACK_CONTENTS(bbl, base);
	} else {
	    knownSource2 = false;
	}

	psw[PSW_OF_INDEX] = PSW_OFF;
	psw[PSW_CF_INDEX] = PSW_OFF;

	if (knownSource2 && knownSource1) {
	    unsigned int temp = unsignedSource1 & unsignedSource2;
	    if (temp & (1 << ((ADDRESS_SIZE * 8) - 1))) {
		psw[PSW_SF_INDEX] = 1;
	    } else {
		psw[PSW_SF_INDEX] = 0;
	    }
	    if (temp == 0) {
		psw[PSW_ZF_INDEX] = 1;
	    } else {
		psw[PSW_ZF_INDEX] = 0;
	    }
	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   INC
	 */

    case CLASS_INC:
	if (destReg) {
	    AssignRegister(bbl, instr->i_dest->a_base_reg,
		REG_CONTENTS(bbl, instr->i_dest->a_base_reg) + 1,
		SLOT_TYPE_CONSTANT, REG_FLAGS(bbl,
		    instr->i_dest->a_base_reg) & SLOT_ADDRESS_FLAG);
	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    AssignStack(bbl, base, STACK_CONTENTS(bbl, base) + 1,
		SLOT_TYPE_CONSTANT, STACK_FLAGS(bbl, base) & SLOT_ADDRESS_FLAG);
	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   DEC
	 */

    case CLASS_DEC:
	if (destReg) {
	    AssignRegister(bbl, instr->i_dest->a_base_reg,
		REG_CONTENTS(bbl, instr->i_dest->a_base_reg) - 1,
		SLOT_TYPE_CONSTANT, REG_FLAGS(bbl,
		    instr->i_dest->a_base_reg) & SLOT_ADDRESS_FLAG);
	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    AssignStack(bbl, base, STACK_CONTENTS(bbl, base) - 1,
		SLOT_TYPE_CONSTANT, STACK_FLAGS(bbl, base) & SLOT_ADDRESS_FLAG);
	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   SHL
	 */

    case CLASS_SHL:

	if (!src1Imm) {
	    goto noEval;
	}

	NO_EVAL_IF_DEST_IS_ADDRESS();

	if (destReg) {
	    AssignRegister(bbl, instr->i_dest->a_base_reg,
		REG_CONTENTS(bbl, instr->i_dest->a_base_reg) <<
		instr->i_src->a_displacement, SLOT_TYPE_CONSTANT, 0);

	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    AssignStack(bbl, base,
		STACK_CONTENTS(bbl, base) << instr->i_src->a_displacement,
		SLOT_TYPE_CONSTANT, 0);

	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   SHR
	 */

    case CLASS_SHR:

	if (!src1Imm) {
	    goto noEval;
	}

	NO_EVAL_IF_DEST_IS_ADDRESS()

	if (destReg) {
	    AssignRegister(bbl, instr->i_dest->a_base_reg,
		REG_CONTENTS(bbl, instr->i_dest->a_base_reg) >>
		instr->i_src->a_displacement, SLOT_TYPE_CONSTANT, 0);

	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    AssignStack(bbl, base,
		STACK_CONTENTS(bbl, base) >> instr->i_src->a_displacement,
		SLOT_TYPE_CONSTANT, 0);

	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   XOR
	 */

    case CLASS_XOR:

	NO_EVAL_IF_SOURCE1_IS_ADDRESS();
	NO_EVAL_IF_DEST_IS_ADDRESS();

	if (destReg) {
	    regContents = REG_CONTENTS(bbl, instr->i_dest->a_base_reg);
	    if (src1Imm) {
		if (src1Addr) {
		    goto noEval;
		}
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    instr->i_src->a_displacement ^ regContents,
		    SLOT_TYPE_CONSTANT, 0);
	    } else if (src1Stack) {
		base = myBase - instr->i_src->a_displacement / ADDRESS_SIZE;
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    STACK_CONTENTS(bbl, base) ^ regContents, STACK_TYPE(bbl,
			base), 0);
	    } else if (src1Reg) {
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    REG_CONTENTS(bbl,
			instr->i_src->a_base_reg) ^ regContents,
		    REG_TYPE(bbl, instr->i_src->a_base_reg), 0);
	    } else {
		goto noEval;
	    }

	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    stackContents = STACK_CONTENTS(bbl, base);

	    if (src1Imm) {
		if (src1Addr) {
		    goto noEval;
		}
		AssignStack(bbl, base,
		    stackContents ^ instr->i_src->a_displacement,
		    SLOT_TYPE_CONSTANT, 0);
	    } else if (src1Reg) {
		int reg = instr->i_src->a_base_reg;
		AssignStack(bbl, base, REG_CONTENTS(bbl, reg) ^ stackContents,
		    REG_TYPE(bbl, reg), 0);
	    } else if (src1Stack) {
		int base2 = myBase - instr->i_src->a_displacement/ADDRESS_SIZE;
		AssignStack(bbl, base, STACK_CONTENTS(bbl,
			base2) ^ stackContents, STACK_TYPE(bbl, base2), 0);
	    } else {
		goto noEval;
	    }

	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   OR
	 */

    case CLASS_OR:

	NO_EVAL_IF_SOURCE1_IS_ADDRESS();
	NO_EVAL_IF_DEST_IS_ADDRESS();

	if (destReg) {
	    regContents = REG_CONTENTS(bbl, instr->i_dest->a_base_reg);
	    if (src1Imm) {
		if (src1Addr) {
		    goto noEval;
		}
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    instr->i_src->a_displacement | regContents,
		    SLOT_TYPE_CONSTANT, 0);
	    } else if (src1Stack) {
		base = myBase - instr->i_src->a_displacement / ADDRESS_SIZE;
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    STACK_CONTENTS(bbl, base) | regContents, STACK_TYPE(bbl,
			base), 0);
	    } else if (src1Reg) {
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    REG_CONTENTS(bbl,
			instr->i_src->a_base_reg) | regContents,
		    REG_TYPE(bbl, instr->i_src->a_base_reg), 0);
	    } else {
		goto noEval;
	    }

	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    stackContents = STACK_CONTENTS(bbl, base);
	    if (src1Imm) {
		if (src1Addr) {
		    goto noEval;
		}
		AssignStack(bbl, base,
		    stackContents | instr->i_src->a_displacement,
		    SLOT_TYPE_CONSTANT, 0);
	    } else if (src1Reg) {
		int reg = instr->i_src->a_base_reg;
		AssignStack(bbl, base, REG_CONTENTS(bbl, reg) | stackContents,
		    REG_TYPE(bbl, reg), 0);
	    } else if (src1Stack) {
		int base2 = myBase - instr->i_src->a_displacement/ADDRESS_SIZE;
		AssignStack(bbl, base, STACK_CONTENTS(bbl,
			base2) | stackContents, STACK_TYPE(bbl, base2), 0);
	    } else {
		goto noEval;
	    }

	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   AND
	 */

    case CLASS_AND:

	NO_EVAL_IF_SOURCE1_IS_ADDRESS();
	NO_EVAL_IF_DEST_IS_ADDRESS();

	if (destReg) {
	    regContents = REG_CONTENTS(bbl, instr->i_dest->a_base_reg);
	    if (src1Imm) {
		if (src1Addr) {
		    goto noEval;
		}
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    instr->i_src->a_displacement & regContents,
		    SLOT_TYPE_CONSTANT, 0);
	    } else if (src1Stack) {
		base = myBase - instr->i_src->a_displacement / ADDRESS_SIZE;
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    STACK_CONTENTS(bbl, base) & regContents, STACK_TYPE(bbl,
			base), 0);
	    } else if (src1Reg) {
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    REG_CONTENTS(bbl,
			instr->i_src->a_base_reg) & regContents,
		    REG_TYPE(bbl, instr->i_src->a_base_reg), 0);
	    } else {
		goto noEval;
	    }

	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    stackContents = STACK_CONTENTS(bbl, base);
	    if (src1Imm) {
		if (src1Addr) {
		    goto noEval;
		}
		AssignStack(bbl, base,
		    stackContents & instr->i_src->a_displacement,
		    SLOT_TYPE_CONSTANT, 0);
	    } else if (src1Reg) {
		int reg = instr->i_src->a_base_reg;
		AssignStack(bbl, base, REG_CONTENTS(bbl, reg) & stackContents,
		    REG_TYPE(bbl, reg), 0);
	    } else if (src1Stack) {
		int base2 = myBase - instr->i_src->a_displacement/ADDRESS_SIZE;
		AssignStack(bbl, base, STACK_CONTENTS(bbl,
			base2) & stackContents, STACK_TYPE(bbl, base2), 0);
	    } else {
		goto noEval;
	    }

	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   NOT
	 */

    case CLASS_NOT:

	NO_EVAL_IF_DEST_IS_ADDRESS();

	if (destReg) {
	    AssignRegister(bbl, instr->i_dest->a_base_reg,
		~(REG_CONTENTS(bbl, instr->i_dest->a_base_reg)),
		SLOT_TYPE_CONSTANT, 0);
	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    AssignStack(bbl, base, ~(STACK_CONTENTS(bbl, base)),
		SLOT_TYPE_CONSTANT, 0);
	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   NEG
	 */

    case CLASS_NEG:

	NO_EVAL_IF_DEST_IS_ADDRESS();

	if (destReg) {
	    AssignRegister(bbl, instr->i_dest->a_base_reg,
		REG_CONTENTS(bbl, instr->i_dest->a_base_reg) * -1,
		SLOT_TYPE_CONSTANT, 0);
	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    AssignStack(bbl, base, STACK_CONTENTS(bbl, base) * -1,
		SLOT_TYPE_CONSTANT, 0);
	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   JMP
	 */

    case CLASS_JUMP:
	break;

	/*
	 *                                   JA
	 */

    case CLASS_JA:
	if ((psw[PSW_CF_INDEX] == PSW_OFF) && (psw[PSW_ZF_INDEX] == PSW_OFF)) {
	    deleteEdge = BBLGetFalseEdge(bbl);
	} else if ((psw[PSW_CF_INDEX] == PSW_SET)
	    || (psw[PSW_ZF_INDEX] == PSW_SET)) {
	    deleteEdge = BBLGetTrueEdge(bbl);
	}
	goto deleteConditionalEdge;

	/*
	 *                                   JAE
	 */

    case CLASS_JAE:
	if (psw[PSW_CF_INDEX] == PSW_OFF) {
	    deleteEdge = BBLGetFalseEdge(bbl);
	} else if (psw[PSW_CF_INDEX] == PSW_SET) {
	    deleteEdge = BBLGetTrueEdge(bbl);
	}
	goto deleteConditionalEdge;

	/*
	 *                                   JB
	 */

    case CLASS_JB:
	if (psw[PSW_CF_INDEX] == PSW_SET) {
	    deleteEdge = BBLGetFalseEdge(bbl);
	} else if (psw[PSW_CF_INDEX] == PSW_OFF) {
	    deleteEdge = BBLGetTrueEdge(bbl);
	}
	goto deleteConditionalEdge;

	/*
	 *                                   JBE
	 */

    case CLASS_JBE:
	if ((psw[PSW_CF_INDEX] == PSW_SET) || (psw[PSW_ZF_INDEX] == PSW_SET)) {
	    deleteEdge = BBLGetFalseEdge(bbl);
	} else if ((psw[PSW_CF_INDEX] == PSW_OFF)
	    && (psw[PSW_ZF_INDEX] == PSW_OFF)) {
	    deleteEdge = BBLGetTrueEdge(bbl);
	}
	goto deleteConditionalEdge;

	/*
	 *                                   JCXZ
	 */

    case CLASS_JCXZ:
	break;

	/*
	 *                                   JE
	 */

    case CLASS_JE:
	if (psw[PSW_ZF_INDEX] == PSW_SET) {
	    deleteEdge = BBLGetFalseEdge(bbl);
	} else if (psw[PSW_ZF_INDEX] == PSW_OFF) {
	    deleteEdge = BBLGetTrueEdge(bbl);
	}
	goto deleteConditionalEdge;

	/*
	 *                                   JG
	 */

    case CLASS_JG:
	if ((psw[PSW_ZF_INDEX] == PSW_OFF)
	    && (psw[PSW_OF_INDEX] == psw[PSW_SF_INDEX])) {
	    deleteEdge = BBLGetFalseEdge(bbl);
	} else if ((psw[PSW_ZF_INDEX] == PSW_SET)
	    || (psw[PSW_OF_INDEX] != psw[PSW_SF_INDEX])) {
	    deleteEdge = BBLGetTrueEdge(bbl);
	}
	goto deleteConditionalEdge;

	/*
	 *                                   JGE
	 */

    case CLASS_JGE:
        if (psw[PSW_SF_INDEX] == PSW_UNKNOWN ||
            psw[PSW_OF_INDEX] == PSW_UNKNOWN) {
            goto deleteConditionalEdge;
        }
        if (psw[PSW_SF_INDEX] == psw[PSW_OF_INDEX]) {
            deleteEdge = BBLGetFalseEdge(bbl);
        } else if (psw[PSW_SF_INDEX] != psw[PSW_OF_INDEX]) {
            deleteEdge = BBLGetTrueEdge(bbl);
        }
	goto deleteConditionalEdge;

	/*
	 *                                   JL
	 */

    case CLASS_JL:
        if (psw[PSW_SF_INDEX] == PSW_UNKNOWN ||
            psw[PSW_OF_INDEX] == PSW_UNKNOWN) {
            goto deleteConditionalEdge;
        }        
        if (psw[PSW_SF_INDEX] != psw[PSW_OF_INDEX]) {
            deleteEdge = BBLGetFalseEdge(bbl);
        } else if (psw[PSW_SF_INDEX] == psw[PSW_OF_INDEX]) {
            deleteEdge = BBLGetTrueEdge(bbl);
        }
        goto deleteConditionalEdge;

        /*
         *                                   JLE
         */

    case CLASS_JLE:
        if ((psw[PSW_ZF_INDEX] == PSW_SET)
            || (psw[PSW_OF_INDEX] != psw[PSW_SF_INDEX])) {
            deleteEdge = BBLGetFalseEdge(bbl);
        } else if ((psw[PSW_ZF_INDEX] == PSW_OFF)
            && (psw[PSW_OF_INDEX] == psw[PSW_SF_INDEX])) {
            deleteEdge = BBLGetTrueEdge(bbl);
        }
        goto deleteConditionalEdge;

        /*
         *                                   JNE
         */

    case CLASS_JNE:
        if (psw[PSW_ZF_INDEX] == PSW_OFF) {
            deleteEdge = BBLGetFalseEdge(bbl);
        } else if (psw[PSW_ZF_INDEX] == PSW_SET) {
            deleteEdge = BBLGetTrueEdge(bbl);
        }
        goto deleteConditionalEdge;

        /*
         *                                   JNO
         */

    case CLASS_JNO:
        break;

        /*
         *                                   JNP
         */

    case CLASS_JNP:
        break;

        /*
         *                                   JNS
         */

    case CLASS_JNS:
        break;

        /*
         *                                   JO
         */

    case CLASS_JO:
        break;

        /*
         *                                   JP
         */

    case CLASS_JP:
        break;

        /*
         *                                   JS
         */

    case CLASS_JS:
        break;

    deleteConditionalEdge:

        /* The no flow flags are set so that we know that
         * there is no control flow through that edge. If we are not
         * sure about the control flow, as is the case directly
         * below then we remove the flag. Added these lines of code
         * because when you MergePredecessors alone then we should
         * be aware what predecessors to merge and what not to
         * Merge.... :)
         * Cause : Self looped basic block has failed in Constant Propagation
         * algorithm
         */
        if (deleteEdge == NULL) {
            /*
             * The checks for if the edge is present or not present was added
             * because the CFG generated in the following case was missing the
             * true edge and false edge.
             * d665f:  cmp    $0x2,%edx
             * d6662:  je     d6664
             * d6664:  mov    0xfffffffc(%ebp),%ebx
             * d6667:  leave
             * d6668:  ret
             * I know the fix has to go into CFG building. It is collapsing
             * both the edges into ET_UNCOND. 
             */
            if (BBLGetTrueEdge(bbl)) {
                EDGE_REMOVE_FLAG(BBLGetTrueEdge(bbl),
                    EDGE_NO_FLOW_FLAG);
            }
            if (BBLGetFalseEdge(bbl)) {
                EDGE_REMOVE_FLAG(BBLGetFalseEdge(bbl),
                    EDGE_NO_FLOW_FLAG);
            }
            break;
        }

    EDGE_SET_FLAG(deleteEdge, EDGE_NO_FLOW_FLAG);
	if (deleteEdge->e_type == ET_FALSE) {
	    keepEdge = BBLGetTrueEdge(bbl);
	}
	if (deleteEdge->e_type == ET_TRUE) {
	    keepEdge = BBLGetFalseEdge(bbl);
	}

    target = keepEdge->e_to;
    
	if (resolveConditionalJumps == false) {
	    break;
	}

	pswBitvector = ComputePSWUsed(instr);
	for (i = 0; i < 32; i++) {
	    if ((psw[i] == PSW_UNKNOWN) && (pswBitvector & (1 << i))) {
		goto dontDeleteEdge;
	    }
	}
    
/*     if (prevFunction != INSTR_FUNCTION(instr)) { */
/*         prevFunction = INSTR_FUNCTION(instr) ; */
/*         tnode = FunCreatePredominatorTree(prevFunction); */
/*     } */
    
	CASSERT(keepEdge != NULL, "[EvaluateInstruction]"
	    " Could not find true/false edge in conditional jump.\n");
	CASSERT(bbl->b_last == instr, "[EvaluateInstruction]"
	    " Conditional jump not the last instruction in the block!\n");

	instrText = PrintInstruction(instr);

    REPORT(RP_LGLOOP, "%08X/%-5d Eliminated (%s) %s", instr->i_address,
        instr->i_id, edgeTypeToString(deleteEdge->e_type), instrText);

	conditionalsResolved++;
	free(instrText);

	BBLRemoveEdges(bbl);
	BBLAddEdge(bbl, target, ET_NORMAL);
	bbl->b_type = BT_NORMAL;
    
#if 0
    long long edgeWeight = 0, deletedWeight = 0;
    int deletedEdge = deleteEdge->e_type;

    deletedWeight = deleteEdge->e_edgeWeight;    

    bool backEdge = false;
    
    if (BBLDominatesBBL(tnode, deleteEdge->e_to, bbl) ||
        (deleteEdge->e_to == bbl)) {
        backEdge = true;
        edgeWeight = keepEdge->e_edgeWeight;
    } else {
        edgeWeight =
            keepEdge->e_edgeWeight + deleteEdge->e_edgeWeight;
    }
    
	target = keepEdge->e_to;
	BBLRemoveEdges(bbl);

	BBLAddEdge(bbl, target, ET_NORMAL)->e_edgeWeight = edgeWeight;
	bbl->b_type = BT_NORMAL;
	
    /* The jump has to be added only if it is not a
     * fall through block. Else skip adding the jump.
     * Do not merge the blocks until you know that it has
     * only one predecessor left. 
     */
    if (target != bbl->b_next) {
        jinstr = InstrCreateJMP();
        jinstr->i_Bblock = bbl;

        InsertInstrBeforeInstr(jinstr, instr);

        BBLRemoveEdgeTyped(bbl, target, ET_NORMAL);
        BBLAddEdge(bbl, target, ET_UNCOND)->e_edgeWeight
            =edgeWeight;
        bbl->b_type = BT_UNCOND;
    }

    if (bbl->b_clone) {
        Block *origBlock = bbl->b_clone;
        if (deletedEdge == ET_FALSE) {
            if (!backEdge) {
                BBLGetTrueEdge(origBlock)->e_edgeWeight -=
                    deletedWeight;
            }
            BBLGetFalseEdge(origBlock)->e_edgeWeight +=
                deletedWeight;
        } else {
            BBLGetTrueEdge(origBlock)->e_edgeWeight +=
                deletedWeight;
            if (!backEdge) {
                BBLGetFalseEdge(origBlock)->e_edgeWeight -=
                    deletedWeight;
            }
        }
    }
    
	InstrDestroy(instr);
    tnode = FunCreatePredominatorTree(prevFunction);
#endif
    
    if (target->b_type != BT_DATABLOCK) {
        if (target != bbl->b_next) {
            jinstr = InstrCreateJMP();
            jinstr->i_Bblock = bbl;

            jinstr->i_next = instr;
            jinstr->i_prev = instr->i_prev;

            if (instr->i_prev) {
                instr->i_prev->i_next = jinstr;
            }

            instr->i_prev = jinstr;

            bbl->b_type = BT_UNCOND;
            BBLRemoveEdgeTyped(bbl, target, ET_NORMAL);
            BBLAddEdge(bbl, target, ET_UNCOND);
        }
        InstrDestroy(instr);
    } else {
        if (keepEdge->e_type == ET_TRUE) {
            jinstr = InstrCreateJMP();
            jinstr->i_Bblock = bbl;

            jinstr->i_next = instr->i_next;
            jinstr->i_prev = instr->i_prev;
            if (instr->i_prev) {
                instr->i_prev->i_next = jinstr;
            }

            instr->i_prev = jinstr;

            ADDRESS addr;
            Instruction *tinstr = 0;
            if (OP_TYPE((instr->i_src), 
                    OPERAND_DISP_R_FLAG)) {
                addr = instr->i_src->a_displacement +
                    instr->i_address + instr->i_length;
                tinstr = BacktraceToInstr(addr);
                AddPatchInstrOffsetToInstr(tinstr,
                    addr - tinstr->i_address, jinstr, 
                    jinstr->i_src);
                OPERAND_SET_FLAG(jinstr->i_src, OPERAND_RELOCATED);
            } else {
                printf("[MRRR] Add an absolute patch !\n");
            }
            
            
            bbl->b_type = BT_UNCOND;
            BBLRemoveEdgeTyped(bbl, target, ET_NORMAL);
            BBLAddEdge(bbl, target, ET_UNCOND);

            InstrDestroy(instr);
        }
    }
    
dontDeleteEdge:

	break;

	/*
	 *                                   STD
	 */

    case CLASS_STD:
    case CLASS_STI:
    case CLASS_CLD:
    case CLASS_CLI:
    case CLASS_STC:
    case CLASS_CLC:
    case CLASS_CSTD:
	goto noEval;
	break;


	/*
	 *                                   MOV
	 */

    case CLASS_MOV:

	if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    if (src1Imm) {
		AssignStack(bbl, base,
		    instr->i_src->a_displacement,
		    SLOT_TYPE_CONSTANT,
		    (OperandIsAddress(instr->i_src) ? SLOT_ADDRESS_FLAG : 0));
	    } else if (src1Reg) {
		AssignStack(bbl, base, REG_CONTENTS(bbl,
			instr->i_src->a_base_reg), REG_TYPE(bbl,
			instr->i_src->a_base_reg), REG_FLAGS(bbl,
			instr->i_src->a_base_reg) & SLOT_ADDRESS_FLAG);
		if (src1FullReg) {
		    AssignStackPtr(bbl, base,
			RegisterToSlotFlag(instr->i_src->a_base_reg) |
			    (REG_FLAGS(bbl, instr->i_src->a_base_reg)), 0);
		}
	    } else if (src1Stack) {
		int base2 = myBase - instr->i_src->a_displacement/ADDRESS_SIZE;
		AssignStack(bbl, base, STACK_CONTENTS(bbl, base2),
		    STACK_TYPE(bbl, base2), STACK_FLAGS(bbl,
			base2) & SLOT_ADDRESS_FLAG);
		AssignStackPtr(bbl, base, SLOT_STACK_FLAG | (STACK_FLAGS(bbl,
			    base2)), base2);
	    } else {
		goto noEval;
	    }

	} else if (destReg) {
	    int reg = RegisterToIndex(instr->i_dest->a_base_reg);
	    if (src1Imm) {
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    instr->i_src->a_displacement,
		    SLOT_TYPE_CONSTANT,
		    (OperandIsAddress(instr->i_src) ? SLOT_ADDRESS_FLAG : 0));
	    } else if (src1Stack) {
		base = myBase - instr->i_src->a_displacement / ADDRESS_SIZE;
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    STACK_CONTENTS(bbl, base), STACK_TYPE(bbl, base),
		    STACK_FLAGS(bbl, base) & SLOT_ADDRESS_FLAG);
		if (destFullReg) {
		    AssignRegPtr(bbl, reg, SLOT_STACK_FLAG | (STACK_FLAGS(bbl,
				base)), base);
		}
	    } else if (src1Reg) {
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    REG_CONTENTS(bbl, instr->i_src->a_base_reg),
		    REG_TYPE(bbl, instr->i_src->a_base_reg),
		    REG_FLAGS(bbl,instr->i_src->a_base_reg)&SLOT_ADDRESS_FLAG);
		if (src1FullReg && destFullReg) {
		    AssignRegPtr(bbl, reg,
			RegisterToSlotFlag(instr->i_src->a_base_reg)
			    | (REG_FLAGS(bbl, instr->i_src->a_base_reg)), 0);
		}
	    } else {
		goto noEval;
	    }

	} else {
	    goto noEval;
	}

	break;

	/*
	 *                                   ADD
	 */

    case CLASS_ADD:

	NO_EVAL_IF_SOURCE1_IS_ADDRESS();
	NO_EVAL_IF_DEST_IS_ADDRESS();

	if (destReg) {
	    regContents = REG_CONTENTS(bbl, instr->i_dest->a_base_reg);
	    if (src1Imm) {
		if (src1Addr) {
		    goto noEval;
		}
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    instr->i_src->a_displacement + regContents,
		    SLOT_TYPE_CONSTANT, 0);
	    } else if (src1Stack) {
		base = myBase - instr->i_src->a_displacement / ADDRESS_SIZE;
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    STACK_CONTENTS(bbl, base) + regContents, STACK_TYPE(bbl,
			base), 0);
	    } else if (src1Reg) {
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    REG_CONTENTS(bbl,
			instr->i_src->a_base_reg) + regContents,
		    REG_TYPE(bbl, instr->i_src->a_base_reg), 0);
	    } else {
		goto noEval;
	    }

	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    stackContents = STACK_CONTENTS(bbl, base);
	    if (src1Imm) {
		if (src1Addr) {
		    goto noEval;
		}
		AssignStack(bbl, base,
		    stackContents + instr->i_src->a_displacement,
		    SLOT_TYPE_CONSTANT, 0);
	    } else if (src1Reg) {
		int reg = instr->i_src->a_base_reg;
		AssignStack(bbl, base, REG_CONTENTS(bbl, reg) + stackContents,
		    REG_TYPE(bbl, reg), 0);
	    } else if (src1Stack) {
		int base2 = myBase - instr->i_src->a_displacement/ADDRESS_SIZE;
		AssignStack(bbl, base, STACK_CONTENTS(bbl,
			base2) + stackContents, STACK_TYPE(bbl, base2), 0);
	    } else {
		goto noEval;
	    }
	} else {
	    goto noEval;
	}

	break;


	/*
	 *                                   SUB
	 */

    case CLASS_SUB:

	NO_EVAL_IF_SOURCE1_IS_ADDRESS();
	NO_EVAL_IF_DEST_IS_ADDRESS();

	if (destReg) {
	    regContents = REG_CONTENTS(bbl, instr->i_dest->a_base_reg);
	    if (src1Imm) {
		if (src1Addr) {
		    goto noEval;
		}
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    regContents - instr->i_src->a_displacement,
		    SLOT_TYPE_CONSTANT, 0);
	    } else if (src1Stack) {
		base = myBase - instr->i_src->a_displacement / ADDRESS_SIZE;
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    regContents - STACK_CONTENTS(bbl, base), STACK_TYPE(bbl,
			base), 0);
	    } else if (src1Reg) {
		AssignRegister(bbl, instr->i_dest->a_base_reg,
		    regContents - REG_CONTENTS(bbl,
			instr->i_src->a_base_reg), REG_TYPE(bbl,
			instr->i_src->a_base_reg), 0);
	    } else {
		goto noEval;
	    }

	} else if (destStack) {
	    base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
	    stackContents = STACK_CONTENTS(bbl, base);
	    if (src1Imm) {
		if (src1Addr) {
		    goto noEval;
		}
		AssignStack(bbl, base,
		    stackContents - instr->i_src->a_displacement,
		    SLOT_TYPE_CONSTANT, 0);
	    } else if (src1Reg) {
		int reg = instr->i_src->a_base_reg;
		AssignStack(bbl, base, stackContents - REG_CONTENTS(bbl, reg),
		    REG_TYPE(bbl, reg), 0);
	    } else if (src1Stack) {
		int base2 = myBase - instr->i_src->a_displacement/ADDRESS_SIZE;
		AssignStack(bbl, base, stackContents - STACK_CONTENTS(bbl,
			base2), STACK_TYPE(bbl, base2), 0);
	    } else {
		goto noEval;
	    }

	} else {
	    goto noEval;
	}

	break;



    default:

	/*
	 * If we don't know how to evaluate the instruction, then we should set
	 * all the destination registers to SLOT_TYPE_UNKNOWN, since they are
	 * probably going to get defined when the instruction really executes.
	 * If the destination is somewhere on the stack then we should also set
	 * the slot to unknown.
	 */

noEval:
	numInstructionsEvaluated = numInstructionsEvaluated - 1;
	numInstructionsNotEvaluated = numInstructionsNotEvaluated + 1;

	if (instr->i_dest) {
	    if (destReg) {
		AssignRegister(bbl, instr->i_dest->a_base_reg, 0,
		    SLOT_TYPE_UNKNOWN, 0);
	    }
	    if (destStack) {
		base = myBase - instr->i_dest->a_displacement / ADDRESS_SIZE;
		AssignStack(bbl, base, 0, SLOT_TYPE_UNKNOWN, 0);
		if (INSTR_HAS_FLAG(instr, INSTR_USES_FP_BUT_COMPENSATED_FLAG)) {
		    BBLStackBecomesUnknown(bbl);
		}
	    }
	}

	if (InstrHasImplicitDest(instr)) {
	    AssignRegister(bbl, INSTR_IMPLICIT_DEST1(instr), 0,
		SLOT_TYPE_UNKNOWN, 0);
	}

	if (InstrHasImplicitDest2(instr)) {
	    AssignRegister(bbl, INSTR_IMPLICIT_DEST2(instr), 0,
		SLOT_TYPE_UNKNOWN, 0);
	}

	pushBytes = InstrPushBytes(instr);
	if (pushBytes != 0) {
	    for (i = 0; i < pushBytes / ADDRESS_SIZE; i++) {
		AssignStack(bbl, stackBase + 1 + i, 0, SLOT_TYPE_UNKNOWN, 0);
	    }
	}

	if (instr->i_prefixes) {
	    BBLRegsBecomeUnknown(bbl);
	}

	break;

    }					/* end switch over instruction type */

    return target;
}



/*
 *  RegisterToIndex(aReg) --
 *
 *  Returns the index of 'aReg' in the register snapshot for basic blocks.
 */
int RegisterToIndex(int aReg) {

    switch (aReg) {

    case reg_a:
	return REGISTER_INDEX_EAX;
	break;
    case reg_b:
	return REGISTER_INDEX_EBX;
	break;
    case reg_c:
	return REGISTER_INDEX_ECX;
	break;
    case reg_d:
	return REGISTER_INDEX_EDX;
	break;
    case reg_sp:
	return REGISTER_INDEX_ESP;
	break;
    case reg_bp:
	return REGISTER_INDEX_EBP;
	break;
    case reg_di:
	return REGISTER_INDEX_EDI;
	break;
    case reg_si:
	return REGISTER_INDEX_ESI;
	break;

	/* 32-bit general purpose registers */
    case reg_a | reg_32:
	return REGISTER_INDEX_EAX;
	break;
    case reg_b | reg_32:
	return REGISTER_INDEX_EBX;
	break;
    case reg_c | reg_32:
	return REGISTER_INDEX_ECX;
	break;
    case reg_d | reg_32:
	return REGISTER_INDEX_EDX;
	break;
    case reg_si | reg_32:
	return REGISTER_INDEX_ESI;
	break;
    case reg_di | reg_32:
	return REGISTER_INDEX_EDI;
	break;
    case reg_bp | reg_32:
	return REGISTER_INDEX_EBP;
	break;
    case reg_sp | reg_32:
	return REGISTER_INDEX_ESP;
	break;

	/* 16-bit general purpose registers */
    case reg_a | reg_16:
	return REGISTER_INDEX_EAX;
	break;
    case reg_b | reg_16:
	return REGISTER_INDEX_EBX;
	break;
    case reg_c | reg_16:
	return REGISTER_INDEX_ECX;
	break;
    case reg_d | reg_16:
	return REGISTER_INDEX_EDX;
	break;
    case reg_si | reg_16:
	return REGISTER_INDEX_ESI;
	break;
    case reg_di | reg_16:
	return REGISTER_INDEX_EDI;
	break;
    case reg_bp | reg_16:
	return REGISTER_INDEX_EBP;
	break;
    case reg_sp | reg_16:
	return REGISTER_INDEX_ESP;
	break;

	/* 8-bit general purpose registers */
    case reg_a | reg_8l:
	return REGISTER_INDEX_EAX;
	break;
    case reg_b | reg_8l:
	return REGISTER_INDEX_EBX;
	break;
    case reg_c | reg_8l:
	return REGISTER_INDEX_ECX;
	break;
    case reg_d | reg_8l:
	return REGISTER_INDEX_EDX;
	break;
    case reg_a | reg_8h:
	return REGISTER_INDEX_EAX;
	break;
    case reg_b | reg_8h:
	return REGISTER_INDEX_EBX;
	break;
    case reg_c | reg_8h:
	return REGISTER_INDEX_ECX;
	break;
    case reg_d | reg_8h:
	return REGISTER_INDEX_EDX;
	break;

	/* Special case for the indirect dx register */
    case reg_d | reg_16 | reg_indirect:
	return REGISTER_INDEX_EDX;
	break;

    }					/* end switch over register */

    printf("[RegisterToIndex] Could not map register to index.\n");

    return 0;
}



/*
 *  IndexToRegister(idx) --
 *
 *  Provides the reverse functionality of RegisterToIndex.
 */
int IndexToRegister(int idx) {
    switch (idx) {
    case REGISTER_INDEX_EAX:  return reg_a | reg_32;
    case REGISTER_INDEX_EBX:  return reg_b | reg_32;
    case REGISTER_INDEX_ECX:  return reg_c | reg_32;
    case REGISTER_INDEX_EDX:  return reg_d | reg_32;
    case REGISTER_INDEX_ESI:  return reg_si | reg_32;
    case REGISTER_INDEX_EDI:  return reg_di | reg_32;
    case REGISTER_INDEX_ESP:  return reg_sp | reg_32;
    case REGISTER_INDEX_EBP:  return reg_bp | reg_32;
    default:
	printf("[IndexToRegister] Give index not in the range [0-7].\n");
    }
    return -1;
}



/*
 *  OperandIsUndefinedOrUnknown(instr,op) --
 *
 *  Returns true if the operand is a stack slot or a register that is of type
 *  SLOT_TYPE_UNDEFINED or SLOT_TYPE_UNKNOWN.
 */
static bool OperandIsUndefinedOrUnknown(Instruction *instr, Operand *op) {
    Block *bbl = instr->i_Bblock;

    if (op == NULL) {
	return false;
    }

    if (!PROPAGATE_ADDRESSES) {
	if (OperandIsAddress(op)) {
	    return true;
	}
    }

    if (OperandIsGPRegister(op)) {
	if (REG_TYPE(bbl, op->a_base_reg) == SLOT_TYPE_UNKNOWN ||
	    REG_TYPE(bbl, op->a_base_reg) == SLOT_TYPE_UNDEFINED) {
	    return true;
	}
    }

    if (OperandAccessesStackFromFramePointer(op)) {
	base = myBase - (op->a_displacement / ADDRESS_SIZE);
	if (STACK_TYPE(bbl, base) == SLOT_TYPE_UNKNOWN ||
	    STACK_TYPE(bbl, base) == SLOT_TYPE_UNDEFINED) {
	    return true;
	}
    }

    if (OperandIsAbsoluteMemoryLocation(op)) {
	return true;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_INDIRECT_FLAG)) {
	return true;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)
	&& !OperandAccessesStackFromFramePointer(op)) {
	return true;
    }

    return false;
}



/*
 *  RegisterIsUndefinedOrUnknown(instr,aReg) --
 *
 *  Returns true if the register is undefined or unknown
 *  in the basic block which contains the instruction 'instr'
 */
static bool RegisterIsUndefinedOrUnknown(Instruction *instr, int aReg) {
    int regIndex;

    if (RegisterIsFloatingPoint(aReg)) {
	return true;
    }

    if (aReg == reg_all) {
	return true;
    }

    regIndex = RegisterToIndex(aReg);

    if ((instr->i_Bblock->b_regshot[regIndex].v_type) == SLOT_TYPE_UNKNOWN) {
	return true;
    }

    if ((instr->i_Bblock->b_regshot[regIndex].v_type) == SLOT_TYPE_UNDEFINED) {
	return true;
    }

    return false;
}

/*
 *  ReplaceOperandWithConstant(instr,op) --
 *
 *  Attempts to replace a use of the stack or a register with the constant that
 *  is in that location.
 */
static int ReplaceOperandWithConstant(Instruction *instr, Operand *op) {
    int base2, val = 0, type, flags = 0;
    char *oldInstr = PrintInstruction(instr);
    char oldInstrBytes[16];
    Block *bbl = instr->i_Bblock;
    int numSimplified = 0;

    InstrAssemble(oldInstrBytes, instr);

    if (OPERAND_HAS_FLAG(op, OPERAND_INDIRECT_FLAG)) {
        goto replaceCleanup;
    }

    if (OperandAccessesStackFromFramePointer(op)) {
	base2 = myBase - op->a_displacement / ADDRESS_SIZE;
	if (base2 >= 0) {
	    val = STACK_CONTENTS(bbl, base2);
	    type = STACK_TYPE(bbl, base2);
	    flags = STACK_FLAGS(bbl, base2);
	} else {
	    type = SLOT_TYPE_UNKNOWN;
	}

	if (type == SLOT_TYPE_CONSTANT) {
	    if (ReplaceStack(instr, op, val)) {
            if (WANT_REPORT(RP_LGLOOP)) {
                char *newInstr = PrintInstruction(instr);
                REPORT(RP_LGLOOP, "%08X/%-5d Replaced %s with %s in bbl %d",
                    instr->i_address, instr->i_id, oldInstr, newInstr,
                    instr->i_Bblock->b_id);
                free(newInstr);
            }            
		if (flags & SLOT_ADDRESS_FLAG) {
		    OPERAND_SET_FLAG(op, OPERAND_RELOCATABLE_ABS_FLAG);
		    CreateUPMForAddressPropagation(instr, op, val);
		    INSTR_SET_FLAG(instr, INSTR_RELOCATABLE_FLAG);			 		    
		}
		numSimplified++;
	    } else {
            DEBUG('q',
                "%08X/%-5d Failed to replace %s with const %d in %s",
                instr->i_address, instr->i_id, oldInstr, val,
                op == instr->i_src ? "source1" : "source2");
	    }
	}
    }

    if (OperandIsGPRegister(op)) {
	if (OperandIsStackPointer(op)) {
	    goto replaceCleanup;
	}

	val = REG_CONTENTS(bbl, op->a_base_reg);
	type = REG_TYPE(bbl, op->a_base_reg);
	flags = REG_FLAGS(bbl, op->a_base_reg);
	if (type == SLOT_TYPE_CONSTANT) {
	    if (ReplaceReg(instr, op, val)) {
            if (WANT_REPORT(RP_LGLOOP)) {
                char *newInstr = PrintInstruction(instr);
                REPORT(RP_LGLOOP, "%08X/%-5d Replaced %s with %s in bbl %d",
                    instr->i_address, instr->i_id, oldInstr, newInstr,
                    instr->i_Bblock->b_id);
                free(newInstr);
            }            
		if (flags & SLOT_ADDRESS_FLAG) {
		    OPERAND_SET_FLAG(op, OPERAND_RELOCATABLE_ABS_FLAG);
		    if (val == 0) {
			    op->a_displacement = 0;
		    } else {
			    CreateUPMForAddressPropagation(instr, op, val);
			    INSTR_SET_FLAG(instr,INSTR_RELOCATABLE_FLAG);			 
		    }
		}
		numSimplified++;
	    } else {
            DEBUG('q',
                "%08X/%-5d Failed to replace %s with const %d in %s",
                instr->i_address, instr->i_id, oldInstr, val,
                op == instr->i_src ? "source1" : "source2");
	    }
	}
    }
    // cml: it looks to me like the purpose of this is to update the
    // offset in the instruction where the operand lies, however this is
    // unnecessary with the new upm system since it actually tracks the
    // operands and not their offsets
// if (numSimplified > 0) {
//    InstrAssemble(newInstrBytes, instr);
//    UpdateUPM(oldInstrBytes, newInstrBytes, instr, upms);
// }

replaceCleanup:
    free(oldInstr);
    return numSimplified;
}

/*
 *  ReplaceOperandWithRegister(instr,op) --
 *
 *  Replaces a memory reference with a register use when possible.
 *  Currently, we only try and replace memory references in the stack frame.
 */
static int ReplaceOperandWithRegister(Instruction *instr, Operand *op) {
    Block *bbl = instr->i_Bblock;
    Function *fun = INSTR_FUNCTION(instr);
    char *oldInstr = PrintInstruction(instr);
    char oldInstrBytes[16];
    int numSimplified = 0;
    int reg = -1;
    int base2;
    int i;

    InstrAssemble(oldInstrBytes, instr);

    if ((OPERAND_HAS_FLAG(op, OPERAND_INDIRECT_FLAG))
	|| !(OperandAccessesStackFromFramePointer(op))) {
	free(oldInstr);
	return numSimplified;
    }

    base2 = myBase - op->a_displacement / ADDRESS_SIZE;

    /*
     * First, lets check if any of the x86-registers.have the same value as the
     * stack location. If so, we can swap out the memory reference with this
     * register.
     */

    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	if (SLOT_HAS_FLAG_DIRECT(bbl->b_regshot[i], SLOT_STACK_FLAG)) {
	    if (bbl->b_regshot[i].v_stackOffset == base2) {
		reg = IndexToRegister(i);
		break;
	    }
	}
    }

    /*
     * If none of the registers were aliased to this stack location, we can
     * next check if the stack location is aliased to any of the registers.
     * This can be done by examining the flags that may be set (e.g:
     * SLOT_EAX_FLAG) for this stack location. If one of the flags is set, then
     * the stack location is aliasing a register and we can interchange the
     * two.
     */

    if (reg == -1) {
	if (base2 <=
	    ((fun->f_stackFrameSize + fun->f_callerFrameSize) / ADDRESS_SIZE)) {
	    if (base2 > 0) {
		reg = FlagsToReg(bbl->b_stackshot[base2].v_flags);
	    }
	}
    }

    if (reg != -1) {
	if (ReplaceStackWithRegister(instr, op, reg)) {
	    numSimplified++;
	    if (WANT_REPORT(RP_LGLOOP)) {
            char *newInstr = PrintInstruction(instr);
            REPORT(RP_LGLOOP, "%08X/%-5d Replaced %s with %s",
                instr->i_address, instr->i_id, oldInstr, newInstr);
            free(newInstr);
	    }
	} else {
	    DEBUG('q',
		"%08X/%-5d Failed to replace %s with reg %d in %s",
		instr->i_address, instr->i_id, oldInstr, reg,
		op == instr->i_src ? "source1" : "source2");
	}
    }

    free(oldInstr);

    return numSimplified;
}



/*
 *  ReplaceStack(instr,op,val) --
 *
 *  Replaces the use of a stack frame reference with an immediate value.
 */
static bool ReplaceStack(Instruction *instr, Operand *op, int val) {
    int oldVal = op->a_displacement;
    UINT32 oldFlags = op->a_flags;
    Operand *dest = instr->i_dest;
    UINT32 destFlags = 0;

    if (dest) {
	destFlags = dest->a_flags;
    }

    OPERAND_SET_FLAG(op, OPERAND_IMM_FLAG);
    op->a_base_reg = 0;
    op->a_displacement = val;
    OPERAND_REMOVE_FLAG(op, OPERAND_OPG_FLAG);
    OPERAND_REMOVE_FLAG(op, OPERAND_OPREG_FLAG);
    OPERAND_REMOVE_FLAG(op, OPERAND_REGISTER_INDIRECT);
    OPERAND_REMOVE_FLAG(op, OPERAND_REGISTER_MODRM);
    OPERAND_REMOVE_FLAG(op, OPERAND_BASE_FLAG);

    if (!isValidInstr(instr)) {
	if (dest && OperandIsGPRegister(dest)) {
	    if (OPERAND_HAS_FLAG(dest, OPERAND_OPG_FLAG)) {
		OPERAND_REMOVE_FLAG(dest, OPERAND_OPG_FLAG);
		OPERAND_REMOVE_FLAG(dest, OPERAND_VMODE_FLAG);
		OPERAND_REMOVE_FLAG(dest, OPERAND_BMODE_FLAG);
		OPERAND_REMOVE_FLAG(dest, OPERAND_WMODE_FLAG);
		OPERAND_REMOVE_FLAG(dest, OPERAND_DMODE_FLAG);
		OPERAND_SET_FLAG(dest, OPERAND_OPREG_FLAG);
		if (!isValidInstr(instr)) {
		    goto restoreStackInstr;
		}
	    } else if (OPERAND_HAS_FLAG(dest, OPERAND_OPREG_FLAG)) {
		OPERAND_REMOVE_FLAG(dest, OPERAND_OPREG_FLAG);
		OPERAND_SET_FLAG(dest, OPERAND_OPG_FLAG);
		if (!isValidInstr(instr)) {
		    goto restoreStackInstr;
		}
	    }
	} else {
	    goto restoreStackInstr;
	}
    }

    UpdateInstr(instr);
    return true;

restoreStackInstr:

    op->a_base_reg = reg_bp | reg_32;
    op->a_displacement = oldVal;
    op->a_flags = oldFlags;
    if (dest) {
	dest->a_flags = destFlags;
    }

    return false;
}



/*
 *  ReplaceReg(instr,op,val) --
 *
 *  Replaces the use of a register with an immediate value.
 */
static bool ReplaceReg(Instruction *instr, Operand *op, int val) {
    int oldVal = op->a_displacement;
    int oldReg;
    UINT32 oldFlags = op->a_flags;

    oldReg = op->a_base_reg;

    OPERAND_SET_FLAG(op, OPERAND_IMM_FLAG);
    op->a_displacement = val;
    op->a_base_reg = 0;
    OPERAND_REMOVE_FLAG(op, OPERAND_BASE_FLAG);
    OPERAND_REMOVE_FLAG(op, OPERAND_OPG_FLAG);
    OPERAND_REMOVE_FLAG(op, OPERAND_OPREG_FLAG);
    OPERAND_REMOVE_FLAG(op, OPERAND_REGISTER_INDIRECT);
    OPERAND_REMOVE_FLAG(op, OPERAND_REGISTER_MODRM);
    OPERAND_REMOVE_FLAG(op, OPERAND_BASE_FLAG);

    if (!isValidInstr(instr)) {
	op->a_base_reg = oldReg;
	op->a_displacement = oldVal;
	op->a_flags = oldFlags;
	return false;
    }

    UpdateInstr(instr);

    return true;
}



/*
 *  BBLCopyRegshot(bbl) --
 *
 *  Copies the register snapshot at the end of 'bbl' and returns a pointer to
 *  the copy.
 */
static Value *BBLCopyRegshot(Block *bbl) {
    int i;

    Value *result = zalloc(NUMBER_OF_GP_REGISTERS * sizeof(Value));

    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	result[i].v_type = bbl->b_regshot[i].v_type;
	result[i].v_value = bbl->b_regshot[i].v_value;
    }

    return result;
}



/*
 *  BBLCopyStackshot(bbl) --
 *
 *  Copies the stack snapshot at the end of 'bbl' and returns a pointer to
 *  the copy.
 */
static Value *BBLCopyStackshot(Block *bbl) {
    int sz = (bbl->b_fun->f_stackFrameSize +
	bbl->b_fun->f_callerFrameSize) / ADDRESS_SIZE;
    int i;

    Value *result = zalloc(sz * sizeof(Value));

    for (i = 0; i < sz; i++) {
	result[i].v_type = bbl->b_stackshot[i].v_type;
	result[i].v_value = bbl->b_stackshot[i].v_value;
    }

    return result;
}



/*
 *  BBLRegContentsAreEqual(bbl,regs) --
 *
 *  Compares the values/types of the registers in the value slot array 'regs'
 *  with the values/types in the basic block 'bbl'. If there are no
 *  discrepancies then true is returned.
 */
static bool BBLRegContentsAreEqual(Block *bbl, Value *regs) {
    int i;

    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	if (!ValueSlotsAreEqual(&bbl->b_regshot[i], &regs[i])) {
	    return false;
	}
    }

    return true;
}



/*
 *  BBLStackContentsAreEqual(bbl,stack) --
 *
 *  Compares the values/types of the stack in the value slot array 'stack'
 *  with the values/types in the basic block 'bbl'. If there are no
 *  discrepancies then true is returned.
 */
static bool BBLStackContentsAreEqual(Block *bbl, Value *stack) {
    int sz = (bbl->b_fun->f_stackFrameSize +
	bbl->b_fun->f_callerFrameSize) / ADDRESS_SIZE;
    int i;

    for (i = 0; i < sz; i++) {
	if (!ValueSlotsAreEqual(&bbl->b_stackshot[i], &stack[i])) {
	    return false;
	}
    }

    return true;
}



/*
 *  ValueSlotsAreEqual(x,y) --
 *
 *  Returns true if the two supplied value slots are equal.
 */
static bool ValueSlotsAreEqual(Value *x, Value *y) {

    if (x->v_type == SLOT_TYPE_UNDEFINED) {
	if (y->v_type == SLOT_TYPE_UNDEFINED) {
	    return true;
	}
	return false;
    }

    if (x->v_type == SLOT_TYPE_UNKNOWN) {
	if (y->v_type == SLOT_TYPE_UNKNOWN) {
	    return true;
	}
	return false;
    }

    if (x->v_value != y->v_value) {
	return false;
    }

    return true;
}



/*
 *  BBLStackBecomesUnknown(bbl) --
 *
 *  Sets all the stack slots to unknown in the basic block.
 */
static void BBLStackBecomesUnknown(Block *bbl) {
    Function *fun = bbl->b_fun;
    int i;

    for (i = 0;
	i < (fun->f_stackFrameSize + fun->f_callerFrameSize) / ADDRESS_SIZE;
	i++)
	AssignStack(bbl, i, 0, SLOT_TYPE_UNKNOWN, 0);

    return;
}



/*
 *  BBLRegsBecomeUnknown(bbl) --
 *
 *  Sets all the register slots to unknown in the basic block.
 */
static void BBLRegsBecomeUnknown(Block *bbl) {
    int i;

    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	bbl->b_regshot[i].v_value = 0;
	bbl->b_regshot[i].v_type = SLOT_TYPE_UNKNOWN;
	bbl->b_regshot[i].v_flags = 0;
    }

    return;
}



/*
 *  InstrDefinesNothing(instr) --
 *
 *  Returns true if the instruction defines nothing (excluding PSW).
 */
bool InstrDefinesNothing(Instruction *instr) {
    X86op *disEntry = InstrGetOpcodeEntry(instr);

    if (instr->i_dest) {
	return false;
    }

    if (disEntry->x_psw & S1D) {
	if (instr->i_src) {
	    return false;
	}
    }

    if (disEntry->x_psw & S2D) {
	if (instr->i_src2) {
	    return false;
	}
    }

    if (INSTR_IMPLICIT_DEST1(instr)) {
	return false;
    }
    if (INSTR_IMPLICIT_DEST2(instr)) {
	return false;
    }

    return true;
}



/*
 *  BBLGetConditionalEqualityInformation(bbl) --
 *
 *  If any predecessors of 'bbl' branched on an equality jump to 'bbl' after
 *  doing a comparison with a register or memory location, then we can infer
 *  some things about the contents of these locations.
 */
static void BBLGetConditionalEqualityInformation(Block *bbl) {
    Block *pred, *target;
    Instruction *iterator;
    Edge *propagateEdge = NULL;
    PSW_BITVECTOR used;

	//return;
    /*
     * If control reached the block multiple ways, then it is unlikely
     * we will be able to infer anything. There is a chance that all
     * predecessors used the same comparison/jump combination, but the
     * case will be not considered below.
     */

    if (BBLNumberOfPredecessors(bbl) > 1) {
	return;
    }

    if (BBLNumberOfPredecessors(bbl) == 0) {
	return;
    }

    pred = NODE_FROM(bbl->b_pred);
    iterator = pred->b_last;

    if ((pred == NULL) || (iterator == NULL)) {
	return;
    }

    if (iterator->i_Bblock->b_type != BT_BRANCH) {
	return;
    }

    if (iterator->i_class == CLASS_JE) {
	propagateEdge = BBLGetTrueEdge(pred);
    } else if (iterator->i_class == CLASS_JNE) {
	propagateEdge = BBLGetFalseEdge(pred);
    }

    if (propagateEdge == NULL) {
	return;
    }

    if (pred->b_first == iterator) {
	return;
    }

    used = ComputePSWUsed(iterator);
    iterator = iterator->i_prev;

    target = propagateEdge->e_to;

    if (target != bbl) {
	return;
    }

    while (iterator->i_Bblock == pred) {
	if (used & ComputePSWDefined(iterator)) {

	    /*
	     *                            CMP IDIOM
	     */

	    if (iterator->i_class == CLASS_CMP) {
		if (OperandIsImmediate(iterator->i_src)) {
		    if (OperandIsGPRegister(iterator->i_src2)) {
			AssignRegister(target,
			    iterator->i_src2->a_base_reg,
			    iterator->i_src->a_displacement,
			    SLOT_TYPE_CONSTANT,
			    (OperandIsAddress(iterator->i_src) ?
				SLOT_ADDRESS_FLAG : 0));
		    }
		    if (OperandAccessesStackFromFramePointer(iterator->i_src2)){
			if (((iterator->i_src2->a_displacement) %
				ADDRESS_SIZE) == 0) {
			    base = myBase -
				iterator->i_src2->a_displacement / ADDRESS_SIZE;
			    AssignStack(target, base,
				iterator->i_src->a_displacement,
				SLOT_TYPE_CONSTANT,
				(OperandIsAddress(iterator->i_src) ?
				    SLOT_ADDRESS_FLAG : 0));
			}
		    }
		}
		if (OperandIsImmediate(iterator->i_src2)) {
		    if (OperandIsGPRegister(iterator->i_src)) {
			AssignRegister(target,
			    iterator->i_src->a_base_reg,
			    iterator->i_src2->a_displacement,
			    SLOT_TYPE_CONSTANT,
			    (OperandIsAddress(iterator->i_src2) ?
				SLOT_ADDRESS_FLAG : 0));
		    }
		    if (OperandAccessesStackFromFramePointer(iterator->i_src)) {
			if (
			(iterator->i_src->a_displacement % ADDRESS_SIZE) == 0) {
			    base = myBase -
				iterator->i_src->a_displacement / ADDRESS_SIZE;
			    AssignStack(target, base,
				iterator->i_src2->a_displacement,
				SLOT_TYPE_CONSTANT,
				(OperandIsAddress(iterator->i_src2) ?
				    SLOT_ADDRESS_FLAG : 0));
			}
		    }
		}
		return;

	    } else {
		/*
		 *  Not CMP
		 */
		return;
	    }
	}			/* end test for setting bit that jump uses */
	iterator = iterator->i_prev;
    }				/* end iteration over instructions in bbl */

    return;
}



/*
 *  BBLStackAndRegistersBecomeUnknown(bbl) --
 *
 *  Sets all stack and register slots to type SLOT_UNKNOWN in the basic block
 */
static void BBLStackAndRegistersBecomeUnknown(Block *bbl) {
    BBLStackBecomesUnknown(bbl);
    BBLRegsBecomeUnknown(bbl);
}



/*
 *  InitializeStackAndRegs(bbl) --
 *
 *  Initializes the model of the stack and registers
 *  for the supplied basic block.
 */
static void InitializeStackAndRegs(Block *bbl) {
    int stackType;
    Function *fun = bbl->b_fun;
    int i;

    if (BBLHasHellPredecessor(bbl) ||
	FUNCTION_HAS_FLAG(bbl->b_fun, FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG)) {
	stackType = SLOT_TYPE_UNKNOWN;
    } else {
	stackType = SLOT_TYPE_UNDEFINED;
    }

    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	if (fun->f_stackFrameSize <= MAXIMUM_AR_SIZE) {
	    for (i = 0;
		i <
		(fun->f_stackFrameSize +
		    fun->f_callerFrameSize) / ADDRESS_SIZE; i++) {
		bbl->b_stackshot[i].v_type = stackType;
	    }
	}
    }

    if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
	for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	    bbl->b_regshot[i].v_type = stackType;
	}
    }

    BBL_SET_FLAG(bbl, BBL_TEMP_FLAG);

    return;
}



/*
 *  CreateUPMForAddressPropagation(instr,op,val) --
 *
 *  When an address is propagated into an instruction, we need to create a
 *  patch entry for this address to make sure it is updated at relocation time.
 */
static void CreateUPMForAddressPropagation(Instruction *instr, Operand *op,
    int val) {
    Section *sec = NULL;

    if (AddressToInstr(val) != NULL) {
        AddPatchInstrAddressIntoInstr(AddressToInstr(val), instr, op);
    } else {
        /* 
         * Addresses properly. Inlined the code instead of calling
         * AddressToSection(). Failed to propagate address constants
         * before this properly
         */

        FOREACH_SECTION(sec) {
            if ((val < (UINT32) (sec->s_address + (int)(sec->s_newdatasize)))
                &&(val >= (UINT32)(sec->s_address))) {
                break;
            }
        }

        CASSERT(sec != NULL, "AddressToInstr() was NULL");
    
        AddPatchSectionAddressIntoInstr(sec, val - sec->s_address, instr, op);
    }
}



/*
 *  RegisterToSlotFlag(aReg) --
 *
 *  Converts 'aReg' into a slot flag, and returns the result.
 */
static int RegisterToSlotFlag(int aReg) {

    switch (aReg) {
    case (reg_a | reg_32):
	return SLOT_EAX_FLAG;
	break;
    case (reg_b | reg_32):
	return SLOT_EBX_FLAG;
	break;
    case (reg_c | reg_32):
	return SLOT_ECX_FLAG;
	break;
    case (reg_d | reg_32):
	return SLOT_EDX_FLAG;
	break;
    case (reg_si | reg_32):
	return SLOT_ESI_FLAG;
	break;
    case (reg_di | reg_32):
	return SLOT_EDI_FLAG;
	break;
    case (reg_sp | reg_32):
	return SLOT_ESP_FLAG;
	break;
    case (reg_bp | reg_32):
	return SLOT_EBP_FLAG;
	break;
    default:
	printf("[RegisterToSlotFlag] Unrecognized register %d.\n", aReg);
	break;
    }

    printf("[RegisterToSlotFlag] Unrecognized register %d.\n", aReg);

    return -1;
}



/*
 *  ReportCPResults(funName,constants,registers,branches) --
 *
 *  ReportCPResults() should be called with the name of the function, along
 *  with the number of constants propagated, the number of registers
 *  propagated (loads/stores forwarded) and the number of conditional jumps
 *  resolved. The constant REPORT_CP_HEADER should be used as the last three
 *  arguments to print out the header for the table. The constant
 *  REPORT_CP_TRAILER should be used as the last three arguments to print
 *  out summary information.
 */
static void ReportCPResults(char *funName, int constants, int registers,
    int branches) {
    static int totalSimplified = 0;
    static int totalConditionals = 0;
    static int totalForwards = 0;
    bool trailer, header;
    char *name = zalloc(strlen(funName) + 5);

    trailer = false;
    if (constants == REPORT_CP_TRAILER) {
	trailer = true;
    }

    header = false;
    if (constants == REPORT_CP_HEADER) {
	header = true;
    }

    strcpy(name, funName);
    if (strlen(funName) >= 30) {
	name[30] = 0;
    }

    if (header) {
        totalSimplified = totalConditionals = totalForwards = 0;
        REPORT(RP_SMLOOP, "          Function             | Constants |"
            " Forwards | Branches");
        REPORT(RP_SMLOOP, "===============================|===========|"
            "==========|==============");
        free(name);
        return;
    }

    if (!trailer) {
	totalSimplified += constants;
	totalConditionals += branches;
	totalForwards += registers;
    } else {
	constants = totalSimplified;
	registers = totalForwards;
	branches = totalConditionals;
	sprintf(name, "[Total]");
    }

    if ((constants > 0) || (registers > 0) || (branches > 0)) {
        REPORT(RP_SMLOOP, "%-30s |      %-4d |    %-4d  |    %-4d",
            name, constants, registers, branches);
    }

    if (trailer) {
        REPORT(RP_DETAIL, "%5d constants propagated", totalSimplified);
        REPORT(RP_DETAIL, "%5d registers propagated", totalForwards);
        REPORT(RP_DETAIL, "%5d cond jmps eliminated", totalConditionals);
    }

    free(name);

    return;
}



/*
 *  ReplaceStackWithRegister(instr,op,reg) --
 *
 *  Replaces a stack frame reference with the register 'reg'.
 */
static bool ReplaceStackWithRegister(Instruction *instr, Operand *op, int reg) {
    Operand *newOp;

    /*
     * Try #1: First try using the OP_E addressing mode.
     * We'll create a new operand and save the old one.
     */

    newOp = OperandCreate();
    newOp->a_base_reg = reg;
    newOp->a_flags =
	(OPERAND_BASE_FLAG | OPERAND_VMODE_FLAG | OPERAND_REGISTER_MODRM);
    if (instr->i_src == op) {
	instr->i_src = newOp;
    }
    if (instr->i_src2 == op) {
	instr->i_src2 = newOp;
    }
    if (isValidInstr(instr)) {
	UpdateInstr(instr);
	return true;
    }
    if (instr->i_src == newOp) {
	instr->i_src = op;
    }
    if (instr->i_src2 == newOp) {
	instr->i_src2 = op;
    }


    /*
     * Try #2: We'll try out the OP_REG addressing mode.
     */

    newOp->a_flags =
	(OPERAND_BASE_FLAG | OPERAND_VMODE_FLAG | OPERAND_OPREG_FLAG);
    if (instr->i_src == op) {
	instr->i_src = newOp;
    }
    if (instr->i_src2 == op) {
	instr->i_src2 = newOp;
    }
    if (isValidInstr(instr)) {
	UpdateInstr(instr);
	return true;
    }
    if (instr->i_src == newOp) {
	instr->i_src = op;
    }
    if (instr->i_src2 == newOp) {
	instr->i_src2 = op;
    }

    /*
     * Try #3: We'll try out the OP_G addressing mode.
     */

    newOp->a_flags =
	(OPERAND_BASE_FLAG | OPERAND_VMODE_FLAG | OPERAND_OPG_FLAG);
    if (instr->i_src == op) {
	instr->i_src = newOp;
    }
    if (instr->i_src2 == op) {
	instr->i_src2 = newOp;
    }
    if (isValidInstr(instr)) {
	UpdateInstr(instr);
	return true;
    }
    if (instr->i_src == newOp) {
	instr->i_src = op;
    }
    if (instr->i_src2 == newOp) {
	instr->i_src2 = op;
    }

    return false;
}



/*
 *  UpdateFlags(instr) --
 *
 *  Updates the constant propagation flag for 'instr'.
 */
static void UpdateFlags(Instruction *instr) {
    Block *bbl = instr->i_Bblock;
    Function *fun = INSTR_FUNCTION(instr);
    Function *callee;
    Edge *funDirect;
    int tmp[MAXIMUM_AR_SIZE * 2];
    REGISTER_BITVECTOR regsDef;
    int i, j, sz, unknownIndex, callOffset;

    sz = (fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE;

    for (i = 0; i <= sz; i++) {
	tmp[i] = false;
    }

    ComputeStackDefinedLiberal(instr, tmp,
	fun->f_callerFrameSize / ADDRESS_SIZE,
	(fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE);
    regsDef = ComputeRegistersDefinedLiberal(instr);

    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	if ((regsDef & (1 << i))) {
	    if (RegisterIsFramePointer(BitRegisterToRegister(i))) {
		for (j = 0; j < NUMBER_OF_GP_REGISTERS; j++) {
		    bbl->b_regshot[j].v_stackOffset = 0;
		    SLOT_REMOVE_FLAG_DIRECT(bbl->b_regshot[j], SLOT_STACK_FLAG);
		}
		break;
	    }
	}
    }

    if ((instr->i_prev) && isCallClass(instr->i_prev)) {
	funDirect = BBLGetFunDirectEdge(instr->i_prev->i_Bblock);

	if (funDirect == NULL) {
	    regsDef |= ALL_REGISTERS;
	    for (i = 0; i <= sz; i++) {
		tmp[i] = true;
	    }

	} else {
	    regsDef |= funDirect->e_to->b_fun->f_regsDefined;
	    callee = funDirect->e_to->b_fun;
	    callOffset = InstrBPSPDiff(instr->i_prev);

	    if ((callee->f_trashDepth == -1)
		|| (callOffset == STACK_UNKNOWN)) {
		for (i = 0; i <= sz; i++) {
		    tmp[i] = true;
		}
	    } else {
		unknownIndex = (callOffset / ADDRESS_SIZE) + 2 -
		    (callee->f_trashDepth / ADDRESS_SIZE) +
		    (INSTR_FUNCTION(instr)->f_callerFrameSize / ADDRESS_SIZE);
		for (i = 0; i <= sz; i++) {
		    if (i >= unknownIndex) {
			tmp[i] = true;
		    }
		}
	    }
	}

    }

    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	for (i = 0;
	    i <=
	    (fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE;
	    i++) {
	    if (SLOT_HAS_FLAG_DIRECT(bbl->b_stackshot[i], SLOT_STACK_FLAG)) {
		if (bbl->b_stackshot[i].v_stackOffset && tmp[i]) {
		    SLOT_REMOVE_FLAG_DIRECT(bbl->b_stackshot[i],
			SLOT_STACK_FLAG);
		    bbl->b_stackshot[i].v_stackOffset = 0;
		}
	    }
	}
    }

    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	if (SLOT_HAS_FLAG_DIRECT(bbl->b_regshot[i], SLOT_STACK_FLAG)) {
	    int stackSlot = bbl->b_regshot[i].v_stackOffset;
	    if ((stackSlot > 0) && (stackSlot <= ((fun->f_callerFrameSize) +
			fun->f_stackFrameSize) / ADDRESS_SIZE)
		&& tmp[stackSlot]) {
		SLOT_REMOVE_FLAG_DIRECT(bbl->b_regshot[i], SLOT_STACK_FLAG);
		bbl->b_regshot[i].v_stackOffset = 0;
	    }
	}
    }

    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	if ((regsDef & (1 << i))) {
	    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
		for (j = 0;
		    j <=
		    (fun->f_callerFrameSize +
			fun->f_stackFrameSize) / ADDRESS_SIZE; j++) {
		    if (SLOT_HAS_FLAG_DIRECT(bbl->b_stackshot[j],
			    RegisterToSlotFlag(BitRegisterToRegister(i)))) {
			SLOT_REMOVE_FLAG_DIRECT(bbl->b_stackshot[j],
			    RegisterToSlotFlag(BitRegisterToRegister(i)));
		    }
		}
	    }
	    for (j = 0; j < NUMBER_OF_GP_REGISTERS; j++) {
		if (SLOT_HAS_FLAG_DIRECT(bbl->b_regshot[j],
			RegisterToSlotFlag(BitRegisterToRegister(i)))) {
		    SLOT_REMOVE_FLAG_DIRECT(bbl->b_regshot[j],
			RegisterToSlotFlag(BitRegisterToRegister(i)));
		}
	    }
	}
    }

    return;
}



/*
 *  FlagsToReg(flags) --
 *
 *  Converts the slot flags into a register.
 */
static int FlagsToReg(int flags) {
    int reg = -1;

    if (flags & SLOT_EAX_FLAG) {
	CASSERT(reg == -1, "[FlagsToReg] Multiple registers flagged.\n");
	reg = reg_a | reg_32;
    }

    if (flags & SLOT_EBX_FLAG) {
	CASSERT(reg == -1, "[FlagsToReg] Multiple registers flagged.\n");
	reg = reg_b | reg_32;
    }

    if (flags & SLOT_ECX_FLAG) {
	CASSERT(reg == -1, "[FlagsToReg] Multiple registers flagged.\n");
	reg = reg_c | reg_32;
    }

    if (flags & SLOT_EDX_FLAG) {
	CASSERT(reg == -1, "[FlagsToReg] Multiple registers flagged.\n");
	reg = reg_d | reg_32;
    }

    if (flags & SLOT_ESI_FLAG) {
	CASSERT(reg == -1, "[FlagsToReg] Multiple registers flagged.\n");
	reg = reg_si | reg_32;
    }

    if (flags & SLOT_EDI_FLAG) {
	CASSERT(reg == -1, "[FlagsToReg] Multiple registers flagged.\n");
	reg = reg_di | reg_32;
    }

    if (flags & SLOT_EBP_FLAG) {
	CASSERT(reg == -1, "[FlagsToReg] Multiple registers flagged.\n");
	reg = reg_bp | reg_32;
    }

    if (flags & SLOT_ESP_FLAG) {
	CASSERT(reg == -1, "[FlagsToReg] Multiple registers flagged.\n");
	reg = reg_sp | reg_32;
    }

    return reg;
}



/*
 *  AssignStackPtr(bbl,stackLoc,flags,stackPtr) --
 *
 *  This assigns, to the stack location 'stackLoc', the bitvector 'flags'.
 *  If flags is SLOT_STACK_FLAG, meaning this location is aliased to
 *  another stack location, then that location appears in the 'stackPtr'
 *  field.
 */
static void AssignStackPtr(Block *bbl, int stackLoc, int flags, int stackPtr) {
    Function *fun = bbl->b_fun;

    if (bbl->b_stackshot == NULL) {
	return;
    }

    if ((stackLoc < 0) || (stackPtr < 0)) {
	return;
    }
    if ((stackLoc >=
	    ((fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE))
	|| (stackPtr >=
	    ((fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE)))
	return;

    bbl->b_stackshot[stackLoc].v_stackOffset = stackPtr;
    bbl->b_stackshot[stackLoc].v_flags = flags;

    return;
}



/*
 *  AssignRegPtr(bbl,regIndex,flags,stackPtr) --
 *
 *  This assigns, to the register at index 'regIndex', the bitvector
 *  'flags. If flags is SLOT_STACK_FLAG, meaning this location is aliased
 *  to another stack location, then that location appears in the 'stackPtr'
 *  field.
 */
static void AssignRegPtr(Block *bbl, int regIndex, int flags, int stackPtr) {
    Function *fun = bbl->b_fun;

    CASSERT(regIndex < NUMBER_OF_GP_REGISTERS,
	"[AssignRegPtr] Register index is too large.\n");
    CASSERT(regIndex >= 0, "[AssignRegPtr] Register index is negative.\n");

    if ((stackPtr >=
	    ((fun->f_callerFrameSize +
		    fun->f_stackFrameSize) / ADDRESS_SIZE)) || (stackPtr < 0))
	return;

    bbl->b_regshot[regIndex].v_stackOffset = stackPtr;
    bbl->b_regshot[regIndex].v_flags = flags;

    return;
}
