#include <stdlib.h>
#include <string.h>
#include "utilities.h"
#include "d_graph.h"



/*
 *  CreateDGraph(max_size) --
 */
Digraph *CreateDGraph(int max_size) {
    Digraph *new_graph;

    /* Allocate memory.  Return NULL if unable. */
    new_graph = alloc(sizeof(Digraph));
    if (new_graph == NULL) {
	return NULL;
    }

    new_graph->d_vertices = alloc(sizeof(DigraphVertex) * max_size);
    if (new_graph->d_vertices == NULL) {
	free(new_graph);
	return NULL;
    }

    new_graph->d_edges = alloc(sizeof(DigraphEdge) * max_size * max_size);
    if (new_graph->d_edges == NULL) {
	free(new_graph->d_vertices);
	free(new_graph);
	return NULL;
    }

    new_graph->d_successors = alloc(sizeof(DigraphEdge *) * max_size);
    if (new_graph->d_successors == NULL) {
	free(new_graph->d_edges);
	free(new_graph->d_vertices);
	free(new_graph);
	return NULL;
    }

    new_graph->d_predecessors = alloc(sizeof(DigraphEdge *) * max_size);
    if (new_graph->d_predecessors == NULL) {
	free(new_graph->d_successors);
	free(new_graph->d_edges);
	free(new_graph->d_vertices);
	free(new_graph);
	return NULL;
    }

    /* Set the graph to it's initial state */
    new_graph->d_size = max_size;
    ClearDGraph(new_graph);

    return new_graph;
}



/*
 *  ClearDGraph(graph) --
 */
void ClearDGraph(Digraph *graph) {
    int i;

    if (graph == NULL) {
	return;
    }

    /* Set the edges, successors, and predecessor arrays to contain all 0's */
    memset(graph->d_edges, 0, sizeof(DigraphEdge)*graph->d_size*graph->d_size);
    memset(graph->d_successors, 0, sizeof(DigraphEdge *) * graph->d_size);
    memset(graph->d_predecessors, 0, sizeof(DigraphEdge *) * graph->d_size);
    memset(graph->d_vertices, 0, sizeof(DigraphVertex *) * graph->d_size);

    /* Set up the vertex array */
    for (i = 0; i < graph->d_size; i++) {
	graph->d_vertices[i].dv_id = i;
	graph->d_vertices[i].dv_next = &graph->d_vertices[i + 1];
    }
    graph->d_vertices[graph->d_size - 1].dv_next = NULL;
}



/*
 *  DestroyDGraph(graph) --
 */
void DestroyDGraph(Digraph *graph) {
    if (graph == NULL) {
	return;
    }

    free(graph->d_successors);
    free(graph->d_predecessors);
    free(graph->d_edges);
    free(graph->d_vertices);
    free(graph);
}



/*
 *  CreateEdgeDGraph(graph,s,d,weight) --
 */
void CreateEdgeDGraph(Digraph *graph, DigraphVertex *s, DigraphVertex *d,
    UINT8 weight) {
    /* Set up the edge and its weight */
    DigraphEdge *new_edge = DGRAPH_GET_EDGE(graph, s, d);
    new_edge->di_source = s;
    new_edge->di_target = d;
    new_edge->di_weight = weight;
    new_edge->di_exists = true;

    /* By adding an edge we create a new successor on the source vertex ... */
    new_edge->di_next_successor = graph->d_successors[s->dv_id];
    graph->d_successors[s->dv_id] = new_edge;

    /* ... and a new predecessor for the destination vertex. */
    new_edge->di_next_predecessor = graph->d_predecessors[d->dv_id];
    graph->d_predecessors[d->dv_id] = new_edge;
}



/*
 *  DGRAPH_IN_DEGREE(graph,s) --
 */
UINT32 DGRAPH_IN_DEGREE(Digraph *graph, DigraphVertex *s) {
    UINT32 count = 0;
    DigraphEdge *pred = graph->d_predecessors[s->dv_id];
    for (; pred != NULL; pred = pred->di_next_predecessor) {
	count++;
    }
    return count;
}



/*
 *  DGRAPH_OUT_DEGREE(graph,s) --
 */
UINT32 DGRAPH_OUT_DEGREE(Digraph *graph, DigraphVertex *s) {
    UINT32 count = 0;
    DigraphEdge *succ = graph->d_successors[s->dv_id];
    for (; succ != NULL; succ = succ->di_next_successor) {
	count++;
    }
    return count;
}



/*
 *  PrintGraph(graph) --
 */
void PrintGraph(Digraph *graph) {
    int x, y;
    printf("   ");
    for (x = 0; x < graph->d_size; x++) {
	printf("%d ", x);
    }
    for (y = 0; y < graph->d_size; y++) {
	printf("\n%d: ", y);
	for (x = 0; x < graph->d_size; x++) {
	    printf("%d ", (DGRAPH_GET_EDGE(graph,
		&graph->d_vertices[y], &graph->d_vertices[x])->di_weight));
	}
    }
    printf("\n");
}
