/*
 *  Provides support for disassembling a list of instructions. The stream
 *  is disassembled by calling the system dependency handler, which in
 *  turn dispatches a disassembler for the correct architecture.  The
 *  disassembler will fill in the instruction that has been pre-allocated,
 *  and return its length.
 */
#include "global.h"
#include "disassemble.h"
#include "hash.h"
#include "utilities.h"
#include "flowgraph.h"
#include "sysdep.h"
#include "linkage.h"
#include "opt.unreachable.h"
#include "relocate.h"
#include "upm.h"
#include "instr.h"
#include "symbol.h"
#include "x86-classes.h"

// #define DISA_DEBUG 		// enable disassembler debugging

/* Returns true if the operand has a relocation associated with it */
static bool OperandHasRelocation(Operand *op);

/* Disassembles a basic block's worth of instructions */
static void DisassembleBlock(ADDRESS offset, int datasize, UINT8 *data);

/* Sorts a list of instructions based on address */
static Instruction *SortInstructionList(Instruction *inslist);

/* Used by qsort to compare instructions based on their address */
static int InstrCompareAddress(const void *first, const void *second);

static void show_disassembly_stats(int iteration);
static int isSpecialDataSymbol(Symbol *sym);

static void ReportFlist();

#if DISA_DEBUG
    static void convert_untouched_regions_to_data();
    static void show_untouched_regions();
    static void show_data_entries();
    static void AddDataEntry(DataBlock *de);
    static DataBlock *CreateDataEntry(ADDRESS startAddress, int size);

    static DataBlock *de_list;	// data bytes in the instruction stream
#endif

static void DisassembleAllFunctions(ADDRESS pc, int datasize, UINT8 *data);
static void DisassembleRelocationTargets(ADDRESS pc, int datasize, UINT8 *data);
static void DisassembleStubs(ADDRESS pc, int datasize, UINT8 *data);

static void addInstrToIList(Instruction *instr);



/*
 *  Special data symbols that must be ignored during disassembly,
 *  but be kept in the symbol table.
 */
static char *disa_ignore_syms[] = {
    "pg0",
    "pg1",
    "swapper_pg_dir",
    "stack_start",
    "ready",
    "idt_descr",
    "gdt_descr",
    "idt",
    "gdt",
    "cpu_gdt_descr",
    "_text",
    "_etext",
};



/*
 *  Symbols identifying kernel regions that should not be disassembled.
 */
static DataSymbol ds_list[] = {	
    { "pg0",		0x1000,         0x1000 },
    { "pg1",		0x1000,         0x1000 },
    { "swapper_pg_dir",	0x1000,         0x1000 },
    { "stack_start",	8,              0x00 },
    // split in the 2.4.xx kernel into 2 parts 
    // the idt_descr and idt
    { "idt_descr",	2,              0x00 },     
    { "idt",	6,              0x00 },     
    { "gdt_descr",	2,              0x00 },     
    { "gdt",	6,              0x00 },     
    { "int_msg",	20,             0x00 },
    { "cpu_gdt_descr",	0xddc,          0x00 },
    { "empty_zero_page", 0x1000,        0x1000 },
    { "wakeup_data",	2,              0x00 },
    { "real_save_gdt",	6,              0x00 },
    { "real_save_cr3",	4,              0x00 },
    { "real_save_cr4",	4,              0x00 },
    //{ "vide",		10,             0x00 },
    { "ready",		1,              0x00 },
    { "check_x87",	38,             0x00 },
    { "wakeup_code",	0x100,          0x00 },
    { "check_hlt",	0x3f,           0x00 },
    //    { "yyparse",	0x3d8b,           0x00 },    
};



/*
 *  Symbols identifying calls that terminate the program and do not return.
 *  Only kernel symbols are in this list at the current time.
 */
static char *TERMINATING_CALLS[] = {
    "panic",
};



static int disassembly_region_size;



/*
 *  This array is used to keep track of bytes which were disassembled during the
 *  disassembly process. At the moment only one bit is used, which leaves the
 *  remaining seven free for any other use (eg. marking entry points).
 */
static UINT8 DISASSEMBLY_BYTE_ARRAY[0x1000000];	//#%#%#% CHANGE THIS MR

#define SET_DISASSEMBLED_FLAG(x) x |= 1

#define NEW_TO_OLD(x) (x + 0x120a50)		//#%#%#% MR TO REMOVE
#define prev_value(x) (x > 0xc02775b0 ? NEW_TO_OLD(x) : x)



/*
 *  CreateNewFlistEntry(startAddr,size,sym) -- map data symbol to flist.
 */
static void CreateNewFlistEntry(ADDRESS startAddr, int size, Symbol *sym) {

    FList *newFlist;
    ADDRESS endAddr = startAddr + size;
    int codeOffset;

    newFlist = zalloc(sizeof(FList));
    newFlist->fl_startAddress = startAddr;
    newFlist->fl_endAddress = endAddr - 1;
    newFlist->fl_name =
	salloc(GetStrongestSymbolForValue(sym->sy_value)->sy_name);
    codeOffset = startAddr - SECTION_REF_CODE->s_address;
    newFlist->fl_data = zalloc(sizeof(UINT8) * (endAddr - startAddr));
    memcpy(newFlist->fl_data, &(SECTION_REF_CODE->s_data)[codeOffset],
	endAddr - startAddr);

    REPORT(RP_TRIVIAL, "Created new FLIST entry: Name(%s), start(%x), end(%x)",
	newFlist->fl_name, newFlist->fl_startAddress, newFlist->fl_endAddress);

    if (dontTouchFuns == NULL) {
	dontTouchFuns = newFlist;
    } else {
	newFlist->fl_next = dontTouchFuns;
	dontTouchFuns->fl_prev = newFlist;
	dontTouchFuns = newFlist;
    }

}



/*
 *  ConvertSpecialDataSymbolsIntoDataInstructions
 *
 *  The special data symbols correspond to data in the text
 *  segment of the kernel. This function, goes through and creates
 *  data instructions which cover these symbols.
 */
static void ConvertSpecialDataSymbolsIntoDataInstructions() {
    int i;
    bool found = false;
    Symbol *sym;
    Instruction *instr;

    REPORT(RP_MINOR, WITHFUNC("Protecting special regions"));

    FOREACH_SYMBOL(sym) {
	// Find the data symbol.
	for (i = 0; i < ARRAYSIZE(ds_list); i++) {
	    found = false;
	    if (strcmp(sym->sy_name, ds_list[i].ds_name) == 0 &&
            sym->sy_section == SECTION_REF_CODE) {
		found = true;
		break;
	    }
	}
	if (!found) {
	    continue;
	} else {
            // check if we need to align this data instr according
            // to the section alignment.
            if ((sym->sy_value % SECTION_REF_CODE->s_alignment) == 0) {
                if (!ds_list[i].ds_alignment) {
                    ds_list[i].ds_alignment = 1;
                }
            }

	    REPORT(RP_SMLOOP,
		"Creating data instr for %s from 0x%x with size %d, align - %s",
		sym->sy_name, sym->sy_value, ds_list[i].ds_datasize, 
                ds_list[i].ds_alignment?"Y":"N");

	    instr = CreateDataInstr(sym->sy_value - SECTION_REF_CODE->s_address,
		ds_list[i].ds_datasize,
		&(SECTION_REF_CODE->s_data)[
		    sym->sy_value - SECTION_REF_CODE->s_address]);
	    // since Flist Addresses are used for a ton of things later on ...
	    // create an Flist entry for each special symbol.
	    CreateNewFlistEntry(sym->sy_value, ds_list[i].ds_datasize, sym);
            
            // if we do, then set the alignment flag.
            // Actually set in src/relocate.c:UpdateInstrAddresses()
            if (ds_list[i].ds_alignment) {
                INSTR_SET_FLAG(instr, INSTR_NEEDS_ALIGNMENT);

                if (SECTION_REF_CODE->s_alignment > ds_list[i].ds_alignment) {
                    instr->i_alignment = SECTION_REF_CODE->s_alignment;
                } else {
                    instr->i_alignment = ds_list[i].ds_alignment;
                }
            }

	    InstrInsertEntry(instr);
	    if (instr->i_length > MaxInstrLength) {
		MaxInstrLength = instr->i_length;
	    }
	    addInstrToIList(instr);
	    REPORT(RP_SMDETAIL,
		"Added data instruction at 0x%x of size %d bytes",
		instr->i_address, instr->i_length);
	}
    }

    return;
}



#ifdef DISA_DEBUG
/*
 *  convert_untouched_regions_to_data() --
 *
 *  Adds all untouched locations in the kernel binary to the  data entry list.
 *
 *  N.B.: At the moment I read everything as a byte -- ref: more than 2
 *  relocations in the data entry. THIS MAY NEED TO CHANGE.
 */
static void convert_untouched_regions_to_data() {

    int pointer = 0;

    while (pointer < SECTION_REF_CODE->s_datasize) {

	if (DISASSEMBLY_BYTE_ARRAY[pointer] == 0) {
	    AddDataEntry(CreateDataEntry(pointer +
		    SECTION_REF_CODE->s_address, 1));
	    pointer++;
	} else {
	    pointer++;
	}
    }
}
#endif					// DISA_DEBUG



#ifdef DISA_DEBUG
/*
 *  show_untouched_regions() --
 *
 *  Utility function. Goes through the entire byte-map and indicates
 *  the regions of code which were not disassembled.
 *  A region is defined as a contiguous series of bytes which were not
 *  disassembled.
 */
static void show_untouched_regions() {

    int region_size = 0;
    int region_start_offset = 0;
    int pointer = region_start_offset;
    int counter = 0;

    REPORT(RP_LGLOOP, "Untouched regions:");
    REPORT(RP_LGLOOP, "S.No \t Start Address \t Size");
    while (pointer < disassembly_region_size) {

	if (DISASSEMBLY_BYTE_ARRAY[pointer] == 0) {
	    // reset the values.
	    region_start_offset = pointer;
	    region_size = 0;
	    // count the non-disassembled bytes.
	    while (DISASSEMBLY_BYTE_ARRAY[pointer] == 0) {
		region_size++;
		pointer++;
	    }
	    REPORT(RP_LGLOOP, "%8.8d \t 0x%8.8x \t %8.8d", ++counter,
		region_start_offset + SECTION_REF_CODE->s_address, region_size);
	} else {
	    pointer++;
	}
    }
}
#endif				// DISA_DEBUG



/*
 *  show_disassembly_stats(iteration) --
 *
 *  Utility function to show the number of undisassembled bytes.
 */
static void show_disassembly_stats(int iteration) {

    int undisassembled_counter = 0;
    int i;

    for (i = 0; i < disassembly_region_size; i++) {
	if (DISASSEMBLY_BYTE_ARRAY[i] == 0) {
	    undisassembled_counter++;
	}
    }

    REPORT(RP_DETAIL, "Iteration %d: covered all but %5.2f%% (%d / %d bytes)",
	iteration, 100.0 * undisassembled_counter / disassembly_region_size,
	undisassembled_counter, disassembly_region_size);
}



/*
 *  isSpecialDataSymbol(sym) --
 *
 *  Special symbols in the text section (SYMBOL_TYPE = NO_TYPE) can point to
 *  either stub-code or data.  Since there is no "good" way to distinguish
 *  either, for the moment the special symbols which point to data are
 *  special-cased and so ignored during disassembly.
 */
int isSpecialDataSymbol(Symbol *sym) {
    int i;
    for (i = 0; i < ARRAYSIZE(disa_ignore_syms); i++) {
	if (strcmp(sym->sy_name, disa_ignore_syms[i]) == 0) {
	    return 1;
	}
    }
    return 0;
}



/*
 *  TargetAddressIsRelocation(targetAddress) --
 *
 *
 *  the target is not disassemblable.
 *  NB : This is a bad way of doing it (N^N) -- eventually move this out and
 *  optimize with table based lookup or hash.
 */
static int TargetAddressIsRelocation(ADDRESS targetAddress) {

    Relocation *reloc;

    FOREACH_RELOCATION(reloc) {
	if (targetAddress == RELOCATION_ADDRESS(reloc)) {
	    return 1;
	}
    }
    return 0;
}



/*
 *  PointsIntoInstruction(targetAddress) --
 *
 *  Given an address, checks if there is any instruction that
 *  contains the target address.
 */
int PointsIntoInstruction(ADDRESS targetAddress) {

    Instruction *temp_instr;
    int i = MaxInstrLength;

    // don't start at 0 since in that will be an instruction and not into.
    for (i = 1; i < MaxInstrLength; i++) {
	temp_instr = AddressToInstr(targetAddress - i);
	if (temp_instr != NULL) {
	    if (targetAddress < temp_instr->i_address+(temp_instr->i_length)) {
		REPORT(RP_TRIVIAL, "%x points into instruction %x",
		    targetAddress, temp_instr->i_address);
		return 1;
	    }
	}
    }
    return 0;
}



#if DISA_DEBUG
/*
 *  CreateDataEntry(startAddress,size) --
 *
 *  Create a new DataBlock based on starting address and size.
 */
static DataBlock *CreateDataEntry(ADDRESS startAddress, int size) {

    DataBlock *de = alloc(sizeof(DataBlock));
    int i;

    // set the startAddress and size values.
    de->db_startAddress = startAddress;
    de->db_size = size;

    de->db_data = alloc(size);
    memcpy(de->db_data,
	&SECTION_REF_CODE->s_data[startAddress - SECTION_REF_CODE->s_address],
	size);

    // Finally touch corresponding data in the disassembly_byte_array to reflect
    // that it has been seen.
    for (i = startAddress - SECTION_REF_CODE->s_address; i < size; i++) {
	SET_DISASSEMBLED_FLAG(DISASSEMBLY_BYTE_ARRAY[i]);
    }

    return de;

}
#endif				// DISA_DEBUG



#if DISA_DEBUG
/*
 *  AddDataEntry(de) --
 *
 *  Add a data entry to the list of data entries.
 */
static void AddDataEntry(DataBlock *de) {


    if (NULL == de_list) {		// first entry
	de_list = de;
	de_list->db_prev = de;
	de_list->db_next = de;
    } else {				// other entries
	de->db_next = de_list;
	de->db_prev = de_list->db_prev;
	de->db_prev->db_next = de;
	de_list->db_prev = de;
    }
}
#endif				// DISA_DEBUG



#ifdef DISA_DEBUG
/*
 *  show_data_entries() --
 *
 *  Goes through the list of data-entries and displays each one in
 *  the following format.
 */
static void show_data_entries() {

    DataBlock *temp_de = de_list;
    int s_no = 0;

    printf(" ############################## SHOWING DATA ENTRIES"
	" ############################### \n");
    printf(" ###################################################"
	"################################ \n");
    printf(" \t S.No \t Address \t Offset \t Size \n");
    printf(" ###################################################"
	"################################ \n");

    if (de_list != NULL) {
	do {
	    printf(" \t %d \t 0x%8.8x \t 0x%8.8x \t 0x%8.8x\n",
		++s_no, temp_de->db_startAddress,
		temp_de->db_startAddress - SECTION_REF_CODE->s_address,
		temp_de->db_size);
	    temp_de = temp_de->it_next;
	} while (temp_de != de_list);
	printf(" ###################################################"
	    "################################ \n");

    }
}
#endif				// DISA_DEBUG



/*
 *  CreateDataInstr(offset, datasize, data) -- creates a data instruction.
 *
 *  Note: InstrCreate should add it to the i_list.
 */
Instruction *CreateDataInstr(ADDRESS offset, int datasize, UINT8 *data) {
    Instruction *new_instr = InstrCreate();

    INSTR_SET_FLAG(new_instr, INSTR_DATA_BYTE);
    new_instr->i_address = offset + SECTION_REF_CODE->s_address;
    new_instr->i_length = datasize;

    new_instr->i_literal_data = alloc(sizeof(UINT8) * datasize);
    memcpy(new_instr->i_literal_data, data, datasize);

    return new_instr;
}



/*
 *  addInstrToIList(instr) -- add instruction to the instruction list i_list.
 */
void addInstrToIList(Instruction *instr) {

    if (i_list == NULL) {
	i_list = instr;
    } else {
	instr->i_next = i_list;
	i_list->i_prev = instr;
	i_list = instr;
    }
}



/*
 *  isTerminatingSymbol(sym_name) -- returns true if symbol is in special list.
 */
bool isTerminatingSymbol(char *sym_name) {
    int i;

    for (i = 0; i < ARRAYSIZE(TERMINATING_CALLS); i++) {
	if (strcmp(sym_name, TERMINATING_CALLS[i]) == 0) {
	    return true;
	}
    }
    return false;
}



/*
 *  CheckCallTarget(instr) -- returns true if instr is a non-returning call.
 */
static bool CheckCallTarget(Instruction *instr) {
    Relocation *reloc;
    Relocation *target_reloc = NULL;
    ADDRESS dest_address;

    // 1. Find the correct relocation.
    FOREACH_RELOCATION(reloc) {
	if (reloc->r_section == SECTION_REF_CODE) {
	    dest_address = reloc->r_section->s_address + (reloc->r_offset);
	    if ((dest_address > instr->i_address) &&
		(dest_address < (instr->i_address + instr->i_length))) {
		target_reloc = reloc;
		break;
	    }
	}
    }

    // 2. If you couldn't find the relocation, then possibly a relative
    //    offset conditional transfer.
    if (target_reloc == NULL) {
	int call_offset;
	Symbol *sym;
	int target_value;
	int found = 0;

	if (OPERAND_HAS_FLAG(instr->i_src, OPERAND_INDIRECT_FLAG)) {
	    // case 1: Could we be seeing an indirect call?
	    return false;
	} else {
	    // case 2: Or else, a PC relative offset ?

	    // 1. Get the offset
	    // 2. Compute the intended target : offset + Instr_PC + Instr_len.
	    // 3. Find symbol associated with target.
	    // 4. Create Relocation ? <-- at the moment, trust CFG edges,
	    //                    but eventually check this assumption.
	    call_offset = loadLE(
		&(SECTION_REF_CODE->s_data)[
		    instr->i_address + 1 - SECTION_REF_CODE->s_address]);
	    target_value = instr->i_address + instr->i_length + call_offset;

	    FOREACH_SYMBOL(sym) {
		if (sym->sy_value == target_value) {
		    found++;
		}
	    }
	    if (!found) {
		REPORT(RP_TRIVIAL,
		    "Unable to find symbol pointing to %x in instr (@%x)",
		    prev_value(target_value), instr->i_address);
	    }
	}
    } else if (isTerminatingSymbol(target_reloc->r_symbol->sy_name)) {
	return true;
    }

    return false;
}



/*
 *  OperandHasRelocation(op) --
 *
 *  Returns true if the operand has a relocation in it.
 */
static bool OperandHasRelocation(Operand *op) {
    if (op == NULL) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_RELOCATABLE_ABS_FLAG)) {
	return true;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_RELOCATABLE_PC_FLAG)) {
	return true;
    }
    return false;
}



/*
 *  SortInstructionList(inslist) --
 *
 *  Sorts a linked-list of instructions by their address in the executable.
 *  Returns the new head of the list.
 */
static Instruction *SortInstructionList(Instruction *inslist) {
    Instruction *iterator;
    int instrCount = 0;
    Instruction **instrArray;
    int offset, i;

    for (iterator = inslist; iterator != NULL; iterator = iterator->i_next) {
	instrCount++;
    }

    instrArray = zalloc(sizeof(Instruction *) * instrCount);

    offset = 0;
    for (iterator = inslist; iterator != NULL; iterator = iterator->i_next) {
	instrArray[offset++] = iterator;
    }

    // Sort the instruction array
    qsort(instrArray, instrCount, sizeof(Instruction *), InstrCompareAddress);

    // Re-link the entries in their new order
    instrArray[0]->i_prev = NULL;
    for (i = 1; i < instrCount; i++) {
	instrArray[i-1]->i_next = instrArray[i];
	instrArray[i]->i_prev = instrArray[i-1];
    }
    instrArray[i-1]->i_next = NULL;

    return instrArray[0];
}



/*
 *  InstrCompareAddress(first,second) --
 *
 *  Used by qsort to compare instructions. An instruction is less than
 *  another if its original address in the executable is less than the
 *  other instruction's.
 */
int InstrCompareAddress(const void *first, const void *second) {
    int addr1 = ((*(Instruction **) first))->i_address;
    int addr2 = ((*(Instruction **) second))->i_address;

    if (addr1 < addr2) {
	return -1;
    }
    if (addr1 > addr2) {
	return 1;
    }

    return 0;
}



/*
 *  ReportFlist() --
 *
 *  Prints out the contents of all flists.
 */
static void ReportFlist() {
    FList *fiterator;

    REPORT(RP_SMLOOP, "Functions not disassembled:");
    FOREACH_FLIST(fiterator) {
	REPORT(RP_SMLOOP, "%08X - %08X  (%3d bytes)  %s",
	    fiterator->fl_startAddress, fiterator->fl_endAddress,
	    fiterator->fl_endAddress - fiterator->fl_startAddress + 1,
	    fiterator->fl_name);
    }
    return;
}



/*
 *  DisassembleSection(code_section) --
 *
 *  Called by src/plto.c to disassemble the text section
 */
Instruction *DisassembleSection(Section *code_section) {

    return disassemble(code_section->s_address,
	code_section->s_datasize, code_section->s_data);
}



#if DISA_DEBUG
/*
 *  show_instructions() -- show all instructions that have been disassembled.
 */
static void show_instructions() {
    Instruction *instr;
    int counter = 0;

    REPORT(RP_TRIVIAL, "All disassembled instructions:");
    FOREACH_INSTR(instr) {
	REPORT(RP_TRIVIAL, " %d 0x%8.8x %s",
 	    ++counter, instr->i_address, PrintInstruction(instr));
    }
}
#endif					// DISA_DEBUG



/*
 *  disassemble(pc, size, data) -- the main disassembly function.
 */
Instruction *disassemble(ADDRESS pc, int datasize, UINT8 *data) {
    Instruction *result;

    ConvertSpecialDataSymbolsIntoDataInstructions();
    DisassembleAllFunctions(pc, datasize, data);
    show_disassembly_stats(1);
    DisassembleStubs(pc, datasize, data);
    show_disassembly_stats(2);
    DisassembleRelocationTargets(pc, datasize, data);
    show_disassembly_stats(3);
    result = i_list = SortInstructionList(i_list);
    startins = AddressToInstr(startAddress);
    ReportFlist();
    return result;
}



/*
 *  DisassembleAllFunctions(pc, size, data) -- recursively disasm functions.
 *
 *  This function identifies functions through the symbol table and performs
 *  recursive disassembly.
 */
static void DisassembleAllFunctions(ADDRESS pc, int datasize, UINT8 *data) {
    Symbol *sym;

    REPORT(RP_MINOR, WITHFUNC("Disassembling functions"));

    FOREACH_SYMBOL(sym) {
	if (!!(sym->sy_flags & SYMBOL_FUNCTION_FLAG)) {
	    DisassembleBlock(
		sym->sy_value - SECTION_REF_CODE->s_address, datasize, data);
	}
    }
}



/*
 *  DisassembleStubs(pc, size, data) -- disassemble unusual code fragments.
 *
 *  Stubs are defined as pieces of code that do not follow the
 *  conventions seen in regular functions -- for example, hand-written
 *  assembly routines, kernel entry points, etc.  We identify stubs
 *  by looking at the symbol table -- since these could, as in the
 *  case of the Linux kernel, themselves be entry points and thus
 *  targets for recursive disassembly.  Because these are not
 *  marked as functions, they end up as entries in the symbol table
 *  with NOTYPE.  (N.B.: Not all NOTYPE symbols in the text section
 *  are stubs -- see isSpecialDataSymbol().)
 */
static void DisassembleStubs(ADDRESS pc, int datasize, UINT8 *data) {
    Symbol *sym;

    REPORT(RP_MINOR, WITHFUNC("Disassembling stubs"));

    FOREACH_SYMBOL(sym) {
	// check if the symbol points to a location within the code section.
	if (sym->sy_value >= SECTION_REF_CODE->s_address &&
	    sym->sy_value <= SECTION_REF_CODE->s_address +
		SECTION_REF_CODE->s_datasize) {

	    // check if it is a symbol of type NOTYPE.
	    if ((!!(sym->sy_flags & SYMBOL_NOTYPE_FLAG))
		&& !isSpecialDataSymbol(sym)) {
		DisassembleBlock(sym->sy_value - SECTION_REF_CODE->s_address,
		    datasize, data);
	    }
	}
    }
}



/*
 *  DisassembleRelocationTargets(pc, size, data) -- scan relocations and disasm.
 *
 *  Goes through all relocations and checks if they point to a
 *  disassemble-able piece of code. If they do, perform recursive
 *  disassaembly as before.
 */
static void DisassembleRelocationTargets(ADDRESS pc, int size, UINT8 *data) {
    Relocation *reloc;
    ADDRESS targetAddress;

    REPORT(RP_MINOR, WITHFUNC("Disassembling relocation targets"));

    FOREACH_RELOCATION(reloc) {
	// Does it point into an ignored section ??
	if (reloc->r_section->s_address == 0) {
	    continue;
	}

	if (isSpecialDataSymbol(reloc->r_symbol)) {
	    // If its a data symbol, then don't disassemble target.
	    continue;
	}

	targetAddress = loadLE(&(reloc->r_section->s_data)[reloc->r_offset]);
	if (PointsIntoInstruction(targetAddress)) {
	    REPORT(RP_TRIVIAL,
		"Ignoring reloc [@ %s::%x] pointing into instr (0x%8.8x)",
		reloc->r_section->s_name, reloc->r_offset, targetAddress);
	    continue;
	}
	if (SECTION_CONTAINS(SECTION_REF_CODE, targetAddress)) {
	    if (TargetAddressIsRelocation(targetAddress)) {
		// If the relocation points to another relocation, then we may
		// be looking at a jump table, so don't follow this disassembly
		// target.
		//
		// A better idea would be to convert it to a data instr.
		// How do you know the size of the table --
		// Wait until complete disassembly,
		// revisit and locate next valid instr
		WARNING(
		    "Reloc [@ %s::%x] points to reloc -- jump tbl (0x%8.8x)?",
		    reloc->r_section->s_name, reloc->r_offset, targetAddress);
		continue;
	    }
	    DisassembleBlock(targetAddress - SECTION_REF_CODE->s_address,
		size, data);
	}
    }
}



#define DONT_WANT_TO_HANDLE(x) \
    ((x->i_class == CLASS_NONE) && (x->i_type == TYPE_NONE))

/*
 *  DisassembleBlock(offset, size, data) -- disassemble one fragment of code.
 *
 *  Disassembles a block of code, adding the instructions into the
 *  instruction list and the hash table. The routine calls itself
 *  recursively to disassemble all reachable portions of code.
 */
static void DisassembleBlock(ADDRESS offset, int datasize, UINT8 *data) {
    Instruction *instr;
    int instr_size;
    int temp_i;

    if (GetFlistFunAtAddress(offset + SECTION_REF_CODE->s_address)) {
	return;
    }

    while (offset < datasize) {
	Instruction *instr2 =
	    AddressToInstr(SECTION_REF_CODE->s_address + offset);

	// If we've already decoded the block, the its instruction
	// will appear in the hash table, and we shouldn't decode it again.
	if (instr2 != NULL) {
	    return;
	}

	instr = InstrCreate();
    
	ADDRESS t_pc = 0;
	t_pc = offset + SECTION_REF_CODE->s_address;
    
	instr_size =
	    DisassembleInstr(offset + SECTION_REF_CODE->s_address,
	    data + offset, instr);

	// This is a good time to check if the instr was disassembled correctly
	if ((instr_size < 0) ||
	    PointsIntoInstruction(offset + (SECTION_REF_CODE->s_address))
	    || DONT_WANT_TO_HANDLE(instr)) {

	    FList *flist_entry;
	    // if (already in flist) then just ignore and return !!!
	    if (GetFlistFunAtAddress(offset + SECTION_REF_CODE->s_address)) {
		return;
	    } else {
		flist_entry = AddFlist(offset + SECTION_REF_CODE->s_address);
	    }


	    if (PointsIntoInstruction(offset + (SECTION_REF_CODE->s_address))) {
		REPORT(RP_TRIVIAL,
		    "Flisting bcos it Points into Instruction %x\n", t_pc);
	    }

	    if (DONT_WANT_TO_HANDLE(instr)) {
		REPORT(RP_TRIVIAL,
		    "Flisting bcos we don't want to handle instruction %x\n",
		    t_pc);
	    }

	    if (instr_size < 0) {
		REPORT(RP_TRIVIAL,
		    "Flisting bcos instr size is less than 0 %x\n", t_pc);
	    }        
        
	    REPORT(RP_TRIVIAL,
		"Found an Flist (%-30s) -- from region 0x%8x --> 0x%8x , %8d",
		flist_entry->fl_name, flist_entry->fl_startAddress,
		flist_entry->fl_endAddress,
		flist_entry->fl_endAddress - flist_entry->fl_startAddress);

	    // see if any instructions from this function were disassembled
	    // if they were, dequeue them from i_list.
	    {
		Instruction *rem_instr = i_list;
		if (i_list == NULL) {
		    break;
		}

		// go through the i_list and look at the address of each instr.
		// if any instructions lies within a data region --- REMOVE !
		while (rem_instr != NULL) {
		    Instruction *next_instr = NULL;

		    if (rem_instr->i_next) {
		        next_instr = rem_instr->i_next;
		    }

		    // if the instr lies within an flist then remove it
		    if ((rem_instr->i_address >= flist_entry->fl_startAddress)
		    && (rem_instr->i_address <= flist_entry->fl_endAddress)) {

			if (rem_instr->i_next) {
			    rem_instr->i_next->i_prev = rem_instr->i_prev;
			}
			if (rem_instr->i_prev) {
			    rem_instr->i_prev->i_next = rem_instr->i_next;
			}

			// update the i_list if required.
			if (rem_instr == i_list) {
			    i_list = next_instr;
			}
			InstrRemoveEntry(rem_instr);
			InstrFree(rem_instr);
		    }
		    rem_instr = next_instr;
		}
	    }

	    // Finally create a data instruction corresponding to the
	    // given flist.
	    instr = CreateDataInstr(
		flist_entry->fl_startAddress - SECTION_REF_CODE->s_address,
		flist_entry->fl_endAddress - flist_entry->fl_startAddress + 1,
		flist_entry->fl_data);

	    flist_entry->fl_newStartAddress = flist_entry->fl_startAddress;
	    flist_entry->fl_newEndAddress = flist_entry->fl_endAddress;

	    InstrInsertEntry(instr);
	    instr_size = instr->i_length;
	    //   MAX_INSTR_LENGTH needs to be updated !!!
	    if (instr->i_length > MaxInstrLength) {
		MaxInstrLength = instr->i_length;
	    }
	    addInstrToIList(instr);
	    return;
	}

	// check if its a special case UD2A instruction from the kernel
	if (KERNEL_MODE && (instr->i_class == CLASS_UD2)) {
	    REPORT(RP_LGLOOP, "Encountered UD2A instruction @ %x",
		instr->i_address);
	    instr = CreateDataInstr(
		instr->i_address - SECTION_REF_CODE->s_address,
		8,
		&(SECTION_REF_CODE->s_data)[(instr->i_address) -
		    (SECTION_REF_CODE->s_address)]);
	    instr->i_class = CLASS_UD2;
	    instr_size = 8;
	}

	if ((instr->i_class == CLASS_PREFETCH)) {
	    REPORT(RP_LGLOOP, "Encountered UD2A instruction @ %x",
            instr->i_address);
	    instr = CreateDataInstr(
            instr->i_address - SECTION_REF_CODE->s_address,
            3,
            &(SECTION_REF_CODE->s_data)[(instr->i_address) -
                (SECTION_REF_CODE->s_address)]);
	    instr->i_class = CLASS_PREFETCH;
	    instr_size = 3;
	}

        if (KERNEL_MODE && (instr->i_class == CLASS_LJUMP1)) {
            REPORT(RP_LGLOOP, "Encountered LJMP1 instruction @ %x",
        	instr->i_address);
            instr = CreateDataInstr(
        	instr->i_address - SECTION_REF_CODE->s_address,
        	7,
        	&(SECTION_REF_CODE->s_data)[(instr->i_address) -
        	    (SECTION_REF_CODE->s_address)]);
            instr_size = 7;
            instr->i_class = CLASS_LJUMP1;
        }

        //  if (KERNEL_MODE && (instr->i_class == CLASS_LJUMP)) {
        //      REPORT(RP_LGLOOP, "Encountered LJMP instruction @ %x",
        //  	instr->i_address);
        //      instr = CreateDataInstr(
        //  	instr->i_address - SECTION_REF_CODE->s_address,
        //  	2,
        //  	&(SECTION_REF_CODE->s_data)[(instr->i_address) -
        //  	    (SECTION_REF_CODE->s_address)]);
        //      instr_size = 2;
        //  }

	InstrInsertEntry(instr);
	instr->i_length = instr_size;

	// Add instr to the i_list
	addInstrToIList(instr);

	// Set a flag if the instruction belongs to a certain group of classes
	// NB: Inherited from old code, not sure if still relevant.
	InstrSetClassFlags(instr);

	if (OperandHasRelocation(instr->i_dest) ||
	    OperandHasRelocation(instr->i_src) ||
	    OperandHasRelocation(instr->i_src2)) {
	    INSTR_SET_FLAG(instr, INSTR_RELOCATION_FLAG);
	}

	for (temp_i = offset; temp_i < offset + instr_size; temp_i++) {
	    SET_DISASSEMBLED_FLAG(DISASSEMBLY_BYTE_ARRAY[temp_i]);
	}

	offset = offset + instr_size;

	if (offset > disassembly_region_size) {
	    disassembly_region_size = offset;
	}
	//
	//   If we're dealing with a call/jump/branch instruction, then we'll
	//   recursively decode the target block. Note that call instructions
	//   should keep decoding, since control returns to the instruction
	//   following the call.
	//
	if (isPCChangingClass(instr)) {
	    Operand *op = instr->i_src;

	    if (isHaltClass(instr)) {
		return;
	    }
	    if (isReturnClass(instr)) {
		return;
	    }

	    // In case we encounter a call class instruction,
	    // no need to disassemble the target.
	    if (isCallClass(instr)) {
		if (CheckCallTarget(instr)) {
		    REPORT(RP_TRIVIAL, "call to non-returning function @ 0x%x",
			prev_value(instr->i_address));
		    return;
		} else {
		    continue;
		}
	    }
	    if (isSyscallClass(instr)) {
		continue;
	    }

	    if (op != NULL) {
		if (!OperandUsesRegister(op)) {
		    if (OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
			DisassembleBlock(op->a_displacement -
			    SECTION_REF_CODE->s_address, datasize, data);
		    } else if (OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG)) {
			DisassembleBlock(op->a_displacement + offset, datasize,
			    data);
		    }
		}
	    }

	    if (!isCallClass(instr) && !isConditionalInstr(instr)) {
		return;
	    }
	}
    }
    return;
}
