/*
 *  A module for constructing dominator relationships between blocks.
 *  Functions for constructing loop-lists and determining loop
 *  relationships are also present here.
 */
#include "dominator.h"
#include "global.h"
#include "utilities.h"
#include "flowgraph.h"
#include "cache.analyze.h"
#include "linkage.h"

void TestDominatorCode();
static void CreateLoop(TreeNode *tree, Block *head, Block *tail);
static void LinkLoops(TreeNode *tree);
static bool IsBackEdge(TreeNode *tree, Edge *edge);
static void RecurseCreatePostdominatorTree(TreeNode *root, TreeNode *current,
    Block *bbl);
static void RecurseCreatePredominatorTree(TreeNode *root, TreeNode *current,
    Block *bbl);
static LinkedList *GetTraversal(TreeNode *tree,
    void *(*list_add) (LinkedList *list, void *val),
    void *(*list_remove) (LinkedList *list));



/*
 *  BBLIsInALoop(bbl) --
 *
 *  Returns whether or not the specified basic block is in a loop based on the
 *  stored dominator and loop information for the basic block.
 *
 *  bbl: the basic block to check for loop containment
 *  returns: true if bbl is in a loop
 */
bool BBLIsInALoop(Block *bbl) {
    return bbl->b_loopHead != NULL;
}



/*
 *  BBLIsInSameLoopAsBBL(bbl1,bbl2) --
 *
 *  Returns whether or not the two basic blocks are in the SAME loop. If bbl2 is
 *  in a loop that is nested within bbl1, this function will not identify them
 *  as being in the same loop.
 *
 *  bbl1: the first basic block
 *  bbl2: the second basic block
 *  returns: true if bbl1 and bbl2 are in the SAME loop
 */
bool BBLIsInSameLoopAsBBL(Block *bbl1, Block *bbl2) {
    if (bbl1 == bbl2) {
        return true;
    }

    if (bbl1->b_loopHead == bbl2 || bbl2->b_loopHead == bbl1) {
        return true;
    }

    if (bbl1->b_loopHead != NULL
	&& bbl2->b_loopHead != NULL
	&& bbl1->b_loopHead == bbl2->b_loopHead) {
            return true;
    }

    return false;
}



/*
 *  FunCreatePredominatorTree(fun) --
 *
 *  Constructs a predominator tree for the specified function.
 *
 *  fun: the function to create the dominator tree for
 *  returns: the predominator tree for the specified function
 */
TreeNode *FunCreatePredominatorTree(Function *fun) {
    TreeNode *root;
    Block *dummy;
    bool renumber_nodes;

    renumber_nodes = false;

    // some basic blocks are not actually in the bbl_list (BAD!!) so this
    // assigns BBL_IDs as the actual memory address of the basic block
    // to ensure that they are all unique. This also accounts for those
    // basic block that were not created correctly ie those that did not
    // get created with BBLCreate()
    if (renumber_nodes) {
	FOREACH_BBL(dummy) {
	    dummy->b_id = (int) dummy;
	    dummy->b_tmp = dummy->b_id;
	}
    }

    dummy = zalloc(sizeof(Block));
    dummy->b_id = -1;
    BBLAddEdge(dummy, fun->f_entry, ET_NORMAL);

    root = zalloc(sizeof(TreeNode));
    root->t_bbl = dummy;
    root->t_parent = NULL;
    root->t_next_child = NULL;
    root->t_children = NULL;

    RecurseCreatePredominatorTree(root, root, fun->f_entry);

    root = root->t_children;
    BBLRemoveEdges(dummy);
    root->t_parent = NULL;

    free(dummy);

    if (renumber_nodes) {
	FOREACH_BBL(dummy) {
	    dummy->b_id = dummy->b_tmp;
	}
    }

    return root;
}



/*
 *  RecurseCreatePredominatorTree(root,current,bbl) --
 */
static void RecurseCreatePredominatorTree(TreeNode *root, TreeNode *current,
    Block *bbl) {
    TreeNode *new_node, *ancestor, *existing;

    ancestor = existing = NULL;

    if (BBLIsAncestorOfNode(current, bbl)) {
	return;
    }

    existing = GetExistingNodeFor(root, bbl);

    if (existing != NULL) {
	// if there is already a node for this basic block in the tree we want
	// to find the nearest common ancestor for these two
	ancestor = GetNearestCommonAncestor(existing, current);

	CASSERT(ancestor != NULL, "No common ancestor exists");

	if (ancestor == existing->t_parent) {
	    return;
	}

	if (existing->t_parent != NULL) {
	    // remove the child node from it's old parent
	    RemoveChildNode(existing->t_parent, existing);

	    // move the existing node up in the tree s.t. it's new parent
	    // is the nearest common ancestor
	    existing->t_parent = ancestor;
	}
	// add the existing node to the nearest common ancestor
	AddChildNode(ancestor, existing);
    } else {
	// if there is no common ancestor then we will just go ahead and add
	// this basic block into the tree
	EdgeNode *niterator;

	// create a new node for the current basic block
	new_node = DominatorNodeCreate();
	new_node->t_bbl = bbl;
	AddChildNode(current, new_node);

	// for each successor in biterator we want to make sure
	// that it gets added into the dominator tree
	FOREACH_SUCC_IN_BBL(bbl, niterator) {
	    if (NODE_TYPE(niterator) == ET_HELL) {
		continue;
	    }

	    if (isInterproceduralEdge(niterator->en_edge) &&
		(NODE_TYPE(niterator) != ET_ENTRY) &&
		(NODE_TYPE(niterator) != ET_EXIT)) {
		continue;
	    }

	    if (NODE_TYPE(niterator) == ET_RETURN) {
		continue;
	    }

	    RecurseCreatePredominatorTree(root, new_node, NODE_TO(niterator));
	}
    }					// end else no common ancestor

}



/*
 *  FunCreatePostdominatorTree(fun) --
 *
 *  Constructs a postdominator tree for the specified function.
 *
 *  fun: the function to create the dominator tree for
 *  returns: the postdominator tree for the specified function
 */
TreeNode *FunCreatePostdominatorTree(Function *fun) {
    TreeNode *root;
    bool renumber_nodes;
    Block *dummy, *exit_bbl;

    renumber_nodes = false;

    // some basic blocks are not actually in the bbl_list (BAD!!) so this
    // assigns BBL_IDs as the actual memory address of the basic block
    // to ensure that they are all unique. This also accounts for those
    // basic block that were not created correctly ie those that did not
    // get created with BBLCreate()
    if (renumber_nodes) {
	FOREACH_BBL(dummy) {
	    dummy->b_id = (int) dummy;
	    dummy->b_tmp = dummy->b_id;
	}
    }

    if (fun->f_exit->b_pred != NULL) {
	exit_bbl = fun->f_exit;
    } else {
	Block *biterator;

	exit_bbl = NULL;

	FOREACH_BBL_IN_FUN(fun, biterator) {
	    if (!isCallClass(biterator->b_last)) {
		EdgeNode *node;

		FOREACH_SUCC_IN_BBL(biterator, node) {
		    if (((NODE_TO(node)) != HELL) &&
                        ((NODE_TO(node))->b_fun) != fun) {
			if (exit_bbl != NULL) {
			    WARNING("two exit blocks for %s", fun->f_name);
			} else {
			    exit_bbl = biterator;
			}
		    }
		}
	    }
	}
    }

    if (exit_bbl == NULL) {
	REPORT(RP_SMLOOP,
	    "Could not compute post-dominator info for %s", fun->f_name);
	return NULL;
    }

    dummy = zalloc(sizeof(Block));
    dummy->b_id = -1;
    BBLAddEdge(dummy, fun->f_entry, ET_NORMAL);

    root = zalloc(sizeof(TreeNode));
    root->t_bbl = dummy;
    root->t_parent = NULL;
    root->t_next_child = NULL;
    root->t_children = NULL;

    RecurseCreatePostdominatorTree(root, root, exit_bbl);

    root = root->t_children;
    BBLRemoveEdges(dummy);
    root->t_parent = NULL;

    free(dummy);

    if (renumber_nodes) {
	FOREACH_BBL(dummy) {
	    dummy->b_id = dummy->b_tmp;
	}
    }

    return root;
}



/*
 *  RecurseCreatePostdominatorTree(root,current,bbl) --
 */
static void RecurseCreatePostdominatorTree(TreeNode *root, TreeNode *current,
    Block *bbl) {
    TreeNode *new_node, *ancestor, *existing;

    ancestor = existing = NULL;

    if (BBLIsAncestorOfNode(current, bbl)) {
	return;
    }

    existing = GetExistingNodeFor(root, bbl);

    if (existing != NULL) {
	// if there is already a node for this basic block in the tree,
	// we want to find the nearest common ancestor for these two
	ancestor = GetNearestCommonAncestor(existing, current);

	CASSERT(ancestor != NULL, "No common ancestor exists");

	if (ancestor == existing->t_parent) {
	    return;
	}

	if (existing->t_parent != NULL) {
	    // remove the child node from it's old parent
	    RemoveChildNode(existing->t_parent, existing);

	    // move the existing node up in the tree s.t. it's new parent
	    // is the nearest common ancestor
	    existing->t_parent = ancestor;
	}
	// add the existing node to the nearest common ancestor
	AddChildNode(ancestor, existing);
    } else {
	// if there is no common ancestor then we will just go ahead and add
	// this basic block into the tree
	EdgeNode *niterator;

	// create a new node for the current basic block
	new_node = DominatorNodeCreate();
	new_node->t_bbl = bbl;
	AddChildNode(current, new_node);

	// for each successor in biterator we want to make sure
	// that it gets added into the dominator tree
	FOREACH_PRED_IN_BBL(bbl, niterator) {
	    if (NODE_TYPE(niterator) == ET_HELL) {
		continue;
	    }

	    if (isInterproceduralEdge(niterator->en_edge) &&
		(NODE_TYPE(niterator) != ET_EXIT) &&
		(NODE_TYPE(niterator) != ET_ENTRY)) {
		continue;
	    }

	    if (NODE_TYPE(niterator) == ET_RETURN) {
		continue;
	    }

	    RecurseCreatePostdominatorTree(root, new_node,
		NODE_FROM(niterator));
	}

    }					// end else no common ancestor

}



/*
 *  DominatorNodeCreate() --
 *
 *  Creates and returns a new dominator node.
 *  returns: and returns a new dominator node.
 */
TreeNode *DominatorNodeCreate() {
    TreeNode *node;

    node = zalloc(sizeof(TreeNode));
    node->t_bbl = NULL;
    node->t_parent = NULL;
    node->t_children = NULL;
    node->t_next_child = NULL;

    return node;
}



/*
 *  RemoveChildNode(parent,child) --
 *
 *  Removes the specified child node from the specified parent dominator node.
 *
 *  parent: the parent that child should be removed from
 *  child: the child of the parent that should be removed
 */
void RemoveChildNode(TreeNode *parent, TreeNode *child) {
    TreeNode *itr;

    if (parent->t_children == child) {
	parent->t_children = parent->t_children->t_next_child;
	return;
    }

    itr = parent->t_children;
    while ((itr != NULL) && (itr->t_next_child != child)) {
	itr = itr->t_next_child;
    }

    if (itr != NULL) {
	itr->t_next_child = itr->t_next_child->t_next_child;
    }
}



/*
 *  GetRootNode(tree) --
 *
 *  Functions
 *
 *  Returns the root node of tree be traversing upwards through parent nodes.
 *
 *  tree: some dominator node in the tree of interest
 *  returns: the root node of the tree containing tree
 */
TreeNode *GetRootNode(TreeNode *tree) {
    if (tree == NULL) {
	return NULL;
    }

    while (tree->t_parent != NULL) {
	tree = tree->t_parent;
    }

    return tree;
}



/*
 *  GetExistingNodeFor(tree,bbl) --
 *
 *  Gets a pointer to any existing node for the specified basic block using tree
 *  as the root node of the tree to search.
 *
 *  tree: the dominator tree to search for prior existence of bbl
 *  bbl: the basic block to find in the tree
 *  returns: a pointer to the existing dominator node for bbl or NULL if there
 *  isn't one
 */
TreeNode *GetExistingNodeFor(TreeNode *tree, Block *bbl) {
    TreeNode *child, *existing;

    if (tree == NULL) {
	return NULL;
    }

    if (tree->t_bbl == bbl) {
	return tree;
    }

    FOREACH_NODE_CHILD(tree, child) {
	existing = GetExistingNodeFor(child, bbl);

	if (existing != NULL) {
	    return existing;
	}
    }

    return NULL;
}



/*
 *  GetNearestCommonAncestor(node1,node2) --
 *
 *  Returns a pointer to the nearest common ancestor of node1 and node2.
 *
 *  node1: the first node of the pair to be tested
 *  node2: the second node of the pair to be tested
 *  returns: a pointer to the nearest common ancestor of node1 and node2 or NULL
 *  if one does not exist.
 */
TreeNode *GetNearestCommonAncestor(TreeNode *node1, TreeNode *node2) {
    TreeNode *child;

    // if either of the nodes are NULL then we know that there is no common
    // ancestor
    if ((node1 == NULL) || (node2 == NULL)) {
	return NULL;
    }

    if (node1 == node2) {
	return node1;
    }
    // see if node1 is an ancestor for node2
    child = GetExistingNodeFor(node1, node2->t_bbl);
    if (child != NULL) {
	return node1;
    }
    // see if node2 is an ancestor of node1
    child = GetExistingNodeFor(node2, node1->t_bbl);
    if (child != NULL) {
	return node2;
    }
    // we now know that there neither of these two nodes are ancestors of each
    // other so we need to look at their parents to see what the nearest common
    // ancestor is between their parents
    return GetNearestCommonAncestor(node1->t_parent, node2->t_parent);
}



/*
 *  AddChildNode(tree,new_node) --
 *
 *  Adds the specified node as a child to the node specified by tree.
 *
 *  tree: the node that new_node should be added as a child to
 *  new_node: the node to add as a child to tree
 */
void AddChildNode(TreeNode *tree, TreeNode *new_node) {
    TreeNode *itr;

    if (tree->t_children == NULL) {
	tree->t_children = new_node;
    } else {
	DOMINATOR_TREE_LAST_CHILD(tree, itr);

	itr->t_next_child = new_node;
    }

    new_node->t_next_child = NULL;
    new_node->t_parent = tree;
}



/*
 *  BBLIsAncestorOfNode(tree,bbl) --
 *
 *  Returns whether or not the specified basic block is an ancestor of the
 *  specified dominator tree node.
 *
 *  tree: the node to test against
 *  bbl: the basic block being tested for ancestorship
 *  returns: true if bbl is an ancestor of tree
 */
bool BBLIsAncestorOfNode(TreeNode *tree, Block *bbl) {
    TreeNode *itr;

    // just check every ancestor in the tree and check to see if the basic
    // block that it encapsulates is the specified one
    itr = tree;
    while (itr != NULL) {
	if (itr->t_bbl == bbl) {
	    return true;
	}

	itr = itr->t_parent;
    }

    return false;
}



/*
 *  FindAllNodesForFunHelper(list,tree,is_the_one)node) --
 */
void FindAllNodesForFunHelper(LinkedList *list,
    TreeNode *tree, bool (*is_the_one) (TreeNode *node)) {
    TreeNode *child;

    if (tree == NULL) {
	return;
    }

    if (is_the_one(tree)) {
	ll_add(list, (void *) tree);
    }

    FOREACH_NODE_CHILD(tree, child) {
	FindAllNodesForFunHelper(list, child, is_the_one);
    }
}



/*
 *  FindAllNodesForFun(tree,is_the_one)node) --
 *
 *  Returns a linked list of all tree nodes for which is_the_one returns
 *  true for during a traversal of the specified tree.
 *
 *  tree: the dominator tree to traverse
 *  is_the_one: a function that will return true for desirable nodes
 *  returns: a linked list containing each of the tree nodes that
 *  is_the_one returned true for
 */
LinkedList *FindAllNodesForFun(TreeNode *tree,
    bool (*is_the_one) (TreeNode *node)) {
    LinkedList *list;

    list = ll_new(LIST_UNSORTED);

    FindAllNodesForFunHelper(list, tree, is_the_one);

    return list;
}



/*
 *  FindFirstNodeForFun(tree,is_the_one)node) --
 *
 *  Returns the first node for which is_the_one returns true in a
 *  traversal of the specified tree.
 *
 *  tree: the dominator tree to traverse
 *  is_the_one: a function that will return true for desirable nodes
 *  returns: the first node for which is_the_one returns true for
 */
TreeNode *FindFirstNodeForFun(TreeNode *tree,
    bool (*is_the_one) (TreeNode *node)) {
    LinkedList *list;
    TreeNode *node;

    list = FindAllNodesForFun(tree, is_the_one);

    node = NULL;

    if (ll_size(list) > 0) {
	node = ll_get(list, 0);
    }
    ll_free(list);

    return node;
}



/*
 *  BBLDominatesBBL(tree,bbl1,bbl2) --
 *
 *  Returns whether or not one basic block in a dominator tree dominates
 *  another basic block.
 *
 *  tree: the dominator tree
 *  bbl1: the basic block that is suspected of dominating
 *  bbl2: the basic block that is suspected of being dominated
 *  returns: true if bbl1 dominates bbl2
 */
bool BBLDominatesBBL(TreeNode *tree, Block *bbl1, Block *bbl2) {
    TreeNode *bbl1_node, *bbl2_node;

    bbl1_node = GetExistingNodeFor(tree, bbl1);
    if (bbl1_node == NULL) {
	//WARNING("given bbl1 does not lie in the specified dominator tree");
	return false;
    }

    bbl2_node = GetExistingNodeFor(tree, bbl2);
    if (bbl2_node == NULL) {
	//WARNING("given bbl2 does not lie in the specified dominator tree");
	return false;
    }

    return NodeDominatesNode(bbl1_node, bbl2_node);
}



/*
 *  NodeDominatesNode(node1,node2) --
 *
 *  Returns whether or not the specified dominator tree node dominates the
 *  node of the specified basic block.
 *
 *  tree: the dominator tree
 *  node1: the dominator tree node that is suspected of dominating
 *  node2: the dominator tree node that is suspected of being dominated
 *  returns: true if node1 dominates node2
 */
bool NodeDominatesNode(TreeNode *node1, TreeNode *node2) {
    while (node2 != NULL) {
	if (node2 == node1) {
	    return true;
	}

	node2 = node2->t_parent;
    }

    return false;
}



/*
 *  GetDepthFirstTraversal(tree) --
 *
 *  Returns a linked list filled with the dominator tree nodes as seen in
 *  a depth first traversal.
 *
 *  tree: the dominator tree to fill the list with
 *  returns: a LinkedList filled with a depth first traversal of the
 *  specified dominator tree
 */
LinkedList *GetDepthFirstTraversal(TreeNode *tree) {
    CASSERT(tree != NULL, "Passed a NULL tree");
    return GetTraversal(tree, &ll_push, &ll_pop);
}



/*
 *  GetBreadthFirstTraversal(tree) --
 *
 *  Returns a linked list filled with the dominator tree nodes as seen in
 *  a breadth first traversal.
 *
 *  tree: the dominator tree to fill the list with
 *  returns: a LinkedList filled with a breadth first traversal of the
 *  specified dominator tree
 */
LinkedList *GetBreadthFirstTraversal(TreeNode *tree) {
    CASSERT(tree != NULL, "Passed a NULL tree");
    return GetTraversal(tree, &ll_enqueue, &ll_dequeue);
}



/*
 *  GetTraversal(tree,list_add)list,val)list_remove)list) --
 */
static LinkedList *GetTraversal(TreeNode *tree,
    void *(*list_add) (LinkedList *list, void *val),
    void *(*list_remove) (LinkedList *list)) {
    TreeNode *node, *itr;
    LinkedList *result, *list;

    result = ll_new(LIST_UNSORTED);
    list = ll_new(LIST_UNSORTED);

    list_add(list, tree);
    while (ll_size(list) > 0) {
	node = (TreeNode *) list_remove(list);

	ll_add(result, node);

	FOREACH_NODE_CHILD(node, itr) {
	    list_add(list, itr);
	}
    }

    ll_free(list);

    return result;
}



/*
 *  IsBackEdge(tree,edge) --
 */
static bool IsBackEdge(TreeNode *tree, Edge *edge) {
    return BBLDominatesBBL(tree, edge->e_to, edge->e_from);
}



/*
 *  CreateLoop(tree,head,tail) --
 */
static void CreateLoop(TreeNode *tree, Block *head, Block *tail) {
    TreeNode *head_node, *tail_node;

    head_node = GetExistingNodeFor(tree, head);
    if (head_node == NULL) {
	WARNING("Specified head does not lie in specified dominator tree");
	return;
    }

    tail_node = GetExistingNodeFor(tree, tail);
    if (tail_node == NULL) {
	WARNING("Specified tail does not lie in specified dominator tree");
	return;
    }

    BBL_SET_FLAG(head, BBL_LOOP_HEADER_FLAG);

    // if we have a single block loop we don't want to set it to be it's
    // own header since it could still be nested
    if (tail == head) {
	return;
    }

    tail->b_loopHead = head;

    while (tail_node != head_node) {
	tail = tail_node->t_bbl;
	tail->b_loopHead = head;

	tail_node = tail_node->t_parent;
    }
}



/*
 *  LinkLoops(tree) --
 */
static void LinkLoops(TreeNode *tree) {
    LinkedList *list;
    Iterator *itr;
    TreeNode *node;
    Block *bbl;

    list = GetBreadthFirstTraversal(tree);
    itr = ll_iterator(list);
    while (itr->it_hasNext(itr)) {
	node = (TreeNode *) itr->it_next(itr);
	bbl = node->t_bbl;

	if (bbl->b_loopHead != NULL) {
	    BBLAddToLoop(bbl, bbl->b_loopHead);
	}
    }
    free(itr);
}



/*
 *  StoreLoopInfo(tree) --
 *
 *  Computes and stores loop information for the specified dominator tree
 *  directly into the auxiliary information for each basic block
 *  concerned.
 *
 *  tree: the dominator tree to compute loop relationships for
 */
void StoreLoopInfo(TreeNode *tree) {
    TreeNode *node;
    LinkedList *list;
    Iterator *itr;
    EdgeNode *predn;
    Block *bbl;

    list = GetBreadthFirstTraversal(tree);

    // clear all of the bbl info from the concerned basic blocks
    itr = ll_iterator(list);
    while (itr->it_hasNext(itr)) {
	node = (TreeNode *) itr->it_next(itr);

	node->t_bbl->b_loopFirst = NULL;
	node->t_bbl->b_loopNext = NULL;
	node->t_bbl->b_loopHead = NULL;
    }
    free(itr);

    itr = ll_iterator(list);
    while (itr->it_hasNext(itr)) {
	node = (TreeNode *) itr->it_next(itr);
	bbl = node->t_bbl;

	FOREACH_PRED_IN_BBL(bbl, predn) {
	    if (isInterproceduralEdge(predn->en_edge)) {
		continue;
	    }

	    if (IsBackEdge(tree, predn->en_edge)) {
		CreateLoop(tree, bbl, NODE_FROM(predn));
	    }
	}
    }
    free(itr);
    ll_free(list);

    LinkLoops(tree);
}



/*
 *  RecursePrintDominatorTree(root,outfile) --
 */
static void RecursePrintDominatorTree(TreeNode *root, FILE *outfile) {
    TreeNode *children;

    if (root == NULL) {
	return;
    }

    if (root->t_parent != NULL) {
	fprintf(outfile, "  b_%d -> b_%d\n",
	    root->t_parent->t_bbl->b_id, root->t_bbl->b_id);
    }

    children = root->t_children;
    while (children != NULL) {
	RecursePrintDominatorTree(children, outfile);
	children = children->t_next_child;
    }
}



/*
 *  PrintDominatorTree(tree,filename) --
 */
static void PrintDominatorTree(TreeNode *tree, const char *filename) {
    FILE *outfile;

    outfile = fopen(filename, "w");
    if (!outfile) {
	WARNING("Could not open outfile: %s", filename);
	return;
    }

    fprintf(outfile, "digraph %s {\n", filename);
    fprintf(outfile, "  center=1;\n");
    fprintf(outfile, "  ratio=fill;\n");

    RecursePrintDominatorTree(tree, outfile);

    fprintf(outfile, "}");
    fclose(outfile);
}



/*
 *  TestDominatorCode() --
 */
void TestDominatorCode() {
    Function *fun;
    Block *bbl1, *bbl2, *bbl3, *bbl4, *bbl5, *bbl6, *bbl7, *bbl8, *bbl9, *bbl10;
    TreeNode *tree;

    REPORT(RP_MAJOR, "Testing dominator code");

#define NEW_BBL(bbl, id) \
    (bbl = BBLCreate(), \
    ((bbl)->b_id) = id, \
    ((bbl)->b_succ) = NULL, \
    ((bbl)->b_pred) = NULL, \
    bbl)
#define INSERT_NEW_BBL(bbl, last, id) \
    { bbl = NEW_BBL(bbl, id); \
    InsertBBLAfterBBL(bbl, last); \
    ((bbl)->b_fun) = fun; }

    fun = FunCreate();
    fun->f_name = "test_fun";
    fun->f_entry = NEW_BBL(bbl1, 99);
    bbl1->b_fun = fun;
    fun->f_exit = NEW_BBL(bbl1, 999);
    bbl1->b_fun = fun;
    fun->f_first = fun->f_last = NEW_BBL(bbl1, 1);
    bbl1->b_fun = fun;

    INSERT_NEW_BBL(bbl2, bbl1, 2);
    INSERT_NEW_BBL(bbl3, bbl2, 3);
    INSERT_NEW_BBL(bbl4, bbl3, 4);
    INSERT_NEW_BBL(bbl5, bbl4, 5);
    INSERT_NEW_BBL(bbl6, bbl5, 6);
    INSERT_NEW_BBL(bbl7, bbl6, 7);
    INSERT_NEW_BBL(bbl8, bbl7, 8);
    INSERT_NEW_BBL(bbl9, bbl8, 9);
    INSERT_NEW_BBL(bbl10, bbl9, 10);

#undef NEW_BBL
#undef INSERT_NEW_BBL

    BBLAddEdge(bbl1, bbl2, ET_TRUE);
    BBLAddEdge(bbl1, bbl3, ET_FALSE);

    BBLAddEdge(bbl2, bbl3, ET_UNCOND);

    BBLAddEdge(bbl3, bbl4, ET_UNCOND);

    BBLAddEdge(bbl4, bbl5, ET_TRUE);
    BBLAddEdge(bbl4, bbl6, ET_FALSE);
    BBLAddEdge(bbl4, bbl3, ET_NORMAL);

    BBLAddEdge(bbl5, bbl7, ET_UNCOND);

    BBLAddEdge(bbl6, bbl7, ET_UNCOND);

    BBLAddEdge(bbl7, bbl8, ET_FALSE);
    BBLAddEdge(bbl7, bbl4, ET_TRUE);

    BBLAddEdge(bbl8, bbl9, ET_TRUE);
    BBLAddEdge(bbl8, bbl10, ET_FALSE);

    BBLAddEdge(bbl9, bbl1, ET_UNCOND);

    BBLAddEdge(bbl10, bbl7, ET_TRUE);
    BBLAddEdge(bbl10, bbl3, ET_FALSE);

    BBLAddEdge(fun->f_entry, bbl1, ET_ENTRY);

    printf("Beginning Dominator Tree Creation\n");
    tree = FunCreatePredominatorTree(fun);
    printf("Done with Dominator Tree Creation\n");

    // PrintFunctionCFG(fun, NULL, NULL, "test_fun_cfg");

    REPORT(RP_MINOR, "Printing dominator tree");
    PrintDominatorTree(tree, "test_fun_dom_tree");

    REPORT(RP_MINOR, "Testing breadth-first code");
    {
	LinkedList *list;
	Iterator *itr;
	TreeNode *node;

	list = GetBreadthFirstTraversal(tree);
	itr = ll_iterator(list);
	while (itr->it_hasNext(itr)) {
	    node = (TreeNode *) itr->it_next(itr);

	    printf("  BBL: %d\n", node->t_bbl->b_id);
	}
    }

    REPORT(RP_MINOR, "Testing depth-first code");
    {
	LinkedList *list;
	Iterator *itr;
	TreeNode *node;

	list = GetDepthFirstTraversal(tree);
	itr = ll_iterator(list);
	while (itr->it_hasNext(itr)) {
	    node = (TreeNode *) itr->it_next(itr);

	    printf("  BBL: %d\n", node->t_bbl->b_id);
	}
    }

    REPORT(RP_MINOR, "Storing loop information");
    StoreLoopInfo(tree);
    {
	LinkedList *list;
	Iterator *itr;
	TreeNode *node;
	Block *bbl, *biterator, *head;

	list = GetBreadthFirstTraversal(tree);
	itr = ll_iterator(list);
	while (itr->it_hasNext(itr)) {
	    node = (TreeNode *) itr->it_next(itr);
	    bbl = node->t_bbl;

	    if (bbl->b_loopFirst != NULL) {
		printf("Loop Head: BBL: %d\n", bbl->b_id);
		head = bbl;

		biterator = bbl;
		while (biterator != NULL) {
		    if (biterator->b_loopHead == bbl) {
			printf("  BBL: %d\n", biterator->b_id);
		    }

		    if (biterator == head) {
			biterator = biterator->b_loopFirst;
		    } else {
			biterator = biterator->b_loopNext;
		    }
		}
	    }
	}
    }
}



/*
 *  DominatorTreeFree(tree) --
 *
 *  Frees up the memory for all nodes in the specified dominator tree.
 *
 *  tree: the dominator tree to free up
 */
void DominatorTreeFree(TreeNode *tree) {
    LinkedList *nodes;
    Iterator *itr;

    nodes = GetBreadthFirstTraversal(tree);
    itr = ll_iterator(nodes);
    while (itr->it_hasNext(itr)) {
	free(itr->it_next(itr));
    }
    free(itr);
    ll_free(nodes);
}



/*
 *  StoreDominatorInfoForFun(fun) --
 *
 *  This function computes all pre-dominator, post-dominator, and loop
 *  information and stores it into the proper portions of each basic
 *  blocks' auxiliary info struct for the specified function.
 *  see also: ->b_preDom
 *  see also: ->b_postDom
 *  see also: ->b_loopHead
 *  see also: ->b_loopNext
 *  see also: ->b_loopFirst
 */
void StoreDominatorInfoForFun(Function *fun) {
    Block *biterator, *bbl;
    TreeNode *tree, *node;
    LinkedList *nodes;
    Iterator *itr;

    FOREACH_BBL_IN_FUN(fun, biterator) {
	biterator->b_preDom = biterator;
	biterator->b_postDom = biterator;
    }

    tree = FunCreatePredominatorTree(fun);
    if (tree != NULL) {
	nodes = GetBreadthFirstTraversal(tree);
	itr = ll_iterator(nodes);
	while (itr->it_hasNext(itr)) {
	    node = (TreeNode *) itr->it_next(itr);
	    bbl = node->t_bbl;
	    bbl->b_preDom = node->t_parent ?  node->t_parent->t_bbl : NULL;
	}
	free(itr);
	ll_free(nodes);

	StoreLoopInfo(tree);
	DominatorTreeFree(tree);
    }

    tree = FunCreatePostdominatorTree(fun);
    if (tree != NULL) {
	nodes = GetBreadthFirstTraversal(tree);
	itr = ll_iterator(nodes);
	while (itr->it_hasNext(itr)) {
	    node = (TreeNode *) itr->it_next(itr);
	    bbl = node->t_bbl;
	    bbl->b_postDom = node->t_parent ?  node->t_parent->t_bbl : NULL;
	}
	free(itr);
	ll_free(nodes);
	DominatorTreeFree(tree);
    }
}



/*
 *  StoreDominatorInfo() --
 *
 *  This function computes all pre-dominator, post-dominator, and loop
 *  information and stores it into the proper portions of each basic
 *  blocks' auxiliary info struct for all functions.
 *  see also: StoreDominatorInfoForFun
 */
void StoreDominatorInfo() {
    Function *fiterator;

    REPORT(RP_MINOR, WITHFUNC("Computing dominator information"));

    FOREACH_FUNCTION(fiterator) {
	StoreDominatorInfoForFun(fiterator);
    }
}



/*
 *  BBLAddToLoop(bbl,head) --
 */
void BBLAddToLoop(Block *bbl, Block *head) {
    bbl->b_loopHead = head;
    bbl->b_loopNext = head->b_loopFirst;
    head->b_loopFirst = bbl;
}
