
/*
 *  Handles the insertion of blocks to profile edge frequencies.
 */
#include "global.h"
#include "edge.profile.h"
#include "flowgraph.h"
#include "utilities.h"
#include "linkage.h"
#include "opt.layout.h"
#include "sysdep.h"
#include "relocate.h"
#include "upm.h"
#include "unistd.h"

static void ProfileBBLs();
static void AddBBLCounter(Block *block);
static void ProfileEdges();
static void ProfileEdge(EdgeNode *node);
static void AddEdgeCounter(Block *block, int jtslot, ADDRESS source,
    ADDRESS dest);
static Edge *DetectDoubleSwitchEdgeForBBL(Block *switchBlock);
static void RemoveDoubleSwitchEdges();

static Block *lastBlock;
static int bblProfileCounterBase;   // offset to bbl profiles in .data
static int edgeProfileCounterBase;  // offset to edge profiles in .data



/*
 *  Profile() --
 *
 *  Profile performs bbl profiling or edge profiling or both by calling the
 *  appropriate functions.
 */
void Profile() {

    /*
     * Entry/Exit/Return/Funcall edges will not be profiled, as they only
     * exist to make things easier. Control cannot actually "flow" along
     * these edges, so inserting our counters between blocks connected by
     * these edges wouldn't have any effect.
     */
    RemoveEntryExitReturnFuncallEdges();
    RemoveEntryExitBlocks();
    RemoveDoubleSwitchEdges();

    ProfInit();

    if (PROFILE_BBLS) {
        ProfileBBLs();
    }

    if (PROFILE_EDGES) {
        ProfileEdges();
    }

    ProfExit(profFilename);

    return;
}



/*
 *  ProfileBBLs() --
 *
 *  ProfileBBLs will insert some counters at the beginning of every basic
 *  block in the flowgraph. These counters will keep tab of how
 *  many times the block was traversed during execution.
 */
void ProfileBBLs() {
    Block *block_iterator;

    bblProfiles = 0;

    bblProfileCounterBase = SECTION_REF_DATA->s_growth +
        SECTION_REF_DATA->s_datasize;

    /* Grow the data size to accommodate the number of blocks we have. */
    IncreaseSectionSize(SECTION_REF_DATA, BBL_PROFILE_DATA_SIZE * numBlocks);

    /* For all blocks, insert counters */
    FOREACH_BBL(block_iterator) {
        if (block_iterator == HELL) {
            continue;
        }
        if (BBL_HAS_FLAG(block_iterator, BBL_PROFILED_FLAG)) {
            continue;
        }
        AddBBLCounter(block_iterator);
    }

    REPORT(RP_DETAIL, "Profiled %d bbls.", bblProfiles);

    return;
}



/*
 *  AddBBLCounter(bbl) --
 *
 *  Profiles a basic block by inserting the following 3 instructions into it:
 *    push PSR/PSW
 *    increment counter
 *    pop PSR/PSW
 *  In most of the cases the profiling code is inserted after the first
 *  instruction in the bbl. It is inserted before the first instruction only
 *  when bbl has one instruction and this instruction is either a pc changing
 *  or a data instruction.
 */
static void AddBBLCounter(Block *block) {
    Instruction *save, *restore, *add, *first;
    int offset;
    EdgeNode *node_iterator;
    bool found;

    first = block->b_first;
    CASSERT(first != NULL, "Passed an empty basic block.");

    save = InstrCreateSave();
    restore = InstrCreateRestore();

    add = InstrCreate32BitMemAdd(bblProfileCounterBase +
        bblProfiles * BBL_PROFILE_DATA_SIZE);
    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA,
        bblProfileCounterBase + bblProfiles * BBL_PROFILE_DATA_SIZE,
        add, add->i_dest);

    add->i_src->a_displacement = 1;

    offset = profile.pc_currentStart;
    profile.pc_currentStart += BBL_PROFILE_DATA_SIZE;
    profile.pc_bblsProf++;

    /* now write the counter info here */
    storeLE(&(SECTION_REF_DATA->s_newdata[offset]), 0);
    storeLE(&(SECTION_REF_DATA->s_newdata[offset + 4]), 0);
    storeLE(&(SECTION_REF_DATA->s_newdata[offset + 8]), block->b_id);

    bblProfiles++;

    /* if bbl contains only one instruction and it is a pc changing or a data
     * instruction, the counter will be added before it */
    if ((block->b_first == block->b_last) && (isPCChangingClass(block->b_last)
        || (INSTR_HAS_FLAG(block->b_last, INSTR_DATA_BYTE)))) {
        InsertInstrBeforeInstr(save, first);
        save->i_address = first->i_address;
    } else { /* otherwise it will be added after the first instruction in bbl */
        InsertInstrAfterInstr(save, first);
    }

    InsertInstrAfterInstr(add, save);
    InsertInstrAfterInstr(restore, add);

    node_iterator = block->b_pred;
    while (node_iterator != NULL) {
        if (NODE_TYPE(node_iterator) == ET_SWITCH) {

            found = false;

            {
                Patch *instr_upm, *tbl_upm;
                LinkedList *upms;
                PatchSectionToInstr *instr_patch;
                PatchInstrToSection *tbl_patch;

                instr_upm = tbl_upm = NULL;

                upms = 
                    UPMTableGetByDest(NODE_FROM(node_iterator)->b_last->i_src);
                if (ll_size(upms) > 0) {
                    instr_upm = ll_removeFirst(upms);
                }
                ll_free(upms);

                if (instr_upm != NULL) {
                    instr_patch = (PatchSectionToInstr *) instr_upm->p_patch;

                    upms = UPMTableGetByDestAddress(instr_patch->p_src_section,
                        instr_patch->p_src_offset +
                        NODE_JT_SLOT(node_iterator) * sizeof(ADDRESS));

                    if (ll_size(upms) > 0) {
                        tbl_upm = ll_removeFirst(upms);
                    }
                    ll_free(upms);

                    if (tbl_upm != NULL) {
                        tbl_patch = (PatchInstrToSection *) tbl_upm->p_patch;
                        found = UPMChangeSourceFromInstrToInstr(tbl_upm,
                            tbl_patch->p_src_instr, block->b_first);
                    }
                }
            }
        }
        node_iterator = node_iterator->en_next;
    }

    return;
}



/*
 *  ProfileEdges() --
 *
 *  ProfileEdges will insert some counters between connected basic
 *  blocks in the flowgraph. These counter-blocks will keep tab of how
 *  many times the edge was traversed during execution.
 */
void ProfileEdges() {
    Block *block_iterator;
    EdgeNode *node_iterator;

    edgeProfiles = 0;

    edgeProfileCounterBase = SECTION_REF_DATA->s_growth +
	SECTION_REF_DATA->s_datasize;

    /* Grow the data size to accommodate the maximum number of edges we have.
     * Note that some edges (pseudo ones, HELL ones) will not get profiles,
     * so this is an over-estimate */
    IncreaseSectionSize(SECTION_REF_DATA, EDGE_PROFILE_DATA_SIZE * globalEID);

    lastBlock = bbl_list;
    while (lastBlock->b_next != NULL) {
	lastBlock = lastBlock->b_next;
    }

    /* For all the remaining edges in the live blocks, insert counters */
    FOREACH_BBL(block_iterator) {
	if (block_iterator == HELL) {
	    continue;
	}
	if (BBL_HAS_FLAG(block_iterator, BBL_PROFILED_FLAG)) {
	    continue;
	}
	node_iterator = block_iterator->b_succ;
	while (node_iterator != NULL) {
	    EdgeNode *next = node_iterator->en_next;
	    ProfileEdge(node_iterator);
	    node_iterator = next;
	}
    }

    REPORT(RP_DETAIL, "Profiled %d edges.", edgeProfiles);

    return;
}



/*
 *  ProfileEdge(node) --
 *
 *  Profiles an edge by inserting a basic block between the two endpoints
 *  and redirecting control transfer to flow through the block. A counter is
 *  then inserted into the block to keep track of edge frequency.
 */
static void ProfileEdge(EdgeNode *node) {
    Block *newBlock = BBLCreate();
    Instruction *jmp = InstrCreateJMP();
    bool found = false;

    if (NODE_TYPE(node) == ET_HELL) {
	return;
    }
    if (NODE_TYPE(node) == ET_FUNLINK) {
	return;
    }

    /* Set up our new profile block */
    INSTR_SET_FLAG(jmp, BBL_START);
    INSTR_SET_FLAG(jmp, BBL_END);
    BBL_SET_FLAG(newBlock, BBL_PROFILED_FLAG);
    jmp->i_Bblock = newBlock;
    newBlock->b_type = BT_UNCOND;
    newBlock->b_first = newBlock->b_last = jmp;

    AddEdgeCounter(newBlock, NODE_JT_SLOT(node),
	NODE_FROM(node)->b_first->i_address, NODE_TO(node)->b_first->i_address);

    /*
     * In the most general, and also the most common case, we will simply
     * add our newblock (complete with counters) either directly after
     * source block, or at the end of the flowgraph.  It will be added
     * directly after the source block when exploiting the fall-through
     * of an ET_NORMAL or ET_FALSE edge can be beneficial.  Otherwise,
     * we'll simply put it at the end of the executable and add an edge
     * to it.
     */

    found = true;
    switch (NODE_TYPE(node)) {

    case ET_NORMAL:
	InsertBBLAfterBBL(newBlock, NODE_FROM(node));
	BBLAddEdge(NODE_FROM(node), newBlock, ET_NORMAL);
	BBLAddEdge(newBlock, NODE_TO(node), ET_UNCOND);
	break;

    case ET_SYSCALL_LINK:
	InsertBBLAfterBBL(newBlock, NODE_FROM(node));
	BBLAddEdge(NODE_FROM(node), newBlock, ET_NORMAL);
	BBLAddEdge(newBlock, NODE_TO(node), ET_SYSCALL_LINK);
	break;

    case ET_UNCOND:
	InsertBBLAfterBBL(newBlock, NODE_FROM(node));
	BBLAddEdge(NODE_FROM(node), newBlock, ET_UNCOND);
	BBLAddEdge(newBlock, NODE_TO(node), ET_UNCOND);
	break;

    case ET_TRUE:
	InsertBBLAfterBBL(newBlock, lastBlock);
	BBLAddEdge(NODE_FROM(node), newBlock, ET_TRUE);
	BBLAddEdge(newBlock, NODE_TO(node), ET_UNCOND);
	break;

    case ET_FALSE:
	InsertBBLAfterBBL(newBlock, NODE_FROM(node));
	BBLAddEdge(NODE_FROM(node), newBlock, ET_FALSE);
	BBLAddEdge(newBlock, NODE_TO(node), ET_UNCOND);
	break;

    case ET_FUNDIRECT:
	InsertBBLAfterBBL(newBlock, lastBlock);
	BBLAddEdge(NODE_FROM(node), newBlock, ET_FUNDIRECT);
	BBLAddEdge(newBlock, NODE_TO(node), ET_UNCOND);
	break;

	/*
	 * Switch edges are tricky.  The strategy used above for other
	 * edges will not work here.  This is because edges of type
	 * ET_SWITCH are traversed when an instruction makes an
	 * jump through a jump table. What we want to do is find the
	 * jump table entry and change it to point to our newblock, and
	 * have our newblock then re-route to the original target. The
	 * relocation item for the jump table entry should also be
	 * removed, so that our hard work isn't undone come patching
	 * time.
	 */

    case ET_SWITCH:

	found = false;

	{
	    Patch *instr_upm, *tbl_upm;
	    LinkedList *upms;
	    PatchSectionToInstr *instr_patch;
	    PatchInstrToSection *tbl_patch;

	    instr_upm = tbl_upm = NULL;

	    upms = UPMTableGetByDest(NODE_FROM(node)->b_last->i_src);
	    if (ll_size(upms) > 0) {
		instr_upm = ll_removeFirst(upms);
	    }
	    ll_free(upms);

	    if (instr_upm != NULL) {
		instr_patch = (PatchSectionToInstr *) instr_upm->p_patch;

		upms = UPMTableGetByDestAddress(instr_patch->p_src_section,
		    instr_patch->p_src_offset +
		    NODE_JT_SLOT(node) * sizeof(ADDRESS));
		if (ll_size(upms) > 0) {
		    tbl_upm = ll_removeFirst(upms);
		}
		ll_free(upms);

		if (tbl_upm != NULL) {
		    tbl_patch = (PatchInstrToSection *) tbl_upm->p_patch;
		    found = UPMChangeSourceFromInstrToInstr(tbl_upm,
			tbl_patch->p_src_instr, newBlock->b_first);
		}
	    }
	}

	if (found) {
	    InsertBBLAfterBBL(newBlock, lastBlock);
	    BBLAddEdge(NODE_FROM(node), newBlock, ET_SWITCH)->e_jumpTableSlot =
		NODE_JT_SLOT(node);
	    BBLAddEdge(newBlock, NODE_TO(node), ET_UNCOND);
	} else {
	    WARNING("Unable to insert profiling data for switch edge"
		" from block at 0x%8.8x. Not rerouting edge.",
		NODE_FROM(node)->b_first->i_address);
	}

	break;

    case ET_DATALINK:
        found = false;
        WARNING("Edge type %s. Not rerouting edge.",
            edgeTypeToString(NODE_TYPE(node)));
        break;

    default:
	found = false;
	WARNING("Edge type %s not known. Not rerouting edge.",
	    edgeTypeToString(NODE_TYPE(node)));
	break;

    }

    if (found) {
	BBLRemoveEdge(NODE_FROM(node), NODE_TO(node));
    }

    return;
}



/*
 *  AddEdgeCounter(block,jtslot,source,dest) --
 *
 *  Adds the appropriate instructions to the profiling-bbl. Note that
 *  the creation of these instructions is highly architecture dependent.
 *  The type of basic block we're trying to create will look like:
 *
 *             Source BBL
 *                  |
 *  ________________|_______________
 *  |                              |
 *  |          push PSR/PSW        |
 *  |        increment counter     |
 *  |           pop PSR/PSW        |
 *  |          jump dest bbl       |
 *  |______________________________|
 *                  |
 *           Destination BBL
 */
static void AddEdgeCounter(Block *block, int jtslot, ADDRESS source,
    ADDRESS dest) {
    Instruction *save, *restore, *add, *first;
    int offset;

    save = InstrCreateSave();
    restore = InstrCreateRestore();

    add = InstrCreate32BitMemAdd(edgeProfileCounterBase +
	edgeProfiles * EDGE_PROFILE_DATA_SIZE);
    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA,
	edgeProfileCounterBase + edgeProfiles * EDGE_PROFILE_DATA_SIZE,
	add, add->i_dest);

    add->i_src->a_displacement = 1;

    offset = profile.pc_currentStart;
    profile.pc_currentStart += EDGE_PROFILE_DATA_SIZE;
    profile.pc_edgsProf++;

    /* now write the counter info here */
    storeLE(&(SECTION_REF_DATA->s_newdata[offset]), 0);
    storeLE(&(SECTION_REF_DATA->s_newdata[offset + 4]), 0);
    storeLE(&(SECTION_REF_DATA->s_newdata[offset + 8]), source);
    storeLE(&(SECTION_REF_DATA->s_newdata[offset + 12]), dest);
    storeLE(&(SECTION_REF_DATA->s_newdata[offset + 16]), jtslot);

    edgeProfiles++;

    first = block->b_first;
    CASSERT(first != NULL, "Passed an empty basic block.");

    InsertInstrBeforeInstr(restore, first);
    InsertInstrBeforeInstr(add, restore);
    InsertInstrBeforeInstr(save, add);

    block->b_first = save;
}



/*
 *  RemoveDoubleSwitchEdges() --
 *
 *  Removes parallel switch edges from the ICFG.
 */
static void RemoveDoubleSwitchEdges() {
    Block *bbl;
    Edge *parallel;

#if 1
    return;
#endif

    FOREACH_BBL(bbl) {
	if (bbl->b_type != BT_SWITCH) {
	    continue;
	}
	while (true) {
	    parallel = DetectDoubleSwitchEdgeForBBL(bbl);
	    if (parallel != NULL) {
		BBLRemoveEdgeTyped(bbl, parallel->e_to, ET_SWITCH);
	    } else {
		break;
	    }
	}
    }

    return;
}



/*
 *  DetectDoubleSwitchEdgeForBBL(switchBlock) --
 *
 *  Given a switch bbl, this function returns a parallel outgoing edge if one
 *  exists, or NULL otherwise.
 */
static Edge *DetectDoubleSwitchEdgeForBBL(Block *switchBlock) {
    Edge *result = NULL;
    EdgeNode *succs;

    CASSERT(switchBlock->b_type == BT_SWITCH, "Given a non-BT_SWITCH block.");

    FOREACH_SUCC_IN_BBL(switchBlock, succs) {
	BBL_REMOVE_FLAG(NODE_TO(succs), BBL_TEMP_FLAG);
    }

    FOREACH_SUCC_IN_BBL(switchBlock, succs) {
	if (BBL_HAS_FLAG(NODE_TO(succs), BBL_TEMP_FLAG)) {
	    if (NODE_TYPE(succs) == ET_SWITCH) {
		result = succs->en_edge;
	    }
	}
	BBL_SET_FLAG(NODE_TO(succs), BBL_TEMP_FLAG);
    }

    return result;
}



/*
 *  ReadProfiles() --
 *
 *  Reads in bbl and edge profiles and assigns weights to the bbls and edges
 *  based on the data read.
 */
void ReadProfiles() {
    Block *biterator;
    EdgeNode *niterator;
    bool found;
    unsigned int bcount, ecount, ccount, jcount, fcount;
    unsigned int boffset, eoffset, coffset, joffset, foffset;
    unsigned int nentries;
    ADDRESS startAddr, endAddr;
    EDGE_WEIGHT_TYPE count;
    UINT32 bbl_id;
    Instruction *source, *dest;
    int i, jtslot;
    ProfileFile *next_profile;
    FILE *profileFile;
    int numBBLs = 1;
    Block **bblArr;

    if (profileFiles == NULL) {
        return;
    }

    /* get the number of basic blocks (the size of array of pointers to them) */
    FOREACH_BBL(biterator) {
        if (biterator->b_id > numBBLs) {
            numBBLs = biterator->b_id;
        }
    }

    bblArr = (Block **)zalloc((++numBBLs) * sizeof(Block *));

    /* fill in the array with the bbls addresses and set weights to 0 */
    FOREACH_BBL(biterator) {
        if (biterator->b_id > numBBLs) {
            WARNING("Block id > number of BBLs");
        }

        if (bblArr[biterator->b_id] != NULL) {
            WARNING("The array element is already full");
        } else {
            bblArr[biterator->b_id] = biterator;
        }
        biterator->b_weight = 0;
        FOREACH_SUCC_IN_BBL(biterator, niterator) {
            NODE_WEIGHT(niterator) = 0;
        }

    }

    /* read and store the profile info from the input file(s) */
    next_profile = profileFiles;
    while (next_profile) {
        REPORT(RP_MINOR, "Loading bbl and edge profile file: %s",
            next_profile->pf_name);

        profileFile = next_profile->pf_file;
        fread(&nentries, ADDRESS_SIZE, 1, profileFile);
        if (nentries != ALTO_MAGIC) {
            WARNING("count file %s does not use proper magic number",
                next_profile->pf_name);
            return;
        }

        fread(&bcount, sizeof(bcount), 1, profileFile);
        fread(&boffset, sizeof(boffset), 1, profileFile);
        fread(&ecount, sizeof(ecount), 1, profileFile);
        fread(&eoffset, sizeof(eoffset), 1, profileFile);
        fread(&ccount, sizeof(ccount), 1, profileFile);
        fread(&coffset, sizeof(coffset), 1, profileFile);
        fread(&jcount, sizeof(jcount), 1, profileFile);
        fread(&joffset, sizeof(joffset), 1, profileFile);
        fread(&fcount, sizeof(fcount), 1, profileFile);
        fread(&foffset, sizeof(foffset), 1, profileFile);
        fread(&nentries, sizeof(nentries), 1, profileFile);

        REPORT(RP_DETAIL, "%10d bblk counters at 0x%x", bcount, boffset);
        REPORT(RP_DETAIL, "%10d edge counters at 0x%x", ecount, eoffset);
        REPORT(RP_DETAIL, "%10d call counters at 0x%x", ccount, coffset);
        REPORT(RP_DETAIL, "%10d jump counters at 0x%x", jcount, joffset);
        REPORT(RP_DETAIL, "%10d func counters at 0x%x", fcount, foffset);
        REPORT(RP_DETAIL, "%10d entries", nentries);

        bblProfiles = bcount;
        edgeProfiles = ecount;

        if (bcount == 0 && ecount == 0) {
            WARNING("Count file %s does not contain any profiling information",
                next_profile->pf_name);
            return;
        } else if (bcount == 0) {
            REPORT(RP_DETAIL, "No bbl profile,"
                " bbl weights will be computed from edge profile");
        } else if (ecount == 0) {
            REPORT(RP_DETAIL, "No edge profile,"
                " edge weights will be computed from bbl profile");
        }

        /* read the bbl profiles */
        fseek(profileFile, boffset, SEEK_SET);
        for (i = 0; i < bcount; i++) {
            fread(&count, sizeof(EDGE_WEIGHT_TYPE), 1, profileFile);
            fread(&bbl_id, ADDRESS_SIZE, 1, profileFile);

            if (bbl_id >= numBBLs) {
                WARNING("bbl id in .counts file is > the highest bbl id"
                    " in the binary");
                continue;
            }

            biterator = bblArr[bbl_id];
            biterator->b_weight += count;
        }

        //free(bblArr);

        /* read the edge profiles */
        fseek(profileFile, eoffset, SEEK_SET);
        for (i = 0; i < ecount; i++) {

            if (feof(profileFile)) {
                break;
            }

            fread(&count, sizeof(EDGE_WEIGHT_TYPE), 1, profileFile);
            fread(&startAddr, ADDRESS_SIZE, 1, profileFile);
            fread(&endAddr, ADDRESS_SIZE, 1, profileFile);
            fread(&jtslot, sizeof(int), 1, profileFile);

            if ((startAddr == 0x0) || (endAddr == 0x0)) {
                break;
            }

            source = AddressToInstr(startAddr);
            dest = AddressToInstr(endAddr);

            if (source == NULL) {
                WARNING("[ReadEdgeProfiles] Could not locate source bbl at"
                    " 0x%8.8x", startAddr);
            }
            if (dest == NULL) {
                WARNING("[ReadEdgeProfiles] Could not locate target bbl at"
                    " 0x%8.8x", endAddr);
            }

            if (count < 0) {
                WARNING("[ReadEdgeProfiles] count < 0");
            }

            if ((source == NULL) || (dest == NULL)) {
                continue;
            }

            found = false;
            biterator = source->i_Bblock;
            FOREACH_SUCC_IN_BBL(biterator, niterator) {
                if (dest == NODE_TO(niterator)->b_first) {
                    if (NODE_JT_SLOT(niterator) != jtslot) {
                        continue;
                    }
                    NODE_WEIGHT(niterator) += count;
                    found = true;
                    if (NODE_TYPE(niterator) == ET_FUNDIRECT) {
                        BBLGetFunlinkEdge(biterator)->e_edgeWeight += count;

                        // we don't give PIC calls entry or exit edges so we'll
                        // just skip those bad boys
                        if (NODE_TO(niterator)->b_fun ==
                            NODE_FROM(niterator)->b_fun) {
                            continue;
                        }

                        BBLGetEdgeToBBL(NODE_TO(niterator)->b_fun->f_exit,
                            BBLGetFunlinkEdge(biterator)->e_to)
                            ->e_edgeWeight += count;
                    }
                }
            }

            if (found == false) {
                WARNING("[ReadEdgeProfiles] Could not find edge from source bbl"
                    " at 0x%8.8x to target bbl 0x%8.8x", startAddr, endAddr);
            }
        }

        next_profile = next_profile->pf_next;
    }

    return;
}



/*
 *  ProfInit() --
 *
 *  Sets up the profiling data structures
 */
void ProfInit() {
    profile.pc_initStart = SECTION_REF_DATA->s_newdatasize;

    IncreaseSectionSize(SECTION_REF_DATA, PROFILE_INIT_SIZE);

    profile.pc_counterStart = SECTION_REF_DATA->s_newdatasize;
    profile.pc_currentStart = profile.pc_counterStart;

}



/*
 *  ProfExit(Filename) --
 *
 *  Inserts code to dump out profiling info to a file
 */
void ProfExit(char *Filename) {
    InsertDumpCode(Filename);
}
