/*
 *  Provides primitives for constructing a flowgraph from instruction
 *  streams.
 *
 *  Notes
 *
 *  Edges: An edge is represented as a pointer to a source and a
 *  destination block.  Since both the source and destination blocks
 *  should have knowledge about their predecessors and successors, both
 *  the source and destination need to have pointers to the edge.
 *  However, if a block B1 has a list of edges, it is inconvenient to
 *  force block B2 to use and maintain the same same list.  Edge nodes are
 *  introduced as a level of indirection between the actual edge and the
 *  block.  Each block maintains a list of predecessor and successor nodes
 *  ((()->b_pred) and (()->b_succ) respectively).  These are merely doubly-
 *  linked lists that contain a pointer to the edge. Thus, a Block B1 having an
 *  edge to a Block B2 will have a pointer somewhere in its successor list
 *  to an edge.  B2 will have a pointer somewhere in its predecessor list
 *  to the same edge.
 *
 *  CheckConsistency: A function that should be called at regular
 *  intervals when testing.  It verifies that the data structures agree
 *  with each other
 */
#include "global.h"
#include "flowgraph.h"
#include "utilities.h"
#include "disassemble.h"
#include "hash.h"
#include "linkage.h"
#include "opt.layout.h"
#include "sysdep.h"
#include "relocate.h"
#include "upm.h"
#include "x86-classes.h"
#include "dominator.h"
#include "clone.h"
#include "stack.analyze.h"
#include "util.h"



static JumpTable *all_jts;		// all jumptables

#define FOREACH_JT(itr) \
    for (itr = all_jts; itr != NULL; itr = ((itr)->jt_next))



static void UpdateCallerForEscapingTarget2(Block *bbl, ADDRESS call_slot);
static void AddUPMHellEdges();
static void AddJumpTableEdges();
static void AddEntryEdge(Block *bbl);
static void DecipherAllJumpTables();
static bool AddJumpTableEdgesForNonJump(Instruction *instr, JumpTable *jt);
static void AddJumpTableEdgesForJump(Instruction *instr, JumpTable *jt);
static void ResolveTargets(Instruction *list);
static bool RemoveConditionalEscapingEdgeForFun(Function *fun);

/* Creates the list of basic blocks based upon the marked leaders */
static void BBLCreateList(Instruction *list);

/* Creates the list of functions based upon the marked leaders */
static void FunCreateList(Block *list);

/* Attaches edges from return statements to abstract exit blocks */
static void FunResolveExitBlocks();

/* Marks all functions using the symbol table */
static void MarkFunctions();

/* Marks all leaders for basic blocks */
static void MarkBasicBlocks(Instruction *list);

/* Creates HELL */
static void SetupHELL();

/* Merges the two basic blocks that are connected by an ET_NORMAL edge */
static void BBLMerge(Block *first, Block *second);

static void RemoveConditionalEscapingEdges();
static void ResolveEntrylessEntries();

static void RemoveHellsForFixup();

static void NormalizeFunLinkEdges(Instruction *iterator) {
    Block *newBBL = 0;
    Instruction *nop = 0;
    
    newBBL = BBLCreate();
    nop = InstrCreateNOP();

    nop->i_Bblock = newBBL;
    newBBL->b_first = newBBL->b_last = nop;
    INSTR_SET_FLAG(nop, BBL_START);
    INSTR_SET_FLAG(nop, BBL_END);    
    newBBL->b_type = BT_NORMAL;
    newBBL->b_fun = INSTR_FUNCTION(iterator);

    /*
     * insert the block between the current block and its
     * next block
     */
    newBBL->b_prev = iterator->i_Bblock;
    newBBL->b_next = iterator->i_Bblock->b_next;

    iterator->i_Bblock->b_next = newBBL;

    if (newBBL->b_next) {
        newBBL->b_next->b_prev = newBBL;
    }

    
    /*
     * insert the nop between the current instruction and its
     * next instruction
     */
    nop->i_next = iterator->i_next;
    nop->i_prev = iterator;
    iterator->i_next = nop;
    if (nop->i_next) {
        nop->i_next->i_prev = nop;
    }

}


/*
 *  BuildControlFlowGraph(list) --
 *
 *  BuildControlFlowGraph should be called once after the instruction
 *  stream is disassembled and linked.  Basic blocks, Functions and HELL
 *  will be constructed, and edges will be added to the control flow
 *  graph.  The targets for any instruction that changes the value of the
 *  pc to something other than the next instruction are resolved by
 *  backtracing.  Jump Tables are resolved by adding edges from the basic
 *  block to all possible targets in the jump table.  Functions are
 *  constructed as linked lists of blocks, and finally the graph is
 *  checked for data structure consistency.
 *
 *  list: the first instruction of the list of instructions that the
 *  control flow graph will be constructed from
 */
void BuildControlFlowGraph(Instruction *list) {

    SetupHELL();                // Initialize HELL, which is Pluto's domain
    bbl_list = HELL;
    f_list = HELL->b_fun;

    MarkBasicBlocks(list);	    // Mark instructions as BBL_START or BBL_END
    MarkFunctions();            // Mark instructions as FUN_START or FUN_END

    BBLCreateList(i_list);	    // Create the initial bbl set
    FunCreateList(bbl_list);	// Create a set of functions

    FunResolveExitBlocks();     // Create ET_EXIT edges for functions
    ResolveTargets(i_list);     // Expand the basic block set based on jumps
    ResolveEntrylessEntries();	// Ensure an entry block for every function

    AddUPMHellEdges();          // Add HELL edges based on UPM information

    DecipherAllJumpTables();    // Look for jump tables
    AddJumpTableEdges();        // Add edges for them

    CheckConsistency();         // Verify that data structures agree
//    MarkFlistCalls();         // Add HELL edges to FListed functions

    BBLOutputStats();           // Display statistics about CFG

    if (WANT_DEBUG(DB_FLOW)) {
        BBLOutputGraph();
    }

    if (KERNEL_MODE) {
        RemoveHellsForFixup();
    }
}



/*
 *  RemoveEscapingEdges() --
 *
 *  Removes all escaping edges that can be removed. This is done in two
 *  major steps: (1) Remove all unconditional escaping edges of the form
 *  leave; jmp X, (2) Remove all conditional escaping edges by inlining
 *  the escaping target.
 */
void RemoveEscapingEdges() {
    REPORT(RP_MINOR, "Removing escaping edges");

    RemoveConditionalEscapingEdges();
    // RemoveUnconditionalEscapingEdges();
}



/*
 *  BBLCreateList(list) --
 *
 *  Takes the marked instructions in the instruction stream and
 *  creates basic blocks and edges between the blocks
 */
void BBLCreateList(Instruction *list) {
    Instruction *iterator;
    Block *bbl;
    Block *cur;

    cur = bbl = bbl_list = NULL;
    FOREACH_INSTR(iterator) {
        if (INSTR_HAS_FLAG(iterator, BBL_START)) {

            /* Create the new block and link it into the end of the list */
            bbl = BBLCreate();
            bbl->b_type = BT_INVALID;
            bbl->b_first = iterator;
            bbl->b_last = iterator;
            if (cur) {
                bbl->b_prev = cur;
                cur->b_next = bbl;
            } else {
                bbl->b_prev = NULL;
            }
            bbl->b_next = NULL;
            cur = bbl;

        } else if (INSTR_HAS_FLAG(iterator, BBL_END)) {
            cur->b_last = iterator;
        }

        iterator->i_Bblock = cur;
    }

    bbl_list = i_list->i_Bblock;

    FOREACH_BBL(bbl) {
        bbl->b_auxinfo = BBLCreateAuxinfo();
    }

    return;
}

bool hadUPMAlready(Operand *op) {
    LinkedList *upms;
    upms = UPMTableGetByDest(op);
    if (ll_size(upms) > 0) {
    	ll_free(upms);
    	return true;
    }
    
    ll_free(upms);
    return false;
}



/*
 *  ResolveTargets(list) --
 *
 *  Adds edges to the CFG based on targets of jmps and calls.
 *
 *  list: the first instruction of the list of instructions that
 *  will be used to resolve any targets
 */
static void ResolveTargets(Instruction *list) {
    Instruction *iterator;        // current instruction being resolved
    Instruction *target;        // target instruction
    ADDRESS addr;

    iterator = list;
    while (iterator != NULL) {
        target = NULL;
        if (isDataInstruction(iterator)) {
            //Just to let us know that function has 
            //a data instruction 
            FUNCTION_SET_FLAG(iterator->i_Bblock->b_fun,
                FUNCTION_DATA_INSTRUCTION_FLAG);

            if (iterator != iterator->i_Bblock->b_first) {
                BBLDivideBlock(iterator->i_Bblock, iterator);
            }
            //represents data in the text section
            iterator->i_Bblock->b_type = BT_DATABLOCK;

            REPORT(RP_TRIVIAL, "Data Instruction %x %s\n",
                iterator->i_address,
                iterator->i_Bblock->b_fun->f_name);
            
            if (iterator->i_Bblock == iterator->i_Bblock->b_fun->f_first) {
                AddEntryEdge(iterator->i_Bblock);
                REPORT (RP_TRIVIAL, "Data function \n");
            }

            if (iterator->i_next) {
                BBLAddEdge(iterator->i_Bblock, HELL, ET_HELL);
                BBLAddEdge(HELL, iterator->i_next->i_Bblock, ET_HELLMAYBE);

                BBLAddEdge(iterator->i_Bblock,
                    iterator->i_next->i_Bblock,
                    ET_DATALINK);

                REPORT(RP_TRIVIAL, "Added Hell edge for data function %s\n",
                    iterator->i_Bblock->b_fun->f_name);
            } else {
                WARNING("[%s:%d]Data instruction has no successor",
                    __FILE__, __LINE__);
            }

            iterator = iterator->i_next;
            continue;
        }
        
        /*
         * If the instruction doesn't change the program counter then an
         * ET_NORMAL edge will be added between it and the next block
         * if it is a basic block boundary.
         */
        if (!isPCChangingClass(iterator)) {
            if (INSTR_HAS_FLAG(iterator, BBL_END) &&
                (!BBLEdgeExists(iterator->i_Bblock,
                    INSTR_BBL_NEXT(iterator)))) {
                if (iterator->i_Bblock->b_next == NULL) {
                    // this is the last BBL in the program
                    BBLAddEdge(iterator->i_Bblock, HELL, ET_HELL);
                } else {
                    // normal case
                    BBLAddEdge(iterator->i_Bblock,
                        iterator->i_Bblock->b_next, ET_NORMAL);
                    /* If the instruction is falling through to
                     * a different function then the we add a
                     * ET_COMPENSATE Edge
                     */
                    if (INSTR_FUNCTION(iterator) !=
                        INSTR_FUNCTION(iterator->i_next)) {
                        BBLAddEdge(INSTR_FUNCTION(iterator->i_next)->f_exit,
                            INSTR_FUNCTION(iterator)->f_exit, ET_COMPENSATE);
                    }                    
                }
            }
            iterator->i_Bblock->b_type = BT_NORMAL;
            iterator = iterator->i_next;
            continue;
        }

        /* Return edges are added in by the caller */
        if (isReturnClass(iterator)) {
            iterator->i_Bblock->b_type = BT_RETURN;
            iterator = iterator->i_next;
            continue;
        }

        if (isSyscallClass(iterator)) {
            Block *new_bbl;
            Instruction *next;

            next = iterator->i_next;

            if ((iterator->i_Bblock->b_last == iterator) &&
                (iterator->i_Bblock->b_first == iterator)) {
                new_bbl = iterator->i_Bblock;
            } else {
                //Split it after the int x80 instruction. Do not
                //split it before
                if (iterator->i_next != iterator->i_next->i_Bblock->b_first) {
                    BBLDivideBlock(iterator->i_Bblock, iterator->i_next);
                }
                new_bbl = iterator->i_Bblock;
            }

            new_bbl->b_type = BT_SYSCALL;
            BBLRemoveEdges(new_bbl);
            BBLAddEdge(new_bbl, new_bbl->b_next, ET_SYSCALL_LINK);

            iterator = next;
            continue;
        }

        if (isHaltClass(iterator)) {
            iterator->i_Bblock->b_type = BT_HALT;
            iterator = iterator->i_next;
            continue;
        }

        /*
         * For conditional edges, we add an ET_FALSE edge (for the fall-through
         * case) to the next bbl in the list. An ET_TRUE edge (for the true
         * case) is added to the target of the displacement given in the
         * instruction.
         */

        if (isConditionalInstr(iterator)) {
            if (!INSTR_HAS_FLAG(iterator, BBL_END)) {
                WARNING("[%s]Jump instruction not marked as BBL_END",
                    iterator, __FUNCTION__);
            }
            BBLAddEdge(iterator->i_Bblock, iterator->i_Bblock->b_next,
                ET_FALSE);
            iterator->i_Bblock->b_type = BT_BRANCH;
        }


        /*
         * At this point we have handled return, syscall, halt and
         * non-PC-changing instructions, and added fall-through edges.
         * Now we need to resolve targets of calls and jumps by
         * looking at operand values.
         */


        /*
         * Operands that don't use a register should be easy to resolve.
         * We will look at the displacement and calculate the target
         * of the control transfer.
         */

        if (!OperandUsesRegister(iterator->i_src)) {

            if (INSTR_HAS_FLAG(iterator, INDIRECT)) {
                BBLAddEdge(iterator->i_Bblock, HELL, ET_HELL);
                addr = 0x0;

            } else if (OP_TYPE((iterator->i_src), OPERAND_DISP_A_FLAG)) {

                // If an absolute address is given in the operand
                // then we can simply make the translation
                addr = iterator->i_src->a_displacement;
                target = AddressToInstr(addr);
                WARNING("Absolute address %x encountered at %x?", addr, 
                        iterator->i_address);
                if (target == NULL) {
                    ReportInstrError ("AddressToInstr returned null target", 
                            iterator);
                }

            } else if (OP_TYPE(iterator->i_src, OPERAND_DISP_R_FLAG)) {

                if (iterator->i_next) {
                    addr   = iterator->i_src->a_displacement +
                                iterator->i_address + iterator->i_length;
                    target = AddressToInstr(addr);
                    if (target == NULL) {
                        REPORT(RP_TRIVIAL,
                            "AddressToInstr returned null target for %x",
                            iterator->i_address);
                    }
                } else {
                    ReportInstrError("Jump instruction is last in list",
                        iterator);
                }
            }

            /* At this point we have gotten the target, unless of course
             * the target didn't exist.  We now will add the appropriate edge
             * or an edge to HELL if we didn't have any target.  A target that
             * lies in the middle of a basic block will cause the block to be
             * divided into two smaller blocks, since the definition of a basic
             * block requires that control flow can only enter from the first
             * instruction.
             */

            if (!target) {

                /*  First check if the target is a data instruction. If it
                 *  is then we need to add an appropriate patch which will
                 *  adjust target instruction to the correct offset into the 
                 *  data instruction.
                 */ 
                if (addr) { 
                    if ((target = BacktraceToInstr(addr))) {
                        if (isDataInstruction(target) && 
                            !hadUPMAlready(iterator->i_src)) {
                            if (OP_TYPE(iterator->i_src, OPERAND_DISP_R_FLAG)) {
                                AddPatchInstrOffsetToInstr(target,
                                    addr - target->i_address, 
                                    iterator, iterator->i_src);
                                OPERAND_SET_FLAG(iterator->i_src, 
                                    OPERAND_AUTOLENGTH_RELOCATED);

                            } else {
                                MALFUNCTION(
                                    "NEED PATCH FOR ABS JUMP INTO DataInstr");
                            }
                        }
                    }
                    target = NULL;
                }

                // If the target was not found, add an edge to HELL
                if (isCallClass(iterator)) {
                    if (!iterator->i_next) {
                        Instruction *nop = 0;
                        nop = InstrCreateNOP();
                        InsertInstrAfterInstr(nop, iterator);
                        BBLDivideBlock(iterator->i_Bblock, nop);
                        WARNING("Call Instruction did not have a successor");
                    }                    
                    /*
                     * the FUNLINK edge cannot be inter-procedural so
                     * normalizing the graph here! what is normalizing?
                     * Just add a nop block after the call and make
                     * the FunLink edge remain within the same. Now
                     * the ET_NORMAL could be interprocedural in which
                     * case you add a ET_COMPENSATE edge.
                     *
                     */                    
                    if (INSTR_FUNCTION(iterator) !=
                        INSTR_FUNCTION(iterator->i_next)) {
                        NormalizeFunLinkEdges(iterator);
                    }
                    
                    BBLAddEdge(iterator->i_Bblock,
                        iterator->i_next->i_Bblock, ET_FUNLINK);
                    BBLAddEdge(HELL, iterator->i_next->i_Bblock, ET_HELLMAYBE);
                    iterator->i_Bblock->b_type = BT_CALL;
                } else {
                    if (!isConditionalInstr(iterator)) {
                        iterator->i_Bblock->b_type = BT_UNCOND;
                    }
                }
                BBLAddEdge(iterator->i_Bblock, HELL, ET_HELL);
                
            } else if ((target) && !INSTR_HAS_FLAG(target, BBL_START)) {

                // If we're jumping into the middle of a Basic Block, then we
                // need to divide the block into two blocks and add appropriate
                // edges
                BBLDivideBlock(target->i_Bblock, target);
            }

            /* Finally, add in the edge */
            if (target) {
                /* For function calls, make sure we add the function link edge,
                 * return edges and the call edge */
                if (isCallClass(iterator)) {
                    CASSERT(iterator->i_next,
                        "Function call made from block without next block.");

                    /*
                     * We add all sorts of edges here.  Since we want functions
                     * to have a single entry block, all calls into the function
                     * regardless of where they enter will be routed through
                     * fun->f_entry.  Thus, we add the following edges:
                     *   (1) Calling block to entry block
                     *   (2) Entry block to target block inside function
                     *   (3) Exit block to block after calling block
                     *   (4) A link between the caller block and block after it
                     *   (5) *NEW* A FUNDIRECT edge between the calling block
                     *                and the target block inside the function
                     */

                    /* Normalize the CFG
                     */
            
                    if (INSTR_FUNCTION(iterator) !=
                        INSTR_FUNCTION(iterator->i_next)) {
                        NormalizeFunLinkEdges(iterator);
                    }
                    // this excludes function calls that are actually just PIC
                    // setups since it technically isn't an entry edge unless
                    // it originates outside of the function
                    // this excludes function calls that are actually just PIC
                    // setups since it technically isn't an entry edge unless
                    // it originates outside of the function
                    if (INSTR_FUNCTION(target) != INSTR_FUNCTION(iterator)) {
                        AddEntryEdge(target->i_Bblock);
                    }

                    BBLAddEdge(iterator->i_Bblock,
                        INSTR_FUNCTION(target)->f_entry, ET_FUNCALL);
                    BBLAddEdge(INSTR_FUNCTION(target)->f_exit,
                        INSTR_BBL_NEXT(iterator), ET_RETURN);

                    BBLAddEdge(iterator->i_Bblock, target->i_Bblock,
                        ET_FUNDIRECT);
                    BBLAddEdge(iterator->i_Bblock,
                        iterator->i_next->i_Bblock, ET_FUNLINK);

                    iterator->i_Bblock->b_type = BT_CALL;
                    // why is this necessary? cml
                    //BBLAddEdge(INSTR_FUNCTION(iterator->i_next)->f_exit,
                    //    HELL, ET_HELL);
                } else if (isConditionalInstr(iterator)) {
                    if (BBLEdgeExists(iterator->i_Bblock, target->i_Bblock) &&
                        !BBLGetEdgeTyped(iterator->i_Bblock, ET_FALSE)) {
                        WARNING(
			    "Conditional Jmp with a different kind of edge\n");
                    } else {
                        BBLAddEdge(iterator->i_Bblock,target->i_Bblock,ET_TRUE);
                    }
                } else if (isUnconditionalInstr(iterator)) {
                    BBLAddEdge(iterator->i_Bblock, target->i_Bblock, ET_UNCOND);
                    iterator->i_Bblock->b_type = BT_UNCOND;
                } else {
                    ReportInstrError
                        ("Jump neither conditional or unconditional",
                        iterator);
                }
                
                if (INSTR_FUNCTION(iterator) != INSTR_FUNCTION(target)) {
                    if (!isCallClass(iterator)) {
                        BBLAddEdge(INSTR_FUNCTION(target)->f_exit,
                            INSTR_FUNCTION(iterator)->f_exit, ET_COMPENSATE);
                        if (BBLGetEdgeToBBL(INSTR_FUNCTION(target)->f_entry,
                                target->i_Bblock) == NULL) {
                            BBLAddEdge(INSTR_FUNCTION(target)->f_entry,
                                target->i_Bblock, ET_ENTRY);
                        }
                    }
                }
            }

        } else {

            /*
             * Otherwise the instruction has an operand containing a register.
             * For now, we will only consider jumps through jump tables.
             * Other jump/call/ret/branch instructions may exist which can use
             * registers in the operand--they should be taken care of as well.
             * The trick is finding out what the value of the register is
             * at the time of the jump.  Many times the value cannot be
             * determined because of multiple predecessors.  For instructions
             * though jump tables however, all predecessors should be loading
             * the same values.
             */
            if (isJumpClass(iterator)) {
                iterator->i_Bblock->b_type = BT_UNCOND;
                INSTR_SET_FLAG(iterator, INSTR_MAY_HAVE_JT);
                BBLAddEdge(iterator->i_Bblock, HELL, ET_HELL);
                BBLAddEdge(HELL, iterator->i_next->i_Bblock, ET_HELLMAYBE);
            } else if (isCallClass(iterator)) {
                CASSERT(iterator->i_src != NULL,
                    "Call instruction with no operands.");
                BBLAddEdge(iterator->i_Bblock, HELL, ET_HELL);
                BBLAddEdge(iterator->i_Bblock, iterator->i_next->i_Bblock,
                    ET_FUNLINK);
                BBLAddEdge(HELL, iterator->i_next->i_Bblock, ET_HELLMAYBE);
                iterator->i_Bblock->b_type = BT_CALL;
            } else {
                ReportInstrError(
                    "Unconditional non-JT jump with register (unresolved)",
                    iterator);
                ReportInstrError("Added edge to HELL", iterator);
                BBLAddEdge(iterator->i_Bblock, HELL, ET_HELL);
                iterator->i_Bblock->b_type = BT_UNCOND;
            }
        }
        iterator = iterator->i_next;

    }                                // iteration over instructions [while loop]
}



/*
 *  BBLAddEdge(source,destination,type) --
 *
 *  Given two basic blocks, BBLAddEdge simply adds an edge from the
 *  first to the second
 */
Edge *BBLAddEdge(Block *source, Block *destination, UINT8 type) {
    EdgeNode *e1, *e2, *list;
    Edge *newedge;

    e1 = EdgeNodeCreate();
    e2 = EdgeNodeCreate();
    newedge = EdgeCreate();

    CASSERT(source != NULL, "Given NULL source.");
    CASSERT(destination != NULL, "Given NULL destination.");

    /* Set up the pointers to the edge */
    e1->en_edge = e2->en_edge = newedge;

    /* Set up the edge pointers and types */
    newedge->e_type = type;
    newedge->e_to = destination;
    newedge->e_from = source;

    /* Link in e1 to the list of successors for the source block */
    list = source->b_succ;
    if (list == NULL) {
	source->b_succ = e1;
    } else {
	e1->en_next = list;
	list->en_prev = e1;
	source->b_succ = e1;
    }

    /* Link in e2 to the list of predecessors for the destination block */
    list = destination->b_pred;
    if (list == NULL) {
	destination->b_pred = e2;
    } else {
	e2->en_next = list;
	list->en_prev = e2;
	destination->b_pred = e2;
    }

    return newedge;
}


/*
 * BBLMoveSuccEdges(src, dest) -- 
 *
 * Move all the succ edges of src to point to dest
 */
void BBLMoveSuccEdges(Block *src, Block *dest) {
    EdgeNode *iterator;

    dest->b_succ = src->b_succ;

    FOREACH_SUCC_IN_BBL(src, iterator) {
        iterator->en_edge->e_from = dest;
    }
    src->b_succ = NULL;

    return;  
}

/*
 *  BBLDivideBlock(block,instr) --
 *
 *  When the target of a jump instruction is resolved, and it turns out
 *  that the jump is into the middle of a basic block, this function is
 *  called to divide the block up. The target instruction becomes the
 *  first instruction in a new basic block that contains all
 *  instructions appearing after the target in the original basic block.
 *  The original basic block loses the instructions the new one gained.
 *  Since there is no pc-manipulating instruction at the end of the
 *  original block (after the division), we can safely add an edge
 *  connecting the original block to the new block.
 */

void BBLDivideBlock(Block *block, Instruction *instr) {
    Block *newblock;
    Block *original;
    Instruction *iterator;

    if (INSTR_HAS_FLAG(instr, BBL_START)) {
	WARNING("Attempt to BBLDivideBlock"
	    "with an instruction that started a block");
	return;
    }

    CASSERT(instr->i_prev, "No previous instruction.");

    CASSERT(instr->i_prev->i_Bblock == instr->i_Bblock,
	"Corrupted instruction stream. Previous instruction"
	" with different basic block than current instruction.");

    original = instr->i_Bblock;
    newblock = BBLCreate();

    /* Set up the new block */
    newblock->b_type = original->b_type;
    newblock->b_next = original->b_next;
    newblock->b_prev = original;
    if (newblock->b_next) {
	newblock->b_next->b_prev = newblock;
    }
    newblock->b_last = original->b_last;
    newblock->b_first = instr;
    
    /*
     * Instead of making a new copy of all the succ edges for newblock
     * just changing the "from" node of all the succ edges to point to newblock.
     * This wll avoid destroying all the succ edges of original block. 
     */
    BBLMoveSuccEdges(original, newblock);

    newblock->b_fun = original->b_fun;
    newblock->b_flags = original->b_flags;

    /* Flag the instructions appropriately */
    INSTR_SET_FLAG(instr, BBL_START);
    if (instr->i_prev) {
	INSTR_SET_FLAG(instr->i_prev, BBL_END);
    }

    /* Tell the instructions about the change */
    iterator = instr;
    while (true) {
	iterator->i_Bblock = newblock;
	if (INSTR_HAS_FLAG(iterator, BBL_END)) {
	    break;
	}
	iterator = iterator->i_next;
	if (INSTR_HAS_FLAG(iterator, BBL_END)) {
	    iterator->i_Bblock = newblock;
	    break;
	}
    }

    /* Change settings on the original block to reflect the division */
    original->b_type = BT_NORMAL;
    original->b_next = newblock;
    original->b_last = instr->i_prev;

    // Add the edge connecting the original block with the newly created block
    BBLAddEdge(original,newblock,ET_NORMAL)->e_edgeWeight = original->b_weight;

    if (original->b_fun->f_last == original) {
	original->b_fun->f_last = newblock;
    }

    newblock->b_weight = original->b_weight;

/* TAL - ADDED */
    newblock->b_flist = original->b_flist;
    original->b_flist = NULL;

    return;
}



/*
 *  DecipherAllJumpTables() --
 *
 *  This function seeks through each read-only section and identifies
 *  potential jump tables, adding them to a linked list of jump tables as
 *  it goes, then returns that list.  Potential jump tables are identified
 *  by looking at memory locations that are both a source and destination
 *  of some UPMs.
 *  returns: a linked list of JumpTables that were identified as
 *  potential jump tables
 */
void DecipherAllJumpTables() {
    JumpTable *cur_jt;
    Section *seciterator;
    ADDRESS base_offset, offset;
    LinkedList *upms;
    Iterator *itr;
    Location *member;
    Patch *uiterator;

    all_jts = NULL;

    FOREACH_SECTION(seciterator) {
	if (!SECTION_HAS_FLAG(seciterator, SECTION_WRITABLE_FLAG)) {
	    base_offset = 0;

	    // go through the section looking for bases to jump tables
	    while (base_offset < seciterator->s_datasize) {
		if (AddressIsUPMSource(seciterator, base_offset) &&
		    AddressIsUPMDest(seciterator, base_offset)) {
		    // there wasn't an associated size so we have to
		    // figure it out by looking for the next symbol
		    offset = base_offset + sizeof(ADDRESS);
		    while (AddressIsUPMDest(seciterator, offset) &&
			!AddressIsUPMSource(seciterator, offset)) {
			offset += sizeof(ADDRESS);
		    }

		    upms = UPMTableGetBySourceAddress(seciterator, base_offset);

		    // if we can't find anyone that's using this one, we'll just
		    // forget about it
		    if (ll_size(upms) == 0) {
			WARNING("Could not find an instruction"
			    " that uses jump table @ %x(%s)",
			    base_offset, seciterator->s_name);
			ll_free(upms);
			continue;
		    }

		    cur_jt = AddJumpTable(seciterator, base_offset,
			(offset - base_offset) / 4);

		    itr = ll_iterator(upms);
		    while (itr->it_hasNext(itr)) {
			uiterator = (Patch *) itr->it_next(itr);

			BEGIN_FOREACH_UPM_DEST_MEMBER(uiterator, member) {
			    if (UPM_MEMBER_INSTR(member) != NULL) {
				ll_add(cur_jt->jt_users,
				    *UPM_MEMBER_INSTR(member));
			    }
			}
			END_FOREACH_UPM_MEMBER;
		    }
		    free(itr);
		    ll_free(upms);
		}

		base_offset += 1;
	    }
	}
    }
}



/*
 *  AddJumpTableEdges() --
 *
 *  Adds all jump table edges (ET_SWITCH) that can be resolved.
 */
static void AddJumpTableEdges() {
    JumpTable *jiterator, *next;
    Iterator *itr;
    Instruction *user;
    bool result = false;

    REPORT(RP_MINOR, WITHFUNC("Adding jump table edges"));

    jiterator = all_jts;
    while (jiterator != NULL) {
	next = jiterator->jt_next;

	itr = ll_iterator(jiterator->jt_users);

	// iterate through each of the users of this jump table
	while (itr->it_hasNext(itr)) {
	    user = (Instruction *) itr->it_next(itr);

	    if (isJumpClass(user)) {
		AddJumpTableEdgesForJump(user, jiterator);
		result = true;
	    } else if (isCallClass(user)) {
		WARNING("Call through a jump table unimplemented: %s",
		    PrintInstruction(user));
	    } else {
		result |= AddJumpTableEdgesForNonJump(user, jiterator);
	    }

	    if (!result) {
		if (jiterator->jt_prev != NULL) {
		    jiterator->jt_prev->jt_next = jiterator->jt_next;
		}

		if (jiterator->jt_next != NULL) {
		    jiterator->jt_next->jt_prev = jiterator->jt_prev;
		}

		if (jiterator == all_jts) {
		    all_jts = jiterator->jt_next;
		}

		ll_free(jiterator->jt_users);
		free(jiterator);

		break;
	    }
	}
	free(itr);

	jiterator = next;
    }
}



/*
 *  AddJumpTableEdgesForNonJump(instr,jt) -- #%#% NOT IMPLEMENTED.
 *
 *  Adds the jump table edges for basic blocks that use the base address
 *  of the specified jump table which is loaded by the instruction instr.
 *  This function analyzes the usage of the loaded base address and
 *  determines if the address is ultimately used in a control transfer and
 *  if so it adds the jump table edges from those basic blocks based on
 *  the info supplied in the specified jump table.
 *
 *  instr: the non-jump instruction that should be used to resolve
 *  where the jump table edges should be placed
 *  jt: the jump table through which control ultimately transfers
 *  returns: true if the control transfer is determined and the edges are added
 */
static bool AddJumpTableEdgesForNonJump(Instruction *instr, JumpTable *jt) {
    /*
     *  Need to implement this to get the jump tables that are tricky!
     */
    REPORT(RP_SMLOOP, WITHFUNC("NOT IMPLEMENTED"));
    return false;
}



/*
 *  AddJumpTableEdgesForJump(instr,jt) --
 *
 *  Adds the jump table edges for the basic block containing the specified
 *  instruction that is an indirect jump instruction using the information
 *  supplied in the specified JumpTable struct.
 *
 *  instr: the jump instruction whose jump table edges should be added
 *  jt: the JumpTable through which the jump jumps
 */
static void AddJumpTableEdgesForJump(Instruction *instr, JumpTable *jt) {
    Instruction *target;
    int slot;

    if (!isJumpClass(instr) || !INSTR_HAS_FLAG(instr, INDIRECT)) {
	WARNING("Passed a non-jump or non-indirect jump: %s",
	    PrintInstruction(instr));
	return;
    }

    instr->i_Bblock->b_type = BT_SWITCH;

    BBLRemoveEdgeTyped(instr->i_Bblock, HELL, ET_HELL);
    BBLRemoveEdge(HELL, instr->i_next->i_Bblock);        

    for (slot = 0; slot < jt->jt_size; slot++) {
	target = AddressToInstr(loadLE(&jt->jt_section->s_data
		[(jt->jt_base_offset) + sizeof(ADDRESS) * slot]));

	if (target == NULL) {
	    WARNING("Impossible NULL target for the jump table");
	    continue;
	}
	// this should always be true since they should get added in
	// AddUPMHellEdges() unless we have a case where many sites jump
	// through the jump table, in this case only one hell edge will
	// be added to the target, however many call sites will request
	// removal of the target
	if (BBLHasHellPredecessor(target->i_Bblock)) {
	    BBLRemoveEdge(HELL, target->i_Bblock);

	    if (BBLHasEntryPredecessor(target->i_Bblock) &&
		BBLHasHellPredecessor(INSTR_FUNCTION(target)->f_entry)) {
		BBLRemoveEdge(HELL, INSTR_FUNCTION(target)->f_entry);
	    }
	}

	if (!INSTR_HAS_FLAG(target, BBL_START)) {
	    // since the source was not already marked as the beginning of a
	    // basic block we need to go ahead and divide the block and make it
	    // so
	    BBLDivideBlock(target->i_Bblock, target);
	}
	// Finally, add in the edge
	(BBLAddEdge(instr->i_Bblock, target->i_Bblock,
		ET_SWITCH)->e_jumpTableSlot) = slot;
    }
}



/*
 *  BBLCopyEdges(src,dest) -- copies the edges from one list to another.
 */
void BBLCopyEdges(Block *src, Block *dest) {
    EdgeNode *iterator;

    iterator = src->b_succ;

    FOREACH_SUCC_IN_BBL(src, iterator) {
	BBLAddEdge(dest, NODE_TO(iterator), NODE_TYPE(iterator))->e_edgeWeight =
	    NODE_WEIGHT(iterator);
    }

    return;
}



/*
 *  BBLRemoveEdges(block) -- removes all successors from 'block'.
 */
void BBLRemoveEdges(Block *block) {
    EdgeNode *iterator, *next;

    CASSERT(block != NULL, "Passed a NULL block.");

    iterator = block->b_succ;
    while (iterator != NULL) {
	next = iterator->en_next;
	BBLRemoveEdgeTyped(block, NODE_TO(iterator), NODE_TYPE(iterator));
	iterator = next;
    }

    return;
}



/*
 *  BBLRemovePredecessors(block) -- removes all predecessors from 'block'.
 */
void BBLRemovePredecessors(Block *block) {
    EdgeNode *iterator, *next;

    CASSERT(block != NULL, "Passed a NULL block.");

    iterator = block->b_pred;
    while (iterator != NULL) {
	next = iterator->en_next;
	BBLRemoveEdge(NODE_FROM(iterator), block);
	iterator = next;
    }

    return;
}



/*
 *  BBLRemoveEdge(src,dest) --
 *
 *  Given a basic block and an edge, BBLRemoveEdge will remove the
 *  edge from the list of successors on the given block, and from the
 *  list of predecessors on the target block that the edge connects
 *  to.
 */
void BBLRemoveEdge(Block *src, Block *dest) {
    EdgeNode *nodeDest, *nodeSrc, *iterator;
    Edge *edge = NULL;
    bool found = false;

    CASSERT(src != NULL, "Passed a NULL block.");
    CASSERT(dest != NULL, "Passed a NULL edge.");

    nodeDest = nodeSrc = NULL;
    
    FOREACH_SUCC_IN_BBL(src, iterator) {
	if (NODE_TO(iterator) == dest) {
	    if (edge != NULL) {
		switch (NODE_TYPE(iterator)) {
		case ET_ENTRY:
		case ET_HELL:
		case ET_HELLMAYBE:
		case ET_SWITCH:
		case ET_CHAIN:
		    continue;
		    break;
		default:
		    switch (edge->e_type) {
		    case ET_ENTRY:
		    case ET_HELL:
		    case ET_HELLMAYBE:
		    case ET_SWITCH:
		    case ET_CHAIN:
			continue;
			break;
		    default:
			if (edge->e_type != NODE_TYPE(iterator)) {
			    REPORT(RP_SMLOOP,
				"Parallel edges encountered (%d -> %d)!",
				src->b_id, dest->b_id);
			}
			break;
		    }
		}
	    } else {
		nodeSrc = iterator;
		edge = iterator->en_edge;
	    }
	}
    }

    if (!edge) {
        WARNING("No edge present between %s(%d)--->%s(%d)\n",
            blockTypeToString(src->b_type),
            src->b_id,
            blockTypeToString(dest->b_type),
            dest->b_id);
        return;
    }
    //    CASSERT(edge != NULL, "Could not find edge between blocks!");

    FOREACH_PRED_IN_BBL(dest, iterator) {
	if (NODE_FROM(iterator) == src) {
	    nodeDest = iterator;
	    if (edge != iterator->en_edge) {
		continue;
	    }
	    edge = iterator->en_edge;
	    found = true;
	    break;
	}
    }

    CASSERT(found != false, "Could not find edge in destination.");

    CASSERT(nodeDest != NULL, "Could not find destination node.");
    CASSERT(nodeSrc != NULL, "Could not find source node.");

    /* First remove the successor node in the provided block */
    if (!nodeSrc->en_prev) {
	src->b_succ = nodeSrc->en_next;
	if (nodeSrc->en_next) {
	    nodeSrc->en_next->en_prev = NULL;
	}
    } else {
	nodeSrc->en_prev->en_next = nodeSrc->en_next;
	if (nodeSrc->en_next) {
	    nodeSrc->en_next->en_prev = nodeSrc->en_prev;
	}
    }

    /* Now remove the predecessor edge in the target block */
    if (!nodeDest->en_prev) {
	dest->b_pred = nodeDest->en_next;
	if (nodeDest->en_next) {
	    nodeDest->en_next->en_prev = NULL;
	}
    } else {
	nodeDest->en_prev->en_next = nodeDest->en_next;
	if (nodeDest->en_next) {
	    nodeDest->en_next->en_prev = nodeDest->en_prev;
	}
    }

    free(edge);
    free(nodeDest);
    free(nodeSrc);

    return;
}



/*
 *  BBLRemoveEdgeTyped(src,dest,edge_type) --
 *
 *  Given a basic block and an edge type, BBLRemoveEdge will remove
 *  the edge from the list of successors on the given block, and from
 *  the list of predecessors on the target block that the edge
 *  connects to.
 */
void BBLRemoveEdgeTyped(Block *src, Block *dest, int edge_type) {
    EdgeNode *nodeDest, *nodeSrc, *iterator;
    Edge *edge = NULL;

    CASSERT(src != NULL, "Passed a NULL block.");
    CASSERT(dest != NULL, "Passed a NULL edge.");

    nodeDest = nodeSrc = NULL;

    FOREACH_SUCC_IN_BBL(src, iterator) {
	if ((NODE_TO(iterator) == dest) && (NODE_TYPE(iterator) == edge_type)) {
	    nodeSrc = iterator;
	    edge = iterator->en_edge;
	    break;
	}
    }
    
    FOREACH_PRED_IN_BBL(dest, iterator) {
	if (NODE_FROM(iterator) == src && NODE_TYPE(iterator) == edge_type) {
	    nodeDest = iterator;
	    edge = iterator->en_edge;
	    break;
	}
    }

    CASSERT(nodeDest != NULL, "Could not find destination node.");
    CASSERT(nodeSrc != NULL, "Could not find source node.");

    /* First remove the successor node in the provided block */
    if (!nodeSrc->en_prev) {
	src->b_succ = nodeSrc->en_next;
	if (nodeSrc->en_next) {
	    nodeSrc->en_next->en_prev = NULL;
	}
    } else {
	nodeSrc->en_prev->en_next = nodeSrc->en_next;
	if (nodeSrc->en_next) {
	    nodeSrc->en_next->en_prev = nodeSrc->en_prev;
	}
    }

    /* Now remove the predecessor edge in the target block */
    if (!nodeDest->en_prev) {
	dest->b_pred = nodeDest->en_next;
	if (nodeDest->en_next) {
	    nodeDest->en_next->en_prev = NULL;
	}
    } else {
	nodeDest->en_prev->en_next = nodeDest->en_next;
	if (nodeDest->en_next) {
	    nodeDest->en_next->en_prev = nodeDest->en_prev;
	}
    }

    free(edge);
    free(nodeDest);
    free(nodeSrc);

    return;
}



/*
 *  BBLOutputGraph() --
 *
 *  Prints out a textual representation of the control flow graph
 *  by showing the bounding addresses of all basic blocks along with
 *  the edges it has.  The edges are shown as pointers to addresses
 *  that correspond to the first instruction in the target basic
 *  block.
 */
void BBLOutputGraph() {
    Block *iterator;
    EdgeNode *nodeEdge;
    Edge *edge;
    Function *fiterator;
    Symbol *siterator;

    printf("\n\n");
    DEBUG(DB_FLOW, "Control flow graph: functions, blocks, and edges");
    printf("\n");

    /* For each function, get all the basic blocks and print out their edges */
    FOREACH_FUN(fiterator) {

	if (fiterator == HELL->b_fun) {
	    continue;
	}

	/* Display the function bound and name */
	printf("Function[0x%8.8x,0x%8.8x] (",
	    FUNCTION_FIRST_INSTR(fiterator)->i_address,
	    FUNCTION_LAST_INSTR(fiterator)->i_address);
	FOREACH_SYMBOL(siterator) {
	    if (siterator->sy_value ==
		FUNCTION_FIRST_INSTR(fiterator)->i_address) {
		printf("%s", siterator->sy_name);
		break;
	    }
	}
	printf(")\n");

	FOREACH_BBL_IN_FUN(fiterator, iterator) {
	    printf("  BasicBlock {%8.8d}[0x%8.8x,0x%8.8x]\n", iterator->b_id,
		iterator->b_first->i_address, iterator->b_last->i_address);
	    printf("    {");
	    FOREACH_SUCC_IN_BBL(iterator, nodeEdge) {
		edge = nodeEdge->en_edge;
		if (edge->e_to->b_type == BT_HELL) {
		    printf(" HELL[%s] ", edgeTypeToString(edge->e_type));
		} else if (edge->e_type == ET_EXIT) {
		    printf(" EXIT[%s] ", edgeTypeToString(edge->e_type));
		} else if (edge->e_type == ET_FUNCALL) {
		    printf("%s[%s] ",
			lookupFunctionName(edge->e_to->b_fun),
			edgeTypeToString(edge->e_type));
		} else {
		    printf(" 0x%8.8x[%s] ",
			edge->e_to->b_first->i_address,
			edgeTypeToString(edge->e_type));
		}
	    }
	    printf("}\n");
	}
	printf("\n");
    }

    return;
}



/*
 *  MarkFunctions() --
 *
 *  Using information from the symbol table we will mark all functions.
 *  Note that functions are also marked as the beginning and ending of
 *  basic blocks as well.
 */
void MarkFunctions() {

    LinkedList    *fn_list;
    Iterator      *iterator;
    Symbol        *curr;
    ADDRESS       prev;
    Instruction   *inst;

    INSTR_SET_FLAG(i_list, FUN_START);
    INSTR_SET_FLAG(startins, FUN_START);

    fn_list   = GetSortedSymbolList();
    iterator  = ll_iterator(fn_list);
    prev      = ((Symbol *)iterator)->sy_value;
    
    while (iterator->it_hasNext(iterator)) {
        curr = (Symbol *) iterator->it_next(iterator);
        if (!!(curr->sy_flags & SYMBOL_NOTYPE_FLAG)) {
            curr->sy_size = prev - curr->sy_value;
        }
        prev = curr->sy_value;
        if (curr->sy_size == 0) {
            continue;
        }
        
        inst = AddressToInstr(curr->sy_value);
        if (inst == NULL) {
            REPORT(RP_DETAIL, "[MarkFunctions]No instruction for symbol %s",
                curr->sy_name);
            continue;
        }

        /* Mark the instruction */
        INSTR_SET_FLAG(inst, FUN_START);
        INSTR_SET_FLAG(inst, BBL_START);
        if (inst->i_prev) {
            INSTR_SET_FLAG(inst->i_prev, FUN_END);
            INSTR_SET_FLAG(inst->i_prev, BBL_END);
        }
    }

    return;
}



/*
 *  CheckConsistency() --
 *
 *  The functions verifies that all data structures used in the flowgraph
 *  are consistent with one another.  The various checks are described in
 *  detail below:
 *  (1) If a basic block B contains an instruction i, then i has its
 *  basic block field pointing to B
 *  (2) If a function F contains a basic block B, then B  has its
 *  function field pointing to F
 *  (3) For all lists: If X precedes Y then Y follows X
 *  (4) Type constraints: For example, there can only be a function
 *  call out of a block whose type corresponds to a function call
 *  (5) All basic blocks have a first and last instruction
 *  (6) All functions have a first and last basic block
 *  (7) If a Basic Block Z has successors that point to a Basic Block
 *  X then X has predecessors pointing to Z  (Edge consistency)
 *  (8) Edge/Block type consistency
 */
void CheckConsistency() {
#ifdef DEBUG_CONSIST
    Block *biterator, *biterator2;
    Instruction *iiterator, *iiterator2;
    Function *fiterator;
    EdgeNode *eiterator, *eiterator2;


    /* Make sure all basic blocks (exc HELL) have a first and last instruction*/
    FOREACH_BBL(biterator) {
	if (biterator == HELL) {
	    continue;
	}
	CASSERT(biterator->b_first != NULL,
	    "Basic Block without first instruction.");
	CASSERT(biterator->b_last != NULL,
	    "Basic Block without last instruction.");
    }

    /* Check each basic block for linked-list consistency */
    FOREACH_BBL(biterator) {
	if (biterator->b_next) {
	    CASSERT(biterator->b_next->b_prev == biterator,
		"Basic Blocks not consistent.");
	}
	if (biterator->b_prev) {
	    CASSERT(biterator->b_prev->b_next == biterator,
		"Basic Blocks not consistent.");
	}
    }

    /* Check each function for linked-list consistency */
    FOREACH_FUN(fiterator) {
	if (fiterator->f_next) {
	    CASSERT(fiterator->f_next->f_prev == fiterator,
		"Functions not consistent.");
	}
	if (fiterator->f_prev) {
	    CASSERT(fiterator->f_prev->f_next == fiterator,
		"Functions not consistent.");
	}
    }

    /* Check each instruction for linked list consistency */
    FOREACH_INSTR(iiterator) {
	if (iiterator->i_next) {
	    CASSERT(iiterator->i_next->i_prev == iiterator,
		"Instructions not consistent.");
	}
	if (iiterator->i_prev) {
	    CASSERT(iiterator->i_prev->i_next == iiterator,
		"Instructions not consistent.");
	}

	/* also check to make sure that any instruction with a start
	 * flag is the first one in it's block, and any instruction
	 * with an end flag is the last one in it's block. */
	if INSTR_HAS_FLAG
	    (iiterator, BBL_START) {
	    CASSERT(iiterator->i_Bblock->b_first == iiterator,
		"Start Flag not consistent.");
	    if (iiterator->i_Bblock->b_first != iiterator) {
		printf("instruction is %d\n", iiterator->i_id);
	    }
	    }
	if INSTR_HAS_FLAG
	    (iiterator, BBL_END) {
	    CASSERT(iiterator->i_Bblock->b_last == iiterator,
		"End Flag not consistent.");
	    }


    }

    /* Check to make sure that instructions know about the blocks they
     * are contained in */
    FOREACH_BBL(biterator) {
	FOREACH_INSTR_IN_BBL(biterator, iiterator) {
	    CASSERT(iiterator->i_Bblock == biterator,
		"Instructions not consistent with blocks.");
	}
    }

    /* Check to make sure that basic blocks know about the functions they
     * are contained in */
    FOREACH_FUN(fiterator) {
	FOREACH_BBL_IN_FUN(fiterator, biterator) {
	    CASSERT(biterator->b_fun == fiterator,
		"Blocks not consistent with functions.");
	}
    }

    /* Check to see that the blocks know what instructions they contain */
    FOREACH_INSTR(iiterator) {
	bool found = false;
	biterator = iiterator->i_Bblock;
	FOREACH_INSTR_IN_BBL(biterator, iiterator2) {
	    if (iiterator2 == iiterator) {
		found = true;
	    }
	}
	CASSERT(found == true, "Blocks not consistent with instructions.");
    }

    /* Check to see that the functions know what blocks they contain */
    FOREACH_BBL(biterator) {
	bool found = false;
	fiterator = biterator->b_fun;
	if (fiterator == NULL) {
	    continue;
	}
	FOREACH_BBL_IN_FUN(fiterator, biterator2) {
	    if (biterator2 == biterator) {
		found = true;
	    }
	}
	CASSERT(found == true, "Functions not consistent with basic blocks.");
    }

    /* Check edge consistency */
    FOREACH_BBL(biterator) {
	FOREACH_SUCC_IN_BBL(biterator, eiterator) {
	    bool found = false;
	    if (NODE_TYPE(eiterator) == ET_HELL) {
		continue;
	    }
	    biterator2 = NODE_TO(eiterator);
	    FOREACH_PRED_IN_BBL(biterator2, eiterator2) {
		if (NODE_FROM(eiterator2) == biterator) {
		    found = true;
		}
	    }
	    CASSERT(found == true, "Edges not consistent.");
	}
    }

    /* Check for "parallel" edges between blocks -- i.e., two or more
     * edges between a given pair of nodes. */
    FOREACH_BBL(biterator) {
	FOREACH_SUCC_IN_BBL(biterator, eiterator) {
	    FOREACH_SUCC_IN_BBL(biterator, eiterator2) {
		if (eiterator == eiterator2) {
		    continue;
		}
		if (NODE_FROM(eiterator) == NODE_FROM(eiterator2)
		    && NODE_TO(eiterator) == NODE_TO(eiterator2)
		    && NODE_TYPE(eiterator) < ET_SWITCH     // see edge.types.h
		    && NODE_TYPE(eiterator2) < ET_SWITCH) {
		    /*
		     * Position-independent code is implemented using a call
		     * to the next instruction.  In this case, there will be
		     * both a CALL edge and a LINK edge
		     * from the first block to the second.
		     */
		    if ((NODE_TYPE(eiterator) == ET_FUNDIRECT
			    && NODE_TYPE(eiterator2) == ET_FUNLINK)
			|| (NODE_TYPE(eiterator2) == ET_FUNDIRECT
			    && NODE_TYPE(eiterator) == ET_FUNLINK))
			continue;

		    printf("[CheckConsistency] Parallel edges encountered"
			"(%d -> %d)!... fun %d [%s]\n",
			NODE_FROM(eiterator)->b_id, NODE_TO(eiterator)->b_id,
			NODE_FROM(eiterator)->b_fun->f_id,
			NODE_FROM(eiterator)->b_fun->f_name);
		}
	    }
	}
    }

    /* Check to make sure no entry/exit blocks are in the list */
    FOREACH_BBL(biterator) {
	CASSERT(biterator->b_type != BT_EXIT, "Exit block linked in to list.");
	CASSERT(biterator->b_type != BT_ENTRY,"Entry block linked in to list.");
    }

    /* See if instructions have the proper BBL_END/BB_START flags set */
    FOREACH_BBL(biterator) {
	CASSERT(INSTR_HAS_FLAG(biterator->b_last, BBL_END),
	    "Last instruction in basic block without BBL_END flag.");
	CASSERT(INSTR_HAS_FLAG(biterator->b_first, BBL_START),
	    "First instruction in basic block without BBL_START flag.");
	if (!(INSTR_HAS_FLAG((biterator->b_first), BBL_START))) {
	    printf("block number %d, instruction %d\n", biterator->b_id,
		biterator->b_first->i_id);
	}
    }


    /* Check to see if block types are consistent with edge types */
    FOREACH_BBL(biterator) {
	FOREACH_SUCC_IN_BBL(biterator, eiterator) {
	    int type = NODE_TYPE(eiterator);
	    bool passed = true;
	    switch (type) {
	    case ET_HELLMAYBE:
		if (biterator->b_type != BT_HELL) {
		    passed = false;
		    printf("maybe hell? block %d? %d edge\n", biterator->b_id,
			type);
		}
		break;
	    case ET_FUNCALL:
	    case ET_FUNLINK:
	    case ET_FUNDIRECT:
		if (biterator->b_type != BT_CALL) {
		    passed = false;
		    printf("call block %d? %d edge\n", biterator->b_id, type);
		}
		break;
	    case ET_TRUE:
	    case ET_FALSE:
		if (biterator->b_type != BT_BRANCH) {
		    passed = false;
		    printf("branch block %d? %d edge\n", biterator->b_id, type);
		}
		break;
	    case ET_HELL:
		break;
	    case ET_UNCOND:
		if (biterator->b_type != BT_UNCOND) {
		    passed = false;
		    printf("uncond block %d? %d edge\n", biterator->b_id, type);
		}

		break;
	    case ET_RETURN:
		printf("return block %d? %d edge\n", biterator->b_id, type);
		passed = false;
		break;
	    case ET_ENTRY:
		printf("entry block %d? %d edge\n", biterator->b_id, type);
		passed = false;
		break;
	    case ET_EXIT:
		if (biterator->b_type != BT_RETURN) {
		    passed = false;
		    printf("exit block %d? %d edge\n", biterator->b_id, type);
		}
		break;
	    case ET_NORMAL:
		if (biterator->b_type != BT_NORMAL) {
		    passed = false;
		    printf("normal block %d? %d edge\n", biterator->b_id, type);
		}
		break;
	    case ET_SWITCH:
		if (biterator->b_type != BT_SWITCH) {
		    passed = false;
		    printf("switch block %d?\n", biterator->b_id);
		}
		break;
	    case ET_SYSCALL_LINK:
		if (biterator->b_type != BT_SYSCALL) {
		    passed = false;
		    printf("syscall block %d?\n", biterator->b_id);
		}
		break;
	    case ET_INVALID:
	    default:
		printf("got an invalid, or default case:"
		    " BBL = %d, FUN = %d [%s]\n", biterator->b_id,
		    biterator->b_fun->f_id, biterator->b_fun->f_name);
		passed = false;
		break;
	    }
	    CASSERT(passed == true, "Blocks not consistent with edges.");
	}
    }
#endif				// DEBUG_CONSIST

    return;
}



/*
 *  FunCreateList(list) --
 *
 *  Creates functions (lists of basic blocks) by looking at which
 *  instructions are flagged as FUN_START and FUN_END
 */
void FunCreateList(Block *list) {
    Block *iterator;
    Function *fun;
    Function *cur;

    cur = fun = NULL;
    f_list = NULL;

    FOREACH_BBL(iterator) {
        if (INSTR_HAS_FLAG(iterator->b_first, FUN_START)) {
            fun = FunCreate();
            iterator->b_fun = fun;
            fun->f_first = fun->f_last = iterator;
            fun->f_name = salloc(GetStrongestSymbolForValue(
                iterator->b_first->i_address)->sy_name);
            if (cur) {
                fun->f_prev = cur;
                cur->f_next = fun;
            } else {
                fun->f_prev = NULL;
            }
            fun->f_next = NULL;
            cur = fun;

        } else {
            if (INSTR_HAS_FLAG(iterator->b_last, FUN_END)) {
                cur->f_last = iterator;
            }
        }

        iterator->b_fun = cur;
    }

    f_list = bbl_list->b_fun;
}



/*
 *  MarkBasicBlocks(list) --
 *
 *  Raises a flag on all instructions which are members of jump classes
 *  e.g. ret/call/branch/jmp and marks them as the end of basic
 *  blocks.  The instructions right after them are marked as the beginning
 *  of basic blocks.  Eventually the targets of these instructions will
 *  also be marked as the beginning of basic blocks.
 */
void MarkBasicBlocks(Instruction *list) {
    Instruction *iterator;

    iterator = list;
    INSTR_SET_FLAG(startins, BBL_START);

    if (startins->i_prev) {
        INSTR_SET_FLAG(startins->i_prev, BBL_END);
    }

    if (iterator) {
        INSTR_SET_FLAG(iterator, BBL_START);
    }

    FOREACH_INSTR(iterator) {

	if (isPCChangingClass(iterator) || 
            isDataInstruction(iterator)) {
	    INSTR_SET_FLAG(iterator, BBL_END);
	    if (iterator->i_next) {
            INSTR_SET_FLAG(iterator->i_next, BBL_START);
	    }
	}

	if (!iterator->i_next) {
	    INSTR_SET_FLAG(iterator, FUN_END);
	    INSTR_SET_FLAG(iterator, BBL_END);
	    break;
	}
    }
    return;
}



/*
 *  CountStructures() --
 *
 *  Counts the number of instructions, functions, sections and
 *  basic blocks that are being used, as stores the results in
 *  numInstructions, numSections, numFunctions and numBlocks
 *  respectively.
 */
void CountStructures() {
    Block *blocks;
    Instruction *instructions;
    Function *functions;
    Section *sections;

    numSections = numInstructions = numBlocks = numFunctions = 0;

    FOREACH_FUN(functions) {
	numFunctions++;
    }

    FOREACH_BBL(blocks) {
	numBlocks++;
    }

    FOREACH_INSTR(instructions) {
	numInstructions++;
    }

    FOREACH_SECTION(sections) {
	numSections++;
    }

    return;
}



/*
 *  BBLOutputStats() --
 *
 *  Prints out various stats about the control flow graph. Currently
 *  printed are:
 *  (1) Number of instructions, functions and basic blocks
 *  (2) Number of unresolved references (Edges to HELL)
 *  (3) Thenumber and types of different edges
 */
void BBLOutputStats() {
    EdgeNode *eiterator;
    int numHell = 0;

    CountStructures();
    REPORT(RP_MINOR, "Control flow graph:");
    REPORT(RP_MINOR, "%10d instructions", numInstructions);
    REPORT(RP_MINOR, "%10d blocks", numBlocks);
    REPORT(RP_MINOR, "%10d functions", numFunctions);

    FOREACH_PRED_IN_BBL(HELL, eiterator) {
	numHell++;
    }
    REPORT(RP_DETAIL, "%10d HELL predecessors", numHell);

    numHell = 0;
    FOREACH_SUCC_IN_BBL(HELL, eiterator) {
	numHell++;
    }
    REPORT(RP_DETAIL, "%10d HELL successors", numHell);

    return;
}



/*
 *  BBLEdgeExists(src,dest) --
 *
 *  Returns true if there exists an edge from the source to the
 *  destination, or false if such an edge does not exist.
 */
bool BBLEdgeExists(Block *src, Block *dest) {
    EdgeNode *iterator;

    FOREACH_SUCC_IN_BBL(src, iterator) {
	if (NODE_TO(iterator) == dest) {
	    return true;
	}
    }
    return false;
}



/*
 *  BblNew() --
 *
 *  This calls BBLCreate then puts the bbl at the end of the basic block list.
 */
Block *BblNew() {
    Block *bbl;
    Block *iterator;

    bbl = BBLCreate();

    for (iterator = bbl_list; iterator->b_next != NULL;
	iterator = iterator->b_next);

    iterator->b_next = bbl;
    bbl->b_prev = iterator;

    return bbl;
}



/*
 *  FunNew() --
 *
 *  This calls FunCreate, then puts the function at the end of the
 *  function list.
 */
Function *FunNew() {
    Function *newfun;
    Function *iterator;

    newfun = FunCreate();

    for (iterator = f_list; iterator->f_next != NULL;
	iterator = iterator->f_next);

    iterator->f_next = newfun;
    newfun->f_prev = iterator;

    return newfun;
}



/*
 *  EdgeCreate() -- sets up a new edge for use and returns a pointer.
 */
Edge *EdgeCreate() {
    Edge *edge;

    edge = zalloc(sizeof(Edge));
    edge->e_id = globalEID++;

    return edge;
}



/*
 *  EdgeNodeCreate() -- sets up a new edge node for use and returns a pointer.
 */
EdgeNode *EdgeNodeCreate() {
    EdgeNode *node;

    node = zalloc(sizeof(EdgeNode));
    node->en_id = globalEID++;
    return node;
}



/*
 *  OperandCreate() -- allocates the memory for a new operand.
 */
Operand *OperandCreate() {
    return zalloc(sizeof(Operand));
}



/*
 *  FunResolveExitBlocks() --
 *
 *  Scans through all functions and looks for return statements.
 *  Attaches edges from the return statements (possible exit points)
 *  to the exit block associated with the function
 */
void FunResolveExitBlocks() {
    Instruction *iterator;

    FOREACH_INSTR(iterator) {
        if (isReturnClass(iterator)) {
            if (!INSTR_FUNCTION(iterator)) {
                ReportInstrError("Instruction without function", iterator);
            }
            if (!iterator->i_Bblock) {
                ReportInstrError("Instruction without basic block", iterator);
            }
            BBLAddEdge(iterator->i_Bblock,
                INSTR_FUNCTION(iterator)->f_exit, ET_EXIT);
            continue;
        }

        if (isDataInstruction(iterator) &&
            iterator->i_Bblock == INSTR_FUNCTION(iterator)->f_first &&
            iterator->i_Bblock == INSTR_FUNCTION(iterator)->f_last) {

            REPORT(RP_TRIVIAL, "Adding Exit edge to data function %s\n",
                INSTR_FUNCTION(iterator)->f_name);

            BBLAddEdge(iterator->i_Bblock,
                INSTR_FUNCTION(iterator)->f_exit, ET_EXIT);
        }        
    }

    return;
}



/*
 *  SetupHELL() --
 *
 *  Fow now, we don't want to preserve HELL across multiple control
 *  flow graph constructions.  We simply free the memory used by its
 *  edges and set up a new one.
 */
void SetupHELL() {
    Function *hellfun;

    if (HELL) {
	BBLRemoveEdges(HELL);
	free(HELL);
    }
    HELL = BBLCreate();
    hellfun = FunCreate();
    HELL->b_fun = hellfun;
    hellfun->f_first = HELL;
    hellfun->f_last = HELL;
    HELL->b_type = BT_HELL;
    return;
}



/*
 *  AddUPMHellEdges() --
 *
 *  Adds HELL edges from HELL to those basic blocks that have UPMs
 *  pointing to them with undecipherable sources.
 */
static void AddUPMHellEdges() {
    Patch *uiterator;
    Instruction *dest_instr, *src_instr;
    Section *dest_section;
    Iterator *itr;
    Location *src_member, *dest_member;

    REPORT(RP_MINOR, WITHFUNC("Adding UPM HELL edges"));

    BBLAddEdge(HELL, startins->i_Bblock, ET_HELLMAYBE);
    AddEntryEdge(startins->i_Bblock);

    itr = ll_iterator(upm2s_list);
    while (itr->it_hasNext(itr)) {
	uiterator = (Patch *) itr->it_next(itr);

	BEGIN_FOREACH_UPM_SOURCE_MEMBER(uiterator, src_member) {
	    if (UPM_MEMBER_INSTR(src_member) != NULL) {
		src_instr = *UPM_MEMBER_INSTR(src_member);

		/*
		 * Now that we have identified a UPM that sources an
		 * instruction, we need to try and identify how it will be used
		 * and from where.
		 */

		if (!INSTR_HAS_FLAG(src_instr, BBL_START)) {
		    BBLDivideBlock(src_instr->i_Bblock, src_instr);
		}

		dest_section = NULL;
		dest_instr = NULL;
		BEGIN_FOREACH_UPM_DEST_MEMBER(uiterator, dest_member) {
		    if (UPM_MEMBER_SECTION(dest_member) != NULL) {
			/*
			 * Since the instruction address is put into a section
			 * somewhere we need to figure out who uses it.
			 * If it is usedin the same function we need to try
			 * and locate the basic block that uses it - if it is
			 * used in any other spot, we will assume that we
			 * don't know where the HELL it's coming from...get it?
			 * HELL ha ha haa...that's PLTO humor late at
			 * night...I guess we could recursively search for all
			 * consumers of the relocation, but we don't currently
			 * do that.
			 */

			dest_section = *UPM_MEMBER_SECTION(dest_member);

			break;
		    } else if (UPM_MEMBER_INSTR(dest_member) != NULL) {
			if (*UPM_MEMBER_INSTR(dest_member) ==
			    *UPM_MEMBER_INSTR(src_member)) {
			    continue;
			}

			/*
			 * Since the consumer is an instruction, we need to find
			 * out what function actually does something with it.
			 */

			dest_instr = *UPM_MEMBER_INSTR(dest_member);
			break;
		    }
		}			// end for destinations

        END_FOREACH_UPM_MEMBER;

		if (dest_instr != NULL) {

		    if (INSTR_FUNCTION(dest_instr)!=INSTR_FUNCTION(src_instr)) {
			/*
			 * Since the instruction originates from outside of the
			 * target function, we will assume that it is used to
			 * enter the function whether it is a PC changing
			 * instruction or not.
			 */
			AddEntryEdge(src_instr->i_Bblock);
		    }

		    if (!isPCChangingClass(dest_instr)) {
			/*
			 * Here what we'd like to do is see if we can figure
			 * out if a PC changing instruction consumes this
			 * somewhere, and if so add a real edge between that
			 * consumer and the destination.  If we can't find
			 * such a thing then we just add a HELL edge
			 * to the destination. Currently we don't look around,
			 * we just pretend as if we couldn't find the source.
			 */
			BBLAddEdge(HELL, src_instr->i_Bblock, ET_HELLMAYBE);

			if (BBLHasEntryPredecessor(src_instr->i_Bblock)) {
			    BBLAddEdge(HELL,
				INSTR_FUNCTION(src_instr)->f_entry,
				ET_HELLMAYBE);
			    BBLAddEdge(INSTR_FUNCTION(src_instr)->f_exit,
				HELL, ET_HELL);
			}
		    } else {
			ETYPE type;

			if (isConditionalInstr(dest_instr)) {
			    type = ET_TRUE;
			} else if (isCallClass(dest_instr)) {
			    type = ET_FUNDIRECT;
			} else {
			    type = ET_UNCOND;
			}

			BBLAddEdge(dest_instr->i_Bblock, src_instr->i_Bblock,
			    type);
		    }
		} else if (dest_section != NULL) {
		    /*
		     * Since we don't know how the address is being used, we can
		     * say for sure whether or not it will be used to enter the
		     * function, and therefore conservatively say that it is
		     * indeed an entry point.
		     */

		    if (!SECTION_HAS_FLAG(dest_section,SECTION_DEBUGGING_FLAG)){
			AddEntryEdge(src_instr->i_Bblock);
		    }

		    BBLAddEdge(HELL, src_instr->i_Bblock, ET_HELLMAYBE);
		    if (BBLHasEntryPredecessor(src_instr->i_Bblock)) {
			BBLAddEdge(HELL, INSTR_FUNCTION(src_instr)->f_entry,
			    ET_HELLMAYBE);
			BBLAddEdge(INSTR_FUNCTION(src_instr)->f_exit,
			    HELL, ET_HELL);
		    }
		}
	    }
	}				// end for sources
	END_FOREACH_UPM_MEMBER;
    }					// end upm iteration

    free(itr);
}



/*
 *  BBLGetSyscallLinkEdge(bbl) --
 */
Edge *BBLGetSyscallLinkEdge(Block *bbl) {
    EdgeNode *niterator;

    if (bbl->b_type != BT_SYSCALL) {
        WARNING("BBLSyscallLinkEdge: bbl type is not BT_SYSCALL: bbl-id = %d", 
                bbl->b_id);
    }

    FOREACH_SUCC_IN_BBL(bbl, niterator) {
	if (NODE_TYPE(niterator) == ET_SYSCALL_LINK) {
	    return niterator->en_edge;
	}
    }

    return NULL;
}

Edge *BBLGetHaltLinkEdge(Block *block) {
    EdgeNode *niterator;

    if (block->b_type != BT_HALT) {
        WARNING("BBLGetFunlinkEdge: bbl type is not BT_CALL: bbl-id = %d", 
                block->b_id);
    }

    FOREACH_SUCC_IN_BBL(block, niterator) {
	if (NODE_TYPE(niterator) == ET_HALTLINK) {
	    return niterator->en_edge;
	}
    }
    return NULL;
}


/*
 *  BBLGetFunlinkEdge(block) --
 *
 *  Returns an edge of type ET_FUNLINK in the basic block's successor list.
 */
Edge *BBLGetFunlinkEdge(Block *block) {
    EdgeNode *niterator;

    if (block->b_type != BT_CALL) {
        WARNING("BBLGetFunlinkEdge: bbl type is not BT_CALL: bbl-id = %d", 
                block->b_id);
    }

    FOREACH_SUCC_IN_BBL(block, niterator) {
	if (NODE_TYPE(niterator) == ET_FUNLINK) {
	    return niterator->en_edge;
	}
    }
    return NULL;
}



/*
 *  BBLGetFunDirectEdge(block) --
 *
 *  Returns and edge of type ET_FUNDIRECT in the basic block's successor list.
 */
Edge *BBLGetFunDirectEdge(Block *block) {
    EdgeNode *niterator;

    if (block->b_type != BT_CALL) {
        WARNING("BBLGetFunDirectEdge: bbl type is not BT_CALL: bbl-id = %d", 
                block->b_id);
    }

    FOREACH_SUCC_IN_BBL(block, niterator) {
	if (NODE_TYPE(niterator) == ET_FUNDIRECT) {
	    return niterator->en_edge;
	}
    }
    return NULL;
}



/*
 *  BBLGetFuncallEdge(block) --
 *
 *  Returns an edge of type ET_FUNCALL in the basic block's successor list.
 */
Edge *BBLGetFuncallEdge(Block *block) {
    EdgeNode *niterator;

    if (block->b_type != BT_CALL) {
        WARNING("BBLGetFuncallEdge: bbl type is not BT_CALL: bbl-id = %d", 
                block->b_id);
    }

    FOREACH_SUCC_IN_BBL(block, niterator) {
	if (NODE_TYPE(niterator) == ET_FUNCALL) {
	    return niterator->en_edge;
	}
    }
    return NULL;
}



/*
 *  BBLGetUncondEdge(block) --
 *
 *  Returns the first (and hopefully only) edge of type ET_UNCOND going out
 *  from 'block'.
 */
Edge *BBLGetUncondEdge(Block *block) {
    EdgeNode *niterator;
    
    if (block->b_type != BT_UNCOND) {
        WARNING("BBLGetUncondEdge: bbl type is not BT_UNCOND: bbl-id = %d", 
                block->b_id);
    }

    FOREACH_SUCC_IN_BBL(block, niterator) {
	if (NODE_TYPE(niterator) == ET_UNCOND) {
	    return niterator->en_edge;
	}
    }
    return NULL;
}



/*
 *  BBLGetNormalEdge(block) --
 *
 *  Returns the first (and hopefully only) edge of type ET_NORMAL going out
 *  from 'block'.
 */
Edge *BBLGetNormalEdge(Block *block) {
    EdgeNode *niterator;
    
    if (block->b_type != BT_NORMAL) {
        WARNING("BBLGetNormalEdge: bbl type is not BT_NORMAL: bbl-id = %d", 
                block->b_id);
    }

    FOREACH_SUCC_IN_BBL(block, niterator) {
	if (NODE_TYPE(niterator) == ET_NORMAL) {
	    return niterator->en_edge;
	}
    }
    return NULL;
}



/*
 *  BBLGetTrueEdge(block) --
 *
 *  Returns the first (and hopefully only) edge of type ET_TRUE going out
 *  from 'block'.
 */
Edge *BBLGetTrueEdge(Block *block) {
    EdgeNode *niterator;

    if (block->b_type != BT_BRANCH) {
        WARNING("BBLGetTrueEdge: bbl type is not BT_BRANCH: bbl-id = %d", 
                block->b_id);
    }

    FOREACH_SUCC_IN_BBL(block, niterator) {
	if (NODE_TYPE(niterator) == ET_TRUE) {
	    return niterator->en_edge;
	}
    }
    return NULL;
}



/*
 *  BBLGetFalseEdge(block) --
 *
 *  Returns the first (and hopefully only) edge of type ET_FALSE going out
 *  from 'block'.
 */
Edge *BBLGetFalseEdge(Block *block) {
    EdgeNode *niterator;

    if (block->b_type != BT_BRANCH) {
        WARNING("BBLGetFalseEdge: bbl type is not BT_BRANCH: bbl-id = %d", 
                block->b_id);
    }

    FOREACH_SUCC_IN_BBL(block, niterator) {
	if (NODE_TYPE(niterator) == ET_FALSE) {
	    return niterator->en_edge;
	}
    }
    return NULL;
}



/*
 *  BBLGetReturnPredecessor(block) --
 *
 *  Returns the first (and hopefully only) edge of type ET_RETURN coming
 *  in to the supplied block.
 */
Edge *BBLGetReturnPredecessor(Block *block) {
    EdgeNode *niterator;

    FOREACH_PRED_IN_BBL(block, niterator) {
	if (NODE_TYPE(niterator) == ET_RETURN) {
	    return niterator->en_edge;
	}
    }
    return NULL;
}



/*
 *  BBLHasHellSuccessor(bbl) --
 *
 *  Returns true if the supplied basic block has a successor edge
 *  going to HELL.
 */
bool BBLHasHellSuccessor(Block *bbl) {
    EdgeNode *niterator;

    FOREACH_SUCC_IN_BBL(bbl, niterator) {
	if (NODE_TO(niterator) == HELL) {
	    return true;
	}
    }
    return false;
}



/*
 *  BBLHasHellPredecessor(bbl) --
 *
 *  Returns true if the supplied basic block has a predecessor edge
 *  coming from HELL.
 */
bool BBLHasHellPredecessor(Block *bbl) {
    EdgeNode *niterator;

    FOREACH_PRED_IN_BBL(bbl, niterator) {
	if (NODE_TYPE(niterator) == ET_HELL) {
	    return true;
	}
	if (NODE_FROM(niterator) == HELL) {
	    return true;
	}
	if (NODE_TYPE(niterator) == ET_HELLMAYBE) {
	    return true;
	}
    }

    return false;
}



/*
 *  BBLIsFirstBBLInFun(bbl) --
 *
 *  Returns true if the supplied basic block is the first basic block
 *  in its function.
 */
bool BBLIsFirstBBLInFun(Block *bbl) {
    return bbl->b_fun->f_first == bbl;
}



/*
 *  BBLIsLastBBLInFun(bbl) --
 *
 *  Returns true if the supplied basic block is the last basic block
 *  in its function.
 */
bool BBLIsLastBBLInFun(Block *bbl) {
    return bbl->b_fun->f_last == bbl;
}



/*
 *  BBLNumberOfPredecessors(bbl) --
 *
 *  Returns the number of predecessors that 'bbl' has.
 */
int BBLNumberOfPredecessors(Block *bbl) {
    int count = 0;
    EdgeNode *preds;

    FOREACH_PRED_IN_BBL(bbl, preds) {
	count++;
    }

    return count;
}



/*
 *  BBLNumberOfSuccessors(bbl) --
 *
 *  Returns the number of successors that 'bbl' has.
 */
int BBLNumberOfSuccessors(Block *bbl) {
    int count = 0;
    EdgeNode *succs;

    FOREACH_SUCC_IN_BBL(bbl, succs) {
	count++;
    }

    return count;
}



/*
 *  isInterproceduralEdge(edge) --
 *
 *  Returns true if 'edge' is an interprocedural edge.
 *  An interprocedural edge is defined to be an edge originating from a
 *  basic block in a function A whose destination is a basic block in a
 *  function B. A != B.
 */
bool isInterproceduralEdge(Edge *edge) {

    if (edge->e_type == ET_FUNDIRECT) {
	return true;
    }
    if (edge->e_type == ET_RETURN) {
	return true;
    }
    if (edge->e_type == ET_ENTRY) {
	return true;
    }
    if (edge->e_type == ET_EXIT) {
	return true;
    }
    if (edge->e_type == ET_HELL) {
	return true;
    }
    if (edge->e_type == ET_HELLMAYBE) {
	return true;
    }

    if (edge->e_to->b_fun == edge->e_from->b_fun) {
	return false;
    }

    return true;
}



/*
 *  CountFuncalls() -- counts the known function calls in the flowgraph.
 */
long long CountFuncalls() {
    Block *biterator;
    long long numFunCalls = 0;

    FOREACH_BBL(biterator) {
	if (biterator->b_type == BT_CALL) {
	    Edge *funDirect = BBLGetFunDirectEdge(biterator);
	    if (funDirect != NULL) {
		numFunCalls = numFunCalls + funDirect->e_edgeWeight;
	    }
	}
    }

    return numFunCalls;
}



/*
 *  BBLGetEdgeToBBL(src,dest) --
 *
 *  Returns the first edge of any type between the two basic blocks.
 */
Edge *BBLGetEdgeToBBL(Block *src, Block *dest) {
    EdgeNode *niterator;

    FOREACH_SUCC_IN_BBL(src, niterator) {
	if (NODE_TO(niterator) == dest) {
	    return niterator->en_edge;
	}
    }
    return NULL;
}



/*
 *  BBLGetEscapingEdge(bbl) --
 *
 *  Returns the escaping edge for the specified basic block if one exists.
 *
 *  bbl: the basic block that may have the outgoing escaping edge
 *  returns: the escaping edge or NULL if there is not one
 */
Edge *BBLGetEscapingEdge(Block *bbl) {
    EdgeNode *niterator;

    if ((bbl->b_type != BT_UNCOND) &&
	(bbl->b_type != BT_BRANCH) &&
	(bbl->b_type != BT_SWITCH) &&
	(bbl->b_type != BT_NORMAL) && (bbl->b_type != BT_HALT)) {
	return NULL;
    }

    FOREACH_SUCC_IN_BBL(bbl, niterator) {
	if (isInterproceduralEdge(niterator->en_edge)) {
	    return niterator->en_edge;
	}
    }

    return NULL;
}



/*
 *  BBLRemoveIncomingEdges(bbl) -- removes all edges coming into 'bbl'.
 */
void BBLRemoveIncomingEdges(Block *bbl) {
    EdgeNode *preds, *next;

    preds = bbl->b_pred;
    while (preds != NULL) {
	next = preds->en_next;
	BBLRemoveEdge(NODE_FROM(preds), NODE_TO(preds));
	preds = next;
    }

    return;
}



/*
 *  FunNumberOfInstructions(fun) --
 *
 *  Returns the number of instructions inside the function.
 */
int FunNumberOfInstructions(Function *fun) {
    Instruction *instr;
    Block *bbl;
    int number = 0;

    FOREACH_BBL_IN_FUN(fun, bbl) {
	FOREACH_INSTR_IN_BBL(bbl, instr) {
	    number++;
	}
    }
    return number;
}



/*
 *  BBLHasExitSuccessor(bbl) --
 *
 *  Returns whether or not the specified basic block has an EXIT block
 *  successor.
 *
 *  bbl: the basic block to check
 *  returns: true if the specified basic block has an EXIT block successor
 */
bool BBLHasExitSuccessor(Block *bbl) {
    EdgeNode *succn;

    FOREACH_SUCC_IN_BBL(bbl, succn) {
	if (NODE_TO(succn)->b_type == BT_EXIT) {
	    return true;
	}
    }

    return false;
}



/*
 *  BBLHasEntryPredecessor(bbl) --
 *
 *  Returns true if 'bbl' has a predecessor of type BT_ENTRY.
 */
bool BBLHasEntryPredecessor(Block *bbl) {
    EdgeNode *preds;

    FOREACH_PRED_IN_BBL(bbl, preds) {
	if (NODE_FROM(preds)->b_type == BT_ENTRY) {
	    return true;
	}
    }
    return false;
}



/*
 *  BBLHasInterproceduralPredecessor(bbl) --
 *
 *  Returns true if 'bbl' has an incoming interprocedural edge.
 */
bool BBLHasInterproceduralPredecessor(Block *bbl) {
    EdgeNode *preds;

    FOREACH_PRED_IN_BBL(bbl, preds) {
	if (isInterproceduralEdge(preds->en_edge)) {
	    return true;
	}
    }
    return false;
}



/*
 *  InstrExecutedAfterInstr(instr) --
 *
 *  Returns the instruction that will be executed immediately after 'instr', or
 *  NULL if none exists or it cannot be determined.
 */
Instruction *InstrExecutedAfterInstr(Instruction *instr) {

    if (instr->i_Bblock->b_last != instr) {
	return instr->i_next;
    }

    if (BBLNumberOfSuccessors(instr->i_Bblock) != 1) {
	return NULL;
    }

    if (BBLHasHellSuccessor(instr->i_Bblock)) {
	return NULL;
    }

    return NODE_TO(instr->i_Bblock->b_succ)->b_first;
}



/*
 *  MergeBasicBlocks() --
 *
 *  Merges all basic blocks in the control flow graph that are connected
 *  by ET_NORMAL edges that have exactly 1 successor and predecessor.
 */
void MergeBasicBlocks() {
    Block *blocks, *succ;

    REPORT(RP_DETAIL, "Merging basic blocks");
    FOREACH_BBL(blocks) {
	if (blocks->b_type != BT_NORMAL) {
	    continue;
	}
	if (BBLNumberOfSuccessors(blocks) != 1) {
	    continue;
	}
	succ = NODE_TO(blocks->b_succ);
	if (BBLNumberOfPredecessors(succ) != 1) {
	    continue;
	}
	BBLMerge(blocks, succ);
    }

    return;
}



/*
 *  BBLMerge(first,second) --
 *
 *  Merges bbl 'first' with bbl 'second' by collapsing the ET_NORMAL edge
 *  between them.
 */
static void BBLMerge(Block *first, Block *second) {
    Instruction *instr, *instrNext;
    EdgeNode *successors;

    first->b_flist = second->b_flist;
    first->b_type = second->b_type;
    BBLRemoveEdges(first);

    FOREACH_SUCC_IN_BBL(second, successors) {
	BBLAddEdge(first, NODE_TO(successors), NODE_TYPE(successors))
	    ->e_edgeWeight = NODE_WEIGHT(successors);
    }

    instr = second->b_first;

    if (instr != NULL) {
	INSTR_REMOVE_FLAG(instr, BBL_START);
    }

    while ((instr != NULL) && (instr->i_Bblock == second)) {
	instrNext = instr->i_next;
	InstrDisconnect(instr);
	INSTR_REMOVE_FLAG(instr, BBL_START);
	InsertInstrAfterInstr(instr, first->b_last);
	first->b_last = instr;
	instr = instrNext;
    }

    BBLRemoveEdges(second);
    BBLDisconnect(second);
    BBLFree(second);

    return;
}



/*
 *  BBLSumIncomingEdges(bbl) --
 *
 *  Sums the weight of all incoming edges and returns the result
 */
EDGE_WEIGHT_TYPE BBLSumIncomingEdges(Block *bbl) {
    EdgeNode *niterator;
    EDGE_WEIGHT_TYPE result = 0;

    FOREACH_PRED_IN_BBL(bbl, niterator) {
	if (NODE_TYPE(niterator) == ET_ENTRY) {
	    continue;
	}
	if (NODE_TYPE(niterator) == ET_EXIT) {
	    continue;
	}
	if (NODE_TYPE(niterator) == ET_FUNCALL) {
	    continue;
	}
	if (NODE_TYPE(niterator) == ET_FUNLINK) {
	    continue;
	}
	result = result + NODE_WEIGHT(niterator);
    }

    return result;
}



/*
 *  BBLSumOutgoingEdges(bbl) --
 *
 *  Sums the weight of all outgoing edges and returns the result
 */
EDGE_WEIGHT_TYPE BBLSumOutgoingEdges(Block *bbl) {
    EdgeNode *niterator;
    EDGE_WEIGHT_TYPE result = 0;

    FOREACH_SUCC_IN_BBL(bbl, niterator) {
	if (NODE_TYPE(niterator) == ET_ENTRY) {
	    continue;
	}
	if (NODE_TYPE(niterator) == ET_EXIT) {
	    continue;
	}
	if (NODE_TYPE(niterator) == ET_FUNCALL) {
	    continue;
	}
	if (NODE_TYPE(niterator) == ET_FUNLINK) {
	    continue;
	}
	result = result + NODE_WEIGHT(niterator);
    }

    return result;
}



/*
 *  FunNumberOfCallSites(fun) --
 *
 *  Returns the number of callsites in the function 'fun'.
 */
int FunNumberOfCallSites(Function *fun) {
    return BBLNumberOfPredecessors(fun->f_entry);
}



/*
 *  BBLNumberOfInstructions(bbl) --
 *
 *  Returns the number of instructions in 'bbl'.
 */
int BBLNumberOfInstructions(Block *bbl) {
    Instruction *instr;
    int count = 0;

    FOREACH_INSTR_IN_BBL(bbl, instr) {
	count++;
    }
    return count;
}



/*
 *  BBLNumberOfEdgesToBBL(bbl1,bbl2) --
 *
 *  Returns the number of edges from bbl1 to bbl2.
 */
int BBLNumberOfEdgesToBBL(Block *bbl1, Block *bbl2) {
    EdgeNode *successors;
    int count = 0;

    FOREACH_SUCC_IN_BBL(bbl1, successors) {
	if (NODE_TO(successors) == bbl2) {
	    count++;
	}
    }

    return count;
}



/*
 *  BBLNumberOfPredecessorsFromBBL(bbl1,bbl2) --
 *
 *  Returns the number of times bbl2 appears as a predecessor to bbl1.
 */
int BBLNumberOfPredecessorsFromBBL(Block *bbl1, Block *bbl2) {
    EdgeNode *preds;
    int count = 0;

    FOREACH_PRED_IN_BBL(bbl1, preds) {
	if (NODE_FROM(preds) == bbl2) {
	    count++;
	}
    }

    return count;
}



/*
 *  BBLHasFunlinkPredecessor(block) --
 *
 *  Returns true if 'block' has an incoming edge of type ET_FUNLINK
 */
bool BBLHasFunlinkPredecessor(Block *block) {
    EdgeNode *preds;

    FOREACH_PRED_IN_BBL(block, preds) {
	if (NODE_TYPE(preds) == ET_FUNLINK) {
	    return true;
	}
    }

    return false;
}



/*
 *  BBLIsIndirectTarget(bbl) --
 *
 *  Returns true if the BBL is the target of an indirect call.
 */
bool BBLIsIndirectTarget(Block *bbl) {
    if (BBLHasHellPredecessor(bbl)) {
	return true;
    }
    if (BBLHasEntryPredecessor(bbl) &&
	BBLHasHellPredecessor(bbl->b_fun->f_entry)) {
	return true;
    }
    return false;
}



/*
 *  GetPICCalls() --
 *
 *  Returns a linked list of all of the PIC calls that could be identified.
 *  returns: the linked list containing all PIC calls that could be identified
 */
LinkedList *GetPICCalls() {
    LinkedList *calls;
    Instruction *iiterator;

    calls = ll_new(LIST_UNSORTED);

    FOREACH_INSTR(iiterator) {
	if (IsPICCall(iiterator)) {
	    ll_add(calls, iiterator);
	}
    }

    return calls;
}



/*
 *  BBLRemoveOutgoingEdges(bbl) --
 */
void BBLRemoveOutgoingEdges(Block *bbl) {
    EdgeNode *succs, *next;

    succs = bbl->b_succ;
    while (succs != NULL) {
	next = succs->en_next;
	BBLRemoveEdge(NODE_FROM(succs), NODE_TO(succs));
	succs = next;
    }

    return;
}



/*
 *  BBLGetEdgeTyped(block,type) --
 */
Edge *BBLGetEdgeTyped(Block *block, ETYPE type) {
    EdgeNode *niterator;

    FOREACH_SUCC_IN_BBL(block, niterator) {
	if (NODE_TYPE(niterator) == type) {
	    return niterator->en_edge;
	}
    }

    return NULL;
}



/*
 *  BBLGetEdgeToBBLTyped(src,dest,type) --
 *
 *  Returns the edge from src to dest that has the specified type.
 *
 *  src: the basic block the edge originates from
 *  dest: the basic block the edge goes to
 *  type: the edge type to be matched
 *  returns: the edge with the specified type connecting the two basic
 *  blocks, or NULL if no matching edge is found
 */
Edge *BBLGetEdgeToBBLTyped(Block *src, Block *dest, ETYPE type) {
    EdgeNode *niterator;

    FOREACH_SUCC_IN_BBL(src, niterator) {
	if ((NODE_TO(niterator) == dest) && (type == NODE_TYPE(niterator))) {
	    return niterator->en_edge;
	}
    }

    return NULL;
}



/*
 *  InstrIsEscapingJump(instr) --
 */
bool InstrIsEscapingJump(Instruction *instr) {
    return BBLHasEsc2(instr->i_Bblock);
}



/*
 *  MarkFlistCalls() -- Check for all calls that call into an flist.
 *
 *  Iterates through the global instruction list and sets the INSTR_FLIST_FLAG
 *  for an pc changing instruction that references an FLIST function and points
 *  the (()->b_flist) of that instruction's basic block at that FLIST.
 *  see also: INSTR_FLIST_FLAG
 *  see also: (()->b_flist)
 */
void MarkFlistCalls() {
    Patch *upm;
    LinkedList *upms;
    Instruction *iiterator;
    ADDRESS offset;
    Section *section;
    Iterator *itr;
    UPMList *list;
    int i;

    FOREACH_INSTR(iiterator) {
	if (!isPCChangingClass(iiterator) ||
	    INSTR_HAS_FLAG(iiterator, INDIRECT)) {
	    continue;	// don't bother: indirect jump or not PC changing
	}

	INSTR_REMOVE_FLAG(iiterator, INSTR_FLIST_FLAG);
	upms = UPMTableGetByDest(iiterator->i_src);
	itr = ll_iterator(upms);
	while (itr->it_hasNext(itr)) {
	    upm = (Patch *) itr->it_next(itr);

	    list = upm->p_GetSources(upm);
	    for (i = 0; i < list->u_size; i++) {
		if (UPM_MEMBER_SECTION(&list->u_members[i]) == NULL) {
		    continue;
		}

		if ((list->u_size == (i - 1)) ||
		    (UPM_MEMBER_OFFSET(&list->u_members[i + 1]) == NULL)) {
		    continue;
		}

		section = *UPM_MEMBER_SECTION(&list->u_members[i]);
		offset = *UPM_MEMBER_OFFSET(&list->u_members[i + 1]);

		WARNING("data instr in flowgraph treated as regular function");
		INSTR_SET_FLAG(iiterator, INSTR_FLIST_FLAG);
		iiterator->i_Bblock->b_flist = GetFlistAtOffset(offset);
		if (iiterator->i_Bblock->b_flist == NULL) {
		    REPORT(RP_MINOR, "No FLIST for offset %x", offset);
		}

	    }
	    UPMMemberListDestroy(list);
	}
	free(itr);
	ll_free(upms);
    }
}



/*
 *  ReplaceHaltBranchesWithHalts() --
 *
 *  This function iterates through all instructions, replacing any jump or call
 *  instructions to 0 with halt instructions. This simplifies life a great deal
 *  and reduces lots of HELL edges.
 */
void ReplaceHaltBranchesWithHalts() {
    int replaced;
    Instruction *iiterator, *next, *new;

    REPORT(RP_MINOR, WITHFUNC("Converting zero-jumps and calls"));
    replaced = 0;
    iiterator = i_list;
    while (iiterator != NULL) {
	next = iiterator->i_next;

    // CLASS_LJUMP1 is treated as jumpclass instruction
    // and is also a data instruction.
    // figure out later where it should go.
	if ((isJumpClass(iiterator) && !isConditionalInstr(iiterator) &&
            !isDataInstruction(iiterator)) ||
	    isCallClass(iiterator)) {
	    if (iiterator->i_src->a_displacement +
		iiterator->i_address + iiterator->i_length == 0) {
		new = InstrCreateHalt();

		new->i_next = iiterator->i_next;
		new->i_prev = iiterator->i_prev;

		if (iiterator->i_next != NULL) {
		    iiterator->i_next->i_prev = new;
		}

		if (iiterator->i_prev != NULL) {
		    iiterator->i_prev->i_next = new;
		}

		new->i_address = iiterator->i_address;
		InstrRemoveEntry(iiterator);
		InstrInsertEntry(new);

		free(iiterator->i_src);
		free(iiterator);
		replaced++;
	    }
	}

	iiterator = next;
    }

    REPORT(RP_DETAIL, "Replaced %d jmps and calls to 0 with halts", replaced);
}



/*
 *  RemoveUnconditionalEscapingEdgesFromFun(fun,call_slot) --
 */
bool RemoveUnconditionalEscapingEdgesFromFun(Function *fun, ADDRESS call_slot) {
    Instruction *mov, *stc, *clc, *ret, *last, *esc_jmp;
    Block *biterator, *esc_bbl, *esc_target;
    Edge *edge;
    bool hasHell, removed;

    removed = hasHell = false;

    if (BBLHasHellPredecessor(fun->f_entry)) {
	hasHell = true;
    }

    if (!hasHell && !FunHasEsc2(fun)) {
	return false;
    }

    FOREACH_BBL_IN_FUN(fun, biterator) {
	edge = BBLGetEscapingEdge(biterator);

	esc_bbl = biterator;

	if (edge != NULL) {
	    esc_target = edge->e_to;

	    if (!isJumpClass(biterator->b_last) ||
		isConditionalInstr(biterator->b_last)) {
		continue;
	    }

	    stc = InstrCreateSetCarry();
	    ret = InstrCreateReturn();
	    mov = InstrCreate32BitMovImmToMem(0x0, 0x0);
	    AddPatchInstrAddressIntoInstr(esc_target->b_first, mov, mov->i_src);
	    AddPatchSectionAddressIntoInstr(
		SECTION_REF_DATA, call_slot, mov, mov->i_dest);

	    esc_jmp = last = esc_bbl->b_last;
	    if ((esc_bbl->b_last->i_prev != NULL) &&
		isLeaveClass(esc_bbl->b_last->i_prev)) {
		last = esc_bbl->b_last->i_prev;
	    }

	    BBLRemoveEdges(esc_bbl);

	    InsertInstrBeforeInstr(mov, last);
	    InsertInstrBeforeInstr(stc, last);
	    InsertInstrAfterInstr(ret, last);
	    esc_bbl->b_type = BT_RETURN;

	    BBLAddEdge(last->i_Bblock, fun->f_exit, ET_EXIT);

	    InstrDisconnect(esc_jmp);
	    InstrFree(esc_jmp);

	    removed = true;
	} else if (BBLHasExitSuccessor(biterator)) {
	    clc = InstrCreateClearCarry();

	    last = esc_bbl->b_last;
	    if ((esc_bbl->b_last->i_prev != NULL) &&
		isLeaveClass(esc_bbl->b_last->i_prev)) {
		last = esc_bbl->b_last->i_prev;
	    }

	    InsertInstrBeforeInstr(clc, last);
	}
    }

    return removed;
}



/*
 *  cml_tmp(fun,call_slot) --
 */
bool cml_tmp(Function *fun, ADDRESS call_slot) {
    Instruction *mov, *ret, *last, *esc_jmp;
    Block *biterator, *esc_bbl, *esc_target;
    Edge *edge;
// EdgeNode *node;
    bool hasHell, removed, doIt;

    removed = hasHell = false;
    doIt = true;

    if (BBLHasHellPredecessor(fun->f_entry)) {
	hasHell = true;
    }

    if (!hasHell && !FunHasEsc2(fun)) {
	return false;
    }
// FOREACH_PRED_IN_BBL(((fun)->f_entry), node) {
//  if (FunHasEsc2(((NODE_FROM(node))->b_fun))) {
//    REPORT(RP_DETAIL, "Predecessor had escaping edges!");
//    doIt = false;
//  // return false;
//  }
// }

    FOREACH_BBL_IN_FUN(fun, biterator) {
	edge = BBLGetEscapingEdge(biterator);

	if (edge != NULL) {
	    if (FunHasEsc2(edge->e_to->b_fun)) {
		REPORT(RP_DETAIL, "Successor function has escaping edges!");
		doIt = false;
//      return false;
	    }
	}
    }

    FOREACH_BBL_IN_FUN(fun, biterator) {
	edge = BBLGetEscapingEdge(biterator);

	esc_bbl = biterator;

	if ((edge != NULL) && doIt) {
	    esc_target = edge->e_to;

	    if (!isJumpClass(biterator->b_last) ||
		isConditionalInstr(biterator->b_last)) {
		continue;
	    }

	    ret = InstrCreateReturn();
	    mov = InstrCreate32BitMovImmToMem(0x0, 0x0);
	    AddPatchInstrAddressIntoInstr(esc_target->b_first, mov, mov->i_src);
	    AddPatchSectionAddressIntoInstr(
		SECTION_REF_DATA, call_slot, mov, mov->i_dest);

	    esc_jmp = last = esc_bbl->b_last;
	    if ((esc_bbl->b_last->i_prev != NULL) &&
		isLeaveClass(esc_bbl->b_last->i_prev)) {
		last = esc_bbl->b_last->i_prev;
	    }

	    BBLRemoveEdges(esc_bbl);

	    InsertInstrBeforeInstr(mov, last);
	    InsertInstrAfterInstr(ret, last);
	    esc_bbl->b_type = BT_RETURN;

	    BBLAddEdge(last->i_Bblock, fun->f_exit, ET_EXIT);

	    InstrDisconnect(esc_jmp);
	    InstrFree(esc_jmp);

	    removed = true;
	} else if (BBLHasExitSuccessor(biterator)) {
	    Instruction *mov2, *add;
	    int reg;
	    bool free_reg;

	    last = esc_bbl->b_last;
	    if ((esc_bbl->b_last->i_prev != NULL) &&
		isLeaveClass(esc_bbl->b_last->i_prev)) {
		last = esc_bbl->b_last->i_prev;
	    }
	    // get free reg
	    // reg = GetFreeRegisterAtInstr(last);
	    reg = -1;

	    free_reg = true;
	    if (reg == -1) {
		reg = GetARegister();
		free_reg = false;
	    }

	    if (isLeaveClass(last)) {
		// mov (%ebp), B
		mov = InstrCreate32BitMovMemDispToReg(
		    GetFramePointer(), 0x4, reg);
	    } else {
		mov = InstrCreate32BitMovMemDispToReg(
		    GetStackPointer(), 0x8, reg);
	    }

	    add = InstrCreate32BitAddImmToReg(8, reg);
	    mov2 = InstrCreate32BitMovRegToMem(reg);
	    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, call_slot, mov2,
		mov2->i_dest);

	    InsertInstrBeforeInstr(mov, last);
	    InsertInstrAfterInstr(add, mov);
	    InsertInstrAfterInstr(mov2, add);

	    if (!free_reg) {
		// there isn't a free one, we'll push instead
		Instruction *push_pop;

		push_pop = InstrCreatePush(reg);
		InsertInstrBeforeInstr(push_pop, mov);
		push_pop = InstrCreatePop(reg);
		InsertInstrAfterInstr(push_pop, mov2);

		push_pop = InstrCreateSave();
		InsertInstrBeforeInstr(push_pop, mov->i_prev);
		push_pop = InstrCreateRestore();
		InsertInstrAfterInstr(push_pop, mov2->i_next);
	    }
	}
    }

    return removed && doIt;
}



/*
 *  UpdateEscapingFunCallers(fun,call_slot) --
 */
static void UpdateEscapingFunCallers(Function *fun, ADDRESS call_slot) {
    EdgeNode *predn;

    FOREACH_PRED_IN_BBL(fun->f_entry, predn) {
	if (NODE_TYPE(predn) == ET_FUNCALL) {
	    UpdateCallerForEscapingTarget2(NODE_FROM(predn), call_slot);
	}
    }
}



/*
 *  UpdateCallerForEscapingTarget2(bbl,call_slot) --
 */
static void UpdateCallerForEscapingTarget2(Block *bbl, ADDRESS call_slot) {
    Instruction *call_z, *jmp, *og_call, *og_funlink, *pop;
    Block *tmp_bbl;

    og_call = bbl->b_last;

    if (INSTR_HAS_FLAG(og_call, INSTR_TEMP_FLAG)) {
	return;
    }

    og_funlink = og_call->i_next;

    call_z = InstrCreateCallIndirect();
    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA,
	call_slot, call_z, call_z->i_src);
    tmp_bbl = BBLCreate();
    BblAppendInstr(tmp_bbl, call_z);
    InsertBBLAfterBBL(tmp_bbl, og_call->i_Bblock);

    jmp = InstrCreateJMP();
    tmp_bbl = BBLCreate();
    BblAppendInstr(tmp_bbl, jmp);
    InsertBBLAfterBBL(tmp_bbl, call_z->i_Bblock);

    pop = InstrCreate32BitPopToMem();
    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA,
	call_slot, pop, pop->i_dest);
    tmp_bbl = BBLCreate();
    BblAppendInstr(tmp_bbl, pop);
    InsertBBLAfterBBL(tmp_bbl, jmp->i_Bblock);

    jmp->i_Bblock->b_type = BT_UNCOND;
    jmp->i_Bblock->b_fun = INSTR_FUNCTION(og_call);

    call_z->i_Bblock->b_type = BT_CALL;
    call_z->i_Bblock->b_fun = INSTR_FUNCTION(og_call);

    pop->i_Bblock->b_type = BT_NORMAL;
    pop->i_Bblock->b_fun = INSTR_FUNCTION(og_call);

    BBLRemoveEdge(og_call->i_Bblock, og_funlink->i_Bblock);

    // set up the new edges
    BBLAddEdge(og_call->i_Bblock, call_z->i_Bblock, ET_FUNLINK);
    BBLAddEdge(call_z->i_Bblock, jmp->i_Bblock, ET_FUNLINK);
    BBLAddEdge(call_z->i_Bblock, HELL, ET_HELL);
    BBLAddEdge(HELL, call_z->i_Bblock->b_next, ET_HELL);
    BBLAddEdge(HELL, pop->i_Bblock, ET_HELL);
    BBLAddEdge(jmp->i_Bblock, og_funlink->i_Bblock, ET_UNCOND);
    BBLAddEdge(pop->i_Bblock, og_funlink->i_Bblock, ET_NORMAL);

    INSTR_SET_FLAG(og_call, INSTR_TEMP_FLAG);
    INSTR_SET_FLAG(call_z, INSTR_TEMP_FLAG);
}



/*
 *  UpdateIndirectCallers(fun,call_slot) --
 */
static void UpdateIndirectCallers(Function *fun, ADDRESS call_slot) {
    Block *biterator, *next;

    biterator = fun->f_first;
    while ((biterator != NULL) && (biterator->b_fun == fun)) {
	next = biterator->b_next;

	if ((biterator->b_type == BT_CALL) &&
	    INSTR_HAS_FLAG(biterator->b_last, INDIRECT)) {
	    UpdateCallerForEscapingTarget2(biterator, call_slot);
	}

	biterator = next;
    }
}



/*
 *  RemoveUnconditionalEscapingEdges() --
 */
void RemoveUnconditionalEscapingEdges() {
    Function *fiterator;
    Block *biterator;
    ADDRESS call_slot;

    /*
     *  When we remove an unconditional escaping edge we have to keep a few
     *  things in mind. The basic idea behind what we do is to replace
     *  combinations that look like:
     *	    Y:
     *		...
     *		leave
     *		jmp Z
     *  (where Z lies in a function other than Y)
     *  with a regular exit combo (leave/ret). This way we'll get more well
     *  behaved functions. However, if we do this the old target (Z) still
     *  needs to be executed, so we must instrument the caller so that it
     *  calls Z after the escaping function returns.  For example,
     *  a function call in X that looked like this:
     *	    X:
     *		...
     *		call Y
     *		...
     *  can be transformed in a version:
     *	    X:
     *		...
     *		call Y
     *		call Z
     *		...
     * What if, however, Y has a regular return as well as the escaping edge.
     * Consider the following version of Y:
     *	    Y:
     *		...
     *		leave
     *		ret
     *		...
     *		leave
     *		jmp Z
     *  In this case we can simply transform X to call Z upon return from Y
     *  since there is the possibility that Z wouldn't have been called in Y.
     *  In this case we will add a little more information to both the caller
     *  and callee so that X and Y look like this:
     *	    X:             Y:
     *		...            ...
     *		call Y         clc
     *		jae a          leave
     *		call Z         ret
     *	    a:	...            ...
     *		stc
     *		leave
     *		ret
     *  Here the clc instruction clears the carry flag of the PSW, and stc
     *  sets that flag.  We set the flag if we are returning from a point
     *  where we should have gone to Z, and clear it otherwise, and the
     *  caller checks to see if the flag is set upon return from Y - if it
     *  is set (jae is false) it calls Z, otherwise it does not and hence
     *  we preserve the behavior.
     *
     *  Now if Y actually has escaping edges to more than one function,
     *  things are complicated a little more. We have a solution to this
     *  as well, though.  By adding a little more  info we can take care
     *  of this case as well. Consider the following fix:
     *	    X:             Y:
     *		...            ...
     *		call Y         clc
     *		jae a          leave
     *		call *B        ret
     *	    a:	...            ...
     *		stc
     *		mov $Z, B
     *		leave
     *		ret
     *		...
     *		stc
     *		mov $G, B
     *		leave
     *		ret
     *  Now, when Y executes, it may have escaped to either G or Z or
     *  neither, but X will still execute the appropriate one. This has
     *  the unfortunate side effect of set the stack to unknown directly
     *  after our indirect call in X, however it would have been unknown
     *  anyway since Y would have to be termed as misbehaved
     *  due to it's escaping edge.
     *
     *  We're still not done considering things just yet. Consider now a
     *  function with escaping edges and hell edges into it's entry. Now we
     *  aren't sure who is calling us, but we still need to make sure that
     *  all callers will execute the proper function calls to previously
     *  escaping targets when necessary. Here we treat any function with
     *  hell edges into it's entry and an escaping edge in the same manner
     *  as we treated those with multiple exits (of multiple escaping edge
     *  targets), and we instrument all indirect call sites in the same
     *  manner as well. What if, however, the function is not calling a
     *  function with an escaping edge? In that case the condition flag may
     *  be set and then B might try to be indirectly called through but it
     *  shouldn't. So now we go ahead and instrument every function in the
     *  entire program with a hell edge into it's entry so that it clears
     *  the carry flag before exiting.
     *
     *  This should take care of all cases. For ease of implementation we just
     *  treat all escaping edges as if they were the most complicated case,
     *  therefore we instrument all indirect call sites, instrument all returns
     *  in hell targeted functions with clc's at call sites, and replace all
     *  escaping edges with the stc, mov target, B, leave, ret and instrument
     *  all call sites whose callee's are escaping edge funs.
     *
     *  Okay, there might actually be a problem with the stc and clc causing
     *  problems even though it's across function boundaries and shouldn't
     *  matter -- how about this:
     *	    X:             Y:
     *		...            ...
     *		push %eax
     *		call Y         mov (%esp), %eax
     *		mov %eax, B
     *		pop %eax
     *		call *B        ret
     *	    a:	...            ...
     *		...            mov $Z, B
     *		end of fun     leave
     *		add $6,B       ret
     *		jmp *B         ...
     *		mov $G, B
     *		leave
     *		ret
     * Let's try it out...
     */

    FOREACH_FUNCTION(fiterator) {
	FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_TEMP_FLAG);

	FOREACH_BBL_IN_FUN(fiterator, biterator) {
	    INSTR_REMOVE_FLAG(biterator->b_last, INSTR_TEMP_FLAG);
	}
    }

    call_slot = SECTION_REF_DATA->s_datasize + SECTION_REF_DATA->s_growth;
    IncreaseSectionSize(SECTION_REF_DATA, sizeof(ADDRESS) * 0x100);

    FOREACH_FUNCTION(fiterator) {
	UpdateIndirectCallers(fiterator, call_slot);

	// if (RemoveUnconditionalEscapingEdgesFromFun(fiterator, call_slot)) {
	if (cml_tmp(fiterator, call_slot)) {
	    UpdateEscapingFunCallers(fiterator, call_slot);

	    REPORT(RP_DETAIL, "Removed for %s", fiterator->f_name);
	}
    }
}



/*
 *  AddEntryEdge(bbl) --
 *
 *  Adds an entry edge from the specified basic block's function's entry
 *  block to the specified basic block iff there is not already an entry
 *  edge to that basic block.
 *
 *  bbl: the basic block that should receive an entry edge
 */
static void AddEntryEdge(Block *bbl) {
    if (BBLGetEdgeToBBL(bbl->b_fun->f_entry, bbl) == NULL) {
	BBLAddEdge(bbl->b_fun->f_entry, bbl, ET_ENTRY);
    }
}



/*
 *  InstrGetJT(instr) --
 *
 *  Returns the jump table that corresponds to the specified instruction,
 *  if there is one.
 *
 *  instr: the instruction that jumps through the jump table
 *  returns: the jump table associated with the specified instruction
 */
JumpTable *InstrGetJT(Instruction *instr) {
    JumpTable *jiterator;

    FOREACH_JT(jiterator) {
	if (ll_contains(jiterator->jt_users, instr)) {
	    return jiterator;
	}
    }

    return NULL;
}



/*
 *  FunHasEsc2(fun) --
 *
 *  Returns whether or not the specified function has an escaping edge.
 *
 *  fun: the function to check
 *  returns: true if the specified function has an escaping edge
 */
bool FunHasEsc2(Function *fun) {
    Block *biterator;

    FOREACH_BBL_IN_FUN(fun, biterator) {
	if (BBLHasEsc2(biterator)) {
	    return true;
	}
    }
    return false;
}



/*
 *  BBLHasEsc2(bbl) --
 *
 *  Returns whether or not the specified basic block has an escaping edge.
 *
 *  bbl: the basic block to check
 *  returns: true if the specified basic block has an escaping edge
 */
bool BBLHasEsc2(Block *bbl) {
    return BBLGetEscapingEdge(bbl) != NULL;
}



/*
 *  RemoveConditionalEscapingEdges() --
 *
 *  static void RemoveUnconditionalEscapingEdges();
 */
void RemoveConditionalEscapingEdges() {
    Function *fiterator;
    bool changed;

    // we do this so that we can tell if the inlined function's prologue
    // can be removed (we need to know if it has a prologue)
    MarkPrologueFuns();

    changed = true;
    while (changed) {
	changed = false;

	FOREACH_FUNCTION(fiterator) {
	    if (RemoveConditionalEscapingEdgeForFun(fiterator)) {
		REPORT(RP_DETAIL, "Inlined Escaping Edge for %s",
		    fiterator->f_name);
		changed = true;
	    }
	}
    }
}



/*
 *  RemoveConditionalEscapingEdgeForFun(fun) --
 *
 *  Inlines one escaping edge for the specified function if one exists and
 *  returns whether or not inlining was performed.
 *
 *  fun: the function to inline escaping edges in
 *  returns: true if any escaping edges in fun were inlined
 */
static bool RemoveConditionalEscapingEdgeForFun(Function *fun) {
    Function *target_fun, *fun_clone;
    EdgeNode *succn, *next_succ;
    Block *biterator, *last, *next, *target_bbl;
    Edge *edge, *esc_edge;
    Instruction *esc_instr;

    if (!FunHasEsc2(fun)) {
	return false;
    }
// REPORT(RP_LGDETAIL, "Attempting to remove escaping edges from %s",
// ((fun)->f_name));

    esc_edge = NULL;

    FOREACH_BBL_IN_FUN(fun, biterator) {
	edge = BBLGetEscapingEdge(biterator);

	if (edge != NULL) {
	    esc_instr = biterator->b_last;

//   REPORT(RP_LGDETAIL, "Escaping instruction: %s",
//   PrintInstruction(esc_instr));

	    // we only deal with conditional escaping edges in this function
	    if (!isJumpClass(esc_instr) || !isConditionalInstr(esc_instr)) {
		continue;
	    }

	    if (esc_instr->i_Bblock->b_type != BT_BRANCH) {
		continue;
	    }

	    if (BBLGetTrueEdge(esc_instr->i_Bblock) != edge) {
		continue;
	    }

	    esc_edge = edge;
	    break;
	}
    }

    // if we couldn't find a qualifying escaping edge we'll just quit now
    if (esc_edge == NULL) {
//  REPORT(RP_LGDETAIL, No qualifying escaping edge");
	return false;
    }

    target_bbl = esc_edge->e_to;
    target_fun = target_bbl->b_fun;

    if (target_fun->f_first == HELL) {
//  REPORT(RP_LGDETAIL, "target_func is HELL!");

	return false;
    }

    fun_clone = FunClone(target_fun);

    last = fun->f_last;

    // if we happen to have a case where there is a leave then jcc that
    // escapes to a function with a prologue, we can eliminate the leave
    // and prologue
    if (isLeaveClass(last->b_last)) {
	if (FUNCTION_HAS_FLAG(fun_clone, FUNCTION_GCC_PROLOGUE_FLAG)) {
	    InstrDisconnect(last->b_last);
	    InstrDisconnect(target_bbl->b_first);
	    InstrDisconnect(target_bbl->b_first);
	    InstrFree(last->b_last);
	    InstrFree(target_bbl->b_first);
	    InstrFree(target_bbl->b_first);
	}
    }

    BBLRemoveEdge(esc_edge->e_from, esc_edge->e_to);
    BBLAddEdge(esc_edge->e_from, BBLOriginalToClone(target_bbl, fun_clone),
	ET_TRUE);

    // we don't want any entry edges into our newly inlined blocks
    BBLRemoveEdges(fun_clone->f_entry);

    biterator = fun_clone->f_first;
    while (biterator != NULL) {
	next = biterator->b_next;

	InsertBBLAfterBBL(biterator, last);
	// move any exit edges here to the exit block of the new
	// containing function
	succn = biterator->b_succ;
	while (succn != NULL) {
	    next_succ = succn->en_next;

	    if (NODE_TYPE(succn) == ET_EXIT) {
		BBLRemoveEdge(NODE_FROM(succn), NODE_TO(succn));
		BBLAddEdge(biterator, fun->f_exit, ET_RETURN);
	    }

	    succn = next_succ;
	}
	biterator->b_fun = fun;
	last = biterator;

	fun->f_last = last;
	biterator = next;
    }

    free(fun_clone);

    return true;
}



/*
 *  ResolveEntrylessEntries() --
 */
static void ResolveEntrylessEntries() {
    Function *fiterator;
    Block *biterator;

    REPORT(RP_MINOR, WITHFUNC("Resolving entryless entry blocks"));

    FOREACH_FUNCTION(fiterator) {
	if (fiterator->f_entry->b_succ == NULL) {
	    FOREACH_BBL_IN_FUN(fiterator, biterator) {
		if (BBLContainsPrologue(biterator)) {
//        REPORT(RP_LGDETAIL, "Adding an entry edge to a prologue block in %s",
//		((fiterator)->f_name));
		    AddEntryEdge(biterator);
		}
	    }
	}
    }

    FOREACH_FUNCTION(fiterator) {
	if (fiterator->f_entry->b_succ == NULL) {
	    REPORT(RP_SMLOOP, "Adding entry edge to first BBL of %s",
		fiterator->f_name);
	    AddEntryEdge(fiterator->f_first);
	}
    }
}



/*
 *  GetJumpTableForInstr(instr) --
 *
 *  Given an instruction that uses the base address of a jump table, this
 *  function returns the jump table structure that represents that jump
 *  table.
 *
 *  instr: the instruction that uses the base address of a jump table
 *  returns: the jump table that the specified instruction uses or NULL if
 *  one can not be identified
 */
JumpTable *GetJumpTableForInstr(Instruction *instr) {
    JumpTable *jiterator;
    PatchSectionToInstr *patch;
    LinkedList *upms;
    Patch *upm;

    upms = UPMTableGetByDest(instr->i_src);
    if (ll_size(upms) == 0) {
	WARNING("No upms");
	return NULL;
    }

    upm = ll_removeFirst(upms);
    patch = (PatchSectionToInstr *) upm->p_patch;

    FOREACH_JT(jiterator) {
	if ((jiterator->jt_section == patch->p_src_section) &&
	    (jiterator->jt_base_offset == patch->p_src_offset)) {
	    return jiterator;
	}
    }

    return NULL;
}



/*
 *  AddJumpTable(section,offset,size) --
 */
JumpTable *AddJumpTable(Section *section, ADDRESS offset, int size) {
    JumpTable *jt;

    jt = zalloc(sizeof(JumpTable));
    jt->jt_users = ll_new(LIST_UNSORTED);
    jt->jt_next = NULL;
    jt->jt_next = all_jts;
    jt->jt_section = section;
    jt->jt_base_offset = offset;
    jt->jt_size = size;

    if (all_jts != NULL) {
	all_jts->jt_prev = jt;
    }

    all_jts = jt;

    return all_jts;
}



/*
 *  ReportInstrError(problem,instruction) --
 *  a systematic reporting routine for instructions.
 */
void ReportInstrError(char *problem, Instruction *instruction) {
    REPORT(RP_SMLOOP, "InstrError [0x%8.8x]: %s",
	instruction->i_address, problem);
}

//HACKED FUNCTION THERE SHOULD BE A BETTER WAY OF FIGURING THIS OUT
/* This function removes all the hell edges created 
 * by the fixup code for the kernel page fault handler
 */
static void RemoveHellsForFixup() {
    Function *fiterator = 0;
    Block *sblock       = 0;
    EdgeNode *succs = 0, *preds = 0;
    
    FOREACH_FUN(fiterator) {
        if (!strcmp(fiterator->f_name, "rwsem_wake")) {
            continue;
        }
	FOREACH_SUCC_IN_BBL(fiterator->f_entry, succs) {
            sblock = NODE_TO(succs);
            FOREACH_PRED_IN_BBL(sblock, preds) {
                // Why shud a function be nameless
                if (!preds->en_edge) {
                    continue;
                }
                if (!preds->en_edge->e_from) {
                    continue;
                }
  
                if (NODE_FROM(preds) == HELL) {
                    continue;
                }
                if (!preds->en_edge->e_from->b_fun) {
                    continue;
                }

                if (!preds->en_edge->e_from->b_fun->f_name) {
                    continue;
                }
              
                if (!strcmp(NODE_FROM(preds)->b_fun->f_name,
                        "rwsem_wake")) {
                    BBLRemoveEdge(fiterator->f_entry, sblock);
                    WARNING("REmoved entry edge for jump from"
                        " fixup code");
                }
            }
	}
    }
}
