/*
 *  Hashing functions for instructions.
 */
#include "hash.h"
#include "global.h"
#include "sysdep.h"

/*
 *  Hash Mask definitions. These denote the size of the hash tables. They
 *  should be increased if too much time is being spent in these routines.
 *  They should be decreased if too much memory is being used.
 */
#define INSTR_HASH_MASK 0x000fffff	// Instructions: Hash off lower 20 bits
#define RELOC_HASH_MASK 0x0000ffff	// Relocations:  Hash off lower 16 bits

/*
 *  The static declarations of the hash tables follow.
 */
Instruction *InstrHashTable[INSTR_HASH_MASK + 1];
Relocation *RelocHashTable[RELOC_HASH_MASK + 1];

/* Returns the slot in InstrHashTable[] where 'address' should appear */
static int InstrGetHashKey(UINT32 address);

/* Returns the index into RelocHashTable[] where 'address' hashes to */
static int RelocGetHashKey(UINT32 address);



/*
 *  InstrGetHashKey(instrAddress) --
 *
 *  A relatively quick hash that uses the last 16 bits of the address
 */
static int InstrGetHashKey(UINT32 instrAddress) {
    return instrAddress & INSTR_HASH_MASK;
}



/*
 *  InstrInsertEntry(entry) --
 *
 *  Sticks an instruction into the hash table based on its key
 */
void InstrInsertEntry(Instruction *entry) {
    Instruction *chain;
    int key;

    key = InstrGetHashKey(entry->i_address);
    chain = InstrHashTable[key];

    if (chain != NULL) {
	entry->i_hashNext = chain;
    }

    InstrHashTable[key] = entry;

    return;
}



/*
 *  InstrRemoveEntry(entry) --
 *
 *  Removes an entry from the hash table.
 */
void InstrRemoveEntry(Instruction *entry) {
    Instruction *chain;
    int key;

    key = InstrGetHashKey(entry->i_address);
    chain = InstrHashTable[key];
    if (chain == entry) {
	InstrHashTable[key] = entry->i_hashNext;
	return;
    }
    while ((chain->i_hashNext) && (chain->i_hashNext != entry)) {
	chain = chain->i_hashNext;
    }
    if (chain->i_hashNext == NULL) {
	WARNING("Couldn't remove instruction entry from hash table, %x: %s",
	    entry->i_address, PrintInstruction(entry));
    } else {
	chain->i_hashNext = chain->i_hashNext->i_hashNext;
    }

    return;
}



/*
 *  InstrLookupEntry(address) --
 *
 *  Scans through the bin associated with the last 16 bits of the given
 *  address and returns the instruction associated with the address, or
 *  NULL if none is found.
 */
Instruction *InstrLookupEntry(UINT32 address) {
    int key;
    Instruction *chain;

    key = InstrGetHashKey(address);
    chain = InstrHashTable[key];

    while (chain != NULL) {
	if (chain->i_address == address) {
	    return chain;
	}
	if (chain == chain->i_hashNext) {
	    return NULL;
	}
	chain = chain->i_hashNext;
    }

    return NULL;
}



/*
 *  InstrClearHashTable() -- removes all entries from the hash table.
 */
void InstrClearHashTable() {
    Instruction *chain, *instr;
    int i;

    for (i = 0; i <= INSTR_HASH_MASK; i++) {
	chain = InstrHashTable[i];
	while (chain != NULL) {
	    instr = chain;
	    chain = chain->i_hashNext;
	    instr->i_hashNext = NULL;
	}
	InstrHashTable[i] = NULL;
    }

    return;
}



/*
 *  InstrAddAllHashTable() -- adds all instructions into the hash table.
 */
void InstrAddAllHashTable() {
    Instruction *iterator;

    FOREACH_INSTR(iterator) {
	InstrInsertEntry(iterator);
    }
    return;
}



/*
 *  RelocGetHashKey(relocAddress) --
 *
 *  A relatively quick hash that uses the last 12 bits of the reloc
 */
static int RelocGetHashKey(UINT32 relocAddress) {
    return relocAddress & RELOC_HASH_MASK;
}



/*
 *  RelocInsertEntry(entry) --
 *
 *  Sticks an instruction into the hash table based on its key
 */
void RelocInsertEntry(Relocation *entry) {
    Relocation *chain;
    int key;

    key = RelocGetHashKey(ABSOLUTE_RELOCATION_ADDRESS(entry));
    chain = RelocHashTable[key];

    if (chain != NULL) {
	entry->r_hashNext = chain;
    }

    RelocHashTable[key] = entry;

    return;
}



/*
 *  RelocRemoveEntry(entry) -- removes an entry from the hash table.
 */
void RelocRemoveEntry(Relocation *entry) {
    Relocation *chain;
    int key;

    key = RelocGetHashKey(ABSOLUTE_RELOCATION_ADDRESS(entry));

    chain = RelocHashTable[key];
    if (chain == entry) {
	RelocHashTable[key] = entry->r_hashNext;
	return;
    }

    if (chain == NULL) {
	WARNING("Tried to remove non-existent relocation with offset: %x",
	    entry->r_offset);
	return;
    }

    while ((chain->r_hashNext != NULL) && (chain->r_hashNext != entry)) {
	chain = chain->r_hashNext;
    }

    if (chain->r_hashNext == NULL) {
	WARNING("Could not find relocation at %x to remove it",chain->r_offset);
	return;
    }

    chain->r_hashNext = chain->r_hashNext->r_hashNext;

    return;
}



/*
 *  RelocLookupEntry(address) --
 *
 *  Scans through the bin associated with the last 12 bits of the given
 *  address and returns the relocation associated with the address, or
 *  NULL if none is found.
 */
Relocation *RelocLookupEntry(UINT32 address) {
    int key;
    Relocation *chain;

    key = RelocGetHashKey(address);
    chain = RelocHashTable[key];

    while (chain != NULL) {
	if (ABSOLUTE_RELOCATION_ADDRESS(chain) == address) {
	    return chain;
	}

	chain = chain->r_hashNext;
    }

    return NULL;
}



/*
 *  ReportInstrDensityInHashTable() --
 *
 *  Displays information about the instruction hash table.
 */
void ReportInstrDensityInHashTable() {
    Instruction *instr;
    int i;
    int count = 0;
    int total = 0;
    int density[1025];

    for (i = 0; i < 1024; i++) {
	density[i] = 0;
    }

    for (i = 0; i < INSTR_HASH_MASK; i++) {
	instr = InstrHashTable[i];
	while (instr != NULL) {
	    count++;
	    total++;
	    instr = instr->i_hashNext;
	}
	if (count > 1024) {
	    count = 1024;
	}
	density[count] = density[count] + 1;
    }

    for (i = 0; i < ARRAYSIZE(density); i++) {
	REPORT(RP_LGLOOP, 
	    "[ReportInstrDensityInHashTable] %d slots had %d instructions",
	    density[i], i);
    }
}



/*
 *  BuildRelocationHashTable() --
 */
void BuildRelocationHashTable() {
    Relocation *rel;

    FOREACH_RELOCATION(rel) {
	RelocInsertEntry(rel);
    }
}
