/*
 *  A module for adjusting links on bbls, instructions and functions.
 */
#include "global.h"
#include "flowgraph.h"
#include "sysdep.h"
#include "utilities.h"
#include "opt.unreachable.h"
#include "upm.h"



/*
 *  InstrRelinkBasedOnBBL() --
 *
 *  Adjusts the linking of instructions based on the basic block linkage.
 *  Only the first and last instructions in a basic block will be updated,
 *  as this function assumes the ordering of instructions inside a basic
 *  block has not changed.
 */
void InstrRelinkBasedOnBBL() {
    Instruction *iterator;
    Block *biterator;

    /* Unlink the first and last instructions in all basic blocks */
    FOREACH_BBL(biterator) {
	if (biterator->b_first == NULL) {
	    continue;
	}
	biterator->b_last->i_next = NULL;
	biterator->b_first->i_prev = NULL;
    }

    /* Relink the first and last instructions in all basic blocks */
    FOREACH_BBL(biterator) {
	iterator = biterator->b_first;
	if (iterator == NULL) {
	    CASSERT(biterator == HELL,
		"Non-Hell block with no first instruction.");
	    continue;
	}
	if (biterator->b_prev == NULL) {
	    iterator->i_prev = NULL;
	} else {
	    iterator->i_prev = biterator->b_prev->b_last;
	}
	iterator = biterator->b_last;
	if (biterator->b_next == NULL) {
	    iterator->i_next = NULL;
	} else {
	    iterator->i_next = biterator->b_next->b_first;
	}
    }

    return;
}



/*
 *  BBLRelinkBasedOnFun() --
 *
 *  Adjusts the linking of basic blocks based on the function linkage.
 *  Only the first and last blocks in a function will be updated,
 *  as this function assumes the ordering of blocks inside a function
 *  block has not changed.
 */
void BBLRelinkBasedOnFun() {
    Block *biterator;
    Function *fiterator;
    Block *temp;

    /* Unlink the first and last bbls in all funs */
    FOREACH_FUNCTION(fiterator) {
	if (fiterator->f_first == NULL) {
	    continue;
	}
	temp = fiterator->f_last;
	while ((temp->b_next) && (temp->b_next->b_fun == fiterator)) {
	    temp = temp->b_next;
	}
	temp->b_next = NULL;
	fiterator->f_first->b_prev = NULL;
    }

    /* Relink the first and last blocks in all functions */
    FOREACH_FUNCTION(fiterator) {
	biterator = fiterator->f_first;
	if (biterator == NULL) {
	    CASSERT(fiterator == HELL->b_fun,
		"Non-Hell function with no first basic block.");
	    continue;
	}
	if (fiterator->f_prev == NULL) {
	    biterator->b_prev = NULL;
	} else {
	    temp = fiterator->f_prev->f_last;
	    while ((temp->b_next)
		&& (temp->b_next->b_fun == fiterator->f_prev))
		temp = temp->b_next;
	    biterator->b_prev = temp;
	}
	biterator = fiterator->f_last;
	while ((biterator->b_next)
	    && (biterator->b_next->b_fun == fiterator))
	    biterator = biterator->b_next;
	if (fiterator->f_next == NULL) {
	    biterator->b_next = NULL;
	} else {
	    biterator->b_next = fiterator->f_next->f_first;
	}
    }

    i_list = bbl_list->b_first;

    return;
}



/*
 *  RemoveSymbol(sym) --
 */
void RemoveSymbol(Symbol *sym) {
    if (sym->sy_prev) {
	sym->sy_prev->sy_next = sym->sy_next;
    }
    if (sym->sy_next) {
	sym->sy_next->sy_prev = sym->sy_prev;
    }
    if (sym == s_list) {
	s_list = sym->sy_next;
    }
}



/*
 *  RemoveFunSymbols(fun) --
 *
 *  Removes all symbols that lie within a function from the symbol list.
 */
static void RemoveFunSymbols(Function *fun) {
    Symbol *sym;
    ADDRESS start, end;

    if (fun->f_first == NULL || FUNCTION_FIRST_INSTR(fun) == NULL) {
	return;
    }

    start = FUNCTION_FIRST_INSTR(fun)->i_address;
    end = FUNCTION_LAST_INSTR(fun)->i_address +
	FUNCTION_LAST_INSTR(fun)->i_length;
    FOREACH_SYMBOL(sym) {
	if (sym->sy_section != SECTION_REF_CODE) {
	    continue;
	}
	if ((sym->sy_value >= start) && (sym->sy_value < end)) {
	    RemoveSymbol(sym);
	}
    }
}



/*
 *  RemoveBBLSymbols(bbl) --
 *
 *  Remove all symbol that lie with a Basic Block from the symbol list.
 */
static void RemoveBBLSymbols(Block *bbl) {
    Symbol *sym;
    ADDRESS start, end;

    if (bbl->b_first == NULL) {
	return;
    }
    start = bbl->b_first->i_address;
    end = bbl->b_last->i_address + bbl->b_last->i_length;
    FOREACH_SYMBOL(sym) {
	if (sym->sy_section != SECTION_REF_CODE) {
	    continue;
	}
	if ((sym->sy_value >= start) && (sym->sy_value < end)) {
	    RemoveSymbol(sym);
	}
    }
}



/*
 *  RemoveInstrSymbols(instr) --
 *
 *  Removes all symbols that lie within an instruction from the symbol list.
 */
static void RemoveInstrSymbols(Instruction *instr) {
    Symbol *sym;
    ADDRESS start, end;

    start = instr->i_address;
    end = instr->i_address + instr->i_length;
    FOREACH_SYMBOL(sym) {
	if (sym->sy_section != SECTION_REF_CODE) {
	    continue;
	}
	if ((sym->sy_value >= start) && (sym->sy_value < end)) {
	    RemoveSymbol(sym);
	}
    }
}



/*
 *  FunDisconnect(fun) --
 *
 *  Disconnects 'fun' from the control flow graph by adjusting links.
 */
void FunDisconnect(Function *fun) {
    if (fun->f_prev) {
	fun->f_prev->f_next = fun->f_next;
    }
    if (fun->f_next) {
	fun->f_next->f_prev = fun->f_prev;
    }
    if (f_list == fun) {
	f_list = fun->f_next;
    }
    RemoveFunSymbols(fun);
}



/*
 *  BBLDisconnect(block) --
 *
 *  Disconnects a basic block from the graph by adjusting links.
 *  This function does not alter predecessors or successors however.
 */
void BBLDisconnect(Block *block) {

    if (block->b_prev) {
	block->b_prev->b_next = block->b_next;
    }
    if (block->b_next) {
	block->b_next->b_prev = block->b_prev;
    }
    if (bbl_list == block) {
	bbl_list = block->b_next;
    }

    if (block->b_first != NULL) {
	if (block->b_first->i_prev != NULL) {
	    block->b_first->i_prev->i_next = block->b_last->i_next;
	}
    }

    if (block->b_last != NULL) {
	if (block->b_last->i_next != NULL) {
	    block->b_last->i_next->i_prev = block->b_first->i_prev;
	}
    }

    if (block->b_fun == NULL) {
	return;
    }

    if (block->b_fun->f_first == block) {
	if (block->b_fun->f_last == block) {
	    block->b_fun->f_last = NULL;
	    block->b_fun->f_first = NULL;
	} else {
	    block->b_fun->f_first = block->b_next;
	}
    } else if (block->b_fun->f_last == block) {
	if (block->b_fun->f_first == block) {
	    block->b_fun->f_last = NULL;
	    block->b_fun->f_first = NULL;
	} else {
	    block->b_fun->f_last = block->b_prev;
	}
    }
    RemoveBBLSymbols(block);

    return;
}



/*
 *  InstrDisconnect(instr) --
 *
 *  Disconnects an instruction from the graph by adjusting links.
 *  Basic blocks are updated if this instruction was the first or
 *  last in the basic block it belonged to. If the instruction
 *  is both the first and last in the basic block, then its removal
 *  will make the basic block empty.  This function will not
 *  disconnect the basic block to reflect this.
 */
Instruction *InstrDisconnect(Instruction *instr) {
    Block *ibbl;

    if (instr->i_prev) {
	instr->i_prev->i_next = instr->i_next;
    }
    if (instr->i_next) {
	instr->i_next->i_prev = instr->i_prev;
    }
    if (i_list == instr) {
	i_list = instr->i_next;
    }

    ibbl = instr->i_Bblock;
    if (ibbl) {
	if ((INSTR_HAS_FLAG(instr, BBL_END))
		&& instr->i_prev && (instr->i_prev->i_Bblock == ibbl)) {
	    INSTR_REMOVE_FLAG(instr, BBL_END);
	    INSTR_SET_FLAG(instr->i_prev, BBL_END);
	}

	if ((INSTR_HAS_FLAG(instr, BBL_START))
		&& instr->i_next && (instr->i_next->i_Bblock == ibbl)) {
	    INSTR_REMOVE_FLAG(instr, BBL_START);
	    INSTR_SET_FLAG(instr->i_next, BBL_START);
	}
    } else {				/* instr->i_Bblock == NULL */
	return NULL;
    }

    if (ibbl->b_first == instr) {
	if (ibbl->b_last == instr) {
	    ibbl->b_last = NULL;
	    ibbl->b_first = NULL;
	} else {
	    ibbl->b_first = instr->i_next;
	}
    } else if (ibbl->b_last == instr) {
	if (ibbl->b_first == instr) {
	    ibbl->b_last = NULL;
	    ibbl->b_first = NULL;
	} else {
	    ibbl->b_last = instr->i_prev;
	}
    }

    RemoveInstrSymbols(instr);

    return instr;
}



/*
 *  InsertInstrBeforeInstr(newinstr,instr) --
 *
 *  Inserts the given instruction into the control flow graph before the
 *  specified instruction. The new instruction will go in the same basicblock
 *  and function as the instruction to insert before.
 */
void InsertInstrBeforeInstr(Instruction *newinstr, Instruction *instr) {

    CASSERT(newinstr != NULL, "Passed a NULL new instruction.");
    CASSERT(instr != NULL, "Passed a NULL instruction.");

    newinstr->i_Bblock = instr->i_Bblock;
    newinstr->i_next = instr;
    newinstr->i_prev = instr->i_prev;
    if (newinstr->i_prev) {
	newinstr->i_prev->i_next = newinstr;
    }
    instr->i_prev = newinstr;

    if (i_list == instr) {
	i_list = newinstr;
    }

    if (startins == instr) {
	startins = newinstr;
    }

    if (INSTR_HAS_FLAG(instr, BBL_START)) {
	INSTR_REMOVE_FLAG(instr, BBL_START);
	INSTR_SET_FLAG(newinstr, BBL_START);
    }

    if (INSTR_HAS_FLAG(instr, FUN_START)) {
	INSTR_REMOVE_FLAG(instr, FUN_START);
	INSTR_SET_FLAG(newinstr, FUN_START);
    }

    if (instr->i_Bblock != NULL) {
	if (instr->i_Bblock->b_first == instr) {
	    instr->i_Bblock->b_first = newinstr;
	}
    }

    return;
}



/*
 *  BBLPrependInstr(ibbl,iins) --
 *
 *  Puts the given instruction at the beginning of a basic block,
 *  if the block is currently empty, this will set the appropriate
 *  flags on the instruction, and set the first and last pointers for
 *  the basic block to point at this instruction
 */
void BBLPrependInstr(Block *ibbl, Instruction *iins) {
    if (ibbl->b_first == NULL) {
        Block *jbbl;
        Instruction *jins;

	ibbl->b_first = ibbl->b_last = iins;
        iins->i_Bblock = ibbl;
        INSTR_SET_FLAG(iins, (BBL_START | BBL_END));
        /*
         * add iins to the list of instructions: find the next
         * basic block jbbl that contains an instruction, and 
         * add iins just before the first instruction of jbbl.
         */
        for (jbbl = ibbl->b_next;
            jbbl != NULL && jbbl->b_first == NULL;
            jbbl = jbbl->b_next) {
        }

        if (jbbl != NULL) {
            jins = jbbl->b_first;
            iins->i_next = jins;
            iins->i_prev = jins->i_prev;
            if (jins->i_prev) {
                jins->i_prev->i_next = iins;
            }
            jins->i_prev = iins;
        } else {
            WARNING("Instruction not linked into instruction list");
        }

    } else {
	InsertInstrBeforeInstr(iins, ibbl->b_first);
    }

}



/*
 *  InsertInstrAfterInstr(newinstr,instr) --
 *
 *  Inserts the given instruction into the control flow graph after the
 *  specified instruction. The new instruction will go in the same basicblock
 *  and function as the instruction to insert after.
 */
void InsertInstrAfterInstr(Instruction *newinstr, Instruction *instr) {

    CASSERT(newinstr != NULL, "Passed a NULL new instruction.");
    CASSERT(instr != NULL, "Passed a NULL instruction.");

    newinstr->i_Bblock = instr->i_Bblock;
    newinstr->i_prev = instr;
    newinstr->i_next = instr->i_next;
    if (instr->i_next) {
	instr->i_next->i_prev = newinstr;
    }
    instr->i_next = newinstr;

    if (INSTR_HAS_FLAG(instr, BBL_END)) {
	INSTR_REMOVE_FLAG(instr, BBL_END);
	INSTR_SET_FLAG(newinstr, BBL_END);
    }

    if (INSTR_HAS_FLAG(instr, FUN_END)) {
	INSTR_REMOVE_FLAG(instr, FUN_END);
	INSTR_SET_FLAG(newinstr, FUN_END);
    }

    if ((instr->i_Bblock) && (instr->i_Bblock->b_last == instr)) {
	instr->i_Bblock->b_last = newinstr;
    }

    return;
}



/*
 *  InsertBBLAfterBBL(newbbl,block) --
 *
 *  Inserts 'newbbl' after 'block' and updates the instructions in each basic
 *  block to reflect the correct linkage.
 */
void InsertBBLAfterBBL(Block *newbbl, Block *block) {
    Instruction *first, *last;

    CASSERT(newbbl != NULL, "Passed a NULL new block.");
    CASSERT(block != NULL, "Passed a NULL block.");

    newbbl->b_fun = block->b_fun;
    newbbl->b_prev = block;
    newbbl->b_next = block->b_next;
    if (block->b_next) {
	block->b_next->b_prev = newbbl;
    }
    block->b_next = newbbl;

    first = newbbl->b_first;
    last = block->b_last;

    if ((first != NULL) && (last != NULL)) {
	last->i_next = first;
	first->i_prev = last;
    }

    if (block->b_fun->f_last == block) {
	block->b_fun->f_last = newbbl;
    }

    if (newbbl->b_next == NULL) {
	if (newbbl->b_last != NULL) {
	    newbbl->b_last->i_next = NULL;
	}
	return;
    }

    first = newbbl->b_next->b_first;
    last = newbbl->b_last;

    if ((first != NULL) && (last != NULL)) {
	last->i_next = first;
	first->i_prev = last;
    }

    return;
}



/*
 *  InsertBBLBeforeBBL(newbbl,block) --
 *
 *  Adjusts links on the basic blocks so that 'newbbl' falls in the
 *  chain immediately before 'block'. Instruction links are also
 *  updated to reflect the change.  Note that functions first/last
 *  block pointers are not changed here.
 */
void InsertBBLBeforeBBL(Block *newbbl, Block *block) {
    Instruction *first, *last;

    if (bbl_list == block) {
	bbl_list = newbbl;
	if (newbbl->b_first != NULL) {
	    i_list = newbbl->b_first;
	}
    }

    CASSERT(newbbl != NULL, "Passed a NULL new block.");
    CASSERT(block != NULL, "Passed a NULL block.");

    newbbl->b_fun = block->b_fun;
    newbbl->b_prev = block->b_prev;
    newbbl->b_next = block;

    if (block->b_prev) {
	block->b_prev->b_next = newbbl;
    }
    block->b_prev = newbbl;

    last = newbbl->b_last;
    first = block->b_first;

    if ((first != NULL) && (last != NULL)) {
	last->i_next = first;
	first->i_prev = last;
    }

    if (newbbl->b_prev == NULL) {
	if (newbbl->b_first != NULL) {
	    newbbl->b_first->i_prev = NULL;
	}
	return;
    }

    first = newbbl->b_first;
    last = newbbl->b_prev->b_last;

    if ((first != NULL) && (last != NULL)) {
	last->i_next = first;
	first->i_prev = last;
    }

    return;
}



/*
 *  RelocationDisconnect(reloc) --
 *
 *  Disconnects the relocation 'reloc' from the list of relocations.
 */
void RelocationDisconnect(Relocation *reloc) {

    if (reloc == r_list) {
	r_list = reloc->r_next;
    }
    if (reloc->r_next) {
	reloc->r_next->r_prev = reloc->r_prev;
    }
    if (reloc->r_prev) {
	reloc->r_prev->r_next = reloc->r_next;
    }

    return;
}



/*
 *  InsertSectionAfterSection(new_section,section) --
 */
void InsertSectionAfterSection(Section *new_section, Section *section) {

    CASSERT(new_section != NULL, "Cannot insert a NULL section");
    CASSERT(section != NULL, "Cannot insert after a NULL section");

    new_section->s_prev = section;
    new_section->s_next = section->s_next;
    section->s_next = new_section;

    if (new_section->s_next != NULL) {
	new_section->s_next->s_prev = section;
    }
}



/*
 *  InsertSymbolBeforeSymbol(new_symbol,symbol) --
 */
void InsertSymbolBeforeSymbol(Symbol *new_symbol, Symbol *symbol) {

    CASSERT(new_symbol != NULL, "Cannot insert a NULL symbol");
    CASSERT(symbol != NULL, "Cannot insert before a NULL symbol");

    new_symbol->sy_next = symbol;
    new_symbol->sy_prev = symbol->sy_prev;

    if (symbol->sy_prev != NULL) {
	symbol->sy_prev->sy_next = new_symbol;
    }
    symbol->sy_prev = new_symbol;

    if (s_list == symbol) {
	s_list = new_symbol;
    }
}



/*
 *  InsertSymbolAfterSymbol(new_symbol,symbol) --
 */
void InsertSymbolAfterSymbol(Symbol *new_symbol, Symbol *symbol) {

    CASSERT(new_symbol != NULL, "Cannot insert a NULL symbol");
    CASSERT(symbol != NULL, "Cannot insert after a NULL symbol");

    new_symbol->sy_prev = symbol;
    new_symbol->sy_next = symbol->sy_next;
    symbol->sy_next = new_symbol;

    if (new_symbol->sy_next != NULL) {
	new_symbol->sy_next->sy_prev = new_symbol;
    }
}



/*
 *  InsertRelocationBeforeRelocation(new_relocation,relocation) --
 */
void InsertRelocationBeforeRelocation(Relocation *new_relocation,
    Relocation *relocation) {

    CASSERT(new_relocation != NULL, "Cannot insert a NULL relocation");
    CASSERT(relocation != NULL, "Cannot insert before a NULL relocation");

    new_relocation->r_next = relocation;
    new_relocation->r_prev = relocation->r_prev;
    if (relocation->r_prev != NULL) {
	relocation->r_prev->r_next = new_relocation;
    }
    relocation->r_prev = new_relocation;
}



/*
 *  InsertRelocationAfterRelocation(new_relocation,relocation) --
 */
void InsertRelocationAfterRelocation(Relocation *new_relocation,
    Relocation *relocation) {

    CASSERT(new_relocation != NULL, "Cannot insert a NULL relocation");
    CASSERT(relocation != NULL, "Cannot insert after a NULL relocation");

    new_relocation->r_prev = relocation;
    new_relocation->r_next = relocation->r_next;
    relocation->r_next = new_relocation;

    if (new_relocation->r_next != NULL) {
	new_relocation->r_next->r_prev = relocation;
    }
}



/*
 *  InsertSectionLast(new_section,list) --
 */
void InsertSectionLast(Section *new_section, Section *list) {
    Section *siterator;

    FOREACH_SECTION_IN_LIST(list, siterator) {
	if (siterator->s_next == NULL) {
	    break;
	}
    }

    InsertSectionAfterSection(new_section, siterator);
}



/*
 *  InsertSectionAfterSectionNamed(section,name,list) --
 */
void InsertSectionAfterSectionNamed(Section *section,char *name,Section *list){
    Section *siterator;

    FOREACH_SECTION_IN_LIST(list, siterator) {
	if (strcmp((siterator->s_name), name) == 0) {
	    break;
	}
    }

    InsertSectionAfterSection(section, siterator);
}



/*
 *  RemoveSectionFromList(to_remove,list) --
 */
void RemoveSectionFromList(Section *to_remove, Section **list) {
    if (*list == to_remove) {
	*list = to_remove->s_next;
	(*list)->s_prev = NULL;

	return;
    }

    if (to_remove->s_next != NULL) {
	to_remove->s_next->s_prev = to_remove->s_prev;
    }

    if (to_remove->s_prev != NULL) {
	to_remove->s_prev->s_next = to_remove->s_next;
    }
}



/*
 *  RemoveDroppedSectionsFromList(list) --
 */
Section *RemoveDroppedSectionsFromList(Section *list) {
    Section **sections, *seciterator;
    int num_sections, i;

    num_sections = 0;
    FOREACH_SECTION_IN_LIST(list, seciterator) {
	num_sections++;
    }

    sections = zalloc(sizeof(Section *) * num_sections);
    i = 0;
    FOREACH_SECTION_IN_LIST(list, seciterator) {
	sections[i++] = seciterator;
    }

    list = NULL;
    for (i = 0; i < num_sections; i++) {
	if (SECTION_HAS_FLAG(sections[i], SECTION_DROPPED_FLAG)) {
	    REPORT(RP_SMLOOP, "Removing dropped section: %s",
		sections[i]->s_name);
	    continue;
	}

	if (list == NULL) {
	    list = seciterator = sections[i];
	    seciterator->s_next = seciterator->s_prev = NULL;
	} else {
	    InsertSectionAfterSection(sections[i], seciterator);
	    seciterator = sections[i];
	}
    }

    free(sections);

    return list;
}



/*
 *  InsertFunctionAfterFunction(new_fun,fun) --
 */
void InsertFunctionAfterFunction(Function *new_fun, Function *fun) {

    CASSERT(new_fun != NULL, "Cannot insert a NULL function");
    CASSERT(fun != NULL, "Cannot insert after a NULL function");

    new_fun->f_prev = fun;
    new_fun->f_next = fun->f_next;
    if (fun->f_next != NULL) {
	fun->f_next->f_prev = new_fun;
    }
    fun->f_next = new_fun;
}
