/*
 *  liveness.c
 *
 *  A module for liveness computations among registers.  Liveness information
 *  for memory is currently not implemented. This file is not very machine
 *  dependent, as the representation for a set of registers depends on the
 *  architecture at hand. The analysis is context-insensitive.
 *
 *  Our goal is to iteratively compute the fixpoint of the following dataflow
 *  equations:
 *
 *  LiveIn[n] = use[n] + (LiveOut[n] -- def[n]) where n is a BBL
 *  LiveOut[n] = LiveIn[s] for all successors, s, of n where n is a BBL
 *
 *  Subject to the initial conditions:
 *  LiveIn[n] = LiveOut[n] = EMPTY SET
 *
 *  (Register Liveness Analysis of Executable Code, Muth)
 */
#include "global.h"
#include "utilities.h"
#include "liveness.h"
#include "sysdep.h"
#include "linkage.h"
#include "cp.h"
#include "upm.h"
#include "stack.analyze.h"
#include "flowgraph.h"
#include "opt.unreachable.h"
#include "opt.h"
#include "x86-liveness.h"
#include "x86-table-defs.h"
#include "x86-asm.h"
#include "liveness.cs.h"
#include "liveness.ci.h"
#include "x86-classes.h"

#define MERGE_TRASH_DEPTH(a, b) \
    if ((a == -1) || (b == -1)) { \
	a = -1; \
    } else if (a < b) { \
	a = b; \
    }

#define MERGE_USE_DEPTH(a, b) \
    MERGE_TRASH_DEPTH(a, b)

/*
 *  FUNCTIONS FOR REGISTER LIVENESS
 */
/* Returns true if the bitvector is live at the instruction */
extern bool RegBitvectorIsLiveAtInstr(REGISTER_BITVECTOR vec,
    Instruction *instr);

/*
 *  Returns a bitvector with a bit set for each register that this function
 *  may define when called.  Note that this includes any registers that can
 *  be defined by instructions reachable from this function--such as other
 *  functions it may call.
 */
static REGISTER_BITVECTOR FunComputeDefinedRegistersLiberal(Function *fun);

/*
 *  FUNCTIONS FOR PSW LIVENESS
 */
/* Returns true if the bitvector is live at the instruction */
extern bool PSWBitvectorIsLiveAtInstr(PSW_BITVECTOR vec, Instruction *instr);

/*
 *  FUNCTIONS FOR STACK LIVENESS
 */
/* Returns true if the stack defined[] is live at 'instr' */
extern bool StackIsLiveAtInstr(int defined[], Instruction *instr);

/*
 *  FUNCTIONS FOR TRASH/USE DEPTH
 */
/* Computes the depth of the stack frame which 'fun' may overwrite */
static void ComputeFunctionTrashDepth(Function *fun);

/* Computes the depth that the operand may overwrite */
static int ComputeOperandTrashDepth(Operand *op);

/* Computes the depth of the stack frame which 'fun' may access */
static void ComputeFunctionUseDepth(Function *fun);

/* Computes the depth that the operand may access */
static int ComputeOperandUseDepth(Operand *op);

/* Returns true if the instruction defines a location that is live */
extern bool InstrDefinesLiveLocations(Instruction *instr);



/*
 *  LIVENESS ANALYSIS PARAMETERS
 */
static int LIVENESS_ANALYSIS_REGISTER;
static int LIVENESS_ANALYSIS_PSW;
static int LIVENESS_ANALYSIS_STACK;



/*
 *  LivenessParam(char *name) -- translates a parameter value to numeric form.
 */
static int LivenessParam(char *name) {
    switch (StringParameter(name)[0]) {
    	case 'O':  return LIVENESS_NONE;
	case 'I':  return LIVENESS_INSENSITIVE;
	case 'S':  return LIVENESS_SENSITIVE;
	default:   MALFUNCTION("unrecognized value for %s", name);  return 0;
    }
}



/*
 *  InitializeLivenessParameters() -- sets values used in subsequent analysis.
 */
void InitializeLivenessParameters() {
    LIVENESS_ANALYSIS_REGISTER = LivenessParam("RegisterLiveness");
    LIVENESS_ANALYSIS_PSW = LivenessParam("PSWLiveness");
    LIVENESS_ANALYSIS_STACK = LivenessParam("StackLiveness");
    
}



/*
 *  ComputeRegisterLiveness(full) --
 *
 *  Computes register liveness information and stores the results of the
 *  computation in ->b_liveIn and ->b_liveOut
 *  for each basic block.
 */
void ComputeRegisterLiveness(int full) {
    Block *bbl;

    if (LIVENESS_ANALYSIS_REGISTER == LIVENESS_INSENSITIVE) {
	CIComputeRegisterLiveness(full);
    } else if (LIVENESS_ANALYSIS_REGISTER == LIVENESS_SENSITIVE) {
	CSComputeRegisterLiveness(full);
    } else if (LIVENESS_ANALYSIS_REGISTER == LIVENESS_NONE) {
	FOREACH_BBL(bbl) {
	    bbl->b_liveIn = ALL_REGISTERS;
	    bbl->b_liveOut = ALL_REGISTERS;
	}
    } else {
	MALFUNCTION("Unknown register liveness analysis set.");
    }

    return;
}



/*
 *  ComputePSWLiveness(full) --
 *
 *  Computes PSW liveness information and stores the results of the
 *  computation in ->b_liveInPSW and ->b_liveOutPSW
 *  for each basic block.
 */
void ComputePSWLiveness(int full) {
    Block *bbl;

    if (LIVENESS_ANALYSIS_PSW == LIVENESS_INSENSITIVE) {
	CIComputePSWLiveness(full);
    } else if (LIVENESS_ANALYSIS_PSW == LIVENESS_SENSITIVE) {
	CSComputePSWLiveness(full);
    } else if (LIVENESS_ANALYSIS_PSW == LIVENESS_NONE) {
	FOREACH_BBL(bbl) {
	    bbl->b_liveInPSW = ALL_PSW;
	    bbl->b_liveOutPSW = ALL_PSW;
	}
    } else {
	MALFUNCTION("Unknown PSW liveness analysis set.");
    }

    return;
}



/*
 *  ComputeStackLiveness(full) --
 *
 *  Computes stack liveness information and stores the results of the
 *  computation in the stack frame model for each basic block.
 */
void ComputeStackLiveness(int full) {
    Block *bbl;
    Function *fun;
    int i;

    MarkFunsDefiningFramePointer();

    if (LIVENESS_ANALYSIS_STACK == LIVENESS_INSENSITIVE) {
	CIComputeStackLiveness(full);
    } else if (LIVENESS_ANALYSIS_STACK == LIVENESS_SENSITIVE) {
	CSComputeStackLiveness(full);
    } else if (LIVENESS_ANALYSIS_STACK == LIVENESS_NONE) {
	AllocateSnapshots();
	FOREACH_FUNCTION(fun) {
	    int sz =
		(fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE;
	    FOREACH_BBL_IN_FUN(fun, bbl) {
		if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
		    for (i = 0; i <= sz; i++) {
			bbl->b_stackshot[i].v_live = true;
		    }
		}
	    }
	}
    } else {
	MALFUNCTION("Unknown stack liveness analysis set.");
    }

    return;
}



/*
 *  TestRegisterLiveness() --
 *
 *  This function should only be called when testing the correctness of the
 *  register liveness analysis. It looks for dead registers, and inserts
 *  an instruction that will load them with an immediate value (-1).
 */
void TestRegisterLiveness() {
    Block *bbl;
    Instruction *instr;
    int i, reg, temp;

    temp = 0;
    FOREACH_BBL(bbl) {
	for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {

	    if (bbl->b_liveIn & (1 << i)) {
		continue;
	    }

	    reg = BitRegisterToRegister(i);
	    if (RegisterIsFramePointer(reg) || RegisterIsStackPointer(reg)) {
		continue;
	    }

	    temp++;

	    reg = BitRegisterToRegister(i);
	    CASSERT(reg != -1, "No register found for bit.");
	    instr = InstrCreateMovImmToReg(reg, -1);
	    InsertInstrBeforeInstr(instr, bbl->b_first);
	    bbl->b_first = instr;
	}
    }

    REPORT(RP_DETAIL, "%d registers dead at start of bbls [%s]",
	temp, __func__);

    return;
}



/*
 *  RemoveDeadCode() -- removes dead code from the control flow graph.
 */
void RemoveDeadCode() {
    Instruction *instr, *compensate, *nextInstr;
    Function *fun, *nextFun;
    Block *bbl, *nextBBL;
    bool change;
    char *text;
    int numRegsChanged, numStackChanged, totalChanged;
    int round, pswInstr, stackChange, spAdds;
    long long instrCountOrig = CountInstructionsExecuted();
    long long originalMemOps = CountMemoryOperations();
    long long instrCountNew;
    long long pswWeight;
    long long newMemOps;
    PSW_BITVECTOR compensatePSW = ComputePSWDefined(InstrCreateSPInc());

    int i = 0;
    
    REPORT(RP_MINOR, WITHFUNC("Removing dead code"));

    numRegsChanged = pswWeight = numStackChanged = 0;
    totalChanged = round =pswInstr = spAdds = 0;
    change = true;
    
    while (change) {

	round++;
	numRegsChanged = numStackChanged = spAdds = 0;
	change = false;

	fun = nextFun = f_list;
        i = 0;
	while ((fun != NULL) && (nextFun != NULL)) {
	    fun = nextFun;
	    nextFun = fun->f_next;
            i++;

	    bbl = nextBBL = fun->f_first;
	    while ((bbl != NULL) && (nextBBL != NULL)
		    && (nextBBL->b_fun == bbl->b_fun)) {
		bbl = nextBBL;
		nextBBL = bbl->b_next;
		BBL_REMOVE_FLAG(bbl, BBL_TEMP_FLAG);

		instr = nextInstr = bbl->b_first;
		while ((instr != NULL) && (nextInstr != NULL) &&
                    (nextInstr->i_Bblock == instr->i_Bblock)) {
		    instr = nextInstr;
		    nextInstr = instr->i_next;

		    if (isPCChangingClass(instr)) {
			continue;
		    }
		    if (isFPClass(instr)) {
			continue;
		    }

                    if (isDataInstruction(instr)) {
                        continue;
                    }

		    /*
		     * We shouldn't get rid of instructions that assign to the
		     * frame or stack pointer. Interrupts and Traps will
		     * probably use calling conventions, so having ESP/EBP
		     * set to bogus values may produce incorrect binaries.
		     */

		    if (instr->i_dest) {
			if (OperandIsStackPointer(instr->i_dest)) {
			    continue;
			}
			if (OperandIsFramePointer(instr->i_dest)) {
			    continue;
			}
		    }

		    /*
		     * See if the location is live at this instruction.
		     * If so, we can't remove it.
		     */
		    if (InstrDefinesLiveLocations(instr)) {
			continue;
		    }

		    /*
		     * We can't remove the instruction if it defines a PSW bit
		     * that is going to be used before being re-defined.
		     */
		    if (PSWBitvectorIsLiveAfterInstr(
                            ComputePSWDefinedLiberal(instr), instr)) {
                        continue;
		    }

		    stackChange = InstrPushBytes(instr) - InstrPopBytes(instr);

		    if (stackChange	&&
                        PSWBitvectorIsLiveAtInstr(compensatePSW, instr)) {
			if (round == 1) {
			    pswInstr++;
			    pswWeight = pswWeight + instr->i_Bblock->b_weight;
			}
			continue;
		    }

		    if (((instr->i_dest) &&
                            OperandIsGPRegister(instr->i_dest) &&
                            (InstrPushBytes(instr) == 0)) ||
                        !(instr->i_dest)) {
			if (OPT_ELIMINATE_DEAD_REGISTERS) {
			    numRegsChanged++;
			} else {
			    continue;
			}
		    } else {
			if (OPT_ELIMINATE_DEAD_MEMORY) {
			    numStackChanged++;
			} else {
			    continue;
			}
		    }

		    text = PrintInstruction(instr);

		    if (((instr->i_dest) &&
                            OperandIsGPRegister(instr->i_dest) &&
                            (InstrPushBytes(instr) == 0)) ||
                        !(instr->i_dest)) {
			DEBUG(DB_DEADREG, "[RemoveDeadCode-Register]"
			    " Removed instruction at 0x%8.8x (%s) in %s",
			    instr->i_address, text, fun->f_name);
		    } else {
			DEBUG(DB_DEADSTK, "[RemoveDeadCode-Stack]"
			    " Removed instruction at 0x%8.8x (%s) in %s",
			    instr->i_address, text, fun->f_name);
		    }

		    free(text);

		    /*
		     * Mark the basic block so PSW liveness information
		     * is recomputedfor it.
		     */

		    BBL_SET_FLAG(bbl, BBL_TEMP_FLAG);
		    BBL_SET_FLAG(bbl, BBL_REGLIVENESS_FLAG);
		    BBL_SET_FLAG(bbl, BBL_STACKLIVENESS_FLAG);
		    BBL_SET_FLAG(bbl, BBL_PSWLIVENESS_FLAG);

		    /*
		     * If the instruction we're removing has an effect on
		     * the stack, we need to insert another instruction
		     * (hopefully cheaper than the one being removed)
		     * to simulate what the instruction does to the
		     * runtime stack. A check should be inserted to make
		     * sure we aren't tampering with the PSW.
		     */

		    if (stackChange != 0) {
			compensate = InstrCreateSPInc();
			compensate->i_src->a_displacement = -stackChange;
			if (bbl->b_last == instr) {
			    bbl->b_last = compensate;
			}
			InsertInstrAfterInstr(compensate, instr);
                        UPMChangeAllSourcesFromInstrToInstr(instr, compensate);
			spAdds++;
		    }

		    InstrDestroy(instr);
		    change = true;

		}			/* end iteration over instructions */
	    }				/* end iteration over bbls */
	}				/* end iteration over funs */

	/* Re-compute liveness information for blocks that changed */
	ComputeRegisterLiveness(0);
	ComputePSWLiveness(0);
	ComputeStackLiveness(0);

	totalChanged += (numRegsChanged + numStackChanged);

	if ((numRegsChanged + numStackChanged) != 0) {
	    REPORT(RP_SMLOOP, 
		"Dead code removal, round %d: %d instructions removed",
		round, numRegsChanged + numStackChanged);
	    REPORT(RP_SMDETAIL, "  %d (%.0f%%) had dead stack locations;"
		" %d (%.0f%%) had dead registers",
		numStackChanged,
		(100.0 * numStackChanged) / (numStackChanged + numRegsChanged),
		numRegsChanged,
		(100.0 * numRegsChanged) / (numStackChanged + numRegsChanged));
	    if (spAdds > 0) {
		REPORT(RP_SMDETAIL,
		    "  %d instructions inserted to preserve stack height",
		    spAdds);
	    }
	}

    }					/* end fixpoint */

    instrCountNew = CountInstructionsExecuted();
    newMemOps = CountMemoryOperations();

    REPORT(RP_DETAIL, "Total: %d instructions removed", totalChanged);
    REPORT(RP_DETAIL, "  Original number of instructions executed:     %.0f",
	(float) instrCountOrig);
    REPORT(RP_DETAIL, "  Post-Removal number of instructions executed: %.0f",
	(float) instrCountNew);
    if (instrCountNew != instrCountOrig) {
	REPORT(RP_DETAIL, "%f%% percent reduction",
	    (instrCountOrig - instrCountNew) * 100.0 / instrCountOrig);
    }

    REPORT(RP_DETAIL, "  %.0f%% reduction in memory operations (%lld to %lld)",
	(originalMemOps - newMemOps) * 100.0 / originalMemOps,
	originalMemOps, newMemOps);
    REPORT(RP_DETAIL,
	"  %d instructions (wt = %lld) not removed due to live PSW",
	pswInstr, pswWeight);
}



/*
 *  RegBitvectorIsLiveAtInstr(vec,instr) --
 *
 *  Returns true if 'vec' is live at the instruction.
 */
bool RegBitvectorIsLiveAtInstr(REGISTER_BITVECTOR vec, Instruction *instr) {
    Instruction *iterator;
    REGISTER_BITVECTOR live = instr->i_Bblock->b_liveOut;

    iterator = instr->i_Bblock->b_last;
    while (true) {
	if (iterator == instr) {
	    break;
	}
	live &= ~(ComputeRegistersDefined(iterator));
	live |= ComputeRegistersUsed(iterator);
	iterator = iterator->i_prev;
    }

    if (vec & live) {
	return true;
    }

    return false;
}



/*
 *  RegsLiveBitvectorAtInstr(instr) --
 *
 *  This function computes and returns the REGISTER_BITVECTOR with bit
 *  registers set for each register that is live just prior to the
 *  execution of the specified instruction.
 *
 *  instr: the instruction at which liveness should be checked
 *  returns: the bit vector containing set bit registers for each register
 *  that is live at the specified instruction
 */
REGISTER_BITVECTOR RegsLiveBitvectorAtInstr(Instruction *instr) {
    Instruction *iiterator;
    REGISTER_BITVECTOR live;

    live = instr->i_Bblock->b_liveOut;

    iiterator = instr->i_Bblock->b_last;
    while (true) {
	live &= ~(ComputeRegistersDefined(iiterator));
	live |= ComputeRegistersUsed(iiterator);

	if (iiterator == instr) {
	    break;
	}
	iiterator = iiterator->i_prev;
    }

    return live;
}



/*
 *  PSWBitvectorIsLiveAtInstr(vec,instr) --
 *
 *
 *  PSWBitvectorIsLiveAtInstr
 *  Returns true if 'vec' is live after the instruction executes.
 */
bool PSWBitvectorIsLiveAtInstr(PSW_BITVECTOR vec, Instruction *instr) {
    return PSWLiveBitvectorAtInstr(instr) & vec;
}



/*
 *  PSWBitvectorIsLiveAfterInstr(vec,instr) --
 *
 *  Returns whether or not the specified PSW bitvector is live after the
 *  specified instruction (directly after the specified instruction is executed.
 *
 *  vec: the PSW bitvector to check liveness against
 *  instr: the instruction directly after which the PSW bitvector
 *  liveness will be tested
 *  returns: true if the specified PSW bitvector is live after the specified
 *  instruction
 */
bool PSWBitvectorIsLiveAfterInstr(PSW_BITVECTOR vec, Instruction *instr) {
    return PSWLiveBitvectorAfterInstr(instr) & vec;
}



/*
 *  StackIsLiveAtInstr(defined,instr) --
 *
 *  Returns true if the defined[] integer array stack slots that 'instr'
 *  defines has any location which is live after the execution of 'instr'.
 */
extern bool StackIsLiveAtInstr(int defined[], Instruction *instr) {
    Function *fun = INSTR_FUNCTION(instr);
    Block *bbl = instr->i_Bblock;
    int sz = (fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE;
    Instruction *iterator;
    int live[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int temp[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int i;

    for (i = 0; i <= sz; i++) {
	if (bbl->b_stackshot[i].v_live == true) {
	    live[i] = true;
	} else {
	    live[i] = false;
	}
    }

    iterator = instr->i_Bblock->b_last;
    while (true) {

	if (iterator == instr) {
	    break;
	}

	for (i = 0; i <= sz; i++) {
	    temp[i] = false;
	}

	ComputeStackDefined(iterator, temp,
	    fun->f_callerFrameSize / ADDRESS_SIZE, sz);

	for (i = 0; i <= sz; i++) {
	    if ((live[i] == true) && (temp[i] == false)) {
		live[i] = true;
	    } else {
		live[i] = false;
	    }
	}

	for (i = 0; i <= sz; i++) {
	    temp[i] = false;
	}

	ComputeStackUsed(iterator, temp,
	    fun->f_callerFrameSize / ADDRESS_SIZE, sz);

	for (i = 0; i <= sz; i++) {
	    if (temp[i] == true) {
		live[i] = true;
	    }
	}

	iterator = iterator->i_prev;
    }

    for (i = 0; i <= sz; i++) {
	if (defined[i] && live[i]) {
	    return true;
	}
    }

    return false;
}



/*
 *  InstrDefinesLiveLocations(instr) --
 *
 *  Returns true if the instruction defines a location that is live.
 */
bool InstrDefinesLiveLocations(Instruction *instr) {
    int i;
    int stackDefined[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int bp_index, totalSize;
    Function *fun = INSTR_FUNCTION(instr);
    REGISTER_BITVECTOR defined;

    defined = ComputeRegistersDefinedLiberal(instr);

    /*
     * We don't want to remove instructions that define arbitrary locations
     * in memory: e,g. mov $0x5 -> (eax)
     */

    if ((instr->i_dest)
	&& OPERAND_HAS_FLAG(instr->i_dest, OPERAND_INDIRECT_FLAG))
	return true;

    /*
     * Instructions that use registers indirectly are also bad, except if
     * that register is the frame pointer.
     */

    if ((instr->i_dest)
	&& OPERAND_HAS_FLAG(instr->i_dest, OPERAND_REGISTER_INDIRECT)
	&& !RegisterIsFramePointer(instr->i_dest->a_base_reg))
	return true;

    /*
     * We only want to remove pushes/pops/assignments to GP registers and
     * assignments to the runtime stack. Anything else could be dangerous.
     */

    if (instr->i_dest) {
	if ((InstrPushBytes(instr) == 0)
	    && !OperandAccessesStackFromFramePointer(instr->i_dest)
	    && !OperandIsGPRegister(instr->i_dest)
	    && (InstrPopBytes(instr) == 0)) {
		return true;
	}
    }

    /*
     * Removing something like "push %ebp" probably means we're messing with
     * the function prologue. This is special cased at the moment, since
     * liveness information will tell us the instruction assigns to a dead
     * location even though a "leave" in the future may use it.
     */

    if ((InstrPushBytes(instr) == ADDRESS_SIZE) && (instr->i_src) &&
	OperandIsFramePointer(instr->i_src))
	return true;

    /*
     * If we're dealing with a push/pop instruction, then it's ok to be
     * removing the instruction even if ESP is live.
     */

    if (instr->i_class == CLASS_PUSH) {
	defined &= (~BITVECTOR_REGISTER_ESP);
    }
    if (instr->i_class == CLASS_POP) {
	defined &= (~BITVECTOR_REGISTER_ESP);
    }

    /*
     * If the instruction defines any registers that will later be used,
     * then it is not a candidate for removal.
     */

    if (RegBitvectorIsLiveAtInstr(defined, instr)) {
	return true;
    }

    /*
     * If it so happens that control has reached this point and the
     * destination operand is a dead register, we can safely remove say the
     * instruction is fine to remove.
     */

    if ((instr->i_dest) && OperandIsGPRegister(instr->i_dest)) {
	return false;
    }

    /*
     * We're dealing with either a push or an instruction that defines a
     * location somewhere in the runtime stack. We will compute the exact
     * locations that the instruction defines, and see if they are later
     * used by using liveness information.
     */

    totalSize = fun->f_callerFrameSize + fun->f_stackFrameSize;
    bp_index = fun->f_callerFrameSize / ADDRESS_SIZE;

    /*
     * Nothing can be done if no liveness information exists for the
     * function. If this is the case, true is returned to be conservative.
     */

    if (totalSize > MAXIMUM_SIZE_FOR_STACK_LIVENESS) {
	return true;
    }

    if (!FUNCTION_HAS_FLAG(fun, FUNCTION_SNAPSHOT_FLAG)) {
	return true;
    }

    if ((instr->i_dest) && (instr->i_dest->a_displacement) > 0) {
	return true;
    }

    for (i = 0; i <= totalSize / ADDRESS_SIZE; i++) {
	stackDefined[i] = false;
    }

    ComputeStackDefinedLiberal(instr, stackDefined, bp_index,
	totalSize / ADDRESS_SIZE);

    if (StackIsLiveAtInstr(stackDefined, instr)) {
	return true;
    }

    return false;
}



/*
 *  ComputeFunctionTrashDepths() --
 *
 *  Sets the (()->f_trashDepth) field in each function, corresponding to
 *  how far down in the stack (in bytes) the function will overwrite the
 *  stack. A negative number means it will overwrite everything.
 */
void ComputeFunctionTrashDepths() {
    Function *fun;
    Block *bbl;
    Edge *edge;
    bool change = true;

    FOREACH_FUNCTION(fun) {
	ComputeFunctionTrashDepth(fun);
    }

    while (change) {
	change = false;

	FOREACH_FUNCTION(fun) {
	    FOREACH_BBL_IN_FUN(fun, bbl) {
		edge = BBLGetEscapingEdge(bbl);
		if (edge != NULL) {
		    int oldTrashDepth = fun->f_trashDepth;
		    MERGE_TRASH_DEPTH(
			edge->e_to->b_fun->f_trashDepth, fun->f_trashDepth);
		    if (fun->f_trashDepth != oldTrashDepth) {
			change = true;
		    }
		    continue;
		}

		if (bbl->b_type == BT_CALL) {
		    Edge *direct = BBLGetFunDirectEdge(bbl);
		    Function *callee;
		    int offset = InstrBPSPDiff(bbl->b_last);
		    int oldTrashDepth = fun->f_trashDepth;
		    int temp;

		    if (direct == NULL) {
			fun->f_trashDepth = -1;
		    } else {
			callee = direct->e_to->b_fun;

			if (callee == fun) {
			    continue;
			}

			temp = callee->f_trashDepth;
			if (temp != -1) {
			    temp = temp - (offset + ADDRESS_SIZE);
			}

			MERGE_TRASH_DEPTH(fun->f_trashDepth, temp);

			if (fun->f_trashDepth != oldTrashDepth) {
			    change = true;
			}
		    }

		}		/* End test for CALL blocks */
	    }			/* End iteration over basic blocks in 'fun' */
	}			/* End iteration of functions */
    }				/* End fixpoint computation */

    return;
}



/*
 *  IsIndirectStore(instr) --
 */
static bool IsIndirectStore(Instruction *instr) {
    if (!isMovClass(instr)) {
	return false;
    }

    if (instr->i_dest == NULL) {
	return false;
    }

    if (OPERAND_HAS_FLAG(instr->i_dest,
	    OPERAND_SCALE_FLAG | OPERAND_INDEX_FLAG)) {
	return true;
    }

    if (OPERAND_HAS_FLAG(instr->i_dest, OPERAND_REGISTER_INDIRECT)) {
	return true;
    }

    if (OPERAND_HAS_FLAG(instr->i_dest, OPERAND_INDIRECT_FLAG)) {
	return true;
    }

    return false;
}



/*
 *  IsIndirectLoad(instr) --
 */
static bool IsIndirectLoad(Instruction *instr) {
    if (!isMovClass(instr)) {
	return false;
    }

    if (OPERAND_HAS_FLAG(instr->i_src, OPERAND_SCALE_FLAG|OPERAND_INDEX_FLAG)) {
	return true;
    }

    if (OPERAND_HAS_FLAG(instr->i_src, OPERAND_REGISTER_INDIRECT)) {
	return true;
    }

    if (OPERAND_HAS_FLAG(instr->i_src, OPERAND_INDIRECT_FLAG)) {
	return true;
    }

    return false;
}



/*
 *  SearchBlockFrom(instr,reg) --
 *
 *  This functions looks backward in the basic block that instr lies in
 *  for a definition of reg.
 *
 *  instr: the instruction to start backtracking from
 *  reg: the register whose definition will be looked for
 *  returns: 0 if reg is defined by either the stack or frame pointer,
 *  the register number of the source definition otherwise
 */
static int SearchBlockFrom(Instruction **instr, int *reg) {
    Instruction *iiterator;

    iiterator = *instr;
    while (iiterator != NULL && iiterator->i_Bblock == (*instr)->i_Bblock) {
	if (isMovClass(iiterator) &&
		(iiterator->i_dest != NULL) &&
		(OperandIsGPRegister(iiterator->i_dest) &&
		iiterator->i_dest->a_base_reg == *reg)) {

	    if (OperandIsAddress(iiterator->i_src)) {
		return 2;
	    }

	    if ((iiterator->i_src->a_base_reg == GetStackPointer())
		    || (iiterator->i_src->a_base_reg == GetFramePointer())) {
		*reg = 0;
		*instr = iiterator;
		return 1;
	    } else if ((iiterator->i_src->a_base_reg != 0) &&
		(iiterator->i_src->a_index_reg == 0)) {
		*reg = iiterator->i_src->a_base_reg;
		*instr = iiterator;
	    } else if ((iiterator->i_src->a_index_reg != 0) &&
		(iiterator->i_src->a_base_reg == 0)) {
		*reg = iiterator->i_src->a_index_reg;
		*instr = iiterator;
	    } else if ((iiterator->i_src->a_index_reg != 0) &&
		(iiterator->i_src->a_base_reg != 0)) {
		// haven't handled this one quite yet
	    }
	}

	iiterator = iiterator->i_prev;
    }

    return 0;
}



/*
 *  RegisterMayBeDefinedByStack(instr,reg) --
 */
static bool RegisterMayBeDefinedByStack(Instruction *instr, int reg) {
    EdgeNode *node;
    int search_result;

    BBL_SET_FLAG(instr->i_Bblock, BBL_TEMP_FLAG);

    search_result = SearchBlockFrom(&instr, &reg);
    if (search_result == 1) {
	return true;
    } else if (search_result == 2) {
	return false;
    }

    FOREACH_PRED_IN_BBL(instr->i_Bblock, node) {
	if (BBL_HAS_FLAG(NODE_FROM(node), BBL_TEMP_FLAG)) {
	    continue;
	}

	if (((NODE_FROM(node))->b_last) == NULL) {
	    continue;
	}

	if (isInterproceduralEdge(node->en_edge)) {
	    continue;
	}

	if (((NODE_FROM(node))->b_fun) != INSTR_FUNCTION(instr)) {
	    return true;
	}

	if (RegisterMayBeDefinedByStack(NODE_FROM(node)->b_last, reg)) {
	    return true;
	}
    }

    return false;
}



/*
 *  CorrectDepth(instr,op) --
 */
int CorrectDepth(Instruction *instr, Operand *op) {
    bool from_stack;

    // we'll try and correct those indirect stores that don't actually have
    // anything to do with the stack

    from_stack = false;

    // first we'll try to fix any simple ones (i.e., directly off of the
    // stack pointer or frame pointer
    if ((op->a_index_reg == 0) && (op->a_base_reg != 0)) {
	if (op->a_base_reg == GetFramePointer()) {
	    // it is directly off of the frame pointer (this should probably
	    // not ever happen since the ComputeTrashDepth function should
	    // catch all of these

	    // REPORT(RP_LGDETAIL, "corrected depth: %d", op->a_displacement);
	    return op->a_displacement;
	} else if (op->a_base_reg == GetStackPointer()) {
	    if (InstrBPSPDiff(instr) != STACK_UNKNOWN) {
		// it is directly off of the stack pointer and we know the BPSP
		// diff at this instruction

		// REPORT(RP_LGDETAIL, "corrected depth: %d",
		// InstrBPSPDiff(instr) - ((op)->a_displacement) - 4);
		return InstrBPSPDiff(instr) - op->a_displacement - 4;
	    } else {
		return -1;
	    }
	}
    } else if ((op->a_index_reg != 0) &&
	(op->a_base_reg == 0) && (op->a_scale == 1)) {
	if (op->a_index_reg == GetFramePointer()) {
	    // it is directly off of the frame pointer (this should probably
	    // not ever happen since the ComputeTrashDepth function should
	    // catch all of these

	    // REPORT(RP_LGDETAIL, "corrected depth: %d", op->a_displacement);
	    return op->a_displacement;
	} else if (op->a_index_reg == GetStackPointer()) {
	    if (InstrBPSPDiff(instr) != STACK_UNKNOWN) {
		// it is directly off of the stack pointer and we know the BPSP
		// diff at this instruction

		// REPORT(RP_LGDETAIL, "corrected depth: %d",
		// InstrBPSPDiff(instr) - ((op)->a_displacement) - 4);
		return InstrBPSPDiff(instr) - op->a_displacement - 4;
	    } else {
		return -1;
	    }
	}
    }
    // it was not directly off of the stack or frame pointer but off of
    // some other register instead -- so we need to figure out where
    // those register definitions originated

    if (op->a_index_reg != 0) {
	from_stack =
	    RegisterMayBeDefinedByStack(instr, instr->i_dest->a_index_reg);
    }

    if (op->a_base_reg != 0) {
	from_stack |=
	    RegisterMayBeDefinedByStack(instr, instr->i_dest->a_base_reg);
    }

    if (from_stack) {
	return -1;
    }

    return 0;
}



/*
 *  ComputeFunctionTrashDepth(fun) --
 *
 *  Computes the (()->f_trashDepth) field for 'fun'.
 */
static void ComputeFunctionTrashDepth(Function *fun) {
    Block *bbl;
    Instruction *instr;
    int depth = 0;
    int temp;

    if (!FUNCTION_HAS_FLAG(fun, FUNCTION_GCC_PROLOGUE_FLAG)) {
	fun->f_trashDepth = -1;
	return;
    }

    FOREACH_BBL_IN_FUN(fun, bbl) {
	BBL_REMOVE_FLAG(bbl, BBL_TEMP_FLAG);

	FOREACH_INSTR_IN_BBL(bbl, instr) {
	    X86op *disEntry = InstrGetOpcodeEntry(instr);

	    temp = ComputeOperandTrashDepth(instr->i_dest);
	    if ((temp == -1) && (IsIndirectStore(instr))) {
		temp = CorrectDepth(instr, instr->i_dest);
	    }
	    MERGE_TRASH_DEPTH(depth, temp);

	    if (disEntry->x_psw & S1D) {
		temp = ComputeOperandTrashDepth(instr->i_src);
		MERGE_TRASH_DEPTH(depth, temp);
	    }

	    if (disEntry->x_psw & S2D) {
		temp = ComputeOperandTrashDepth(instr->i_src2);

		MERGE_TRASH_DEPTH(depth, temp);
	    }
	}			/* End iteration over instructions in bbl */
    }				/* End iteration of basic blocks in fun */

    fun->f_trashDepth = depth;
}



/*
 *  ComputeOperandTrashDepth(op) --
 *
 *  Computes the maximum depth in the stack that this operand will use.
 */
static int ComputeOperandTrashDepth(Operand *op) {

    if (op == NULL) {
	return 0;
    } else if (OperandWritesToUnknownStackSlot(op)) {
	return -1;
    } else if (OperandAccessesStackFromFramePointer(op)) {
	return op->a_displacement;
    } else {
	return 0;
    }
}



/*
 *  ComputeFunctionUseDepths() --
 *
 *  Sets the f_useDepth field in each function, corresponding to
 *  how far down in the stack (in bytes) the function will access the
 *  stack. A negative number means it will access everything.
 */
void ComputeFunctionUseDepths() {
    Function *fun;
    Block *bbl;
    Edge *edge;
    bool change = true;

    FOREACH_FUNCTION(fun) {
	ComputeFunctionUseDepth(fun);
    }

    while (change) {
	change = false;

	FOREACH_FUNCTION(fun) {
	    FOREACH_BBL_IN_FUN(fun, bbl) {
		edge = BBLGetEscapingEdge(bbl);
		if (edge != NULL) {
		    int oldUseDepth = fun->f_useDepth;
		    MERGE_USE_DEPTH(
			edge->e_to->b_fun->f_useDepth, fun->f_useDepth);

		    if (fun->f_useDepth != oldUseDepth) {
			change = true;
		    }

		    continue;
		}

		if (bbl->b_type == BT_CALL) {
		    Edge *direct = BBLGetFunDirectEdge(bbl);
		    Function *callee;
		    int offset = InstrBPSPDiff(bbl->b_last);
		    int oldUseDepth = fun->f_useDepth;
		    int temp;

		    if (direct == NULL) {
			fun->f_useDepth = -1;
		    } else {
			callee = direct->e_to->b_fun;

			if (callee == fun) {
			    continue;
			}

			temp = callee->f_useDepth;
			if (temp != -1) {
			    temp = temp - (offset + ADDRESS_SIZE);
			}


			MERGE_USE_DEPTH(fun->f_useDepth, temp);

			if (fun->f_useDepth != oldUseDepth) {
			    change = true;
			}
		    }

		}		/* end test for CALL blocks */
	    }			/* end iteration over basic blocks in 'fun' */
	}			/* end iteration of functions */
    }				/* end fixpoint computation */

    return;
}



/*
 *  ComputeFunctionUseDepth(fun) -- computes the f_useDepth field for 'fun'.
 */
static void ComputeFunctionUseDepth(Function *fun) {
    Block *bbl;
    Instruction *instr;
    int depth = 0;
    int temp;

    if (!FUNCTION_HAS_FLAG(fun,FUNCTION_GCC_PROLOGUE_FLAG) || FunHasEsc2(fun)) {
	fun->f_useDepth = -1;
	return;
    }

    FOREACH_BBL_IN_FUN(fun, bbl) {
	FOREACH_INSTR_IN_BBL(bbl, instr) {
	    X86op *disEntry = InstrGetOpcodeEntry(instr);

	    temp = ComputeOperandUseDepth(instr->i_src);
	    if ((temp == -1) && (IsIndirectLoad(instr))) {
		temp = CorrectDepth(instr, instr->i_src);
	    }
	    MERGE_USE_DEPTH(depth, temp);

	    temp = ComputeOperandUseDepth(instr->i_src2);
	    MERGE_USE_DEPTH(depth, temp);

	    if (disEntry->x_psw & DUF) {
		temp = ComputeOperandUseDepth(instr->i_dest);
		MERGE_USE_DEPTH(depth, temp);
	    }

	}			/* End iteration over instructions in bbl */
    }				/* End iteration of basic blocks in fun */

    fun->f_useDepth = depth;

    return;
}



/*
 *  ComputeOperandUseDepth(op) --
 *
 *  Computes the maximum depth in the stack that this operand will use.
 */
static int ComputeOperandUseDepth(Operand *op) {
    return ComputeOperandTrashDepth(op);
}



/*
 *  FunComputeDefinedRegistersLiberal(fun) --
 *
 *  Returns a register bitvector with a bit set when 'fun' may define the
 *  supplied register.
 */
static REGISTER_BITVECTOR FunComputeDefinedRegistersLiberal(Function *fun) {
    REGISTER_BITVECTOR regs = 0;
    Block *bbl;
    Instruction *instr;

    FOREACH_BBL_IN_FUN(fun, bbl) {
	if (BBLHasHellSuccessor(bbl)) {
	    return ALL_REGISTERS;
	}
	FOREACH_INSTR_IN_BBL(bbl, instr) {
	    regs |= ComputeRegistersDefinedLiberal(instr);
	}
    }


    return regs;
}



/*
 *  ComputeRegsThatFunctionsDefineLiberal() --
 *
 *  Performs a fixpoint computation and propagates registers that each
 *  function defines around the control flow graph.
 */
void ComputeRegsThatFunctionsDefineLiberal() {
    bool change = true;
    Function *fun;
    Block *bbl;
    EdgeNode *successors;

    HELL->b_fun->f_regsDefined = ALL_REGISTERS;

    FOREACH_FUNCTION(fun) {
	fun->f_regsDefined = FunComputeDefinedRegistersLiberal(fun);
    }

    while (change) {
	change = false;

	FOREACH_FUNCTION(fun) {
	    FOREACH_BBL_IN_FUN(fun, bbl) {
		FOREACH_SUCC_IN_BBL(bbl, successors) {
		    if (isInterproceduralEdge(successors->en_edge)) {
			Function *target = (NODE_TO(successors))->b_fun;
			REGISTER_BITVECTOR newRegs = fun->f_regsDefined;
			newRegs |= target->f_regsDefined;
			if (fun->f_regsDefined != newRegs) {
			    change = true;
			    fun->f_regsDefined = newRegs;
			}
		    }		/* end test for interprocedural edge */
		}		/* end iteration over successors in bbl */
	    }			/* end iteration over bbls */
	}			/* end iteration over functions */
    }				/* end fixpoint iteration */

    return;
}



/*
 *  ComputeRegistersUsedBeforeDefinedInBBL(bbl) --
 *
 *  Returns a bitvector with the bits set for each register
 *  which this basic block uses before defining.
 */
REGISTER_BITVECTOR ComputeRegistersUsedBeforeDefinedInBBL(Block *bbl) {
    REGISTER_BITVECTOR result = 0;
    Instruction *instr;

    if (bbl == HELL) {
	return ALL_REGISTERS;
    }

    if (bbl->b_type == BT_DATABLOCK) {
        return ALL_REGISTERS;
    }

    instr = bbl->b_last;
    while (instr && (instr->i_Bblock == bbl)) {
	result &= ~(ComputeRegistersDefined(instr));
	result |= ComputeRegistersUsed(instr);
	instr = instr->i_prev;
    }

    return result;
}



/*
 *  ComputeRegistersDefinedInBBL(bbl) --
 *
 *  Returns a bitvector with the bits set for each register
 *  which this basic block defines.
 */
REGISTER_BITVECTOR ComputeRegistersDefinedInBBL(Block *bbl) {
    REGISTER_BITVECTOR result = 0;
    Instruction *instr;

    if (bbl->b_type == BT_DATABLOCK) {
        return ALL_REGISTERS;
    }

    FOREACH_INSTR_IN_BBL(bbl, instr) {
	result |= ComputeRegistersDefined(instr);
    }

    return result;
}



/*
 *  CountDeadRegisters() --
 *
 *  Returns the total number of dead registers at the end of all basic blocks
 *  in the program.
 */
int CountDeadRegisters() {
    int total = 0;
    int i;
    REGISTER_BITVECTOR liveOut;
    Block *bbl;

    FOREACH_BBL(bbl) {
	liveOut = bbl->b_liveOut;
	for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	    if (((1 << i) & liveOut) == 0) {
		total++;
	    }
	}
    }
    return total;
}



/*
 *  PSWLiveBitvectorAtInstr(instr) --
 *
 *  Returns the PSW_BITVECTOR that represents all live PSW flags at instr, i.e.,
 *  directly before execution of the instruction.
 *
 *  instr: the instruction to check for PSW liveness
 *  returns: the PSW_BITVECTOR representing all live PSW flags going into instr
 */
PSW_BITVECTOR PSWLiveBitvectorAtInstr(Instruction *instr) {
    PSW_BITVECTOR live;
    Instruction *iiterator;

    live = instr->i_Bblock->b_liveOutPSW;

    iiterator = instr->i_Bblock->b_last;
    while (true) {
	live &= ~(ComputePSWDefined(iiterator));
	live |= ComputePSWUsed(iiterator);

	if (iiterator == instr) {
	    break;
	}
	iiterator = iiterator->i_prev;
    }

    return live;
}



/*
 *  PSWLiveBitvectorAfterInstr(instr) --
 *
 *  Returns the PSW_BITVECTOR that represents all live PSW flags just after
 *  instr, i.e., directly after execution of the instruction.
 *
 *  instr: the instruction to check for PSW liveness
 *  returns: the PSW_BITVECTOR that represents all live PSW flags going out of
 *  instr
 */
PSW_BITVECTOR PSWLiveBitvectorAfterInstr(Instruction *instr) {
    PSW_BITVECTOR live;
    Instruction *iiterator;

    live = instr->i_Bblock->b_liveOutPSW;

    iiterator = instr->i_Bblock->b_last;
    while (iiterator != instr) {
	live &= ~(ComputePSWDefined(iiterator));
	live |= ComputePSWUsed(iiterator);

	iiterator = iiterator->i_prev;
    }

    return live;
}



/*
 *  GetFreeRegisterAtInstr(instr) --
 *
 *  Given an instruction, this function will attempt to find a free
 *  register (one that is free just before this instruction executes) and
 *  return it. In order to function properly, register liveness must have
 *  already been computed.
 *
 *  instr: the instruction at which the register should be free
 *  returns: the register that is free, or -1 if none are free
 *  see also: ComputeRegisterLiveness
 */
int GetFreeRegisterAtInstr(Instruction *instr) {
    REGISTER_BITVECTOR vec;
    int i;

    vec = RegsLiveBitvectorAtInstr(instr);
    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	// if the bit is not set
	if (!((1 << i) & vec)) {
	    return BitRegisterToRegister(1 << i);
	}
    }

    return -1;
}
