/*
 *  liveness.ci.c
 *
 *  Context insensitive liveness analysis for registers, PSW and memory
 */
#include "global.h"
#include "utilities.h"
#include "liveness.h"
#include "sysdep.h"
#include "linkage.h"
#include "cp.h"
#include "stack.analyze.h"
#include "flowgraph.h"
#include "opt.unreachable.h"
#include "x86-liveness.h"
#include "x86-table-defs.h"
#include "x86-asm.h"
#include "liveness.cs.h"
#include "liveness.ci.h"
#include "opt.h"

/*
 *  FUNCTIONS FOR REGISTER LIVENESS
 */
/* Performs the register liveness fixpoint computation */
int CIComputeRegisterFixpoint();

/* Computes register liveness for the basic block 'bbl' */
static bool ComputeRegisterLivenessForBBL(Block *bbl);

/* Computes the register that are live going into the basic block */
static REGISTER_BITVECTOR ComputeRegisterLiveIn(Block *bbl);

/*
 *  FUNCTIONS FOR PSW LIVENESS
 */
/* Performs the PSW liveness fixpoint computation */
int CIComputePSWFixpoint();

/* Computes PSW liveness for the basic block 'bbl' */
static bool ComputePSWLivenessForBBL(Block *bbl);

/* Computes the PSW flags that are live going into the basic block */
static PSW_BITVECTOR ComputePSWLiveIn(Block *bbl);

/*
 *  FUNCTIONS FOR STACK LIVENESS
 */
/* Performs the runtime stack liveness fixpoint computation */
int CIComputeStackFixpoint();

/* Computes stack liveness for the basic block 'bbl' */
static bool ComputeStackLivenessForBBL(Block *bbl);

/* Computes the stack slots that are live going into the BBL */
static int *ComputeStackLiveIn(Block *bbl, int *result);



/*
 *  CIComputeRegisterLiveness(full) --
 *
 *  When called, ComputeRegisterLiveness will set bits in the space allocated
 *  in the basic block structures for the register bitvectors
 *  ->b_liveIn and ->b_liveOut).
 *  A set bit means that the register is live.
 */
void CIComputeRegisterLiveness(int full) {
    Block *blocks;
    Function *fun;
    int rounds;

    REPORT(RP_MAJOR, "Computing Insensitive Register Liveness");
    
    if (full) {

	/* Set the initial conditions--no registers are live */
	FOREACH_BBL(blocks) {
	    blocks->b_liveIn = 0;
	    blocks->b_liveOut = 0;
	    BBL_SET_FLAG(blocks, BBL_REGLIVENESS_FLAG);
	}

	FOREACH_FUNCTION(fun) {
	    fun->f_entry->b_liveIn = 0;
	    fun->f_entry->b_liveOut = 0;
	    BBL_SET_FLAG(fun->f_entry, BBL_REGLIVENESS_FLAG);
	    fun->f_exit->b_liveIn = 0;
	    fun->f_exit->b_liveOut = 0;
	    BBL_SET_FLAG(fun->f_exit, BBL_REGLIVENESS_FLAG);
	}

	/* Everything is live in HELL */
	HELL->b_liveIn = ALL_REGISTERS;
	HELL->b_liveOut = ALL_REGISTERS;
	BBL_SET_FLAG(HELL, BBL_REGLIVENESS_FLAG);
    }

    /* Begin the fixpoint computation */
    rounds = CIComputeRegisterFixpoint();

    REPORT(RP_SMLOOP,"Context insensitive register liveness: %d rounds",rounds);

}



/*
 *  CIComputeRegisterFixpoint() --
 *
 *  Given a basic block, this function will recursively compute the liveness
 *  of its successors, then set its own live registers based on the results.
 */
int CIComputeRegisterFixpoint() {
    Block *bbl;
    Function *funs;
    int rounds = 0;
    bool flag = true;


    /*
     *        Context Insensitive Register Liveness Analysis
     *        -----------------------------------------------
     * Below is the code for computing the fixpoint of the dataflow
     * equations:
     *
     *             LiveIn[n] = use[n] + (LiveOut[n] - def[n])
     *             LiveOut[n] = LiveIn[s] for all successors, s, of n
     *      Subject to the initial conditions:
     *             LiveIn[n] = LiveOut[n] = EMPTY SET
     *
     */

    while (flag) {

	flag = false;
	rounds++;

	FOREACH_BBL(bbl) {
	    if (ComputeRegisterLivenessForBBL(bbl) == true) {
		flag = true;
	    }
	}

	FOREACH_FUNCTION(funs) {
	    if (ComputeRegisterLivenessForBBL(funs->f_entry) == true) {
		flag = true;
	    }
	    if (ComputeRegisterLivenessForBBL(funs->f_exit) == true) {
		flag = true;
	    }
	}

	if (ComputeRegisterLivenessForBBL(HELL) == true) {
	    flag = true;
	}

    }					/* End fixpoint computation */

    FOREACH_BBL(bbl) {
	bbl->b_liveIn = ComputeRegisterLiveIn(bbl);
    }
    
	FOREACH_FUNCTION(funs) {
        funs->f_entry->b_liveIn = ComputeRegisterLiveIn(funs->f_entry);
        funs->f_exit->b_liveIn = ComputeRegisterLiveIn(funs->f_exit);        
    }
    
    return rounds;
}



/*
 *  ComputeRegisterLivenessForBBL(bbl) --
 *
 *  Computes the liveness for the supplied basic block, and returns true
 *  if there is a change in liveness, or false if there is not.
 */
static bool ComputeRegisterLivenessForBBL(Block *bbl) {
    EdgeNode *predecessors;
    bool flag = false;
    REGISTER_BITVECTOR in;

    if (!BBL_HAS_FLAG(bbl, BBL_REGLIVENESS_FLAG)) {
	return flag;
    }

    BBL_REMOVE_FLAG(bbl, BBL_REGLIVENESS_FLAG);
    flag = true;
    
    in = ComputeRegisterLiveIn(bbl);

    FOREACH_PRED_IN_BBL(bbl, predecessors) {
	Block *from = NODE_FROM(predecessors);
	REGISTER_BITVECTOR new_out = from->b_liveOut | in;

	if (NODE_TYPE(predecessors) == ET_FUNLINK ||
        NODE_TYPE(predecessors) == ET_DATALINK) {
	    continue;
	}

	if (new_out != from->b_liveOut) {
	    from->b_liveOut = new_out;
	    BBL_SET_FLAG(from, BBL_REGLIVENESS_FLAG);
	}
    }				/* End iteration over predecessors of bbl */

    return flag;
}



/*
 *  ComputeRegisterLiveIn(bbl) --
 *
 *  Computes the registers which are live at the start of the basic
 *  block. LiveIn[n] = use[n] + (LiveOut[n] -- def[n]).
 */
static REGISTER_BITVECTOR ComputeRegisterLiveIn(Block *bbl) {
    REGISTER_BITVECTOR def = 0;
    REGISTER_BITVECTOR use = 0;
    Instruction *iterator;

    if (bbl->b_type == BT_DATABLOCK) {
        use = ALL_REGISTERS;
        def = ALL_REGISTERS;
    } else {
        /*
         * 'def' will contain the registers that are defined in the basic
         * block. 'use' will contain the registers that are used before
         * being defined in the basic block.  'use' is calculated by XORing
         * the registers used in an instruction with all the registers
         * defined before it, then ANDing the result to the registers used
         * by the instruction. Example:
         *
         *         Defined before instr:  01000101
         *                Used by instr:  00010100
         *             Defined XOR Used:  01010001
         *  (Defined XOR Used) AND Used:  00010000
         *
         * This gives us what we want, the register in the fourth slot is
         * the only one used before being defined. An instruction which
         * both defines and uses a register will use the register before
         * defining it, so 'use' is calculated before 'def'.
         */

        FOREACH_INSTR_IN_BBL(bbl, iterator) {
            use |=
                ((def ^ ComputeRegistersUsed(iterator)) &
                    ComputeRegistersUsed(iterator));
            def |= ComputeRegistersDefined(iterator);
        }
    }

    /*
     * Knowing use[n], def[n], and LiveOut[n], we can now calculate
     * LiveIn[n]:
     *              LiveIn[n] = use[n] + (LiveOut[n] - def[n])
     */

    return use | (bbl->b_liveOut & ~def);
}



/*
 *  CIComputePSWLiveness(full) --
 *
 *  Sets the fields ->b_liveInPSW and ->b_liveOutPSW
 *  for all basic blocks.
 */
void CIComputePSWLiveness(int full) {
    Block *blocks = 0, *biterator = 0;
    Function *fun = 0;
    int rounds = 0;
    bool change = true;
    EdgeNode *sucessors = 0;

    REPORT(RP_MAJOR, "Computing Insensitive PSW Liveness");
    if (full) {
        /* Set the initial conditions--no PSW bits are live */
        FOREACH_BBL(blocks) {
            if (blocks->b_type == BT_DATABLOCK) {
                printf("[%s:%d] %s\n",
                    __FILE__,
                    __LINE__,
                    blocks->b_fun->f_name);
                blocks->b_liveInPSW = ALL_PSW;
            } else {
                blocks->b_liveInPSW = 0;
            }
            blocks->b_liveOutPSW = 0;
            BBL_SET_FLAG(blocks, BBL_PSWLIVENESS_FLAG);
        }

        FOREACH_FUNCTION(fun) {
            fun->f_entry->b_liveInPSW = 0;
            fun->f_entry->b_liveOutPSW = 0;
            BBL_SET_FLAG(fun->f_entry, BBL_PSWLIVENESS_FLAG);
            fun->f_exit->b_liveInPSW = 0;
            fun->f_exit->b_liveOutPSW = 0;
            BBL_SET_FLAG(fun->f_exit, BBL_PSWLIVENESS_FLAG);
        }
    
    }
    
    /* First compute the PSW live at the entry of each function
     * For this we shall assume that hell is not going to contribute
     * anything. Then we know the union of PSW live at the entry
     * of each function will be the PSW live at HELL nodes
     */

    HELL->b_liveInPSW = 0;
    HELL->b_liveOutPSW = 0;
    BBL_SET_FLAG(HELL, BBL_PSWLIVENESS_FLAG);

    FOREACH_FUNCTION(fun) {
        while (change) {
            change = false;
            FOREACH_BBL_IN_FUN(fun, biterator) {
        	    if (ComputePSWLivenessForBBL(biterator) == true) {
                    change = true;
                }        
            }
        }
    }

    FOREACH_FUNCTION(fun) {
        FOREACH_SUCC_IN_BBL(fun->f_entry, sucessors) {
            biterator = NODE_TO(sucessors);
            HELL->b_liveInPSW |= biterator->b_liveInPSW;
        }
    }

    FOREACH_BBL(biterator) {
        if (BBLHasHellPredecessor(biterator)) {
            HELL->b_liveInPSW |= biterator->b_liveInPSW;
        }
    }
    
    HELL->b_liveOutPSW = HELL->b_liveInPSW;     

    BBL_SET_FLAG(HELL, BBL_PSWLIVENESS_FLAG);

    if (full) {
        FOREACH_BBL(blocks) {
	    if (blocks->b_type == BT_DATABLOCK) {
            blocks->b_liveInPSW = ALL_PSW;
	    } else {
            blocks->b_liveInPSW = 0;
	    }
            blocks->b_liveOutPSW = 0;
            BBL_SET_FLAG(blocks, BBL_PSWLIVENESS_FLAG);
        }

        FOREACH_FUNCTION(fun) {
            fun->f_entry->b_liveInPSW = 0;
            fun->f_entry->b_liveOutPSW = 0;
            BBL_SET_FLAG(fun->f_entry, BBL_PSWLIVENESS_FLAG);
            fun->f_exit->b_liveInPSW = 0;
            fun->f_exit->b_liveOutPSW = 0;
            BBL_SET_FLAG(fun->f_exit, BBL_PSWLIVENESS_FLAG);
        }
    }
    
    /* Begin the fixpoint computation */
    rounds = CIComputePSWFixpoint();

}



/*
 *  CIComputePSWFixpoint() --
 *
 *  Performs a fixpoint computation to compute the live psw flags at the start
 *  and end of each BBL.
 */
int CIComputePSWFixpoint() {
    Block *bbl;
    Function *funs;
    int rounds = 0;
    bool flag = true;


    /*
     *           Context Insensitive PSW Liveness Analysis
     *        -----------------------------------------------
     * Below is the code for computing the fixpoint of the dataflow
     * equations:
     *
     *             LiveIn[n] = use[n] + (LiveOut[n] - def[n])
     *             LiveOut[n] = LiveIn[s] for all successors, s, of n
     *      Subject to the initial conditions:
     *             LiveIn[n] = LiveOut[n] = EMPTY SET
     *
     */

    while (flag) {

	flag = false;
	rounds++;

	FOREACH_BBL(bbl) {
	    if (ComputePSWLivenessForBBL(bbl) == true) {
		flag = true;
	    }
	}

	FOREACH_FUNCTION(funs) {
	    if (ComputePSWLivenessForBBL(funs->f_entry) == true) {
		flag = true;
	    }
	    if (ComputePSWLivenessForBBL(funs->f_exit) == true) {
		flag = true;
	    }
	}

    	if (ComputePSWLivenessForBBL(HELL) == true) {
    	    flag = true;
    	}

    }					/* End fixpoint computation */

    FOREACH_BBL(bbl) {
	bbl->b_liveInPSW = ComputePSWLiveIn(bbl);
    }

    return rounds;

}



/*
 *  ComputePSWLivenessForBBL(bbl) --
 *
 *  Computes the PSW liveness for the supplied basic block
 *  and returns true if there was a change.
 */
static bool ComputePSWLivenessForBBL(Block *bbl) {
    EdgeNode *predecessors;
    bool flag = false;
    REGISTER_BITVECTOR in;

    if (!BBL_HAS_FLAG(bbl, BBL_PSWLIVENESS_FLAG)) {
	return flag;
    }

    BBL_REMOVE_FLAG(bbl, BBL_PSWLIVENESS_FLAG);
    flag = true;
    in = ComputePSWLiveIn(bbl);

    FOREACH_PRED_IN_BBL(bbl, predecessors) {
	Block *from = NODE_FROM(predecessors);
	REGISTER_BITVECTOR new_out = from->b_liveOutPSW | in;

	if (NODE_TYPE(predecessors) == ET_FUNLINK ||
        NODE_TYPE(predecessors) == ET_DATALINK) {
	    continue;
	}

	if (new_out != from->b_liveOutPSW) {
	    from->b_liveOutPSW = new_out;
	    BBL_SET_FLAG(from, BBL_PSWLIVENESS_FLAG);
	}
    }				/* End iteration over predecessors of bbl */

    return flag;
}



/*
 *  ComputePSWLiveIn(bbl) --
 *
 *  Returns a bitvector for the registers that are live coming into 'bbl'.
 */
static PSW_BITVECTOR ComputePSWLiveIn(Block *bbl) {
    PSW_BITVECTOR def = 0;
    PSW_BITVECTOR use = 0;
    Instruction *iterator;

    if (bbl->b_type == BT_HELL) {
        return HELL->b_liveInPSW;
    }

    if (bbl->b_type == BT_DATABLOCK) {
        use = ALL_PSW;
        def = ALL_PSW;
        return use | (bbl->b_liveOutPSW & ~def);
    } 
    /*
     * 'def' will contain the PSW flags that are defined in the basic
     * block. 'use' will contain the PSW flags that are used before
     * being defined in the basic block.  'use' is calculated by XORing
     * the PSW flags used in an instruction with all the f;ags
     * defined before it, then ANDing the result to the flags used
     * by the instruction. Example:
     *
     *         Defined before instr:  01000101
     *                Used by instr:  00010100
     *             Defined XOR Used:  01010001
     *  (Defined XOR Used) AND Used:  00010000
     *
     * This gives us what we want, the flag in the fourth slot is
     * the only one used before being defined. An instruction which
     * both defines and uses a PSW bit will use the bit before
     * defining it, so 'use' is calculated before 'def'.
     */

    FOREACH_INSTR_IN_BBL(bbl, iterator) {
	use |= (def ^ ComputePSWUsed(iterator)) & ComputePSWUsed(iterator);
	def |= ComputePSWDefined(iterator);
    }

    /*
     * Knowing use[n], def[n], and LiveOut[n], we can now calculate
     * LiveIn[n]:
     *              LiveIn[n] = use[n] + (LiveOut[n] - def[n])
     */

    return use | (bbl->b_liveOutPSW & ~def);
}



/*
 *  CIComputeStackLiveness(full) --
 *
 *  Computes liveness information for the models of the stack.
 */
void CIComputeStackLiveness(int full) {
    Block *blocks;
    Function *fun;
    int i, rounds, sz;

    REPORT(RP_MAJOR, "Computing Insensitive Stack Liveness");
    if (full) {
        AllocateSnapshots();
        CalculateSPBPDifference();
    }

    ComputeFunctionUseDepths();

    /*
     * If the function or one of its callers has an activation record that
     * is "too large" then we won't compute liveness information for it.
     */
    if (full) {
	FOREACH_FUNCTION(fun) {
	    FUNCTION_SET_FLAG(fun, FUNCTION_TEMP_FLAG);
	    if (fun->f_stackFrameSize + fun->f_callerFrameSize >
		MAXIMUM_SIZE_FOR_STACK_LIVENESS)
		FUNCTION_REMOVE_FLAG(fun, FUNCTION_TEMP_FLAG);
	    if (!FUNCTION_HAS_FLAG(fun, FUNCTION_SNAPSHOT_FLAG)) {
		FUNCTION_REMOVE_FLAG(fun, FUNCTION_TEMP_FLAG);
	    }
	    FOREACH_BBL_IN_FUN(fun, blocks) {
		if (!BBL_HAS_FLAG(blocks, BBL_STACKSHOT_FLAG)) {
		    FUNCTION_REMOVE_FLAG(fun, FUNCTION_TEMP_FLAG);
		}
	    }
	}

	/* Set the initial conditions--no stack locations are live */
	FOREACH_FUNCTION(fun) {
	    if (!FUNCTION_HAS_FLAG(fun, FUNCTION_TEMP_FLAG)) {
		continue;
	    }
	    FOREACH_BBL_IN_FUN(fun, blocks) {
		for (i = 0; i <= fun->f_stackFrameSize / ADDRESS_SIZE; i++) {
		    blocks->b_stackshot[i].v_live = false;
		    BBL_SET_FLAG(blocks, BBL_STACKLIVENESS_FLAG);
		}
	    }
	}

	FOREACH_FUNCTION(fun) {

	    if (!FUNCTION_HAS_FLAG(fun, FUNCTION_TEMP_FLAG)) {
		continue;
	    }

	    sz = (fun->f_stackFrameSize +
		fun->f_callerFrameSize) / ADDRESS_SIZE;

	    /*
	     * If the function doesn't have the typical prologue/epilogue code
	     * then we really can't make assumptions about what is live and what
	     * isn't, so we'll set everything to live. Note that most functions
	     * are well-behaved (> 90%).
	     */

	    if (!FUNCTION_HAS_FLAG(fun, FUNCTION_WELL_BEHAVED_FLAG) ||
		FUNCTION_HAS_FLAG(fun,
		    FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG)) {
		FOREACH_BBL_IN_FUN(fun, blocks) {
		    for (i = 0; i <= sz; i++) {
			blocks->b_stackshot[i].v_live = true;
		    }
		}
	    }

	    /* Take care of the entry/exit pseudo blocks */
	    for (i = 0; i <= sz; i++) {
		fun->f_entry->b_stackshot[i].v_live = false;
		fun->f_exit->b_stackshot[i].v_live = false;
	    }
	    BBL_SET_FLAG(fun->f_entry, BBL_STACKLIVENESS_FLAG);
	    BBL_SET_FLAG(fun->f_exit, BBL_STACKLIVENESS_FLAG);
	}
    }				/* end setup for the full computation */

    /* Begin the fixpoint computation */
    rounds = CIComputeStackFixpoint();

    REPORT(RP_SMLOOP, "Context insensitive stack liveness: %d rounds", rounds);
}



/*
 *  CIComputeStackFixpoint() --
 *
 *  Iteratively computes the fixpoint of the liveness dataflow equations for
 *  the stack.
 */
int CIComputeStackFixpoint() {
    Block *bbl;
    Function *funs;
    int rounds = 0;
    bool flag = true;

    /*
     *          Context Insensitive Stack Liveness Analysis
     *        -----------------------------------------------
     * Below is the code for computing the fixpoint of the dataflow
     * equations:
     *
     *             LiveIn[n] = use[n] + (LiveOut[n] - def[n])
     *             LiveOut[n] = LiveIn[s] for all successors, s, of n
     *      Subject to the initial conditions:
     *             LiveIn[n] = LiveOut[n] = EMPTY SET
     *
     */

    while (flag) {

	flag = false;
	rounds++;

	FOREACH_FUNCTION(funs) {
	    if (!FUNCTION_HAS_FLAG(funs, FUNCTION_TEMP_FLAG)) {
		continue;
	    }
	    FOREACH_BBL_IN_FUN(funs, bbl) {
		if (ComputeStackLivenessForBBL(bbl) == true) {
		    flag = true;
		}
		if (ComputeStackLivenessForBBL(funs->f_entry) == true) {
		    flag = true;
		}
		if (ComputeStackLivenessForBBL(funs->f_exit) == true) {
		    flag = true;
		}
	    }
	}

    }					/* End fixpoint computation */

    return rounds;

}



/*
 *  ComputeStackLivenessForBBL(bbl) -- computes liveness information for 'bbl'.
 */
static bool ComputeStackLivenessForBBL(Block *bbl) {
    EdgeNode *predecessors;
    bool flag = false;
    int result[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int liveout[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int newout[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int sz = (bbl->b_fun->f_stackFrameSize +
	bbl->b_fun->f_callerFrameSize) / ADDRESS_SIZE;
    int i;

    if (!BBL_HAS_FLAG(bbl, BBL_STACKLIVENESS_FLAG)) {
	return flag;
    }

    if (BBLHasEsc2(bbl)) {
	for (i = 0; i <= sz; i++) {
	    bbl->b_stackshot[i].v_live = true;
	}
    }

    BBL_REMOVE_FLAG(bbl, BBL_STACKLIVENESS_FLAG);
    flag = true;
    ComputeStackLiveIn(bbl, result);

    FOREACH_PRED_IN_BBL(bbl, predecessors) {
	Block *from = NODE_FROM(predecessors);
	bool same = true;

	if (isInterproceduralEdge(predecessors->en_edge)) {
	    continue;
	}

	for (i = 0; i <= sz; i++) {
	    liveout[i] = false;
	    if ((from->b_stackshot[i].v_live) == true) {
		liveout[i] = true;
		newout[i] = true;
	    } else if (result[i] == true) {
		newout[i] = true;
	    } else {
		newout[i] = false;
	    }
	}

	for (i = 0; i <= sz; i++) {
	    if (newout[i] != liveout[i]) {
		same = false;
	    }
	}

	if (same == false) {
	    for (i = 0; i <= sz; i++) {
		from->b_stackshot[i].v_live = newout[i];
	    }
	    BBL_SET_FLAG(from, BBL_STACKLIVENESS_FLAG);
	}
    }			/* End iteration over predecessors of bbl */
    return flag;
}



/*
 *  ComputeStackLiveIn(bbl,result) --
 *
 *  Returns a bitvector for the registers that are live coming into 'bbl'.
 */
static int *ComputeStackLiveIn(Block *bbl, int *result) {
    Instruction *iterator;
    int def[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int use[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int tmp[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int tmp2[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int tmp3[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int tmp4[MAXIMUM_SIZE_FOR_STACK_LIVENESS];
    int sz = (bbl->b_fun->f_stackFrameSize +
	bbl->b_fun->f_callerFrameSize) / ADDRESS_SIZE;
    int i;

    for (i = 0; i <= sz; i++) {
	def[i] = 0;
	use[i] = 0;
    }

    FOREACH_INSTR_IN_BBL(bbl, iterator) {

	for (i = 0; i <= sz; i++) {
	    tmp2[i] = tmp[i] = tmp3[i] = tmp4[i] = result[i] = false;
	}

	ComputeStackUsed(iterator, tmp,
	    bbl->b_fun->f_callerFrameSize / ADDRESS_SIZE, sz);

	for (i = 0; i <= sz; i++) {
	    tmp2[i] = false;
	    if (def[i] == true && tmp[i] == false) {
		tmp2[i] = true;
	    }
	    if (def[i] == false && tmp[i] == true) {
		tmp2[i] = true;
	    }
	}

	for (i = 0; i <= sz; i++) {
	    if (tmp[i] == true && tmp2[i] == true) {
		use[i] = true;
	    }
	}

	ComputeStackDefined(iterator, tmp3,
	    bbl->b_fun->f_callerFrameSize / ADDRESS_SIZE, sz);

	for (i = 0; i <= sz; i++) {
	    if (tmp3[i] == true) {
		def[i] = true;
	    }
	}
    }

    for (i = 0; i <= sz; i++) {
	tmp4[i] = false;
	if (bbl->b_stackshot[i].v_live == true) {
	    if (def[i] == false) {
		tmp4[i] = true;
	    }
	}
    }

    for (i = 0; i <= sz; i++) {
	result[i] = false;
	if (tmp4[i] == true) {
	    result[i] = true;
	}
	if (use[i] == true) {
	    result[i] = true;
	}
    }

    return result;
}
