/*
 *  liveness.cs.c
 *
 *  Context-Sensitive liveness analysis, as described in (1) "Register Liveness
 *  Analysis of Executable Code", and (2) "ALTO: A Platform for Object Code
 *  Modification" by Robert Muth.
 */
#include "global.h"
#include "liveness.cs.h"
#include "liveness.h"
#include "sysdep.h"
#include "utilities.h"
#include "flowgraph.h"
#include "x86-liveness.h"

/* Sets the f_bypass field in each function for registers
 * which bypass the function */
static int CSComputeRegisterBypassSets(int full);

/*
 *  Sets the f->mayuse field in each function for registers which are live
 *  at entry.  This function assumed that the f_bypass field has already
 *  been computed for each fun
 */
static int CSComputeRegisterMayUseSets(int full);

/* Computes use information for the basic block,
 * and propagates results to its predecessors */
static bool CSComputeRegisterMayUseSetForBBL(Block *bbl);

/* Computes bypass information for the basic block,
 * and propagates results to its predecessors */
static bool CSComputeRegisterBypassSetForBBL(Block *bbl);

/*
 *  Performs the register fixpoint computation to attain liveness information.
 *  This function call assumes that the f_bypass and f_mayuse fields
 *  are set appropriately for each function; presumably this is done with
 *  calls to CSComputeRegisterBypassSets() and CSComputeRegisterBypassSets()
 */
static int CSComputeRegisterLivenessFixpoint(int full);

/* Computes liveness information for the basic block,
 *  and propagates results to its predecessors */
static bool CSComputeRegisterLivenessForBBL(Block *bbl);



/*
 *  CSComputeRegisterLiveness(full) --
 *
 *  Sets the bitvectors BBL_LIVE_OUT and BBL_LIVE_IN for each basic block
 *  corresponding to what registers are live at the start and end of the block.
 */
void CSComputeRegisterLiveness(int full) {
    int bypass_rounds, mayuse_rounds, liveness_rounds;
    Block *bbl;

    FOREACH_BBL(bbl) {
	bbl->b_ubd = ComputeRegistersUsedBeforeDefinedInBBL(bbl);
	bbl->b_def = ComputeRegistersDefinedInBBL(bbl);
    }

    /*
     * Compute the BYPASS sets for each function. The BYPASS set for a
     * function is the set of registers which, if live a the exit block
     * of a function, are live at the entry block to the function.
     */
    bypass_rounds = CSComputeRegisterBypassSets(full);

    /*
     * Compute the MAYUSE sets for each function. The MAYUSE set for a function
     * is the set of registers that the entry block for the function may use;
     * that is, the set of all live registers at the entry block.
     */
    mayuse_rounds = CSComputeRegisterMayUseSets(full);

    /*
     * Using the summary information for each function (namely BYPASS and
     * MAYUSE sets), we can now compute liveness information for each basic
     * block in a context-sensitive manner.
     */
    liveness_rounds = CSComputeRegisterLivenessFixpoint(full);

    REPORT(RP_SMLOOP, "Context sensitive liveness: "
	"%d+%d+%d rounds (bypass, may-use, merge)",
	bypass_rounds, mayuse_rounds, liveness_rounds);

    return;
}



/*
 *  CSComputeRegisterMayUseSets(full) --
 *
 *  Computes MAYUSE sets for each function. A register falls into this set
 *  if it can be used before being defined in the function.  More precisely,
 *  if the register is live at the entry block then it is in the MAYUSE set.
 */
static int CSComputeRegisterMayUseSets(int full) {
    Block *bbl;
    Function *fun;
    bool change = true;
    int rounds = 0;

    if (full) {
	FOREACH_BBL(bbl) {
        bbl->b_may_use_in = 0;
        bbl->b_may_use_out = 0;
	    BBL_SET_FLAG(bbl, BBL_REGLIVENESS_FLAG);
	}

	HELL->b_may_use_in = ALL_REGISTERS;
	HELL->b_may_use_out = ALL_REGISTERS;
	BBL_SET_FLAG(HELL, BBL_REGLIVENESS_FLAG);
	HELL->b_fun->f_mayuse = ALL_REGISTERS;

	FOREACH_FUNCTION(fun) {
	    fun->f_entry->b_may_use_in = 0;
	    fun->f_entry->b_may_use_out = 0;
	    fun->f_exit->b_may_use_in = 0;
	    fun->f_exit->b_may_use_out = 0;
	    BBL_SET_FLAG(fun->f_entry, BBL_REGLIVENESS_FLAG);
	    BBL_SET_FLAG(fun->f_exit, BBL_REGLIVENESS_FLAG);
	    fun->f_mayuse = 0;
	}
    }

    /*
     * Compute the fixpoint for the following dataflow equations:
     *       (1) MayUseIn[n]  = use[n] + (MayUseOut[n] - def[n])
     *       (2) MayUseOut[n] = U(MayUseIn[s]), for all s a successor of n,
     *				TYPE(n) != BT_CALL
     *           MayUseOut[n] = MayUse[f] + (ByPass[f] INTERSECT MayUseIn[n'])
     *             if TYPE(n) == BT_CALL, where f = Callee[n]
     *				and n' = ReturnNode[n]
     *       (3) MayUse[f]    = MayUseIn[InitNode[f]]
     * Subject to the initial conditions:
     *       MayUseOut[n] = 0
     *       MayUseIn[n] = 0
     *       MayUse[f] = 0
     *       ByPass[f] = f->f_bypass
     */

    while (change) {

	rounds++;
	change = false;

	FOREACH_BBL(bbl) {
	    if (CSComputeRegisterMayUseSetForBBL(bbl)) {
		change = true;
	    }
	}

	if (CSComputeRegisterMayUseSetForBBL(HELL)) {
	    change = true;
	}

	FOREACH_FUNCTION(fun) {
	    if (CSComputeRegisterMayUseSetForBBL(fun->f_entry)) {
		change = true;
	    }
	    if (CSComputeRegisterMayUseSetForBBL(fun->f_exit)) {
		change = true;
	    }
	    if (fun->f_mayuse != fun->f_entry->b_may_use_in) {
		fun->f_mayuse = fun->f_entry->b_may_use_in;
		change = true;
	    }
	}

    }					/* end fixpoint computation */

    return rounds;
}



/*
 *  CSComputeRegisterMayUseSetForBBL(bbl) --
 *
 *  Computes MAY_USE_IN information for 'bbl' and propagates the results
 *  to all its predecessors.
 */
static bool CSComputeRegisterMayUseSetForBBL(Block *bbl) {
    EdgeNode *predecessors, *successors;
    Function *callee;
    Block *linkBlock;
    bool flag = false;
    REGISTER_BITVECTOR in, out = 0;

    if (!BBL_HAS_FLAG(bbl, BBL_REGLIVENESS_FLAG)) {
	return flag;
    }

    BBL_REMOVE_FLAG(bbl, BBL_REGLIVENESS_FLAG);
    flag = true;

    /*
     * For function call edges, the set of registers that the block may use
     * at its exit is computed by adding the set of registers the callee may
     * use to the set of registers the function link block uses that also
     * bypass the callee. For non function call blocks, we simply union all
     * MayUseIn sets for all our successors.
     */

    if (bbl->b_type == BT_CALL) {
	Edge *direct = BBLGetFunDirectEdge(bbl);
	linkBlock = BBLGetFunlinkEdge(bbl)->e_to;

	if (direct) {
	    callee = direct->e_to->b_fun;
	} else {
	    callee = HELL->b_fun;
	}

	out = callee->f_mayuse | (callee->f_bypass & linkBlock->b_may_use_in);

    } else {
	FOREACH_SUCC_IN_BBL(bbl, successors) {
	    switch (NODE_TYPE(successors)) {
	    case ET_FUNCALL:
	    case ET_FUNDIRECT:
		WARNING("[CSComputeRegisterMayUseSetForBBL]"
		    " Unexpected edge type encountered.");
		break;
	    default:
		out |= NODE_TO(successors)->b_may_use_in;
		break;
	    }
	}
    }

    /*
     * Update our use information for the start of the basic block.  If it
     * differs from the vector that currently sits there, we'll tell all our
     * predecessors that they should recompute their use sets.
     */

    if (bbl->b_type != BT_EXIT) {
	bbl->b_may_use_out = out;
    }
    in = (bbl->b_may_use_out & ~(bbl->b_def)) | bbl->b_ubd;

    if (bbl->b_may_use_in != in) {
	FOREACH_PRED_IN_BBL(bbl, predecessors) {
	    BBL_SET_FLAG(NODE_FROM(predecessors), BBL_REGLIVENESS_FLAG);
	}
    }

    bbl->b_may_use_in = in;

    return flag;
}



/*
 *  CSComputeRegisterBypassSets(full) --
 *
 *  Computes BYPASS sets for each function. A register falls into this if it
 *  is live at the return node and live at any call node calling the function.
 */
static int CSComputeRegisterBypassSets(int full) {
    Block *bbl;
    Function *fun;
    bool change = true;
    int rounds = 0;

    if (full) {
	FOREACH_BBL(bbl) {
        bbl->b_bypass_in = 0;
        bbl->b_bypass_out = 0;
	    BBL_SET_FLAG(bbl, BBL_REGLIVENESS_FLAG);
	}

	HELL->b_bypass_in = ALL_REGISTERS;
	HELL->b_bypass_out = ALL_REGISTERS;
	BBL_SET_FLAG(HELL, BBL_REGLIVENESS_FLAG);
	HELL->b_fun->f_bypass = ALL_REGISTERS;

	FOREACH_FUNCTION(fun) {
	    fun->f_entry->b_bypass_in = 0;
	    fun->f_entry->b_bypass_out = 0;
	    fun->f_exit->b_bypass_in = 0;
	    fun->f_exit->b_bypass_out = ALL_REGISTERS;
	    BBL_SET_FLAG(fun->f_entry, BBL_REGLIVENESS_FLAG);
	    BBL_SET_FLAG(fun->f_exit, BBL_REGLIVENESS_FLAG);
	    fun->f_bypass = 0;
	}
    }

    /*
     * Compute the fixpoint of the following dataflow equations:
     *   (1) ByPassIn[n]  = use[n] + (ByPassOut[n] - def[n])
     *   (2) ByPassOut[n] = UNION(ByPassIn[s]),
     *				s a successor of n if TYPE(n) != BT_CALL
     *                    = (ByPass[Callee[n]] INTERSECT ByPassIn[n']),
     *                      when TYPE(n) == BT_CALL,
     *				and where n' = ReturnNode[n]
     *   (3) ByPass[f]    = ByPassIn[((f)->f_entry)]
     * Subject to the initial conditions:
     *   ByPassOut[n] = 0
     *   ByPassIn[n] = 0
     *   ByPassOut[n] = ALL_REGISTERS when n is a node of type BT_EXIT
     *   ByPass[f] = 0
     */

    while (change) {

	rounds++;
	change = false;

	FOREACH_BBL(bbl) {
	    if (CSComputeRegisterBypassSetForBBL(bbl)) {
		change = true;
	    }
	}

	if (CSComputeRegisterBypassSetForBBL(HELL)) {
	    change = true;
	}

	FOREACH_FUNCTION(fun) {
	    if (CSComputeRegisterBypassSetForBBL(fun->f_entry)) {
		change = true;
	    }
	    if (CSComputeRegisterBypassSetForBBL(fun->f_exit)) {
		change = true;
	    }
	    if (fun->f_bypass != fun->f_entry->b_bypass_in) {
		fun->f_bypass = fun->f_entry->b_bypass_in;
		change = true;
	    }
	}

    }					/* end fixpoint computation */

    return rounds;
}



/*
 *  CSComputeRegisterBypassSetForBBL(bbl) --
 *
 *  Computes BYPASS information for 'bbl' and propagates the results to all its
 *  predecessors.
 */
static bool CSComputeRegisterBypassSetForBBL(Block *bbl) {
    EdgeNode *predecessors, *successors;
    Function *callee;
    Block *linkBlock;
    bool flag = false;
    REGISTER_BITVECTOR in, out = 0;

    if (!BBL_HAS_FLAG(bbl, BBL_REGLIVENESS_FLAG)) {
	return flag;
    }

    BBL_REMOVE_FLAG(bbl, BBL_REGLIVENESS_FLAG);
    flag = true;

    /*
     * For function call edges, the set of registers that the block may bypass
     * at its exit is the registers that bypass the link node that also bypass
     * the callee.
     */

    if (bbl->b_type == BT_CALL) {
        //printf("BBL_ID: %d\n", bbl->b_id);
	Edge *direct = BBLGetFunDirectEdge(bbl);
//        if (BBLGetFunlinkEdge(bbl) == NULL) {
//            printf("this block ID: %d, next block ID: %d\n ",
//                bbl->b_id, bbl->b_next->b_id);
//            //printf("BOO-O-O-OM\n");
//            linkBlock = bbl->b_next;
//        } else {
	    linkBlock = BBLGetFunlinkEdge(bbl)->e_to;
//        }

	if (direct) {
	    callee = direct->e_to->b_fun;
	} else {
	    callee = HELL->b_fun;
	}

	out = callee->f_bypass & linkBlock->b_bypass_in;

    } else {
	FOREACH_SUCC_IN_BBL(bbl, successors) {
	    switch (NODE_TYPE(successors)) {
	    case ET_FUNCALL:
	    case ET_FUNDIRECT:
		WARNING("[CSComputeRegisterBypassSetForBBL]"
		    " Unexpected edge type encountered.");
		break;
	    default:
		out |= NODE_TO(successors)->b_bypass_in;
		break;
	    }
	}
    }

    /*
     * Update our bypass information for the start of the basic block.
     * If it differs from the vector that currently sits there,
     * we'll tell all our predecessors that they should recompute
     * their bypass sets.
     */

    if (bbl->b_type != BT_EXIT) {
	bbl->b_bypass_out = out;
    }
    
    in = (bbl->b_bypass_out & ~(bbl->b_def)) | bbl->b_ubd;

    if (bbl->b_bypass_in != in) {
	FOREACH_PRED_IN_BBL(bbl, predecessors) {
	    BBL_SET_FLAG(NODE_FROM(predecessors), BBL_REGLIVENESS_FLAG);
	}
    }

    bbl->b_bypass_in = in;

    return flag;
}



/*
 *  CSComputeRegisterLivenessFixpoint(full) --
 *
 *  Using MayUse and ByPass sets as summary information for each function,
 *  this function computes liveness information for each basic block
 *  and sets the BBL_LIVE_IN and BBL_LIVE_OUT fields appropriately.
 */
static int CSComputeRegisterLivenessFixpoint(int full) {
    Block *bbl;
    Function *fun;
    bool change = true;
    int rounds = 0;

    REPORT(RP_MAJOR, "Computing ContextSensitive Register Liveness");
    FOREACH_BBL(bbl) {
	bbl->b_liveIn = 0;
	bbl->b_liveOut = 0;
	BBL_SET_FLAG(bbl, BBL_REGLIVENESS_FLAG);
    }

    HELL->b_liveIn = ALL_REGISTERS;
    HELL->b_liveOut = ALL_REGISTERS;
    BBL_SET_FLAG(HELL, BBL_REGLIVENESS_FLAG);

    FOREACH_FUNCTION(fun) {
	fun->f_entry->b_liveIn = 0;
	fun->f_entry->b_liveOut = 0;
	fun->f_exit->b_liveIn = 0;
	fun->f_exit->b_liveOut = 0;
	BBL_SET_FLAG(fun->f_entry, BBL_REGLIVENESS_FLAG);
	BBL_SET_FLAG(fun->f_exit, BBL_REGLIVENESS_FLAG);
    }

    /*
     * Compute the fixpoint for the following dataflow equations:
     *    (1) LiveIn[n] = use[n] + (LiveOut[n] - def[n])
     *    (2) LiveOut[n] = UNION(LiveIn[s]) for all s,
     *			a successor of n when TYPE(n) != BT_CALL
     *                   = MayUse[f] + (ByPass[f] INTERSECT LiveIn[n'])
     *                     when TYPE(n) == BT_CALL, where f is the callee
     *				and n' is the link bbl
     * Subject to the initial conditions:
     *    LiveIn[n] = 0
     *    LiveOut[n] = 0
     *    MayUse[f] = f->mayuse
     *    ByPass[f] = f->bypass
     */

    while (change) {

	rounds++;
	change = false;

	FOREACH_BBL(bbl) {
	    if (CSComputeRegisterLivenessForBBL(bbl)) {
		change = true;
	    }
	}

	if (CSComputeRegisterLivenessForBBL(HELL)) {
	    change = true;
	}

	FOREACH_FUNCTION(fun) {
	    if (CSComputeRegisterLivenessForBBL(fun->f_entry)) {
		change = true;
	    }
	    if (CSComputeRegisterLivenessForBBL(fun->f_exit)) {
		change = true;
	    }
	}

    }					/* end fixpoint computation */

    return rounds;
}



/*
 *  CSComputeRegisterLivenessForBBL(bbl) --
 *
 *  Updates BBL_LIVE_IN and BBL_LIVE_OUT fields for the basic block.
 *  Results are propagated to the predecessors of 'bbl'.
 */
static bool CSComputeRegisterLivenessForBBL(Block *bbl) {
    EdgeNode *predecessors, *successors;
    Function *callee;
    Block *linkBlock;
    bool flag = false;
    REGISTER_BITVECTOR in, out = 0;

    if (!BBL_HAS_FLAG(bbl, BBL_REGLIVENESS_FLAG)) {
	return flag;
    }

    BBL_REMOVE_FLAG(bbl, BBL_REGLIVENESS_FLAG);
    flag = true;

    /*
     * For function call edges, the set of liveOut registers is
     * MayUse[callee] + (ByPass[callee] & LiveIn[linkNode]).
     * For all other nodes, we union the BBL_LIVE_IN fields of the successors.
     */

    if (bbl->b_type == BT_CALL) {
	Edge *direct = BBLGetFunDirectEdge(bbl);
	linkBlock = BBLGetFunlinkEdge(bbl)->e_to;

	if (direct) {
	    callee = direct->e_to->b_fun;
	} else {
	    callee = HELL->b_fun;
	}
	out = callee->f_mayuse | (callee->f_bypass & linkBlock->b_liveIn);

    } else {
	FOREACH_SUCC_IN_BBL(bbl, successors) {
	    switch (NODE_TYPE(successors)) {
	    case ET_FUNCALL:
	    case ET_FUNDIRECT:
		WARNING("[CSComputeRegisterLivenessForBBL]"
		    " Unexpected edge type encountered.");
		break;
	    default:
		out |= NODE_TO(successors)->b_liveIn;
		break;
	    }
	}
    }

    /*
     * Update our liveness information for the start of the basic block.
     * If it differs from the vector that currently sits there,
     * we'll tell all our predecessors that they should
     * recompute their liveness sets.
     */

    bbl->b_liveOut = out;
    in = (bbl->b_liveOut & ~(bbl->b_def)) | bbl->b_ubd;

    if (bbl->b_liveIn != in) {
	FOREACH_PRED_IN_BBL(bbl, predecessors) {
	    BBL_SET_FLAG(NODE_FROM(predecessors), BBL_REGLIVENESS_FLAG);
	}
    }

    bbl->b_liveIn = in;

    return flag;
}



/*
 *  CSComputePSWLiveness(full) --
 *
 *  Sets the bitvectors for the PSW flags which are live
 *  at the start and end of each BBL
 */
void CSComputePSWLiveness(int full) {
    WARNING("Context Sensitive PSW Liveness not implemented\n");
}



/*
 *  CSComputeStackLiveness(full) --
 *
 *  Sets the live bits in the stack snapshots for stack slots
 *  that are live at the start of each BBL
 */
void CSComputeStackLiveness(int full) {
    WARNING("Context Sensitive Stack Liveness not implemented\n");
}
