/*
 *  Drives the optimization processes.
 */
#include "global.h"
#include "lto.h"
#include "utilities.h"
#include "opt.unreachable.h"
#include "opt.layout.h"
#include "edge.profile.h"
#include "dominator.h"
#include "sysdep.h"
#include "sloop.h"
#include "flowgraph.h"
#include "stack.analyze.h"
#include "liveness.h"
#include "cp.h"
#include "upm.h"
#include "relocate.h"
#include "opt.h"
/*
 *  include "opt.syscall.h"
 */
/* Performs a round of CP/LSF and dead code removal,
 * and returns true if anything changed */
static bool OptimizeEasy(bool want_stack_analysis, bool *need_stack_analysis);

/*
 *  Optimize() --
 *
 *  Calls various helper functions to optimize the control flow graph.
 */
void Optimize() {
    bool want_stack_analysis = BooleanParameter("AnalyzeStacks");
    bool need_stack_analysis;

    if (BooleanParameter("RemoveDeadCode")) {
	RemoveUnreachableCode(startins->i_Bblock);
    }

    if (BooleanParameter("FixPrologues")) {
	NormalizeOperations();
    }

    if (OPT_PEEPHOLE) {
	PeepholeOptimize();
    }

    if (want_stack_analysis) {
	MarkWellBehavedFunctions2();
	StoreInstrBPSPDiffForAllFuns();
	MarkFunsAliasingStackPointer();
    }

    need_stack_analysis = false;

    if (BooleanParameter("PropagateConstants")) {
	DumpAllSections();
	PropagateConstants();
	RemoveUnreachableCode(startins->i_Bblock);

	need_stack_analysis = false;
    }

    if (want_stack_analysis && need_stack_analysis) {
	MarkWellBehavedFunctions2();
	StoreInstrBPSPDiffForAllFuns();
	MarkFunsAliasingStackPointer();

	need_stack_analysis = false;
    }

    if (BooleanParameter("ComputeDominators")) {
	StoreDominatorInfo();
    }

    if (BooleanParameter("InlineFunctions")) {
	GenerateSloopOutput(sloopfile, "preinline");

	OptInlineFuns();
	BBLAssignWeights();
	MergeBasicBlocks();

	GenerateSloopOutput(sloopfile, "postinline");

	CountStructures();

	need_stack_analysis = true;
    }

    if (want_stack_analysis && need_stack_analysis) {
	REPORT(RP_MINOR, "Performing stack analysis");
	MarkWellBehavedFunctions2();
	StoreInstrBPSPDiffForAllFuns();
	MarkFunsAliasingStackPointer();

	need_stack_analysis = false;
    }

    OptimizeEasy(want_stack_analysis, &need_stack_analysis);

    if (BooleanParameter("OptimizeJumpTables")) {
	OptimizeJumpTables();
    }

    if (OPT_PEEPHOLE) {
	UpdateFunctions();
	PeepholeOptimize();
    }

    if (BooleanParameter("Compact")) {
        OptCompact();
        RemoveUnreachableCode(startins->i_Bblock);
    }

    GenerateSloopOutput(sloopfile, "preschedule");

    if (BooleanParameter("ScheduleInstructions")) {
	REPORT(RP_MINOR, "Scheduling instructions");
	ScheduleInstructions();
	GenerateSloopOutput(sloopfile, "postschedule");
    }

    if (BBL_LAYOUT_SCHEME != BBL_LAYOUT_OFF) {
	REPORT(RP_MINOR, "Basic block layout");
	BBLOptimizeLayout();
	BBLAssignWeights();
    } else {
//    InstrUpdateEdges();
    }

    if (OPT_PEEPHOLE) {
	ExtendSecondJumps();
    }
// ReportSuspiciousInstrs();

//   // Value Profiles
//
//   if (option_P) {
//      REPORT(RP_MINOR, "Profiling Values");
//      AllocateSnapshots();
//      ValueProfile();
//      BBLOutputStats();
//      REPORT(RP_MINOR, "Finished Value Profiling");
//      return;
//   }
//
//   if (BooleanParameter("SpecializeValues")) {
//     GenerateSloopOutput(sloopfile, "prevaluespec");
//
//      REPORT(RP_MINOR, "Value Specializing");
//      ReadValueProfiles();
//      ValueSpecialize();
//
//      GenerateSloopOutput(sloopfile, "postvaluespec");
//   }
//


//   GenerateSloopOutput(sloopfile, "presyscall");
//
//    REPORT(RP_MINOR, "SYS-CALL CLUSTERING");
//    identify_all_syscalls();
//
//    UpdateInstrAddresses(SECTION_REF_CODE);
//
//    REPORT(RP_MINOR, "SYS-CALL CLUSTERING COMPLETE");
//    GenerateSloopOutput(sloopfile, "postsyscall");
//
}



/*
 *  OptimizeEasy(want_stack_analysis, need_stack_analysis) --
 *
 *  Performs a round of constant propagation, load/store forwarding
 *  and dead-code removal, and returns true if anything changes during
 *  the course of these optimizations.
 */
static bool OptimizeEasy(bool want_stack_analysis, bool *need_stack_analysis) {

    /*
     *  Propagate Constants
     */

    if (BooleanParameter("PropagateConstants")) {
	PropagateConstants();
	RemoveUnreachableCode(startins->i_Bblock);

	*need_stack_analysis = false;
    }

    if (want_stack_analysis && *need_stack_analysis) {
	MarkWellBehavedFunctions2();
	StoreInstrBPSPDiffForAllFuns();
	MarkFunsAliasingStackPointer();

	*need_stack_analysis = false;
    }

    if (BooleanParameter("ForwardValues")) {
	ComputeRegisterLiveness(1);
	OptLoadStoreForward();

	*need_stack_analysis = true;
    }

    if (want_stack_analysis && *need_stack_analysis) {
	MarkWellBehavedFunctions2();
	StoreInstrBPSPDiffForAllFuns();
	MarkFunsAliasingStackPointer();

	*need_stack_analysis = false;
    }

    if (OPT_ELIMINATE_DEAD_REGISTERS || OPT_ELIMINATE_DEAD_MEMORY) {
	ComputeRegisterLiveness(1);
	ComputePSWLiveness(1);
	ComputeStackLiveness(1);
	RemoveDeadCode();

	*need_stack_analysis = true;
    }

    return true;
}
