/*
 *  main.c
 *
 *  Drives the disassembly/assembly/optimization processes. A binary is
 *  read-in, normalized, a flowgraph is constructed, various optimizations
 *  are made, and then the binary is rewritten.
 */

#include <unistd.h>

#include "global.h"
#include "disassemble.h"
#include "flowgraph.h"
#include "hash.h"
#include "utilities.h"
#include "relocate.h"
#include "lto.h"
#include "sloop.h"
#include "sysdep.h"
#include "linkage.h"
#include "edge.profile.h"
#include "opt.layout.h"
#include "dominator.h"
#include "liveness.h"
#include "opt.h"
#include "plto.h"
#include "upm.h"
#include "x86-registers.h"
#include "x86-classes.h"
#include "elfio.h"



static void ReportCommandLine(int argc, char *argv[]);



/*
 *  main(argc,argv) -- overall control.
 */
int main(int argc, char *argv[]) {
    char cwd[PATH_MAX];
    ProfileFile *next_profile;

    Configure(argc, argv);

    if (WANT_DEBUG(DB_UNBUF)) {
    	setvbuf(stdout, NULL, _IOLBF, 0);
    }

    binFile = fopen(binFilename, "rb");	
    if (binFile == NULL) {
    	PFATAL(binFilename);
    }

    // open profile data now, for an early diagnostic if files are not there
    next_profile = profileFiles;
    while (next_profile != NULL) {
        next_profile->pf_file = fopen(next_profile->pf_name, "rb");
        if (next_profile->pf_file == NULL) {
            PFATAL(next_profile->pf_name);
        }
        next_profile = next_profile->pf_next;
    }

    REPORT(RP_MAJOR, "PLTO (University of Arizona) version " VERSION "%s",
	KERNEL_MODE ? " (Charon)" : "");
    ReportCommandLine(argc, argv);
    REPORT(RP_DETAIL, "Working directory: %s", getcwd(cwd, sizeof(cwd)));
    REPORT(RP_DETAIL, "Configuration file: %s", cfgFilename);

    REPORT(RP_MAJOR, "Loading %s", binFilename);
    SetupDataStructures();
    fclose(binFile);

    if (sloopfile) {
	REPORT(RP_MINOR, "Writing SLOOP output file: %s", sloopFilename);
	GenerateSloopOutput(sloopfile, "initial");
    }

    if (PROFILE_BBLS || PROFILE_EDGES) {
	REPORT(RP_MAJOR, "Instrumenting for profiling");
	Profile(); //InsertProfilingCode() or smth
    } else {

        ReadProfiles();

        if (bblProfiles > 0 && edgeProfiles == 0) {
            EdgeComputeWeights();
        } else if (bblProfiles == 0 && edgeProfiles > 0) {
            BBLComputeWeights();
        }

        BBLAssignTemperature();

        BBLMarkEdgesToBeAdjusted();

	GenerateSloopOutput(sloopfile, "preopt");

	if (!OPT_DISABLE_ALL) {
	    REPORT(RP_MAJOR, "Optimizing");
	    Optimize();
	}
    }

    if (!KERNEL_MODE && BooleanParameter("LabelBasicBlocks")) {
        AddBasicBlockSymbols();
    }

    REPORT(RP_MAJOR, "Splitting sections");
    SplitSections();
    FinalizeMemoryImage();

    REPORT(RP_MINOR, "Modifications complete");
    DumpAllSections();

    if (WANT_DEBUG(DB_WADDR)) {		// if WriteAddressTranslations
	WriteAddressTranslationFile(
	    MakeFilename("AddressTranslationFile", binFilename));
    }

    if (WANT_DEBUG(DB_WRTGDB)) {	// if WriteGDBScript
	WriteGDBScript(MakeFilename("GDBScriptFile", binFilename), NULL);
    }

    WriteOutBinary();

    GenerateSloopOutput(sloopfile, "final");

    if (WANT_DEBUG(DB_STATHG)) {	// if GenerateStaticHistograms
	GenerateStaticOpcodeHistogram();
    }

    if (WANT_DEBUG(DB_WGTHG)) {		// if GenerateWeightedHistograms
	GenerateWeightedOpcodeHistogram();
    }

    return 0;
}



/*
 *  ReportCommandLine(argc, argv) -- report the contents of the command line.
 */
static void ReportCommandLine(int argc, char *argv[]) {
    int i, n;
    char *s;

    for (i = n = 0; i < argc; i++) {
	n += strlen(argv[i]) + 1;
    }

    s = alloc(n + 1);
    for (i = n = 0; i < argc; i++) {
	s[n++] = ' ';
	strcpy(s + n, argv[i]);
	n += strlen(argv[i]);
    }

    REPORT(RP_DETAIL, "Command line:%s", s);
    free(s);
}
