/*
 *  misc.c -- a dumping ground for miscellaneous stuff
 */
#include "utilities.h"
#include "global.h"
#include "hash.h"
#include "flowgraph.h"
#include "linkage.h"
#include "sysdep.h"
#include "upm.h"
#include "x86-registers.h"
#include "x86-classes.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>


static char *touchMeNotFuncs[] = {
    "_text",
    "stext",
    "startup_32",
    "check_x87",
    "is486",
    "is386",
    "ready",
    "swapper_pg_dir",
    "stack_start",
    "check_hlt",
    "idt_descr",
    "gdt_descr",
    "idt",
    "gdt",
    "cpu_gdt_descr",
    "checkCPUtype",
    "pg0",
    "pg1",
    "empty_zero_page",
    "L6",
    "rp_sidt",
    "setup_idt",
    "int_msg",
    "ignore_int",
       
    /* new added functions */
       
    /* High-level Kernel Init */
    "start_kernel",
    "page_address_init",
    "setup_arch",
    "setup_per_cpu_areas ",
    "smp_prepare_boot_cpu",
    "sched_init",
    "preempt_disable",
    "build_all_zonelists",
    "page_alloc_init",
    "sort_main_extable",
    "trap_init",
    "rcu_init",
    "init_IRQ",
    "pidhash_init",
    "init_timers",
    "softirq_init",
    "time_init",
    "console_init",
    "vfs_caches_init_early",

    /* page frames */
    "mem_init",
    "mem_map",
    "free_pages_init",
    "set_max_mapnr_init",
    "set_highmem_pages_init",
    "free_all_bootmem",
    "kclist_add",
    "kmem_cache_create",
    "panic",
    "page_is_ram",
    "PageReserved",
    "pfn_to_page",

    "pgtable_cache_init",
    "in_interrupt",
    "kmem_cache_alloc",
    "cache_estimate",

    "kmem_cache_init",
    "setup_per_cpu_pageset",
    "numa_policy_init",
    "calibrate_delay",
    "pidmap_init",

    "prio_tree_init",
    "anon_vma_init",
    "fork_init",
    "proc_caches_init",
    "buffer_init",
    "unnamed_dev_init",
    "key_init",
    "security_init",
    "vfs_caches_init",
    "init_mount_tree",  /* may cause null dereference */
    "radix_tree_init",
    "signals_init",
    "cpuset_init ",
    "check_bugs",
    "acpi_early_init",
    "rest_init",

    "do_basic_setup",
    "do_initcalls",
    "ResetCounter", // kprofile function
    "resetValueTable", // kprofile function
    "csum_partial",
    "csum_partial_copy_generic"
};

bool DonotTouchMe(Function *fun) {
    int i;

    if (strncmp("__initcall", fun->f_name, strlen("__initcall")) == 0) { 
        return true;
    }

    for (i = 0; i < ARRAYSIZE(touchMeNotFuncs); i++) {
        if (strcmp(touchMeNotFuncs[i], fun->f_name) == 0) {
            return true;
        }
    } 
 
    return false;
}

/*
 *  AddressToInstr(address) --
 *
 *  Looks up the address in a hash table and returns the instruction who has
 *  that address, or NULL if none is found
 */
Instruction *AddressToInstr(UINT32 address) {
    return InstrLookupEntry(address);
}


/* To map the an instruction to the address. But the
 * current one. 
 */
Instruction * FinalAddrToInstr(unsigned int addr) {
    Instruction *instr = 0;
    FOREACH_INSTR(instr) {
        if (instr->i_final_address == addr) {
            return instr;
        }
    }

    return NULL;
}



/*
 *  AddressToData(address) --
 *
 *  Given an address, AddressToData takes the contents of the address and
 *  interprets it as an unsigned integer (address). This is similar to
 *  dereferencing a pointer
 */
UINT32 AddressToData(UINT32 address) {
    Section *iterator;
    int x;
    UINT8 *location;
    int offset;

    x = 0;

    FOREACH_SECTION(iterator) {
	if (SECTION_CONTAINS(iterator, address)) {
	    offset = address - iterator->s_address;
	    location = iterator->s_data + offset;
	    x = (*location++ & 0xff) << 0;
	    x |= (*location++ & 0xff) << 8;
	    x |= (*location++ & 0xff) << 16;
	    x |= (*location++ & 0xff) << 24;
	    break;
	}
    }
    return x;

}



/*
 *  AddressToSection(address) --
 *
 *  Given an address, returns a pointer to the segment in which it lies
 */
Section *AddressToSection(UINT32 address) {
    Section *iterator;

    FOREACH_SECTION(iterator) {
	if (SECTION_CONTAINS(iterator, address)) {
	    return iterator;
	}
    }

    return NULL;
}



/*
 *  AddressToPrevSection(address) --
 *
 *  Given an address returns a pointer to the section that starts before
 *  the address.
 */
Section *AddressToPrevSection(ADDRESS address) {
    Section *cur = NULL, *result = NULL;

    FOREACH_SECTION(cur) {
	if ((cur->s_address < address) &&
		(result == NULL || cur->s_address > result->s_address)) {
	    result = cur;
	}
    }
    return result;
}



/*
 *  NewAddressToSection(address) --
 *
 *  Given an address in the final executable, returns a pointer to the
 *  section in which it lies
 */
Section *NewAddressToSection(ADDRESS address) {
    Section *iterator;
    Section *sec = NULL;
    ADDRESS addr = 0x0;

    FOREACH_SECTION(iterator) {
	if ((iterator->s_newaddress <= address) &&
	    (iterator->s_newaddress + iterator->s_newdatasize > address)) {
	    return iterator;
	}
    }

    FOREACH_SECTION(iterator) {
	if (iterator->s_newaddress > addr && iterator->s_newaddress <= address){
	    addr = iterator->s_newaddress;
	    sec = iterator;
	}
    }
    return sec;
}



/*
 *  edgeTypeToString(type) --
 *
 *  Converts an edge type to a string and returns the result.
 */
char *edgeTypeToString(int type) {
    char *result;

    switch (type) {
    case ET_HELL:
	result = "HELL";
	break;
    case ET_FUNCALL:
	result = "FUNCALL";
	break;
    case ET_UNCOND:
	result = "UNCOND";
	break;
    case ET_NORMAL:
	result = "NORMAL";
	break;
    case ET_RETURN:
	result = "RETURN";
	break;
    case ET_SWITCH:
	result = "SWITCH";
	break;
    case ET_FUNLINK:
	result = "FUNLINK";
	break;
    case ET_DATALINK:
        result = "DATALINK";
        break;
    case ET_TRUE:
	result = "true";
	break;
    case ET_FALSE:
	result = "false";
	break;
    case ET_ENTRY:
	result = "ENTRY";
	break;
    case ET_EXIT:
	result = "EXIT";
	break;
    case ET_FUNDIRECT:
	result = "FUNDIRECT";
	break;
    case ET_HELLMAYBE:
	result = "HELLMAYBE";
	break;
    case ET_SYSCALL:
	result = "SYSCALL";
	break;
    case ET_SYSCALL_LINK:
	result = "SYSCALL_LINK";
	break;
    default:
	result = "NONE";
	break;
    }
    return result;
}



/*
 *  blockTypeToString(type) --
 *
 *  Converts a block type to a string and returns the result
 */
char *blockTypeToString(int type) {
    char *result;

    switch (type) {
    case BT_CALL:
	result = "CALL";
	break;
    case BT_NORMAL:
	result = "NORMAL";
	break;
    case BT_BRANCH:
	result = "BRANCH";
	break;
    case BT_UNCOND:
	result = "UNCOND";
	break;
    case BT_RETURN:
	result = "RETURN";
	break;
    case BT_SWITCH:
	result = "SWITCH";
	break;
    case BT_HELL:
	result = "HELL";
	break;
    case BT_ENTRY:
	result = "ENTRY";
	break;
    case BT_EXIT:
	result = "EXIT";
	break;
    case BT_SYSCALL:
	result = "SYSCALL";
	break;
    case BT_DATABLOCK:
	result = "DATABLOCK";
	break;    
    default:
	result = "NONE";
	break;
    }
    return result;
}



/*
 *  lookupFunctionName(fun) --
 *
 *  Returns a pointer to a string containing the label of the function.
 */
char *lookupFunctionName(Function *fun) {
    Symbol *iterator;

    FOREACH_SYMBOL(iterator) {
	if (!(iterator->sy_flags & SYMBOL_FUNCTION_FLAG)) {
	    continue;
	}
	if (iterator->sy_value == fun->f_first->b_first->i_address) {
	    return (char *) iterator->sy_name;
	    break;
	}
    }
    return "Not found";
}



/*
 *  nameToSection(name) --
 *
 *  Given a section name, returns a pointer to the corresponding section.
 */
Section *nameToSection(const char *name) {
    Section *iterator;

    FOREACH_SECTION(iterator) {
	if (strcmp(iterator->s_name, name) == 0) {
	    return iterator;
	}
    }
    return NULL;
}



/*
 *  nameToSymbol(name) --
 *
 *  Iterates over all the symbols in the symbol table and attempts to match the
 *  name.
 */
Symbol *nameToSymbol(char *name) {
    Symbol *siterator;

    FOREACH_SYMBOL(siterator) {
	if (strcmp(siterator->sy_name, name) == 0) {
	    return siterator;
	}
    }

    return NULL;
}



/*
 *  InstrContainingAddress(destination) --
 *
 *  Locates the instruction (if it exists) that contains the given address.
 */
Instruction *InstrContainingAddress(UINT32 destination) {
    int i;

    for (i = 0; i > -16; i--) {
	if (AddressToInstr(destination + i)) {
	    if ((((AddressToInstr(destination + i))->i_length) +
		    ((AddressToInstr(destination + i))->i_address)) >=
		destination)
		return AddressToInstr(destination + i);
	}
    }
    return NULL;
}



/*
 *  DeleteAllInstructionsStartingWith(instr) --
 *
 *  Starting with the given instruction, this function deletes and frees up
 *  the memory used by all instructions after, and including, it.
 */
void DeleteAllInstructionsStartingWith(Instruction *instr) {
    Instruction *iterator;
    Instruction *prevInstr;

    iterator = instr;
    if (iterator->i_prev) {
	iterator->i_prev->i_next = NULL;
    }
    while (iterator != NULL) {
	InstrRemoveEntry(iterator);
	prevInstr = iterator;
	iterator = iterator->i_next;
	InstrDisconnect(prevInstr);
	InstrFree(prevInstr);
    }
    return;
}



/*
 *  operandHasDisplacement(op) --
 *
 *  Returns true if the operand uses a displacement
 */
bool operandHasDisplacement(Operand *op) {
    if (OP_TYPE(op, OPERAND_DISP_R_FLAG)) {
	return true;
    }
    if (OP_TYPE(op, OPERAND_DISP_A_FLAG)) {
	return true;
    }
    if (OP_TYPE(op, OPERAND_IMM_FLAG)) {
	return true;
    }
    return false;
}



/*
 *  IncreaseSectionSize(section,nbytes) --
 *
 *  This assumes that new stuff goes at the end of the old stuff, and it
 *  copies the old stuff into the beginning of the new pointer..
 */
void IncreaseSectionSize(Section *section, int nbytes) {
    unsigned char *newdata = zalloc(section->s_newdatasize + nbytes);

    CASSERT(newdata != NULL, "could not increase section size");

    memcpy(newdata, section->s_newdata, section->s_newdatasize);
    free(section->s_newdata);

    section->s_newdata = newdata;
    section->s_newdatasize += nbytes;
    section->s_growth += nbytes;
}



/*
 *  RelocationFree(reloc) --
 *
 *  Functions for freeing up memory used by blocks, instructions and functions
 */
void RelocationFree(Relocation *reloc) {
    free(reloc);
}



/*
 *  functionFirstInstr(fun) -- returns the first instruction in the function.
 */
Instruction *functionFirstInstr(Function *fun) {
    Block *blocks;

    FOREACH_BBL_IN_FUN(fun, blocks) {
	if (blocks->b_first != NULL) {
	    return blocks->b_first;
	}
    }
    return NULL;
}



/*
 *  functionLastInstr(fun) -- returns the last instruction in the function.
 */
Instruction *functionLastInstr(Function *fun) {
    Block *blocks;

    blocks = fun->f_last;
    while (blocks != NULL) {
	if (blocks->b_last != NULL) {
	    return blocks->b_last;
	}
	blocks = blocks->b_prev;
    }
    return NULL;
}



/*
 *  isOnlyInstrInBBL(instr) --
 *
 *  Returns true if the given instruction is the only one in its basic block
 */
bool isOnlyInstrInBBL(Instruction *instr) {
    if (instr->i_Bblock->b_first != instr) {
	return false;
    }
    if (instr->i_Bblock->b_last != instr) {
	return false;
    }
    return true;
}



/*
 *  IDToInstr(id) --
 */
Instruction *IDToInstr(INT32 id) {
    Instruction *iterator;

    FOREACH_INSTR(iterator) {
	if (iterator->i_id == id) {
	    return iterator;
	}
    }
    return NULL;
}



#define CHKFLAGS() \
    if (!INSTR_HAS_FLAG(instr, INSTR_HAS_CLASS_FLAGS)) \
	InstrSetClassFlags(instr);



/*
 *  isOPOFFClass(instr) --
 */
bool isOPOFFClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_OPOFF);
}



/*
 *  isOPDIRClass(instr) --
 */
bool isOPDIRClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_OPDIR);
}



/*
 *  isJumpClass(instr) --
 */
bool isJumpClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_JUMP);
}



/*
 *  isBranchClass(instr) --
 */
bool isBranchClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_BRANCH);
}



/*
 *  isPCChangingClass(instr) --
 */
bool isPCChangingClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_PC_CHANGING);
}


/*
 *  isDataInstruction(instr) -- returns true if the instrucion has
 *				the INSTR_DATA_BYTE flag set.
 */
bool isDataInstruction(Instruction *instr) {
    return INSTR_HAS_FLAG(instr, INSTR_DATA_BYTE);
}


/*
 *  isNOPClass(instr) --
 */
bool isNOPClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_NOP);
}



/*
 *  isFPClass(instr) --
 */
bool isFPClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_FLOAT);
}



/*
 *  isLoopClass(instr) --
 */
bool isLoopClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_LOOP);
}



/*
 *  isCallClass(instr) --
 */
bool isCallClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_CALL);
}



/*
 *  isReturnClass(instr) --
 */
bool isReturnClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_RETURN);
}



/*
 *  isLeaveClass(instr) --
 */
bool isLeaveClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_LEAVE);
}



/*
 *  isLeaClass(instr) --
 */
bool isLeaClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_LEA);
}



/*
 *  isAddClass(instr) --
 */
bool isAddClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_ADD);
}


/*
 *  isAndClass(instr) --
 */
bool isAndClass(Instruction *instr) {
    CHKFLAGS();
    return (instr->i_class == CLASS_AND) ? true : false;
}


/*
 *  isSubClass(instr) --
 */
bool isSubClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_SUB);
}



/*
 *  isHaltClass(instr) --
 */
bool isHaltClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_HALT);
}



/*
 *  isMovClass(instr) --
 */
bool isMovClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_MOV);
}



/*
 *  isSyscallClass(instr) --
 */
bool isSyscallClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_SYSCALL);
}



/*
 *  isPopClass(instr) --
 */
bool isPopClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_POP);
}



/*
 *  isPushClass(instr) --
 */
bool isPushClass(Instruction *instr) {
    CHKFLAGS();
    return INSTR_HAS_FLAG(instr, INSTR_CLASS_PUSH);
}



/*
 *  BBLClearQuadwords() -- clear fields that used to be in the "quadwords".
 */
void BBLClearQuadwords() {
    Block *bbl;

    FOREACH_BBL(bbl) {
	bbl->b_tmp = 0;
	bbl->b_stackshot = NULL;
	bbl->b_regshot = NULL;
	bbl->b_chainnext = NULL;
	bbl->b_chainprev = NULL;
	bbl->b_chainhead = NULL;
	bbl->b_headnext = NULL;
	bbl->b_headprev = NULL;
	bbl->b_nexttochain = NULL;
    }

    return;
}



/*
 *  InstrClearTemp() --
 */
void InstrClearTemp() {
    Instruction *instr;

    FOREACH_INSTR(instr) {
	instr->i_temp = 0;
    }
    return;
}



/*
 *  BblAppendInstr(ibbl,iins) --
 */
void BblAppendInstr(Block *ibbl, Instruction *iins) {
    if (ibbl->b_last != NULL) {
	InsertInstrAfterInstr(iins, ibbl->b_last);
    } else {
	/* empty bbl, so set up accordingly */
	ibbl->b_last = ibbl->b_first = iins;
    }
    iins->i_Bblock = ibbl;
}



/*
 *  CountInstructionsExecuted() --
 */
long long CountInstructionsExecuted() {
    long long count = 0;
    Instruction *instr;

    FOREACH_INSTR(instr) {
	count += instr->i_Bblock->b_weight;
    }
    return count;
}



/*
 *  GetSymbolForSectionAndValue(section,value) --
 */
Symbol *GetSymbolForSectionAndValue(Section *section, ADDRESS value) {
    Symbol *symiterator;

    FOREACH_SYMBOL(symiterator) {
	if ((symiterator->sy_section == section) &&
	    (symiterator->sy_value == value)) {
	    return symiterator;
	}
    }

    return NULL;
}



/*
 *  AddressToSymbol(addr,name) --
 *
 *  Returns the symbol that resides at a given address.
 *  If more than one symbol resides at that address
 *  name may be used to distinguish them.
 */
Symbol *AddressToSymbol(ADDRESS addr, char *name) {
    Symbol *symbol = NULL;
    Section *sec = AddressToSection(addr);
    if (sec == NULL) {
	return NULL;
    }

    FOREACH_SYMBOL(symbol) {
	if (symbol->sy_section == sec && symbol->sy_value == addr) {
	    if (name != NULL && strcmp(symbol->sy_name, name) == 0) {
		return symbol;
	    }
	}
    }

    return NULL;
}



/*
 *  nameToFunction(name) --
 */
Function *nameToFunction(char *name) {
    Function *fun;

    FOREACH_FUNCTION(fun) {
	if (strcmp(fun->f_name, name) == 0) {
	    return fun;
	}
    }
    return NULL;
}



/*
 *  SectionNameToSection(sections,name) --
 */
Section *SectionNameToSection(Section *sections, char *name) {
    Section *siterator;

    FOREACH_SECTION_IN_LIST(sections, siterator) {
	if (strcmp(siterator->s_name, name) == 0) {
	    return siterator;
	}
    }

    return NULL;
}



/*
 *  GetContainingSection(address,useNewAddress) --
 */
Section *GetContainingSection(ADDRESS address, bool useNewAddress) {
    Section *siterator;
    ADDRESS begin, end;

    FOREACH_SECTION(siterator) {
	begin = (useNewAddress) ?
	    siterator->s_newaddress : siterator->s_address;
	end = (useNewAddress) ?
	    siterator->s_newaddress + siterator->s_newdatasize :
	    siterator->s_address + siterator->s_datasize;

	if (address >= begin && address < end) {
	    return siterator;
	}
    }

    return NULL;
}



/*
 *  DumpAllSections() -- print one line of statistics for every section.
 */
void DumpAllSections() {
    Section *siterator;

    REPORT(RP_SMLOOP, "%-25s %9s %9s %9s %9s",
	"Section Name", "Org Addr", "New Addr", "Og Size", "New Size");
    FOREACH_SECTION(siterator) {
	REPORT(RP_SMLOOP, "%-25s %9x %9x %9d %9d",
	    siterator->s_name,
	    siterator->s_address,
	    siterator->s_newaddress,
	    siterator->s_datasize, siterator->s_newdatasize);
    }
}



/*
 *  OperandOffset(instr,op) --
 */
int OperandOffset(Instruction *instr, Operand *op) {
    UINT8 buf[16], buf2[16];
    int len, i;

    if (op == NULL) {
	return -1;
    }

    /*
     * To find out the offset of the source operand for the destination
     * instruction we first put a known value into the operand, then
     * assemble it into a buffer, then set the operand to a value where
     * each byte differs from the original value and assemble it into
     * a second buffer. We then to a byte by byte comparison on the two
     * buffers until we find a difference and that must be the beginning
     * of the operand since the rest of the instruction will have to remain
     * the same.
     */

    len = InstrAssemble(buf, instr);

    op->a_displacement++;
    len = InstrAssemble(buf2, instr);

    for (i = 0; i < len; i++) {
	if (buf[i] != buf2[i]) {
	    break;
	}
    }

    op->a_displacement--;

    return i;
}



/*
 *  GetOperandForOperandAddress(instr,operand_address) --
 */
Operand *GetOperandForOperandAddress(Instruction *instr,
    ADDRESS operand_address) {

    CASSERT(instr != NULL, "Passed a NULL instruction");

    if (instr->i_src != NULL) {
	if (OperandOffset(instr, instr->i_src) + instr->i_address
	== operand_address) {
	    return instr->i_src;
	}
    }

    if (instr->i_src2 != NULL) {
	if (OperandOffset(instr, instr->i_src2) + instr->i_address
	== operand_address) {
	    return instr->i_src2;
	}
    }

    if (instr->i_dest != NULL) {
	if (OperandOffset(instr, instr->i_dest) + instr->i_address
	== operand_address) {
	    return instr->i_dest;
	}
    }

    return NULL;
}



/*
 *  GetOperandForInstrOffset(instr,offset) --
 */
Operand *GetOperandForInstrOffset(Instruction *instr, int offset) {
    if (OperandOffset(instr, instr->i_src) == offset) {
	return instr->i_src;
    }

    if (OperandOffset(instr, instr->i_src2) == offset) {
	return instr->i_src2;
    }

    if (OperandOffset(instr, instr->i_dest) == offset) {
	return instr->i_dest;
    }

    return NULL;
}



/*
 *  AddNewSymbolForInstr(name,instr,flags,after) --
 */
Symbol *AddNewSymbolForInstr(char *name, Instruction *instr, UINT32 flags,
    Symbol *after) {
    Symbol *sym;

    sym = zalloc(sizeof(Symbol));
    sym->sy_name = salloc(name);
    sym->sy_flags = flags;
    sym->sy_section = SECTION_REF_CODE;

    AddPatchInstrAddressIntoSymbol(instr, sym);
    InsertSymbolAfterSymbol(sym, after ? after : s_list);

    return sym;
}



/*
 *  AddNewFunctionSymbol(fun) --
 */
Symbol *AddNewFunctionSymbol(Function *fun) {
    Symbol *sym;

    sym = zalloc(sizeof(Symbol));
    sym->sy_name = salloc(fun->f_name);
    sym->sy_flags = SYMBOL_FUNCTION_FLAG | SYMBOL_GLOBAL_FLAG;
    sym->sy_section = SECTION_REF_CODE;

    AddPatchFunctionAddressIntoSymbol(fun, sym);
    InsertSymbolAfterSymbol(sym, s_list);

    return sym;
}



/*
 *  AddNewObjectSymbol(name,section,offset,flags) --
 */
Symbol *AddNewObjectSymbol(char *name, Section *section, ADDRESS offset,
    UINT32 flags) {
    Symbol *sym;

    sym = zalloc(sizeof(Symbol));
    sym->sy_name = salloc(name);
    sym->sy_flags = flags;
    sym->sy_section = section;

    AddPatchSectionAddressIntoSymbol(section, offset, sym);
    InsertSymbolAfterSymbol(sym, s_list);

    return sym;
}



/*
 *  AddBasicBlockSymbols() --
 */
void AddBasicBlockSymbols() {
    Block *biterator;
    Symbol *siterator, *prev;
    char name[256];

    FOREACH_SYMBOL(siterator) {
	prev = siterator;		// find last symbol
    }

    FOREACH_BBL(biterator) {
	if (biterator->b_first != NULL) {
	    sprintf(name, "z%04d", biterator->b_id);
	    prev = AddNewSymbolForInstr(name,
		biterator->b_first,
		SYMBOL_GLOBAL_FLAG | SYMBOL_FUNCTION_FLAG, prev);
	}
    }
}



/*
 *  GetSymbolSectionName(symbol) --
 */
char *GetSymbolSectionName(Symbol *symbol) {
    if (symbol->sy_section && !(symbol->sy_flags & SYMBOL_COMMON_FLAG)) {
	return symbol->sy_section->s_name;
    }
    if (symbol->sy_flags & SYMBOL_ABSOLUTE_FLAG) {
	return "Absolute";
    }
    if (symbol->sy_flags & SYMBOL_COMMON_FLAG) {
	return "Common";
    }
    if (symbol->sy_flags & SYMBOL_UNDEFINED_FLAG) {
	return "Undefined";
    }
    return "I don't know";
}



/*
 *  GetFlistAtOffset(offset) --
 *
 *  Returns the FList that contains the specified offset from the
 *  flist section.
 *
 *  offset: the offset from the flist section to check each flist
 *  for containment of
 *  returns: the FList that contains the specified offset from the
 *  flist section
 *  see also: SECTION_REF_FLISTS
 */
FList *GetFlistAtOffset(ADDRESS offset) {
    FList *fliterator, *container;

    container = NULL;
    FOREACH_FLIST(fliterator) {
	if (fliterator->fl_new_offset >= offset) {
	    if (container == NULL) {
		container = fliterator;
	    } else if (fliterator->fl_new_offset > container->fl_new_offset) {
		container = fliterator;
	    }
	}
    }

    return container;
}



/*
 *  GetStrongestSymbolForValue(value) --
 *
 *  Returns the Symbol whose value matches the specified value,
 *  guaranteeing that if there is another symbol with the same value, the
 *  one with the strongest type is returned.
 *
 *  value: the value the target symbol should have
 *  returns: a pointer to the Symbol whose value matches value and
 *  whose type is the strongest possible
 */
Symbol *GetStrongestSymbolForValue(ADDRESS value) {
    Symbol *symiterator, *symbol;

    symbol = NULL;

    FOREACH_SYMBOL(symiterator) {
	if (symiterator->sy_value == value) {
	    // we want the function to have the global name as opposed
	    // to the weak one if we can
	    if (symbol == NULL) {
		symbol = symiterator;
	    } else if ((symiterator->sy_flags & SYMBOL_LOCAL_FLAG) &&
		(symbol->sy_flags & SYMBOL_WEAK_FLAG)) {
		symbol = symiterator;
	    } else if (symiterator->sy_flags & SYMBOL_GLOBAL_FLAG) {
		return symiterator;
	    }
	}
    }

    return symbol;
}



/*
 *  WriteAddressTranslationFile(filename) --
 *
 *  This function writes out a file that contains a mapping of old
 *  addresses to new addresses for each instruction in the binary.  The
 *  mappings appear in plain text in the form 0xXXXXXXXX -> 0xXXXXXXXX.
 *
 *  filename: the name of the output file
 */
void WriteAddressTranslationFile(char *filename) {
    Instruction *iiterator;
    FILE *file;

    REPORT(RP_MINOR, "Writing address translation file: %s", filename);

    file = fopen(filename, "w+");
    if (file == NULL) {
	PFATAL(filename);
    }

    FOREACH_INSTR(iiterator) {
	fprintf(file, "%08x -> %08x\n", iiterator->i_address,
	    iiterator->i_final_address);
    }

    fclose(file);
}



/*
 *  SectionCreate() --
 *
 *  Creates an returns a new Section with all of it's fields set to zero
 *  values whose id number is the next available section id number, and
 *  increments the section id counter.
 *  returns: the new section
 */
Section *SectionCreate() {
    Section *section;
    static int section_ids = 0;

    section = zalloc(sizeof(Section));
    section->s_data = section->s_newdata = NULL;
    section->s_datasize = section->s_newdatasize = 0;
    section->s_address = section->s_newaddress = 0;
    section->s_name = NULL;
    section->s_alignment = 0;
    section->s_flags = 0;
    section->s_growth = 0;
    section->s_section_id = section_ids++;

    return section;
}



/*
 *  WriteGDBScript(file_name,extra_info_fun) --
 */
void WriteGDBScript(char *file_name, void (*extra_info_fun) (FILE *file)) {
    FILE *gdb_file;

    REPORT(RP_MINOR, "Writing GDB script file: %s", file_name);

    gdb_file = fopen(file_name, "w");
    if (gdb_file == NULL) {
	PFATAL(file_name);
    }

    fprintf(gdb_file, "file tr.cml\n");
    fprintf(gdb_file, "set confirm off\n");
    fprintf(gdb_file, "set verbose off\n");
    fprintf(gdb_file, "set pagination off\n");
    fprintf(gdb_file, "display /x $eip\n");
    fprintf(gdb_file, "#break at the first instruction\n");
    fprintf(gdb_file, "b *0x%x\n",
	startins->i_final_address + startins->i_length);
    fprintf(gdb_file, "run\n");

    if (extra_info_fun != NULL) {
	extra_info_fun(gdb_file);
    }

    fclose(gdb_file);
}

