/*
 * A module for code compaction via procedural abstraction and "tail-merging."
 * Author: Saumya Debray
 */

#include "global.h"
#include "bbl.h"
#include "clone.h"
#include "flowgraph.h"
#include "opt.h"
#include "opt.unreachable.h"
#include "stack.analyze.h"
#include "sysdep.h"
#include "linkage.h"
#include "upm.h"
#include "util.h"
#include "utilities.h"
#include "x86-classes.h"
#include "x86-registers.h"

#define FNAME_LEN 1024

/*
 * A hash table of basic blocks.  Basic blocks are distributed amongst this
 * hash table so that two blocks that are in different hash buckets are
 * guaranteed to not be identical.  This is done to reduce the cost of the 
 * O(n^2) comparisons amongst blocks to determine which ones are, in fact,
 * identical.
 */
#define HASHTBLSZ   1024
#define HASHMASK    0x3ff          /* log_2(HASHTBLSZ) = 10 bits */

Block *bbl_hash[HASHTBLSZ];
Block *btmp;

/*
 * A buffer for creating new file names during procedural abstraction.
 */
char buf[FNAME_LEN];

/*
 * the address of a global to hold return addresses.
 */
static ADDRESS retAddr = 0;

/*
 * Linked lists of lists of basic blocks.  Each element on the list points
 * to a linked list of basic blocks (chained using their b_chainprev and
 * b_chainnext fields), each such list consisting of a set of identical
 * blocks.
 */
typedef struct bs {
    Block *bbl;
    struct bs *next;
} bbl_set;

bbl_set *bbl_clone_set = NULL;

int tailmerge_instr_deleted = 0;
int procabs_instr_deleted = 0;
int bbls_merged = 0;
int ret_bbls_merged = 0;
int bbls_procabs = 0;
int newfun_procabs = 0;
/*
 * nullify_bbl_chain_fields() --
 *
 * nullify_bbl_chain_fields sets the fields b_chainhead, b_chainprev,
 * b_chainnext, and b_nexttochain of each basic block to NULL.  We do 
 * this because we "borrow" these fields for our purposes; since they 
 * aren't used until the layout phase, this ought to be OK.
 */
static void nullify_bbl_chain_fields() {
    Block *b0;

    FOREACH_BBL(b0) {
        b0->b_chainhead = NULL;
        b0->b_chainprev = NULL;
        b0->b_chainnext = NULL;
        b0->b_nexttochain = NULL;
    }
}



/*
 * init_tail_merge() --
 * 
 * Initialized data structures for tail merging.
 */
static void init_tail_merge() {
    int i;
    Block *b0;
    Instruction *i0;
    Operand *op;

    for (i = 0; i < HASHTBLSZ; i++) {  /* initialize */
        bbl_hash[i] = NULL;
    }
  
    FOREACH_BBL(b0) {                  /* initialize */
        BBL_REMOVE_FLAG(b0, BBL_MARKED_FLAG);

        if (b0->b_type == BT_BRANCH || b0->b_type == BT_UNCOND) {
            i0 = b0->b_last;
            op = i0->i_src;
            op->a_displacement = 0;
        }
    }

    return;
}



/*
 * init_procabs() --
 * 
 * cleans up after TailMerge() and sets things up for procedural
 * abstraction of identical blocks.
 */
static void init_procabs() {
    int i;
    Block *b0;
    bbl_set *bs, *bs_next;

    for (i = 0; i < HASHTBLSZ; i++) {  /* initialize */
        bbl_hash[i] = NULL;
    }
  
    FOREACH_BBL(b0) {                  /* initialize */
        BBL_REMOVE_FLAG(b0, BBL_MARKED_FLAG);

        b0->b_chainhead = NULL;
        b0->b_chainprev = NULL;
        b0->b_chainnext = NULL;
        b0->b_nexttochain = NULL;
    }

    bs = bbl_clone_set;
    while (bs != NULL) {
        bs_next = bs->next;
        free(bs);
        bs = bs_next;
    }

    bbl_clone_set = NULL;
}



/*
 * BBl_compute_fingerprint(bbl) --
 *
 * Computes a "fingerprint" for a basic block that is then used to identify
 * "similar" blocks for code compaction, and uses this fingerprint to add
 * the block to a hash table of blocks to be considered for code compaction.
 *
 * In essence, a fingerprint is just a 32-bit hash value computed for blocks,
 * such that blocks with different fingerprints are guaranteed to have 
 * different instruction sequences.  It has the following structure:
 *
 *   top 8 bits: 4 bits: block type; 4 bits: no. of instrs in the block
 *                         (saturating at 15);
 *   remaining 24 bits: opcodes from the first 6 instructions in the
 *            block (each opcode encoded in 4 bits).
 */
static void BBl_compute_fingerprint(Block *b0) {
    Instruction *ins0;
    unsigned long fpt, hashval;
    int bits_0_to_9, bits_10_to_19, bits_20_to_29;
    int val, i, shift;
  
    fpt = 0;
    
    val = BBLNumberOfInstructions(b0);
    if (val > 15) {
        val = 15;
    }
    val |= ((b0->b_type & 0xf) << 4);
    fpt |= val << 24;

    i = 6;  /* we look at the first 6 instructions only */
    FOREACH_INSTR_IN_BBL(b0, ins0) {
        val = (ins0->i_class & 0xf);
        shift = 4 * (i - 1);
        fpt |= val << shift;
        i--;
        if (i == 0) {
            break;
        }
    }

    b0->b_fprint = fpt;
    /*
     * Add to the hash table of basic blocks.  We compute the hash value of
     * a block by computing the XOR of bits 0-9, 10-19, and 20-29 of its
     * fingerprint.
     */
    bits_0_to_9 = fpt & 0x3ff;
    bits_10_to_19 = (fpt >> 10) & 0x3ff;
    bits_20_to_29 = (fpt >> 20) & 0x3ff;
    hashval = bits_0_to_9 ^ bits_10_to_19 ^ bits_20_to_29;

    b0->b_nexttochain = bbl_hash[hashval];
    bbl_hash[hashval] = b0;

    return;
}



/*
 * InstrsAreIdentical(ins1,ins2) --
 * 
 * Returns true if instructions ins1 and ins2 are identical.
 */
bool InstrsAreIdentical(Instruction *ins1, Instruction *ins2) {
    if (ins1->i_class != ins2->i_class) {
        return false;
    }
    if (ins1->i_prefixes != ins2->i_prefixes) {
        return false;
    }

    if (ins1->i_src && !ins2->i_src) {
        return false;
    }
    if (ins2->i_src && !ins1->i_src) {
        return false;
    }

    if (ins1->i_src2 && !ins2->i_src2) {
        return false;
    }
    if (ins2->i_src2 && !ins1->i_src2) {
        return false;
    }

    if (ins1->i_dest && !ins2->i_dest) {
        return false;
    }
    if (ins2->i_dest && !ins1->i_dest) {
        return false;
    }

    if (ins1->i_src && !OperandsAreIdentical(ins1->i_src, ins2->i_src)) {
        return false;
    }
    if (ins1->i_src2 && !OperandsAreIdentical(ins1->i_src2, ins2->i_src2)) {
        return false;
    }
    if (ins1->i_dest && !OperandsAreIdentical(ins1->i_dest, ins2->i_dest)) {
        return false;
    }
  
    return true;
}



/*
 * BBLsAreIdentical(bbl1, bbl2) --
 *
 * Returns true if bbl1 and bbl2 are identical.  Two blocks are identical
 * if they have the same sequence of instructions.
 */
bool BBLsAreIdentical(Block *bbl1, Block *bbl2) {
    Instruction *ins1, *ins2;

    if (BBLNumberOfInstructions(bbl1) != BBLNumberOfInstructions(bbl2)) {
        return false;
    }
    /*
     * since we know that bbl1 and bbl2 have the same number of
     * instructions, it's enough to check that ins1 remains within
     * bbl1 -- that implies that ins2 is witin bbl2.
     */
    ins1 = bbl1->b_first;
    ins2 = bbl2->b_first;
    while (ins1 && ins1->i_Bblock == bbl1) {
        if (!InstrsAreIdentical(ins1, ins2)) {
            return false;
        }
        ins1 = ins1->i_next;
        ins2 = ins2->i_next;
    }

    return true;
}



/*
 * BblComputeClones() associates, with each basic block, the set of
 * blocks that are identical to this block.  We "borrow" the b_chainprev
 * and b_chainnext fields of basic blocks, since these aren't used until
 * the layout phase.
 */
static void BblComputeClones() {
    Block *b0, *b1, *btmp;
    bbl_set *bs;
    int i;

    FOREACH_BBL(b0) {
        if (b0->b_type == BT_INVALID
                || b0->b_type == BT_HELL
                || b0->b_type == BT_ENTRY
                || b0->b_type == BT_EXIT
                || b0->b_type == BT_HALT
                || BBLNumberOfInstructions(b0) < 2) {
            continue;
        }

        BBl_compute_fingerprint(b0);
    }

    for (i = 0; i < HASHTBLSZ; i++) {
        for (b0 = bbl_hash[i]; b0; b0 = b0->b_nexttochain) {
            for (b1 = b0->b_nexttochain; b1; b1 = b1->b_nexttochain) {
                /*
                 * since b1 iterates over the basic blocks that come after b0,
                 * if b1 is already part of some chain of blocks C then it
                 * must have been found to be identical to some prior basic 
                 * block B prior to b0. But in this case, b0 and b1 can be 
                 * identical only if b0 is also identical to B, in which case 
                 * b1 is already in B's chain and so a comparison between b0 
                 * and b1 is unnecessary.  In other words, if b1 is already 
                 * part of a chain, we can skip the check and go on.
                 */
                if (b1->b_chainnext || b1->b_chainprev) {
                    continue;
                }
    
                if (b0 == b1 
                    || b0->b_fprint != b1->b_fprint 
                    || !BBLsAreIdentical(b0, b1)) {
                        continue;
                }
                /*
                 * b0 and b1 are identical, and will be chained together using
                 * the b_chainprev and b_chainnext fields.  
                 *
                 * We'll arbitrarily adopt the convention that (the chain 
                 * starting with) b1 will be set to point to the chain starting
                 * with b0.  For this, we find the end of the chain starting 
                 * with b0 and add b1 to its beginning.
                 */
                btmp = b0; 
                while (btmp != NULL && btmp->b_chainprev != NULL) {
                    btmp = btmp->b_chainprev;
                }

                b1->b_chainnext = btmp;
                btmp->b_chainprev = b1;
            }
        }
    }
    /*
     * finally, set the b_chainhead field of each block to point to the start
     * of its chain.  This is to simplify subsequent checking of whether two
     * arbitrary blocks lie on the same chain or not (i.e., have identical
     * instruction sequences).
     */
    FOREACH_BBL(b0) {
        if (b0->b_chainprev == NULL && b0->b_chainnext == NULL) {
            continue;
        }

        if (b0->b_chainprev == NULL) {
            b0->b_chainhead = b0;
            /*
             * update the set of clone-sets.
             */
            bs = alloc(sizeof(struct bs));
            bs->next = bbl_clone_set;
            bbl_clone_set = bs;
            bs->bbl = b0;
        } else {    /* find chain head */
            b1 = b0->b_chainprev;
            while (b1 && b1->b_chainprev) {
                b1 = b1->b_chainprev;
            }
            b0->b_chainhead = b1;
        }
    }

    return;
}    



/*
 * Edges_are_congruent(e0, e1) --
 *
 * Returns true if e0 and e1 are congruent, i.e., the basic blocks they
 * point to have been inferred to be congruent.
 */
static bool Edges_are_congruent(Edge *e0, Edge *e1) {
    Block *b_succ0, *b_succ1;

    b_succ0 = e0->e_to;
    b_succ1 = e1->e_to;
    if (!(b_succ0->b_chainhead == b_succ1->b_chainhead
        && BBL_HAS_FLAG(b_succ0, BBL_MARKED_FLAG)
        && BBL_HAS_FLAG(b_succ1, BBL_MARKED_FLAG))) {
            return false;
    }

    return true;
}



/*
 * Bbls_are_congruent(b0, b1) --
 *
 * Returns true if b0 and b1 are "congruent" -- that is, (1) b0 and b1
 * are identical; and (2) their corresponding successors are congruent.
 */
static bool Bbls_are_congruent(Block *b0, Block *b1) {
    Edge *e_succ0, *e_succ1;
    /*
     * If b0 and b1 are identical, they'll have the same b_chainhead.
     */
    if (b0->b_chainhead == NULL 
        || b1->b_chainhead == NULL
        || b0->b_chainhead != b1->b_chainhead) {
            return false;
    }
    /*
     * At this point we know that b0 and b1 have identical instruction
     * sequences.  Next we need to find out whether their corresponding
     * successors are (known to be) congruent.
     */
    switch (b0->b_type) {
    case BT_UNCOND:
        e_succ0 = BBLGetUncondEdge(b0);
        e_succ1 = BBLGetUncondEdge(b1);
    
        if (e_succ0 == NULL || e_succ1 == NULL) {  /* at least one indirect */
            return e_succ0 == e_succ1;        /* both must be indirect */
        }
        /*
         * At this point, both jumps are direct.
         */
        return Edges_are_congruent(e_succ0, e_succ1);

    case BT_NORMAL:
        return Edges_are_congruent(BBLGetNormalEdge(b0), BBLGetNormalEdge(b1));

    case BT_BRANCH:
        return (Edges_are_congruent(BBLGetTrueEdge(b0), BBLGetTrueEdge(b1))
            && Edges_are_congruent(BBLGetFalseEdge(b0), BBLGetFalseEdge(b1)));

    case BT_RETURN:
        return b0->b_chainhead == b1->b_chainhead;

    case BT_SWITCH:
        WARNING("Compaction: congruence of BT_SWITCH blocks not implemented");
        return false;

    case BT_CALL:
        if (!Edges_are_congruent(BBLGetFunlinkEdge(b0), 
                                BBLGetFunlinkEdge(b1))) {
            return false;
        }

        e_succ0 = BBLGetFunDirectEdge(b0);
        e_succ1 = BBLGetFunDirectEdge(b1);
        if (e_succ0 == NULL || e_succ1 == NULL) {  /* at least one indirect */
            return e_succ0 != e_succ1;     /* both must be indirect */
        }
        /*
         * At this point, both are direct calls.
         */
        return Edges_are_congruent(e_succ0, e_succ1);

    case BT_SYSCALL:
        e_succ0 = BBLGetSyscallLinkEdge(b0);
        e_succ1 = BBLGetSyscallLinkEdge(b1);
        if (!e_succ0 || !e_succ1) {     /* indirect call */
            return false;
        }

        return Edges_are_congruent(e_succ0, e_succ1);

    default:
        return false;
    }

    return false;
}



/*
 * BblDoesNotUseSP(bbl) --
 *
 * returns true if bbl does not use or define the stack pointer.
 */
static bool BblDoesNotTouchSP(Block *bbl) {
    Instruction *ins;
    Operand *op;
    /*
     * An instruction touches the stack pointer if it pushes/pops 
     * the stack.
     */
    FOREACH_INSTR_IN_BBL(bbl, ins) {
        if (ins->i_class == CLASS_PUSH
            || ins->i_class == CLASS_POP
            || ins->i_class == CLASS_ENTER
            || ins->i_class == CLASS_LEAVE
            || ins->i_class == CLASS_PUSHALL
            || ins->i_class == CLASS_POPALL
            || ins->i_class == CLASS_PUSHF
            || ins->i_class == CLASS_PUSHFD
            || ins->i_class == CLASS_POPF
            || ins->i_class == CLASS_POPFD) {
                return false;
        }
        /*
         * An instruction touches the stack pointer if any of its
         * operands mentions the stack pointer.
         */      
        op = ins->i_src;
        if (op != NULL) {
            if ((OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG) 
                    && RegisterIsStackPointer(op->a_base_reg))
                || (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)
                    && RegisterIsStackPointer(op->a_index_reg))) {
                return false;
            }
        }

        op = ins->i_src2;
        if (op != NULL) {
            if ((OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG) 
                    && RegisterIsStackPointer(op->a_base_reg))
                || (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)
                    && RegisterIsStackPointer(op->a_index_reg))) {
                return false;
            }
        }

        op = ins->i_dest;
        if (op != NULL) {
            if ((OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG) 
                    && RegisterIsStackPointer(op->a_base_reg))
                || (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)
                    && RegisterIsStackPointer(op->a_index_reg))) {
                return false;
            }
        }
    }

    return true;
}



/*
 * redirect_edges(b0, b1) -- redirect control from b1 into b0.
 *
 * Replaces the instructions in b1 by an unconditional jump to b0.
 */
static void redirect_edges(Block *b0, Block *b1) {
    Edge *e0, *e1;
    Block *b_exit0, *b_exit1;
    Instruction *ins0, *ins1, *nextins;

    ins0 = InstrCreateJMP();
    ins1 = b1->b_first; 
    InsertInstrBeforeInstr(ins0, ins1);

    if (InstrIsUPMSource(ins1)) {
        UPMChangeAllSourcesFromInstrToInstr(ins1, ins0);
    }

    tailmerge_instr_deleted -= 1;
  
    while (ins1 && ins1->i_Bblock == b1) {
        nextins = ins1->i_next;
        InstrDestroy(ins1);
        ins1 = nextins;
        tailmerge_instr_deleted += 1;
    }
    b1->b_first = b1->b_last = ins0;
    /*
     * Now fix the outgoing edges from b1 and set its type appropriately.
     * If b0 and b1 are of type BT_RETURN, add a compensation edge from b0's 
     * exit block to b1's exit block.
     */
    if (b0->b_type == BT_RETURN) {
        e0 = BBLGetEdgeTyped(b0, ET_EXIT);
        CASSERT(e0 != NULL, "Could not find ET_EXIT edge");
        b_exit0 = e0->e_to;

        e1 = BBLGetEdgeTyped(b1, ET_EXIT);
        CASSERT(e1 != NULL, "Could not find ET_EXIT edge");
        b_exit1 = e1->e_to;
    }

    BBLRemoveEdges(b1);
    BBLAddEdge(b1, b0, ET_UNCOND);
    b1->b_type = BT_UNCOND;

    if (b0->b_type == BT_RETURN) {
        BBLAddEdge(b_exit0, b_exit1, ET_COMPENSATE);
    }

    BBL_SET_FLAG(b0, BBL_MARKED_FLAG);
    BBL_SET_FLAG(b1, BBL_MARKED_FLAG);

    bbls_merged += 1;
    if (b0->b_type == BT_RETURN) {
        ret_bbls_merged += 1;
    }

    return;
}



/*
 * TailMerge() --
 *
 * Carries out tail-merging on functions.  Tail merging is an optimization 
 * where, given two functions that have identical "tails" (subgraphs leading
 * up to returns from the function), one of them is replaced by a jump to
 * the other.
 */
static void TailMerge() {
    Block *b0, *b1;
    bbl_set *bs;
    int niters, nbbls;
    int chg;

    init_tail_merge();

    BblComputeClones();

    for (chg = 1, niters = 1; chg; niters++) {
        chg = 0;
        nbbls = 0;
        for (bs = bbl_clone_set; bs; bs = bs->next) {
            b0 = bs->bbl;
            CASSERT(b0->b_chainhead == b0, "chainhead doesn't match block");

            for (b1 = b0->b_chainnext; b1; b1 = b1->b_chainnext) {
                if (!BBL_HAS_FLAG(b1, BBL_MARKED_FLAG)
                    && Bbls_are_congruent(b0, b1)) {
                    redirect_edges(b0, b1);
                    chg = 1;
                    nbbls++;
                }
            }
        }

        REPORT(RP_SMLOOP, 
            "TailMerging: iteration %d -- %d blocks merged", niters, nbbls);
    }

    REPORT(RP_DETAIL, 
        "TailMerging: %d iters, %d bbls [%d returns] merged, %d ins deleted", 
        niters - 1,
        bbls_merged, 
        ret_bbls_merged,
        tailmerge_instr_deleted);

    return;
}



/* BBLAbstractToFunCall(bbl, fun) --
 *
 * Given a block bbl and a function fun, deletes all the instructions 
 * in bbl and replaces them with a single CALL instruction to fun.
 */
static void BBLAbstractToFunCall(Block *bbl, Function *fun) {
    Block *b1;
    Edge *e0;
    Instruction *i0, *i1, *i_call;
    /*
     * create a CALL instruction and add it to bbl.
     */
    i0 = bbl->b_first; 
    i_call = InstrCreateCALLDisp();
    InsertInstrBeforeInstr(i_call, i0);
    if (InstrIsUPMSource(i0)) {
        UPMChangeAllSourcesFromInstrToInstr(i0, i_call);
    }
    /*
     * delete all the instructions in the block except for the CALL.
     */
    while (i0 != NULL && i0->i_Bblock == bbl) {
        i1 = i0->i_next;
        InstrDestroy(i0);
        i0 = i1;
    }
    /*
     * fix the control flow edges from b_iter.
     */
    e0 = BBLGetNormalEdge(bbl);  /* control returns to b1 */
    b1 = e0->e_to;
    bbl->b_type = BT_CALL;
    BBLRemoveEdges(bbl);
    BBLAddEdge(bbl, b1, ET_FUNLINK);
    BBLAddEdge(bbl, fun->f_entry, ET_FUNCALL);
    BBLAddEdge(bbl, fun->f_first, ET_FUNDIRECT);
    BBLAddEdge(fun->f_exit, b1, ET_RETURN);
  
    return;
}



/*
 * ProcAbs_No_SP_use(bbl) --
 *
 * Given a list of identical basic blocks that do not touch the stack pointer,
 * this function carries out procedural abstraction for them.  This is
 * straightforward in this case because all we need to do is to replace the
 * block being abstracted by a single CALL instruction, and add a RET at the
 * end of the original instruction sequence in the callee.  This has the
 * effect of pushing the return address on the stack and thus changing the
 * stack pointer, but since bbl doesn't use the stack pointer this is OK.
 * Note that writes to the stack frame, e.g., through the frame pointer or
 * indirectly through other registers, is OK.
 */
static void ProcAbs_No_SP_use(Block *bbl) {
    Function *f0;
    Block *b0, *b_iter;
    Instruction *i0;
    int nins;
  
    nins = BBLNumberOfInstructions(bbl);
    /*
     * create the abstracted function.
     */
    f0 = FunCreate();
    snprintf(buf, FNAME_LEN, "$proc_%d", bbl->b_id);
    f0->f_name = salloc(buf);
    InsertFunctionAfterFunction(f0, bbl->b_fun);
    newfun_procabs++;
  
    b0 = BBLClone(bbl);
    b0->b_type = BT_RETURN;
    b0->b_fun = f0;
    f0->f_first = f0->f_last = b0;
    procabs_instr_deleted -= nins;   /* from cloning */
  
    i0 = InstrCreateReturn();
    InsertInstrAfterInstr(i0, b0->b_last);
    procabs_instr_deleted--;

    INSTR_SET_FLAG(b0->b_first, FUN_START);
    INSTR_SET_FLAG(i0, FUN_END);
    /*
     * carry out procedural abstraction for each block on bbl's clone list.
     */
    for (b_iter = bbl; b_iter != NULL; b_iter = b_iter->b_chainnext) {
        if (BBL_HAS_FLAG(b_iter, BBL_HOT_FLAG)) {
            continue;
        }

        bbls_procabs++;
        procabs_instr_deleted += nins - 1;  /* -1 because a CALL is added */
        BBLAbstractToFunCall(b_iter, f0);
        b0->b_weight += b_iter->b_weight;
    }

    return;
}



/*
 * BblSize(bbl) --
 *
 * returns the number of bytes occupied by instructions in the given
 * basic block.  In general this is an estimate because the size of 
 * some instructions may depend on operand values that are not known
 * until the final layout of the code has been determined.
 */
static int BblSize(Block *bbl) {
    Instruction *ins;
    int sz = 0;

    FOREACH_INSTR_IN_BBL(bbl, ins) {
        sz += CalculateInstrLength(ins);
    }

    return sz;
}



/*
 * ProcAbs_SP_Use(bbl) --
 *
 * Carries out procedural abstraction of a set of identical basic blocks
 * that may use/define the stack pointer.  For such blocks, a simple
 * call/return pair for control transfer doesn't work.  Instead, we allocate
 * a global location Loc to hold the return address.  The transformed 
 * code is as follows:
 *
 *   at the caller:
 *                   call abstracted_procedure
 *
 *   at the callee:
 *                   pop  Loc         // Loc <- ret addr
 *                   ... original bbl instructions ...
 *                   jmp *Loc         // return to caller
 */
static void ProcAbs_SP_Use(Block *bbl) {
    Function *f0;
    Block *b0, *b_iter;
    Instruction *i0;
    int nins;

    nins = BBLNumberOfInstructions(bbl);
    /*
     * allocate global for return address.
     */
    if (retAddr == 0) {
        retAddr = SECTION_REF_DATA->s_growth + SECTION_REF_DATA->s_datasize;
        IncreaseSectionSize(SECTION_REF_DATA, ADDRESS_SIZE);
    }
    /*
     * create the abstracted function.
     */
    f0 = FunCreate();
    snprintf(buf, FNAME_LEN, "$proc_%d", bbl->b_id);
    f0->f_name = salloc(buf);
    InsertFunctionAfterFunction(f0, bbl->b_fun);
    newfun_procabs++;
  
    b0 = BBLClone(bbl);
    b0->b_type = BT_UNCOND;
    b0->b_fun = f0;
    f0->f_first = f0->f_last = b0;
    procabs_instr_deleted -= nins;     /* from the cloning */
  
    INSTR_SET_FLAG(b0->b_first, FUN_START);

    i0 = InstrCreate32BitPopToMem();           /* pop retAddr */
    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, retAddr, i0, i0->i_dest);
    InsertInstrBeforeInstr(i0, b0->b_first);
    procabs_instr_deleted--;

    i0 = InstrCreateJMPIndirectAbsMemAddr();   /* jmp *retAddr */
    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, retAddr, i0, i0->i_src);

    InsertInstrAfterInstr(i0, b0->b_last);
    procabs_instr_deleted--;

    INSTR_SET_FLAG(i0, FUN_END);
    /*
     * carry out procedural abstraction for each block on bbl's clone list.
     */
    for (b_iter = bbl; b_iter != NULL; b_iter = b_iter->b_chainnext) {
        if (BBL_HAS_FLAG(b_iter, BBL_HOT_FLAG)) {
            continue;
        }

        bbls_procabs++;
        procabs_instr_deleted += nins - 1;  /* -1 because a CALL is added */
        BBLAbstractToFunCall(b_iter, f0);
        b0->b_weight += b_iter->b_weight;
    }

    return;
}



/*
 * ProcAbs_Identical_Blocks() --
 * 
 * Carries out procedural abstraction of identical basic blocks.  First,
 * blocks with PC-changing instructions at the end are split and the
 * PC-changing instruction moved to a new block.  The BT_NORMAL blocks
 * in the program are then grouped into sets of identical blocks, which
 * are then subjected to procedural abstraction.
 */
static void ProcAbs_Identical_Blocks() {
    Block *b0, *b1;
    Instruction *i0;
    bbl_set *bs;
    int nbbls, bblsz, del_bytes, ins_bytes;
  
    init_procabs();
    /*
     * First, split any blocks that have a PC-changing instruction at the
     * end (blocks containing less than two instructions are ignored).
     */
    FOREACH_BBL(b0) {
        if (BBLNumberOfInstructions(b0) < 2) {
            continue;
        }

        if (b0->b_type == BT_UNCOND
            || b0->b_type == BT_BRANCH
            || b0->b_type == BT_SWITCH
            || b0->b_type == BT_CALL
            || b0->b_type == BT_SYSCALL) {
                i0 = b0->b_last;
                CASSERT(isPCChangingClass(i0),
                    "Last instr of b0 is disappointingly not PC-changing");

            BBLDivideBlock(b0, i0);  /* move i0 into a new block */
        }
    }

    BblComputeClones();

    for (bs = bbl_clone_set; bs; bs = bs->next) {
        bblsz = BblSize(bs->bbl);
        for (nbbls = 0, b1 = bs->bbl; b1 != NULL; b1 = b1->b_chainnext) {
            if (!BBL_HAS_FLAG(b1, BBL_HOT_FLAG)) {
                nbbls++;
            }
        }

        if (BblDoesNotTouchSP(bs->bbl)) {
            /*
             * We add one CALL instruction per block, make one copy of each
             * instruction in the block in the abstracted procedure, and add
             * a RET instruction in the abstracted procedure.
             *
             * The next line of code assumes that a CALL instruction occupies
             * 5 bytes (opcode + 4-byte address), and a RET occupies 1 byte.
             */
            del_bytes = nbbls * bblsz;            /* no. of bytes deleted */
            ins_bytes = (nbbls*5) + bblsz + 1;    /* no. of bytes inserted */
    
            if (ins_bytes < del_bytes) {
                ProcAbs_No_SP_use(bs->bbl);
            }
        } else {
            /*
             * We add one CALL instruction per block at the call site.
             * At the callee, in addition to each instruction in the original
             * block, we add two instructions: a POP and an indirect JMP
             * through an absolute memory address.
             *
             * The next line of code assumes that a CALL instruction occupies
             * 5 bytes, a POP into an absolute memory address is 6 bytes, and
             * an indirect JMP through an absolute memory address is 6 bytes.
             */
            del_bytes = nbbls * bblsz;            /* no. of bytes deleted */
            ins_bytes = (nbbls * 5) + bblsz + 6 + 6;    /* bytes inserted */
            if (ins_bytes < del_bytes) {
                ProcAbs_SP_Use(bs->bbl);
            }
        }
    }

    REPORT(RP_DETAIL,
        "Abstraction: %d bbls abstracted into %d functions; %d ins deleted",
        bbls_procabs,
        newfun_procabs,
        procabs_instr_deleted);

    return;
}



/*
 * OptCompact() --
 *
 * Examines the basic blocks in the program and applies procedural
 * abstraction and tail merging to effect code compaction.
 */
void OptCompact() {
    REPORT(RP_MINOR, WITHFUNC("Compacting"));

    nullify_bbl_chain_fields();        /* initialize */

    TailMerge();

    ProcAbs_Identical_Blocks();

    nullify_bbl_chain_fields();  /* cleanup */

    REPORT(RP_MINOR, "Compaction: deleted %d instructions", 
        tailmerge_instr_deleted + procabs_instr_deleted);

    return;
}
