/*
 *  A module for inlining function calls. Most decisions about inlining
 *  are made by the specific architecture which defines the functions
 *  FunCanBeInlinedFromCallSite(), FunInline() and FunIsInlinable().
 */
#include "global.h"
#include "opt.h"
#include "opt.layout.h"
#include "linkage.h"
#include "sysdep.h"
#include "flowgraph.h"
#include "utilities.h"
#include "hash.h"
#include "relocate.h"
#include "cache.analyze.h"
#include "opt.unreachable.h"
#include "stack.analyze.h"
#include "upm.h"
#include "cp.h"
#include "dominator.h"
#include "potential.h"
#include "x86-classes.h"
#include "clone.h"



static double HOT_SET_CALL_SITE_THRESHOLD;
static double HOT_INSTRS_NEEDED_TO_INLINE;



/*
 *  Inline
 */
typedef struct Inline {
    Instruction *in_callInstr;
    Function *in_callee;
    struct Inline *in_prev;
    struct Inline *in_next;
    double in_weight;
    UINT32 in_id;
} Inline;



static EDGE_WEIGHT_TYPE PotentialBenefitForInlining(
    Instruction *instr, Function *callee);
static EDGE_WEIGHT_TYPE PotentialBenefitForInliningAdjusted(
    Instruction *instr, Function *callee);
static EDGE_WEIGHT_TYPE PotentialBenefitForInliningFromPushAnalysis(Instruction
    *instr, Function *callee);

static int NumberOfConstantArguments(Instruction *instr, Function *fun);

static int inlineCounter;
static int hotInstrWeight;
static void BBLMarkHotCallSites();
static void InlineFun(Inline *i);
static void AddInline(Instruction *callInstr, Function *fun, double weight);
static void RemoveInline(Inline *i);
static bool InlineCalleeFromCallsite(Block *callsite);
static long long numFunCalls;
static long long originalCalls;
static long long newCalls;
static int numInlined;
static Inline *inlineList;



/*
 *  OptInlineFuns() --
 *
 *  Examines all callsites in the program, and inlines those functions
 *  which the backend determines should be inlined.  Note that the decision
 *  can be dependent on the architecture and the cost of calling conventions.
 */
void OptInlineFuns() {
    Function *funs;
    Block *biterator;
    bool funsDestroyed;
    int eligible, extinct, endangered, thriving;

    REPORT(RP_MINOR, WITHFUNC("Inlining functions"));

    HOT_SET_CALL_SITE_THRESHOLD = FloatParameter("InliningHotThreshold");
    HOT_INSTRS_NEEDED_TO_INLINE = IntegerParameter("HotInstrsNeededToInline");

    numInlined = 0;

    numFunCalls = CountFuncalls();
    originalCalls = numFunCalls;

    REPORT(RP_DETAIL, "Before inlining: %lld function calls", numFunCalls);

    if (numFunCalls == 0) {
	return;
    }

    /*
     * First mark all the functions that can be inlined.  Whether or not a
     * function can be inlined is the decision of the back-end, as it will
     * also be doing the inlining process.
     */

    eligible = 0;
    FOREACH_FUNCTION(funs) {
	if (!FUNCTION_HAS_FLAG(funs, FUNCTION_WELL_BEHAVED_FLAG)) {
	    continue;
	}
	if (FUNCTION_HAS_FLAG(funs, FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG)) {
	    continue;
	}
	if (FunIsInlinable(funs)) {
	    FUNCTION_SET_FLAG(funs, FUNCTION_CAN_BE_INLINED_FLAG);
	    eligible++;
	}
    }
    REPORT(RP_DETAIL, "%6d functions eligible for inlining", eligible);

    /*
     * Sort all the callsites by execution frequency (count), and mark hot
     * callsites as such.
     */

    BBLMarkHotCallSites();

    hotInstrWeight = 0;
    FOREACH_BBL(biterator) {
	if (BBL_HAS_FLAG(biterator, BBL_HOT_FLAG)) {
	    if (biterator->b_weight < hotInstrWeight) {
		hotInstrWeight = biterator->b_weight;
	    }
	    if (hotInstrWeight == 0) {
		hotInstrWeight = biterator->b_weight;
	    }
	}
    }

    FOREACH_BBL(biterator) {
	if (biterator->b_type == BT_CALL) {
	    if (InlineCalleeFromCallsite(biterator)) {
		Edge *direct = BBLGetFunDirectEdge(biterator);
		int pot = PotentialBenefitForInliningAdjusted(biterator->b_last,
		    direct->e_to->b_fun);
		AddInline(biterator->b_last,
		    direct->e_to->b_fun, direct->e_edgeWeight + pot);
	    }
	}
    }

    REPORT(RP_SMLOOP,
	"Callsite Instrs     Calls   Benefit    %%  B1   CA/A   Callee");
    while (inlineList != NULL) {
	InlineFun(inlineList);
    }

    newCalls = CountFuncalls();
    REPORT(RP_DETAIL, "After inlining: %lld function calls", newCalls);
    if (newCalls != originalCalls) {
	REPORT(RP_DETAIL, "  %.0f%% reduction in number of function calls",
	    (float) 100.0 * (originalCalls - newCalls) / originalCalls);
    }

    /*
     * Over the course of inlining, some functions may have become extinct.
     * If a function no longer has any callsites, we will destroy it.
     */
    extinct = endangered = thriving = 0;
    FOREACH_FUNCTION(funs) {
	if (BBLNumberOfPredecessors(funs->f_entry) > 1) {
	    thriving++;
	} else if (BBLNumberOfPredecessors(funs->f_entry) == 0) {
	    extinct++;
	} else {
	    endangered++;
	}
    }

    FOREACH_BBL(biterator) {
	BBL_REMOVE_FLAG(biterator, BBL_LOOP_HEADER_FLAG);
	biterator->b_loopFirst = biterator->b_loopNext =
	    biterator->b_loopHead = NULL;
    }


    funsDestroyed = true;
    while (funsDestroyed) {
	funsDestroyed = false;
	FOREACH_FUNCTION(funs) {
	    if (FunctionIsExtinct(funs)) {
		REPORT(RP_LGLOOP, "now extinct: %s", funs->f_name);
		FunDestroy(funs);
		funsDestroyed = true;
	    }
	}
    }

    REPORT(RP_DETAIL, "%6d functions became extinct", extinct);
    REPORT(RP_DETAIL, "%6d functions are endangered", endangered);
    REPORT(RP_DETAIL, "%6d functions are thriving", thriving);

    return;
}



/*
 *  InlineFun(i) --
 *
 *  Inlines 'fun' from its callsite at 'callInstr'.
 */
static void InlineFun(Inline *i) {
    Instruction *callInstr = i->in_callInstr;
    Function *fun = i->in_callee;
    Function *callFun = INSTR_FUNCTION(callInstr);
    Edge *funDirect = BBLGetFunDirectEdge(callInstr->i_Bblock);
    Inline *anInline;
    Block *biterator;
    float pot;
    char str1[20], str2[20];

    strncpy(str1, INSTR_FUNCTION(callInstr)->f_name, 19);
    strncpy(str2, fun->f_name, 19);
    str1[18] = str2[18] = 0;

    pot = PotentialBenefitForInliningAdjusted(callInstr, fun);

    REPORT(RP_SMLOOP, "%08X %6d %9.0f %9.0f %4.2f  %s%s %3d/%-3d %s",
	callInstr->i_address, FunNumberOfInstructions(fun),
	(float) funDirect->e_edgeWeight, pot,
	funDirect->e_edgeWeight * 100.0 / originalCalls,
	FUNCTION_HAS_FLAG(fun, FUNCTION_BALANCED_STACK_FLAG) ? "Y" : "N",
	FunNumberOfCallSites(fun) == 1 ? "Y" : "N",
	NumberOfConstantArguments(callInstr, fun),
	FunNumberOfArguments(fun), str2);

    HOT_SET_CALL_SITE_THRESHOLD -=
	(funDirect->e_edgeWeight * 1.0) / originalCalls;
    numFunCalls -= funDirect->e_edgeWeight;

    numInlined++;

    FunInline(callInstr);
    UpdateFunction(callFun);

    FOREACH_BBL_IN_FUN(callFun, biterator) {
	if (BBL_HAS_FLAG(biterator, BBL_REGSHOT_FLAG)) {
	    free(biterator->b_regshot);
	}
	if (BBL_HAS_FLAG(biterator, BBL_STACKSHOT_FLAG)) {
	    free(biterator->b_stackshot);
	}
	BBL_REMOVE_FLAG(biterator, BBL_REGSHOT_FLAG);
	BBL_REMOVE_FLAG(biterator, BBL_STACKSHOT_FLAG);
	if (FUNCTION_HAS_FLAG(callFun, FUNCTION_SNAPSHOT_FLAG)) {
	    BBLAllocateStackshot(biterator);
	}
	BBLAllocateRegshot(biterator);
    }

    StoreDominatorInfoForFun(callFun);

    RemoveInline(i);

    anInline = inlineList;
    while (anInline != NULL) {
	Inline *next = anInline->in_next;
	if (INSTR_FUNCTION(anInline->in_callInstr) == callFun) {
	    RemoveInline(anInline);
	} else if (INSTR_FUNCTION(anInline->in_callInstr) == fun) {
	    RemoveInline(anInline);
	}
	anInline = next;
    }

    if (FunIsInlinable(callFun)) {
	FUNCTION_SET_FLAG(callFun, FUNCTION_CAN_BE_INLINED_FLAG);
    }
    if (!FUNCTION_HAS_FLAG(callFun, FUNCTION_GCC_PROLOGUE_FLAG)) {
	FUNCTION_REMOVE_FLAG(callFun, FUNCTION_CAN_BE_INLINED_FLAG);
    }

// if (!FUNCTION_HAS_FLAG(callFun, FUNCTION_BALANCED_STACK_FLAG)) {
//     FUNCTION_REMOVE_FLAG(callFun, FUNCTION_CAN_BE_INLINED_FLAG);
//  }

    FOREACH_BBL_IN_FUN(fun, biterator) {
	if (biterator->b_type == BT_CALL) {
	    if (InlineCalleeFromCallsite(biterator)) {
		Edge *direct = BBLGetFunDirectEdge(biterator);
		int p = PotentialBenefitForInliningAdjusted(biterator->b_last,
		    direct->e_to->b_fun);
		AddInline(biterator->b_last,
		    direct->e_to->b_fun, direct->e_edgeWeight + p);
	    }
	}
    }

    FOREACH_BBL_IN_FUN(callFun, biterator) {
	if (biterator->b_type == BT_CALL) {
	    if (InlineCalleeFromCallsite(biterator)) {
		Edge *direct = BBLGetFunDirectEdge(biterator);
		int p = PotentialBenefitForInliningAdjusted(biterator->b_last,
		    direct->e_to->b_fun);
		AddInline(biterator->b_last,
		    direct->e_to->b_fun, direct->e_edgeWeight + p);
	    }
	}
    }

    FUNCTION_SET_FLAG(callFun, FUNCTION_CP_FLAG);

    return;
}



/*
 *  BBLMarkHotCallSites() --
 *
 *  This function will flag all callsites (granularity of a basic block) that
 *  are considered hot.  That is, they account for more than X percent of the
 *  total calls made in the program.  X can be defined on the command line.
 *  Chains are used to link basic blocks together in a non-increasing order
 *  of hotness. The hottest basic block in the program is returned.
 */
static void BBLMarkHotCallSites() {
    Block *biterator;
    EDGE_WEIGHT_TYPE sum = 0;
    EDGE_WEIGHT_TYPE current = 0;
    int bblCount = 0;
    Block **bblArray;
    int offset, i;

    FOREACH_BBL(biterator) {
	if (biterator->b_type == BT_CALL) {
	    sum = sum + biterator->b_weight;
	    bblCount++;
	}
    }

    bblArray = zalloc(sizeof(Block *) * bblCount);

    offset = 0;
    FOREACH_BBL(biterator) {
	if (biterator->b_type == BT_CALL) {
	    bblArray[offset++] = biterator;
	}
    }

    CASSERT(offset == bblCount, "Block counts did not match.");

    /* Sort the basic block array */
    qsort(bblArray, bblCount, sizeof(Block *), BBLCompare);

    for (i = 0; i < bblCount; i++) {
	bblArray[i]->b_chainnext = NULL;
	bblArray[i]->b_chainprev = NULL;
    }

    for (i = 0; i < bblCount; i++) {
	biterator = bblArray[i];
	if (i != (bblCount - 1)) {
	    biterator->b_chainnext = bblArray[i + 1];
	} else {
	    biterator->b_chainnext = NULL;
	}
	if (i != 0) {
	    biterator->b_chainprev = bblArray[i - 1];
	} else {
	    biterator->b_chainprev = NULL;
	}
    }

    offset = 0;
    while (current < (HOT_SET_CALL_SITE_THRESHOLD * sum)) {
	if (((bblArray[offset])->b_type) == BT_CALL) {
	    BBL_SET_FLAG(bblArray[offset], BBL_HOT_CALL_SITE_FLAG);
	    current = current + ((bblArray[offset])->b_weight);
	}
	offset++;
    }

    CASSERT(offset < bblCount, "Wrote past array bounds.");
    CASSERT(current >= (HOT_SET_CALL_SITE_THRESHOLD * sum),
	"Hot call sites not set appropriately.");

    free(bblArray);

    return;
}



/*
 *  AddInline(callInstr,callee,weight) --
 *
 *  Adds the (callInstr, callee) pair to the list of inlinings to take place.
 */
static void AddInline(Instruction *callInstr, Function *callee, double weight) {
    Inline *iterator = inlineList;
    Inline *anInline = zalloc(sizeof(Inline));

    anInline->in_callee = callee;
    anInline->in_callInstr = callInstr;
    anInline->in_weight = weight;
    anInline->in_id = inlineCounter++;

    if (iterator == NULL) {
	inlineList = anInline;
	return;
    }

    while (iterator->in_weight >= weight) {
	if (iterator->in_next == NULL) {
	    iterator->in_next = anInline;
	    anInline->in_prev = iterator;
	    anInline->in_next = NULL;
	    return;
	} else {
	    iterator = iterator->in_next;
	}
    }

    if (iterator == inlineList) {
	anInline->in_next = inlineList;
	inlineList->in_prev = anInline;
	inlineList = anInline;
    } else {
	anInline->in_next = iterator;
	anInline->in_prev = iterator->in_prev;
	iterator->in_prev = anInline;
	anInline->in_prev->in_next = anInline;
    }

    return;
}



/*
 *  RemoveInline(i) --
 *
 *  Removes a (callInstr, callee) pair from the list of inlinings to take place.
 */
static void RemoveInline(Inline *i) {

    if (inlineList == i) {
	inlineList = i->in_next;
    }
    if (i->in_prev) {
	i->in_prev->in_next = i->in_next;
    }
    if (i->in_next) {
	i->in_next->in_prev = i->in_prev;
    }

    return;
}



/*
 *  InlineCalleeFromCallsite(callsite) --
 *
 *  Returns true if we're going to inline at the callsite.
 */
static bool InlineCalleeFromCallsite(Block *callsite) {
    Edge *funDirect;
    Function *fun;
    Instruction *callInstr;
    int numInstrs, numCallSites;
    EDGE_WEIGHT_TYPE potential;

    /*
     * If a function direct edge doesn't exist, then we have a function call to
     * HELL. We will not bother trying to do anything in this case.
     */

    callInstr = callsite->b_last;

    funDirect = BBLGetFunDirectEdge(callsite);
    if (funDirect == NULL) {
	DEBUG(DB_INFAIL, "[%d] Cannot resolve target of CALL at 0x%8.8x",
	    callInstr->i_id, callInstr->i_address);
	return false;
    }

    /*
     * We shouldn't inline functions that are never called.
     */

    if (funDirect->e_edgeWeight == 0) {
	return false;
    }

    fun = funDirect->e_to->b_fun;

    /*
     * Also don't inline functions that don't have a balanced stack, a gcc
     * prologue and the exit combo flag.
     */

    if (!FUNCTION_HAS_FLAG(fun, FUNCTION_CAN_BE_INLINED_FLAG)) {
	DEBUG(DB_INFAIL,
	    "[%d] CALL at 0x%8.8x calls %s, which cannot be inlined",
	    callInstr->i_id, callInstr->i_address, fun->f_name);
	return false;
    }

    if (InstrBPSPDiff(callInstr) == STACK_UNKNOWN) {
	DEBUG(DB_INFAIL,
	    "[%d] Cannot inline at 0x%8.8x due to unknown stack height",
	    callInstr->i_id, callInstr->i_address);
	return false;
    }

    /*
     * We won't allow recursive calls to be inlined, since this could
     * potentially have disastrous icache effects.
     */

    if (fun == INSTR_FUNCTION(callInstr)) {
	DEBUG(DB_INFAIL,"[%d] Cannot inline recursive function call at 0x%8.8x",
	    callInstr->i_id, callInstr->i_address);
	return false;
    }

    /*
     * Now we know the function can be inlined successfully.
     * We will inline it if any of the following conditions hold:
     *    a) The function has just one callsite
     *    b) The function has fewer than INLINE_ALWAYS_INSTRUCTION_CUTOFF instrs
     *    c) The function is hot enough, and cache effects are acceptable
     */

    numInstrs = FunNumberOfInstructions(fun);
    numCallSites = FunNumberOfCallSites(fun);

    if (numInstrs <= INLINE_ALWAYS_INSTRUCTION_CUTOFF || numCallSites == 1) {
	return true;
    }

    /*
     * If the callsite frequency is not greater than some constant threshold,
     * we won't bother inlining.
     */

    potential = PotentialBenefitForInliningAdjusted(callInstr, fun);
    potential = potential + funDirect->e_edgeWeight * 2;

    if (potential < (hotInstrWeight * HOT_INSTRS_NEEDED_TO_INLINE)) {
	return false;
    }

    CASSERT(isCallClass(callInstr),
	"Last instruction in BT_CALL block not a call instruction.");

    return FunCanBeInlinedFromCallSite(fun, callInstr) &&
	GoodIdeaToInlineFunFromCallsite(fun, callInstr);
}



/*
 *  PotentialBenefitForInliningAdjusted(instr,callee) --
 */
static EDGE_WEIGHT_TYPE PotentialBenefitForInliningAdjusted(Instruction *instr,
    Function *callee) {

    float scale = 0;
    EDGE_WEIGHT_TYPE total;

#if 1
    return 0;
#endif

    total = callee->f_first->b_weight;
    if (total != 0) {
	scale = ((float) instr->i_Bblock->b_weight) / ((float) total);
    }

    return ((EDGE_WEIGHT_TYPE) PotentialBenefitForInlining(instr, callee)) -
	((EDGE_WEIGHT_TYPE) (scale *
	    ComputePotentialAtInstr(FUNCTION_FIRST_INSTR(callee))));
}



/*
 *  PotentialBenefitForInlining(instr,callee) --
 */
static EDGE_WEIGHT_TYPE PotentialBenefitForInlining(Instruction *instr,
    Function *callee) {
    float scale = 0;
    EDGE_WEIGHT_TYPE total;

    total = callee->f_first->b_weight;
    if (total != 0) {
	scale = ((float) instr->i_Bblock->b_weight) / ((float) total);
    }

    return (EDGE_WEIGHT_TYPE) (scale *
	PotentialBenefitForInliningFromPushAnalysis(instr, callee));
}



/*
 *  NumberOfConstantArguments(instr,callee) --
 */
static int NumberOfConstantArguments(Instruction *instr, Function *callee) {
    Instruction *iterator = instr;
    int i = 2 * ADDRESS_SIZE;
    int found = 0;
    int constants = 0;
    Block *bbl = instr->i_Bblock;
    int args = FunNumberOfArguments(callee);

    while (found < args) {
	if (iterator->i_class == CLASS_PUSH) {
	    if ((OperandIsImmediate(iterator->i_src))) {
		AddKnownStackSlot(i, FUNCTION_FIRST_INSTR(callee));
		constants++;
	    }
	    found++;
	    i = i + ADDRESS_SIZE;
	}

	/*
	 * Locate the instruction to examine next. If we're at the top
	 * of a basic block, we can only move to a predecessor block
	 * to analyze pushes if the current block has only one
	 * predecessor.
	 */

	if (iterator->i_Bblock->b_first == iterator) {
	    if (BBLNumberOfPredecessors(bbl) == 1) {
		bbl = NODE_FROM(bbl->b_pred);
		iterator = bbl->b_last;
	    } else {
		return constants;
	    }
	} else {
	    iterator = iterator->i_prev;
	}
    }

    return constants;
}

static EDGE_WEIGHT_TYPE PotentialBenefitForInliningFromPushAnalysis(Instruction
    *instr, Function *callee) {

    Instruction *iterator = instr;
    int i = 2 * ADDRESS_SIZE;
    int found = 0;
    Block *bbl = instr->i_Bblock;
    int args = FunNumberOfArguments(callee);

    while (found < args) {
	if (iterator->i_class == CLASS_PUSH) {
	    if ((OperandIsImmediate(iterator->i_src))) {
		AddKnownStackSlot(i, FUNCTION_FIRST_INSTR(callee));
	    }
	    found++;
	    i = i + ADDRESS_SIZE;
	}

	/*
	 * Locate the instruction to examine next. If we're at the top
	 * of a basic block, we can only move to a predecessor block
	 * to analyze pushes if the current block has only one
	 * predecessor.
	 */

	if (iterator->i_Bblock->b_first == iterator) {
	    if (BBLNumberOfPredecessors(bbl) == 1) {
		bbl = NODE_FROM(bbl->b_pred);
		iterator = bbl->b_last;
	    } else {
		return ComputePotentialAtInstr(FUNCTION_FIRST_INSTR(callee));
	    }
	} else {
	    iterator = iterator->i_prev;
	}

    }

    return ComputePotentialAtInstr(FUNCTION_FIRST_INSTR(callee));
}
