/*
 *  opt.jt.c -- for optimizing jump tables
 */
#include "global.h"
#include "opt.h"
#include "utilities.h"
#include "sysdep.h"
#include "flowgraph.h"
#include "linkage.h"
#include "opt.unreachable.h"

/* Adds a single test for the common case in the JT in 'jumpBLock'
 * given by 'node' */
static void AddCheapTest(Block *jumpBlock, EdgeNode *node);



/*
 *  OptimizeJumpTables() --
 *
 *  Optimizes jumps through jump tables that have skewed distributions by
 *  inserting cheap test(s) for the common cases.
 */
void OptimizeJumpTables() {
    Block *biterator;
    EdgeNode *successors;
    EDGE_WEIGHT_TYPE wt;
    float cutoff = FloatParameter("JTSingleTestCutoff");

    REPORT(RP_MINOR, WITHFUNC("Optimizing jump tables"));

    REPORT(RP_LGLOOP, "  Jump Address    |    Weight     |     Slot(s)     |"
	"    Frequency   |   Type");
    REPORT(RP_LGLOOP, "==================|===============|=================|"
	"================|==============");

    FOREACH_BBL(biterator) {

	if (biterator->b_type != BT_SWITCH) {
	    continue;
	}

	if (biterator->b_weight < 100) {
	    continue;
	}

	wt = BBLSumOutgoingEdges(biterator);

	/*
	 * First see if we can insert a single test for a very common single
	 * case.
	 */

	FOREACH_SUCC_IN_BBL(biterator, successors) {
	    if (NODE_WEIGHT(successors) > (cutoff * wt)) {
		AddCheapTest(biterator, successors);
	    }
	}

    }					/* end iteration over basic blocks */

    return;
}



/*
 *  AddCheapTest(jumpBlock,node) --
 *
 *  Adds a single cheap test from 'jumpBlock' for the common case given in
 *  'node'.
 */
static void AddCheapTest(Block *jumpBlock, EdgeNode *node) {
    Block *newBlock;
    int jtSlot;
    double freq = ((double) NODE_WEIGHT(node) / (double) jumpBlock->b_weight);
    ADDRESS pointer, baseAddress;
    Instruction *boundsCheck = InstrLocateJTBoundsCheck(jumpBlock->b_last);
    Instruction *cmp, *je, *nop, *target;
    double wt = boundsCheck->i_Bblock->b_weight;

    if (boundsCheck == NULL) {
	return;
    }

    newBlock = BBLCreate();
    newBlock->b_weight = wt;
    jtSlot = NODE_JT_SLOT(node);
    baseAddress = InstrLocateJTBaseAddress(jumpBlock->b_last);
    pointer = AddressToData(baseAddress + (ADDRESS_SIZE * jtSlot));
    target = AddressToInstr(pointer);

    if (target == NULL) {
	return;
    }

    cmp = InstrCreateCmpRegToValue(boundsCheck->i_src->a_base_reg, jtSlot);
    je = InstrCreateJe();
    nop = InstrCreateNOP();

    newBlock->b_last = newBlock->b_first = cmp;
    cmp->i_Bblock = newBlock;
    INSTR_SET_FLAG(cmp, BBL_START);
    INSTR_SET_FLAG(je, BBL_END);

    newBlock->b_type = BT_BRANCH;
    InsertInstrAfterInstr(je, cmp);
    InsertBBLBeforeBBL(newBlock, boundsCheck->i_Bblock);
    if (boundsCheck->i_Bblock->b_fun->f_first == boundsCheck->i_Bblock) {
	boundsCheck->i_Bblock->b_fun->f_first = newBlock;
	BBL_REMOVE_FLAG(boundsCheck->i_Bblock, FUN_START);
	BBL_SET_FLAG(newBlock, FUN_START);
    }
    InsertInstrBeforeInstr(nop, boundsCheck);
    BBLDivideBlock(boundsCheck->i_Bblock, boundsCheck);
    BBLRemoveEdges(nop->i_Bblock);
    BBLAddEdge(nop->i_Bblock, newBlock, ET_NORMAL)->e_edgeWeight = wt;
    BBLAddEdge(newBlock, boundsCheck->i_Bblock, ET_FALSE)->e_edgeWeight =
	(1.0 - freq) * wt;
    BBLAddEdge(newBlock, target->i_Bblock, ET_TRUE)->e_edgeWeight = freq * wt;
    InstrDestroy(nop);

    NODE_WEIGHT(node) = 0;

    REPORT(RP_LGLOOP, "   0x%8.8x     |     %.0f    |        %d        |"
	"      %-2.2f      |  Single Test", jumpBlock->b_last->i_address,
	(float) jumpBlock->b_weight, jtSlot, (float) (freq * 100.0));

    return;
}
