/*
 *  opt.layout.c
 *
 *  Functions for changing the basic block layout and structure of the control
 *  flow graph.  These optimizations are intended to improve cache hit rates
 *  and locality.  This optimization should be performed _very_ last because it
 *  ruins a lot of the data structures set up. For instance, we have no notion
 *  of what a 'function' is after layout has been done.
 *
 *  Notes:
 *  1) There should be checks to see if the condition can in fact be switched.
 *  The x86 has some jump instructions like jecxz (jump when register eCX is
 *  0) that have no opposites.  Luckily it doesn't seem the compiler generates
 *  any of these.  If it turns out that a condition cannot be switched then an
 *  unconditional jump should be inserted.
 *
 *  2) On the x86 it turns out that adjacent jumps can share the same branch
 *  target buffer entry if their addresses are the same for bits 2-31. This
 *  means that when we have to add an unconditional jumps after a conditional
 *  jump to preserve both true/false edges when neither has a fall-through,
 *  we may possibly be doing some damage.  Unfortunately we cannot check
 *  something like this until final addresses have been calculated.  Thus, it
 *  may be better to insert two NOPS between adjacent jumps to be safe. It
 *  should be noted that no real noticeable performance hit is seen from the
 *  BTB misses due to this phenomenon, but there are definitely a noticeable
 *  difference in the number of misses.
 *
 *  3) The Figure of Merit analysis has not been checked thoroughly, but it
 *  should produce a better-than-average layout. A greedy layout scheme is
 *  almost always superior though.
 *
 *  4) The Hold/Cold/Zero set construction has not been tested thorough, but it
 *  should produce a better-than-average layout.  The standard greedy (Pettis-
 *  Hansen) algorithm is almost always a better choice.
 */
#include "global.h"
#include "opt.layout.h"
#include "flowgraph.h"
#include "utilities.h"
#include "linkage.h"
#include "sysdep.h"
#include "opt.unreachable.h"
#include "dominator.h"
#include "NDG.h"

#define EDGE_SOLO_ADJUST     1.50

/*
 *  Initializes the basic block chains so that each basic block is the head and
 *  tail of its own chain
 */
static void BBLSetupChains();

/*
 *  Initializes the basic block chains so that each bbl is the head/tail of its
 *  own chain. Only basic block within the supplied function are considered.
 */
static void BBLSetupChainsForFun(Function *fun);

/* Constructs basic block chains by linking together bbls based on
 * aggregate chain weights */
static void BBLBuildChains();

/*
 *  After the chains have been constructed we want to relink all
 *  the basic blocks so that their layout in the final executable
 *  is based upon the chaining order
 */
static void BBLLinkBlocksFromChains();

/* Does the same as the above function, but only for 'fun' */
static void BBLLinkBlocksFromChainsForFun(Function *fun);

/*
 *  After the function chains have been constructed, FunLinkBlocksFromChains()
 *  links together functions based on their chain orderings
 */
static void FunLinkBlocksFromChains();

/* Removes all edges of type ET_CHAIN from basic blocks */
static void BBLRemoveChainEdges();

/* Returns true if there exists a chain edge (ET_CHAIN)
 * between the src and dest block */
static bool ChainEdgeExists(Block *src, Block *dest);

/*
 *  Returns the next edge in the given basic block that
 *  doesn't have a chain edge associated with it
 */
static Edge *BBLGetNextEdgeForChain(Block *block);

/*
 *  Returns the aggregate chain weight between two basic block chains.
 *  'src' and 'dest' are assumed to be the head of two chains
 */
static EDGE_WEIGHT_TYPE EdgeAggregateChainWeight(Block *src, Block *dest);

/* Sorts the edges */
static void EdgeSort();

/* Sorts the linked list of edges based on weight. */
static int EdgeCompareGreedy();

/*
 *  Often times we want to know if two basic blocks are in the same
 *  'temperature' set.  BBLInSameTemperatureSetAsBBL() will return true
 *  if both blocks are.
 */
static bool BBLInSameTemperatureSetAsBBL(Block *bbl1, Block *bbl2);

/*
 *  Connects the destination chain at the end of the source chain, and updates
 *  chain headers and such.
 */
static void BBLConnectChains(Block *sourceChain, Block *destinationChain);

/*
 *  Connects the destination chain at the end of the source chain, and updates
 *  chain headers and such.
 */
static void FunConnectChains(Function *sourceChain, Function *destinationChain);

/*
 *  Links chains together so that there is one chain
 *  for each temperature group: hot, cold and zero
 */
static void BBLTemperatureChain();

/* The first basic block chain in the list of chains */
static Block *chain;

/* The first function in the list of chains */
static Function *fchain;

/* Linked list of edges in sorted order */
static Edge *edge_list;	

static void FunsPositioning();
static Block *BBLSplitChainsForFun(Function *fun);
static void BBLSplitChainsAppend(Block *fluff_head);

/*
 *  BBLOptimizeLayout() --
 *
 *  Layout the basic blocks so that the number of branches is reduced
 *  for the most common execution sequences.  Basic block chains are
 *  constructed based on edge weights. High-weight edges between two
 *  blocks results in the blocks being placed closed together.  The
 *  chains of blocks are then optimized for locality by looking at the
 *  aggregate weight of edges between chains.  Basic blocks and
 *  instructions are then re-linked to reflect the optimized layout.
 *  Finally, a pass is made over the instruction sequence to make sure
 *  that fall-throughs can still fall-through.  During this process
 *  some unconditional jump instructions are inserted, and some are
 *  also removed if we have introduced new fall-throughs.
 */
void BBLOptimizeLayout() {
    Function *fun;

    BBLClearQuadwords();
    CheckConsistency();
    RemoveEntryExitReturnFuncallEdges();// Get rid of edges to fake blocks
    RemoveEntryExitBlocks();		// Get rid of fake blocks

    /*
     * For greedy and figure-of-merit layout schemes, we'll take the same
     * course of action. The only difference is that we'll compute Figures
     * of Merit (when using the FOM scheme) and then sort the edges based on
     * this figure rather than the weight of edge.
     */

    if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_GREEDY) {
    	BBLSetupChains();	// Let each block be head and tail of a chain
	BBLBuildChains();	// Attach blocks together based on edge weights
    	BBLLinkBlocksFromChains();	// Use chain linkage to layout bbls
	InstrRelinkBasedOnBBL();	// Update instrs to reflect bbl movement
	InstrUpdateEdges();		// Update for changed fall-throughs
    }

    /*
     * If we're laying out only functions, the only change we make is that
     * we ignore basic blocks chains and instead relink functions based on
     * the function chains that were constructed.
     */

    //    BBL_LAYOUT_SCHEME = BBL_LAYOUT_FUNS_ONLY;
    if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_FUNS_ONLY) {
	BBLSetupChains();	// Let each block be head and tail of a chain
	BBLBuildChains();	// Attach blocks together based on edge weights
	FunLinkBlocksFromChains();
	BBLRelinkBasedOnFun();
	InstrRelinkBasedOnBBL();
	InstrUpdateEdges();
    }

    /*
     * For the Pettis-Hansen layout scheme, we don't want to consider
     * inter-procedural edges; that is, we want to lay out the basic blocks
     * only on the function level. The functions can then later be
     * re-arranged among themselves to improve icache performance.
     */


    if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_PH || 
        BBL_LAYOUT_SCHEME == BBL_LAYOUT_PH_WITH_SPLIT) {
	/* 
	 * By default, set FUNCTION_LAYOUT_FLAG on all functions; 
	 * but should be easy to use this flag to choose to do layout on 
	 * predefined functions.
	 */
	FOREACH_FUNCTION(fun) {	
	    /*
	     * There are a problem of handling _init and _fini function:
	     * _init and _fini are in .init and .fini sections. PLTO 
	     * treats them as regular functions (assume them in .text section).
	     * but we can not move these functions around inside .text section.
	     * Simple solution is just to avoid these functions. 
	     * A more generic solution, would be keeping track of the original 
	     * sections after elf parsing, and performing intra-section layout
	     */
	    if (strcmp (fun->f_name, "_init") == 0 || 
		strcmp (fun->f_name, "_fini") == 0) {
		continue;
	    }

	    FUNCTION_SET_FLAG(fun, FUNCTION_LAYOUT_FLAG);
	}

        /*
         * Function Positioning
	 */
	FunsPositioning ();

	/*
	 * Update instrs to reflect bbl movement
	 */
	BBLRelinkBasedOnFun();
	InstrRelinkBasedOnBBL();
	InstrUpdateEdges();	

	/* BBL Layout */

	/* fluff list used for function splitting */
	Block *fluff_head, *fluff_tail; 
	Block *fhead = NULL, *ftail = NULL; 
	fluff_head = fluff_tail = NULL;

	FOREACH_FUNCTION(fun) {
	    if (! FUNCTION_HAS_FLAG(fun, FUNCTION_LAYOUT_FLAG)) {
		continue;
	    }

	    BBLSetupChainsForFun(fun);
	    BBLBuildChains();

	    if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_PH_WITH_SPLIT) {
		/*
		 * Function splitting. Remove all fluff blocks in function
		 * from the bbl chains and link them together.
		 */
		fhead = BBLSplitChainsForFun (fun);
		if (fhead) {
		    /* append fluff list in the fun into global fluff list */
		    /* get the tail of the list */
		    for (ftail = fhead; ftail->b_chainnext; 
			    ftail = ftail->b_chainnext);

		    if (fluff_head == NULL) {
			fluff_head = fhead;
			fluff_tail = ftail;
		    } else {
			fluff_tail->b_chainnext = fhead;
			fhead->b_chainprev = fluff_tail;
			fluff_tail = ftail;
		    }
		}
	    }

	    BBLLinkBlocksFromChainsForFun(fun);	    
	}

	BBLRelinkBasedOnFun();	
	InstrRelinkBasedOnBBL();
	
	if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_PH_WITH_SPLIT) {
	    BBLSplitChainsAppend (fluff_head);
	}

	InstrUpdateEdges();	// Update edges for changed fall-throughs
	bbl_list = f_list->f_first;
	i_list = bbl_list->b_first;
    }

    return;
}

static int function_cmp(void *fun1, void *fun2) {
    return strcmp (((Function *)fun1)->f_name, ((Function *)fun2)->f_name);
}

static void function_print(void *fun) {
    printf ("%s ", ((Function *)fun)->f_name);
}

static void FunsPositioning() {
    Function *fun, *caller;
    Block *from;
    EdgeNode *predNode;
    NDG *graph;
    NDGNode *node;
    NDGElem *elem;

    /* 
     * A Non-directed graph (NDG) is used to represent the function call 
     * graph.
     */

    graph = NDGCreate ();
    graph->data_cmp = &function_cmp;
    graph->print_data = &function_print;

    FOREACH_FUNCTION (fun) {
	if (! FUNCTION_HAS_FLAG(fun, FUNCTION_LAYOUT_FLAG)) {
	    continue;
	}
	FOREACH_PRED_IN_BBL (fun->f_first, predNode) {
	    /*
	     * scan all precedence edges of fun (potential a call),
	     * if the edge's weight is larger than 0, add the edge into
	     * NDG.
	     */
	    from = NODE_FROM(predNode);
	    if (from && NODE_WEIGHT(predNode)) {
		caller = from->b_fun;	    
		if (strcmp (caller->f_name, "_init") == 0 ||
		    strcmp (caller->f_name, "_fini") == 0) {
		    continue;
		}

		NDGAddEdge (graph, caller, fun, NODE_WEIGHT(predNode));
	    }
	}
    }

    NDGInitWeight (graph);
    NDGSort (graph);
    
    /* 
     * Displace functions based on the order for NDG sort
     */

    /* store the location where we are going to put next function */
    Function *f_insert; 

    /*
     * Each node represents a list of functions which are sorted
     */
    for (node = graph->nodes; node; node = node->next) {
	f_insert = NULL;	
	for (elem = node->elems; elem; elem = elem->next) {
	    fun = (Function *)(elem->data);
	    /* 
	     * First remove fun from f_list and the place we start to insert 
	     * functions is the location where the first function in the list
	     * used to be. .
	     */
	    if (fun->f_prev) {
		if (f_insert == NULL) {
		    f_insert = fun->f_prev;
		}		    
		fun->f_prev->f_next = fun->f_next;		
	    }
	    if (fun->f_next) {
		if (f_insert == NULL) {
		    f_insert = fun->f_next;
		}		    
		fun->f_next->f_prev = fun->f_prev;
	    }
	    if (f_list == fun) {
		f_list = fun->f_next;
	    }
	    
	    fun->f_next = NULL;
	    fun->f_prev = NULL;
	    InsertFunctionAfterFunction(fun, f_insert);
	    f_insert = fun;
	}
    }

    return; 
}


/*
 *  BBLSetupChains() --
 *
 *  We will initially create one chain for each basic block, since each basic
 *  block should start out being its own head and tail.
 */
static void BBLSetupChains() {
    Block *biterator;
    Function *fiterator;

    FOREACH_FUNCTION(fiterator) {
	/* Set up each function as its own chain */
	fiterator->f_chainnext = NULL;
	fiterator->f_chainprev = NULL;
	fiterator->f_headnext = fiterator->f_next;
	fiterator->f_chainhead = fiterator;
	fiterator->f_nexttochain = fiterator;
	if (fiterator->f_next != NULL) {
	    fiterator->f_next->f_headprev = fiterator;
	}
    }

    FOREACH_BBL(biterator) {
	if (biterator == HELL) {
	    continue;
	}

	/* Check data structure consistency to make sure all blocks have at
	 * least one edge.  Exceptions are blocks that end with a HALT instr. */
	if (biterator->b_succ == NULL) {
	    if (biterator->b_type != BT_RETURN) {
		if (biterator->b_type != BT_HALT) {
		    WARNING("Basic block encountered without any edges");
		}
	    }
	}

	/* Set up each basic block as its own chain */
	biterator->b_chainnext = NULL;
	biterator->b_chainprev = NULL;
	biterator->b_headnext = biterator->b_next;
	biterator->b_chainhead = biterator;
	biterator->b_nexttochain = biterator;
	if (biterator->b_next != NULL) {
	    biterator->b_next->b_headprev = biterator;
	}
    }

    chain = bbl_list;
    fchain = f_list;

    return;
}



/*
 *  BBLSetupChainsForFun(fun) --
 *
 *  We will initially create one chain for each basic block, since each basic
 *  block should start out being its own head and tail. This is done for all
 *  bbls in the supplied function.
 */
static void BBLSetupChainsForFun(Function *fun) {
    Block *biterator;

    FOREACH_BBL_IN_FUN(fun, biterator) {
	/* Set up each basic block as its own chain */
	biterator->b_chainnext = NULL;
	biterator->b_chainprev = NULL;

	if (biterator->b_next == NULL) {
	    biterator->b_headnext = NULL;
	} else if (biterator->b_next->b_fun == fun) {
	    biterator->b_headnext = biterator->b_next;
	} else {
	    biterator->b_headnext = NULL;
	}

	biterator->b_chainhead = biterator;
	biterator->b_nexttochain = biterator;
	if (biterator->b_next != NULL) {
	    biterator->b_next->b_headprev = biterator;
	}
    }
    chain = fun->f_first;
    fchain = fun;

    return;
}

/*
 *  BBLBuildChains() --
 *
 *  Until the tree is empty, keep removing and deleting the largest element.
 *  We aren't interested in exit/entry/hell edges, because the basic blocks
 *  they link to will not end up in the final binary.  Call edges will be
 *  taken care of by ET_FUNDIRECT edges as opposed to the entry edges. The
 *  basic block chains will only be linked if the source block is at the
 *  end of a chain and the destination block is at the beginning of a chain.
 */
static void BBLBuildChains() {
    Block *biterator, *src, *dest, *last;
    Edge *largest;
    Edge *edge;
    int round, chains, change, temp;
    bool edgeExists;

    change = 0;
    chains = 0;

    biterator = chain;
    while (biterator != NULL) {
	chains++;
	biterator = biterator->b_headnext;
    }

    round = 1;
    while (chains > 1) {

	BBLRemoveChainEdges();
	edgeExists = false;

	/*
	 * The following tests check convergence conditions for the layout
	 * construction.  Attaching chains together becomes increasingly
	 * (non-linear) expensive as the number of blocks per chain increases.
	 * This is usually little improvement seen after 2 or 3 rounds, but
	 * usually we can safely let it proceed 7-9 rounds without taking
	 * too long.
	 */

	if (BBL_LAYOUT_SCHEME != BBL_LAYOUT_PH) {
	    if (round > 2) {
		if (change < DELTA_CHAIN) {
		    break;
		}
		if (round > MAXIMUM_ROUNDS) {
		    break;
		}
		if ((chains / change) > DELTA_PERCENT) {
		    break;
		}
	    }
	} else {
	    if (round > 2) {
		if (change == 0) {
		    break;
		}
	    }
	}

	/* Add edges between the chains */
	biterator = chain;
	while (biterator != NULL) {

	    /* Locate the last block in the current chain */
	    last = biterator;
	    while (last->b_chainnext != NULL) {
		last = last->b_chainnext;
	    }

	    while (true) {
		edge = BBLGetNextEdgeForChain(biterator);
		if (edge == NULL || edge->e_edgeWeight == 0) {
		    /* 
		     * We skip weight 0 edge because it doesn't affect
		     * bbl layout.
		     */
		    break;
		}
		edgeExists = true;
		(BBLAddEdge(last, edge->e_to->b_chainhead,
			ET_CHAIN)->e_edgeWeight) =
		    EdgeAggregateChainWeight(biterator,
		    edge->e_to->b_chainhead);
		CASSERT(NODE_TYPE(last->b_succ) == ET_CHAIN,
		    "ET_CHAIN not first edge. Edges corrupt.");
	    }
	    biterator = biterator->b_headnext;
	}

	if (!edgeExists) {
	    break;
	}

	/* EdgeSort will either sort based on FOM or weight.  */
	EdgeSort();

	while (true) {
	    Function *srcFun, *destFun;

	    largest = edge_list;
	    if (largest == NULL) {
		break;
	    }

	    edge_list = edge_list->e_next;

	    src = largest->e_from;
	    dest = largest->e_to;

	    srcFun = src->b_fun;
	    destFun = dest->b_fun;

	    if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_PH) {
		CASSERT(srcFun == destFun, "Pettis-Hansen layout used but"
		    " basic blocks chosen do not lie in the same function.");
	    }

	    if (srcFun->f_chainhead != destFun->f_chainhead) {
		if (!srcFun->f_chainnext) {
		    if (!destFun->f_chainprev) {
			FunConnectChains(srcFun, destFun);
		    }
		} else if (!srcFun->f_chainprev) {
		    if (!destFun->f_chainnext) {
			FunConnectChains(destFun, srcFun);
		    }
		}
	    }

	    if (src->b_chainnext || dest->b_chainprev) {
		continue;
	    }

	    if (BBL_HAS_FLAG(dest, BBL_LOOP_HEADER_FLAG)
		&& BBLIsInSameLoopAsBBL(src, dest)) {
		
    		continue;
        }

	    /*
	     * We have a source block that is the tail of its chain, and a
	     * destination block that is the head of the chain.  They will be
	     * merged together by appending the target chain to the source
	     * chain.  If they already lie within the same chain then we
	     * have nothing to do.
	     */

	    if (src->b_chainhead == dest->b_chainhead) {
		continue;
	    }
	    
	    BBLConnectChains(src, dest);

	}

	biterator = chain;
	temp = 0;
	while (biterator != NULL) {
	    temp++;
	    biterator = biterator->b_headnext;
	}
	change = chains - temp;
	chains = temp;

	if (BBL_LAYOUT_SCHEME != BBL_LAYOUT_PH) {
	    REPORT(RP_SMLOOP, "BBL layout, round %d: %d chains remain",
		round, chains);
	}

	round++;

    }					/* End construction of chains */

    if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_PH) {
	BBLRemoveChainEdges();
    }
    
    return;
}



/*
 *  FunLinkBlocksFromChains() --
 *
 *  The chains have been created but the functions still have the original
 *  linkage that the flowgraph imposed upon them. FunLinkBlocksFromChains
 *  will re-adjust the layout (next and prev links) based on the chains
 *  constructed (chainNext and chainPrev links). First the chains will
 *  be linked so we are left with 1 giant chain.
 */
static void FunLinkBlocksFromChains() {
    Function *funs, *heads;

    heads = funs = f_list;

    while (funs != NULL) {
	heads = funs->f_next;
	funs->f_prev = funs->f_next = NULL;
	funs = heads;
    }

    heads = funs = f_list = fchain;

    while (heads != NULL) {
	funs = heads;
	while (funs != NULL) {

	    /*
	     * First take care of the previous linkage.  Note that we may have
	     * to backtrack to the previously visited chain if we're currently
	     * at the first func in a chain. When we do backtrack, the function
	     * we're looking for is the last one in that chain.
	     */

	    if (funs->f_chainprev == NULL) {
		Function *fiterator;
		Function *found = NULL;
    
		fiterator = heads->f_headprev;
		if (fiterator != NULL) {
		    Function *fun_iterator;
		    fun_iterator = fiterator;
		    while (fun_iterator->f_chainnext != NULL) {
			fun_iterator = fun_iterator->f_chainnext;
		    }
		    found = fun_iterator;
    	    	}
		funs->f_prev = found;
    	    } else {
		funs->f_prev = funs->f_chainprev;
    	    }

	    /*
	     * Adjust the 'next' linkage.  Again, examining the next chain
	     * may be necessary if the current block is a tail of a chain.
	     */

	    if (funs->f_chainnext == NULL) {
		if (heads->f_headnext != NULL) {
		    funs->f_next = heads->f_headnext;
		} else {
		    funs->f_next = NULL;
		}
	    } else {
		funs->f_next = funs->f_chainnext;
    	    }

	    funs = funs->f_chainnext;
	}		/* End iteration over the functions in a chain */

	heads = heads->f_headnext;
    }			/* End iteration over the chain list */

    return;
}



/*
 *  BBLLinkBlocksFromChains() --
 *
 *  The chains have been created but the blocks still have the original
 *  linkage that the flowgraph imposed upon them. BBLLinkBlocksFromChains
 *  will re-adjust the layout (next and prev links) based on the chains
 *  constructed (chainNext and chainPrev links). First the chains will
 *  be linked so we are left with 1 giant chain.
 */
static void BBLLinkBlocksFromChains() {
    Block *blocks, *heads;

    heads = blocks = bbl_list;

    if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_ALTO) {
	BBLTemperatureChain();
    }

    while (blocks != NULL) {
	heads = blocks->b_next;
	blocks->b_prev = blocks->b_next = NULL;
	blocks = heads;
    }

    heads = blocks = bbl_list = chain;

    while (heads != NULL) {
	blocks = heads;
	while (blocks != NULL) {

	    /*
	     * First take care of the previous linkage.  Note that we may have
	     * to backtrack to the previously visited chain if we're currently
	     * at the first block in a chain. When we do backtrack, the block
	     * we're looking for is the last one in that chain.
	     */

	    if (blocks->b_chainprev == NULL) {
		Block *biterator;
		Block *found = NULL;

		biterator = heads->b_headprev;
		if (biterator != NULL) {
		    Block *block_iterator;
		    block_iterator = biterator;
		    while (block_iterator->b_chainnext != NULL) {
			block_iterator = block_iterator->b_chainnext;
		    }
		    found = block_iterator;
		}
		blocks->b_prev = found;
	    } else {
		blocks->b_prev = blocks->b_chainprev;
	    }

	    /*
	     * Adjust the 'next' linkage.  Again, examining the next chain
	     * may be necessary if the current block is a tail of a chain.
	     */

	    if (blocks->b_chainnext == NULL) {
		if (heads->b_headnext != NULL) {
		    blocks->b_next = heads->b_headnext;
		} else {
		    blocks->b_next = NULL;
		}
	    } else {
		blocks->b_next = blocks->b_chainnext;
	    }

	    blocks = blocks->b_chainnext;
	}		/* End iteration over the blocks in a chain */

	heads = heads->b_headnext;
    }			/* End iteration over the chain list */

    return;
}

static void BBLSplitChainsAppend(Block *fluff_head) {
    Block *lastbbl, *bbl;
    int i;

    /* 
     * append the fluff list at the end of bbl_list 
     * but since .fini section is treated as .text and 
     * locates at the end of bbl_list, we want to add fluff list just 
     * before .fini section to avoid long jump.
     *
     */
    if (fluff_head) {
	for (bbl = bbl_list; bbl->b_next; bbl = bbl->b_next) {
	    if (strcmp (((bbl->b_next)->b_fun)->f_name, "_fini") == 0) {
		break;
	    }
	}
	
	lastbbl = bbl;
	for (i = 0, bbl = fluff_head; bbl; bbl = bbl->b_chainnext, i++) {
	    InsertBBLAfterBBL(bbl, lastbbl);
	    lastbbl = bbl;
	}
    }
}

/* 
 * BBLSplitChainsForFun ()
 * 
 * After BBL chains created, remove fluff blocks (basic blocks did not 
 * executed) from BBL chains and link them togehter (fluff list).
 */
Block * BBLSplitChainsForFun(Function *fun) {
    Block *blocks, *bbl, *heads, *headnext, *chainnext;
    Block *fhead = NULL, *ftail = NULL;

    heads = chain;
    while (heads) {
	headnext = heads->b_headnext;
	for (bbl = heads; bbl; bbl = bbl->b_chainnext) {
	    if (bbl->b_weight != 0) {
		break;
	    }
	}

	if (bbl == NULL) {
	    /* the current chain is a fluff chain (never been executed 
	     * during according to the profile data; 	     
	     * remove it from chain list 
	     */
	    if (heads->b_headprev) {
		heads->b_headprev->b_headnext = heads->b_headnext;
	    } 
	    
	    if (heads->b_headnext) {
		heads->b_headnext->b_headprev = heads->b_headprev;
	    }

	    if (heads == chain) {
		chain = heads->b_headnext;
	    }

	    blocks = heads;
	    /* link them to fluff list; */
	    while (blocks != NULL) {	
		chainnext = blocks->b_chainnext;
		/* dont break the original link */
/* 		blocks->b_prev = NULL;	 */
/* 		blocks->b_next = NULL; */

		blocks->b_chainprev = NULL;	
		blocks->b_chainnext = NULL;		

		if (fhead == NULL) {
		    fhead = ftail = blocks;
		} else {
		    ftail->b_chainnext = blocks;
		    blocks->b_chainprev = ftail;
		    ftail = blocks;
		}
	    
		blocks = chainnext;
	    }    
	}
	heads = headnext; 
    }

    if (chain == NULL) {
	/* If all blocks in fun are in fluff list
	 * it means the whole fun has been executed. 
	 * we dont need to reorder its blocks. Also
	 * we avoid putting them into global fluff list,
	 * FunsPositioning should handle this function. 
	 */
	return NULL;
    } else {
	return fhead;
    }
}


/*
 *  BBLLinkBlocksFromChainsForFun(fun) --
 *
 *  The chains have been created but the blocks still have the original
 *  linkage that the flowgraph imposed upon them. BBLLinkBlocksFromChains
 *  will re-adjust the layout (next and prev links) based on the chains
 *  constructed (chainNext and chainPrev links). First the chains will
 *  be linked so we are left with 1 giant chain. This transformation is
 *  performed assuming the bbl chain consists of blocks only from 'fun'
 */
static void BBLLinkBlocksFromChainsForFun(Function *fun) {
    Block *bbl;
    Block *blocks, *heads;

    /* 
     * If we are doing function splitting, chain can be NULL because
     * if the function has not been executed at all, all the bbl chains 
     * functions will be moved into fluff chains. If it is that case, we 
     * just keep the function as it is.
     */
    if (chain == NULL) {
	return;
    }
    fun->f_first = chain;

    FOREACH_BBL_IN_FUN(fun, bbl) {
	bbl->b_prev = bbl->b_next = NULL;
    }

    heads = blocks = chain;

    while (heads != NULL) {
	blocks = heads;
	
	while (blocks != NULL) {

	    /*
	     * First take care of the previous linkage.  Note that we may have
	     * to backtrack to the previously visited chain if we're currently
	     * at the first block in a chain. When we do backtrack, the block
	     * we're looking for is the last one in that chain.
	     */

	    if (blocks->b_chainprev == NULL) {
		Block *biterator;
		Block *found = NULL;

		biterator = heads->b_headprev;
		if (biterator != NULL) {
		    Block *block_iterator;
		    block_iterator = biterator;
		    while (block_iterator->b_chainnext != NULL) {
			block_iterator = block_iterator->b_chainnext;
		    }
		    found = block_iterator;
		}
		blocks->b_prev = found;
	    } else {
		blocks->b_prev = blocks->b_chainprev;
	    }

	    /*
	     * Adjust the 'next' linkage.  Again, examining the next chain
	     * may be necessary if the current block is a tail of a chain.
	     */

	    if (blocks->b_chainnext == NULL) {
		if (heads->b_headnext != NULL) {
		    blocks->b_next = heads->b_headnext;
		} else {
		    blocks->b_next = NULL;
		}
	    } else {
		blocks->b_next = blocks->b_chainnext;
	    }

	    if ((blocks->b_chainnext == NULL)
		&& (blocks->b_chainhead->b_headnext == NULL))
		fun->f_last = blocks;

	    blocks = blocks->b_chainnext;
	}			/* End iteration over the blocks in a chain */

	heads = heads->b_headnext;
    }				/* End iteration over the chain list */

    
    return;
}



/*
 *  InstrUpdateEdges() --
 *
 *  Iterates over all basic blocks and changes edges based on
 *  fall-throughs that may have been introduced or eliminated by basic
 *  block reordering.
 */
void InstrUpdateEdges() {
    Instruction *iterator;
    Block *biterator;
    EdgeNode *niterator;

    int jumpsAddedFromCalls = 0;
    int jumpsAddedFromBranches = 0;
    int jumpsAddedFromNormals = 0;
    int jumpsRemoved = 0;

    float jumpsExecuted = 0;

    i_list = bbl_list->b_first;

    FOREACH_BBL(biterator) {
	iterator = biterator->b_last;

	if (iterator == NULL) {
	    CASSERT(biterator == HELL, "Basic block without last instruction.");
	    continue;
	}

	if (biterator->b_type == BT_BRANCH) {
	    EdgeNode *FALSENode = NULL;
	    EdgeNode *TRUENode = NULL;

	    FOREACH_SUCC_IN_BBL(biterator, niterator) {
		if (NODE_TYPE(niterator) == ET_FALSE) {
		    CASSERT(FALSENode == NULL,
			"Two ET_FALSE edges found in one bbl.");
		    FALSENode = niterator;
		}
		if (NODE_TYPE(niterator) == ET_TRUE) {
		    CASSERT(TRUENode == NULL,
			"Two ET_TRUE edges found in one bbl.");
		    TRUENode = niterator;
		}
	    }

	    CASSERT(TRUENode != NULL, "No TRUE edge found in BT_BRANCH block");
	    CASSERT(FALSENode != NULL,"No FALSE edge found in BT_BRANCH block");

	    /* See if the condition needs to be switched */

	    if ((iterator->i_next == NULL) ||
		(NODE_TO(FALSENode) != iterator->i_next->i_Bblock)) {

		/*
		 * The blocks have been constructed such that this conditional
		 * branch does not have either of its target blocks following
		 * the branch.  For this case we need to add an unconditional
		 * jump at the end so the false edge will still execute
		 * correctly.
		 */

		if ((iterator->i_next == NULL) ||
		    (NODE_TO(TRUENode) != iterator->i_next->i_Bblock)) {
		    Block *newblock = BBLCreate();
		    Instruction *jmp = InstrCreateJMP();

		    INSTR_SET_FLAG(jmp, BBL_START);
		    INSTR_SET_FLAG(jmp, BBL_END);
		    jmp->i_Bblock = newblock;
		    newblock->b_type = BT_UNCOND;
		    newblock->b_first = newblock->b_last = jmp;

		    jumpsAddedFromBranches++;
		    jumpsExecuted = jumpsExecuted + NODE_WEIGHT(FALSENode);

		    BBLAddEdge(biterator, newblock, ET_FALSE)->e_edgeWeight =
			NODE_WEIGHT(FALSENode);
		    BBLAddEdge(newblock, NODE_TO(FALSENode), ET_UNCOND)
			->e_edgeWeight = NODE_WEIGHT(FALSENode);

		    BBLRemoveEdgeTyped(biterator, NODE_TO(FALSENode), ET_FALSE);
		    InsertBBLAfterBBL(newblock, biterator);

		    if (biterator->b_loopHead != NULL) {
			BBLAddToLoop(newblock, biterator->b_loopHead);
		    }

		    /*
		     * The true edge target block resides directly after the
		     * branch block.  In order to let control flow fall through,
		     * we need to switch the fall and the true edges, and also
		     * switch the switch the condition on the jump instruction.
		     */

		} else {
		    NODE_TYPE(TRUENode) = ET_FALSE;
		    NODE_TYPE(FALSENode) = ET_TRUE;
		    jumpsExecuted -= NODE_WEIGHT(TRUENode);
		    jumpsExecuted += NODE_WEIGHT(FALSENode);
		    INSTR_SET_FLAG(iterator, INSTR_SWITCH_CONDITION_FLAG);
		}

	    }
	}

	/*
	 * We may have separated two adjacent blocks who were connected by an
	 * ET_NORMAL edge. To fix things up, we need to add an unconditional jmp
	 * at the end of the block.
	 */


	if (biterator->b_type == BT_NORMAL) {
	    EdgeNode *normalEdge = NULL;

	    FOREACH_SUCC_IN_BBL(biterator, niterator) {
		if (NODE_TYPE(niterator) == ET_NORMAL) {
		    CASSERT(normalEdge == NULL,
			"Two ET_NORMAL edges found in one bbl.");
		    normalEdge = niterator;
		}
	    }

	    if (normalEdge == NULL) {
		WARNING("[InstrUpdateEdges]"
		    " No ET_NORMAL edge found in BT_NORMAL block.");
		WARNING("  Last instr: %x: %s, bbl_type: %x\n",
		    biterator->b_last->i_address,
		    PrintInstruction(biterator->b_last),
		    biterator->b_type);
		continue;
	    }

	    if ((iterator->i_next == NULL) ||
		(NODE_TO(normalEdge) != iterator->i_next->i_Bblock)) {

		jumpsAddedFromNormals++;
		jumpsExecuted = jumpsExecuted + NODE_WEIGHT(normalEdge);

		InsertInstrAfterInstr(InstrCreateJMP(), iterator);
		NODE_TYPE(normalEdge) = ET_UNCOND;
		biterator->b_type = BT_UNCOND;
		biterator->b_last = iterator->i_next;
	    }
	}

	/*
	 * It also may be the case that two basic blocks connected by an uncond.
	 * jmp are now adjacent, and we can exploit the fact that there is a
	 * fall-through. We will transform the unconditional jump edge into an
	 * ET_NORMAL edge, and remove the jmp instruction.
	 */

	if (biterator->b_type == BT_UNCOND) {
	    EDGE_WEIGHT_TYPE weight;

	    if (iterator->i_next == NULL) {
		continue;
	    }

	    /* See if we can exploit the fall-through */
	    if ((biterator->b_succ != NULL) &&
		(NODE_TO(biterator->b_succ) == iterator->i_next->i_Bblock)) {
		jumpsRemoved++;
		CASSERT(BBLNumberOfSuccessors(biterator) == 1,
		    "BT_UNCOND block with successors != 1.");
		jumpsExecuted = jumpsExecuted - NODE_WEIGHT(biterator->b_succ);

		biterator->b_type = BT_NORMAL;
		weight = BBLGetEdgeToBBL
		    (biterator, iterator->i_next->i_Bblock)->e_edgeWeight;
		BBLRemoveEdge(biterator, iterator->i_next->i_Bblock);
		(BBLAddEdge
		    (biterator, iterator->i_next->i_Bblock,
			ET_NORMAL)->e_edgeWeight) = weight;
		InstrDestroy(iterator);
	    } else if (isConditionalInstr(iterator)) {
		InsertInstrBeforeInstr(InstrCreateJMP(), iterator);
		InstrDestroy(iterator);
	    }
	}

	/*
	 * If the basic block to which control returns after a function call
	 * no longer follows the block which makes the call, then an
	 * unconditional jump needs to be added to patch things up correctly.
	 */

	if (biterator->b_type == BT_CALL) {
	    Edge *funlink = BBLGetFunlinkEdge(biterator);
	    CASSERT(funlink != NULL,
		"No ET_FUNLINK edge found in BT_CALL block.");
	    if ((iterator->i_next == NULL) ||
		(funlink->e_to != iterator->i_next->i_Bblock)) {
		Block *newblock = BBLCreate();
		Instruction *jmp = InstrCreateJMP();
		Block *linkBlock = funlink->e_to;

		jumpsAddedFromCalls++;
		jumpsExecuted = jumpsExecuted + funlink->e_edgeWeight;

		INSTR_SET_FLAG(jmp, BBL_START);
		INSTR_SET_FLAG(jmp, BBL_END);
		jmp->i_Bblock = newblock;
		newblock->b_type = BT_UNCOND;
		newblock->b_first = newblock->b_last = jmp;

		/* Edge maintenance */
		BBLRemoveEdgeTyped(biterator, linkBlock, ET_FUNLINK);
		BBLAddEdge(biterator, newblock, ET_FUNLINK)->e_edgeWeight = 0;
		BBLAddEdge(newblock, linkBlock, ET_UNCOND)->e_edgeWeight =
		    biterator->b_weight;
		InsertBBLAfterBBL(newblock, biterator);
		if (biterator->b_loopHead != NULL) {
		    BBLAddToLoop(newblock, biterator->b_loopHead);
		}

	    }
	}

	if (biterator->b_type == BT_SYSCALL) {
	    Edge *syslink = BBLGetSyscallLinkEdge(biterator);
	    if ((iterator->i_next == NULL) ||
		(syslink->e_to != iterator->i_next->i_Bblock)) {
		Block *newblock = BBLCreate();
		Instruction *jmp = InstrCreateJMP();
		Block *linkBlock = syslink->e_to;

		// jumpsAddedFromCalls++;
		jumpsExecuted = jumpsExecuted + syslink->e_edgeWeight;

		INSTR_SET_FLAG(jmp, BBL_START);
		INSTR_SET_FLAG(jmp, BBL_END);
		jmp->i_Bblock = newblock;
		newblock->b_type = BT_UNCOND;
		newblock->b_first = newblock->b_last = jmp;

		/* Edge maintenance */
		BBLRemoveEdgeTyped(biterator, linkBlock, ET_SYSCALL_LINK);
		BBLAddEdge(biterator, newblock, ET_SYSCALL_LINK)->e_edgeWeight =
		    0;
		BBLAddEdge(newblock, linkBlock, ET_UNCOND)->e_edgeWeight =
		    biterator->b_weight;
		InsertBBLAfterBBL(newblock, biterator);

		if (biterator->b_loopHead != NULL) {
		    BBLAddToLoop(newblock, biterator->b_loopHead);
		}
	    }
	}
    }					/* End iteration over basic blocks */

    /* Spit out information about unconditional jumps */
    REPORT(RP_DETAIL, "%6d unconditional jumps removed", jumpsRemoved);
    REPORT(RP_DETAIL, "%6d unconditional jumps added",
	jumpsAddedFromCalls + jumpsAddedFromBranches + jumpsAddedFromNormals);
    REPORT(RP_DETAIL, "   %6d added to preserve ET_NORMAL fallthroughs",
	jumpsAddedFromNormals);
    REPORT(RP_DETAIL, "   %6d added to preserve ET_FALSE fallthroughs",
	jumpsAddedFromBranches);
    REPORT(RP_DETAIL, "   %6d added to preserve ET_CALL return fallthroughs",
	jumpsAddedFromCalls);
    REPORT(RP_DETAIL, "%6.0f net additional unconditional jumps executed",
	jumpsExecuted);

    return;
}



/*
 *  RemoveEntryExitReturnFuncallEdges() --
 *
 *  Entry, Exit, Return and Funcall edges don't interest us because they have
 *  source or destination blocks that don't actually show up in the executable.
 *  FUNDIRECT edges will allow functions to be placed close to their callees.
 */
void RemoveEntryExitReturnFuncallEdges() {
    Block *biterator;
    Function *fiterator;
    EdgeNode *niterator, *next;

    FOREACH_FUNCTION(fiterator) {
	BBLRemoveEdges(fiterator->f_entry);
	BBLRemoveEdges(fiterator->f_exit);
    }

    FOREACH_BBL(biterator) {
	niterator = biterator->b_succ;
	while (niterator != NULL) {
	    next = niterator->en_next;
	    if ((NODE_TYPE(niterator) == ET_FUNCALL) ||
		(NODE_TYPE(niterator) == ET_RETURN) ||
		(NODE_TYPE(niterator) == ET_EXIT) ||
		(NODE_TYPE(niterator) == ET_ENTRY))
		BBLRemoveEdgeTyped(biterator, NODE_TO(niterator),
		    NODE_TYPE(niterator));
	    niterator = next;
	}
    }

    return;
}



/*
 *  RemoveEntryExitBlocks() --
 *
 *  Removes BT_EXIT and BT_ENTRY blocks from the list of basic blocks.
 *  One should call RemoveEntryExitReturnFuncallEdges before making a
 *  call to this function.
 */
void RemoveEntryExitBlocks() {
    Block *biterator;

    FOREACH_BBL(biterator) {
	if ((biterator->b_type != BT_ENTRY) && (biterator->b_type != BT_EXIT)) {
	    continue;
	}
	BBLDisconnect(biterator);
	BBLFree(biterator);
    }

    return;
}



/*
 *  BBLGetNextEdgeForChain(block) --
 *
 *  Iterates over the list of edges in a basic block chain, and finds the
 *  next one that does not yet have an ET_CHAIN edge associated with it.
 */
static Edge *BBLGetNextEdgeForChain(Block *block) {
    EdgeNode *node;
    Block *biterator, *last;

    last = block;
    while (last->b_chainnext != NULL) {
	last = last->b_chainnext;
    }

    biterator = block->b_nexttochain;
    while (biterator != NULL) {
	FOREACH_SUCC_IN_BBL(biterator, node) {
	    if (NODE_TYPE(node) == ET_CHAIN) {
		continue;
	    }
	    if (NODE_TYPE(node) == ET_HELL) {
		continue;
	    }
	    if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_PH) {
		if (isInterproceduralEdge(node->en_edge)) {
		    continue;
		}
	    }
	    if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_ALTO) {
		if (!BBLInSameTemperatureSetAsBBL(NODE_FROM(node),
			NODE_TO(node))) {
		    continue;
		}
	    }
	    if (NODE_TO(node)->b_chainhead == NODE_FROM(node)->b_chainhead) {
		continue;
	    }
	    if (!ChainEdgeExists(last, NODE_TO(node)->b_chainhead)) {
		return node->en_edge;
	    }
	}
	biterator = biterator->b_chainnext;
	block->b_nexttochain = biterator;
    }
    return NULL;
}



/*
 *  EdgeAggregateChainWeight(src,dest) --
 *
 *  Computes and returns the aggregate weight from one basic block
 *  chain to another.
 */
static EDGE_WEIGHT_TYPE EdgeAggregateChainWeight(Block *src, Block *dest) {
    EDGE_WEIGHT_TYPE weight;
    Block *biterator;
    EdgeNode *nodes;

    weight = 0;
    biterator = src;
    while (biterator != NULL) {
	FOREACH_SUCC_IN_BBL(biterator, nodes) {
	    if (NODE_TO(nodes)->b_chainhead == dest) {
		weight = weight + NODE_WEIGHT(nodes);
	    }
	}
	biterator = biterator->b_chainnext;
    }

    return weight;
}



/*
 *  ChainEdgeExists(src,dest) --
 *
 *  Returns true if there exists a chain edge between the two given chains.
 */
static bool ChainEdgeExists(Block *src, Block *dest) {
    EdgeNode *nodes;

    FOREACH_SUCC_IN_BBL(src, nodes) {
	if ((NODE_TO(nodes) == dest) && (NODE_TYPE(nodes) == ET_CHAIN)) {
	    return true;
	}
    }
    return false;
}



/*
 *  BBLRemoveChainEdges() --
 *
 *  Removes all edges of type ET_CHAIN from the control flow graph.
 */
static void BBLRemoveChainEdges() {
    Block *blocks, *heads;
    EdgeNode *niterator, *next;

    heads = chain;
    while (heads != NULL) {
	blocks = heads;
	while (blocks != NULL) {
	    niterator = blocks->b_succ;
	    while (niterator != NULL) {
		next = niterator->en_next;
		if (NODE_TYPE(niterator) == ET_CHAIN) {
		    BBLRemoveEdge(blocks, NODE_TO(niterator));
		}
		niterator = next;
	    }
	    blocks = blocks->b_chainnext;
	}
	heads->b_nexttochain = heads;
	heads = heads->b_headnext;
    }
    return;
}



/*
 *  EdgeSort() --
 *
 *  Sorts all the edges linking basic blocks, and chains them together
 *  such that the edge at the start of the list (edge_list) has the
 *  smallest weight, and all the edges following it have weights in a
 *  nondecreasing order.
 */
static void EdgeSort() {
    Edge *eiterator, *current;
    Edge **eiteratorArray;
    Block *biterator;
    Function *fun;
    EdgeNode *niterator;
    int count = 0;
    int offset = 0;
    int i;

    if (BBL_LAYOUT_SCHEME == BBL_LAYOUT_PH) {
	fun = chain->b_fun;
	FOREACH_BBL_IN_FUN(fun, biterator) {
	    FOREACH_SUCC_IN_BBL(biterator, niterator) {
		if (NODE_TYPE(niterator) == ET_CHAIN) {
		    if (edge_list != NULL) {
			niterator->en_edge->e_next = edge_list;
			edge_list->e_prev = niterator->en_edge;
			edge_list = niterator->en_edge;
		    } else {
			edge_list = niterator->en_edge;
			edge_list->e_next = edge_list->e_prev = NULL;
		    }
		}
	    }
	}
    } else {
	FOREACH_BBL(biterator) {
	    FOREACH_SUCC_IN_BBL(biterator, niterator) {
		if (NODE_TYPE(niterator) == ET_CHAIN) {
		    if (edge_list != NULL) {
			niterator->en_edge->e_next = edge_list;
			edge_list->e_prev = niterator->en_edge;
			edge_list = niterator->en_edge;
		    } else {
			edge_list = niterator->en_edge;
			edge_list->e_next = edge_list->e_prev = NULL;
		    }
		}
	    }
	}
    }

    /* Count the number of edges we have */
    eiterator = edge_list;
    while (eiterator != NULL) {
	count++;
	eiterator = eiterator->e_next;
    }

    eiteratorArray = zalloc(sizeof(Edge *) * count);

    /* Put the edges in an array */
    eiterator = edge_list;
    offset = 0;
    while (eiterator != NULL) {
	eiteratorArray[offset++] = eiterator;
	eiterator = eiterator->e_next;
    }

    /* Sort the edge array */
    qsort(eiteratorArray, count, sizeof(Edge *), EdgeCompareGreedy);

    for (i = 0; i < count; i++) {
	eiteratorArray[i]->e_next = NULL;
	eiteratorArray[i]->e_prev = NULL;
    }

    /* Re-link edges based on the sorted order in the array */
    edge_list = current = eiteratorArray[0];
    for (i = 1; i < count; i++) {
	eiteratorArray[i]->e_prev = current;
	current->e_next = eiteratorArray[i];
	current = current->e_next;
    }

    free(eiteratorArray);

    return;
}



/*
 *  EdgeCompareGreedy(first,second) --
 *
 *  Returns a number greater than 1 if the first edge is less than the
 *  second edge. Returns a number less than 1 if the second edge is
 *  less than the first edge.  Returns 0 if both edges have equal
 *  weight. Note that this will sort the list in a non-increasing order
 */
static int EdgeCompareGreedy(const void *first, const void *second) {
    long long weightFirst, weightSecond;

    weightFirst = (*(Edge **) first)->e_edgeWeight;
    weightSecond = (*(Edge **) second)->e_edgeWeight;

    if (EDGE_HAS_FLAG((*(Edge **) first), EDGE_ONLY_IN_FLAG) ||
	EDGE_HAS_FLAG((*(Edge **) first), EDGE_ONLY_OUT_FLAG))
	weightFirst = weightFirst * EDGE_SOLO_ADJUST;

    if (EDGE_HAS_FLAG((*(Edge **) second), EDGE_ONLY_IN_FLAG) ||
	EDGE_HAS_FLAG((*(Edge **) second), EDGE_ONLY_OUT_FLAG))
	weightSecond = weightSecond * EDGE_SOLO_ADJUST;

    if (weightFirst > weightSecond) {
	return -1;
    }
    if (weightFirst < weightSecond) {
	return 1;
    }

    return 0;
}



/*
 *  BBLMarkEdgesToBeAdjusted() --
 *
 *  The inner-doings of this function are based upon empirically gathered
 *  data. We note that Pettis/Hansen does not always produce the best
 *  basic block layout.  Consider the following example:
 *         _________         _________
 *         |       |         |       |
 *         | jcc X |         |       |
 *         ---------         ---------
 *           /   \               |
 *   False  /     \ True         / Normal
 *  wt=0   /       \  wt=1000   / wt = 999
 *        /         \          /
 *       /           \        /
 *      /             \      /
 *                     \    /
 *                    --------
 *                    |      |
 *                    --------
 *
 *  Using Pettis/Hansen, we should choose the true edge and switch the
 *  condition on the jump statement to make it a fallthrough.  Then we
 *  will add an unconditional jump to the end of the normal block.  The
 *  observation is that not-adding an unconditional jump is better than
 *  non-executing a conditional jump. But how much better? Twice? 1.5
 *  times? The conditional jump still has to be fetched, decoded, executed
 *  (1 clock cycle), while the non-existent unconditional jump has none
 *  of these costs.  Thus, we would like to give preference to normal
 *  edges when their weight is close to that of another edge also going to
 *  the same block. The heuristic used is to give preference (1.5 times
 *  weight) to edges if they are the only edge going in to, or leaving, a
 *  basic block.
 */
void BBLMarkEdgesToBeAdjusted() {
    Block *biterator;

    /* Mark the lone-edges */
    FOREACH_BBL(biterator) {
	if (BBLNumberOfSuccessors(biterator) == 1) {
	    EDGE_SET_FLAG(biterator->b_succ->en_edge, EDGE_ONLY_OUT_FLAG);
	}
	if (BBLNumberOfPredecessors(biterator) == 1) {
	    EDGE_SET_FLAG(biterator->b_pred->en_edge, EDGE_ONLY_IN_FLAG);
	}
    }

    return;
}



/*
 *  BBLInSameTemperatureSetAsBBL(bbl1,bbl2) --
 *
 *  This function will return true iff the two supplied basic blocks are the
 *  same 'temperature' e.g. hot, cold or zero.
 */
static bool BBLInSameTemperatureSetAsBBL(Block *bbl1, Block *bbl2) {

    if (BBL_HAS_FLAG(bbl1, BBL_HOT_FLAG)) {
	if (BBL_HAS_FLAG(bbl2, BBL_HOT_FLAG)) {
	    return true;
	} else {
	    return false;
	}
    }

    if (BBL_HAS_FLAG(bbl1, BBL_COLD_FLAG)) {
	if (BBL_HAS_FLAG(bbl2, BBL_COLD_FLAG)) {
	    return true;
	} else {
	    return false;
	}
    }

    if (BBL_HAS_FLAG(bbl1, BBL_ZERO_FLAG)) {
	if (BBL_HAS_FLAG(bbl2, BBL_ZERO_FLAG)) {
	    return true;
	} else {
	    return false;
	}
    }

    WARNING("[BBLInSameTemperatureSetAsBBL]"
	" No temperature group found for bbls.");
    return false;
}



/*
 *  GetNumUnknownEdges() --
 *
 *  This function computes the number of edges in the CFG having a weight of
 *  -1, which flags that it is unknown. It prints the WARNING message if the
 *  weight is negative (< -1).
 */
static int GetNumUnknownEdges() {
    int unknown_edges = 0;
    Block *biterator;
    EdgeNode *niterator;
    Function *fiterator;

    FOREACH_BBL(biterator) {
        FOREACH_SUCC_IN_BBL(biterator, niterator) {
            if (NODE_WEIGHT(niterator) == -1) {
                unknown_edges++;
            } else if (NODE_WEIGHT(niterator) < -1) {
                WARNING("Negative edge weight (%lld) "
                    "from bbl %d (%lld) to bbl %d (%lld)",
                    NODE_WEIGHT(niterator),
                    NODE_FROM(niterator)->b_id,
                    NODE_FROM(niterator)->b_weight,
                    NODE_TO(niterator)->b_id,
                    NODE_TO(niterator)->b_weight);
            }
        }
    }

    FOREACH_FUNCTION(fiterator) {
        Block *entry = fiterator->f_entry;
        Block *exit = fiterator->f_exit;

        FOREACH_SUCC_IN_BBL(entry, niterator) {
            if (NODE_WEIGHT(niterator) == -1) {
                unknown_edges++;
            } else if (NODE_WEIGHT(niterator) < -1) {
                WARNING("Negative edge weight (%lld) from entry bbl %d",
                    NODE_WEIGHT(niterator), entry->b_id);
            }
        }

        FOREACH_SUCC_IN_BBL(exit, niterator) {
            if (NODE_WEIGHT(niterator) == -1) {
                unknown_edges++;
            } else if (NODE_WEIGHT(niterator) < -1) {
                WARNING("Negative edge weight (%lld) from exit bbl %d",
                    NODE_WEIGHT(niterator), exit->b_id);
            }
        }
    }

    FOREACH_SUCC_IN_BBL(HELL, niterator) {
        if (NODE_WEIGHT(niterator) == -1) {
            unknown_edges++;
        } else if (NODE_WEIGHT(niterator) < -1) {
            WARNING("Negative edge weight (%lld) from HELL bbl %d",
                    NODE_WEIGHT(niterator), HELL->b_id);
        }
    }

    return unknown_edges;
}



/*
 *  PrintUnknownEdgeTypes() --
 *
 *  This function iterates through all edges and for the ones with unknown
 *  weight computes the number of them by types as well as total. Only edges
 *  from BT_BRANCH, BT_SWITCH and BT_EXIT are counted. For the other edge types
 *  weights are usually computed.
 */
static void PrintUnknownEdgeTypes() {
    int unknown_edges = 0;
    int bt_branch = 0;
    int bt_switch = 0;
    int bt_exit = 0;
    int bt_xxx = 0;
    Block *biterator;
    EdgeNode *niterator;
    Function *fiterator;

    FOREACH_BBL(biterator) {
        FOREACH_SUCC_IN_BBL(biterator, niterator) {
            if (NODE_WEIGHT(niterator) == -1) {
                unknown_edges++;
                if (biterator->b_type == BT_BRANCH) {
                    bt_branch++;
                } else if (biterator->b_type == BT_SWITCH) {
                    bt_switch++;
                }
            }
        }
    }

    FOREACH_FUNCTION(fiterator) {
        Block *entry = fiterator->f_entry;
        Block *exit = fiterator->f_exit;

        FOREACH_SUCC_IN_BBL(entry, niterator) {
            if (NODE_WEIGHT(niterator) == -1) {
                unknown_edges++;
            }
        }

        FOREACH_SUCC_IN_BBL(exit, niterator) {
            if (NODE_WEIGHT(niterator) == -1) {
                unknown_edges++;
                bt_exit++;
            }
        }
    }

    FOREACH_SUCC_IN_BBL(HELL, niterator) {
        if (NODE_WEIGHT(niterator) == -1) {
            unknown_edges++;
        }
    }

    bt_xxx = unknown_edges - bt_branch - bt_switch - bt_exit;
    REPORT(RP_SMLOOP, " from: BT_BRANCH - %d, BT_SWITCH - %d, BT_EXIT - %d,"
        " ? - %d", bt_branch, bt_switch, bt_exit, bt_xxx);

    return;
}



/*
 *  ResetEdgeWeights(new_weight) --
 *
 *  Resets the weight in all edges to new_weight. Returns the number of edges
 *  traversed.
 */
static int ResetEdgeWeights(EDGE_WEIGHT_TYPE new_weight) {
    int edges_traversed = 0;
    Block *biterator;
    EdgeNode *niterator;
    Function *fiterator;

    FOREACH_BBL(biterator) {
        FOREACH_SUCC_IN_BBL(biterator, niterator) {
            NODE_WEIGHT(niterator) = new_weight;
            edges_traversed++;
        }
    }

    FOREACH_FUNCTION(fiterator) {
        Block *entry = fiterator->f_entry;
        Block *exit = fiterator->f_exit;

        FOREACH_SUCC_IN_BBL(entry, niterator) {
            NODE_WEIGHT(niterator) = new_weight;
            edges_traversed++;
        }

        FOREACH_SUCC_IN_BBL(exit, niterator) {
            NODE_WEIGHT(niterator) = new_weight;
            edges_traversed++;
        }
    }

    FOREACH_SUCC_IN_BBL(HELL, niterator) {
        NODE_WEIGHT(niterator) = new_weight;
        edges_traversed++;
    }

    return edges_traversed;
}



/*
 *  EdgeComputeWeights() --
 *
 *  For each edge BBLComputeEdges computes its weight based on the weights of
 *  the basic blocks it connects.
 */
void EdgeComputeWeights() {
    Block *biterator;
    EdgeNode *niterator;
    Function *fiterator;
    int total_edges;
    int edge_weights_updated;
    int edges_remains;

    REPORT(RP_MINOR, "Computing edge weights");

    /* set all edge weights to -1 */
    total_edges = edges_remains = ResetEdgeWeights(-1);
    REPORT(RP_DETAIL, "Total number of edges: %7d", GetNumUnknownEdges());

    /*
     * For all non-fake basic blocks with weight 0 set all incoming and
     * outgoing edge weights to be 0.
     */
    edge_weights_updated = 0;
    FOREACH_BBL(biterator) {
        if (biterator->b_weight == 0) {
            FOREACH_PRED_IN_BBL(biterator, niterator) {
                if (NODE_WEIGHT(niterator) == -1) {
                    edge_weights_updated++;
                    NODE_WEIGHT(niterator) = 0;
                }
            }

            FOREACH_SUCC_IN_BBL(biterator, niterator) {
                if (NODE_WEIGHT(niterator) == -1) {
                    edge_weights_updated++;
                    NODE_WEIGHT(niterator) = 0;
                }
            }
        }
    }

    edges_remains -= edge_weights_updated;
    REPORT(RP_SMLOOP, "Set 0-bbl edge weights to 0 in %d edges, %d remains.",
        edge_weights_updated, edges_remains);

    /* set all ET_FUNCALL and ET_ENTRY edges to 0 */
    edge_weights_updated = 0;
    FOREACH_FUNCTION(fiterator) {
        Block *entry = fiterator->f_entry;

        FOREACH_PRED_IN_BBL(entry, niterator) {
            if (NODE_WEIGHT(niterator) == -1) {
                edge_weights_updated++;
                NODE_WEIGHT(niterator) = 0;
            }
        }

        FOREACH_SUCC_IN_BBL(entry, niterator) {
            if (NODE_WEIGHT(niterator) == -1) {
                edge_weights_updated++;
                NODE_WEIGHT(niterator) = 0;
            }
        }
    }

    edges_remains -= edge_weights_updated;
    REPORT(RP_SMLOOP, "Set entry bbl edge weights to 0 in %d edges, "
        "%d remains.", edge_weights_updated, edges_remains);

    /*
     * Set all FUNLINK edges to 0, since execution never falls through from
     * the caller to the next block. Since, it still returns there back from
     * the callee, assign the weight of caller to either ET_RETURN, if present,
     * or to ET_HELLMAYBE. Also, since now fake edges of BT_RETURN will have
     * weights, assign caller's weight to BT_RETURN, too.
     */
    edge_weights_updated = 0;
    FOREACH_BBL(biterator) {
        FOREACH_SUCC_IN_BBL(biterator, niterator) {
            if (NODE_WEIGHT(niterator) == -1) {
                if ((NODE_TYPE(niterator) == ET_FUNLINK)) {
                    EdgeNode *niterator2;
                    Block *b_fall_through = NODE_TO(niterator);
                    edge_weights_updated++;
                    NODE_WEIGHT(niterator) = 0;
                    FOREACH_PRED_IN_BBL(b_fall_through, niterator2) {
                        if (NODE_TYPE(niterator2) == ET_RETURN) {
                            Block *b_exit = NODE_FROM(niterator2);
                            b_exit->b_weight += biterator->b_weight;
                            if (NODE_WEIGHT(niterator2) == -1) {
                                edge_weights_updated++;
                                NODE_WEIGHT(niterator2) = biterator->b_weight;
                            }
                        }
                        if (NODE_TYPE(niterator2) == ET_HELLMAYBE) {
                            if (NODE_WEIGHT(niterator2) == -1) {
                                edge_weights_updated++;
                                NODE_WEIGHT(niterator2) = biterator->b_weight;
                            }
                        }
                    }
                }
            }
        }
    }

    edges_remains -= edge_weights_updated;
    REPORT(RP_SMLOOP, "Set FUNLINK and RETURN edge weights in %d edges, "
        "%d remains.", edge_weights_updated, edges_remains);

    /* repeat the below until no new edge weights have been computed */
    do {
        int unknown_edges_before = GetNumUnknownEdges();
        edge_weights_updated = 0;

        /*
         * For each bbl with only one unknown incoming/outgoing edge, compute
         * its weight as a difference between the bbl weight and the weights of
         * the rest of the incoming/outgoing edges.
         */
        FOREACH_BBL(biterator) {
            int num_unknown_edges = 0;
            EDGE_WEIGHT_TYPE weight_of_known_edges = 0;
            EdgeNode *unknown_node = NULL;

            FOREACH_PRED_IN_BBL(biterator, niterator) {
                if (NODE_WEIGHT(niterator) == -1) {
                    num_unknown_edges++;
                    unknown_node = niterator;
                } else {
                    weight_of_known_edges += NODE_WEIGHT(niterator);
                }
            }

            if (num_unknown_edges == 1) {
                edge_weights_updated++;
                NODE_WEIGHT(unknown_node) =
                    biterator->b_weight - weight_of_known_edges;
            }

            num_unknown_edges = 0;
            weight_of_known_edges = 0;
            unknown_node = NULL;

            FOREACH_SUCC_IN_BBL(biterator, niterator) {
                if (NODE_WEIGHT(niterator) == -1) {
                    num_unknown_edges++;
                    unknown_node = niterator;
                } else {
                    weight_of_known_edges += NODE_WEIGHT(niterator);
                }
            }

            if (num_unknown_edges == 1) {
                edge_weights_updated++;
                NODE_WEIGHT(unknown_node) =
                    biterator->b_weight - weight_of_known_edges;
            }
        }

        /*
         * If all the bbl weight has been distributed between the
         * incoming/outgoing edges, but we still have unknown weights for the
         * rest of the incoming/outgoing edges, their weights must be 0.
         */
        FOREACH_BBL(biterator) {
            int num_unknown_edges = 0;
            EDGE_WEIGHT_TYPE weight_of_known_edges = 0;

            FOREACH_PRED_IN_BBL(biterator, niterator) {
                if (NODE_WEIGHT(niterator) == -1) {
                    num_unknown_edges++;
                } else {
                    weight_of_known_edges += NODE_WEIGHT(niterator);
                }
            }

            if (num_unknown_edges > 0
                && weight_of_known_edges == biterator->b_weight) {
                FOREACH_PRED_IN_BBL(biterator, niterator) {
                    if (NODE_WEIGHT(niterator) == -1) {
                        edge_weights_updated++; 
                        NODE_WEIGHT(niterator) = 0;
                    }
                }
            }

            num_unknown_edges = 0;
            weight_of_known_edges = 0;

            FOREACH_SUCC_IN_BBL(biterator, niterator) {
                if (NODE_WEIGHT(niterator) == -1) {
                    num_unknown_edges++;
                } else {
                    weight_of_known_edges += NODE_WEIGHT(niterator);
                }
            }

            if (num_unknown_edges > 0
                && weight_of_known_edges == biterator->b_weight) {
                FOREACH_SUCC_IN_BBL(biterator, niterator) {
                    if (NODE_WEIGHT(niterator) == -1) {
                        edge_weights_updated++;
                        NODE_WEIGHT(niterator) = 0;
                    }
                }
            }
        }

        edge_weights_updated = unknown_edges_before - GetNumUnknownEdges();
        edges_remains -= edge_weights_updated;
        REPORT(RP_SMLOOP, "Updated edge weights in %d in/out edges, "
            "%d remains.", edge_weights_updated, edges_remains);
    } while (edge_weights_updated > 0);

    /* repeat the below until no new edge weights have been computed */
    do {
        int unknown_edges_before = GetNumUnknownEdges();
        edge_weights_updated = 0;

        /*
         * For each exit bbl with only one unknown incoming edge, compute
         * its weight as a difference between the bbl weight and the weights of
         * the rest of the incoming/outgoing edges.
         */
        FOREACH_FUNCTION(fiterator) {
            Block *exit_bbl = fiterator->f_exit;
            int num_unknown_preds = 0;
            EDGE_WEIGHT_TYPE weight_of_known_preds = 0;
            EdgeNode *unknown_pred = NULL;

            FOREACH_PRED_IN_BBL(exit_bbl, niterator) {
                if (NODE_WEIGHT(niterator) == -1) {
                    num_unknown_preds++;
                    unknown_pred = niterator;
                } else {
                    weight_of_known_preds += NODE_WEIGHT(niterator);
                }

                if (num_unknown_preds == 1) {
                    if (exit_bbl->b_weight >= weight_of_known_preds) {
                        edge_weights_updated++;
                        NODE_WEIGHT(unknown_pred) =
                            exit_bbl->b_weight - weight_of_known_preds;
                    }
                }
            }
        }

        /*
         * If all the exit bbl weight has been distributed between the
         * incoming edges, but we still have unknown weights for the
         * rest of them, their weights must be 0.
         */
        FOREACH_FUNCTION(fiterator) {
            Block *exit_bbl = fiterator->f_exit;
            int num_unknown_preds = 0;
            EDGE_WEIGHT_TYPE weight_of_known_preds = 0;

            FOREACH_PRED_IN_BBL(exit_bbl, niterator) {
                if (NODE_WEIGHT(niterator) == -1) {
                    num_unknown_preds++;
                } else {
                    weight_of_known_preds += NODE_WEIGHT(niterator);
                }
            }

            if (num_unknown_preds > 0 && 
                weight_of_known_preds >= exit_bbl->b_weight) {
                FOREACH_PRED_IN_BBL(exit_bbl, niterator) {
                    if (NODE_WEIGHT(niterator) == -1) {
                        edge_weights_updated++;
                        NODE_WEIGHT(niterator) = 0;
                    }
                }
            }
        }

        edge_weights_updated = unknown_edges_before - GetNumUnknownEdges();
        edges_remains -= edge_weights_updated;
        REPORT(RP_SMLOOP, "Updated edge weights in %d exit edges, %d remains.",
            edge_weights_updated, edges_remains);
    } while (edge_weights_updated > 0);

    /*
     * The last resort - to set all unknown HELL and HELLMAYBE edges to 0,
     * since we are not profiling HELL anyways.
     */
    edge_weights_updated = 0;
    FOREACH_PRED_IN_BBL(HELL, niterator) {
        if (NODE_WEIGHT(niterator) == -1) {
            edge_weights_updated++; 
            NODE_WEIGHT(niterator) = 0;
        }
    }
    FOREACH_SUCC_IN_BBL(HELL, niterator) {
        if (NODE_WEIGHT(niterator) == -1) {
            edge_weights_updated++;        
            NODE_WEIGHT(niterator) = 0;
        }
    }
    edges_remains -= edge_weights_updated;
    REPORT(RP_SMLOOP, "Set HELL edge weights to %d in %d edges, %d remains.",
        0, edge_weights_updated, edges_remains);
    REPORT(RP_DETAIL, "Edge weights computed: %7d", total_edges-edges_remains);
    REPORT(RP_DETAIL, "Edge weights unknown:  %7d", GetNumUnknownEdges());
    PrintUnknownEdgeTypes();

    return;
}



/*
 *  BBLComputeWeights() --
 *
 *  For each basic block BBLComputeWeights will add up the weights of the
 *  predecessor edges, then add up the weights of the successor edges, then
 *  assign to the block the highest of the two sums.
 */
void BBLComputeWeights() {
    Block *biterator;

    REPORT(RP_MINOR, "Computing bbl weights");

    FOREACH_BBL(biterator) {
        biterator->b_weight = 0;
        BBLUpdateWeight(biterator);
    }

    return;
}



/*
 *  BBLAssignTemperature() --
 *
 *  BBLAssignTemperature constructs hot/cold/zero sets based on the weights of
 *  the basic blocks and on the hot code threshold.
 */
void BBLAssignTemperature() {
    Block *biterator;
    EDGE_WEIGHT_TYPE sum = 0;
    EDGE_WEIGHT_TYPE current = 0;
    int bblCount = 0;
    Block **bblArray;
    int offset;
    float thresh = FloatParameter("BBLHotThreshold");

    REPORT(RP_MINOR, "Assigning BBL temperature");

    FOREACH_BBL(biterator) {
        BBL_REMOVE_FLAG(biterator,
            BBL_HOT_FLAG | BBL_ZERO_FLAG | BBL_COLD_FLAG);
	sum = sum + biterator->b_weight;
	bblCount = bblCount + 1;
    }

    bblArray = zalloc(sizeof(Block *) * bblCount);

    offset = 0;
    FOREACH_BBL(biterator) {
	bblArray[offset] = biterator;
	offset++;
    }

    CASSERT(offset == bblCount, "Block counts did not match.");

    /* Sort the basic block array */
    qsort(bblArray, bblCount, sizeof(Block *), BBLCompare);

    lowest_hot_weight = (bblArray[0])->b_weight;
    offset = 0;
    while (current < (thresh * sum)) {
	BBL_SET_FLAG(bblArray[offset], BBL_HOT_FLAG);
	current = current + ((bblArray[offset])->b_weight);
	if (((bblArray[offset])->b_weight) < lowest_hot_weight) {
	    lowest_hot_weight = (bblArray[offset])->b_weight;
	}
	offset++;
    }

    CASSERT(offset < bblCount, "Wrote past array bounds.");
    CASSERT(current >= (thresh * sum), "Hot blocks not set appropriately.");

    while (true) {

	if (((bblArray[offset])->b_weight) == 0) {
	    break;
	}

	CASSERT(offset < bblCount, "Wrote past array bounds.");

	BBL_SET_FLAG(bblArray[offset], BBL_COLD_FLAG);
	offset++;
    }

    if (offset >= bblCount) {
	WARNING("[BBLAssignTemperature] Suspicious--No zero set blocks.");
    }

    CASSERT(offset < bblCount, "Wrote past array bounds.");

    while (offset < bblCount) {
	BBL_SET_FLAG(bblArray[offset], BBL_ZERO_FLAG);
	offset++;
    }

    free(bblArray);

    return;
}



/*
 *  BBLAssignWeights() --
 *
 *  BBLAssignWeights will add up the weights of the predecessor edges for all
 *  blocks, and assign to the blocks a weight (equal to the sum of the
 *  weights of the predecessors). Additionally, if BBL_LAYOUT_SCHEME is
 *  the scheme used in the alpha link-time optimizer, hot/cold/zero sets will
 *  be constructed at this point.
 */
void BBLAssignWeights() {
    BBLComputeWeights();
    BBLAssignTemperature();
    return;
}



/*
 *  BBLCompare(first,second) --
 *
 *  Returns a number greater than 1 if the first bbl is less than the
 *  second bbl. Returns a number less than 1 if the second bbl is
 *  less than the first bbl.  Returns 0 if both bbls have equal
 *  weight. Note that this will sort the list in a non-increasing order
 */
int BBLCompare(const void *first, const void *second) {

    if (((*(Block **) first)->b_weight) > ((*(Block **) second)->b_weight)) {
	return -1;
    }

    if (((*(Block **) first)->b_weight) < ((*(Block **) second)->b_weight)) {
	return 1;
    }

    return 0;
}



/*
 *  FunConnectChains(src,dest) -- connects two fun chains together.
 */
static void FunConnectChains(Function *src, Function *dest) {
    Function *fiterator;

    /* Find the end of the source chain */
    fiterator = src;
    while (fiterator->f_chainnext != NULL) {
	fiterator = fiterator->f_chainnext;
    }

    /* Link the destination at the end of the source chain */
    fiterator->f_chainnext = dest;
    dest->f_chainprev = fiterator;

    /* Tell all the blocks about their new head */
    fiterator = dest;
    while (fiterator != NULL) {
	fiterator->f_chainhead = src->f_chainhead;
	fiterator = fiterator->f_chainnext;
    }

    /* Update the chain headers */
    if (dest->f_headprev) {
	dest->f_headprev->f_headnext = dest->f_headnext;
    }
    if (dest->f_headnext) {
	dest->f_headnext->f_headprev = dest->f_headprev;
    }
    if (fchain == dest) {
	fchain = fchain->f_headnext;
    }
    dest->f_headnext = dest->f_headprev = NULL;

    return;
}



/*
 *  BBLConnectChains(src,dest) -- connects two bbl chains together.
 */
static void BBLConnectChains(Block *src, Block *dest) {
    Block *biterator;

    /* Find the end of the source chain */
    biterator = src;
    while (biterator->b_chainnext != NULL) {
	biterator = biterator->b_chainnext;
    }

    /* Link the destination at the end of the source chain */
    biterator->b_chainnext = dest;
    dest->b_chainprev = biterator;

    /* Tell all the blocks about their new head */
    biterator = dest;
    while (biterator != NULL) {
	biterator->b_chainhead = src->b_chainhead;
	biterator = biterator->b_chainnext;
    }

    /* Update the chain headers */
    if (dest->b_headprev) {
	dest->b_headprev->b_headnext = dest->b_headnext;
    }
    if (dest->b_headnext) {
	dest->b_headnext->b_headprev = dest->b_headprev;
    }
    if (chain == dest) {
	chain = chain->b_headnext;
    }
    dest->b_headnext = dest->b_headprev = NULL;

    return;
}



/*
 *  BBLTemperatureChain() --
 *
 *  Links the chain based on their temperature (hot, cold, zero). After
 *  this function has been called, three chains should remain: one for
 *  each group.
 */
static void BBLTemperatureChain() {
    Block *blocks, *head;
    Block *firstHot, *firstCold, *firstZero, *next;

    firstHot = firstCold = firstZero = next = NULL;

    blocks = chain;
    while (blocks != NULL) {
	if (BBL_HAS_FLAG(blocks, BBL_HOT_FLAG)) {
	    firstHot = blocks;
	    break;
	}
	blocks = blocks->b_headnext;
    }

    blocks = chain;
    while (blocks != NULL) {
	if (BBL_HAS_FLAG(blocks, BBL_COLD_FLAG)) {
	    firstCold = blocks;
	    break;
	}
	blocks = blocks->b_headnext;
    }

    blocks = chain;
    while (blocks != NULL) {
	if (BBL_HAS_FLAG(blocks, BBL_ZERO_FLAG)) {
	    firstZero = blocks;
	    break;
	}
	blocks = blocks->b_headnext;
    }

    blocks = chain;
    CASSERT(firstHot != NULL, "No HOT  blocks found.");
    CASSERT(firstCold != NULL, "No COLD blocks found.");
    CASSERT(firstZero != NULL, "No ZERO blocks found.");

    blocks = firstHot->b_headnext;
    while (blocks != NULL) {
	next = blocks->b_headnext;
	if (BBL_HAS_FLAG(blocks, BBL_HOT_FLAG)) {
	    BBLConnectChains(firstHot, blocks);
	}
	blocks = next;
    }

    if (firstCold != NULL) {
	blocks = firstCold->b_headnext;
	while (blocks != NULL) {
	    next = blocks->b_headnext;
	    if (BBL_HAS_FLAG(blocks, BBL_COLD_FLAG)) {
		BBLConnectChains(firstCold, blocks);
	    }
	    blocks = next;
	}
    }

    blocks = firstZero->b_headnext;
    while (blocks != NULL) {
	next = blocks->b_headnext;
	if (BBL_HAS_FLAG(blocks, BBL_ZERO_FLAG)) {
	    BBLConnectChains(firstZero, blocks);
	}
	blocks = next;
    }

    blocks = head = chain;
    while (blocks != NULL) {
	CASSERT(BBLInSameTemperatureSetAsBBL(blocks, head),
	    "Incorrect temperature set linkage.");
	blocks = blocks->b_chainnext;
    }

    blocks = head = chain->b_headnext;
    while (blocks != NULL) {
	CASSERT(BBLInSameTemperatureSetAsBBL(blocks, head),
	    "Incorrect temperature set linkage.");
	blocks = blocks->b_chainnext;
    }

    blocks = head = chain->b_headnext->b_headnext;
    while (blocks != NULL) {
	CASSERT(BBLInSameTemperatureSetAsBBL(blocks, head),
	    "Incorrect temperature set linkage.");
	blocks = blocks->b_chainnext;
    }

    BBLConnectChains(firstHot, firstCold);
    BBLConnectChains(firstHot, firstZero);

    return;
}



/*
 *  FunctionHasHotBBL(fun) --
 *
 *  Returns true if 'fun' contains a basic block that is hot.
 */
bool FunctionHasHotBBL(Function *fun) {
    Block *bbl;

    FOREACH_BBL_IN_FUN(fun, bbl) {
	if (BBL_HAS_FLAG(bbl, BBL_HOT_FLAG)) {
	    return true;
	}
    }
    return false;
}



/*
 *  BBLUpdateWeight(block) --
 *
 *  Updates the weight for the basic block. weight = max(incoming weights,
 *  outgoing weights);
 */
void BBLUpdateWeight(Block *block) {
    EDGE_WEIGHT_TYPE incoming = BBLSumIncomingEdges(block);
    EDGE_WEIGHT_TYPE outgoing = BBLSumOutgoingEdges(block);

    if (incoming >= outgoing) {
	block->b_weight = incoming;
    } else {
	block->b_weight = outgoing;
    }

    return;
}
