/*
 *  Memory optimizations.
 */
#include "global.h"
#include "flowgraph.h"
#include "x86-classes.h"
#include "sysdep.h"
#include "opt.unreachable.h"
#include "opt.h"
#include "liveness.h"
#include "linkage.h"
#include "stack.analyze.h"
#include "x86-registers.h"
#include "cp.h"

#define LOAD_LATENCY  2

static int n = 0, nhot = 0;
static double wt = 0.0, totwt = 0.0;

extern bool RegBitvectorIsLiveAtInstr(REGISTER_BITVECTOR vec,
    Instruction *instr);
extern int registerToBitRegister(int someRegister);
extern int ComputeMyStackGainForInstr(Instruction *instr);



/*
 *  KillRedundantLoad(iins) --
 */
static void KillRedundantLoad(Instruction *iins) {
    n++;
    wt += iins->i_Bblock->b_weight;
    if (BBL_HAS_FLAG(iins->i_Bblock, BBL_HOT_FLAG)) {
	nhot++;
    }

    InstrDestroy(iins);
}



/*
 *  InstrLoadsRegFromMemory(iins) --
 *
 *  Returns true for MOVE instructions that load a value from memory into
 *  a register.
 */
bool InstrLoadsRegFromMemory(Instruction *iins) {
    Operand *src, *dst;

    CASSERT(iins != NULL, "null argument");

    if (iins->i_class != CLASS_MOV) {
	return false;
    }

    src = iins->i_src;
    dst = iins->i_dest;
    /*
     * for a load instruction, we want the source operand to be
     * a memory location.
     */
    if (!(OPERAND_HAS_FLAG(src, OPERAND_REGISTER_INDIRECT)
	    || OPERAND_HAS_FLAG(src, OPERAND_INDIRECT_FLAG)
	    || OperandIsAbsoluteMemoryLocation(src)))
	return false;
    /*
     * for a load instruction, we want the destination operand to be
     * a register.
     */
    if (!OPERAND_HAS_FLAG(dst, OPERAND_BASE_FLAG)
	|| OPERAND_HAS_FLAG(dst, OPERAND_DISP_R_FLAG)
	|| OPERAND_HAS_FLAG(dst, OPERAND_INDEX_FLAG)
	|| OPERAND_HAS_FLAG(dst, OPERAND_SCALE_FLAG)
	|| OPERAND_HAS_FLAG(dst, OPERAND_IMM_FLAG)
	|| OPERAND_HAS_FLAG(dst, OPERAND_DISP_A_FLAG)
	|| OPERAND_HAS_FLAG(dst, OPERAND_REGISTER_INDIRECT))
	return false;

    return true;
}



/*
 *  InstrStoresRegToMemory(iins) --
 *
 *  Returns true for MOVE instructions that store a value from a register
 *  into memory.
 */
bool InstrStoresRegToMemory(Instruction *iins) {
    Operand *src, *dst;

    CASSERT(iins != NULL, "null argument");

    if (iins->i_class != CLASS_MOV) {
	return false;
    }

    src = iins->i_src;
    dst = iins->i_dest;
    /*
     * for a store instruction, we want the destination operand to be
     * a memory location.
     */

    if (!(OPERAND_HAS_FLAG(dst, OPERAND_REGISTER_INDIRECT)
	    || OPERAND_HAS_FLAG(dst, OPERAND_DISP_R_FLAG)
	    || OPERAND_HAS_FLAG(dst, OPERAND_DISP_A_FLAG)))
	return false;
    /*
     * for a register store instruction, we want the source operand to be
     * a register.
     */
    if (!OPERAND_HAS_FLAG(src, OPERAND_BASE_FLAG)
	|| OPERAND_HAS_FLAG(src, OPERAND_DISP_R_FLAG)
	|| OPERAND_HAS_FLAG(src, OPERAND_INDEX_FLAG)
	|| OPERAND_HAS_FLAG(src, OPERAND_SCALE_FLAG)
	|| OPERAND_HAS_FLAG(src, OPERAND_IMM_FLAG)
	|| OPERAND_HAS_FLAG(src, OPERAND_DISP_A_FLAG)
	|| OPERAND_HAS_FLAG(src, OPERAND_REGISTER_INDIRECT))
	return false;

    return true;
}



/*
 *  InstrStoresToMemory(iins) --
 *
 *  Returns true for MOVE instructions that store a value into memory.
 */
bool InstrStoresToMemory(Instruction *iins) {
    Operand *dst;

    CASSERT(iins != NULL, "null argument");

    if (iins->i_class == CLASS_PUSH) {
	return true;
    }

    dst = iins->i_dest;
    if (!dst) {
	return false;
    }

    if (OPERAND_HAS_FLAG(dst, OPERAND_REGISTER_INDIRECT)) {
	return true;
    }
    if (OPERAND_HAS_FLAG(dst, OPERAND_DISP_R_FLAG)) {
	return true;
    }
    if (OPERAND_HAS_FLAG(dst, OPERAND_DISP_A_FLAG)) {
	return true;
    }

    return false;
}



/*
 *  OperandsAreIdentical(op1,op2) --
 *
 *  Returns true if op1 and op2 are identical operands.
 */
bool OperandsAreIdentical(Operand *op1, Operand *op2) {
    if (op1->a_flags != op2->a_flags) {
	return false;
    }
    if (op1->a_base_reg != op2->a_base_reg) {
	return false;
    }
    if (op1->a_index_reg != op2->a_index_reg) {
	return false;
    }
    if (op1->a_scale != op2->a_scale) {
	return false;
    }
    if (op1->a_displacement != op2->a_displacement) {
	return false;
    }
    if (op1->a_relocation != op2->a_relocation) {
	return false;
    }

    return true;
}



/*
 *  OperandsMayAlias(op1,op2) --
 *
 *  Returns true if the two (memory) operands may possibly be aliases.
 */
bool OperandsMayAlias(Operand *op1, Operand *op2) {
    bool op1IsStk, op2IsStk, op1IsAbs, op2IsAbs;

    op1IsStk = (OperandAccessesStackFromFramePointer(op1)
	|| OperandAccessesStackFromStackPointer(op1));

    op2IsStk = (OperandAccessesStackFromFramePointer(op2)
	|| OperandAccessesStackFromStackPointer(op2));

    op1IsAbs = OperandIsAbsoluteMemoryLocation(op1);
    op2IsAbs = OperandIsAbsoluteMemoryLocation(op2);

    if (op1IsStk && op2IsAbs) {
	return false;
    }
    if (op2IsStk && op1IsAbs) {
	return false;
    }

    if (op1->a_base_reg == op2->a_base_reg
	&& op1->a_index_reg == op2->a_index_reg
	&& op1->a_displacement != op2->a_displacement)
	return false;

    return true;
}



/*
 *  StkLocIsDeallocated(offset,iins) --
 *
 *  Returns true if the offset supplied, relative to the frame pointer,
 *  represents a deallocated stack location at the specified instruction.
 *
 *  Given that this function is intended for use in load/store forwarding,
 *  i.e., situations of the form
 *
 *  (1) load/store from mem_loc_1
 *  <some code>
 *  (2) load from mem_loc_1
 *
 *  we're interested in knowing whether mem_loc_1 is a stack location that
 *  gets deallocated in the code fragment <some code>: if so, instr (2) can't
 *  re-use the value loaded/stored by instr (1).  Because of this, if the
 *  stack pointer is not modified in <some code>, we consider mem_loc_1 to be
 *  not-deallocated.  Thus, this function indicates that location mem_loc_1
 *  is not deallocated if the instruction specified does not modify the stack
 *  pointer.  This behavior should be taken into account if this function is
 *  used elsewhere.
 */
bool StkLocIsDeallocated(int offset, Instruction *iins) {
    int bp_sp_diff = InstrBPSPDiff(iins);

    if (offset >= 0) {
	return false;
    }

    if (!InstrDefinesReg(iins, reg_sp)) {
	return false;
    }

    if (bp_sp_diff == STACK_UNKNOWN) {
	return true;
    }

    bp_sp_diff += ComputeMyStackGainForInstr(iins);

    if (offset < 0 && -offset > bp_sp_diff) {
	return true;
    }

    return false;
}



/*
 *  DefiningInstr(iins) --
 *
 *  Given an instruction iins that loads a register from memory, trace back
 *  along an extended basic block until we find an instruction jins that is a
 *  load from the same location as iins, and return jins.  If such an instr
 *  cannot be found within the extended basic block containing iins, return
 *  NULL.
 *
 *  MR -- Correction :
 *  The original procedure doesn't take into account the case when
 *  the 'iins' load location is based on the value of '%esp'.
 *  For example:
 *		mpeg_play.reloc : INSTR_ID(25083-25088) **
 *
 *		25083   movl 44(%esp,1)<- %eax
 *		   84   pushl %edx
 *		   85	pushl $8202
 *		   86   pushl %edi
 *		   87   pushl $1
 *		   88   mov  %eax <- 44(%esp,1)
 *
 *  Load-Store forwarding naively, deletes {25088}, since the locations
 *  seem to be the same. But note -- 44(esp,1) is loaded wrt esp -- which
 *  gets updated because of the PUSH instrs !!!
 *
 *  The new version explicitly checks for this case.
 */
static Instruction *DefiningInstr(Instruction *iins) {
    Block *jbbl;
    Instruction *jins;
    int basereg, indexreg;
    int offset;
    int stack_flag = 0;

    if (iins->i_src) {
	// --MR --
	// Is source1 on the stack through esp ??
	// if (iins->i_src->a_base_reg == IndexToRegister(REGISTER_INDEX_ESP)) {
	if (OperandAccessesStackFromStackPointer(iins->i_src)) {
	    stack_flag = 1;
	}

	if (OPERAND_HAS_FLAG(iins->i_src, OPERAND_BASE_FLAG)
	    && InstrDefinesReg(iins, iins->i_src->a_base_reg)) {
	    return NULL;
	}
	if (OPERAND_HAS_FLAG(iins->i_src, OPERAND_INDEX_FLAG)
	    && InstrDefinesReg(iins, iins->i_src->a_index_reg)) {
	    return NULL;
	}
    }

    if (iins->i_src2) {
	// --MR --
	// Is source2 on the stack through esp ??
	// if (iins->i_src2->a_base_reg==IndexToRegister(REGISTER_INDEX_ESP)) {
	if (OperandAccessesStackFromStackPointer(iins->i_src)) {
	    stack_flag = 1;
	}
	if (OPERAND_HAS_FLAG(iins->i_src2, OPERAND_BASE_FLAG)
	    && InstrDefinesReg(iins, iins->i_src2->a_base_reg)) {
	    return NULL;
	}
	if (OPERAND_HAS_FLAG(iins->i_src2, OPERAND_INDEX_FLAG)
	    && InstrDefinesReg(iins, iins->i_src2->a_index_reg)) {
	    return NULL;
	}
    }

    basereg = iins->i_src->a_base_reg;
    indexreg = iins->i_src->a_index_reg;

    for (jbbl = iins->i_Bblock, jins = iins->i_prev;
	jbbl != NULL;
	jbbl = NODE_FROM(jbbl->b_pred), jins = jbbl->b_last) {
#if 1
	/* This shouldn't be necessary, but is here to get around what seems to
	 * be a bug in flowgraph construction. */
	if (jbbl->b_type == BT_CALL && jbbl != iins->i_Bblock) {
	    return NULL;
	}
#endif

	while (jins != NULL && jins->i_Bblock == jbbl) {

	    if (InstrPushBytes(jins) && (stack_flag == 1)) {
		return NULL;
	    }

	    if (OperandAccessesStackFromFramePointer(iins->i_src)) {
		offset = iins->i_src->a_displacement;
		if (StkLocIsDeallocated(offset, jins)) {
		    return NULL;
		}
	    }

	    if (((InstrLoadsRegFromMemory(jins)
			&& OperandsAreIdentical(iins->i_src, jins->i_src))
		    || (InstrStoresRegToMemory(jins)
			&& OperandsAreIdentical(iins->i_src, jins->i_dest)))
		&& !(InstrDefinesReg(jins, basereg)
		    || InstrDefinesReg(jins, indexreg))) {
		return jins;
	    }

	    if (InstrStoresToMemory(jins)) {
		if (jins->i_class == CLASS_PUSH) {

		} else {
		    if (OperandsMayAlias(iins->i_src, jins->i_dest)) {
			return NULL;
		    }
		}
	    }

	    if (jins->i_dest) {
		if (OperandsAreIdentical(jins->i_dest, iins->i_dest)
		    || OperandsAreIdentical(jins->i_dest, iins->i_src)
		    || InstrDefinesReg(jins, basereg)
		    || InstrDefinesReg(jins, indexreg)) {
		    return NULL;
		}
	    }

	    jins = jins->i_prev;
	}				/* while */

	if (BBLNumberOfPredecessors(jbbl) != 1) {
	    return NULL;
	}
    }
    /* NOT REACHED */
    return NULL;
}



/*
 *  FixLivenessInfo(ibbl,jbbl,reg) --
 *
 *  FixLivenessInfo(ibbl, jbbl, reg) is called with basic block ibbl a
 *  descendant of jbbl, such that along the backward path from ibbl to jbbl
 *  each basic block has a single predecessor: i.e., jbbl...ibbl forms an
 *  extended basic block.  This routine traverses this path, adding register
 *  reg to the set of live registers at each basic block along the way, from
 *  IN(ibbl) back all the way up to OUT(jbbl).
 */
static void FixLivenessInfo(Block *ibbl, Block *jbbl, int reg) {
    Block *xbbl;
    REGISTER_BITVECTOR regs;

    if (ibbl == jbbl) {
	return;				// nothing to be done
    }

    regs = registerToBitRegister(reg);
    /*
     * Given the assumption that each basic block on the backward path
     * from ibbl to jbbl has exactly one predecessor, we just work our
     * way back.  A more careful person would check that this assumption
     * is actually satisfied along the way.
     */
    for (xbbl = ibbl; true; xbbl = NODE_FROM(xbbl->b_pred)) {

	if (xbbl != ibbl) {
	    xbbl->b_liveOut |= regs;
	}
	if (xbbl != jbbl) {
	    xbbl->b_liveIn |= regs;
	}

	if (xbbl == jbbl) {
	    break;
	}
    }
}



/*
 *  FwdLoadStore(ibbl,phase) --
 */
static void FwdLoadStore(Block *ibbl, int phase) {
    Block *ubbl, *xbbl;
    Instruction *iins, *jins, *xins, *def_iins, *def_jins, *mvins, *killins;
    int ireg, jreg;
    int iindreg, jindreg;
    bool ireg_is_live, bailout;
    int done = 0;
    
    killins = NULL;
    FOREACH_INSTR_IN_BBL(ibbl, iins) {
	if (!done) {
	    done = 1;
	}
	if (killins) {
	    KillRedundantLoad(killins);
	    killins = NULL;
	}

	if (!InstrLoadsRegFromMemory(iins)) {
	    continue;
	}

	// If an instr has prefixes -- then all registers become unknown.
	//                                                  -- "cp.c:1370"
	if (iins->i_prefixes) {
	    continue;
	}

	if (phase == 1) {
	    totwt += iins->i_Bblock->b_weight;
	}

	jins = DefiningInstr(iins);
	if (jins == NULL) {
	    continue;
	}
	if (jins->i_prefixes) {
	    continue;
	}

	/*
	 * we've found an instruction jins that loads from the same location
	 * as iins.  We now have to see whether it's possible to eliminate
	 * the load operation iins.
	 */
	ireg = iins->i_dest->a_base_reg;
	iindreg = iins->i_src->a_index_reg;

	if (InstrLoadsRegFromMemory(jins)) {
	    jreg = jins->i_dest->a_base_reg;
	    jindreg = jins->i_dest->a_index_reg;
	} else {
	    jreg = jins->i_src->a_base_reg;
	    jindreg = jins->i_src->a_index_reg;
	}

	def_iins = def_jins = NULL;
	bailout = false;
	ubbl = NULL;

	for (xbbl = ibbl, xins = iins->i_prev;
	    xbbl != NULL && !bailout;
	    xbbl = NODE_FROM(xbbl->b_pred), xins = xbbl->b_last) {
	    while (xins != NULL && xins->i_Bblock == xbbl) {
		if (xins->i_class == CLASS_STOS ||
                    xins->i_class == CLASS_FSTPL) {	// store to string
		    bailout = true;
		    break;
		}

		if (xins == jins) {
		    if (!def_iins && !def_jins) {
			if (jreg != ireg) {
			    /*
			     * we have to insert a reg-to-reg move: this can
			     * be done anywhere between iins and jins.  But
			     * given that we have an extended bbl, we know
			     * that the execution freq of iins is no greater
			     *  than that of jins (the converse may not
			     * be true), we insert it immediately before iins.
			     */
			    mvins = InstrCreateMovRegToReg(jreg, ireg);
				// ireg <- jreg
			    InsertInstrBeforeInstr(mvins, iins);
			}
			ubbl = iins->i_Bblock;
			FixLivenessInfo(ubbl, jins->i_Bblock, jreg);
		    } else if (phase > 1 && def_iins && !def_jins) {
			if (LOAD_LATENCY * ibbl->b_weight >=
				def_iins->i_Bblock->b_weight) {
			    REPORT(RP_LGDETAIL, "defIins: %d", def_iins->i_id);
			    mvins = InstrCreateMovRegToReg(jreg, ireg);
				// ireg <- jreg
			    InsertInstrBeforeInstr(mvins, iins);
			    ubbl = def_iins->i_Bblock;
			    FixLivenessInfo(ubbl, jins->i_Bblock, jreg);
			}
		    } else if (phase > 1 && def_jins && !def_iins) {
			if (LOAD_LATENCY * ibbl->b_weight >=
				def_jins->i_Bblock->b_weight) {
			    REPORT(RP_LGDETAIL, "defJins: %d", def_jins->i_id);
			    mvins = InstrCreateMovRegToReg(jreg, ireg);
			    // ireg <- jreg
			    InsertInstrBeforeInstr(mvins, def_jins);
			    ubbl = def_jins->i_Bblock;
			    FixLivenessInfo(ibbl, ubbl, ireg);
			    FixLivenessInfo(ubbl, jins->i_Bblock, jreg);
			}
		    }

		    if (ubbl) {
			REPORT(RP_LGLOOP, "Eliminated ins %d"
			    " [ bbl %d ] --> ins %d [bbl %d]: wt %ld",
			    iins->i_id, iins->i_Bblock->b_id,
			    jins->i_id, jins->i_Bblock->b_id,
			    (long) iins->i_Bblock->b_weight);
			killins = iins;
		    }

		    break;
		}

		if (InstrDefinesReg(xins, ireg)) {
		    def_iins = xins;
		}

		if (InstrDefinesReg(xins, jreg)) {
		    if (xins->i_prev && xins->i_prev->i_Bblock == xbbl) {
			ireg_is_live = RegBitvectorIsLiveAtInstr(
			    registerToBitRegister(ireg), xins->i_prev);
		    } else {
			ireg_is_live =
			    ((registerToBitRegister(ireg) & xbbl->b_liveIn) !=
			    0);
		    }
		    if (ireg_is_live) {
			bailout = true;
			break;
		    } else {
			def_jins = xins;
		    }
		}

		xins = xins->i_prev;
	    }				/* while */

	    if (BBLNumberOfPredecessors(xbbl) != 1
		|| NODE_FROM(xbbl->b_pred) == NULL)
		break;
	}
    }
}



/*
 *  OptLoadStoreForward() --
 *
 *  Load/store forwarding is done in two phases.  In the first phase, we
 *  process situations of the form
 *
 *  mov*  reg1 <- loc1
 *  ...
 *  mov*  reg1 <- loc1
 *
 *  In this case, if the second load instruction can be eliminated, we save
 *  an instruction without incurring any additional cost.  After this, in
 *  the second phase, we process situations of the form
 *
 *  mov*  reg1 <- loc1
 *  ...
 *  mov*  reg2 <- loc1
 *
 *  In this case, even if the second load instruction can be eliminated, we
 *  have to add a register-to-register move instruction, which reduces our
 *  gains somewhat.  The reason for splitting this into phases is that they
 *  can sometimes interfere with each other, and if they do we always prefer
 *  the first situation over the second.
 */
void OptLoadStoreForward() {
    Function *ifun;
    Block *ibbl;

    REPORT(RP_MINOR, WITHFUNC("Load/store forwarding"));

    FOREACH_FUN(ifun) {
	FOREACH_BBL_IN_FUN(ifun, ibbl) {
            FwdLoadStore(ibbl, 1);	// phase 1
            FwdLoadStore(ibbl, 2);	// phase 2
	}
    } 

    REPORT(RP_DETAIL,
	"Eliminated %d instructions [%d hot]: total weight %.0f (%.2f%%)",
	n, nhot, wt, (totwt == 0.0 ? 0.0 : wt / totwt * 100.0));
}

void OptLoadStoreForwardForFun(Function *fun) {
    Block *ibbl;

    if (!fun) {
        WARNING(WITHFUN("Function is invalid"));
        return;
    }
    
    REPORT(RP_MINOR, WITHFUN("Load/store forwarding for function %s"),
        fun->f_name);    
    FOREACH_BBL_IN_FUN(fun, ibbl) {
        FwdLoadStore(ibbl, 1);	// phase 1
        FwdLoadStore(ibbl, 2);	// phase 2
    }
    
}
