/*
 *  opt.peephole.c
 *
 *  Notes: RemoveDoubleJumps seems to break some programs. Needs to be fixed
 *  before the call is part of the standard execution sequence.
 */
#include "global.h"
#include "opt.h"
#include "sysdep.h"
#include "flowgraph.h"
#include "utilities.h"
#include "opt.unreachable.h"
#include "x86-classes.h"
#include "linkage.h"
#include "liveness.h"
#include "x86-registers.h"

/* Eliminates moves that do nothing.
 * An example of such a mov is "mov eax <- eax" */
static void EliminateStupidMoves();

/* Eliminates jumps tp jumps.
 * The first jump is rerouted to the target of the second jump */
void RemoveDoubleJumps();

/* Gets rid of things like adding 0 to a register, multiplying by 1, etc */
static void EliminateStupidMathOps();

/* Destroys pushes that are deallocated before being used */
static void EliminateUnusedPushes();

/* Converts conditional jumps and moves into conditional moves */
static void CreateConditionalMoves();

/* Returns true if the false branch block is a conditional move candidate */
static bool FalseBranchIsCmovCandidate(Block *bbl);

/*
 *  Eliminates some special cases of jumps that are stupid.  Examples
 *  include multiple chains of "test" instructions on the same condition
 */
static void EliminateStupidJumps();

/*
 *  Coalesces adds. If two adds are in the same basic block, and their result
 *  is not used between them, then we can coalesce them into one add.
 *  The same works for subtracts
 */
static void CoalesceAdds();

/*
 *  Registers the peephole optimization to be reported.
 *  This function should be called with the name of the peephole optimization
 */
static void RegisterPeephole(char *optName);

/* Returns the first occurrence of 'push %ebp" in 'block', or NULL */
Instruction *BBLFindFirstPushOfFramePointer(Block *block);

/* Returns the first occurrence of 'mov %esp -> %ebp' in 'block', or NULL */
Instruction *BBLFindFirstMovStackPointerToFramePointer(Block *block);

static int peepRound;

#define PEEPHOLE_END   "endPeephole"
#define PEEPHOLE_START "startPeephole"



/*
 *  NormalizeOperations() --
 *
 *  Normalizes some operations to make instructions more uniform
 */
void NormalizeOperations() {
    Function *fun;
    Block *first;
    Instruction *push = NULL;
    Instruction *mov = NULL;
    Instruction *iterator;
    REGISTER_BITVECTOR used;
    REGISTER_BITVECTOR def;
    int ebp, esp;

    REPORT(RP_MINOR, WITHFUNC("Standardizing function prologues"));

    /*
     * We want each function to have a standard prologue. For a binary
     * compiled with gcc, this prologue is typically a push of the frame
     * pointer followed by a "mov %ebp <- %esp" operation. Some functions
     * may have other instructions that have been scheduled before or in
     * between the prologue, however.  The following code seeks to push
     * the function prologue upwards so that it appears all by itself.
     */

    FOREACH_FUNCTION(fun) {
	first = fun->f_first;
	push = BBLFindFirstPushOfFramePointer(first);

	if (!push || (FUNCTION_FIRST_INSTR(fun) == push)
	    || (push->i_Bblock != first))
	    continue;

	ebp = FRAME_POINTER;
	esp = STACK_POINTER;

	iterator = FUNCTION_FIRST_INSTR(fun);
	while (1) {

	    if (iterator == push) {
		InstrDisconnect(iterator);
		InsertInstrBeforeInstr(push, FUNCTION_FIRST_INSTR(fun));
		break;
	    }

	    used = ComputeRegistersUsed(iterator);
	    def = ComputeRegistersDefinedLiberal(iterator);
	    if (RegisterIsLiveInBitvector(ebp, used | def)) {
		break;
	    }
	    if (RegisterIsLiveInBitvector(esp, used | def)) {
		break;
	    }
	    iterator = iterator->i_next;
	}
    }

    FOREACH_FUNCTION(fun) {
	first = fun->f_first;
	push = BBLFindFirstPushOfFramePointer(first);
	mov = BBLFindFirstMovStackPointerToFramePointer(first);

	if (first->b_first == push) {
	    if ((mov != NULL) && (first->b_first->i_next != mov)) {

		iterator = push->i_next;
		ebp = FRAME_POINTER;
		esp = STACK_POINTER;

		while (iterator && (iterator->i_Bblock == first)) {

		    if (iterator == mov) {
			InstrDisconnect(iterator);
			InsertInstrAfterInstr(mov, push);
			break;
		    }

		    used = ComputeRegistersUsed(iterator);
		    def = ComputeRegistersDefinedLiberal(iterator);
		    if (RegisterIsLiveInBitvector(ebp, used | def)) {
			break;
		    }
		    if (RegisterIsLiveInBitvector(esp, used | def)) {
			break;
		    }
		    iterator = iterator->i_next;
		}
	    }
	}
    }

    return;
}



/*
 *  PeepholeOptimize() --
 *
 *  Performs various peephole optimizations on the control flow graph.
 */
void PeepholeOptimize() {
    int maxIterations = IntegerParameter("PeepholeIterations");

    REPORT(RP_MINOR, WITHFUNC("Peephole optimizing"));
    RegisterPeephole(PEEPHOLE_START);

    for (peepRound = 0; peepRound < maxIterations; peepRound++) {

	if (PEEPHOLE_STUPID_MOVES) {
	    EliminateStupidMoves();
	}

	if (PEEPHOLE_STUPID_JUMPS) {
	    EliminateStupidJumps();
	    //   RemoveDoubleJumps();
	}

	if (PEEPHOLE_COALESCE_ADDS) {
	    CoalesceAdds();
	}

	if (PEEPHOLE_STUPID_MATH) {
	    EliminateStupidMathOps();
	}

	if (PEEPHOLE_UNUSED_PUSHES) {
	    EliminateUnusedPushes();
	}

	if (PEEPHOLE_CONDITIONAL_MOVE) {
	    CreateConditionalMoves();
	}

    }					/* end peephole iterations */

    RegisterPeephole(PEEPHOLE_END);
    RemoveUnreachableCode(startins->i_Bblock);

    return;
}



/*
 *  RemoveDoubleJumps() --
 *
 *  Consider an instruction sequence of the form jmp -> target1, where
 *  target1 has jmp -> target2. It seems like a good idea to change the
 *  first jump to jmp -> target2 and avoid the extra jmp overhead
 *  completely. What about if we have jmp -> target1, where target1 is
 *  call functionFoo. Can we change the jmp -> target1 to simply call
 *  functionFoo? Likely not, as we need to guarantee the return of control
 *  flow is the same after the function returns. Thus, we only optimize
 *  the flow for unconditional jumps and ET_TRUE edges.
 */
void RemoveDoubleJumps() {
    Instruction *iterator;
    EdgeNode *niterator;
    Block *targetBlock;

    RegisterPeephole("RemoveDoubleJumps");
    FOREACH_INSTR(iterator) {

	if (!isJumpClass(iterator)) {
	    continue;
	}
	if (!isUnconditionalInstr(iterator)) {
	    continue;
	}
	if (iterator->i_Bblock->b_first != iterator) {
	    continue;
	}
	if (NODE_TYPE(iterator->i_Bblock->b_succ) == ET_HELL) {
	    continue;
	}
	if (BBLHasHellPredecessor(iterator->i_Bblock)) {
	    continue;
	}

	targetBlock = NODE_TO(iterator->i_Bblock->b_succ);

	niterator = iterator->i_Bblock->b_pred;
	while (niterator != NULL) {
	    EdgeNode *next;
	    next = niterator->en_next;

	    if ((NODE_TYPE(niterator) != ET_TRUE) &&
		(NODE_TYPE(niterator) != ET_UNCOND)) {
		niterator = next;
		continue;
	    }

	    /*
	     * Check if by adding this edge we may create a parallel edge
	     * situation. If so, convert the block from BT_BRANCH to BT_UNCOND.
	     */

	    if ((((NODE_FROM(niterator))->b_type) == BT_BRANCH) &&
		((BBLGetFalseEdge(NODE_FROM(niterator))->e_to) == targetBlock))
	    {
		Block *bbl = NODE_FROM(niterator);
		bbl->b_type = BT_UNCOND;
		BBLAddEdge(bbl, targetBlock, ET_UNCOND)->e_edgeWeight =
		    bbl->b_weight;
		BBLRemoveEdgeTyped(bbl, iterator->i_Bblock, ET_TRUE);
		BBLRemoveEdgeTyped(bbl, (BBLGetFalseEdge(bbl)->e_to), ET_FALSE);
		niterator = next;
	    } else {
		Block *bbl = NODE_FROM(niterator);
		EDGE_WEIGHT_TYPE wt = NODE_WEIGHT(niterator);
		(BBLAddEdge(bbl, NODE_TO(iterator->i_Bblock->b_succ),
		    NODE_TYPE(niterator))->e_edgeWeight) = wt;
		BBLRemoveEdge(bbl, iterator->i_Bblock);
		iterator->i_Bblock->b_weight = iterator->i_Bblock->b_weight -wt;
		NODE_WEIGHT(iterator->i_Bblock->b_succ) =
		    iterator->i_Bblock->b_weight;
		niterator = next;
	    }
	}
    }

    return;
}



/*
 *  EliminateStupidMoves() --
 *
 *  Removes instructions like "mov %eax <- %eax"
 *  that don't actually do anything.
 */
static void EliminateStupidMoves() {
    Instruction *instr, *nextInstr;
    long long registersRemoved = 0;
    long long stackRemoved = 0;

    RegisterPeephole("EliminateStupidMoves");

    /*
     * Eliminate stupid moves between registers
     */

    FOREACH_INSTR(instr) {
	nextInstr = instr->i_next;
	if (!isMovClass(instr)) {
	    continue;
	}
	if (!instr->i_src || !instr->i_dest) {
	    continue;
	}
	if (!OperandIsGPRegister(instr->i_dest)) {
	    continue;
	}
	if (!OperandIsGPRegister(instr->i_src)) {
	    continue;
	}
	if (instr->i_dest->a_base_reg != instr->i_src->a_base_reg) {
	    continue;
	}
	registersRemoved = registersRemoved + instr->i_Bblock->b_weight;
	InstrDestroy(instr);
	instr = nextInstr;
    }

    /*
     * Eliminate stupid moves between stack locations
     */

    FOREACH_INSTR(instr) {
	nextInstr = instr->i_next;
	if (!isMovClass(instr)) {
	    continue;
	}
	if (!instr->i_src || !instr->i_dest) {
	    continue;
	}
	if (!OperandAccessesStackFromFramePointer(instr->i_dest)) {
	    continue;
	}
	if (!OperandAccessesStackFromFramePointer(instr->i_src)) {
	    continue;
	}
	if (instr->i_dest->a_displacement != instr->i_src->a_displacement) {
	    continue;
	}
	stackRemoved = stackRemoved + instr->i_Bblock->b_weight;
	InstrDestroy(instr);
	instr = nextInstr;
    }

    return;
}



/*
 *  EliminateStupidJumps() --
 *
 *  Attempts to eliminate/redirect various kinds of stupid jumps that may
 *  appear in the control flow graph. Most of the examples were added after
 *  seeing them in code.
 */
static void EliminateStupidJumps() {
    return;
}



/*
 *  CoalesceAdds() -- coalesces adds (and subtracts) to registers.
 */
static void CoalesceAdds() {
    Block *bbl;
    Instruction *instr, *iterator, *iterator2;
    Instruction *killInstr = NULL;
    PSW_BITVECTOR defined;
    bool eliminateOperation = false;

    RegisterPeephole("CoalesceAdds");

    /*
     * We look for two adds/subs inside a basic block where the result
     * is not used between them. The PSW also must not be used.
     */

    FOREACH_BBL(bbl) {
	instr = bbl->b_first;
	while (instr && (instr->i_Bblock == bbl)) {

	    if ((instr->i_class != CLASS_ADD)
		&& (instr->i_class != CLASS_SUB)) {
		instr = instr->i_next;
		continue;
	    }

	    if (!OperandIsImmediate(instr->i_src)) {
		instr = instr->i_next;
		continue;
	    }

	    if (!OperandIsGPRegister(instr->i_dest)) {
		instr = instr->i_next;
		continue;
	    }

	    if (!OPERAND_HAS_FLAG(instr->i_src, OPERAND_VMODE_FLAG)) {
		instr = instr->i_next;
		continue;
	    }

	    if (abs(instr->i_src->a_displacement) > 0x0fffffff) {
		instr = instr->i_next;
		continue;
	    }

	    /*
	     * The first add/sub has been found. Now we scan through
	     * and try to find another.
	     */

	    iterator = instr->i_next;
	    while (iterator && (iterator->i_Bblock == bbl)) {

		if (instr->i_class != iterator->i_class) {
		    iterator = iterator->i_next;
		    continue;
		}

		if (instr->i_dest->a_base_reg != iterator->i_dest->a_base_reg) {
		    iterator = iterator->i_next;
		    continue;
		}

		if (!OperandIsImmediate(iterator->i_src)) {
		    iterator = iterator->i_next;
		    continue;
		}

		if (!OperandIsGPRegister(iterator->i_dest)) {
		    iterator = iterator->i_next;
		    continue;
		}

		if (abs(iterator->i_src->a_displacement) > 0x0fffffff) {
		    iterator = iterator->i_next;
		    continue;
		}

		iterator2 = instr->i_next;
		defined = ComputePSWDefinedLiberal(instr);
		eliminateOperation = true;
		while (iterator2 != iterator->i_next) {
		    if (ComputePSWUsed(iterator2) && defined) {
			eliminateOperation = false;
		    }
		    iterator2 = iterator2->i_next;
		}

		iterator2 = instr->i_next;
		while (iterator2 != iterator) {
		    if (InstrUsesReg(iterator2, instr->i_dest->a_base_reg)) {
			eliminateOperation = false;
		    }
		    if (RegisterIsStackPointer(instr->i_dest->a_base_reg)) {

			if ((iterator2->i_dest)
			    && OPERAND_HAS_FLAG(iterator2->i_dest,
				OPERAND_REGISTER_INDIRECT)) {
			    eliminateOperation = false;
			}

			if ((iterator2->i_src)
			    && OPERAND_HAS_FLAG(iterator2->i_src,
				OPERAND_REGISTER_INDIRECT)) {
			    eliminateOperation = false;
			}

			if ((iterator2->i_src2)
			    && OPERAND_HAS_FLAG(iterator2->i_src2,
				OPERAND_REGISTER_INDIRECT)) {
			    eliminateOperation = false;
			}
		    }
		    iterator2 = iterator2->i_next;
		}

		if (eliminateOperation == true) {
		    instr->i_src->a_displacement = instr->i_src->a_displacement
			+ iterator->i_src->a_displacement;
		    killInstr = iterator;
		    break;
		}

		iterator = iterator->i_next;
	    }			/* end search for a matching instruction */

	    if (eliminateOperation == true) {
		eliminateOperation = false;
		InstrDestroy(killInstr);
	    } else {
		instr = instr->i_next;
	    }

	}			/* end iteration over instructions in bbl */
    }				/* end iteration over blocks */

    return;
}



/*
 *  RegisterPeephole(optName) --
 *
 *  Registers the peephole optimization to be reported on output.
 *  This function should be called with the name of the optimization.
 */
static void RegisterPeephole(char *optName) {
    static long long instructionsBefore, instructionsAfter;
    float change, prior = 0.0, after = 0.0;
    static long long originalInstructions;
    static char *nextName, *currentName;

    instructionsAfter = CountInstructionsExecuted();

    if (strcmp(optName, PEEPHOLE_START) == 0) {
	originalInstructions = CountInstructionsExecuted();
	nextName = currentName = NULL;
    } else {
	currentName = nextName;
	nextName = zalloc(strlen(optName) + 10);
	sprintf(nextName, "[%s:%d]", optName, peepRound);
	prior = instructionsBefore;
	after = instructionsAfter;
    }

    change = (prior == 0) ? 0.0 : ((prior - after) / prior) * 100.0;
    instructionsBefore = instructionsAfter;

    /*
     * If the optimization was the start initializer,
     * we'll display the header for peephole optimizations.
     */

    if ((currentName == NULL) && (nextName == NULL)) {
	REPORT(RP_DETAIL, "  Optimization               | Insns Prior |"
	    " Insns After | Change");
	REPORT(RP_SMLOOP, "=========================================="
	    "======================");
	return;
    } else if (currentName != NULL) {
	REPORT(RP_SMLOOP, "%-26s |%12.0f |%12.0f | %.3f %%",
	    currentName, prior, after, change);
    }

    if (strcmp(optName, PEEPHOLE_END) == 0) {
	change = (originalInstructions == 0) ? 0.0 :
	    (originalInstructions - instructionsAfter) * 100.0 /
	    originalInstructions;
	REPORT(RP_DETAIL, "  %-26s |%12.0f |%12.0f | %.3f %%",
	    "Total", (float) originalInstructions, (float) instructionsAfter,
	    change);
    }

    return;
}



/*
 *  EliminateStupidMathOps() --
 *
 *  Gets rid of extraneous operations that have no effect.
 */
static void EliminateStupidMathOps() {
    Instruction *instr, *instrNext;

    RegisterPeephole("EliminateStupidMath");

    instr = i_list;
    while (instr != NULL) {
	instrNext = instr->i_next;

	switch (instr->i_class) {

	case CLASS_ADD:
	case CLASS_SUB:
	    if (OperandIsImmediate(instr->i_src)
		&& (instr->i_src->a_displacement == 0)) {
		if (!PSWBitvectorIsLiveAfterInstr(ComputePSWDefinedLiberal
			(instr), instr)) {
		    InstrDestroy(instr);
		}
	    }
	    break;

	}				/* end switch over instruction class */

	instr = instrNext;
    }					/* end iteration over instructions */

    return;
}



/*
 *  EliminateUnusedPushes() --
 *
 *  Removes PUSH instructions
 *  if the stack is immediately deallocated after them.
 */
static void EliminateUnusedPushes() {
    Instruction *instr, *instr2, *instrNext;

    RegisterPeephole("EliminateUnusedPushes");

    instr = i_list;
    while (instr != NULL) {
	instrNext = instr->i_next;
	if ((instr->i_class == CLASS_PUSH)
	    && (instr->i_Bblock->b_last != instr)) {
	    instr2 = instr->i_next;
	    if ((instr2->i_class == CLASS_ADD)
		&& OperandIsGPRegister(instr2->i_dest)
		&& (RegisterIsStackPointer(instr2->i_dest->a_base_reg))
		&& (OperandIsImmediate(instr2->i_src))
		&& (instr2->i_src->a_displacement > InstrPushBytes(instr))) {
		instr2->i_src->a_displacement -= InstrPushBytes(instr);
		InstrDestroy(instr);
	    }
	}
	instr = instrNext;
    }

    return;
}



/*
 *  CreateConditionalMoves() --
 *
 *  Creates conditional moves by removing conditional jumps and control
 *  transfer with the cmov instructions.
 */
static void CreateConditionalMoves() {
    Block *bbl, *trueNode, *falseNode;
    Instruction *cmov, *jmp, *mov;

    RegisterPeephole("CreateConditionalMoves");

    FOREACH_BBL(bbl) {

	cmov = NULL;
	if (bbl->b_type == BT_BRANCH) {

	    trueNode = BBLGetTrueEdge(bbl)->e_to;
	    falseNode = BBLGetFalseEdge(bbl)->e_to;

	    if (FalseBranchIsCmovCandidate(falseNode)
		&& BBLEdgeExists(falseNode, trueNode)) {

		mov = falseNode->b_last;
		jmp = bbl->b_last;

		if (!OperandIsGPRegister(mov->i_src)) {
		    cmov = InstrCreateCmovMemToReg();
		} else {
		    cmov = InstrCreateCmovRegToReg();
		}

		switch (jmp->i_class) {
		case CLASS_JA:
		    cmov->i_class = CLASS_CMOVBE;
		    break;
		case CLASS_JAE:
		    cmov->i_class = CLASS_CMOVB;
		    break;
		case CLASS_JB:
		    cmov->i_class = CLASS_CMOVAE;
		    break;
		case CLASS_JBE:
		    cmov->i_class = CLASS_CMOVA;
		    break;
		case CLASS_JE:
		    cmov->i_class = CLASS_CMOVNE;
		    break;
		case CLASS_JG:
		    cmov->i_class = CLASS_CMOVLE;
		    break;
		case CLASS_JGE:
		    cmov->i_class = CLASS_CMOVL;
		    break;
		case CLASS_JL:
		    cmov->i_class = CLASS_CMOVGE;
		    break;
		case CLASS_JLE:
		    cmov->i_class = CLASS_CMOVG;
		    break;
		case CLASS_JNE:
		    cmov->i_class = CLASS_CMOVE;
		    break;
		case CLASS_JNO:
		    cmov->i_class = CLASS_CMOVO;
		    break;
		case CLASS_JNP:
		    cmov->i_class = CLASS_CMOVP;
		    break;
		case CLASS_JNS:
		    cmov->i_class = CLASS_CMOVS;
		    break;
		case CLASS_JO:
		    cmov->i_class = CLASS_CMOVNO;
		    break;
		case CLASS_JP:
		    cmov->i_class = CLASS_CMOVNP;
		    break;
		case CLASS_JS:
		    cmov->i_class = CLASS_CMOVNS;
		    break;
		default:
		    MALFUNCTION("Unexpected class %d encountered",jmp->i_class);
		    break;
		}

		cmov->i_dest->a_base_reg = mov->i_dest->a_base_reg;

		if (OperandIsGPRegister(mov->i_src)) {
		    cmov->i_src->a_base_reg = mov->i_src->a_base_reg;
		} else {
		    cmov->i_src->a_base_reg = mov->i_src->a_base_reg;
		    cmov->i_src->a_index_reg = mov->i_src->a_index_reg;
		    cmov->i_src->a_displacement = mov->i_src->a_displacement;
		    cmov->i_src->a_scale = mov->i_src->a_scale;
		    cmov->i_src->a_base_reg = mov->i_src->a_base_reg;
		}
		UpdateInstr(cmov);

		BBLRemoveEdges(bbl);
		BBLRemoveEdges(falseNode);
		bbl->b_type = BT_NORMAL;
		BBLAddEdge(bbl, trueNode, ET_NORMAL);
		InstrDestroy(falseNode->b_first);
		InsertInstrBeforeInstr(cmov, bbl->b_last);
		InstrDestroy(bbl->b_last);
	    }
	}
    }

    return;
}



/*
 *  FalseBranchIsCmovCandidate(bbl) --
 *
 *  Returns true if the false branch is a candidate for a conditional move.
 */
static bool FalseBranchIsCmovCandidate(Block *bbl) {
    Instruction *instr = bbl->b_first;

    if (bbl->b_first != bbl->b_last) {
	return false;
    }

    if (!isMovClass(instr)) {
	return false;
    }

    if (!OperandIsGPRegister(instr->i_dest)) {
	return false;
    }

    if (!OperandIsGPRegister(instr->i_src) &&
	!OperandAccessesStackFromFramePointer(instr->i_src))
	return false;

    if (BBLNumberOfPredecessors(bbl) > 1) {
	return false;
    }

    return true;
}



/*
 *  ExtendSecondJumps() --
 *
 *  For jumps that lie next to each other, this routine will make the second
 *  jump of the two take up more space (5-6 bytes).
 */
void ExtendSecondJumps() {
    Instruction *instr;

    FOREACH_INSTR(instr) {
	if (isPCChangingClass(instr)) {
	    if (instr->i_next && isPCChangingClass(instr->i_next)) {
		if (instr->i_next->i_dest) {
		    OperandUseLargestBytemode(instr->i_next->i_dest);
		}
	    }
	}
    }

    return;
}



/*
 *  BBLFindFirstPushOfFramePointer(block) --
 *
 *  Returns the first occurrence of 'push %ebp' in the basic block
 *  if one exists.
 */
Instruction *BBLFindFirstPushOfFramePointer(Block *block) {
    Instruction *iterator;

    FOREACH_INSTR_IN_BBL(block, iterator) {
	if (iterator->i_class != CLASS_PUSH) {
	    continue;
	}
	if (iterator->i_src == NULL) {
	    continue;
	}
	if (!OperandIsFramePointer(iterator->i_src)) {
	    continue;
	}
	return iterator;
    }

    return NULL;
}



/*
 *  BBLFindFirstMovStackPointerToFramePointer(block) --
 *
 *  Returns the first occurrence of 'mov %esp -> %ebp' in the bbl
 *  if one exists.
 */
Instruction *BBLFindFirstMovStackPointerToFramePointer(Block *block) {
    Instruction *iterator;

    FOREACH_INSTR_IN_BBL(block, iterator) {
	if (iterator->i_class != CLASS_MOV) {
	    continue;
	}
	if (iterator->i_src == NULL) {
	    continue;
	}
	if (iterator->i_dest == NULL) {
	    continue;
	}
	if (!OperandIsStackPointer(iterator->i_src)) {
	    continue;
	}
	if (!OperandIsFramePointer(iterator->i_dest)) {
	    continue;
	}
	return iterator;
    }

    return NULL;
}
