/*
 *  opt.schedule.c
 *
 *  Instruction scheduler
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "global.h"
#include "sysdep.h"
#include "opt.h"
#include "d_graph.h"
#include "utilities.h"
#include "upm.h"

static bool ScheduleBB(Block *bb);
static Digraph *CreateDependenceDAG(int num_instrs, Block *bb,
    SchedPoints *points);
static Digraph *CreatePairingGraph(int num_instrs, Block *bb);
static void CalcDAGHeight(Digraph *graph, Digraph *pairing_graph,
    DigraphVertex *vertex);
static bool ScheduleInstruction(DigraphVertex *node, int current_cycle,
    Digraph *dep_dag, SchedNode **root_list, SchedNode **scheduled_list);
static DigraphVertex *GetBestInstruction(SchedNode *candidates,
    Digraph *dep_graph, Digraph *pairing_graph);
static UINT32 SchCalculatePoints(DigraphVertex *node, Digraph *dep_graph,
    Digraph *pairing_graph);
static void OutputInstructions(SchedNode *scheduled_list, Block *bb);

/*
 *  TODO: Should be moved to architecture dependent with SchCalculatePoints
 */
#define PAIRING_WEIGHT 1
#define SUCCESSOR_WEIGHT 1
#define HEIGHT_WEIGHT 1
#define UOPS_WEIGHT 100



/*
 *  SchCalculatePoints(node,dep_graph,pairing_graph) --
 *
 *  This function is used to make the selection of which vertex we want
 *  to select next for scheduling.  We'll associate a point value with
 *  each vertex depending on how badly we want to schedule this vertex.
 *  When given a choice the algorithm should pick the vertex with the highest
 *  point value.  This function returns the point value for one instruction,
 *  represented by 'node'.
 *
 *  TODO: Move to architecture dependent when done.
 */
UINT32 SchCalculatePoints(DigraphVertex *node, Digraph *dep_graph,
    Digraph *pairing_graph) {
    UINT32 points;
    DigraphEdge *succ;

    points = PAIRING_WEIGHT * SCH_POINTS_PAIRINGS(node);
    points += HEIGHT_WEIGHT * SCH_POINTS_HEIGHT(node);
    if (DGRAPH_OUT_DEGREE(pairing_graph, node) != 0) {
	points += UOPS_WEIGHT;
    }

    succ = dep_graph->d_successors[node->dv_id];
    for (; succ != NULL; succ = succ->di_next_successor) {
	points += SUCCESSOR_WEIGHT * succ->di_weight;
    }

    return points;
}



/*
 *  ScheduleInstructions() --
 *
 *  This function runs the basic block scheduling algorithm on each basic
 *  block in the program.
 */
void ScheduleInstructions() {
    Function *func;
    Block *bb;
    bool result;

    /*
     * We skip the _start function because of special case issues that
     * would be hard to fix and easy to ignore.
     */
    for (func = f_list; func != NULL; func = func->f_next) {
	// TODO: Bad hack here.  Fix
	if (strcmp(func->f_name, "_start") != 0) {
	    for (bb = func->f_first;
		    bb != func->f_last && bb != NULL;
		    bb = bb->b_next) {
		result = ScheduleBB(bb);
		assert(result);
	    }
	    result = ScheduleBB(bb);
	    assert(result);
	}
    }
}

/*
 *  This block should free up all memory used by ScheduleBB in case
 *  of error, without risk of seg faulting.
 */
#define CLEANUP_MEMORY(val) \
    { SchDestroyInstrList(&root_list); \
    SchDestroyInstrList(&run_candidate_list); \
    SchDestroyInstrList(&scheduled_list); \
    DestroyDGraph(dep_dag); \
    DestroyDGraph(pairing_graph); \
    if (point_list != NULL) free(point_list); \
    assert(val == true); \
    return val; }



/*
 *  ScheduleBB(bb) --
 *
 *  Reschedule the instructions a single basic block.
 *
 *  The algorithm works by first creating a dependency DAG for the basic block
 *  as well as a pairing graph.  An edge exists in the dependency DAG from
 *  vertex A to vertex B if instruction A must be run before instruction B.
 *  An edge exists in the pairing graph from vertex A to vertex B if
 *  instructions A and B can be paired up so they run in parallel.
 *
 *  Every vertex in the dependency DAG without any successors is available to
 *  be scheduled as the next instruction.  We'll use root_list to hold the
 *  instruction we may be able to schedule.
 *
 *  After we 'pick' an instruction to schedule we'll see if we can also pick an
 *  instruction that pairs with the first.  We 'pick' and instruction base on a
 *  series of heuristics.
 */
static bool ScheduleBB(Block *bb) {
    Digraph *dep_dag = NULL, *pairing_graph = NULL;
    Instruction *instr = NULL, *first_instr = NULL, *new_first_instr = NULL;
    DigraphVertex *node = NULL;
    SchedPoints *point_list = NULL;

    /*
     * The root_list will contain all vertices which are legal to run at
     * some point.  The run_candidate_list will be built from the root_list
     * and will contain instructions that can be run without causing a
     * latency delay.  The scheduled_list will be an ordered list of the
     * instructions as we've decided to run them.
     */
    SchedNode *root_list = SchInitInstrList();
    SchedNode *run_candidate_list = SchInitInstrList();
    SchedNode *scheduled_list = SchInitInstrList();
    SchedNode *runner = NULL;

    int num_instrs = 0;
    int num_instrs_in_list;
    int current_cycle = 0;
    int instrs_moved = 0;
    int num_pairable = NumInstrsPairable();
    int i;
    bool runnable, success;
    DigraphEdge *pred;


    if (bb == NULL) {
	return true;
    }
    /*
     * We need to know the number of instructions in this block so
     * we can correctly allocate enough space.
     */
    for (instr = bb->b_first; instr != bb->b_last; instr = instr->i_next) {
	num_instrs++;
    }
    num_instrs++;

    /*
     * Allocate and setup the dependence DAG and the Pairing Graph .
     * Also allocate the point list, which contains fields of data needed
     * by each vertex.
     */
    point_list = alloc(sizeof(SchedPoints) * num_instrs);
    if (point_list == NULL) {
	CLEANUP_MEMORY(false);
    }
    dep_dag = CreateDependenceDAG(num_instrs, bb, point_list);
    if (dep_dag == NULL) {
	CLEANUP_MEMORY(false);
    }
    pairing_graph = CreatePairingGraph(num_instrs, bb);
    if (pairing_graph == NULL) {
	CLEANUP_MEMORY(false);
    }

    /*
     * We'll look through each vertex, if it's on the top of our dependence DAG
     * we need to add it to our list of root vertices, which we'll need later.
     * We'll also use the vertex for a height calculation we'll need later.
     */
    for (node = dep_dag->d_vertices; node != NULL; node = node->dv_next) {
	if (!(dep_dag->d_predecessors[node->dv_id] != NULL)) {
	    root_list = SchAddInstrList(root_list, node);
	    if (root_list == NULL) {
		CLEANUP_MEMORY(false);
	    }
	    CalcDAGHeight(dep_dag, pairing_graph, node);
	}
    }

    /*
     * The core part of this scheduler algorithm.
     */
    while (instrs_moved < num_instrs) {
	/*
	 * Find all instructions that can be run without stalling and add them
	 * to the run_candidate_list
	 */
	num_instrs_in_list = 0;
	for (runner = root_list; runner != NULL; runner = runner->sn_next) {
	    runnable = true;
	    pred = dep_dag->d_predecessors[runner->sn_vertex->dv_id];
	    for (; pred != NULL; pred = pred->di_next_predecessor) {
		if (pred->di_weight + SCH_POINTS_CYCLE(pred->di_source)
			> current_cycle) {
		    runnable = false;
		    break;
		}
	    }
	    if (runnable) {
		run_candidate_list =
		    SchAddInstrList(run_candidate_list, runner->sn_vertex);
		if (run_candidate_list == NULL) {
		    CLEANUP_MEMORY(false);
		}
		num_instrs_in_list++;
	    }
	}

	/*
	 * We'll handle the trivial cases where we have one or zero instructions
	 * in the run_candidate_list here.
	 */
	if (num_instrs_in_list == 0) {
	    current_cycle++;
	    continue;
	}
	if (num_instrs_in_list == 1) {
	    node = run_candidate_list->sn_vertex;
	    success = ScheduleInstruction(node, current_cycle++, dep_dag,
		&root_list, &scheduled_list);
	    if (!success) {
		CLEANUP_MEMORY(false);
	    }
	    SchDestroyInstrList(&run_candidate_list);
	    instrs_moved++;
	    continue;
	}
	/*
	 * Now for the big part.  We've got several instructions to be scheduled
	 * and we need to decide which one we want.
	 */

	node = GetBestInstruction(run_candidate_list, dep_dag, pairing_graph);
	success = ScheduleInstruction(node, current_cycle, dep_dag, &root_list,
	    &scheduled_list);
	if (!success) {
	    CLEANUP_MEMORY(false);
	}
	instrs_moved++;
	run_candidate_list = SchRemoveInstrList(run_candidate_list, node);

	/*
	 * It's possible that we may pair up with more instructions in the
	 * candidate_list and be able to run all instructions in the same
	 * cycle.  So, we'll remove every instruction we can't pair with from
	 * the candidate_list, and then select the best instructions left.
	 */
	runner = run_candidate_list;
	while (runner != NULL) {
	    if (!DGRAPH_HAS_EDGE(pairing_graph, node, runner->sn_vertex)) {
		run_candidate_list =
		    SchRemoveInstrList(run_candidate_list, runner->sn_vertex);
		runner = run_candidate_list;
	    } else {
		runner = runner->sn_next;
	    }
	}

	/*
	 * Everything left pairs with the first instruction, so we'll pick
	 * the best ones left.  We've already selected one instruction to run,
	 * so we search for num_pairable - 1 instructions to pair with it.
	 */
	for (i = 0; i < num_pairable - 1; i++) {
	    node =
		GetBestInstruction(run_candidate_list, dep_dag, pairing_graph);
	    if (node != NULL) {
		if (!ScheduleInstruction(node, current_cycle, dep_dag,
			&root_list, &scheduled_list)) {
		    CLEANUP_MEMORY(false);
		}
		run_candidate_list =
		    SchRemoveInstrList(run_candidate_list, node);
		instrs_moved++;
	    }
	}
	current_cycle++;
	SchDestroyInstrList(&run_candidate_list);
    }


    /*
     * Unfortunatly we added our new instructions the scheduled_list by
     * putting the last instructions first in the list.  We need to reverse the
     * order of the list before trying to schedule it.
     */

    scheduled_list = SchReverseInstrList(scheduled_list);

    /*
     * The UPM tables need to be updated.  Since they are supposed to point to
     * the first instruction in a basic block, and we may have moved that first
     * instruction away.
     */
    first_instr = bb->b_first;
    new_first_instr = SCH_POINTS_INSTR(scheduled_list->sn_vertex);
    if (first_instr != new_first_instr) {
	UPMChangeAllSourcesFromInstrToInstr(first_instr, new_first_instr);
    }

    /*
     * Now all that's left is to do the update the instructions
     */

    OutputInstructions(scheduled_list, bb);

    CLEANUP_MEMORY(true);
}



/*
 *  CreateDependenceDAG(num_instrs,bb,points) --
 *
 *  Instruction A points to Instruction B in the DAG if A has to execute
 *  before B for program correctness.
 */
static Digraph *CreateDependenceDAG(int num_instrs, Block *bb,
    SchedPoints *points) {
    Digraph *dep_dag;
    Instruction *instr, *dep_instr;
    UINT32 instr_count, dep_instr_count;
    DigraphVertex *vertex;
    int i = 0;

    dep_dag = CreateDGraph(num_instrs);
    if (dep_dag == NULL) {
	/* Not enough memory. */
	return NULL;
    }

    /*
     * Assign each vertex an element in the points list.
     * Each vertex will use that element to store data about itself.
     */
    memset(points, 0, sizeof(SchedPoints) * num_instrs);
    for (vertex = dep_dag->d_vertices; vertex!=NULL; vertex = vertex->dv_next) {
	vertex->dv_data = &points[i++];
    }

    /*
     * We'll build the dependence graph by looking backward from each
     * instruction until we find all instructions it depends on.
     * Time wise this should be one of the kickers for this algorithm,
     * since it runs in O(n^2).
     */
    for (instr = bb->b_first, instr_count = 0;
	instr_count < num_instrs; instr = instr->i_next, instr_count++) {
	for (dep_instr = instr->i_next, dep_instr_count = instr_count + 1;
	    dep_instr_count < num_instrs;
	    dep_instr = dep_instr->i_next, dep_instr_count++) {
	    if (InstrIsDependent(instr, dep_instr)) {
		CreateEdgeDGraph(dep_dag,
		    &(dep_dag->d_vertices[instr_count]),
		    &(dep_dag->d_vertices[dep_instr_count]),
		    InstrDelayBetween(instr, dep_instr));
	    }
	}
	SCH_POINTS_INSTR((&(dep_dag->d_vertices[instr_count]))) = instr;
    }
    return dep_dag;
}



/*
 *  CreatePairingGraph(num_instrs,bb) --
 *
 *  The pairing graph is a directed graph where instructions are represented by
 *  vertices.  There exists an edge between instruction 'a' and instruction 'b'
 *  if after running 'a' we can pair up with instruction 'b' and run both at the
 *  same time.
 */
static Digraph *CreatePairingGraph(int num_instrs, Block *bb) {
    Digraph *pair_graph;
    Instruction *instr_a, *instr_b;
    int a, b;

    pair_graph = CreateDGraph(num_instrs);
    if (pair_graph == NULL) {
	/* Not enough memory, return an error. */
	return NULL;
    }

    for (a = 0, instr_a = bb->b_first; a < num_instrs;
	instr_a = instr_a->i_next, a++) {
	for (b = 0, instr_b = bb->b_first; b < num_instrs;
	    instr_b = instr_b->i_next, b++) {
	    if ((a != b) && (InstrsPair(instr_a, instr_b))) {
		CreateEdgeDGraph(pair_graph, (&(pair_graph->d_vertices[a])),
		    (&(pair_graph->d_vertices[b])), 0);
	    }
	}
    }

    return pair_graph;
}



/*
 *  max(a,b) -- return the larger of two UINT32 values.
 */
static UINT32 max(UINT32 a, UINT32 b) {
    if (a > b) {
	return a;
    }
    return b;
}



/*
 *  CalcDAGHeight(graph,pairing_graph,vertex) --
 *
 *  As part of this algorithm we need to know the height of the dependence
 *  DAG at any vertex and the number of pairable instructions that lie
 *  below any vertex.  This is you standard recursive algorithm for calculating
 *  such things.
 */
static void CalcDAGHeight(Digraph *graph, Digraph *pairing_graph,
    DigraphVertex *vertex) {
    UINT32 height = 0, pairing = 0;
    DigraphEdge *succ;

    if (SCH_POINTS_HEIGHT(vertex) != 0) {
	return;
    }

    succ = graph->d_successors[vertex->dv_id];
    for (; succ != NULL; succ = succ->di_next_successor) {
	CalcDAGHeight(graph, pairing_graph, succ->di_target);
	pairing += SCH_POINTS_PAIRINGS(succ->di_target);
	height = max(height, SCH_POINTS_HEIGHT(succ->di_target));
    }

    SCH_POINTS_HEIGHT(vertex) = height + 1;
    SCH_POINTS_PAIRINGS(vertex) = pairing + DGRAPH_IN_DEGREE(
	pairing_graph, (&(pairing_graph->d_vertices[vertex->dv_id])));
}



/*
 *  ScheduleInstruction(node,current_cycle,dep_dag,root_list,scheduled_list) --
 *
 *  A helper function for ScheduleBB.  This function takes almost every local
 *  variable from ScheduleBB and updates it all to account for the
 *  instruction represented by 'node' to be scheduled at cycle 'current_cycle'.
 *
 *  Returns false if we were unable to schedule (out of memory) and true if
 *  we were successful.
 */
static bool ScheduleInstruction(DigraphVertex *node, int current_cycle,
    Digraph *dep_dag, SchedNode **root_list, SchedNode **scheduled_list) {
    bool is_free;
    DigraphEdge *pred, *succ;

    /* First we'll update the fields of node to account for being scheduled */
    assert(SCH_POINTS_SCHEDULED(node) == false);

    SCH_POINTS_SCHEDULED(node) = true;
    SCH_POINTS_CYCLE(node) = current_cycle++;

    /*
     * We need to remove node from the root_list and add it to the
     * scheduled_list.
     */
    *root_list = SchRemoveInstrList(*root_list, node);
    *scheduled_list = SchAddInstrList(*scheduled_list, node);
    if (*scheduled_list == NULL) {
	return false;
    }

    /*
     * Now that we've scheduled this node some it may be legal to schedule
     * some if it's successors.  So, for each successor if all predecessors
     * have been scheduled then we can add the successor to our root_list,
     * which indicates it can now be scheduled.
     */
    succ = dep_dag->d_successors[node->dv_id];
    for (; succ != NULL; succ = succ->di_next_successor) {
	pred = dep_dag->d_predecessors[succ->di_target->dv_id];
	is_free = true;
	for (; pred != NULL; pred = pred->di_next_predecessor) {
	    if (!SCH_POINTS_SCHEDULED(pred->di_source)) {
		is_free = false;
		break;
	    }
	}
	if (is_free) {
	    *root_list = SchAddInstrList(*root_list, succ->di_target);
	    if (*root_list == NULL) {
		return false;
	    }
	}
    }

    return true;
}



/*
 *  GetBestInstruction(candidates,dep_graph,pairing_graph) --
 *
 *  This function will select the 'best' instruction for us to schedule next,
 *  given a list of possible instructions.  We'll iterate over each instruction
 *  in 'candidates', if the instruction doesn't have a point value associated
 *  with it we'll compute one for it.  Then, we'll select the instruction with
 *  the highest point value as the 'best'.
 */
static DigraphVertex *GetBestInstruction(SchedNode *candidates,
    Digraph *dep_graph, Digraph *pairing_graph) {
    SchedNode *runner;
    DigraphVertex *vertex, *max_vertex;
    if (candidates == NULL) {
	return NULL;
    }
    max_vertex = candidates->sn_vertex;
    for (runner = candidates; runner != NULL; runner = runner->sn_next) {
	vertex = runner->sn_vertex;
	if (SCH_POINTS_POINTS(vertex) == 0) {
	    SCH_POINTS_POINTS(vertex) =
		SchCalculatePoints(vertex, dep_graph, pairing_graph);
	}
	if (SCH_POINTS_POINTS(vertex) > SCH_POINTS_POINTS(max_vertex)) {
	    max_vertex = vertex;
	}
    }
    return max_vertex;
}



/*
 *  OutputInstructions(scheduled_list,bb) --
 *
 *  After everything is said and done we come up with a basic block, 'bb'
 *  and a rescheduled list of instructions, 'scheduled_list'.  This function
 *  reorders the instructions in 'bb' to match the order represented by
 *  'scheduled_list'.
 */
void OutputInstructions(SchedNode *scheduled_list, Block *bb) {
    SchedNode *runner;
    Instruction *instr = NULL, *prev, *first, *last;

    prev = bb->b_first->i_prev;

    first = SCH_POINTS_INSTR(scheduled_list->sn_vertex);
    last = bb->b_last->i_next;

    runner = scheduled_list;
    for (; runner != NULL; runner = runner->sn_next) {
	instr = SCH_POINTS_INSTR(runner->sn_vertex);
	instr->i_prev = prev;
	if (prev != NULL) {
	    prev->i_next = instr;
	}
	prev = instr;
    }
    instr->i_next = last;
    if (last != NULL) {
	last->i_prev = instr;
    }

    bb->b_first = first;
    bb->b_last = instr;

}
