/*
 *  opt.specialize.c
 */
#include "global.h"
#include "opt.h"
#include "sysdep.h"
#include "utilities.h"
#include "potential.h"
#include "opt.unreachable.h"
#include "flowgraph.h"
#include "linkage.h"
#include "clone.h"

#define COST_OF_SPECIALIZATION_TEST 3

/*
 *  Specialization
 */
typedef struct Specialization {
    struct Specialization *sz_next;
    struct Specialization *sz_prev;
    Instruction *sz_instr;
    Operand *sz_op;
    EDGE_WEIGHT_TYPE sz_potential;
    int sz_id;
    int sz_value;
} Specialization;


static Specialization *specializationList;

static int specializationCounter;
static EDGE_WEIGHT_TYPE hotInstrWeight;

static void AddSpecializationPoint(Instruction *instr, Operand *op,
    EDGE_WEIGHT_TYPE pot);
static void RemoveSpecialization(Specialization *aSpec);

static bool ShouldSpecializeOperand(Instruction *instr, Operand *op,
    ProfStats *vpt, double threshold);
static void Specialize(Specialization *spec);

static void MarkSpecializationSet(Instruction *instr);



/*
 *  ValueSpecialize(threshold) --
 *
 *  Locates points at which value profiling is profitable, and adds them to
 *  the candidate pool for specialization.
 */
void ValueSpecialize(double threshold) {
    ProfStats *vpt;
    Instruction *instr;
    Block *biterator;
    Operand *op = NULL;

    /* Compute the weight of a 'hot' instruction. */
    hotInstrWeight = 0;
    FOREACH_BBL(biterator) {
	if (BBL_HAS_FLAG(biterator, BBL_HOT_FLAG)) {
	    if (biterator->b_weight < hotInstrWeight) {
		hotInstrWeight = biterator->b_weight;
	    }
	    if (hotInstrWeight == 0) {
		hotInstrWeight = biterator->b_weight;
	    }
	}
    }

    /*
     * Check out each point that was profiled, and see what the expected
     * savings of that point are. If the savings are high enough, insert
     * the point into the list of points to be specialized.
     */

    FOREACH_VPT(vpt) {
	EDGE_WEIGHT_TYPE potential = 0;

	instr = IDToInstr(vpt->ps_instr_id);

	if (vpt->ps_op == SRC1) {
	    op = instr->i_src;
	} else if (vpt->ps_op == SRC2) {
	    op = instr->i_src2;
	} else if (vpt->ps_op == DEST) {
	    op = instr->i_dest;
	}

	if (ShouldSpecializeOperand(instr, op, vpt, threshold)) {
	    AddSpecializationPoint(instr, op, potential);
	}
    }

    while (specializationList != NULL) {
	Specialize(specializationList);
    }

    return;
}



/*
 *  Specialize(spec) --
 *
 *  Carries out specialization for a program point/register pair. Control flow
 *  graph manipulations, such as the cloning, is done in here, and edges are
 *  fixed and re-weighted. The specialization point is then removed from the
 *  list of points to specialize.
 */
static void Specialize(Specialization *spec) {
    Block *clonedBlocks;
    Instruction *cmp;
    Instruction *jmp;
    Instruction *nop;
    Instruction *iterator;
    Block *firstCloned;
    int i, count;
    Instruction *instr = spec->sz_instr;
    Operand *op = spec->sz_op;

    cmp = jmp = NULL;

    if (OperandIsGPRegister(op)) {
	ComputeRegisterPotential(op->a_base_reg, instr);
	cmp = InstrCreateCmpRegToValue(op->a_base_reg, spec->sz_value);
    } else if (OperandAccessesStackFromFramePointer(op)) {
	ComputeStackPotential(op->a_displacement, instr);
	cmp = InstrCreateCmpStackToValue(op->a_displacement, spec->sz_value);
    } else {
	MALFUNCTION("unexpected error.");
    }

    CASSERT(cmp != NULL, "CMP instruction was NULL.");

    MarkSpecializationSet(instr);
    clonedBlocks = FunCloneMarkedBlocks(INSTR_FUNCTION(instr));

    /*
     * Although the first block into which the cmp/test is being inserted
     * is marked to be cloned, we really only want to clone the portion
     * below the specialization point. The following code removes the
     * cloned instructions above the specialization point, by counting how
     * how many there are, then removing them from the cloned block.
     */

    count = 0;
    iterator = instr->i_Bblock->b_first;
    while (iterator != instr) {
	count++;
	iterator = iterator->i_next;
    }

    firstCloned = BBLOriginalToCloneFromList(instr->i_Bblock, clonedBlocks);
    iterator = firstCloned->b_first;
    for (i = 0; i < count; i++) {
	Instruction *next = iterator->i_next;
	InstrDestroy(iterator);
	iterator = next;
    }

    /*
     * Create the test and jump and insert them before the instruction.
     */

    nop = InstrCreateNOP();
    jmp = InstrCreateJe();

    InsertInstrBeforeInstr(nop, instr);
    InsertInstrBeforeInstr(cmp, instr);
    InsertInstrBeforeInstr(jmp, instr);

    /*
     * Divide the basic block so that 'instr' is the first block in the
     * newly formed block. Then remove edges going out of the block with
     * the 'cmp' and 'test', and create the true/false edges.
     */

    BBLDivideBlock(instr->i_Bblock, instr);
    BBLRemoveEdge(jmp->i_Bblock, instr->i_Bblock);
    BBLAddEdge(jmp->i_Bblock, instr->i_Bblock, ET_TRUE);
    BBLAddEdge(jmp->i_Bblock, firstCloned, ET_FALSE);

    REPORT(RP_LGLOOP, "Specialize: %d   0x%8.8x %s  %.0f",
	spec->sz_id, instr->i_address,
	PrintInstruction(instr), (float) spec->sz_potential);

    RemoveSpecialization(spec);
}



/*
 *  AddSpecializationPoint(instr,op,pot) --
 */
static void AddSpecializationPoint(Instruction *instr, Operand *op,
    EDGE_WEIGHT_TYPE pot) {
    Specialization *iterator = specializationList;
    Specialization *aSpec = zalloc(sizeof(Specialization));

    aSpec->sz_id = specializationCounter++;
    aSpec->sz_potential = pot;
    aSpec->sz_op = op;
    aSpec->sz_instr = instr;

    if (iterator == NULL) {
	specializationList = aSpec;
	return;
    }

    while (iterator->sz_potential >= pot) {
	if (iterator->sz_next == NULL) {
	    iterator->sz_next = aSpec;
	    aSpec->sz_prev = iterator;
	    aSpec->sz_next = NULL;
	    return;
	} else {
	    iterator = iterator->sz_next;
	}
    }

    if (iterator == specializationList) {
	aSpec->sz_next = specializationList;
	specializationList->sz_prev = aSpec;
	specializationList = aSpec;
    } else {
	aSpec->sz_next = iterator;
	aSpec->sz_prev = iterator->sz_prev;
	iterator->sz_prev = aSpec;
	aSpec->sz_prev->sz_next = aSpec;
    }

    return;
}



/*
 *  MarkSpecializationSet(instr) --
 */
static void MarkSpecializationSet(Instruction *instr) {

    Instruction *iterator;
    Block *biterator;
    EdgeNode *successors;
    bool change = true;

    FOREACH_BBL_IN_FUN(INSTR_FUNCTION(instr), biterator) {
	BBL_REMOVE_FLAG(biterator, BBL_MARKED_FLAG);
	FOREACH_INSTR_IN_BBL(biterator, iterator) {
	    if (INSTR_HAS_FLAG(iterator, INSTR_ELIMINATABLE_FLAG)) {
		BBL_SET_FLAG(biterator, BBL_MARKED_FLAG);
	    }
	}
    }

    BBL_SET_FLAG(instr->i_Bblock, BBL_MARKED_FLAG);

    while (change) {
	change = false;
	FOREACH_BBL_IN_FUN(INSTR_FUNCTION(instr), biterator) {

	    if (BBL_HAS_FLAG(biterator, BBL_MARKED_FLAG)) {
		continue;
	    }

	    FOREACH_SUCC_IN_BBL(biterator, successors) {
		if (BBL_HAS_FLAG(NODE_TO(successors), BBL_MARKED_FLAG)) {
		    change = true;
		    BBL_SET_FLAG(biterator, BBL_MARKED_FLAG);
		}
	    }
	}
    }

    return;
}



/*
 *  ShouldSpecializeOperand(instr,op,vpt,threshold) --
 */
static bool ShouldSpecializeOperand(Instruction *instr, Operand *op,
    ProfStats *vpt, double threshold) {
    EDGE_WEIGHT_TYPE potential = 0, cost, neededSavings;

    if (OperandIsGPRegister(op)) {
	potential = ComputeRegisterPotential(op->a_base_reg, instr);
    } else if (OperandAccessesStackFromFramePointer(op)) {
	potential = ComputeStackPotential(op->a_displacement, instr);
    } else {
	MALFUNCTION("Unexpected operand type encountered");
    }

    neededSavings = threshold * hotInstrWeight;
    cost = instr->i_Bblock->b_weight * COST_OF_SPECIALIZATION_TEST;
    potential = potential * vpt->ps_top_dist;

    if ((potential - cost) > neededSavings) {
	return true;
    }

    return false;
}



/*
 *  RemoveSpecialization(aSpec) --
 */
static void RemoveSpecialization(Specialization *aSpec) {

    if (specializationList == aSpec) {
	specializationList = aSpec->sz_next;
    }
    if (aSpec->sz_prev) {
	aSpec->sz_prev->sz_next = aSpec->sz_next;
    }
    if (aSpec->sz_next) {
	aSpec->sz_next->sz_prev = aSpec->sz_prev;
    }

    return;
}
