/*
 *  Unreachable-code elimination and object destruction module.
 */
#include "global.h"
#include "linkage.h"
#include "opt.unreachable.h"
#include "utilities.h"
#include "flowgraph.h"
#include "hash.h"
#include "sysdep.h"
#include "relocate.h"
#include "relocate.h"
#include "upm.h"

/* Marks all blocks that are live, by setting their BBL_TEMP_FLAG */
static void BBLMarkLive(Block *bbl);

/* Removes all blocks that are not live from the control flow graph */
static bool KillUnreachableBlocks();

int NOPSKilled;			// number of NOP(0x90) instructions removed
int instructionsKilled;		// total number of instructions removed
int basicBlocksKilled;		// total number of basic blocks removed
int functionsKilled;		// total number of functions removed

typedef struct HellCall {
    struct HellCall *hc_next;
    Block *hc_from;
    Block *hc_to;
} HellCall;


static HellCall *tmp_hell_calls;

static void IndirectHellRemoveTemporary();

static void IndirectHellRestore();



/*
 *  RemoveUnreachableCode(bbl) --
 *
 *  Remove unreachable basic blocks from the control flow graph.  A depth first
 *  search is used, and return blocks are ignored since they have return edges
 *  to all possible callees.  If we followed the edge, we might hit a function
 *  that's unreachable.  The meat of this function lies in BBLMarkLive()
 */
void RemoveUnreachableCode(Block *bbl) {
    Function *fiterator, *next;
    Block *biterator;
    bool cont = true;
    int round;

    CASSERT(bbl != NULL, "Passed a NULL block.");
    REPORT(RP_MINOR, WITHFUNC("Removing unreachable code"));

    round = 1;

    /*
     * Removing one block may result in another block being a removal candidate.
     * For this reason, we need to continue to unmark/mark/remove blocks until
     * none are being removed. Typically 3-4 rounds is enough to get all the
     * blocks.
     */

    while (cont == true) {
	instructionsKilled = basicBlocksKilled = functionsKilled = NOPSKilled =
	    0;

	/* Unmark all blocks as being live */
	FOREACH_BBL(biterator) {
	    BBL_REMOVE_FLAG(biterator, BBL_LIVE_FLAG);
	}
	BBL_REMOVE_FLAG(HELL, BBL_LIVE_FLAG);

	IndirectHellRemoveTemporary();

	/* Mark all functions to catch blocks that are called indirectly  */
	FOREACH_FUNCTION(fiterator) {
	    BBLMarkLive(fiterator->f_first);
	}

	IndirectHellRestore();

	cont = KillUnreachableBlocks();

	/* If all the basic blocks in a function are killed,
	 * then we can kill the function */
	fiterator = f_list;
	while (fiterator != NULL) {
	    next = fiterator->f_next;

	    bool dead = true;
	    FOREACH_BBL_IN_FUN(fiterator, biterator) {
		if (BBL_HAS_FLAG(biterator, BBL_LIVE_FLAG)) {
		    dead = false;
		}
	    }
	    if (dead == true) {
		functionsKilled++;
		DEBUG(DB_KILLS, "Killed function %s", fiterator->f_name);
		FunDestroy(fiterator);
	    }

	    fiterator = next;
	}

	REPORT(RP_DETAIL, "Round %d: killed %d instructions (%d NOPs)",
	    round, instructionsKilled, NOPSKilled);
	REPORT(RP_DETAIL, "Round %d: killed %d functions and %d basic blocks",
	    round, functionsKilled, basicBlocksKilled);
	CountStructures();
	if (instructionsKilled + functionsKilled + basicBlocksKilled > 0) {
	    BBLOutputStats();
	}

	round++;

    }

    CheckConsistency();

    return;
}



/*
 *  BBLMarkLive(bbl) --
 *
 *  Marks all reachable basic blocks as live by using a depth-first search.
 */
static void BBLMarkLive(Block *bbl) {
    EdgeNode *niterator;

    BBL_SET_FLAG(bbl, BBL_LIVE_FLAG);
    FOREACH_SUCC_IN_BBL(bbl, niterator) {
	if (NODE_TYPE(niterator) == ET_RETURN) {
	    continue;
	}

	if (!BBL_HAS_FLAG(NODE_TO(niterator), BBL_LIVE_FLAG)) {
	    BBLMarkLive(NODE_TO(niterator));
	}
    }
}



/*
 *  KillUnreachableBlocks() --
 *
 *  Kills any unreachable block (those not marked as being live) from the graph.
 */
static bool KillUnreachableBlocks() {
    Block *biterator;
    Instruction *iterator, *next;
    bool destroy = false;

    FOREACH_BBL(biterator) {
	if (!BBL_HAS_FLAG(biterator, BBL_LIVE_FLAG)) {
	    BBLRemoveEdges(biterator);
	    BBLRemovePredecessors(biterator);
	}
    }

    FOREACH_BBL(biterator) {
	if (!BBL_HAS_FLAG(biterator, BBL_LIVE_FLAG)) {
	    iterator = biterator->b_first;
	    while ((iterator != NULL) && (iterator->i_Bblock == biterator)) {
		next = iterator->i_next;
		InstrDestroy(iterator);
		iterator = next;
	    }
	    destroy = true;
	}
    }

    return destroy;
}



/*
 *  InstrDestroy(instr) --
 *
 *  Disconnects killed instructions from the graph, and attempts to
 *  remove edges from HELL to possible targets.  Since HELL has edges
 *  (parallel edges, too) to all possible blocks gathered from
 *  relocation points, we can remove some of the edges if we remove
 *  relocation points through instruction elimination.
 *  returns true if the instruction was successfully removed.
 */
bool InstrDestroy(Instruction *instr) {
    Instruction *linkInstr;
    EdgeNode *predecessors;
    bool callClass = false;

    if (instr->i_Bblock->b_type == BT_CALL) {
	callClass = true;
	if (isCallClass(instr)) {
	    if (INSTR_HAS_FLAG(instr, INDIRECT)) {
		if (BBLGetEdgeToBBLTyped(HELL, instr->i_Bblock->b_next,
			ET_HELLMAYBE) != NULL) {
		    BBLRemoveEdgeTyped(HELL, instr->i_Bblock->b_next,
			ET_HELLMAYBE);
		}
	    }
	}
    }

    if (instr->i_Bblock->b_last == instr) {
	if (BBLNumberOfSuccessors(instr->i_Bblock) == 0) {
	    if (instr->i_Bblock->b_next) {
		linkInstr = instr->i_Bblock->b_next->b_first;
	    } else {
		linkInstr = NULL;
	    }
	} else {
	    if (instr->i_Bblock->b_type == BT_CALL) {
		linkInstr = BBLGetFunlinkEdge(instr->i_Bblock)->e_to->b_first;
	    } else {
		linkInstr = NODE_TO(instr->i_Bblock->b_succ)->b_first;
	    }
	}
    } else {
	linkInstr = instr->i_next;
    }


    if (instr == startins) {
	CASSERT (linkInstr != NULL, 
	    "Destroying start instruction, and no link instruction!");
	startins = linkInstr;
    }
// if (linkInstr == NULL) {
    InstrDestroyUPMs(instr);
// }
// else {
// // WARNING("Changing upm sources to a link instruction! Dangerous!!");
//   UPMChangeAllSourcesFromInstrToInstr(instr, linkInstr);
// }

    if (isOnlyInstrInBBL(instr)) {

	/*
	 * We take all the predecessors of this basic block and redirect
	 * them to the basic block of the linking instruction.
	 */

	if (callClass) {
	    Edge *funDirect = BBLGetFunDirectEdge(instr->i_Bblock);
	    Edge *funLink = BBLGetFunlinkEdge(instr->i_Bblock);

	    if (funLink != NULL) {
		if (funDirect == NULL) {
		    // it's possible that this call was to a removed function,
		    // and in that case it will have had it's fundirect edge
		    // removed already but does not go to HELL
		    if (BBLGetEdgeToBBL(instr->i_Bblock, HELL) != NULL) {
			BBLRemoveEdge(instr->i_Bblock, HELL);
			BBLRemoveEdge(HELL, funLink->e_to);
		    }
		    BBLRemoveEdge(instr->i_Bblock, funLink->e_to);
		} else {
		    REPORT(RP_LGDETAIL, "Destroy %x: %s",
			instr->i_address, PrintInstruction(instr));
		    // for recursive calls the entry and exit may have already
		    // been removed and therefore there will be no edges in
		    // that case (this happens when the exit block comes before
		    // this instruction per the layout)
		    if (INSTR_FUNCTION(instr) == funDirect->e_to->b_fun) {
			if (BBLGetEdgeToBBL(funDirect->e_to->b_fun->f_exit,
				funLink->e_to) != NULL) {
			    BBLRemoveEdge(
				funDirect->e_to->b_fun->f_exit, funLink->e_to);
			}
			if (BBLGetEdgeToBBL(instr->i_Bblock,
				funDirect->e_to->b_fun->f_entry) != NULL) {
			    BBLRemoveEdge(instr->i_Bblock,
				funDirect->e_to->b_fun->f_entry);
			}
		    } else {
			BBLRemoveEdge(
			    funDirect->e_to->b_fun->f_exit, funLink->e_to);
			BBLRemoveEdge(
			    instr->i_Bblock, funDirect->e_to->b_fun->f_entry);
		    }

		    BBLRemoveEdge(instr->i_Bblock, funDirect->e_to);
		    BBLRemoveEdge(instr->i_Bblock, funLink->e_to);
		}
	    }
	}

	predecessors = instr->i_Bblock->b_pred;
	while ((predecessors != NULL) && (linkInstr != NULL)) {
	    EdgeNode *next = predecessors->en_next;
	    Edge *existingEdge;
	    Block *from = NODE_FROM(predecessors);
	    UINT32 type = NODE_TYPE(predecessors);
	    EDGE_WEIGHT_TYPE weight = NODE_WEIGHT(predecessors);

	    BBLRemoveEdgeTyped(NODE_FROM(predecessors), NODE_TO(predecessors),
		type);

	    existingEdge = BBLGetEdgeToBBL(from, linkInstr->i_Bblock);
	    if (existingEdge && (type == ET_FALSE)
		&& (existingEdge->e_type == ET_TRUE)) {
		from->b_type = BT_UNCOND;
		BBLRemoveEdge(from, linkInstr->i_Bblock);
		BBLAddEdge(from, linkInstr->i_Bblock, ET_UNCOND)->e_edgeWeight =
		    weight;
	    } else if (existingEdge && (type == ET_TRUE)
		&& (existingEdge->e_type == ET_FALSE)) {
		from->b_type = BT_UNCOND;
		BBLRemoveEdge(from, linkInstr->i_Bblock);
		BBLAddEdge(from, linkInstr->i_Bblock, ET_UNCOND)->e_edgeWeight =
		    weight;
	    } else {
		BBLAddEdge(from, linkInstr->i_Bblock, type)->e_edgeWeight =
		    weight;
	    }

	    predecessors = next;
	}

	BBLRemoveEdges(instr->i_Bblock);
	BBLRemovePredecessors(instr->i_Bblock);
	BBLDisconnect(instr->i_Bblock);
	BBLFree(instr->i_Bblock);
    }

    if (instr->i_address != MAGIC_NUMBER) {
	InstrRemoveEntry(instr);
    }

    InstrDisconnect(instr);
    instructionsKilled++;
    if (isNOPClass(instr)) {
	NOPSKilled++;
    }

    DEBUG(DB_KILLS, "[0x%8.8x] Instruction killed", instr->i_address);

    if (!isPCChangingClass(instr)) {
	InstrFree(instr);
	return true;
    }

    InstrFree(instr);

    return true;
}



/*
 *  FunctionIsExtinct(fun) -- returns true if the function is extinct.
 */
bool FunctionIsExtinct(Function *fun) {
    Block *bbl;
    EdgeNode *preds;

    if (BBLNumberOfPredecessors(fun->f_entry) > 0) {
	return false;
    }

    FOREACH_BBL_IN_FUN(fun, bbl) {
	if (BBLHasHellPredecessor(bbl)) {
	    return false;
	}
	FOREACH_PRED_IN_BBL(bbl, preds) {
	    if (NODE_TYPE(preds) == ET_RETURN) {
		continue;
	    }
	    if (((NODE_FROM(preds))->b_fun) != fun) {
		return false;
	    }
	}
    }

    return true;
}



/*
 *  FunDestroy(fun) --
 *
 *  Disconnects the function from the control flow graph, and frees up all
 *  memory that is used by it.
 */
void FunDestroy(Function *fun) {
    Block *blocks, *next;
    Instruction *instr, *nextInstr;

    FunDisconnect(fun);
    FunDestroyUPMs(fun);

    blocks = fun->f_first;
    while (blocks != NULL) {
	next = blocks->b_next;
	instr = blocks->b_first;
	while ((instr != NULL) && instr->i_Bblock == blocks) {
	    nextInstr = instr->i_next;
	    InstrDestroy(instr);
	    instr = nextInstr;
	}
	blocks = next;
	if (blocks->b_fun != fun) {
	    break;
	}
    }

    BBLRemoveEdges(fun->f_entry);
    BBLRemoveEdges(fun->f_exit);
    BBLRemovePredecessors(fun->f_entry);
    BBLRemovePredecessors(fun->f_exit);
    BBLFree(fun->f_entry);
    BBLFree(fun->f_exit);

    return;
}



/*
 *  IndirectHellRemoveTemporary() --
 *
 *  Removes all HELL edges from indirect calls to HELL and the
 *  corresponding HELL edge back to the funlink successor.  We do this
 *  just before we mark basic blocks for liveness, otherwise dead code may
 *  be considered live because all edges in HELL are traversed.
 *  see also: tmp_hell_calls
 *  see also: HellCall
 */
static void IndirectHellRemoveTemporary() {
    Edge *edge;
    Block *biterator;
    HellCall *last, *newone;

    last = tmp_hell_calls = zalloc(sizeof(HellCall));
    tmp_hell_calls->hc_next = NULL;

    FOREACH_BBL(biterator) {
	if (biterator->b_type == BT_CALL) {
	    if (INSTR_HAS_FLAG(biterator->b_last, INDIRECT)) {
		edge = BBLGetFunlinkEdge(biterator);

		newone = zalloc(sizeof(HellCall));
		newone->hc_from = biterator;
		newone->hc_to = edge->e_to;
		last->hc_next = newone;
		newone->hc_next = NULL;

		last = newone;

		BBLRemoveEdgeTyped(biterator, HELL, ET_HELL);
		BBLRemoveEdgeTyped(HELL, edge->e_to, ET_HELLMAYBE);
	    }
	}
    }
}



/*
 *  IndirectHellRestore() --
 *
 *  Restores all HELL edges from indirect calls to HELL and the
 *  corresponding HELL edge back to the funlink successor.  We do this
 *  just after we mark basic blocks for liveness to maintain correctness
 *  for the rest of the operation of the program.
 *  see also: tmp_hell_calls
 *  see also: HellCall
 */
static void IndirectHellRestore() {
    HellCall *tchiterator, *next;

    tchiterator = tmp_hell_calls->hc_next;
    free(tmp_hell_calls);

    while (tchiterator != NULL) {
	next = tchiterator->hc_next;

	if (BBL_HAS_FLAG(tchiterator->hc_from, BBL_LIVE_FLAG)) {
	    BBLAddEdge(tchiterator->hc_from, HELL, ET_HELL);
	    BBLAddEdge(HELL, tchiterator->hc_to, ET_HELLMAYBE);
	}

	free(tchiterator);

	tchiterator = next;
    }
}

