/*
 *  Drives the disassembly/assembly/optimization processes. A binary is
 *  read-in, normalized, a flowgraph is constructed, various optimizations
 *  are made, and then the binary is rewritten.
 */

#define GLOBAL		// causes globals to be defined by this file

#include "global.h"
#include "disassemble.h"
#include "flowgraph.h"
#include "hash.h"
#include "utilities.h"
#include "relocate.h"
#include "lto.h"
#include "sloop.h"
#include "sysdep.h"
#include "linkage.h"
#include "edge.profile.h"
#include "opt.layout.h"
#include "dominator.h"
#include "liveness.h"
#include "opt.h"
#include "upm.h"
#include "time.h"
#include "liveness.h"
#include "stack.analyze.h"
#include "elfio.h"



/*
 *  FinalizeMemoryImage() --
 *
 *  'Finalizes' the memory image associated with the code sections. Usage
 *  of this function need not be limited to true finalization since it may
 *  be called multiple times (in this sense the name is misleading) but it
 *  must be called directly before writing out the binary. This function
 *  updates all addresses, section sizes, UPMs, instruction sizes, etc.
 */
void FinalizeMemoryImage() {
    Section *seciterator;

    REPORT(RP_MINOR, WITHFUNC("Recomputing addresses"));
    sec_list = RemoveDroppedSectionsFromList(sec_list);

    REPORT(RP_DETAIL, "Assigning addresses");
    RecalculateAddresses();

    REPORT(RP_DETAIL, "Reassembling instructions");
    FOREACH_SECTION(seciterator) {
	if (SECTION_WAS_MERGED(seciterator)) {
	    AssembleCode(seciterator);
	}
    }
}

/*
 *  UpdateOldInstrAddresses() --
 */
void UpdateOldInstrAddresses() {
    Instruction *iiterator;

    FinalizeMemoryImage();

    InstrClearHashTable();

    FOREACH_INSTR(iiterator) {
	iiterator->i_address = iiterator->i_final_address;
	InstrInsertEntry(iiterator);
    }

    startAddress = startins->i_address;
}

/*
 *  UpdateOldSectionAddresses() --
 */
void UpdateOldSectionAddresses() {
    Section *seciterator;

    FOREACH_SECTION(seciterator) {
	seciterator->s_address = seciterator->s_newaddress;
    }
}

/*
 * Normalizes the control flow graph to make sure the cmp
 * and branch instructions are placed consecutively.
 * - somu
 */
static void NormalizeCmpAndBranch() {
    Block *biterator = 0;
    Instruction *instr = 0;
    REGISTER_BITVECTOR regUsed, regDefined;
    REGISTER_BITVECTOR regUsed1, regDefined1;
    int tcount = 0, ncount = 0;

    REPORT(RP_MINOR, WITHFUN("Normalizing cmp and jumps"));

    FOREACH_BBL(biterator) {
        regDefined = 0;
        // It has to have a compare and branch
        // in the same basic block. Hence if there
        // is just a branch in the basic block then
        // we do nothing.
        
        if (biterator->b_type == BT_BRANCH &&
            biterator->b_first != biterator->b_last) {

            // If the immediately previous instruction does not
            // define PSW find the instruction which does
            
            if (!ComputePSWDefined(biterator->b_last->i_prev)) {
                tcount++;
                instr = biterator->b_last->i_prev;
                
                while (instr->i_Bblock == biterator) {
                    if (ComputePSWDefined(instr)) {
                        break;
                    }
                    regDefined |= ComputeRegistersDefined(instr);
                    regUsed1   |= ComputeRegistersUsed(instr);
                    instr = instr->i_prev;
                }

                if (instr->i_Bblock != biterator) {
                    continue;
                }

                if ((instr->i_src && OperandIsMemory(instr->i_src)) ||
                    (instr->i_src2 && OperandIsMemory(instr->i_src2))) {
                    continue;
                }
                
                regUsed = ComputeRegistersUsed(instr);
                regDefined1 = ComputeRegistersDefined(instr);

                // Here we check for Write After Read ,
                // Read After Write and Write After Write
                // Hazards
                
                if (!(regUsed & regDefined) &&
                    !(regDefined1 & regUsed1) &&
                    !(regDefined1 & regDefined) &&
                    (instr->i_Bblock == biterator)) {
                    ncount++;
                    // Change the UPM sources for the instruction
                    if (InstrIsUPMSource(instr)) {
                        if (INSTR_HAS_FLAG(instr, BBL_START)) {
                            UPMChangeAllSourcesFromInstrToInstr(instr,
                                instr->i_next);
                        } else {
                            WARNING(WITHFUN("Instr is upm source "
                                        "but not first in bbl"));
                        }
                    }
                    instr = InstrDisconnect(instr);
                    InsertInstrBeforeInstr(instr, biterator->b_last);
                }
            }
        }
    }
    REPORT(RP_MINOR, WITHFUN("Normalized %d cmp and jumps out of the total %d"),
        ncount, tcount);    
}


/*
 * Brings pop ebp/leave and ret closer
 */

static void NormalizeRetBlocks() {
    Block *biterator = 0;
    Instruction *instr = 0, *prevInstr = 0;
    REGISTER_BITVECTOR regUsed, regDefined;
    REGISTER_BITVECTOR regUsed1, regDefined1;
    int tcount = 0, ncount = 0;

    REPORT(RP_MINOR, WITHFUN("Normalizing leave/pop and rets"));
    FOREACH_BBL(biterator) {
        regDefined = 0;
        regDefined1 = 0;
        regUsed1   = 0;
        regUsed    = 0;
        // If it is just a return nothing much
        // can be done about it
        if (biterator->b_type == BT_RETURN &&
            biterator->b_first != biterator->b_last) {
            
            prevInstr = biterator->b_last->i_prev;

            // if the leave/pop and return pair are together
            // please continue
            if (isLeaveClass(prevInstr) ||
                (isPopClass(prevInstr)
                    && OperandIsFramePointer(prevInstr->i_dest))) {
                continue;
            }
            
            // If the immediately previous instruction does not
            // define PSW find the instruction which does
            tcount++;
            instr = prevInstr;
                
            while (instr->i_Bblock == biterator) {
                if (isLeaveClass(instr) ||
                    (isPopClass(instr) &&
                        OperandIsFramePointer(instr->i_dest))) {
                    break;
                }
                regDefined |= ComputeRegistersDefined(instr);
                regUsed1   |= ComputeRegistersUsed(instr);
                instr = instr->i_prev;
            }

            if (instr->i_Bblock != biterator) {
                continue;
            }

            regUsed = ComputeRegistersUsed(instr);
            regDefined1 = ComputeRegistersDefined(instr);

            // Here we check for Write After Read ,
            // Read After Write and Write After Write
            // Hazards
                
            if (!(regUsed & regDefined) &&
                !(regDefined1 & regUsed1) &&
                !(regDefined1 & regDefined) &&
                (instr->i_Bblock == biterator)) {
                ncount++;
                // Change the UPM sources for the instruction
                if (InstrIsUPMSource(instr)) {
                    if (INSTR_HAS_FLAG(instr, BBL_START)) {
                        UPMChangeAllSourcesFromInstrToInstr(instr,
                                                            instr->i_next);
                    } else {
                        WARNING(WITHFUN("Instr is upm source "
                                    "but not first in bbl"));
                    }
                }
                instr = InstrDisconnect(instr);
                InsertInstrBeforeInstr(instr, biterator->b_last);
            }
        }
    }
    REPORT(RP_MINOR, WITHFUN("Normalized %d leave/pop and ret"
                " out of the total %d"), ncount, tcount);    
}

static bool inlineEscJumps(Function *fun) {
    Block *bbl = 0, *biterator = 0;
    Edge *escape = 0, *tempEdge = 0;
    Function *target = 0;
    static int numberOfEscapingEdgesFixed = 0;
    static int totalEscapingEdges = 0;

    if (!fun) {
        REPORT(RP_MINOR, WITHFUN("Total Escaping Edges Encounterd %d"),
            totalEscapingEdges);
        REPORT(RP_MINOR, WITHFUN("Total Escaping Edges Fixed %d"),
            numberOfEscapingEdgesFixed);
        numberOfEscapingEdgesFixed = 0;
        totalEscapingEdges = 0;
        return true;
    }
    
    FOREACH_BBL_IN_FUN(fun, bbl) {
        if ((escape = BBLGetEscapingEdge(bbl))) {
            target = escape->e_to->b_fun;
            totalEscapingEdges++;
            if (!target->f_name) {
                continue;
            }
            if (strstr(target->f_name, ".text.lock")) {
                numberOfEscapingEdgesFixed++;
                FOREACH_BBL_IN_FUN(target, biterator) {
                    BBL_REMOVE_FLAG(biterator, BBL_MARKED_FLAG);
                }

                biterator = escape->e_to;
                
                while (biterator->b_type != BT_UNCOND) {
                    BBL_SET_FLAG(biterator, BBL_MARKED_FLAG);
                    biterator = biterator->b_next;
                }
                
                BBL_SET_FLAG(biterator, BBL_MARKED_FLAG);
                tempEdge = BBLGetUncondEdge(biterator);

                if (tempEdge) {
                    BBLRemoveEdge(fun->f_entry, tempEdge->e_to);
                }
            }
        }
    }
    return true;
}


/*
 * Function which performs any normalization of
 * CFGs should be done here
 */
static void NormalizeCFG() {
    Function *fun = 0;

    REPORT(RP_MAJOR, "Normalizing flowgraph");

    if (!KERNEL_MODE) {
        NormalizeCmpAndBranch();
    }
    
    NormalizeRetBlocks();

    if (KERNEL_MODE) {
        FOREACH_FUN(fun) {
            if (!DonotTouchMe(fun)) {
                inlineEscJumps(fun);
            }
        }
        inlineEscJumps(0);
    }
}


/*
 *  SetupDataStructures() -- set up initial PLTO data structures.
 *
 *	1. Load program binary.
 *	2. Create an opcode-hash table -- this initializes the disassembler.
 *	3. Disassemble the binary.
 *	4. Creates UPMs.
 *	5. Remove Call 0 type instructions (for simplifying call graph).
 *	6. Build ControlFlowGraph.
 *	7. Create additional UPMs for symbol table entries.
 *	8. Update the instruction & section addresses.
 */
void SetupDataStructures() {

    SetupPltoDataStructures();


    REPORT(RP_MAJOR, "Disassembling code");
    BuildOpcodeHashTable();
    upm2s_list = ll_new(LIST_UNSORTED);
    i_list = DisassembleSection(SECTION_REF_CODE);


    InitializeUPMs();


    ReplaceHaltBranchesWithHalts();

    REPORT(RP_MAJOR, "Constructing flowgraph");
    BuildControlFlowGraph(i_list);

    // Puts the compares next to the
    // branches
    NormalizeCFG();

    AddSymbolUPMs();
    UpdateOldInstrAddresses();

    UpdateOldSectionAddresses();

}



/*
 *  CML_TestStackAnalysis() --
 */
void CML_TestStackAnalysis() {
    Function *fiterator;
    int known_trash, known_use;
    long long total;
    clock_t start, end;

    /*
     * Get the timing info for the analysis.
     */
    total = 0;
    start = clock();
    MarkWellBehavedFunctions2();
    StoreInstrBPSPDiffForAllFuns();
    end = clock();
    fprintf(stdout,
	"[%s:%d] Stack Analysis Start: %ld, End: %ld, Elapsed: %ld\n",
	__FILE__, __LINE__, start * 1000 / CLOCKS_PER_SEC,
	end * 1000 / CLOCKS_PER_SEC, (end - start) * 1000 / CLOCKS_PER_SEC);
    total += end - start;

    PrintStackAnalysisSummary();

    start = clock();
    ComputeFunctionTrashDepths();
    end = clock();
    fprintf(stdout,
	"[%s:%d] Trash Depths Start: %ld, End: %ld, Elapsed: %ld\n",
	__FILE__, __LINE__, start * 1000 / CLOCKS_PER_SEC,
	end * 1000 / CLOCKS_PER_SEC, (end - start) * 1000 / CLOCKS_PER_SEC);
    total += end - start;

    start = clock();
    ComputeFunctionUseDepths();
    end = clock();
    fprintf(stdout,
	"[%s:%d] Use Depths Start: %ld, End: %ld, Elapsed: %ld\n",
	__FILE__, __LINE__, start * 1000 / CLOCKS_PER_SEC,
	end * 1000 / CLOCKS_PER_SEC, (end - start) * 1000 / CLOCKS_PER_SEC);
    total += end - start;

    fprintf(stdout,
	"[%s:%d] Total Elapsed Time for All: %lld\n",
	__FILE__, __LINE__, total * 1000 / CLOCKS_PER_SEC);


    /*
     * Get the statistics on the trash and use depths
     */

    known_trash = known_use = 0;
    FOREACH_FUNCTION(fiterator) {
	if (fiterator->f_trashDepth >= 0) {
	    known_trash++;
	} else {
//    REPORT(RP_LGDETAIL, "Unknown Trash Depth for %s", ((fiterator)->f_name));
	}

	if (fiterator->f_useDepth >= 0) {
	    known_use++;
	}
    }

    fprintf(stdout,
	"[%s:%d] Functions with known trash depths: %d\n",
	__FILE__, __LINE__, known_trash);
    fprintf(stdout,
	"[%s:%d] Functions with known use depths: %d\n",
	__FILE__, __LINE__, known_use);
}



/*
 *  CodeSection(address) -- returns the section associated with a code address.
 */
Section *CodeSection(ADDRESS address) {

    Section *sec = SECTION_REF_CODE;
    if (address < sec->s_newaddress) {
	return NULL;
    }
    while (sec != NULL && address >= sec->s_newaddress + sec->s_newdatasize) {
	sec = sec->s_next;
    }
    return sec;
}



/*
 *  SplitInstrs() -- marks instructions that begin new (split) code sections.
 */
void SplitInstrs() {
    Section *seciterator;

    REPORT(RP_DETAIL, "Finding boundaries");
    FOREACH_SECTION_IN_LIST(mrg_list, seciterator) {
	Symbol *sym = seciterator->s_label;
	Instruction *instr = InstrLookupEntry(sym->sy_value);
	INSTR_SET_FLAG(instr, INSTR_STARTS_SECTION);
	seciterator->s_instr = instr;
    }
}



/*
 *  SplitCode() -- Recreates multiple sections from the single code section.
 */
static void SplitCode() {
    Section *seciterator, *last;
    int length;

    REPORT(RP_DETAIL, "Splitting code");

    /* Restore the original name of the first section */
    SECTION_REF_CODE->s_name = SECTION_REF_CODE->s_label->sy_name;
    REPORT(RP_SMLOOP, "Split at %8X  %s", SECTION_REF_CODE->s_newaddress,
	SECTION_REF_CODE->s_name);

    /* Restore section symbols and set section addresses from them */
    FOREACH_SECTION_IN_LIST(mrg_list, seciterator) {
	Symbol *symbol = seciterator->s_label;
	symbol->sy_flags &= ~SYMBOL_NOTYPE_FLAG;
	symbol->sy_flags |= SYMBOL_SECTION_FLAG;
	seciterator->s_newaddress = symbol->sy_value;
	REPORT(RP_SMLOOP, "Split at %8X  %s",
	    seciterator->s_newaddress, seciterator->s_name);
    }

    /* Copy section contents from the merged code section */
    FOREACH_SECTION_IN_LIST(mrg_list, seciterator) {

	int offset = seciterator->s_newaddress
	    - (SECTION_REF_CODE->s_newaddress);
	if (seciterator->s_next != NULL) {
	    length = seciterator->s_next->s_newaddress
		- seciterator->s_newaddress;
	} else {
	    length = SECTION_REF_CODE->s_newdatasize - offset;
	}

	seciterator->s_newdatasize = length;
	seciterator->s_newdata = alloc(seciterator->s_newdatasize);
	memcpy(seciterator->s_newdata,
	    SECTION_REF_CODE->s_newdata + offset, length);

	last = seciterator;
    }

    /* Link the merged sections behind the original code section */
    last->s_next = SECTION_REF_CODE->s_next;
    last->s_next->s_prev = last;
    mrg_list->s_prev = SECTION_REF_CODE;
    SECTION_REF_CODE->s_next = mrg_list;

    /* Resize the code section to exclude portions moved out */
    SECTION_REF_CODE->s_newdatasize =
	mrg_list->s_newaddress - SECTION_REF_CODE->s_newaddress;

}



/*
 *  SplitSymbols() -- adjusts symbols for the section split.
 */
static void SplitSymbols() {
    Symbol *sym;

    REPORT(RP_DETAIL, "Splitting symbols");
    FOREACH_SYMBOL(sym) {
	if (sym->sy_section == SECTION_REF_CODE) {
	    sym->sy_section = CodeSection(sym->sy_value);
	}
    }
}



/*
 *  SplitRelocs() -- adjusts relocations for the code split.
 */
static void SplitRelocs() {
    Relocation *rlc;
    ADDRESS base = SECTION_REF_CODE->s_newaddress;

    REPORT(RP_DETAIL, "Splitting relocations");
    FOREACH_RELOCATION(rlc) {
	if (rlc->r_section == SECTION_REF_CODE) {
	    rlc->r_section = CodeSection(base + rlc->r_offset);
	}
    }
}



/*
 *  SplitSections() --
 *
 *  Splits the code section back into its original parts as determined
 *  by the mrg_list of merged-away sections and the symbols they track.
 *  Upon return, the restored sections immediately follow SECTION_REF_CODE
 *  and can be distinguished, if necessary, by non-null MERGETO fields.
 */
void SplitSections() {

    if (mrg_list == NULL) {
	REPORT(RP_MINOR, "NO SECTIONS TO SPLIT");
	return;
    }

    UpdateOldInstrAddresses();

    REPORT(RP_MINOR, "Splitting sections");
    SplitInstrs();
    SplitCode();
    SplitSymbols();
    SplitRelocs();
}
