/*
 *  potential.c
 *
 *  Functions for computing the potential savings from knowing register or
 *  stack values We can compute the potential in a few different ways:
 *
 *  List Scheduling:
 *  Suppose we know the value of a location, l, right before the
 *  instruction 'instr' in basic block 'bbl'. We can figure out which
 *  instructions can be eliminated as a result of knowing l by walking
 *  through the instructions in 'bbl'. Then, we can figure out which new
 *  locations became known as a result of knowing l. Call this set L_bbl.
 *  We can propagate information about L_bbl to all the realizable
 *  successors of bbl that are also dominated by bbl.  Then, for any block
 *  for which all predecessors have been analyzed (using the method listed
 *  above), we can analyze it and propagate the results as described above.
 *  This method doesn't use any sort of fixpoint computation, and since
 *  the set of blocks to analyze is restricted by the dominator relation,
 *  the analysis is not very time consuming. It is also not very accurate
 *  in some circumstances, as some blocks with instructions that could be
 *  eliminated are never examined due to the list scheduling scheme. In
 *  particular, consider a loop, where the loop header is never analyzed
 *  since it requires the end of loop to first be examined (which in turn
 *  requires the loop header to be examined). So there is a dependence
 *  relationship that cannot easily be resolved. For situations where this
 *  may not be desirable, other methods for computing potential may be more
 *  desirable.
 */
#include "global.h"
#include "potential.h"
#include "sysdep.h"
#include "utilities.h"
#include "cp.h"
#include "flowgraph.h"
#include "x86-registers.h"

#define SLOT_KNOWN_FLAG        (1 << 12)
#define SLOT_UNDEFINED_FLAG    (1 << 13)
#define SLOT_UNKNOWN_FLAG      (1 << 14)

/* Machine-dependent */
#define RETURN_REGISTER        (reg_a | reg_32)

#define EXCLUSIVE   15
#define INCLUSIVE   17

typedef struct Known {
    struct Known *kn_prev;
    struct Known *kn_next;
    bool kn_isRegister;
    Instruction *kn_instr;
    int kn_value;
} Known;


/* Updates the snapshot in instr's basic block to set all locations
 * defined by 'instr' as not known */
static void UpdateNotKnown(Instruction *instr, PSW_BITVECTOR *currentPSW);

/* Updates the snapshot in instr's basic block to set all locations
 * defined by 'instr' as known */
static void UpdateKnown(Instruction *instr, PSW_BITVECTOR *currentPSW);

/* Returns true if the operand contains a known constant value */
static bool OperandIsKnown(Instruction *instr, Operand *op);

/* Returns true if all the PSW bits used by 'instr' are known in the
 * bitvector currentPSW */
static bool PSWIsKnown(Instruction *instr, PSW_BITVECTOR currentPSW);

/* Returns true if something changed in the basic block containing 'instr'. */
static bool EvaluateBBL(Instruction *instr);

static void MergePredecessors(Block *bbl);

/*
 *  Returns the potential associated with knowing any locations set with
 *  SLOT_KNOWN_FLAG before 'instr' is executed.
 */
static EDGE_WEIGHT_TYPE ComputePotential(Instruction *instr);

static void InitializeBBLFlags(Block *bbl);

/*
 *  If definitions of 'op' were previously recorded as being eliminable,
 *  this function subtracts their weight from 'potential'.  This function
 *  is used to remove previously-credited potential when we find out that
 *  the use of some location has multiple reaching definitions
 */
static void RemovePotentialForReachableDefinitions(Instruction *instr,
    Operand *op);

/* Returns the number of definitions that can reach locations used by 'op' */
static EDGE_WEIGHT_TYPE ReachableDefinitions(
    Instruction *instr, Operand *op, bool discredit);

static EDGE_WEIGHT_TYPE ReachableDefinitionsOfRegister(
    Instruction *instr, int reg, int includeInstr, bool discredit);
static EDGE_WEIGHT_TYPE ReachableDefinitionsOfStack(
    Instruction *instr, int loc, int includeInstr, bool discredit);

static Value *BBLCopyStackFlags(Block *bbl);
static Value *BBLCopyRegisterFlags(Block *bbl);
static bool BBLRegFlagsAreEqual(Value *one, Block *bbl);
static bool BBLStackFlagsAreEqual(Value *one, Block *bbl);

static void AddKnownLocationToList(Known *newLoc);

static int definedSlotsPot[(MAXIMUM_AR_SIZE * 2) + 1];

static Known *knownList;



/*
 *  AddKnownRegister(aReg,instr) --
 *
 *  Adds a known register to the basic block which contains 'instr'.
 */
void AddKnownRegister(int aReg, Instruction *instr) {
    Known *newLoc;
    Block *bbl = instr->i_Bblock;
    int regIndex = RegisterToIndex(aReg);

    if (!BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
	return;
    }

    bbl->b_regshot[regIndex].v_flags |= SLOT_KNOWN_FLAG;

    newLoc = zalloc(sizeof(Known));
    newLoc->kn_isRegister = true;
    newLoc->kn_value = aReg;
    newLoc->kn_instr = instr;
    AddKnownLocationToList(newLoc);

    return;
}



/*
 *  AddKnownStackSlot(loc,instr) --
 *
 *  Adds a known stack slot to the basic block which contains 'instr'.
 */
void AddKnownStackSlot(int loc, Instruction *instr) {
    Known *newLoc;
    Block *bbl = instr->i_Bblock;
    Function *fun = INSTR_FUNCTION(instr);
    int size = (fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE;

    if (!BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	return;
    }

    loc = (fun->f_callerFrameSize / ADDRESS_SIZE) - (loc / ADDRESS_SIZE);

    if ((loc < 0) || (loc > size)) {
	return;
    }

    bbl->b_stackshot[loc].v_flags |= SLOT_KNOWN_FLAG;

    newLoc = zalloc(sizeof(Known));
    newLoc->kn_isRegister = false;
    newLoc->kn_value = loc;
    newLoc->kn_instr = instr;
    AddKnownLocationToList(newLoc);

    return;
}



/*
 *  ComputePotentialAtInstr(instr) --
 *
 *  Assumes locations have been loaded with AddKnownRegister() and
 *  AddKnownStackSlot(), and computes the potential starting at 'instr'.
 */
EDGE_WEIGHT_TYPE ComputePotentialAtInstr(Instruction *instr) {
    return ComputePotential(instr);
}



/*
 *  ComputeRegisterPotential(aReg,instr) --
 *
 *  Returns the benefit associated with knowing the constant value
 *  in register 'aReg' before the supplied instruction.
 *  The benefit is computed transitively.
 */
EDGE_WEIGHT_TYPE ComputeRegisterPotential(int aReg, Instruction *instr) {

    AddKnownRegister(aReg, instr);

    return ComputePotentialAtInstr(instr);
}



/*
 *  ComputeStackPotential(loc,instr) --
 *
 *  Returns the benefit associated with knowing the constant value
 *  in the stack location 'loc' before the supplied instruction.
 *  The benefit is computed transitively.
 */
EDGE_WEIGHT_TYPE ComputeStackPotential(int loc, Instruction *instr) {

    AddKnownStackSlot(loc, instr);

    return ComputePotentialAtInstr(instr);
}



/*
 *  ComputePotential(instr) --
 *
 *  Assumes that slots have been loaded with the SLOT_KNOWN_FLAG.
 *  Starting at 'instr', the SLOT_KNOWN_FLAG is propagated along the CFG
 *  and the potential benefit from being able to evaluate instructions
 *  is returned.
 */
EDGE_WEIGHT_TYPE ComputePotential(Instruction *instr) {
    Block *biterator;
    Instruction *iterator;
    EdgeNode *predecessors;
    Known *kiterator;
    EdgeNode *successors;
    bool change = true;
    EDGE_WEIGHT_TYPE potential = 0;

    /*
     * Initially, set all locations to be undefined (TOP),
     * and set only the block containing 'instr' to be evaluated.
     * Other blocks will be set as the known values change.
     */

    FOREACH_BBL_IN_FUN(INSTR_FUNCTION(instr), biterator) {
	BBL_REMOVE_FLAG(biterator, BBL_TEMP_FLAG);
	InitializeBBLFlags(biterator);
	FOREACH_INSTR_IN_BBL(biterator, iterator) {
	    INSTR_REMOVE_FLAG(iterator, INSTR_ELIMINATABLE_FLAG);
	}
    }
    BBL_SET_FLAG(instr->i_Bblock, BBL_TEMP_FLAG);

    /*
     * Begin the fixpoint computation. For all marked blocks,
     * propagate information through and update known values.
     * If things change, mark all successors to be visited
     * on the next iteration so they receive the updated information.
     */

    while (change) {
	change = false;
	FOREACH_BBL_IN_FUN(INSTR_FUNCTION(instr), biterator) {
	    if (BBL_HAS_FLAG(biterator, BBL_TEMP_FLAG)) {
		bool BBLChange = EvaluateBBL(biterator->b_first);
		BBL_REMOVE_FLAG(biterator, BBL_TEMP_FLAG);
		if (BBLChange == true) {
		    FOREACH_SUCC_IN_BBL(biterator, successors) {
			BBL_SET_FLAG(NODE_TO(successors), BBL_TEMP_FLAG);
		    }
		    change = true;
		}
	    }
	}			/* end iteration over blocks in function */
    }				/* end fixpoint iteration */

    /*
     * If the function returns a value, and the register is live at the end
     * of the function, we will have incorrectly marked the return value
     * assignment as eliminable. This piece of code assumes liveness analysis
     * has been carried out, and it unmarks the assignments to the return
     * value register.
     */

    FOREACH_PRED_IN_BBL(INSTR_FUNCTION(instr)->f_exit, predecessors) {

	FOREACH_BBL_IN_FUN(INSTR_FUNCTION(instr), biterator) {
	    BBL_REMOVE_FLAG(biterator, BBL_MARKED_FLAG);
	}

	biterator = NODE_FROM(predecessors);

	if (biterator != NULL) {
	    if (RegisterIsLiveInBitvector(RETURN_REGISTER,
		    biterator->b_liveOut)) {
		ReachableDefinitionsOfRegister(biterator->b_last,
		    RETURN_REGISTER, EXCLUSIVE, true);
	    }
	}

    }

    /*
     * Some number (maybe 0) of instructions is now marked as eliminatable.
     * We should sum up their weights and figure out the total expected savings.
     */

    FOREACH_BBL_IN_FUN(INSTR_FUNCTION(instr), biterator) {
	FOREACH_INSTR_IN_BBL(biterator, iterator) {
	    if (INSTR_HAS_FLAG(iterator, INSTR_ELIMINATABLE_FLAG)) {
		potential = potential + iterator->i_Bblock->b_weight;
	    }
	}
    }

    /* Deallocate list of known (value,program point) pairs before returning */
    kiterator = knownList;
    while (kiterator != NULL) {
	Known *next = kiterator->kn_next;
	free(kiterator);
	kiterator = next;
    }
    knownList = NULL;

    return potential;
}



/*
 *  EvaluateBBL(instr) --
 *
 *  Iterates over all the instructions in the basic block and propagates the
 *  SLOT_KNOWN_FLAG around. If 'mergePreds' is true, then first the snapshot is
 *  constructed by merging the known locations in all predecessor blocks.
 *  Returns true if something in the BBL changed.
 */
static bool EvaluateBBL(Instruction *instr) {
    Instruction *iterator;
    bool result;
    Value *regsCopy = NULL;
    Value *stackCopy = NULL;
    REGISTER_BITVECTOR currentPSW = 0;
    Block *bbl = instr->i_Bblock;

    if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
	regsCopy = BBLCopyRegisterFlags(bbl);
    }

    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	stackCopy = BBLCopyStackFlags(bbl);
    }

    MergePredecessors(instr->i_Bblock);

    /*
     * Set the PSW so that no bits are known.
     * See comment at the top of this file for details
     */
    currentPSW = 0;

    /*
     * Now that all known values have been merged, we can walk through the
     * instructions in this block and determine which are eliminatable.
     * At the end we will have constructed a new snapshot of the known
     * values at the end of the block.
     */

    iterator = instr;
    while (true) {
	Known *kiterator = knownList;

	while (kiterator != NULL) {

	    /*
	     * If the instruction being evaluated was one loaded with a
	     * known value, we will set the slot to known regardless
	     * of what flag may have been set coming in.
	     */

	    if (kiterator->kn_instr == iterator) {
		if ((kiterator->kn_isRegister)
		    && BBL_HAS_FLAG(iterator->i_Bblock, BBL_REGSHOT_FLAG)) {
		    int regIndex = RegisterToIndex(kiterator->kn_value);
		    iterator->i_Bblock->b_regshot[regIndex].v_flags &=
			~SLOT_UNKNOWN_FLAG;
		    iterator->i_Bblock->b_regshot[regIndex].v_flags &=
			~SLOT_UNDEFINED_FLAG;
		    iterator->i_Bblock->b_regshot[regIndex].v_flags |=
			SLOT_KNOWN_FLAG;
		} else if (BBL_HAS_FLAG(iterator->i_Bblock,BBL_STACKSHOT_FLAG)){
		    int sindex = kiterator->kn_value;
		    iterator->i_Bblock->b_stackshot[sindex].v_flags &=
			~SLOT_UNKNOWN_FLAG;
		    iterator->i_Bblock->b_stackshot[sindex].v_flags &=
			~SLOT_UNDEFINED_FLAG;
		    iterator->i_Bblock->b_stackshot[sindex].v_flags |=
			SLOT_KNOWN_FLAG;
		}
	    }
	    kiterator = kiterator->kn_next;
	}

	/*
	 * Check if we have an instruction that uses a known operand,
	 * but cannot be simplified.  If this is the case,
	 * we'll have to do some more discrediting.
	 */

	if ((iterator->i_src)
	    && !OperandIsKnown(iterator, iterator->i_src)) {
	    if (InstrUsesDest(iterator)
		&& OperandIsKnown(iterator, iterator->i_dest))
		RemovePotentialForReachableDefinitions(iterator,
		    iterator->i_dest);
	}

	if (isJumpClass(iterator) && (PSWIsKnown(iterator, currentPSW))
		&& isConditionalInstr(iterator)) {
	    goto ifKnown;
	} else if (InstrPopBytes(iterator) == ADDRESS_SIZE) {
	    UpdateNotKnown(iterator, &currentPSW);
	} else if (iterator->i_src
		&& !OperandIsKnown(iterator, iterator->i_src)) {
	    UpdateNotKnown(iterator, &currentPSW);
	} else if (iterator->i_src2
		&& !OperandIsKnown(iterator, iterator->i_src2)) {
	    UpdateNotKnown(iterator, &currentPSW);
	} else if (iterator->i_dest && InstrUsesDest(iterator)
		&& !OperandIsKnown(iterator, iterator->i_dest)) {
	    UpdateNotKnown(iterator, &currentPSW);
	} else if (!PSWIsKnown(iterator, currentPSW)) {
	    UpdateNotKnown(iterator, &currentPSW);

	/*
	 * Finally, we need to make sure that the location in question has only
	 * one reachable definition. If it has two, then we probably won't be
	 * able to remove this instruction. For instance, consider the
	 * situation:
	 *
	 *                r1 <- 3       r1 <- 10
	 *                   \            /
	 *                    \          /
	 *                    r2 = r1 + 1
	 *
	 * Just because r1 is known at the add instruction doesn't mean we can
	 * eliminate it.
	 */

	/*
	 * So, if it happens that there are multiple reaching definitions, then
	 * we will stop the propagation at this point, and discredit the
	 * benefit that was given to the definitions (since the instruction
	 * that uses their value has multiple reaching definitions, and hence
	 * could not be eliminated in a liveness/dead code analysis).
	 */

	} else if (iterator->i_src
	&& (ReachableDefinitions(iterator, iterator->i_src, false) > 1)) {
	    RemovePotentialForReachableDefinitions(iterator, iterator->i_src);
	} else if (iterator->i_src2
	&& (ReachableDefinitions(iterator, iterator->i_src2, false) > 1)) {
	    RemovePotentialForReachableDefinitions(iterator, iterator->i_src2);
	} else if (iterator->i_dest && InstrUsesDest(iterator)
	&& (ReachableDefinitions(iterator, iterator->i_dest, false) > 1)) {
	    RemovePotentialForReachableDefinitions(iterator, iterator->i_dest);
	} else if ((INSTR_FUNCTION(iterator)->f_last == iterator->i_Bblock)
	&& (InstrPopBytes(iterator) == ADDRESS_SIZE)) {
	    UpdateNotKnown(iterator, &currentPSW);

	} else {

	    /*
	     * Otherwise, we're pretty sure that the instruction can be
	     * evaluated and eliminated.
	     */
ifKnown:
	    UpdateKnown(iterator, &currentPSW);
	}

	iterator = iterator->i_next;
	if ((iterator == NULL) || (iterator->i_Bblock != bbl)) {
	    break;
	}
    }

    result = false;

    if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
	if (!BBLRegFlagsAreEqual(regsCopy, bbl)) {
	    result = true;
	}
	free(regsCopy);
    }

    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	if (!BBLStackFlagsAreEqual(stackCopy, bbl)) {
	    result = true;
	}
	free(stackCopy);
    }

    return result;
}



/*
 *  UpdateNotKnown(instr,currentPSW) --
 *
 *  Removes the SLOT_KNOWN_FLAGS for the destination of the given instruction,
 *  and returns a psw bitvector with the appropriate flags set for the bits we
 *  still know after executing the instruction.
 */
static void UpdateNotKnown(Instruction *instr, PSW_BITVECTOR *currentPSW) {
    REGISTER_BITVECTOR definedRegs;
    int i;
    Block *bbl = instr->i_Bblock;
    Function *fun = INSTR_FUNCTION(instr);
    int size = (fun->f_stackFrameSize + fun->f_callerFrameSize) / ADDRESS_SIZE;
    int definedSlots[MAXIMUM_AR_SIZE * 2];

    for (i = 0; i <= size; i++) {
	definedSlots[i] = false;
    }

    ComputeStackDefined(instr, definedSlots,
	fun->f_callerFrameSize / ADDRESS_SIZE, size);
    definedRegs = ComputeRegistersDefinedLiberal(instr);

    /*
     * Set the registers defined by this instruction to contain BOT. If any
     * flags change, return true.
     */

    if (BBL_HAS_FLAG(instr->i_Bblock, BBL_REGSHOT_FLAG)) {
	for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	    if ((definedRegs & (1 << i))) {
		int regshotIndex = RegisterToIndex(BitRegisterToRegister(i));
		bbl->b_regshot[regshotIndex].v_flags &= ~SLOT_KNOWN_FLAG;
		bbl->b_regshot[regshotIndex].v_flags &= ~SLOT_UNDEFINED_FLAG;
		bbl->b_regshot[regshotIndex].v_flags |= SLOT_UNKNOWN_FLAG;
	    }
	}
    }

    /*
     * Set the stack locations defined by this instruction to contain BOT. If
     * any flags change, return true.
     */

    if (BBL_HAS_FLAG(instr->i_Bblock, BBL_STACKSHOT_FLAG)) {
	for (i = 0; i <= size; i++) {
	    if (definedSlots[i] == true) {
		bbl->b_stackshot[i].v_flags &= ~SLOT_KNOWN_FLAG;
		bbl->b_stackshot[i].v_flags &= ~SLOT_UNDEFINED_FLAG;
		bbl->b_stackshot[i].v_flags |= SLOT_UNKNOWN_FLAG;
	    }
	}
    }

    if (INSTR_HAS_FLAG(instr, INSTR_ELIMINATABLE_FLAG)) {
	INSTR_REMOVE_FLAG(instr, INSTR_ELIMINATABLE_FLAG);
    }

    *currentPSW = *currentPSW & (~(ComputePSWDefinedLiberal(instr)));

    return;
}



/*
 *  UpdateKnown(instr,currentPSW) --
 *
 *  Updates all locations defined by 'instr' to contain the SLOT_KNOWN_FLAG.
 *  The weight of the instruction is also added to 'potential', and the known
 *  psw bits are returned. We assume that since the instruction can be evaluated
 *  then the PSW bits are known; this is not always the case though, especially
 *  for things like the overflow and carry flags, which depend entirely on the
 *  sign and size of the quantities being operated on.
 */
static void UpdateKnown(Instruction *instr, PSW_BITVECTOR *currentPSW) {
    REGISTER_BITVECTOR definedRegs;
    int i;
    Block *bbl = instr->i_Bblock;
    Function *fun = INSTR_FUNCTION(instr);
    int size = (fun->f_stackFrameSize + fun->f_callerFrameSize) / ADDRESS_SIZE;
    int definedSlots[MAXIMUM_AR_SIZE];

    for (i = 0; i <= size; i++) {
	definedSlots[i] = false;
    }

    ComputeStackDefined(instr, definedSlots,
	fun->f_callerFrameSize / ADDRESS_SIZE, size);
    definedRegs = ComputeRegistersDefined(instr);

    /*
     * Set all registers defined by 'instr' to contain the KNOWN flag. Returns
     * true if some flags changed after doing so.
     */

    if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
	for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	    if ((definedRegs & (1 << i))) {
		int regshotIndex = RegisterToIndex(BitRegisterToRegister(i));
		if (RegisterIsStackPointer(BitRegisterToRegister(i))) {
		    bbl->b_regshot[regshotIndex].v_flags |= SLOT_UNKNOWN_FLAG;
		    bbl->b_regshot[regshotIndex].v_flags &=~SLOT_UNDEFINED_FLAG;
		    bbl->b_regshot[regshotIndex].v_flags &=~SLOT_KNOWN_FLAG;
		} else {
		    bbl->b_regshot[regshotIndex].v_flags &=~SLOT_UNKNOWN_FLAG;
		    bbl->b_regshot[regshotIndex].v_flags &=~SLOT_UNDEFINED_FLAG;
		    bbl->b_regshot[regshotIndex].v_flags |= SLOT_KNOWN_FLAG;
		}
	    }
	}			/* end iteration over GP registers */
    }				/* end test for containing register snapshot */
    /*
     * Set all stack locations defined by 'instr' to contain the KNOWN flag.
     * Returns true if some flags changed after doing so.
     */
    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	for (i = 0; i <= size; i++) {
	    if (definedSlots[i] == true) {
		bbl->b_stackshot[i].v_flags &= ~SLOT_UNKNOWN_FLAG;
		bbl->b_stackshot[i].v_flags &= ~SLOT_UNDEFINED_FLAG;
		bbl->b_stackshot[i].v_flags |= SLOT_KNOWN_FLAG;
	    }
	}
    }

    /*
     * If the instruction has no destination or defines some ambiguous memory
     * location, we can just return. A caveat is a jump instruction, which can
     * still be eliminated by the constant propagator if the PSW bits are known.
     * A check for this is performed.
     */

    *currentPSW = *currentPSW | ComputePSWDefined(instr);

    /* Disallow stores to memory to be marked as eliminable */
    if ((instr->i_dest) && OperandIsMemory(instr->i_dest)) {
	return;
    }

    /* If the instruction is a NOP or something weird that sets PSW bits,
     * don't let it be marked as eliminable */
    if (!instr->i_dest && !instr->i_src) {
	return;
    }

    /* Pushes are stores to memory, and we cannot eliminate them either */
    if (InstrPushBytes(instr) > 0) {
	return;
    }

    INSTR_SET_FLAG(instr, INSTR_ELIMINATABLE_FLAG);

    return;
}



/*
 *  OperandIsKnown(instr,op) --
 *
 *  Returns true if the location(s) used by the supplied operand have the
 *  SLOT_KNOWN_FLAG set. If any location used by the operand does not have
 *  this flag set, false is returned.
 */
static bool OperandIsKnown(Instruction *instr, Operand *op) {
    Block *bbl = instr->i_Bblock;
    Function *fun = INSTR_FUNCTION(instr);

    if (OperandIsImmediate(op) && !OperandIsAddress(op)) {
	return true;
    }

    if (OperandIsGPRegister(op)) {
	int regIndex = RegisterToIndex(op->a_base_reg);
	if (SLOT_HAS_FLAG_DIRECT(bbl->b_regshot[regIndex], SLOT_KNOWN_FLAG)) {
	    return true;
	} else {
	    return false;
	}
    }

    /*
     * Addendum: For the SIB addressing mode, the operand can be evaluated
     * by the constant propagator if the instruction is an LEA and both
     * index and base regs are known. This is captured by the following
     * code.
     */

    if (isLeaClass(instr)) {
	int baseIndex, indexIndex;
	int baseReg = op->a_base_reg;
	int indexReg = op->a_index_reg;

	if (!BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
	    return false;
	}

	if (baseReg != 0) {
	    baseIndex = RegisterToIndex(baseReg);
	    if (!SLOT_HAS_FLAG_DIRECT(bbl->b_regshot[baseIndex],
		    SLOT_KNOWN_FLAG)) {
		return false;
	    }
	}

	if (indexReg != 0) {
	    indexIndex = RegisterToIndex(indexReg);

	    if (!SLOT_HAS_FLAG_DIRECT(bbl->b_regshot[indexIndex],
		    SLOT_KNOWN_FLAG)) {
		return false;
	    }
	}

	return true;
    }

    if (OperandAccessesStackFromFramePointer(op)) {
	int size =
	    (fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE;
	int base = fun->f_callerFrameSize / ADDRESS_SIZE;
	int slotIndex = base - (op->a_displacement / ADDRESS_SIZE);

	if (slotIndex < 0 || slotIndex > size) {
	    return false;
	}

	if (!BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	    return false;
	}

	if (SLOT_HAS_FLAG_DIRECT(bbl->b_stackshot[slotIndex],SLOT_KNOWN_FLAG)) {
	    return true;
	} else {
	    return false;
	}
    }

    return false;
}



/*
 *  PSWIsKnown(instr,currentPSW) --
 *
 *  Returns true if all bits that 'instr' uses in the program status word are
 *  set in 'currentPSW'.
 */
static bool PSWIsKnown(Instruction *instr, PSW_BITVECTOR currentPSW) {
    PSW_BITVECTOR result = ComputePSWUsed(instr) & (~currentPSW);

    if (result != 0) {
	return false;
    }

    return true;
}



/*
 *  RemovePotentialForReachableDefinitions(instr,op) --
 *
 *  This function takes back credit previously awarded to instructions defining
 *  locations used by 'op'.
 */
static void RemovePotentialForReachableDefinitions(Instruction *instr,
    Operand *op) {

    ReachableDefinitions(instr, op, true);

    return;
}



/*
 *  ReachableDefinitions(instr,op,discredit) --
 *
 *  Returns the number of reachable definitions that define the location
 *  'op' uses.
 */
static EDGE_WEIGHT_TYPE ReachableDefinitions(Instruction *instr, Operand *op,
    bool discredit) {
    Block *biterator;

    FOREACH_BBL_IN_FUN(INSTR_FUNCTION(instr), biterator) {
	BBL_REMOVE_FLAG(biterator, BBL_MARKED_FLAG);
    }

    if (OperandIsMemory(op)) {
	return ReachableDefinitionsOfStack(
	    instr, op->a_displacement, EXCLUSIVE, discredit);
    } else {
	return ReachableDefinitionsOfRegister(
	    instr, op->a_base_reg, EXCLUSIVE, discredit);
    }

    return -1;
}



/*
 *  ReachableDefinitionsOfRegister(instr,reg,includeInstr,discredit) --
 *
 *  Returns the number of reachable definitions of register 'reg'
 *  coming into 'instr'.
 */
static EDGE_WEIGHT_TYPE ReachableDefinitionsOfRegister(
Instruction *instr, int reg, int includeInstr, bool discredit) {

    Instruction *iterator;
    EdgeNode *preds;
    iterator = instr;

    /* We will only visit each basic block once.
     * A visited BBL will have its BBL_MARKED_FLAG flag set. */

    if (!BBL_HAS_FLAG(instr->i_Bblock, BBL_REGSHOT_FLAG)) {
	return 0;
    }

    BBL_SET_FLAG(instr->i_Bblock, BBL_MARKED_FLAG);

    while (true) {

	/*
	 * If the instruction defines the register we're interested in, then
	 * return 1--signaling that we have 1 reachable definition.
	 */

	if ((iterator != instr) || (includeInstr == INCLUSIVE)) {
	    if (InstrDefinesReg(iterator, reg)) {
		if (discredit) {
		    if (INSTR_HAS_FLAG(iterator, INSTR_ELIMINATABLE_FLAG)) {
			INSTR_REMOVE_FLAG(iterator, INSTR_ELIMINATABLE_FLAG);
			    return 1;
		    } else {
			return 1;
		    }
		} else {
		    return 1;
		}
	    }
	}

	/*
	 * If the instruction does not define the register we're interested in,
	 * then we'll walk backward through the basic block. If we cannot walk
	 * backward any further, then we must call this routine recursively on
	 * all predecessors of the block, and return the sum of their results.
	 */

	if (iterator->i_Bblock->b_first == iterator) {
	    EDGE_WEIGHT_TYPE sum = 0;
	    FOREACH_PRED_IN_BBL(iterator->i_Bblock, preds) {
		if (isInterproceduralEdge(preds->en_edge)) {
		    continue;
		}
		if (BBL_HAS_FLAG(NODE_FROM(preds), BBL_MARKED_FLAG)) {
		    continue;
		}
		if (((NODE_FROM(preds))->b_last) == NULL) {
		    continue;
		}
		sum = sum + ReachableDefinitionsOfRegister(
		    NODE_FROM(preds)->b_last, reg, INCLUSIVE, discredit);
	    }
	    return sum;
	}
	iterator = iterator->i_prev;
    }
}



/*
 *  ReachableDefinitionsOfStack(instr,loc,includeInstr,discredit) --
 *
 *  Returns the number of reachable definitions of the stack slot 'loc'
 *  coming into 'instr'.  In general, the stack analysis is too conservative
 *  when dealing with indirect loads and stores; most of the time these
 *  loads/stores do not overwrite locations on the stack.  The correctness
 *  of this module is affected it we overestimate the benefit is a few cases;
 *  for this reason, the function is not conservative in its treatment of
 *  stores to locations that may point to anywhere.
 */
static EDGE_WEIGHT_TYPE ReachableDefinitionsOfStack(
Instruction *instr, int loc, int includeInstr, bool discredit) {
    Instruction *iterator;
    EdgeNode *preds;
    int i;
    int funSize = INSTR_FUNCTION(instr)->f_stackFrameSize / ADDRESS_SIZE;
    int callerSize = INSTR_FUNCTION(instr)->f_callerFrameSize / ADDRESS_SIZE;
    int totalSize = funSize + callerSize;
    int locIndex = (loc / ADDRESS_SIZE) + callerSize;

    if (!BBL_HAS_FLAG(instr->i_Bblock, BBL_STACKSHOT_FLAG)) {
	return 0;
    }

    for (i = 0; i <= totalSize; i++) {
	definedSlotsPot[i] = false;
    }

    iterator = instr;

    /* We will only visit each basic block once.
     * A visited BBL will have its BBL_MARKED_FLAG flag set. */

    BBL_SET_FLAG(instr->i_Bblock, BBL_MARKED_FLAG);

    while (true) {

	/*
	 * If the instruction defines the location we're interested in, then
	 * return 1--signaling that we have 1 reachable definition.
	 */

	ComputeStackDefined(iterator, definedSlotsPot, callerSize, totalSize);

	if ((iterator != instr) || (includeInstr == INCLUSIVE)) {
	    if ((locIndex > 0) && (locIndex <= totalSize)) {
		if (definedSlotsPot[locIndex] == true) {
		    if (discredit) {
			if (INSTR_HAS_FLAG(iterator, INSTR_ELIMINATABLE_FLAG)) {
			    INSTR_REMOVE_FLAG(iterator,
				INSTR_ELIMINATABLE_FLAG);
				return 1;
			} else {
			    return 1;
			}
		    } else {
			return 1;
		    }
		}
	    }
	}

	/*
	 * If the instruction does not define the register we're interested in,
	 * then we'll walk backward through the basic block. If we cannot walk
	 * backward any further, then we must call this routine recursively on
	 * all predecessors of the block, and return the sum of their results.
	 */

	if (iterator->i_Bblock->b_first == iterator) {
	    EDGE_WEIGHT_TYPE sum = 0;
	    FOREACH_PRED_IN_BBL(iterator->i_Bblock, preds) {
		if (isInterproceduralEdge(preds->en_edge)) {
		    continue;
		}
		if (BBL_HAS_FLAG(NODE_FROM(preds), BBL_MARKED_FLAG)) {
		    continue;
		}
		if (((NODE_FROM(preds))->b_last) == NULL) {
		    continue;
		}
		sum = sum + ReachableDefinitionsOfStack(
		    NODE_FROM(preds)->b_last, loc, INCLUSIVE, discredit);
	    }
	    return sum;
	}
	iterator = iterator->i_prev;
    }
}



/*
 *  InitializeBBLFlags(bbl) --
 *
 *  Sets all slot flags to be undefined in 'bbl'
 */
static void InitializeBBLFlags(Block *bbl) {
    int i;
    Function *fun;
    int size;

    fun = bbl->b_fun;
    size = (fun->f_callerFrameSize + fun->f_stackFrameSize) / ADDRESS_SIZE;

    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	for (i = 0; i <= size; i++) {
	    bbl->b_stackshot[i].v_flags &= ~SLOT_KNOWN_FLAG;
	    bbl->b_stackshot[i].v_flags &= ~SLOT_UNKNOWN_FLAG;
	    bbl->b_stackshot[i].v_flags |= SLOT_UNDEFINED_FLAG;
	}
    }

    if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
	for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	    bbl->b_regshot[i].v_flags &= ~SLOT_KNOWN_FLAG;
	    bbl->b_regshot[i].v_flags &= ~SLOT_UNKNOWN_FLAG;
	    bbl->b_regshot[i].v_flags |= SLOT_UNDEFINED_FLAG;
	}
    }

    return;
}



/*
 *  AddKnownLocationToList(newLoc) --
 */
static void AddKnownLocationToList(Known *newLoc) {

    if (knownList == NULL) {
	knownList = newLoc;
	return;
    }

    knownList->kn_next = newLoc;
    newLoc->kn_prev = knownList;
    knownList = newLoc;

    return;
}



/*
 *  MergePredecessors(bbl) --
 *
 *  MergePredecessors() performs a join at the top of 'bbl' with all its
 *  predecessors.  Constants/registers are propagated through, and the
 *  basic block's snapshots are updated to reflect its view of the world
 *  before the first instruction in the block is executed.
 *
 *  Sets 'bbl's stack and register snapshots to the result attained
 *  from merging its predecessors
 */
static void MergePredecessors(Block *bbl) {
    int numberKnown, numberBot, numberTop, i;
    EdgeNode *predecessors;
    Function *fn = bbl->b_fun;
    int fnsize = (fn->f_stackFrameSize + fn->f_callerFrameSize) / ADDRESS_SIZE;

    numberKnown = numberBot = numberTop = 0;

    /*
     * Construct the known locations at the start of this basic block by merging
     * the known locations at the end of all predecessors blocks. This first
     * loop merges the stack locations. meet(TOP,KNOWN) = KNOWN.
     * meet(TOP, BOT) = BOT. meet(TOP, TOP) = TOP. meet(KNOWN, BOT) = BOT.
     * meet(KNOWN, KNOWN) = KNOWN. meet(BOT, BOT) = BOT.
     */

    if (BBL_HAS_FLAG(bbl, BBL_STACKSHOT_FLAG)) {
	for (i = 0; i <= fnsize; i++) {

	    FOREACH_PRED_IN_BBL(bbl, predecessors) {
		Block *from = NODE_FROM(predecessors);

		if (from->b_type == BT_ENTRY) {
		    continue;
		}
		if (from->b_type == BT_EXIT) {
		    continue;
		}

		/*
		 * count up the number of TOP, BOT, and KNOWN predecessors.
		 */

		if (from->b_fun != bbl->b_fun) {
		    if (INTERPROCEDURAL_CONSERVATIVE) {
			numberBot++;
		    } else {
			numberTop++;
		    }
		} else if (!BBL_HAS_FLAG(from, BBL_STACKSHOT_FLAG)) {
		    numberBot++;
		} else if (SLOT_HAS_FLAG_DIRECT((from->b_stackshot[i]),
			SLOT_KNOWN_FLAG)) {
		    numberKnown++;
		} else if (SLOT_HAS_FLAG_DIRECT((from->b_stackshot[i]),
			SLOT_UNKNOWN_FLAG)) {
		    numberBot++;
		} else if (SLOT_HAS_FLAG_DIRECT((from->b_stackshot[i]),
			SLOT_UNDEFINED_FLAG)) {
		    numberTop++;
		} else {
		    WARNING("BBL %d location %d has no slot flag.",
			from->b_id, i);
		}

	    }			/* end iteration over predecessors */

	    /* Clear all the flags for this location. */
	    bbl->b_stackshot[i].v_flags &= ~SLOT_KNOWN_FLAG;
	    bbl->b_stackshot[i].v_flags &= ~SLOT_UNKNOWN_FLAG;
	    bbl->b_stackshot[i].v_flags &= ~SLOT_UNDEFINED_FLAG;

	    /* Set new flags for this location based on rules discussed above */
	    if (numberBot > 0) {
		bbl->b_stackshot[i].v_flags |= SLOT_UNKNOWN_FLAG;
	    } else if (numberKnown == 0) {
		bbl->b_stackshot[i].v_flags |= SLOT_UNDEFINED_FLAG;
	    } else {
		bbl->b_stackshot[i].v_flags |= SLOT_KNOWN_FLAG;
	    }

	}			/* end iteration over each stack location */
    }				/* end merge of stack */
    /*
     * Construct the known locations at the start of this basic block by merging
     * the known locations at the end of all predecessors blocks. This loop
     * merges the registers, and has the same rules for meet as discussed above.
     */
    numberKnown = numberBot = numberTop = 0;

    if (BBL_HAS_FLAG(bbl, BBL_REGSHOT_FLAG)) {
	for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {

	    FOREACH_PRED_IN_BBL(bbl, predecessors) {
		Block *from = NODE_FROM(predecessors);

		if (from->b_type == BT_ENTRY) {
		    continue;
		}
		if (from->b_type == BT_EXIT) {
		    continue;
		}

		/*
		 * count up the number of TOP, BOT, and KNOWN predecessors.
		 */

		if (from->b_fun != bbl->b_fun) {
		    if (INTERPROCEDURAL_CONSERVATIVE) {
			numberBot++;
		    } else {
			numberTop++;
		    }
		} else if (!BBL_HAS_FLAG(from, BBL_REGSHOT_FLAG)) {
		    numberBot++;
		} else if (SLOT_HAS_FLAG_DIRECT((from->b_regshot[i]),
			SLOT_KNOWN_FLAG)) {
		    numberKnown++;
		} else if (SLOT_HAS_FLAG_DIRECT((from->b_regshot[i]),
			SLOT_UNKNOWN_FLAG)) {
		    numberBot++;
		} else if (SLOT_HAS_FLAG_DIRECT((from->b_regshot[i]),
			SLOT_UNDEFINED_FLAG)) {
		    numberTop++;
		} else {
		    WARNING("BBL %d register %d has no slot flag.",
			from->b_id, i);
		}

	    }				/* end iteration over predecessors */

	    /* Clear all the flags for this location. */
	    bbl->b_regshot[i].v_flags &= ~SLOT_KNOWN_FLAG;
	    bbl->b_regshot[i].v_flags &= ~SLOT_UNKNOWN_FLAG;
	    bbl->b_regshot[i].v_flags &= ~SLOT_UNDEFINED_FLAG;

	    /* Set new flags for this location based on rules discussed above */
	    if (numberBot > 0) {
		bbl->b_regshot[i].v_flags |= SLOT_UNKNOWN_FLAG;
	    } else if (numberKnown == 0) {
		bbl->b_regshot[i].v_flags |= SLOT_UNDEFINED_FLAG;
	    } else {
		bbl->b_regshot[i].v_flags |= SLOT_KNOWN_FLAG;
	    }

	}				/* end iteration over each register */
    }					/* end merge of registers */
    return;
}



/*
 *  BBLCopyStackFlags(bbl) --
 */
static Value *BBLCopyStackFlags(Block *bbl) {
    int sz = (bbl->b_fun->f_stackFrameSize +
	bbl->b_fun->f_callerFrameSize) / ADDRESS_SIZE;
    int i;

    Value *result = zalloc((sz + 1) * sizeof(Value));

    for (i = 0; i < sz; i++) {
	result[i].v_flags = bbl->b_stackshot[i].v_flags;
    }

    return result;
}



/*
 *  BBLCopyRegisterFlags(bbl) --
 */
static Value *BBLCopyRegisterFlags(Block *bbl) {
    int i;

    Value *result = zalloc(NUMBER_OF_GP_REGISTERS * sizeof(Value));

    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	result[i].v_flags = bbl->b_regshot[i].v_flags;
    }

    return result;
}



/*
 *  BBLRegFlagsAreEqual(regs,bbl) --
 */
static bool BBLRegFlagsAreEqual(Value *regs, Block *bbl) {
    int i;

    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	if (regs[i].v_flags != bbl->b_regshot[i].v_flags) {
	    return false;
	}
    }

    return true;
}



/*
 *  BBLStackFlagsAreEqual(stack,bbl) --
 */
static bool BBLStackFlagsAreEqual(Value *stack, Block *bbl) {
    int sz = (bbl->b_fun->f_stackFrameSize + bbl->b_fun->f_callerFrameSize) /
	ADDRESS_SIZE;
    int i;

    for (i = 0; i < sz; i++) {
	if (stack[i].v_flags != bbl->b_stackshot[i].v_flags) {
	    return false;
	}
    }

    return true;
}
