/*
 *  potential.h
 *
 *  For computing the potential of knowing registers and stack locations.
 *
 *  There are two ways to compute the potential. If you are only interested in
 *  the benefit for knowing ONE location, use Compute{Register,Stack}Potential()
 *  and supply the appropriate location. If you need more than one location,
 *  use AddKnown{Register,StackSlot} to set the appropriate flags, then use
 *  ComputePotentialAtInstr() to get the benefit.
 */
#ifndef POTENTIAL_H
#define POTENTIAL_H

/*
 *  In most optimizations we have to be conservative. The benefit analysis is
 *  slightly different, however, in that it only computers a number that is the
 *  expected savings. If this number is wrong, an optimization may make a poor
 *  decision, but incorrect code will not be produced. Given that we don't need
 *  to return the exact number all the time, we don't need to treat indirect
 *  memory stores conservatively. In particular, we don't have to assume that
 *  they overlap any memory region and destroy the contents of the stack.
 *  Also, since this analysis is intraprocedural, we also can assume that
 *  function calls obey standard calling conventions.
 *  INTERPROCEDURAL_CONSERVATIVE should be defined as true if one wishes
 *  function calls to be treated conservatively (assume they overwrite all
 *  registers and stack locations).
 *
 *  (Old comment said: INDIRECT_CONSERVATIVE should be defined
 *  as true if one wishes to treat indirect stores conservatively
 *  (e.g., assume they overlap any memory region.).
 *  However, INDIRECT_CONSERVATIVE is not referenced anywhere.)
 */
#define INTERPROCEDURAL_CONSERVATIVE      true

/*
 *  Returns the benefit (in terms of number of cycles) from knowing the value
 *  of register 'aReg' before executing the instruction given in 'instr'
 */
EDGE_WEIGHT_TYPE ComputeRegisterPotential(int aReg, Instruction *instr);

/*
 *  Returns the benefit (in terms of number of cycles) from knowing the value
 *  of a stack location at offset 'stackLoc' from the frame ptr before
 *  executing the instruction given in 'instr'
 */
EDGE_WEIGHT_TYPE ComputeStackPotential(int stackLoc, Instruction *instr);

/* Sets the flags in instr's basic block so that aReg is a known location */
void AddKnownRegister(int aReg, Instruction *instr);

/* Sets the flag in instr's basic block so that loc is a known stack location */
void AddKnownStackSlot(int loc, Instruction *instr);

/*
 *  Assuming some locations have been marked as known, this function computed
 *  the benefit for knowing those locations.
 */
EDGE_WEIGHT_TYPE ComputePotentialAtInstr(Instruction *instr);

#endif				// POTENTIAL_H
