/*
 *  print.c
 *
 *  A module for printing various structures
 */
#include "global.h"
#include "print.h"
#include "sysdep.h"
#include "flowgraph.h"
#include "utilities.h"



/*
 *  PrintRegBitvector(rbv) --
 *
 *
 *  PrintRegBitvector
 *  Prints out the name of the register given its position in the bit
 *  vector.
 */
void PrintRegBitvector(REGISTER_BITVECTOR rbv) {
    int i;

    for (i = 0; i < NUMBER_OF_GP_REGISTERS; i++) {
	printf("%s ", PrintRegister(rbv & (1 << i)));
    }
}



/*
 *  PrintBbl(ibbl) --
 *
 *  Prints out information about the basic block and all of its
 *  instructions.
 */
void PrintBbl(Block *ibbl) {
    EdgeNode *iedg;
    Block *jbbl;
    Instruction *iins;
    char bbltype = '?';

    CASSERT(ibbl != NULL, "NULL argument");

    switch (ibbl->b_type) {
    case BT_INVALID:
	bbltype = '#';
	break;
    case BT_NORMAL:
	bbltype = 'N';
	break;
    case BT_BRANCH:
	bbltype = 'B';
	break;
    case BT_UNCOND:
	bbltype = 'U';
	break;
    case BT_RETURN:
	bbltype = 'R';
	break;
    case BT_SWITCH:
	bbltype = 'S';
	break;
    case BT_HELL:
	bbltype = 'H';
	break;
    case BT_ENTRY:
	bbltype = 'e';
	break;
    case BT_EXIT:
	bbltype = 'E';
	break;
    case BT_CALL:
	bbltype = 'C';
	break;
    case BT_HALT:
	bbltype = 'h';
	break;
    case BT_SYSCALL:
	bbltype = 's';
	break;
    case BT_DATABLOCK:
	bbltype = 'D';
	break;
    default:
	bbltype = '?';
	break;
    }
    printf("BBL: %d   TYPE: %c   FUN: %d [ %s ]  COUNT: %lld\n", ibbl->b_id,
	bbltype, ibbl->b_fun->f_id, ibbl->b_fun->f_name, ibbl->b_weight);

    printf("PREDS:");
    FOREACH_PRED_IN_BBL(ibbl, iedg) {
	jbbl = NODE_FROM(iedg);
	printf(" %d [%s]", jbbl->b_id, edgeTypeToString(NODE_TYPE(iedg)));
	if (isInterproceduralEdge(iedg->en_edge)) {
	    printf("^");
	}
        printf("@%ld", (long) NODE_WEIGHT(iedg));
    }
    printf("\n");

    printf("SUCCS:");
    FOREACH_SUCC_IN_BBL(ibbl, iedg) {
	jbbl = NODE_TO(iedg);
	printf(" %d [%s]", jbbl->b_id, edgeTypeToString(NODE_TYPE(iedg)));
	if (isInterproceduralEdge(iedg->en_edge)) {
	    printf("^");
	}
        printf("@%ld", (long) NODE_WEIGHT(iedg));
    }
    printf("\n");

    printf("LiveIN:  ");
    PrintRegBitvector(ibbl->b_liveIn);
    printf("\n");
    printf("LiveOUT: ");
    PrintRegBitvector(ibbl->b_liveOut);
    printf("\n");

    FOREACH_INSTR_IN_BBL(ibbl, iins) {
	printf("  [%d]    [0x%8.8x -> 0x%8.8x] [%d bytes]    %s\n", 
	    iins->i_id,
	    iins->i_address,
	    iins->i_final_address, 
	    CalculateInstrLength(iins),
	    PrintInstruction(iins));
    }

    printf("-----------------------------------------------"
	"-----------------------\n");
}



/*
 *  ShowBbl(n) --
 *
 *  Displays information about the function with id 'n'.
 */
void ShowBbl(int n) {
    Block *ibbl;

    FOREACH_BBL(ibbl) {
	if (ibbl->b_id == n) {
	    PrintBbl(ibbl);
	    return;
	}
    }
}



/*
 *  ShowFun(n) --
 */
void ShowFun(int n) {
    Function *ifun;
    Block *ibbl;
    int k = 0;

    FOREACH_FUNCTION(ifun) {
	if (ifun->f_id != n) {
	    continue;
	}

	printf("############### FUN %d : %s ####################\n",
	    ifun->f_id, ifun->f_name);

	printf("BBLS:");
	FOREACH_BBL_IN_FUN(ifun, ibbl) {
	    k += printf(" %d", ibbl->b_id);
	    if (k > 70) {
		k = 0;
		printf("\n     ");
	    }
	}

	FOREACH_BBL_IN_FUN(ifun, ibbl) {
	    PrintBbl(ibbl);
	}

	break;
    }
}



/*
 *  PrintSymbol(symbol) --
 */
void PrintSymbol(Symbol *symbol) {
    printf("[Symbol| %s:%s %8x\n", GetSymbolSectionName(symbol),
	symbol->sy_name, symbol->sy_value);
}
