/*
 *  Routines for manipulating relocations and instructions lie within.
 *  The order in which these functions are applied to the instructions is
 *  quite important.  One may note the following:
 *
 *  Updating a relative offset for a PC changing instruction may change
 *  the instruction length.  This means instruction addresses will have to
 *  be recalculated periodically.  In addition, changing a relative
 *  displacement may alter the instruction in such a way that other
 *  relative displacements need to be changed. e.g: consider updating jmp
 *  0xd4 to jmp 0x0104.  All of a sudden the instruction uses an extra
 *  byte.  Any relative displacement somewhere else in the instruction
 *  stream that used the 1 byte representation now needs to be updated.
 *  And these updates could cause even more instructions to be
 *  out-of-sync!  Thus, a good strategy is to iterate over the
 *  instructions, updating their displacements, until none of them change.
 */
#include "global.h"
#include "relocate.h"
#include "utilities.h"
#include "flowgraph.h"
#include "linkage.h"
#include "sysdep.h"
#include "hash.h"
#include "upm.h"
#include "plto.h"
#include "elfio.h"
#include "disassemble.h"


static int callNumber;



/*
 *  MovFlistFunctionsToDataInstrs() -- converts all FLIST functions.
 *
 *  Converts all the functions in FLISTs to data instructions.
 */
void MovFlistFunctionsToDataInstrs() {
    FList *fliterator;

    FOREACH_FLIST(fliterator) {
	fliterator->fl_newStartAddress = fliterator->fl_startAddress;
	fliterator->fl_newEndAddress   = fliterator->fl_endAddress;
	CreateDataInstr(
	    fliterator->fl_startAddress - SECTION_REF_CODE->s_address,
	    fliterator->fl_endAddress - fliterator->fl_startAddress,
	    fliterator->fl_data);
    }
}


/* 
 * CreatePaddingByte (char) -- 
 * 
 * Adds 1 byte of padding whose contents are specified by the input parameter
 */
Instruction * CreatePaddingByte(const char content) {
    Instruction *new_instr = InstrCreate();
	
    INSTR_SET_FLAG(new_instr, INSTR_DATA_BYTE);
    INSTR_SET_FLAG(new_instr, INSTR_PADDING_BYTE); 	
    new_instr->i_length = 1;
    new_instr->i_literal_data = zalloc(sizeof(UINT8));
    memcpy(new_instr->i_literal_data, &content, 1);

    return new_instr;
}



/*
 *  UpdateInstrAddresses(sec) --
 *
 *  Goes through the list of instructions and recomputes their virtual
 *  memory addresses based upon where they will lie in the final binary.
 *  Returns the total length.
 */
int UpdateInstrAddresses(Section *sec) {
    Instruction *iterator;
    int starting_address = sec->s_newaddress;
    int offset = 0;

    FOREACH_INSTR_IN_SECTION(sec, iterator) {

	if (iterator->i_length <= 0) {
	    iterator->i_length = CalculateInstrLength(iterator);
	}

	if (iterator->i_length <= 0) {
	    ReportInstrError("Bad length", iterator);
	}

	iterator->i_final_address = starting_address + offset;

        if (INSTR_HAS_FLAG(iterator, INSTR_NEEDS_ALIGNMENT)) {
            Instruction *ref_instr = iterator->i_prev;
            while (iterator->i_final_address % iterator->i_alignment) {
                Instruction *nop_instr;

                if (INSTR_HAS_FLAG(ref_instr, INSTR_DATA_BYTE)) {
                    nop_instr = CreatePaddingByte(0x00);
                } else {
                    nop_instr = CreatePaddingByte(0x90);
                }
                //Instruction *nop_instr = InstrCreateNOP();
                InsertInstrAfterInstr(nop_instr, ref_instr);
                InstrInsertEntry(nop_instr);
                offset++;
                iterator->i_final_address++;
                if (ref_instr->i_Bblock) {
                    REPORT(RP_TRIVIAL,
                        "Added NOP to func (%s) for aligning DataInst to %x\n",
                        ref_instr->i_Bblock->b_fun->f_name,
                        iterator->i_final_address);
                } else {
                    REPORT(RP_TRIVIAL,
                        "Added NOP after instr %x - aligning DataInst to %x\n",
                        ref_instr->i_address,
                        iterator->i_final_address);
                }
                
            }
        }

        offset += iterator->i_length;
    }
    return offset;
}



/*
 *  CalculateJumpDisplacement(instr,offset_found) --
 *
 *  Given some instruction which is a PC relative jump, calculate what the
 *  new displacement for the jump is in the new executable.  If no new
 *  displacement was need we'll set offset_found to false, otherwise true.
 *
 *  Note that this calculates the displacement from the instruction, not the
 *  address of the target.
 */
static ADDRESS CalculateJumpDisplacement(
Instruction *instr, bool *offset_found) {
    ADDRESS offset = 0;
    EdgeNode *niterator;

    *offset_found = false;
    /*
     * First we look at all the edges and find out which one is going to take
     * us to the jump target.  ET_FUNCALL edges are a little tricky, because
     * functions all have entry basic blocks that in turn point to instructions
     * inside the function.  If we have an ET_FUNCALL edge then we look at the
     * entry block and find which edge points to the block we're looking for.
     */

    FOREACH_SUCC_IN_BBL(instr->i_Bblock, niterator) {
	if ((niterator->en_edge->e_type == ET_TRUE) ||
	    (niterator->en_edge->e_type == ET_UNCOND) ||
	    (niterator->en_edge->e_type == ET_FUNDIRECT) ||
	    (niterator->en_edge->e_type == ET_HELL)) {
	    if (*offset_found == true) {
		continue;
	    }

	    *offset_found = true;

	    /* Function call hell edges */
	    if (NODE_TYPE(niterator) == ET_HELL) {
		if ((instr->i_src) && OperandUsesRegister(instr->i_src)) {
		    offset = instr->i_src->a_displacement;
		} else if (!INSTR_HAS_FLAG(instr, INSTR_FLIST_FLAG)) {
		    // occasionally the static linker misses a
		    // symbol resolution or includes code that
		    // references something that doesn't
		    // actually need to be included (sometimes
		    // code required for dynamic linking is
		    // embedded in code that we still need,
		    // but the code it calls is unneeded) --
		    // calls to this kind of code based on an
		    // unresolved symbol end up being call 0
		    // instructions which would just die if
		    // ever executed but do not ever execute,
		    // so here we just set the offset that
		    // would be needed to make it call 0
		    offset = (instr->i_final_address + instr->i_length) * -1;
		} else {
		    // if we're calling an flist then this instruction will be
		    // handled using UPMs and we do not want to return an offset
		    *offset_found = false;
		    return -1;
		}

	    } else if (NODE_TYPE(niterator) == ET_FUNDIRECT) {

		// Function calls are an interesting case.  Since
		// we use a level of indirection to implement
		// function entry blocks, all function calls to a
		// function will have a FUNCALL edge to the same
		// block.  The entry block then attaches edges to
		// all possible function blocks where control flow
		// may enter. Thus, we cannot directly see the
		// target of a function call.  In fact, we have no
		// idea what the target is by just looking at
		// edges, since the function entry block may have
		// multiple edges leading into the function and we
		// don't know which one to use. Also, we cannot
		// use the displacement in the call instruction to
		// calculate the new address, because the
		// displacement is what we're going to change.
		// Once we change it, the next time through we
		// will use the new displacement and calculate a
		// bad offset.  It will go on forever like this.
		// Thus, a new edge type "ET_FUNDIRECT" is used to
		// link the calling block to the target block. We
		// look at the new address of the edge's
		// destination and simply make the translation.

		EdgeNode *found = NULL;

		found = niterator;

		if (found == NULL) {
		    ReportInstrError("Could not find ET_FUNDIRECT edge"
			" in a block with an ET_FUNCALL edge", instr);
		}

		if (NODE_TO(found)->b_first->i_final_address == 0) {
		    WARNING
			("Instruction not in final executable referenced: 0x%x",
			NODE_TO(found)->b_first->i_address);
		}

		offset = NODE_TO(found)->b_first->i_final_address -
		    (instr->i_final_address + instr->i_length);
		/* Non-function call edges */
	    } else {
		if (NODE_TO(niterator)->b_first->i_final_address == 0) {
		    WARNING
			("Instruction not in final executable referenced: 0x%x",
			NODE_TO(niterator)->b_first->i_final_address);
		}

		offset = NODE_TO(niterator)->b_first->i_final_address -
		    (instr->i_final_address + instr->i_length);
	    }
	}
    }

    return offset;
}



/*
 *  UpdatePCRelativeJumps(iiterator) --
 *
 *  Updates all PC relative jumps based on their addresses and the addresses
 *  of their successors. Note that at this point the code segment has not
 *  been written to a memory image (since the size of it may change still).
 *  Thus, we need to update operands instead of patching actual addresses.
 *  This means that we cannot use the displacements provided in the operands
 *  and then make old address -> new address translations.  This is mainly
 *  because multiple passes are made over the instructions until none of
 *  them change.  If we change the displacement in one pass, then we will
 *  use an already-updated displacement in the next pass and get bad
 *  translations. Setting a flag that the displacement has been changed is
 *  not sufficient either, because it may need to be changed twice if
 *  any bytemodes are updated in other instructions.
 */
bool UpdatePCRelativeJumps(Instruction *iiterator) {
    Instruction *iterator;
    ADDRESS offset;
    int numChanged;
    int roundNumber = 0;
    bool offset_found;
    bool changed = true;
    bool lengthChanged = false;
    int newLength, oldLength;
    offset = 0;

    REPORT(RP_SMLOOP, WITHFUNC("Updating relative jumps"));

    callNumber++;

    while (changed) {
	changed = false;
	numChanged = 0;
	roundNumber++;
	UpdateSectionSizes(true);
	iterator = iiterator;
	while (iterator != NULL) {
	    /* If a jump instruction with a relative displacement, update it */
	    if ((!isSyscallClass(iterator)) &&
		(isPCChangingClass(iterator)) &&
		(iterator->i_src != NULL) &&
		(OP_TYPE(iterator->i_src, OPERAND_DISP_R_FLAG)) &&
		(!OPERAND_HAS_FLAG(iterator->i_src, OPERAND_RELOCATED)) &&
		(!OPERAND_HAS_FLAG(iterator->i_src, 
			OPERAND_AUTOLENGTH_RELOCATED))) {

		offset = CalculateJumpDisplacement(iterator, &offset_found);

		/* Only update offset (and possibly bytemode) if it changed */
		if (offset_found && iterator->i_src->a_displacement != offset) {
		    DEBUG(DB_JUMPS,
		    "Adjusted displacement in 0x%8.8x from 0x%8.8x to 0x%8.8x",
			iterator->i_address, iterator->i_src->a_displacement,
			offset);
		    OP_DISP_R(iterator->i_src, offset);
		    oldLength = iterator->i_length;
		    UpdateInstr(iterator);
		    newLength = iterator->i_length;
		    if (oldLength != newLength) {
			lengthChanged = true;
		    }
		    changed = true;
		    numChanged++;
		}

	    }
	    iterator = iterator->i_next;
	}			/* end iteration over instructions */

	REPORT(RP_DETAIL, "Round %d.%d complete; %d instructions modified",
	    callNumber, roundNumber, numChanged);

    }				/* end updating PC relative displacements */

    return lengthChanged;
}



/*
 *  UpdateSectionSizes(useNew) --
 *
 *  Updates the section sizes and virtual memory addresses based on present
 *  knowledge.  Instruction addresses are also updated, if useNew is true.
 */
void UpdateSectionSizes(bool useNew) {
    Section *siterator;
    int location_counter;

    if (useNew) {
	location_counter = sec_list->s_newaddress;
    } else {
	location_counter = sec_list->s_address;
    }

    FOREACH_SECTION(siterator) {
	// .data sections need to go on their own page, there is also something
	// else that the linker does so we'll do it too
	/*
	 * ld's linker script has the following to say about this:
	 * Adjust the address for the data segment.  We want to adjust up to
	 * the same address within the page on the next page up.
	 * . = ALIGN (0x1000) - ((0x1000 - .) & (0x1000 - 1));
	 * . = DATA_SEGMENT_ALIGN (0x1000, 0x1000);
	 *
	 * macro defs:
	 * ALIGN(x) = (. + x - 1) & ~(x - 1)
	 * DATA_SEGMENT_ALIGN(x) = (ALIGN(x) + (. & (x - 1)))
	 */

#define ALIGN(x, loc) ((loc + x - 1) & ~(x - 1))
#define DATA_SEGMENT_ALIGN(x, loc) (ALIGN(x, loc) + (loc & (x - 1)))

        
        if (strstr(siterator->s_name, ".page_aligned")) {
            REPORT(RP_DETAIL, "Aligning %s at page boundary\n",
                siterator->s_name);
            location_counter = ALIGN(PAGE_SIZE, location_counter);
        } else if (!KERNEL_MODE && siterator == SECTION_REF_DATA) {
	    location_counter = ALIGN(PAGE_SIZE, location_counter) -
		((PAGE_SIZE - location_counter) & (PAGE_SIZE - 1));
	    location_counter = DATA_SEGMENT_ALIGN(PAGE_SIZE, location_counter);
	} else if (KERNEL_MODE && 
                    !strcmp(siterator->s_name, ".data.init_task")) {
            location_counter = ALIGN(PAGE_SIZE, location_counter); 
        } else {
	    location_counter = ALIGN(siterator->s_alignment, location_counter);
	} 

#undef ALIGN
#undef DATA_SEGMENT_ALIGN

	// if no space in the memory image should allocated for the section then
	// we shouldn't assign it a virtual address
	if (!SECTION_HAS_FLAG(siterator, SECTION_OCCUPIES_VM_FLAG)) {
	    continue;
	}

	siterator->s_newaddress = location_counter;

	// if we are updating new addresses we should also update section sizes
	if (useNew) {
	    if (SECTION_WAS_MERGED(siterator)) {
		siterator->s_newdatasize = UpdateInstrAddresses(siterator);
	    } else {
		siterator->s_newdatasize =
		    siterator->s_datasize + siterator->s_growth;
	    }
	    location_counter += siterator->s_newdatasize;
	} else {
	    siterator->s_address = location_counter;
	    location_counter += siterator->s_datasize;
	}
    }					// end section iteration

    return;
}



/*
 *  GetNextSection() --
 *
 *  Returns the lowest-addressed, unmarked, section in the linked-list of
 *  sections.
 */
Section *GetNextSection() {
    Section *iterator, *found;

    found = NULL;
    FOREACH_SECTION(iterator) {
	if (!SECTION_HAS_FLAG(iterator, SECTION_MARKED_FLAG)) {
	    if ((found == NULL) || (iterator->s_address < found->s_address)) {
		found = iterator;
	    }
	}
    }

    if (found) {
	SECTION_SET_FLAG(found, SECTION_MARKED_FLAG);
    }
    return found;
}



/*
 *  UnmarkAllSections() --
 *
 *  Turns off the flag in the sections that are used for marking them.
 */
void UnmarkAllSections() {
    Section *iterator;

    FOREACH_SECTION(iterator) {
	SECTION_REMOVE_FLAG(iterator, SECTION_MARKED_FLAG);
    }

    return;
}



/*
 *  GetFirstSection() --
 *
 *  Returns the lowest-addressed section in the linked-list of sections
 */
Section *GetFirstSection() {
    Section *iterator, *found;

    found = NULL;
    FOREACH_SECTION(iterator) {
	if ((found == NULL) || ((iterator->s_address < found->s_address)
	    && (iterator->s_address != 0)))
	found = iterator;
    }

    return found;
}



/*
 *  GetLastSection() --
 *
 *  Returns the highest-addressed section in the linked-list of sections
 */
Section *GetLastSection() {
    Section *iterator, *found;

    found = NULL;
    FOREACH_SECTION(iterator) {
	if ((found == NULL) || ((iterator->s_address > found->s_address)
	    && (iterator->s_address != 0)))
	found = iterator;
    }

    return found;
}



/*
 *  BacktraceToInstr(addr) --
 *
 *  Given an address, BacktraceToInstr() will look backward through the
 *  bitstream until it hits the first instruction. The intuition is that 'addr'
 *  is an address contained within this instruction.
 */
Instruction *BacktraceToInstr(UINT32 addr) {
    Instruction *instr;
    int offset;

    instr = NULL;
    offset = 0;
    while ((instr == NULL) && (abs(offset) < MaxInstrLength)) {
	instr = AddressToInstr(addr + offset);
	if (instr) {
	    instr->i_temp = abs(offset);
	}
	offset--;
    }
    return instr;
}



/*
 *  RemoveRelocationByAddress(address) --
 *
 *  Scans through the relocation list and removes the first relocation which
 *  matches the given address.
 */
bool RemoveRelocationByAddress(UINT32 address) {
    Relocation *reloc;

    reloc = RelocLookupEntry(address);
    if (reloc == NULL) {
	return false;
    }

    RemoveRelocation(reloc);
    return true;
}



/*
 *  RemoveRelocation(reloc) --
 *
 *  Removes/disconnects the relocation and frees up all memory used by it.
 */
void RemoveRelocation(Relocation *reloc) {

    if (reloc->r_next) {
	reloc->r_next->r_prev = reloc->r_prev;
    }
    if (reloc->r_prev) {
	reloc->r_prev->r_next = reloc->r_next;
    }
    if (reloc == r_list) {
	r_list = reloc->r_next;
    }

    RelocRemoveEntry(reloc);
    RelocationFree(reloc);
}



/*
 *  GetFlistFunAtAddress(addr) --
 *
 *  Returns the FList from the global list of FLists that contains
 *  the address addr as determined by the original address of the flist.
 *
 *  addr: the address to check each FLIST for containment of
 *  returns: the FList that contains the address addr
 *  see also: dontTouchFuns
 */
FList *GetFlistFunAtAddress(ADDRESS addr) {
    FList *fiterator;

    FOREACH_FLIST(fiterator) {
	if ((fiterator->fl_startAddress <= addr)
	&& (fiterator->fl_endAddress >= addr)) {
	    return fiterator;
	}
    }

    return NULL;
}



/*
 *  GetFlistFunAtNewAddress(addr) --
 *
 *  Returns the FList from the global list of FLists that contains
 *  the address addr as determined by the new address of the flist.
 *
 *  addr: the address to check each FLIST for containment of
 *  returns: the FList that contains the address addr
 *  see also: dontTouchFuns
 */
FList *GetFlistFunAtNewAddress(ADDRESS addr) {
    FList *fiterator;

    FOREACH_FLIST(fiterator) {
	if ((fiterator->fl_newStartAddress <= addr)
	    && (fiterator->fl_newEndAddress >= addr))
	    return fiterator;
    }

    return NULL;
}



/*
 *  addr_compare(left,right) --
 *
 *  this will reverse sort function symbols by address
 */
static int addr_compare(void *left, void *right) {
    return (((Symbol *) right)->sy_value) - (((Symbol *) left)->sy_value);
}



/*
 *  GetSortedSymbolList() - return sorted list of symbols.
 *
 *  Creates a list of symbols, and sorts it if one
 *  has not been initialized already.
 */
LinkedList *GetSortedSymbolList() {
    Symbol *curr;
    Iterator *iter;
    static LinkedList *fun_syms;
    static bool fun_syms_initialized;

    // do we have the sorted list ??
    if (fun_syms_initialized) {
        return fun_syms;
    }

    // first time -- initialize and sort the list.
    fun_syms_initialized = true;
    fun_syms = ll_new(LIST_UNSORTED);
    fun_syms->ll_compare = &addr_compare;

    long long i = 0;
    FOREACH_SYMBOL(curr) {
        if ((!!(curr->sy_flags & SYMBOL_FUNCTION_FLAG)) ||
            ((!!(curr->sy_flags & SYMBOL_NOTYPE_FLAG)) &&
                (curr->sy_section == SECTION_REF_CODE))) {
            ll_add(fun_syms, curr);
            i++;
        }
    }

    REPORT(RP_TRIVIAL, "Added %ld symbols to symbol list bfore sorting\n",i);
    ll_sort2(fun_syms);

    REPORT(RP_TRIVIAL, "Sorted symbol list:");
    REPORT(RP_TRIVIAL, "  %8s %-30s %8s", "value", "symbol", "size");
    iter = ll_iterator(fun_syms);
    while (iter->it_hasNext(iter)) {
	curr = (Symbol *) iter->it_next(iter);
	REPORT(RP_TRIVIAL, "  %8x %-30s %8x",
	    curr->sy_value, curr->sy_name, curr->sy_size);
    }

    return fun_syms;
}



/*
 *  AddFlist(addr) --
 *
 *  Adds a function to the flist -- list of undisassembled functions.
 */
FList *AddFlist(ADDRESS addr) {
    ADDRESS startAddr, endAddr;
    Symbol *siterator;
    FList *newFlist;
    int codeOffset;

    newFlist = GetFlistFunAtAddress(addr);

    if (newFlist != NULL) {
	return newFlist;
    }

    // check to see if there is a matching function symbol with size information
    // as well (this is so we don't have to try and guess the function boundary)
    FOREACH_SYMBOL(siterator) {
	if (siterator->sy_flags & SYMBOL_FUNCTION_FLAG) {
	    startAddr = siterator->sy_value;
	    endAddr = startAddr + siterator->sy_size;

	    if (addr >= startAddr && addr <= endAddr) {
		REPORT(RP_SMLOOP, "Found symbol for flist %x: %s",
		    addr, siterator->sy_name);
		break;
	    } else {
		startAddr = endAddr = -1;	// reset the bounds
	    }
	}
    }

    // if there was no size information associated with the containing function
    // we will have to search for the bounds manually
    if (siterator == NULL) {
	LinkedList *f_syms;
	Iterator *iterator;
	Symbol *curr, *prev;

	f_syms = GetSortedSymbolList();
	iterator = ll_iterator(f_syms);
	prev = NULL;

	// the idea is to look for the first function whose starting address
	// is lower than the target address (since we search from the back of
	// a sorted list we can be sure that this is the containing function)
	while (iterator->it_hasNext(iterator)) {
	    curr = (Symbol *) (iterator->it_next(iterator));

	    startAddr = curr->sy_value;
	    // if this is the function it belongs in,
	    // we find the ending address of this function
	    if (addr >= startAddr) {
		// if this is not the last function we will take the beginning
		// address of the previous (logically the next function) to be
		// the ending address of this function
		if (prev != NULL) {
		    endAddr = prev->sy_value;
		} else {
		    // if this is the last function then we will assume that the
		    // function spans the remaining portion of the code section
		    endAddr = SECTION_REF_CODE->s_datasize +
			SECTION_REF_CODE->s_address;
		}

		siterator = curr;
		break;
	    }

	    prev = curr;
	}
    }

    if (siterator == NULL) {
	WARNING("Could not find bounding function for %x, not adding FLIST!",
	    addr);
	return NULL;
    }
    // create the flist with the new found info about the function
    newFlist = zalloc(sizeof(FList));
    newFlist->fl_startAddress = startAddr;
    newFlist->fl_endAddress = endAddr - 1;
    newFlist->fl_name =
	salloc(GetStrongestSymbolForValue(siterator->sy_value)->sy_name);
    codeOffset = startAddr - SECTION_REF_CODE->s_address;
    newFlist->fl_data = zalloc(endAddr - startAddr);
    memcpy(newFlist->fl_data, &SECTION_REF_CODE->s_data[codeOffset],
	endAddr - startAddr);

    REPORT(RP_SMDETAIL, "New flist entry: %8x %8x %s",
	newFlist->fl_startAddress, newFlist->fl_endAddress, newFlist->fl_name);

    if (dontTouchFuns == NULL) {
	dontTouchFuns = newFlist;
    } else {
	newFlist->fl_next = dontTouchFuns;
	dontTouchFuns->fl_prev = newFlist;
	dontTouchFuns = newFlist;
    }

    return newFlist;
}



/*
 *  NameToFlist(name) --
 *
 *  Returns the FList in the global list of FLists whose
 *  name matches the specified name.
 *
 *  name: the name to match
 *  returns: the FList whose name matches the specified name
 *  see also: dontTouchFuns
 */
FList *NameToFlist(char *name) {
    FList *flist;

    FOREACH_FLIST(flist) {
	if (strcmp(flist->fl_name, name) == 0) {
	    return flist;
	}
    }

    return NULL;
}



/*
 *  RecalculateAddresses() --
 *
 *  This function recalculates all address in the binary, including
 *  instructions, symbol values, and flists. This function relinks all
 *  instructions based on basic block layout, sorts the sections by the
 *  order in which they will be written out, then does a fixed-point
 *  update of all PC relative instructions and UPMs (the fixed-point is
 *  necessary because both the PC relative instructions and any relative
 *  UPMs may change byte modes of instructions and may therefore require
 *  instruction lengths to change and hence change addresses). Finally
 *  FLIST addresses are updated to reflect the new position of the FLIST
 *  section.
 *  see also: SECTION_REF_FLISTS
 */
void RecalculateAddresses() {
    int iterations, changed;

    REPORT(RP_DETAIL, "Recalculating section sizes and addresses");

    MarkFlistCalls();
    InstrClearTemp();
    InstrRelinkBasedOnBBL();

    SortSectionsByWriteOrder();
    sec_list->s_newaddress = ComputeFirstSectionAddress();
    UpdateSectionSizes(true);

    iterations = 0;

    changed = 1;
    while (changed > 0) {
	changed = 0;
	iterations++;

	REPORT(RP_SMLOOP, "Beginning round %d", iterations);
	UpdateSectionSizes(true);

	UpdatePCRelativeJumps(i_list);
	UpdateSectionSizes(true);

	changed = ApplyUPMs(upm2s_list);
	REPORT(RP_SMLOOP, "Round %d complete; %d UPMs modified",
	    iterations, changed);
    }

}



/*
 *  IsPICCall(instr) --
 *
 *  Returns whether or not the the specified instruction is a PIC call.
 *
 *  instr: the instruction to check
 *  returns: true if the specified instruction is a PIC call.
 */
bool IsPICCall(Instruction *instr) {
    Instruction *target;

    if (!isCallClass(instr)) {
	return false;
    }

    target = AddressToInstr(instr->i_address +
	instr->i_length + instr->i_src->a_displacement);

    return target == instr->i_next;
}



/*
 *  GetGOTInstrForPICCall(instr) --
 *
 *  Returns the GOT instruction for the specified PIC call. The GOT
 *  instruction is the instruction that is actually patched.
 *
 *  instr: the instruction to check
 *  returns: the GOT instruction that corresponds to the specified PIC
 *  call, or NULL if there isn't one
 */
Instruction *GetGOTInstrForPICCall(Instruction *instr) {
    if ((instr->i_next == NULL) || (instr->i_next->i_next == NULL)) {
	return NULL;
    }

    return instr->i_next->i_next;
}



/*
 *  GetPICCallForGOTInstr(instr) --
 *
 *  Returns the PIC call that corresponds to the specified instruction
 *  that is assumed to have a GOT relocation/UPM patching into it.
 *
 *  instr: the instruction to check
 *  returns: the PIC call that corresponds to the specified instruction, or
 *  NULL if there isn't one
 */
Instruction *GetPICCallForGOTInstr(Instruction *instr) {
    if (instr->i_prev == NULL) {
	return NULL;
    }

    if (instr->i_prev->i_prev) {
	return NULL;
    }

    if (!IsPICCall(instr->i_prev->i_prev)) {
	return NULL;
    }

    return instr->i_prev->i_prev;
}
