/*
 *  sloop.c
 *
 *  Generates a textual representation of the CFG to be used by Sloop.
 */
#define START       "%StartSloop"
#define END         "%EndSloop"
#define FUNSTART    "%FUNCTION"
#define CALLSITES   "%CALLSITES"
#define CALLER      "%CALLER"
#define BBL         "%BBL"
#define COUNT       "%COUNT"
#define INSTR       "%I"
#define EDGES       "%EDGES"
#define CHECKED     "%CHECKED"
#define LOOP        "%LOOP"
#define LOOPHEADER  "%LOOPHEADER"

#include "global.h"
#include "sloop.h"
#include "utilities.h"
#include "sysdep.h"
#include "dominator.h"
#include "flowgraph.h"

static char buffer[1024], buffer2[64];

static int sloopExecutedAlready;


/* Writes out the loop. The assumption is that 'bbl' is the loop header */
void BBLSloopOutputLoop(Block *header, FILE *fp);

/* Writes out SLOOP output for 'block' */
void BBLSloopOutput(Block *block, FILE *fp);

/* Returns a non-NULL name for the function */
static char *FunctionConcreteName(Function *fun);



/*
 *  GenerateSloopOutput(fp,phase) --
 *
 *  Generates a sloop output file for the control flow graph.
 */
void GenerateSloopOutput(FILE *fp, char *phase) {
    Function *fiterator;
    Block *biterator;
    EdgeNode *niterator;

    if (!sloopfile) {
	return;
    }

    StoreDominatorInfo();

    fprintf(fp, "%s\n", START);

    fprintf(fp, "%%PHASE{%s}\n", phase);

    fprintf(fp, "\n%%FUNCTION{HELL}{-1}{\n");
    fprintf(fp, "\t%%CHECKED{no}\n");
    fprintf(fp, "\t%%BBL{-1}{\n\t\t%%COUNT{-1}\n\t}\n}\n\n");

    FOREACH_FUN(fiterator) {
	fprintf(fp, "%s{%s}{%d}{\n\t%s{\n", FUNSTART,
	    FunctionConcreteName(fiterator), fiterator->f_id, CALLSITES);

	FOREACH_PRED_IN_BBL(fiterator->f_entry, niterator) {
	    Block *caller = NULL;
	    Block *callee = NULL;
	    EdgeNode *niterator2;

	    caller = NODE_FROM(niterator);
	    if (NODE_TYPE(niterator) == ET_HELL
	        || NODE_TYPE(niterator) == ET_HELLMAYBE) {
	            FOREACH_SUCC_IN_BBL(fiterator->f_entry, niterator2) {
		        if (BBLHasHellPredecessor(NODE_TO(niterator2))) {
		            callee = NODE_TO(niterator2);
                            fprintf(fp, "\t\t%s{%s{%d},%d,%d,%d}\n", CALLER, 
                                "-HELL-", -1, -1, callee->b_id, -1);
		        }
	            }
	    } else {
	        FOREACH_SUCC_IN_BBL(caller, niterator2) {
		    if (NODE_TYPE(niterator2) == ET_FUNDIRECT) {
		        callee = NODE_TO(niterator2);
		        break;
		    }
	        }

	        fprintf(fp, "\t\t%s{%s{%d},%d,%d,%d}\n", CALLER,
		    FunctionConcreteName(caller->b_fun),
		    caller->b_fun->f_id, caller->b_id,
		    callee->b_id, (int) NODE_WEIGHT(niterator2));
	    }
	}

	fprintf(fp, "\t}\n\t%s{no}\n", CHECKED);

	/* a count of -99999 indicates pseudo-blocks that needn't be shown */
	fprintf(fp, "\t%s{%d}{\n\t\t%s{-99999}\n\t}\n",
	    BBL, fiterator->f_entry->b_id, COUNT);
	fprintf(fp, "\t%s{%d}{\n\t\t%s{-99999}\n\t}\n",
	    BBL, fiterator->f_exit->b_id, COUNT);

	FOREACH_BBL_IN_FUN(fiterator, biterator) {
	    BBL_REMOVE_FLAG(biterator, BBL_TEMP_FLAG);
	}

	FOREACH_BBL_IN_FUN(fiterator, biterator) {
	    if (biterator->b_loopFirst != NULL) {
		if (!BBL_HAS_FLAG(biterator, BBL_TEMP_FLAG)) {
		    BBLSloopOutputLoop(biterator, fp);
		}
	    }
	}

	FOREACH_BBL_IN_FUN(fiterator, biterator) {
	    if (BBL_HAS_FLAG(biterator, BBL_TEMP_FLAG)) {
		continue;
	    }
	    BBLSloopOutput(biterator, fp);
	}

	fprintf(fp, "%s{", EDGES);

	FOREACH_BBL_IN_FUN(fiterator, biterator) {
	    FOREACH_SUCC_IN_BBL(biterator, niterator) {
		if (NODE_TYPE(niterator) == ET_FUNDIRECT) {
		    continue;
		}
		if (NODE_TYPE(niterator) == ET_FUNCALL) {
		    continue;
		}
		if (NODE_TYPE(niterator) == ET_EXIT) {
		    continue;
		}
		if (NODE_TYPE(niterator) == ET_HELL) {
		    fprintf(fp, "{%d -> -1 : %d}",
                        biterator->b_id, (int) NODE_WEIGHT(niterator));
		} else {
		    fprintf(fp, "{%d -> %d : %d}",
                        biterator->b_id, ((NODE_TO(niterator))->b_id),
                        (int) NODE_WEIGHT(niterator));
		}
	    }
	}

	biterator = fiterator->f_entry;
	FOREACH_SUCC_IN_BBL(biterator, niterator) {
	    if (NODE_TYPE(niterator) == ET_FUNDIRECT) {
		continue;
	    }
	    if (NODE_TYPE(niterator) == ET_HELL) {
                fprintf(fp, "{%d -> -1 : -1}", biterator->b_id);
	    } else {
                fprintf(fp, "{%d -> %d : -1}",
		    biterator->b_id, ((NODE_TO(niterator))->b_id));
            }
	}

	fprintf(fp, "}\n}\n\n");
    }

    fprintf(fp, "%s\n", END);

    fflush(fp);

    return;
}



/*
 *  FunctionConcreteName(fun) --
 *
 *  Returns a concrete name for the function. The function's name, unless it
 *  is NULL, will be returned. If the name is NULL, then a bogus identifier
 *  will be returned.
 */
static char *FunctionConcreteName(Function *fun) {
    if (fun->f_name == NULL) {
	return "SomeFunction";
    }
    if (strlen(fun->f_name) <= 0) {
	return "SomeFunction";
    }
    if (strstr(fun->f_name, ".") != NULL) {
	return "SomeFunction";
    }
    return fun->f_name;
}



/*
 *  BBLSloopOutputLoop(header,fp) --
 *
 *  Writes out the loop beginning with loop header 'header' to the sloop output
 *  file.
 */
void BBLSloopOutputLoop(Block *header, FILE *fp) {
    Block *biterator;
    EdgeNode *niterator;
    Function *fiterator = header->b_fun;

    fprintf(fp, "\t%s{%d}{\n", LOOP, header->b_id);
    fprintf(fp, "\t\t%s{\n", LOOPHEADER);
    BBLSloopOutput(header, fp);
    fprintf(fp, "\t}\n");

    biterator = header->b_loopFirst;
    while (biterator != NULL) {
	if (!BBL_HAS_FLAG(biterator, BBL_TEMP_FLAG)) {
	    if (biterator->b_loopFirst != NULL) {
		BBLSloopOutputLoop(biterator, fp);
	    } else {
		BBLSloopOutput(biterator, fp);
	    }
	}
	biterator = biterator->b_loopNext;
    }

    fprintf(fp, "\t%s{", EDGES);

    FOREACH_BBL_IN_FUN(fiterator, biterator) {
	if (!BBLIsInSameLoopAsBBL(biterator, header)) {
	    continue;
	}
	FOREACH_SUCC_IN_BBL(biterator, niterator) {
	    if (NODE_TYPE(niterator) == ET_FUNDIRECT) {
		continue;
	    }
	    if (NODE_TYPE(niterator) == ET_FUNCALL) {
		continue;
	    }
	    if (NODE_TYPE(niterator) == ET_EXIT) {
	        continue;
	    }
	    if (NODE_TYPE(niterator) == ET_HELL) {
	        fprintf(fp, "{%d -> -1 : %d}",
		    biterator->b_id, (int) NODE_WEIGHT(niterator));
		
	    } else {
	        fprintf(fp, "{%d -> %d : %d}",
		    biterator->b_id, ((NODE_TO(niterator))->b_id),
		    (int) NODE_WEIGHT(niterator));
            }
	}
    }

    fprintf(fp, "}\n\n");
    fprintf(fp, "\t}\n");

    return;
}



/*
 *  BBLSloopOutput(biterator,fp) --
 *
 *  Writes out sloop output for the supplied basic block
 */
void BBLSloopOutput(Block *biterator, FILE *fp) {
    Instruction *iterator;
    Edge *fundirect, *e0;
    char *instrText, *ptr, *instr;

    BBL_SET_FLAG(biterator, BBL_TEMP_FLAG);

    fprintf(fp, "\t%s{%d}%s{\n\t\t%s{%d}\n",
	BBL, biterator->b_id,
	BBL_HAS_FLAG(biterator, BBL_HOT_FLAG) ? "*" : "",
	COUNT, (int) biterator->b_weight);

    FOREACH_INSTR_IN_BBL(biterator, iterator) {
	if (sloopExecutedAlready == 3) {
	    instrText = PrintInstructionFinal(iterator);
	} else {
	    instrText = PrintInstruction(iterator);
	}

	instr = instrText;

	if ((biterator->b_type == BT_BRANCH || biterator->b_type == BT_UNCOND)
            && iterator == biterator->b_last) {

            if (biterator->b_type == BT_BRANCH) {
                e0 = BBLGetTrueEdge(biterator);
            } else {
                e0 = BBLGetUncondEdge(biterator);
            }

            if (e0) {
                strcpy(buffer, instrText);
                for (ptr = buffer; *ptr != '\0'; ptr++) {
                    /* skip to the end */
                }
                sprintf(ptr, " --> B%d", e0->e_to->b_id);
                instr = buffer;
            }
        } else if (biterator->b_type == BT_CALL 
                    && iterator == biterator->b_last) {
	    strcpy(buffer, instrText);
	    for (ptr = buffer; !isspace(*ptr) && *ptr != '\0'; ptr++) {
		/* skip ahead to the first blank */
	    }
	    for (; isspace(*ptr) && *ptr != '\0'; ptr++) {
		/* now skip to the first non-blank, the target of the call */
	    }

	    fundirect = BBLGetFunDirectEdge(biterator);
	    if (fundirect) {
		strcpy(ptr, fundirect->e_to->b_fun->f_name);
		sprintf(buffer2, ":B%d", fundirect->e_to->b_id);
		strcat(ptr, buffer2);
	    } else {
                while (*ptr != '\0') {
                    ptr++;     /* skip to the end */
	        }
                strcpy(ptr, " -HELL-");
	    }

	    instr = buffer;
	}

	if (sloopExecutedAlready == 3) {
	    fprintf(fp, "\t\t%s{ %d  0x%8.8x  %s }%s\n", INSTR,
		iterator->i_id, iterator->i_final_address, instr, INSTR);
	} else {
	    fprintf(fp, "\t\t%s{ %d  0x%8.8x  %s }%s\n", INSTR,
		iterator->i_id, iterator->i_address, instr, INSTR);
	}

	free(instrText);
    }

    fprintf(fp, "\t}\n");

    return;
}
