/*
 *  A module for performing stack analysis. Well-behaved functions: A
 *  function is well-behaved iff it has both a valid prologue, valid
 *  epilogue, does not have escaping edges, and does not maliciously
 *  define the frame pointer.
 */
#include "global.h"
#include "utilities.h"
#include "linkage.h"
#include "flowgraph.h"
#include "stack.analyze.h"
#include "opt.h"
#include "sysdep.h"
#include "x86-classes.h"

#define ESCAPING_EDGE                  (1 << 0)
#define STACK_NOT_BALANCED             (1 << 1)
#define STACK_BALANCED                 (1 << 2)
#define HELL_PREDECESSOR               (1 << 3)
#define INSTR_UNKNOWN                  (1 << 4)
#define NOT_WELL_BEHAVED               (1 << 5)

#define NOT_WELL_BEHAVED_CONTRIBUTIONS STACK_UNKNOWN

/*
 *  Returns 1 if the stack pushes the same number of bytes as it pops along all
 *  possible execution paths
 */
static int FunctionHasBalancedStack(Function *fun);

/* Returns true if, at every return point, the net stack gain is 0 */
static bool FunctionNetStackGainIsZero(Function *fun);

/* Returns the net gain in bytes that 'instr' contributes to the stack */
int ComputeMyStackGainForInstr(Instruction *instr);

/* Calculates the difference between ESP end EBP for 'fun'
 * and updates all bbls in the function */
static void CalculateSPBPDifferenceForFun(Function *fun);

/* Returns true if an instruction in 'fun' aliases the stack pointer */
bool FunAliasesStackPointer(Function *fun);

/* Returns true if an instruction in 'bbl' aliases the stack pointer */
bool BBLAliasesStackPointer(Block *bbl);

/* Returns true if the destination operand in the instruction
 * gets the contents of a stack pointer */
bool InstrAliasesStackPointer(Instruction *instr);

/* Returns the displacement which the instruction synchronizes
 * the stack and the frame pointer to */
int InstrSynchronizesStack(Instruction *instr);

/* Computes the net stack gain at the end of 'bbl' */
static int ComputeNetStackAtEndOfBBL(Block *bbl);



/*
 *  MarkWellBehavedFunctions() --
 *
 *  Sets the FUNCTION_WELL_BEHAVED_FLAG on all functions that leave the
 *  stack as they found it.
 */
void MarkWellBehavedFunctions() {
    Function *funs;
    EdgeNode *preds, *succs;
    int wellBehaved, balancedStack, totalBalanced;
    int totalHellPred, totalEscapingEdges, totalNotBalanced, totalInstrUnknown;
    bool change = true;

    REPORT(RP_DETAIL, "Marking well-behaved functions");

    /*
     *                     PASS 1: LEAVE/RET COMBO
     * We require that a well-behaved function use a leave followed by
     * a return to restore the stack to its previous state. A function
     * which does this is a candidate for a well-behaved function.
     */

    FOREACH_FUNCTION(funs) {
	FUNCTION_REMOVE_FLAG(funs, FUNCTION_TEMP_FLAG);
	FUNCTION_SET_FLAG(funs, FUNCTION_WELL_BEHAVED_FLAG);

	FOREACH_PRED_IN_BBL(funs->f_exit, preds) {
	    if (NODE_TYPE(preds) == ET_COMPENSATE) {
		continue;
	    }
		
	    if (BBLContainsExitCombo(NODE_FROM(preds))) {
		FUNCTION_SET_FLAG(funs, FUNCTION_LEAVE_RET_COMBO);
	    }
	}

	if (FUNCTION_HAS_FLAG(funs, FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG)) {
	    FUNCTION_REMOVE_FLAG(funs, FUNCTION_WELL_BEHAVED_FLAG);
	}

	/*
	 * Set the FUNCTION_GCC_PROLOGUE_FLAG for functions that have the
	 * standard entry instructions of the form:
	 *       push %ebp
	 *       mov  %esp->%ebp
	 */

	FUNCTION_SET_FLAG(funs, FUNCTION_GCC_PROLOGUE_FLAG);
	FOREACH_SUCC_IN_BBL(funs->f_entry, succs) {
	    if (!BBLContainsPrologue(NODE_TO(succs))) {
		FUNCTION_REMOVE_FLAG(funs, FUNCTION_GCC_PROLOGUE_FLAG);
	    }
	}
    }

    FOREACH_FUNCTION(funs) {
	FUNCTION_SET_FLAG(funs, FUNCTION_TEMP_FLAG);
	if (!FUNCTION_HAS_FLAG(funs, FUNCTION_GCC_PROLOGUE_FLAG)) {
	    FUNCTION_REMOVE_FLAG(funs, FUNCTION_WELL_BEHAVED_FLAG);
	}
    }

    balancedStack = totalEscapingEdges = totalHellPred =
	totalNotBalanced = totalBalanced = totalInstrUnknown = wellBehaved = 0;

    while (change) {
	change = false;
	FOREACH_FUNCTION(funs) {
	    if (FUNCTION_HAS_FLAG(funs, FUNCTION_TEMP_FLAG)) {
		int temp2 = FunctionHasBalancedStack(funs);
		if (temp2 & NOT_WELL_BEHAVED) {
		    if (FUNCTION_HAS_FLAG(funs, FUNCTION_WELL_BEHAVED_FLAG)) {
			change = true;
			FOREACH_PRED_IN_BBL(funs->f_entry, preds) {
			    Function *predFun = NODE_FROM(preds)->b_fun;
			    if (!FUNCTION_HAS_FLAG(predFun,
				    FUNCTION_GCC_PROLOGUE_FLAG)
				|| !FUNCTION_HAS_FLAG(predFun,
				    FUNCTION_LEAVE_RET_COMBO)) {
				FUNCTION_SET_FLAG(NODE_FROM(preds)->b_fun,
				    FUNCTION_TEMP_FLAG);
				FUNCTION_REMOVE_FLAG(NODE_FROM(preds)->b_fun,
				    FUNCTION_WELL_BEHAVED_FLAG);
			    }
			}
			FUNCTION_REMOVE_FLAG(funs, FUNCTION_WELL_BEHAVED_FLAG);
		    }
		}
	    }
	}
    }

    FOREACH_FUNCTION(funs) {
	int temp2 = FunctionHasBalancedStack(funs);

	/*
	 *          Update stats about the stack for all funs
	 */

	if (temp2 & STACK_BALANCED) {
	    totalBalanced++;
	    FUNCTION_SET_FLAG(funs, FUNCTION_BALANCED_STACK_FLAG);
	}
	if (temp2 & ESCAPING_EDGE) {
	    totalEscapingEdges++;
	}
	if (temp2 & HELL_PREDECESSOR) {
	    totalHellPred++;
	}
	if (temp2 & INSTR_UNKNOWN) {
	    totalInstrUnknown++;
	}
	if (temp2 & STACK_NOT_BALANCED) {
	    totalNotBalanced++;
	}
	if (FUNCTION_HAS_FLAG(funs, FUNCTION_WELL_BEHAVED_FLAG)) {
	    wellBehaved++;
	}
    }

    REPORT(RP_DETAIL, "%5d functions with balanced stack [%.0f%%]",
	totalBalanced, (totalBalanced * 100.0) / numFunctions);
    REPORT(RP_DETAIL, "%5d functions with escaping edges [%.0f%%]",
	totalEscapingEdges, (totalEscapingEdges * 100.0) / numFunctions);
    REPORT(RP_DETAIL, "%5d functions had HELL as a predecessor [%.0f%%]",
	totalHellPred, (totalHellPred * 100.0) / numFunctions);
    REPORT(RP_DETAIL, "%5d functions produced UNKNOWN [%.0f%%]",
	totalInstrUnknown, (totalInstrUnknown * 100.0) / numFunctions);
    REPORT(RP_DETAIL, "%5d functions with not balanced stacks [%.0f%%]",
	totalNotBalanced, (totalNotBalanced * 100.0) / numFunctions);
    REPORT(RP_DETAIL, "%5d are well-behaved. [%.0f%%]",
	wellBehaved, (float) (wellBehaved) * 100 / numFunctions);
}



/*
 *  FunctionHasBalancedStack(fun) --
 *
 *  Returns true if 'fun' is well-behaved; e.g, if it leaves the stack
 *  as it found it.
 */
static int FunctionHasBalancedStack(Function *fun) {
    Block *bbl;
    EdgeNode *successors;
    EdgeNode *predecessors;
    bool change = true;
    bool first = true;
    int result = 0;

    /*
     * Functions with escaping edges will not be considered
     * as well-behaved, since we might not be able to determine what
     * is happening with the stack.
     */

    if (FunHasEsc2(fun)) {
	result |= ESCAPING_EDGE;
    }

    FOREACH_BBL_IN_FUN(fun, bbl) {
	bbl->b_tmp = STACK_UNDEFINED;
	BBL_REMOVE_FLAG(bbl, BBL_TEMP_FLAG);

	if (BBLHasHellPredecessor(bbl)) {
	    if (!BBLHasEntryPredecessor(bbl)) {
		result |= HELL_PREDECESSOR;
		result |= STACK_NOT_BALANCED;
		bbl->b_tmp = STACK_UNKNOWN;
	    }
	}
    }

    FOREACH_SUCC_IN_BBL(fun->f_entry, successors) {
	NODE_TO(successors)->b_tmp = 0;
    }

    /*
     * Below is a computation that iteratively tests to see if the stack is
     * balanced along all execution paths.  The analysis terminates when
     * nothing changes.
     */

    while (change) {

	change = false;

	FOREACH_BBL_IN_FUN(fun, bbl) {
	    int inGain;

	    first = true;
	    FOREACH_PRED_IN_BBL(bbl, predecessors) {
		Block *pred = NODE_FROM(predecessors);
		inGain = ComputeNetStackAtEndOfBBL(pred);

		/*
		 * The analysis should work inductively on the functions, so we
		 * need not consider interprocedural edges.
		 */

		if ((NODE_TYPE(predecessors) == ET_ENTRY)
		    || (NODE_TYPE(predecessors) == ET_EXIT))
		    continue;

		if (isInterproceduralEdge(predecessors->en_edge)) {
		    continue;
		}

		/*
		 * If we are dealing with the first predecessor, then set our
		 * net stack gain to the stack gain coming out of the
		 * predecessor block.
		 */

		if (first) {
		    if (inGain == STACK_UNDEFINED) {
			continue;
		    }

		    first = false;

		    if (bbl->b_tmp == STACK_UNKNOWN) {
			continue;
		    }

		    if (bbl->b_tmp != inGain) {
			change = true;
		    }

		    bbl->b_tmp = inGain;

		    if (bbl->b_tmp == STACK_UNKNOWN) {
			result |= STACK_NOT_BALANCED;
		    }

		} else {
		    /*
		     * If we are not dealing with the first predecessor, then
		     * make sure that the net stack gain for the current basic
		     * block will be the same if we use the incoming stack gain
		     * from the current predecessor.
		     */
		    if (inGain == STACK_UNDEFINED) {
			continue;
		    }

		    if (bbl->b_tmp == STACK_UNKNOWN) {
			continue;
		    }

		    if (bbl->b_tmp != inGain) {
			change = true;
			bbl->b_tmp = STACK_UNKNOWN;
			result |= STACK_NOT_BALANCED;
		    }
		}

	    }			/* End iteration over predecessors */
	}			/* End iteration over each basic block */
    }				/* End fixpoint computation loop */

    if (!(result & STACK_NOT_BALANCED)) {
	result |= STACK_BALANCED;
    }

    if (!FUNCTION_HAS_FLAG(fun, FUNCTION_LEAVE_RET_COMBO) ||
	!FUNCTION_HAS_FLAG(fun, FUNCTION_GCC_PROLOGUE_FLAG)) {
	if (!FunctionNetStackGainIsZero(fun)) {
	    result |= NOT_WELL_BEHAVED;
	}
    }

    return result;
}



/*
 *  ComputeNetStackAtEndOfBBL(bbl) --
 *
 *  ComputeNetStackAtEndOfBBL returns the net stack gain at the end of the
 *  basic block.
 */
static int ComputeNetStackAtEndOfBBL(Block *bbl) {
    int outGain, synch, netGain, instrGain;
    Instruction *instr;

    netGain = bbl->b_tmp;
    outGain = netGain;

    if (netGain == STACK_UNDEFINED) {
	return STACK_UNDEFINED;
    }

    FOREACH_INSTR_IN_BBL(bbl, instr) {
	synch = InstrSynchronizesStack(instr);

	if (synch != STACK_UNKNOWN) {
	    if (FUNCTION_HAS_FLAG(bbl->b_fun, FUNCTION_GCC_PROLOGUE_FLAG)) {
		outGain = -1 * synch;
	    } else {
		outGain = STACK_UNKNOWN;
	    }
	    continue;
	}

	if (outGain == STACK_UNKNOWN) {
	    continue;
	}

	instrGain = ComputeMyStackGainForInstr(instr);
	if (instrGain == STACK_UNKNOWN) {
	    outGain = STACK_UNKNOWN;
	} else {
	    outGain = outGain + instrGain;
	}
    }

    return outGain;
}



/*
 *  InstrSynchronizesStack(instr) --
 *
 *  Returns the displacement to which EBP/ESP are synchronized.
 */
int InstrSynchronizesStack(Instruction *instr) {

    /*
     * Below we try to detect instructions of the form:
     *    lea    0xffffffec(%ebp) -> %esp
     */

    if (isLeaClass(instr) && (instr->i_dest) && (instr->i_src) &&
	OperandIsStackPointer(instr->i_dest) &&
	RegisterIsFramePointer(instr->i_src->a_base_reg) &&
	!OPERAND_HAS_FLAG(instr->i_src, OPERAND_INDEX_FLAG) &&
	!OPERAND_HAS_FLAG(instr->i_src, OPERAND_SCALE_FLAG) &&
	(OPERAND_HAS_FLAG(instr->i_src, OPERAND_DISP_A_FLAG) ||
	    OPERAND_HAS_FLAG(instr->i_src, OPERAND_DISP_R_FLAG))) {
	return instr->i_src->a_displacement;
    }

    if (isMovClass(instr) && (instr->i_dest) && (instr->i_src) &&
	OperandIsFramePointer(instr->i_dest)
	&& OperandIsStackPointer(instr->i_src)) {
	return 0;
    }

    return STACK_UNKNOWN;
}



/*
 *  ComputeMyStackGainForInstr(instr) --
 *
 *  Returns the net gain that this instruction contributes to the height of the
 *  stack.
 */
int ComputeMyStackGainForInstr(Instruction *instr) {

    if ((instr->i_dest) && OperandIsStackPointer(instr->i_dest)) {
	if (isAddClass(instr) && instr->i_src && instr->i_src->a_displacement) {
	    return -1 * instr->i_src->a_displacement;
	} else if (
	isSubClass(instr) && instr->i_src && instr->i_src->a_displacement) {
	    return instr->i_src->a_displacement;
	} else {
	    return STACK_UNKNOWN;
	}
    }

    if (InstrPopBytes(instr) != 0) {
	if (isReturnClass(instr) && (instr->i_src)) {
	    return (InstrPopBytes(instr) * -1) - instr->i_src->a_displacement;
	} else {
	    return InstrPopBytes(instr) * -1;
	}
    }

    if ((instr->i_dest) && OperandIsFramePointer(instr->i_dest)) {
	if (isAddClass(instr) && (instr->i_src)
		&& instr->i_src->a_displacement) {
	    return instr->i_src->a_displacement;
	} else if (isSubClass(instr) && instr->i_src
		&& (instr->i_src->a_displacement)) {
	    return -1 * instr->i_src->a_displacement;
	} else {
	    return STACK_UNKNOWN;
	}
    }

    if (isCallClass(instr)) {
	if (BBLGetFunDirectEdge(instr->i_Bblock) == NULL) {
	    return NOT_WELL_BEHAVED_CONTRIBUTIONS;
	} else if (!FUNCTION_HAS_FLAG(BBLGetFunDirectEdge(instr->i_Bblock)
		->e_to->b_fun, FUNCTION_WELL_BEHAVED_FLAG)) {
	    return NOT_WELL_BEHAVED_CONTRIBUTIONS;
	} else {
	    return 0;
	}
    }

    if (InstrPushBytes(instr) != 0) {
	return InstrPushBytes(instr);
    }

    return 0;
}



/*
 *  ComputeStackFrameSize(fun) --
 *
 *  Returns the maximum number of bytes used at any given time in the
 *  stack frame for this function. 0 is returned to denote an unbounded
 *  limit (e.g, a loop which results in a net gain on the stack).
 *  This function should be called after MarkWellBehavedFunctions was
 *  called, so that the temporary storage that holds net stack gains can be
 *  re-used.
 */
int ComputeStackFrameSize(Function *fun) {
    Block *bbl;
    Instruction *instr;
    int max = 0;
    int diff;

    if (!FUNCTION_HAS_FLAG(fun, FUNCTION_BALANCED_STACK_FLAG)) {
	return max;
    }

    if (!FUNCTION_HAS_FLAG(fun, FUNCTION_GCC_PROLOGUE_FLAG)) {
	return max;
    }

    FOREACH_BBL_IN_FUN(fun, bbl) {
	FOREACH_INSTR_IN_BBL(bbl, instr) {
	    diff = InstrBPSPDiff(instr);
	    if (diff > max) {
		max = diff;
	    }
	}
    }

    return max + ADDRESS_SIZE;
}



/*
 *  CalculateSPBPDifference() --
 *
 *  Computes the difference between the frame and stack pointer for all
 *  basic blocks (the beginning of them), and stores the result in
 *  ->b_BPSPdiff.
 */
void CalculateSPBPDifference() {
    Function *fun;

    FOREACH_FUNCTION(fun) {
	CalculateSPBPDifferenceForFun(fun);
    }

    return;
}



/*
 *  CalculateSPBPDifferenceForFun(fun) --
 *
 *  Computes the difference between the frame and stack pointer for all
 *  basic blocks (the beginning of them) in 'fun', and stores the result in
 *  ->b_BPSPdiff.
 */
void CalculateSPBPDifferenceForFun(Function *fun) {
    Block *bbl;

    FOREACH_BBL_IN_FUN(fun, bbl) {
	if (FUNCTION_HAS_FLAG(bbl->b_fun, FUNCTION_WELL_BEHAVED_FLAG)) {
	    bbl->b_BPSPdiff = bbl->b_tmp;
	    if (((bbl->b_BPSPdiff % ADDRESS_SIZE) != 0) &&
		(bbl->b_BPSPdiff != STACK_UNKNOWN)) {
	    }
	} else {
	    bbl->b_BPSPdiff = STACK_UNKNOWN;
	}
    }

    return;
}



/*
 *  MarkFunsAliasingStackPointer() --
 *
 *  Sets the flag FUNCTION_ALIASES_STACK_POINTER if the function contains
 *  an instruction which aliases ESP or EBP.
 */
void MarkFunsAliasingStackPointer() {
    Function *fun;

    FOREACH_FUNCTION(fun) {
	if (FunAliasesStackPointer(fun)) {
	    FUNCTION_SET_FLAG(fun, FUNCTION_ALIASES_STACK_POINTER_FLAG);
	}
    }

    return;
}



/*
 *  FunAliasesStackPointer(fun) --
 *
 *  Returns true if an instruction inside of the supplied function
 *  aliases a stack or frame pointer.
 */
bool FunAliasesStackPointer(Function *fun) {
    Block *bbl;

    FOREACH_BBL_IN_FUN(fun, bbl) {
	if (BBLAliasesStackPointer(bbl)) {
	    return true;
	}
    }
    return false;
}



/*
 *  BBLAliasesStackPointer(bbl) --
 *
 *  Returns true if an instruction inside of the supplied basic block
 *  aliases a stack or frame pointer.
 */
bool BBLAliasesStackPointer(Block *bbl) {
    Instruction *instr;

    FOREACH_INSTR_IN_BBL(bbl, instr) {
	if (InstrAliasesStackPointer(instr)) {
	    return true;
	}
    }
    return false;
}



/*
 *  InstrAliasesStackPointer(instr) --
 *
 *  Returns true if 'instr' may alias the stack or frame pointer.
 */
bool InstrAliasesStackPointer(Instruction *instr) {
    bool sourceSP, sourceFP;

    if (!instr->i_dest) {
	return false;
    }

    /*
     * If the destination operand is %ebp or %esp then this instruction is
     * not aliasing any pointer into the stack, since by convention ESP and
     * EBP point into the stack anyways.
     */

    if (OperandIsStackPointer(instr->i_dest)) {
	return false;
    }

    if (OperandIsFramePointer(instr->i_dest)) {
	return false;
    }

    if (isLeaClass(instr)
	&& (RegisterIsFramePointer(instr->i_src->a_base_reg)
	    || RegisterIsStackPointer(instr->i_src->a_base_reg)
	    || RegisterIsFramePointer(instr->i_src->a_index_reg)
	    || RegisterIsStackPointer(instr->i_src->a_index_reg)))
	return true;

    /*
     * If either of the source operands use the contents of EBP or ESP
     * then the instruction is aliasing a pointer into the stack, and true
     * is returned.
     */

    sourceSP = false;
    sourceFP = false;
    if ((instr->i_src) && OperandUsesContentsOfStackPointer(instr->i_src)) {
	sourceSP = true;
    } else if (instr->i_src && OperandUsesContentsOfFramePointer(instr->i_src)){
	sourceFP = true;
    }else if(instr->i_src2 && OperandUsesContentsOfStackPointer(instr->i_src2)){
	sourceSP = true;
    }else if(instr->i_src2 && OperandUsesContentsOfFramePointer(instr->i_src2)){
	sourceFP = true;
    }

    if (sourceFP || sourceSP) {
	return true;
    }


    /* ALERT: Add in case for indirect references, e.g. mov eax <- *0x804b551 */


    return false;
}



/*
 *  BBLContainsExitCombo(bbl) --
 *
 *  Returns true if the basic block ends in the LEAVE/RET combination.
 */
bool BBLContainsExitCombo(Block *bbl) {
    if (!isReturnClass(bbl->b_last)) {
	return false;
    }
    if (bbl->b_last->i_prev == NULL) {
	return false;
    }
    if (!isLeaveClass(bbl->b_last->i_prev)) {
	return false;
    }
    if (isReturnClass(bbl->b_last)) {
	if (bbl->b_last->i_src) {
	    return false;
	}
    }
    return true;
}



/*
 *  InstrPotentiallyWritesToStack(instr) --
 *
 *  Returns true if there is a chance the instruction may write to a memory
 *  location on the stack. An indirect write through register or memory could
 *  possibly write to the stack.
 */
bool InstrPotentiallyWritesToStack(Instruction *instr) {
    Operand *dest = instr->i_dest;

    if (dest == NULL) {
	return false;
    }

    if (OPERAND_HAS_FLAG(dest, OPERAND_INDIRECT_FLAG)) {
	return true;
    }

    if (OPERAND_HAS_FLAG(dest, OPERAND_REGISTER_INDIRECT)) {
	return true;
    }

    return false;
}



/*
 *  InstrBPSPDiff(instr) --
 *
 *  Returns the difference between the BP and SP right before the instruction.
 */
int InstrBPSPDiff(Instruction *instr) {
    Instruction *iterator;
    Block *bbl = instr->i_Bblock;
    int diff = bbl->b_BPSPdiff;
    int gain;
    int synch;

    if (INSTR_HAS_FLAG(instr, INSTR_BPSP_DIFF_STORED_FLAG)) {
	return instr->i_temp;
    }

    if (diff == STACK_UNKNOWN) {
	return STACK_UNKNOWN;
    }

    FOREACH_INSTR_IN_BBL(bbl, iterator) {

	if (iterator == instr) {
	    return diff;
	}

	synch = InstrSynchronizesStack(iterator);
	if (synch != STACK_UNKNOWN) {
	    diff = -1 * synch;
	} else {
	    gain = ComputeMyStackGainForInstr(iterator);
	    if (gain == STACK_UNKNOWN) {
		diff = gain;
	    } else if (diff != STACK_UNKNOWN) {
		diff = diff + gain;
	    }
	}

    }			/* end iteration over instructions in basic block */

    WARNING("[InstrBPSPDiff] Control reached end of function.");

    return -1;
}



/*
 *  UpdateFunction(fun) --
 *
 *  Updates the stack height for 'fun', and updates flags as well.
 */
void UpdateFunction(Function *fun) {
    int temp;
    EdgeNode *pred, *succ;

    FUNCTION_SET_FLAG(fun, FUNCTION_LEAVE_RET_COMBO);
    FOREACH_PRED_IN_BBL(fun->f_exit, pred) {
	if (NODE_TYPE(pred) == ET_COMPENSATE) {
	    continue;
	}	 	  
	if (!BBLContainsExitCombo(NODE_FROM(pred))) {
	    FUNCTION_REMOVE_FLAG(fun, FUNCTION_LEAVE_RET_COMBO);
	}
    }

    FUNCTION_SET_FLAG(fun, FUNCTION_GCC_PROLOGUE_FLAG);
    FOREACH_SUCC_IN_BBL(fun->f_entry, succ) {
	if (!BBLContainsPrologue(NODE_TO(succ))) {
	    FUNCTION_REMOVE_FLAG(fun, FUNCTION_GCC_PROLOGUE_FLAG);
	    FUNCTION_REMOVE_FLAG(fun, FUNCTION_WELL_BEHAVED_FLAG);
	}
    }

    temp = FunctionHasBalancedStack(fun);
    if (temp & STACK_BALANCED) {
	FUNCTION_SET_FLAG(fun, FUNCTION_BALANCED_STACK_FLAG);
    } else {
	FUNCTION_REMOVE_FLAG(fun, FUNCTION_BALANCED_STACK_FLAG);
    }
    if (temp & NOT_WELL_BEHAVED) {
	FUNCTION_REMOVE_FLAG(fun, FUNCTION_WELL_BEHAVED_FLAG);
    }

    CalculateSPBPDifferenceForFun(fun);
    StoreInstrBPSPDiffForFunction(fun);

    return;
}



/*
 *  FunctionNetStackGainIsZero(fun) --
 *
 *  Returns true if the function, at all exit points, has a net stack gain of
 *  zero.
 */
static bool FunctionNetStackGainIsZero(Function *fun) {
    Block *bbl;
    EdgeNode *preds;

    FOREACH_PRED_IN_BBL(fun->f_exit, preds) {
	bbl = NODE_FROM(preds);
	if (ComputeNetStackAtEndOfBBL(bbl) != (ADDRESS_SIZE * -2)) {
	    return false;
	}
    }

    return true;
}



/*
 *  ComputeFunctionAndCallerSizes() --
 *
 *  Sets the field FUNCTION_SIZE and (()->f_callerFrameSize) in every function
 *  to denote the maximum stack frame size in the function, and the maximum size
 *  of its callers.
 */
void ComputeFunctionAndCallerSizes() {
    Function *fun;
    INT32 callerStack;
    EdgeNode *predecessors;
    Block *ibbl;
    Instruction *instr;
    int base;

    HELL->b_fun->f_stackFrameSize = 0;

    FOREACH_FUNCTION(fun) {
	fun->f_stackFrameSize = ComputeStackFrameSize(fun);
    }

    FOREACH_FUNCTION(fun) {
	callerStack = 0;
	FOREACH_PRED_IN_BBL(fun->f_entry, predecessors) {
	    if (NODE_FROM(predecessors)->b_fun->f_stackFrameSize > callerStack){
		callerStack = NODE_FROM(predecessors)->b_fun->f_stackFrameSize;
	    }
	}

	/* if callerStack == 0, then we can inspect the instructions to see
	 * how far the caller stack must extend in order to be legal here */
	if (callerStack == 0) {
	    FOREACH_BBL_IN_FUN(fun, ibbl) {
		FOREACH_INSTR_IN_BBL(ibbl, instr) {
		    if ((instr->i_dest) &&
			OperandAccessesStackFromFramePointer(instr->i_dest)) {
			base = instr->i_dest->a_displacement;
			if (base > callerStack) {
			    callerStack = base;
			}
		    }

		    if ((instr->i_src) &&
			OperandAccessesStackFromFramePointer(instr->i_src)) {
			base = instr->i_src->a_displacement;
			if (base > callerStack) {
			    callerStack = base;
			}
		    }

		    if ((instr->i_src2) &&
			OperandAccessesStackFromFramePointer(instr->i_src2)) {
			base = instr->i_src2->a_displacement;
			if (base > callerStack) {
			    callerStack = base;
			}
		    }

		}
	    }
	}

	fun->f_callerFrameSize = callerStack;

    }

    return;
}



/*
 *  CountMemoryOperations() --
 *
 *  Returns the weighted count of all memory operations in the executable.
 */
long long CountMemoryOperations() {
    Instruction *instr;
    long long weight = 0;

    FOREACH_INSTR(instr) {
	if ((InstrPushBytes(instr) > 0) || (InstrPopBytes(instr) > 0)) {
	    weight = weight + instr->i_Bblock->b_weight;
	}
	if (instr->i_dest) {
	    if (OperandIsMemory(instr->i_dest)) {
		weight = weight + instr->i_Bblock->b_weight;
	    }
	}
	if (instr->i_src) {
	    if (OperandIsMemory(instr->i_src)) {
		weight = weight + instr->i_Bblock->b_weight;
	    }
	}
	if (instr->i_src2) {
	    if (OperandIsMemory(instr->i_src2)) {
		weight = weight + instr->i_Bblock->b_weight;
	    }
	}
    }

    return weight;
}



/*
 *  UpdateFunctions() --
 *
 *  Updates the stack height for all functions and re-sets their flags.
 */
void UpdateFunctions() {
    Function *fun;

    FOREACH_FUNCTION(fun) {
	UpdateFunction(fun);
    }
}



/*
 *  StoreInstrBPSPDiffForFunction(fun) --
 *
 *  Stores the result of calling InstrBPSPDiff() on every instruction in the
 *  instruction's temp field for the supplied function
 */
void StoreInstrBPSPDiffForFunction(Function *fun) {
    Instruction *instr;
    Block *bbl;

    FOREACH_BBL_IN_FUN(fun, bbl) {
	int diff = bbl->b_BPSPdiff;
	int gain;
	int synch;

	FOREACH_INSTR_IN_BBL(bbl, instr) {

	    instr->i_temp = diff;
	    INSTR_SET_FLAG(instr, INSTR_BPSP_DIFF_STORED_FLAG);

	    synch = InstrSynchronizesStack(instr);
	    if (synch != STACK_UNKNOWN) {
		diff = -1 * synch;
	    } else {
		gain = ComputeMyStackGainForInstr(instr);
		if (gain == STACK_UNKNOWN) {
		    diff = gain;
		} else if (diff != STACK_UNKNOWN) {
		    diff = diff + gain;
		}
	    }
	}			/* end iteration over instructions in bbl */
    }				/* end iteration over bbls in function */

    return;
}



/*
 *  StoreInstrBPSPDiff() --
 *
 *  Stores the result of calling InstrBPSPDiff() on every instruction in the
 *  instruction's temp field
 */
void StoreInstrBPSPDiff() {
    Function *fun;

    FOREACH_FUNCTION(fun) {
	StoreInstrBPSPDiffForFunction(fun);
    }

    return;
}



/*
 *  MarkFunsDefiningFramePointer() --
 *
 *  Sets FUNCTION_REDEFINES_FP_FLAG in all functions where %ebp is defined
 *  outside of the epilogue or prologue. Sets FUNCTION_MALICIOUSLY_REDEFINES_
 *  FP_FLAG in all functions where %ebp is defined outside of the epilogue
 *  or prologue, and the re-definition was not introduced by the inlining
 *  mechanism to compensate for a use of %ebp.
 */
void MarkFunsDefiningFramePointer() {
    Function *fiterator;
    Instruction *iiterator;
    Block *biterator;
    EdgeNode *niterator;

    FOREACH_FUNCTION(fiterator) {
	FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_REDEFINES_FP_FLAG);
	FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG);

	FOREACH_BBL_IN_FUN(fiterator, biterator) {
	    BBL_REMOVE_FLAG(biterator, BBL_MARKED_FLAG);
	}

	if (FUNCTION_HAS_FLAG(fiterator, FUNCTION_GCC_PROLOGUE_FLAG)) {
	    FOREACH_SUCC_IN_BBL(fiterator->f_entry, niterator) {
		BBL_SET_FLAG(NODE_TO(niterator), BBL_MARKED_FLAG);
	    }
	}

	if (FUNCTION_HAS_FLAG(fiterator, FUNCTION_EXIT_COMBO_FLAG)) {
	    FOREACH_PRED_IN_BBL(fiterator->f_exit, niterator) {
		BBL_SET_FLAG(NODE_FROM(niterator), BBL_MARKED_FLAG);
	    }
	}

	FOREACH_BBL_IN_FUN(fiterator, biterator) {
	    if (BBL_HAS_FLAG(biterator, BBL_MARKED_FLAG)) {
		continue;
	    }

	    FOREACH_INSTR_IN_BBL(biterator, iiterator) {
		if (INSTR_HAS_FLAG(iiterator, INSTR_DEFINES_FP_NICELY)) {
		    continue;
		}

		if ((iiterator->i_dest != NULL) &&
		    (OperandIsFramePointer(iiterator->i_dest))) {
		    FUNCTION_SET_FLAG(fiterator, FUNCTION_REDEFINES_FP_FLAG);

		    if (!INSTR_HAS_FLAG((iiterator->i_prev),
			    INSTR_USES_FP_BUT_COMPENSATED_FLAG)) {
			FUNCTION_SET_FLAG(fiterator,
			    FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG);
		    }
		}
	    }
	}
    }					// end foreach function
}



/*
 *  BBLContainsPrologue(bbl) --
 *
 *  Returns whether or not the basic block begins with the standard
 *  function prologue code.
 *
 *  bbl: the basic block to check
 *  returns: true if the basic block begins with the prologue.
 */
bool BBLContainsPrologue(Block *bbl) {
    Instruction *first = bbl->b_first;
    Instruction *second = first->i_next;

#define IS_SECOND(x) (isPushClass(x) && OperandIsFramePointer((x)->i_src))

#define IS_FIRST(x) \
    (isMovClass(x) && OperandIsFramePointer((x)->i_dest) \
    && OperandIsStackPointer((x)->i_src))

    if (IS_SECOND(first)) {
	if (IS_FIRST(second)) {
	    return true;
	}
    }

    return false;

#undef IS_SECOND
#undef IS_FIRST
}



/*
 *  FunNumberOfArguments(fun) --
 *
 *  Returns the number of arguments that 'fun' takes. This is a rough estimate.
 */
int FunNumberOfArguments(Function *fun) {
    Block *biterator;
    Instruction *iterator;
    int deepest = 0;

    FOREACH_BBL_IN_FUN(fun, biterator) {
	FOREACH_INSTR_IN_BBL(biterator, iterator) {
	    if ((iterator->i_src)
		&& OperandAccessesStackFromFramePointer(iterator->i_src)) {
		if (iterator->i_src->a_displacement > deepest) {
		    deepest = iterator->i_src->a_displacement;
		}
	    }		/* end test for source1 being in the runtime stack */
	    if ((iterator->i_src2)
		&& OperandAccessesStackFromFramePointer(iterator->i_src2)) {
		if (iterator->i_src2->a_displacement > deepest) {
		    deepest = iterator->i_src2->a_displacement;
		}
	    }		/* end test for source2 being in the runtime stack */
	}		/* end iteration over instructions in bbl */
    }			/* end iteration over bbls in function */

    if (deepest <= ADDRESS_SIZE) {
	return 0;
    }

    /* The first slot down is actually the return address */
    deepest = deepest - ADDRESS_SIZE;
    return deepest / ADDRESS_SIZE;
}



/*
 *  ComputeStackContributionForInstr(instr) --
 */
int ComputeStackContributionForInstr(Instruction *instr) {
    if (isLeaveClass(instr)) {
	return STACK_UNKNOWN;
    }

    if (isCallClass(instr)) {
	return 0;
    }

    if ((instr->i_dest != NULL) && OperandIsStackPointer(instr->i_dest)) {
	if (isAddClass(instr) && OperandIsImmediate(instr->i_src)) {
	    return -1 * instr->i_src->a_displacement;
	} else if (isSubClass(instr) && OperandIsImmediate(instr->i_src)) {
	    return instr->i_src->a_displacement;
	} else {
	    return STACK_UNKNOWN;
	}
    }

    if (InstrPopBytes(instr) != 0) {
	return InstrPopBytes(instr) * -1;
    }

    if (InstrPushBytes(instr) != 0) {
	return InstrPushBytes(instr);
    }

    return 0;
}



/*
 *  ComputeStackContributionsForBBL(bbl) --
 */
int ComputeStackContributionsForBBL(Block *bbl) {
    Instruction *iiterator;
    int gain, diff;

    gain = 0;

    FOREACH_INSTR_IN_BBL(bbl, iiterator) {
	diff = ComputeStackContributionForInstr(iiterator);

	if (diff == STACK_UNKNOWN) {
	    return STACK_UNKNOWN;
	}

	gain += diff;
    }

    return gain;
}



/*
 *  StoreStackContributionsForAllBBLsInFun(fun) --
 */
void StoreStackContributionsForAllBBLsInFun(Function *fun) {
    Block *biterator;

    FOREACH_BBL_IN_FUN(fun, biterator) {
	biterator->b_stack_cont = ComputeStackContributionsForBBL(biterator);
    }
}



/*
 *  StoreStackContributionsForAllBBLs() --
 */
void StoreStackContributionsForAllBBLs() {
    Block *biterator;

    FOREACH_BBL(biterator) {
	biterator->b_stack_cont = ComputeStackContributionsForBBL(biterator);
    }
}



/*
 *  UpdateBBLNetStackGain(bbl) --
 */
bool UpdateBBLNetStackGain(Block *bbl) {
    EdgeNode *predn;
    bool changed;
    int og_bbl_gain, computed_cont;

    changed = false;

    og_bbl_gain = bbl->b_tmp;
    bbl->b_tmp = STACK_UNDEFINED;

    FOREACH_PRED_IN_BBL(bbl, predn) {
	// there's the possibility that we get an escaping edge into this
	// basic block, however we don't treat it specially because the jump
	// that escapes should also have a correct net stack gain and
	// therefore we'll get a good value

	if (NODE_TYPE(predn) == ET_RETURN) {
	    continue;
	}

	if (NODE_FROM(predn)->b_tmp == STACK_UNDEFINED) {
	    continue;
	}

	computed_cont = NODE_FROM(predn)->b_stack_cont;

	if (bbl->b_first != NULL) {
	    DEBUG(DB_STKG, "looking at %x", bbl->b_first->i_address);
	}
	// if this predecessor of this block has an outgoing HELL edge or calls
	// a function that is not well-behaved then we don't know what the net
	// stack gain really is
	if (NODE_FROM(predn) == HELL) {
	    DEBUG(DB_STKG, "hell");
	    computed_cont = STACK_UNKNOWN;
	} else if ((NODE_FROM(predn)->b_last != NULL) &&
	    (NODE_FROM(predn)->b_last->i_prev != NULL) &&
	    (isLeaveClass(NODE_FROM(predn)->b_last->i_prev)) &&
	    FUNCTION_HAS_FLAG(((NODE_FROM(predn))->b_fun),
		FUNCTION_GCC_PROLOGUE_FLAG)) {
	    // if the preceding block is an 'exit' block (both return and
	    // escaping edge) and uses a leave as well as a standard prologue,
	    // we can safely say that the net stack gain I will encounter from
	    // it is just the overhead of the instructions following the leave
	    // (i.e., the ret or jump)
	    DEBUG(DB_STKG, "pred is a leave something");
	    // if the previous basic block had a leave/jmp escaping combo
	    // we know that the net stack contribution is actually just
	    // the stack synch, so we treat it as such
	    if (BBLGetEscapingEdge(NODE_FROM(predn)) != NULL) {
		bbl->b_tmp = StackJoin(4, bbl->b_tmp);
		continue;
	    } else if (bbl == bbl->b_fun->f_exit) {
		bbl->b_tmp = StackJoin(0, bbl->b_tmp);
		continue;
	    }
	} else if (predn->en_edge->e_type == ET_FUNDIRECT) {
	    if (((NODE_TO(predn))->b_fun) == ((NODE_FROM(predn))->b_fun)) {
		DEBUG(DB_STKG, "fundirect pic");
		computed_cont += 4;
	    } else {
		DEBUG(DB_STKG, "fundirect, skipping!");
		continue;
	    }
	} else if (predn->en_edge->e_type == ET_FUNLINK) {
	    Edge *edge;

	    edge = BBLGetFunDirectEdge(NODE_FROM(predn));

	    if (edge != NULL) {
		// pic code will add 4 to the stack
		if (bbl == edge->e_to) {
		    computed_cont += 4;
		    DEBUG(DB_STKG, "pic");
		} else if (edge->e_to->b_fun == bbl->b_fun) {
		    // recursive code we'll just assume that the function is
		    // well-behaved for the function call -- if it's not
		    // we'll find out somewhere else
		    DEBUG(DB_STKG, "recursive!");
		} else {
		    DEBUG(DB_STKG, "none of the above");
		}
	    } else if ((edge == NULL) ||
		!FUNCTION_HAS_FLAG(edge->e_to->b_fun,
		    FUNCTION_BALANCED_STACK_FLAG)) {
		computed_cont = STACK_UNKNOWN;
	    }
	}

	DEBUG(DB_STKG, "adding net_gain: %x comp: %x, node_type: %x",
	    NODE_FROM(predn)->b_tmp, computed_cont, NODE_TYPE(predn));
	DEBUG(DB_STKG, "joining pred: %x: %x and bbl: %x: %x",
	    NODE_FROM(predn),
	    StackAdd(NODE_FROM(predn)->b_tmp, computed_cont),
	    bbl, bbl->b_tmp);
	bbl->b_tmp = StackJoin(StackAdd(NODE_FROM(predn)->b_tmp,
	    computed_cont), bbl->b_tmp);
    }

    return og_bbl_gain != bbl->b_tmp;
}



/*
 *  StoreNetStackContributionsForFun(fun) --
 */
void StoreNetStackContributionsForFun(Function *fun) {
    Block *biterator;
    bool changed;
    int iterations;

// REPORT(RP_LGDETAIL, "Storing Net Stack Contributions for %s", fun->f_name);

    FOREACH_BBL_IN_FUN(fun, biterator) {
	biterator->b_tmp = STACK_UNDEFINED;
    }
    fun->f_entry->b_tmp = 4;
    fun->f_exit->b_tmp = STACK_UNDEFINED;

    iterations = 0;

    changed = true;
    while (changed) {
	changed = false;
	iterations++;

	FOREACH_BBL_IN_FUN(fun, biterator) {
	    changed |= UpdateBBLNetStackGain(biterator);
	}

	changed |= UpdateBBLNetStackGain(fun->f_exit);
    }

    FOREACH_BBL_IN_FUN(fun, biterator) {
	if (biterator->b_tmp == STACK_UNDEFINED) {
	    WARNING("UNDEFINED STACK GAIN AT %x."
		" Suggest removing unreachable code.",
		biterator->b_first->i_address);
	    biterator->b_tmp = STACK_UNKNOWN;
	}
    }
}



/*
 *  StoreInstrBPSPDiffForFun(fun) --
 */
void StoreInstrBPSPDiffForFun(Function *fun) {
    Block *biterator;
    Instruction *iiterator;
    bool known;
    int diff, instr_diff;

    known = FUNCTION_HAS_FLAG(fun, FUNCTION_GCC_PROLOGUE_FLAG);

    FOREACH_BBL_IN_FUN(fun, biterator) {
	known = FUNCTION_HAS_FLAG(fun, FUNCTION_GCC_PROLOGUE_FLAG);

	if (known) {
	    diff = biterator->b_tmp;

	    if (diff != STACK_UNKNOWN) {
		diff -= 8;
	    } else {
		known = false;
	    }
	} else {
	    diff = biterator->b_tmp;
	}

	FOREACH_INSTR_IN_BBL(biterator, iiterator) {
	    iiterator->i_temp = diff;
	    INSTR_SET_FLAG(iiterator, INSTR_BPSP_DIFF_STORED_FLAG);

	    if (diff == STACK_UNKNOWN) {
		continue;
	    }

	    instr_diff = ComputeStackContributionForInstr(iiterator);

	    if (instr_diff != STACK_UNKNOWN) {
		diff += instr_diff;
	    } else {
		diff = STACK_UNKNOWN;
	    }
	}
    }
}



/*
 *  StoreInstrBPSPDiffForAllFuns() --
 */
void StoreInstrBPSPDiffForAllFuns() {
    Function *fiterator;

    FOREACH_FUNCTION(fiterator) {
	StoreInstrBPSPDiffForFun(fiterator);
    }
}



/*
 *  MarkExitComboFuns() --
 */
void MarkExitComboFuns() {
    Function *fiterator;
    EdgeNode *predn;

    FOREACH_FUNCTION(fiterator) {
	FUNCTION_SET_FLAG(fiterator, FUNCTION_EXIT_COMBO_FLAG);

	FOREACH_PRED_IN_BBL(fiterator->f_exit, predn) {
	    if (!BBLContainsExitCombo(NODE_FROM(predn))) {
		FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_EXIT_COMBO_FLAG);
	    }
	}
    }
}



/*
 *  MarkPrologueFuns() --
 */
void MarkPrologueFuns() {
    Function *fiterator;
    EdgeNode *succn;

    FOREACH_FUNCTION(fiterator) {
	FUNCTION_SET_FLAG(fiterator, FUNCTION_GCC_PROLOGUE_FLAG);

	if (fiterator->f_entry->b_succ == NULL) {
	    FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_GCC_PROLOGUE_FLAG);
	} else {
	    FOREACH_SUCC_IN_BBL(fiterator->f_entry, succn) {
		if (!BBLContainsPrologue(NODE_TO(succn))) {
		    FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_GCC_PROLOGUE_FLAG);
		}
	    }
	}
    }
}



/*
 *  StoreNetStackContributionsForAllFuns() --
 */
void StoreNetStackContributionsForAllFuns() {
    Function *fiterator;
    bool changed;
    int og_fun_gain;

    StoreStackContributionsForAllBBLs();

    changed = true;
    while (changed) {
	changed = false;

	FOREACH_FUNCTION(fiterator) {
	    og_fun_gain = fiterator->f_exit->b_tmp;

	    StoreNetStackContributionsForFun(fiterator);

	    if (og_fun_gain != fiterator->f_exit->b_tmp) {
		changed = true;
	    }

	    if (fiterator->f_exit->b_tmp == 0) {
		FUNCTION_SET_FLAG(fiterator, FUNCTION_BALANCED_STACK_FLAG);
	    }
	}
    }
}



/*
 *  MarkWellBehavedFunctions2() --
 */
void MarkWellBehavedFunctions2() {
    Function *fiterator;
    bool changed;
    int og_fun_gain;

    StoreStackContributionsForAllBBLs();

    MarkPrologueFuns();
    MarkExitComboFuns();
    MarkFunsDefiningFramePointer();

    // anything with both the gcc prologue and proper exit combination is
    // well behaved if it doesn't have any escaping edges or maliciously
    // define the frame pointer (almost no functions maliciously define the
    // frame pointer, however, not identifying an exit combo results in our
    // analysis thinking they are malicious...)
    FOREACH_FUNCTION(fiterator) {
	FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_WELL_BEHAVED_FLAG);
	FUNCTION_REMOVE_FLAG(fiterator, FUNCTION_BALANCED_STACK_FLAG);

	if (FUNCTION_HAS_FLAG(fiterator, FUNCTION_GCC_PROLOGUE_FLAG) &&
	    FUNCTION_HAS_FLAG(fiterator, FUNCTION_EXIT_COMBO_FLAG) &&
	    !FUNCTION_HAS_FLAG(fiterator,
		FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG)
	    && !FunHasEsc2(fiterator)) {
	    FUNCTION_SET_FLAG(fiterator, FUNCTION_WELL_BEHAVED_FLAG);
	    FUNCTION_SET_FLAG(fiterator, FUNCTION_BALANCED_STACK_FLAG);
	}
    }

    changed = true;
    while (changed) {
	changed = false;

	FOREACH_FUNCTION(fiterator) {
	    og_fun_gain = fiterator->f_exit->b_tmp;

	    StoreNetStackContributionsForFun(fiterator);

	    if (og_fun_gain != fiterator->f_exit->b_tmp) {
		changed = true;
	    }

	    if (fiterator->f_exit->b_tmp == 0) {
		FUNCTION_SET_FLAG(fiterator, FUNCTION_BALANCED_STACK_FLAG);
	    }

	    if (FunHasEsc2(fiterator)) {
		Block *biterator;
		Edge *edge;
		bool all_balanced_escapers;

		all_balanced_escapers = true;

		FOREACH_BBL_IN_FUN(fiterator, biterator) {
		    edge = BBLGetEscapingEdge(biterator);

		    if (edge != NULL) {
			if ((biterator->b_last->i_prev == NULL) ||
			    !isLeaveClass(biterator->b_last->i_prev)) {
			    all_balanced_escapers = false;
			} else {
			    if (!FUNCTION_HAS_FLAG(edge->e_to->b_fun,
				    FUNCTION_BALANCED_STACK_FLAG)) {
				REPORT(RP_LGDETAIL,
				    "Setting %s to unbalanced because"
				    " escaping target %s is unbalanced",
				    fiterator->f_name,
				    edge->e_to->b_fun->f_name);
				all_balanced_escapers = false;
			    }
			}
		    }
		}

		if (!all_balanced_escapers
		    && FUNCTION_HAS_FLAG(fiterator,
			FUNCTION_BALANCED_STACK_FLAG)) {
		    FUNCTION_REMOVE_FLAG(fiterator,
			FUNCTION_BALANCED_STACK_FLAG);
		} else if (all_balanced_escapers) {
		    if (fiterator->f_exit->b_tmp == STACK_UNDEFINED) {
			if (!FUNCTION_HAS_FLAG(fiterator,
				FUNCTION_BALANCED_STACK_FLAG)) {
			    changed = true;
			}

			FUNCTION_SET_FLAG(fiterator,
			    FUNCTION_BALANCED_STACK_FLAG);
		    }
		}
	    }
	}
    }

    PrintStackAnalysisSummary();
}



/*
 *  StackAdd(a,b) --
 */
int StackAdd(int a, int b) {
    if ((a == STACK_UNKNOWN) || (b == STACK_UNKNOWN)) {
	return STACK_UNKNOWN;
    }

    if (a == STACK_UNDEFINED) {
	return b;
    }

    if (b == STACK_UNDEFINED) {
	return a;
    }

    return a + b;
}



/*
 *  StackJoin(a,b) --
 *
 *  Returns the join of its two arguments, according to the following
 *  lattice:
 *
 *        STACK_UNKNOWN
 *   ... /      |      \ ...
 *  ... -1      0       1 ...
 *   ... \      |      / ...
 *       STACK_UNDEFINED
 */
int StackJoin(int a, int b) {
    if (a == STACK_UNDEFINED) {
	return b;
    }
    if (b == STACK_UNDEFINED) {
	return a;
    }

    if (a == STACK_UNKNOWN || b == STACK_UNKNOWN) {
	return STACK_UNKNOWN;
    }

    if (a == b) {
	return a;
    }

    return STACK_UNKNOWN;
}



/*
 *  PrintStackAnalysisSummary() --
 */
void PrintStackAnalysisSummary() {
    Function *fiterator;
    int total,
	well_behaved,
	balanced_stack, no_prologue, no_exit_combo, no_exits, mal_fp, esc_edge;

    REPORT(RP_DETAIL, "Stack analysis summary");

    total = well_behaved = balanced_stack = no_prologue = no_exit_combo =
	no_exits = mal_fp = esc_edge = 0;

    FOREACH_FUNCTION(fiterator) {
	total++;

	if (FUNCTION_HAS_FLAG(fiterator, FUNCTION_WELL_BEHAVED_FLAG)) {
	    REPORT(RP_LGLOOP, "Function %-10s is well-behaved",
		fiterator->f_name);
	    well_behaved++;
	    continue;
	} else if (FUNCTION_HAS_FLAG(fiterator, FUNCTION_BALANCED_STACK_FLAG)) {
	    REPORT(RP_LGLOOP, "Function %-10s has balanced stack",
		fiterator->f_name);
	    balanced_stack++;
	} else {
	    if (!FUNCTION_HAS_FLAG(fiterator, FUNCTION_GCC_PROLOGUE_FLAG)) {
		REPORT(RP_LGLOOP, "Function %-10s has no prologue",
		    fiterator->f_name);
		no_prologue++;
	    }

	    if (!FUNCTION_HAS_FLAG(fiterator, FUNCTION_EXIT_COMBO_FLAG)) {
		REPORT(RP_LGLOOP, "Function %-10s has no exit combo",
		    fiterator->f_name);
		no_exit_combo++;
	    }

	    if (FUNCTION_HAS_FLAG(fiterator,
		    FUNCTION_MALICIOUSLY_REDEFINES_FP_FLAG)) {
		REPORT(RP_LGLOOP, "Function %-10s maliciously redefines FP",
		    fiterator->f_name);
		mal_fp++;
	    }

	    if (FunHasEsc2(fiterator)) {
		REPORT(RP_LGLOOP, "Function %-10s has escaping edges",
		    fiterator->f_name);
		esc_edge++;
	    }

	    if (fiterator->f_exit->b_pred == NULL) {
		REPORT(RP_LGLOOP, "Function %-10s has no exit block",
		    fiterator->f_name);
		no_exits++;
	    }
	}
    }

#define PERCENTAGE(x, y) (x), 100 * ((double)x) / (y)

    REPORT(RP_DETAIL, "Total Functions:               %5d", total);
    REPORT(RP_DETAIL, "Well Behaved Functions:        %5d (%lf%%)",
	PERCENTAGE(well_behaved, total));
    REPORT(RP_DETAIL, "Balanced Stack:                %5d (%lf%%)",
	PERCENTAGE(balanced_stack + well_behaved, total));
    REPORT(RP_DETAIL, "Unbalanced Stack:              %5d (%lf%%)",
	PERCENTAGE(total - well_behaved - balanced_stack, total));
    REPORT(RP_DETAIL, "  Escaping Edge(s):            %5d (%lf%%) %5d (%lf%%)",
	PERCENTAGE(esc_edge, total),
	PERCENTAGE(esc_edge, total - well_behaved - balanced_stack));
    REPORT(RP_DETAIL, "  No Prologue:                 %5d (%lf%%) %5d (%lf%%)",
	PERCENTAGE(no_prologue, total),
	PERCENTAGE(no_prologue, total - well_behaved - balanced_stack));
    REPORT(RP_DETAIL, "  No Exit Combo:               %5d (%lf%%) %5d (%lf%%)",
	PERCENTAGE(no_exit_combo, total),
	PERCENTAGE(no_exit_combo, total - well_behaved - balanced_stack));
    REPORT(RP_DETAIL, "  Malicious FP Define(s):      %5d (%lf%%) %5d (%lf%%)",
	PERCENTAGE(mal_fp, total),
	PERCENTAGE(mal_fp, total - well_behaved - balanced_stack));
    REPORT(RP_DETAIL, "  No Exit Block:               %5d (%lf%%) %5d (%lf%%)",
	PERCENTAGE(no_exits, total),
	PERCENTAGE(no_exits, total - well_behaved - balanced_stack));
}
