/*
 *  sysdep.c
 *
 *  System dependency handler. Everything that is architecture,
 *  system or executable dependent should go in the handler.  It
 *  will re-route the functions to the appropriate machine-dependent
 *  handlers based on what architecture is defined.
 */
#include "global.h"
#include "sysdep.h"
#include "upm.h"

#ifdef X86
    #include "x86-asm.h"
    #include "x86-classes.h"
    #include "x86.h"
    #include "x86-liveness.h"
    #include "x86-table-defs.h"
    #include "x86-util.h"
    #include "../x86/x86-dis.h"
    #include "../x86/x86-histogram.h"
    #include "../x86/x86-icreate.h"
    #include "../x86/x86-inline.h"
    #include "../x86/x86-iprint.h"
    #include "../x86/x86-jt.h"
    #include "../x86/x86-op.h"
    #include "../x86/x86-profdump.h"
    #include "../x86/x86-schedule.h"
#endif



/*
 *  CalculateInstrLength(instr) --
 *
 *  Returns the length of the instruction, but does not change anything
 *  in the instruction's internal fields.
 */
int CalculateInstrLength(Instruction *instr) {
    return xCalculateInstrLength(instr);
}



/*
 *  UpdateInstrLength(instr) --
 *
 *  Sets the length of the instruction by calling CalculateInstrLength
 */
void UpdateInstrLength(Instruction *instr) {
    xUpdateInstrLength(instr);
}



/*
 *  isValidInstr(instr) --
 *
 *  Returns true if the instruction can be represented by a particular
 *  architecture.
 */
bool isValidInstr(Instruction *instr) {
    return xIsValidInstr(instr);
}



/*
 *  InstrLocateJTBaseAddress(instr) --
 *
 *  Given an instruction that jumps through a jump table,
 *  InstrLocateJTBaseAddress will trace backward through predecessor
 *  instructions and attempt to find the base address of the table.
 */
UINT32 InstrLocateJTBaseAddress(Instruction *instr) {
    return xInstrLocateJTBaseAddress(instr);
}



/*
 *  InstrLocateJTBaseInstr(instr) --
 *
 *  Given an instruction that jumps through a jump table,
 *  InstrLocateJTBaseInstr returns the instruction that loads/uses
 *  the base address of the jump table.
 */
Instruction *InstrLocateJTBaseInstr(Instruction *instr) {
    return xInstrLocateJTBaseInstr(instr);
}



/*
 *  InstrLocateJTBoundsCheck(instr) --
 *
 *  Given an instruction that jumps through a jump table,
 *  InstrLocateJTBoundsCheck will trace backward and return
 *  the instruction that performs the bounds check on the JT
 */
Instruction *InstrLocateJTBoundsCheck(Instruction *instr) {
    return xInstrLocateJTBoundsCheck(instr);
}



/*
 *  InstrLocateJTSize(instr) --
 *
 *  Given an instruction that jumps through a jump table,
 *  InstrLocateJTSize will trace backward through predecessor instructions
 *  and attempt to find the size of the table
 */
UINT32 InstrLocateJTSize(Instruction *instr) {
    return xInstrLocateJTSize(instr);
}



/*
 *  InsertDumpCode(filename) --
 *
 *  Inserts code to dump out nbytes of profiling data starting from
 *  address start.  Usually this code is inserted as a call from an
 *  _exit or exit routine
 */
void InsertDumpCode(char *filename) {
    xInsertDumpCode(filename);
}



/*
 *  DisassembleInstr(pc,b,instr) --
 *
 *  Disassembles the data pointed to by 'data', and fills in 'instr' with
 *  relevant information.
 */
int DisassembleInstr(ADDRESS pc, UINT8 *b, Instruction *instr) {
    return xDisassembleInstr(pc, b, instr);
}



/*
 *  InstrPushBytes(instr) --
 *
 *  Returns the number of bytes which the instr pushes on the runtime stack
 */
int InstrPushBytes(Instruction *instr) {
    return xInstrPushBytes(instr);
}



/*
 *  InstrPopBytes(instr) --
 *
 *  Returns the number of bytes which the instruction pops off the runtime stack
 */
int InstrPopBytes(Instruction *instr) {
    return xInstrPopBytes(instr);
}



/*
 *  InstrCreateNOP() --
 *
 *  Returns a NOP
 */
Instruction *InstrCreateNOP() {
    return xInstrCreateNOP();
}



/*
 *  InstrCreateJMP() --
 *
 *  Returns an Unconditional Jump
 */
Instruction *InstrCreateJMP() {
    return xInstrCreateJMP();
}



/*
 *  InstrCreateSave() --
 *
 *  Returns an instruction to push the PSR/PSW onto the stack
 */
Instruction *InstrCreateSave() {
    return xInstrCreateSave();
}



/*
 *  InstrCreateSPInc() --
 *
 *  Returns an instruction to increment the stack pointer.
 *  The immediate field is 0.
 */
Instruction *InstrCreateSPInc() {
    return xInstrCreateSPInc();
}



/*
 *  InstrCreateSPDec() --
 *
 *  Returns an instruction to decrement the stack pointer.
 *  The immediate field is 0.
 */
Instruction *InstrCreateSPDec() {
    return xInstrCreateSPDec();
}



/*
 *  InstrCreateRestore() --
 *
 *  Returns an instruction to pop top of stack into PSR/PSW.
 */
Instruction *InstrCreateRestore() {
    return xInstrCreateRestore();
}



/*
 *  InstrCreate32BitMemINC(addr) --
 *
 *  Returns an instruction to increment a 32-bit value
 *  at the given memory address.
 */
Instruction *InstrCreate32BitMemINC(ADDRESS addr) {
    return _InstrCreate32BitMemINC(addr);
}



/*
 *  InstrCreate32BitMemAdc(addr) --
 *
 *  Returns an instruction that adds a 32-bit value
 *  (+ the carry flag) to a memory address.
 */
Instruction *InstrCreate32BitMemAdc(ADDRESS addr) {
    return _InstrCreate32BitMemAdc(addr);
}



/*
 *  InstrCreate32BitMemAdd(addr) --
 *
 *  Returns an instruction that adds a 32-but constant to a memory address
 */
Instruction *InstrCreate32BitMemAdd(ADDRESS addr) {
    return _InstrCreate32BitMemAdd(addr);
}



/*
 *  InstrCreateMovImmToReg(reg,value) --
 *
 *  Returns an instruction that moves an immediate value to a GP register
 */
Instruction *InstrCreateMovImmToReg(int reg, int value) {
    return xInstrCreateMovImmToReg(reg, value);
}



/*
 *  InstrCreatePushAll() --
 *
 *  Returns an instruction that pushs all registers onto the stack
 */
Instruction *InstrCreatePushAll() {
    return xInstrCreatePushAll();
}



/*
 *  InstrCreatePush(reg) --
 *
 *  Returns an instruction that pushs the value in register reg onto the
 *  stack
 */
Instruction *InstrCreatePush(int reg) {
    return xInstrCreatePush(reg);
}



/*
 *  InstrCreatePop(reg) --
 *
 *  Returns an instruction that pops the top value into register reg
 */
Instruction *InstrCreatePop(int reg) {
    return xInstrCreatePop(reg);
}



/*
 *  InstrCreatePopAll() --
 *
 *  Returns an instruction that pops all register (EXCEPT ESI) from stack
 */
Instruction *InstrCreatePopAll() {
    return xInstrCreatePopAll();
}



/*
 *  InstrCreatePushImm(value) --
 *
 *  Returns an instruction that pushes an immediate value onto the stack
 */
Instruction *InstrCreatePushImm(int value) {
    return xInstrCreatePushImm(value);
}



/*
 *  InstrCreatePushMem(address) --
 *
 *  Returns an instruction that pushes the contents of a memory location onto
 *  the stack
 */
Instruction *InstrCreatePushMem(int address) {
    return xInstrCreatePushMem(address);
}



/*
 *  InstrCreatePushMemBase(reg) --
 *
 *  Returns an instruction that pushes the contents of memory addressed by
 *  the contents of reg onto the stack.
 */
Instruction *InstrCreatePushMemBase(int reg) {
    return xInstrCreatePushMemBase(reg);
}



/*
 *  InstrCreatePushMemDisp8Base(disp8, reg) --
 *
 *  Returns an instruction that pushes the contents of memory addressed by
 *  the contents of reg + disp8 onto the stack.
 */
Instruction *InstrCreatePushMemDisp8Base(INT8 disp8, int reg) {
    return xInstrCreatePushMemDisp8Base(disp8, reg);
}



/*  
 *  InstrCreatePushMemDisp32Base(disp32, reg) --
 *
 *  Returns an instruction that pushes the contents of memory addressed by
 *  the contents of reg + disp32 onto the stack.
 */
Instruction *InstrCreatePushMemDisp32Base(INT32 disp32, int reg) {
    return xInstrCreatePushMemDisp32Base(disp32, reg);
}



/*
 *  InstrCreateCmpRegToValue(reg,value) --
 *
 *  Returns an instruction that compares a register to an immediate value
 */
Instruction *InstrCreateCmpRegToValue(int reg, int value) {
    return xInstrCreateCmpRegToValue(reg, value);
}



/*
 *  InstrCreateCmovRegToReg() --
 *
 *  Creates a conditional register to register move instruction
 */
Instruction *InstrCreateCmovRegToReg() {
    return xInstrCreateCmovRegToReg();
}



/*
 *  InstrCreateCmovMemToReg() --
 *
 *  Creates a conditional memory to register move instruction
 */
Instruction *InstrCreateCmovMemToReg() {
    return xInstrCreateCmovMemToReg();
}



/*
 *  InstrCreateCmpStackToValue(reg,value) --
 *
 *  Returns an instruction that compares a stack offset to an immediate value
 */
Instruction *InstrCreateCmpStackToValue(int reg, int value) {
    return xInstrCreateCmpStackToValue(reg, value);
}



/*
 *  InstrCreateCmpImm8WithReg(imm8, reg) --
 *
 *  Returns an instruction that compares one byte immediate with the contents
 *  of register reg
 */
Instruction *InstrCreateCmpImm8WithReg(INT8 imm8, int reg) {
    return xInstrCreateCmpImm8WithReg(imm8, reg);
}



/*
 *  InstrCreateCmpImm32WithReg(imm32, reg) --
 *
 *  Returns an instruction that compares four byte immediate with the contents
 *  of register reg
 */
Instruction *InstrCreateCmpImm32WithReg(INT32 imm32, int reg) {
    return xInstrCreateCmpImm32WithReg(imm32, reg);
}



/*
 *  InstrCreateCmpImm8WithMemAbs(imm8, offset) --
 *
 *  Returns an instruction that compares one byte immediate with the contents
 *  of memory location addressed by offset.
 */
Instruction *InstrCreateCmpImm8WithMemAbs(INT8 imm8, int offset) {
    return xInstrCreateCmpImm8WithMemAbs(imm8, offset);
}



/*
 *  InstrCreateCmpImm32WithMemAbs(imm32, offset) --
 *
 *  Returns an instruction that compares four byte immediate with the contents
 *  of memory location addressed by offset.
 */
Instruction *InstrCreateCmpImm32WithMemAbs(INT32 imm32, int offset) {
    return xInstrCreateCmpImm32WithMemAbs(imm32, offset);
}



/*
 *  InstrCreateCmpImm8WithMemBase(imm8, reg) --
 *
 *  Returns an instruction that compares one byte immediate with the contents
 *  of memory location addressed by the contents of reg.
 */
Instruction *InstrCreateCmpImm8WithMemBase(INT8 imm8, int reg) {
    return xInstrCreateCmpImm8WithMemBase(imm8, reg);
}



/*
 *  InstrCreateCmpImm32WithMemBase(imm32, reg) --
 *
 *  Returns an instruction that compares one byte immediate with the contents
 *  of memory location addressed by the contents of reg.
 */
Instruction *InstrCreateCmpImm32WithMemBase(INT32 imm32, int reg) {
    return xInstrCreateCmpImm32WithMemBase(imm32, reg);
}



/*
 *  InstrCreateCmpImm8WithMemDisp8Base(imm8, disp8, reg) --
 *
 *  Returns an instruction that compares one byte immediate with the contents
 *  of memory location addressed by the contents of reg and disp8.
 */
Instruction *InstrCreateCmpImm8WithMemDisp8Base(INT8 imm8, INT8 disp8,
    int reg) {
    return xInstrCreateCmpImm8WithMemDisp8Base(imm8, disp8, reg);
}



/*
 *  InstrCreateCmpImm32WithMemDisp8Base(imm32, disp8, reg) --
 *
 *  Returns an instruction that compares four byte immediate with the contents
 *  of memory location addressed by the contents of reg and disp8.
 */
Instruction *InstrCreateCmpImm32WithMemDisp8Base(INT32 imm32, INT8 disp8,
    int reg) {
    return xInstrCreateCmpImm32WithMemDisp8Base(imm32, disp8, reg);
}



/*
 *  InstrCreateCmpImm8WithMemDisp32Base(imm8, disp32, reg) --
 *
 *  Returns an instruction that compares one byte immediate with the contents
 *  of memory location addressed by the contents of reg and disp32.
 */
Instruction *InstrCreateCmpImm8WithMemDisp32Base(INT8 imm8, INT32 disp32,
    int reg) {
    return xInstrCreateCmpImm8WithMemDisp32Base(imm8, disp32, reg);
}



/*
 *  InstrCreateCmpImm32WithMemDisp32Base(imm32, disp32, reg) --
 *
 *  Returns an instruction that compares four byte immediate with the contents
 *  of memory location addressed by the contents of reg and disp32.
 */
Instruction *InstrCreateCmpImm32WithMemDisp32Base(INT32 imm32, INT32 disp32,
    int reg) {
    return xInstrCreateCmpImm32WithMemDisp32Base(imm32, disp32, reg);
}



/*
 *  InstrCreateCmpImm8WithMemBaseIndex(imm8, base, index, scale) --
 *
 *  Returns an instruction that compares one byte immediate with the contents
 *  of memory location addressed by the contents of base register, index
 *  register and scale.
 */
Instruction *InstrCreateCmpImm8WithMemBaseIndex(INT8 imm8, int base,
    int index, int scale) {
    return xInstrCreateCmpImm8WithMemBaseIndex(imm8, base, index, scale);
}



/*
 *  InstrCreateCmpImm32WithMemBaseIndex(imm32, base, index, scale) --
 *
 *  Returns an instruction that compares four byte immediate with the contents
 *  of memory location addressed by the contents of base register, index
 *  register and scale.
 */
Instruction *InstrCreateCmpImm32WithMemBaseIndex(INT32 imm32, int base,
    int index, int scale) {
    return xInstrCreateCmpImm32WithMemBaseIndex(imm32, base, index, scale);
}



/*
 *  InstrCreateJne() --
 *
 *  Returns an instruction that will jump if the ZF=0
 */
Instruction *InstrCreateJne() {
    return xInstrCreateJne();
}



/*
 *  InstrCreateJe() --
 *
 *  Returns an instruction that will jump if the ZF=1
 */
Instruction *InstrCreateJe() {
    return xInstrCreateJe();
}



/*
 *  InstrCreateMovRegToReg(srcreg,destreg) --
 *
 *  Creates an instruction that moves the 'srcreg' register into 'destreg'
 */
//Instruction *InstrCreateMovRegToReg(int srcreg, int destreg) {
//    return xInstrCreateMovRegToReg(srcreg, destreg);
//}



/*
 *  UpdateInstr(instr) --
 *
 *  Updates information about the instruction.  This function is a little
 *  ambiguous, and some architectures may to define it as simply doing nothing.
 *  The function is called after changes are made to the instruction.
 *  For instance, if a displacement is changed from 0x56 to 0x58,
 *  we call this function.  false is returned if something
 *  terrible happened that prevented the instruction from being updated.
 *  Errors here are usually fatal and should not be ignored.
 */
bool UpdateInstr(Instruction *instr) {
    return xUpdateInstr(instr);
}



/*
 *  OperandUsesRegister(op) --
 *
 *  Returns true if the operand uses a register
 */
bool OperandUsesRegister(Operand *op) {
    return xOperandUsesRegister(op);
}



/*
 *  PrintInstruction(instr) --
 *
 *  Returns a textual representation of the instruction
 */
char *PrintInstruction(Instruction *instr) {
    return xPrintInstruction(instr);
}



/*
 *  PrintInstructionFinal(instr) --
 *
 *  Returns a textual representation of the instruction, using its final address
 */
char *PrintInstructionFinal(Instruction *instr) {
    return xPrintInstructionFinal(instr);
}



/*
 *  PrintRegister(rbv) --
 *
 *  Returns a textual representation of a bitregister
 */
char *PrintRegister(int rbv) {
    return xPrintRegister(rbv);
}



/*
 *  ReportSuspiciousInstrs() --
 *
 *  This function is ambiguous.  It gives specific architectures the chance
 *  to scan through the instruction list and report anything that looks odd
 */
void ReportSuspiciousInstrs() {
    xReportSuspiciousInstrs();
}



/*
 *  isConditionalInstr(inst) --
 */
bool isConditionalInstr(Instruction *inst) {
    return xIsConditionalInstr(inst);
}



/*
 *  isUnconditionalInstr(inst) --
 */
bool isUnconditionalInstr(Instruction *inst) {
    return xIsUnconditionalInstr(inst);
}



/*
 *  RegisterIs32Bit(aReg) --
 *
 *  Returns true if the register is a 32-bit register
 */
bool RegisterIs32Bit(int aReg) {
    return _RegisterIs32Bit(aReg);
}



/*
 *  RegisterIs16Bit(aReg) --
 *
 *  Returns true if the register is a 16-bit register
 */
bool RegisterIs16Bit(int aReg) {
    return _RegisterIs16Bit(aReg);
}



/*
 *  RegisterIs8Bit(aReg) --
 *
 *  Returns true if the register is a 8-bit register
 */
bool RegisterIs8Bit(int aReg) {
    return _RegisterIs8Bit(aReg);
}



/*
 *  AlignLoops() --
 *
 *  Aligns loops on cache-line boundaries
 */
bool AlignLoops() {
    return xAlignLoops();
}



/*
 *  FunIsInlinable(aFun) --
 *
 *  Returns true if the given function is inlinable
 */
bool FunIsInlinable(Function *aFun) {
    return xFunIsInlinable(aFun);
}



/*
 *  FunInline(callInstr) --
 *
 *  Inlines the function call made from callInstr
 */
void FunInline(Instruction *callInstr) {
    xFunInline(callInstr);
}



/*
 *  FunCanBeInlinedFromCallSite(fun,callInstr) --
 *
 *  Returns true if the given function can be inlined from the call site
 */
bool FunCanBeInlinedFromCallSite(Function *fun, Instruction *callInstr) {
    return xFunCanBeInlinedFromCallSite(fun, callInstr);
}



/*
 *  getIcacheSize() --
 *
 *  Returns the size of the instruction cache
 */
int getIcacheSize() {
    return xGetIcacheSize();
}



/*
 *  getDcacheSize() --
 *
 *  Returns the size of the data cache
 */
int getDcacheSize() {
    return xGetDcacheSize();
}



/*
 *  getMegahertz() --
 *
 *  Returns the megahertz rating of the processor
 */
int getMegahertz() {
    return xGetMegahertz();
}



/*
 *  getIcacheMissPenalty() --
 *
 *  Returns the penalty, expressed as the number of clock ticks,
 *  for an icache miss
 */
int getIcacheMissPenalty() {
    return xGetIcacheMissPenalty();
}



/*
 *  getCallPenalty() --
 *
 *  Returns the penalty, expressed as the number of clock ticks, for a CALL
 */
int getCallPenalty() {
    return xGetCallPenalty();
}



/*
 *  BuildOpcodeHashTable() --
 *
 *  Builds opcode hash tables
 */
void BuildOpcodeHashTable() {
    xBuildOpcodeHashTable();
}



/*
 *  InstrAssemble(buffer,instr) --
 *
 *  Fills in 'buffer' with the assembled instruction and returns its length
 */
int InstrAssemble(unsigned char *buffer, Instruction *instr) {
    return xInstrAssemble(buffer, instr);
}



/*
 *  ComputeRegistersDefined(instr) --
 *
 *  Returns a register bitvector, with bits set for all registers that
 *  'instr' defines
 */
REGISTER_BITVECTOR ComputeRegistersDefined(Instruction *instr) {
    return xComputeRegistersDefined(instr);
}



/*
 *  ComputeRegistersDefinedLiberal(instr) --
 *
 *  Returns a register bitvector, with bits set for all registers that
 *  'instr' completely or partially may define
 */
REGISTER_BITVECTOR ComputeRegistersDefinedLiberal(Instruction *instr) {
    return xComputeRegistersDefinedLiberal(instr);
}



/*
 *  ComputeRegistersUsed(instr) --
 *
 *  Returns a register bitvector, with bits set for all registers that
 *  'instr' completely defines
 */
REGISTER_BITVECTOR ComputeRegistersUsed(Instruction *instr) {
    return xComputeRegistersUsed(instr);
}



/*
 *  ComputePSWUsed(instr) --
 *
 *  Returns a PSW bitvector with a bit set for each PSW flag 'instr' uses
 */
PSW_BITVECTOR ComputePSWUsed(Instruction *instr) {
    return xComputePSWUsed(instr);
}



/*
 *  ComputePSWDefined(instr) --
 *
 *  Returns a PSW bitvector with a bit set for each PSW flag
 *  that is defined by 'instr'
 */
PSW_BITVECTOR ComputePSWDefined(Instruction *instr) {
    return xComputePSWDefined(instr);
}



/*
 *  ComputePSWDefinedLiberal(instr) --
 *
 *  Returns a PSW bitvector with a bit set for each PSW flag
 *  that may be defined by 'instr'
 */
PSW_BITVECTOR ComputePSWDefinedLiberal(Instruction *instr) {
    return xComputePSWDefinedLiberal(instr);
}



/*
 *  ComputeStackUsed(instr,slots,bp_index,arsize) --
 *
 *  Fills in the integer array with true or false
 *  depending on which stack slots 'instr' uses
 */
void ComputeStackUsed(
Instruction *instr, int *slots, int bp_index, int arsize) {
    xComputeStackUsed(instr, slots, bp_index, arsize);
}



/*
 *  ComputeStackDefined(instr,slots,bp_index,arsize) --
 *
 *  Fills in the integer array with true or false
 *  depending on which stack slots 'instr' defines
 */
void ComputeStackDefined(
Instruction *instr, int *slots, int bp_index, int arsize) {
    xComputeStackDefined(instr, slots, bp_index, arsize);
}



/*
 *  ComputeStackDefinedLiberal(instr,slots,bp_index,arsize) --
 *
 *  Fills in the integer array with true or false
 *  depending on which stack slots 'instr' may defines
 */
void ComputeStackDefinedLiberal(
Instruction *instr, int *slots, int bp_index, int arsize) {
    xComputeStackDefinedLiberal(instr, slots, bp_index, arsize);
}



/*
 *  RegisterIsLiveInBitvector(aReg,vec) --
 *
 *  Returns true if the register is live in the bitvector
 */
bool RegisterIsLiveInBitvector(int aReg, REGISTER_BITVECTOR vec) {
    return xRegisterIsLiveInBitvector(aReg, vec);
}



/*
 *  RegisterIsFramePointer(aReg) --
 *
 *  Returns true if the register is a frame pointer
 */
bool RegisterIsFramePointer(int aReg) {
    return xRegisterIsFramePointer(aReg);
}



/*
 *  RegisterIsStackPointer(aReg) --
 *
 *  Returns true if the register is a stack pointer
 */
bool RegisterIsStackPointer(int aReg) {
    return xRegisterIsStackPointer(aReg);
}



/*
 *  OperandIsGPRegister(op) --
 *
 *  Returns true if the operand is a general purpose register
 */
bool OperandIsGPRegister(Operand *op) {
    return xOperandIsGPRegister(op);
}



/*
 *  InstrSetClassFlags(instr) --
 *
 *  Sets class flags in the instruction
 */
void InstrSetClassFlags(Instruction *instr) {
    xInstrSetClassFlags(instr);
}



/*
 *  InstrHasImplicitDest(instr) --
 *
 *  Returns true if the instruction has an implicit destination
 */
bool InstrHasImplicitDest(Instruction *instr) {
    return xInstrHasImplicitDest(instr);
}



/*
 *  InstrHasImplicitDest2(instr) --
 *
 *  Returns true if the instruction has an implicit second destination
 */
bool InstrHasImplicitDest2(Instruction *instr) {
    return xInstrHasImplicitDest2(instr);
}



/*
 *  InstrImplicitDest(instr) --
 *
 *  Returns the implicit register that the instruction defines
 */
UINT32 InstrImplicitDest(Instruction *instr) {
    return xInstrImplicitDest(instr);
}



/*
 *  InstrImplicitDest2(instr) --
 *
 *  Returns the second implicit register that the instruction defines
 */
UINT32 InstrImplicitDest2(Instruction *instr) {
    return _InstrImplicitDest2(instr);
}



/*
 *  InstrHasImplicitSrc(instr) --
 *
 *  Returns true if the instruction has an implicit source
 */
bool InstrHasImplicitSrc(Instruction *instr) {
    return xInstrHasImplicitSrc(instr);
}



/*
 *  InstrHasImplicitSrc2(instr) --
 *
 *  Returns true if the instruction has an implicit second source
 */
bool InstrHasImplicitSrc2(Instruction *instr) {
    return xInstrHasImplicitSrc2(instr);
}



/*
 *  InstrImplicitSrc(instr) --
 *
 *  Returns the implicit register that the instruction uses
 */
UINT32 InstrImplicitSrc(Instruction *instr) {
    return xInstrImplicitSrc(instr);
}



/*
 *  InstrImplicitSrc2(instr) --
 *
 *  Returns the second implicit register that the instruction uses
 */
UINT32 InstrImplicitSrc2(Instruction *instr) {
    return _InstrImplicitSrc2(instr);
}



/*
 *  InstrUsesDest(instr) --
 *
 *  Returns true if the instruction uses its destination operand
 *  as a source operand
 */
bool InstrUsesDest(Instruction *instr) {
    return xInstrUsesDest(instr);
}



/*
 *  GenerateStaticOpcodeHistogram() --
 *
 *  Generates a static opcode histogram table
 */
void GenerateStaticOpcodeHistogram() {
    xGenerateStaticOpcodeHistogram();
}



/*
 *  GenerateWeightedOpcodeHistogram() --
 *
 *  Generates a weighted opcode histogram table
 */
void GenerateWeightedOpcodeHistogram() {
    xGenerateWeightedOpcodeHistogram();
}



/*
 *  OperandAccessesStackFromFramePointer(op) --
 *
 *  Returns true if the operand accesses the stack by using the frame pointer
 */
bool OperandAccessesStackFromFramePointer(Operand *op) {
    return xOperandAccessesStackFromFramePointer(op);
}



/*
 *  OperandAccessesStackFromFramePointerWithSIB(op) --
 *
 *  Returns true if the operand accesses the stack using the frame pointer
 *  and an SIB byte
 */
bool OperandAccessesStackFromFramePointerWithSIB(Operand *op) {
    return xOperandAccessesStackFromFramePointerWithSIB(op);
}



/*
 *  OperandUseLargestBytemode(op) --
 *
 *  Updates the operand to use the largest bytemode possible
 */
void OperandUseLargestBytemode(Operand *op) {
    xOperandUseLargestBytemode(op);
}



/*
 *  OperandIsImmediate(op) --
 *
 *  Returns true if the operand is an immediate value
 */
bool OperandIsImmediate(Operand *op) {
    return xOperandIsImmediate(op);
}



/*
 *  OperandAccessesStackFromStackPointer(op) --
 *
 *  Returns true if the operand accesses the stack by using the stack pointer
 */
bool OperandAccessesStackFromStackPointer(Operand *op) {
    return xOperandAccessesStackFromStackPointer(op);
}



/*
 *  OperandIsAddress(op) --
 *
 *  Returns true if the operand is an address
 */
bool OperandIsAddress(Operand *op) {
    return xOperandIsAddress(op);
}



/*
 *  OperandUsesContentsOfStackPointer(op) --
 *
 *  Returns true if 'op' uses the contents of ESP
 */
bool OperandUsesContentsOfStackPointer(Operand *op) {
    return xOperandUsesContentsOfStackPointer(op);
}



/*
 *  OperandUsesContentsOfFramePointer(op) --
 *
 *  Returns true if 'op' uses the contents of EBP
 */
bool OperandUsesContentsOfFramePointer(Operand *op) {
    return xOperandUsesContentsOfFramePointer(op);
}



/*
 *  OperandIsAbsoluteMemoryLocation(op) --
 *
 *  Returns true if the operand is an absolute memory location
 */
bool OperandIsAbsoluteMemoryLocation(Operand *op) {
    return xOperandIsAbsoluteMemoryLocation(op);
}



/*
 *  OperandIsStackPointer(op) --
 *
 *  Returns true if 'op' is ESP
 */
bool OperandIsStackPointer(Operand *op) {
    return xOperandIsStackPointer(op);
}



/*
 *  OperandIsMemory(op) --
 *
 *  Returns true if 'op' is a memory location
 */
bool OperandIsMemory(Operand *op) {
    return xOperandIsMemory(op);
}



/*
 *  OperandIsFramePointer(op) --
 *
 *  Returns true if 'op' is EBP
 */
bool OperandIsFramePointer(Operand *op) {
    return xOperandIsFramePointer(op);
}



/*
 *  OperandWritesToUnknownStackSlot(op) --
 *
 *  Returns true if 'op' accesses memory that may be the runtime stack
 */
bool OperandWritesToUnknownStackSlot(Operand *op) {
    return xOperandWritesToUnknownStackSlot(op);
}



/*
 *  BitRegisterToRegister(i) --
 *
 *  Returns the register that the bit 'i' (from a bitvector) corresponds to
 */
int BitRegisterToRegister(int i) {
    return xBitRegisterToRegister(i);
}



/*
 *  InstrDefinesPSW(instr) --
 *
 *  Returns true if the instruction can affect the processor status word
 */
bool InstrDefinesPSW(Instruction *instr) {
    return xInstrDefinesPSW(instr);
}



/*
 *  InstrIsDependent(a,b) --
 *
 *  Returns true if there is a data or control dependency between a and b
 */
bool InstrIsDependent(Instruction *a, Instruction *b) {
    return xInstrIsDependent(a, b);
}



/*
 *  InstrDelayBetween(a,b) --
 *
 *  Returns how long the pipeline may delay between executing a then b
 */
UINT8 InstrDelayBetween(Instruction *a, Instruction *b) {
    return xInstrDelayBetween(a, b);
}



/*
 *  InstrsPair(a,b) --
 *
 *  Returns true if a and b can pair and execute in the same cycle
 */
bool InstrsPair(Instruction *a, Instruction *b) {
    return xInstrsPair(a, b);
}



/*
 *  NumInstrsPairable() --
 *
 *  Returns the maximum number of instructions that can be paired at once
 */
UINT32 NumInstrsPairable() {
    return xNumInstrsPairable();
}



/*
 *  InstrCreate2ByteDoNothing() --
 *
 *  Creates a 2-byte instruction that has the effect of a NOP
 */
Instruction *InstrCreate2ByteDoNothing() {
    return _InstrCreate2ByteDoNothing();
}



/*
 *  InstrUsesReg(iins,areg) --
 *
 *  Reg(iins, areg) Returns true if instruction iins uses register areg
 *  as [part of] a source operand.
 */
bool InstrUsesReg(Instruction *iins, int areg) {
    return xInstrUsesReg(iins, areg);
}



/*
 *  InstrDefinesReg(iins,areg) --
 *
 *  Returns true if instruction iins defines [part of] register areg.
 */
bool InstrDefinesReg(Instruction *iins, int areg) {
    return xInstrDefinesReg(iins, areg);

}



/*
 *  GetOperandLocation(instr,op) --
 *
 *  Given an instruction and an operand returns the location of the operand
 *  in the instruction encoding as an offset from the start of the instruction.
 */
int GetOperandLocation(Instruction *instr, Operand *op) {
    return xGetOperandLocation(instr, op);
}



/*
 *  ReplaceNonsupportedInstructions() --
 *
 *  Replaces instructions that are not supported
 *  with equivalent supported instructions
 */
void ReplaceNonsupportedInstructions() {
    xReplaceNonsupportedInstructions();
}



/*
 *  PICGetALUForCall(call_instr) --
 *
 *  Given a CALL instruction that is suspected to be part of a PIC calculation
 *  that gets the current PC, this function will return the corresponding ALU
 *  instruction.  Return NULL if this is not a PIC related call.
 */
Instruction *PICGetALUForCall(Instruction *call_instr) {
    return xPICGetALUForCall(call_instr);
}



/*
 *  InstrCreate32BitMovMemToReg(reg) --
 *
 *  Builds an instruction that moves a memory location to a register
 */
Instruction *InstrCreate32BitMovMemToReg(int reg) {
    return _InstrCreate32BitMovMemToReg(reg);
}



/*
 *  AssembleCode(code_section) --
 */
void AssembleCode(Section *code_section) {
    xAssembleCode(code_section);
}



/*
 *  InstrCreateMovMemToReg(scale,index_reg,base_reg,dest_reg) --
 */
Instruction *InstrCreateMovMemToReg(int scale,
    UINT32 index_reg, UINT32 base_reg, UINT32 dest_reg) {
    return xInstrCreateMovMemToReg(scale, index_reg, base_reg, dest_reg);
}



/*
 *  InstrCreateSubImmFromReg(imm_value,reg) --
 */
Instruction *InstrCreateSubImmFromReg(int imm_value, UINT32 reg) {
    return xInstrCreateSubImmFromReg(imm_value, reg);
}					// InstrCreateSubImmFromReg



/*
 *  InstrCreateMovAbsoluteMemToReg(memloc,reg) --
 */
Instruction *InstrCreateMovAbsoluteMemToReg(ADDRESS memloc, int reg) {
    return xInstrCreateMovAbsoluteMemToReg(memloc, reg);
}



/*
 *  InstrCreate32BitMovRegToMem(reg) --
 */
Instruction *InstrCreate32BitMovRegToMem(int reg) {
    return _InstrCreate32BitMovRegToMem(reg);
}



/*
 *  PrintPSWFlag(flag) --
 */
void PrintPSWFlag(int flag) {
    return xPrintPSWFlag(flag);
}



/*
 *  InstrCreate32BitMovImmToMem(mem_loc,imm) --
 */
Instruction *InstrCreate32BitMovImmToMem(ADDRESS mem_loc, int imm) {
    return _InstrCreate32BitMovImmToMem(mem_loc, imm);
}



/*
 *  InstrCreate32BitAddRegToMem(mem_loc,reg) --
 */
Instruction *InstrCreate32BitAddRegToMem(ADDRESS mem_loc, int reg) {
    return _InstrCreate32BitAddRegToMem(mem_loc, reg);
}



/*
 *  InstrCreate32BitSubRegFromMem(mem_loc,reg) --
 */
Instruction *InstrCreate32BitSubRegFromMem(ADDRESS mem_loc, int reg) {
    return _InstrCreate32BitSubRegFromMem(mem_loc, reg);
}



/*
 *  InstrCreate32BitAddImmToMem(mem_loc,imm) --
 */
Instruction *InstrCreate32BitAddImmToMem(ADDRESS mem_loc, int imm) {
    return _InstrCreate32BitAddImmToMem(mem_loc, imm);
}



/*
 *  InstrCreate32BitAddMemToReg(mem_loc,reg) --
 */
Instruction *InstrCreate32BitAddMemToReg(ADDRESS mem_loc, UINT32 reg) {
    return _InstrCreate32BitAddMemToReg(mem_loc, reg);
}



/*
 *  InstrCreateSubMemFromReg(mem_loc,reg) --
 */
Instruction *InstrCreateSubMemFromReg(ADDRESS mem_loc, UINT32 reg) {
    return xInstrCreateSubMemFromReg(mem_loc, reg);
}



/*
 *  InstrCreateSubImmFromMem(imm,mem_loc) --
 */
Instruction *InstrCreateSubImmFromMem(int imm, ADDRESS mem_loc) {
    return xInstrCreateSubImmFromMem(imm, mem_loc);
}



/*
 *  InstrCreate16BitAddRegToMem(mem_loc,reg) --
 */
Instruction *InstrCreate16BitAddRegToMem(ADDRESS mem_loc, int reg) {
    return _InstrCreate16BitAddRegToMem(mem_loc, reg);
}



/*
 *  InstrCreate16BitSubRegFromMem(mem_loc,reg) --
 */
Instruction *InstrCreate16BitSubRegFromMem(ADDRESS mem_loc, int reg) {
    return _InstrCreate16BitSubRegFromMem(mem_loc, reg);
}



/*
 *  InstrCreateLeave() --
 */
Instruction *InstrCreateLeave() {
    return xInstrCreateLeave();
}



/*
 *  InstrCreateReturn() --
 */
Instruction *InstrCreateReturn() {
    return xInstrCreateReturn();
}



/*
 *  InstrCreate32BitAddImmToReg(imm,reg) --
 */
Instruction *InstrCreate32BitAddImmToReg(int imm, int reg) {
    return _InstrCreate32BitAddImmToReg(imm, reg);
}



/*
 *  InstrCreateShortJb() --
 */
Instruction *InstrCreateShortJb() {
    return xInstrCreateShortJb();
}



/*
 *  InstrCreateShortJae() --
 */
Instruction *InstrCreateShortJae() {
    return xInstrCreateShortJae();
}



/*
 *  InstrCreateShortJe() --
 */
Instruction *InstrCreateShortJe() {
    return xInstrCreateShortJe();
}



/*
 *  InstrCreateShortJne() --
 */
Instruction *InstrCreateShortJne() {
    return xInstrCreateShortJne();
}



/*
 *  InstrCreateShortJbe() --
 */
Instruction *InstrCreateShortJbe() {
    return xInstrCreateShortJbe();
}



/*
 *  InstrCreateShortJa() --
 */
Instruction *InstrCreateShortJa() {
    return xInstrCreateShortJa();
}



/*
 *  InstrCreateShortJs() --
 */
Instruction *InstrCreateShortJs() {
    return xInstrCreateShortJs();
}   



/*
 *  InstrCreateShortJns() --
 */
Instruction *InstrCreateShortJns() {
    return xInstrCreateShortJns();
}



/*
 *  InstrCreateShortJl() --
 */
Instruction *InstrCreateShortJl() {
    return xInstrCreateShortJl();
}



/*
 *  InstrCreateShortJge() --
 */
Instruction *InstrCreateShortJge() {
    return xInstrCreateShortJge();
}



/*
 *  InstrCreateShortJle() --
 */
Instruction *InstrCreateShortJle() {
    return xInstrCreateShortJle();
}



/*
 *  InstrCreateShortJng() --
 */
Instruction *InstrCreateShortJng() {
    return xInstrCreateShortJle();
}



/*
 *  InstrCreateShortJg() --
 */
Instruction *InstrCreateShortJg() {
    return xInstrCreateShortJg();
}



/*
 *  InstrCreateShortJnle() --
 */
Instruction *InstrCreateShortJnle() {
    return xInstrCreateShortJg();
}



/*
 *  InstrCreateShortJecxz() --
 */
Instruction *InstrCreateShortJecxz() {
    return xInstrCreateShortJecxz();
}



/*
 *  InstrCreateNearJb() --
 */
Instruction *InstrCreateNearJb() {
    return xInstrCreateNearJb();
}



/*
 *  InstrCreateNearJae() --
 */
Instruction *InstrCreateNearJae() {
    return xInstrCreateNearJae();
}



/*
 *  InstrCreateNearJbe() --
 */
Instruction *InstrCreateNearJbe() {
    return xInstrCreateNearJbe();
}



/*
 *  InstrCreateNearJa() --
 */
Instruction *InstrCreateNearJa() {
    return xInstrCreateNearJa();
}



/*
 *  InstrCreateNearJs() --
 */
Instruction *InstrCreateNearJs() {
    return xInstrCreateNearJs();
}



/*
 *  InstrCreateNearJns() --
 */
Instruction *InstrCreateNearJns() {
    return xInstrCreateNearJns();
}



/*
 *  InstrCreateNearJl() --
 */
Instruction *InstrCreateNearJl() {
    return xInstrCreateNearJl();
}



/*
 *  InstrCreateNearJge() --
 */
Instruction *InstrCreateNearJge() {
    return xInstrCreateNearJge();
}



/*
 *  InstrCreateNearJle() --
 */
Instruction *InstrCreateNearJle() {
    return xInstrCreateNearJle();
}



/*
 *  InstrCreateNearJg() --
 */
Instruction *InstrCreateNearJg() {
    return xInstrCreateNearJg();
}



/*
 *  InstrCreate32BitShlImm(reg,times) --
 */
Instruction *InstrCreate32BitShlImm(UINT32 reg, int times) {
    return _InstrCreate32BitShlImm(reg, times);
}



/*
 *  InstrCreate32BitShr(reg,value) --
 */
Instruction *InstrCreate32BitShr(int reg, int value) {
    return _InstrCreate32BitShr(reg, value);
}



/*
 *  InstrCreate32BitShlCL(reg) --
 */
Instruction *InstrCreate32BitShlCL(UINT32 reg) {
    return xInstrCreate32BitShlCL(reg);
}



/*
 *  InstrCreate32BitShrCL(reg) --
 */
Instruction *InstrCreate32BitShrCL(UINT32 reg) {
    return xInstrCreate32BitShrCL(reg);
}



/*
 *  InstrCreateSubRegFromReg(src_reg,dest_reg) --
 */
Instruction *InstrCreateSubRegFromReg(UINT32 src_reg, UINT32 dest_reg) {
    return xInstrCreateSubRegFromReg(src_reg, dest_reg);
}



/*
 *  InstrCreateIMULRegReg(src_reg,dest_reg) --
 */
Instruction *InstrCreateIMULRegReg(UINT32 src_reg, UINT32 dest_reg) {
    return xInstrCreateIMULRegReg(src_reg, dest_reg);
}



/*  
 *  InstrCreateInc(reg) --
 */
Instruction *InstrCreateInc(UINT32 reg) {
    return xInstrCreateInc(reg);
}



/*
 *  InstrCreateDec(reg) --
 */
Instruction *InstrCreateDec(UINT32 reg) {
    return xInstrCreateDec(reg);
}



/*
 *  InstrCreateCmpRegToReg(reg1,reg2) --
 */
Instruction *InstrCreateCmpRegToReg(UINT32 reg1, UINT32 reg2) {
    return xInstrCreateCmpRegToReg(reg1, reg2);
}



/*
 *  InstrCreateHalt() --
 */
Instruction *InstrCreateHalt() {
    return xInstrCreateHalt();
}



/*
 *  InstrCreateCALLDisp() --
 */
Instruction *InstrCreateCALLDisp() {
    return xInstrCreateCALLDisp();
}



/*
 *  InstrCreateCallNearMemBase(reg) --
 */
Instruction *InstrCreateCallNearMemBase(int reg) {
    return xInstrCreateCallNearMemBase(reg);
}



/*
 *  InstrCreateSetCarry() --
 */
Instruction *InstrCreateSetCarry() {
    return xInstrCreateSetCarry();
}



/*
 *  InstrCreateClearCarry() --
 */
Instruction *InstrCreateClearCarry() {
    return xInstrCreateClearCarry();
}



/*
 *  InstrCreateJumpOnNoCarry() --
 */
Instruction *InstrCreateJumpOnNoCarry() {
    return xInstrCreateJumpOnNoCarry();
}



/*
 *  InstrCreateCallIndirect() --
 */
Instruction *InstrCreateCallIndirect() {
    return xInstrCreateCallIndirect();
}



/*
 *  InstrCreateJMPIndirectAbsMemAddr() --
 */
Instruction *InstrCreateJMPIndirectAbsMemAddr() {
    return xInstrCreateJMPIndirectAbsMemAddr();
}



/*
 *  InstrCreateAddRegToReg(src_reg,dest_reg) --
 */
Instruction *InstrCreateAddRegToReg(int src_reg, int dest_reg) {
    return xInstrCreateAddRegToReg(src_reg, dest_reg);
}



/*
 *  InstrCreateAddImm8ToReg(imm8,dest_reg) --
 */
Instruction *InstrCreateAddImm8ToReg(INT8 imm8, int dest_reg) {
    return xInstrCreateAddImm8ToReg(imm8, dest_reg);
}



/*
 *  InstrCreateAddImm32ToReg(imm32,dest_reg) --
 */
Instruction *InstrCreateAddImm32ToReg(INT32 imm32, int dest_reg) {
    return xInstrCreateAddImm32ToReg(imm32, dest_reg);
}



/*
 *  InstrCreateSubImm8FromReg(imm8,dest_reg) --
 */
Instruction *InstrCreateSubImm8FromReg(INT8 imm8, int dest_reg) {
    return xInstrCreateSubImm8FromReg(imm8, dest_reg);
}



/*
 *  InstrCreateSubImm32FromReg(imm32,dest_reg) --
 */
Instruction *InstrCreateSubImm32FromReg(INT32 imm32, int dest_reg) {
    return xInstrCreateSubImm32FromReg(imm32, dest_reg);
}



/*
 *  InstrCreateOrRegWithMemBaseIndex(src,base,index,scale) --
 */
Instruction *InstrCreateOrRegWithMemBaseIndex(int src, int base, int index,
    UINT8 scale) {
    return xInstrCreateOrRegWithMemBaseIndex(src, base, index, scale);
}



/*
 *  InstrCreateOrMemBaseIndexWithReg(base,index,scale,dest) --
 */
Instruction *InstrCreateOrMemBaseIndexWithReg(int base, int index, UINT8 scale,
    int dest) {
    return xInstrCreateOrMemBaseIndexWithReg(base, index, scale, dest);
}



/*
 *  InstrCreateAndImmWithReg(imm,dest_reg) --
 */
Instruction *InstrCreateAndImmWithReg(int imm, int dest_reg) {
    return xInstrCreateAndImmWithReg(imm, dest_reg);
}



/*
 *  InstrCreateXorRegWithMemDisp8Base(src,dest_disp,dest) --
 */
Instruction *InstrCreateXorRegWithMemDisp8Base(int src, INT8 dest_disp,
    int dest) {
    return xInstrCreateXorRegWithMemDisp8Base(src, dest_disp, dest);
}



/*
 *  InstrCreateXorRegWithMemDisp32Base(src,dest_disp,dest) --
 */
Instruction *InstrCreateXorRegWithMemDisp32Base(int src, INT32 dest_disp,
    int dest) {
    return xInstrCreateXorRegWithMemDisp32Base(src, dest_disp, dest);
}



/*
 *  InstrCreate32BitMovMemDispToReg(s_reg,off,d_reg) --
 */
Instruction *InstrCreate32BitMovMemDispToReg(int s_reg, INT8 off, int d_reg) {
    return _InstrCreate32BitMovMemDispToReg(s_reg, off, d_reg);
}



/*
 *  InstrCreateMovRegToReg(srcreg,destreg) --
 *
 *  Creates an instruction that moves the contents of the 'srcreg' register
 *  into the 'destreg'.
 */
Instruction *InstrCreateMovRegToReg(int srcreg, int destreg) {
    return xInstrCreateMovRegToReg(srcreg, destreg);
}



/*
 *  InstrCreateMovRegToMemBase(src,dest) --
 */
Instruction *InstrCreateMovRegToMemBase(int src, int dest) {
    return xInstrCreateMovRegToMemBase(src, dest);
}



/*
 *  InstrCreateMovMemBaseToReg(src,dest) --
 */
Instruction *InstrCreateMovMemBaseToReg(int src, int dest) {
    return xInstrCreateMovMemBaseToReg(src, dest);
}



/*
 *  InstrCreateMovRegToMemDisp8Base(src,dest_disp,dest) --
 */
Instruction *InstrCreateMovRegToMemDisp8Base(int src, INT8 dest_disp,
    int dest) {
    return xInstrCreateMovRegToMemDisp8Base(src, dest_disp, dest);
}



/*
 *  InstrCreateMovRegToMemDisp32Base(src,dest_disp,dest) --
 */
Instruction *InstrCreateMovRegToMemDisp32Base(int src, INT32 dest_disp,
    int dest) {
    return xInstrCreateMovRegToMemDisp32Base(src, dest_disp, dest);
}



/*
 *  InstrCreateMovMemDisp8BaseToReg(src_disp,src, dest) --
 */
Instruction *InstrCreateMovMemDisp8BaseToReg(INT8 src_disp, int src,
    int dest) {
    return xInstrCreateMovMemDisp8BaseToReg(src_disp, src, dest);
}



/*
 *  InstrCreateMovMemDisp32BaseToReg(src_disp,src, dest) --
 */
Instruction *InstrCreateMovMemDisp32BaseToReg(INT32 src_disp, int src,
    int dest) {
    return xInstrCreateMovMemDisp32BaseToReg(src_disp, src, dest);
}



/*
 *  InstrCreateMovRegToMemBaseIndex(src,base,index,scale) --
 */
Instruction *InstrCreateMovRegToMemBaseIndex(int src, int base, int index,
    UINT8 scale) {
    return xInstrCreateMovRegToMemBaseIndex(src, base, index, scale);
}



/*
 *  InstrCreateMovMemBaseIndexToReg(base,index,scale,dest) --
 */
Instruction *InstrCreateMovMemBaseIndexToReg(int base, int index, UINT8 scale,
    int dest) {
    return xInstrCreateMovMemBaseIndexToReg(base, index, scale, dest);
}



/*
 *  InstrCreateMovImmToMemBase(imm,dest_reg) --
 *  
 *  Returns an instruction that moves an immediate value to a memory address
 *  located in dest_reg.
 */
Instruction *InstrCreateMovImmToMemBase(int imm, int dest_reg) {
    return xInstrCreateMovImmToMemBase(imm, dest_reg);
}



/*
 *  InstrCreateMovImmToMemDisp8Base(imm,disp,dest) --
 */
Instruction *InstrCreateMovImmToMemDisp8Base(int imm, INT8 disp, int dest) {
    return xInstrCreateMovImmToMemDisp8Base(imm, disp, dest);
}



/*
 *  InstrCreateMovImmToMemDisp32Base(imm,disp,dest) --
 */
Instruction *InstrCreateMovImmToMemDisp32Base(int imm, INT32 disp, int dest) {
    return xInstrCreateMovImmToMemDisp32Base(imm, disp, dest);
}



/*  
 *  InstrCreateMovImmToMemBaseIndex(imm,base,index,scale) --
 */
Instruction *InstrCreateMovImmToMemBaseIndex(int imm, int base, int index,
    int scale) {
    return xInstrCreateMovImmToMemBaseIndex(imm, base, index, scale);
}



/*
 *  InstrCreateMovzReg8ToReg32(src_reg,dest_reg) --
 */
Instruction *InstrCreateMovzReg8ToReg32(UINT32 src_reg, UINT32 dest_reg) {
    return xInstrCreateMovzReg8ToReg32(src_reg, dest_reg);
}



/*
 *  InstrCreate32BitPopToMem() --
 */
Instruction *InstrCreate32BitPopToMem() {
    return _InstrCreate32BitPopToMem();
}



/*
 *  InstrCreateDiv(divisor) --
 */
Instruction *InstrCreateDiv(int divisor) {
    return xInstrCreateDiv(divisor);
}



/*
 *  InstrCreateSyscall() --
 */
Instruction *InstrCreateSyscall() {
    return xInstrCreateSyscall();
}



/*
 *  InstrCreateTestRegWithReg(reg1,reg2) --
 */
Instruction *InstrCreateTestRegWithReg(UINT32 reg1, UINT32 reg2) {
    return xInstrCreateTestRegWithReg(reg1, reg2);
}



/*
 *  InstrCreateTestImmWithReg(imm,reg) --
 */
Instruction *InstrCreateTestImmWithReg(int imm, UINT32 reg) {
    return xInstrCreateTestImmWithReg(imm, reg);
}



/*
 *  InstrCreateLea(src,offset,dest) --
 */
Instruction *InstrCreateLea(UINT32 src, int offset, UINT32 dest) {
    return xInstrCreateLea(src, offset, dest);
}



/*
 *  InstrCreateSeteReg(reg) --
 */
Instruction *InstrCreateSeteReg(UINT32 reg) {
    return xInstrCreateSeteReg(reg);
}



/*
 *  InstrCreateSetzReg(reg) --
 */
Instruction *InstrCreateSetzReg(UINT32 reg) {
    return xInstrCreateSeteReg(reg);
}



/*
 *  InstrCreate8bitDivByReg(divisor) --
 */
Instruction *InstrCreate8bitDivByReg(int divisor) {
    return xInstrCreate8bitDivByReg(divisor);
}



/*
 *  InstrCreate32bitDivByReg(divisor) --
 */
Instruction *InstrCreate32bitDivByReg(int divisor) {
    return xInstrCreate32bitDivByReg(divisor);
}



/*
 *  InstrCreate8bitIDivByReg(divisor) --
 */
Instruction *InstrCreate8bitIDivByReg(int divisor) {
    return xInstrCreate8bitIDivByReg(divisor);
}



/*
 *  InstrCreate32bitIDivByReg(divisor) --
 */
Instruction *InstrCreate32bitIDivByReg(int divisor) {
    return xInstrCreate32bitIDivByReg(divisor);
}



/*
 *  InstrCreate8bitMulByReg(multiplier) --
 */
Instruction *InstrCreate8bitMulByReg(int multiplier) {
    return xInstrCreate8bitMulByReg(multiplier);
}



/*
 *  InstrCreate32bitMulByReg(multiplier) --
 */
Instruction *InstrCreate32bitMulByReg(int multiplier) {
    return xInstrCreate32bitMulByReg(multiplier);
}



/*
 *  InstrCreate8bitIMulByReg(multiplier) --
 */
Instruction *InstrCreate8bitIMulByReg(int multiplier) {
    return xInstrCreate8bitIMulByReg(multiplier);
}



/*
 *  InstrCreate32bitIMulByReg(multiplier) --
 */
Instruction *InstrCreate32bitIMulByReg(int multiplier) {
    return xInstrCreate32bitIMulByReg(multiplier);
}



/*
 *  InstrCreateIMulImm8ByReg(imm8, reg) --
 */
Instruction *InstrCreateIMulImm8ByReg(INT8 imm8, int reg) {
    return xInstrCreateIMulImm8ByReg(imm8, reg);
}



/*
 *  InstrCreateIMulImm32ByReg(imm32, reg) --
 */
Instruction *InstrCreateIMulImm32ByReg(INT32 imm32, int reg) {
    return xInstrCreateIMulImm32ByReg(imm32, reg);
}



/*
 *  InstrCreateJmpNearAbsDisp32Index(disp32, index, scale) --
 */
Instruction *InstrCreateJmpNearAbsDisp32Index(INT32 disp32, int index,
    int scale) {
    return xInstrCreateJmpNearAbsDisp32Index(disp32, index, scale);
}



/*
 *  GetStackPointer() --
 */
int GetStackPointer() {
    return xGetStackPointer();
}



/*
 *  GetFramePointer() --
 */
int GetFramePointer() {
    return xGetFramePointer();
}



/*
 *  GetARegister() --
 */
int GetARegister() {
    return xGetARegister();
}



/*
 *  GetSyscallNoForSyscall(syscall) --
 */
int GetSyscallNoForSyscall(Instruction *syscall) {
    return xGetSyscallNoForSyscall(syscall);
}



/*
 *  RemoveSyscallNoForSyscall(syscall) --
 */
void RemoveSyscallNoForSyscall(Instruction *syscall) {
    return xRemoveSyscallNoForSyscall(syscall);
}
