/*
 *  upm.c -- see upm.h for details of this file.
 */

#include "global.h"
#include "upm.h"
#include "relocate.h"
#include "hash.h"
#include "utilities.h"
#include "flowgraph.h"
#include "sysdep.h"
#include "plto.h"
#include "linkage.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

typedef struct UPMHash {
    void *uh_key1;
    void *uh_key2;
    Patch *uh_upm;
    struct UPMHash *uh_hash_next;
    struct UPMHash *uh_hash_prev;
} UPMHash;


#define UPM_HASH_SIZE (0x1000)
#define UPM_HASH_MASK (0x0fff)

static UPMHash *upm2_dest_table[UPM_HASH_SIZE];
static UPMHash *upm2_src_table[UPM_HASH_SIZE];
static UPMHash *upm2_dest_table_addr[UPM_HASH_SIZE];
static UPMHash *upm2_src_table_addr[UPM_HASH_SIZE];

static int UPMHash1(void *key);
static int UPMHash2(void *key1, void *key2);
static void UPMTableInsert(void *key1, void *key2, Patch *upm, UPMHash **table);
static LinkedList *UPMTableGetAll(void *key1, void *key2, UPMHash **table);
static void UPMTableRemove(void *key1, void *key2, UPMHash **table);
static void UPMTableRemovePair(void *k1, void *k2, Patch *upm, UPMHash **table);



/*
 *  UPMHash1(key) --
 *
 *  Given a key, this function returns the hash index for this key based
 *  upon the current hash size. This function is the helper hash function
 *  for UPMHash2.
 *
 *  key: the key to be hashed
 *  returns: the hash index based on the hash size and key
 *  see also: UPM_HASH_SIZE
 *  see also: UPMHash2
 */
static int UPMHash1(void *key) {
    return abs(((INT32) key + ((0xf0 & (INT32) key) >> 4) +
	    ((0xf00 & (INT32) key) >> 8)) & UPM_HASH_MASK);
}



/*
 *  UPMHash2(key1,key2) --
 *
 *  Given two keys, this function returns the hash index based on the two
 *  keys and the current hash size. This function relies on UPMHash1 to
 *  achieve the final hash value.
 *
 *  key1: the first key component to be hashed
 *  key2: the second key component to be hashed
 *  returns: the hash index based on the hash size and key
 *  see also: UPM_HASH_SIZE
 *  see also: UPMHash1
 */
static int UPMHash2(void *key1, void *key2) {
    if (key1 != key2) {
	return UPMHash1((void *) ((int) key1 + (int) key2));
    }

    return UPMHash1(key1);
}



/*
 *  AddressIsUPMDest(section,offset) --
 *
 *  Returns whether or not the specified address is the destination of at
 *  least one UPM.
 *
 *  section: the section to check
 *  offset: the offset from section that makes the address to check
 *
 *  returns: true if the address formed from the specified section and
 *  offset is the destination of at least one UPM
 */
bool AddressIsUPMDest(Section *section, ADDRESS offset) {
    LinkedList *upms;

    upms = UPMTableGetByDestAddress(section, offset);
    if (ll_size(upms) > 0) {
	ll_free(upms);
	return true;
    }

    ll_free(upms);
    return false;
}



/*
 *  InstrIsUPMSource(instr) --
 *
 *  Returns whether or not the specified instruction is the source of at
 *  least one UPM.
 *
 *  instr: the instruction to check
 *  returns: true if the instruction is the source of at least one UPM
 */
bool InstrIsUPMSource(Instruction *instr) {
    LinkedList *upms;

    upms = UPMTableGetBySource(instr);
    if (ll_size(upms) > 0) {
	ll_free(upms);
	return true;
    }

    ll_free(upms);
    return false;
}



/*
 *  AddressIsUPMSource(section,offset) --
 *
 *  Returns whether or not the specified address is the source of at least
 *  one UPM.
 *
 *  section: the section to check
 *  offset: the offset from section that makes the address to check
 *  returns: true if the address formed from the specified section and
 *  offset is the source of at least one UPM.
 */
bool AddressIsUPMSource(Section *section, ADDRESS offset) {
    LinkedList *upms;

    upms = UPMTableGetBySourceAddress(section, offset);
    if (ll_size(upms) > 0) {
	ll_free(upms);
	return true;
    }

    ll_free(upms);
    return false;
}



/*
 *  UPMTableInsertByDestAddress(section,offset,upm) --
 *
 *  Inserts the specified UPM by the destination address formed from the
 *  specified section and offset from that section.
 *
 *  section: the destination section
 *  offset: the offset from section that will form the destination address
 *  upm: the UPM to insert
 */
void UPMTableInsertByDestAddress(Section *section, ADDRESS offset, Patch *upm) {
    UPMTableInsert(section, (void *) offset, upm, upm2_dest_table_addr);
}



/*
 *  UPMTableInsertBySourceAddress(section,offset,upm) --
 *
 *  Inserts the specified UPM by the source address formed from the
 *  specified section and offset from that section.
 *
 *  section: the source section
 *  offset: the offset from section that will form the source address
 *  upm: the UPM to insert
 */
void UPMTableInsertBySourceAddress(
Section *section, ADDRESS offset, Patch *upm) {
    UPMTableInsert(section, (void *) offset, upm, upm2_src_table_addr);
}



/*
 *  UPMTableGetFirstBySource(key) --
 *
 *  Given a pointer to a struct that is a source of a UPM, this function
 *  looks for all UPMs that were inserted by this key as a source and
 *  returns the first such UPM.
 *
 *  key: the key that is suspected of being a source of a UPM
 *  returns: the first UPM found to be associated with the specified key
 *  as a source
 *  see also: UPMTableInsertBySource
 *  see also: UPMTableGetBySource
 */
Patch *UPMTableGetFirstBySource(void *key) {
    LinkedList *all;
    Patch *upm;

    upm = NULL;

    all = UPMTableGetAll(key, key, upm2_src_table);
    if (ll_size(all) != 0) {
	upm = ll_removeFirst(all);
    }

    ll_free(all);

    return upm;
}



/*
 *  UPMTableGetBySourceAddress(section,offset) --
 *
 *  Returns a linked list of all of the UPMs that have the specified
 *  source address as a source.
 *
 *  section: the source section
 *  offset: the offset from section that will form the source address
 *  upm: the UPM to insert
 */
LinkedList *UPMTableGetBySourceAddress(Section *section, ADDRESS offset) {
    return UPMTableGetAll(section, (void *) offset, upm2_src_table_addr);
}



/*
 *  UPMTableGetByDestAddress(section,offset) --
 *
 *  Returns a list of all of the UPMs that have the specified destination
 *  address as a destination.
 *
 *  section: the destination section
 *  offset: the offset from section that will form the destination address
 *  upm: the UPM to insert
 */
LinkedList *UPMTableGetByDestAddress(Section *section, ADDRESS offset) {
    return UPMTableGetAll(section, (void *) offset, upm2_dest_table_addr);
}



/*
 *  UPMTableInsertBySource(key,upm) --
 *
 *  Given a pointer to a struct that is a source of the UPM, inserts the
 *  upm into a table.
 *
 *  key: a source of upm that upm should be inserted by
 *  upm: the UPM to be inserted
 */
void UPMTableInsertBySource(void *key, Patch *upm) {
    UPMTableInsert(key, key, upm, upm2_src_table);
}



/*
 *  UPMTableInsertByDest(key,upm) --
 *
 *  Given a pointer to a struct that is a destination of a UPM, this
 *  function inserts the upm into a table.
 *  UPM.
 *
 *  key: a destination of upm that upm should be inserted by
 *  upm: the UPM to be inserted
 */
void UPMTableInsertByDest(void *key, Patch *upm) {
    UPMTableInsert(key, key, upm, upm2_dest_table);
}



/*
 *  UPMTableGetBySource(key) --
 *
 *  Given a pointer to a struct that is a source of a UPM, this function
 *  looks for all UPMs that were inserted by this key as a source and
 *  returns a LinkedList containing all those found.
 *
 *  key: the key that is suspected of being a source of a UPM
 *  returns: a LinkedList containing all UPMs that were inserted by key
 *  as a source
 *  see also: UPMTableInsertBySource
 */
LinkedList *UPMTableGetBySource(void *key) {
    return UPMTableGetAll(key, key, upm2_src_table);
}



/*
 *  UPMTableGetByDest(key) --
 *
 *  Given a pointer to a struct that is a source of a UPM, this function
 *  looks for all UPMs that were inserted by this key as a destination
 *  and returns a LinkedList containing all those found.
 *
 *  key: the key that is suspected of being a destination of a UPM
 *  returns: a LinkedList containing all UPMs that were inserted by key as
 *  a destination
 *  see also: UPMTableInsertByDest
 */
LinkedList *UPMTableGetByDest(void *key) {
    return UPMTableGetAll(key, key, upm2_dest_table);
}



/*
 *  UPMGetDestInstr(upm) --
 *
 *  Given a UPM, this function will return the destination instruction,
 *  iff the UPM has only one instruction destination.
 *
 *  upm: the UPM to find a destination of
 *  returns: the instruction that is the sole destination of upm, NULL if
 *  there is no instruction destination or more than one instruction
 *  destination
 */
Instruction *UPMGetDestInstr(Patch *upm) {
    Instruction *instr, *dest;
    UPMList *destinations;
    int i;

    dest = instr = NULL;

    destinations = upm->p_GetDestinations(upm);
    for (i = 0; i < destinations->u_size; i++) {
	if (UPM_MEMBER_INSTR(&destinations->u_members[i]) != NULL) {
	    instr = *UPM_MEMBER_INSTR(&destinations->u_members[i]);
	}

	if (instr != NULL) {
	    if (dest != NULL) {
		return NULL;
	    } else {
		dest = instr;
	    }
	}
    }

    return dest;
}



/*
 *  UPMGetDestSection(upm,offset) --
 *
 *  Given a UPM, this function will return the destination section and
 *  fill in the offset, iff the UPM has only one section destination.
 *
 *  upm: the UPM to find a destination of
 *  returns: the section that is the sole destination of upm, NULL if there
 *  is no section destination or more than one section destination
 */
Section *UPMGetDestSection(Patch *upm, ADDRESS *offset) {
    Section *section, *dest;
    UPMList *destinations;
    int i;

    dest = section = NULL;

    destinations = upm->p_GetDestinations(upm);
    for (i = 0; i < destinations->u_size; i++) {
	if (UPM_MEMBER_SECTION(&destinations->u_members[i]) != NULL) {
	    section = *UPM_MEMBER_SECTION(&destinations->u_members[i]);
	}

	if (section != NULL) {
	    if (dest != NULL) {
		return NULL;
	    } else {
		dest = section;
		i++;
		*offset = *UPM_MEMBER_OFFSET(&destinations->u_members[i]);
	    }
	}
    }

    return dest;
}



/*
 *  UPMGetSourceSection(upm,offset) --
 *
 *  Given a UPM, this function will return the source section and
 *  fill in the offset, iff the UPM has only one section source.
 *
 *  upm: the UPM to find a source of
 *  returns: the section that is the sole source of upm, NULL if there
 *  is no section source or more than one section source
 */
Section *UPMGetSourceSection(Patch *upm, ADDRESS *offset) {
    Section *section, *source;
    UPMList *sources;
    int i;

    source = section = NULL;

    sources = upm->p_GetDestinations(upm);
    for (i = 0; i < sources->u_size; i++) {
	if (UPM_MEMBER_SECTION(&sources->u_members[i]) != NULL) {
	    section = *UPM_MEMBER_SECTION(&sources->u_members[i]);
	}

	if (section != NULL) {
	    if (source != NULL) {
		return NULL;
	    } else {
		source = section;
		i++;
		*offset = *UPM_MEMBER_OFFSET(&sources->u_members[i]);
	    }
	}
    }

    return source;
}



/*
 *  UPMTableInsert(key1,key2,upm,table) --
 *
 *  Given two keys a upm and a upm hash table, this function inserts upm
 *  into the specified table based on the specified keys. If the insertion
 *  only needs one key, this function should be called using that key as
 *  both the first and second key.
 *
 *  key1: the first key component
 *  key2: the second key component
 *  upm: the UPM to be inserted
 *  table: the hash table to insert into
 */
static void UPMTableInsert(void *key1, void *key2, Patch *upm, UPMHash **table){
    UPMHash *new_node;
    int index;

    new_node = zalloc(sizeof(UPMHash));
    new_node->uh_key1 = key1;
    new_node->uh_key2 = key2;
    new_node->uh_upm = upm;

    index = UPMHash2(key1, key2);

    if (table[index] == NULL) {
	table[index] = new_node;
	new_node->uh_hash_next = new_node->uh_hash_prev = NULL;
    } else {
	new_node->uh_hash_next = table[index];
	new_node->uh_hash_prev = NULL;
	table[index]->uh_hash_prev = new_node;
	table[index] = new_node;
    }
}



/*
 *  UPMTableGetAll(key1,key2,table) --
 *
 *  Returns a LinkedList of all of the entries in the specified hash table
 *  that have key1 and key2 as their keys.
 *
 *  key1: the first key component
 *  key2: the second key component
 *  table: the UPM hash table to look in
 *  returns: a LinkedList containing all of the entries in table that have
 *  keys that match the specified keys
 */
static LinkedList *UPMTableGetAll(void *key1, void *key2, UPMHash **table) {
    UPMHash *node;
    LinkedList *upms;
    int index;

    upms = ll_new(LIST_UNSORTED);

    index = UPMHash2(key1, key2);

    node = table[index];
    while (node != NULL) {
	if ((node->uh_key1 == key1) && (node->uh_key2 == key2)) {
	    if (!ll_contains(upms, node->uh_upm)) {
		ll_add(upms, node->uh_upm);
	    }
	}

	node = node->uh_hash_next;
    }

    return upms;
}



/*
 *  UPMTableRemovePair(key1,key2,upm,table) --
 *
 *  Given two keys and a UPM, this function will remove only those
 *  entries from table that have both key1 and key2 as their keys and have
 *  upm as the inserted item. If only one key is available to be supplied
 *  then it should be supplied as both the first and second keys.
 *
 *  key1: the first key component
 *  key2: the second key component
 *  upm: the UPM to be matched as the inserted item
 *  table: the UPM hash table to remove from
 */
static void UPMTableRemovePair(
void *key1, void *key2, Patch *upm, UPMHash **table) {
    UPMHash *node, *next;
    int index;

    index = UPMHash2(key1, key2);

    node = table[index];
    while (node != NULL) {
	next = node->uh_hash_next;

	if ((node->uh_key1 == key1) && (node->uh_key2 == key2) &&
	    (node->uh_upm == upm)) {
	    if (node == table[index]) {
		table[index] = (table[index])->uh_hash_next;

		if (table[index] != NULL) {
		    table[index]->uh_hash_prev = NULL;
		}
	    } else {
		if (node->uh_hash_prev != NULL) {
		    node->uh_hash_prev->uh_hash_next = node->uh_hash_next;
		}

		if (node->uh_hash_next != NULL) {
		    node->uh_hash_next->uh_hash_prev = node->uh_hash_prev;
		}
	    }

	    free(node);
	}

	node = next;
    }
}



/*
 *  UPMTableRemove(key1,key2,table) --
 *
 *  Given two keys and a UPM hash table, this function removes any
 *  entries in the table that have the specified keys as keys. If only one
 *  key is available to be supplied, it should be supplied as both the
 *  first and second keys.
 *
 *  key1: the first key component
 *  key2: the second key component
 *  table: the UPM hash table to insert into
 */
static void UPMTableRemove(void *key1, void *key2, UPMHash **table) {
    UPMHash *node, *next;
    int index;

    index = UPMHash2(key1, key2);
    node = table[index];
    while (node != NULL) {
	next = node->uh_hash_next;

	if ((node->uh_key1 == key1) && (node->uh_key2 == key2)) {
	    if (node == table[index]) {
		table[index] = (table[index])->uh_hash_next;

		if (table[index] != NULL) {
		    table[index]->uh_hash_prev = NULL;
		}
	    } else {
		if (node->uh_hash_prev != NULL) {
		    node->uh_hash_prev->uh_hash_next = node->uh_hash_next;
		}

		if (node->uh_hash_next != NULL) {
		    node->uh_hash_next->uh_hash_prev = node->uh_hash_prev;
		}
	    }

	    free(node);
	}

	node = next;
    }
}



/*
 *  UPMFree(upm) --
 *
 *  General Purpose UPM Functions
 *
 *  Frees the memory allocated for the specified UPM.
 *
 *  upm: the UPM whose memory should be freed
 */
void UPMFree(Patch *upm) {
    free(upm->p_patch);
    free(upm);
}



/*
 *  UPMCreate(patch,Apply,CloneToInstr,GetSources,GetDestinations) --
 *
 *  Creates and returns a new UPM that will use the specified patch,
 *  apply function and clone instruction.
 *
 *  patch: the patch that the UPM will use
 *  Apply: the apply function that the UPM should use
 *  CloneToInstr: the clone function that the UPM should use
 *  GetSources: the function that will return all sources for the UPM
 *  GetDestinations: the function that will return all destinations of the UPM
 *  returns: the newly created UPM
 */
Patch *UPMCreate(void *patch, bool (*Apply) (Patch *upm),
void (*CloneToInstr) (Patch *upm, Instruction *dest_instr),
UPMList *(*GetSources) (Patch *upm),
UPMList *(*GetDestinations) (Patch *upm)) {
    Patch *upm;

    upm = zalloc(sizeof(Patch));
    upm->p_patch = patch;
    upm->p_Apply = Apply;
    upm->p_CloneToInstr = CloneToInstr;
    upm->p_GetSources = GetSources;
    upm->p_GetDestinations = GetDestinations;

    return upm;
}



/*
 *  UPMMemberListCreate(num_members) --
 *
 *  Given a number that represents the number of members that the list
 *  should contain, this function creates and returns a
 *  UPMList that has that many elements allocated for it's
 *  internal member list and has the correct list size set.
 *
 *  num_members: the number of Locations that the newly
 *  created list should have
 *  returns: the initialized UPMList
 */
UPMList *UPMMemberListCreate(int num_members) {
    UPMList *list;
    Location *members;

    members = zalloc(sizeof(Location) * num_members);
    list = zalloc(sizeof(UPMList));

    list->u_size = num_members;
    list->u_members = members;

    return list;
}



/*
 *  UPMMemberListDestroy(list) --
 *
 *  Given a pointer to a UPMList, this function
 *  deallocates the memory for this list and it's internal list.
 */
void UPMMemberListDestroy(UPMList *list) {
    free(list->u_members);
    free(list);
}



/*
 *  UPMChangeSourceFromSectionAddressToSectionAddress
 *  (upm,old_sec,old_offset,new_sec,new_offset) --
 *
 *  Given a UPM, a section and offset that are at least one of the
 *  sources of the specified UPM, this function changes all sources of
 *  the upm that pointed to the original section to point to the new
 *  offset from the new section instead.
 *
 *  upm: the UPM to be modified
 *  old_sec: the original section
 *  old_offset: the original offset into the original section
 *  new_sec: the new section
 *  new_offset: the new offset into the new section
 */
void UPMChangeSourceFromSectionAddressToSectionAddress(
Patch *upm, Section *old_sec, ADDRESS old_offset,
Section *new_sec, ADDRESS new_offset) {
    UPMList *sources;
    Location *maybe_sec, *maybe_addr;
    int i;

    sources = upm->p_GetSources(upm);
    for (i = 0; i < sources->u_size - 1; i++) {
	maybe_sec = &sources->u_members[i];
	maybe_addr = &sources->u_members[i + 1];

	if ((UPM_MEMBER_SECTION(maybe_sec) != NULL) &&
	    (UPM_MEMBER_OFFSET(maybe_addr) != NULL) &&
	    (*UPM_MEMBER_SECTION(maybe_sec) == old_sec) &&
	    (*UPM_MEMBER_OFFSET(maybe_addr) == old_offset)) {
	    *UPM_MEMBER_SECTION(maybe_sec) = new_sec;
	    *UPM_MEMBER_OFFSET(maybe_addr) = new_offset;

	    UPMTableRemove(*UPM_MEMBER_SECTION(maybe_sec),
		(void *) *UPM_MEMBER_OFFSET(maybe_addr), upm2_src_table_addr);
	    UPMTableInsertBySourceAddress(new_sec, new_offset, upm);
	}
    }

    UPMMemberListDestroy(sources);
}



/*
 *  UPMChangeDestFromSectionAddressToSectionAddress
 *  (upm,old_sec,old_offset,new_sec,new_offset) --
 *
 *  Given a UPM, a section and offset that are at least one of the
 *  destinations of the specified UPM, this function changes all
 *  destinations of the upm that pointed to the original section to point
 *  to the new offset from the new section instead.
 *
 *  upm: the UPM to be modified
 *  old_sec: the original section
 *  old_offset: the original offset into the original section
 *  new_sec: the new section
 *  new_offset: the new offset into the new section
 */
void UPMChangeDestFromSectionAddressToSectionAddress(
Patch *upm, Section *old_sec, ADDRESS old_offset,
Section *new_sec, ADDRESS new_offset) {
    UPMList *destinations;
    Location *maybe_sec, *maybe_addr;
    int i;

    destinations = upm->p_GetDestinations(upm);
    for (i = 0; i < destinations->u_size - 1; i++) {
	maybe_sec = &destinations->u_members[i];
	maybe_addr = &destinations->u_members[i + 1];

	if ((UPM_MEMBER_SECTION(maybe_sec) != NULL) &&
	    (UPM_MEMBER_OFFSET(maybe_addr) != NULL) &&
	    (*UPM_MEMBER_SECTION(maybe_sec) == old_sec) &&
	    (*UPM_MEMBER_OFFSET(maybe_addr) == old_offset)) {
	    *UPM_MEMBER_SECTION(maybe_sec) = new_sec;
	    *UPM_MEMBER_OFFSET(maybe_addr) = new_offset;

	    UPMTableRemove(*UPM_MEMBER_SECTION(maybe_sec),
		(void *) *UPM_MEMBER_OFFSET(maybe_addr), upm2_dest_table_addr);
	    UPMTableInsertByDestAddress(new_sec, new_offset, upm);
	}
    }

    UPMMemberListDestroy(destinations);
}



/*
 *  UPMChangeDestFromInstrToInstr(upm,old,old_op,new,new_op) --
 *
 *  Given a UPM, an instruction and operand that make up one or more
 *  destination(s) and a new instruction and operand, this function
 *  'repoints' any of the destination members of the upm that point at the
 *  old pair to point to the new pair.
 *
 *  upm: the UPM to modify
 *  old_instr: the instruction that upm originally points to
 *  old_op: the operand in old_instr that upm originally points to
 *  new_instr: the instruction that the function will be repointed to
 *  new_op: the operand in new_instr that upm will be repointed to
 */
void UPMChangeDestFromInstrToInstr(Patch *upm,
Instruction *old, Operand *old_op, Instruction *new, Operand *new_op) {
    UPMList *destinations;
    Location *member;
    int i;

    destinations = upm->p_GetDestinations(upm);
    for (i = 0; i < destinations->u_size; i++) {
	member = &destinations->u_members[i];

	if ((UPM_MEMBER_INSTR(member) != NULL) &&
	    (*UPM_MEMBER_INSTR(member) == old)) {
	    UPMTableRemove(*UPM_MEMBER_INSTR(member), *UPM_MEMBER_INSTR(member),
		upm2_dest_table);
	    UPMTableInsertByDest(new, upm);
	    *UPM_MEMBER_INSTR(member) = new;
	}

	if ((UPM_MEMBER_OPERAND(member) != NULL) &&
	    (*UPM_MEMBER_OPERAND(member) == old_op)) {
	    UPMTableRemove(old_op, old_op, upm2_dest_table);
	    UPMTableInsertByDest(new_op, upm);
	    *UPM_MEMBER_OPERAND(member) = new_op;
	}
    }

    UPMMemberListDestroy(destinations);
}



/*
 *  UPMChangeSourceFromInstrToInstr(upm,old,new) --
 *
 *  Given a UPM, and instruction that is a source of upm and another
 *  instruction, this function changes all sources of upm that point to
 *  old to point to new_instr instead. This function is untested.
 *
 *  upm: the UPM to redirect
 *  old: the instruction that we are moving the UPM from
 *  new_instr: the instruction that we are moving the UPM to
 */
bool UPMChangeSourceFromInstrToInstr(
Patch *upm, Instruction *old, Instruction *new) {
    UPMList *sources;
    Location *member;
    int i;
    bool changed;

    changed = false;

    sources = upm->p_GetSources(upm);
    for (i = 0; i < sources->u_size; i++) {
	member = &sources->u_members[i];

	if ((UPM_MEMBER_INSTR(member) != NULL) &&
	    (*UPM_MEMBER_INSTR(member) == old)) {
	    UPMTableRemovePair(*UPM_MEMBER_INSTR(member),
		*UPM_MEMBER_INSTR(member), upm, upm2_src_table);
	    UPMTableInsertBySource(new, upm);
	    *UPM_MEMBER_INSTR(member) = new;
	    changed = true;
	}
    }

    UPMMemberListDestroy(sources);

    return changed;
}



/*
 *  UPMChangeAllSourcesFromInstrToInstr(old_instr,new_instr) --
 *
 *  This function repoints all UPMs associated with old_instr at
 *  new_instr instead. If the instructions aren't of the same type that's
 *  OK, since this function merely looks at all UPMs that source
 *  old_instr and modify them to source new_instr instead.
 *
 *  old_instr: the instruction whose UPMs should be 'repointed'
 *  new_instr: the instruction that old_instr's UPMs will be
 *  pointed to
 */
void UPMChangeAllSourcesFromInstrToInstr(
Instruction *old_instr, Instruction *new_instr) {
    LinkedList *upms;
    Iterator *itr;
    Patch *upm;

    upms = UPMTableGetBySource(old_instr);
    itr = ll_iterator(upms);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);

	UPMChangeSourceFromInstrToInstr(upm, old_instr, new_instr);
    }

    free(itr);
    ll_free(upms);
}



/*
 *  UPMChangeAllDestinationsFromInstrToInstr(old_instr,new_instr) --
 *
 *  This function moves any UPMs associated with old_instr to be
 *  associated with new_instr instead. This function can be dangerous
 *  depending on how it is used. This function repoints any UPM
 *  associated with the first source operand of old_instr to point to the
 *  first source operand of new_instr, the second source of old_instr to
 *  point to the second source of new_instr and so on -- so if there isn't
 *  a meaningful relationship between the operands of the two
 *  instructions, one may get undesirable results.
 *
 *  old_instr: the instruction whose UPMs should be moved
 *  new_instr: the instruction that should receive the UPMs
 */
void UPMChangeAllDestinationsFromInstrToInstr(
Instruction *old_instr, Instruction *new_instr) {
    LinkedList *upms;
    Iterator *itr;
    Patch *upm;

    upms = UPMTableGetByDest(old_instr->i_src);
    itr = ll_iterator(upms);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);

	UPMChangeDestFromInstrToInstr(upm, old_instr, old_instr->i_src,
	    new_instr, new_instr->i_src);
    }
    free(itr);
    ll_free(upms);

    upms = UPMTableGetByDest(old_instr->i_src2);
    itr = ll_iterator(upms);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);

	UPMChangeDestFromInstrToInstr(upm, old_instr, old_instr->i_src2,
	    new_instr, new_instr->i_src2);
    }
    free(itr);
    ll_free(upms);

    upms = UPMTableGetByDest(old_instr->i_dest);
    itr = ll_iterator(upms);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);

	UPMChangeDestFromInstrToInstr(upm, old_instr, old_instr->i_dest,
	    new_instr, new_instr->i_dest);
    }
    free(itr);
    ll_free(upms);
}



/*
 *  ApplyUPMs(upm2s) --
 *
 *  Iterates through the specified linked list containing UPMs and calls
 *  the Apply function for each of them.
 *
 *  upm2_list: the LinkedList of UPMs to be applied
 *  returns: the number of UPMs from upm2_list whose patch value was a
 *  change from it's previous value (i.e., the number of patches whose
 *  apply functions returned true)
 */
int ApplyUPMs(LinkedList *upm2s) {
    Iterator *itr;
    Patch *uiterator;
    int changed;

    changed = 0;

    itr = ll_iterator(upm2s);

    while (itr->it_hasNext(itr)) {
	uiterator = (Patch *) itr->it_next(itr);

	if ((uiterator->p_Apply) (uiterator)) {
	    changed++;
	}
    }

    free(itr);

    return changed;
}



/*
 *  FunDestroyUPMs(fun) --
 *
 *  Given a Function, this function removes and destroys all UPMs
 *  that either have this instruction as a source or as a destination.
 *
 *  fun: the function whose UPMs should be destroyed
 */
void FunDestroyUPMs(Function *fun) {
    Patch *upm;
    LinkedList *upms, *removed;
    Iterator *itr;
    UPMList *list;
    int i;

// REPORT(RP_TRIVIAL, "Destroying %s", fun->f_name);

    removed = ll_new(LIST_UNSORTED);

    /*
     * We get all of the UPMs that patch into this function, and remove
     * each of them by each of the objects they've been inserted into the
     * tables by, then destroy the UPM.
     */

    upms = UPMTableGetByDest(fun);
    itr = ll_iterator(upms);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);

	if (ll_contains(removed, upm)) {
	    continue;
	}
//  REPORT(RP_TRIVIAL, "removing upm: %p", upm);

	ll_remove(upm2s_list, upm);
	ll_add(removed, upm);
	list = upm->p_GetDestinations(upm);
	for (i = 0; i < list->u_size; i++) {
	    UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		*(ADDRESS **) list->u_members[i].l_member, upm2_dest_table);
	}
	UPMMemberListDestroy(list);
	list = upm->p_GetSources(upm);
	for (i = 0; i < list->u_size; i++) {
	    UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		*(ADDRESS **) list->u_members[i].l_member, upm2_dest_table);
	}
	UPMMemberListDestroy(list);
    }
    free(itr);
    ll_free(upms);


    /*
     * We get all of the UPMs that patch from this function, and remove
     * each of them by each of the objects they've been inserted into the
     * tables by, then destroy the UPM.
     */

    upms = UPMTableGetBySource(fun);
    itr = ll_iterator(upms);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);

	if (ll_contains(removed, upm)) {
	    continue;
	}
//  REPORT(RP_TRIVIAL, "removing upm: %p", upm);
	ll_remove(upm2s_list, upm);
	ll_add(removed, upm);
	list = upm->p_GetDestinations(upm);
	for (i = 0; i < list->u_size; i++) {
	    UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		*(ADDRESS **) list->u_members[i].l_member, upm2_src_table);
	}
	UPMMemberListDestroy(list);

	list = upm->p_GetSources(upm);
	for (i = 0; i < list->u_size; i++) {
	    UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		*(ADDRESS **) list->u_members[i].l_member, upm2_src_table);
	}
	UPMMemberListDestroy(list);
    }
    free(itr);
    ll_free(upms);

    itr = ll_iterator(removed);
    while (itr->it_hasNext(itr)) {
	UPMFree((Patch *) itr->it_next(itr));
    }
    free(itr);
    ll_free(removed);
}



/*
 *  InstrDestroyUPMs(instr) --
 *
 *  Given an instruction, this function removes and destroys all UPMs
 *  that either have this instruction as a source or as a destination.
 *
 *  instr: the instruction whose UPMs should be destroyed
 */
void InstrDestroyUPMs(Instruction *instr) {
    Patch *upm;
    LinkedList *upms, *removed;
    Iterator *itr;
    UPMList *list;
    int i;

    removed = ll_new(LIST_UNSORTED);

    /*
     * We get all of the UPMs that patch into this instruction, and remove
     * each of them by each of the objects they've been inserted into the
     * tables by, then destroy the UPM.
     */

    if (isDataInstruction(instr)) {
        printf("removing data instruction related UPMS\n");
        upms = UPMTableGetByDest(instr);
        
        itr = ll_iterator(upms);
        while (itr->it_hasNext(itr)) {
            upm = (Patch *) itr->it_next(itr);

            if (ll_contains(removed, upm)) {
                continue;
            }

            ll_remove(upm2s_list, upm);
            ll_add(removed, upm);
            list = upm->p_GetDestinations(upm);
            for (i = 0; i < list->u_size; i++) {
                UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
                    *(ADDRESS **) list->u_members[i].l_member, upm2_dest_table);
            }
            UPMMemberListDestroy(list);
            list = upm->p_GetSources(upm);
            for (i = 0; i < list->u_size; i++) {
                UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
                    *(ADDRESS **) list->u_members[i].l_member, upm2_src_table);
            }
            UPMMemberListDestroy(list);
        }
        free(itr);
        ll_free(upms);        
    }
    
    if (instr->i_src != NULL) {
	upms = UPMTableGetByDest(instr->i_src);
	itr = ll_iterator(upms);
	while (itr->it_hasNext(itr)) {
	    upm = (Patch *) itr->it_next(itr);

	    if (ll_contains(removed, upm)) {
		continue;
	    }

	    ll_remove(upm2s_list, upm);
	    ll_add(removed, upm);
	    list = upm->p_GetDestinations(upm);
	    for (i = 0; i < list->u_size; i++) {
		UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		    *(ADDRESS **) list->u_members[i].l_member, upm2_dest_table);
	    }
	    UPMMemberListDestroy(list);
	    list = upm->p_GetSources(upm);
	    for (i = 0; i < list->u_size; i++) {
		UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		    *(ADDRESS **) list->u_members[i].l_member, upm2_src_table);
	    }
	    UPMMemberListDestroy(list);
	}
	free(itr);
	ll_free(upms);
    }

    if (instr->i_src2 != NULL) {
	upms = UPMTableGetByDest(instr->i_src2);
	itr = ll_iterator(upms);
	while (itr->it_hasNext(itr)) {
	    upm = (Patch *) itr->it_next(itr);

	    if (ll_contains(removed, upm)) {
		continue;
	    }
//    REPORT(RP_TRIVIAL, "removing upm: %p", upm);

	    ll_remove(upm2s_list, upm);
	    ll_add(removed, upm);
	    list = upm->p_GetDestinations(upm);
	    for (i = 0; i < list->u_size; i++) {
		UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		    *(ADDRESS **) list->u_members[i].l_member, upm2_dest_table);
	    }
	    UPMMemberListDestroy(list);
	    list = upm->p_GetSources(upm);
	    for (i = 0; i < list->u_size; i++) {
		UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		    *(ADDRESS **) list->u_members[i].l_member, upm2_src_table);
	    }
	    UPMMemberListDestroy(list);
	}
	free(itr);
	ll_free(upms);
    }

    if (instr->i_dest != NULL) {
	upms = UPMTableGetByDest(instr->i_dest);
	itr = ll_iterator(upms);
	while (itr->it_hasNext(itr)) {
	    upm = (Patch *) itr->it_next(itr);

	    if (ll_contains(removed, upm)) {
		continue;
	    }
//    REPORT(RP_TRIVIAL, "removing upm: %p", upm);

	    ll_remove(upm2s_list, upm);
	    ll_add(removed, upm);
	    list = upm->p_GetDestinations(upm);
	    for (i = 0; i < list->u_size; i++) {
		UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		    *(ADDRESS **) list->u_members[i].l_member, upm2_dest_table);
	    }
	    UPMMemberListDestroy(list);
	    list = upm->p_GetSources(upm);
	    for (i = 0; i < list->u_size; i++) {
		UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		    *(ADDRESS **) list->u_members[i].l_member, upm2_src_table);
	    }
	    UPMMemberListDestroy(list);
	}
	free(itr);
	ll_free(upms);
    }

    /*
     * We get all of the UPMs that patch from this instruction, and remove
     * each of them by each of the objects they've been inserted into the
     * tables by, then destroy the UPM.
     */
    upms = UPMTableGetBySource(instr);
    itr = ll_iterator(upms);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);

	if (ll_contains(removed, upm)) {
	    continue;
	}
//  REPORT(RP_TRIVIAL, "removing upm: %p", upm);
	ll_remove(upm2s_list, upm);
	ll_add(removed, upm);
	list = upm->p_GetDestinations(upm);
	for (i = 0; i < list->u_size; i++) {
	    UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		*(ADDRESS **) list->u_members[i].l_member, upm2_dest_table);
	}
	UPMMemberListDestroy(list);

	list = upm->p_GetSources(upm);
	for (i = 0; i < list->u_size; i++) {
	    UPMTableRemove(*(ADDRESS **) list->u_members[i].l_member,
		*(ADDRESS **) list->u_members[i].l_member, upm2_src_table);
	}
	UPMMemberListDestroy(list);
    }
    free(itr);
    ll_free(upms);

    itr = ll_iterator(removed);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);
	UPMFree(upm);
    }
    free(itr);
    ll_free(removed);
}



/*
 *  CloneInstrUPMsToInstr(src_instr,dest_instr) --
 *
 *  Given an instruction that may have UPMs associated with it and
 *  another instruction, this function clones any UPMs associated with
 *  the operands of the first into to the same operands of the section
 *  instruction.
 *
 *  src_instr: the instruction whose UPMs will be cloned
 *  dest_instr: the instruction that will receive the newly cloned
 *  UPMs
 */
void CloneInstrUPMsToInstr(Instruction *src_instr, Instruction *dest_instr) {
    Patch *upm;
    LinkedList *upms;
    Iterator *itr;

    /*
     * Here we get all upms that have source1 as a destination, and for
     * each of those, we clone it to the new destination instruction. One
     * might argue that we don't need to iterate through the upms since
     * only one UPM should have the operand as a destination, but that's
     * not true.  It is possible that a new patch function modifies some
     * aspect of the operand other than the displacement (i.e., maybe one
     * upm modifies the operand's flags, while another modifies the
     * displacement - in this case there would be two upms and both would
     * need to be cloned).
     *
     * We then do the same thing for source2 and dest.
     */

    upms = UPMTableGetByDest(src_instr->i_src);
    itr = ll_iterator(upms);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);

	if (upm->p_CloneToInstr == NULL) {
	    WARNING("No clone defined for patch");
	} else {
	    upm->p_CloneToInstr(upm, dest_instr);
	}
    }
    free(itr);
    ll_free(upms);

    upms = UPMTableGetByDest(src_instr->i_src2);
    itr = ll_iterator(upms);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);

	if (upm->p_CloneToInstr == NULL) {
	    WARNING("No clone defined for patch");
	} else {
	    upm->p_CloneToInstr(upm, dest_instr);
	}
    }
    free(itr);
    ll_free(upms);

    upms = UPMTableGetByDest(src_instr->i_dest);
    itr = ll_iterator(upms);
    while (itr->it_hasNext(itr)) {
	upm = (Patch *) itr->it_next(itr);

	if (upm->p_CloneToInstr == NULL) {
	    WARNING("No clone defined for patch");
	} else {
	    upm->p_CloneToInstr(upm, dest_instr);
	}
    }
    free(itr);
    ll_free(upms);
}



/*
 *  DestroyUPMs(upm2s) --
 */
void DestroyUPMs(LinkedList *upm2s) {
    Patch *uiterator;
    Iterator *itr;

    if (upm2s == NULL) {
	REPORT(RP_LGDETAIL, "UPM List is NULL");
	return;
    }

    itr = ll_iterator(upm2s);
    while (itr->it_hasNext(itr)) {
	uiterator = (Patch *) itr->it_next(itr);

	UPMFree(uiterator);
    }

    ll_clear(upm2s);
}



/*
 *  AddSymbolUPMs() --
 *
 *  Adds patches for each symbol currently in the global list of Symbols.
 *
 *  see also: s_list
 *  see also: AddPatchFunctionAddressIntoSymbol
 *  see also: AddPatchSectionAddressIntoSymbol
 *  see also: AddPatchInstrAddressIntoSymbol
 */
void AddSymbolUPMs() {
    Symbol *symiterator;
    Instruction *instr;
    Section *section;

    FOREACH_SYMBOL(symiterator) {
	if (symiterator->sy_flags & SYMBOL_UNDEFINED_FLAG) {
	    continue;
	}

	// _etext is a special symbol which indicates the end
 	// of the text section.
	if ((strcmp(symiterator->sy_name, "_etext") == 0)) {
	    AddPatchSectionEndIntoSymbol(SECTION_REF_CODE,
				symiterator);
	    //symiterator->sy_section = NULL;
	    continue;
	}

	if ((strcmp(symiterator->sy_name, "_edata") == 0)) {
	    AddPatchSectionEndIntoSymbol(SECTION_REF_DATA,
				symiterator);
	    //symiterator->sy_section = NULL;
	    continue;
	}
        
	instr = AddressToInstr(symiterator->sy_value);

	if (instr != NULL) {
	    if (symiterator->sy_flags & SYMBOL_FUNCTION_FLAG) {
		AddPatchFunctionAddressIntoSymbol(INSTR_FUNCTION(instr),
		    symiterator);
	    } else if (symiterator->sy_flags & SYMBOL_SECTION_FLAG) {
		AddPatchSectionAddressIntoSymbol(SECTION_REF_CODE, 0,
		    symiterator);
	    } else {
		AddPatchInstrAddressIntoSymbol(instr, symiterator);
	    }
	} else if (symiterator->sy_value != 0) {
	    FList *flist;

	    flist = GetFlistFunAtAddress(symiterator->sy_value);
	    if (flist != NULL) {
		WARNING("Need to add patch for symbol %s",symiterator->sy_name);
		// was:  AddPatchSectionAddressIntoSymbol(SECTION_REF_FLISTS,
		//           flist->fl_new_offset, symiterator);
	    } else if (symiterator->sy_flags & SYMBOL_OBJECT_END_FLAG) {
		section = AddressToSection(symiterator->sy_value - 1);
		AddPatchSectionEndIntoSymbol(section, symiterator);
	    } else {
		section = AddressToSection(symiterator->sy_value);
		if (section != NULL) {
		    AddPatchSectionAddressIntoSymbol(section,
			symiterator->sy_value - section->s_address,
			symiterator);
		} else {
		    if ((symiterator->sy_flags & SYMBOL_ABSOLUTE_FLAG) &&
			symiterator->sy_section) {
			AddPatchSectionAddressIntoSymbol(symiterator->sy_section,
                            symiterator->sy_value -
                            symiterator->sy_section->s_address,
                            symiterator);
		    }
		}
	    }
	}
    }
}



/*
 *  patchSectionAddressIntoInstr(upm) --
 *
 *  Patches the final address of some section location into an instruction.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchSectionAddressIntoInstr
 */
bool patchSectionAddressIntoInstr(Patch *upm) {
    PatchSectionToInstr *patch;
    ADDRESS og_val, new_val;

    patch = (PatchSectionToInstr *) upm->p_patch;

    CASSERT(patch->p_dest_instr != NULL, "Patching into a NULL instruction");
    CASSERT(patch->p_src_section != NULL, "Patching into a NULL section");
    CASSERT(patch->p_dest_op != NULL, "Patching into a NULL operand");

    og_val = patch->p_dest_op->a_displacement;
    new_val = patch->p_src_section->s_newaddress + patch->p_src_offset;

    if (og_val != new_val) {
	patch->p_dest_op->a_displacement = new_val;
	UpdateInstr(patch->p_dest_instr);

	return true;
    }

    return false;
}



/*
 *  patchSectionAddressToInstrClone(upm,dest_instr) --
 */
static void patchSectionAddressToInstrClone(
Patch *upm, Instruction *dest_instr) {
    PatchSectionToInstr *patch;
    Operand *op;

    patch = (PatchSectionToInstr *) upm->p_patch;

    if (patch->p_dest_op == patch->p_dest_instr->i_src) {
	op = dest_instr->i_src;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_src2) {
	op = dest_instr->i_src2;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_dest) {
	op = dest_instr->i_dest;
    }

    AddPatchSectionAddressIntoInstr(patch->p_src_section,
	patch->p_src_offset, dest_instr, op);
}



/*
 *  patchSectionAddressIntoInstrGetSources(upm) --
 */
UPMList *patchSectionAddressIntoInstrGetSources(Patch *upm) {
    PatchSectionToInstr *patch;
    UPMList *list;

    patch = (PatchSectionToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_offset;

    return list;
}



/*
 *  patchSectionAddressIntoInstrGetDestinations(upm) --
 */
UPMList *patchSectionAddressIntoInstrGetDestinations(Patch *upm) {
    PatchSectionToInstr *patch;
    UPMList *list;

    patch = (PatchSectionToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_instr;

    list->u_members[1].l_type = UPM_OPERAND_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_op;

    return list;
}



/*
 *  AddPatchSectionAddressIntoInstr(src_section,src_offset,dest_instr,dest_op)--
 *
 *  Adds a UPM into the global UPM list that will patch the final
 *  address of src_offset from src_section into the specified operand of
 *  the specified instruction.
 *
 *  src_section: the section containing the location whose address
 *  will be examined
 *  src_offset: the offset from src_section that will be examined
 *  dest_instr: the instruction whose operand will be updated
 *  dest_op: the operand whose displacement will be updated
 *  returns: a pointer to the newly added UPM
 *  see also: PatchSectionToInstr
 *  see also: upm2s_list
 */
Patch *AddPatchSectionAddressIntoInstr(Section *src_section,
ADDRESS src_offset, Instruction *dest_instr, Operand *dest_op) {
    PatchSectionToInstr *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchSectionToInstr));
    patch->p_src_section = src_section;
    patch->p_src_offset = src_offset;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_op = dest_op;

    upm = UPMCreate(patch,
	&patchSectionAddressIntoInstr,
	&patchSectionAddressToInstrClone,
	&patchSectionAddressIntoInstrGetSources,
	&patchSectionAddressIntoInstrGetDestinations);

    UPMTableInsertByDest(dest_op, upm);
    UPMTableInsertBySourceAddress(src_section, src_offset, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchInstrSectionOffsetIntoInstr(upm) --
 *
 *  Patches the offset between a section location and an instruction
 *  into an instruction.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchInstrSectionOffsetIntoInstr
 */
bool patchInstrSectionOffsetIntoInstr(Patch *upm) {
    PatchSectionInstrToInstr *patch;
    ADDRESS og_val, new_val;

    patch = (PatchSectionInstrToInstr *) upm->p_patch;

    CASSERT(patch->p_src_instr != NULL, "src_instr == NULL");
    CASSERT(patch->p_src_section != NULL, "src_section == NULL");
    CASSERT(patch->p_dest_instr != NULL, "dest_instr == NULL");
    CASSERT(patch->p_dest_op != NULL, "dest_op == NULL");

    og_val = patch->p_dest_op->a_displacement;
    new_val = patch->p_src_section->s_newaddress + patch->p_src_offset -
	(patch->p_src_instr->i_final_address + patch->p_src_instr->i_length);

    if (og_val != new_val) {
	patch->p_dest_op->a_displacement = new_val;
	UpdateInstr(patch->p_dest_instr);
	return true;
    }

    return false;
}



/*
 *  patchInstrSectionOffsetIntoInstrClone(upm,dest_instr) --
 */
static void patchInstrSectionOffsetIntoInstrClone(
Patch *upm, Instruction *dest_instr) {
    PatchSectionInstrToInstr *patch;
    Operand *op;

    patch = (PatchSectionInstrToInstr *) upm->p_patch;

    if (patch->p_dest_op == patch->p_dest_instr->i_src) {
	op = dest_instr->i_src;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_src2) {
	op = dest_instr->i_src2;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_dest) {
	op = dest_instr->i_dest;
    }

    if (patch->p_src_instr == patch->p_dest_instr) {
	AddPatchInstrSectionOffsetIntoInstr(patch->p_src_section,
	    patch->p_src_offset, dest_instr, dest_instr, op);
    } else {
	AddPatchInstrSectionOffsetIntoInstr(patch->p_src_section,
	    patch->p_src_offset, patch->p_src_instr, dest_instr, op);
    }

}



/*
 *  patchInstrSectionOffsetIntoInstrGetSources(upm) --
 */
UPMList *patchInstrSectionOffsetIntoInstrGetSources(Patch *upm) {
    UPMList *list;
    PatchSectionInstrToInstr *patch;

    patch = (PatchSectionInstrToInstr *) upm->p_patch;
    list = UPMMemberListCreate(3);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_offset;

    list->u_members[2].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[2].l_member = &patch->p_src_instr;

    return list;
}



/*
 *  patchInstrSectionOffsetIntoInstrGetDestinations(upm) --
 */
UPMList *patchInstrSectionOffsetIntoInstrGetDestinations(Patch *upm) {
    UPMList *list;
    PatchSectionInstrToInstr *patch;

    patch = (PatchSectionInstrToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_instr;

    list->u_members[1].l_type = UPM_OPERAND_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_op;

    return list;
}



/*
 *  AddPatchInstrSectionOffsetIntoInstr
 *  (src_section,src_offset,src_instr,dest_instr,dest_op) --
 *
 *  Adds a UPM into the global UPM list that will patch the offset from
 *  src_instr to src_offset from src_section into the specified operand of
 *  the specified destination instruction. The offset is calculated as a
 *  PC relative displacement, i.e., address(src_instr) -
 *  (address(src_section) + src_offset.
 *
 *  src_section: the section that would be the target of
 *  the PC relative instruction
 *  src_offset: the offset from src_section that would be the target
 *  of the PC relative instruction
 *  src_instr: the instruction that would be the origin of the
 *  PC relative instruction
 *  dest_instr: the instruction whose operand will be updated
 *  dest_op: the operand whose displacement will be updated
 *  returns: a pointer to the newly added UPM
 *  see also: PatchSectionInstrToInstr
 *  see also: upm2s_list
 */
Patch *AddPatchInstrSectionOffsetIntoInstr(
Section *src_section, ADDRESS src_offset,
Instruction *src_instr, Instruction *dest_instr, Operand *dest_op) {
    PatchSectionInstrToInstr *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchSectionInstrToInstr));
    patch->p_src_section = src_section;
    patch->p_src_offset = src_offset;
    patch->p_src_instr = src_instr;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_op = dest_op;

    upm = UPMCreate(patch,
	&patchInstrSectionOffsetIntoInstr,
	&patchInstrSectionOffsetIntoInstrClone,
	&patchInstrSectionOffsetIntoInstrGetSources,
	&patchInstrSectionOffsetIntoInstrGetDestinations);

    UPMTableInsertBySource(src_instr, upm);
    UPMTableInsertByDest(dest_op, upm);
    UPMTableInsertBySourceAddress(src_section, src_offset, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchInstrSectionOffsetIntoSectionAddress(upm) --
 *
 *  Patches the offset between a section location and an instruction
 *  into a section location.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchInstrSectionOffsetIntoSectionAddress
 */
bool patchInstrSectionOffsetIntoSectionAddress(Patch *upm) {
    PatchSectionInstrToSection *patch;
    ADDRESS og_val, new_val;

    patch = (PatchSectionInstrToSection *) upm->p_patch;

    CASSERT(patch->p_src_section != NULL, "src_section == NULL");
    CASSERT(patch->p_dest_section != NULL, "dest_section == NULL");
    CASSERT(patch->p_src_instr != NULL, "src_instr == NULL");

    og_val = loadLE(&patch->p_dest_section->s_newdata[patch->p_dest_offset]);
    new_val = patch->p_src_instr->i_final_address -
	(patch->p_src_section->s_newaddress + patch->p_src_offset);

    if (og_val != new_val) {
	storeLE(&patch->p_dest_section->s_newdata
	    [patch->p_dest_offset], new_val);

	return true;
    }

    return false;
}



/*
 *  patchInstrSectionOffsetIntoSectionAddressGetSources(upm) --
 */
UPMList *patchInstrSectionOffsetIntoSectionAddressGetSources(Patch *upm) {
    UPMList *list;
    PatchSectionInstrToSection *patch;

    patch = (PatchSectionInstrToSection *) upm->p_patch;
    list = UPMMemberListCreate(3);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_instr;

    list->u_members[1].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_section;

    list->u_members[2].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[2].l_member = &patch->p_src_offset;

    return list;
}



/*
 *  patchInstrSectionOffsetIntoSectionAddressGetDestinations(upm) --
 */
UPMList *patchInstrSectionOffsetIntoSectionAddressGetDestinations(Patch *upm) {
    UPMList *list;
    PatchSectionInstrToSection *patch;

    patch = (PatchSectionInstrToSection *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_offset;

    return list;
}



/*
 *  AddPatchInstrSectionOffsetIntoSectionAddress
 *  (src_instr,src_section,src_offset,dest_section,dest_offset) --
 *
 *  Adds a UPM into the global UPM list that will patch the offset from
 *  src_offset from src_section to src_instr into the specified offset
 *  from dest_section. The offset is calculated as a PC relative displacement,
 *  i.e., (address(src_section) + src_offset - address(src_instr).
 *
 *  src_instr: the instruction that would be the target of the
 *  PC relative instruction
 *  src_section: the section that would be the origin of
 *  the PC relative instruction
 *  src_offset: the offset from src_section that would be the origin
 *  of the PC relative instruction
 *  dest_section: the section that will be patched
 *  dest_offset: the offset from dest_section where the patch will occur
 *  returns: a pointer to the newly added UPM
 *  see also: PatchSectionInstrToSection
 *  see also: upm2s_list
 */
Patch *AddPatchInstrSectionOffsetIntoSectionAddress(
Instruction *src_instr, Section *src_section, ADDRESS src_offset,
Section *dest_section, ADDRESS dest_offset) {
    PatchSectionInstrToSection *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchSectionInstrToSection));
    patch->p_src_instr = src_instr;
    patch->p_src_section = src_section;
    patch->p_src_offset = src_offset;
    patch->p_dest_section = dest_section;
    patch->p_dest_offset = dest_offset;

    upm = UPMCreate(patch,
	&patchInstrSectionOffsetIntoSectionAddress,
	NULL,
	&patchInstrSectionOffsetIntoSectionAddressGetSources,
	&patchInstrSectionOffsetIntoSectionAddressGetDestinations);

    UPMTableInsertBySource(src_instr, upm);
    UPMTableInsertBySourceAddress(src_section, src_offset, upm);
    UPMTableInsertByDestAddress(dest_section, dest_offset, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchInstrAddressIntoSectionAddress(upm) --
 *
 *  Patches an instruction address into a section location.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchInstrAddressIntoSectionAddress
 */
bool patchInstrAddressIntoSectionAddress(Patch *upm) {
    PatchInstrToSection *patch;
    ADDRESS og_val, new_val;

    patch = (PatchInstrToSection *) upm->p_patch;

    CASSERT(patch->p_dest_section != NULL, "dest_section == NULL");
    CASSERT(patch->p_src_instr != NULL, "src_instr == NULL");

    og_val = loadLE(&patch->p_dest_section->s_newdata[patch->p_dest_offset]);
    new_val = patch->p_src_instr->i_final_address;

    if (og_val != new_val) {
	// patch the address of the source instr into the section address
	storeLE(&patch->p_dest_section->s_newdata
	    [patch->p_dest_offset], new_val);
	return true;
    }

    return false;
}



/*
 *  patchInstrAddressIntoSectionAddressGetSources(upm) --
 */
UPMList *patchInstrAddressIntoSectionAddressGetSources(Patch *upm) {
    UPMList *list;
    PatchInstrToSection *patch;

    patch = (PatchInstrToSection *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_instr;

    return list;
}



/*
 *  patchInstrAddressIntoSectionAddressGetDestinations(upm) --
 */
UPMList *patchInstrAddressIntoSectionAddressGetDestinations(Patch *upm) {
    UPMList *list;
    PatchInstrToSection *patch;

    patch = (PatchInstrToSection *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_offset;

    return list;
}



/*
 *  AddPatchInstrAddressIntoSectionAddress(src_instr,dest_section,dest_offset)--
 *
 *  Adds a UPM into the global UPM list that will patch the final
 *  address of src_instr into dest_offset from dest_section.
 *
 *  src_instr: the instruction whose address will be examined
 *  dest_section: the section that will be patched
 *  dest_offset: the offset from dest_section where the patch will occur
 *  returns: a pointer to the newly added UPM
 *  see also: PatchInstrToSection
 *  see also: upm2s_list
 */
Patch *AddPatchInstrAddressIntoSectionAddress(
Instruction *src_instr, Section *dest_section, ADDRESS dest_offset) {
    PatchInstrToSection *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchInstrToSection));
    patch->p_src_instr = src_instr;
    patch->p_dest_section = dest_section;
    patch->p_dest_offset = dest_offset;
    upm = UPMCreate(patch,
	&patchInstrAddressIntoSectionAddress,
	NULL,
	&patchInstrAddressIntoSectionAddressGetSources,
	&patchInstrAddressIntoSectionAddressGetDestinations);

    UPMTableInsertBySource(src_instr, upm);
    UPMTableInsertByDestAddress(dest_section, dest_offset, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchSectionSectionOffsetIntoSectionAddress(upm) --
 *
 *  Patches the offset from one section to another section into a
 *  section location.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchSectionSectionOffsetIntoSectionAddress
 */
bool patchSectionSectionOffsetIntoSectionAddress(Patch *upm) {
    PatchSectionToSectionAux *patch;
    ADDRESS og_val, new_val;

    patch = (PatchSectionToSectionAux *) upm->p_patch;
    og_val = loadLE(&patch->p_dest_section->s_newdata[patch->p_dest_offset]);
    new_val = patch->p_src_section->s_newaddress + patch->p_src_offset -
	(patch->p_aux_section->s_newaddress + patch->p_aux_offset);

    if (og_val != new_val) {
	storeLE(
	    &patch->p_dest_section->s_newdata[patch->p_dest_offset], new_val);
	return true;
    }

    return false;
}



/*
 *  patchSectionSectionOffsetIntoSectionAddressGetSources(upm) --
 */
UPMList *patchSectionSectionOffsetIntoSectionAddressGetSources(Patch *upm) {
    UPMList *list;
    PatchSectionToSectionAux *patch;

    patch = (PatchSectionToSectionAux *) upm->p_patch;
    list = UPMMemberListCreate(4);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_offset;

    list->u_members[2].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[2].l_member = &patch->p_aux_section;

    list->u_members[3].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[3].l_member = &patch->p_aux_offset;

    return list;
}



/*
 *  patchSectionSectionOffsetIntoSectionAddressGetDestinations(upm) --
 */
UPMList *patchSectionSectionOffsetIntoSectionAddressGetDestinations(
Patch *upm) {
    UPMList *list;
    PatchSectionToSectionAux *patch;

    patch = (PatchSectionToSectionAux *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_offset;

    return list;
}



/*
 *  AddPatchSectionSectionOffsetIntoSectionAddress
 *  (src_section,src_offset,aux_section,aux_offset,dest_section,dest_offset) --
 *
 *  Adds a UPM into the global UPM list that will patch the offset from
 *  aux_offset from aux_section to src_offset from src_section into the
 *  specified offset from dest_section. The offset is calculated as a PC
 *  relative displacement, i.e., (address(src_section) + src_offset -
 *  (address(aux_section) + aux_offset).
 *
 *  src_section: the section that would be the target of
 *  the PC relative instruction
 *  src_offset: the offset from src_section that would be the target
 *  of the PC relative instruction
 *  aux_section: the section that would be the origin of
 *  the PC relative instruction
 *  aux_offset: the offset from src_section that would be the origin
 *  of the PC relative instruction
 *  dest_section: the section that will be patched
 *  dest_offset: the offset from dest_section where the patch will occur
 *  returns: a pointer to the newly added UPM
 *  see also: PatchSectionToSectionAux
 *  see also: upm2s_list
 */
Patch *AddPatchSectionSectionOffsetIntoSectionAddress(
Section *src_section, ADDRESS src_offset,
Section *aux_section, ADDRESS aux_offset,
Section *dest_section, ADDRESS dest_offset) {
    PatchSectionToSectionAux *patch;
    Patch *upm;


    patch = zalloc(sizeof(PatchSectionToSectionAux));
    patch->p_src_section = src_section;
    patch->p_src_offset = src_offset;
    patch->p_aux_section = aux_section;
    patch->p_aux_offset = aux_offset;
    patch->p_dest_section = dest_section;
    patch->p_dest_offset = dest_offset;

    upm = UPMCreate(patch,
	&patchSectionSectionOffsetIntoSectionAddress,
	NULL,
	&patchSectionSectionOffsetIntoSectionAddressGetSources,
	&patchSectionSectionOffsetIntoSectionAddressGetDestinations);

    UPMTableInsertBySourceAddress(src_section, src_offset, upm);
    UPMTableInsertBySourceAddress(aux_section, aux_offset, upm);
    UPMTableInsertByDestAddress(dest_section, dest_offset, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchSectionAddressIntoSectionAddress(upm) --
 *
 *  Patches a section location address into a section location.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 */
bool patchSectionAddressIntoSectionAddress(Patch *upm) {
    PatchSectionToSection *patch;
    ADDRESS og_val, new_val;

    patch = (PatchSectionToSection *) upm->p_patch;

    CASSERT(patch->p_src_section != NULL, "Patching null section");
    CASSERT(patch->p_dest_section != NULL, "Patching into null section");

    og_val = loadLE(&patch->p_dest_section->s_newdata[patch->p_dest_offset]);
    new_val = patch->p_src_section->s_newaddress + patch->p_src_offset;

    if (og_val != new_val) {
	storeLE(&patch->p_dest_section->s_newdata
	    [patch->p_dest_offset], new_val);
	return true;
    }

    return false;
}



/*
 *  patchSectionAddressIntoSectionAddressGetSources(upm) --
 */
UPMList *patchSectionAddressIntoSectionAddressGetSources(Patch *upm) {
    UPMList *list;
    PatchSectionToSection *patch;

    patch = (PatchSectionToSection *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_offset;

    return list;
}



/*
 *  patchSectionAddressIntoSectionAddressGetDestinations(upm) --
 */
UPMList *patchSectionAddressIntoSectionAddressGetDestinations(Patch *upm) {
    UPMList *list;
    PatchSectionToSection *patch;

    patch = (PatchSectionToSection *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_offset;

    return list;
}



/*
 *  AddPatchSectionAddressIntoSectionAddress
 *  (src_section,src_offset,dest_section,dest_offset) --
 *
 *  Adds a UPM into the global UPM list that will patch the final
 *  address of src_offset from src_section into dest_offset from
 *  dest_section.
 *
 *  src_section: the section that will be examined
 *  src_offset: the offset from src_section that will be examined
 *  dest_section: the section that will be patched
 *  dest_offset: the offset from dest_section where the patch will occur
 *  returns: a pointer to the newly added UPM
 *  see also: PatchSectionToSection
 *  see also: upm2s_list
 */
Patch *AddPatchSectionAddressIntoSectionAddress(Section *src_section,
ADDRESS src_offset, Section *dest_section, ADDRESS dest_offset) {
    PatchSectionToSection *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchSectionToSection));
    patch->p_src_section = src_section;
    patch->p_src_offset = src_offset;
    patch->p_dest_section = dest_section;
    patch->p_dest_offset = dest_offset;

    upm = UPMCreate(patch,
	&patchSectionAddressIntoSectionAddress,
	NULL,
	&patchSectionAddressIntoSectionAddressGetSources,
	&patchSectionAddressIntoSectionAddressGetDestinations);

    UPMTableInsertBySourceAddress(src_section, src_offset, upm);
    UPMTableInsertByDestAddress(dest_section, dest_offset, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchSectionAddressIntoSymbol(upm) --
 *
 *  Patches a section location address into a symbol's value field.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchSectionAddressIntoSymbol
 */
bool patchSectionAddressIntoSymbol(Patch *upm) {
    PatchSectionToSymbol *patch;
    ADDRESS og_val, new_val;

    patch = (PatchSectionToSymbol *) upm->p_patch;

    CASSERT(patch->p_src_section != NULL, "Patching from a null section");
    CASSERT(patch->p_dest_sym != NULL,"Patching a null function into a symbol");

    og_val = patch->p_dest_sym->sy_value;

    new_val = patch->p_src_section->s_newaddress + patch->p_src_offset;

    if (og_val != new_val) {
	patch->p_dest_sym->sy_value = new_val;
	return true;
    }

    return false;
}



/*
 *  patchSectionAddressIntoSymbolGetSources(upm) --
 */
UPMList *patchSectionAddressIntoSymbolGetSources(Patch *upm) {
    UPMList *list;
    PatchSectionToSymbol *patch;

    patch = (PatchSectionToSymbol *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_offset;

    return list;
}



/*
 *  patchSectionAddressIntoSymbolGetDestinations(upm) --
 */
UPMList *patchSectionAddressIntoSymbolGetDestinations(Patch *upm) {
    UPMList *list;
    PatchSectionToSymbol *patch;

    patch = (PatchSectionToSymbol *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_SYMBOL_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_sym;

    return list;
}



/*
 *  AddPatchSectionAddressIntoSymbol(src_section,src_offset,dest_sym) --
 *
 *  Adds a UPM into the global UPM list that will patch the final
 *  address of src_offset from src_section into the value of dest_sym.
 *
 *  src_section: the section that will be examined
 *  src_offset: the offset from src_section that will be examined
 *  dest_sym: the symbol whose value will be updated
 *  returns: a pointer to the newly added UPM
 *  see also: PatchSectionToSymbol
 *  see also: upm2s_list
 */
Patch *AddPatchSectionAddressIntoSymbol(
Section *src_section, ADDRESS src_offset, Symbol *dest_sym) {
    Patch *upm;
    PatchSectionToSymbol *patch;

    patch = zalloc(sizeof(PatchSectionToSymbol));
    patch->p_src_section = src_section;
    patch->p_src_offset = src_offset;
    patch->p_dest_sym = dest_sym;


    upm = UPMCreate(patch,
	&patchSectionAddressIntoSymbol,
	NULL,
	&patchSectionAddressIntoSymbolGetSources,
	&patchSectionAddressIntoSymbolGetDestinations);

    UPMTableInsertBySourceAddress(src_section, src_offset, upm);
    UPMTableInsertByDest(dest_sym, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchInstrInstrOffsetIntoInstr(upm) --
 *
 *  Patches the offset between two instructions into an instruction.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchInstrInstrOffsetIntoInstr
 */
bool patchInstrInstrOffsetIntoInstr(Patch *upm) {
    PatchInstrToInstrAux *patch;
    ADDRESS og_val, new_val;

    patch = (PatchInstrToInstrAux *) upm->p_patch;

    CASSERT(patch->p_src_instr != NULL, "NULL src_instr");
    CASSERT(patch->p_dest_instr != NULL, "NULL dest_instr");
    CASSERT(patch->p_aux_instr != NULL, "NULL aux_instr");
    CASSERT(patch->p_dest_op != NULL, "NULL dest_op");

    og_val = patch->p_dest_op->a_displacement;
    new_val = patch->p_src_instr->i_final_address -
	patch->p_aux_instr->i_final_address - patch->p_aux_instr->i_length;

    if (og_val != new_val) {
	patch->p_dest_op->a_displacement = new_val;
	UpdateInstr(patch->p_dest_instr);
	return true;
    }

    return false;
}



/*
 *  patchInstrInstrOffsetToInstrClone(upm,dest_instr) --
 */
static void patchInstrInstrOffsetToInstrClone(
Patch *upm, Instruction *dest_instr) {
    PatchInstrToInstrAux *patch;
    Operand *op;

    patch = (PatchInstrToInstrAux *) upm->p_patch;

    if (patch->p_dest_op == patch->p_dest_instr->i_src) {
	op = dest_instr->i_src;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_src2) {
	op = dest_instr->i_src2;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_dest) {
	op = dest_instr->i_dest;
    }

    AddPatchInstrInstrOffsetIntoInstr(patch->p_src_instr,
	patch->p_aux_instr, dest_instr, op);
}



/*
 *  patchInstrInstrOffsetIntoInstrGetSources(upm) --
 */
UPMList *patchInstrInstrOffsetIntoInstrGetSources(Patch *upm) {
    UPMList *list;
    PatchInstrToInstrAux *patch;

    patch = (PatchInstrToInstrAux *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_instr;

    list->u_members[1].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_aux_instr;

    return list;
}



/*
 *  patchInstrInstrOffsetIntoInstrGetDestinations(upm) --
 */
UPMList *patchInstrInstrOffsetIntoInstrGetDestinations(Patch *upm) {
    UPMList *list;
    PatchInstrToInstrAux *patch;

    patch = (PatchInstrToInstrAux *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_instr;

    list->u_members[1].l_type = UPM_OPERAND_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_op;

    return list;
}



/*
 *  AddPatchInstrInstrOffsetIntoInstr(src_instr,aux_instr,dest_instr,dest_op) --
 *
 *  Adds a UPM into the global UPM list that will patch the offset from
 *  src_instr to aux_instr into the specified operand of the specified
 *  destination instruction. The offset is calculated as a PC relative
 *  displacement, i.e., address(src_instr) -- (length(aux_instr) +
 *  address(aux_instr)).
 *
 *  src_instr: the the instruction that would be the target of
 *  the PC relative instruction
 *  aux_instr: the instruction that would be the origin of the
 *  PC relative instruction
 *  dest_instr: the instruction whose operand will be updated
 *  dest_op: the operand whose displacement will be updated
 *  returns: a pointer to the newly added UPM
 *  see also: PatchInstrToInstrAux
 *  see also: upm2s_list
 */
Patch *AddPatchInstrInstrOffsetIntoInstr(Instruction *src_instr,
Instruction *aux_instr, Instruction *dest_instr, Operand *dest_op) {
    Patch *upm;
    PatchInstrToInstrAux *patch;

    patch = zalloc(sizeof(PatchInstrToInstrAux));
    patch->p_src_instr = src_instr;
    patch->p_aux_instr = aux_instr;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_op = dest_op;

    upm = UPMCreate(patch,
	&patchInstrInstrOffsetIntoInstr,
	&patchInstrInstrOffsetToInstrClone,
	&patchInstrInstrOffsetIntoInstrGetSources,
	&patchInstrInstrOffsetIntoInstrGetDestinations);


    if (aux_instr != dest_instr) {
	UPMTableInsertBySource(aux_instr, upm);
    }

    UPMTableInsertByDest(dest_op, upm);
    UPMTableInsertBySource(src_instr, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchInstrAddressIntoInstr(upm) --
 *
 *  Patching Functions
 *
 *  Patches the final address of an instruction into another instruction.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchInstrAddressIntoInstr
 */
bool patchInstrAddressIntoInstr(Patch *upm) {
    PatchInstrToInstr *patch;
    ADDRESS og_val, new_val;

    patch = (PatchInstrToInstr *) upm->p_patch;

    CASSERT(patch->p_src_instr != NULL, "Patching a NULL instruction.");
    CASSERT(patch->p_dest_instr != NULL, "Patching into a NULL instruction.");
    CASSERT(patch->p_dest_op != NULL, "Patching into a NULL operand.");
    CASSERT(UPMTableGetByDest(patch->p_dest_op) != NULL, "WHAAAA: DEST_OP");
    CASSERT(UPMTableGetByDest(patch->p_dest_instr)!=NULL, "WHAAAA: DEST_INSTR");
    CASSERT(UPMTableGetBySource(patch->p_src_instr)!=NULL, "WHAAAA: SRC_INSTR");

    og_val = patch->p_dest_op->a_displacement;
    new_val = patch->p_src_instr->i_final_address;

    if (og_val != new_val) {
	patch->p_dest_op->a_displacement = new_val;

	UpdateInstr(patch->p_dest_instr);

	return true;
    }

    return false;
}



/*
 *  patchInstrAddressIntoInstrClone(upm,dest_instr) --
 */
void patchInstrAddressIntoInstrClone(Patch *upm, Instruction *dest_instr) {
    PatchInstrToInstr *patch;
    Operand *op;

    patch = (PatchInstrToInstr *) upm->p_patch;

    if (patch->p_dest_op == patch->p_dest_instr->i_src) {
	op = dest_instr->i_src;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_src2) {
	op = dest_instr->i_src2;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_dest) {
	op = dest_instr->i_dest;
    }

    AddPatchInstrAddressIntoInstr(patch->p_src_instr, dest_instr, op);
}



/*
 *  patchInstrAddressIntoInstrGetSources(upm) --
 */
UPMList *patchInstrAddressIntoInstrGetSources(Patch *upm) {
    UPMList *list;
    PatchInstrToInstr *patch;

    patch = (PatchInstrToInstr *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_instr;

    return list;
}



/*
 *  patchInstrAddressIntoInstrGetDestinations(upm) --
 */
UPMList *patchInstrAddressIntoInstrGetDestinations(Patch *upm) {
    UPMList *list;
    PatchInstrToInstr *patch;

    patch = (PatchInstrToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_instr;

    list->u_members[1].l_type = UPM_OPERAND_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_op;

    return list;
}



/*
 *  AddPatchInstrAddressIntoInstr(src_instr,dest_instr,dest_op) --
 *
 *  UPM Creation Functions
 *
 *  Adds a UPM into the global UPM list that will patch the final
 *  address of src_instr into the specified operand of dest_instr.
 *
 *  src_instr: the instruction whose address will be examined
 *  dest_instr: the instruction whose operand will be updated
 *  dest_op: the operand whose displacement will be updated
 *  returns: a pointer to the newly added UPM
 *  see also: PatchInstrToInstr
 *  see also: upm2s_list
 */
Patch *AddPatchInstrAddressIntoInstr(
Instruction *src_instr, Instruction *dest_instr, Operand *dest_op) {
    PatchInstrToInstr *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchInstrToInstr));
    patch->p_src_instr = src_instr;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_op = dest_op;

    upm = UPMCreate(patch,
	&patchInstrAddressIntoInstr,
	&patchInstrAddressIntoInstrClone,
	&patchInstrAddressIntoInstrGetSources,
	&patchInstrAddressIntoInstrGetDestinations);

    UPMTableInsertBySource(src_instr, upm);
    UPMTableInsertByDest(dest_op, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchFunctionAddressIntoSymbol(upm) --
 *
 *  Patches a function address into a symbol's value field.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchFunctionAddressIntoSymbol
 */
bool patchFunctionAddressIntoSymbol(Patch *upm) {
    PatchFunctionToSymbol *patch;
    ADDRESS og_val, new_val;

    patch = (PatchFunctionToSymbol *) upm->p_patch;

    CASSERT(patch->p_dest_sym != NULL, "Patching into a null symbol!");
    CASSERT(patch->p_src_function != NULL, "Patching a null function");

    if ((patch->p_src_function->f_first == NULL) ||
	(FUNCTION_FIRST_INSTR(patch->p_src_function) == NULL)) {
	// when unreachable code is removed
	WARNING("Cannot find first instruction for %s",
	    patch->p_src_function->f_name);
	return false;
    }

    og_val = patch->p_dest_sym->sy_value;

// if (output_binary_type == BINARY_RELOCATABLE_TYPE) {
//  new_val = FUNCTION_START(((patch)->p_src_function))
//      - ((SECTION_REF_CODE)->s_newaddress);
// }
// else {
    new_val = FUNCTION_START(patch->p_src_function);
// }

    if (og_val != new_val) {
	patch->p_dest_sym->sy_value = new_val;
	patch->p_dest_sym->sy_size = FUNCTION_END(patch->p_src_function) -
	    FUNCTION_START(patch->p_src_function) + 1;
	return true;
    }

    return false;
}



/*
 *  patchFunctionAddressIntoSymbolGetSources(upm) --
 */
UPMList *patchFunctionAddressIntoSymbolGetSources(Patch *upm) {
    UPMList *list;
    PatchFunctionToSymbol *patch;

    patch = (PatchFunctionToSymbol *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_FUNCTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_function;

    return list;
}



/*
 *  patchFunctionAddressIntoSymbolGetDestinations(upm) --
 */
UPMList *patchFunctionAddressIntoSymbolGetDestinations(Patch *upm) {
    UPMList *list;
    PatchFunctionToSymbol *patch;

    patch = (PatchFunctionToSymbol *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_SYMBOL_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_sym;

    return list;
}



/*
 *  AddPatchFunctionAddressIntoSymbol(src_fun,dest_sym) --
 *
 *  Adds a UPM into the global UPM list that will patch the final
 *  address of src_fun into the value of dest_sym.
 *
 *  src_fun: the function whose address will be examined
 *  dest_sym: the symbol whose value will be updated
 *  returns: a pointer to the newly added UPM
 *  see also: PatchFunctionToSymbol
 *  see also: upm2s_list
 */
Patch *AddPatchFunctionAddressIntoSymbol(Function *src_fun, Symbol *dest_sym) {
    Patch *upm;
    PatchFunctionToSymbol *patch;

    patch = zalloc(sizeof(PatchFunctionToSymbol));
    patch->p_src_function = src_fun;
    patch->p_dest_sym = dest_sym;

    upm = UPMCreate(patch,
	&patchFunctionAddressIntoSymbol,
	NULL,
	&patchFunctionAddressIntoSymbolGetSources,
	&patchFunctionAddressIntoSymbolGetDestinations);

    UPMTableInsertBySource(src_fun, upm);
    UPMTableInsertByDest(dest_sym, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchInstrAddressIntoSymbol(upm) --
 *
 *  Patches an instruction address into a symbol's value field.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchInstrAddressIntoSymbol
 */
bool patchInstrAddressIntoSymbol(Patch *upm) {
    PatchInstrToSymbol *patch;
    ADDRESS og_val, new_val;

    patch = (PatchInstrToSymbol *) upm->p_patch;

    CASSERT(patch->p_src_instr != NULL,
	"Patching a null instruction into a symbol");
    CASSERT(patch->p_dest_sym != NULL,
	"Patching a null function into a symbol");

    og_val = patch->p_dest_sym->sy_value;

// if (output_binary_type == BINARY_RELOCATABLE_TYPE) {
//  new_val = ((((patch)->p_src_instr))->i_final_address)
//      - ((SECTION_REF_CODE)->s_newaddress);
// }
// else {
    new_val = patch->p_src_instr->i_final_address;
// }

    if (og_val != new_val) {
	patch->p_dest_sym->sy_value = new_val;

	return true;
    }

    return false;
}



/*
 *  patchInstrAddressIntoSymbolGetSources(upm) --
 */
UPMList *patchInstrAddressIntoSymbolGetSources(Patch *upm) {
    UPMList *list;
    PatchInstrToSymbol *patch;

    patch = (PatchInstrToSymbol *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_instr;

    return list;
}



/*
 *  patchInstrAddressIntoSymbolGetDestinations(upm) --
 */
UPMList *patchInstrAddressIntoSymbolGetDestinations(Patch *upm) {
    UPMList *list;
    PatchInstrToSymbol *patch;

    patch = (PatchInstrToSymbol *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_SYMBOL_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_sym;

    return list;
}



/*
 *  AddPatchInstrAddressIntoSymbol(src_instr,dest_sym) --
 *
 *  Adds a UPM into the global UPM list that will patch the final
 *  address of src_instr into the value of dest_sym.
 *
 *  src_instr: the instruction whose address will be examined
 *  dest_sym: the symbol whose value will be updated
 *  returns: a pointer to the newly added UPM
 *  see also: PatchInstrToSymbol
 *  see also: upm2s_list
 */
Patch *AddPatchInstrAddressIntoSymbol(
Instruction *src_instr, Symbol *dest_sym) {
    Patch *upm;
    PatchInstrToSymbol *patch;

    patch = zalloc(sizeof(PatchInstrToSymbol));
    patch->p_src_instr = src_instr;
    patch->p_dest_sym = dest_sym;

    upm = UPMCreate(patch,
	&patchInstrAddressIntoSymbol,
	NULL,
	&patchInstrAddressIntoSymbolGetSources,
	&patchInstrAddressIntoSymbolGetDestinations);

    UPMTableInsertBySource(src_instr, upm);
    UPMTableInsertByDest(dest_sym, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchSectionEndIntoInstr(upm) --
 *
 *  Patches the final address of some section end into an instruction.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchSectionEndIntoInstr
 */
bool patchSectionEndIntoInstr(Patch *upm) {
    PatchSectionToInstr *patch;
    ADDRESS og_val, new_val;

    patch = (PatchSectionToInstr *) upm->p_patch;

    CASSERT(patch->p_dest_instr != NULL, "Patching into a NULL instruction");
    CASSERT(patch->p_src_section != NULL, "Patching into a NULL section");
    CASSERT(patch->p_dest_op != NULL, "Patching into a NULL operand");

    og_val = patch->p_dest_op->a_displacement;

    //Modified by somu to fix _end _edata
    new_val = patch->p_src_section->s_newaddress +
	patch->p_src_section->s_newdatasize +
	patch->p_src_offset;

    if (og_val != new_val) {
	patch->p_dest_op->a_displacement = new_val;
	UpdateInstr(patch->p_dest_instr);
	return true;
    }

    return false;
}



/*
 *  patchSectionEndIntoInstrClone(upm,dest_instr) --
 */
static void patchSectionEndIntoInstrClone(Patch *upm, Instruction *dest_instr) {
    PatchSectionToInstr *patch;
    Operand *op;

    patch = (PatchSectionToInstr *) upm->p_patch;

    if (patch->p_dest_op == patch->p_dest_instr->i_src) {
	op = dest_instr->i_src;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_src2) {
	op = dest_instr->i_src2;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_dest) {
	op = dest_instr->i_dest;
    }

    AddPatchSectionEndIntoInstr(patch->p_src_section, dest_instr,
        op, patch->p_src_offset);
}



/*
 *  patchSectionEndIntoInstrGetSources(upm) --
 */
UPMList *patchSectionEndIntoInstrGetSources(Patch *upm) {
    PatchSectionToInstr *patch;
    UPMList *list;

    patch = (PatchSectionToInstr *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_section;

    return list;
}



/*
 *  patchSectionEndIntoInstrGetDestinations(upm) --
 */
UPMList *patchSectionEndIntoInstrGetDestinations(Patch *upm) {
    PatchSectionToInstr *patch;
    UPMList *list;

    patch = (PatchSectionToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_instr;

    list->u_members[1].l_type = UPM_OPERAND_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_op;

    return list;
}



/*
 *  AddPatchSectionEndIntoInstr(src_section,dest_instr,dest_op) --
 *
 *  Adds a UPM into the global UPM list that will patch the final
 *  address of the end of src_section into the value of dest_op in
 *  dest_instr.
 *
 *  src_section: the section that will be examined
 *  dest_instr: the instruction that will be patched
 *  dest_op: the operand that will be updated
 *  returns: a pointer to the newly added UPM
 *  see also: PatchSectionToInstr
 *  see also: upm2s_list
 */
Patch *AddPatchSectionEndIntoInstr(
Section *src_section, Instruction *dest_instr, Operand *dest_op, int addendum) {
    PatchSectionToInstr *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchSectionToInstr));
    patch->p_src_section = src_section;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_op = dest_op;
    //Added by Somu 
    patch->p_src_offset = addendum;

    upm = UPMCreate(patch,
	&patchSectionEndIntoInstr,
	&patchSectionEndIntoInstrClone,
	&patchSectionEndIntoInstrGetSources,
	&patchSectionEndIntoInstrGetDestinations);

    UPMTableInsertByDest(dest_op, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchSectionEndIntoSectionAddress(upm) --
 *
 *  Patches the end address of a section into a section location.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value it replaces
 *  see also: AddPatchSectionEndIntoInstr
 */
bool patchSectionEndIntoSectionAddress(Patch *upm) {
    PatchSectionToSection *patch;
    ADDRESS og_val, new_val;

    patch = (PatchSectionToSection *) upm->p_patch;

    CASSERT(patch->p_src_section != NULL, "Patching a NULL section");
    CASSERT(patch->p_dest_section != NULL, "Patching into a NULL section");

    og_val = loadLE(&patch->p_dest_section->s_newdata[patch->p_dest_offset]);
    new_val = patch->p_src_section->s_newaddress +
	patch->p_src_section->s_newdatasize;

    if (og_val != new_val) {
	storeLE(&patch->p_dest_section->s_newdata
	    [patch->p_dest_offset], new_val);
	return true;
    }

    return false;
}



/*
 *  patchSectionEndIntoSectionAddressGetSources(upm) --
 */
UPMList *patchSectionEndIntoSectionAddressGetSources(Patch *upm) {
    UPMList *list;
    PatchSectionToSection *patch;

    patch = (PatchSectionToSection *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_section;

    return list;
}



/*
 *  patchSectionEndIntoSectionAddressGetDestinations(upm) --
 */
UPMList *patchSectionEndIntoSectionAddressGetDestinations(Patch *upm) {
    UPMList *list;
    PatchSectionToSection *patch;

    patch = (PatchSectionToSection *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_offset;

    return list;
}



/*
 *  AddPatchSectionEndIntoSectionAddress(src_section,dest_section,dest_offset)--
 *
 *  Adds a UPM into the global UPM list that will patch the final
 *  address of the end of src_section into dest_offset from
 *  dest_section.
 *
 *  src_section: the section that will be examined
 *  dest_section: the section that will be patched
 *  dest_offset: the offset from dest_section where the patch will occur
 *  returns: a pointer to the newly added UPM
 *  see also: PatchSectionToSection
 *  see also: upm2s_list
 */
Patch *AddPatchSectionEndIntoSectionAddress(
Section *src_section, Section *dest_section, ADDRESS dest_offset) {
    PatchSectionToSection *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchSectionToSection));
    patch->p_src_section = src_section;
    patch->p_dest_section = dest_section;
    patch->p_dest_offset = dest_offset;

    upm = UPMCreate(patch,
	&patchSectionEndIntoSectionAddress,
	NULL,
	&patchSectionEndIntoSectionAddressGetSources,
	&patchSectionEndIntoSectionAddressGetDestinations);

    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchSectionEndIntoSymbol(upm) --
 *
 *  Patches the end address of a section into a symbol value.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value replaces
 *  see also: AddPatchSectionEndIntoSymbol
 */
bool patchSectionEndIntoSymbol(Patch *upm) {
    PatchSectionToSymbol *patch;
    ADDRESS og_val, new_val;

    patch = (PatchSectionToSymbol *) upm->p_patch;

    CASSERT(patch->p_src_section != NULL, "Patching from a null section");
    CASSERT(patch->p_dest_sym != NULL,"Patching a null function into a symbol");

    og_val = patch->p_dest_sym->sy_value;

    new_val = patch->p_src_section->s_newaddress +
	patch->p_src_section->s_newdatasize;

    if (og_val != new_val) {
	patch->p_dest_sym->sy_value = new_val;
	return true;
    }

    return false;
}



/*
 *  patchSectionEndIntoSymbolGetSources(upm) --
 */
UPMList *patchSectionEndIntoSymbolGetSources(Patch *upm) {
    UPMList *list;
    PatchSectionToSymbol *patch;

    patch = (PatchSectionToSymbol *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_section;

    return list;
}



/*
 *  patchSectionEndIntoSymbolGetDestinations(upm) --
 */
UPMList *patchSectionEndIntoSymbolGetDestinations(Patch *upm) {
    UPMList *list;
    PatchSectionToSymbol *patch;

    patch = (PatchSectionToSymbol *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_SYMBOL_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_sym;

    return list;
}



/*
 *  AddPatchSectionEndIntoSymbol(src_section,dest_sym) --
 *
 *  Adds a UPM into the global UPM list that will patch the final
 *  address of the end of src_section into the value of dest_sym.
 *
 *  src_section: the section that will be examined
 *  dest_sym: the symbol whose value will be updated
 *  returns: a pointer to the newly added UPM
 *  see also: PatchSectionToSymbol
 *  see also: upm2s_list
 */
Patch *AddPatchSectionEndIntoSymbol(Section *src_section, Symbol *dest_sym) {
    Patch *upm;
    PatchSectionToSymbol *patch;

    patch = zalloc(sizeof(PatchSectionToSymbol));
    patch->p_src_section = src_section;
    patch->p_dest_sym = dest_sym;

    upm = UPMCreate(patch,
	&patchSectionEndIntoSymbol,
	NULL,
	&patchSectionEndIntoSymbolGetSources,
	&patchSectionEndIntoSymbolGetDestinations);

    UPMTableInsertByDest(dest_sym, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchGOTOffsetIntoInstr(upm) --
 *
 *  Patches the offset from the beginning of an instruction to the global
 *  offset table into an instruction.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value replaces
 *  see also: AddPatchGOTOffsetIntoInstr
 */
bool patchGOTOffsetIntoInstr(Patch *upm) {
    PatchInstrToInstrAux *patch;
    ADDRESS old_value, new_value;

    patch = (PatchInstrToInstrAux *) upm->p_patch;

    CASSERT(SECTION_REF_GOT != NULL, "SECTION_REF_GOT == NULL");
    CASSERT(patch->p_src_instr != NULL, "Source instruction == NULL");
    CASSERT(patch->p_dest_instr != NULL, "Destination instruction == NULL");
    CASSERT(patch->p_dest_op != NULL, "Destination operand == NULL");

    old_value = patch->p_dest_op->a_displacement;
    new_value = SECTION_REF_GOT->s_newaddress -
	patch->p_src_instr->i_final_address - patch->p_src_instr->i_length;

    if (old_value != new_value) {
	patch->p_dest_op->a_displacement = new_value;
	return true;
    }

    return false;
}



/*
 *  patchGOTOffsetIntoInstrClone(upm,dest_instr) --
 */
void patchGOTOffsetIntoInstrClone(Patch *upm, Instruction *dest_instr) {
    PatchInstrToInstrAux *patch;
    Operand *op;

    patch = (PatchInstrToInstrAux *) upm->p_patch;

    if (patch->p_dest_op == patch->p_dest_instr->i_src) {
	op = dest_instr->i_src;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_src2) {
	op = dest_instr->i_src2;
    } else if (patch->p_dest_op == patch->p_dest_instr->i_dest) {
	op = dest_instr->i_dest;
    }

    AddPatchGOTOffsetIntoInstr(patch->p_src_instr, dest_instr, op);
}



/*
 *  patchGOTOffsetIntoInstrGetSources(upm) --
 */
UPMList *patchGOTOffsetIntoInstrGetSources(Patch *upm) {
    PatchInstrToInstrAux *patch;
    UPMList *list;

    patch = (PatchInstrToInstrAux *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &SECTION_REF_GOT;

    list->u_members[1].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_instr;

    return list;
}



/*
 *  patchGOTOffsetIntoInstrGetDestinations(upm) --
 */
UPMList *patchGOTOffsetIntoInstrGetDestinations(Patch *upm) {
    PatchInstrToInstrAux *patch;
    UPMList *list;

    patch = (PatchInstrToInstrAux *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_instr;

    list->u_members[1].l_type = UPM_OPERAND_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_op;

    return list;
}



/*
 *  AddPatchGOTOffsetIntoInstr(pic_call,dest_instr,dest_op) --
 *
 *  Adds a UPM into the global UPM list that will patch the offset from
 *  the final address of pic_call + len(pic_call) to the global offset
 *  table into the specified destination operand of the specified
 *  destination instruction.
 *
 *  pic_call: the call instruction for the PIC
 *  dest_instr: the instruction whose operand will be patched
 *  dest_op: the operand that will be updated
 *  returns: a pointer to the newly added UPM
 *  see also: PatchInstrToInstrAux
 *  see also: upm2s_list
 */
Patch *AddPatchGOTOffsetIntoInstr(
Instruction *pic_call, Instruction *dest_instr, Operand *dest_op) {
    PatchInstrToInstrAux *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchInstrToInstrAux));
    patch->p_src_instr = pic_call;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_op = dest_op;

    upm = UPMCreate(patch,
	&patchGOTOffsetIntoInstr,
	&patchGOTOffsetIntoInstrClone,
	&patchGOTOffsetIntoInstrGetSources,
	&patchGOTOffsetIntoInstrGetDestinations);

    UPMTableInsertBySourceAddress(SECTION_REF_GOT, 0, upm);
    UPMTableInsertBySource(pic_call, upm);
    UPMTableInsertByDest(dest_op, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchInstrEndAddressIntoSectionAddress(upm) --
 *
 *  Patches the address of the end of an instruction into
 *  a location in some section.
 *
 *  upm: the UPM that will be used to complete the patch
 *  returns: true if the patch value is different from the value replaces
 *  see also: AddPatchInstrEndAddressIntoSectionAddress
 */
bool patchInstrEndAddressIntoSectionAddress(Patch *upm) {
    PatchInstrToSection *patch;
    ADDRESS old_value, new_value;

    patch = upm->p_patch;

    CASSERT(patch->p_src_instr != NULL, "src_instr == NULL");
    CASSERT(patch->p_dest_section != NULL, "dest_section == NULL");

    old_value = loadLE(&patch->p_dest_section->s_newdata[patch->p_dest_offset]);
    new_value =
	patch->p_src_instr->i_final_address + patch->p_src_instr->i_length;

    if (old_value != new_value) {
	storeLE(&patch->p_dest_section->s_newdata
	    [patch->p_dest_offset], new_value);
	return true;
    }

    return false;
}



/*
 *  patchInstrEndAddressIntoSectionAddressGetSources(upm) --
 */
UPMList *patchInstrEndAddressIntoSectionAddressGetSources(Patch *upm) {
    UPMList *list;
    PatchInstrToSection *patch;

    patch = (PatchInstrToSection *) upm->p_patch;
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_instr;

    return list;
}



/*
 *  patchInstrEndAddressIntoSectionAddressGetDestinations(upm) --
 */
UPMList *patchInstrEndAddressIntoSectionAddressGetDestinations(Patch *upm) {
    UPMList *list;
    PatchInstrToSection *patch;

    patch = (PatchInstrToSection *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_section;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_offset;

    return list;
}



/*
 *  AddPatchInstrEndAddressIntoSectionAddress(
 *  src_instr,dest_section,dest_offset) --
 *
 *  Adds a UPM into the global UPM list that will patch the final address
 *  of the end of the specified instruction into the specified offset from
 *  the specified section.
 *
 *  src_instr: the instruction whose ending address will be examined
 *  dest_section: the section that will be updated
 *  dest_offset: the offset into section where the update will take place
 *  returns: a pointer to the newly added UPM
 *  see also: PatchInstrToSection
 *  see also: upm2s_list
 */
Patch *AddPatchInstrEndAddressIntoSectionAddress(
Instruction *src_instr, Section *dest_section, ADDRESS dest_offset) {
    PatchInstrToSection *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchInstrToSection));
    patch->p_src_instr = src_instr;
    patch->p_dest_section = dest_section;
    patch->p_dest_offset = dest_offset;

    upm = UPMCreate(patch,
	&patchInstrEndAddressIntoSectionAddress,
	NULL,
	&patchInstrEndAddressIntoSectionAddressGetSources,
	&patchInstrEndAddressIntoSectionAddressGetDestinations);

    UPMTableInsertByDestAddress(dest_section, dest_offset, upm);
    UPMTableInsertBySource(src_instr, upm);
    ll_add(upm2s_list, upm);

    return upm;
}


/*
 *  patchDataInstrToDataInstr(upm) -- patch from one data instr to another.
 */
bool patchDataInstrToDataInstr(Patch *upm) {
    PatchDataInstrToDataInstr *patch;
    ADDRESS old_value, new_value;

    patch = upm->p_patch;

    if (patch->p_src_instr == NULL) {
	WARNING("src_instr == NULL");
	return false;
    }

    if (patch->p_dest_instr == NULL) {
	WARNING("dest_instr == NULL");
	return false;
    }

    memcpy(&old_value,
	&(patch->p_dest_instr->i_literal_data)[patch->p_dest_patch_offset],
	sizeof(old_value));
    new_value = patch->p_src_instr->i_final_address + patch->p_src_patch_offset;

    if (old_value != new_value) {
	memcpy(&patch->p_dest_instr->i_literal_data[patch->p_dest_patch_offset],
	    &new_value, sizeof(new_value));
	return true;
    }
    return false;
}



/*
 *  patchInstrDataInstrToDataInstrGetSources(upm) --
 */
UPMList *patchInstrDataInstrToDataInstrGetSources(Patch *upm) {
    UPMList *list;
    PatchDataInstrToDataInstr *patch;

    patch = (PatchDataInstrToDataInstr *) (upm->p_patch);
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &(patch->p_src_instr);

    return list;
}



/*
 *  patchInstrDataInstrToDataInstrGetDestinations(upm) --
 */
UPMList *patchInstrDataInstrToDataInstrGetDestinations(Patch *upm) {
    UPMList *list;
    PatchDataInstrToDataInstr *patch;

    patch = (PatchDataInstrToDataInstr *) (upm->p_patch);
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &(patch->p_dest_instr);

    return list;
}



/*
 *  AddPatchDataInstrToDataInstr(src, src_offset, dest, dest_offset) --
 */
Patch *AddPatchDataInstrToDataInstr(Instruction *src_instr, int src_offset,
    Instruction *dest_instr, int dest_offset) {

    PatchDataInstrToDataInstr *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchDataInstrToDataInstr));
    patch->p_src_instr = src_instr;
    patch->p_src_patch_offset = src_offset;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_patch_offset = dest_offset;

    upm = UPMCreate(patch,
	&patchDataInstrToDataInstr,
	NULL,
	&patchInstrDataInstrToDataInstrGetSources,
	&patchInstrDataInstrToDataInstrGetDestinations);

    //SOMU added this    
    UPMTableInsertByDest(dest_instr, upm);
    UPMTableInsertBySource(src_instr, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchDataInstrToSectionAddress(upm) -- patch data instr with section addr
 */
bool patchDataInstrToSectionAddress(Patch *upm) {
    PatchDataInstrToSectionAddress *patch;
    ADDRESS og_val, new_val;

    patch = (PatchDataInstrToSectionAddress *) (upm->p_patch);
    if (patch->p_dest_section == NULL) {
	WARNING("dest_section == NULL");
	return false;
    }
    if (patch->p_src_instr == NULL) {
	WARNING("src_instr == NULL");
	return false;
    }
    og_val =
	loadLE(&(patch->p_dest_section->s_newdata)[patch->p_dest_patch_offset]);
    new_val = patch->p_src_instr->i_final_address + patch->p_src_patch_offset;
    if (og_val != new_val) {
	storeLE(&(patch->p_dest_section->s_newdata)[patch->p_dest_patch_offset],
	    new_val);
	return true;
    }
    return false;
}



/*
 *  patchInstrDataInstrToSectionAddressGetSources(upm) --
 */
UPMList *patchInstrDataInstrToSectionAddressGetSources(Patch *upm) {
    UPMList *list;
    PatchDataInstrToSectionAddress *patch;

    patch = (PatchDataInstrToSectionAddress *) (upm->p_patch);
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &(patch->p_src_instr);

    // list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    // list->u_members[1].l_member = patch->p_src_patch_offset;

    return list;
}



/*
 *  patchInstrDataInstrToSectionAddressGetDestinations(upm) --
 */
UPMList *patchInstrDataInstrToSectionAddressGetDestinations(Patch *upm) {
    UPMList *list;
    PatchDataInstrToSectionAddress *patch;

    patch = (PatchDataInstrToSectionAddress *) (upm->p_patch);
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &(patch->p_dest_section);

    return list;
}



/*
 *  AddPatchDataInstrToSectionAddress(src, src_offset, dest, dest_offset) --
 */
Patch *AddPatchDataInstrToSectionAddress(Instruction *src_instr, int src_offset,
    Section *dest_section, int dest_offset) {

    PatchDataInstrToSectionAddress *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchDataInstrToSectionAddress));
    patch->p_src_instr = src_instr;
    patch->p_src_patch_offset = src_offset;
    patch->p_dest_section = dest_section;
    patch->p_dest_patch_offset = dest_offset;

    upm = UPMCreate(patch,
	&patchDataInstrToSectionAddress,
	NULL,
	&patchInstrDataInstrToSectionAddressGetSources,
	&patchInstrDataInstrToSectionAddressGetDestinations);

    //SOMU added this
    UPMTableInsertByDestAddress(dest_section, dest_offset, upm);
    UPMTableInsertBySource(src_instr, upm);
    ll_add(upm2s_list, upm);

    return upm;
}



/*
 *  patchDataInstrOffsetIntoSectionAddress(upm) --
 *  patch section address with pointer to data instruction
 */
bool patchDataInstrOffsetIntoSectionAddress(Patch *upm) {
    PatchSectionInstrToSection *patch;
    ADDRESS og_val, new_val;

    patch = (PatchSectionInstrToSection *) (upm->p_patch);

    if (patch->p_src_section == NULL) {
	WARNING("src_section == NULL");
	return false;
    }
    if (patch->p_dest_section == NULL) {
	WARNING("dest_section == NULL");
	return false;
    }
    if (patch->p_src_instr == NULL) {
	WARNING("src_instr == NULL");
	return false;
    }

    og_val =
	loadLE(&(patch->p_dest_section->s_newdata)[(patch->p_dest_offset)]);
    new_val = (patch->p_src_instr->i_final_address + patch->p_src_offset) -
	(patch->p_dest_section->s_newaddress + patch->p_dest_offset);

    if (og_val != new_val) {
	storeLE(&(patch->p_dest_section->s_newdata)[(patch->p_dest_offset)],
	    new_val);
	return true;
    }
    return false;
}



/*
 *  patchDataInstrOffsetIntoSectionAddressGetSources(upm) --
 */
UPMList *patchDataInstrOffsetIntoSectionAddressGetSources(Patch *upm) {
    UPMList *list;
    PatchSectionInstrToSection *patch;

    patch = (PatchSectionInstrToSection *) (upm->p_patch);
    list = UPMMemberListCreate(3);
    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &(patch->p_src_instr);
    list->u_members[1].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[1].l_member = &(patch->p_src_section);
    list->u_members[2].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[2].l_member = &(patch->p_src_offset);

    return list;
}



/*
 *  patchDataInstrOffsetIntoSectionAddressGetDestinations(upm) --
 */
UPMList *patchDataInstrOffsetIntoSectionAddressGetDestinations(Patch *upm) {
    UPMList *list;
    PatchSectionInstrToSection *patch;

    patch = (PatchSectionInstrToSection *) (upm->p_patch);
    list = UPMMemberListCreate(2);
    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &(patch->p_dest_section);
    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &(patch->p_dest_offset);

    return list;
}


/*
 *  AddPatchDestInstrOffsetIntoSectionAddress(si, so, ds, do) --
 */
Patch *AddPatchDestInstrOffsetIntoSectionAddress(Instruction *src_instr,
    Section *src_section,
    ADDRESS src_offset, Section *dest_section, ADDRESS dest_offset) {
    PatchSectionInstrToSection *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchSectionInstrToSection));
    patch->p_src_instr = src_instr;
    patch->p_src_section = src_section;
    patch->p_src_offset = src_offset;
    patch->p_dest_section = dest_section;
    patch->p_dest_offset = dest_offset;

    upm = UPMCreate(patch,
	&patchDataInstrOffsetIntoSectionAddress,
	NULL,
	&patchDataInstrOffsetIntoSectionAddressGetSources,
	&patchDataInstrOffsetIntoSectionAddressGetDestinations);

    ll_add(upm2s_list, upm);

    return upm;
}


/*
 *  patchDataSectionOffsetIntoDataInstr(upm) -- patch data inst with sect offset
 */
bool patchDataSectionOffsetIntoDataInstr(Patch *upm) {
    PatchDataSectionOffsetIntoDataInstr *patch;
    ADDRESS og_val, new_val;

    patch = (PatchDataSectionOffsetIntoDataInstr *) (upm->p_patch);

    if (patch->p_src_section == NULL) {
	WARNING("src_section == NULL");
	return false;
    }
    if (patch->p_dest_instr == NULL) {
	WARNING("dest_instr== NULL");
	return false;
    }

    memcpy(&og_val,
        &patch->p_dest_instr->i_literal_data[patch->p_dest_offset],
        sizeof(og_val));
    
    // Added by SOMU
    new_val = patch->p_src_section->s_newaddress + patch->p_src_offset;

    if (og_val != new_val) {
	memcpy(&(patch->p_dest_instr->i_literal_data)[patch->p_dest_offset],
	    &new_val, sizeof(new_val));
	return true;
    }
    return false;
}



/*
 *  AddPatchDataSectionOffsetIntoDataInstrGetSources(upm) --
 */
UPMList *AddPatchDataSectionOffsetIntoDataInstrGetSources(Patch *upm) {
    UPMList *list;
    PatchDataSectionOffsetIntoDataInstr *patch;

    patch = (PatchDataSectionOffsetIntoDataInstr *) (upm->p_patch);
    list = UPMMemberListCreate(2);
    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &(patch->p_src_section);
    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &(patch->p_src_offset);

    return list;
}



/*
 *  AddPatchDataSectionOffsetIntoDataInstrGetDestinations(upm) --
 */
UPMList *AddPatchDataSectionOffsetIntoDataInstrGetDestinations(Patch *upm) {
    UPMList *list;
    PatchDataSectionOffsetIntoDataInstr *patch;

    patch = (PatchDataSectionOffsetIntoDataInstr *) (upm->p_patch);
    list = UPMMemberListCreate(2);
    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &(patch->p_dest_instr);
    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &(patch->p_dest_offset);

    return list;
}



/*
 *  AddPatchDataSectionOffsetIntoDataInstr(ss, so, di, do) --
 */
Patch *AddPatchDataSectionOffsetIntoDataInstr(Section *src_section,
    int src_offset, Instruction *dest_instr, int dest_offset) {
    Patch *upm;
    PatchDataSectionOffsetIntoDataInstr *patch;

    patch = zalloc(sizeof(PatchDataSectionOffsetIntoDataInstr));
    patch->p_src_section = src_section;
    patch->p_src_offset = src_offset;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_offset = dest_offset;

    upm = UPMCreate(patch,
	&patchDataSectionOffsetIntoDataInstr,
	NULL,
	&AddPatchDataSectionOffsetIntoDataInstrGetSources,
	&AddPatchDataSectionOffsetIntoDataInstrGetDestinations);

    UPMTableInsertByDest(dest_instr, upm);
    UPMTableInsertBySourceAddress(src_section, src_offset, upm);    
    ll_add(upm2s_list, upm);

    return upm;
}


/*
 *  patchInstrOffsetIntoInstr(upm) --
 *
 *  Patches offset between an instruction and the middle
 *  of another.
 */
bool patchInstrOffsetIntoInstr(Patch *upm) {
    PatchInstrOffsetToInstr *patch; 
    int og_val, new_val;

    patch = (PatchInstrOffsetToInstr *) upm->p_patch;

    CASSERT(patch->p_src_instr != NULL, "NULL src_instr");
    CASSERT(patch->p_dest_instr != NULL, "NULL dest_instr");
    CASSERT(patch->p_dest_op != NULL, "NULL dest_op");

    og_val  = patch->p_dest_op->a_displacement;
    //og_val  = patch->p_dest_instr->i_src->a_displacement;
    new_val = patch->p_src_instr->i_final_address + patch->p_src_offset
    		- patch->p_dest_instr->i_final_address 
	      	- patch->p_dest_instr->i_length;
			    
    if (og_val != new_val) {
	patch->p_dest_op->a_displacement = new_val;
	//patch->p_dest_instr->i_src->a_displacement = new_val;
	UpdateInstr(patch->p_dest_instr);
	return true;
    }
    return false;
}



/*
 *  patchInstrInstrOffsetToInstrClone(upm,dest_instr) --
 */
static void patchInstrOffsetToInstrClone(
Patch *upm, Instruction *dest_instr) {
	MALFUNCTION("not implemented !!! \n");	
}



/*
 *  patchInstrOffsetIntoInstrGetSources(upm) --
 */
UPMList *patchInstrOffsetIntoInstrGetSources(Patch *upm) {
    UPMList *list;
    PatchInstrOffsetToInstr *patch; 

    patch = (PatchInstrOffsetToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_instr;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_offset;

    return list;
}



/*
 *  patchInstrOffsetIntoInstrGetDestinations(upm) --
 */
UPMList *patchInstrOffsetIntoInstrGetDestinations(Patch *upm) {
    UPMList *list;
    PatchInstrOffsetToInstr *patch; 

    patch = (PatchInstrOffsetToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_instr;

    list->u_members[1].l_type = UPM_OPERAND_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_op;

    return list;
}



/*
 *  AddPatchInstrOffsetIntoInstr(src_instr,src_offset,dest_instr,dest_op) --
 *
 *  Patches into dest_instr (dest_op), the relative distance to 
 *  src_instr+src_offset. This is useful for example, to patch relative
 *  calls without relocation information into the middle of data instructions.
 */
Patch *AddPatchInstrOffsetToInstr(Instruction *src_instr,
int src_offset , Instruction *dest_instr, Operand *dest_op) {
    Patch *upm;
    PatchInstrOffsetToInstr *patch; 

    patch = zalloc(sizeof(PatchInstrOffsetToInstr));
    patch->p_src_instr  = src_instr;
    patch->p_src_offset = src_offset;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_op    = dest_op;

    upm = UPMCreate(patch,
	&patchInstrOffsetIntoInstr,
	&patchInstrOffsetToInstrClone,
	&patchInstrOffsetIntoInstrGetSources,
	&patchInstrOffsetIntoInstrGetDestinations);

    UPMTableInsertByDest(dest_op, upm);
    UPMTableInsertBySource(src_instr, upm);
    ll_add(upm2s_list, upm);

    return upm;
}

/*
 *  patchInstrOffsetIntoInstr(upm) --
 *
 *  Patches offset between an instruction and the middle
 *  of another.
 */
bool patchDataInstrAddressIntoInstr(Patch *upm) {
    PatchDataInstrAddressToInstr *patch; 
    int og_val, new_val;

    patch = (PatchDataInstrAddressToInstr *) upm->p_patch;

    CASSERT(patch->p_src_instr != NULL, "NULL src_instr");
    CASSERT(patch->p_dest_instr != NULL, "NULL dest_instr");
    CASSERT(patch->p_dest_op != NULL, "NULL dest_op");

    og_val  = patch->p_dest_op->a_displacement;
    //og_val  = patch->p_dest_instr->i_src->a_displacement;
    new_val = patch->p_src_instr->i_final_address + patch->p_src_offset;
			    
    if (og_val != new_val) {
	patch->p_dest_op->a_displacement = new_val;
	//patch->p_dest_instr->i_src->a_displacement = new_val;
	UpdateInstr(patch->p_dest_instr);
	return true;
    }
    
    return false;
}


/*
 *  patchInstrInstrOffsetToInstrClone(upm,dest_instr) --
 */
static void patchDataInstrAddressToInstrClone(
Patch *upm, Instruction *dest_instr) {
    MALFUNCTION("not implemented !!! \n");	
}



/*
 *  patchInstrOffsetIntoInstrGetSources(upm) --
 */
UPMList *patchDataInstrAddressIntoInstrGetSources(Patch *upm) {
    UPMList *list;
    PatchDataInstrAddressToInstr *patch; 

    patch = (PatchDataInstrAddressToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_instr;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_offset;

    return list;
}



/*
 *  patchInstrOffsetIntoInstrGetDestinations(upm) --
 */
UPMList *patchDataInstrAddressIntoInstrGetDestinations(Patch *upm) {
    UPMList *list;
    PatchDataInstrAddressToInstr *patch; 

    patch = (PatchDataInstrAddressToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_instr;

    list->u_members[1].l_type = UPM_OPERAND_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_op;

    return list;
}



/*
 *  AddPatchInstrOffsetIntoInstr(src_instr,src_offset,dest_instr,dest_op) --
 *
 *  Patches into dest_instr (dest_op), the relative distance to 
 *  src_instr+src_offset. This is useful for example, to patch relative
 *  calls without relocation information into the middle of data instructions.
 */
Patch *AddPatchDataInstrAddressToInstr(Instruction *src_instr,
int src_offset , Instruction *dest_instr, Operand *dest_op) {
    Patch *upm;
    PatchDataInstrAddressToInstr *patch; 

    patch = zalloc(sizeof(PatchDataInstrAddressToInstr));
    patch->p_src_instr  = src_instr;
    patch->p_src_offset = src_offset;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_op    = dest_op;

    upm = UPMCreate(patch,
	&patchDataInstrAddressIntoInstr,
	&patchDataInstrAddressToInstrClone,
	&patchDataInstrAddressIntoInstrGetSources,
	&patchDataInstrAddressIntoInstrGetDestinations);

    UPMTableInsertByDest(dest_op, upm);
    UPMTableInsertBySource(src_instr, upm);
    ll_add(upm2s_list, upm);

    return upm;
}

/*
 *  patchSymbolValueIntoInstr(upm)
 *  Patches the value of a symbol into the instruction
 *  History : Added for absolute symbols which were
 *  bindings to address of sections ( like _etext) 
 */
bool patchSymbolValueIntoInstr(Patch *upm) {
    PatchSymbolValueToInstr *patch; 
    int og_val, new_val;

    patch = (PatchSymbolValueToInstr *) upm->p_patch;

    CASSERT(patch->p_src_sym != NULL, "NULL src_instr");
    CASSERT(patch->p_dest_instr != NULL, "NULL dest_instr");
    CASSERT(patch->p_dest_op != NULL, "NULL dest_op");

    og_val  = patch->p_dest_op->a_displacement;
    new_val = patch->p_src_offset + patch->p_src_sym->sy_value;
			    
    if (og_val != new_val) {
        patch->p_dest_op->a_displacement = new_val;
        UpdateInstr(patch->p_dest_instr);
        return true;
    }
    
    return false;
}


/*
 *  patchInstrInstrOffsetToInstrClone(upm,dest_instr) --
 */
static void patchSymbolValueIntoInstrClone(
Patch *upm, Instruction *dest_instr) {
    MALFUNCTION("not implemented !!! \n");	
}



/*
 *  patchInstrOffsetIntoInstrGetSources(upm) --
 */
UPMList *patchSymbolValueIntoInstrGetSources(Patch *upm) {
    UPMList *list;
    PatchSymbolValueToInstr *patch; 

    patch = (PatchSymbolValueToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SYMBOL_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_sym;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_offset;
    
    return list;
}



/*
 *  patchInstrOffsetIntoInstrGetDestinations(upm) --
 */
UPMList *patchSymbolValueIntoInstrGetDestinations(Patch *upm) {
    UPMList *list;
    PatchSymbolValueToInstr *patch; 

    patch = (PatchSymbolValueToInstr *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_instr;

    list->u_members[1].l_type = UPM_OPERAND_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_op;

    return list;
}



/*
 *  AddPatchSymbolValueIntoInstr(...)
 *
 *  Patches into dest_instr (dest_op), symbol relocation
 *  to the new value of the symbol 
 * 
 */
Patch *AddPatchSymbolValueIntoInstr(Symbol *sym, ADDRESS addend,
    Instruction *dest_instr, Operand *dest_op) {

    Patch *upm;
    PatchSymbolValueToInstr *patch; 

    patch = zalloc(sizeof(PatchSymbolValueToInstr));
    patch->p_src_sym  = sym;
    patch->p_src_offset = addend;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_op    = dest_op;


    upm = UPMCreate(patch,
        &patchSymbolValueIntoInstr,
        &patchSymbolValueIntoInstrClone,
        &patchSymbolValueIntoInstrGetSources,
        &patchSymbolValueIntoInstrGetDestinations);

    UPMTableInsertByDest(dest_op, upm);
    UPMTableInsertBySource(sym, upm);    

    ll_add(upm2s_list, upm);

    return upm;
}

/******************************************************************************/
/*
 *  patchSymbolValueIntoSectionAddress(upm)
 *  Patches the value of a symbol into the instruction
 *  History : Added for absolute symbols which were
 *  bindings to address of sections ( like _etext) 
 */
bool patchSymbolValueIntoSectionAddress(Patch *upm) {
    PatchSymbolValueToSectionAddress *patch; 
    int og_val, new_val;

    patch = (PatchSymbolValueToSectionAddress *) upm->p_patch;

    CASSERT(patch->p_src_sym != NULL, "NULL src_instr");

    og_val =
        loadLE(&patch->p_dest_section->s_newdata[patch->p_dest_patch_offset]);
    new_val = patch->p_src_offset + patch->p_src_sym->sy_value;
			    
    if (og_val != new_val) {
	storeLE(&patch->p_dest_section->s_newdata
	    [patch->p_dest_patch_offset], new_val);
        return true;
    }
    
    return false;
}


/*
 *  patchInstrInstrOffsetToInstrClone(upm,dest_instr) --
 */
static void patchSymbolValueIntoSectionAddressClone(
Patch *upm, Instruction *dest_instr) {
    MALFUNCTION("not implemented !!! \n");	
}



/*
 *  patchInstrOffsetIntoInstrGetSources(upm) --
 */
UPMList *patchSymbolValueIntoSectionAddressGetSources(Patch *upm) {
    UPMList *list;
    PatchSymbolValueToSectionAddress *patch; 

    patch = (PatchSymbolValueToSectionAddress *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SYMBOL_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_src_sym;

    list->u_members[1].l_type = UPM_OFFSET_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_src_offset;
    
    return list;
}



/*
 *  patchInstrOffsetIntoInstrGetDestinations(upm) --
 */
UPMList *patchSymbolValueIntoSectionAddressGetDestinations(Patch *upm) {
    UPMList *list;
    PatchSymbolValueToSectionAddress *patch; 

    patch = (PatchSymbolValueToSectionAddress *) upm->p_patch;
    list = UPMMemberListCreate(2);

    list->u_members[0].l_type = UPM_SECTION_MEMBER_TYPE;
    list->u_members[0].l_member = &patch->p_dest_section;

    list->u_members[1].l_type = UPM_OPERAND_MEMBER_TYPE;
    list->u_members[1].l_member = &patch->p_dest_patch_offset;

    return list;
}



/*
 *  AddPatchSymbolValueIntoSectionAddress(...)
 *
 *  Patches into dest_instr (dest_op), symbol relocation
 *  to the new value of the symbol 
 * 
 */
Patch *AddPatchSymbolValueIntoSectionAddress(Symbol *sym, ADDRESS addend,
    Section *dest_section, int dest_offset) {

    Patch *upm;
    PatchSymbolValueToSectionAddress *patch; 

    patch = zalloc(sizeof(PatchSymbolValueToSectionAddress));
    patch->p_src_sym  = sym;
    patch->p_src_offset = addend;
    patch->p_dest_section = dest_section;
    patch->p_dest_patch_offset = dest_offset;


    upm = UPMCreate(patch,
        &patchSymbolValueIntoSectionAddress,
        &patchSymbolValueIntoSectionAddressClone,
        &patchSymbolValueIntoSectionAddressGetSources,
        &patchSymbolValueIntoSectionAddressGetDestinations);

    UPMTableInsertByDest(dest_section, upm);
    UPMTableInsertBySource(sym, upm);    

    ll_add(upm2s_list, upm);

    return upm;
}



/******************************************************************************/

/*
 *  patchDataInstrToDataInstr(upm) -- patch from one data instr to another.
 */
bool patchDataInstrToDataInstrRelative(Patch *upm) {
    PatchDataInstrToDataInstr *patch;
    ADDRESS old_value, new_value;

    patch = upm->p_patch;

    if (patch->p_src_instr == NULL) {
        WARNING("src_instr == NULL");
        return false;
    }

    if (patch->p_dest_instr == NULL) {
        WARNING("dest_instr == NULL");
        return false;
    }

    memcpy(&old_value,
        &(patch->p_dest_instr->i_literal_data)[patch->p_dest_patch_offset],
        sizeof(old_value));

    // the -4 here is under the assumption that the address is going to be
    // 4 byte
    new_value = patch->p_src_instr->i_final_address +
        patch->p_src_patch_offset -
        patch->p_dest_instr->i_final_address -
        patch->p_dest_patch_offset - 4;    

    if (old_value != new_value) {
        memcpy(&patch->p_dest_instr->i_literal_data[patch->p_dest_patch_offset],
            &new_value, sizeof(new_value));
        return true;
    }
    return false;
}



/*
 *  patchInstrDataInstrToDataInstrGetSources(upm) --
 */
UPMList *patchInstrDataInstrToDataInstrRelativeGetSources(Patch *upm) {
    UPMList *list;
    PatchDataInstrToDataInstr *patch;

    patch = (PatchDataInstrToDataInstr *) (upm->p_patch);
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &(patch->p_src_instr);

    return list;
}



/*
 *  patchInstrDataInstrToDataInstrGetDestinations(upm) --
 */
UPMList *patchInstrDataInstrToDataInstrRelativeGetDestinations(Patch *upm) {
    UPMList *list;
    PatchDataInstrToDataInstr *patch;

    patch = (PatchDataInstrToDataInstr *) (upm->p_patch);
    list = UPMMemberListCreate(1);

    list->u_members[0].l_type = UPM_INSTR_MEMBER_TYPE;
    list->u_members[0].l_member = &(patch->p_dest_instr);

    return list;
}



/*
 *  AddPatchDataInstrToDataInstrRelative(src, src_offset, dest, dest_offset) --
 */
Patch *AddPatchDataInstrToDataInstrRelative(Instruction *src_instr,
    int src_offset, Instruction *dest_instr, int dest_offset) {

    PatchDataInstrToDataInstr *patch;
    Patch *upm;

    patch = zalloc(sizeof(PatchDataInstrToDataInstr));
    patch->p_src_instr = src_instr;
    patch->p_src_patch_offset = src_offset;
    patch->p_dest_instr = dest_instr;
    patch->p_dest_patch_offset = dest_offset;

    upm = UPMCreate(patch,
        &patchDataInstrToDataInstrRelative,
        NULL,
        &patchInstrDataInstrToDataInstrRelativeGetSources,
        &patchInstrDataInstrToDataInstrRelativeGetDestinations);

    //SOMU added this
    UPMTableInsertByDest(dest_instr, upm);
    UPMTableInsertByDest(src_instr, upm);        
    ll_add(upm2s_list, upm);

    return upm;
}
