/*  dem2pgm.c -- convert DEM file to PGM file  */

/* lumped together from topovista 3.0 code */



#include <ctype.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <unistd.h>

#define SET_BINARY_MODE(f)

#  define FOPEN(path,mode) fopen(path,mode)
#  define FDOPEN(filedes,mode) fdopen(filedes,mode)
#  define FCLOSE(file)  fclose(file)
#  define FREAD(ptr,size,nmemb,file) fread(ptr,size,nmemb,file)
#  define FGETS(s,size,file) fgets(s,size,file)
#  define GETC(file) getc(file)
#  define UNGETC(c,file) ungetc(c,file)
typedef FILE *FILETYPE;

#ifndef PI
#define PI 3.1415926535897932385
#endif

#define ARCSEC_PER_RAD ((360 * 60 * 60) / (2.0 * PI))
#define METERS_PER_ARCSEC (1.0e7 / (90 * 60 * 60))
#define METERS_PER_FEET 0.3048

#ifndef MAX
#define MAX(a, b) (((a)>(b))?(a):(b))
#endif

#ifndef MIN
#define MIN(a, b) (((a)>(b))?(b):(a))
#endif

typedef struct { double u, v; } UV;

void nospace(void);
FILETYPE open4read( char *path );



/*  dem.h -- structures and prototype for reading DEM files  */


/*
 * ============== USGS DEM format ==============
 */
#define MAXROWS 1420		/* the biggest DEM size I've seen */
#define MAXCOLS 1201

#define NAD27 1
#define WGS72 2
#define WGS84 3
#define NAD83 4
#define OLD_HAWAII_DATUM 5
#define PUERTO_RICO_DATUM 6

#define GEO_SYSTEM 0
#define UTM_SYSTEM 1

typedef struct {
  int rows, cols;
  double dz2m;		/* scale from data units to meters */
  int dzoffset;		/* add this to data item then scale */
  unsigned short dzmax;	/* data item maximum (always >= 0) */

  /* information from "A" record */
  char name[145];
  int level;
  int elevation_pattern;
  int reference_system;
  int zone;
  double map_projection_parameter[15];	/* always zero */
  int ground_unit;
  int elevation_unit;
  int n_polygon;
  UV corner[4];
  double minElevation, maxElevation;
  double axis_angle;
  int accuracy;
  double resolution[3];
  double cellx, celly;
  int datum;		/* one of the above datums (default=NAD27) */

  /* information from "B" records */
  int rowsInProfile[MAXCOLS], colsInProfile[MAXCOLS];
  UV firstUVInProfile[MAXCOLS];
  double firstElevationInProfile[MAXCOLS];
  double minElevationInProfile[MAXCOLS], maxElevationInProfile[MAXCOLS];
  unsigned short data[MAXCOLS][MAXROWS];

  /* information from "C" record */
  int C2;
  int C2x, C2y, C2z;
  int C2sample_size;
  int C5;
  int C5x, C5y, C5z;
  int C5sample_size;
} Dem;

Dem *readDEM(char *filename);

/*
 * ============== Elevation Model =================
 */
typedef struct {
  int width, height;
  double dz2m;
  int dzoffset;
  unsigned short dzmax;

  int zone;
  int datum;
  int ref_system;	/* either GEO_SYSTEM or UTM_SYSTEM */
  UV groundO;		/* minimum UTM or arcsec corner */
  UV groundDelta;	/* change in UTM or arcsec per change in x and y */

  /***** longtitude and latitude are in arc seconds****/
  /***** order is longtitude=U, latitude=V         ****/
  UV swUV, nwUV, neUV, seUV;	/* longitude,latitude of 4 corners */
  UV swXY, nwXY, neXY, seXY;	/* x,y of 4 corners */
  unsigned short *data;

  /* modified by Xiangyu Zhang */
  float *normal;
} Em;







void usage(char *progname)
{
  fprintf(stderr,
	  "usage: %s [DEMfile [PGMfile]]\n", progname);
  exit(1);
}


int main(int argc, char *argv[])
{
  int i, j;
  FILE *out;
  int opti;
  char *inName,*outName;
  Dem *dem;
	
  opti = 1;
  if (argc > opti) {
    inName = argv[opti];
    opti++;
  } else {
    inName = NULL;
  }
  if (argc > opti) {
    outName = argv[opti];
    opti++;
  } else {
    outName = NULL;
  }
  if (argc > opti) usage(argv[0]);

  if (outName) {
    out = fopen(outName,"wb");
    if (out==NULL) {
      fprintf(stderr, "Can't open %s for write.\n",outName);
      exit(1);
    }
  } else {
    out = stdout;
  }

  dem = readDEM(inName);

  if (dem == NULL)
    exit(1);
	
//  fprintf(stderr, "%s\n", dem->name);
//  fprintf(stderr, "Writing...");
//  fflush(stderr);
	
  fprintf(out, "P2\n");
  fprintf(out, "# produced by dem2pgm\n");
  fprintf(out, "%d %d\n", dem->cols, dem->rows);
  fprintf(out, "%d\n", (int)(dem->maxElevation));
  for (i = dem->rows - 1; i >= 0; --i) {
    for(j = 0; j < dem->cols; ++j) {
      fprintf(out, "%u ", dem->data[j][i]);
    }
    fprintf(out, "\n");
  }
//  fprintf(stderr, "done\n");
	
  return 0;
}

void nospace(void)
{
  fprintf(stderr,"Out of Memory.\n");
  exit(1);
}



FILETYPE open4read( char *path )
{
  if( path == NULL ) {
    SET_BINARY_MODE(stdin);
    return FDOPEN(fileno(stdin), "rb");
  } else {
    return FOPEN(path, "rb");
  }
}




/*  dem.c -- functions for handling USGS Digital Elevation Model files  */


static Dem *tryReadDEM(FILETYPE f);
static int getDEMInt(FILETYPE f);
static int getDEMInt2(FILETYPE f);
static int getDEMInt6(FILETYPE f);
static double getDEMDouble12(FILETYPE f);
static double getDEMDouble24(FILETYPE f);



Dem *readDEM(char *filename)
{
  Dem *dem;
  FILETYPE f;

  if( (f = open4read(filename)) == NULL ) return NULL;
  dem = tryReadDEM(f);
  FCLOSE(f);
  return dem;
}



static Dem *tryReadDEM(FILETYPE f)
{
  Dem *dem;
  int i,j,k;
  int col;
  int row_id, col_id;
  int colsInProfiles, numProfiles;
  UV min, max;
  int firstIndex;
  int dzoffset;
  int truncated;
  int tmp;
  unsigned short dzmax;

  dem = (Dem *)malloc(sizeof(Dem));
  if (dem == NULL) {
    fprintf(stderr,"Can't allocate space for Dem structure.\n");
    return(NULL);
  }
	
  /* A records */
	
  /* A 1 */
  if (!FGETS(dem->name, 145, f)) {
    fprintf(stderr, "Empty input file.\n");
    free(dem);
    return NULL;
  }
	
  /* A 2 */
  dem->level = getDEMInt6(f);
	
  /* A 3 */
  dem->elevation_pattern = getDEMInt6(f);
	
  /* A 4 */
  dem->reference_system = getDEMInt6(f);

  /* A 5 */
  dem->zone = getDEMInt6(f);
	
  /* A 6 */
  for (i = 0; i < 15; i++) {
    dem->map_projection_parameter[i] = getDEMDouble24(f);
  }
	
  /* A 7 */
  dem->ground_unit = getDEMInt6(f);
	
  /* A 8 */
  dem->elevation_unit = getDEMInt6(f);
  if (dem->elevation_unit == 1) {
    dem->dz2m = METERS_PER_FEET;
  } else {
    dem->dz2m = 1.0;
  }
	
  /* A 9 */
  dem->n_polygon = getDEMInt6(f);
	
  /* A 10 */
  for (i = 0; i < 4; i++) {
    dem->corner[i].u = getDEMDouble24(f);
    dem->corner[i].v = getDEMDouble24(f);
    if (i == 0) {
      min.u = dem->corner[i].u;
      max.u = dem->corner[i].u;
      min.v = dem->corner[i].v;
      max.v = dem->corner[i].v;
    } else {
      min.u = MIN(min.u, dem->corner[i].u);
      max.u = MAX(max.u, dem->corner[i].u);
      min.v = MIN(min.v, dem->corner[i].v);
      max.v = MAX(max.v, dem->corner[i].v);
    }
  }
	
  /* A 11 */
  dem->minElevation = getDEMDouble24(f);
  dem->maxElevation = getDEMDouble24(f);

  /* A 12 */
  dem->axis_angle = getDEMDouble24(f);		/* assumed to be 0.0 */
	
  /* A 13 */
  dem->accuracy = getDEMInt6(f);
	
  /* A 14 */
  for (i = 0; i < 3; i++) {
    dem->resolution[i] = getDEMDouble12(f);
  }
  if (dem->resolution[2] == 0) {
    fprintf(stderr,"Error reading DEM file.\n");
    free(dem);
    return NULL;
  }
  dem->dz2m *= dem->resolution[2];	/* scale by accuracy */
  dzoffset = (int)(dem->minElevation/dem->resolution[2])-1;
  if (dzoffset < -32767) dzoffset=0-1;
	
  /* A 15 */
  colsInProfiles = getDEMInt6(f);
  if (colsInProfiles != 1) {
    fprintf(stderr, "Cannot handle more than one column profiles.\n");
    free(dem);
    return NULL;
  }

  numProfiles = getDEMInt6(f);
  if (numProfiles > MAXCOLS) {
    fprintf(stderr, "Too many profiles (%d) in data set.\n",numProfiles);
    free(dem);
    return NULL;
  }
	
  /* A remaining */
  /*
    All USGS DEM's lacking the new type A record, elements 16-29, are NAD 27.
    All NIMA DEM's and DTED's lacking datum descriptors are WGS 72.
    All DEM's having the new type A record elements have datums as
    indicated in type A record, element 27.

    1=North American Datum 1927 (NAD 27)
    2=World Geodetic System 1972 (WGS 72)
    3=WGS 84
    4=NAD 83
    5=Old Hawaii Datum
    6=Puerto Rico Datum

    Since I don't know how to determine a NIMA DEM or DTED, I assume NAD 27
    (for UTM) and WGS 72 (for GEO) unless told otherwise.
  */
  if( dem->reference_system == UTM_SYSTEM ) {
    dem->datum = NAD27;
  } else {
    dem->datum = WGS72;
  }
  for (i = 865; i <= 1024; i++) {
    if( i == 891 ) {
      j = getDEMInt2(f);
      if( j > 0 ) {
	dem->datum = j;
      }
      i += 1;
    } else if (GETC(f) == '\n')
      break;
  }

  /* B records */
		
  dem->cols = numProfiles;
  dem->rows = (int)(1 + (max.v - min.v)/dem->resolution[1]);
		
  if (dem->rows > MAXROWS) {
    fprintf(stderr, "Too many rows (%d) in DEM.\n",dem->rows);
    free(dem);
    return NULL;
  }

  truncated = 0;
  dzmax = 0;
  for (col = 0; col < numProfiles; ++col) {
    for (i = 0; i < dem->rows; ++i) {
      dem->data[col][i] = 0;
    }
			
    /* B 1 */
    row_id = getDEMInt(f);
    col_id = getDEMInt(f);
    if (col_id != col + 1 || row_id != 1) {
      fprintf(stderr, "Profile %d claims to be %d.\n", col + 1, col_id);
      free(dem);
      return NULL;
    }
			
    /* B 2 */
    dem->rowsInProfile[col] = getDEMInt6(f);
    dem->colsInProfile[col] = getDEMInt6(f);
			
    /* B 3 */
    dem->firstUVInProfile[col].u = getDEMDouble24(f);
    dem->firstUVInProfile[col].v = getDEMDouble24(f);
    firstIndex = (int)((dem->firstUVInProfile[col].v - min.v)/dem->resolution[1]);
			
    /* B 4 */
    dem->firstElevationInProfile[col] = getDEMDouble24(f);
			
    /* B 5 */
    dem->minElevationInProfile[col] = getDEMDouble24(f);
    dem->maxElevationInProfile[col] = getDEMDouble24(f);
			
    /* B 6 */
#ifdef DEBUG
    fprintf(stderr, "Profile %d starts with row (%lf - %lf)/%lf = %d.\n",
	    col + 1, dem->firstUVInProfile[col].v, min.v, dem->resolution[1],
	    firstIndex);
#endif
			
    for (i = firstIndex; i < firstIndex + dem->rowsInProfile[col]; i++) {
      if (i >= dem->rows) {
	fprintf(stderr, "Too many rows in DEM profile %d.\n", col + 1);
	free(dem);
	return NULL;
      }

      /* WARNING some DEM files have record delimiters (like '\n')
	 that may break a single elevation into two.  We do not handle
	 these types of DEM files.
	 */
      tmp = getDEMInt(f);

      if (tmp != -32767) {		/* -32767 ==> no data */
	if (tmp < dzoffset+1) {		/* the min elevation was wrong */
#ifdef DEBUG
	  fprintf(stderr,"Rescaling for new minimum elevation.\n");
#endif
	  for (j = 0; j <= col; ++j) {
	    for (k = 0; k < dem->rows; ++k) {
	      if (dem->data[j][k] != 0) {
		if (dem->data[j][k] + dzoffset-tmp+1 > USHRT_MAX) {
		  truncated = 1;
		  dem->data[j][k] = USHRT_MAX;
		} else {
		  dem->data[j][k] += (unsigned short)(dzoffset-tmp+1);
		}
		dzmax = MAX(dzmax,dem->data[j][k]);
	      }
	    }
	  }
	  dzoffset = tmp-1;
	}

	if (tmp-dzoffset > USHRT_MAX) {
	  truncated = 1;
	  dem->data[col][i] = USHRT_MAX;
	} else {
	  dem->data[col][i] = (unsigned short)(tmp-dzoffset);
	}
	dzmax = MAX(dzmax,dem->data[col][i]);
      }
    }
			
  }
		
  if (truncated) {
    fprintf(stderr,"**Warning** Some elevations have been truncated.\n");
  }
  dem->dzoffset = dzoffset;
  dem->dzmax = dzmax;

  /* C record */
		
  if (dem->accuracy != 0) {
			
    /* C 1 */
    dem->C2 = getDEMInt6(f);
			
    /* C 2 */
    dem->C2x = getDEMInt6(f);
    dem->C2y = getDEMInt6(f);
    dem->C2z = getDEMInt6(f);
			
    /* C 3 */
    dem->C2sample_size = getDEMInt6(f);
			
    /* C 4 */
    dem->C5 = getDEMInt6(f);
			
    /* C 5 */
    dem->C5x = getDEMInt6(f);
    dem->C5y = getDEMInt6(f);
    dem->C5z = getDEMInt6(f);
			
    /* C 6 */
    dem->C5sample_size = getDEMInt6(f);
  }
		
  return(dem);
}


static void getDEMD2E(int n, char *buf, FILETYPE f)
{
  int i;
	
  for (i = 0; i < n; ++i) {
    buf[i] = GETC(f);
    if (buf[i] == 'D') {
      buf[i] = 'E';
    }
  }
  buf[n] = '\0';
}



/* ? characters */
static int getDEMInt(FILETYPE f)
{
  int i = 0, c;
  int sgn = 1;
	
  while( (c = GETC(f)) != EOF && isspace(c) )
    ;
  if( c == '-' ) {
    sgn = -1;
    c = GETC(f);
  } else if( c == '+' ) {
    c = GETC(f);
  }
  while( isdigit(c) ) {
    i = 10 * i + (c - '0');
    c = GETC(f);
  }
  UNGETC(c,f);
  return sgn * i;
}


static void getDEMNoCR(int n, char *buf, FILETYPE f)
{
  int i;
	
  for (i = 0; i < n; ++i) {
    buf[i] = GETC(f);
    if (buf[i] == '\n') {
      --i;
    }
  }
  buf[n] = '\0';
	
}


/* 2 characters */
static int getDEMInt2(FILETYPE f)
{
  char buf[3];
	
  getDEMNoCR(2, buf, f);
  return atoi(buf);
}


/* 6 characters */
static int getDEMInt6(FILETYPE f)
{
  char buf[7];
	
  getDEMNoCR(6, buf, f);
  return atoi(buf);
}


/* 12 characters */
static double getDEMDouble12(FILETYPE f)
{
  char buf[13];
	
  getDEMD2E(12, buf, f);
  return atof(buf);
}



/* 24 characters */
static double getDEMDouble24(FILETYPE f)
{
  char buf[25];
	
  getDEMD2E(24, buf, f);
  return atof(buf);
}
