/* analyze functions, loops etc. to compute assorted information */

#include "sloop.h"

extern bbllist bbl_table[];
extern gobjptr allfuncs, allloops, allbbls;
extern long totedges, totunknown;
extern int cg_threshold;

/*********************************************************************/
extern bblptr bbl_lookup(long bblno);
extern funcptr func_lookup(long fno);
void define_edge_info(edgeptr eptr, gobjptr allbbls);
edgeptr update_edge_info(objptr optr, edgeptr eptr, edgeptr alledges);
int get_entry_pt(edgeptr epts, funcptr func);
void set_call_sites(funcptr fptr);
void mark_critical_paths(funcptr fptr);
/*********************************************************************/

void analyze(f)
funcptr f;
{
  Processed(f) = TRUE;
  define_edge_info(EdgeList(f), allbbls);
  update_edge_info(Contents(f), EdgeList(f), EdgeList(f));
  set_call_sites(f);
  mark_critical_paths(f);
}

void define_edge_info(eptr, allbbls)
edgeptr eptr;
gobjptr allbbls;
{
  bblptr from, to;
  bbllist xbbl;
  long from_id, to_id, currid, hashval;
  gobjptr gobj;
  edgeptr e0, e;
  edgeinfo einfo;

  for (e0 = eptr ; e0; e0 = Next(e0)) {
    from_id = FromId(e0); 
    to_id = ToId(e0);
    from = to = NULL;

    from = bbl_lookup(from_id);
    to   = bbl_lookup(to_id);

    if (!from || !to) {
      fprintf(stderr, "error: can't find endpoint(s) of edge %d -> %d\n",
	      from_id, to_id);
      exit(-1);
    }
    /*
     * allocate a new edgeinfo node, fill it in, and link it in.
     */
    einfo = (edgeinfo)NewNode(einfo);
    EFrom(einfo)    = from;
    ETo(einfo)      = to;
    ELink(einfo)    = BSR(from) ? 1 : 0;
    Count(einfo)    = Count(e0);
    EProb(einfo)    = -1.0;
    EdgeInfo(e0) = einfo;
    /*
     * Add to as a successor of from, and from as a predecessor of to.
     */
    e = (edgeptr)NewNode(e);
    FromId(e) = from_id;
    ToId(e) = to_id;
    EdgeInfo(e) = einfo;
    Next(e) = OutEdges(from);
    OutEdges(from) = e;
    NumSuccs(from) += 1;
    
    e = (edgeptr)NewNode(e);
    FromId(e) = from_id;
    ToId(e) = to_id;
    EdgeInfo(e) = einfo;
    Next(e) = InEdges(to);
    InEdges(to) = e;
    NumPreds(to) += 1;
  }
  /*
   * Now compute and fill in edge probabilities.
   */
  for (e0 = eptr; e0; e0 = Next(e0)) {
    if (ECount(e0) > 0 && Count(From(e0)) > 0) {
      Prob(e0) = (float)ECount(e0) / (float)Count(From(e0));
    }
    else {
      Prob(e0) = 0.0;
    }
  } /* for */
}

/*
 * update_edge_info() recursively traverses the body of a function,
 * filling in edge information.
 */
edgeptr update_edge_info(optr, eptr, alledges)
objptr optr;
edgeptr eptr, alledges;
{
  objptr bodyobj;
  edgeptr e0, e1, etmp, seen;

  /*
   * First, fill in the edgeinfo field for each edge.
   */
  for (e0 = eptr; e0; e0 = Next(e0)) {
    for (e1 = alledges; e1; e1 = Next(e1)) {
      if (FromId(e0) == FromId(e1) && ToId(e0) == ToId(e1)) {
	EdgeInfo(e0) = EdgeInfo(e1);
	break;
      }
    }
  }
  /*
   * Next, determine which edges at this level in fact belong to a
   * nested structure, and therefore aren't really "valid" at this
   * level.
   */
  seen = NULL;
  for (bodyobj = optr; bodyobj; bodyobj = Next(bodyobj)) {
    if (IsLoopNode(bodyobj)) {
      e0 = update_edge_info(Contents(LoopNode(bodyobj)), EdgeList(LoopNode(bodyobj)), alledges);
      for (etmp = e0; etmp && Next(etmp); etmp = Next(etmp)) 
	;  /* run down to the end */
      if (etmp) {
	assert(!Next(etmp));
	Next(etmp) = seen;
      }
      seen = e0;  /* merge e0 with seen */
    }
  }
  /*
   * Now go through eptr and eliminate edges that belong to nested structures.
   */
  for (e1 = eptr; e1; e1 = Next(e1)) {
    Valid(e1) = TRUE;
    for (etmp = seen; etmp; etmp = Next(etmp)) {
      if (FromId(e1) == FromId(etmp) && ToId(e1) == ToId(etmp)) {
	Valid(e1) = FALSE;
      }
    }
    if (Valid(e1)) {  /* new edge: add it to seen */
      etmp = (edgeptr)NewNode(etmp);
      FromId(etmp) = FromId(e1);
      ToId(etmp) = ToId(e1);
      EdgeInfo(etmp) = EdgeInfo(e1);
      Next(etmp) = seen;
      seen = etmp;
    }
  }
  return seen;
}

/*
 * set_call_sites() takes a pointer to a function and defines its
 * call sites.  If the function has a defined list of call-sites
 * already (from the input file) this function does nothing.
 * Otherwise, it traverses the body of the function looking
 * for basic blocks with no predecessors: all such blocks are added
 * to the list of entry points.  If no such blocks are found, the
 * first block in the list of body objects is takes as the entry
 * point.
 */
void set_call_sites(fptr)
funcptr fptr;
{
  callerptr cptr;
  objptr optr;
  callerptr epts, currep, nextep;

  if (CallerList(fptr)) {  /* there is a list of call sites */
    return;
  }

  epts = NULL;
#if 0
  for (optr = Contents(fptr); optr; optr = Next(optr)) {
    if (IsBBlNode(optr) && NumPreds(BblNode(optr)) == 0) {
      nextep = (callerptr)NewNode(nextep);
      if (!nextep) error("Out of memory!");
      FuncName(nextep)    = "???";
      BblCallFrom(nextep) = -1;
      BblCallTo(nextep)   = Id(BblNode(optr));
      Count(nextep)       = 0;
      Attach(epts, currep, nextep);
    }
  }
#endif
  if (!epts) {   /* no blocks without predecessors */
    optr = Contents(fptr);
    nextep = (callerptr)NewNode(nextep);
    if (!nextep) error("Out of memory!");
    FuncName(nextep)    = "???";
    BblCallFrom(nextep) = -1;
    BblCallTo(nextep)   = IsBBlNode(optr) ? Id(BblNode(optr)) : Id(Hdr(LoopNode(optr)));
    Count(nextep)       = 0;
    Attach(epts, currep, nextep);
    fprintf(stderr,
	    "Warning: can't find entry point for function %s, assuming it's block B%d\n", 
	    FuncName(fptr), 
	    BblCallTo(nextep));
  }
  CallerList(fptr) = epts;
}

/*
 * mark_critical_paths() starts at the entry points of a function and marks
 * the critical path(s) in the function.  The algorithm for this is as follows:
 *
 * 1. if A is an entry block then A is on the critical path.
 *
 * 2. if A is on the critical path then each successor B of A such that the
 *    edge A->B has maximal probability among the outedges of A is on the
 *    critical path (and A->B is a critical edge).
 *
 * 3. if A is a loop header for a loop L and B is an exit block for loop L such
 *    that there's an edge from B to some block C not in L, and the edge B->C
 *    is such that C has maximal execution count among all such edges, then B
 *    is on the critical path and C is on the critical path.
 *
 *    Note that in handling edges leaving loops, it isn't enough to consider
 *    simply maximal-probability edges from a block in L on the critical path,
 *    from criterion 2, to a block not in L: it's possible that for every exit
 *    block of L that satisfies this condition, the edges that leave the loop
 *    have probability 0 (see loop L6594 of ijpeg for an example of this).
 *
 * 4. There's no guarantee that criteria (2) and (3) will result in a connected
 *    sequence of blocks, i.e., a path (see loop L6594 of ijpeg).  To address
 *    this, we need to throw in predecessors of nodes added by criterion (3)
 *    if necessary.  However, for this it isn't enough to simply consider the
 *    probabilities of the in-edges of these nodes, since a high-probability
 *    in-edge need not be on a frequently executed path.  So we add in each
 *    predecessor with maximal weight.
 */
void mark_critical_paths(fptr)
funcptr fptr;
{
  objptr optr;
  bbllist rootblk, currblk, nextblk, loophdrs;
  bblptr bptr;
  callerptr cptr;
  edgeptr eptr, e0;
  float maxprob;
  int change, done, maxwt, maxcnt, ebbl;

  rootblk = NULL;
  loophdrs = NULL;
  for (cptr = CallerList(fptr); cptr; cptr = Next(cptr)) {
    bptr = bbl_lookup(BblCallTo(cptr));
    if (bptr && Count(bptr) > 0) { 
      nextblk = (bbllist)NewNode(nextblk);
      if (!nextblk) error("Out of memory!");
      GBBlNode(nextblk) = bptr;
      Next(nextblk) = rootblk;
      rootblk = nextblk;
    }
  }

  change = TRUE;
  while (change) {
    change = FALSE;
    while (rootblk) {
      bptr = GBBlNode(rootblk);
      currblk = rootblk;
      rootblk = Next(currblk);
      if (Critical(bptr)) continue;
      Critical(bptr) = TRUE;
      /*
       * find outedge with maximal probability, then follow all
       * outedges that have this probability.
       */
      maxprob = 0.0;
      for (eptr = OutEdges(bptr); eptr; eptr = Next(eptr)) {
	if (Prob(eptr) > maxprob) {
	  maxprob = Prob(eptr);
	}
      }
      if (maxprob == 0.0) continue;  /* never taken: ignore */
      for (eptr = OutEdges(bptr); eptr; eptr = Next(eptr)) {
	if (Prob(eptr) == maxprob) {
	  IsCriticalEdge(eptr) = TRUE;
	  nextblk = (bbllist)NewNode(nextblk);
	  if (!nextblk) error("Out of memory!");
	  GBBlNode(nextblk) = To(eptr);
	  Next(nextblk) = rootblk;     /* add to queue */
	  rootblk = nextblk;
	}
      } /* for */
      if (GLoopNode(GBBlNode(currblk)) &&
	  GBBlNode(currblk) == Hdr(GLoopNode(GBBlNode(currblk)))) {  /* loop header */
	Next(currblk) = loophdrs;
	loophdrs = currblk;
	change = TRUE;
      }
    }
    /*
     * at this point, rootblk == NULL; moreover change == TRUE iff
     * some loop header was added to the loophdrs queue in this
     * iteration.  We go and process the loophdrs queue.
     */
    while (loophdrs) {
      bptr = GBBlNode(loophdrs);
      currblk = loophdrs;
      loophdrs = Next(loophdrs);
      maxcnt = 0;
      for (eptr = EdgeList(GLoopNode(bptr)); eptr; eptr = Next(eptr)) {
	/*
	 * find an exit edge from this loop.  For this, eptr needs to satisfy:
	 *
	 * 1. it starts in the current loop;
	 * 2. it ends outside the loop; and
	 * 3. it doesn't end in a loop nested inside the current loop.
	 */
	if (GLoopNode(From(eptr)) && 
	    Hdr(GLoopNode(From(eptr))) == bptr &&  /* starts in the current loop... */
	    GLoopNode(From(eptr)) != GLoopNode(To(eptr)) && /* ends outside it */
	    !(GLoopNode(To(eptr)) && EnclLoop(GLoopNode(To(eptr))) == GLoopNode(From(eptr)))
	    ) {
	  if (Count(To(eptr)) > maxcnt) {
	    maxcnt = Count(To(eptr));
	  }
	}
      }
      if (maxcnt == 0) continue;  /* never executed: ignore */
      for (eptr = EdgeList(GLoopNode(bptr)); eptr; eptr = Next(eptr)) {
	if (GLoopNode(From(eptr)) && 
	    Hdr(GLoopNode(From(eptr))) == bptr &&
	    GLoopNode(From(eptr)) != GLoopNode(To(eptr)) &&
	    Prob(eptr) > 0.0 &&     /* this edge is actually taken */
	    Count(To(eptr)) == maxcnt) {
	  IsCriticalEdge(eptr) = TRUE;
	  bptr = From(eptr);
	  if (!Critical(bptr)) {
	    /*
	     * now work backwards until a critical block is found.
	     */
	    done = FALSE;
	    while (!done) {
	      Critical(bptr) = TRUE;
	      maxwt = 0;
	      for (e0 = InEdges(bptr); e0; e0 = Next(e0)) {
		if (Critical(From(e0))) {
		  done = TRUE;
		  IsCriticalEdge(e0) = TRUE;
		  break;
		}
		if (Weight(From(e0)) > maxwt) {
		  maxwt = Weight(From(e0));
		}
	      }
	      if (!done) {
		for (e0 = InEdges(bptr); e0; e0 = Next(e0)) {
		  if (Weight(From(e0)) == maxwt) {
		    bptr = From(e0);
		    IsCriticalEdge(e0) = TRUE;
		    break;
		  }
		}
	      }
	    }	      
	  }
	  if (!Critical(To(eptr))) {
	    nextblk = (bbllist)NewNode(nextblk);
	    GBBlNode(nextblk) = To(eptr);
	    Next(nextblk) = rootblk;
	    rootblk = nextblk;
	    change = TRUE;
	  }
	}
      }
    }
  }   /* while (change) */
}

void mk_callgraph()
{
  gobjptr optr;
  funcptr fptr;
  callerptr cptr;
  bblptr callerb, calleeb;
  cgptr cg0;

  for (optr = allfuncs; optr; optr = Next(optr)) {
    for (cptr = CallerList(GFuncNode(optr)); cptr; cptr = Next(cptr)) {
      fptr = func_lookup(CallerId(cptr));
      if (!fptr) {
	error("could not find function: %s\n", FuncName(cptr));
      }
      if (Visited(fptr)) {
	/*
	 * if the visited bit is set, it must be that the user doesn't
	 * want to see this function and stuff it calls
	 */
	continue;
      }

      /* construct a node representing this call */
      cg0 = (cgptr)NewNode(cg0);
      if (!cg0) error("out of memory!");
      callerb = bbl_lookup(BblCallFrom(cptr));
      calleeb = bbl_lookup(BblCallTo(cptr));
      if (!callerb || !calleeb) {
	error("could not find one of the bbls: %d and %d\n", 
	      BblCallFrom(cptr), BblCallTo(cptr));
      }
      Caller(cg0) = callerb;
      Callee(cg0) = calleeb;
      Count(cg0)  = Count(cptr);
      if (Count(cg0) > cg_threshold) {
	Significant(fptr) = TRUE;
      }
      /* attach this node to the caller's callee list */
      Next(cg0) = CalleeList(fptr);
      CalleeList(fptr) = cg0;
    }
  }
}

