/*
 *  HashTable.c -- A hash table with genericity, automatic memory freeing,
 *  and iteration.
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "utilities.h"



/*
 *  mod(m,n) -- computes m % n in [0, m) like % SHOULD do.
 *
 *  Works for negative numbers.
 */
static int mod(int m, int n) {
    int result = m % n;
    if (result < 0) {
	result += n;
    }
    return result;
}



/*
 *  refHashCode(obj) -- hash function for pointers.
 *  Divides by 4 so not all indices are multiples of 4.
 */
int refHashCode(void *obj) {
    return (int) obj / sizeof(void *);
}



/*
 *  intHashCode(ptr) -- hash function for ints.
 *  Just uses int's value.
 */
int intHashCode(void *ptr) {
    return (int) ptr;
}



/*
 *  strHashCode(ptr) --
 *
 *
 *  Returns a hashcode for this string. The hashcode for a
 *  String is computed as
 *
 *  s[0]*31^(n-1) + s[1]*31^(n-2) + ... + s[n-1]
 *
 *  where s[i] is the i'th character of the string, n is the length of
 *  the string, and ^ indicates exponentiation.
 *  (The hash value of the empty string is zero.)
 */
int strHashCode(void *ptr) {
    int i;
    int len;
    int key = 0;
    char *str = (char *) ptr;

    assert(ptr);

    len = strlen(str);
    for (i = 0; i < len; i++) {
	key = 19 * key + str[i];
    }

    return key;
}



/*
 *  ht_itrHasNext(itr) --
 */
static int ht_itrHasNext(Iterator *itr) {
    HashTable *table;
    HashEntry *entry;
    int index;
    int ii;

    assert(itr);
    table = (HashTable *) itr->it_dataStructure;
    entry = (HashEntry *) itr->it_opaque1;
    index = (int) itr->it_opaque2;

    if (entry->h_next) {
	return 1;
    }

    for (ii = index + 1; ii < table->ht_capacity; ii++) {
	if (table->ht_values[ii].h_next) {
	    return 1;
	}
    }

    return 0;
}



/*
 *  ht_itrNext(itr) --
 */
static void *ht_itrNext(Iterator *itr) {
    HashTable *table;
    HashEntry *entry;
    int index;
    int ii;

    assert(itr);
    table = (HashTable *) itr->it_dataStructure;
    entry = (HashEntry *) itr->it_opaque1;
    index = (int) itr->it_opaque2;

    if (entry->h_next) {
	entry = entry->h_next;
	itr->it_opaque1 = entry;
	return itr->it_current = entry->h_value;
    } else {
	for (ii = index + 1; ii < table->ht_capacity; ii++) {
	    if (table->ht_values[ii].h_next) {
		itr->it_opaque1 = table->ht_values[ii].h_next;
		itr->it_opaque2 = (void *) ii;
		return itr->it_current = table->ht_values[ii].h_next->h_value;
	    }
	}

	return itr->it_current = NULL;
    }
}



/*
 *  ht_itrRemove(itr) --
 */
static void *ht_itrRemove(Iterator *itr) {
    HashTable *table;
    HashEntry *entry;
    HashEntry *curr;
    void *value;
    int index;

    assert(itr);
    table = (HashTable *) itr->it_dataStructure;
    entry = (HashEntry *) itr->it_opaque1;
    index = (int) itr->it_opaque2;

    curr = &table->ht_values[index];
    while (curr->h_next && curr->h_next != entry) {
	curr = curr->h_next;
    }
    itr->it_opaque1 = curr;

    value = entry->h_value;
    curr->h_next = entry->h_next;
    free(entry);

    table->ht_size--;
    return value;
}



/*
 *  ht_new(keyType,valueType,capacity,load) --
 *
 *  Creates and returns an empty hash table of the given size.
 */
HashTable *ht_new(HTYPE keyType, HTYPE valueType, int capacity, double load) {

    int (*compareKeyFunc) (void *, void *) =
	(keyType == HASH_STR) ? strCompare : refCompare;

    int (*hashFunc) (void *) = (keyType == HASH_STR) ?
	strHashCode : (keyType == HASH_PTR) ? refHashCode : intHashCode;

    int (*compareValueFunc) (void *, void *) =
	(valueType == HASH_STR) ? strCompare : refCompare;

    void (*printKeyFunc) (void *) = (keyType == HASH_STR) ?
	strPrint : (keyType == HASH_PTR) ? refPrint : intPrint;

    void (*printValueFunc) (void *) = (valueType == HASH_STR) ?
	strPrint : (valueType == HASH_PTR) ? refPrint : intPrint;

    void (*freeValueFunc) (void *) =
	(valueType == HASH_STR) ? free : (valueType == HASH_PTR) ? free : NULL;

    return ht_newGeneric(capacity, load, hashFunc, compareKeyFunc,
	compareValueFunc, printKeyFunc, printValueFunc, freeValueFunc);

}



/*
 *  HashTable(valueType,capacity,load,hashFunc,compareKeyFunc,printKeyFunc) --
 *
 *  Creates and returns an empty hash table of the given size.
 */
HashTable *ht_newGenericKey(HTYPE valueType, int capacity, double load,
int (*hashFunc) (void *), int (*compareKeyFunc) (void *, void *),
void (*printKeyFunc) (void *)) {

    int (*compareValueFunc) (void *, void *) =
	(valueType == HASH_STR) ? strCompare : refCompare;

    void (*printValueFunc) (void *) = (valueType == HASH_STR) ? strPrint :
	(valueType == HASH_PTR) ? refPrint : intPrint;

    void (*freeValueFunc) (void *) =
	(valueType == HASH_STR) ? free : (valueType == HASH_PTR) ? free : NULL;

    return ht_newGeneric(capacity, load, hashFunc, compareKeyFunc,
	compareValueFunc, printKeyFunc, printValueFunc, freeValueFunc);

}



/*
 *  ht_newGenericValue(
 *  keyType,capacity,load,compareValueFunc,printValueFunc,freeValueFunc) --
 */
HashTable *ht_newGenericValue(HTYPE keyType, int capacity, double load,
int (*compareValueFunc) (void *, void *), void (*printValueFunc) (void *),
void (*freeValueFunc) (void *)) {

    int (*compareKeyFunc) (void *, void *) =
	(keyType == HASH_STR) ? strCompare : refCompare;

    int (*hashFunc) (void *) = (keyType == HASH_STR) ?
	strHashCode : (keyType == HASH_PTR) ? refHashCode : intHashCode;

    void (*printKeyFunc) (void *) = (keyType == HASH_STR) ?
	strPrint : (keyType == HASH_PTR) ? refPrint : intPrint;

    return ht_newGeneric(capacity, load, hashFunc, compareKeyFunc,
	compareValueFunc, printKeyFunc, printValueFunc, freeValueFunc);
}



/*
 *  ht_newGeneric(
 *  capacity,load,hashFunc,compareKeyFunc,compareValueFunc,
 *  printKeyFunc,printValueFunc,freeValueFunc) --
 */
HashTable *ht_newGeneric(int capacity, double load, int (*hashFunc) (void *),
int (*compareKeyFunc) (void*, void*), int (*compareValueFunc) (void*, void*),
void (*printKeyFunc) (void*), void (*printValueFunc) (void*),
void (*freeValueFunc) (void*)) {

    HashTable *table = NULL;
    assert(capacity > 0 && load > 0.0 && compareValueFunc);

    table = zalloc(sizeof(HashTable));
    table->ht_values = zalloc(capacity * sizeof(HashEntry));
    table->ht_capacity = capacity;
    table->ht_load = load;
    table->ht_hashCode = hashFunc ? hashFunc : intHashCode;
    table->ht_compareKey = compareKeyFunc ? compareKeyFunc : refCompare;
    table->ht_compareValue = compareValueFunc ? compareValueFunc : refCompare;
    table->ht_printKey = printKeyFunc ? printKeyFunc : intPrint;
    table->ht_printValue = printValueFunc ? printValueFunc : refPrint;
    table->ht_free = freeValueFunc ? freeValueFunc : NULL;
    return table;
}



/*
 *  ht_free(table) --
 *
 *  Unallocates all memory previously allocated to the given hash table.
 *  Does NOT free the memory stored in the keys or values in the table.
 */
void ht_free(HashTable *table) {
    int i = 0;
    HashEntry *e = NULL;
    HashEntry *next = NULL;
    assert(table);

    for (i = 0; i < table->ht_capacity; i++) {
	e = &table->ht_values[i];
	while (e->h_next) {
	    next = e->h_next->h_next;
	    free(e->h_next);
	    e->h_next = next;
	}
    }
    free(table->ht_values);
}



/*
 *  ht_freeWithElements(table) --
 *
 *  Unallocates all memory previously allocated to the given hash table,
 *  and the memory used by the table's element values.
 */
void ht_freeWithElements(HashTable *table) {
    int i;
    HashEntry *e;
    HashEntry *next;
    assert(table);

    for (i = 0; i < table->ht_capacity; i++) {
	e = &table->ht_values[i];
	while (e->h_next) {
	    next = e->h_next->h_next;
	    if (table->ht_free) {
		table->ht_free(e->h_value);
	    }
	    free(e->h_next);
	    e->h_next = next;
	}
    }
    free(table->ht_values);
}



/*
 *  ht_clearAndFreeElements(table) -- removes all values from the hash table.
 */
void ht_clearAndFreeElements(HashTable *table) {
    int i;
    HashEntry *e;
    HashEntry *next;

    assert(table);
    for (i = 0; i < table->ht_capacity; i++) {
	e = &table->ht_values[i];
	while (e->h_next) {
	    next = e->h_next->h_next;
	    if (table->ht_free) {
		table->ht_free(e->h_value);
	    }
	    free(e->h_next);
	    e->h_next = next;
	}
    }
    memset(table->ht_values, 0, table->ht_capacity * sizeof(HashEntry));
    table->ht_size = 0;
}



/*
 *  ht_clear(table) -- removes all values from the hash table.
 */
void ht_clear(HashTable *table) {
    int i;
    HashEntry *e;
    HashEntry *next;
    assert(table);

    for (i = 0; i < table->ht_capacity; i++) {
	e = &table->ht_values[i];
	while (e->h_next) {
	    next = e->h_next->h_next;
	    free(e->h_next);
	    e->h_next = next;
	}
    }
    memset(table->ht_values, 0, table->ht_capacity * sizeof(HashEntry));
    table->ht_size = 0;
}



/*
 *  ht_containsKey(table,key) --
 *
 *  Returns 1 if the given hash table contains the given key; 0 if not.
 */
int ht_containsKey(HashTable *table, void *key) {
    int code = 0;
    HashEntry *e = NULL;
    assert(table);
    assert(table);

    code = mod(table->ht_hashCode(key), table->ht_capacity);
    e = &table->ht_values[code];
    while (e->h_next) {
	e = e->h_next;
	if (e->h_hash == code && !table->ht_compareKey(key, e->h_key)) {
	    return 1;
	}
    }
    return 0;
}



/*
 *  ht_containsValue(table,value) --
 *
 *  Returns 1 if the given hash table contains the given value; 0 if not.
 */
int ht_containsValue(HashTable *table, void *value) {
    int i = 0;
    HashEntry *e = NULL;
    assert(table);

    for (i = 0; i < table->ht_capacity; i++) {
	e = &table->ht_values[i];
	while (e->h_next) {
	    e = e->h_next;
	    if (table->ht_compareValue(e->h_value, value)) {
		return 1;
	    }
	}
    }
    return 0;
}



/*
 *  ht_equals(table1,table2) --
 *
 *  Returns 1 if the given hash tables contain the same set of values.
 *  Doesn't consider keys, just the values stored at them.
 */
int ht_equals(HashTable *table1, HashTable *table2) {
    int i;
    HashEntry *e = NULL;
    assert(table1 && table2);

    if (table1->ht_size != table2->ht_size) {
	return 0;
    }
    for (i = 0; i < table1->ht_capacity; i++) {
	e = &table1->ht_values[i];
	while (e->h_next) {
	    e = e->h_next;
	    if (!ht_containsValue(table2, e->h_value)) {
		return 0;
	    }
	}
    }

    for (i = 0; i < table2->ht_capacity; i++) {
	e = &table2->ht_values[i];
	while (e->h_next) {
	    e = e->h_next;
	    if (!ht_containsValue(table1, e->h_value)) {
		return 0;
	    }
	}
    }
    return 1;
}



/*
 *  ht_get(table,key) --
 *
 *  Returns the element hashed by the given key in this table.
 *  Assumes the hash key is a pointer.
 *  If no element is hashed by this key, returns NULL.
 */
void *ht_get(HashTable *table, void *key) {
    int code;
    HashEntry *e = NULL;
    assert(table);

    code = mod(table->ht_hashCode(key), table->ht_capacity);
    e = &table->ht_values[code];
    while (e->h_next) {
	e = e->h_next;
	if (e->h_hash == code && !table->ht_compareKey(key, e->h_key)) {
	    return e->h_value;
	}
    }
    return NULL;
}



/*
 *  ht_isEmpty(table) --
 *
 *  Returns whether or not the given hash table contains no elements.
 */
int ht_isEmpty(HashTable *table) {
    assert(table);
    return table->ht_size == 0;
}



/*
 *  ht_iterator(table) --
 */
Iterator *ht_iterator(HashTable *table) {
    Iterator *itr;
    assert(table);

    itr = zalloc(sizeof(Iterator));
    itr->it_dataStructure = table;
    itr->it_next = ht_itrNext;
    itr->it_hasNext = ht_itrHasNext;
    itr->it_remove = ht_itrRemove;
    itr->it_opaque1 = &table->ht_values[0];
    itr->it_opaque2 = NULL;
    return itr;
}



/*
 *  ht_put(table,key,value) --
 */
void *ht_put(HashTable *table, void *key, void *value) {
    void *prevValue = NULL;
    HashEntry *e = NULL;
    HashEntry *curr = NULL;
    assert(table);

    if (ht_containsKey(table, key)) {
	prevValue = ht_remove(table, key);
    }
    if ((table->ht_size + 1) >= table->ht_capacity * table->ht_load) {
	ht_rehash(table);
    }
    e = zalloc(sizeof(HashEntry));
    e->h_hash = mod(table->ht_hashCode(key), table->ht_capacity);
    e->h_key = key;
    e->h_value = value;
    curr = &table->ht_values[e->h_hash];
    while (curr->h_next) {
	curr = curr->h_next;
    }
    curr->h_next = e;
    table->ht_size++;
    return prevValue;
}



/*
 *  ht_putAll(table,otherTable) --
 */
void ht_putAll(HashTable *table, HashTable *otherTable) {
    assert(0);
}



/*
 *  ht_rehash(table)
 *
 *  doubles the size of the given hash table and remaps all of its values.
 */
void ht_rehash(HashTable *table) {
    int i = 0;
    int oldCapacity = 0;
    HashEntry *oldValues = NULL;
    HashEntry *e = NULL;
    HashEntry *prev = NULL;
    assert(table);

    oldCapacity = table->ht_capacity;
    oldValues = table->ht_values;
    table->ht_size = 0;
    table->ht_capacity *= 2;
    table->ht_values = zalloc(table->ht_capacity * sizeof(HashEntry));

    for (i = 0; i < oldCapacity; i++) {
	e = oldValues[i].h_next;
	while (e) {
	    ht_put(table, e->h_key, e->h_value);
	    prev = e;
	    e = e->h_next;
	    free(prev);
	}
    }
    free(oldValues);
}



/*
 *  ht_remove(table,key) --
 */
void *ht_remove(HashTable *table, void *key) {
    int code = 0;
    void *value = NULL;
    HashEntry *e = NULL;
    HashEntry *prev = NULL;
    assert(table);

    code = mod(table->ht_hashCode(key), table->ht_capacity);
    e = &table->ht_values[code];

    while (e->h_next) {
	prev = e;
	e = e->h_next;
	if (e->h_hash == code && !table->ht_compareKey(key, e->h_key)) {
	    value = e->h_value;
	    prev->h_next = e->h_next;
	    free(e);
	    table->ht_size--;
	    break;
	}
    }
    return value;
}



/*
 *  ht_size(table) --
 *
 *  returns the number of values stored in the given hash table.
 */
int ht_size(HashTable *table) {
    assert(table);
    return table->ht_size;
}



/*
 *  ht_print(table) -- prints the given hash table on standard output.
 */
void ht_print(HashTable *table) {
    int i;
    HashEntry *e = NULL;
    assert(table);

    printf("HashTable(size=%d, capacity=%d):\n",
	table->ht_size, table->ht_capacity);
    for (i = 0; i < table->ht_capacity; i++) {
	e = &table->ht_values[i];
	if (e->h_next) {
	    printf("  [%3d]: ", i);
	    while (e->h_next) {
		e = e->h_next;
		printf("(");
		table->ht_printKey(e->h_key);
		printf(" --> ");
		table->ht_printValue(e->h_value);
		printf(")");
	    }
	    printf("\n");
	}
    }
    return;

}

/*
 *  int main() {
 *
 *  HashTable *table = ht_new(HASH_INT, HASH_STR, 4, 0.75);
 *  HashTable *table2 = ht_new(HASH_STR, HASH_STR, 4, 0.75);
 *  Iterator *itr;
 *  HashTable *table2 = ht_new(50, DEFAULT_HASH_LOAD);
 *  char *one = "one";
 *  char *four = "four";
 *
 *  ht_put(table,  (void*)1, one);
 *  ht_put(table,  (void*)3, "three");
 *  ht_put(table, (void*)11, "eleven");
 *  ht_put(table,  (void*)7, "seven");
 *  ht_put(table, (void*)15, "fifteen");
 *  ht_put(table, (void*)1, "ONE");
 *  ht_print(table);
 *  printf("do i have one? %d\n", ht_containsValue(table, one));
 *  printf("do i have four? %d\n", ht_containsValue(table, four));
 *  printf("do i have key 11? %d\n", ht_containsKey(table, (void*)11));
 *  printf("do i have key 2? %d\n", ht_containsKey(table, (void*)2));
 *  printf("do i have key twenty? %d\n", ht_containsKey(table, "twenty"));
 *  printf("I got \"%s\"\n", (char*)ht_get(table, (void*)7));
 *  ht_put(table2, "one", "one");
 *  ht_put(table2, "two", "two");
 *  ht_put(table2, "three", "three");
 *  ht_put(table2, "four", "four");
 *  ht_put(table2, "five", "five");
 *  ht_put(table2, "six", "six");
 *  ht_print(table2);
 *  printf("do i have key five? %d\n", ht_containsKey(table2, "five"));
 *  printf("do i have key seven? %d\n", ht_containsKey(table2, "seven"));
 *  ht_rehash(table2);
 *  ht_print(table2);
 *  printf("itr:");
 *  itr = ht_iterator(table2);
 *  while (itr->it_hasNext(itr)) {
 *      printf("\nelement: ");
 *      table->ht_printValue(one = (char*)itr->h_next(itr));
 *      if (strchr(one, 'e'))
 *      itr->it_remove(itr);
 *  }
 *  printf("\n\n");
 *  itr = ht_iterator(table2);
 *  while (itr->it_hasNext(itr)) {
 *      printf("\nelement: ");
 *      table->ht_printValue(itr->h_next(itr));
 *  }
 *  printf("\n");
 *  ht_remove(table, (void*)11);
 *  ht_print(table);
 *  ht_print(table2);
 *  ht_free(table);
 *  ht_free(table2);
 *  return 0;
 *  }
 */
