/*
 *  LinkedList.c -- A fully functional, sortable LinkedList with set operations.
 *  Also supports iteration.
 */
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "utilities.h"



typedef struct Link {
    void *lk_data;
    struct Link *lk_prev;
    struct Link *lk_next;
} Link;



#define LL_LOCK()
#define LL_UNLOCK()



/* STATIC HELPER FUNCTIONS */



/*
 *  newLink(prev,data,next) --
 *  Creates and returns a new empty linked list node.
 */
static Link *newLink(Link *prev, void *data, Link *next) {

    Link *result = zalloc(sizeof(Link));
    result->lk_data = data;
    result->lk_prev = prev;
    result->lk_next = next;
    return result;
}



/*
 *  ll_itrHasNext(itr) --
 */
static int ll_itrHasNext(Iterator *itr) {
    LinkedList *list;
    Link *link;

    assert(itr);

    list = (LinkedList *) itr->it_dataStructure;
    link = (Link *) itr->it_opaque1;

    return link && link->lk_next && link->lk_next != list->ll_tail;
}



/*
 *  ll_itrNext(itr) --
 */
static void *ll_itrNext(Iterator *itr) {
    LinkedList *list;
    Link *link;

    assert(itr);

    list = (LinkedList *) itr->it_dataStructure;
    link = (Link *) itr->it_opaque1;

    itr->it_opaque1 = link = link->lk_next;
    return itr->it_current = link->lk_data;
}



/*
 *  ll_itrRemove(itr) --
 */
static void *ll_itrRemove(Iterator *itr) {
    Link *link;
    void *result;
    LinkedList *list;

    assert(itr && itr->it_opaque1);

    list = (LinkedList *) itr->it_dataStructure;
    link = (Link *) itr->it_opaque1;

    result = link->lk_data;
    itr->it_opaque1 = link->lk_prev;

    // remove node
    LL_LOCK();
    link->lk_next->lk_prev = link->lk_prev;
    link->lk_prev->lk_next = link->lk_next;
    free(link);
    list->ll_size--;
    LL_UNLOCK();

    return result;
}



/* CONSTRUCTORS / DESTRUCTORS */



/*
 *  ll_new(type) --
 *
 *  Initializes a new empty linked list assumed to hold addresses.
 */
LinkedList *ll_new(LTYPE type) {
    return ll_newPtr(type);
}



/*
 *  ll_newInt(type) --
 *
 *  Initializes a new empty linked list of integers.
 */
LinkedList *ll_newInt(LTYPE type) {
    return ll_newGeneric(type, refCompare, intPrint, NULL);
}



/*
 *  ll_newPtr(type) --
 *
 *  Initializes a new empty linked list assumed to hold addresses.
 */
LinkedList *ll_newPtr(LTYPE type) {
    return ll_newGeneric(type, refCompare, refPrint, free);
}



/*
 *  ll_newStr(type) --
 *
 *  Initializes a new empty linked list.
 */
LinkedList *ll_newStr(LTYPE type) {
    return ll_newGeneric(type, strCompare, strPrint, free);
}



/*
 *  ll_newGeneric(type,compareFunc)printFunc)freeFunc) --
 *
 *
 *  Initializes a new empty linked list of given type and
 *  with the given function to pairwise compare/sort its elements.
 *
 *  The function is expected to return:
 *  -1 if the first argument is 'less than' the second argument,
 *  0 if they are equal,
 *  1 if the first argument is 'greater than' the second argument.
 */
LinkedList *ll_newGeneric(LTYPE type, int (*compareFunc) (void *, void *),
    void (*printFunc) (void *), void (*freeFunc) (void *)) {

    LinkedList *list = zalloc(sizeof(LinkedList));
    list->ll_head = newLink(NULL, NULL, NULL);
    list->ll_tail = newLink(list->ll_head, NULL, NULL);
    list->ll_head->lk_next = list->ll_tail;
    list->ll_size = 0;
    list->ll_type = type;
    list->ll_lock = 1;
    list->ll_compare = compareFunc ? compareFunc : refCompare;
    list->ll_print = printFunc ? printFunc : refPrint;
    list->ll_free = freeFunc ? freeFunc : NULL;

    return list;
}



/*
 *  ll_free(list) --
 *
 *  Frees the given list (doesn't free the elements themselves).
 */
void ll_free(LinkedList *list) {
    Link *current;
    Link *next;

    assert(list);

    // free each node
    LL_LOCK();
    for (current = list->ll_head; current != NULL; current = next) {
	next = current->lk_next;
	free(current);
    }

    free(list);
    LL_UNLOCK();
}



/*
 *  ll_freeWithElements(list) --
 *
 *  Frees the given list, and calls its free function to free every element
 *  contained inside it.
 *  #%#% ACTUALLY, IT APPEARS AS IF THAT HAS BEEN DISABLED.
 */
void ll_freeWithElements(LinkedList *list) {
    Link *current;
    Link *next;

    assert(list);

    // free each node
    LL_LOCK();
    for (current = list->ll_head->lk_next;
    current != list->ll_tail; current = next) {
	next = current->lk_next;
//  if (list->ll_free)
//    list->ll_free(current->lk_data);
    }

    ll_free(list);
    LL_UNLOCK();
}



/* PROCEDURE DECLARATIONS */



/*
 *  ll_add(list,val) -- add an element to the end of the list.
 */
void *ll_add(LinkedList *list, void *val) {
    assert(list);
    return (list->ll_type == LIST_SORTED)
	? ll_insertSorted(list, val)
	: ll_addLast(list, val);
}



/*
 *  ll_addFirst(list,val) --
 */
void *ll_addFirst(LinkedList *list, void *val) {
    Link *theNewLink = NULL;

    assert(list);

    LL_LOCK();

    theNewLink = newLink(list->ll_head, val, list->ll_head->lk_next);
    list->ll_head->lk_next->lk_prev = theNewLink;
    list->ll_head->lk_next = theNewLink;
    list->ll_size++;

    LL_UNLOCK();

    return val;
}



/*
 *  ll_addLast(list,val) --
 */
void *ll_addLast(LinkedList *list, void *val) {
    Link *theNewLink = NULL;

    assert(list);

    LL_LOCK();

    theNewLink = newLink(list->ll_tail->lk_prev, val, list->ll_tail);
    list->ll_tail->lk_prev->lk_next = theNewLink;
    list->ll_tail->lk_prev = theNewLink;
    list->ll_size++;

    LL_UNLOCK();

    return val;
}



/*
 *  ll_addAll(list,list2) -- appends list2 to list1.
 */
void ll_addAll(LinkedList *list, LinkedList *list2) {
    Link *current;

    assert(list && list2);

    // check for invalid arguments
    if (!list || !list2) {
	return;
    }

    LL_LOCK();

    // iterate through list2, adding each piece to list
    for (current = list2->ll_head->lk_next; current != list2->ll_tail;
	current = current->lk_next)
	ll_add(list, current->lk_data);
    LL_UNLOCK();
}



/*
 *  ll_backmap(list,func) --
 *
 *
 *  This function applies the function func on every element
 *  in the list but in reverse order
 */
void ll_backmap(LinkedList *list, void (*func) (void **)) {
    Link *current;

    assert(list && func);

    // check for invalid arguments
    if (!list || !func) {
	return;
    }

    // iterate through list, applying func to each element
    for (current = list->ll_tail->lk_prev; current != list->ll_head;
	current = current->lk_prev)
	func(&current->lk_data);
}



/*
 *  ll_clear(list) -- remove all elements from the given list.
 */
void ll_clear(LinkedList *list) {
    LL_LOCK();
    while (!ll_isEmpty(list)) {
	ll_removeFirst(list);
    }
    LL_UNLOCK();
}



/*
 *  ll_clearAndFreeElements(list) --
 *
 *  Calls given list's free function to free every element
 *  contained inside it.  Also clears list to have 0 elements.
 */
void ll_clearAndFreeElements(LinkedList *list) {
    Link *current;
    Link *next;

    assert(list);

    // free each node
    LL_LOCK();
    for (current = list->ll_head->lk_next;
    current != list->ll_tail; current = next) {
	next = current->lk_next;
//  if (list->ll_free)
//    list->ll_free(current->lk_data);
    }

    // clear
    while (!ll_isEmpty(list)) {
	ll_removeFirst(list);
    }

    LL_UNLOCK();
}



/*
 *  ll_contains(list,val) --
 *
 *  Returns whether the list contains the given value.
 *  Compares by the list's given comparison function.
 */
int ll_contains(LinkedList *list, void *val) {
    assert(list);
    return ll_indexOf(list, val) >= 0;
}



/*
 *  ll_difference(list1,list2) --
 *
 *  Returns the set difference of the two lists (which represent sets).
 *  The difference is the elements in list1 that are not in list2.
 */
LinkedList *ll_difference(LinkedList *list1, LinkedList *list2) {
    Link *current = NULL;
    LinkedList *result = NULL;

    assert(list1 && list2);

    // make new empty list using list1's type and function ptrs
    result = ll_newGeneric(
	list1->ll_type, list1->ll_compare, list1->ll_print, list1->ll_free);

    // add all elements of list1 that are not in list2
    for (current = list1->ll_head->lk_next; current != list1->ll_tail;
	    current = current->lk_next) {
	if (!ll_contains(list2, current->lk_data)) {
	    ll_add(result, current->lk_data);
	}
    }

    return result;
}



/*
 *  ll_equals(list1,list2) --
 *
 *  Returns whether the two given linked lists are memberwise equal,
 *  using the comparison function of list1 to compare elements pairwise.
 *
 *  If the lists are sets, the order is not considered; only the elements.
 */
int ll_equals(LinkedList *list1, LinkedList *list2) {
    int i, n;

    assert(list1 && list2);

    n = ll_size(list1);
    if (n != ll_size(list2)) {
	return 0;
    }

    if (list1->ll_type == LIST_SET && list2->ll_type == LIST_SET) {
	// do not have to be in same order, just have to be present
	for (i = 0; i < n; i++) {
	    if (!ll_contains(list1, ll_get(list2, i))
		    || !ll_contains(list2, ll_get(list1, i))) {
		return 0;
	    }
	}
    } else {
	for (i = 0; i < n; i++) {
	    if (list1->ll_compare(ll_get(list1, i), ll_get(list2, i))) {
		return 0;
	    }
	}
    }

    return 1;
}



/*
 *  ll_first(list) -- returns the first element of the list.
 */
void *ll_first(LinkedList *list) {
    return ll_get(list, 0);
}



/*
 *  ll_get(list,index) --
 *
 *  Return the element at index index
 *  Must check that index < numElements head
 */
void *ll_get(LinkedList *list, int index) {
    int i;
    Link *current;

    assert(list && 0 <= index && index < list->ll_size);

    // check for invalid arguments
    if (!list || !(0 <= index && index < list->ll_size)) {
	return NULL;
    }

    // iterate through list to find location to get
    current = list->ll_head->lk_next;
    for (i = 0; i < index; i++) {
	current = current->lk_next;
    }

    return current->lk_data;
}



/*
 *  ll_indexOf(list,val) --
 *
 *  Lookup val in list using list's compare function for equality test
 *  Return index if present, otherwise -1
 */
int ll_indexOf(LinkedList *list, void *val) {
    int i = 0;
    Link *current;

    assert(list);

    // check for invalid arguments
    if (!list) {
	return -1;
    }

    // iterate through list to find val, if present
    for (current = list->ll_head->lk_next;
    current != list->ll_tail; current = current->lk_next, i++) {
	if (!list->ll_compare(current->lk_data, val)) {
	    return i;
	}
    }

    return -1;
}



/*
 *  ll_insert(list,val) -- add an element to the start of the list.
 */
void *ll_insert(LinkedList *list, void *val) {
    assert(list);
    return (list->ll_type == LIST_SORTED)
	? ll_insertSorted(list, val)
	: ll_insertAt(list, val, 0);
}



/*
 *  ll_insertAt(list,val,index) -- add an element at the given list index.
 */
void *ll_insertAt(LinkedList *list, void *val, int index) {
    int i = 0;
    Link *current = NULL;
    Link *theNewLink = NULL;

    assert(list && 0 <= index && index <= ll_size(list));

    // check for invalid parameters
    // (includes duplicate check if list is a set)
    if (!(0 <= index && index <= ll_size(list))
	|| (list->ll_type == LIST_SET && ll_contains(list, val)))
	return val;

    LL_LOCK();

    // iterate through to find place to insert this new link
    for (current = list->ll_head->lk_next;
	current != list->ll_tail && i < index; current = current->lk_next, i++);

    // put the new node here
    theNewLink = newLink(current->lk_prev, val, current);
    theNewLink->lk_prev->lk_next = theNewLink;
    theNewLink->lk_next->lk_prev = theNewLink;
    list->ll_size++;

    LL_UNLOCK();
    return val;
}



/*
 *  ll_insertSorted(list,val) -- add an element at the given index of the list.
 */
void *ll_insertSorted(LinkedList *list, void *val) {
    int i = 0;
    Link *current = NULL;
    Link *theNewLink = NULL;

    assert(list);

    // duplicate check if list is a set
    if (list->ll_type == LIST_SET && ll_contains(list, val)) {
	return val;
    }

    LL_LOCK();

    // iterate through to find place to insert this new link
    // (insert it after all nodes that are <= it
    for (current = list->ll_head->lk_next;
	current != list->ll_tail && list->ll_compare(current->lk_data,val) <= 0;
	current = current->lk_next, i++);

    // put the new node here
    theNewLink = newLink(current->lk_prev, val, current);
    theNewLink->lk_prev->lk_next = theNewLink;
    theNewLink->lk_next->lk_prev = theNewLink;
    list->ll_size++;

    LL_UNLOCK();
    return val;
}



/*
 *  ll_intersection(list1,list2) --
 *
 *  Returns the set intersection of the two lists (which represent sets).
 *  The intersection is all elements in BOTH list1 and list2.
 */
LinkedList *ll_intersection(LinkedList *list1, LinkedList *list2) {
    Link *current = NULL;
    LinkedList *result = NULL;

    assert(list1 && list2);

    // make new empty list using list1's type and function ptrs
    result = ll_newGeneric(
	list1->ll_type, list1->ll_compare, list1->ll_print, list1->ll_free);

    // add all elements of list1 that are also in list2
    for (current = list1->ll_head->lk_next; current != list1->ll_tail;
	current = current->lk_next)
	if (ll_contains(list2, current->lk_data)) {
	    ll_add(result, current->lk_data);
	}

    // add all elements of list2 that are also in list1
    for (current = list2->ll_head->lk_next; current != list2->ll_tail;
	current = current->lk_next)
	if (ll_contains(list1, current->lk_data)) {
	    ll_add(result, current->lk_data);
	}

    return result;
}



/*
 *  ll_isEmpty(list) --
 *
 * Returns 1 if the list has no elements, 0 if it has elements.
 */
int ll_isEmpty(LinkedList *list) {
    return ll_size(list) == 0;
}



/*
 *  ll_iterator(list) --
 */
Iterator *ll_iterator(LinkedList *list) {
    Iterator *itr;

    assert(list);
    itr = zalloc(sizeof(Iterator));
    itr->it_dataStructure = list;
    itr->it_next = ll_itrNext;
    itr->it_hasNext = ll_itrHasNext;
    itr->it_remove = ll_itrRemove;
    itr->it_opaque1 = list->ll_head;

    return itr;
}



/*
 *  ll_last(list) -- returns the last element of the list.
 */
void *ll_last(LinkedList *list) {
    return ll_get(list, ll_size(list) - 1);
}



/*
 *  ll_map(list,func) --
 *
 *  This function applies the function func on every element in the list.
 */
void ll_map(LinkedList *list, void (*func) (void **)) {
    Link *current;

    assert(list && func);

    // check for invalid arguments
    if (!list || !func) {
	return;
    }

    // iterate through list, applying func to each element
    for (current = list->ll_head->lk_next;
    current != list->ll_tail; current = current->lk_next) {
	func(&current->lk_data);
    }
}



/*
 *  ll_print(list) --
 *
 *  Prints this linked list, using the given function pointer to print
 *  each argument.
 */
void ll_print(LinkedList *list) {
    int i = 0;

    if (!list) {
	printf("NULL\n");
    } else {
	printf("LIST: [");

	if (ll_size(list) > 0) {
	    list->ll_print(ll_get(list, 0));

	    for (i = 1; i < ll_size(list); i++) {
		printf(", ");
		list->ll_print(ll_get(list, i));
	    }
	}

	printf("]\n");
    }
}



/*
 *  ll_remove(list,val) -- remove an element.
 */
void *ll_remove(LinkedList *list, void *val) {
    Link *current;
    assert(list);

    for (current = list->ll_head->lk_next;
	current != list->ll_tail; current = current->lk_next) {
	if (current->lk_data == val) {
	    break;
	}
    }

    if (current != list->ll_tail) {
	// remove node
	current->lk_next->lk_prev = current->lk_prev;
	current->lk_prev->lk_next = current->lk_next;
	free(current);
	list->ll_size--;

	return val;
    }

    return NULL;
}



/*
 *  ll_removeAt(list,index) --
 *
 *  Remove an element at the specified index
 *  Must check that index < numElements head
 */
void *ll_removeAt(LinkedList *list, int index) {
    int i;
    Link *current = NULL;
    void *result = NULL;

    assert(list && 0 <= index && index < list->ll_size);

    // check for valid argument
    if (!list || !(0 <= index && index < list->ll_size)) {
	return result;
    }

    // iterate through list to find location to remove
    LL_LOCK();
    current = list->ll_head->lk_next;
    for (i = 0; i < index; i++) {
	current = current->lk_next;
    }
    result = current->lk_data;

    if (current) {
	// remove node
	current->lk_next->lk_prev = current->lk_prev;
	current->lk_prev->lk_next = current->lk_next;
	free(current);
	list->ll_size--;
    }

    LL_UNLOCK();
    return result;
}



/*
 *  ll_removeFirst(list) -- removes and returns first element of list.
 */
void *ll_removeFirst(LinkedList *list) {
    return ll_removeAt(list, 0);
}



/*
 *  ll_removeLast(list) -- removes and returns last element of list.
 */
void *ll_removeLast(LinkedList *list) {
    return ll_removeAt(list, ll_size(list) - 1);
}



/*
 *  ll_setEquals(list1, list2) --
 *
 *  Returns whether the two linked lists, which represent sets,
 *  are memberwise equal.  Note that order is not considered, only contents.
 */

bool ll_setEquals(LinkedList *list1, LinkedList *list2) {
    int i, n;

    assert(list1 && list2);

    /* if their sizes differ, they're unequal */
    n = ll_size(list1);
    if (n != ll_size(list2)) {
	return false;
    }

    /* if one is missing an element from the other, they're unequal */
    for (i = 0; i < n; i++) {
	if (!ll_contains(list2, ll_get(list1, i))) {
	    return false;
	}
	if (!ll_contains(list1, ll_get(list2, i))) {
	    return false;
	}
    }

    return true;
}



/*
 *  ll_size(list) -- returns the number of elements in the list.
 */
int ll_size(LinkedList *list) {
    assert(list);

    // check for invalid arguments
    if (!list) {
	return -1;
    }

    return list->ll_size;
}



/*
 *  ll_sort2(list) --
 *
 *  Sorts using a selection sort by first transferring the linked list into
 *  an array.
 */
void ll_sort2(LinkedList *list) {
    int low, i, j, size;
    Link *links[ll_size(list)], *current;
    void *item;

    size = ll_size(list);

    for (i = 0, current = list->ll_head->lk_next; current != list->ll_tail;
	current = current->lk_next, i++) {
	links[i] = current;
    }

    // do a selection sort to order the data of the nodes
    for (i = 0; i < size; i++) {
	low = i;
	for (j = i + 1; j < size; j++) {
	    if (list->ll_compare(links[j]->lk_data, links[low]->lk_data) < 0) {
		low = j;
	    }
	}

	if (low != i) {
	    item = links[i]->lk_data;
	    links[i]->lk_data = links[low]->lk_data;
	    links[low]->lk_data = item;
	}
    }
}



/*
 *  ll_sort(list) --
 *
 *  Sorts the linked list using its compare function for equality test.
 *  Uses a crappy selection sort because I do not wish to implement
 *  anything tougher!
 */
void ll_sort(LinkedList *list) {
    int i, j, size, smallestIndex;
    void *el_i;
    void *el_j;
    void *el_smallest;

    assert(list);

    LL_LOCK();
    size = ll_size(list);
    for (i = 0; i < size; i++) {
	// find smallest element
	el_i = ll_get(list, i);
	el_smallest = el_i;
	smallestIndex = i;

	for (j = i + 1; j < size; j++) {
	    el_j = ll_get(list, j);

	    if (list->ll_compare(el_j, el_smallest) < 0) {
		el_smallest = el_j;
		smallestIndex = j;
	    }
	}

	// swap
	if (smallestIndex != i) {
	    ll_removeAt(list, i);
	    ll_insertAt(list, el_smallest, i);
	    ll_removeAt(list, smallestIndex);
	    ll_insertAt(list, el_i, smallestIndex);
	}
    }

    LL_UNLOCK();
}



/*
 *  ll_union(list1,list2) --
 *
 *  Returns the set union of the two linked lists (which represent sets).
 *  The union is all elements in either list1 or list2 or both.
 */
LinkedList *ll_union(LinkedList *list1, LinkedList *list2) {
    Link *current = NULL;
    LinkedList *result = NULL;

    assert(list1 && list2);

    // make new empty list using list1's type and function ptrs
    result = ll_newGeneric(
	list1->ll_type, list1->ll_compare, list1->ll_print, list1->ll_free);

    // add all elements of list1
    for (current = list1->ll_head->lk_next; current != list1->ll_tail;
	    current = current->lk_next) {
	if (!ll_contains(result, current->lk_data)) {
	    ll_add(result, current->lk_data);
	}
    }

    // add all elements of list2 not already present
    for (current = list2->ll_head->lk_next; current != list2->ll_tail;
	    current = current->lk_next) {
	if (!ll_contains(result, current->lk_data)) {
	    ll_add(result, current->lk_data);
	}
    }

    return result;
}



/*
 *  ll_enqueue(list,val) --
 */
void *ll_enqueue(LinkedList *list, void *val) {
    assert(list);
    assert(list->ll_type == LIST_UNSORTED);

    return ll_addLast(list, val);
}



/*
 *  ll_dequeue(list) --
 */
void *ll_dequeue(LinkedList *list) {
    assert(list);
    assert(list->ll_type == LIST_UNSORTED);

    return ll_removeFirst(list);
}



/*
 *  ll_push(list,val) --
 */
void *ll_push(LinkedList *list, void *val) {
    assert(list);
    assert(list->ll_type == LIST_UNSORTED);

    return ll_addFirst(list, val);
}



/*
 *  ll_pop(list) --
 */
void *ll_pop(LinkedList *list) {
    assert(list);
    assert(list->ll_type == LIST_UNSORTED);

    return ll_removeFirst(list);
}



/*
 *  Test driver.
 *  int main(void)
 *  {
 *  char arr[7] = {'h', 'e', 'l', 'l', 'o', '!', '\0'};
 *  LinkedList *ll;
 *  int i;
 *  char *str1, *str2;
 *  Iterator *itr;
 *
 *  for (i = 0; i < 1; i++)
 *  {
 *  ll = ll_newStr(LIST_UNSORTED);
 *  str1 = malloc(8);
 *  strcpy(str1, "hello!");
 *  ll_add(ll, str1);
 *  str2 = malloc(8);
 *  strcpy(str2, "byebye");
 *  ll_add(ll, str2);
 *
 *  ll_freeWithElements(ll);
 *
 *  ll_add(ll, "marty");
 *  ll_add(ll, "jessica");
 *  ll_add(ll, "ursula");
 *  ll_add(ll, "zheng");
 *  ll_add(ll, "shauna");
 *  ll_add(ll, "hello!");
 *  ll_add(ll, "mary");
 *  ll_add(ll, "susan");
 *  ll_add(ll, "lindsey");
 *  ll_add(ll, "kelly");
 *  ll_add(ll, "anna");
 *  ll_print(ll);
 *
 *  printf("has hello? %d \n", ll_contains(ll, arr));
 *
 *  ll_remove(ll, arr);
 *  ll_print(ll);
 *  printf("has hello? %d \n", ll_contains(ll, arr));
 *
 *  ll_sort(ll);
 *  ll_print(ll);
 *  itr = ll_iterator(ll);
 *  while (itr->it_hasNext(itr))
 *  {
 *  printf("\nelement: ");
 *  ll->ll_print(str1 = (char*)itr->lk_next(itr));
 *  if (strchr(str1, 'm'))
 *  itr->it_remove(itr);
 *  }
 *  printf("\n\n");
 *
 *  itr = ll_iterator(ll);
 *  while (itr->hasNext(itr))
 *  {
 *  printf("\nelement: ");
 *  ll->ll_print(str1 = (char*)itr->lk_next(itr));
 *  }
 *  printf("\n");
 *  }
 *
 *  return 0;
 *  }
 */
