/*
 *  output.c -- Output functions.
 *
 *  Errors and warnings are sent to stderr, and are also echoed on stdout
 *  if it is not the terminal (where presumably the stderr copy suffices).
 *
 *  Reports and debugging messages are output or suppressed depending on
 *  globals set by command-line options.
 */

#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include "util.h"

extern int verbosityLevel;
extern bool noWarnings;
extern bool showTimestamps;
extern bool debugFlags[128];

static char *timestamp();



/*
 *  pfatal(s) -- abort after calling perror(s).
 */
void pfatal(char *s) {
    fflush(stdout);
    perror(s);
    exit(1);
}



/*
 *  warning(fmt,args) -- issue warning message.
 */
void warning(char *fmt, ...) {

    va_list args;
    va_start(args, fmt);

    if (noWarnings) {
    	return;
    }

    if (! isatty(fileno(stdout))) {
	fprintf(stdout, "%sWARNING: ", timestamp());
	vfprintf(stdout, fmt, args);
	fprintf(stdout, "\n");
    }

    fflush(stdout);

    fprintf(stderr, "WARNING: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");

    va_end(args);
}



/*
 *  fatal(fmt,args) -- diagnose error and exit.
 */
void fatal(char *fmt, ...) {

    va_list args;
    va_start(args, fmt);

    if (! isatty(fileno(stdout))) {
	fprintf(stdout, "\n%sFATAL: ", timestamp());
	vfprintf(stdout, fmt, args);
	fprintf(stdout, "\n");
    }

    fflush(stdout);

    fprintf(stderr, "FATAL: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");

    va_end(args);
    exit(1);
}



/*
 *  malfunction(func,file,line,fmt,args) -- diagnose malfunction and exit.
 *
 *  A malfunction is an error in PLTO itself, rather than its input.
 */
void malfunction(const char *func, char *file, int line, char *fmt, ...) {

    va_list args;
    va_start(args, fmt);

    if (! isatty(fileno(stdout))) {
	fprintf(stdout,
	    "\n%sMalfunction detected in function %s at line %d of %s:\n",
	    timestamp(), func, line, file);
	fprintf(stdout, "%s", timestamp());
	vfprintf(stdout, fmt, args);
	fprintf(stdout, "\n");
    }

    fflush(stdout);

    fprintf(stderr, "Malfunction detected in function %s at line %d of %s:\n",
	func, line, file);
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");

    va_end(args);
    exit(1);
}



/*
 *  report(level,fmt,args) -- issue report message of the given level.
 *
 *  Should be called only after level has been validated by the REPORT macro.
 */
void report(int level, char *fmt, ...) {
    static bool first = true;			// first call?

    // space before major & minor milestones iff more detail is also wanted
    if (level <= RP_MINOR && level < verbosityLevel && !first) {
	if (level == RP_MAJOR && verbosityLevel > RP_MINOR) {
	    fprintf(stdout, "_______________________________________"
		"_______________________________________\n");
	}
	fprintf(stdout, "\n");
    }

    // print the report message
    va_list args;
    va_start(args, fmt);
    fprintf(stdout, "%s%d  ", timestamp(), level);
    if (level > RP_DETAIL) {
	fprintf(stdout, "            " + 18 - 2 * level);	// indent
    }
    vfprintf(stdout, fmt, args);
    fprintf(stdout, "\n");
    va_end(args);

    // check that the macro was used properly to protect this call
    if (! WANT_REPORT(level)) {
        MALFUNCTION("eligibility failure on preceding report message");
    }

    first = false;
}



/*
 *  debug(c,fmt,args) -- issue debugging message of class c.
 *
 *  Should be called only if either the level or the debugging class c
 *  has been selected.  This is checked by the DEBUG macro.
 */
void debug(char c, char *fmt, ...) {

    // print the debug message
    va_list args;
    va_start(args, fmt);
    fprintf(stdout, "%s%c ", timestamp(), c);
    vfprintf(stdout, fmt, args);
    fprintf(stdout, "\n");
    va_end(args);

    // check that the macro was used properly to protect this call
    if (! WANT_DEBUG(c)) {
        MALFUNCTION("eligibility failure on preceding debug message");
    }
}



/*
 *  timestamp() -- return a timestamp string in a static buffer.
 *
 *  Returns an empty string if timestamps are not wanted.
 */
static char *timestamp() {
    static char buf[20];

    if (showTimestamps) {
	sprintf(buf, "%5.2f ", clock() / (double) CLOCKS_PER_SEC);
	return buf;
    } else {
	return "";
    }
}
