/*
 *  x86/asm.c
 *
 *  An assembler/binary writer that takes instructions stored in the internal
 *  lists and converts them to instruction codes
 *
 *  Notes:
 *  The general encoding of an instruction looks somewhat like the following:
 *  (taken from Chapter 2 Vol 2 of the Intel Architecture manual)
 *
 *  ----------------------------------------------------------------------
 *  | Prefixes | Opcode | Opcode2 | MODRM  | SIB    | Displcmt | Immed   |
 *  | (1 to 4) | 1 byte | 1 byte  | 1 byte | 1 byte | (1,2,4)  | (1,2,4) |
 *  ----------------------------------------------------------------------
 *
 *  Prefixes are optional, as is the second byte of the opcode, the
 *  MODR/M byte, the SIB byte and any displacement or immediate value.
 *  Any instruction that references memory uses a MODR/M byte.
 */
#include "global.h"
#include "x86-asm.h"
#include "x86-prefix.h"
#include "x86-hash.h"
#include "utilities.h"
#include "x86-util.h"
#include "x86-dis.h"
#include "x86-table-defs.h"
#include "x86-op.h"
#include "x86-registers.h"
#include "x86-classes.h"
#include "sysdep.h"
#include "x86-histogram.h"
#include "hash.h"
#include "upm.h"
#include "flowgraph.h"

#define NOP 0x90

/* Puts the instruction prefixes in the instruction stream */
static int WritePrefixes(Instruction *instr, unsigned char *buffer);

/* Puts the opcode(s) in the instruction stream */
static int WriteOpcode(Opcode *opcode, unsigned char *buffer,
    Instruction *instr);

/* Puts the MODRM byte in the instruction stream */
static int WriteMODRM(unsigned char *buffer);

/* Puts the Scale-Index-Base byte in the instruction stream */
static int WriteSIB(unsigned char *buffer);

/* Puts the displacement in the instruction stream */
static int WriteDisplacement(unsigned char *buffer);

/* Puts the immediate value in the instruction stream */
static int WriteImmediate(unsigned char *buffer);

/* Various operations used for handling different operand types. */
static void HandleOperandE(Operand *op, int bytemode, int dflag);
static void HandleOperandJ(Operand *op, int bytemode, int dflag);
static void HandleOperandG(Operand *op, int bytemode, int dflag);
static void HandleOperandsI(Operand *op, int bytemode, int dflag);
static void HandleOperandOFF(Operand *op, int bytemode, int dflag);
static void HandleOperandDIR(Operand *op, int bytemode, int dflag);
static void HandleOperandI(Operand *op, int bytemode, int dflag);
static void HandleOperandSTi(Operand *op, int bytemode, int dflag);
static void HandleOperandC(Operand *op, int bytemode, int dflag);
static void HandleOperandSEG(Operand *op, int bytemode, int dflag);
static void HandleOperandDBG(Operand *op, int bytemode, int dflag);

/* Takes a register and converts its effective register address.
 * (See 2-6 of Vol 2.) */
static int RegisterToRM(int the_reg);

/*
 *  Given an operand, this function will dispatch the operand
 *  to one of its helper routines (which have similar names, see above).
 */
static void HandleOperand(Operand *operand, op_rtn routine, int bytemode,
    Instruction *i, int dflag);

/*
 *  GLOBAL VARIABLES
 *  HASMODRM: true if the current instruction has a modrm byte
 *  HASSIB:   true if the current instruction has an SIB byte
 *  MOD:      The value of the current MOD bits in the MODR/M byte
 *  REG:      The value of the current REG bits in the MODR/M byte
 *  RM:       The value of the current R/M bits in the MODR/M byte
 *  SCALE:    The scale factor of the current instruction using an SIB byte
 *  INDEX:    The index register of the current instruction using an SIB byte
 *  BASE:     The base register of the current instruction using an SIB byte
 *  DISPLACEMENT: The displacement bytes for the current instruction.
 *  IMMEDIATE:The immediate field for the current instruction
 *  displacementSize: The number of bytes the displacement will take up.
 *  immediateSize:    The number of bytes the immediate field will take up.
 *  displacementLoc:  The location of the displacement in the instruction.
 *  immediateLoc:     The location of the immediate val in the instruction.
 */
static bool DISPLACEMENT_GROWTH_BYTES;
static bool HASMODRM, HASSIB;
static int MOD, Reg, RM, SCALE, INDEX, BASE, DISPLACEMENT, IMMEDIATE;
static int displacementSize, immediateSize;
static int displacementLoc, immediateLoc;



/*
 *  xInstrAssemble(buffer,instr) --
 *
 *  Assemble the instruction "instr" into its machine code
 *  representation and write it into the memory location "buffer".
 */
int xInstrAssemble(unsigned char *buffer, Instruction *instr) {
    Operand *opp1, *opp2, *opp3;
    int bytemode1, bytemode2, bytemode3;
    op_rtn op1, op2, op3;
    Opcode *opcode;
    int dflag;
    int length;

    // if the instruction has databytes associated with it,
    // copy the data and return the length.
    if (INSTR_HAS_FLAG(instr, INSTR_DATA_BYTE)) {
	int i = 0;
	for (; i < instr->i_length; i++) {
	    buffer[i] = instr->i_literal_data[i];
	}
	return instr->i_length;
    }

    // if the instruction is marked as having literal bytes associated
    // with it then we will just return those bytes instead of
    // trying to disassemble it
    if (INSTR_HAS_FLAG(instr, INSTR_LITERAL_BYTES_FLAG)) {
	int i = 0;
	for (; i < instr->i_length; i++) {
	    buffer[i] = instr->i_literal_data[i];
	}

	return instr->i_length;
    }					// (cml)

    opp1 = opp2 = opp3 = NULL;
    bytemode1 = bytemode2 = bytemode3 = -1;
    op1 = op2 = op3 = NULL;
    HASMODRM = HASSIB = false;
    MOD = RM = Reg = SCALE = INDEX = BASE = length = 0;
    displacementSize = immediateSize = -1;
    DISPLACEMENT_GROWTH_BYTES = 0;

    /* If dflag is set then the operands are 32 bit, otherwise 16 bit.
     * Most common case is 32 */
    dflag = 1;
    if (instr->i_prefixes & PREFIX_OPERAND) {
	dflag = 0;
    }

    if (INSTR_HAS_FLAG(instr, INSTR_SWITCH_CONDITION_FLAG)) {
	InstrSwitchCondition(instr);
	INSTR_REMOVE_FLAG(instr, INSTR_SWITCH_CONDITION_FLAG);
    }

    /* Get the operands if they exist */
    if (INSTR_HAS_FLAG(instr, DEST_FLAG)) {
	opp1 = instr->i_dest;
    }
    if (INSTR_HAS_FLAG(instr, SOURCE1_FLAG)) {
	opp2 = instr->i_src;
    }
    if (INSTR_HAS_FLAG(instr, SOURCE2_FLAG)) {
	opp3 = instr->i_src2;
    }

    /* Bytemode is the size of the operands, e.g; 0, 8, 16 or 32 bits */
    bytemode1 = GetByteMode(opp1, dflag);
    bytemode2 = GetByteMode(opp2, dflag);
    bytemode3 = GetByteMode(opp3, dflag);

    /* Determine which operand-handling routines are applicable */
    if (opp1) {
	op1 = GetOpRoutine(instr, opp1, bytemode1, bytemode2);
    }
    if (opp2) {
	op2 = GetOpRoutine(instr, opp2, bytemode2, bytemode1);
    }
    if (opp3) {
	op3 = GetOpRoutine(instr, opp3, bytemode3, bytemode1);
    }

    opcode = PatternMatchOpcode(instr->i_class, bytemode1, op1,
	bytemode2, op2, bytemode3, op3, instr);

    if (opcode == NULL) {
	ReportInstrError("Could not find opcode", instr);
    }

    /* Call the appropriate routines for all the operands */
    if (opp1) {
	HandleOperand(opp1, op1, bytemode1, instr, dflag);
    }
    if (opp2) {
	HandleOperand(opp2, op2, bytemode2, instr, dflag);
    }
    if (opp3) {
	HandleOperand(opp3, op3, bytemode3, instr, dflag);
    }

    /* Presumably we have all the information needed to
     * write out the instruction. */
    length += WritePrefixes(instr, &buffer[length]);
    length += WriteOpcode(opcode, &buffer[length], instr);
    if (HASMODRM) {
	length += WriteMODRM(&buffer[length]);
    }
    if (HASSIB) {
	length += WriteSIB(&buffer[length]);
    }
    if (displacementSize != -1) {
//       if (DISPLACEMENT_GROWTH_BYTES > 0)
//          UpdateUPMWithGrowth(instr, length, DISPLACEMENT_GROWTH_BYTES);
	displacementLoc = length;
	length += WriteDisplacement(&buffer[length]);
    }
    if (immediateSize != -1) {
	immediateLoc = length;
	length += WriteImmediate(&buffer[length]);
    }

    return length;
}



/*
 *  OperandOnlyUsesBaseRegister(op) --
 *
 *  Returns true if the given operand ONLY uses a base register.
 */
bool OperandOnlyUsesBaseRegister(Operand *op) {
    if (op == NULL) {
	return false;
    }
    if (OP_TYPE(op, OPERAND_INDEX_FLAG)) {
	return false;
    }
    if (OP_TYPE(op, OPERAND_SCALE_FLAG)) {
	return false;
    }
    if (OP_TYPE(op, OPERAND_DISP_A_FLAG)) {
	return false;
    }
    if (OP_TYPE(op, OPERAND_DISP_R_FLAG)) {
	return false;
    }
    if (OP_TYPE(op, OPERAND_IMM_FLAG)) {
	return false;
    }
    if (OP_TYPE(op, OPERAND_BASE_FLAG)) {
	return true;
    }
    return false;
}



/*
 *  PatternMatchOpcode(class,bytemode1,op1,bytemode2,op2,bytemode3,op3,instr) --
 *
 *  Given an instruction class and the bytemode and routines used by its
 *  operands, PatternMatchOpcode will scan through the opcode tables
 *  and find the right opcode to use.
 */
Opcode *PatternMatchOpcode(UINT16 class, int bytemode1, op_rtn op1,
    int bytemode2, op_rtn op2, int bytemode3, op_rtn op3, Instruction *instr) {
    Opcode *opcode;
    OpcodeHash *oHash;
    int reg;
    UINT8 primary;

    reg = primary = 0;
    opcode = zalloc(sizeof(Opcode));


    oHash = X86RetrieveOpcode(
	instr, bytemode1, op1, bytemode2, op2, bytemode3, op3);

    if (oHash == NULL) {
	free(opcode);
	return NULL;
    }

    opcode->o_h = oHash;

    /*
     * We have 6 possible cases, one for each table in x86-table.h.
     */

    if (oHash->oh_flags & OPCODE_FLAG_DIS386) {
	primary = oHash->oh_index1 * 8 + oHash->oh_index2;
	opcode->o_primary = primary;
	opcode->o_need_modrm = onebyte_has_modrm[primary];
	opcode->o_table = OPCODE_TABLE_DIS386;
	if (!HASMODRM) {
	    HASMODRM = onebyte_has_modrm[primary];
	}
	return opcode;
    }

    if (oHash->oh_flags & OPCODE_FLAG_DIS3862BYTE) {
	primary = oHash->oh_index1 * 8 + oHash->oh_index2;
	opcode->o_primary = OPCODE_EXTENSION;
	opcode->o_secondary = primary;
	opcode->o_need_secondary = true;
	opcode->o_table = OPCODE_TABLE_DIS386_TWOBYTE;
	opcode->o_need_modrm = twobyte_has_modrm[primary];
	if (!HASMODRM) {
	    HASMODRM = twobyte_has_modrm[primary];
	}
	return opcode;
    }

    if (oHash->oh_flags & OPCODE_FLAG_GRPS) {
	opcode->o_need_modrm = true;
	HASMODRM = true;
	opcode->o_reg = Reg = oHash->oh_index2;
	opcode->o_table = OPCODE_TABLE_GRPS;
	if (oHash->oh_flags & OPCODE_FLAG_TWOBYTE) {
	    opcode->o_need_secondary = true;
	    opcode->o_primary = OPCODE_EXTENSION;
	    opcode->o_secondary = oHash->oh_index3;
	} else {
	    opcode->o_need_secondary = false;
	    opcode->o_primary = oHash->oh_index3;
	}
	return opcode;
    }

    /*
     * We must have a floating point instruction at this point.
     */

    HASMODRM = true;	// All floating point instructions have a MODR/M byte
    opcode->o_need_modrm = true;

    if (oHash->oh_flags & OPCODE_FLAG_FLOATREG) {
	opcode->o_primary = 0xd8 + oHash->oh_index1;
	opcode->o_reg = Reg = oHash->oh_index2;
	opcode->o_mod = 3;
	opcode->o_table = OPCODE_TABLE_FLOAT_REG;
	MOD = 3;
	return opcode;
    }

    if (oHash->oh_flags & OPCODE_FLAG_FLOATMEM) {
	opcode->o_primary = 0xd8 + oHash->oh_index1;
	opcode->o_reg = Reg = oHash->oh_index2;
	opcode->o_table = OPCODE_TABLE_FLOAT_MEM;
	return opcode;
    }

    if (oHash->oh_flags & OPCODE_FLAG_FGRPS) {
	RM = oHash->oh_index2;
	opcode->o_primary = 0xd8 + oHash->oh_index3 / 8;
	opcode->o_mod = MOD = 3;
	opcode->o_table = OPCODE_TABLE_FGRPS;
	Reg = opcode->o_reg = oHash->oh_index3 % 8;
	return opcode;
    }

    WARNING("PatternMatchOpcode returing NULL");
    return NULL;
}



/*
 *  SpitOutOpcodeData(opcode,instr) --
 *
 *  Used primarily for debugging.
 *  Prints the opcodes and MODRM byte to standard out.
 */
void SpitOutOpcodeData(Opcode *opcode, Instruction *instr) {

    printf("[0x%8.8x] Primary Opcode: %2.2x   ",
	instr->i_address, opcode->o_primary);
    if (opcode->o_need_secondary) {
	printf("Secondary Opcode: %2.2x   ", opcode->o_secondary);
    }
    if (opcode->o_need_modrm) {
	printf("MODR/M(reg): %2.2x", opcode->o_reg);
    }
    printf("\n");

    return;
}



/*
 *  WritePrefixes(instr,buffer) --
 *
 *  Writes all prefixes of an instruction into the buffer.
 */
static int WritePrefixes(Instruction *instr, unsigned char *buffer) {
    int offset = 0;

    if (instr->i_prefixes & PREFIX_FWAIT) {
	buffer[offset++] = X86_PREFIX_FWAIT;
    }
    if (instr->i_prefixes & PREFIX_REPZ) {
	buffer[offset++] = X86_PREFIX_REPZ;
    }
    if (instr->i_prefixes & PREFIX_REPNZ) {
	buffer[offset++] = X86_PREFIX_REPNZ;
    }
    if (instr->i_prefixes & PREFIX_LOCK) {
	buffer[offset++] = X86_PREFIX_LOCK;
    }
    if (instr->i_prefixes & PREFIX_CS) {
	buffer[offset++] = X86_PREFIX_CS;
    }
    if (instr->i_prefixes & PREFIX_SS) {
	buffer[offset++] = X86_PREFIX_SS;
    }
    if (instr->i_prefixes & PREFIX_DS) {
	buffer[offset++] = X86_PREFIX_DS;
    }
    if (instr->i_prefixes & PREFIX_ES) {
	buffer[offset++] = X86_PREFIX_ES;
    }
    if (instr->i_prefixes & PREFIX_FS) {
	buffer[offset++] = X86_PREFIX_FS;
    }
    if (instr->i_prefixes & PREFIX_GS) {
	buffer[offset++] = X86_PREFIX_GS;
    }
    if (instr->i_prefixes & PREFIX_OPERAND) {
	buffer[offset++] = X86_PREFIX_OPERAND;
    }
    if (instr->i_prefixes & PREFIX_ADDRESS) {
	buffer[offset++] = X86_PREFIX_ADDRESS;
    }
    return offset;
}



/*
 *  WriteOpcode(opcode,buffer,instr) --
 *
 *  Writes the primary and maybe the secondary opcode to the memory
 *  that represents what the binary will look like.
 */
static int WriteOpcode(Opcode *opcode, unsigned char *buffer,
    Instruction *instr) {
    int offset = 0;
    long long weight = 0;

    if (instr->i_Bblock != NULL) {
	weight = instr->i_Bblock->b_weight;
    }

    /*
     *         For weighted histograms we will keep track of assembled opcodes
     */

    if (WANT_DEBUG(DB_WGTHG)) {		// if GenerateWeightedHistograms
	switch (opcode->o_table) {
	case OPCODE_TABLE_DIS386:
	    onebyteOpcodesWeighted[opcode->o_primary] =
		onebyteOpcodesWeighted[opcode->o_primary] + weight;
	    break;
	case OPCODE_TABLE_DIS386_TWOBYTE:
	    twobyteOpcodesWeighted[opcode->o_secondary] =
		twobyteOpcodesWeighted[opcode->o_secondary] + weight;
	    break;
	case OPCODE_TABLE_GRPS:
	    grpsOpcodesWeighted[opcode->o_h->oh_index1 * 8 +
		opcode->o_h->oh_index2] =
		grpsOpcodesWeighted[opcode->o_h->oh_index1 *
		8 + opcode->o_h->oh_index2] + weight;
	    break;
	case OPCODE_TABLE_FGRPS:
	    fgrpsOpcodesWeighted[opcode->o_h->oh_index1 * 8 +
		opcode->o_h->oh_index2] =
		fgrpsOpcodesWeighted[opcode->o_h->oh_index1 *
		8 + opcode->o_h->oh_index2] + weight;
	    break;
	case OPCODE_TABLE_FLOAT_MEM:
	    floatmemOpcodesWeighted[(opcode->o_primary - 0xd8) * 8 +
		opcode->o_reg] =
		floatmemOpcodesWeighted[(opcode->o_primary - 0xd8) * 8 +
		opcode->o_reg] + weight;
	    break;
	case OPCODE_TABLE_FLOAT_REG:
	    floatregOpcodesWeighted[(opcode->o_primary - 0xd8) * 8 +
		opcode->o_reg] =
		floatregOpcodesWeighted[(opcode->o_primary - 0xd8) * 8 +
		opcode->o_reg] + weight;
	    break;
	default:
	    REPORT(RP_DETAIL,
		WITHFUNC("Could not discern which table opcode came from"));
	}
    }

    buffer[offset++] = opcode->o_primary;

    if (opcode->o_need_secondary) {
	buffer[offset++] = opcode->o_secondary;
    }

    return offset;
}



/*
 *  HandleOperand(op,routine,bytemode,instr,dflag) --
 *
 *  This function takes an operand and dispatches it to another routine designed
 *  specifically for operands of its type.
 */
static void HandleOperand(Operand *op, op_rtn routine, int bytemode,
    Instruction *instr, int dflag) {

    if (routine == OP_E) {
	HandleOperandE(op, bytemode, dflag);
    } else if (routine == OP_J) {
	HandleOperandJ(op, bytemode, dflag);
    } else if (routine == OP_G) {
	HandleOperandG(op, bytemode, dflag);
    } else if (routine == OP_sI) {
	HandleOperandsI(op, bytemode, dflag);
    } else if (routine == OP_OFF) {
	HandleOperandOFF(op, bytemode, dflag);
    } else if (routine == OP_DIR) {
	HandleOperandDIR(op, bytemode, dflag);
    } else if (routine == OP_indirE) {
	HandleOperandE(op, bytemode, dflag);
    } else if (routine == OP_I) {
	HandleOperandI(op, bytemode, dflag);
    } else if (routine == OP_REG) {
	return;
    } else if (routine == OP_ESDI) {
	return;
    } else if (routine == OP_DSSI) {
	return;
    } else if (routine == OP_ST) {
	return;
    } else if (routine == OP_STi) {
	HandleOperandSTi(op, bytemode, dflag);
    } else if(routine == OP_SEG) {
	HandleOperandSEG(op, bytemode, dflag);
    } else if(routine == OP_C) {
	HandleOperandC(op, bytemode, dflag);
    } else if(routine == OP_D) {
	HandleOperandDBG(op, bytemode, dflag);
    } else {
	ReportInstrError("Could not find operand routine", instr);
    }
    return;
}



/*
 *  HandleOperandC(op,bytemode,dflag) -- handle special-case CR registers
 */
static void HandleOperandC(Operand *op, int bytemode, int dflag) {
    Reg = op->a_base_reg - reg_cr0;
    REPORT(RP_TRIVIAL, "Handled CR Operand --- by setting REG to %d", Reg);
    return;
}


/*
 *  HandleOperandSEG(op,bytemode,dflag) -- handle segment registers
 */
static void HandleOperandSEG(Operand *op, int bytemode, int dflag) {
    //RM = RegisterToRM(op->a_base_reg);
    Reg = op->a_base_reg - reg_es;
    REPORT(RP_TRIVIAL, "Handled Operand SEG --- by setting REG to %d", Reg);
    return;
}



/*
 *  HandleOperandDBG(op,bytemode,dflag) -- handle debug registers
 */
static void HandleOperandDBG(Operand *op, int bytemode, int dflag) {
    RM = RegisterToRM(op->a_base_reg);
    Reg = op->a_base_reg;
    REPORT(RP_TRIVIAL, "Handled CR Operand --- by setting REG to %d", Reg);
    return;
}



/*
 *  HandleOperandE(op,bytemode,dflag) --
 *
 *  This operand handling function is the parallel of the function OP_E used in
 *  the disassembly process. It handles all operands that use MOD/RM bytes in
 *  their addressing modes.
 */
static void HandleOperandE(Operand *op, int bytemode, int dflag) {

    /*
     * If an operand only uses a base register then we don't have to worry about
     * displacements and such.  All we have to do is set to the MOD bits to 11B
     * (see Chapter 2, Vol 2 of the intel manual) and encode the register
     */

    if (OperandOnlyUsesBaseRegister(op)) {
	if (OP_TYPE(op, OPERAND_REGISTER_MODRM)) {
	    MOD = 3;
	} else {
	    MOD = 0;
	}
	RM = RegisterToRM(op->a_base_reg);
	return;
    }

    /*
     * We next see if the operand has an SIB byte following the MODRM byte.  If
     * so, we get the SCALE value and encode it (again, the SIB table provides
     * this information. We also encode the scaled index and the base register.
     * A special case is when there is no index to be scaled (and presumably the
     * scale factor is 0).
     */

    if (OperandHasSIB(op)) {
	HASSIB = true;
	RM = 4;
	SCALE = op->a_scale;

	/* Encode the scale. The encoding is neat because it's log base 2
	 * of the actual scale value. */
	switch (SCALE) {
	case 0:
	case 1:
	    SCALE = 0;
	    break;
	case 2:
	    SCALE = 1;
	    break;
	case 4:
	    SCALE = 2;
	    break;
	case 8:
	    SCALE = 3;
	    break;
	default:
	    SCALE = 0;
	    break;
	}

	/* Encode the index and base registers. They use the same values for
	 * encoding that the MODR/M tables use, so we can use the MOD/RM
	 * function RegisterToRM */
	if (op->a_index_reg) {
	    INDEX = RegisterToRM(op->a_index_reg);
	} else {
	    INDEX = 4;
	}

	if (OP_TYPE(op, OPERAND_BASE_FLAG)) {
	    BASE = RegisterToRM(op->a_base_reg);
	} else {
	    BASE = 5;
	}

	/* An operand can have the form disp32[--] where [--] denotes an SIB
	 * byte following the MODRM byte. disp16[--] and disp8[--] are also
	 * valid. In these cases, we have to worry about a displacement
	 * in addition to the Scale-Index-Base. */

	if (OP_TYPE(op, OPERAND_DISP_R_FLAG)) {
	    MOD = 1;
	    displacementSize = 8;
	    DISPLACEMENT = op->a_displacement;
	    if ((DISPLACEMENT < CHAR_MIN) || (DISPLACEMENT > CHAR_MAX)) {
		MOD = 2;
		displacementSize = 32;
		DISPLACEMENT_GROWTH_BYTES = 3;
		OPERAND_SET_FLAG(op, OPERAND_DISP_A_FLAG);
		OPERAND_REMOVE_FLAG(op, OPERAND_DISP_R_FLAG);
	    }
	} else if ((OP_TYPE(op, OPERAND_DISP_A_FLAG)) &&
	    ((BASE != 5) || (op->a_base_reg == (reg_bp | reg_32)))) {
	    MOD = 2;
	    displacementSize = 32;
	    DISPLACEMENT = op->a_displacement;
	} else if ((OP_TYPE(op, OPERAND_DISP_A_FLAG)) && (BASE == 5)) {
	    MOD = 0;
	    displacementSize = 32;
	    DISPLACEMENT = op->a_displacement;
	} else {
	    MOD = 0;
	}

	return;
    }

    /*
     * If the operand provides no base register then it must be just a
     * displacement. e.g. jmp 0x12341234.
     */

    if (!OP_TYPE(op, OPERAND_BASE_FLAG)) {
	DISPLACEMENT = op->a_displacement;
	displacementSize = 32;
	if (HASMODRM) {
	    MOD = 0;
	    RM = 5;
	}
	return;
    }

    /*
     * If the operand provides no displacement (including immediate value!)
     * then the encoding is simply a register
     */

    if (!OP_TYPE(op,OPERAND_DISP_R_FLAG) && !OP_TYPE(op,OPERAND_DISP_A_FLAG)) {
	MOD = 0;
	RM = RegisterToRM(op->a_base_reg);
	return;
    }

    /*
     * The following are the two most general cases for encoding.  When a
     * relative displacement is provided we will assume it's an 8 bit field
     * and encode the the byte from the table.  If the displacement is
     * absolute, the only change we make is the table section
     * (MOD 2 instead of MOD 1)
     */

    if (OP_TYPE(op, OPERAND_DISP_R_FLAG)) {
	displacementSize = 8;
	DISPLACEMENT = op->a_displacement;
	MOD = 1;
	RM = RegisterToRM(op->a_base_reg);
	if ((DISPLACEMENT < CHAR_MIN) || (DISPLACEMENT > CHAR_MAX)) {
	    MOD = 2;
	    displacementSize = 32;
	    DISPLACEMENT_GROWTH_BYTES = 3;
	    OPERAND_SET_FLAG(op, OPERAND_DISP_A_FLAG);
	    OPERAND_REMOVE_FLAG(op, OPERAND_DISP_R_FLAG);
	}
	return;
    }

    if (OP_TYPE(op, OPERAND_DISP_A_FLAG)) {
	displacementSize = 32;
	DISPLACEMENT = op->a_displacement;
	MOD = 2;
	RM = RegisterToRM(op->a_base_reg);
	return;
    }

    MALFUNCTION("Could not resolve OP_E operand.");
}



/*
 *  RegisterToRM(the_reg) --
 *
 *  Given a register (see x86-registers.h), this function converts it into the
 *  appropriate encoding used for bits 0, 1 and 2 of the MODRM byte.
 */
static int RegisterToRM(int the_reg) {
    switch (the_reg) {
    case reg_a | reg_32:
    case reg_a | reg_16:
    case reg_a | reg_8l:
	return 0;
	break;
    case reg_c | reg_32:
    case reg_c | reg_16:
    case reg_c | reg_8l:
	return 1;
	break;
    case reg_d | reg_32:
    case reg_d | reg_16:
    case reg_d | reg_8l:
	return 2;
	break;
    case reg_b | reg_32:
    case reg_b | reg_16:
    case reg_b | reg_8l:
	return 3;
	break;
    case reg_sp | reg_32:
    case reg_sp | reg_16:
    case reg_a | reg_8h:
	return 4;
	break;
    case reg_bp | reg_32:
    case reg_bp | reg_16:
    case reg_c | reg_8h:
	return 5;
	break;
    case reg_si | reg_32:
    case reg_si | reg_16:
    case reg_d | reg_8h:
	return 6;
	break;
    case reg_di | reg_32:
    case reg_di | reg_16:
    case reg_b | reg_8h:
	return 7;
	break;
    default:
	return -1;
    }
    return -1;
}



/*
 *  HandleOperandJ(op,bytemode,dflag) --
 *
 *  Handles operands that are targets for jump instructions.  The parallel
 *  of this function used in the disassembly processes is OP_J.
 */
static void HandleOperandJ(Operand *op, int bytemode, int dflag) {
    CASSERT(displacementSize == -1, "Already a displacement!");

    if (bytemode == v_mode) {
	if (dflag) {
	    displacementSize = 32;
	} else {
	    displacementSize = 16;
	}
    }
    if (bytemode == w_mode) {
	displacementSize = 16;
    }
    if (bytemode == b_mode) {
	displacementSize = 8;
    }

    DISPLACEMENT = op->a_displacement;
    return;
}



/*
 *  HandleOperandG(op,bytemode,dflag) --
 *
 *  Handles operands that are just registers.  Often this is the second op
 *  routine used when OP_E is used for the first operand. The parallel of
 *  this function used in the disassembly process is OP_G.
 */
static void HandleOperandG(Operand *op, int bytemode, int dflag) {
    Reg = RegisterToRM(op->a_base_reg);
    //RM = RegisterToRM(op->a_base_reg);
    return;
}



/*
 *  HandleOperandsI(op,bytemode,dflag) --
 *
 *  Takes care of sign-extended immediate values.  Since we already have the
 *  immediate values in their sign-extended form, we can just chop off the bits
 *  that aren't needed when we write out the operand. The parallel of this
 *  function used in the disassembly process is OP_sI.
 */
static void HandleOperandsI(Operand *op, int bytemode, int dflag) {
    CASSERT(OP_TYPE(op, OPERAND_IMM_FLAG), "Operand lacks IMM flag.");
    if (bytemode == b_mode) {
	immediateSize = 8;
    } else if (bytemode == w_mode) {
	immediateSize = 16;
    } else {				/* v_mode */
	if (dflag) {
	    immediateSize = 32;
	} else {
	    immediateSize = 16;
	}
    }
    IMMEDIATE = op->a_displacement;
    return;
}



/*
 *  HandleOperandOFF(op,bytemode,dflag) --
 *
 *  OP_OFF is the parallel of this function that we use when disassembling.
 *  Basically it handles relative displacements from call instructions
 *  and the like.
 */
static void HandleOperandOFF(Operand *op, int bytemode, int dflag) {
    displacementSize = 32;
    DISPLACEMENT = op->a_displacement;
    return;
}



/*
 *  HandleOperandDIR(op,bytemode,dflag) --
 *
 *  OP_DIR is the parallel of this function used in the disassembly process. It
 *  is intended to be used for instructions that select their own segment, but
 *  rarely is that used if ever. We assume that we're dealing with a 32 bit
 *  relative offset.
 */
static void HandleOperandDIR(Operand *op, int bytemode, int dflag) {
    displacementSize = 32;
    DISPLACEMENT = op->a_displacement;
    return;
}



/*
 *  HandleOperandI(op,bytemode,dflag) --
 *
 *  Takes care of operands specifying an immediate value.
 *  The parallel of this function used in the disassembly process is OP_I.
 */
static void HandleOperandI(Operand *op, int bytemode, int dflag) {
    if (bytemode == b_mode) {
	immediateSize = 8;
    } else if (bytemode == w_mode) {
	immediateSize = 16;
    } else {				/* v_mode */
	if (dflag) {
	    immediateSize = 32;
	} else {
	    immediateSize = 16;
	}
    }
    IMMEDIATE = op->a_displacement;
    return;
}



/*
 *  WriteMODRM(buffer) --
 *
 *  Writes the MODR/M byte to the memory area that will be used to create
 *  the segment in the binary.
 */
static int WriteMODRM(unsigned char *buffer) {
    UINT8 modrm = 0;
    int offset = 0;

    modrm |= (MOD << 6);
    modrm |= (Reg << 3);
    modrm |= RM;
    MOD = Reg = RM = 0;
    HASMODRM = false;
    buffer[offset++] = modrm;

    return offset;
}



/*
 *  WriteSIB(buffer) --
 *
 *  Writes the SIB byte to the memory area that will be used to create
 *  the segment in the binary.
 */
static int WriteSIB(unsigned char *buffer) {
    int offset = 0;
    UINT8 sib = 0;

    sib |= (SCALE << 6);
    sib |= (INDEX << 3);
    sib |= BASE;
    SCALE = INDEX = BASE = 0;
    HASSIB = false;
    buffer[offset++] = sib;

    return offset;
}



/*
 *  WriteDisplacement(buffer) --
 *
 *  Writes out the displacement used by the instruction.  This function
 *  assumes that the architecture is consistent Little Endian.
 */
static int WriteDisplacement(unsigned char *buffer) {
    int offset = 0;
    long long minDisp = 0;
    long long maxDisp = 0;

    if (displacementSize == 8) {
	minDisp = CHAR_MIN;
	maxDisp = CHAR_MAX;
	buffer[offset++] = (UINT8) DISPLACEMENT;
    } else if (displacementSize == 16) {
	minDisp = SHRT_MIN;
	maxDisp = SHRT_MAX;
	storeLE2(&buffer[offset], (INT32) DISPLACEMENT);
	offset = offset + 2;
    } else if (displacementSize == 32) {
	minDisp = INT_MIN;
	maxDisp = INT_MAX;
	storeLE(&buffer[offset], (INT32) DISPLACEMENT);
	offset = offset + 4;
    }

    if ((DISPLACEMENT < minDisp) || (DISPLACEMENT > maxDisp)) {
	WARNING("[WriteDisplacement] Could not encode %d in %d bits.",
	    DISPLACEMENT, displacementSize);
    }

    displacementSize = -1;
    DISPLACEMENT = 0;

    return offset;
}



/*
 *  WriteImmediate(buffer) --
 *
 *  Writes out the immediate value used by the instruction.  This function
 *  assumes that the architecture is consistent Little Endian.
 */
static int WriteImmediate(unsigned char *buffer) {
    int offset = 0;

    if (immediateSize == 8) {
	buffer[offset++] = (UINT8) IMMEDIATE;
    } else if (immediateSize == 16) {
	storeLE2(&buffer[offset], (INT32) IMMEDIATE);
	offset = offset + 2;
    } else if (immediateSize == 32) {
	storeLE(&buffer[offset], (INT32) IMMEDIATE);
	offset = offset + 4;
    }

    immediateSize = -1;
    IMMEDIATE = 0;
    return offset;
}



/*
 *  HandleOperandSTi(op,bytemode,dflag) --
 *
 *  HandleOperandSTi is used for floating-point instructions that reference
 *  registers offset from the top of the stack.  The encoding is written
 *  into the R/M bits of the MODR/M field and is simply the offset from the
 *  top of the stack.
 */
void HandleOperandSTi(Operand *op, int bytemode, int dflag) {
    MOD = 11;
    switch (op->a_base_reg) {
	case reg_st0:  RM = 0;  break;
	case reg_st1:  RM = 1;  break;
	case reg_st2:  RM = 2;  break;
	case reg_st3:  RM = 3;  break;
	case reg_st4:  RM = 4;  break;
	case reg_st5:  RM = 5;  break;
	case reg_st6:  RM = 6;  break;
	case reg_st7:  RM = 7;  break;
	default:
	    MALFUNCTION("Could not encode base register in STi handler.");
    }
}



/*
 *  OpcodeEntryMatches(
 *  entry,op1,bytemode1,op2,bytemode2,op3,bytemode3,type,class) --
 *
 *  Returns true if the given information about an instruction
 *  matches the given opcode-table entry.
 */
bool OpcodeEntryMatches(X86op *entry, op_rtn op1, int bytemode1, op_rtn op2,
int bytemode2, op_rtn op3, int bytemode3, int type, int class) {

    if (entry->x_class != class) {
	return false;
    }
    if (entry->x_type != type) {
	return false;
    }
    if (entry->x_bytemode1 != bytemode1) {
	return false;
    }
    if (entry->x_bytemode2 != bytemode2) {
	return false;
    }
    if (entry->x_bytemode3 != bytemode3) {
	return false;
    }
    if (entry->x_op1 != op1) {
	return false;
    }
    if (entry->x_op2 != op2) {
	return false;
    }
    if (entry->x_op3 != op3) {
	return false;
    }

    return true;
}



/*
 *  xCalculateInstrLength(iterator) --
 *
 *  Calculates the length of the instruction by assembling it.
 */
int xCalculateInstrLength(Instruction *iterator) {
    unsigned char *buffer = alloc(iterator->i_length);
    return InstrAssemble(buffer, iterator);
}



/*
 *  xUpdateInstrLength(instr) --
 *
 *  Sets the length on 'instr' by pattern matching the opcode in the
 *  opcode table, and determining operand lengths.
 */
void xUpdateInstrLength(Instruction *instr) {
    instr->i_length = xCalculateInstrLength(instr);
    return;
}



/*
 *  InstrSwitchCondition(instr) --
 *
 *  Switches the class of the given instruction to change the jump
 *  condition. e.g. a jump on '>' will become a jump on '<=', or '!>'
 */
void InstrSwitchCondition(Instruction *instr) {
    int new_class;
    switch (instr->i_class) {
    case CLASS_JA:
	new_class = CLASS_JNA;
	break;
    case CLASS_JAE:
	new_class = CLASS_JNAE;
	break;
    case CLASS_JB:
	new_class = CLASS_JNB;
	break;
    case CLASS_JBE:
	new_class = CLASS_JNBE;
	break;
    case CLASS_JCXZ:
	new_class = MAGIC_NUMBER;
	break;
    case CLASS_JE:
	new_class = CLASS_JNE;
	break;
    case CLASS_JG:
	new_class = CLASS_JNG;
	break;
    case CLASS_JGE:
	new_class = CLASS_JNGE;
	break;
    case CLASS_JL:
	new_class = CLASS_JNL;
	break;
    case CLASS_JLE:
	new_class = CLASS_JNLE;
	break;
    case CLASS_JNE:
	new_class = CLASS_JE;
	break;
    case CLASS_JNO:
	new_class = CLASS_JO;
	break;
    case CLASS_JNP:
	new_class = CLASS_JP;
	break;
    case CLASS_JNS:
	new_class = CLASS_JS;
	break;
    case CLASS_JO:
	new_class = CLASS_JNO;
	break;
    case CLASS_JP:
	new_class = CLASS_JNP;
	break;
    case CLASS_JS:
	new_class = CLASS_JNS;
	break;
    default:
	new_class = MAGIC_NUMBER;
	break;
    }

    if (new_class == MAGIC_NUMBER) {
	MALFUNCTION("[InstrSwitchCondition] Could not switch condition");
    }

    instr->i_class = new_class;

    return;
}



/*
 *  InstrGetOpcodeEntry(instr) --
 *
 *  Returns the opcode entry corresponding to what 'instr' would use when
 *  assembled.
 */
X86op *InstrGetOpcodeEntry(Instruction *instr) {
    Operand *opp1, *opp2, *opp3;
    int bytemode1, bytemode2, bytemode3;
    op_rtn op1, op2, op3;
    OpcodeHash *oHash;

    opp1 = opp2 = opp3 = NULL;
    bytemode1 = bytemode2 = bytemode3 = -1;
    op1 = op2 = op3 = NULL;

    /* Get the operands if they exist */
    if (INSTR_HAS_FLAG(instr, DEST_FLAG)) {
	opp1 = instr->i_dest;
    }
    if (INSTR_HAS_FLAG(instr, SOURCE1_FLAG)) {
	opp2 = instr->i_src;
    }
    if (INSTR_HAS_FLAG(instr, SOURCE2_FLAG)) {
	opp3 = instr->i_src2;
    }

    /* Bytemode is the size of the operands, e.g; 0, 8, 16 or 32 bits */
    bytemode1 = GetByteMode(opp1, (instr->i_prefixes & PREFIX_OPERAND) ? 0 : 1);
    bytemode2 = GetByteMode(opp2, (instr->i_prefixes & PREFIX_OPERAND) ? 0 : 1);
    bytemode3 = GetByteMode(opp3, (instr->i_prefixes & PREFIX_OPERAND) ? 0 : 1);

    /* Determine which operand-handling routines are applicable */
    if (opp1) {
	op1 = GetOpRoutine(instr, opp1, bytemode1, bytemode2);
    }
    if (opp2) {
	op2 = GetOpRoutine(instr, opp2, bytemode2, bytemode1);
    }
    if (opp3) {
	op3 = GetOpRoutine(instr, opp3, bytemode3, bytemode1);
    }

    oHash = X86RetrieveOpcode(
	instr, bytemode1, op1, bytemode2, op2, bytemode3, op3);

    if (oHash == NULL) {
	return NULL;
    }

    return oHash->oh_entry;
}



/*
 *  xInstrPushBytes(instr) --
 *
 *  Returns the number of bytes that this instruction will push on the
 *  stack at runtime.
 */
int xInstrPushBytes(Instruction *instr) {
    X86op *opcode = InstrGetOpcodeEntry(instr);

    if ((opcode->x_pushBytes % 4) != 0) {
	printf("0x%8.8x\n", instr->i_address);
    }

    CASSERT(instr != NULL, "Could not locate an opcode");

    if (opcode->x_class == CLASS_PUSH) {
	// if this push happens to have the operand override prefix then
	// the ->x_pushBytes will give us the incorrect value since it
	// gives the number of bytes associated with the opcode, not accounting
	// for the instruction prefix
	if (instr->i_prefixes & PREFIX_OPERAND) {
	    return 2;
	}
    }

    return opcode->x_pushBytes;
}



/*
 *  xInstrPopBytes(instr) --
 *
 *  Returns the number of bytes that this instruction will pop off the
 *  stack at runtime.
 */
int xInstrPopBytes(Instruction *instr) {
    X86op *opcode = InstrGetOpcodeEntry(instr);

    if ((opcode->x_popBytes % 4) != 0) {
	printf("0x%8.8x\n", instr->i_address);
    }

    if ((instr->i_class == CLASS_RET) && instr->i_src) {
	return opcode->x_popBytes + instr->i_src->a_displacement;
    }

    CASSERT(instr != NULL, "Could not locate an opcode");

    if (opcode->x_class == CLASS_POP) {
	// if this push happens to have the operand override prefix then
	// the ->x_popBytes will give us the incorrect value since it
	// gives the number of bytes associated with the opcode, not accounting
	// for the instruction prefix
	if (instr->i_prefixes & PREFIX_OPERAND) {
	    return 2;
	}
    }

    return opcode->x_popBytes;
}



/*
 *  xInstrHasImplicitDest(instr) --
 *
 *  Returns true if the instruction uses an implicit destination
 *  operand.
 */
bool xInstrHasImplicitDest(Instruction *instr) {
    int reg = DIS386_IMPLICIT_DEST(InstrGetOpcodeEntry(instr));
    if (reg) {
	return true;
    }
    return false;
}



/*
 *  xInstrHasImplicitDest2(instr) --
 *
 *  Returns true if the instruction uses a second implicit destination
 *  operand.
 */
bool xInstrHasImplicitDest2(Instruction *instr) {
    int reg = DIS386_IMPLICIT_DEST2(InstrGetOpcodeEntry(instr));
    if (reg) {
	return true;
    }
    return false;
}



/*
 *  xInstrImplicitDest(instr) --
 *
 *  Returns the first implicit destination register the instruction uses.
 */
UINT32 xInstrImplicitDest(Instruction *instr) {
    return DIS386_IMPLICIT_DEST(InstrGetOpcodeEntry(instr));
}



/*
 *  xInstrImplicitDest2(instr) --
 *
 *  Returns the second implicit destination register the instruction uses.
 */
UINT32 _InstrImplicitDest2(Instruction *instr) {
    return DIS386_IMPLICIT_DEST2(InstrGetOpcodeEntry(instr));
}



/*
 *  xInstrHasImplicitSrc(instr) --
 *
 *  Returns true if the instruction uses an implicit source operand
 */
bool xInstrHasImplicitSrc(Instruction *instr) {
    int reg = DIS386_IMPLICIT_SOURCE1(InstrGetOpcodeEntry(instr));
    if (reg && (reg != no_reg)) {
	return true;
    }
    return false;
}



/*
 *  xInstrHasImplicitSrc2(instr) --
 *
 *  Returns true if the instruction uses a second implicit source operand.
 */
bool xInstrHasImplicitSrc2(Instruction *instr) {
    int reg = DIS386_IMPLICIT_SOURCE2(InstrGetOpcodeEntry(instr));
    if (reg && (reg != no_reg)) {
	return true;
    }
    return false;
}



/*
 *  xInstrImplicitSrc(instr) --
 *
 *  Returns the first implicit source register the instruction uses.
 */
UINT32 xInstrImplicitSrc(Instruction *instr) {
    return DIS386_IMPLICIT_SOURCE1(InstrGetOpcodeEntry(instr));
}



/*
 *  _InstrImplicitSrc2(instr) --
 *
 *  Returns the second implicit source register the instruction uses.
 */
UINT32 _InstrImplicitSrc2(Instruction *instr) {
    return DIS386_IMPLICIT_SOURCE2(InstrGetOpcodeEntry(instr));
}



/*
 *  xGetOperandLocation(instr,op) --
 *  Returns the starting offset of an operand in an instruction.
 *
 *  Given an instruction and an operand returns the location of the operand
 *  in the instruction encoding as an offset from the start of the instruction.
 *
 *  The location of an operand only makes sense for immediate and displacement
 *  value.  Other operand types are likely to have an operand share a byte
 *  with the opcode or another operand.  We're usually calling this function in
 *  order to find a spot for a relocation, so that's okay.  We don't want to
 *  relocate those operands anyways.
 *
 *  Returns -1 if the operand doesn't have a location.
 */
int xGetOperandLocation(Instruction *instr, Operand *op) {
    unsigned char buffer[128];

    displacementLoc = immediateLoc = 0;
    InstrAssemble(buffer, instr);

    if (OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG)) {
	return immediateLoc;
    } else if (OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG) ||
	OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
	return displacementLoc;
    }
    return -1;
}



/*
 *  xAssembleCode(code_section) --
 */
void xAssembleCode(Section *code_section) {
    Section *s;
    Instruction *iterator;
    int last_offset, current_offset;
    UINT8 *data;

    s = code_section;

    /*
     * At this point we are examining a section that has instructions.
     * The basic strategy will be to determine what type
     * the operands are, e.g: memory/32 bit registers/8 bit register,
     * then from that look up the opcode in the same table used
     * in the disassembly process (x86-table.h). This should allow
     * for instructions to change operand types (e.g. constant propagation
     * changing a register to an immediate value) and be updated appropriately.
     */

    data = zalloc(s->s_newdatasize);
    current_offset = 0;
    last_offset = current_offset = 0;

    /*
     * Our basic pattern is:
     *   (1) Get operands for the instruction
     *   (2) Figure out what addressing modes are used
     *   (3) Pattern match the instruction class + addressing modes against
     *       classes of instructions located in x86-table.h and get an opcode
     *   (4) Write out prefixes
     *   (5) Write out opcode(s)
     *   (6) Write out operands
     */

    FOREACH_INSTR_IN_SECTION(code_section, iterator) {
	int length;

	last_offset = current_offset;
	UpdateInstrLength(iterator);
	length = InstrAssemble(&data[current_offset], iterator);
	current_offset = current_offset + length;

	/* If ASM_DEBUG is defined then we will do a byte by byte comparison
	 * of the generated binary to the original binary.  This is useful for
	 * determining correctness of the assembler */
#ifdef ASM_DEBUG
	{
	    int i, j;
	    for (i = 0; i < (current_offset - last_offset); i++) {
		if (data[last_offset + i] != s->s_data[last_offset + i]) {
		    printf("Difference in instructions\n");
		    printf("[0x%8.8x] Original:", iterator->i_address);
		    for (j = 0; j < (current_offset - last_offset); j++) {
			printf(" %2.2x", s->s_data[last_offset + j]);
		    }
		    printf("\n[0x%8.8x] Rewrite: ", iterator->i_address);
		    for (j = 0; j < (current_offset - last_offset); j++) {
			printf(" %2.2x", data[last_offset + j]);
		    }
		    printf("\n");
		}
	    }
	}
#endif

	if ((iterator->i_next)
	    && !INSTR_HAS_FLAG(iterator->i_next, INSTR_STARTS_SECTION))
	    if ((current_offset - last_offset) !=
		(iterator->i_next->i_final_address - iterator->i_final_address))
		ReportInstrError
		    ("[ASSEMBLER] Encoded instruction with suspicious length",
		    iterator);

    }				/* end iteration over instruction list */

    s->s_newdata = data;

    return;
}
