/*
 *  Print i386 instructions for GDB, the GNU debugger.
 *  Copyright (C) 1988, 89, 91, 93, 94, 95, 96, 97, 1998
 *  Free Software Foundation, Inc.
 *
 *  Modified by the PLTO project,
 *  Department of Computer Science,
 *  University of Arizona, 2004-2005.
 *
 *
 *  This file is part of GDB.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place -- Suite 330, Boston, MA 02111-1307, USA.
 */
#include <assert.h>

#include "global.h"
#include "x86-dis.h"
#include "x86-table.h"
#include "utilities.h"
#include "x86-prefix.h"
#include "x86-registers.h"
#include "x86-classes.h"
#include "x86-histogram.h"
#include "hash.h"
#include "flowgraph.h"

int putop(char *template, int aflag, int dflag);
void dofloat(int aflag, int dflag);
int get16();
int get32();
void ckprefix();
void SetOperandFlag(Operand *op, int flag, op_rtn *rtn);
static void MarkRelocationOperands(Operand *op, UINT32 strt, int range);

UINT8 *start_codep;		// start of code region
UINT8 *codep;			// cur position in code
Operand *the_op;		// cur operand that is being disassembled

Instruction *the_instr;		// cur instruction
static int mod;
static int rm;
static int reg;
static int prefixes;
static int start_pc;
static int offset;

int regs_32bit[] = {
    reg_a | reg_32, reg_c | reg_32, reg_d | reg_32, reg_b | reg_32,
    reg_sp | reg_32, reg_bp | reg_32, reg_si | reg_32, reg_di | reg_32
};

int regs_16bit[] = {
    reg_a | reg_16, reg_c | reg_16, reg_d | reg_16, reg_b | reg_16,
    reg_sp | reg_16, reg_bp | reg_16, reg_si | reg_16, reg_di | reg_16
};

int regs_8bit[] = {
    reg_a | reg_8l, reg_c | reg_8l, reg_d | reg_8l, reg_b | reg_8l,
    reg_a | reg_8h, reg_c | reg_8h, reg_d | reg_8h, reg_b | reg_8h
};

int regs_seg[] = {
    reg_es, reg_cs, reg_ss, reg_ds, reg_fs, reg_gs
};

int regs_cr[] = {
    reg_cr0, 0, reg_cr2, reg_cr3, reg_cr4
};



/*
 *  ckprefix() --
 */
void ckprefix() {
    prefixes = 0;

    /* Keep getting prefixes until the code doesn't correspond to
     * any known prefix.  The limit on the number of prefixes is
     * 4, so this loop should iterate a maximum of 4 times. If a
     * prefix is discovered then advance the instruction pointer a
     * byte.  If a prefix is not discovered, leave it because that
     * means it's pointing at the opcode, which is what we want. */

    while (1) {
	switch (*codep) {
	case X86_PREFIX_REPZ:
	    prefixes |= PREFIX_REPZ;
	    break;
	case X86_PREFIX_REPNZ:
	    prefixes |= PREFIX_REPNZ;
	    break;
	case X86_PREFIX_LOCK:
	    prefixes |= PREFIX_LOCK;
	    break;
	case X86_PREFIX_CS:
	    prefixes |= PREFIX_CS;
	    break;
	case X86_PREFIX_SS:
	    prefixes |= PREFIX_SS;
	    break;
	case X86_PREFIX_DS:
	    prefixes |= PREFIX_DS;
	    break;
	case X86_PREFIX_ES:
	    prefixes |= PREFIX_ES;
	    break;
	case X86_PREFIX_FS:
	    prefixes |= PREFIX_FS;
	    break;
	case X86_PREFIX_GS:
	    prefixes |= PREFIX_GS;
	    break;
	case X86_PREFIX_OPERAND:
	    prefixes |= PREFIX_OPERAND;
	    break;
	case X86_PREFIX_ADDRESS:
//          WARNING("[Disassembling] Found an address prefix.");
	    prefixes |= PREFIX_ADDRESS;
	    break;
	case X86_PREFIX_FWAIT:
	    prefixes |= PREFIX_FWAIT;
	    break;
	default:
	    return;
	}
	codep++;
    }
}

int decode_insn_x86(ADDRESS pc, UINT8 *b, Instruction *instr, int aflag,
    int dflag);



/*
 *  xDisassembleInstr(pc,b,instr) --
 */
int xDisassembleInstr(ADDRESS pc, UINT8 *b, Instruction *instr) {
    return decode_insn_x86(pc, b, instr, 1, 1);
}



/*
 *  decode_insn_x86(pc,b,instr,aflag,dflag) --
 */
int decode_insn_x86(ADDRESS pc, UINT8 *b, Instruction *instr, int aflag,
    int dflag) {

    X86op *dp;
    unsigned char need_modrm;

    UINT8 *inbuf = b;
    the_instr = instr;
    start_pc = pc;
    start_codep = inbuf;
    codep = inbuf;
    offset = 0;
    the_instr->i_flags = 0;

    /* First we parse out the prefixes if they exist. */
    ckprefix();

    /* Turn on the bits in the prefix field for the instruction corresponding
     * to whatever we found as a prefix */
    the_instr->i_prefixes = prefixes;

    if (prefixes & PREFIX_OPERAND) {
	dflag ^= 1;
    }

    if (prefixes & PREFIX_ADDRESS) {
	aflag ^= 1;
    }

    /*
     * At this point codep is pointing to the byte containing the opcode for
     * the instruction.
     */

    if (*codep == 0x0f) {
	// If the opcode is 0x0f then that means the opcode is actually two
	// bytes, and the real interpretation lies in the next byte
	dp = &dis386_twobyte[*++codep];
	offset++;
	if (dp->x_name != NULL) {
	    twobyteOpcodes[*codep] = twobyteOpcodes[*codep] + 1;
	}
	need_modrm = twobyte_has_modrm[*codep];

    } else {
	// Otherwise we have a normal one-byte opcode.
	dp = &dis386[*codep];
	if (dp->x_name != NULL) {
	    onebyteOpcodes[*codep] = onebyteOpcodes[*codep] + 1;
	}
	need_modrm = onebyte_has_modrm[*codep];
	
    }

    if (dp->x_class == CLASS_ERROR) {
	// can't handle this instruction, so the safe thing to do is 
	// let the disassembler flist the function.
	return -1;
    }

    codep++;
    offset++;

    /*
     * At this point codep is pointing to the MODR/M byte if it is encoded
     * in the instruction.  If not, then it may be pointing at SIB byte or
     * displacement.
     */

    if (need_modrm) {
	mod = (*codep >> 6) & 3;
	reg = (*codep >> 3) & 7;
	rm = *codep & 7;
	INSTR_SET_FLAG(instr, EXTENDED_OPCODE);
    }

    if (dp->x_name == NULL && dp->x_bytemode1 == FLOATCODE) {

	/* If we have a floating floating instruction, handle it accordingly */
	dofloat(aflag, dflag);

    } else {
	/* Otherwise it's a normal non-floating point instruction. */

	/* If we still haven't found an instruction, then the bits in the REG
	 * section of the MODR/M byte point us to a group placeholder (grps)
	 * where we might find an instruction */
	if (dp->x_name == NULL) {
	    grpsOpcodes[(dp->x_bytemode1 * 8) + reg] =
		grpsOpcodes[(dp->x_bytemode1 * 8) + reg] + 1;
	    dp = &grps[dp->x_bytemode1][reg];
	}

	/*
	 * By now dp definitely contains a valid instructions and maybe
	 * even some operands.
	 */

	/* Fill in the instruction with the type and class */
	instr->i_type = dp->x_type;
	instr->i_class = dp->x_class;

	/* Get the operand size.  Although we don't do anything with it */
	putop(dp->x_name, aflag, dflag);

	/* Parse the 3 operands (if they exist) */
	if (dp->x_op1) {
	    int oldoffset = offset;
	    instr->i_dest = OperandCreate();
	    the_op = instr->i_dest;
	    the_op->a_flags = 0;
	    INSTR_SET_FLAG(instr, DEST_FLAG);
	    SetOperandFlag(the_op, dp->x_bytemode1, &dp->x_op1);
	    (*dp->x_op1) (dp->x_bytemode1, aflag, dflag);
	    MarkRelocationOperands(the_op, start_pc + oldoffset,
		offset - oldoffset);
	}

	if (dp->x_op2) {
	    int oldoffset = offset;
	    instr->i_src = OperandCreate();
	    the_op = instr->i_src;
	    the_op->a_flags = 0;
	    INSTR_SET_FLAG(instr, SOURCE1_FLAG);
	    SetOperandFlag(the_op, dp->x_bytemode2, &dp->x_op2);
	    (*dp->x_op2) (dp->x_bytemode2, aflag, dflag);
	    MarkRelocationOperands(the_op, start_pc + oldoffset,
		offset - oldoffset);
	}

	if (dp->x_op3) {
	    int oldoffset = offset;
	    instr->i_src2 = OperandCreate();
	    the_op = instr->i_src2;
	    the_op->a_flags = 0;
	    INSTR_SET_FLAG(instr, SOURCE2_FLAG);
	    SetOperandFlag(the_op, dp->x_bytemode3, &dp->x_op3);
	    (*dp->x_op3) (dp->x_bytemode3, aflag, dflag);
	    MarkRelocationOperands(the_op, start_pc + oldoffset,
		offset - oldoffset);
	}

    }

    if (pc != 0) {
	instr->i_address = pc;
    } else {
	instr->i_address = MAGIC_NUMBER;
    }

    return codep - inbuf;
}



/*
 *  dofloat(aflag,dflag) --
 */
void dofloat(int aflag, int dflag) {
    X86op *dp;
    unsigned char floatop;

    floatop = codep[-1];

    if (mod != 3) {
	dp = &float_mem[(floatop - 0xd8) * 8 + reg];
	floatmemOpcodes[(floatop - 0xd8) * 8 + reg] =
	    floatmemOpcodes[(floatop - 0xd8) * 8 + reg] + 1;
	the_instr->i_type = dp->x_type;
	the_instr->i_class = dp->x_class;

	if (dp->x_op1) {
	    int oldoffset = offset;
	    the_instr->i_dest = OperandCreate();
	    the_op = the_instr->i_dest;
	    the_op->a_flags = 0;
	    INSTR_SET_FLAG(the_instr, DEST_FLAG);
	    SetOperandFlag(the_op, dp->x_bytemode1, &dp->x_op1);
	    (*dp->x_op1) (dp->x_bytemode1, aflag, dflag);
	    MarkRelocationOperands(the_op, start_pc + oldoffset,
		offset - oldoffset);
	}

	if (dp->x_op2) {
	    int oldoffset = offset;
	    the_instr->i_src = OperandCreate();
	    the_op = the_instr->i_src;
	    the_op->a_flags = 0;
	    INSTR_SET_FLAG(the_instr, SOURCE1_FLAG);
	    SetOperandFlag(the_op, dp->x_bytemode2, &dp->x_op2);
	    (*dp->x_op2) (dp->x_bytemode2, aflag, dflag);
	    MarkRelocationOperands(the_op, start_pc + oldoffset,
		offset - oldoffset);
	}

	if (dp->x_op3) {
	    int oldoffset = offset;
	    the_instr->i_src2 = OperandCreate();
	    the_op = the_instr->i_src2;
	    the_op->a_flags = 0;
	    INSTR_SET_FLAG(the_instr, SOURCE2_FLAG);
	    SetOperandFlag(the_op, dp->x_bytemode3, &dp->x_op3);
	    (*dp->x_op3) (dp->x_bytemode3, aflag, dflag);
	    MarkRelocationOperands(the_op, start_pc + oldoffset,
		offset - oldoffset);
	}

	return;
    }
    codep++;
    offset++;

    dp = &float_reg[floatop - 0xd8][reg];

    if (dp->x_name == NULL) {
	fgrpsOpcodes[dp->x_bytemode1 * 8 + rm] =
	    fgrpsOpcodes[dp->x_bytemode1 * 8 + rm] + 1;
	dp = &fgrps[dp->x_bytemode1][rm];
	the_instr->i_type = dp->x_type;
	the_instr->i_class = dp->x_class;
	return;

    } else {

	the_instr->i_type = dp->x_type;
	the_instr->i_class = dp->x_class;
	floatregOpcodes[8 * (floatop - 0xd8) + reg] =
	    floatregOpcodes[8 * (floatop - 0xd8) + reg] + 1;

	if (dp->x_op1) {
	    int oldoffset = offset;
	    the_instr->i_dest = OperandCreate();
	    the_op = the_instr->i_dest;
	    the_op->a_flags = 0;
	    INSTR_SET_FLAG(the_instr, DEST_FLAG);
	    SetOperandFlag(the_op, dp->x_bytemode1, &dp->x_op1);
	    (*dp->x_op1) (dp->x_bytemode1, aflag, dflag);
	    MarkRelocationOperands(the_op, start_pc + oldoffset,
		offset - oldoffset);
	}

	if (dp->x_op2) {
	    int oldoffset = offset;
	    the_instr->i_src = OperandCreate();
	    the_op = the_instr->i_src;
	    the_op->a_flags = 0;
	    INSTR_SET_FLAG(the_instr, SOURCE1_FLAG);
	    SetOperandFlag(the_op, dp->x_bytemode2, &dp->x_op2);
	    (*dp->x_op2) (dp->x_bytemode2, aflag, dflag);
	    MarkRelocationOperands(the_op, start_pc + oldoffset,
		offset - oldoffset);
	}

    }
}



/*
 *  putop(template,aflag,dflag) --
 */
int putop(char *template, int aflag, int dflag) {
    char *p;
    for (p = template; *p; p++) {
	switch (*p) {
	default:
	    return -1;
	    break;
	case 'S':
	    if (dflag) {
		return 32;
	    } else {
		return 16;
	    }
	    break;
	case 'W':
	    if (dflag) {
		return 16;
	    } else {
		return 8;
	    }
	    break;
	}
    }
    return -1;
}



/*
 *  get32() --
 */
int get32() {
    int x = 0;

    x = *codep++ & 0xff;
    x |= (*codep++ & 0xff) << 8;
    x |= (*codep++ & 0xff) << 16;
    x |= (*codep++ & 0xff) << 24;
    offset = offset + 4;
    return x;
}



/*
 *  get16() --
 */
int get16() {
    int x = 0;

    x = *codep++ & 0xff;
    x |= (*codep++ & 0xff) << 8;
    offset = offset + 2;
    return x;
}



/*
 *  OP_ST(ignore,aflag,dflag) --
 */
int OP_ST(int ignore, int aflag, int dflag) {
    OP_BASE(the_op, reg_st);
    return 0;
}



/*
 *  OP_STi(ignore,aflag,dflag) --
 */
int OP_STi(int ignore, int aflag, int dflag) {
    switch (rm) {
    case 0:
	OP_BASE(the_op, reg_st0);
	break;
    case 1:
	OP_BASE(the_op, reg_st1);
	break;
    case 2:
	OP_BASE(the_op, reg_st2);
	break;
    case 3:
	OP_BASE(the_op, reg_st3);
	break;
    case 4:
	OP_BASE(the_op, reg_st4);
	break;
    case 5:
	OP_BASE(the_op, reg_st5);
	break;
    case 6:
	OP_BASE(the_op, reg_st6);
	break;
    case 7:
	OP_BASE(the_op, reg_st7);
	break;
    }
    return 0;
}



/*
 *  OP_indirE(bytemode,aflag,dflag) --
 */
int OP_indirE(int bytemode, int aflag, int dflag) {
    INSTR_SET_FLAG(the_instr, INDIRECT);
    OPERAND_SET_FLAG(the_op, OPERAND_INDIRECT_FLAG);
    return OP_E(bytemode, aflag, dflag);
}



/*
 *  OP_PREFETCH(bytemode,aflag,dflag) --
 */
int OP_PREFETCH(int bytemode, int aflag, int dflag) {
    return OP_E(bytemode, aflag, dflag);
}



/*
 *  OP_E(bytemode,aflag,dflag) --
 *
 *  Get information about the operands.  Chapter 2 of the Intel Manual Vol 2
 *  is especially useful here, specifically the tables.
 */
int OP_E(int bytemode, int aflag, int dflag) {
    int disp;
    int havesib;
    int havebase;
    int base;
    int idx = 0;
    int scale = 0;

    havesib = 0;
    havebase = 1;
    base = rm;

    /* Currently examining the MODR/M byte */
    codep++;
    offset++;


    /* if mod == 3 then the operand is simply a register, and the register
     * is encoded in the R/M bits.  We look it up in the table and set
     * things accordingly */
    if (mod == 3) {
	OP_SET_FLAG(the_op, OPERAND_REGISTER_MODRM);
	switch (bytemode) {
	case b_mode:
	    OP_BASE(the_op, regs_8bit[rm]);
	    break;
	case w_mode:
	    OP_BASE(the_op, regs_16bit[rm]);
	    break;
	case v_mode:
	case d_mode:
	    if (dflag) {
		OP_BASE(the_op, regs_32bit[rm]);
	    } else {
		OP_BASE(the_op, regs_16bit[rm]);
	    }
	    break;
	case 0:
	    REPORT(RP_TRIVIAL, "[OP_E] returning -1 @ %x", offset);
	    return -1;		// this is to allow for simulated disassembly
				// that certain other tools may utilize (cml)
	default:
	    MALFUNCTION("unexpected bytemode");
	}
	return 0;
    }

    /* MOD was not "11", so a memory addressing mode is used */
    disp = 0;

    if (aflag) {			 // aflag = 32 bit address mode

	/* For 32 bit addressing modes, an R/M of 4
	 * means an SIB follows the MOD/RM byte */
	if (base == 4) {
	    havesib = 1;
	    scale = (*codep >> 6) & 3;
	    idx = (*codep >> 3) & 7;
	    base = *codep & 7;
	    codep++;
	    offset++;
	}

	switch (mod) {

	    /* All effective addresses for MOD 00 have no displacement--
	     * except when R/M is 5 */
	case 0:
	    /* If in fact R/M is 5, a displacement follows */
	    if (base == 5) {
		havebase = 0;
		disp = get32();
		OP_DISP_A(the_op, disp);
	    }
	    break;

	    /* For MOD bits of 01, displacement follows in the next 8 bits */
	case 1:
	    disp = *codep++;
	    offset++;
	    /* This is pretty confusing. Since 8 bit displacements are sign
	     * extended and then added to the index, we will just sign extend
	     * any number that's supposed to be negative by subtracting 256
	     * from it, then we can add it to the index.
	     * disp & 0x80 is simply checking the sign bit to see if it's 1
	     * (negative number) */
	    if ((disp & 0x80) != 0) {
		disp -= 0x100;
	    }
	    OP_DISP_R(the_op, disp);
	    break;
	    /* For 32 bit displacements we will just read the next 4 bytes */
	case 2:
	    disp = get32();
	    OP_DISP_A(the_op, disp);
	    break;
	}

	/* The following code makes references to "index 4" which is really
	 * the index to the ESP register. The ESP register, by definition,
	 * cannot be used as an index in an addressing mode. */

	if (havebase || (havesib && (idx != 4 || scale != 0))) {
	    OP_SET_FLAG(the_op, OPERAND_REGISTER_INDIRECT);
	    if (havebase) {
		OP_BASE(the_op, regs_32bit[base]);
	    }
	    if (havesib) {
		if (idx != 4) {
		    OP_INDEX(the_op, regs_32bit[idx]);
		}
		OP_SCALE(the_op, 1 << scale);
	    }
	}

    } else {

	/* Otherwise we have a 16 bit address mode */

	switch (mod) {
	case 0:
	    if (rm == 6) {
		havebase = 0;
		disp = get16();
		if ((disp & 0x8000) != 0) {
		    disp -= 0x10000;
		}
	    }
	    break;
	case 1:
	    disp = *codep++;
	    offset++;
	    if ((disp & 0x80) != 0) {
		disp -= 0x100;
	    }
	    break;
	case 2:
	    disp = get16();
	    if ((disp & 0x8000) != 0) {
		disp -= 0x10000;
	    }
	    break;
	}

	if (mod != 0 || rm == 6) {
	    OP_DISP_A(the_op, disp);
	}

	if (havebase || (havesib && (idx != 4 || scale != 0))) {
	    if (havebase) {
		OP_BASE(the_op, regs_32bit[base]);
	    }
	    if (havesib) {
		if (idx != 4) {
		    OP_INDEX(the_op, regs_32bit[idx]);
		}
		OP_SCALE(the_op, 1 << scale);
	    }
	}

    }
    return 0;
}



/*
 *  OP_G(bytemode,aflag,dflag) --
 */
int OP_G(int bytemode, int aflag, int dflag) {
    OP_SET_FLAG(the_op, OPERAND_OPG_FLAG);
    switch (bytemode) {
    case b_mode:
	OP_BASE(the_op, regs_8bit[reg]);
	break;
    case w_mode:
	OP_BASE(the_op, regs_16bit[reg]);
	break;
    case d_mode:
	OP_BASE(the_op, regs_32bit[reg]);
	break;
    case v_mode:
	if (dflag) {
	    OP_BASE(the_op, regs_32bit[reg]);
	} else {
	    OP_BASE(the_op, regs_16bit[reg]);
	}
	break;
    default:
	MALFUNCTION("unexpected bytemode");
    }
    return 0;
}



/*
 *  OP_REG(code,aflag,dflag) --
 */
int OP_REG(int code, int aflag, int dflag) {
    OP_SET_FLAG(the_op, OPERAND_OPREG_FLAG);
    switch (code) {
    case indir_dx_reg:
	OPERAND_SET_FLAG(the_op, OPERAND_REGISTER_INDIRECT);
	OP_BASE(the_op, reg_d | reg_16 | reg_indirect);
	break;
    case ax_reg:
    case cx_reg:
    case dx_reg:
    case bx_reg:
    case sp_reg:
    case bp_reg:
    case si_reg:
    case di_reg:
	OP_BASE(the_op, regs_16bit[code - ax_reg]);
	break;
    case es_reg:
    case ss_reg:
    case cs_reg:
    case ds_reg:
    case fs_reg:
    case gs_reg:
	OP_BASE(the_op, regs_seg[code - es_reg]);
	break;
    case al_reg:
    case ah_reg:
    case cl_reg:
    case ch_reg:
    case dl_reg:
    case dh_reg:
    case bl_reg:
    case bh_reg:
	OP_BASE(the_op, regs_8bit[code - al_reg]);
	break;
    case eAX_reg:
    case eCX_reg:
    case eDX_reg:
    case eBX_reg:
    case eSP_reg:
    case eBP_reg:
    case eSI_reg:
    case eDI_reg:
	if (dflag) {
	    OP_BASE(the_op, regs_32bit[code - eAX_reg]);
	} else {
	    OP_BASE(the_op, regs_16bit[code - eAX_reg]);
	}
	break;
    default:
	MALFUNCTION("unexpected code");
    }
    return 0;
}



/*
 *  OP_I(bytemode,aflag,dflag) --
 */
int OP_I(int bytemode, int aflag, int dflag) {
    switch (bytemode) {
    case b_mode:
	OP_IMM(the_op, *codep++ & 0xff);
	offset++;
	break;
    case v_mode:
	if (dflag) {
	    OP_IMM(the_op, get32());
	} else {
	    OP_IMM(the_op, get16());
	}
	break;
    case w_mode:
	OP_IMM(the_op, get16());
	break;
    case d_mode:
	OP_IMM(the_op, get32());
	break;
    default:
	MALFUNCTION("unexpected bytemode");
    }
    return 0;
}



/*
 *  OP_sI(bytemode,aflag,dflag) --
 */
int OP_sI(int bytemode, int aflag, int dflag) {
    int op;

    OP_SET_FLAG(the_op, OPERAND_SIGN_EXTEND_FLAG);

    switch (bytemode) {
    case b_mode:
	op = *codep++;
	offset++;
	if ((op & 0x80) != 0) {
	    op -= 0x100;
	}
	break;
    case v_mode:
	if (dflag) {
	    op = get32();
	} else {
	    op = get16();
	    if ((op & 0x8000) != 0) {
		op -= 0x10000;
	    }
	}
	break;
    case w_mode:
	op = get16();
	if ((op & 0x8000) != 0) {
	    op -= 0x10000;
	}
	break;
    default:
	MALFUNCTION("unexpected bytemode");
    }
    OP_IMM(the_op, op);
    return 0;
}



/*
 *  OP_J(bytemode,aflag,dflag) --
 */
int OP_J(int bytemode, int aflag, int dflag) {
    int disp;
    int mask = -1;

    switch (bytemode) {
    case b_mode:
	disp = *codep++;
	offset++;
	if ((disp & 0x80) != 0) {
	    disp -= 0x100;
	}
	break;
    case v_mode:
	if (dflag) {
	    disp = get32();
	} else {
	    disp = get16();
	    if ((disp & 0x8000) != 0) {
		disp -= 0x10000;
	    }
	    /* for some reason, a data16 prefix on a jump instruction
	     * means that the pc is masked to 16 bits after the
	     * displacement is added!  */
	    mask = 0xffff;
	}
	break;
    default:
	MALFUNCTION("unexpected bytemode");
    }
    OP_DISP_R(the_op, disp);
				// ** EDIT - Fix me - what about the mask?
    disp = (start_pc + codep - start_codep + disp) & mask;
				// ** EDIT - should this be absolute?

    return 0;
}



/*
 *  OP_SEG(dummy,aflag,dflag) -- encode segment register operand.
 */
int OP_SEG(int dummy, int aflag, int dflag) {
    OP_SET_FLAG(the_op, OPERAND_SEGMENT_REG_FLAG);
    if (mod <= 3) {
        OP_BASE(the_op, regs_seg[reg]);
        return 0;
    }
    MALFUNCTION("[OP_SEG] unexpected MODRM byte");
    return 0;
}

/*
 *  OP_DIR(size,aflag,dflag) --
 */
int OP_DIR(int size, int aflag, int dflag) {
    int seg, off;

//  From binutils ...
// if (sizeflag & DFLAG)
//    {
//      offset = get32 ();
//      seg = get16 ();
//    }
// else
//    {
//      offset = get16 ();
//      seg = get16 ();
//    }
       
    switch (size) {
    case lptr:
	if (dflag) {
	    off = get32();
	    seg = get16();
	} else {
	    off = get16();
	    seg = get16();
	}
	OP_DISP_R(the_op, off);
	break;
    case v_mode:
	if (aflag) {
	    off = get32();
	} else {
	    off = get16();
	    if ((off & 0x8000) != 0) {
		off -= 0x10000;
	    }
	}
	OP_DISP_R(the_op, off);
	off = start_pc + codep - start_codep + off;
	break;
    case d_mode:
	if (aflag) {
	    off = get32();
	    break;
	}
//    default:
//	MALFUNCTION("unexpected bytemode"); // Igor 10/07, for junk insertion
    }
    return 0;
}



/*
 *  OP_OFF(bytemode,aflag,dflag) --
 */
int OP_OFF(int bytemode, int aflag, int dflag) {
    int off;

    if (aflag) {
	off = get32();
    } else {
	off = get16();
    }

    OP_DISP_R(the_op, off);
    return 0;
}



/*
 *  OP_ESDI(dummy,aflag,dflag) --
 */
int OP_ESDI(int dummy, int aflag, int dflag) {
    OP_SET_FLAG(the_op, OPERAND_SIDI_FLAG);
    OP_SET_FLAG(the_op, OPERAND_REGISTER_INDIRECT);
    OP_BASE(the_op, reg_di | reg_32);
    return 0;
}



/*
 *  OP_DSSI(dummy,aflag,dflag) --
 */
int OP_DSSI(int dummy, int aflag, int dflag) {
    OP_SET_FLAG(the_op, OPERAND_SIDI_FLAG);
    OP_SET_FLAG(the_op, OPERAND_REGISTER_INDIRECT);
    OP_BASE(the_op, reg_si | reg_32);
    return 0;
}



/*
 *  OP_C(dummy,aflag,dflag) -- encode control register operand.
 */
int OP_C(int dummy, int aflag, int dflag) {
    OP_SET_FLAG(the_op, OPERAND_CONTROL_REG_FLAG);
    OP_BASE(the_op, regs_cr[reg]);
    return 0;
}



/*
 *  OP_D(dummy,aflag,dflag) -- encode debug register operand.
 */
int OP_D(int dummy, int aflag, int dflag) {
    OP_SET_FLAG(the_op, OPERAND_DEBUG_REG_FLAG);
    OP_BASE(the_op, reg);
    return 0;
}



/*
 *  OP_T(dummy,aflag,dflag) --
 */
int OP_T(int dummy, int aflag, int dflag) {
    codep++;			// skip mod/rm
    offset++;
    return -1;			// this is to allow for simulated disassembly
				// that certain other tools may utilize (cml)
}



/*
 *  OP_rm(bytemode,aflag,dflag) -- encode an RM-specified operand.
 */
int OP_rm(int bytemode, int aflag, int dflag) {

    if (mod == 3) {
	OP_SET_FLAG(the_op, OPERAND_OPREG_FLAG);
    }

    switch (bytemode) {
    case d_mode:
	OP_BASE(the_op, regs_32bit[rm]);
	break;
    case w_mode:
	OP_BASE(the_op, regs_16bit[rm]);
	break;
    }
    return 0;
}



/*
 *  OP_MMX(bytemode,aflag,dflag) --
 */
int OP_MMX(int bytemode, int aflag, int dflag) {
    return -1;			// this is to allow for simulated disassembly
				// that certain other tools may utilize (cml)
}



/*
 *  OP_EM(bytemode,aflag,dflag) --
 */
int OP_EM(int bytemode, int aflag, int dflag) {
    if (mod != 3) {
	return OP_E(bytemode, aflag, dflag);
    }

    codep++;
    offset++;
    return -1;			// this is to allow for simulated disassembly
				// that certain other tools may utilize (cml)
}



/*
 *  OP_MS(bytemode,aflag,dflag) --
 */
int OP_MS(int bytemode, int aflag, int dflag) {
    ++codep;
    offset++;
    return -1;			// this is to allow for simulated disassembly
				// that certain other tools may utilize (cml)
}



/*
 *  OP_3DNow(bytemode,aflag,dflag) --
 */
int OP_3DNow(int bytemode, int aflag, int dflag) {
    return -1;			// this is to allow for simulated disassembly
				// that certain other tools may utilize (cml)
}



/*
 *  OP_3DNowSuffix(bytemode,aflag,dflag) --
 */
int OP_3DNowSuffix(int bytemode, int aflag, int dflag) {
    return -1;			// this is to allow for simulated disassembly
				// that certain other tools may utilize (cml)
}



/*
 *  SetOperandFlag(op,bytemode,rtn) --
 */
void SetOperandFlag(Operand *op, int bytemode, op_rtn *rtn) {
    /* First set the bytemode */
    if (bytemode == v_mode) {
	OP_SET_FLAG(op, OPERAND_VMODE_FLAG);
    }
    if (bytemode == w_mode) {
	OP_SET_FLAG(op, OPERAND_WMODE_FLAG);
    }
    if (bytemode == b_mode) {
	OP_SET_FLAG(op, OPERAND_BMODE_FLAG);
    }
    if (bytemode == d_mode) {
	OP_SET_FLAG(op, OPERAND_DMODE_FLAG);
    }

    return;
}



/*
 *  IndexToOneByteOpcode(idx) --
 */
X86op IndexToOneByteOpcode(int idx) {
    return dis386[idx];
}



/*
 *  IndexToTwoByteOpcode(idx) --
 */
X86op IndexToTwoByteOpcode(int idx) {
    return dis386_twobyte[idx];
}



/*
 *  IndexToGRPSOpcode(idx) --
 */
X86op IndexToGRPSOpcode(int idx) {
    return grps[idx / 8][idx % 8];
}



/*
 *  IndexToFGRPSOpcode(idx) --
 */
X86op IndexToFGRPSOpcode(int idx) {
    return fgrps[idx / 8][idx % 8];
}



/*
 *  IndexToFloatMemOpcode(idx) --
 */
X86op IndexToFloatMemOpcode(int idx) {
    return float_mem[idx];
}



/*
 *  IndexToFloatRegOpcode(idx) --
 */
X86op IndexToFloatRegOpcode(int idx) {
    return float_reg[idx / 8][idx % 8];
}



/*
 *  MarkRelocationOperands(op,strt,range) --
 */
static void MarkRelocationOperands(Operand *op, UINT32 strt, int range) {
    int i;
    Relocation *r;

    CASSERT(range >= 0, "range < 0.");

    for (i = 0; i < range; i++) {
	r = RelocLookupEntry(strt + i);
	if (r != NULL) {
	    op->a_relocation = r;
	    if (RELOCATION_HAS_FLAG(r, RELOCATION_PC_RELATIVE_FLAG)) {
		OPERAND_SET_FLAG(the_op, OPERAND_RELOCATABLE_PC_FLAG);
		if (RELOCATION_HAS_FLAG(r, RELOCATION_GOTPC_FLAG)) {
		    OPERAND_SET_FLAG(the_op, OPERAND_RELOCATABLE_GOTPC_FLAG);
		}
	    } else {
		OPERAND_SET_FLAG(the_op, OPERAND_RELOCATABLE_ABS_FLAG);
	    }
	}
    }

    return;
}
