/*
 *  x86/hash.c -- X86 Hashing module.
 */
#include "global.h"
#include "utilities.h"
#include "x86-classes.h"
#include "x86-asm.h"
#include "x86-hash.h"
#include "x86-table.h"

#define HashVal(class, bmode1, bmode2, bmode3) \
    (((class) ^ ((bmode1) << 0) ^ ((bmode2) << 3) ^ ((bmode3) << 6)) &0x3ff)

OpcodeHash *opcodeHash[1024];	// NUMBER_OF_CLASSES + 10


/* 
 * SanityCheckForDuplicates(hash) -- Identifies duplicate hash chain elements
 */
bool SanityCheckForDuplicates(OpcodeHash *hash, X86op *entry) {
    OpcodeHash *temp = hash;
    
    while (temp != NULL) {
        if (OpcodeEntryMatches(temp->oh_entry, 
        	entry->x_op1, entry->x_bytemode1,
        	entry->x_op2, entry->x_bytemode2,
        	entry->x_op3, entry->x_bytemode3,
        	entry->x_type, entry->x_class)) {
		if (entry->x_class != CLASS_ERROR &&
			entry->x_class != CLASS_PACK) {
        	    MALFUNCTION(
		        "Identical opcode entries in InstrHashTable [%s,%s]",
        		entry->x_name,
        		temp->oh_entry->x_name);
                }
        	return false;
        }
        temp = temp->oh_hashNext;
    }
    return true;
}

/*
 *  X86InsertOpcode(index1,index2,index3,entry,type) --
 *
 *  Inserts a dis386 entry from x86-table.h with the given indices
 *  into the opcode table.
 */
void X86InsertOpcode(
int index1, int index2, int index3, X86op *entry, int type) {
    OpcodeHash *newHash;
    int index;

    index = HashVal(entry->x_class, entry->x_bytemode1, entry->x_bytemode2,
	entry->x_bytemode3);

    newHash = zalloc(sizeof(OpcodeHash));
    newHash->oh_index1 = index1;
    newHash->oh_index2 = index2;
    newHash->oh_index3 = index3;
    newHash->oh_entry = entry;

    // Before adding entry make sure that there are no irrecoverable
    // collision in the chain.
    // SanityCheckForDuplicates(opcodeHash[index], entry);

    newHash->oh_hashNext = opcodeHash[index];
    newHash->oh_flags |= type;
    opcodeHash[index] = newHash;

    return;
}



/*
 *  X86RetrieveOpcode(instr,bytemode1,op1,bytemode2,op2,bytemode3,op3) --
 *  Retrieves the requested entry from the table.
 */
OpcodeHash *X86RetrieveOpcode(Instruction *instr, int bytemode1, op_rtn op1,
    int bytemode2, op_rtn op2, int bytemode3, op_rtn op3) {

    OpcodeHash *chain;
    int index;

    index = HashVal(instr->i_class, bytemode1, bytemode2, bytemode3);
    chain = opcodeHash[index];

    while (chain != NULL) {
	if (OpcodeEntryMatches(chain->oh_entry, op1, bytemode1, op2,
		bytemode2, op3, bytemode3, instr->i_type, instr->i_class)) {
	    return chain;
	}
	chain = chain->oh_hashNext;
    }

    return NULL;

}



/*
 *  X86BuildHashTable() --
 *
 *  This functions builds the hash table for the X86 opcodes.
 *  It should be called one time when the program first starts up.
 */
void X86BuildHashTable() {
    int i, j, k, q;

    /*
     *    Insert everything in dis386[] and dis386_twobyte[]
     */

    for (i = 0; i < 32; i++) {
	for (j = 0; j < 8; j++) {
	    X86InsertOpcode(i, j, 0, &dis386[(i * 8) + j], OPCODE_FLAG_DIS386);
	    X86InsertOpcode(i, j, 0, &dis386_twobyte[(i * 8) + j],
		OPCODE_FLAG_DIS3862BYTE);
	}
    }

    /*
     *               Insert everything in grps[]
     */

    for (i = 0; i < 20; i++) {
	for (j = 0; j < 8; j++) {
	    for (k = 0; k <= 0xff; k++) {
		if (dis386[k].x_name == NULL && dis386[k].x_bytemode1 == i) {
		    X86InsertOpcode(i, j, k, &grps[i][j], OPCODE_FLAG_GRPS);
		}
	    }
	    for (k = 0; k <= 0xff; k++) {
		if ((dis386_twobyte[k].x_name == NULL)
		    && (dis386_twobyte[k].x_bytemode1 == i))
		    X86InsertOpcode(i, j, k, &grps[i][j],
			OPCODE_FLAG_GRPS | OPCODE_FLAG_TWOBYTE);
	    }
	}
    }

    /*
     *          Insert everything in float_mem[]
     */

    for (i = 0; i < 8; i++) {
	for (j = 0; j < 8; j++) {
	    X86InsertOpcode(i, j, 0, &float_mem[(i * 8) + j],
		OPCODE_FLAG_FLOATMEM);
	}
    }

    /*
     *          Insert everything in float_reg[]
     */

    for (i = 0; i < 8; i++) {
	for (j = 0; j < 8; j++) {
	    X86InsertOpcode(i, j, 0, &float_reg[i][j], OPCODE_FLAG_FLOATREG);
	}
    }

    /*
     *             Insert everything in fgrps[]
     */

    for (i = 0; i < 9; i++) {
	for (j = 0; j < 8; j++) {
	    for (k = 0; k < 8; k++) {
		for (q = 0; q < 8; q++) {
		    if ((float_reg[k][q].x_name == NULL)
			&& (float_reg[k][q].x_bytemode1 == i))
			X86InsertOpcode(i, j, k * 8 + q, &fgrps[i][j],
			    OPCODE_FLAG_FGRPS);

		}
	    }
	}
    }
}
