/*
 *  x86/icreate.c
 *
 *  Creates instructions corresponding to general operations that
 *  some machine-independent optimizations may wish to use.
 */
#include "global.h"
#include "flowgraph.h"
#include "x86-classes.h"
#include "x86-icreate.h"
#include "x86-liveness.h"
#include "utilities.h"
#include "linkage.h"
#include "x86-registers.h"
#include "x86-dis.h"
#include "upm.h"
#include "relocate.h"
#include "sysdep.h"



/*
 *  xInstrCreateNOP() --
 *
 *  Creates and returns an instruction that has no effect on the state of
 *  the processor.  The instruction is guaranteed to be 1 byte in length,
 */
Instruction *xInstrCreateNOP() {
    Instruction *nop = InstrCreate();

    nop->i_class = CLASS_NOP;
    nop->i_type = TYPE_NONE;
    nop->i_length = 1;
    nop->i_address = MAGIC_NUMBER;

    return nop;
}



/*
 *  xInstrCreateJMP() --
 *
 *  Creates and returns an unconditional jump instruction.
 */
Instruction *xInstrCreateJMP() {
    Instruction *instr = InstrCreate();
    Operand *op;

    op = OperandCreate();
    OP_SET_FLAG(op, OPERAND_DISP_R_FLAG);
    OP_SET_FLAG(op, OPERAND_VMODE_FLAG);
    OP_DISP_A(op, 0);

    instr->i_class = CLASS_JUMP;
    instr->i_type = TYPE_UNCONDITIONAL;
    instr->i_src = op;
    instr->i_address = MAGIC_NUMBER;

    INSTR_SET_FLAG(instr, SOURCE1_FLAG);

    return instr;
}



/*
 *  xInstrCreateSave() --
 *
 *  Creates and returns an instruction that will save the state of
 *  the processor status word (or the processor status register).
 *  The x86 implementation is to push the EFLAGS register onto the
 *  stack.
 */
Instruction *xInstrCreateSave() {
    Instruction *instr = InstrCreate();

    instr->i_class = CLASS_PUSHF;
    instr->i_type = TYPE_NONE;
    instr->i_address = MAGIC_NUMBER;
    instr->i_length = 1;

    return instr;
}



/*
 *  xInstrCreateRestore() --
 *
 *  Creates and returns an instruction that will restore the state of the
 *  processor status word (or the processor status register). The x86
 *  implementation involves pushing the top of the stack into the EFLAGS
 *  register.
 */
Instruction *xInstrCreateRestore() {
    Instruction *instr = InstrCreate();

    instr->i_class = CLASS_POPF;
    instr->i_type = TYPE_NONE;
    instr->i_address = MAGIC_NUMBER;
    instr->i_length = 1;

    return instr;
}



/*
 *  _InstrCreate32BitMemINC(addr) --
 *
 *  Creates and returns an instruction that will increment a 32 bit value
 *  at a given memory address by 1.
 */
Instruction *_InstrCreate32BitMemINC(ADDRESS addr) {
    Instruction *instr = InstrCreate();
    Operand *op;

    op = OperandCreate();

    instr->i_class = CLASS_INC;
    instr->i_type = TYPE_NONE;
    instr->i_address = MAGIC_NUMBER;
    INSTR_SET_FLAG(instr, DEST_FLAG);
    instr->i_dest = op;

    OP_SET_FLAG(op, OPERAND_VMODE_FLAG);
    OP_DISP_A(op, 0);

    return instr;
}



/*
 *  _InstrCreate32BitMemAdd(addr) --
 *
 *  Creates and returns an instruction that will add to a 32 bit value
 *  to a given memory address by.
 */
Instruction *_InstrCreate32BitMemAdd(ADDRESS addr) {
    Instruction *instr = InstrCreate();
    UINT8 buffer[16];

    memset(buffer, 0, 16);
    buffer[0] = 0x81;
    buffer[1] = 0x05;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  _InstrCreate32BitMemAdc(addr) --
 *
 *  Creates and returns an instruction that will add to a 32 bit value
 *  along with the carry flag if it is set, to a given memory address.
 */
Instruction *_InstrCreate32BitMemAdc(ADDRESS addr) {
    Instruction *instr = InstrCreate();
    UINT8 buffer[16];

    memset(buffer, 0, 16);
    buffer[0] = 0x83;
    buffer[1] = 0x15;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateCALLDisp() --
 *
 *  Creates and returns a CALL instruction that can be used with relative
 *  displacements.
 */
Instruction *xInstrCreateCALLDisp() {
    Instruction *call = InstrCreate();
    Operand *op;

    op = OperandCreate();
    call->i_type = TYPE_UNCONDITIONAL;
    call->i_class = CLASS_CALL;
    call->i_src = op;
    INSTR_SET_FLAG(call, SOURCE1_FLAG);
    call->i_address = MAGIC_NUMBER;

    OP_SET_FLAG(op, OPERAND_VMODE_FLAG);
    OP_DISP_R(op, 0);

    return call;
}



/*  
 *  xInstrCreateCallNearMemBase() --
 *
 *  Creates and returns a call instruction that uses
 */
Instruction *xInstrCreateCallNearMemBase(int reg) {
    UINT8 buf[16];
    Instruction *instr = InstrCreate();

    CASSERT(((reg > 32) && (reg <= 38)),
        "Invalid register for indirect call.\n");

    buf[0] = 0xff;
    buf[1] = 0x10;

    instr->i_length = xDisassembleInstr(0, buf, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;

    return instr;
}



/*
 *  xInstrCreateMovBPToSP() --
 *
 *  Creates and returns an instruction that will move the frame pointer
 *  to the stack pointer.
 */
Instruction *xInstrCreateMovBPToSP() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x89;
    buffer[1] = 0xec;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  _InstrCreate1ByteDoNothing() --
 *
 *  Creates and returns an instruction taking up 1 byte
 *  that will have no effect on the program.
 */
Instruction *_InstrCreate1ByteDoNothing() {
    return xInstrCreateNOP();
}



/*
 *  _InstrCreate2ByteDoNothing() --
 *
 *  Creates and returns an instruction taking up 2 bytes
 *  that will have no effect on the program.
 */
Instruction *_InstrCreate2ByteDoNothing() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x89;
    buffer[1] = 0xf6;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  _InstrCreate3ByteDoNothing() --
 *
 *  Creates and returns an instruction taking up 3 bytes
 *  that will have no effect on the program.
 */
Instruction *_InstrCreate3ByteDoNothing() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8d;
    buffer[1] = 0x76;
    buffer[2] = 0x00;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  _InstrCreate4ByteDoNothing() --
 *
 *  Creates and returns an instruction taking up 4 bytes
 *  that will have no effect on the program.
 */
Instruction *_InstrCreate4ByteDoNothing() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8d;
    buffer[1] = 0x74;
    buffer[2] = 0x26;
    buffer[3] = 0x00;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  _InstrCreate7ByteDoNothing() --
 *
 *  Creates and returns an instruction taking up 7 bytes
 *  that will have no effect on the program.
 */
Instruction *_InstrCreate7ByteDoNothing() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8d;
    buffer[1] = 0xb4;
    buffer[2] = 0x26;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;
    buffer[6] = 0x00;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateSPInc() --
 *
 *  Creates and returns an instruction that will shrink (increment) the stack.
 *  It is left to the calling function to fill in the amount of shrinkage.
 */
Instruction *xInstrCreateSPInc() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0xc4;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateSPDec() --
 *
 *  Creates and returns an instruction that will grow (decrement) the stack.
 *  It is left to the calling function to fill in the amount of growth.
 */
Instruction *xInstrCreateSPDec() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0xec;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  _InstrCreate32BitMovMemToReg(reg) --
 */
Instruction *_InstrCreate32BitMovMemToReg(int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8b;
    buffer[1] = 0x25;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = reg;
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  _InstrCreate32BitMovRegToMem(reg) --
 */
Instruction *_InstrCreate32BitMovRegToMem(int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x89;
    buffer[1] = 0x1d;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;

    return instr;
}



/*
 *  _InstrCreate32BitAddToReg(reg,value) --
 */
Instruction *_InstrCreate32BitAddToReg(int reg, int value) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0xc0;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = reg;
    instr->i_src->a_displacement = value;
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateMovImmToReg(reg,value) --
 */
Instruction *xInstrCreateMovImmToReg(int reg, int value) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    if (RegisterIs8Bit(reg)) {
	buffer[0] = 0xb0;
	buffer[1] = 0x00;
	buffer[2] = 0x00;
	buffer[3] = 0x00;
	buffer[4] = 0x00;
    } else if (RegisterIs16Bit(reg)) {
	buffer[0] = 0x66;		// operand size prefix
	buffer[1] = 0xb8;
	buffer[2] = 0x00;
	buffer[3] = 0x00;
	buffer[3] = 0x00;
	buffer[4] = 0x00;
    } else {
	buffer[0] = 0xb8;
	buffer[1] = 0x00;
	buffer[2] = 0x00;
	buffer[3] = 0x00;
	buffer[4] = 0x00;
    }

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = reg;
    instr->i_src->a_displacement = value;
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreatePushAll() --
 */
Instruction *xInstrCreatePushAll() {
    Instruction *instr = InstrCreate();

    instr->i_class = CLASS_PUSHALL;
    instr->i_type = TYPE_NONE;
    instr->i_address = MAGIC_NUMBER;


    return instr;
}



/*
 *  xInstrCreateBPAdd() --
 *
 *
 *  InstrCreateBPAdd
 *  Creates and returns an instruction that will add an immediate value to
 *  the frame pointer.
 */
Instruction *xInstrCreateBPAdd() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0xc5;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreatePush(reg) --
 */
Instruction *xInstrCreatePush(int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    CASSERT(((reg > 32) && (reg <= 40)), "Invalid register to push");

    buffer[0] = 0x55;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;

    return instr;
}



/*
 *  xInstrCreatePushStack(offset) --
 *
 *  XXX hack, fix this.
 */
Instruction *xInstrCreatePushStack(int offset) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();


    buffer[0] = 0xff;
    buffer[1] = 0x75;
    storeLE(&buffer[2], offset);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    instr->i_src->a_base_reg = 0x7 | reg_32;

    return instr;
}



/*
 *  xInstrCreateCmpStackToValue(offset,value) --
 *
 *  XXX hack, fix this.
 */
Instruction *xInstrCreateCmpStackToValue(int offset, int value) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0xbd;
    storeLE(&buffer[2], offset);
    storeLE(&buffer[6], value);


    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    instr->i_src->a_base_reg = 0x7 | reg_32;

    return instr;
}

Instruction *xInstrCreateCmpStackStackToValue(int offset, int value) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0x7c;
    // storeLE(&buffer[2], offset);
    // storeLE(&buffer[4], value);


    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    instr->i_src->a_base_reg = 0x8 | reg_32;
    instr->i_src->a_displacement = offset;    
    instr->i_src2->a_displacement = value;

    return instr;
}

Instruction *xInstrCreateCmpMemToValue() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0x3d;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    return instr;    
}


/*
 *  xInstrCreatePop(reg) --
 */
Instruction *xInstrCreatePop(int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    CASSERT(((reg > 32) && (reg <= 40)), "Invalid register to pop");

    buffer[0] = 0x5b;
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    instr->i_dest->a_base_reg = reg;

    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreatePopAll() --
 */
Instruction *xInstrCreatePopAll() {
    Instruction *instr = InstrCreate();

    instr->i_class = CLASS_POPALL;
    instr->i_type = TYPE_NONE;
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateCmpRegToValue(reg,value) --
 */
Instruction *xInstrCreateCmpRegToValue(int reg, int value) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0xfd;

    storeLE(&buffer[2], value);

    instr->i_length = xDisassembleInstr(0, buffer, instr);


    /* set the register to be the correct one here */
    instr->i_src->a_base_reg = reg | reg_32;

    instr->i_src2->a_displacement = value;

    instr->i_address = MAGIC_NUMBER;

    return instr;

}



/*
 *  xInstrCreateCmpImm8WithReg(imm8, reg) --
 */
Instruction *xInstrCreateCmpImm8WithReg(INT8 imm8, int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x83;
    buffer[1] = 0xf8;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src2->a_displacement = imm8;

    return instr;
}



/*
 *  xInstrCreateCmpImm32WithReg(imm32, reg) --
 */
Instruction *xInstrCreateCmpImm32WithReg(INT32 imm32, int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0xf8;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src2->a_displacement = imm32;

    return instr;
}



/*
 *  xInstrCreateCmpImm8WithMemAbs(imm8, offset) --
 */
Instruction *xInstrCreateCmpImm8WithMemAbs(INT8 imm8, int offset) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x83;
    buffer[1] = 0x3d;

    storeLE(&buffer[2], offset);
    storeLE(&buffer[6], imm8);
  
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateCmpImm32WithMemAbs(imm32, offset) --
 */
Instruction *xInstrCreateCmpImm32WithMemAbs(INT32 imm32, int offset) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0x3d;

    storeLE(&buffer[2], offset);
    storeLE(&buffer[6], imm32);
  
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateCmpImm8WithMemBase(imm8, reg) --
 */
Instruction *xInstrCreateCmpImm8WithMemBase(INT8 imm8, int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x83;
    buffer[1] = 0x38;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src2->a_displacement = imm8;

    return instr;
}



/*
 *  xInstrCreateCmpImm32WithMemBase(imm32, reg) --
 */
Instruction *xInstrCreateCmpImm32WithMemBase(INT32 imm32, int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0x38;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src2->a_displacement = imm32;

    return instr;
}



/*
 *  xInstrCreateCmpImm8WithMemDisp8Base(imm8, disp8, reg) --
 */
Instruction *xInstrCreateCmpImm8WithMemDisp8Base(INT8 imm8, INT8 disp8,
    int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x83;
    buffer[1] = 0x78;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src->a_displacement = disp8;
    instr->i_src2->a_displacement = imm8;

    return instr;
}



/*
 *  xInstrCreateCmpImm32WithMemDisp8Base(imm32, disp8, reg) --
 */
Instruction *xInstrCreateCmpImm32WithMemDisp8Base(INT32 imm32, INT8 disp8,
    int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0x78;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src->a_displacement = disp8;
    instr->i_src2->a_displacement = imm32;

    return instr;
}



/*
 *  xInstrCreateCmpImm8WithMemDisp32Base(imm8, disp32, reg) --
 */
Instruction *xInstrCreateCmpImm8WithMemDisp32Base(INT8 imm8, INT32 disp32,
    int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x83;
    buffer[1] = 0xb8;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src->a_displacement = disp32;
    instr->i_src2->a_displacement = imm8;

    return instr;
}



/*
 *  xInstrCreateCmpImm32WithMemDisp32Base(imm32, disp32, reg) --
 */
Instruction *xInstrCreateCmpImm32WithMemDisp32Base(INT32 imm32, INT32 disp32,
    int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0xb8;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src->a_displacement = disp32;
    instr->i_src2->a_displacement = imm32;

    return instr;
}



/*
 *  xInstrCreateCmpImm8WithMemBaseIndex(imm8, base, index, scale) --
 */
Instruction *xInstrCreateCmpImm8WithMemBaseIndex(INT8 imm8, int base,
    int index, int scale) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x83;
    buffer[1] = 0x3c;
    buffer[2] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src2->a_displacement = imm8;
    
    instr->i_src->a_base_reg = base;
    instr->i_src->a_index_reg = index;
    instr->i_src->a_scale = scale;

    return instr;
}



/*
 *  xInstrCreateCmpImm32WithMemBaseIndex(imm32, base, index, scale) --
 */
Instruction *xInstrCreateCmpImm32WithMemBaseIndex(INT32 imm32, int base,
    int index, int scale) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();
    
    buffer[0] = 0x81;
    buffer[1] = 0x3c;
    buffer[2] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src2->a_displacement = imm32;

    instr->i_src->a_base_reg = base;
    instr->i_src->a_index_reg = index;
    instr->i_src->a_scale = scale;

    return instr;
}



/*
 *  xInstrCreateJne() --
 */
Instruction *xInstrCreateJne() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x0f;
    buffer[1] = 0x85;
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateJe() --
 */
Instruction *xInstrCreateJe() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x0f;
    buffer[1] = 0x84;
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateCmovRegToReg() --
 *
 *  Creates and returns a conditional register-to-register move instruction.
 */
Instruction *xInstrCreateCmovRegToReg() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x0f;
    buffer[1] = 0x49;
    buffer[2] = 0xe5;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateCmovMemToReg() --
 *
 *  Creates and returns a conditional memory-to-register move instruction.
 */
Instruction *xInstrCreateCmovMemToReg() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x0f;
    buffer[1] = 0x49;
    buffer[2] = 0x75;
    buffer[3] = 0x0c;
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreatePushImm(value) --
 *
 *  Creates a push that puts an immediate value on the stack.
 */
Instruction *xInstrCreatePushImm(int value) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x68;
    buffer[1] = 0x0;
    buffer[2] = 0x0;
    buffer[3] = 0x0;
    buffer[4] = 0x0;

    storeLE(&buffer[1], value);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreatePushMem(address) --
 *
 *  Creates a push that puts the contents of a memory location onto the stack.
 */
Instruction *xInstrCreatePushMem(int address) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xff;
    buffer[1] = 0x35;
    buffer[2] = 0x0;
    buffer[3] = 0x0;
    buffer[4] = 0x0;
    buffer[5] = 0x0;

    storeLE(&buffer[2], address);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreatePushIndMem(reg) --
 */
Instruction *xInstrCreatePushIndMem(int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xff;
    buffer[1] = 0x30;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    return instr;
}



/*
 *  xInstrCreatePushMemBase(reg) --
 */
Instruction *xInstrCreatePushMemBase(int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xff;
    buffer[1] = 0x34;
    buffer[2] = 0x20;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    return instr;
}



/*
 *  xInstrCreatePushIndMemDisp(reg,value) --
 */
Instruction *xInstrCreatePushIndMemDisp(int reg, int value) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xff;
    buffer[1] = 0x74;
    buffer[2] = 0x24;
    buffer[3] = (INT8) value;


    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    OP_DISP_R(instr->i_src, value);
    return instr;
}



/*
 *  xInstrCreatePushMemDisp8Base(disp8, reg) --
 */
Instruction *xInstrCreatePushMemDisp8Base(INT8 disp8, int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xff;
    buffer[1] = 0x74;
    buffer[2] = 0x20;
    
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src->a_displacement = disp8;

    return instr;
}   



/*
 *  xInstrCreatePushMemDisp32Base(disp32, reg) --
 */
Instruction *xInstrCreatePushMemDisp32Base(INT32 disp32, int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xff;
    buffer[1] = 0xb4;
    buffer[2] = 0x20;
    
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src->a_displacement = disp32;

    return instr;
} 



/*
 *  xInstrCreateAddImmToIndirectReg(imm_value,reg) --
 */
Instruction *xInstrCreateAddImmToIndirectReg(int imm_value, int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0x04;
    buffer[2] = 0x24;
    buffer[3] = 0x0;
    buffer[4] = 0x0;
    buffer[5] = 0x0;
    buffer[6] = 0x0;

    storeLE(&buffer[3], imm_value);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = reg;

    return instr;

}



/*
 *  xInstrCreateSafeStackAdd() --
 *
 *  Decrement the %esp register without affecting EFLAGS.
 *
 *  leal -16(%esp), %esp
 */
Instruction *xInstrCreateSafeStackAdd() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8d;
    buffer[1] = 0x64;
    buffer[2] = 0x24;
    buffer[3] = 0xf0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  _InstrCreate32BitLea() --
 */
Instruction *_InstrCreate32BitLea() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8d;
    buffer[1] = 0x8c;
    buffer[3] = 0x98;
    buffer[4] = 0x00;
    buffer[5] = 0x00;
    buffer[6] = 0x00;
    buffer[7] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;


    return instr;
}



/*
 *  _InstrCreate32BitAnd(reg,value) --
 */
Instruction *_InstrCreate32BitAnd(int reg, int value) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x25;
    buffer[1] = 0x00;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;

    storeLE(&buffer[1], value);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*
 *  _InstrCreate32BitShlImm(reg,times) --
 */
Instruction *_InstrCreate32BitShlImm(UINT32 reg, int times) {
    UINT8 buffer[3];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xc1;
    buffer[1] = 0xe0;
    buffer[2] = times;

    instr->i_length = xDisassembleInstr(0, buffer, instr);

    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*
 *  _InstrCreate32BitShr(reg,value) --
 */
Instruction *_InstrCreate32BitShr(int reg, int value) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xc1;
    buffer[1] = 0xe8;
    buffer[2] = (UINT8) value;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*
 *  xInstrCreate32BitShlCL(reg) --
 */
Instruction *xInstrCreate32BitShlCL(UINT32 reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xd3;
    buffer[1] = 0xe0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*
 *  xInstrCreate32BitShrCL(reg) --
 */
Instruction *xInstrCreate32BitShrCL(UINT32 reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xd3;
    buffer[1] = 0xe8;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*  
 *  xInstrCreateAddRegToReg(src_reg, dest_reg) --
 */
Instruction *xInstrCreateAddRegToReg(int src_reg, int dest_reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x01;
    buffer[1] = 0xc0;
    
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src_reg;
    instr->i_dest->a_base_reg = dest_reg;

    return instr;
}



/*  
 *  xInstrCreateAddImm8ToReg(imm8, dest_reg) --
 */
Instruction *xInstrCreateAddImm8ToReg(INT8 imm8, int dest_reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x83;
    buffer[1] = 0xc0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_displacement = imm8;
    instr->i_dest->a_base_reg = dest_reg;

    return instr;
}



/*  
 *  xInstrCreateAddImm32ToReg(imm32, dest_reg) --
 */
Instruction *xInstrCreateAddImm32ToReg(INT32 imm32, int dest_reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();
    
    buffer[0] = 0x81;
    buffer[1] = 0xc0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_displacement = imm32;
    instr->i_dest->a_base_reg = dest_reg;

    return instr;
}



/*  
 *  xInstrCreateSubImm8FromReg(imm8, dest_reg) --
 */
Instruction *xInstrCreateSubImm8FromReg(INT8 imm8, int dest_reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x83;
    buffer[1] = 0xe8;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_displacement = imm8;
    instr->i_dest->a_base_reg = dest_reg;

    return instr;
}
    


/*  
 *  xInstrCreateSubImm32FromReg(imm32, dest_reg) --
 */
Instruction *xInstrCreateSubImm32FromReg(INT32 imm32, int dest_reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();
    
    buffer[0] = 0x81;
    buffer[1] = 0xe8;
    
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_displacement = imm32;
    instr->i_dest->a_base_reg = dest_reg;

    return instr;
}



/*  
 *  xInstrCreateOrRegWithMemBaseIndex(src,base,index,scale) --
 */ 
Instruction *xInstrCreateOrRegWithMemBaseIndex(int src, int base, int index,
    UINT8 scale) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x09; 
    buffer[1] = 0x04;
    buffer[2] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    
    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = base;
    instr->i_dest->a_index_reg = index;
    instr->i_dest->a_scale = scale;

    return instr;
}


    
/*  
 *  xInstrCreateOrMemBaseIndexWithReg(base,index,scale,dest) --
 */ 
Instruction *xInstrCreateOrMemBaseIndexWithReg(int base, int index, UINT8 scale,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();
    
    buffer[0] = 0x0b;
    buffer[1] = 0x04;
    buffer[2] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = base;
    instr->i_src->a_index_reg = index;
    instr->i_src->a_scale = scale;
    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*  
 *  xInstrCreateAndImmWithReg(imm, dest_reg) --
 */
Instruction *xInstrCreateAndImmWithReg(int imm, int dest_reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    buffer[1] = 0xe0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_displacement = imm;
    instr->i_dest->a_base_reg = dest_reg;

    return instr;
}



/*
 *  shortInstrCreateXorRegWithMemDisp8Base(src, dest_disp, dest) --
 */
Instruction *shortInstrCreateXorRegWithMemDisp8Base(int src, INT8 dest_disp,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x31;
    buffer[1] = 0x40;

    storeLE(buffer + 2, dest_disp);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}

/*
 *  xInstrCreateXorRegWithMemDisp8Base(src, dest_disp, dest) --
 */
Instruction *xInstrCreateXorRegWithMemDisp8Base(int src, INT8 dest_disp,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x31;
    buffer[1] = 0x44;
    buffer[2] = 0x20;

    storeLE(buffer + 3, dest_disp);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*
 *  shortInstrCreateXorRegWithMemDisp32Base(src, dest_disp, dest) --
 */
Instruction *shortInstrCreateXorRegWithMemDisp32Base(int src, INT32 dest_disp,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x31;
    buffer[1] = 0x80;

    storeLE(buffer + 2, dest_disp);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}

/*
 *  xInstrCreateXorRegWithMemDisp32Base(src, dest_disp, dest) --
 */
Instruction *xInstrCreateXorRegWithMemDisp32Base(int src, INT32 dest_disp,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x31;
    buffer[1] = 0x84;
    buffer[2] = 0x20;

    storeLE(buffer + 3, dest_disp);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*
 *  _InstrCreate32BitMovMemDispToReg(s_reg,off,d_reg) --
 */
Instruction *_InstrCreate32BitMovMemDispToReg(int s_reg, INT8 off, int d_reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8b;
    buffer[1] = 0x44;
    buffer[2] = 0x20;
    buffer[3] = off;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = s_reg;
    instr->i_dest->a_base_reg = d_reg;

    return instr;
}



/*
 *  xInstrCreateMovRegToReg(srcreg,destreg) --
 *
 *  Creates and returns a register-to-register move.
 */
Instruction *xInstrCreateMovRegToReg(int srcreg, int destreg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x89;
    buffer[1] = 0xc0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = srcreg;
    instr->i_dest->a_base_reg = destreg;

    return instr;
}



/*
 *  xInstrCreateMovRegToMemBase(src,dest) --
 */
Instruction *xInstrCreateMovRegToMemBase(int src, int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x89;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}

/*
 *  longInstrCreateMovRegToMemBase(src,dest) --
 */ 
Instruction *longInstrCreateMovRegToMemBase(int src, int dest) {
    UINT8 buffer[16]; 
    Instruction *instr = InstrCreate();

    buffer[0] = 0x89;
    buffer[1] = 0x04;
    buffer[2] = 0x20;
    
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*
 *  xInstrCreateMovMemBaseToReg(src, dest) --
 */
Instruction *xInstrCreateMovMemBaseToReg(int src, int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8b;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}

/*
 *  longInstrCreateMovMemBaseToReg(src, dest) --
 */
Instruction *longInstrCreateMovMemBaseToReg(int src, int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8b;
    buffer[1] = 0x04;
    buffer[2] = 0x20;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*
 *  xInstrCreateMovRegToMemDisp8Base(src,dest_disp,dest) --
 */
Instruction *xInstrCreateMovRegToMemDisp8Base(int src, INT8 dest_disp,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x89;
    buffer[1] = 0x44;
    buffer[2] = 0x20;
    
    storeLE(buffer + 3, dest_disp);
    
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*
 *  xInstrCreateMovRegToMemDisp32Base(src,dest_disp,dest) --
 */
Instruction *xInstrCreateMovRegToMemDisp32Base(int src, INT32 dest_disp,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x89;
    buffer[1] = 0x84;
    buffer[2] = 0x20;

    storeLE(buffer + 3, dest_disp);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*
 *  shortInstrCreateMovMemDisp8BaseToReg(src_disp,src,dest) --
 */
Instruction *shortInstrCreateMovMemDisp8BaseToReg(INT8 src_disp, int src,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8b;
    buffer[1] = 0x40;

    storeLE(buffer + 2, src_disp);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}

/*
 *  xInstrCreateMovMemDisp8BaseToReg(src_disp,src,dest) --
 */
Instruction *xInstrCreateMovMemDisp8BaseToReg(INT8 src_disp, int src,
    int dest) { 
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8b;
    buffer[1] = 0x44;
    buffer[2] = 0x20;

    storeLE(buffer + 3, src_disp);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*
 *  shortInstrCreateMovMemDisp32BaseToReg(src_disp,src,dest) --
 */
Instruction *shortInstrCreateMovMemDisp32BaseToReg(INT32 src_disp, int src,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8b;
    buffer[1] = 0x80;

    storeLE(buffer + 2, src_disp);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}

/*
 *  xInstrCreateMovMemDisp32BaseToReg(src_disp,src,dest) --
 */
Instruction *xInstrCreateMovMemDisp32BaseToReg(INT32 src_disp, int src,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8b;
    buffer[1] = 0x84;
    buffer[2] = 0x20;

    storeLE(buffer + 3, src_disp);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*
 *  xInstrCreateMovRegToMemBaseIndex(src,base,index,scale) --
 */
Instruction *xInstrCreateMovRegToMemBaseIndex(int src, int base, int index,
    UINT8 scale) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x89;
    buffer[1] = 0x04;
    buffer[2] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src;
    instr->i_dest->a_base_reg = base;
    instr->i_dest->a_index_reg = index;
    instr->i_dest->a_scale = scale;

    return instr;
}



/*
 *  xInstrCreateMovMemBaseIndexToReg(base,index,scale,dest) --
 */
Instruction *xInstrCreateMovMemBaseIndexToReg(int base, int index, UINT8 scale,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8b;
    buffer[1] = 0x04;
    buffer[2] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = base;
    instr->i_src->a_index_reg = index;
    instr->i_src->a_scale = scale;
    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*  
 *  xInstrCreateMovImmToMemBase(imm, dest_reg) --
 */
Instruction *xInstrCreateMovImmToMemBase(int imm, int dest_reg) { 
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xc7;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_displacement = imm;
    instr->i_dest->a_base_reg = dest_reg;

    return instr;
} 



/*
 *  xInstrCreateMovImmToMemDisp8Base(imm,disp,dest) --
 */
Instruction *xInstrCreateMovImmToMemDisp8Base(int imm, INT8 disp,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xc7;
    buffer[1] = 0x44;
    buffer[2] = 0x20;

    storeLE(buffer + 3, disp);
    storeLE(buffer + 4, imm);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*
 *  xInstrCreateMovImmToMemDisp32Base(imm,disp,dest) --
 */
Instruction *xInstrCreateMovImmToMemDisp32Base(int imm, INT32 disp,
    int dest) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xc7;
    buffer[1] = 0x84;
    buffer[2] = 0x20;

    storeLE(buffer + 3, disp);
    storeLE(buffer + 7, imm);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = dest;

    return instr;
}



/*
 *  xInstrCreateMovImmToMemBaseIndex(imm,base,index,scale) --
 */
Instruction *xInstrCreateMovImmToMemBaseIndex(int imm, int base, int index,
    int scale) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xc7;
    buffer[1] = 0x04;
    buffer[2] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_displacement = imm;

    instr->i_dest->a_base_reg = base;
    instr->i_dest->a_index_reg = index;
    instr->i_dest->a_scale = scale;

    return instr;
}



/*
 *  xInstrCreateMovzReg8ToReg32(src_reg, dest_reg) --
 */
Instruction *xInstrCreateMovzReg8ToReg32(UINT32 src_reg, UINT32 dest_reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x0f;
    buffer[1] = 0xb6;
    buffer[2] = 0xc0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src_reg;
    instr->i_dest->a_base_reg = dest_reg;

    return instr;
}



/*
 *  xInstrCreateIMULRegReg(src_reg,dest_reg) --
 */
Instruction *xInstrCreateIMULRegReg(UINT32 src_reg, UINT32 dest_reg) {
    UINT8 buffer[3];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x0f;
    buffer[1] = 0xaf;
    buffer[2] = 0xc0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);

    instr->i_src->a_base_reg = src_reg;
    instr->i_dest->a_base_reg = dest_reg;

    return instr;
}



/*
 *  xInstrCreateSubRegFromReg(src_reg,dest_reg) --
 */
Instruction *xInstrCreateSubRegFromReg(UINT32 src_reg, UINT32 dest_reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x29;
    buffer[1] = 0xc0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_dest->a_base_reg = dest_reg;
    instr->i_src->a_base_reg = src_reg;

    return instr;
}					// InstrCreateSubRegFromReg



#define RegTox86Reg(x) \
    (((x)&0xf)==reg_a ? 0 : ((x)&0xf)==reg_c ? 1 : ((x)&0xf)==reg_d ? 2 :  \
    ((x)&0xf)==reg_b ? 3 : ((x)&0xf)==reg_sp ? 4 : ((x)&0xf)==reg_bp ? 5 :  \
    ((x)&0xf)==reg_si ? 6 : ((x)&0xf)==reg_di ? 7 : -1)



/*
 *  xInstrCreateInc(reg) --
 */
Instruction *xInstrCreateInc(UINT32 reg) {
    UINT8 buffer[1];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x40 + RegTox86Reg(reg);

    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateDec(reg) --
 */
Instruction *xInstrCreateDec(UINT32 reg) {
    UINT8 buffer[1];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x48 + RegTox86Reg(reg);

    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateCmpRegToReg(reg1,reg2) --
 */
Instruction *xInstrCreateCmpRegToReg(UINT32 reg1, UINT32 reg2) {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x39;
    buffer[1] = 0xc0 | (RegTox86Reg(reg1) << 3) | RegTox86Reg(reg2);

    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateAddRegToReg1(reg1,reg2) --
 */
Instruction *xInstrCreateAddRegToReg1(UINT32 reg1, UINT32 reg2) {
    UINT8 buffer[2];
    Instruction *instr;

    buffer[0] = 0x03;
    buffer[1] = 0xc0 | (RegTox86Reg(reg2) << 3) | RegTox86Reg(reg1);

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateBitTest(reg,position) --
 */
Instruction *xInstrCreateBitTest(UINT32 reg, int position) {
    UINT8 buffer[4];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0xba;
    buffer[2] = (3 << 6) | (4 << 3) | RegTox86Reg(reg);
    buffer[3] = position;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateShortJb() --
 */
Instruction *xInstrCreateShortJb() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x72;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJae() --
 */
Instruction *xInstrCreateShortJae() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x73;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJe() --
 */
Instruction *xInstrCreateShortJe() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x74;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJne() --
 */
Instruction *xInstrCreateShortJne() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x75;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJbe() --
 */
Instruction *xInstrCreateShortJbe() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x76;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJa() --
 */
Instruction *xInstrCreateShortJa() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x77;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJs() --
 */
Instruction *xInstrCreateShortJs() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x78;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJns() --
 */
Instruction *xInstrCreateShortJns() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x79;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJl() --
 */
Instruction *xInstrCreateShortJl() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x7c;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJge() --
 */
Instruction *xInstrCreateShortJge() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x7d;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJle() --
 */
Instruction *xInstrCreateShortJle() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x7e;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJg() --
 */
Instruction *xInstrCreateShortJg() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x7f;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateShortJecxz() --
 */
Instruction *xInstrCreateShortJecxz() {
    UINT8 buffer[2];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xe3;
    buffer[1] = 0x00;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    return instr;
}



/*
 *  xInstrCreateNearJb() --
 */
Instruction *xInstrCreateNearJb() {
    UINT8 buffer[6];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0x82;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateNearJae() --
 */
Instruction *xInstrCreateNearJae() {
    UINT8 buffer[6];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0x83;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateNearJbe() --
 */
Instruction *xInstrCreateNearJbe() {
    UINT8 buffer[6];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0x86;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateNearJa() --
 */
Instruction *xInstrCreateNearJa() {
    UINT8 buffer[6];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0x87;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateNearJs() --
 */
Instruction *xInstrCreateNearJs() {
    UINT8 buffer[6];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0x88;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateNearJns() --
 */
Instruction *xInstrCreateNearJns() {
    UINT8 buffer[6];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0x89;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateNearJl() --
 */
Instruction *xInstrCreateNearJl() {
    UINT8 buffer[6];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0x8c;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateNearJge() --
 */
Instruction *xInstrCreateNearJge() {
    UINT8 buffer[6];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0x8d;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateNearJle() --
 */
Instruction *xInstrCreateNearJle() {
    UINT8 buffer[6];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0x8e;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateNearJg() --
 */
Instruction *xInstrCreateNearJg() {
    UINT8 buffer[6];
    Instruction *instr;

    buffer[0] = 0x0f;
    buffer[1] = 0x8f;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  xInstrCreateMovMemToReg(scale,index_reg,base_reg,dest_reg) --
 */
Instruction *xInstrCreateMovMemToReg(int scale,
    UINT32 index_reg, UINT32 base_reg, UINT32 dest_reg) {
    UINT8 buffer[16];
    Instruction *instr;

    buffer[0] = 0x8b;
    buffer[1] = 0x04;
    if (base_reg != -1) {
	buffer[1] |= 0x80;
    }
    buffer[2] = 0x05;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;
    buffer[6] = 0x00;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    instr->i_src->a_index_reg = index_reg;
    instr->i_dest->a_base_reg = dest_reg;
    instr->i_src->a_scale = scale;

    if (base_reg != -1) {
	instr->i_src->a_base_reg = base_reg;
    }

    return instr;
}



/*
 *  xInstrCreateSubImmFromReg(imm_value,reg) --
 */
Instruction *xInstrCreateSubImmFromReg(int imm_value, UINT32 reg) {
    UINT8 buffer[6];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x81;
    //          mod = 11b      reg = 101b
    buffer[1] = (0x3 << 6) | (0x5 << 3);
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    storeLE(buffer + 2, imm_value);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*
 *  xInstrCreateMovAbsoluteMemToReg(memloc,reg) --
 */
Instruction *xInstrCreateMovAbsoluteMemToReg(ADDRESS memloc, int reg) {
    UINT8 buffer[6];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8b;
    //          mod = 00b, reg = reg, r/m = 101b
    buffer[1] = 0x05;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    storeLE(buffer + 2, memloc);

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*
 *  _InstrCreate32BitMovImmToMem(mem_loc,imm) --
 */
Instruction *_InstrCreate32BitMovImmToMem(ADDRESS mem_loc, int imm) {
    Instruction *instr;
    UINT8 buffer[10];

    instr = InstrCreate();
    buffer[0] = 0xc7;
    buffer[1] = 0x05;
    storeLE(buffer + 2, mem_loc);
    storeLE(buffer + 6, imm);

    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  _InstrCreate32BitAddRegToMem(mem_loc,reg) --
 */
Instruction *_InstrCreate32BitAddRegToMem(ADDRESS mem_loc, int reg) {
    Instruction *instr;
    UINT8 buffer[6];

    instr = InstrCreate();
    buffer[0] = 0x01;
    buffer[1] = 0x05;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_src->a_base_reg = reg_32 | reg;
    instr->i_dest->a_displacement = mem_loc;

    return instr;
}



/*
 *  _InstrCreate32BitSubRegFromMem(mem_loc,reg) --
 */
Instruction *_InstrCreate32BitSubRegFromMem(ADDRESS mem_loc, int reg) {
    Instruction *instr;
    UINT8 buffer[6];

    instr = InstrCreate();
    buffer[0] = 0x29;
    buffer[1] = 0x05;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_src->a_base_reg = reg_32 | reg;
    instr->i_dest->a_displacement = mem_loc;

    return instr;
}



/*
 *  _InstrCreate32BitAddImmToMem(mem_loc,imm) --
 */
Instruction *_InstrCreate32BitAddImmToMem(ADDRESS mem_loc, int imm) {
    Instruction *instr;
    UINT8 buffer[10];

    buffer[0] = 0x81;
    buffer[1] = 0x05;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_src->a_displacement = imm;
    instr->i_dest->a_displacement = mem_loc;

    return instr;
}



/*
 *  _InstrCreate32BitAddMemToReg(mem_loc,reg) --
 */
Instruction *_InstrCreate32BitAddMemToReg(ADDRESS mem_loc, UINT32 reg) {
    Instruction *instr;
    UINT8 buffer[16];

    buffer[0] = 0x03;
    buffer[1] = 0x05;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_src->a_displacement = mem_loc;
    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*
 *  xInstrCreateSubMemFromReg(mem_loc,reg) --
 */
Instruction *xInstrCreateSubMemFromReg(ADDRESS mem_loc, UINT32 reg) {
    Instruction *instr;
    UINT8 buffer[16];

    buffer[0] = 0x2b;
    buffer[1] = 0x05;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    instr->i_src->a_displacement = mem_loc;
    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*
 *  xInstrCreateSubImmFromMem(imm,mem_loc) --
 */
Instruction *xInstrCreateSubImmFromMem(int imm, ADDRESS mem_loc) {
    Instruction *instr;
    UINT8 buffer[16];

    buffer[0] = 0x81;
    buffer[1] = 0x2d;

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_src->a_displacement = imm;
    instr->i_src->a_displacement = mem_loc;

    return instr;
}



/*
 *  _InstrCreate16BitAddRegToMem(mem_loc,reg) --
 */
Instruction *_InstrCreate16BitAddRegToMem(ADDRESS mem_loc, int reg) {
    Instruction *instr;
    UINT8 buffer[16];

    instr = InstrCreate();
    buffer[0] = 0x66;
    buffer[1] = 0x01;
    buffer[2] = 0x05;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_src->a_base_reg = reg_16 | reg;
    instr->i_dest->a_displacement = mem_loc;

    return instr;
}



/*
 *  _InstrCreate16BitSubRegFromMem(mem_loc,reg) --
 */
Instruction *_InstrCreate16BitSubRegFromMem(ADDRESS mem_loc, int reg) {
    Instruction *instr;
    UINT8 buffer[16];

    instr = InstrCreate();
    buffer[0] = 0x66;
    buffer[1] = 0x29;
    buffer[2] = 0x05;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_src->a_base_reg = reg_16 | reg;
    instr->i_dest->a_displacement = mem_loc;

    return instr;
}



/*
 *  xInstrCreateLeave() --
 */
Instruction *xInstrCreateLeave() {
    Instruction *instr;
    UINT8 buffer[1];
    
    buffer[0] = 0xc9;
    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    
    return instr;
}



/*
 *  xInstrCreateReturn() --
 */
Instruction *xInstrCreateReturn() {
    Instruction *instr;
    UINT8 buffer[1];

    buffer[0] = 0xc3;
    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buffer, instr);

    return instr;
}



/*
 *  _InstrCreate32BitAddImmToReg(imm,reg) --
 */
Instruction *_InstrCreate32BitAddImmToReg(int imm, int reg) {
    UINT8 buffer[16];
    Instruction *instr;

    buffer[0] = 0x81;
    buffer[1] = 0xc0;
    buffer[2] = 0x00;
    buffer[3] = 0x00;
    buffer[4] = 0x00;
    buffer[5] = 0x00;

    instr = InstrCreate();

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_src->a_displacement = imm;
    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*
 *  xInstrCreateHalt() --
 */
Instruction *xInstrCreateHalt() {
    Instruction *instr;
    UINT8 buffer[16];

    buffer[0] = 0xf4;

    instr = InstrCreate();
    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    return instr;
}



/*
 *  xInstrCreateSetCarry() --
 */
Instruction *xInstrCreateSetCarry() {
    Instruction *instr;
    UINT8 buffer[16];

    buffer[0] = 0xf9;

    instr = InstrCreate();
    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    return instr;
}



/*
 *  xInstrCreateClearCarry() --
 */
Instruction *xInstrCreateClearCarry() {
    Instruction *instr;
    UINT8 buffer[16];

    buffer[0] = 0xf8;

    instr = InstrCreate();
    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    return instr;
}



/*
 *  xInstrCreateJumpOnNoCarry() --
 */
Instruction *xInstrCreateJumpOnNoCarry() {
    Instruction *instr;
    UINT8 buffer[16];

    buffer[0] = 0x73;

    instr = InstrCreate();
    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    return instr;
}



/*
 *  xInstrCreateCallIndirect() --
 */
Instruction *xInstrCreateCallIndirect() {
    UINT8 buf[16];
    Instruction *instr;

    buf[0] = 0xff;
    buf[1] = 0x15;

    instr = InstrCreate();
    instr->i_length = DisassembleInstr(0, buf, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    return instr;
}



/*
 *  xInstrCreateJMPIndirectAbsMemAddr() --
 */
Instruction *xInstrCreateJMPIndirectAbsMemAddr() {
    UINT8 buf[16];
    Instruction *instr;

    buf[0] = 0xff;
    buf[1] = 0x25;

    instr = InstrCreate();
    instr->i_length = DisassembleInstr(0, buf, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    return instr;
}



/*
 *  _InstrCreate32BitPopToMem() --
 */
Instruction *_InstrCreate32BitPopToMem() {
    UINT8 buf[16];
    Instruction *instr;

    buf[0] = 0x8f;
    buf[1] = 0x05;

    instr = InstrCreate();
    instr->i_length = DisassembleInstr(0, buf, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    return instr;
}



/*
 *  xInstrCreateDiv(divisor) --
 */
Instruction *xInstrCreateDiv(int divisor) {
    Instruction *instr;
    UINT8 buf[] = {
	0xf7, 0xf8
    };

    instr = InstrCreate();
    instr->i_length = DisassembleInstr(0, buf, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_dest->a_base_reg = divisor;

    return instr;
}



/*
 *  xInstrCreateSyscall() --
 */
Instruction *xInstrCreateSyscall() {
    Instruction *instr;
    UINT8 buf[] = {
	0xcd, 0x80
    };

    instr = InstrCreate();
    instr->i_length = xDisassembleInstr(0, buf, instr);

    return instr;
}



/*
 *  Instruction *xInstrCreateTestRegWithReg(reg1, reg2) --
 */
Instruction *xInstrCreateTestRegWithReg(UINT32 reg1, UINT32 reg2) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x85;
    buffer[1] = 0xc0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg1;
    instr->i_src2->a_base_reg = reg2;

    return instr;
}



/*
 *  xInstrCreateTestImmWithReg(imm, reg) --
 */
Instruction *xInstrCreateTestImmWithReg(int imm, UINT32 reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xf7;
    buffer[1] = 0xc0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = reg;
    instr->i_src2->a_displacement = imm;

    return instr;
}



/*
 *  xInstrCreateLea(src_reg,offset,dest_reg)--
 */
Instruction *xInstrCreateLea(UINT32 src_reg, int offset, UINT32 dest_reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x8d;
    buffer[1] = 0x80;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_base_reg = src_reg;
    instr->i_src->a_displacement = offset;
    instr->i_dest->a_base_reg = dest_reg;

    return instr;
}



/*
 *  xInstrCreateSeteReg(reg) --
 */
Instruction *xInstrCreateSeteReg(UINT32 reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x0f;
    buffer[1] = 0x94;
    buffer[2] = 0xc0;

    instr->i_length = xDisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_dest->a_base_reg = reg;

    return instr;
}



/*
 *  xInstrCreate8bitDivByReg(divisor) --
 */
Instruction *xInstrCreate8bitDivByReg(int divisor) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xf6;
    buffer[1] = 0xf0;

    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_src->a_base_reg = divisor;

    return instr;
}



/*
 *  xInstrCreate32bitDivByReg(divisor) --
 */
Instruction *xInstrCreate32bitDivByReg(int divisor) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xf7;
    buffer[1] = 0xf0;

    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_src->a_base_reg = divisor;
    
    return instr;
}



/*
 *  xInstrCreate8bitIDivByReg(divisor) --
 */
Instruction *xInstrCreate8bitIDivByReg(int divisor) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xf6;
    buffer[1] = 0xf8;

    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_src->a_base_reg = divisor;

    return instr;
}



/*
 *  xInstrCreate32bitIDivByReg(divisor) --
 */
Instruction *xInstrCreate32bitIDivByReg(int divisor) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xf7;
    buffer[1] = 0xf8;

    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_src->a_base_reg = divisor;

    return instr;
}



/*
 *  xInstrCreate8bitMulByReg(divisor) --
 */
Instruction *xInstrCreate8bitMulByReg(int divisor) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xf6;
    buffer[1] = 0xe0;

    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_src->a_base_reg = divisor;

    return instr;
}



/*
 *  xInstrCreate32bitMulByReg(divisor) --
 */
Instruction *xInstrCreate32bitMulByReg(int divisor) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xf7;
    buffer[1] = 0xe0;

    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_src->a_base_reg = divisor;

    return instr;
}



/*
 *  xInstrCreate8bitIMulByReg(divisor) --
 */
Instruction *xInstrCreate8bitIMulByReg(int divisor) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xf6;
    buffer[1] = 0xe8;

    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_src->a_base_reg = divisor;

    return instr;
}



/*
 *  xInstrCreate32bitIMulByReg(divisor) --
 */
Instruction *xInstrCreate32bitIMulByReg(int divisor) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0xf7;
    buffer[1] = 0xe8;

    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_src->a_base_reg = divisor;

    return instr;
}

Instruction *xInstrCreateCall() {

    UINT8 buf[16];

    Instruction *instr;

    buf[0] = 0xe8;
    buf[1] = 0x00;
    buf[2] = 0x00;
    buf[3] = 0x00;
    buf[4] = 0x00;

    instr = InstrCreate();
    instr->i_length = DisassembleInstr(0, buf, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    return instr;
}


/*
 *  xInstrCreateIMulImm8ByReg(imm8, reg) --
 */
Instruction * xInstrCreateIMulImm8ByReg(INT8 imm8, int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x6b;
    buffer[1] = 0xc0;

    storeLE(&buffer[2], imm8);
   
    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_src->a_base_reg = reg;
    instr->i_dest->a_base_reg = reg;
    
    return instr;
}



/*
 *  xInstrCreateIMulImm32ByReg(imm32, reg) --
 */
Instruction * xInstrCreateIMulImm32ByReg(INT32 imm32, int reg) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    buffer[0] = 0x69;
    buffer[1] = 0xc0;

    storeLE(&buffer[2], imm32);

    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;
    InstrSetClassFlags(instr);

    instr->i_src->a_base_reg = reg;
    instr->i_dest->a_base_reg = reg;

    return instr;
} 



/*
 *  xInstrCreateJmpNearAbsDisp32Index(disp32, index, scale) --
 */
Instruction *xInstrCreateJmpNearAbsDisp32Index(INT32 disp32, int index,
    int scale) {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();

    CASSERT(((index > 32) && (index <= 39)), "Invalid index register\n");
    CASSERT(((scale == 1) || (scale == 2) || (scale == 4) || (scale == 8)),
        "Invalid scale\n");

    buffer[0] = 0xff;
    buffer[1] = 0x24;
    buffer[2] = 0x05;

    storeLE(&buffer[3], disp32);

    instr->i_length = DisassembleInstr(0, buffer, instr);
    instr->i_address = MAGIC_NUMBER;

    instr->i_src->a_index_reg = index;
    instr->i_src->a_scale = scale;

    return instr;
}



int PC_CHANGING_CLASSES[] = {
    CLASS_JUMP, CLASS_BRANCH, CLASS_RET, CLASS_CALL,
    CLASS_LOOP, CLASS_LCALL, CLASS_LJUMP, CLASS_LRET,
    CLASS_JA, CLASS_JAE, CLASS_JB, CLASS_JBE,
    CLASS_JC, CLASS_JCXZ, CLASS_JECXZ, CLASS_JE,
    CLASS_JG, CLASS_JGE, CLASS_JL, CLASS_JLE,
    CLASS_JNA, CLASS_JNAE, CLASS_JNA, CLASS_JNB,
    CLASS_JNBE, CLASS_JNC, CLASS_JNE, CLASS_JNG,
    CLASS_JNGE, CLASS_JNL, CLASS_JNLE, CLASS_JNO,
    CLASS_JNP, CLASS_JNS, CLASS_JNZ, CLASS_JO,
    CLASS_JP, CLASS_JPE, CLASS_JPO, CLASS_JS,
    CLASS_JZ, CLASS_LOOPNE, CLASS_LOOPE, CLASS_HALT,
    CLASS_INT, CLASS_INT3, CLASS_INTO, CLASS_LJUMP1,
};

int JUMP_CLASSES[] = {
    CLASS_JA, CLASS_JAE, CLASS_JB, CLASS_JBE,
    CLASS_JC, CLASS_JCXZ, CLASS_JECXZ, CLASS_JE,
    CLASS_JG, CLASS_JGE, CLASS_JL, CLASS_JLE,
    CLASS_JNA, CLASS_JNAE, CLASS_JNA, CLASS_JNB,
    CLASS_JNBE, CLASS_JNC, CLASS_JNE, CLASS_JNG,
    CLASS_JNGE, CLASS_JNL, CLASS_JNLE, CLASS_JNO,
    CLASS_JNP, CLASS_JNS, CLASS_JNZ, CLASS_JO,
    CLASS_JP, CLASS_JPE, CLASS_JPO, CLASS_JS,
    CLASS_JZ, CLASS_JUMP, CLASS_LJUMP, CLASS_LJUMP1,
    CLASS_LOOP, CLASS_LOOPNE, CLASS_LOOPE
};

int CALL_CLASSES[] = { CLASS_LCALL, CLASS_CALL };

int LOOP_CLASSES[] = { CLASS_LOOP, CLASS_LOOPNE, CLASS_LOOPE };

int RETURN_CLASSES[] = { CLASS_LRET, CLASS_RET };

int BRANCH_CLASSES[] = { CLASS_BRANCH };

int HALT_CLASSES[] = { CLASS_HALT };

int LEAVE_CLASSES[] = { CLASS_LEAVE };

int LEA_CLASSES[] = { CLASS_LEA };

int ADD_CLASSES[] = { CLASS_ADD };

int SUB_CLASSES[] = { CLASS_SUB };

int NOP_CLASSES[] = { CLASS_NOP, CLASS_FNOP };

int FLOAT_CLASSES[] = {
    CLASS_FADD, CLASS_FMUL, CLASS_FCOM, CLASS_FCOMP,
    CLASS_FSUB, CLASS_FSUBR, CLASS_FDIV, CLASS_FDIVR,
    CLASS_FLD, CLASS_FXCH, CLASS_FCMOVB, CLASS_FCMOVE,
    CLASS_FCMOVBE, CLASS_FCMOVU, CLASS_FUCOMI, CLASS_FCOMI,
    CLASS_FCMOVNB, CLASS_FCMOVNE, CLASS_FCMOVNBE, CLASS_FCMOVNU,
    CLASS_FFREE, CLASS_FST, CLASS_FSTP, CLASS_FUCOM,
    CLASS_FSUBRP, CLASS_FDIVP, CLASS_FDIVRP, CLASS_FUCOMIP,
    CLASS_FUCOMP, CLASS_FADDP, CLASS_FMULP, CLASS_FSUBP,
    CLASS_FCOMIP, CLASS_FLDENV, CLASS_FLDCW, CLASS_FNSTENV,
    CLASS_FNSTCW, CLASS_FIADD, CLASS_FIMUL, CLASS_FICOM,
    CLASS_FICOMP, CLASS_FISUB, CLASS_FISUBR, CLASS_FIDIV,
    CLASS_FIDIVR, CLASS_FILD, CLASS_FIST, CLASS_FISTP,
    CLASS_FLDL, CLASS_FRSTOR, CLASS_FNSAVE, CLASS_FNSTSW,
    CLASS_FBLD, CLASS_FBSTP, CLASS_FLDT, CLASS_FSTPT,
    CLASS_FNOP, CLASS_FCHS, CLASS_FABS, CLASS_FTST,
    CLASS_FXAM, CLASS_FLD1, CLASS_FLDL2T, CLASS_FLDL2E,
    CLASS_FLDPI, CLASS_FLDLG2, CLASS_FLDLN2, CLASS_FLDZ,
    CLASS_F2XM1, CLASS_FYL2X, CLASS_FPTAN, CLASS_FPATAN,
    CLASS_FXTRACT, CLASS_FPREM1, CLASS_FDECSTP, CLASS_FINCSTP,
    CLASS_FPREM, CLASS_FYL2XP1, CLASS_FSQRT, CLASS_FSINCOS,
    CLASS_FRNDINT, CLASS_FSCALE, CLASS_FSIN, CLASS_FCOS,
    CLASS_FUCOMPP, CLASS_FNCLEX, CLASS_FNINIT, CLASS_FCOMPP,
    CLASS_FADDS, CLASS_FMULS, CLASS_FCOMS, CLASS_FCOMPS,
    CLASS_FSUBS, CLASS_FSUBRS, CLASS_FDIVS, CLASS_FDIVRS,
    CLASS_FLDS, CLASS_FSTS, CLASS_FSTPS, CLASS_FILDLL,
    CLASS_FISTPLL, CLASS_FIMULL, CLASS_FMULL, CLASS_FIADDL,
    CLASS_FICOML, CLASS_FICOMPL, CLASS_FISUBL, CLASS_FISUBRL,
    CLASS_FIDIVL, CLASS_FIDIVRL, CLASS_FILDL, CLASS_FISTL,
    CLASS_FISTPL, CLASS_FADDL, CLASS_FCOML, CLASS_FCOMP,
    CLASS_FSUBRL, CLASS_FSUBL, CLASS_FDIVL, CLASS_FDIVRL,
    CLASS_FSTL, CLASS_FSTPL
};

int OP_OFF_CLASSES[] = { CLASS_MOV };

int OP_DIR_CLASSES[] = { CLASS_LJUMP, CLASS_CALL, CLASS_LCALL, CLASS_LJUMP1 };

int MOV_CLASSES[] = { CLASS_MOV, };

int SYSCALL_CLASSES[] = { CLASS_INT, CLASS_INT3, CLASS_INTO };



/*
 *  xInstrSetClassFlags(instr) --
 */
void xInstrSetClassFlags(Instruction *instr) {

    if (memberOf(instr->i_class, PC_CHANGING_CLASSES,
	    ARRAYSIZE(PC_CHANGING_CLASSES)))
	INSTR_SET_FLAG(instr, INSTR_CLASS_PC_CHANGING);

    if (memberOf(instr->i_class, JUMP_CLASSES, ARRAYSIZE(JUMP_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_JUMP);
    }

    if (memberOf(instr->i_class, LEA_CLASSES, ARRAYSIZE(LEA_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_LEA);
    }

    if (memberOf(instr->i_class, RETURN_CLASSES, ARRAYSIZE(RETURN_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_RETURN);
    }

    if (memberOf(instr->i_class, ADD_CLASSES, ARRAYSIZE(ADD_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_ADD);
    }

    if (memberOf(instr->i_class, SUB_CLASSES, ARRAYSIZE(SUB_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_SUB);
    }

    if (memberOf(instr->i_class, CALL_CLASSES, ARRAYSIZE(CALL_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_CALL);
    }

    if (memberOf(instr->i_class, BRANCH_CLASSES, ARRAYSIZE(BRANCH_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_BRANCH);
    }

    if (memberOf(instr->i_class, LOOP_CLASSES, ARRAYSIZE(LOOP_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_LOOP);
    }

    if (memberOf(instr->i_class, FLOAT_CLASSES, ARRAYSIZE(FLOAT_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_FLOAT);
    }

    if (memberOf(instr->i_class, HALT_CLASSES, ARRAYSIZE(HALT_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_HALT);
    }

    if (memberOf(instr->i_class, LEAVE_CLASSES, ARRAYSIZE(LEAVE_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_LEAVE);
    }

    if (memberOf(instr->i_class, OP_DIR_CLASSES, ARRAYSIZE(OP_DIR_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_OPDIR);
    }

    if (memberOf(instr->i_class, OP_OFF_CLASSES, ARRAYSIZE(OP_OFF_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_OPOFF);
    }

    if (memberOf(instr->i_class, NOP_CLASSES, ARRAYSIZE(NOP_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_NOP);
    }

    if (memberOf(instr->i_class, MOV_CLASSES, ARRAYSIZE(MOV_CLASSES))) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_MOV);
    }

    if (memberOf(instr->i_class, SYSCALL_CLASSES, ARRAYSIZE(SYSCALL_CLASSES)))
	INSTR_SET_FLAG(instr, INSTR_CLASS_SYSCALL);

    if (instr->i_class == CLASS_PUSH) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_PUSH);
    }

    if (instr->i_class == CLASS_POP) {
	INSTR_SET_FLAG(instr, INSTR_CLASS_POP);
    }

    INSTR_SET_FLAG(instr, INSTR_HAS_CLASS_FLAGS);

    return;
}

Instruction *xInstrCreatePushFlags() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();
    buffer[0] = 0x9c;  
    (instr)->i_length = xDisassembleInstr(0, buffer, instr);
    (instr)->i_address = MAGIC_NUMBER;
    return instr;
}

Instruction *xInstrCreatePopFlags() {
    UINT8 buffer[16];
    Instruction *instr = InstrCreate();
    buffer[0] = 0x9d;
    (instr)->i_length = xDisassembleInstr(0, buffer, instr);
    (instr)->i_address = MAGIC_NUMBER;
    return instr;
}

