/*
 *  x86/inline.c -- performs inlining on the x86.
 */
#include "global.h"
#include "x86-inline.h"
#include "dominator.h"
#include "utilities.h"
#include "sysdep.h"
#include "flowgraph.h"
#include "linkage.h"
#include "opt.unreachable.h"
#include "x86-classes.h"
#include "x86-registers.h"
#include "opt.layout.h"
#include "stack.analyze.h"
#include "x86-icreate.h"
#include "clone.h"

/*
 *  Normalizes a cloned function so that the base pointer isn't redefined,
 *  and LEAVEs are replaced with appropriate instructions
 */
static void CleanupClone(Instruction *callInstr, Function *fun);

/*
 *  Adds add/sub instructions before 'instr'
 *  to make the instruction seem as if it had the correct EBP
 */
static void CompensateForFramePointerUse(int offset, Instruction *instr);

/* Replaces the leave instruction with something semantically equivalent */
static void CompensateForLeave(int offset, Instruction *leave);

/* New instructions generated for syscall clustering. */
static void CleanupSyscallClone(Instruction *callInstr, Function *fun);
static void ReplaceLeaveRet(Instruction *leave);

/* Standard CALL instructions only implicitly push the RA */
#define STANDARD_CALL_RA_BYTES 0x4



/*
 *  xFunIsInlinable(aFun) --
 *
 *  Returns true if the supplied function can be inlined, false if
 *  the supplied function cannot be inlined.  We cannot inline a
 *  function if:
 *  1) The exit block has no predecessors. e.g. no return
 *  blocks in the function.  This should handle the case for
 *  tail-call optimization that we cannot do anything with.
 *  2) The function uses setjmp or longjmp
 */
bool xFunIsInlinable(Function *aFun) {
    Block *biterator;

    if (aFun->f_exit->b_pred == NULL) {
	return false;
    }

    if (FunHasEsc2(aFun)) {
	return false;
    }

    FOREACH_BBL_IN_FUN(aFun, biterator) {
	if (BBLHasHellPredecessor(biterator)) {
	    if (biterator->b_prev == NULL) {
		return false;
	    }
	    if (biterator->b_prev->b_type != BT_CALL) {
		return false;
	    }
	    if (!BBLHasHellSuccessor(biterator->b_prev)) {
		return false;
	    }
	    if (BBLNumberOfPredecessorsFromBBL(biterator, HELL) > 1) {
		return false;
	    }
	}
    }

    return true;
}



/*
 *  xFunCanBeInlinedFromCallSite(fun,callSite) --
 *
 *  Returns true if the supplied function can be inlined when called
 *  from the instruction 'callSite'.
 */
bool xFunCanBeInlinedFromCallSite(Function *fun, Instruction *callSite) {

    return FunIsInlinable(fun);
}



/*
 *  xFunInline(callInstr) -- inlines the function called by the 'callInstr'.
 */
void xFunInline(Instruction *callInstr) {
    Function *funClone;
    Function *funOrig;
    Function *callerFunction;
    Block *targetBBL;
    Edge *callEdge;
    EdgeNode *successors;
    EdgeNode *predecessors;
    Instruction *spIncrement;	// add $0x4, %esp
    Instruction *spDecrement;	// sub $0x4, %esp
    Instruction *nextInstr;	// Instr following callInstr
    Instruction *prevInstr;	// Instr preceding callInstr
    Block *biterator;
    Block *firstBlock;
    Block *newReturnPoint;
    Instruction *iterator;
    bool INCREMENT_SIMULATED = false;
    bool DECREMENT_SIMULATED = false;
    UINT32 linkWeight;
    float factor;

    iterator = xInstrCreateMovBPToSP();

    callEdge = BBLGetFunDirectEdge(callInstr->i_Bblock);
    callerFunction = INSTR_FUNCTION(callInstr);

    if (callEdge == NULL) {
	WARNING("[X86:FunInline] Trying to inline a function from"
	    "a callsite[0x%8.8x] with no ET_FUNCALL edge.",
	    callInstr->i_address);
	return;
    }

    spIncrement = spDecrement = NULL;

    targetBBL = callEdge->e_to;
    funOrig = targetBBL->b_fun;
    funClone = FunClone(funOrig);

    FOREACH_PRED_IN_BBL(funClone->f_exit, predecessors) {
	Instruction *last = (NODE_FROM(predecessors))->b_last;
	if (last->i_class == CLASS_RET) {
	    if (last->i_prev->i_class == CLASS_POP) {
		if (OperandIsFramePointer(last->i_prev->i_dest)) {
		    INSTR_SET_FLAG(last->i_prev, INSTR_DEFINES_FP_NICELY);
		}
	    }
	}
    }

    UpdateFunction(funClone);
    if (INLINE_USE_CALLERS_FP) {
	CleanupClone(callInstr, funClone);
    }
    nextInstr = callInstr->i_next;
    prevInstr = callInstr->i_prev;

    linkWeight = callEdge->e_edgeWeight;

    /*
     * We want to weight the edges of the cloned function to represent
     * what the weights may look like in an inlined version.  Note that
     * the cloned function has the same weights as the original function.
     * What we will do is multiple each weight by (callsite freq /
     * total pred freq), which is some number less than 1.  This
     * represents the frequency that the callsite we're considered may
     * contribute to the execution count of each block.  We will also
     * multiple each edge weight in the original function by 1 -
     * (callsite freq / total pred freq) to account for the adjusted
     * weights that will occur after inlining.
     */

    /* Update our cloned function */
    factor = ((float) callEdge->e_edgeWeight) / ((float) targetBBL->b_weight);
    FOREACH_BBL_IN_FUN(funClone, biterator) {
	FOREACH_SUCC_IN_BBL(biterator, successors) {
	    NODE_WEIGHT(successors) = NODE_WEIGHT(successors) * factor;
	}
	if (factor < REMOVE_HOT_FACTOR) {
	    BBL_REMOVE_FLAG(biterator, BBL_HOT_FLAG);
	    BBL_REMOVE_FLAG(biterator, BBL_HOT_CALL_SITE_FLAG);
	}
    }

    FOREACH_BBL_IN_FUN(funClone, biterator) {
	BBLUpdateWeight(biterator);
    }

    /* Update our original function */
    factor = 1.0 - factor;
    FOREACH_BBL_IN_FUN(funOrig, biterator) {
	FOREACH_SUCC_IN_BBL(biterator, successors) {
	    NODE_WEIGHT(successors) = NODE_WEIGHT(successors) * factor;
	}
	if (factor < REMOVE_HOT_FACTOR) {
	    BBL_REMOVE_FLAG(biterator, BBL_HOT_FLAG);
	    BBL_REMOVE_FLAG(biterator, BBL_HOT_CALL_SITE_FLAG);
	}
    }

    FOREACH_BBL_IN_FUN(funOrig, biterator) {
	BBLUpdateWeight(biterator);
    }

    /* Remove all ENTRY/EXIT edges from the clone */
    BBLRemoveEdges(funClone->f_entry);
    BBLRemoveIncomingEdges(funClone->f_exit);

    if (!INLINE_USE_CALLERS_FP) {

	/*
	 * We add in some stack pointer manipulations here to simulate what
	 * happens Behind the Scenes during a function call.  The return
	 * address is implicitly pushed onto the stack after a CALL has been
	 * executed, and it is implicitly popped off the stack when a
	 * corresponding RET is executed.  There are other things that can
	 * can be pushed/popped implicitly during a CALL (such as the call
	 * gate selector, perhaps some stuff to do with priv. levels), but
	 * we will ignore them for now. The idea is to grow the stack after
	 * CALL would have been executed, and shrink it after the RET would
	 * have been executed.  Then, anything else that accesses the stack
	 * will be correct when assuming that the return address is there--
	 * or at least the space for it has been allocated.  Alternatively,
	 * it may be feasible to change all instructions that reference the
	 * stack by updating their displacements. This would save us from
	 * having to add instructions, but it may be difficult to do.
	 */

	/*
	 * Maybe we can get lucky and abuse a stack pointer manipulation
	 * residing after our call instruction, instead of adding a new
	 * pointer increment. If there is already an SP add present, we can
	 * change the displacement by 4 to simulate the RA being popped.
	 * This case happens quite frequently.
	 */

	if ((nextInstr->i_class == CLASS_ADD) &&
	    (nextInstr->i_dest->a_base_reg == (reg_sp | reg_32)) &&
	    (nextInstr->i_src != NULL) &&
	    (BBLHasHellPredecessor(nextInstr->i_Bblock) == false) &&
	    (OPERAND_HAS_FLAG(nextInstr->i_src, OPERAND_IMM_FLAG))) {
	    nextInstr->i_src->a_displacement =
		nextInstr->i_src->a_displacement + STANDARD_CALL_RA_BYTES;
	    INCREMENT_SIMULATED = true;
	}

	/*
	 * We will try to get lucky again and abuse a stack pointer
	 * manipulation before the call instruction.  If a subtract is
	 * already present, we can again just add 4 to it simulate the RA
	 * being pushed.  This case happens more frequently than the above
	 * case, and most of the time we end up not having to add an extra
	 * instruction.
	 */

	if ((prevInstr->i_class == CLASS_SUB) &&
	    (prevInstr->i_dest->a_base_reg == (reg_sp | reg_32)) &&
	    (prevInstr->i_src != NULL) &&
	    (BBLHasHellPredecessor(prevInstr->i_Bblock) == false) &&
	    (OPERAND_HAS_FLAG(prevInstr->i_src, OPERAND_IMM_FLAG))) {
	    prevInstr->i_src->a_displacement =
		prevInstr->i_src->a_displacement + STANDARD_CALL_RA_BYTES;
	    DECREMENT_SIMULATED = true;
	}

	/*
	 * Ok, maybe we didn't get lucky and we'll have to add in an extra
	 * instruction. We want the stack to grow by 4 bytes so that any
	 * following instructions accessing the stack will be Good To Go.
	 */

	if (DECREMENT_SIMULATED == false) {
	    spDecrement = InstrCreateSPDec();
	    spDecrement->i_src->a_displacement = STANDARD_CALL_RA_BYTES;

	    InsertInstrBeforeInstr(spDecrement,
		funClone->f_first->b_first);
	    funClone->f_first->b_first = spDecrement;
	    BBLDivideBlock(funClone->f_first,
		funClone->f_first->b_first->i_next);
	    funClone->f_first = spDecrement->i_Bblock;
	}

	/*
	 * If we didn't get to exploit an ADD SP after the call instruction
	 * then we need to insert an instruction.  What follows is sort of
	 * tricky, but appears to work.
	 */

	if (INCREMENT_SIMULATED == false) {
	    spIncrement = InstrCreateSPInc();
	    spIncrement->i_src->a_displacement = STANDARD_CALL_RA_BYTES;

	    InsertInstrAfterInstr(spIncrement, funClone->f_last->b_last);

	    funClone->f_last->b_last = spIncrement;
	    BBLDivideBlock(funClone->f_last, funClone->f_last->b_last);
	    funClone->f_last = spIncrement->i_Bblock;

	    /*
	     * Currently we have something that looks like:
	     *
	     *             ---------------------------------
	     *             |      Function's Last BBL      |
	     *             |                               |
	     *             ---------------------------------
	     *                             |
	     *                             | ET_NORMAL
	     *                             |
	     *             ---------------------------------
	     *         ---|        SP Increment BBL        |---
	     *         |   ---------------------------------  |
	     *         |             /     |     \            |
	     *         |            /      |      \           |
	     *
	     * Unfortunately that's not really what we want, since the SP
	     * Increment BBL should fall through to the function link BBL
	     * from the callsite. We will rip out the edges on the increment
	     * BBL and attach them to the block above it. This seems to work
	     * well enough.
	     */

	    BBLRemoveEdge(spIncrement->i_Bblock->b_prev, spIncrement->i_Bblock);
	    BBLCopyEdges(spIncrement->i_Bblock, spIncrement->i_Bblock->b_prev);
	    BBLRemoveEdges(spIncrement->i_Bblock);
	    BBLAddEdge(spIncrement->i_Bblock, callInstr->i_Bblock->b_next,
		ET_NORMAL);
	    spIncrement->i_Bblock->b_type = BT_NORMAL;
	}

    }

    /*
     * Replace all RET instructions with unconditional jumps to the end
     * of the function (the spDecrement instruction).
     */

    if (INLINE_USE_CALLERS_FP) {
	newReturnPoint = nextInstr->i_Bblock;
    } else if (INCREMENT_SIMULATED == true) {
	newReturnPoint = nextInstr->i_Bblock;
    } else {
	newReturnPoint = funClone->f_last;
    }

    FOREACH_BBL_IN_FUN(funClone, biterator) {
	FOREACH_INSTR_IN_BBL(biterator, iterator) {
	    if (isReturnClass(iterator)) {

		/*
		 *                  Add a JMP in place of all RETs
		 */

		// BBLRemoveEdge(iterator->i_Bblock, funClone->f_exit);
		if (iterator->i_next != NULL) {
		    InsertInstrAfterInstr(InstrCreateJMP(), iterator);
		    iterator->i_Bblock->b_last = iterator->i_next;
		    (BBLAddEdge(iterator->i_Bblock, newReturnPoint,
			    ET_UNCOND)->e_edgeWeight)
			= iterator->i_Bblock->b_weight;
		    iterator->i_Bblock->b_type = BT_UNCOND;
		} else {
		    (BBLAddEdge(iterator->i_Bblock, newReturnPoint,
			    ET_NORMAL)->e_edgeWeight)
			= iterator->i_Bblock->b_weight;
		    iterator->i_Bblock->b_type = BT_NORMAL;
		}
		InstrDestroy(iterator);
	    }
	}
    }

    /*
     *                Disconnect the call instruction
     */

    firstBlock = callInstr->i_Bblock;

    BBLRemoveEdges(callInstr->i_Bblock);
    BBLAddEdge(callInstr->i_Bblock, funClone->f_first, ET_NORMAL);
    callInstr->i_Bblock->b_type = BT_NORMAL;
    NODE_WEIGHT(callInstr->i_Bblock->b_succ) = linkWeight;

    /*
     *  Remove the edge from the callee's exit block to the function
     *  link block.
     */

    BBLRemoveEdge(funOrig->f_exit, nextInstr->i_Bblock);

    /*
     *     Inline the function by connecting the basic blocks together
     */

    biterator = funClone->f_first;
    while (biterator != NULL) {
	Block *next = biterator->b_next;
	biterator->b_fun = callerFunction;
	InsertBBLAfterBBL(biterator, firstBlock);
	firstBlock = biterator;
	biterator = next;
    }

    if (!INLINE_USE_CALLERS_FP) {
	if (DECREMENT_SIMULATED == false) {
	    INSTR_REMOVE_FLAG(spDecrement, FUN_START);
	    INSTR_REMOVE_FLAG(spDecrement, FUN_END);
	}
	if (INCREMENT_SIMULATED == false) {
	    INSTR_REMOVE_FLAG(spIncrement, FUN_START);
	    INSTR_REMOVE_FLAG(spIncrement, FUN_END);
	}
    }

    InstrDestroy(callInstr);

    return;
}



/*
 *  CleanupClone(callInstr,fun) -- cleans up and optimizes the cloned function.
 */
void CleanupClone(Instruction *callInstr, Function *fun) {
    Block *bbl;
    Instruction *instr;
    Operand *src, *dst;
    int elim;

    /*
     * offset is how much we will add to each instruction of the form
     * DISPLACEMENT(%ebp). ADDRESS_SIZE is tacked on to account for
     * the push before the frame pointer is set up (we are removing
     * the instruction that sets up the frame pointer).
     */

    int offset = InstrBPSPDiff(callInstr);

    CASSERT(offset != STACK_UNKNOWN, "Call instruction with unknown stack.");
    CASSERT(offset >= 0, "Call instruction with negative stack offset.");

    offset = offset + ADDRESS_SIZE;

    CASSERT(FUNCTION_HAS_FLAG(fun, FUNCTION_GCC_PROLOGUE_FLAG),
	"Cloned function does not have the GCC prologue.");

    /*
     * Next, we need to replace the LEAVE instruction with
     * something that has the same semantics. If a "pop %ebp" is
     * used in place of the leave, we should also get rid of that.
     */

    FOREACH_BBL_IN_FUN(fun, bbl) {
	instr = bbl->b_first;
	while ((instr != NULL) && (instr->i_Bblock == bbl)) {
	    Instruction *next = instr->i_next;
	    if (isLeaveClass(instr)) {
		CompensateForLeave(offset, instr);
		InstrDestroy(instr);
	    } else if ((instr->i_class == CLASS_POP)
		&& OperandIsFramePointer(instr->i_dest)) {
		Instruction *compensate = InstrCreateSPInc();
		compensate->i_src->a_displacement = ADDRESS_SIZE;
		InsertInstrBeforeInstr(compensate, instr);
		InstrDestroy(instr);
	    }
	    instr = next;
	}
    }


    /*
     * First we look for the "mov %esp -> %ebp" instruction and
     * remove it.
     */

    INSTR_REMOVE_FLAG(FUNCTION_LAST_INSTR(fun), FUN_END);
    INSTR_REMOVE_FLAG(FUNCTION_FIRST_INSTR(fun), FUN_START);

    bbl = fun->f_first;
    instr = bbl->b_first;
    instr = instr->i_next;

    CASSERT(isMovClass(instr), "Cloned function with strange GCC prologue.");
    CASSERT(instr->i_dest, "Cloned function with strange GCC prologue.");
    CASSERT(instr->i_src, "Cloned function with strange GCC prologue.");
    CASSERT(OperandIsFramePointer(instr->i_dest),
	"Cloned function with strange GCC prologue.");
    CASSERT(OperandIsStackPointer(instr->i_src),
	"Cloned function with strange GCC prologue.");

    src = instr->i_src;
    dst = instr->i_dest;

    elim = 1;
    /*
     * We want both src and dst to be "just" registers -- i.e., have
     * _only_ OPERAND_BASE_FLAG set.  If not, set elim to 0, indicating
     * that this instruction should not be eliminated.
     */
    if (!isMovClass(instr)) {
	elim = 0;
    }
    if (instr->i_src2 != NULL) {
	elim = 0;
    }

    if (!(OPERAND_HAS_FLAG(src, OPERAND_BASE_FLAG)
	    && !OPERAND_HAS_FLAG(src, OPERAND_DISP_R_FLAG)
	    && !OPERAND_HAS_FLAG(src, OPERAND_INDEX_FLAG)
	    && !OPERAND_HAS_FLAG(src, OPERAND_SCALE_FLAG)
	    && !OPERAND_HAS_FLAG(src, OPERAND_IMM_FLAG)
	    && !OPERAND_HAS_FLAG(src, OPERAND_DISP_A_FLAG)
	    && !OPERAND_HAS_FLAG(src, OPERAND_REGISTER_INDIRECT))) {
	elim = 0;
    }
    if (!(OPERAND_HAS_FLAG(dst, OPERAND_BASE_FLAG)
	    && !OPERAND_HAS_FLAG(dst, OPERAND_DISP_R_FLAG)
	    && !OPERAND_HAS_FLAG(dst, OPERAND_INDEX_FLAG)
	    && !OPERAND_HAS_FLAG(dst, OPERAND_SCALE_FLAG)
	    && !OPERAND_HAS_FLAG(dst, OPERAND_IMM_FLAG)
	    && !OPERAND_HAS_FLAG(dst, OPERAND_DISP_A_FLAG)
	    && !OPERAND_HAS_FLAG(dst, OPERAND_REGISTER_INDIRECT))) {
	elim = 0;
    }

    if (src->a_base_reg != (reg_sp | reg_32)) {	// src reg is not %esp
	elim = 0;
    }
    if (dst->a_base_reg != (reg_bp | reg_32)) {	// dst reg is not %ebp
	elim = 0;
    }

    if (elim) {
	InstrDestroy(instr);
    }

    /*
     * Next we locate all instructions that use EBP indirectly with
     * a displacement and update the displacement.
     */

    FOREACH_BBL_IN_FUN(fun, bbl) {
	instr = bbl->b_first;

	while ((instr != NULL) && (instr->i_Bblock == bbl)) {
	    Instruction *next = instr->i_next;

	    if (instr->i_dest) {
		if (OperandAccessesStackFromFramePointer(instr->i_dest)
		|| OperandAccessesStackFromFramePointerWithSIB(instr->i_dest)) {
		    if (instr->i_dest->a_displacement <= 0) {
			instr->i_dest->a_displacement -= offset;
		    } else {
			instr->i_dest->a_displacement -=
			    (offset + ADDRESS_SIZE);
		    }
		}
	    }

	    if (instr->i_src) {
		if ((OperandAccessesStackFromFramePointer(instr->i_src)
			||
			OperandAccessesStackFromFramePointerWithSIB
			(instr->i_src)) && !isLeaClass(instr)) {
		    if (instr->i_src->a_displacement <= 0) {
			instr->i_src->a_displacement -= offset;
		    } else {
			instr->i_src->a_displacement -= (offset + ADDRESS_SIZE);
		    }
		}
	    }

	    if (instr->i_src2) {
		if ((OperandAccessesStackFromFramePointer(instr->i_src2)
			||
			OperandAccessesStackFromFramePointerWithSIB
			(instr->i_src2)) && !isLeaClass(instr)) {
		    if (instr->i_src2->a_displacement <= 0) {
			instr->i_src2->a_displacement -= offset;
		    } else {
			instr->i_src2->a_displacement -= (offset+ADDRESS_SIZE);
		    }
		}
	    }

	    if (((instr->i_src)
		    && OperandUsesContentsOfFramePointer(instr->i_src))
		|| (instr->i_src2
		    && OperandUsesContentsOfFramePointer(instr->i_src2)))
		CompensateForFramePointerUse(offset, instr);

	    if (isLeaClass(instr) && (instr->i_src)
		&&
		(RegisterIsFramePointer(instr->i_src->a_base_reg)
		    ||
		    RegisterIsFramePointer(instr->i_src->a_index_reg))) {
		if (instr->i_src->a_displacement < 0) {
		    instr->i_src->a_displacement -= offset;
		} else if (instr->i_src->a_displacement > 0) {
		    instr->i_src->a_displacement -= (offset + ADDRESS_SIZE);
		} else {
		    CompensateForFramePointerUse(offset, instr);
		    goto compensationFPDone;
		}
	    }

	    if (isLeaClass(instr) && (instr->i_src2)
		&&
		(RegisterIsFramePointer(instr->i_src2->a_base_reg)
		    ||
		    RegisterIsFramePointer(instr->i_src2->a_index_reg))) {
		if (instr->i_src2->a_displacement < 0) {
		    instr->i_src2->a_displacement -= offset;
		} else if (instr->i_src2->a_displacement > 0) {
		    instr->i_src2->a_displacement -= (offset + ADDRESS_SIZE);
		} else {
		    CompensateForFramePointerUse(offset, instr);
		}
	    }

compensationFPDone:

	    UpdateInstrLength(instr);
	    instr = next;

	}		/* end iteration over instructions */
    }			/* end iteration over basic blocks in the function */
}



/*
 *  CompensateForFramePointerUse(offset,instr) --
 *
 *  CompensateForFramePointerUse() adds in instructions before and after
 *  'instr' to change the frame pointer so it appears as if it was
 *  set up correctly when the function was cloned. Since we actually get
 *  rid of the instruction that sets it up (mov esp -> ebp), we need to
 *  decrement EBP appropriately before a use of it, and increment it
 *  afterward.
 */
static void CompensateForFramePointerUse(int offset, Instruction *instr) {
    Block *bbl = instr->i_Bblock;
    Instruction *sub;
    Instruction *add;

    if (INSTR_HAS_FLAG(instr, INSTR_DONT_COMPENSATE_FOR_FP_USE_FLAG)) {
#if 1
	INSTR_REMOVE_FLAG(instr, INSTR_DONT_COMPENSATE_FOR_FP_USE_FLAG);
#endif
	return;
    }

    sub = xInstrCreateBPAdd();
    add = xInstrCreateBPAdd();

    sub->i_src->a_displacement = offset * -1;
    add->i_src->a_displacement = offset;

    if (bbl->b_first == instr) {
	bbl->b_first = sub;
    }
    if (bbl->b_last == instr) {
	bbl->b_last = add;
    }

    InsertInstrBeforeInstr(sub, instr);
    InsertInstrAfterInstr(add, instr);
    INSTR_SET_FLAG(instr, INSTR_USES_FP_BUT_COMPENSATED_FLAG);
}



/*
 *  CompensateForLeave(offset,leave) --
 *
 *  We want to get rid of "leave" instructions in inlined functions. To
 *  do so, we add the appropriate amount to the stack pointer.
 */
static void CompensateForLeave(int offset, Instruction *leave) {
    Instruction *instr;
    Instruction *mov;
    Instruction *sub;
    int diff = InstrBPSPDiff(leave);

    if ((diff + ADDRESS_SIZE) == 0) {
	return;
    }

    /*
     * If we know how high the stack grew when it hit the leave
     * instruction, then we can add in a simple add to ESP.
     */

    if (diff != STACK_UNKNOWN) {
	instr = xInstrCreateSPInc();
	instr->i_src->a_displacement = diff + ADDRESS_SIZE;
	if (leave->i_Bblock->b_first == leave) {
	    leave->i_Bblock->b_first = instr;
	}
	InsertInstrBeforeInstr(instr, leave);
	return;
    }

    /*
     * If we don't know how high the stack grew, then we can use two
     * instructions to reset the stack pointer:
     *      mov esp <- ebp
     *      sub esp <- stack offset @ callsite
     */

    mov = xInstrCreateMovBPToSP();
    sub = xInstrCreateSPDec();
    INSTR_SET_FLAG(mov, INSTR_DONT_COMPENSATE_FOR_FP_USE_FLAG);
    sub->i_src->a_displacement = offset - ADDRESS_SIZE;

    if (leave->i_Bblock->b_first == leave) {
	leave->i_Bblock->b_first = mov;
    }
    InsertInstrBeforeInstr(mov, leave);
    InsertInstrAfterInstr(sub, mov);

    return;
}



/*
 *  InlineSyscall(callInstr) --
 *  Inlines the function(syscall) called by the 'callInstr'.
 */
void InlineSyscall(Instruction *callInstr) {
    Function *funClone;
    Function *funOrig;
    Function *callerFunction;
    Block *targetBBL;
    Edge *callEdge;
    EdgeNode *successors;
    EdgeNode *predecessors;
    Instruction *spIncrement;	// add $0x4, %esp
    Instruction *spDecrement;	// sub $0x4, %esp
    Instruction *nextInstr;	// Instr following callInstr
    Instruction *prevInstr;	// Instr preceding callInstr
    Block *biterator;
    Block *firstBlock;
    Block *newReturnPoint;
    Instruction *iterator;
    bool INCREMENT_SIMULATED = false;
    bool DECREMENT_SIMULATED = false;
    UINT32 linkWeight;
    float factor;

    iterator = xInstrCreateMovBPToSP();

    callEdge = BBLGetFunDirectEdge(callInstr->i_Bblock);
    callerFunction = INSTR_FUNCTION(callInstr);

    if (callEdge == NULL) {
	WARNING("[X86:FunInline] Trying to inline a function"
	    " from a callsite[0x%8.8x] with no ET_FUNCALL edge.",
	    callInstr->i_address);
	return;
    }

    spIncrement = spDecrement = NULL;

    targetBBL = callEdge->e_to;
    funOrig = targetBBL->b_fun;
    funClone = FunClone(funOrig);

    REPORT(RP_LGLOOP, "Inlining %s <-- %s",
	callerFunction->f_name, funOrig->f_name);

    FOREACH_PRED_IN_BBL(funClone->f_exit, predecessors) {
	Instruction *last = (NODE_FROM(predecessors))->b_last;
	if (last->i_class == CLASS_RET) {
	    if (last->i_prev->i_class == CLASS_POP) {
		if (OperandIsFramePointer(last->i_prev->i_dest)) {
		    INSTR_SET_FLAG(last->i_prev, INSTR_DEFINES_FP_NICELY);
		}
	    }
	}
    }

    UpdateFunction(funClone);
    if (INLINE_USE_CALLERS_FP) {
	CleanupSyscallClone(callInstr, funClone);
    }
    nextInstr = callInstr->i_next;
    prevInstr = callInstr->i_prev;

    linkWeight = callEdge->e_edgeWeight;

    /*
     * We want to weight the edges of the cloned function to represent
     * what the weights may look like in an inlined version.  Note that
     * the cloned function has the same weights as the original function.
     * What we will do is multiple each weight by (callsite freq /
     * total pred freq), which is some number less than 1.  This
     * represents the frequency that the callsite we're considered may
     * contribute to the execution count of each block.  We will also
     * multiple each edge weight in the original function by 1 -
     * (callsite freq / total pred freq) to account for the adjusted
     * weights that will occur after inlining.
     */

    /* Update our cloned function */
    factor = ((float) callEdge->e_edgeWeight) / ((float) targetBBL->b_weight);
    FOREACH_BBL_IN_FUN(funClone, biterator) {
	FOREACH_SUCC_IN_BBL(biterator, successors) {
	    NODE_WEIGHT(successors) = NODE_WEIGHT(successors) * factor;
	}
	if (factor < REMOVE_HOT_FACTOR) {
	    BBL_REMOVE_FLAG(biterator, BBL_HOT_FLAG);
	    BBL_REMOVE_FLAG(biterator, BBL_HOT_CALL_SITE_FLAG);
	}
    }

    FOREACH_BBL_IN_FUN(funClone, biterator) {
	BBLUpdateWeight(biterator);
    }

    /* Update our original function */
    factor = 1.0 - factor;
    FOREACH_BBL_IN_FUN(funOrig, biterator) {
	FOREACH_SUCC_IN_BBL(biterator, successors) {
	    NODE_WEIGHT(successors) = NODE_WEIGHT(successors) * factor;
	}
	//   if (factor < REMOVE_HOT_FACTOR) {
	if (factor < -9999) {
	    BBL_REMOVE_FLAG(biterator, BBL_HOT_FLAG);
	    BBL_REMOVE_FLAG(biterator, BBL_HOT_CALL_SITE_FLAG);
	}
    }


    FOREACH_BBL_IN_FUN(funOrig, biterator) {
	BBLUpdateWeight(biterator);
    }

    /* Remove all ENTRY/EXIT edges from the clone */
    BBLRemoveEdges(funClone->f_entry);
    BBLRemoveIncomingEdges(funClone->f_exit);

    /*
     * Replace all RET instructions with unconditional jumps to the end
     * of the function (the spDecrement instruction).
     */

    if (INLINE_USE_CALLERS_FP) {
	newReturnPoint = nextInstr->i_Bblock;
    } else {
	if (INCREMENT_SIMULATED == true) {
	    newReturnPoint = nextInstr->i_Bblock;
	} else {
	    newReturnPoint = funClone->f_last;
	}
    }

    FOREACH_BBL_IN_FUN(funClone, biterator) {
	FOREACH_INSTR_IN_BBL(biterator, iterator) {
	    if (isReturnClass(iterator)) {

		/*
		 *                  Add a JMP in place of all RETs
		 */

		// BBLRemoveEdge(iterator->i_Bblock, funClone->f_exit);
		if (iterator->i_next != NULL) {
		    InsertInstrAfterInstr(InstrCreateJMP(), iterator);
		    iterator->i_Bblock->b_last = iterator->i_next;
		    (BBLAddEdge(iterator->i_Bblock, newReturnPoint,
			    ET_UNCOND)->e_edgeWeight)
			= iterator->i_Bblock->b_weight;
		    iterator->i_Bblock->b_type = BT_UNCOND;
		} else {
		    (BBLAddEdge(iterator->i_Bblock, newReturnPoint,
			    ET_NORMAL)->e_edgeWeight)
			= iterator->i_Bblock->b_weight;
		    iterator->i_Bblock->b_type = BT_NORMAL;
		}
		InstrDestroy(iterator);
	    }
	}
    }

    /*
     *                Disconnect the call instruction
     */

    firstBlock = callInstr->i_Bblock;

    BBLRemoveEdges(callInstr->i_Bblock);
    BBLAddEdge(callInstr->i_Bblock, funClone->f_first, ET_NORMAL);
    callInstr->i_Bblock->b_type = BT_NORMAL;
    NODE_WEIGHT(callInstr->i_Bblock->b_succ) = linkWeight;

    /*
     *  Remove the edge from the callee's exit block to the function
     *  link block.
     */

    BBLRemoveEdge(funOrig->f_exit, nextInstr->i_Bblock);

    /*
     *     Inline the function by connecting the basic blocks together
     */

    biterator = funClone->f_first;
    while (biterator != NULL) {
	Block *next = biterator->b_next;
	biterator->b_fun = callerFunction;
	InsertBBLAfterBBL(biterator, firstBlock);
	firstBlock = biterator;
	biterator = next;
    }

    if (!INLINE_USE_CALLERS_FP) {
	if (DECREMENT_SIMULATED == false) {
	    INSTR_REMOVE_FLAG(spDecrement, FUN_START);
	    INSTR_REMOVE_FLAG(spDecrement, FUN_END);
	}
	if (INCREMENT_SIMULATED == false) {
	    INSTR_REMOVE_FLAG(spIncrement, FUN_START);
	    INSTR_REMOVE_FLAG(spIncrement, FUN_END);
	}
    }

    InstrDestroy(callInstr);

    return;
}



/*
 *  ReplaceLeaveRet(leave) --
 *
 *  Replaces the LEAVE-RETURN combo with appropriate instruction pair.
 *  MR  :: THIS SEEMS UNNECESSARY ...
 */
void ReplaceLeaveRet(Instruction *leave) {

    /*  Instruction *mov_ins, *pop_ins; */

    /* mov_ins = xInstrCreateMovBPToSP();
     * pop_ins = xInstrCreatePop(IndexToRegister(REGISTER_INDEX_EBP));
     *
     * InsertInstrBeforeInstr(mov_ins,leave);
     * InsertInstrAfterInstr(pop_ins,mov_ins);
     */


}



/*
 *  CleanupSyscallClone(callInstr,fun) --
 *
 *  Cleans up and optimizes the cloned function.
 */
void CleanupSyscallClone(Instruction *callInstr, Function *fun) {
    Block *bbl;
    Instruction *instr;
    /* Operand *src, *dst; */
/*    int elim; */

    /*
     * Next, we need to replace the LEAVE instruction with
     * something that has the same semantics. If a "pop %ebp" is
     * used in place of the leave, we should also get rid of that.
     */

    /*
     * MR :
     *          Replace the LEAVE and RETURN combo with
     *
     *          leave           -->             mov %ebp,%esp
     *          ret                             pop %ebp
     */

    FOREACH_BBL_IN_FUN(fun, bbl) {
	instr = bbl->b_first;
	while ((instr != NULL) && (instr->i_Bblock == bbl)) {
	    Instruction *next = instr->i_next;
	    if (isLeaveClass(instr)) {
		if (isReturnClass(next)) {
		    ReplaceLeaveRet(instr);
		}
	    }
	    instr = next;
	}
    }

}
