/*
 *  Print i386 instructions for GDB, the GNU debugger.
 *  Copyright (C) 1988, 89, 91, 93, 94, 95, 96, 97, 1998
 *  Free Software Foundation, Inc.
 *
 *  Modified by the PLTO project,
 *  Department of Computer Science,
 *  University of Arizona, 2004-2005.
 */
#include <assert.h>

#include "global.h"
#include "utilities.h"
#include "x86-prefix.h"
#include "x86-table.h"
#include "x86-registers.h"
#include "x86-classes.h"

static int Print_OP_E(int, int, int);
static int Print_OP_G(int, int, int);
static int Print_OP_I(int, int, int);
static int Print_OP_indirE(int, int, int);
static int Print_OP_sI(int, int, int);
static int Print_OP_REG(int, int, int);
static int Print_OP_J(int, int, int);
static int Print_OP_DIR(int, int, int);
static int Print_OP_OFF(int, int, int);
static int Print_OP_ESDI(int, int, int);
static int Print_OP_DSSI(int, int, int);
static int Print_OP_SEG(int, int, int);
static int Print_OP_C(int, int, int);
static int Print_OP_D(int, int, int);
static int Print_OP_T(int, int, int);
static int Print_OP_rm(int, int, int);
static int Print_OP_ST(int, int, int);
static int Print_OP_STi(int, int, int);
static int Print_OP_MMX(int, int, int);
static int Print_OP_EM(int, int, int);
static int Print_OP_MS(int, int, int);
static int Print_OP_3DNow(int, int, int);
static int Print_OP_3DNowSuffix(int, int, int);
static int Print_OP_PREFETCH(int, int, int);

static int RouteThroughPrintingRoutine(op_rtn aFun, int bytemode, int flag1,
    int flag2);

static void putop(char *template, int aflag, int dflag);
static void addHex(int num);
static void dofloat(int aflag, int dflag);
static int get16();
static int get32();
static void ckprefix();
static void addSpace();

UINT8 *start_codep;		// start of code region
UINT8 *codep;			// cur position in code
char *instrText;		// textual representation of the instruction

int mod;
int rm;
int reg;
int prefixes;
int offset;
int start_pc;

static char *names32[] = {
    "%eax", "%ecx", "%edx", "%ebx", "%esp", "%ebp", "%esi", "%edi",
};
static char *names16[] = {
    "%ax", "%cx", "%dx", "%bx", "%sp", "%bp", "%si", "%di",
};
static char *names8[] = {
    "%al", "%cl", "%dl", "%bl", "%ah", "%ch", "%dh", "%bh",
};
static char *names_seg[] = {
    "%es", "%cs", "%ss", "%ds", "%fs", "%gs", "%?", "%?",
};
static char *names_cr[] = {
    "%CR0", "%?", "%CR2", "%CR3", "%CR4", "%?", "%?", "%?",
};



/*
 *  ckprefix() --
 */
static void ckprefix() {
    prefixes = 0;

    while (1) {
	switch (*codep) {
	case X86_PREFIX_REPZ:
	    strcpy(&instrText[offset], "repz: ");
	    offset += 6;
	    break;
	case X86_PREFIX_REPNZ:
	    strcpy(&instrText[offset], "repnz: ");
	    offset += 7;
	    break;
	case X86_PREFIX_LOCK:
	    strcpy(&instrText[offset], "lock: ");
	    offset += 6;
	    break;
	case X86_PREFIX_CS:
	    strcpy(&instrText[offset], "cs: ");
	    offset += 4;
	    break;
	case X86_PREFIX_SS:
	    strcpy(&instrText[offset], "ss: ");
	    offset += 4;
	    break;
	case X86_PREFIX_DS:
	    strcpy(&instrText[offset], "ds: ");
	    offset += 4;
	    break;
	case X86_PREFIX_ES:
	    strcpy(&instrText[offset], "es: ");
	    offset += 4;
	    break;
	case X86_PREFIX_FS:
	    strcpy(&instrText[offset], "fs: ");
	    offset += 4;
	    break;
	case X86_PREFIX_GS:
	    strcpy(&instrText[offset], "gs: ");
	    offset += 4;
	    break;
	case X86_PREFIX_OPERAND:
	    strcpy(&instrText[offset], "op: ");
	    offset += 4;
	    prefixes |= PREFIX_OPERAND;
	    break;
	case X86_PREFIX_ADDRESS:
	    strcpy(&instrText[offset], "addr: ");
	    offset += 6;
	    prefixes |= PREFIX_ADDRESS;
	    break;
	case X86_PREFIX_FWAIT:
	    strcpy(&instrText[offset], "fwait: ");
	    offset += 7;
	    break;
	default:
	    return;
	}
	codep++;
    }
}



/*
 *  i386InstrToText(pc,b) --
 */
char *i386InstrToText(ADDRESS pc, UINT8 *b) {

    X86op *dp;
    unsigned char need_modrm;

    UINT8 *inbuf = b;
    int aflag, dflag;
    start_pc = pc;
    start_codep = inbuf;
    codep = inbuf;
    instrText = zalloc(128);
    offset = 0;

    aflag = dflag = 1;
    //ckprefix(); //tmp hack, return this back //Igor - 04/20/06

    if (prefixes & PREFIX_OPERAND) {
	dflag ^= 1;
    }

    if (prefixes & PREFIX_ADDRESS) {
	aflag ^= 1;
    }

    if (*codep == 0x0f) {
	dp = &dis386_twobyte[*++codep];
	need_modrm = twobyte_has_modrm[*codep];
    } else {
	/* Otherwise we have a normal one-byte opcode. */
	dp = &dis386[*codep];
	need_modrm = onebyte_has_modrm[*codep];
    }
    codep++;

    if (need_modrm) {
	mod = (*codep >> 6) & 3;
	reg = (*codep >> 3) & 7;
	rm = *codep & 7;
    }

    /* If we have a floating floating instruction, handle it accordingly */
    if (dp->x_name == NULL && dp->x_bytemode1 == FLOATCODE) {
	dofloat(aflag, dflag);
    } else {
	/* Otherwise it's a normal non-floating point instruction. */

	/* If we still haven't found an instruction, then the bits in the REG
	 * section of the MODR/M byte point us to a group placeholder (grps)
	 * where we might find an instruction */
	if (dp->x_name == NULL) {
	    dp = &grps[dp->x_bytemode1][reg];
	}

	strcpy(&instrText[offset], dp->x_name);
	offset = offset + strlen(dp->x_name);
	putop(&instrText[0], aflag, dflag);
	addSpace();

	if (dp->x_op1) {
	    RouteThroughPrintingRoutine(dp->x_op1,dp->x_bytemode1,aflag,dflag);
	}

	if (dp->x_op2) {
	    if (dp->x_op1) {
		strcpy(&instrText[offset], " <- ");
		offset = offset + 4;
	    }
	    RouteThroughPrintingRoutine(dp->x_op2,dp->x_bytemode2,aflag,dflag);
	}

	if (dp->x_op3) {
	    if (dp->x_op2) {
		strcpy(&instrText[offset++], ",");
	    }
	    RouteThroughPrintingRoutine(dp->x_op3,dp->x_bytemode3,aflag,dflag);
	}

    }

    return instrText;
}



/*
 *  putop(template,aflag,dflag) --
 *
 *  Capital letters in template are macros.
 */
static void putop(char *template, int aflag, int dflag) {
    int i, len;

    len = strlen(instrText);

    for (i = 0; i < len; i++) {
	switch (instrText[i]) {
	case 'S':
	    if (dflag) {
		sprintf(&instrText[i], "l");
	    } else {
		sprintf(&instrText[i], "w");
	    }
	    break;
	case 'W':
	    if (dflag) {
		sprintf(&instrText[i], "w");
	    } else {
		sprintf(&instrText[i], "b");
	    }
	    break;
	}
    }
}



/*
 *  dofloat(aflag,dflag) --
 */
static void dofloat(int aflag, int dflag) {
    X86op *dp;
    unsigned char floatop;

    floatop = codep[-1];

    if (mod != 3) {
	dp = &float_mem[(floatop - 0xd8) * 8 + reg];
	strcpy(&instrText[offset], dp->x_name);
	offset = offset + strlen(dp->x_name);
	putop(&instrText[0], aflag, dflag);
	addSpace();

	if (dp->x_op1) {
	    RouteThroughPrintingRoutine(dp->x_op1,dp->x_bytemode1,aflag,dflag);
	}

	if (dp->x_op2) {
	    if (dp->x_op1) {
		strcpy(&instrText[offset], " <- ");
		offset = offset + 4;
	    }
	    RouteThroughPrintingRoutine(dp->x_op2,dp->x_bytemode2,aflag,dflag);
	}

	if (dp->x_op3) {
	    RouteThroughPrintingRoutine(dp->x_op3,dp->x_bytemode3,aflag,dflag);
	}

	return;
    }
    codep++;

    dp = &float_reg[floatop - 0xd8][reg];

    if (dp->x_name == NULL) {
	dp = &fgrps[dp->x_bytemode1][rm];
	strcpy(&instrText[offset], dp->x_name);
	offset = offset + strlen(dp->x_name);
	putop(&instrText[0], aflag, dflag);
	addSpace();
	return;
    }

    strcpy(&instrText[offset], dp->x_name);
    offset = offset + strlen(dp->x_name);
    putop(&instrText[0], aflag, dflag);
    addSpace();

    if (dp->x_op1) {
	RouteThroughPrintingRoutine(dp->x_op1, dp->x_bytemode1, aflag, dflag);
    }

    if (dp->x_op2) {
	if (dp->x_op1) {
	    strcpy(&instrText[offset], " <- ");
	    offset = offset + 4;
	}
	RouteThroughPrintingRoutine(dp->x_op2, dp->x_bytemode2, aflag, dflag);
    }

}



/*
 *  get32() --
 */
static int get32() {
    int x = 0;

    x = *codep++ & 0xff;
    x |= (*codep++ & 0xff) << 8;
    x |= (*codep++ & 0xff) << 16;
    x |= (*codep++ & 0xff) << 24;
    return x;
}



/*
 *  get16() --
 */
static int get16() {
    int x = 0;

    x = *codep++ & 0xff;
    x |= (*codep++ & 0xff) << 8;
    return x;
}



/*
 *  Print_OP_ST(ignore,aflag,dflag) --
 */
static int Print_OP_ST(int ignore, int aflag, int dflag) {
    strcpy(&instrText[offset], "%st");
    offset = offset + 3;
    return 0;
}



/*
 *  Print_OP_STi(ignore,aflag,dflag) --
 */
static int Print_OP_STi(int ignore, int aflag, int dflag) {
    strcpy(&instrText[offset], "%st");
    offset = offset + 3;
    sprintf(&instrText[offset], "(%d)", rm);
    offset = offset + 3;
    return 0;
}



/*
 *  Print_OP_indirE(bytemode,aflag,dflag) --
 */
static int Print_OP_indirE(int bytemode, int aflag, int dflag) {
    strcpy(&instrText[offset], "*");
    offset = offset + 1;
    return Print_OP_E(bytemode, aflag, dflag);
}



/*
 *  Print_OP_PREFETCH(bytemode,aflag,dflag) --
 */
static int Print_OP_PREFETCH(int bytemode, int aflag, int dflag) {
    return Print_OP_E(bytemode, aflag, dflag);
}



/*
 *  Print_OP_E(bytemode,aflag,dflag) --
 *
 *  Get information about the operands.  Chapter 2 of the Intel Manual Vol 2
 *  is especially useful here, specifically the tables
 */
static int Print_OP_E(int bytemode, int aflag, int dflag) {
    int disp;
    int havesib;
    int havebase;
    int base;
    int idx = 0;
    int scale = 0;

    havesib = 0;
    havebase = 1;
    base = rm;

    /* Currently examining the MODR/M byte */
    codep++;
    /* Currently examining the SIB or displacement byte(s) */

    /* if mod == 3 then the operand is simply a register, and the register
     * is encoded in the R/M bits.  We look it up in the table and set
     * things accordingly */
    if (mod == 3) {
	switch (bytemode) {
	case b_mode:
	    strcpy(&instrText[offset], names8[rm]);
	    offset = offset + strlen(names8[rm]);
	    break;
	case w_mode:
	    strcpy(&instrText[offset], names16[rm]);
	    offset = offset + strlen(names16[rm]);
	    break;
	case v_mode:
	    if (dflag) {
		strcpy(&instrText[offset], names32[rm]);
		offset = offset + strlen(names32[rm]);
	    } else {
		strcpy(&instrText[offset], names16[rm]);
		offset = offset + strlen(names16[rm]);
	    }
	    break;
	default:
	    MALFUNCTION("unexpected bytemode");
	}
	return 0;
    }

    disp = 0;

    /* aflag = 32 bit address mode */
    if (aflag) {
	/* For 32 bit addressing modes, an R/M of 4 means an SIB follows
	 * the MOD/RM byte */
	if (base == 4) {
	    havesib = 1;
	    scale = (*codep >> 6) & 3;
	    idx = (*codep >> 3) & 7;
	    base = *codep & 7;
	    codep++;
	}

	switch (mod) {
	case 0:
	    if (base == 5) {
		havebase = 0;
		disp = get32();
	    }
	    break;
	case 1:
	    disp = *codep++;
	    if ((disp & 0x80) != 0) {
		disp -= 0x100;
	    }
	    break;
	case 2:
	    disp = get32();
	    break;
	}

	if (mod != 0 || base == 5) {
	    addHex(disp);
	}

	if (havebase || (havesib && (idx != 4 || scale != 0))) {
	    strcpy(&instrText[offset], "(");
	    offset = offset + 1;
	    if (havebase) {
		strcpy(&instrText[offset], names32[base]);
		offset = offset + strlen(names32[base]);
	    }
	    if (havesib) {
		if (idx != 4) {
		    strcpy(&instrText[offset], ",");
		    offset = offset + 1;
		    strcpy(&instrText[offset], names32[idx]);
		    offset = offset + strlen(names32[idx]);
		}
		sprintf(&instrText[offset], ",%d", 1 << scale);
		offset = offset + 2;
	    }
	    strcpy(&instrText[offset], ")");
	    offset = offset + 1;
	}
    }

    return 0;

}



/*
 *  Print_OP_G(bytemode,aflag,dflag) --
 */
static int Print_OP_G(int bytemode, int aflag, int dflag) {
    switch (bytemode) {
    case b_mode:
	strcpy(&instrText[offset], names8[reg]);
	offset = offset + strlen(names8[reg]);
	break;
    case w_mode:
	strcpy(&instrText[offset], names16[reg]);
	offset = offset + strlen(names16[reg]);
	break;
    case d_mode:
	strcpy(&instrText[offset], names32[reg]);
	offset = offset + strlen(names32[reg]);
	break;
    case v_mode:
	if (dflag) {
	    strcpy(&instrText[offset], names32[reg]);
	    offset = offset + strlen(names32[reg]);

	} else {
	    strcpy(&instrText[offset], names16[reg]);
	    offset = offset + strlen(names16[reg]);
	}
	break;
    default:
	MALFUNCTION("unexpected bytemode");
    }
    return 0;
}



/*
 *  Print_OP_REG(code,aflag,dflag) --
 */
static int Print_OP_REG(int code, int aflag, int dflag) {
    char *s;

    switch (code) {
    case indir_dx_reg:
	s = "(%dx)";
	break;
    case ax_reg:
    case cx_reg:
    case dx_reg:
    case bx_reg:
    case sp_reg:
    case bp_reg:
    case si_reg:
    case di_reg:
	s = names16[code - ax_reg];
	break;
    case es_reg:
    case ss_reg:
    case cs_reg:
    case ds_reg:
    case fs_reg:
    case gs_reg:
	s = names_seg[code - es_reg];
	break;
    case al_reg:
    case ah_reg:
    case cl_reg:
    case ch_reg:
    case dl_reg:
    case dh_reg:
    case bl_reg:
    case bh_reg:
	s = names8[code - al_reg];
	break;
    case eAX_reg:
    case eCX_reg:
    case eDX_reg:
    case eBX_reg:
    case eSP_reg:
    case eBP_reg:
    case eSI_reg:
    case eDI_reg:
	if (dflag) {
	    s = names32[code - eAX_reg];
	} else {
	    s = names16[code - eAX_reg];
	}
	break;
    default:
	MALFUNCTION("unexpected bytemode");
    }

    strcpy(&instrText[offset], s);
    offset = offset + strlen(s);

    return 0;
}



/*
 *  Print_OP_I(bytemode,aflag,dflag) --
 */
static int Print_OP_I(int bytemode, int aflag, int dflag) {
    int op;

    switch (bytemode) {
    case b_mode:
	op = *codep++ & 0xff;
	break;
    case v_mode:
	if (dflag) {
	    op = get32();
	} else {
	    op = get16();
	}
	break;
    case w_mode:
	op = get16();
	break;
    default:
	MALFUNCTION("unexpected bytemode");
    }

    sprintf(&instrText[offset], "$");
    offset++;
    addHex(op);
    return 0;
}



/*
 *  Print_OP_sI(bytemode,aflag,dflag) --
 */
static int Print_OP_sI(int bytemode, int aflag, int dflag) {
    int op;

    switch (bytemode) {
    case b_mode:
	op = *codep++;
	if ((op & 0x80) != 0) {
	    op -= 0x100;
	}
	break;
    case v_mode:
	if (dflag) {
	    op = get32();
	} else {
	    op = get16();
	    if ((op & 0x8000) != 0) {
		op -= 0x10000;
	    }
	}
	break;
    case w_mode:
	op = get16();
	if ((op & 0x8000) != 0) {
	    op -= 0x10000;
	}
	break;
    default:
	MALFUNCTION("unexpected bytemode");
    }

    sprintf(&instrText[offset], "$");
    offset++;
    addHex(op);
    return 0;
}



/*
 *  Print_OP_J(bytemode,aflag,dflag) --
 */
static int Print_OP_J(int bytemode, int aflag, int dflag) {
    int disp;
    int mask = -1;

    switch (bytemode) {
    case b_mode:
	disp = *codep++;
	if ((disp & 0x80) != 0) {
	    disp -= 0x100;
	}
	break;
    case v_mode:
	if (dflag) {
	    disp = get32();
	} else {
	    disp = get16();
	    if ((disp & 0x8000) != 0) {
		disp -= 0x10000;
	    }
	    /* for some reason, a data16 prefix on a jump instruction
	     * means that the pc is masked to 16 bits after the
	     * displacement is added!  */
	    mask = 0xffff;
	}
	break;
    default:
	MALFUNCTION("unexpected bytemode");
    }

    disp = start_pc + codep - start_codep + disp;
    addHex(disp);
    return 0;
}



/*
 *  Print_OP_SEG(dummy,aflag,dflag) --
 */
static int Print_OP_SEG(int dummy, int aflag, int dflag) {
    strcpy(&instrText[offset], names_seg[reg-1]);
    offset += strlen(names_seg[reg]) + 1;
    return 0;
}



/*
 *  Print_OP_CR(dummy,aflag,dflag) -- does nothing.
 */
/*static*/ int Print_OP_CR(int dummy, int aflag, int dflag) {
    return 0;
}



/*
 *  Print_OP_DIR(size,aflag,dflag) --
 */
static int Print_OP_DIR(int size, int aflag, int dflag) {
    int disp;

    switch (size) {
    case v_mode:
	if (aflag) {
	    disp = get32();
	} else {
	    disp = get16();
	    if ((disp & 0x8000) != 0) {
		disp -= 0x10000;
	    }
	}
	disp = start_pc + codep - start_codep + disp;
	addHex(disp);
	break;
    default:
	MALFUNCTION("unexpected size");
    }
    return 0;
}



/*
 *  Print_OP_OFF(bytemode,aflag,dflag) --
 */
static int Print_OP_OFF(int bytemode, int aflag, int dflag) {
    int disp;

    if (aflag) {
	disp = get32();
    } else {
	disp = get16();
    }

    addHex(disp);

    return 0;
}



/*
 *  Print_OP_ESDI(dummy,aflag,dflag) --
 */
static int Print_OP_ESDI(int dummy, int aflag, int dflag) {
    strcpy(&instrText[offset], "%es:(");
    offset = offset + 5;
    if (aflag) {
	strcpy(&instrText[offset], "%edi");
	offset = offset + 4;
    } else {
	strcpy(&instrText[offset], "%di");
	offset = offset + 3;
    }
    strcpy(&instrText[offset], ")");
    offset++;

    return 0;
}



/*
 *  Print_OP_DSSI(dummy,aflag,dflag) --
 */
static int Print_OP_DSSI(int dummy, int aflag, int dflag) {
    strcpy(&instrText[offset], "%ds:(");
    offset = offset + 5;
    if (aflag) {
	strcpy(&instrText[offset], "%esi");
	offset = offset + 4;
    } else {
	strcpy(&instrText[offset], "%si");
	offset = offset + 3;
    }
    sprintf(&instrText[offset], ")");
    offset++;
    return 0;
}



/*
 *  Print_OP_C(dummy,aflag,dflag) --
 */
static int Print_OP_C(int dummy, int aflag, int dflag) {
    strcpy(&instrText[offset], names_cr[reg]);
    offset += strlen(names_cr[reg]) + 1;
    return 0;
}



/*
 *  Print_OP_D(dummy,aflag,dflag) --
 */
static int Print_OP_D(int dummy, int aflag, int dflag) {
    codep++;				// skip mod/rm
    MALFUNCTION("Print_OP_D abort");
    return 0;
}



/*
 *  Print_OP_T(dummy,aflag,dflag) --
 */
static int Print_OP_T(int dummy, int aflag, int dflag) {
    codep++;				// skip mod/rm
    MALFUNCTION("Print_OP_T abort");
    return 0;
}



/*
 *  Print_OP_rm(bytemode,aflag,dflag) --
 */
static int Print_OP_rm(int bytemode, int aflag, int dflag) {
    char *s;
    switch (bytemode) {
    case d_mode:
	s = names32[rm];
	break;
    case w_mode:
	s = names16[rm];
	break;
    default:
	s = "";
	break;
    }

    sprintf(&instrText[offset], s);
    offset = offset + strlen(s);
    return 0;
}



/*
 *  Print_OP_MMX(bytemode,aflag,dflag) --
 */
static int Print_OP_MMX(int bytemode, int aflag, int dflag) {
    MALFUNCTION("Print_OP_MMX abort");
    return 0;
}



/*
 *  Print_OP_EM(bytemode,aflag,dflag) --
 */
static int Print_OP_EM(int bytemode, int aflag, int dflag) {
    if (mod != 3) {
	return OP_E(bytemode, aflag, dflag);
    }

    codep++;
    MALFUNCTION("Print_OP_EM abort");
    return 0;
}



/*
 *  Print_OP_MS(bytemode,aflag,dflag) --
 */
static int Print_OP_MS(int bytemode, int aflag, int dflag) {
    ++codep;
    MALFUNCTION("Print_OP_MS abort");
    return 0;
}



/*
 *  Print_OP_3DNow(bytemode,aflag,dflag) --
 */
static int Print_OP_3DNow(int bytemode, int aflag, int dflag) {
    MALFUNCTION("Print_OP_3DNow abort");
    return 0;
}



/*
 *  Print_OP_3DNowSuffix(bytemode,aflag,dflag) --
 */
static int Print_OP_3DNowSuffix(int bytemode, int aflag, int dflag) {
    MALFUNCTION("Print_OP_3DNowSuffix abort");
    return 0;
}



/*
 *  addSpace() --
 */
static void addSpace() {
    sprintf(&instrText[offset], " ");
    offset = offset + 1;
}



/*
 *  addHex(num) --
 */
static void addHex(int num) {
    char foo[16];

    if (num < 0) {
	sprintf(foo, "%x", num);
	sprintf(&instrText[offset], "%x", num);
    } else if (num < 0x8048000) {
	sprintf(foo, "%x", num);
	sprintf(&instrText[offset], "%x", num);
    } else {
	sprintf(foo, "0x%8.8x", num);
	sprintf(&instrText[offset], "0x%8.8x", num);
    }
    offset = offset + strlen(foo);
}



/*
 *  RouteThroughPrintingRoutine(aFun,bytemode,flag1,flag2) --
 */
int RouteThroughPrintingRoutine(op_rtn aFun,int bytemode,int flag1,int flag2) {

    if (aFun == OP_E) {
	return Print_OP_E(bytemode, flag1, flag2);
    } else if (aFun == OP_G) {
	return Print_OP_G(bytemode, flag1, flag2);
    } else if (aFun == OP_I) {
	return Print_OP_I(bytemode, flag1, flag2);
    } else if (aFun == OP_indirE) {
	return Print_OP_indirE(bytemode, flag1, flag2);
    } else if (aFun == OP_sI) {
	return Print_OP_sI(bytemode, flag1, flag2);
    } else if (aFun == OP_REG) {
	return Print_OP_REG(bytemode, flag1, flag2);
    } else if (aFun == OP_J) {
	return Print_OP_J(bytemode, flag1, flag2);
    } else if (aFun == OP_DIR) {
	return Print_OP_DIR(bytemode, flag1, flag2);
    } else if (aFun == OP_OFF) {
	return Print_OP_OFF(bytemode, flag1, flag2);
    } else if (aFun == OP_ESDI) {
	return Print_OP_ESDI(bytemode, flag1, flag2);
    } else if (aFun == OP_DSSI) {
	return Print_OP_DSSI(bytemode, flag1, flag2);
    } else if (aFun == OP_SEG) {
	return Print_OP_SEG(bytemode, flag1, flag2);
    } else if (aFun == OP_C) {
	return Print_OP_C(bytemode, flag1, flag2);
    } else if (aFun == OP_D) {
	return Print_OP_D(bytemode, flag1, flag2);
    } else if (aFun == OP_T) {
	return Print_OP_T(bytemode, flag1, flag2);
    } else if (aFun == OP_rm) {
	return Print_OP_rm(bytemode, flag1, flag2);
    } else if (aFun == OP_ST) {
	return Print_OP_ST(bytemode, flag1, flag2);
    } else if (aFun == OP_STi) {
	return Print_OP_STi(bytemode, flag1, flag2);
    } else if (aFun == OP_MMX) {
	return Print_OP_MMX(bytemode, flag1, flag2);
    } else if (aFun == OP_EM) {
	return Print_OP_EM(bytemode, flag1, flag2);
    } else if (aFun == OP_MS) {
	return Print_OP_MS(bytemode, flag1, flag2);
    } else if (aFun == OP_3DNow) {
	return Print_OP_3DNow(bytemode, flag1, flag2);
    } else if (aFun == OP_3DNowSuffix) {
	return Print_OP_3DNowSuffix(bytemode, flag1, flag2);
    } else if (aFun == OP_PREFETCH) {
	return Print_OP_PREFETCH(bytemode, flag1, flag2);
    } else {
	WARNING("[RouteThroughPrintingRoutine] Failed to find a routine.");
    }

    return 0;
}



/*
 *  xPrintPSWFlag(flag) --
 */
void xPrintPSWFlag(int flag) {
    switch (flag) {
    case PSW_OF:
	printf("OF");
	break;
    case PSW_DF:
	printf("DF");
	break;
    case PSW_IF:
	printf("IF");
	break;
    case PSW_TF:
	printf("TF");
	break;
    case PSW_SF:
	printf("SF");
	break;
    case PSW_ZF:
	printf("ZF");
	break;
    case PSW_AF:
	printf("AF");
	break;
    case PSW_PF:
	printf("PF");
	break;
    case PSW_CF:
	printf("CF");
	break;
    default:
	printf("<?>");
    }
}
