/*
 *  x86/jt.c
 *
 *  A module for extracting jump table information out of instruction
 *  sequences.  I've been dissatisfied with how this works from day 1.
 *  For now, we only detect jumps of the form jmp *BASE_ADDR(,%eax,4)
 */
#include "global.h"
#include "flowgraph.h"
#include "x86-jt.h"
#include "x86-registers.h"
#include "x86-classes.h"
#include "x86-op.h"
#include "sysdep.h"
#include "utilities.h"



/*
 *  xInstrLocateJTBaseAddress(instr) --
 *
 *  Given an instruction, InstrLocateJTBaseAddress will scan back through the
 *  preceding instructions trying to find the base address of the jump table.
 *  Once the top of the basic block is hit, we can proceed no further since
 *  the state of things probably becomes nondeterministic.
 */
UINT32 xInstrLocateJTBaseAddress(Instruction *instr) {
    Instruction *inst = xInstrLocateJTBaseInstr(instr);

    if (inst == NULL) {
	return 0;
    }

    if ((inst == instr) && (inst->i_src)
	&& inst->i_src->a_displacement
	&& OP_TYPE(inst->i_src, OPERAND_DISP_A_FLAG))
	return inst->i_src->a_displacement;

    return inst->i_src->a_displacement;
}



/*
 *  xInstrLocateJTBoundsCheck(instr) --
 *  Returns the instruction which does the bounds check on the jump table.
 */
Instruction *xInstrLocateJTBoundsCheck(Instruction *instr) {
    UINT32 registerToFind;	// register we are tracing. see x86-registers.h
    Operand *op;
    Instruction *iterator;

    iterator = instr;
    CASSERT(iterator != NULL, "Given a NULL instruction.");
    op = instr->i_src;
    CASSERT(op != NULL, "Given instr with no destination op.");

    registerToFind = op->a_base_reg;

    if (registerToFind == 0) {
	registerToFind = op->a_index_reg;
    }

    CASSERT(registerToFind != 0,
	"Could not locate the register through which the jump is performed.");

    while (true) {
	op = iterator->i_dest;
	if (op && ((op->a_base_reg == registerToFind)
		|| (op->a_index_reg == registerToFind)))
	    return NULL;

	op = iterator->i_src;
	if (op && OperandIsGPRegister(op)
	    && (op->a_base_reg == registerToFind))
	    if ((iterator->i_class == CLASS_CMP) && (iterator->i_src2)) {
		return iterator;
	    }

	if (INSTR_HAS_FLAG(iterator, BBL_START)) {
	    if (BBLNumberOfPredecessors(iterator->i_Bblock) != 1 ||
		BBLHasEntryPredecessor(iterator->i_Bblock))
		return NULL;
	    iterator = (NODE_FROM(INSTR_BBL_PRED(iterator)))->b_last;
	} else {
	    iterator = iterator->i_prev;
	}

    }

    return NULL;
}



/*
 *  xInstrLocateJTSize(instr) --
 *
 *  InstrLocateJTBSize will scan backward through the sequence of
 *  instructions and attempt to locate the size of the jump table.  Much of
 *  the work done in InstrLocateJTBaseAddress is also done here.  This is
 *  because we need to trace a register back to the predecessor blocks of
 *  the instruction.
 */
UINT32 xInstrLocateJTSize(Instruction *instr) {
    Instruction *instr2 = InstrLocateJTBoundsCheck(instr);

    if (instr2 == NULL) {
	return 0;
    }

    return instr2->i_src2->a_displacement + 1;
}



/*
 *  xInstrLocateJTBaseInstr(instr) --
 *
 *  Returns the instruction that uses/loads the address that 'instr' uses to
 *  jump through a jump table.
 */
Instruction *xInstrLocateJTBaseInstr(Instruction *instr) {
    Instruction *iterator;
    UINT32 registerToFind;
	// The register number we are tracing. see x86-registers.h
    Operand *op, *src;

    iterator = instr;
    CASSERT(iterator != NULL, "Given a NULL instruction.");
    op = instr->i_src;
    CASSERT(op != NULL, "Given instr with no destination op.");

    registerToFind = op->a_base_reg;

    if (registerToFind == 0) {
	registerToFind = op->a_index_reg;
    }

    CASSERT(registerToFind != 0,
	"Could not locate the register through which the jump is performed.");

    /*
     * We will try to find an instruction of the form:
     *		mov  0x01234567(%ebx), %ebx
     * because thats how jumps through the jump table are loaded.
     * After we find it, we will continue to backtrace through the end
     * of the basic block tracing the register we found as the index
     * in the mov instruction (ebx in the above example).
     */

    if ((op->a_displacement) && OP_TYPE(op, OPERAND_DISP_A_FLAG)) {
	return instr;
    }

    while (true) {
	op = iterator->i_dest;
	if (op && (op->a_base_reg == registerToFind)) {
	    src = iterator->i_src;
	    if (src == NULL) {
		return NULL;
	    }
	    if ((src->a_base_reg == 0) && (src->a_index_reg == 0)) {
		return NULL;
	    }
	    if (iterator->i_class != CLASS_MOV) {
		return NULL;
	    }
	    if (!OP_TYPE(src, OPERAND_DISP_A_FLAG)) {
		return NULL;
	    }
	    return iterator;
	}

	if (INSTR_HAS_FLAG(iterator, BBL_START)) {
	    break;
	}

	iterator = iterator->i_prev;

    }					/* end while */

    return NULL;
}
