/*
 *  x86/liveness.c -- Register liveness analysis on the x86.
 */
#include "global.h"
#include "x86-liveness.h"
#include "sysdep.h"
#include "x86-registers.h"
#include "cp.h"
#include "stack.analyze.h"
#include "utilities.h"
#include "x86-table-defs.h"
#include "x86-asm.h"
#include "x86-classes.h"
#include "flowgraph.h"
#include "x86-registers.h"
#include "x86-prefix.h"
#include "x86-table.h"

#define UNKNOWN_TRUE  5
#define UNKNOWN_FALSE 6

#define REPEAT_REGISTER (reg_c | reg_32)

/* Returns a bitvector in which all registers are used */
static REGISTER_BITVECTOR UseAllRegisters();

/* Returns a bitvector in which all registers are defined */
static REGISTER_BITVECTOR DefineAllRegisters();

/* Returns a register bitvector indicating bits set by 'instr' */
static REGISTER_BITVECTOR ComputeRegistersDefinedHelper(Instruction *instr);

/* SetUsedRegister sets the register bit for 'aRegister'
 * in the bitvector 'currentVector' */
static int SetUsedRegister(int aRegister, int currentVector);

/* SetDefinedRegister sets the register bit for 'aRegister'
 * in the bitvector 'currentVector' */
static int SetDefinedRegister(int aRegister, int currentVector);

/* Converts register from x86-registers.h to a bit in the register bitvector */
extern int registerToBitRegister(int someRegister);

/* Fills in the liveness information for stack slots */
static void ComputeStackLocationUsedByOperand(Operand *op, int *slots,
    int bp_index, int arsize, int flg);

static bool DEFINE_CONSERVATIVE;



/*
 *  xComputeRegistersUsed(instr) --
 *
 *  ComputeRegistersUsed will return a register bitvector with bits set
 *  for each register that the instruction uses.
 */
REGISTER_BITVECTOR xComputeRegistersUsed(Instruction *instr) {
    REGISTER_BITVECTOR newVector = 0;
    Operand *op;
    X86op *disEntry = InstrGetOpcodeEntry(instr);

    CASSERT(disEntry != NULL, "Could not find opcode for instruction.");

    /*
     *                         EXPLICIT SOURCE1
     */

    if (INSTR_HAS_FLAG(instr, SOURCE1_FLAG)) {
	op = instr->i_src;
	if (OperandUsesRegister(op)) {
	    if (OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
		newVector = SetUsedRegister(op->a_base_reg, newVector);
	    }
	    if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		newVector = SetUsedRegister(op->a_index_reg, newVector);
	    }
	    if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)
		&& !OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		MALFUNCTION("Operand uses a register,"
			" but neither INDEX or REG flags set.");
	    }
	}				/* End test for register */
    }


    /* End source1 test */
    /*
     *                         EXPLICIT SOURCE2
     */
    if (INSTR_HAS_FLAG(instr, SOURCE2_FLAG)) {
	op = instr->i_src2;
	if (OperandUsesRegister(op)) {
	    if (OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
		newVector = SetUsedRegister(op->a_base_reg, newVector);
	    }
	    if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		newVector = SetUsedRegister(op->a_index_reg, newVector);
	    }
	    if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)
		&& !OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		MALFUNCTION("Operand uses a register,"
			" but neither INDEX or REG flags set.");
	    }
	}				/* End test for register */
    }


    /* End source2 test */
    /*
     *                        EXPLICIT DESTINATION
     */
    if (INSTR_HAS_FLAG(instr, DEST_FLAG)) {
	op = instr->i_dest;
	CASSERT(op != NULL, "DEST_FLAG but no destination operand.");
	if (OperandUsesRegister(op)) {
	    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
		if (OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
		    newVector = SetUsedRegister(op->a_base_reg, newVector);
		}
		if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		    newVector = SetUsedRegister(op->a_index_reg, newVector);
		}
		if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)
		    && !OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		    MALFUNCTION("Operand uses a register,"
			" but neither INDEX or REG flags set.");
		}
	    }			/* end test for register indirect flag */
	}			/* end test for register */
    }				/* end explicit destination test */
    /*
     *                          IMPLICIT SOURCE1
     */
    if (DIS386_IMPLICIT_SOURCE1(disEntry) != 0) {
	if (DIS386_IMPLICIT_SOURCE1(disEntry) == reg_all) {
	    newVector = UseAllRegisters();
	} else {
	    newVector =
		SetUsedRegister(DIS386_IMPLICIT_SOURCE1(disEntry), newVector);
	}
    }

    /*
     *                          IMPLICIT SOURCE2
     */

    if (DIS386_IMPLICIT_SOURCE2(disEntry) != 0) {
	if (DIS386_IMPLICIT_SOURCE2(disEntry) == reg_all) {
	    newVector = UseAllRegisters();
	} else {
	    newVector =
		SetUsedRegister(DIS386_IMPLICIT_SOURCE2(disEntry), newVector);
	}
    }

    /*
     *                          DESTINATION USED
     */

    if (disEntry->x_psw & DUF) {
	op = instr->i_dest;
	CASSERT(op != NULL, "DUF flag with no destination operand.");
	if (OperandUsesRegister(op)) {
	    if (OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
		newVector = SetUsedRegister(op->a_base_reg, newVector);
	    }
	    if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		newVector = SetUsedRegister(op->a_index_reg, newVector);
	    }
	    if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)
		&& !OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		MALFUNCTION("Operand uses a register,"
			" but neither INDEX or REG flags set.");
	    }
	}				/* End test for register */
    }



    /* End destination-as-source test */
    /*
     *                         INSTRUCTION PREFIXES
     */
    if ((instr->i_prefixes & PREFIX_REPZ)
	|| (instr->i_prefixes & PREFIX_REPNZ))
	newVector = SetUsedRegister(REPEAT_REGISTER, newVector);


    return newVector;
}



/*
 *  ComputeRegistersDefinedHelper(instr) --
 *
 *  ComputeRegistersDefinedHelper will return a register bitvector with
 *  bits set for each register that the instruction defines.
 */
static REGISTER_BITVECTOR ComputeRegistersDefinedHelper(Instruction *instr) {
    REGISTER_BITVECTOR newVector = 0;
    Operand *op;
    X86op *disEntry = InstrGetOpcodeEntry(instr);

    // if the instruction is flagged as literal bytes then in some cases the
    // op code is not able to be found and things break, this is a HACK and
    // needs fixin' (cml)
    if (INSTR_HAS_FLAG(instr, INSTR_LITERAL_BYTES_FLAG)) {
	if (instr->i_literal_data[0] == 0x0f) {
	} else {
	    disEntry = &dis386[instr->i_literal_data[0]];
	}
    }

    if (disEntry == NULL) {
	return -1;			// (cml)
    }

    /*
     *                          EXPLICIT DESTINATION
     */

    if (INSTR_HAS_FLAG(instr, DEST_FLAG)) {
	op = instr->i_dest;
	if (OperandUsesRegister(op)
	    && !OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	    if (OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
		newVector = SetDefinedRegister(op->a_base_reg, newVector);
	    }
	    if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		newVector = SetDefinedRegister(op->a_index_reg, newVector);
	    }
	    if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)
		&& !OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		MALFUNCTION("Operand uses a register,"
			" but neither INDEX or REG flags set.");
	    }
	}				/* End test for register */
    }


    /* End destination test */
    /*
     *                          IMPLICIT DEST1
     */
    if (DIS386_IMPLICIT_DEST1(disEntry) != 0) {
	if (DIS386_IMPLICIT_DEST1(disEntry) == reg_all) {
	    newVector = DefineAllRegisters();
	} else {
	    newVector =
		SetDefinedRegister(DIS386_IMPLICIT_DEST1(disEntry), newVector);
	}
    }

    /*
     *                          IMPLICIT DEST2
     */

    if (DIS386_IMPLICIT_DEST2(disEntry) != 0) {
	if (DIS386_IMPLICIT_DEST2(disEntry) == reg_all) {
	    newVector = DefineAllRegisters();
	} else {
	    newVector =
		SetDefinedRegister(DIS386_IMPLICIT_DEST2(disEntry), newVector);
	}
    }

    /*
     *                        SOURCE1 DEFINED
     */

    if (disEntry->x_psw & S1D) {
	op = instr->i_src;
	CASSERT(op != NULL, "S1D flag but no source1 operand.");
	if (OperandUsesRegister(op)
	    && !OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	    if (OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
		newVector = SetDefinedRegister(op->a_base_reg, newVector);
	    }
	    if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		newVector = SetDefinedRegister(op->a_index_reg, newVector);
	    }
	    if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)
		&& !OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		MALFUNCTION("Operand uses a register,"
			" but neither INDEX or REG flags set.");
	    }
	}				/* End test for register */
    }



    /* End source used as destination test */
    /*
     *                        SOURCE2 DEFINED
     */
    if (disEntry->x_psw & S2D) {
	op = instr->i_src2;
	CASSERT(op != NULL, "S2D flag but no source2 operand.");
	if (OperandUsesRegister(op)
	    && !OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	    if (OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
		newVector = SetDefinedRegister(op->a_base_reg, newVector);
	    }
	    if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		newVector = SetDefinedRegister(op->a_index_reg, newVector);
	    }
	    if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)
		&& !OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
		MALFUNCTION("Operand uses a register,"
			" but neither INDEX or REG flags set.");
	    }
	}				/* End test for register */
    }



    /* End source2 used as destination test */
    /*
     *                         INSTRUCTION PREFIXES
     */
    if ((instr->i_prefixes & PREFIX_REPZ)
	|| (instr->i_prefixes & PREFIX_REPNZ)) {
	if (DEFINE_CONSERVATIVE == true) {
	    return 0;
	} else {
	    newVector = SetDefinedRegister(REPEAT_REGISTER, newVector);
	}
    }

    if (DEFINE_CONSERVATIVE == true) {
	if (isConditionalMove(instr)) {
	    newVector = newVector &
		(~(registerToBitRegister(instr->i_dest->a_base_reg)));
	}
    }

    return newVector;
}



/*
 *  SetUsedRegister(aRegister,currentVector) --
 *
 *  Sets the bit for the register specified by 'aRegister' in the bitvector
 *  'currentVector' and returns the new resulting bitvector.
 */
static int SetUsedRegister(int aRegister, int currentVector) {
    return currentVector | registerToBitRegister(aRegister);
}



/*
 *  SetDefinedRegister(aRegister,currentVector) --
 *
 *  Sets the bit for the register specified by 'aRegister' in the bitvector
 *  'currentVector' and returns the new resulting bitvector.
 */
static int SetDefinedRegister(int aRegister, int currentVector) {

    if (DEFINE_CONSERVATIVE == true) {
	if (RegisterIs32Bit(aRegister)) {
	    return currentVector | registerToBitRegister(aRegister);
	} else {
	    return currentVector;
	}
    }

    return currentVector | registerToBitRegister(aRegister);
}



/*
 *  registerToBitRegister(someRegister) --
 *
 *  Converts the register index (from x86-registers.h) into its bit in the
 *  register bitvector.
 */
extern int registerToBitRegister(int someRegister) {

    switch (someRegister) {

	/* These should never be used, but are here just in case */
    case reg_a:
	return BITVECTOR_REGISTER_EAX;
	break;
    case reg_b:
	return BITVECTOR_REGISTER_EBX;
	break;
    case reg_c:
	return BITVECTOR_REGISTER_ECX;
	break;
    case reg_d:
	return BITVECTOR_REGISTER_EDX;
	break;
    case reg_si:
	return BITVECTOR_REGISTER_ESI;
	break;
    case reg_di:
	return BITVECTOR_REGISTER_EDI;
	break;
    case reg_bp:
	return BITVECTOR_REGISTER_EBP;
	break;
    case reg_sp:
	return BITVECTOR_REGISTER_ESP;
	break;

	/* 32-bit general purpose registers */
    case reg_a | reg_32:
	return BITVECTOR_REGISTER_EAX;
	break;
    case reg_b | reg_32:
	return BITVECTOR_REGISTER_EBX;
	break;
    case reg_c | reg_32:
	return BITVECTOR_REGISTER_ECX;
	break;
    case reg_d | reg_32:
	return BITVECTOR_REGISTER_EDX;
	break;
    case reg_si | reg_32:
	return BITVECTOR_REGISTER_ESI;
	break;
    case reg_di | reg_32:
	return BITVECTOR_REGISTER_EDI;
	break;
    case reg_bp | reg_32:
	return BITVECTOR_REGISTER_EBP;
	break;
    case reg_sp | reg_32:
	return BITVECTOR_REGISTER_ESP;
	break;

	/* 16-bit general purpose registers */
    case reg_a | reg_16:
	return BITVECTOR_REGISTER_EAX;
	break;
    case reg_b | reg_16:
	return BITVECTOR_REGISTER_EBX;
	break;
    case reg_c | reg_16:
	return BITVECTOR_REGISTER_ECX;
	break;
    case reg_d | reg_16:
	return BITVECTOR_REGISTER_EDX;
	break;
    case reg_si | reg_16:
	return BITVECTOR_REGISTER_ESI;
	break;
    case reg_di | reg_16:
	return BITVECTOR_REGISTER_EDI;
	break;
    case reg_bp | reg_16:
	return BITVECTOR_REGISTER_EBP;
	break;
    case reg_sp | reg_16:
	return BITVECTOR_REGISTER_ESP;
	break;

	/* 8-bit general purpose registers */
    case reg_a | reg_8l:
	return BITVECTOR_REGISTER_EAX;
	break;
    case reg_b | reg_8l:
	return BITVECTOR_REGISTER_EBX;
	break;
    case reg_c | reg_8l:
	return BITVECTOR_REGISTER_ECX;
	break;
    case reg_d | reg_8l:
	return BITVECTOR_REGISTER_EDX;
	break;
    case reg_a | reg_8h:
	return BITVECTOR_REGISTER_EAX;
	break;
    case reg_b | reg_8h:
	return BITVECTOR_REGISTER_EBX;
	break;
    case reg_c | reg_8h:
	return BITVECTOR_REGISTER_ECX;
	break;
    case reg_d | reg_8h:
	return BITVECTOR_REGISTER_EDX;
	break;

    case reg_st:
	return BITVECTOR_REGISTER_ST;
	break;
    case reg_st0:
	return BITVECTOR_REGISTER_ST;
	break;
    case reg_st1:
	return BITVECTOR_REGISTER_ST;
	break;
    case reg_st2:
	return BITVECTOR_REGISTER_ST;
	break;
    case reg_st3:
	return BITVECTOR_REGISTER_ST;
	break;
    case reg_st4:
	return BITVECTOR_REGISTER_ST;
	break;
    case reg_st5:
	return BITVECTOR_REGISTER_ST;
	break;
    case reg_st6:
	return BITVECTOR_REGISTER_ST;
	break;
    case reg_st7:
	return BITVECTOR_REGISTER_ST;
	break;

    case reg_es:
        return BITVECTOR_REGISTER_ES;
        break;
    case reg_cs:
        return BITVECTOR_REGISTER_CS;
        break;
    case reg_ss:
        return BITVECTOR_REGISTER_SS;
        break;
    case reg_ds:
        return BITVECTOR_REGISTER_DS;
        break;
    case reg_fs:
        return BITVECTOR_REGISTER_FS;
        break;
    case reg_gs:
        return BITVECTOR_REGISTER_GS;
        break;   

	/* Special case for the indirect dx register */
    case reg_d | reg_16 | reg_indirect:
	return BITVECTOR_REGISTER_EDX;
	break;

    default:
	WARNING("[registerToBitRegister]"
	    " Could not map register %d to bit register.\n", someRegister);
	break;

    }					/* end switch over register */

    return BITVECTOR_REGISTER_NONE;

}



/*
 *  DefineAllRegisters() --
 *  Returns a bitvector in which all registers are defined.
 */
static REGISTER_BITVECTOR DefineAllRegisters() {
    return (BITVECTOR_REGISTER_EAX | BITVECTOR_REGISTER_EBX |
	BITVECTOR_REGISTER_ECX | BITVECTOR_REGISTER_EDX | BITVECTOR_REGISTER_EBP
	| BITVECTOR_REGISTER_ESP | BITVECTOR_REGISTER_EDI |
	BITVECTOR_REGISTER_ESI);
}



/*
 *  UseAllRegisters() --
 *  Returns a bitvector in which all registers are used.
 */
static REGISTER_BITVECTOR UseAllRegisters() {
    return (BITVECTOR_REGISTER_EAX | BITVECTOR_REGISTER_EBX |
	BITVECTOR_REGISTER_ECX | BITVECTOR_REGISTER_EDX | BITVECTOR_REGISTER_EBP
	| BITVECTOR_REGISTER_ESP | BITVECTOR_REGISTER_EDI |
	BITVECTOR_REGISTER_ESI);
}



/*
 *  xBitRegisterToRegister(i) --
 *
 *  Does the opposite of registerToBitVector, and returns the register that the
 *  bit 'i' in the register bitvector corresponds to.
 */
int xBitRegisterToRegister(int i) {

    if ((1 << i) == BITVECTOR_REGISTER_EAX) {
	return reg_a | reg_32;
    }
    if ((1 << i) == BITVECTOR_REGISTER_EBX) {
	return reg_b | reg_32;
    }
    if ((1 << i) == BITVECTOR_REGISTER_ECX) {
	return reg_c | reg_32;
    }
    if ((1 << i) == BITVECTOR_REGISTER_EDX) {
	return reg_d | reg_32;
    }
    if ((1 << i) == BITVECTOR_REGISTER_ESP) {
	return reg_sp | reg_32;
    }
    if ((1 << i) == BITVECTOR_REGISTER_EBP) {
	return reg_bp | reg_32;
    }
    if ((1 << i) == BITVECTOR_REGISTER_ESI) {
	return reg_si | reg_32;
    }
    if ((1 << i) == BITVECTOR_REGISTER_EDI) {
	return reg_di | reg_32;
    }
    if ((1 << i) == BITVECTOR_REGISTER_ST) {
	return reg_st | reg_32;
    }

    return -1;
}



/*
 *  _RegisterIs32Bit(aReg) --
 *  Returns true if the given register is a 32 bit register.
 */
bool _RegisterIs32Bit(int aReg) {

    switch (aReg) {
    case reg_a | reg_32:
	return true;
	break;
    case reg_b | reg_32:
	return true;
	break;
    case reg_c | reg_32:
	return true;
	break;
    case reg_d | reg_32:
	return true;
	break;
    case reg_si | reg_32:
	return true;
	break;
    case reg_di | reg_32:
	return true;
	break;
    case reg_sp | reg_32:
	return true;
	break;
    case reg_bp | reg_32:
	return true;
	break;
    default:
	return false;
	break;
    }

    return false;
}



/*
 *  _RegisterIs16Bit(aReg) --
 *  Returns true if the given register is a 16 bit register.
 */
bool _RegisterIs16Bit(int aReg) {

    switch (aReg) {
    case reg_a | reg_16:
	return true;
	break;
    case reg_b | reg_16:
	return true;
	break;
    case reg_c | reg_16:
	return true;
	break;
    case reg_d | reg_16:
	return true;
	break;
    case reg_si | reg_16:
	return true;
	break;
    case reg_di | reg_16:
	return true;
	break;
    case reg_sp | reg_16:
	return true;
	break;
    case reg_bp | reg_16:
	return true;
	break;
    default:
	return false;
	break;
    }

    return false;
}



/*
 *  _RegisterIs8Bit(aReg) --
 *  Returns true if the given register is a 8 bit register.
 */
bool _RegisterIs8Bit(int aReg) {

    switch (aReg) {
    case reg_a | reg_8h:
	return true;
	break;
    case reg_a | reg_8l:
	return true;
	break;
    case reg_b | reg_8h:
	return true;
	break;
    case reg_b | reg_8l:
	return true;
	break;
    case reg_c | reg_8h:
	return true;
	break;
    case reg_c | reg_8l:
	return true;
	break;
    case reg_d | reg_8h:
	return true;
	break;
    case reg_d | reg_8l:
	return true;
	break;
    default:
	return false;
	break;
    }

    return false;
}



/*
 *  xComputePSWUsed(instr) --
 *  Returns a PSW_BITVECTOR with a bit set for each flag that 'instr' may use.
 */
PSW_BITVECTOR xComputePSWUsed(Instruction *instr) {
    return ((InstrGetOpcodeEntry(instr)->x_pswUsed) & (~(PSW_PSEUDO_FLAGS)));

}



/*
 *  xComputePSWDefined(instr) --
 *  Returns a PSW_BITVECTOR with a bit set for each flag that 'instr' defines.
 */
PSW_BITVECTOR xComputePSWDefined(Instruction *instr) {
    X86op *disEntry;

    if ((instr->i_prefixes&PREFIX_REPZ) || (instr->i_prefixes&PREFIX_REPNZ)) {
        if (instr->i_class != CLASS_CMPS &&
            instr->i_class != CLASS_SCAS) {
            return 0;
        }
        disEntry = InstrGetOpcodeEntry(instr);
        return disEntry->x_psw & ~PSW_PSEUDO_FLAGS;
    } else {
	// if the instruction is flagged as literal bytes then in some cases the
	// op code is not able to be found and things break, this is a HACK and
	// needs fixin' (cml)
	if (INSTR_HAS_FLAG(instr, INSTR_LITERAL_BYTES_FLAG)) {
	    if (instr->i_literal_data[0] == 0x0f) {
	    } else {
		disEntry = &dis386[instr->i_literal_data[0]];
	    }
	} else {
	    disEntry = InstrGetOpcodeEntry(instr);
	}

	if (disEntry == NULL) {
	    return -1;
	}

	return disEntry->x_psw & ~PSW_PSEUDO_FLAGS;
    }
}



/*
 *  xComputePSWDefinedLiberal(instr) --
 *
 *  Returns a PSW_BITVECTOR with a bit set for each flag that 'instr' may
 *  define.
 */
PSW_BITVECTOR xComputePSWDefinedLiberal(Instruction *instr) {
    return ((InstrGetOpcodeEntry(instr)->x_psw) & (~(PSW_PSEUDO_FLAGS)));
}



/*
 *  xComputeStackUsed(instr,slots,bp_index,arsize) --
 *
 *  Given an instruction, an array of stack slots and an index for EBP into the
 *  array, ComputeStackUsed will fill in the slots with true if the instruction
 *  may use that location on the stack.  false will be filled in when
 *  an instruction does definitely not use that slots on the stack.
 */
void xComputeStackUsed(Instruction *instr,int *slots,int bp_index,int arsize) {
    X86op *disEntry = InstrGetOpcodeEntry(instr);
    Edge *callEdge;
    Function *callee;
    Function *fun = INSTR_FUNCTION(instr);
    int i;

    CASSERT(disEntry != NULL, "Could not find opcode for instruction.");

    /*
     * Function calls will be treated as using everything on the stack, since
     * we don't know what it might use.
     */

    if (isCallClass(instr)) {
	int unknownIndex;
	int callOffset = InstrBPSPDiff(instr);
	callEdge = BBLGetFunDirectEdge(instr->i_Bblock);
	callee = NULL;
	if (callEdge) {
	    callee = callEdge->e_to->b_fun;
	}

	if ((callee == NULL) || (callee->f_useDepth == -1) ||
	    (callEdge == NULL) || (callOffset == STACK_UNKNOWN)) {
	    for (i = 0; i <= arsize; i++) {
		slots[i] = true;
	    }
	    return;

	} else {
	    unknownIndex = (callOffset / ADDRESS_SIZE) + 2 -
		(callee->f_useDepth / ADDRESS_SIZE) +
		(fun->f_callerFrameSize / ADDRESS_SIZE);

	    for (i = 0;
		i <
		(fun->f_stackFrameSize +
		    fun->f_callerFrameSize) / ADDRESS_SIZE; i++) {
		if (i >= unknownIndex) {
		    slots[i] = true;
		}
	    }
	}
    }

    if (isFPClass(instr) && ((instr->i_src && instr->i_src->a_displacement)
	    || (instr->i_src2 && instr->i_src2->a_displacement))) {
	for (i = 0; i <= arsize; i++) {
	    slots[i] = true;
	}
	return;
    }



    /*
     * Hell edges should be treated as using all stack locations as well, since
     * HELL represents unknown control transfer. We'll let the instructions
     * that generate the HELL edges use everything.
     */

    if (isPCChangingClass(instr) && BBLHasHellSuccessor(instr->i_Bblock)) {
	for (i = 0; i <= arsize; i++) {
	    slots[i] = true;
	}
	return;
    }

    /*
     * If the instruction is surrounded by other instructions that redefine the
     * frame pointer because we inserted compensation code for a use of EBP
     * during inlining, then we should say that it uses all slots on the stack
     * to be safe. Why? Consider the following example:
     *
     *     805cb10:       56                      push   %esi
     *     805cb11:       81 c5 fc ff ff ff       add    $0xfffffffc,%ebp
     *     805cb17:       55                      push   %ebp
     *     805cb18:       81 c5 04 00 00 00       add    $0x4,%ebp
     *
     * Since at 805cb10 and 805cb17 the value of (ESP - EBP) is the same, we
     * would assign these pushes to the same location. But in reality they don't
     * write to the same location. To be conservative, we'll say 805cb17 uses
     * all of the stack and defines none of the stack.
     */

    if (INSTR_HAS_FLAG(instr, INSTR_USES_FP_BUT_COMPENSATED_FLAG)) {
	for (i = 0; i <= arsize; i++) {
	    slots[i] = true;
	}
	return;
    }



    /*
     * Pop instructions implicitly use the stack.
     */

    if ((instr->i_class == CLASS_POPF)
	|| (instr->i_class == CLASS_POPFD)
	|| (instr->i_class == CLASS_POP) || isLeaveClass(instr)) {
	int diff = InstrBPSPDiff(instr);
	CASSERT(instr->i_class != CLASS_POPALL, "popa encountered.");
	if ((diff == STACK_UNKNOWN) || ((bp_index + diff / ADDRESS_SIZE) < 0)
	    || ((bp_index + diff / ADDRESS_SIZE) >= arsize)) {
	    for (i = 0; i <= arsize; i++) {
		slots[i] = true;
	    }
	    return;
	}
	CASSERT((bp_index + diff / ADDRESS_SIZE) >= 0,
	    "POP/LEAVE defines bad location.");
	slots[bp_index + diff / ADDRESS_SIZE] = true;
	return;
    }

    if (instr->i_src) {
	ComputeStackLocationUsedByOperand(instr->i_src, slots, bp_index,
	    arsize, UNKNOWN_TRUE);
    }
    if (instr->i_src2) {
	ComputeStackLocationUsedByOperand(instr->i_src2, slots, bp_index,
	    arsize, UNKNOWN_TRUE);
    }
    if ((instr->i_dest)
	&& OPERAND_HAS_FLAG(instr->i_dest, OPERAND_INDIRECT_FLAG))
	ComputeStackLocationUsedByOperand(instr->i_dest, slots, bp_index,
	    arsize, UNKNOWN_TRUE);
    if (disEntry->x_psw & DUF) {
	ComputeStackLocationUsedByOperand(instr->i_dest, slots, bp_index,
	    arsize, UNKNOWN_TRUE);
    }

    return;
}



/*
 *  xComputeStackDefinedLiberal(instr,slots,bp_index,arsize) --
 *
 *  Fills in the integer array 'slots' with true/false entries if the
 *  instruction 'instr' may define the stack location.
 */
void xComputeStackDefinedLiberal(Instruction *instr, int *slots, int bp_index,
    int arsize) {

    DEFINE_CONSERVATIVE = false;
    xComputeStackDefined(instr, slots, bp_index, arsize);
    DEFINE_CONSERVATIVE = true;

    return;
}



/*
 *  xComputeStackDefined(instr,slots,bp_index,arsize) --
 *
 *  Given an instruction, an array of stack slots and an index for EBP into the
 *  array, ComputeStackDefined will fill in the slots with true if the
 *  instruction may define that location on the stack.  false will be filled in
 *  when an instruction does definitely not define that slots on the stack.
 */
void xComputeStackDefined(Instruction *instr, int *slots, int bp_index,
    int arsize) {
    X86op *disEntry = InstrGetOpcodeEntry(instr);
    int i;
    int unknown;

    CASSERT(disEntry != NULL, "Could not find opcode for instruction.");

    /*
     * Returns will be treated as defining everything on the stack, since
     * the activation record is wiped after executing it.
     */

    if (isReturnClass(instr)) {
	for (i = 0; i <= arsize; i++) {
	    slots[i] = true;
	}
	return;
    }

    /*
     * If the instruction is surrounded by other instructions that redefine the
     * frame pointer because we inserted compensation code for a use of EBP
     * during inlining, then we should say that it uses all slots on the stack
     * to be safe. Why? Consider the following example:
     *
     *     805cb10:       56                      push   %esi
     *     805cb11:       81 c5 fc ff ff ff       add    $0xfffffffc,%ebp
     *     805cb17:       55                      push   %ebp
     *     805cb18:       81 c5 04 00 00 00       add    $0x4,%ebp
     *
     * Since at 805cb10 and 805cb17 the value of (ESP - EBP) is the same, we
     * would assign these pushes to the same location. But in reality they don't
     * write to the same location. To be conservative, we'll say 805cb17 uses
     * all of the stack and defines none of the stack.
     */

    if (INSTR_HAS_FLAG(instr, INSTR_USES_FP_BUT_COMPENSATED_FLAG)) {
	if (DEFINE_CONSERVATIVE == false) {
	    for (i = 0; i <= arsize; i++) {
		slots[i] = true;
	    }
	}
	return;
    }

    /*
     * Push instructions implicitly define the stack.
     */

    if ((instr->i_class == CLASS_PUSH) || (instr->i_class == CLASS_PUSHF) ||
	    (instr->i_class == CLASS_PUSHFD)) {
	int diff = InstrBPSPDiff(instr);
	CASSERT(instr->i_class != CLASS_PUSHALL, "pusha encountered.");

	if ((diff == STACK_UNKNOWN)
	    || ((bp_index + 1 + diff / ADDRESS_SIZE) < 0)
	    || ((bp_index + 1 + diff / ADDRESS_SIZE) >= arsize)) {
	    if (DEFINE_CONSERVATIVE == false) {
		for (i = 0; i <= arsize; i++) {
		    slots[i] = true;
		}
	    }
	    return;
	}

	CASSERT((bp_index + 1 + diff / ADDRESS_SIZE) >= 0,
	    "Accessing negative slot in liveness array.");

	slots[bp_index + 1 + diff / ADDRESS_SIZE] = true;
	return;
    }

    unknown = DEFINE_CONSERVATIVE ? UNKNOWN_FALSE : UNKNOWN_TRUE;

    if (instr->i_dest) {
	ComputeStackLocationUsedByOperand(instr->i_dest, slots, bp_index,
	    arsize, unknown);
    }
    if (disEntry->x_psw & S1D) {
	ComputeStackLocationUsedByOperand(instr->i_src, slots, bp_index,
	    arsize, unknown);
    }
    if (disEntry->x_psw & S2D) {
	ComputeStackLocationUsedByOperand(instr->i_src2, slots, bp_index,
	    arsize, unknown);
    }

    return;
}



/*
 *  ComputeStackLocationUsedByOperand(op,slots,bp_index,arsize,flg) --
 *
 *  Fills in the array 'slots' with true/false if the operand operates
 *  on that stack location.  If the location cannot be determined,
 *  all slots will be set to either false or true depending on 'flg'.
 */
void ComputeStackLocationUsedByOperand(Operand *op, int *slots, int bp_index,
    int arsize, int flg) {
    int i;

    /*
     * Indirect operands are hard to deal with. We either must be able to know
     * the contents of a memory location or a register. In general this is
     * hard to determine, so we are conservative and fill in everything.
     */

    if (OPERAND_HAS_FLAG(op, OPERAND_INDIRECT_FLAG)) {
	if (flg == UNKNOWN_TRUE) {
	    for (i = 0; i <= arsize; i++) {
		slots[i] = true;
	    }
	}
	return;
    }

    if (!OperandUsesRegister(op)) {
	return;
    }

    if (OperandAccessesStackFromFramePointer(op)) {
	int idx = op->a_displacement / ADDRESS_SIZE;

	if (((bp_index - idx) < 0) || ((bp_index - idx) >= arsize)
	    || ((op->a_displacement % ADDRESS_SIZE) != 0)) {
	    if (flg == UNKNOWN_TRUE) {
		for (i = 0; i <= arsize; i++) {
		    slots[i] = true;
		}
	    }
	    return;
	}

	if (OPERAND_HAS_FLAG(op, OPERAND_BMODE_FLAG)) {
	    if (flg == UNKNOWN_TRUE) {
		slots[bp_index - idx] = true;
	    }
	    return;
	}

	slots[bp_index - idx] = true;
	return;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	if (OPERAND_HAS_FLAG(op, OPERAND_RELOCATABLE_ABS_FLAG)) {
	    return;
	}
	if (flg == UNKNOWN_TRUE) {
	    for (i = 0; i <= arsize; i++) {
		slots[i] = true;
	    }
	}
	return;
    }
}



/*
 *  xRegisterIsLiveInBitvector(aReg,vec) --
 *  Returns true if the bit for register 'aReg' is set in the bitvector 'vec'.
 */
bool xRegisterIsLiveInBitvector(int aReg, REGISTER_BITVECTOR vec) {
    if ((registerToBitRegister(aReg) & vec)) {
	return true;
    }
    return false;
}



/*
 *  xComputeRegistersDefinedLiberal(instr) --
 *
 *  Returns a register bitvector with the bits set for the registers that
 *  'instr' may partially or completely define.
 */
REGISTER_BITVECTOR xComputeRegistersDefinedLiberal(Instruction *instr) {
    REGISTER_BITVECTOR vec;

    DEFINE_CONSERVATIVE = false;
    vec = ComputeRegistersDefinedHelper(instr);
    DEFINE_CONSERVATIVE = true;

    return vec;
}



/*
 *  xComputeRegistersDefined(instr) --
 *
 *  Returns a register bitvector with the bits set for the registers that
 *  'instr' completely defines.
 */
REGISTER_BITVECTOR xComputeRegistersDefined(Instruction *instr) {
    REGISTER_BITVECTOR vec;

    DEFINE_CONSERVATIVE = true;
    vec = ComputeRegistersDefinedHelper(instr);

    return vec;
}
