/*
 *  x86/op.c -- Routines for operand queries
 */
#include "global.h"
#include "x86-registers.h"
#include "x86-op.h"
#include "x86-table-defs.h"
#include "x86-util.h"
#include "utilities.h"
#include "sysdep.h"



/*
 *  GetByteMode(op,dflag) --
 *
 *  Returns either v_mode, b_mode, d_mode or w_mode (16/32, 8, 32, 16 bits
 *  respectively) based on flags set in the operand. A valid bytemode can
 *  also be a register. See x86-table.h for values
 */
int GetByteMode(Operand *op, int dflag) {
    if (op == NULL) {
	return 0;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_RELOCATED)) {
	return v_mode;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_VMODE_FLAG)) {
	return v_mode;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_BMODE_FLAG)) {
	return b_mode;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_WMODE_FLAG)) {
	return w_mode;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_DMODE_FLAG)) {
	return d_mode;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_OPREG_FLAG)) {
	int base;
	base = op->a_base_reg;

	switch (base) {
	    /* 32-bit general purpose registers */
	case reg_a | reg_32:
	    return eAX_reg;
	    break;
	case reg_b | reg_32:
	    return eBX_reg;
	    break;
	case reg_c | reg_32:
	    return eCX_reg;
	    break;
	case reg_d | reg_32:
	    return eDX_reg;
	    break;
	case reg_si | reg_32:
	    return eSI_reg;
	    break;
	case reg_di | reg_32:
	    return eDI_reg;
	    break;
	case reg_bp | reg_32:
	    return eBP_reg;
	    break;
	case reg_sp | reg_32:
	    return eSP_reg;
	    break;

	    /* 16-bit general purpose registers */
	case reg_a | reg_16:
	    if (!dflag) {
		return eAX_reg;
	    } else {
		return ax_reg;
	    }
	    break;
	case reg_b | reg_16:
	    if (!dflag) {
		return eBX_reg;
	    } else {
		return bx_reg;
	    }
	    break;
	case reg_c | reg_16:
	    if (!dflag) {
		return eCX_reg;
	    } else {
		return cx_reg;
	    }
	    break;
	case reg_d | reg_16:
	    if (!dflag) {
		return eDX_reg;
	    } else {
		return dx_reg;
	    }
	    break;
	case reg_si | reg_16:
	    if (!dflag) {
		return eSI_reg;
	    } else {
		return si_reg;
	    }
	    break;
	case reg_di | reg_16:
	    if (!dflag) {
		return eDI_reg;
	    } else {
		return di_reg;
	    }
	    break;
	case reg_bp | reg_16:
	    if (!dflag) {
		return eBP_reg;
	    } else {
		return bp_reg;
	    }
	    break;
	case reg_sp | reg_16:
	    if (!dflag) {
		return eSP_reg;
	    } else {
		return sp_reg;
	    }
	    break;

	    /* 8-bit general purpose registers */
	case reg_a | reg_8l:
	    return al_reg;
	    break;
	case reg_b | reg_8l:
	    return bl_reg;
	    break;
	case reg_c | reg_8l:
	    return cl_reg;
	    break;
	case reg_d | reg_8l:
	    return dl_reg;
	    break;
	case reg_a | reg_8h:
	    return ah_reg;
	    break;
	case reg_b | reg_8h:
	    return bh_reg;
	    break;
	case reg_c | reg_8h:
	    return ch_reg;
	    break;
	case reg_d | reg_8h:
	    return dh_reg;
	    break;

	    /* Special case for the indirect dx register */
	case reg_d | reg_16 | reg_indirect:
	    return indir_dx_reg;
	    break;

	}
    }
    return 0;
}



/*
 *  OperandClearBytemodeFlags(op) --
 *
 *  Turn off all possible bytemodes the instruction may have.
 *  Presumably the calling function will turn a different one on.
 */
void OperandClearBytemodeFlags(Operand *op) {
    op->a_flags = op->a_flags & ~OPERAND_BMODE_FLAG;
    op->a_flags = op->a_flags & ~OPERAND_VMODE_FLAG;
    op->a_flags = op->a_flags & ~OPERAND_DMODE_FLAG;
    op->a_flags = op->a_flags & ~OPERAND_WMODE_FLAG;
}



/*
 *  xOperandUsesRegister(op) --
 *  Returns true if the operand uses a register (either index or base).
 */
bool xOperandUsesRegister(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
	return true;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
	return true;
    }

    return false;
}



/*
 *  getOperandBytes(op,dflag) --
 *  Returns the number of bytes the operand will take up in the binary.
 */
int getOperandBytes(Operand *op, int dflag) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (OPERAND_HAS_FLAG(op, OPERAND_VMODE_FLAG)) {
	if (dflag) {
	    return 4;
	} else {
	    return 2;
	}
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_DMODE_FLAG)) {
	return 4;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_WMODE_FLAG)) {
	return 2;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_BMODE_FLAG)) {
	return 1;
    }

    return -1;
}



/*
 *  xOperandIsGPRegister(op) --
 *  Returns true if the operand is a general purpose register.
 */
bool xOperandIsGPRegister(Operand *op) {
    int the_reg = op->a_base_reg;

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_SCALE_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	return false;
    }
    if (RegisterIsFloatingPoint(the_reg)) {
	return false;
    }

    return true;
}



/*
 *  GetOpRoutine(instr,op,bytemode1,bytemode2) --
 *
 *  A bunch of special case tests to determine what op_rtn the
 *  instruction should use.
 */
op_rtn GetOpRoutine(Instruction *instr, Operand *op, int bytemode1,
    int bytemode2) {

    CASSERT(instr != NULL, "Passed null instruction.");

    if (op == NULL) {
	return NULL;
    }

//    if (OPERAND_HAS_FLAG(op, OPERAND_CONTROL_REG_FLAG)) {
//	return OP_C;
//    }

    if (OPERAND_HAS_FLAG(op, OPERAND_DEBUG_REG_FLAG)) {
	REPORT(RP_TRIVIAL, "Operand has Debug Reg Flag Set");
	return OP_D;
    }

    if (INSTR_HAS_FLAG(instr, INDIRECT)) {
	return OP_indirE;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_OPG_FLAG)) {
	return OP_G;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG)) {
	CASSERT(!OPERAND_HAS_FLAG(op, OPERAND_SCALE_FLAG),
	    "Immediate with scale.");
	CASSERT(!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG),
	    "Immediate with base.");
	CASSERT(!OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG),
	    "Immediate with index.");
	if (OPERAND_HAS_FLAG(op, OPERAND_SIGN_EXTEND_FLAG)) {
	    return OP_sI;
	}
	return OP_I;
    }

    if (instr->i_class == 0x22) {
        return OP_PREFETCH;
    }
    
    if (OperandHasSIB(op)) {
	return OP_E;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_SIDI_FLAG) &&
	((op->a_base_reg == reg_di) ||
	    (op->a_base_reg == (reg_di | reg_32)) ||
	    (op->a_base_reg == (reg_di | reg_16)) ||
	    (op->a_base_reg == (reg_di | reg_8l)) ||
	    (op->a_base_reg == (reg_di | reg_8h))))
	return OP_ESDI;

    if (OPERAND_HAS_FLAG(op, OPERAND_SIDI_FLAG) &&
	((op->a_base_reg == reg_si) ||
	    (op->a_base_reg == (reg_si | reg_32)) ||
	    (op->a_base_reg == (reg_si | reg_16)) ||
	    (op->a_base_reg == (reg_si | reg_8l)) ||
	    (op->a_base_reg == (reg_si | reg_8h))))
	return OP_DSSI;

    if (isOPOFFClass(instr)) {
	if ((bytemode2 == al_reg) || (bytemode2 == eAX_reg)) {
	    return OP_OFF;
	}
    }

    if (isOPDIRClass(instr)) {
	if (OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG)) {
	    return OP_DIR;
	}
    }

    if (isJumpClass(instr)) {
	return OP_J;
    }


    if (op->a_base_reg == reg_st) {
	return OP_ST;
    }

    if (isSTRegister(op->a_base_reg)) {
	return OP_STi;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_OPREG_FLAG)) {
	return OP_REG;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_SEGMENT_REG_FLAG)) {
	REPORT(RP_TRIVIAL, "CONTROL_REG_FLAG SET !!!");
	return OP_SEG;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_CONTROL_REG_FLAG)) {
	REPORT(RP_TRIVIAL, "CONTROL_REG_FLAG SET [SRC1] !!! %d", OP_C);
	return OP_C;
    }


    return OP_E;
}



/*
 *  OperandHasSIB(op) --
 *
 *  Returns true if the operand uses a Scale-Index-Base byte or false
 *  if it doesn't.
 */
bool OperandHasSIB(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (OPERAND_HAS_FLAG(op, OPERAND_SCALE_FLAG)) {
	return true;
    }

    return false;
}



/*
 *  xOperandAccessesStackFromFramePointer(op) --
 *
 *  Returns true if the operand accesses the stack using EBP.
 *  Examples are:  0xfffffff4(%ebp), 0x4(%ebp) and (%ebp).
 */
bool xOperandAccessesStackFromFramePointer(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (!OperandUsesRegister(op)) {
	return false;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {

	if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
	    return false;
	}

	if (OPERAND_HAS_FLAG(op, OPERAND_SCALE_FLAG)) {
	    return false;
	}

	if (OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
	    if (OPERAND_HAS_FLAG(op, OPERAND_RELOCATABLE_ABS_FLAG)) {
		return false;
	    }
	}

	if (!OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG)) {
	    if (!OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
		return false;
	    }
	}

	if (RegisterIsFramePointer(op->a_base_reg)) {
	    return true;
	}

    }

    return false;
}



/*
 *  xOperandAccessesStackFromFramePointerWithSIB(op) --
 *
 *  Returns true if the Operand accesses the stack using the frame pointer in
 *  SIB mode.
 */
bool xOperandAccessesStackFromFramePointerWithSIB(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (!OPERAND_HAS_FLAG(op, OPERAND_SCALE_FLAG)) {
	return false;
    }

    if (!OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
	return false;
    }

    if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
	return false;
    }

    if (!OperandUsesRegister(op)) {
	return false;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {

	if (OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
	    if (OPERAND_HAS_FLAG(op, OPERAND_RELOCATABLE_ABS_FLAG)) {
		return false;
	    }
	}

	if (!OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG)) {
	    if (!OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
		return false;
	    }
	}

	if (RegisterIsFramePointer(op->a_base_reg)) {
	    return true;
	}

	if (RegisterIsFramePointer(op->a_index_reg)) {
	    return true;
	}

    }

    return false;
}



/*
 *  xOperandAccessesStackFromStackPointer(op) --
 *
 *  Returns true if the operand accesses the stack using ESP.
 *  Examples are:  0xfffffff4(%esp), 0x4(%esp) and (%esp).
 */
bool xOperandAccessesStackFromStackPointer(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (!OperandUsesRegister(op)) {
	return false;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
	    return false;
	}
//      -- MR --
//      Why is this not a valid AccessesStackFromStackPointer
//          mov 44(%esp,1) <-- eax     [REF: mpeg_play.reloc, XWindows]
//      Unsetting this flag check does the trick.
//      if (OPERAND_HAS_FLAG(op, OPERAND_SCALE_FLAG))
//         return false;
//      -- MR --
	if (OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
	    if (OPERAND_HAS_FLAG(op, OPERAND_RELOCATABLE_ABS_FLAG)) {
		return false;
	    }
	}
	if (!OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG)) {
	    if (!OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
		return false;
	    }
	}
	if (RegisterIsStackPointer(op->a_base_reg)) {
	    return true;
	}
    }

    return false;
}



/*
 *  xOperandIsImmediate(op) --
 *  Returns true if the operand is an immediate value (constant).
 */
bool xOperandIsImmediate(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (OperandUsesRegister(op)) {
	return false;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG)) {
	return true;
    }

    return false;

}



/*
 *  xOperandIsAddress(op) -- returns true if the operand is an address.
 */
bool xOperandIsAddress(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (OperandUsesRegister(op)) {
	return false;
    }

    if (!OPERAND_HAS_FLAG(op, OPERAND_RELOCATABLE_ABS_FLAG)) {
	return false;
    }

    if (!OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG)) {
	return false;
    }

    return true;

}



/*
 *  xOperandUsesContentsOfStackPointer(op) --
 *
 *  If the supplied operand uses the contents of the stack pointer register,
 *  true is returned. Otherwise false is returned. By 'use', we mean a direct
 *  reference such as:
 *	mov eax <- esp
 *	add ebx <- esp + 4
 *	mov 0x8051000 <- esp
 *  In all the examples, another register or memory location is getting the
 *  value of the stack pointer. The following example should return false:
 *	mov eax <- 0x4(esp)
 *  Note that an instruction like "lea eax <- 0x4(esp)" really uses the
 *  contents, but will return false. Since this function doesn't know about
 *  the class of the instruction, any function calling this should check
 *  beforehand.
 */
bool xOperandUsesContentsOfStackPointer(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (!OperandUsesRegister(op)) {
	return false;
    }

    if (!OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	if (RegisterIsStackPointer(op->a_base_reg)) {
	    return true;
	}
	if (RegisterIsStackPointer(op->a_index_reg)) {
	    return true;
	}
    }

    return false;
}



/*
 *  xOperandUsesContentsOfFramePointer(op) --
 *
 *  If the supplied operand uses the contents of the frame pointer register,
 *  true is returned. Otherwise false is returned. By 'use', we mean a direct
 *  reference such as:
 *	mov eax <- ebp
 *	add ebx <- ebp + 4
 *	mov 0x8051000 <- ebp
 *  In all the examples, another register or memory location is getting the
 *  value of the frame pointer. The following example should return false:
 *	mov eax <- 0x4(ebp)
 *  Note that an instruction like "lea eax <- 0x4(ebp)" really uses the
 *  contents, but will return false. Since this function doesn't know about
 *  the class of the instruction, any function calling this should check
 *  beforehand.
 */
bool xOperandUsesContentsOfFramePointer(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (!OperandUsesRegister(op)) {
	return false;
    }

    if (!OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	if (RegisterIsFramePointer(op->a_base_reg)) {
	    return true;
	}
	if (RegisterIsFramePointer(op->a_index_reg)) {
	    return true;
	}
    }

    return false;
}



/*
 *  xOperandIsFramePointer(op) --
 *  Returns true if the operand is the frame pointer.
 */
bool xOperandIsFramePointer(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_SCALE_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	return false;
    }

    return RegisterIsFramePointer(op->a_base_reg);
}



/*
 *  xOperandIsStackPointer(op) --
 *  Returns true if the operand is the stack pointer.
 */
bool xOperandIsStackPointer(Operand *op) {

    CASSERT(op != NULL, "Passed a NULL operand.");

    if (!OPERAND_HAS_FLAG(op, OPERAND_BASE_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_INDEX_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_SCALE_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	return false;
    }

    return RegisterIsStackPointer(op->a_base_reg);
}



/*
 *  xOperandIsAbsoluteMemoryLocation(op) --
 *  Returns true if the operand is an absolute memory location.
 */
bool xOperandIsAbsoluteMemoryLocation(Operand *op) {
    if (!OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)
	&& !OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG))
	return false;
    if (!OPERAND_HAS_FLAG(op, OPERAND_RELOCATABLE_ABS_FLAG)) {
	return false;
    }
    if (OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG)) {
	return false;
    }
    if (OperandUsesRegister(op)) {
	return false;
    }
    if (OperandHasSIB(op)) {
	return false;
    }
    return true;
}



/*
 *  xOperandWritesToUnknownStackSlot(op) --
 *
 *  Returns true if the operand may write to the runtime stack,
 *  but we can't determine for sure if it does.
 */
bool xOperandWritesToUnknownStackSlot(Operand *op) {

    if (OPERAND_HAS_FLAG(op, OPERAND_INDIRECT_FLAG)) {
	return true;
    }

    if (!OperandUsesRegister(op)) {
	return false;
    }

    if (OperandAccessesStackFromFramePointer(op)) {
	return false;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	return true;
    }

    return false;
}



/*
 *  xOperandIsMemory(op) --
 *  Returns true if the operand is ultimately a memory location.
 */
bool xOperandIsMemory(Operand *op) {

    if (OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG)) {
	return false;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_INDIRECT_FLAG)) {
	return true;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_REGISTER_INDIRECT)) {
	return true;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_RELOCATABLE_ABS_FLAG)) {
	return true;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_DISP_A_FLAG)) {
	return true;
    }

    return false;
}



/*
 *  xOperandUseLargestBytemode(op) --
 *  Sets the operand to use the largest number of bytes possible.
 */
void xOperandUseLargestBytemode(Operand *op) {

    if (OperandUsesRegister(op)) {
	return;
    }

    OperandClearBytemodeFlags(op);
    OP_SET_FLAG(op, OPERAND_VMODE_FLAG);

    return;
}
