/*
 *  x86/profdump.c
 *
 *  Inserts code to dump edge profiles gathered at runtime.
 *  Code is very much machine dependent. The heart of this module
 *  lies in InsertDumpBlocks, which adds the routine for dumping
 *  the edge-profiling counters and basic block addresses.
 *
 *  Notes:
 *
 *  1) The memory addresses associated with the edges in libc_write,
 *  libc_open and libc_close will be changed when the profiles are
 *  being dumped.  Thus, the profiles for these three routines may
 *  not be entirely accurate.
 */
#include "global.h"
#include "flowgraph.h"
#include "linkage.h"
#include "utilities.h"
#include "x86-icreate.h"
#include "x86-profdump.h"
#include "x86-dis.h"
#include "x86-registers.h"
#include "x86-classes.h"
#include "relocate.h"
#include "upm.h"
#include "sysdep.h"
#include "cp.h"

#define DEBUG_CONSIST 0

void CreateDumpBlocks(Block *block, char *Filename);

static int edgeProfileFilenameBase;	// offset to filename in .rodata



/*
 *  xInsertDumpCode(Filename) --
 */
void xInsertDumpCode(char *Filename) {
    unsigned char *pstart;
    unsigned int offset;
    Function *dump_fun;
    Symbol *exitSym;
    Instruction *exitInstr, *lastInstr, *jmp;
    Block *newBlock, *lastBlock;
    UINT32 exitAddr;

    exitSym = EXIT_SYMBOL;

    // Make room for the filename we will add
    edgeProfileFilenameBase =
	SECTION_REF_RODATA->s_datasize + SECTION_REF_RODATA->s_growth;

    if (exitSym == NULL) {
	FATAL("[X86: InsertDumpCode] Could not locate exit symbol.");
    }

    exitAddr = exitSym->sy_value;
    exitInstr = AddressToInstr(exitAddr);
    if (exitInstr == NULL) {
	FATAL("[X86: InsertDumpCode]"
	    " Could not locate instruction for exit symbol.");
    }

    lastBlock = bbl_list;
    while (lastBlock->b_next != NULL) {
	lastBlock = lastBlock->b_next;
    }

    /*
     * Insert the call to dump profiles before the last instruction in this
     * basic block, so that its profiling code gets executed and we have the
     * correct basic block counter.
     */
    lastInstr = exitInstr->i_Bblock->b_last;
    BBLDivideBlock(exitInstr->i_Bblock, lastInstr);
    jmp = xInstrCreateCALLDisp();
    InsertInstrAfterInstr(jmp, lastInstr->i_prev);

    dump_fun = FunNew();
    newBlock = BBLCreate();
    dump_fun->f_first = newBlock;
    dump_fun->f_last = newBlock;

    dump_fun->f_name = zalloc(strlen("profilingdumpfcn"));
    strcpy(dump_fun->f_name, "profilingdumpfcn");
    BBLAddEdge(dump_fun->f_entry, newBlock, ET_ENTRY);
    AddNewFunctionSymbol(dump_fun);

    CASSERT(jmp->i_Bblock == exitInstr->i_Bblock,
	"Could not add jump instruction to same block as exit instr.");

    BBLRemoveEdge(jmp->i_Bblock, jmp->i_next->i_Bblock);
    jmp->i_Bblock->b_type = BT_CALL;
    BBLAddEdge(jmp->i_Bblock, newBlock, ET_FUNDIRECT);
    BBLAddEdge(jmp->i_Bblock, jmp->i_next->i_Bblock, ET_FUNLINK);
    INSTR_SET_FLAG(jmp, BBL_END);
    exitInstr->i_Bblock->b_last = jmp;
    InsertBBLAfterBBL(newBlock, lastBlock);
    newBlock->b_fun = dump_fun;

    /* write the numbers of each of the counters (and magic number) here */
    pstart = &(SECTION_REF_DATA->s_newdata[profile.pc_initStart]);
    storeLE(pstart, ALTO_MAGIC);

    if (profile.pc_bblsProf == 0) {
        offset = 0;
    } else {
        offset = PROFILE_INIT_SIZE;
    }
    storeLE(pstart + 4, profile.pc_bblsProf);
    storeLE(pstart + 8, offset);

    if (profile.pc_edgsProf == 0) {
        offset = 0;
    } else {
        offset = PROFILE_INIT_SIZE +
            profile.pc_bblsProf * BBL_PROFILE_DATA_SIZE;
    }
    storeLE(pstart + 12, profile.pc_edgsProf);
    storeLE(pstart + 16, offset);

    storeLE(pstart + 20, profile.pc_callsProf);
    storeLE(pstart + 24, 0);
    storeLE(pstart + 28, profile.pc_jmpsProf);
    storeLE(pstart + 32, 0);
    storeLE(pstart + 36, profile.pc_valsProf);
    storeLE(pstart + 40, 0);
    storeLE(pstart + 44, CONFIG_NENTRIES);

    CreateDumpBlocks(newBlock, Filename);

    return;
}



/*
 *  CreateDumpBlocks(block,Filename) --
 *
 *  CreateDumpBlocks strives to create a function that will dump the contents
 *  of all the edge counters, along with the basic block addresses, to a file.
 */
void CreateDumpBlocks(Block *block, char *Filename) {
    unsigned char instr[16];

    Instruction *bbl1_instr1 = InstrCreate();
    Instruction *bbl1_instr2 = InstrCreate();
    Instruction *bbl1_instr3 = InstrCreate();
    Instruction *bbl1_instr4 = InstrCreate();
    Instruction *bbl1_instr5 = InstrCreate();
    Instruction *bbl1_instr6 = InstrCreate();
    Instruction *bbl1_instr7 = InstrCreate();
    Instruction *bbl1_instr7_2 = InstrCreate();
    Instruction *bbl1_instr8 = InstrCreate();
    Instruction *bbl1_instr9 = InstrCreate();
    Instruction *bbl2_instr1 = InstrCreate();
    Instruction *bbl2_instr2 = InstrCreate();
    Instruction *bbl2_instr3 = InstrCreate();
    Instruction *bbl2_instr6 = InstrCreate();
    Instruction *bbl2_2_instr1 = InstrCreate();
    Instruction *bbl2_2_instr2 = InstrCreate();
    Instruction *bbl2_2_instr3 = InstrCreate();
    Instruction *bbl2_2_instr4 = InstrCreate();
    Instruction *bbl6_instr1 = InstrCreate();
    Instruction *bbl6_instr2 = InstrCreate();
    Instruction *bbl7_instr1 = InstrCreate();
    Instruction *bbl7_instr2 = InstrCreate();
    Instruction *bbl7_instr3 = InstrCreate();
    Instruction *bbl7_instr4 = InstrCreate();
    Instruction *bbl7_instr5 = InstrCreate();
    Instruction *bbl7_instr6 = InstrCreate();

    Block *bbl1 = block;
    Block *bbl2 = BBLCreate();
    Block *bbl2_2 = BBLCreate();
    Block *bbl6 = BBLCreate();
    Block *bbl7 = BBLCreate();

    Instruction *libc_open;
    Instruction *libc_write;
    Instruction *libc_close;

    Symbol *openSym;
    Symbol *writeSym;
    Symbol *closeSym;
    int offset;

    /*
     * The routine will make use of three external functions: __libc_open,
     * __libc_write and __libc_close.  We will find the addresses by
     * using symbols (hopefully).
     */

    openSym = nameToSymbol(LIBC_OPEN);
    writeSym = nameToSymbol(LIBC_WRITE);
    closeSym = nameToSymbol(LIBC_CLOSE);

    if (openSym == NULL) {
	FATAL("[X86: CreateDumpBlocks] Could not find symbol " LIBC_OPEN);
    }
    if (writeSym == NULL) {
	FATAL("[X86: CreateDumpBlocks] Could not find symbol " LIBC_WRITE);
    }
    if (closeSym == NULL) {
	FATAL("[X86: CreateDumpBlocks] Could not find symbol " LIBC_CLOSE);
    }

    libc_open = AddressToInstr(openSym->sy_value);
    libc_write = AddressToInstr(writeSym->sy_value);
    libc_close = AddressToInstr(closeSym->sy_value);

    if (libc_open == NULL) {
	FATAL(
	"[X86: CreateDumpBlocks] Could not find instruction for " LIBC_OPEN);
    }
    if (libc_write == NULL) {
	FATAL(
	"[X86: CreateDumpBlocks] Could not find instruction for " LIBC_WRITE);
    }
    if (libc_close == NULL) {
	FATAL(
	"[X86: CreateDumpBlocks] Could not find instruction for " LIBC_CLOSE);
    }

    /* write the file name here */
    IncreaseSectionSize(SECTION_REF_RODATA, strlen(Filename) + 1);

    for (offset = 0; offset < strlen(Filename); offset++) {
	SECTION_REF_RODATA->s_newdata[edgeProfileFilenameBase + offset] =
	    Filename[offset];
    }
    SECTION_REF_RODATA->s_newdata[edgeProfileFilenameBase + offset] = 0x0;

    /*
     * Block 1:          push    %ebp
     *                   mov     %esp, %ebp
     *                   sub     $0x8, %esp
     *                   push    %edi
     *                   push    %esi
     *                   push    %ebx
     *                   push    $0x41
     *                   push    Address for filename string
     *                   call    __libc_open
     */

    bbl1->b_first = bbl1_instr1;
    bbl1_instr1->i_Bblock = bbl1;

    /* push %ebp */
    instr[0] = 0x55;
    bbl1_instr1->i_length = xDisassembleInstr(0, instr, bbl1_instr1);
    INSTR_SET_FLAG(bbl1_instr1, BBL_START);

    /* mov %esp, %ebp */
    instr[0] = 0x89;
    instr[1] = 0xe5;
    bbl1_instr2->i_length = xDisassembleInstr(0, instr, bbl1_instr2);

    /* sub $0x8, %esp */
    instr[0] = 0x83;
    instr[1] = 0xec;
    instr[2] = 0x08;
    bbl1_instr3->i_length = xDisassembleInstr(0, instr, bbl1_instr3);

    /* push %edi */
    instr[0] = 0x57;
    bbl1_instr4->i_length = xDisassembleInstr(0, instr, bbl1_instr4);

    /* push %esi */
    instr[0] = 0x56;
    bbl1_instr5->i_length = xDisassembleInstr(0, instr, bbl1_instr5);

    /* push %ebx */
    instr[0] = 0x53;
    bbl1_instr6->i_length = xDisassembleInstr(0, instr, bbl1_instr6);

    /* set file permission to rwxrwxrwx */
    /* push $0x1ff */
    instr[0] = 0x68;
    storeLE(&(instr[1]), 0x1ff);
    bbl1_instr7->i_length = xDisassembleInstr(0, instr, bbl1_instr7);


    /* push $0x41 */
    instr[0] = 0x6a;
    instr[1] = 0x41;
    bbl1_instr7_2->i_length = xDisassembleInstr(0, instr, bbl1_instr7_2);

    /* push address for filename string */
    instr[0] = 0x68;
    instr[1] = 0x00;
    instr[2] = 0x00;
    instr[3] = 0x00;
    instr[4] = 0x00;
    bbl1_instr8->i_length = xDisassembleInstr(0, instr, bbl1_instr8);

    AddPatchSectionAddressIntoInstr(SECTION_REF_RODATA, edgeProfileFilenameBase,
	bbl1_instr8, bbl1_instr8->i_src);

    instr[0] = 0xe8;
    instr[1] = 0x00;
    instr[2] = 0x00;
    instr[3] = 0x00;
    instr[4] = 0x00;
    bbl1_instr9->i_length = xDisassembleInstr(0, instr, bbl1_instr9);

    InsertInstrAfterInstr(bbl1_instr2, bbl1_instr1);
    InsertInstrAfterInstr(bbl1_instr3, bbl1_instr2);
    InsertInstrAfterInstr(bbl1_instr4, bbl1_instr3);
    InsertInstrAfterInstr(bbl1_instr5, bbl1_instr4);
    InsertInstrAfterInstr(bbl1_instr6, bbl1_instr5);
    InsertInstrAfterInstr(bbl1_instr7, bbl1_instr6);
    InsertInstrAfterInstr(bbl1_instr7_2, bbl1_instr7);
    InsertInstrAfterInstr(bbl1_instr8, bbl1_instr7_2);
    InsertInstrAfterInstr(bbl1_instr9, bbl1_instr8);

    bbl1->b_last = bbl1_instr9;
    INSTR_SET_FLAG(bbl1_instr9, BBL_END);
    bbl1->b_type = BT_CALL;
    BBLAddEdge(bbl1, INSTR_FUNCTION(libc_open)->f_first, ET_FUNDIRECT);
    BBLAddEdge(bbl1, bbl2, ET_FUNLINK);

    /*
     * Block 2:            mov  %eax, %ebx
     *                     movl $0x0, 0xfffffffc(%ebp)
     *                     add  $0x8, %esp
     *                     movl $(counterBase + 8), 0xfffffff8(%ebp)
     *                     mov  $(counterBase + 4), %edi
     *                     mov  $(counterBase + 0), %esi
     */

    bbl2->b_first = bbl2_instr1;
    bbl2_instr1->i_Bblock = bbl2;

    /* mov %eax, %ebx */
    instr[0] = 0x89;
    instr[1] = 0xc3;
    bbl2_instr1->i_length = xDisassembleInstr(0, instr, bbl2_instr1);
    INSTR_SET_FLAG(bbl2_instr1, BBL_START);

    /* movl $0x0, 0xfffffffc(%ebp) */
    instr[0] = 0xc7;
    instr[1] = 0x45;
    instr[2] = 0xfc;
    instr[3] = 0x00;
    instr[4] = 0x00;
    instr[5] = 0x00;
    instr[6] = 0x00;
    bbl2_instr2->i_length = xDisassembleInstr(0, instr, bbl2_instr2);

    /* add $0x8, %esp */
    instr[0] = 0x83;
    instr[1] = 0xc4;
    instr[2] = 0x08;
    bbl2_instr3->i_length = xDisassembleInstr(0, instr, bbl2_instr3);


    /* mov $(counterBase + 0), %esi */
    instr[0] = 0xbe;
    instr[1] = 0x00;
    instr[2] = 0x00;
    instr[3] = 0x00;
    instr[4] = 0x00;
    bbl2_instr6->i_length = xDisassembleInstr(0, instr, bbl2_instr6);

    AddPatchSectionAddressIntoInstr(SECTION_REF_DATA, profile.pc_initStart,
	bbl2_instr6, bbl2_instr6->i_src);

    InsertInstrAfterInstr(bbl2_instr2, bbl2_instr1);
    InsertInstrAfterInstr(bbl2_instr3, bbl2_instr2);


    InsertInstrAfterInstr(bbl2_instr6, bbl2_instr3);

    bbl2->b_last = bbl2_instr6;
    INSTR_SET_FLAG(bbl2_instr6, BBL_END);

    bbl2->b_type = BT_NORMAL;
    BBLAddEdge(bbl2, bbl2_2, ET_NORMAL);


    /*
     * Block 2_2:          push $0x4
     *                     push %esi
     *                     push %ebx
     *                     call __libc_write
     */


    bbl2_2->b_first = bbl2_2_instr1;
    bbl2_2_instr1->i_Bblock = bbl2_2;

    /* push $0x4 */
    instr[0] = 0x68;
    storeLE(&(instr[1]), profile.pc_currentStart - profile.pc_initStart);
    bbl2_2_instr1->i_length = xDisassembleInstr(0, instr, bbl2_2_instr1);
    INSTR_SET_FLAG(bbl2_2_instr1, BBL_START);

    /* push %esi */
    instr[0] = 0x56;
    bbl2_2_instr2->i_length = xDisassembleInstr(0, instr, bbl2_2_instr2);

    /* push %ebx */
    instr[0] = 0x53;
    bbl2_2_instr3->i_length = xDisassembleInstr(0, instr, bbl2_2_instr3);

    instr[0] = 0xe8;
    instr[1] = 0x00;
    instr[2] = 0x00;
    instr[3] = 0x00;
    instr[4] = 0x00;
    bbl2_2_instr4->i_length = xDisassembleInstr(0, instr, bbl2_2_instr4);

    InsertInstrAfterInstr(bbl2_2_instr2, bbl2_2_instr1);
    InsertInstrAfterInstr(bbl2_2_instr3, bbl2_2_instr2);
    InsertInstrAfterInstr(bbl2_2_instr4, bbl2_2_instr3);

    bbl2_2->b_last = bbl2_2_instr4;
    INSTR_SET_FLAG(bbl2_2_instr4, BBL_END);
    bbl2_2->b_type = BT_CALL;
    BBLAddEdge(bbl2_2, INSTR_FUNCTION(libc_write)->f_first, ET_FUNDIRECT);
    BBLAddEdge(bbl2_2, bbl6, ET_FUNLINK);

    /*
     * Block 6:             push %ebx
     *                      call __libc_close
     */

    bbl6->b_first = bbl6_instr1;
    bbl6_instr1->i_Bblock = bbl6;

    /* push %ebx */
    instr[0] = 0x53;
    bbl6_instr1->i_length = xDisassembleInstr(0, instr, bbl6_instr1);
    INSTR_SET_FLAG(bbl6_instr1, BBL_START);

    /* call __libc_close */
    instr[0] = 0xe8;
    instr[1] = 0x00;
    instr[2] = 0x00;
    instr[3] = 0x00;
    instr[4] = 0x00;
    bbl6_instr2->i_length = xDisassembleInstr(0, instr, bbl6_instr2);

    InsertInstrAfterInstr(bbl6_instr2, bbl6_instr1);

    bbl6->b_last = bbl6_instr2;
    INSTR_SET_FLAG(bbl6_instr2, BBL_END);
    bbl6->b_type = BT_CALL;
    BBLAddEdge(bbl6, INSTR_FUNCTION(libc_close)->f_first, ET_FUNDIRECT);
    BBLAddEdge(bbl6, bbl7, ET_FUNLINK);

    /*
     * Block 7:             lea  0xffffffec(%ebp), %esp)
     *                      pop  %ebx
     *                      pop  %esi
     *                      pop  %edi
     *                      leave
     *                      ret
     */

    bbl7->b_first = bbl7_instr1;
    bbl7_instr1->i_Bblock = bbl7;

    /* lea 0xffffffec(%ebp), %esp) */
    instr[0] = 0x8d;
    instr[1] = 0x65;
    instr[2] = 0xec;
    bbl7_instr1->i_length = xDisassembleInstr(0, instr, bbl7_instr1);
    INSTR_SET_FLAG(bbl7_instr1, BBL_START);

    /* pop %ebx */
    instr[0] = 0x5b;
    bbl7_instr2->i_length = xDisassembleInstr(0, instr, bbl7_instr2);

    /* pop %esi */
    instr[0] = 0x5e;
    bbl7_instr3->i_length = xDisassembleInstr(0, instr, bbl7_instr3);

    /* pop %edi */
    instr[0] = 0x5f;
    bbl7_instr4->i_length = xDisassembleInstr(0, instr, bbl7_instr4);

    /* leave */
    instr[0] = 0xc9;
    bbl7_instr5->i_length = xDisassembleInstr(0, instr, bbl7_instr5);

    /* ret */
    instr[0] = 0xc3;
    bbl7_instr6->i_length = xDisassembleInstr(0, instr, bbl7_instr6);

    InsertInstrAfterInstr(bbl7_instr2, bbl7_instr1);
    InsertInstrAfterInstr(bbl7_instr3, bbl7_instr2);
    InsertInstrAfterInstr(bbl7_instr4, bbl7_instr3);
    InsertInstrAfterInstr(bbl7_instr5, bbl7_instr4);
    InsertInstrAfterInstr(bbl7_instr6, bbl7_instr5);


    bbl7->b_last = bbl7_instr6;
    INSTR_SET_FLAG(bbl7_instr6, BBL_END);
    bbl7->b_type = BT_RETURN;

    /*
     * This concludes the construction of our edge-dumping profile
     * routine. Now we will link the blocks together and return.
     */

    InsertBBLAfterBBL(bbl2, bbl1);
    InsertBBLAfterBBL(bbl2_2, bbl2);

    InsertBBLAfterBBL(bbl6, bbl2_2);
    InsertBBLAfterBBL(bbl7, bbl6);

#if DEBUG_CONSIST
    CheckConsistency();
    printf("checked consistency again\n");
    fflush(stdout);
#endif

    return;
}
