/*
 *  x86/schedule.c -- X86 Instruction Scheduler
 *
 *  Notes:
 *  1) Functions should always be aligned on 4-byte boundaries
 *  2) A loop entry should be aligned on a 16-byte boundary if it is within
 *     8 bytes of that boundary.  Important: Aligning loop headers carelessly
 *     will result in NOPs getting executed a lot.  The loop header may be
 *     executed from a fall-through from another block.
 *  3) A label that follows an unconditional branch should be aligned on a
 *     16-byte boundary when it is less than 8 bytes from that boundary.
 */
#define LOOP_ALIGN_BYTES                16

#ifdef ALIGN_LOOPS_ALWAYS
    #define MAXIMUM_LOOP_ALIGN_DISTANCE  15
#else
    #define MAXIMUM_LOOP_ALIGN_DISTANCE  7
#endif

#define FUNCTION_ALIGN_BYTES            16
#define MAXIMUM_FUNCTION_ALIGN_DISTANCE 15

#include "global.h"
#include "x86-icreate.h"
#include "flowgraph.h"
#include "utilities.h"
#include "linkage.h"
#include "opt.unreachable.h"
#include "dominator.h"
#include "sysdep.h"
#include "x86-classes.h"
#include "x86-asm.h"
#include "x86-table-defs.h"
#include "liveness.h"

/* Aligns the basic block on an 'align'-byte boundary */
int BBLAlign(Block *bbl, int offset, int maxDist, int align);

/* Inserts NOPs so that adjacent jumps won't conflict in the BTB */
void SeparateAdjacentJumps();

void CoalesceInstructions();

bool InstrIsDependent(Instruction *a, Instruction *b);

UINT8 InstrDelayBetween(Instruction *a, Instruction *b);

bool InstrsPair(Instruction *a, Instruction *b);

UINT32 NumInstrsPairable();

extern int registerToBitRegister(int someRegister);
extern bool RegisterIs32Bit(int aReg);



/*
 *  xAlignLoops() --
 *
 *  Aligns loop headers, instructions following unconditional jumps,
 *  and even functions on their appropriate boundaries. This function
 *  will return true if anything was changed, or false if nothing
 *  was changed.
 */
bool xAlignLoops() {
    Block *bbl;
    int offset = 0;
    bool change = false;
    bool doLoops = BooleanParameter("AlignLoopHeaders");
    bool doFuncs = BooleanParameter("AlignFunctions");

    if (doLoops) {

	FOREACH_BBL(bbl) {
	    Block *candidate, *prev;

	    if (BBL_HAS_FLAG(bbl, BBL_LOOP_HEADER_FLAG)) {
		prev = bbl->b_prev;
		candidate = bbl;

		/*
		 * If the loop header can be reached via a fallthrough,
		 * let the loop header be the block above it. This will
		 * make sure we don't insert any garbage
		 * instructions that actually get executed a bunch.
		 */

		while (prev && BBLIsInSameLoopAsBBL(prev, candidate) &&
		    (prev->b_type == BT_NORMAL || prev->b_type == BT_BRANCH)) {
		    candidate = prev;
		    prev = candidate->b_prev;
		}

		BBL_SET_FLAG(candidate, BBL_LOOP_ALIGN_ME_FLAG);

	    }

	    /* Align the instructions following unconditional branches
	     * with the same constraints as loop headers */
	    if (bbl->b_type == BT_UNCOND) {
		if ((bbl->b_next)
		    && (!BBL_HAS_FLAG(bbl->b_next, BBL_ALIGN_PAD_FLAG))) {
		    BBL_SET_FLAG(bbl->b_next, BBL_LOOP_ALIGN_ME_FLAG);
		}
	    }

	}
    }

    /* Now all basic blocks have been marked, so align those that are marked */
    FOREACH_BBL(bbl) {
	if (INSTR_HAS_FLAG(bbl->b_first, FUN_START)) {
	    if (doFuncs) {
		offset = BBLAlign(bbl, offset, MAXIMUM_FUNCTION_ALIGN_DISTANCE,
		    FUNCTION_ALIGN_BYTES);
	    }
	}
	if (BBL_HAS_FLAG(bbl, BBL_LOOP_ALIGN_ME_FLAG)) {
	    if (doLoops) {
		offset = BBLAlign(bbl, offset, MAXIMUM_LOOP_ALIGN_DISTANCE,
		    LOOP_ALIGN_BYTES);
	    }
	}
	if (offset != 0) {
	    change = true;
	}
    }

    return change;
}



/*
 *  BBLAlign(block,offset,maxDistance,align) --
 *
 *  Aligns 'block' on an 'align'-byte boundary if it falls within 'maxDistance'
 *  of that boundary.  'offset' is updated to reflect the number of bytes
 *  added during this alignment pass.
 */
int BBLAlign(Block *block, int offset, int maxDistance, int align) {
    Block *newblock;
    Instruction *instr;
    int i;
    int alignmentNeeded;
    int alignmentCurrent;

    /* Check the preceding block and see if it contains some alignment
     * we already added.  If so, then get rid of the block and count
     * the number of bytes we removed */
    alignmentCurrent = 0;
    if (block->b_prev && BBL_HAS_FLAG(block->b_prev, BBL_ALIGN_PAD_FLAG)) {
	FOREACH_INSTR_IN_BBL(block->b_prev, instr) {
	    alignmentCurrent = alignmentCurrent + instr->i_length;
	}
	instr = block->b_prev->b_first;
	while (instr && (instr->i_Bblock == block->b_prev)) {
	    Instruction *next = instr->i_next;
	    InstrDestroy(instr);
	    instr = next;
	}
    }

    /* Update the offset to reflect what we removed */
    offset = offset - alignmentCurrent;

    /* See how much we would need to add to align it on the boundary */
    alignmentNeeded = align - (block->b_first->i_final_address+offset) % align;

    /* If it's already appropriately aligned, then we have nothing to do */
    if (alignmentNeeded == align) {
	return offset;
    }

    /* If we would have to add too much alignment, then forget it */
    if (alignmentNeeded > maxDistance) {
	return offset;
    }

    offset = offset + alignmentNeeded;

    newblock = BBLCreate();

    /* Keep adding alignment, choosing the largest DoNothingInstruction at each
     * pass, until it is aligned correctly */
    for (i = 0; i < alignmentNeeded;) {
	Instruction *pad;

	switch (alignmentNeeded - i) {

	case 1:
	    pad = _InstrCreate1ByteDoNothing();
	    break;
	case 2:
	    pad = _InstrCreate2ByteDoNothing();
	    break;
	case 3:
	    pad = _InstrCreate3ByteDoNothing();
	    break;
	case 4:
	case 5:
	case 6:
	    pad = _InstrCreate4ByteDoNothing();
	    break;
	case 7:
	case 8:
	case 9:
	case 10:
	case 11:
	case 12:
	case 13:
	case 14:
	case 15:
	case 16:
	case 17:
	case 18:
	case 19:
	case 20:
	case 21:
	case 22:
	case 23:
	case 24:
	case 25:
	case 26:
	case 27:
	case 28:
	case 29:
	case 30:
	case 31:
	case 32:
	    pad = _InstrCreate7ByteDoNothing();
	    break;
	default:
	    pad = NULL;
	    WARNING("[AlignLoops] Could not satisfy alignment.");
	    break;
	}

	if (newblock->b_first != NULL) {
	    InsertInstrBeforeInstr(pad, newblock->b_first);
	    newblock->b_first = pad;
	} else {
	    pad->i_Bblock = newblock;
	    newblock->b_first = newblock->b_last = pad;
	    INSTR_SET_FLAG(pad, BBL_START);
	    INSTR_SET_FLAG(pad, BBL_END);
	}

	i = i + pad->i_length;
    }

    BBL_SET_FLAG(newblock, BBL_ALIGN_PAD_FLAG);
    InsertBBLBeforeBBL(newblock, block);

    return offset;
}



/*
 *  SeparateAdjacentJumps() --
 *
 *  Having adjacent jump instructions can cause multiple instructions to
 *  share the same slot in the branch target buffer.  The buffer uses bits
 *  2-31 to identify the instruction, so if the 2 lower order bits of the
 *  address are the same then two instructions can overlap.  We can't
 *  determine if this is the case until final addresses have been calculated.
 *  Thus, we play it safe and separate all adjacent jumps by inserting a
 *  two-byte NOP.
 */
void SeparateAdjacentJumps() {
    Instruction *iterator, *pad;

    FOREACH_INSTR(iterator) {

	if (!isPCChangingClass(iterator)) {
	    continue;
	}
	if (!iterator->i_next) {
	    continue;
	}
	if (!isPCChangingClass(iterator->i_next)) {
	    continue;
	}

	/* Insert a two-byte NOP */
	pad = _InstrCreate2ByteDoNothing();
	InsertInstrBeforeInstr(pad, iterator->i_next);
	pad->i_Bblock->b_first = pad;

    }

    return;
}



/*
 *  CoalesceInstructions() --
 *
 *  Coalesces adjacent adds/subs to the same register
 */
void CoalesceInstructions() {
    return;
}



/*
 *  OperandSize(op) --
 *
 *  Returns the number of bytes an operand may access.
 *  May be slightly conservative in some cases.
 */
static int OperandSize(Operand *op) {
    if (OPERAND_HAS_FLAG(op, OPERAND_BMODE_FLAG)) {
	return 1;
    } else if (OPERAND_HAS_FLAG(op, OPERAND_WMODE_FLAG)) {
	return 4;
    } else if (OPERAND_HAS_FLAG(op, OPERAND_DMODE_FLAG)) {
	return 8;
    } else {
	return 32;
    }
}



/*
 *  OperandsAliasMemory(op1,op2) --
 *
 *  Slightly modified version of a function Debray wrote to test if two operands
 *  may alias the same memory location.
 */
static bool OperandsAliasMemory(Operand *op1, Operand *op2) {
    bool op1IsStk, op2IsStk, op1IsAbs, op2IsAbs;
    int op_size, op_diff;

    if ((op1 == NULL) || (op2 == NULL)) {
	return false;
    }
    if (!OperandIsMemory(op1) || !OperandIsMemory(op2)) {
	return false;
    }

    op1IsStk = (OperandAccessesStackFromFramePointer(op1) ||
	OperandAccessesStackFromStackPointer(op1));

    op2IsStk = (OperandAccessesStackFromFramePointer(op2) ||
	OperandAccessesStackFromStackPointer(op2));

    op1IsAbs = OperandIsAbsoluteMemoryLocation(op1);
    op2IsAbs = OperandIsAbsoluteMemoryLocation(op2);

    if (op1IsStk && op2IsAbs) {
	return false;
    }
    if (op2IsStk && op1IsAbs) {
	return false;
    }

    if (op1->a_base_reg == op2->a_base_reg &&
	op1->a_index_reg == op2->a_index_reg) {
	op_size = (OperandSize(op1) > OperandSize(op2)) ?
	    OperandSize(op1) : OperandSize(op2);
	op_diff = abs(op1->a_displacement - op2->a_displacement);
	if (op_diff > op_size) {
	    return false;
	}
    }

    return true;
}



/*
 *  InstrAccessStackImplicit(a) --
 *
 *  Returns true if an instruction may access the stack, even though it's
 *  operands are not memory locations.
 */
static bool InstrAccessStackImplicit(Instruction *a) {
    return ((a->i_class == CLASS_PUSH) || (a->i_class == CLASS_PUSHALL)
	|| (a->i_class == CLASS_PUSHF) || (a->i_class == CLASS_POP)
	|| (a->i_class == CLASS_POPALL) || (a->i_class == CLASS_POPF));
}



/*
 *  InstrsAliasMemory(a,b) --
 *
 *  Returns true if the two instructions may possibly access the same
 *  memory location.
 */
static bool InstrsAliasMemory(Instruction *a, Instruction *b) {
    bool a_access_mem, b_access_mem;

    a_access_mem = ((a->i_dest != NULL && OperandIsMemory(a->i_dest)) ||
	(a->i_src != NULL && OperandIsMemory(a->i_src)) ||
	(a->i_src2 != NULL && OperandIsMemory(a->i_src2)));
    b_access_mem = ((b->i_dest != NULL && OperandIsMemory(b->i_dest)) ||
	(b->i_src != NULL && OperandIsMemory(b->i_src)) ||
	(b->i_src2 != NULL && OperandIsMemory(b->i_src2)));

    // TODO: We could be a little less conservative here
    if ((InstrAccessStackImplicit(b) && a_access_mem) ||
	(InstrAccessStackImplicit(a) && b_access_mem)) {
	return true;
    }

    if ((!a_access_mem) || (!b_access_mem)) {
	return false;
    }

    if (OperandsAliasMemory(a->i_dest, b->i_src) ||
	OperandsAliasMemory(a->i_dest, b->i_src2) ||
	OperandsAliasMemory(a->i_dest, b->i_dest)) {
	return true;
    }

    if (OperandsAliasMemory(b->i_dest, a->i_src) ||
	OperandsAliasMemory(b->i_dest, a->i_src2) ||
	OperandsAliasMemory(b->i_dest, a->i_dest)) {
	return true;
    }

    return false;
}



/*
 *  SchedulerSpecialCase(instr) --
 *
 *  If we have to have bad hacks we might as well put them all in one place.
 *  We'll note special instructions here that can not be scheduled around for
 *  whatever reasons, usually because they're a special case in some way.
 *
 *  We need to return true anytime one of these is a special case.
 */
static bool SchedulerSpecialCase(Instruction *instr) {
    // fldcw changes the mode the FPU operates in, and shouldn't be moved.
    if (instr->i_class == CLASS_FLDCW) {
	return true;
    }

    /*
     * Changes to %esp shouldn't be reordered since an interrupt may write
     * to the stack, so %esp should always be correct.
     */
    if ((instr->i_dest != NULL)
	&& (OperandIsStackPointer(instr->i_dest))) {
	return true;
    }

    return false;
}



/*
 *  xInstrIsDependent(a,b) --
 *
 *  This function should return true if there is a data or control dependency
 *  between instructions 'a' and 'b'.  In other words if 'a' is an instruction
 *  in a program, and 'b' is the next instruction, will switching the order of
 *  'a' and 'b' change the correctness of the program?
 *
 *  This function can assume that having 'a' come before 'b' produces
 *  a correct ordering.
 */
bool xInstrIsDependent(Instruction *a, Instruction *b) {
    REGISTER_BITVECTOR a_defn, b_defn;
    REGISTER_BITVECTOR a_use, b_use;
    PSW_BITVECTOR a_psw_defn, b_psw_defn;
    PSW_BITVECTOR a_psw_use, b_psw_use;

    /*  char *an, *bn;
     * an = PrintInstruction(a);
     * bn = PrintInstruction(b); */

    a_defn = ComputeRegistersDefinedLiberal(a);
    b_defn = ComputeRegistersDefinedLiberal(b);
    a_use = ComputeRegistersUsed(a);
    b_use = ComputeRegistersUsed(b);

    /*
     * If there exists any of the standard dependencies between
     * the registers we can't reorder.
     */
    if (a_defn & b_use) {
	return true;
    }
    if (a_use & b_defn) {
	return true;
    }
    if (a_defn & b_defn) {
	return true;
    }

    /*
     * If one of the instructions is a jump, call, or return
     * we can't reorder.
     */
    if (isConditionalInstr(a) || isUnconditionalInstr(a) ||
	isConditionalInstr(b) || isUnconditionalInstr(b) ||
	(a->i_class == CLASS_RET) || (b->i_class == CLASS_RET)) {
	return true;
    }

    /*
     * PSW dependencies are computed here.  We may get overly
     * conservative estimates if we counted all write / write
     * dependencies since many instructions modify the PSW
     * as a side effect.  So we use some liveness
     * analysis to eliminate some of these false dependencies.
     */
    a_psw_defn = ComputePSWDefined(a);
    b_psw_defn = ComputePSWDefined(b);
    a_psw_use = ComputePSWUsed(a);
    b_psw_use = ComputePSWUsed(b);

    if (a_psw_defn & b_psw_use) {
	return true;
    }
    if (a_psw_use & b_psw_defn) {
	return true;
    }
    if (PSWBitvectorIsLiveAtInstr(a_psw_defn & b_psw_defn, b)) {
	return true;
    }

    /*
     * If one instruction writes to a memory location another instruction
     * may use/write then we have a dependency.
     */
    if (InstrsAliasMemory(a, b)) {
	return true;
    }

    if (SchedulerSpecialCase(a) || SchedulerSpecialCase(b)) {
	return true;
    }

    /*
     * We haven't been able to find a dependency, so we'll have
     * to assume there isn't one.
     */
    return false;
}



/*
 *  OperandsMayStall(a,b) --
 *
 *  Returns true if the operands 'a' and 'b' meet the conditions to
 *  cause a partial register stall.
 */
static bool OperandsMayStall(Operand *a, Operand *b) {
    int reg_a, reg_b;

    if ((a == NULL) || (b == NULL)) {
	return false;
    }
    if (!OperandUsesRegister(a) || !OperandUsesRegister(b)) {
	return false;
    }

    reg_a = a->a_base_reg;
    reg_b = b->a_base_reg;

    if ((reg_a == 0) || (reg_b == 0)) {
	return false;
    }

    if ((registerToBitRegister(reg_a) & registerToBitRegister(reg_b)) &&
	(!RegisterIs32Bit(reg_a) && RegisterIs32Bit(reg_b))) {
	return true;
    }

    return false;
}



/*
 *  xInstrDelayBetween(a,b) --
 *
 *  Returns the number of cycles that may pass between the execution of
 *  instruction B.
 */
UINT8 xInstrDelayBetween(Instruction *a, Instruction *b) {
    int base_time;

    /*
     * All integer instructions have a latency equal to their number
     * of micro ops, except for multiplication and division.
     */
    base_time = InstrGetOpcodeEntry(a)->x_uops;
    if (base_time == COMPLEX) {
	return base_time;
    }

    /*
     * Special cases for multiplication and division.
     */
    if ((a->i_class == CLASS_MUL) || (a->i_class == CLASS_IMUL)) {
	base_time += 3;
    }
    if ((a->i_class == CLASS_DIV) || (a->i_class == CLASS_IDIV)) {
	base_time += 7;
    }

    if (b == NULL) {
	return base_time;
    }

    /*
     * If we read a 32 bit register after writing to a 8/16 bit subregister
     * then we concur a partial register stall, which is very nasty.
     */
    if (OperandsMayStall(a->i_dest, b->i_src) ||
	OperandsMayStall(a->i_dest, b->i_src2)) {
	base_time += 7;
    }

    return base_time;
}



/*
 *  xInstrsPair(a,b) --
 *
 *  We can run instruction 'b' in a pairing with instruction 'a'
 *  if 'a' has <= 4 micro-ops and 'b' takes 1 micro-op.
 */
bool xInstrsPair(Instruction *a, Instruction *b) {
    X86op *a_entry, *b_entry;

    a_entry = InstrGetOpcodeEntry(a);
    b_entry = InstrGetOpcodeEntry(b);

    return ((a_entry->x_uops > 1) &&
	(a_entry->x_uops <= 4) && (b_entry->x_uops == 1));
}



/*
 *  xNumInstrsPairable() --
 *
 *  We can schedule three instructions in the 4-1-1 manner on the P6.
 *  See section 3.6.4 of the Intel Architecture Optimization Manual for more
 *  details.
 */
UINT32 xNumInstrsPairable() {
    return 3;
}
