/*
 *  x86/util.c -- General purpose x86 instructions
 */
#include "global.h"
#include "utilities.h"
#include "x86-prefix.h"
#include "x86-asm.h"
#include "x86-util.h"
#include "x86-hash.h"
#include "x86-iprint.h"
#include "x86-op.h"
#include "x86-registers.h"
#include "x86-classes.h"
#include "sysdep.h"
#include "linkage.h"
#include "math.h"
#include "flowgraph.h"
#include "x86-liveness.h"
#include "x86-condition-codes.h"
#include "clone.h"
#include "upm.h"

#define MAXIMUM_BYTEMODE_CHANGES 10
#define CALL_COST_IN_TICKS 2

static Instruction *PICGetNextInstr(Instruction *instr,
    Instruction *call_instr);
static bool IsPICReturnInstr(Instruction *instr);
static bool ISPICStackReaderInstr(Instruction *instr, UINT16 *reg);
static bool IsPICALUInstr(Instruction *instr, int pc_register);
static bool IsPICALUInstrForFLIST(Instruction *instr);



/*
 *  numberOfPrefixes(instr) --
 *  Returns the number of one-byte prefixes that are on the given instruction.
 */
int numberOfPrefixes(Instruction *instr) {
    int num = 0;

    if (instr->i_prefixes & PREFIX_FWAIT) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_REPZ) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_REPNZ) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_LOCK) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_CS) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_SS) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_DS) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_ES) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_FS) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_GS) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_OPERAND) {
	num++;
    }
    if (instr->i_prefixes & PREFIX_ADDRESS) {
	num++;
    }

    if (num > 4) {
	ReportInstrError("More than 4 prefixes", instr);
    }

    return num;
}



/*
 *  UpdateBytemode(instr,op) --
 *
 *  UpdateBytemode looks at the displacement in the operand and decides which
 *  opcode to use.  It does this much in the same way instructions are
 *  re-assembled.  Scans through the opcode table are performed until
 *  a suitable instruction format is uncovered.  The instruction is updated
 *  to reflect this format.  The byte-constraint of the operand is chosen
 *  optimally--that is, it starts out looking for an instruction that will
 *  let the displacement be represented in the smallest number of bytes.
 *  If one is not found, the constraint is relaxed and we will settle
 *  for larger instructions.
 */
bool UpdateBytemode(Instruction *instr, Operand *op) {
    op_rtn rtn1, rtn2, rtn3;
    int bytemode, bytemode1, bytemode2, bytemode3;
    int dflag = 1;
    int initialSpace, finalSpace;

    if (OperandUsesRegister(op)) {
	if (!OperandAccessesStackFromFramePointer(op)) {
	    if (!OperandAccessesStackFromFramePointerWithSIB(op)) {
		return true;
	    }
	}
    }

    if (instr->i_prefixes & PREFIX_OPERAND) {
	dflag = 0;
    }

    initialSpace = getOperandBytes(op, dflag);
    if (initialSpace <= 0) {
	printf("Operand with bad size @  0x%8.8x\n", instr->i_address);
    }
    //CASSERT(initialSpace > 0, "Operand with bad size.");

    /*
     * If a relocation was added to this operand it needs to be relaxed
     * to the relocation size.
     */
    if (OPERAND_HAS_FLAG(op, OPERAND_RELOCATED)) {
	OperandClearBytemodeFlags(op);
	OP_SET_FLAG(op, OPERAND_VMODE_FLAG);
	return true;
    }

    /* See how many bytes are needed. Prefixes might need to be examined here */
    if (op->a_displacement <= CHAR_MAX && op->a_displacement >= CHAR_MIN) {
	bytemode = b_mode;
    } else if (op->a_displacement <= SHRT_MAX && op->a_displacement>=SHRT_MIN) {
	bytemode = w_mode;
    } else if (op->a_displacement <= INT_MAX && op->a_displacement >= INT_MIN) {
	bytemode = v_mode;
    } else {
	WARNING
	    ("[UpdateBytemode] Could not find appropriate size for operand.");
	return false;
    }

    if (instr->i_temp >= MAXIMUM_BYTEMODE_CHANGES) {
	bytemode = v_mode;
    } else {
	instr->i_temp = instr->i_temp + 1;
    }

    if ((bytemode == v_mode) && OP_TYPE(op, OPERAND_VMODE_FLAG)) {
	return true;
    }
    if ((bytemode == w_mode) && OP_TYPE(op, OPERAND_WMODE_FLAG)) {
	return true;
    }
    if ((bytemode == b_mode) && OP_TYPE(op, OPERAND_BMODE_FLAG)) {
	return true;
    }

    /* Keep relaxing the constraint until we find something or die */
    while (true) {

	if (op == instr->i_dest) {
	    bytemode1 = bytemode;
	    bytemode2 = GetByteMode(instr->i_src, dflag);
	    bytemode3 = GetByteMode(instr->i_src2, dflag);
	} else if (op == instr->i_src) {
	    bytemode1 = GetByteMode(instr->i_dest, dflag);
	    bytemode2 = bytemode;
	    bytemode3 = GetByteMode(instr->i_src2, dflag);
	} else if (op == instr->i_src2) {
	    bytemode1 = GetByteMode(instr->i_dest, dflag);
	    bytemode2 = GetByteMode(instr->i_src, dflag);
	    bytemode3 = bytemode;
	} else {
	    WARNING("[UpdateBytemode]"
		" Supplied operand does not match instructions operands.");
	    return false;
	}

	rtn1 = GetOpRoutine(instr, instr->i_dest, bytemode1, bytemode2);
	rtn2 = GetOpRoutine(instr, instr->i_src, bytemode2, bytemode1);
	rtn3 = GetOpRoutine(instr, instr->i_src2, bytemode3, bytemode1);

	if (PatternMatchOpcode(instr->i_class, bytemode1, rtn1, bytemode2,
		rtn2, bytemode3, rtn3, instr) == NULL) {
	    if (bytemode == v_mode) {
		return false;
	    }

	    /* Relax the restrictions */
	    if (bytemode == b_mode) {
		bytemode = w_mode;
	    } else {
		bytemode = v_mode;
	    }

	} else {
	    OperandClearBytemodeFlags(op);
	    if (bytemode == b_mode) {
		OP_SET_FLAG(op, OPERAND_BMODE_FLAG);
	    }
	    if (bytemode == w_mode) {
		OP_SET_FLAG(op, OPERAND_WMODE_FLAG);
	    }
	    if (bytemode == v_mode) {
		OP_SET_FLAG(op, OPERAND_VMODE_FLAG);
	    }
	    break;
	}

    }					/* end while */

    /* It is interesting to know when bytemodes change */
    finalSpace = getOperandBytes(op, dflag);
    CASSERT(finalSpace > 0, "Operand with bad size.");
    if (finalSpace != initialSpace) {
	REPORT(RP_TRIVIAL, "Instruction at 0x%8.8x changed from %d to %d bytes",
	    instr->i_address, initialSpace, finalSpace);
	xUpdateInstrLength(instr);
    }

    return true;
}



/*
 *  xIsValidInstr(i) --
 *
 *  Returns true if the instruction can be assembled into machine code.
 */
bool xIsValidInstr(Instruction *i) {
    op_rtn rtn1, rtn2, rtn3;
    int bytemode1, bytemode2, bytemode3;
    int dflag = 1;
    Instruction *instr = InstrCreate();
    UINT32 flags;

    /*
     * What we do is create a copy of the instruction and try to assemble
     * it. We create the copy instead of assembling the original because
     * some of the flags/bytemodes might need to be massaged in a call to
     * UpdateInstr(). Note that InstrClone() is not used since we don't want
     * to do things like create UPM entries for the instruction, or stick it
     * in hash tables, etc.
     */

    instr->i_length = i->i_length;
    instr->i_class = i->i_class;
    instr->i_type = i->i_type;
    instr->i_prefixes = i->i_prefixes;
    instr->i_flags = i->i_flags;

    if (i->i_dest == NULL) {
	instr->i_dest = NULL;
    } else {
	flags = i->i_dest->a_flags;
	instr->i_dest = OperandClone(i->i_dest);
	instr->i_dest->a_flags = flags;
    }

    if (i->i_src == NULL) {
	instr->i_src = NULL;
    } else {
	flags = i->i_src->a_flags;
	instr->i_src = OperandClone(i->i_src);
	instr->i_src->a_flags = flags;
    }

    if (i->i_src2 == NULL) {
	instr->i_src2 = NULL;
    } else {
	flags = i->i_src2->a_flags;
	instr->i_src2 = OperandClone(i->i_src2);
	instr->i_src2->a_flags = flags;
    }

    if (instr->i_prefixes & PREFIX_OPERAND) {
	dflag = 0;
    }

    bytemode1 = GetByteMode(instr->i_dest, dflag);
    bytemode2 = GetByteMode(instr->i_src, dflag);
    bytemode3 = GetByteMode(instr->i_src2, dflag);
    rtn1 = GetOpRoutine(instr, instr->i_dest, bytemode1, bytemode2);
    rtn2 = GetOpRoutine(instr, instr->i_src, bytemode2, bytemode1);
    rtn3 = GetOpRoutine(instr, instr->i_src2, bytemode3, bytemode1);

    if (PatternMatchOpcode(instr->i_class, bytemode1, rtn1,
	    bytemode2, rtn2, bytemode3, rtn3, instr) == NULL)
	goto isValidFalse;

    if (!UpdateInstr(instr)) {
	goto isValidFalse;
    }

    return true;

isValidFalse:

    if (instr->i_dest) {
	free(instr->i_dest);
    }
    if (instr->i_src) {
	free(instr->i_src);
    }
    if (instr->i_src2) {
	free(instr->i_src2);
    }

    free(instr);

    return false;
}



/*
 *  xUpdateInstr(instr) --
 *
 *  Updates various things in the instruction structure. Currently only
 *  the bytemode and length are updated. False is returned if the
 *  instruction could not be updated.
 */
bool xUpdateInstr(Instruction *instr) {
    bool result = true;

    if (instr->i_dest != NULL) {
	if (!UpdateBytemode(instr, instr->i_dest)) {
	    result = false;
	}
    }
    if (instr->i_src != NULL) {
	if (!UpdateBytemode(instr, instr->i_src)) {
	    result = false;
	}
    }
    if (instr->i_src2 != NULL) {
	if (!UpdateBytemode(instr, instr->i_src2)) {
	    result = false;
	}
    }

    instr->i_length = CalculateInstrLength(instr);

    return result;
}



/*
 *  xReportSuspiciousInstrs() --
 *  Prints any occurrences of strange instructions to standard out.
 */
void xReportSuspiciousInstrs() {
    Instruction *iterator;

    FOREACH_INSTR(iterator) {
	if ((iterator->i_class == CLASS_ADD)
	    && (iterator->i_length == 2)) {
	    if (iterator->i_src && iterator->i_dest) {
		if (iterator->i_src->a_base_reg == (reg_a | reg_8l)) {
		    if (iterator->i_dest->a_base_reg == (reg_a | reg_32)) {
			REPORT(RP_SMLOOP, "SUSPICIOUS INSTRUCTION at 0x%8.8x",
			    iterator->i_address);
		    }
		}
	    }
	}
    }

    return;
}



/*
 *  isSTRegister(the_reg) --
 *  Returns true if the given register is a floating pointer register.
 */
bool isSTRegister(int the_reg) {
    if ((the_reg == reg_st0) ||
	(the_reg == reg_st1) ||
	(the_reg == reg_st2) ||
	(the_reg == reg_st3) ||
	(the_reg == reg_st4) ||
	(the_reg == reg_st5) ||
	(the_reg == reg_st6) ||
	(the_reg == reg_st7) ||
	(the_reg == reg_st))
	return true;
    return false;
}

/*
 *  isSegmentRegister(the_reg) --
 *  Returns true if the given register is a segment register
 */
bool isSegmentRegister(int the_reg) {
    if ((the_reg == reg_cs) ||
	(the_reg == reg_ss) ||
	(the_reg == reg_es) ||
	(the_reg == reg_ds) ||
	(the_reg == reg_fs) ||
	(the_reg == reg_gs))
	return true;
    return false;
}


bool isControlRegister(int the_reg) {
    if ((the_reg == reg_cr0) ||
	(the_reg == reg_cr4) ||
	(the_reg == reg_cr2) ||
	(the_reg == reg_cr3))
	return true;
    return false;
}

/*
 *  xIsConditionalInstr(inst) --
 *  Returns true if the given instruction is a conditional branch.
 */
bool xIsConditionalInstr(Instruction *inst) {
    if (inst->i_type == TYPE_CONDITIONAL) {
	return true;
    }
    return false;
}



/*
 *  xIsUnconditionalInstr(inst) --
 *  Returns true if the given instruction is an unconditional branch.
 */
bool xIsUnconditionalInstr(Instruction *inst) {
    if (inst->i_type == TYPE_UNCONDITIONAL) {
	return true;
    }
    return false;
}



/*
 *  xPrintInstruction(instr) --
 *
 *  Returns a pointer to a string that is a textual representation of the
 *  given instruction.
 */
char *xPrintInstruction(Instruction *instr) {
    static unsigned char buffer[50];

    InstrAssemble(buffer, instr);
    if (INSTR_HAS_FLAG(instr, INSTR_DATA_BYTE)) {
	return "Data bytes instruction";
    } else {
	return i386InstrToText(instr->i_final_address, buffer);
    }
}



/*
 *  xPrintInstructionFinal(instr) --
 *
 *  Returns a pointer to a string that is a textual representation of the
 *  given instruction. Any PC relative displacement is calculated using
 *  the instructions final address.
 */
char *xPrintInstructionFinal(Instruction *instr) {
    unsigned char buffer[16];
    char *result;

    InstrAssemble(buffer, instr);
    result = i386InstrToText(instr->i_final_address, buffer);
    return result;
}



/*
 *  xGetIcacheSize() --
 *  Returns the size of the L1 instruction cache on the processor.
 */
int xGetIcacheSize() {
    return ICACHE_SIZE;
}



/*
 *  xGetDcacheSize() --
 *  Returns the size of the L1 data cache on the processor.
 */
int xGetDcacheSize() {
    return DCACHE_SIZE;
}



/*
 *  xGetMegahertz() -- returns the megahertz rating of the processor.
 */
int xGetMegahertz() {
    return MHZ;
}



/*
 *  xGetIcacheMissPenalty() --
 *  Returns the cost, in clock ticks, of a miss in the icache on this processor.
 */
int xGetIcacheMissPenalty() {
    return MHZ * NANOSECONDS_PER_ICACHE_MISS * 0.001;
}



/*
 *  xBuildOpcodeHashTable() --
 *  Builds the opcode hash table for quick lookups.
 */
void xBuildOpcodeHashTable() {
    X86BuildHashTable();
}



/*
 *  xGetCallPenalty() --
 *  Returns the cost, in clock ticks, of a call instruction on this processor.
 */
int xGetCallPenalty() {
    return CALL_COST_IN_TICKS;
}



/*
 *  xRegisterIsStackPointer(aReg) --
 *  Returns true if the supplied register is the stack pointer (esp).
 */
bool xRegisterIsStackPointer(int aReg) {
    if ((aReg == reg_sp) || (aReg == (reg_sp | reg_32)) ||
	(aReg == (reg_sp | reg_16)) || (aReg == (reg_sp | reg_8l)) ||
	(aReg == (reg_sp | reg_8h))) {
	return true;
    }
    return false;
}



/*
 *  xRegisterIsFramePointer(aReg) --
 *  Returns true if the supplied register is the frame pointer (ebp).
 */
bool xRegisterIsFramePointer(int aReg) {
    if ((aReg == reg_bp) || (aReg == (reg_bp | reg_32)) ||
	(aReg == (reg_bp | reg_16)) || (aReg == (reg_bp | reg_8l)) ||
	(aReg == (reg_bp | reg_8h))) {
	return true;
    }
    return false;
}



/*
 *  xInstrDefinesPSW(instr) --
 *  Returns true if the instruction can define the PSW.
 */
bool xInstrDefinesPSW(Instruction *instr) {
    if ((InstrGetOpcodeEntry(instr)->x_pswUsed)) {
	return true;
    }
    return false;
}



/*
 *  ConvertRegister(aReg) --
 *
 *  Converts the x86-table.h register representation (al_reg) into the
 *  PLTO register representation (reg_a | reg_8l).
 */
int ConvertRegister(int aReg) {

    switch (aReg) {
    case all_regs:
	return reg_all;
	break;
    case no_reg:
	return 0;
	break;
    case eAX_reg:
	return reg_a | reg_32;
	break;
    case eBX_reg:
	return reg_b | reg_32;
	break;
    case eCX_reg:
	return reg_c | reg_32;
	break;
    case eDX_reg:
	return reg_d | reg_32;
	break;
    case eSI_reg:
	return reg_si | reg_32;
	break;
    case eDI_reg:
	return reg_di | reg_32;
	break;
    case eSP_reg:
	return reg_sp | reg_32;
	break;
    case eBP_reg:
	return reg_bp | reg_32;
	break;
    case al_reg:
	return reg_a | reg_8l;
	break;
    case bl_reg:
	return reg_b | reg_8l;
	break;
    case cl_reg:
	return reg_c | reg_8l;
	break;
    case dl_reg:
	return reg_d | reg_8l;
	break;
    case ah_reg:
	return reg_a | reg_8h;
	break;
    case bh_reg:
	return reg_b | reg_8h;
	break;
    case ch_reg:
	return reg_c | reg_8h;
	break;
    case dh_reg:
	return reg_d | reg_8h;
	break;
    case ax_reg:
	return reg_a | reg_16;
	break;
    case bx_reg:
	return reg_b | reg_16;
	break;
    case cx_reg:
	return reg_c | reg_16;
	break;
    case dx_reg:
	return reg_d | reg_16;
	break;
    case sp_reg:
	return reg_sp | reg_16;
	break;
    case bp_reg:
	return reg_bp | reg_16;
	break;
    case si_reg:
	return reg_si | reg_16;
	break;
    case di_reg:
	return reg_di | reg_16;
	break;
    case indir_dx_reg:
	return reg_d | reg_indirect | reg_16;
	break;
    case st0_reg:
	return reg_st0;
	break;
    case st1_reg:
	return reg_st1;
	break;
    case st2_reg:
	return reg_st2;
	break;
    case st3_reg:
	return reg_st3;
	break;
    case st4_reg:
	return reg_st4;
	break;
    case st5_reg:
	return reg_st5;
	break;
    case st6_reg:
	return reg_st6;
	break;
    case st7_reg:
	return reg_st7;
	break;
    default:
	WARNING("[ConvertRegister] Could not convert %d.", aReg);
    }

    return 0;
}



/*
 *  RegisterIsFloatingPoint(aReg) --
 *  Returns true if the register is a floating point register.
 */
bool RegisterIsFloatingPoint(int aReg) {
    return isSTRegister(aReg);
}

bool RegisterIsSegmentRegister(int aReg) {
    return isSegmentRegister(aReg);
}

bool RegisterIsControlRegister(int aReg) {
    return isControlRegister(aReg);
}

/*
 *  xInstrUsesReg(iins,areg) --
 *
 *  InstrUsesReg(iins, areg) Returns true if instruction iins uses register
 *  areg as [part of] an operand.
 */
bool xInstrUsesReg(Instruction *iins, int areg) {
    if (!areg) {
	return false;
    }
    return RegisterIsLiveInBitvector(areg, ComputeRegistersUsed(iins)) != 0;

}



/*
 *  xInstrDefinesReg(iins,areg) --
 *
 *  InstrDefinesReg(iins, areg) Returns true if instruction iins defines
 *  register areg.
 */
bool xInstrDefinesReg(Instruction *iins, int areg) {
    if (!areg) {
	return false;
    }
    return (RegisterIsLiveInBitvector(areg,
	    ComputeRegistersDefinedLiberal(iins)) != 0);

}



/*
 *  xPrintRegister(i) --
 *
 *  Returns a pointer to a textual representation of the bitregister i
 */
char *xPrintRegister(int i) {
    char *regname;

    switch (i) {
        /* These should never be used, but are here just in case */
    case reg_a:  regname = "%eax"; break; 
    case reg_b:  regname = "%ebx"; break;
    case reg_c:  regname = "%ecx"; break;
    case reg_d:  regname = "%edx"; break;
    case reg_si: regname = "%esi"; break;
    case reg_di: regname = "%edi"; break;
    case reg_bp: regname = "%ebp"; break; 
    case reg_sp: regname = "%esp"; break;
    
    case reg_es:  regname = "es"; break; 
    case reg_cs:  regname = "cs"; break;
    case reg_ss:  regname = "ss"; break;
    case reg_ds:  regname = "ds"; break;
    case reg_fs:  regname = "fs"; break;
    case reg_gs:  regname = "gs"; break;
         
        /* 32-bit general purpose registers */
    case reg_a | reg_32:  regname = "%eax"; break;
    case reg_b | reg_32:  regname = "%ebx"; break;
    case reg_c | reg_32:  regname = "%ecx"; break;
    case reg_d | reg_32:  regname = "%edx"; break;
    case reg_si | reg_32: regname = "%esi"; break;
    case reg_di | reg_32: regname = "%edi"; break;
    case reg_bp | reg_32: regname = "%ebp"; break;
    case reg_sp | reg_32: regname = "%esp"; break;

        /* 16-bit general purpose registers */
    case reg_a | reg_16:  regname = "%ax"; break;
    case reg_b | reg_16:  regname = "%bx"; break;
    case reg_c | reg_16:  regname = "%cx"; break;
    case reg_d | reg_16:  regname = "%dx"; break;
    case reg_si | reg_16: regname = "%si"; break;
    case reg_di | reg_16: regname = "%di"; break;
    case reg_bp | reg_16: regname = "%bp"; break;
    case reg_sp | reg_16: regname = "%sp"; break;

        /* 8-bit general purpose registers */
    case reg_a | reg_8l: regname = "%al"; break;
    case reg_b | reg_8l: regname = "%bl"; break;
    case reg_c | reg_8l: regname = "%cl"; break;
    case reg_d | reg_8l: regname = "%dl"; break;

    case reg_a | reg_8h: regname = "%ah"; break;
    case reg_b | reg_8h: regname = "%bh"; break;
    case reg_c | reg_8h: regname = "%ch"; break;
    case reg_d | reg_8h: regname = "%dh"; break;

    case reg_st:
    case reg_st0:
    case reg_st1:
    case reg_st2:
    case reg_st3:
    case reg_st4:
    case reg_st5:
    case reg_st6:
    case reg_st7:      regname = "st"; break;

        /* Special case for the indirect dx register */
    case reg_d | reg_16 | reg_indirect: regname = "[%dx]"; break;
    default:
	regname = "<?>";
	break;
    }

    return regname;
}



/*
 *  CvtJcc2SETcc_compl(iins) --
 *
 *  This function takes a conditional branch instruction and returns the
 *  SETcc/CMOVcc condition code bits corresponding to the complement of that
 *  instruction's condition, e.g., given a JP instruction (jump if PF = 1) it
 *  returns the condition codes for NP (PF = 0).
 */
UINT8 CvtJcc2SETcc_compl(Instruction *iins) {
    int ccbits;

    switch (iins->i_class) {
    case CLASS_JO:
	ccbits = NO;
	break;
    case CLASS_JNO:
	ccbits = O;
	break;
    case CLASS_JB:
	ccbits = NB;
	break;
    case CLASS_JAE:
	ccbits = NAE;
	break;
    case CLASS_JE:
	ccbits = NE;
	break;
    case CLASS_JNE:
	ccbits = E;
	break;
    case CLASS_JBE:
	ccbits = NBE;
	break;
    case CLASS_JA:
	ccbits = NA;
	break;
    case CLASS_JS:
	ccbits = NS;
	break;
    case CLASS_JNS:
	ccbits = S;
	break;
    case CLASS_JP:
	ccbits = NP;
	break;
    case CLASS_JNP:
	ccbits = P;
	break;
    case CLASS_JL:
	ccbits = NL;
	break;
    case CLASS_JNL:
	ccbits = L;
	break;
    case CLASS_JLE:
	ccbits = NLE;
	break;
    case CLASS_JG:
	ccbits = NG;
	break;

    default:
	ccbits = -1;
    }

    return ccbits;
}



/*
 *  xReplaceNonsupportedInstructions() --
 *  Replaces nonsupported instructions with equivalent ones that are supported.
 */
void xReplaceNonsupportedInstructions() {
    Instruction *iterator, *newJmp1, *newJmp2;
    Block *bbl, *newBlock1, *newBlock2, *TRUEBlock, *FALSEBlock;

    FALSEBlock = TRUEBlock = NULL;

    FOREACH_INSTR(iterator) {

	if (iterator->i_class == CLASS_JCXZ) {
	    bbl = iterator->i_Bblock;

	    TRUEBlock = BBLGetTrueEdge(bbl)->e_to;
	    FALSEBlock = BBLGetFalseEdge(bbl)->e_to;

	    CASSERT(TRUEBlock != NULL, "JECXZ block without true edge.");
	    CASSERT(FALSEBlock != NULL, "JECXZ block without false edge.");

	    newBlock1 = BBLCreate();
	    newJmp1 = InstrCreateJMP();
	    newJmp1->i_Bblock = newBlock1;
	    newBlock1->b_first = newBlock1->b_last = newJmp1;
	    InsertBBLAfterBBL(newBlock1, bbl);
	    newBlock1->b_type = BT_UNCOND;
	    INSTR_SET_FLAG(newJmp1, BBL_END);
	    INSTR_SET_FLAG(newJmp1, BBL_START);

	    newBlock2 = BBLCreate();
	    newJmp2 = InstrCreateJMP();
	    newJmp2->i_Bblock = newBlock2;
	    newBlock2->b_first = newBlock2->b_last = newJmp2;
	    InsertBBLAfterBBL(newBlock2, newBlock1);
	    newBlock2->b_type = BT_UNCOND;
	    INSTR_SET_FLAG(newJmp2, BBL_END);
	    INSTR_SET_FLAG(newJmp2, BBL_START);

	    BBLRemoveEdges(bbl);
	    BBLAddEdge(bbl, newBlock1, ET_FALSE);
	    BBLAddEdge(bbl, newBlock2, ET_TRUE);
	    BBLAddEdge(newBlock1, FALSEBlock, ET_UNCOND);
	    BBLAddEdge(newBlock2, TRUEBlock, ET_UNCOND);
	}
    }					/* end iteration over instructions */

    return;
}



/*
 *  isConditionalMove(instr) --
 *
 *  Returns true if the instruction is a conditional move.
 */
bool isConditionalMove(Instruction *instr) {

    if (instr->i_class == CLASS_CMOVO) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVNO) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVB) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVAE) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVE) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVNE) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVBE) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVA) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVS) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVNS) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVP) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVNP) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVL) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVGE) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVLE) {
	return true;
    }
    if (instr->i_class == CLASS_CMOVG) {
	return true;
    }

    return false;
}



/*
 *  xInstrUsesDest(instr) --
 *
 *  Returns true if the instruction implicitly uses the destination operand.
 */
bool xInstrUsesDest(Instruction *instr) {
    if ((InstrGetOpcodeEntry(instr)->x_psw & DUF)) {
	return true;
    }
    return false;
}



/*  The following functions deal with identifying PIC
 *  (Position Independent Code).
 *
 *  Ideally we wouldn't have to worry about PIC, since it's
 *  usually used in dynamic libraries, which we don't support.
 *  However, some hand-coded assembly function (call_gmon_start)
 *  and C++ functions (__register_frame_info) in glibc use PIC
 *  for whatever reasons (They probably did it to make our lives
 *  all the more painful).
 *
 *  Part of the requirements of PIC means no absolute addresses
 *  are allowed in the code.  Thus everything must be referenced
 *  with relative addresses.  However, we can't load data from an
 *  address relative to the current PC.  Thus PIC code computes an
 *  absolute address by storing the difference between a piece of data
 *  and the loading instruction inside the PIC code, and dynamically
 *  computing and adding the current PC to this value.
 *
 *  This usually looks like the following:
 *
 *  L1: mov (%esp,1),%ebx       // Store the top of the stack in ebx
 *	ret
 *	.
 *	.
 *	.
 *	call L1
 *	add $0x2, %ebx
 *
 *  There is usually a relocation added to the 0x2 value.  This relocation
 *  is the difference between the start of the relocation (in the middle of the
 *  add) and the piece of data we're interested in.  The '$0x2' represents the
 *  distance between the PC of the instruction after the call (which is stored
 *  in %ebx) and the relocation.
 *
 *  If we insert new instructions between the call and add, then the '$0x2'
 *  no longer represents the distance from the instruction after the call
 *  to the relocation.  Thus we need to identify these sequences and take
 *  appropriate measures.
 *
 *  Given a call instruction that is suspected of being part of a PIC
 *  sequence, this function returns the corresponding ALU operation with
 *  that sequence.  Returns NULL if the instruction is not a PIC operation.
 *
 *  In the above example, if given the 'call L1' this would return the
 *  'add $0x2, %ebx'.
 */



/*
 *  xPICGetALUForCall(call_instr) --
 *
 *  Returns the ALU operation that goes with a call in a
 *  PIC instruction sequence.  See the function for a more
 *  detailed comment.
 */
Instruction *xPICGetALUForCall(Instruction *call_instr) {
    Instruction *stack_reader = NULL;
    Instruction *ret_instr = NULL;
    Instruction *current_instr;
    bool instr_valid;
    UINT16 reg;

    if (INSTR_HAS_FLAG(call_instr, INSTR_FLIST_FLAG)) {
	if (IsPICALUInstrForFLIST(call_instr->i_next)) {
	    return call_instr->i_next;
	}
	return NULL;
    }

    current_instr = call_instr;
    while (true) {
	current_instr = PICGetNextInstr(current_instr, call_instr);
	if (current_instr == NULL) {
	    return NULL;
	}

	instr_valid = false;
	if (stack_reader == NULL && ISPICStackReaderInstr(current_instr,&reg)) {
	    stack_reader = current_instr;
	    instr_valid = true;
	}
	if (ret_instr == NULL && stack_reader != NULL &&
	    IsPICReturnInstr(current_instr)) {
	    ret_instr = current_instr;
	    instr_valid = true;
	}
	if (ret_instr != NULL && IsPICALUInstr(current_instr, reg)) {
	    return current_instr;
	}

	if (!instr_valid) {
	    return NULL;
	}
    }
}



/*
 *  PICGetNextInstr(instr,call_instr) --
 */
static Instruction *PICGetNextInstr(Instruction *instr,Instruction *call_instr){
    Edge *edge;
    Instruction *next_instr;


    if (instr == call_instr) {
	edge = BBLGetFunDirectEdge(call_instr->i_Bblock);
	if (edge == NULL) {
	    return NULL;
	}
	return edge->e_to->b_first;
    }

    if (instr->i_class == CLASS_RET) {
	return call_instr->i_next;
    }
    if (instr->i_next == NULL) {
	return NULL;
    }

    edge = BBLGetEdgeToBBL(instr->i_Bblock, instr->i_next->i_Bblock);

    if (instr->i_next->i_Bblock == instr->i_Bblock) {
	next_instr = instr->i_next;
    } else if (edge != NULL && edge->e_type == ET_NORMAL) {
	next_instr = instr->i_next;
    } else {
	return NULL;
    }
    return next_instr;
}



/*
 *  IsPICReturnInstr(instr) --
 */
static bool IsPICReturnInstr(Instruction *instr) {
    if (instr->i_class == CLASS_POP &&
	instr->i_dest != NULL && OperandIsGPRegister(instr->i_dest)) {
	return true;
    }

    if (instr->i_class == CLASS_RET) {
	return true;
    }
    return false;
}



/*
 *  ISPICStackReaderInstr(instr,reg) --
 *
 *  Returns true if the given instruction loads the value
 *  at the top of the stack and stores it in a register.
 *  Sets *reg to be the corresponding register.
 */
static bool ISPICStackReaderInstr(Instruction *instr, UINT16 *reg) {
    Operand *op;

    if (instr->i_class == CLASS_POP && IsPICReturnInstr(instr)) {
	*reg = instr->i_dest->a_base_reg;
	return true;
    }

    if (instr->i_src != NULL &&
	OPERAND_HAS_FLAG(instr->i_src, OPERAND_BASE_FLAG) &&
	RegisterIsStackPointer(instr->i_src->a_base_reg)) {
	op = instr->i_src;
    }
    if (instr->i_src2 != NULL &&
	OPERAND_HAS_FLAG(instr->i_src2, OPERAND_BASE_FLAG) &&
	RegisterIsStackPointer(instr->i_src2->a_base_reg)) {
	op = instr->i_src2;
    }
    if (op == NULL) {
	return false;
    }

    if (instr->i_dest != NULL && OperandIsGPRegister(instr->i_dest)) {
	*reg = instr->i_dest->a_base_reg;
    } else {
	return false;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_SCALE_FLAG) &&
	!OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG) &&
	!OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG) && op->a_scale == 1) {
	return true;
    }

    if (OPERAND_HAS_FLAG(op, OPERAND_DISP_R_FLAG) &&
	!OPERAND_HAS_FLAG(op, OPERAND_SCALE_FLAG) &&
	!OPERAND_HAS_FLAG(op, OPERAND_IMM_FLAG) &&
	op->a_displacement == ADDRESS_SIZE) {
	return true;
    }

    return false;
}



/*
 *  IsPICALUInstr(instr,pc_register) --
 *
 *  Returns true if the given instruction matches the PIC
 *  criteria for being the ALU operation by:
 *  1. Being and ADD or SUB instruction.
 *  2. Reading from the register which contains the current PC.
 *  3. Having a PC relative relocation.
 *
 *  This would correspond to the ADD instruction the the
 *  above example.
 */
static bool IsPICALUInstr(Instruction *instr, int pc_register) {
    Operand *op = NULL;

    /*
     * We're looking for ADD or SUB instructions (usually ADD)
     */
    if ((instr->i_class != CLASS_ADD) && (instr->i_class != CLASS_SUB)) {
	return false;
    }

    /*
     * The instruction should also write to the register that the
     * instruction pointer was stored in
     */
    if (instr->i_dest != NULL &&
	OperandUsesRegister(instr->i_dest) &&
	instr->i_dest->a_base_reg == pc_register) {
	op = instr->i_dest;
    }
    if (op == NULL) {
	return false;
    }

    /*
     * The instruction will also have a PC-Relative relocation associated
     * with one of the sources.
     */
    if ((instr->i_src != NULL &&
	    OPERAND_HAS_FLAG(instr->i_src, OPERAND_RELOCATABLE_PC_FLAG))
	|| (instr->i_src2 != NULL
	    && OPERAND_HAS_FLAG(instr->i_src2, OPERAND_RELOCATABLE_PC_FLAG))) {
	return true;
    }

    return false;
}



/*
 *  IsPICALUInstrForFLIST(instr) --
 *
 *  If the code called by the call state was an  FLIST,
 *  then we're not sure what the register being poped into
 *  was.
 */
static bool IsPICALUInstrForFLIST(Instruction *instr) {
    if (instr == NULL) {
	return false;
    }

    /*
     * We're looking for ADD or SUB instructions (usually ADD)
     */
    if ((instr->i_class != CLASS_ADD) && (instr->i_class != CLASS_SUB)) {
	return false;
    }

    /*
     * The instruction should also write to the register that the
     * instruction pointer was stored in
     */
    if (instr->i_dest == NULL || !OperandUsesRegister(instr->i_dest)) {
	return false;
    }

    /*
     * The instruction will also have a GOTPC-Relative relocation associated
     * with one of the sources.
     */
    if ((instr->i_src != NULL &&
	    OPERAND_HAS_FLAG(instr->i_src, OPERAND_RELOCATABLE_GOTPC_FLAG))
	|| (instr->i_src2 != NULL
	    && OPERAND_HAS_FLAG(instr->i_src2,OPERAND_RELOCATABLE_GOTPC_FLAG))){
	return true;
    }

    return false;
}



/*
 *  xGetFramePointer() --
 *
 *  Returns the frame pointer.
 *  returns: the frame pointer
 */
int xGetFramePointer() {
    return reg_bp | reg_32;
}



/*
 *  xGetStackPointer() --
 *
 *  Returns the stack pointer.
 *  returns: the stack pointer.
 */
int xGetStackPointer() {
    return reg_sp | reg_32;
}



/*
 *  xGetARegister() --
 *
 *  Returns some arbitrary general purpose register, currently ebx.
 *  returns: some arbitrary general purpose register.
 */
int xGetARegister() {
    return reg_b | reg_32;
}



/*
 *  xGetSyscallRegister() --
 */
int xGetSyscallRegister() {
    return reg_a | reg_32;
}



/*
 *  xGetSyscallNoForSyscall(syscall) --
 *
 *  Returns the system call number associated with the specified system
 *  call by tracking back to the definition of the system call register.
 *
 *  syscall: the system call instruction
 *  returns: the system call number associated with the specified system
 *  call or -1 if one can not be identified
 */
int xGetSyscallNoForSyscall(Instruction *syscall) {
    Instruction *iiterator;

    if (!isSyscallClass(syscall)) {
	return -1;
    }

    iiterator = syscall->i_prev;
    while ((iiterator != NULL) &&
	(iiterator->i_Bblock == syscall->i_Bblock->b_prev)) {
	if ((iiterator->i_src != NULL) &&
	    (iiterator->i_dest != NULL) &&
	    OperandIsImmediate(iiterator->i_src) &&
	    (iiterator->i_dest->a_base_reg == xGetSyscallRegister()))
	{
	    return iiterator->i_src->a_displacement;
	}

	iiterator = iiterator->i_prev;
    }

    return -1;
}



/*
 *  xRemoveSyscallNoForSyscall(syscall) --
 *
 *  Removes the instruction from the instruction stream that loads the
 *  system call number for the specified system call.
 *
 *  syscall: the system call instruction
 */
void xRemoveSyscallNoForSyscall(Instruction *syscall) {
    Instruction *iiterator;

    if (!isSyscallClass(syscall)) {
	WARNING("Called with non-syscall instruction");
	return;
    }

    iiterator = syscall->i_prev;
    while ((iiterator != NULL) &&
	(iiterator->i_Bblock == syscall->i_Bblock->b_prev)) {
	if ((iiterator->i_src != NULL) &&
	    (iiterator->i_dest != NULL) &&
	    OperandIsImmediate(iiterator->i_src) &&
	    (iiterator->i_dest->a_base_reg == xGetSyscallRegister())) {
	    if (iiterator->i_Bblock->b_last == iiterator->i_Bblock->b_first) {
		InsertInstrBeforeInstr(InstrCreateNOP(), iiterator);
		UPMChangeAllSourcesFromInstrToInstr(iiterator,
		    iiterator->i_prev);
	    }
	    InstrDisconnect(iiterator);
	    InstrFree(iiterator);
	    return;
	}

	iiterator = iiterator->i_prev;
    }
}
